theory GHCRewriteSystem
(*
  This is an initial attempt to interpret 'evaluate' as a rewrite system over 'expr'.
  The idea is that we create a relation 'legalSteps' such that:
    1. evaluate only performs legal Steps:
      evaluate \<subseteq> legalSteps
    2. legalSteps and evaluate agree on the normal forms:
      normalForms = {x . evaluate x = WHNF x} = {x . \<forall> y . \<not> legalSteps x y}
    3. if a term x has a normal form after n steps, doing a legalStep preserves this:
      evaluate^n x \<in> normalForms \<Longrightarrow> \<forall> y . legalSteps x y \<longrightarrow> (evaluate^n y = evaluate^n x)
    ... probably some more stuff, s.t. "legalSteps" can be used 
  For 3. to hold, we use that normal forms do not get evaluated further:
      x \<in> normalForms \<Longrightarrow> evaluate x = x

  The above statements are still somewhat informal:
    In the above we treat the type "expr" as if it were the same type as "expr executionResult".
    In the above we do not take the context ("bind list") into account.
    We have to disallow certain non-well-formed "expr executionResult" forms.
      In particular, we only allow "WHNF x" if "evaluate x = WHNF x".
      We also wish to disallow x if "\<exists> i. evaluate x = UnboundVariable i",
      and all terms that would evaluate to such terms,
      but we also need to allow ourselves to fill in variable values later.
*)
imports Main GHCCore
begin

type_synonym stepInd = "bind list \<Rightarrow> expr \<Rightarrow> expr \<Rightarrow> bool"

(* This step function is not correct yet, as it does not deal with alpha-steps.
   Also, we might need to split it up into simpler relations whose union gives legalStepInd *)
(* RT: perhaps better modelled as "inductive" *)
fun legalStepInd :: "stepInd \<Rightarrow> stepInd" where
"legalStepInd legalStep bs e1 e2 = (
  (\<exists> i e11 e12. e1 = App (Lam i e11) e12 \<and> e2 = Let (NonRec i e12) e11
     \<and> \<not> overlappingBinding (NonRec i e12) bs)
  \<or> (\<exists> i. e1 = Var i \<and> findBinding i bs = Step e2) (* See Note of evaluateHead *)
  \<or> (\<exists> e11 e12 e21 . e1 = App e11 e12 \<and> e2 = App e21 e12
     \<and> legalStep bs e11 e21)
  \<or> (\<exists> e11 e12 e22 . e1 = App e11 e12 \<and> e2 = App e11 e22
     \<and> legalStep bs e12 e22)
  \<or> (\<exists> i e12 e22. e1 = Lam i e12 \<and> e2 = Lam i e22 \<and> legalStep bs e12 e22)
  \<or> (\<exists> b e12 e22. e1 = Let b e12 \<and> e2 = Let b e22
     \<and> \<not> overlappingBinding b bs \<and> legalStep (Cons b bs) e12 e22)
  \<or> (\<exists> e11 b e13 e21 e23 ar1 ar2 eNF. e1 = Case e11 b e13 \<and> e2 = Case e21 b e23
     \<and> legalStep bs e11 e21 \<and> legalStep bs e21 eNF
     \<and> Step ar1 = matchAlts eNF e13 \<and> Step ar2 = matchAlts eNF e23
     \<and> legalStep (Cons (NonRec b e21) bs) ar1 ar2)
  \<or> (\<exists> e11 v. e1 = TickOrCast e11 v \<and> legalStep bs e11 e2))"

(* legalStepInd only contains positive 'legalStep' occurrences,
   so it can be shown to be monotonic using this syntactic property.
   We're not using that argument in this proof though, since fastforce gets it fast enough.
*)
lemma legalStepInd_preMono[simp] :
      assumes "\<And> bs e1 e2 . x bs e1 e2 \<le> y bs e1 e2"
      shows  "legalStepInd x bs e1 e2 \<le> legalStepInd y bs e1 e2"
using assms by fastforce

lemma legalStepInd_mono: "mono legalStepInd"
proof
  fix x y :: stepInd
  assume "x \<le> y"
  hence "\<And> bs e1 e2 . x bs e1 e2 \<le> y bs e1 e2" using le_funD by fastforce
  hence "\<And> bs e1 e2 . legalStepInd x bs e1 e2 \<le> legalStepInd y bs e1 e2"
  using legalStepInd_preMono by blast
  thus "legalStepInd x \<le> legalStepInd y" by (simp add: le_funI)
qed

definition legalStep where "legalStep = lfp legalStepInd"
lemma legalStep[simp] : "lfp legalStepInd = legalStepInd legalStep"
using legalStep_def by (simp add:def_lfp_unfold legalStepInd_mono)


lemma lfp_thm:
assumes "legalStepInd (\<lambda> bs e1 e2. False) bs e1 e2"
shows "legalStep bs e1 e2"
proof -
  have "(\<lambda> bs e1 e2. False) \<le> legalStep" by (simp add: le_funI)
  hence "legalStepInd legalStep bs e1 e2" using assms by auto
  thus ?thesis using legalStep_def[symmetric] legalStep[symmetric] by metis
qed

fun legalStepsInd :: "stepInd \<Rightarrow> stepInd" where
"legalStepsInd legalSteps bs e1 e2
  = (legalStepInd legalSteps bs e1 e2 \<or> (e1 = e2)
    \<or> (\<exists> e3. legalSteps bs e1 e3 \<and> legalSteps bs e3 e2))"

(* In the proof of legalStepsInd_mono, we use that legalStep is mono.
   This gives a slightly more elaborate proof than simply unfolding legalStep.
   Just using fastforce can get it in the latter case, at the cost of taking ~10sec *)
context
notes legalStepInd.simps[simp del] begin
lemma legalStepsInd_mono[simp] : "mono legalStepsInd"
proof
  fix x y :: stepInd
  assume "x \<le> y"
  hence x_le_y : "\<And> bs e1 e2 . x bs e1 e2 \<le> y bs e1 e2" using le_funD by fastforce
  hence "\<And> bs e1 e2 . legalStepInd x bs e1 e2 \<le> legalStepInd y bs e1 e2"
  using legalStepInd_preMono by blast
  hence "\<And> bs e1 e2 . legalStepsInd x bs e1 e2 \<le> legalStepsInd y bs e1 e2"
  using x_le_y by fastforce
  thus "legalStepsInd x \<le> legalStepsInd y" by (simp add: le_funI)
qed
end

definition legalSteps where "legalSteps = lfp legalStepsInd"
lemma legalSteps : "lfp legalStepsInd = legalStepsInd legalSteps"
using legalSteps_def by (simp add:def_lfp_unfold)

lemma legalSteps_rfx [simp] : "legalSteps bs e e"
by (metis legalSteps legalStepsInd.elims(3) legalSteps_def)

(* this doesn't hold yet, but it shows what I am trying to achieve *)
thm evaluate.elims
lemma legalStepIndGEevaluate :
  assumes "evaluateHead e bs = Step e2"
  shows "legalStepInd legalSteps bs e e2"
proof(cases e)
  case(App a b)
  hence "legalStepInd legalSteps bs (App a b) e2" sorry
  thus ?thesis using App by blast
next
  case(Let bnd e)
  show ?thesis using Let sorry
next
  case(Case e bndr alts)
  show ?thesis using Case sorry
qed(insert assms,simp_all)

lemma resWHNF_eq :
 shows "evaluate nf bs = IsWHNF x \<Longrightarrow> nf = x"
oops

lemma endTerm : 
 assumes "evaluate nf bs = WHNF nf"
 shows "\<And> b. legalStep bs nf b \<Longrightarrow> nf = b"
oops
end
