(*
Author:  Sebastiaan Joosten (2016-2017)
Author:  René Thiemann (2016-2017)
Author:  Akihisa Yamada (2016-2017)
License: LGPL (see file COPYING.LESSER)
*)

(* Nearly a complete copy of Integer_Arithmetic where integers have been replaced by rational numbers. *)

theory Rational_Arithmetic
imports
  Ord.Formula
  Deriving.Compare_Generator
  Certification_Monads.Check_Monad
  Ord.Non_Inf_Orders
  Ord.Show_Literal_Polynomial
begin

locale RA_locale
begin
datatype ty = BoolT | RatT
datatype sig = LessF | LeF | SumF nat | ConstF (the_ConstF: rat) | ProdF nat | EqF
datatype "val" = Rat (to_rat: rat) | Bool (to_bool: bool)
no_notation inter (infixl "Int" 70)

fun type_of_fun :: "sig \<Rightarrow> ty list \<times> ty" where
  "type_of_fun LessF = ([RatT,RatT], BoolT)"
| "type_of_fun LeF = ([RatT,RatT], BoolT)"
| "type_of_fun EqF = ([RatT,RatT], BoolT)"
| "type_of_fun (SumF n) = (replicate n RatT, RatT)"
| "type_of_fun (ConstF _) = ([], RatT)"
| "type_of_fun (ProdF n) = (replicate n RatT, RatT)" 

fun Values_of_type :: "ty \<Rightarrow> val set" where
  "Values_of_type BoolT = range Bool"
| "Values_of_type RatT = range Rat"

lemma mem_Values_of_typeE[elim]:
  assumes "v \<in> Values_of_type ty"
      and "\<And>i. ty = RatT \<Longrightarrow> v = Rat i \<Longrightarrow> thesis"
      and "\<And>b. ty = BoolT \<Longrightarrow> v = Bool b \<Longrightarrow> thesis"
  shows "thesis" using assms by (cases ty, auto)

lemma mem_Values_of_typeD[dest]:
  assumes "v \<in> Values_of_type ty"
  shows "(ty = RatT \<longrightarrow> (\<exists>i. v = Rat i)) \<and> (ty = BoolT \<longrightarrow> (\<exists>b. v = Bool b))"
  by (insert assms, auto)

fun "I" :: "sig \<Rightarrow> val list \<Rightarrow> val" where
  "I LessF [x, y] = Bool (to_rat x < to_rat y)"
| "I LeF [x, y] = Bool (to_rat x \<le> to_rat y)"
| "I EqF [x, y] = Bool (to_rat x = to_rat y)"
| "I (SumF n) xs = Rat (sum_list (map to_rat xs))"
| "I (ConstF x) [] = Rat x"
| "I (ProdF n) xs = Rat (prod_list (map to_rat xs))"
| "I _ _ = undefined"

sublocale prelogic
  where I = I
    and type_of_fun = type_of_fun
    and Values_of_type = Values_of_type
    and Bool_types = "{BoolT}"
    and to_bool = to_bool
    and type_fixer= "TYPE(_)".

abbreviation const where "const x \<equiv> Fun (ConstF x) []"

end

(* don't do this before "sublocale prelogic"; otherwise you lose term RA.eval *)
interpretation RA: RA_locale.

type_synonym 'v RA_exp = "(RA.sig,'v,RA.ty) exp"
type_synonym 'v RA_formula = "(RA.sig,'v,RA.ty) exp formula"

instantiation RA.sig :: showl
begin
fun showsl_sig where 
  "showsl_sig RA.LessF = showsl_lit (STR ''<'')" 
| "showsl_sig RA.LeF = showsl_lit (STR ''<='')" 
| "showsl_sig RA.EqF = showsl_lit (STR ''='')" 
| "showsl_sig (RA.SumF n) = showsl_lit (STR ''+'')" 
| "showsl_sig (RA.ProdF n) = showsl_lit (STR ''*'')" 
| "showsl_sig (RA.ConstF n) = showsl n" 
definition "showsl_list (xs :: RA.sig list) = default_showsl_list showsl xs"
instance ..
end

instantiation RA.ty :: showl
begin
fun showsl_ty where 
  "showsl_ty RA.BoolT = showsl_lit (STR ''Bool'')" 
| "showsl_ty RA.RatT = showsl_lit (STR ''Rat'')" 
definition "showsl_list (xs :: RA.ty list) = default_showsl_list showsl xs"
instance ..
end
derive (linorder) compare rat
derive compare_order RA.val

context RA_locale
begin

lemma has_type_induct[consumes 1, case_names Less Lesseq Equals Sum Const Prod Var]:
  assumes "e :\<^sub>f ty"
  and LessCase: "\<And> ta tb.
       ta :\<^sub>f RatT \<Longrightarrow> tb :\<^sub>f RatT \<Longrightarrow> P ta RatT \<Longrightarrow> P tb RatT \<Longrightarrow> P (Fun LessF [ta,tb]) BoolT"
  and LesseqCase: "\<And> ta tb.
       ta :\<^sub>f RatT \<Longrightarrow> tb :\<^sub>f RatT \<Longrightarrow> P ta RatT \<Longrightarrow> P tb RatT \<Longrightarrow> P (Fun LeF [ta,tb]) BoolT"
  and EqualsCase: "\<And> ta tb.
       ta :\<^sub>f RatT \<Longrightarrow> tb :\<^sub>f RatT \<Longrightarrow> P ta RatT \<Longrightarrow> P tb RatT \<Longrightarrow> P (Fun EqF [ta,tb]) BoolT"
  and SumCase: "\<And> ts n.
       (\<And> t. t \<in> set ts \<Longrightarrow> t :\<^sub>f RatT) \<Longrightarrow> (\<And> t. t \<in> set ts \<Longrightarrow> P t RatT) \<Longrightarrow> P (Fun (SumF n) ts) RatT"
  and ConstCase: "\<And> x. P (const x) RatT"
  and ProdCase: "\<And> ts n. 
       (\<And> t. t \<in> set ts \<Longrightarrow> t :\<^sub>f RatT) \<Longrightarrow> (\<And> t. t \<in> set ts \<Longrightarrow> P t RatT) \<Longrightarrow> P (Fun (ProdF n) ts) RatT"
  and VarCase: "\<And> x ty. P (Var (x,ty)) ty"
  shows "P e ty"
using assms(1) proof(induct e ty rule: has_type.induct)
  case (1 x ty) with VarCase show ?case by(cases x, auto)
next
  case Fun: (2 f ls ty)
  show ?case
  proof(cases f)
    case [simp]: LessF
    from Fun(2) obtain ta tb where [simp]: "ls = [ta,tb]" by (cases ls rule: list_3_cases, auto)
    from Fun(2) have at: "ta :\<^sub>f RatT" and bt: "tb :\<^sub>f RatT" and ty: "ty = BoolT" by auto
    with Fun.hyps[of 0] Fun.hyps[of 1] LessCase show ?thesis by auto
  next
    case [simp]: LeF
    from Fun(2) obtain ta tb where [simp]: "ls = [ta,tb]" by (cases ls rule: list_3_cases, auto)
    from Fun(2) have at: "ta :\<^sub>f RatT" and bt: "tb :\<^sub>f RatT" and ty: "ty = BoolT" by auto
    with Fun.hyps[of 0] Fun.hyps[of 1] LesseqCase show ?thesis by auto
  next
    case [simp]: EqF
    from Fun(2) obtain ta tb where [simp]: "ls = [ta,tb]" by (cases ls rule: list_3_cases, auto)
    from Fun(2) have at: "ta :\<^sub>f RatT" and bt: "tb :\<^sub>f RatT" and ty: "ty = BoolT" by auto
    with Fun.hyps[of 0] Fun.hyps[of 1] EqualsCase show ?thesis by auto
  next
    case [simp]: (SumF n)
    {
      fix l
      assume "l \<in> set ls" 
      then obtain i where l: "l = ls ! i" and i: "i < length ls" unfolding set_conv_nth by auto
      from Fun(2) l i have lty: "l :\<^sub>f RatT" by auto
      from Fun.hyps[OF i] lty i l Fun(2) have "P l RatT" by auto
      note lty this
    }
    from SumCase[OF this, of ls n] Fun(2) show ?thesis by auto
  next
    case [simp]: (ConstF x)
    with Fun have [simp]: "ls = []" by auto
    show ?thesis using Fun ConstCase by simp
  next
    case [simp]: (ProdF n)
    {
      fix l
      assume "l \<in> set ls" 
      then obtain i where l: "l = ls ! i" and i: "i < length ls" unfolding set_conv_nth by auto
      from Fun(2) l i have lty: "l :\<^sub>f RatT" by auto
      from Fun.hyps[OF i] lty i l Fun(2) have "P l RatT" by auto
      note lty this
    }
    from ProdCase[OF this, of ls n] Fun(2) show ?thesis by auto
  qed
qed

sublocale logic
  where I = I
    and type_of_fun = type_of_fun
    and Values_of_type = Values_of_type
    and Bool_types = "{BoolT}"
    and to_bool = to_bool
    and type_fixer = "TYPE(_)"
proof
  fix f ds
  show "length ds = length (param_types f) \<Longrightarrow>
       (\<And>i. i < length ds \<Longrightarrow> ds ! i \<in> Values_of_type (param_types f ! i)) \<Longrightarrow>
       I f ds \<in> Values_of_type (return_type f)"
  by (cases f; induct f ds rule: RA.I.induct, auto)
  show "Values_of_type ty \<noteq> {}" for ty by(cases ty, auto)
qed

lemmas [simp] = less_val_def compare_val_def lt_of_comp_def comparator_of_def

lemma less_RA_value_simps[simp]:
 "\<And>x y. Bool x > Bool y \<longleftrightarrow> x > y"
 "\<And>x y. Bool x > Rat y"
 "\<And>x y. Rat x > Rat y \<longleftrightarrow> x > y"
 "\<And>x y. Rat x > Bool y \<longleftrightarrow> False"
proof-
  show "Bool x > Bool y \<longleftrightarrow> x > y" for x y by (cases x; cases y, auto)
qed auto

lemmas [simp del] = less_val_def compare_val_def lt_of_comp_def comparator_of_def

lemma val_le[simp]: "(x :: val) \<ge> y \<longleftrightarrow> x = y \<or> x > y" by auto

end

subsection \<open>Non-term, polynomial version\<close>

type_synonym 'v rpoly = "('v,rat)poly" 

context
  notes [[typedef_overloaded]]
begin
datatype 'v poly_constraint = 
  Poly_Ge "'v rpoly" (* meaning: p \<ge> 0 *)
| Poly_Eq "'v rpoly" (* meaning: p = 0 *) 
| Poly_Gt "'v rpoly" (* meaning: p > 0 *) 
end

context RA_locale begin

fun showsl_poly_constraint :: "'v :: {showl,linorder} poly_constraint \<Rightarrow> showsl" where 
  "showsl_poly_constraint (Poly_Ge p) = (showsl_poly p o showsl_lit (STR '' >= 0''))"
| "showsl_poly_constraint (Poly_Eq p) = (showsl_poly p o showsl_lit (STR '' = 0''))"
| "showsl_poly_constraint (Poly_Gt p) = (showsl_poly p o showsl_lit (STR '' > 0''))"

fun interpret_poly_constraint :: "('v :: linorder \<Rightarrow> rat) \<Rightarrow> 'v poly_constraint \<Rightarrow> bool" where
  "interpret_poly_constraint f (Poly_Ge p) = (eval_poly f p \<ge> 0)" 
| "interpret_poly_constraint f (Poly_Eq p) = (eval_poly f p = 0)" 
| "interpret_poly_constraint f (Poly_Gt p) = (eval_poly f p > 0)" 
  
fun vars_poly_constraint :: "'v :: linorder poly_constraint \<Rightarrow> 'v set" where
  "vars_poly_constraint (Poly_Ge p) = poly_vars p"
| "vars_poly_constraint (Poly_Eq p) = poly_vars p"
| "vars_poly_constraint (Poly_Gt p) = poly_vars p"

fun vars_poly_constraint_list :: "'v :: linorder poly_constraint \<Rightarrow> 'v list" where
  "vars_poly_constraint_list (Poly_Ge p) = poly_vars_list p"
| "vars_poly_constraint_list (Poly_Eq p) = poly_vars_list p"
| "vars_poly_constraint_list (Poly_Gt p) = poly_vars_list p"
  
lemma vars_poly_constraint_list[simp]: "set (vars_poly_constraint_list c) = vars_poly_constraint c" 
  by (cases c, auto)

fun RA_exp_to_tpoly :: "'v RA_exp \<Rightarrow> ('v,rat) tpoly" where
 "RA_exp_to_tpoly (Var (a,ty)) = PVar a" |
 "RA_exp_to_tpoly (Fun (SumF _) as) = PSum (map RA_exp_to_tpoly as)" |
 "RA_exp_to_tpoly (Fun (ConstF a) []) = PNum a" |
 "RA_exp_to_tpoly (Fun (ProdF _) as) = PMult (map RA_exp_to_tpoly as)"
    
definition RA_exp_to_poly :: "'v :: linorder RA_exp \<Rightarrow> ('v,rat) poly" where
  "RA_exp_to_poly = poly_of o RA_exp_to_tpoly" 

fun RA_exp_to_poly_constraint :: "'v :: linorder RA_exp \<Rightarrow> 'v poly_constraint" where
  "RA_exp_to_poly_constraint (Fun LeF [a, b]) = Poly_Ge (poly_minus (RA_exp_to_poly b) (RA_exp_to_poly a))" 
| "RA_exp_to_poly_constraint (Fun EqF [a, b]) = Poly_Eq (poly_minus (RA_exp_to_poly b) (RA_exp_to_poly a))" 
| "RA_exp_to_poly_constraint (Fun LessF [a, b]) = Poly_Gt (poly_minus (RA_exp_to_poly b) (RA_exp_to_poly a))" 

lemma RA_exp_to_tpoly:
  assumes \<alpha>: "assignment \<alpha>"
  and "e :\<^sub>f RatT"
  shows "\<lbrakk>e\<rbrakk>\<alpha> = Rat (eval_tpoly (\<lambda>a. to_rat (\<alpha> (a, RatT))) (RA_exp_to_tpoly e))"
proof -
  have "e :\<^sub>f RatT \<Longrightarrow> ?thesis" 
  proof(induct e RatT rule:has_type_induct)
    case (Var x)
    from \<alpha> obtain i where "\<alpha> (x, RatT) = Rat i" by force
    then show ?case by auto
  next
    case (Sum es n)
    then show ?case by (auto, induct es, auto)
  next
    case (Prod es n)
    then show ?case by (auto, induct es, auto)
  qed auto
  then show ?thesis using assms by simp
qed
  
lemma RA_exp_to_poly:
  assumes \<alpha>: "assignment \<alpha>"
  and "e :\<^sub>f RatT"
shows "\<lbrakk>e\<rbrakk>\<alpha> = Rat (eval_poly (\<lambda>a. to_rat (\<alpha> (a, RatT))) (RA_exp_to_poly e))"
  using RA_exp_to_tpoly[OF assms] unfolding RA_exp_to_poly_def o_def poly_of .
  
lemma RA_exp_to_poly_constraint: 
  assumes \<alpha>: "assignment \<alpha>"
  and e: "is_bool e"
  shows "to_bool (\<lbrakk>e\<rbrakk>\<alpha>) = (interpret_poly_constraint (\<lambda>a. to_rat (\<alpha> (a, RatT))) (RA_exp_to_poly_constraint e))"
proof -
  have "e :\<^sub>f BoolT \<Longrightarrow> is_Fun e \<Longrightarrow> ?thesis"
    by (induct e BoolT rule: has_type_induct, insert RA_exp_to_poly[OF \<alpha>], auto)
  then show ?thesis using e[unfolded is_bool_def] by auto
qed

fun showsl_RA_exp where
  "showsl_RA_exp (Fun LessF [s,t]) = showsl_RA_exp s o showsl_lit (STR '' < '') o showsl_RA_exp t"
| "showsl_RA_exp (Fun LeF [s,t]) = showsl_RA_exp s o showsl_lit (STR '' <= '') o showsl_RA_exp t"
| "showsl_RA_exp (Fun EqF [s,t]) = showsl_RA_exp s o showsl_lit (STR '' = '') o showsl_RA_exp t"
| "showsl_RA_exp e = showsl_poly (RA_exp_to_poly e)"

definition "showsl_RA_clause = showsl_list_gen showsl_RA_exp (STR ''FALSE'') (STR '''') (STR '' || '') (STR '''')" 

definition "showsl_RA_formula = showsl_list_gen (showsl_paren o showsl_RA_clause) (STR ''TRUE'') (STR '''') (STR '' && '') (STR '''')" 

definition def_checker_main where
  "def_checker_main x ty l1 r1 l2 r2 = (is_bool (Fun LeF [l1, r1]) \<and>
    r1 :\<^sub>f ty \<and> is_bool (Fun LeF [l2, r2]) \<and> Var (x,ty) = l1 \<and> l1 = r2 \<and> l2 = r1 \<and> (x,ty) \<notin> vars_term r1)" 

definition def_checker_main' where
  "def_checker_main' x ty l1 r1 = (is_bool (Fun EqF [l1, r1]) \<and>
    r1 :\<^sub>f ty \<and> Var (x,ty) = l1 \<and> (x,ty) \<notin> vars_term r1)" 

definition def_checker_eq where "def_checker_eq x ty eq =
  (case eq of Atom e1 \<Rightarrow> (case e1 of Fun EqF [l1,r1] \<Rightarrow> 
              check (def_checker_main' x ty l1 r1 \<or> def_checker_main' x ty r1 l1)
              (showsl_lit (STR ''Non-equation not supported'')) | _ \<Rightarrow> error (showsl_lit (STR ''Invalid literal'')))
            | _ \<Rightarrow> error (showsl_lit (STR ''Invalid clause'')))" 

definition def_checker :: "(sig, 'v :: {showl, linorder}, ty) definability_checker"
(* unfortunately, a nice pattern matching doesn't go well for proving... *)
where "def_checker x ty \<phi> \<equiv>
 (case \<phi> of Conjunction \<phi>s \<Rightarrow>
   (case \<phi>s of [] \<Rightarrow> succeed
    | [eq] \<Rightarrow> def_checker_eq x ty eq
    | [lt1, lt2] \<Rightarrow>
     (case lt1 of Atom e1 \<Rightarrow>
       (case e1 of Fun LeF [l1,r1] \<Rightarrow>
         (case lt2 of Atom e2 \<Rightarrow>
           (case e2 of Fun LeF [l2,r2] \<Rightarrow>
              check (def_checker_main x ty l1 r1 l2 r2 \<or> def_checker_main x ty l2 r2 l1 r1)
              (showsl_lit (STR ''Non-equation not supported''))
             | _ \<Rightarrow> error (showsl_lit (STR ''Invalid 2nd literal''))
           )
          | _ \<Rightarrow> error (showsl_lit (STR ''Invalid 2nd clause''))
         )
        | _ \<Rightarrow> error (showsl_lit (STR ''Invalid 1st literal''))
        )
      | _ \<Rightarrow> error (showsl_lit (STR ''Invalid 1st clause''))
     )
   | _ \<Rightarrow> error (showsl_lit (STR ''Conjunction of more than two''))
  )
 | _ \<Rightarrow> def_checker_eq x ty \<phi>)
 <+? (\<lambda> s. showsl_lit (STR ''error in definibility checker for '') o showsl x o showsl_lit (STR '' on formula\<newline>  '') o showsl_formula.showsl_formula showsl_RA_exp \<phi> o showsl_nl o s)"

sublocale definability_checker
  where I = I
    and type_of_fun = type_of_fun
    and Values_of_type = Values_of_type
    and Bool_types = "{BoolT}"
    and to_bool = to_bool
    and definability_checker = "def_checker :: (_,'v :: {showl, linorder},_) definability_checker"
    and type_fixer = "TYPE(_)"
proof
  fix \<alpha> :: "'v \<times> ty \<Rightarrow> val" and x :: 'v and ty \<phi>
  assume ok: "isOK (def_checker x ty \<phi>)" and alpha: "assignment \<alpha>"
  note ok = ok[unfolded def_checker_def, simplified]
  { fix l1 r1 l2 r2
    assume "def_checker_main x ty l1 r1 l2 r2"
    note ok = this[unfolded def_checker_main_def]
    then have "is_bool (Fun LeF [l1, r1])" and "is_bool (Fun LeF [l2, r2])" by auto
    from ok have id: "l1 = Var (x,ty)" "r2 = Var (x,ty)" "l2 = r1" and fresh: "(x,ty) \<notin> vars_term r1" by auto
    from fresh have id2: "(\<lbrakk>r1\<rbrakk> (\<alpha>((x, ty) := a))) = (\<lbrakk>r1\<rbrakk> \<alpha>)" for a by simp
    from ok have "r1 :\<^sub>f ty" by auto
    with alpha have r1: "\<lbrakk>r1\<rbrakk> \<alpha> \<in> Values_of_type ty"
      using eval_types by auto
    have "(\<exists>a \<in> Values_of_type ty.
    \<alpha>((x, ty) := a) \<Turnstile> Atom (Fun LeF [l1, r1]) \<and>
    \<alpha>((x, ty) := a) \<Turnstile> Atom (Fun LeF [l2, r2]))"
      unfolding id using r1 by (intro bexI[of _ "\<lbrakk>r1\<rbrakk> \<alpha>"], auto simp: id2)
  } note * = this
  { fix l1 r1
    assume "def_checker_main' x ty l1 r1"
    note ok = this[unfolded def_checker_main'_def]
    then have "is_bool (Fun EqF [l1, r1])" by auto
    from ok have id: "l1 = Var (x,ty)" and fresh: "(x,ty) \<notin> vars_term r1" by auto
    from fresh have id2: "(\<lbrakk>r1\<rbrakk> (\<alpha>((x, ty) := a))) = (\<lbrakk>r1\<rbrakk> \<alpha>)" for a by simp
    from ok have "r1 :\<^sub>f ty" by auto
    with alpha have r1: "\<lbrakk>r1\<rbrakk> \<alpha> \<in> Values_of_type ty"
      using eval_types by auto
    have "\<exists>a \<in> Values_of_type ty.
      to_rat (\<lbrakk>l1\<rbrakk> (\<alpha>((x, ty) := a))) = to_rat (\<lbrakk>r1\<rbrakk> (\<alpha>((x, ty) := a)))"
      unfolding id using r1 by (intro bexI[of _ "\<lbrakk>r1\<rbrakk> \<alpha>"], auto simp: id2)
  } note ** = this
  {
    fix eq
    assume ok: "isOK(def_checker_eq x ty eq)"
    note ok = ok[unfolded def_checker_eq_def]
    from ok obtain l1 r1 where [simp]: "eq = Atom (Fun EqF [l1,r1])"
      by (auto split: list.splits sig.splits formula.splits term.splits)
    note **[dest!]
    note ok = ok[simplified]
    from ok have "\<exists>a\<in>Values_of_type ty. \<alpha>((x, ty) := a) \<Turnstile> eq" by force
  } note *** = this
  show "\<exists>a \<in> Values_of_type ty. \<alpha>((x, ty) := a) \<Turnstile> \<phi>"
  proof (cases \<phi>)
    case [simp]: (Conjunction \<phi>s)
    note ok = ok[simplified]
    show ?thesis
    proof (cases \<phi>s rule: list_4_cases)
      case Nil with ok show ?thesis by (cases ty, auto)
    next
      case [simp]: (2 cl1 cl2)
      note ok = ok[simplified]
      from ok obtain l1 r1 where [simp]: "cl1 = Atom (Fun LeF [l1,r1])"
        by (auto split: list.splits sig.splits formula.splits term.splits)
      with ok obtain l2 r2 where [simp]: "cl2 = Atom (Fun LeF [l2,r2])"
        by (auto split: list.splits sig.splits formula.splits term.splits)
      note ok = ok[simplified]
      note *[dest!]
      from ok show ?thesis by auto
    next
      case [simp]: (1 eq)
      show ?thesis using ok ***[of eq] by auto
    qed (insert ok, auto)
  next
    case [simp]: (Atom eq)
    with ok ***[of "Atom eq"] show ?thesis by auto
  qed (insert ok, auto simp: def_checker_eq_def)
qed
end

end
