module Uint : sig
  type t = int
  val dflt_size : Big_int.big_int
  val less : t -> t -> bool
  val less_eq : t -> t -> bool
  val set_bit : t -> Big_int.big_int -> bool -> t
  val shiftl : t -> Big_int.big_int -> t
  val shiftr : t -> Big_int.big_int -> t
  val shiftr_signed : t -> Big_int.big_int -> t
  val test_bit : t -> Big_int.big_int -> bool
  val int_mask : int
  val int32_mask : int32
  val int64_mask : int64
end = struct

type t = int

(* Can be replaced with Sys.int_size in OCaml 4.03.0 *)
let dflt_size_int = 
  let rec f n = if n=0 then 0 else f (n / 2) + 1 
  in f min_int;;

let dflt_size = Big_int.big_int_of_int dflt_size_int;;

(* negative numbers have their highest bit set, 
   so they are greater than positive ones *)
let less x y =
  if x<0 then
    y<0 && x<y
  else y < 0 || x < y;;

let less_eq x y =
  if x < 0 then
    y < 0 &&  x <= y
  else y < 0 || x <= y;;

let set_bit x n b =
  let mask = 1 lsl (Big_int.int_of_big_int n)
  in if b then x lor mask
     else x land (lnot mask);;

let shiftl x n = x lsl (Big_int.int_of_big_int n);;

let shiftr x n = x lsr (Big_int.int_of_big_int n);;

let shiftr_signed x n = x asr (Big_int.int_of_big_int n);;

let test_bit x n = x land (1 lsl (Big_int.int_of_big_int n)) <> 0;;

let int_mask =
  if dflt_size_int < 32 then lnot 0 else 0xFFFFFFFF;;

let int32_mask = 
  if dflt_size_int < 32 then Int32.pred (Int32.shift_left Int32.one dflt_size_int) 
  else Int32.of_string "0xFFFFFFFF";;

let int64_mask = 
  if dflt_size_int < 64 then Int64.pred (Int64.shift_left Int64.one dflt_size_int) 
  else Int64.of_string "0xFFFFFFFFFFFFFFFF";;

end;; (*struct Uint*)

module Uint32 : sig
  val less : int32 -> int32 -> bool
  val less_eq : int32 -> int32 -> bool
  val set_bit : int32 -> Big_int.big_int -> bool -> int32
  val shiftl : int32 -> Big_int.big_int -> int32
  val shiftr : int32 -> Big_int.big_int -> int32
  val shiftr_signed : int32 -> Big_int.big_int -> int32
  val test_bit : int32 -> Big_int.big_int -> bool
end = struct

(* negative numbers have their highest bit set, 
   so they are greater than positive ones *)
let less x y =
  if Int32.compare x Int32.zero < 0 then
    Int32.compare y Int32.zero < 0 && Int32.compare x y < 0
  else Int32.compare y Int32.zero < 0 || Int32.compare x y < 0;;

let less_eq x y =
  if Int32.compare x Int32.zero < 0 then
    Int32.compare y Int32.zero < 0 && Int32.compare x y <= 0
  else Int32.compare y Int32.zero < 0 || Int32.compare x y <= 0;;

let set_bit x n b =
  let mask = Int32.shift_left Int32.one (Big_int.int_of_big_int n)
  in if b then Int32.logor x mask
     else Int32.logand x (Int32.lognot mask);;

let shiftl x n = Int32.shift_left x (Big_int.int_of_big_int n);;

let shiftr x n = Int32.shift_right_logical x (Big_int.int_of_big_int n);;

let shiftr_signed x n = Int32.shift_right x (Big_int.int_of_big_int n);;

let test_bit x n =
  Int32.compare 
    (Int32.logand x (Int32.shift_left Int32.one (Big_int.int_of_big_int n)))
    Int32.zero
  <> 0;;

end;; (*struct Uint32*)

module Bits_Integer : sig
  val and_pninteger : Big_int.big_int -> Big_int.big_int -> Big_int.big_int
  val or_pninteger : Big_int.big_int -> Big_int.big_int -> Big_int.big_int
  val shiftl : Big_int.big_int -> Big_int.big_int -> Big_int.big_int
  val shiftr : Big_int.big_int -> Big_int.big_int -> Big_int.big_int
  val test_bit : Big_int.big_int -> Big_int.big_int -> bool
end = struct

let and_pninteger bi1 bi2 =
  Big_int.and_big_int bi1
    (Big_int.xor_big_int
      (Big_int.pred_big_int
        (Big_int.shift_left_big_int Big_int.unit_big_int
           (max (Big_int.num_digits_big_int bi1 * Nat.length_of_digit)
                (Big_int.num_digits_big_int bi2 * Nat.length_of_digit))))
      (Big_int.pred_big_int (Big_int.minus_big_int bi2)));;

let or_pninteger bi1 bi2 =
  Big_int.pred_big_int (Big_int.minus_big_int
    (Big_int.and_big_int
      (Big_int.xor_big_int
         (Big_int.pred_big_int
           (Big_int.shift_left_big_int Big_int.unit_big_int
              (max (Big_int.num_digits_big_int bi1 * Nat.length_of_digit)
                   (Big_int.num_digits_big_int bi2 * Nat.length_of_digit))))
         bi1)
      (Big_int.pred_big_int (Big_int.minus_big_int bi2))));;

(* We do not need an explicit range checks here,
   because Big_int.int_of_big_int raises Failure 
   if the argument does not fit into an int. *)
let shiftl x n = Big_int.shift_left_big_int x (Big_int.int_of_big_int n);;

let shiftr x n = Big_int.shift_right_big_int x (Big_int.int_of_big_int n);;

let test_bit x n = 
  Big_int.eq_big_int 
    (Big_int.extract_big_int x (Big_int.int_of_big_int n) 1) 
    Big_int.unit_big_int

end;; (*struct Bits_Integer*)

module Ceta : sig
  type nat
  type char
  type ('a, 'b) sum
  type ('a, 'b) lab
  type ('a, 'b) claim
  type cert_result = Certified | Unsupported of string | Error of string
  val certifyproof : string -> cert_result
end = struct

type int = Int_of_integer of Big_int.big_int;;

let rec integer_of_int (Int_of_integer k) = k;;

let rec equal_inta
  k l = Big_int.eq_big_int (integer_of_int k) (integer_of_int l);;

type 'a equal = {equal : 'a -> 'a -> bool};;
let equal _A = _A.equal;;

let equal_int = ({equal = equal_inta} : int equal);;

let rec times_inta
  k l = Int_of_integer
          (Big_int.mult_big_int (integer_of_int k) (integer_of_int l));;

type 'a times = {times : 'a -> 'a -> 'a};;
let times _A = _A.times;;

type 'a dvd = {times_dvd : 'a times};;

let times_int = ({times = times_inta} : int times);;

let dvd_int = ({times_dvd = times_int} : int dvd);;

let rec uminus_inta
  k = Int_of_integer (Big_int.minus_big_int (integer_of_int k));;

let zero_inta : int = Int_of_integer Big_int.zero_big_int;;

let rec less_int
  k l = Big_int.lt_big_int (integer_of_int k) (integer_of_int l);;

let rec abs_inta i = (if less_int i zero_inta then uminus_inta i else i);;

type 'a abs = {abs : 'a -> 'a};;
let abs _A = _A.abs;;

let abs_int = ({abs = abs_inta} : int abs);;

type num = One | Bit0 of num | Bit1 of num;;

let one_inta : int = Int_of_integer (Big_int.big_int_of_int 1);;

type 'a one = {onea : 'a};;
let onea _A = _A.onea;;

let one_int = ({onea = one_inta} : int one);;

let rec sgn_inta
  i = (if equal_inta i zero_inta then zero_inta
        else (if less_int zero_inta i then one_inta
               else uminus_inta one_inta));;

type 'a sgn = {sgn : 'a -> 'a};;
let sgn _A = _A.sgn;;

let sgn_int = ({sgn = sgn_inta} : int sgn);;

let rec minus_inta
  k l = Int_of_integer
          (Big_int.sub_big_int (integer_of_int k) (integer_of_int l));;

let rec plus_inta
  k l = Int_of_integer
          (Big_int.add_big_int (integer_of_int k) (integer_of_int l));;

type 'a uminus = {uminus : 'a -> 'a};;
let uminus _A = _A.uminus;;

type 'a minus = {minus : 'a -> 'a -> 'a};;
let minus _A = _A.minus;;

type 'a zero = {zerob : 'a};;
let zerob _A = _A.zerob;;

type 'a plus = {plus : 'a -> 'a -> 'a};;
let plus _A = _A.plus;;

type 'a semigroup_add = {plus_semigroup_add : 'a plus};;

type 'a cancel_semigroup_add =
  {semigroup_add_cancel_semigroup_add : 'a semigroup_add};;

type 'a ab_semigroup_add = {semigroup_add_ab_semigroup_add : 'a semigroup_add};;

type 'a cancel_ab_semigroup_add =
  {ab_semigroup_add_cancel_ab_semigroup_add : 'a ab_semigroup_add;
    cancel_semigroup_add_cancel_ab_semigroup_add : 'a cancel_semigroup_add;
    minus_cancel_ab_semigroup_add : 'a minus};;

type 'a monoid_add =
  {semigroup_add_monoid_add : 'a semigroup_add; zero_monoid_add : 'a zero};;

type 'a comm_monoid_add =
  {ab_semigroup_add_comm_monoid_add : 'a ab_semigroup_add;
    monoid_add_comm_monoid_add : 'a monoid_add};;

type 'a cancel_comm_monoid_add =
  {cancel_ab_semigroup_add_cancel_comm_monoid_add : 'a cancel_ab_semigroup_add;
    comm_monoid_add_cancel_comm_monoid_add : 'a comm_monoid_add};;

type 'a mult_zero = {times_mult_zero : 'a times; zero_mult_zero : 'a zero};;

type 'a semigroup_mult = {times_semigroup_mult : 'a times};;

type 'a semiring =
  {ab_semigroup_add_semiring : 'a ab_semigroup_add;
    semigroup_mult_semiring : 'a semigroup_mult};;

type 'a semiring_0 =
  {comm_monoid_add_semiring_0 : 'a comm_monoid_add;
    mult_zero_semiring_0 : 'a mult_zero; semiring_semiring_0 : 'a semiring};;

type 'a semiring_0_cancel =
  {cancel_comm_monoid_add_semiring_0_cancel : 'a cancel_comm_monoid_add;
    semiring_0_semiring_0_cancel : 'a semiring_0};;

type 'a ab_semigroup_mult =
  {semigroup_mult_ab_semigroup_mult : 'a semigroup_mult};;

type 'a comm_semiring =
  {ab_semigroup_mult_comm_semiring : 'a ab_semigroup_mult;
    semiring_comm_semiring : 'a semiring};;

type 'a comm_semiring_0 =
  {comm_semiring_comm_semiring_0 : 'a comm_semiring;
    semiring_0_comm_semiring_0 : 'a semiring_0};;

type 'a comm_semiring_0_cancel =
  {comm_semiring_0_comm_semiring_0_cancel : 'a comm_semiring_0;
    semiring_0_cancel_comm_semiring_0_cancel : 'a semiring_0_cancel};;

type 'a power = {one_power : 'a one; times_power : 'a times};;

type 'a monoid_mult =
  {semigroup_mult_monoid_mult : 'a semigroup_mult;
    power_monoid_mult : 'a power};;

type 'a numeral =
  {one_numeral : 'a one; semigroup_add_numeral : 'a semigroup_add};;

type 'a semiring_numeral =
  {monoid_mult_semiring_numeral : 'a monoid_mult;
    numeral_semiring_numeral : 'a numeral;
    semiring_semiring_numeral : 'a semiring};;

type 'a zero_neq_one =
  {one_zero_neq_one : 'a one; zero_zero_neq_one : 'a zero};;

type 'a semiring_1 =
  {semiring_numeral_semiring_1 : 'a semiring_numeral;
    semiring_0_semiring_1 : 'a semiring_0;
    zero_neq_one_semiring_1 : 'a zero_neq_one};;

type 'a semiring_1_cancel =
  {semiring_0_cancel_semiring_1_cancel : 'a semiring_0_cancel;
    semiring_1_semiring_1_cancel : 'a semiring_1};;

type 'a comm_monoid_mult =
  {ab_semigroup_mult_comm_monoid_mult : 'a ab_semigroup_mult;
    monoid_mult_comm_monoid_mult : 'a monoid_mult;
    dvd_comm_monoid_mult : 'a dvd};;

type 'a comm_semiring_1 =
  {comm_monoid_mult_comm_semiring_1 : 'a comm_monoid_mult;
    comm_semiring_0_comm_semiring_1 : 'a comm_semiring_0;
    semiring_1_comm_semiring_1 : 'a semiring_1};;

type 'a comm_semiring_1_cancel =
  {comm_semiring_0_cancel_comm_semiring_1_cancel : 'a comm_semiring_0_cancel;
    comm_semiring_1_comm_semiring_1_cancel : 'a comm_semiring_1;
    semiring_1_cancel_comm_semiring_1_cancel : 'a semiring_1_cancel};;

type 'a comm_semiring_1_cancel_crossproduct =
  {comm_semiring_1_cancel_comm_semiring_1_cancel_crossproduct :
     'a comm_semiring_1_cancel};;

type 'a semiring_no_zero_divisors =
  {semiring_0_semiring_no_zero_divisors : 'a semiring_0};;

type 'a semiring_1_no_zero_divisors =
  {semiring_1_semiring_1_no_zero_divisors : 'a semiring_1;
    semiring_no_zero_divisors_semiring_1_no_zero_divisors :
      'a semiring_no_zero_divisors};;

type 'a semiring_no_zero_divisors_cancel =
  {semiring_no_zero_divisors_semiring_no_zero_divisors_cancel :
     'a semiring_no_zero_divisors};;

type 'a group_add =
  {cancel_semigroup_add_group_add : 'a cancel_semigroup_add;
    minus_group_add : 'a minus; monoid_add_group_add : 'a monoid_add;
    uminus_group_add : 'a uminus};;

type 'a ab_group_add =
  {cancel_comm_monoid_add_ab_group_add : 'a cancel_comm_monoid_add;
    group_add_ab_group_add : 'a group_add};;

type 'a ring =
  {ab_group_add_ring : 'a ab_group_add;
    semiring_0_cancel_ring : 'a semiring_0_cancel};;

type 'a ring_no_zero_divisors =
  {ring_ring_no_zero_divisors : 'a ring;
    semiring_no_zero_divisors_cancel_ring_no_zero_divisors :
      'a semiring_no_zero_divisors_cancel};;

type 'a neg_numeral =
  {group_add_neg_numeral : 'a group_add; numeral_neg_numeral : 'a numeral};;

type 'a ring_1 =
  {neg_numeral_ring_1 : 'a neg_numeral; ring_ring_1 : 'a ring;
    semiring_1_cancel_ring_1 : 'a semiring_1_cancel};;

type 'a ring_1_no_zero_divisors =
  {ring_1_ring_1_no_zero_divisors : 'a ring_1;
    ring_no_zero_divisors_ring_1_no_zero_divisors : 'a ring_no_zero_divisors;
    semiring_1_no_zero_divisors_ring_1_no_zero_divisors :
      'a semiring_1_no_zero_divisors};;

type 'a comm_ring =
  {comm_semiring_0_cancel_comm_ring : 'a comm_semiring_0_cancel;
    ring_comm_ring : 'a ring};;

type 'a comm_ring_1 =
  {comm_ring_comm_ring_1 : 'a comm_ring;
    comm_semiring_1_cancel_comm_ring_1 : 'a comm_semiring_1_cancel;
    ring_1_comm_ring_1 : 'a ring_1};;

type 'a semidom =
  {comm_semiring_1_cancel_semidom : 'a comm_semiring_1_cancel;
    semiring_1_no_zero_divisors_semidom : 'a semiring_1_no_zero_divisors};;

type 'a idom =
  {comm_ring_1_idom : 'a comm_ring_1;
    ring_1_no_zero_divisors_idom : 'a ring_1_no_zero_divisors;
    semidom_idom : 'a semidom;
    comm_semiring_1_cancel_crossproduct_idom :
      'a comm_semiring_1_cancel_crossproduct};;

let plus_int = ({plus = plus_inta} : int plus);;

let semigroup_add_int = ({plus_semigroup_add = plus_int} : int semigroup_add);;

let cancel_semigroup_add_int =
  ({semigroup_add_cancel_semigroup_add = semigroup_add_int} :
    int cancel_semigroup_add);;

let ab_semigroup_add_int =
  ({semigroup_add_ab_semigroup_add = semigroup_add_int} :
    int ab_semigroup_add);;

let minus_int = ({minus = minus_inta} : int minus);;

let cancel_ab_semigroup_add_int =
  ({ab_semigroup_add_cancel_ab_semigroup_add = ab_semigroup_add_int;
     cancel_semigroup_add_cancel_ab_semigroup_add = cancel_semigroup_add_int;
     minus_cancel_ab_semigroup_add = minus_int}
    : int cancel_ab_semigroup_add);;

let zero_int = ({zerob = zero_inta} : int zero);;

let monoid_add_int =
  ({semigroup_add_monoid_add = semigroup_add_int; zero_monoid_add = zero_int} :
    int monoid_add);;

let comm_monoid_add_int =
  ({ab_semigroup_add_comm_monoid_add = ab_semigroup_add_int;
     monoid_add_comm_monoid_add = monoid_add_int}
    : int comm_monoid_add);;

let cancel_comm_monoid_add_int =
  ({cancel_ab_semigroup_add_cancel_comm_monoid_add =
      cancel_ab_semigroup_add_int;
     comm_monoid_add_cancel_comm_monoid_add = comm_monoid_add_int}
    : int cancel_comm_monoid_add);;

let mult_zero_int =
  ({times_mult_zero = times_int; zero_mult_zero = zero_int} : int mult_zero);;

let semigroup_mult_int =
  ({times_semigroup_mult = times_int} : int semigroup_mult);;

let semiring_int =
  ({ab_semigroup_add_semiring = ab_semigroup_add_int;
     semigroup_mult_semiring = semigroup_mult_int}
    : int semiring);;

let semiring_0_int =
  ({comm_monoid_add_semiring_0 = comm_monoid_add_int;
     mult_zero_semiring_0 = mult_zero_int; semiring_semiring_0 = semiring_int}
    : int semiring_0);;

let semiring_0_cancel_int =
  ({cancel_comm_monoid_add_semiring_0_cancel = cancel_comm_monoid_add_int;
     semiring_0_semiring_0_cancel = semiring_0_int}
    : int semiring_0_cancel);;

let ab_semigroup_mult_int =
  ({semigroup_mult_ab_semigroup_mult = semigroup_mult_int} :
    int ab_semigroup_mult);;

let comm_semiring_int =
  ({ab_semigroup_mult_comm_semiring = ab_semigroup_mult_int;
     semiring_comm_semiring = semiring_int}
    : int comm_semiring);;

let comm_semiring_0_int =
  ({comm_semiring_comm_semiring_0 = comm_semiring_int;
     semiring_0_comm_semiring_0 = semiring_0_int}
    : int comm_semiring_0);;

let comm_semiring_0_cancel_int =
  ({comm_semiring_0_comm_semiring_0_cancel = comm_semiring_0_int;
     semiring_0_cancel_comm_semiring_0_cancel = semiring_0_cancel_int}
    : int comm_semiring_0_cancel);;

let power_int = ({one_power = one_int; times_power = times_int} : int power);;

let monoid_mult_int =
  ({semigroup_mult_monoid_mult = semigroup_mult_int;
     power_monoid_mult = power_int}
    : int monoid_mult);;

let numeral_int =
  ({one_numeral = one_int; semigroup_add_numeral = semigroup_add_int} :
    int numeral);;

let semiring_numeral_int =
  ({monoid_mult_semiring_numeral = monoid_mult_int;
     numeral_semiring_numeral = numeral_int;
     semiring_semiring_numeral = semiring_int}
    : int semiring_numeral);;

let zero_neq_one_int =
  ({one_zero_neq_one = one_int; zero_zero_neq_one = zero_int} :
    int zero_neq_one);;

let semiring_1_int =
  ({semiring_numeral_semiring_1 = semiring_numeral_int;
     semiring_0_semiring_1 = semiring_0_int;
     zero_neq_one_semiring_1 = zero_neq_one_int}
    : int semiring_1);;

let semiring_1_cancel_int =
  ({semiring_0_cancel_semiring_1_cancel = semiring_0_cancel_int;
     semiring_1_semiring_1_cancel = semiring_1_int}
    : int semiring_1_cancel);;

let comm_monoid_mult_int =
  ({ab_semigroup_mult_comm_monoid_mult = ab_semigroup_mult_int;
     monoid_mult_comm_monoid_mult = monoid_mult_int;
     dvd_comm_monoid_mult = dvd_int}
    : int comm_monoid_mult);;

let comm_semiring_1_int =
  ({comm_monoid_mult_comm_semiring_1 = comm_monoid_mult_int;
     comm_semiring_0_comm_semiring_1 = comm_semiring_0_int;
     semiring_1_comm_semiring_1 = semiring_1_int}
    : int comm_semiring_1);;

let comm_semiring_1_cancel_int =
  ({comm_semiring_0_cancel_comm_semiring_1_cancel = comm_semiring_0_cancel_int;
     comm_semiring_1_comm_semiring_1_cancel = comm_semiring_1_int;
     semiring_1_cancel_comm_semiring_1_cancel = semiring_1_cancel_int}
    : int comm_semiring_1_cancel);;

let comm_semiring_1_cancel_crossproduct_int =
  ({comm_semiring_1_cancel_comm_semiring_1_cancel_crossproduct =
      comm_semiring_1_cancel_int}
    : int comm_semiring_1_cancel_crossproduct);;

let semiring_no_zero_divisors_int =
  ({semiring_0_semiring_no_zero_divisors = semiring_0_int} :
    int semiring_no_zero_divisors);;

let semiring_1_no_zero_divisors_int =
  ({semiring_1_semiring_1_no_zero_divisors = semiring_1_int;
     semiring_no_zero_divisors_semiring_1_no_zero_divisors =
       semiring_no_zero_divisors_int}
    : int semiring_1_no_zero_divisors);;

let semiring_no_zero_divisors_cancel_int =
  ({semiring_no_zero_divisors_semiring_no_zero_divisors_cancel =
      semiring_no_zero_divisors_int}
    : int semiring_no_zero_divisors_cancel);;

let uminus_int = ({uminus = uminus_inta} : int uminus);;

let group_add_int =
  ({cancel_semigroup_add_group_add = cancel_semigroup_add_int;
     minus_group_add = minus_int; monoid_add_group_add = monoid_add_int;
     uminus_group_add = uminus_int}
    : int group_add);;

let ab_group_add_int =
  ({cancel_comm_monoid_add_ab_group_add = cancel_comm_monoid_add_int;
     group_add_ab_group_add = group_add_int}
    : int ab_group_add);;

let ring_int =
  ({ab_group_add_ring = ab_group_add_int;
     semiring_0_cancel_ring = semiring_0_cancel_int}
    : int ring);;

let ring_no_zero_divisors_int =
  ({ring_ring_no_zero_divisors = ring_int;
     semiring_no_zero_divisors_cancel_ring_no_zero_divisors =
       semiring_no_zero_divisors_cancel_int}
    : int ring_no_zero_divisors);;

let neg_numeral_int =
  ({group_add_neg_numeral = group_add_int; numeral_neg_numeral = numeral_int} :
    int neg_numeral);;

let ring_1_int =
  ({neg_numeral_ring_1 = neg_numeral_int; ring_ring_1 = ring_int;
     semiring_1_cancel_ring_1 = semiring_1_cancel_int}
    : int ring_1);;

let ring_1_no_zero_divisors_int =
  ({ring_1_ring_1_no_zero_divisors = ring_1_int;
     ring_no_zero_divisors_ring_1_no_zero_divisors = ring_no_zero_divisors_int;
     semiring_1_no_zero_divisors_ring_1_no_zero_divisors =
       semiring_1_no_zero_divisors_int}
    : int ring_1_no_zero_divisors);;

let comm_ring_int =
  ({comm_semiring_0_cancel_comm_ring = comm_semiring_0_cancel_int;
     ring_comm_ring = ring_int}
    : int comm_ring);;

let comm_ring_1_int =
  ({comm_ring_comm_ring_1 = comm_ring_int;
     comm_semiring_1_cancel_comm_ring_1 = comm_semiring_1_cancel_int;
     ring_1_comm_ring_1 = ring_1_int}
    : int comm_ring_1);;

let semidom_int =
  ({comm_semiring_1_cancel_semidom = comm_semiring_1_cancel_int;
     semiring_1_no_zero_divisors_semidom = semiring_1_no_zero_divisors_int}
    : int semidom);;

let idom_int =
  ({comm_ring_1_idom = comm_ring_1_int;
     ring_1_no_zero_divisors_idom = ring_1_no_zero_divisors_int;
     semidom_idom = semidom_int;
     comm_semiring_1_cancel_crossproduct_idom =
       comm_semiring_1_cancel_crossproduct_int}
    : int idom);;

let rec less_eq_int
  k l = Big_int.le_big_int (integer_of_int k) (integer_of_int l);;

type 'a ord = {less_eq : 'a -> 'a -> bool; less : 'a -> 'a -> bool};;
let less_eq _A = _A.less_eq;;
let less _A = _A.less;;

type 'a abs_if =
  {abs_abs_if : 'a abs; minus_abs_if : 'a minus; uminus_abs_if : 'a uminus;
    zero_abs_if : 'a zero; ord_abs_if : 'a ord};;

let ord_int = ({less_eq = less_eq_int; less = less_int} : int ord);;

let abs_if_int =
  ({abs_abs_if = abs_int; minus_abs_if = minus_int; uminus_abs_if = uminus_int;
     zero_abs_if = zero_int; ord_abs_if = ord_int}
    : int abs_if);;

type 'a semiring_char_0 = {semiring_1_semiring_char_0 : 'a semiring_1};;

type 'a ring_char_0 =
  {semiring_char_0_ring_char_0 : 'a semiring_char_0;
    ring_1_ring_char_0 : 'a ring_1};;

let semiring_char_0_int =
  ({semiring_1_semiring_char_0 = semiring_1_int} : int semiring_char_0);;

let ring_char_0_int =
  ({semiring_char_0_ring_char_0 = semiring_char_0_int;
     ring_1_ring_char_0 = ring_1_int}
    : int ring_char_0);;

type 'a quasi_order = {ord_quasi_order : 'a ord};;

type 'a weak_order = {quasi_order_weak_order : 'a quasi_order};;

type 'a preorder = {ord_preorder : 'a ord};;

type 'a order =
  {preorder_order : 'a preorder; weak_order_order : 'a weak_order};;

let quasi_order_int = ({ord_quasi_order = ord_int} : int quasi_order);;

let weak_order_int =
  ({quasi_order_weak_order = quasi_order_int} : int weak_order);;

let preorder_int = ({ord_preorder = ord_int} : int preorder);;

let order_int =
  ({preorder_order = preorder_int; weak_order_order = weak_order_int} :
    int order);;

let ceq_inta : (int -> int -> bool) option = Some equal_inta;;

type 'a ceq = {ceq : ('a -> 'a -> bool) option};;
let ceq _A = _A.ceq;;

let ceq_int = ({ceq = ceq_inta} : int ceq);;

type ('a, 'b) phantom = Phantom of 'b;;

type set_impla = Set_Choose | Set_Collect | Set_DList | Set_RBT | Set_Monada;;

let set_impl_inta : (int, set_impla) phantom = Phantom Set_RBT;;

type 'a set_impl = {set_impl : ('a, set_impla) phantom};;
let set_impl _A = _A.set_impl;;

let set_impl_int = ({set_impl = set_impl_inta} : int set_impl);;

type 'a linorder = {order_linorder : 'a order};;

let linorder_int = ({order_linorder = order_int} : int linorder);;

type 'a idom_abs_sgn =
  {abs_idom_abs_sgn : 'a abs; sgn_idom_abs_sgn : 'a sgn;
    idom_idom_abs_sgn : 'a idom};;

let idom_abs_sgn_int =
  ({abs_idom_abs_sgn = abs_int; sgn_idom_abs_sgn = sgn_int;
     idom_idom_abs_sgn = idom_int}
    : int idom_abs_sgn);;

type 'a ordered_ab_semigroup_add =
  {ab_semigroup_add_ordered_ab_semigroup_add : 'a ab_semigroup_add;
    order_ordered_ab_semigroup_add : 'a order};;

type 'a strict_ordered_ab_semigroup_add =
  {ordered_ab_semigroup_add_strict_ordered_ab_semigroup_add :
     'a ordered_ab_semigroup_add};;

type 'a ordered_cancel_ab_semigroup_add =
  {cancel_ab_semigroup_add_ordered_cancel_ab_semigroup_add :
     'a cancel_ab_semigroup_add;
    strict_ordered_ab_semigroup_add_ordered_cancel_ab_semigroup_add :
      'a strict_ordered_ab_semigroup_add};;

type 'a ordered_semigroup_mult_zero =
  {semigroup_mult_ordered_semigroup_mult_zero : 'a semigroup_mult;
    order_ordered_semigroup_mult_zero : 'a order;
    mult_zero_ordered_semigroup_mult_zero : 'a mult_zero};;

type 'a ordered_comm_monoid_add =
  {comm_monoid_add_ordered_comm_monoid_add : 'a comm_monoid_add;
    ordered_ab_semigroup_add_ordered_comm_monoid_add :
      'a ordered_ab_semigroup_add};;

type 'a ordered_semiring =
  {ordered_comm_monoid_add_ordered_semiring : 'a ordered_comm_monoid_add;
    semiring_ordered_semiring : 'a semiring};;

type 'a ordered_semiring_0 =
  {ordered_semigroup_mult_zero_ordered_semiring_0 :
     'a ordered_semigroup_mult_zero;
    ordered_semiring_ordered_semiring_0 : 'a ordered_semiring;
    semiring_0_ordered_semiring_0 : 'a semiring_0};;

type 'a ordered_cancel_semiring =
  {ordered_cancel_ab_semigroup_add_ordered_cancel_semiring :
     'a ordered_cancel_ab_semigroup_add;
    ordered_semiring_0_ordered_cancel_semiring : 'a ordered_semiring_0;
    semiring_0_cancel_ordered_cancel_semiring : 'a semiring_0_cancel};;

type 'a ordered_ab_semigroup_add_imp_le =
  {ordered_cancel_ab_semigroup_add_ordered_ab_semigroup_add_imp_le :
     'a ordered_cancel_ab_semigroup_add};;

type 'a strict_ordered_comm_monoid_add =
  {comm_monoid_add_strict_ordered_comm_monoid_add : 'a comm_monoid_add;
    strict_ordered_ab_semigroup_add_strict_ordered_comm_monoid_add :
      'a strict_ordered_ab_semigroup_add};;

type 'a ordered_cancel_comm_monoid_add =
  {ordered_cancel_ab_semigroup_add_ordered_cancel_comm_monoid_add :
     'a ordered_cancel_ab_semigroup_add;
    ordered_comm_monoid_add_ordered_cancel_comm_monoid_add :
      'a ordered_comm_monoid_add;
    strict_ordered_comm_monoid_add_ordered_cancel_comm_monoid_add :
      'a strict_ordered_comm_monoid_add};;

type 'a ordered_ab_semigroup_monoid_add_imp_le =
  {cancel_comm_monoid_add_ordered_ab_semigroup_monoid_add_imp_le :
     'a cancel_comm_monoid_add;
    ordered_ab_semigroup_add_imp_le_ordered_ab_semigroup_monoid_add_imp_le :
      'a ordered_ab_semigroup_add_imp_le;
    ordered_cancel_comm_monoid_add_ordered_ab_semigroup_monoid_add_imp_le :
      'a ordered_cancel_comm_monoid_add};;

type 'a ordered_ab_group_add =
  {ab_group_add_ordered_ab_group_add : 'a ab_group_add;
    ordered_ab_semigroup_monoid_add_imp_le_ordered_ab_group_add :
      'a ordered_ab_semigroup_monoid_add_imp_le};;

type 'a ordered_ring =
  {ordered_ab_group_add_ordered_ring : 'a ordered_ab_group_add;
    ordered_cancel_semiring_ordered_ring : 'a ordered_cancel_semiring;
    ring_ordered_ring : 'a ring};;

let ordered_ab_semigroup_add_int =
  ({ab_semigroup_add_ordered_ab_semigroup_add = ab_semigroup_add_int;
     order_ordered_ab_semigroup_add = order_int}
    : int ordered_ab_semigroup_add);;

let strict_ordered_ab_semigroup_add_int =
  ({ordered_ab_semigroup_add_strict_ordered_ab_semigroup_add =
      ordered_ab_semigroup_add_int}
    : int strict_ordered_ab_semigroup_add);;

let ordered_cancel_ab_semigroup_add_int =
  ({cancel_ab_semigroup_add_ordered_cancel_ab_semigroup_add =
      cancel_ab_semigroup_add_int;
     strict_ordered_ab_semigroup_add_ordered_cancel_ab_semigroup_add =
       strict_ordered_ab_semigroup_add_int}
    : int ordered_cancel_ab_semigroup_add);;

let ordered_semigroup_mult_zero_int =
  ({semigroup_mult_ordered_semigroup_mult_zero = semigroup_mult_int;
     order_ordered_semigroup_mult_zero = order_int;
     mult_zero_ordered_semigroup_mult_zero = mult_zero_int}
    : int ordered_semigroup_mult_zero);;

let ordered_comm_monoid_add_int =
  ({comm_monoid_add_ordered_comm_monoid_add = comm_monoid_add_int;
     ordered_ab_semigroup_add_ordered_comm_monoid_add =
       ordered_ab_semigroup_add_int}
    : int ordered_comm_monoid_add);;

let ordered_semiring_int =
  ({ordered_comm_monoid_add_ordered_semiring = ordered_comm_monoid_add_int;
     semiring_ordered_semiring = semiring_int}
    : int ordered_semiring);;

let ordered_semiring_0_int =
  ({ordered_semigroup_mult_zero_ordered_semiring_0 =
      ordered_semigroup_mult_zero_int;
     ordered_semiring_ordered_semiring_0 = ordered_semiring_int;
     semiring_0_ordered_semiring_0 = semiring_0_int}
    : int ordered_semiring_0);;

let ordered_cancel_semiring_int =
  ({ordered_cancel_ab_semigroup_add_ordered_cancel_semiring =
      ordered_cancel_ab_semigroup_add_int;
     ordered_semiring_0_ordered_cancel_semiring = ordered_semiring_0_int;
     semiring_0_cancel_ordered_cancel_semiring = semiring_0_cancel_int}
    : int ordered_cancel_semiring);;

let ordered_ab_semigroup_add_imp_le_int =
  ({ordered_cancel_ab_semigroup_add_ordered_ab_semigroup_add_imp_le =
      ordered_cancel_ab_semigroup_add_int}
    : int ordered_ab_semigroup_add_imp_le);;

let strict_ordered_comm_monoid_add_int =
  ({comm_monoid_add_strict_ordered_comm_monoid_add = comm_monoid_add_int;
     strict_ordered_ab_semigroup_add_strict_ordered_comm_monoid_add =
       strict_ordered_ab_semigroup_add_int}
    : int strict_ordered_comm_monoid_add);;

let ordered_cancel_comm_monoid_add_int =
  ({ordered_cancel_ab_semigroup_add_ordered_cancel_comm_monoid_add =
      ordered_cancel_ab_semigroup_add_int;
     ordered_comm_monoid_add_ordered_cancel_comm_monoid_add =
       ordered_comm_monoid_add_int;
     strict_ordered_comm_monoid_add_ordered_cancel_comm_monoid_add =
       strict_ordered_comm_monoid_add_int}
    : int ordered_cancel_comm_monoid_add);;

let ordered_ab_semigroup_monoid_add_imp_le_int =
  ({cancel_comm_monoid_add_ordered_ab_semigroup_monoid_add_imp_le =
      cancel_comm_monoid_add_int;
     ordered_ab_semigroup_add_imp_le_ordered_ab_semigroup_monoid_add_imp_le =
       ordered_ab_semigroup_add_imp_le_int;
     ordered_cancel_comm_monoid_add_ordered_ab_semigroup_monoid_add_imp_le =
       ordered_cancel_comm_monoid_add_int}
    : int ordered_ab_semigroup_monoid_add_imp_le);;

let ordered_ab_group_add_int =
  ({ab_group_add_ordered_ab_group_add = ab_group_add_int;
     ordered_ab_semigroup_monoid_add_imp_le_ordered_ab_group_add =
       ordered_ab_semigroup_monoid_add_imp_le_int}
    : int ordered_ab_group_add);;

let ordered_ring_int =
  ({ordered_ab_group_add_ordered_ring = ordered_ab_group_add_int;
     ordered_cancel_semiring_ordered_ring = ordered_cancel_semiring_int;
     ring_ordered_ring = ring_int}
    : int ordered_ring);;

type 'a zero_less_one =
  {order_zero_less_one : 'a order;
    zero_neq_one_zero_less_one : 'a zero_neq_one};;

let zero_less_one_int =
  ({order_zero_less_one = order_int;
     zero_neq_one_zero_less_one = zero_neq_one_int}
    : int zero_less_one);;

let rec sgn_integer
  k = (if Big_int.eq_big_int k Big_int.zero_big_int then Big_int.zero_big_int
        else (if Big_int.lt_big_int k Big_int.zero_big_int
               then (Big_int.minus_big_int (Big_int.big_int_of_int 1))
               else (Big_int.big_int_of_int 1)));;

let rec apsnd f (x, y) = (x, f y);;

let rec comp f g = (fun x -> f (g x));;

let rec divmod_integer
  k l = (if Big_int.eq_big_int k Big_int.zero_big_int
          then (Big_int.zero_big_int, Big_int.zero_big_int)
          else (if Big_int.eq_big_int l Big_int.zero_big_int
                 then (Big_int.zero_big_int, k)
                 else comp (comp apsnd Big_int.mult_big_int) sgn_integer l
                        (if Big_int.eq_big_int (sgn_integer k) (sgn_integer l)
                          then Big_int.quomod_big_int (Big_int.abs_big_int k)
                                 (Big_int.abs_big_int l)
                          else (let (r, s) =
                                  Big_int.quomod_big_int (Big_int.abs_big_int k)
                                    (Big_int.abs_big_int l)
                                  in
                                 (if Big_int.eq_big_int s Big_int.zero_big_int
                                   then (Big_int.minus_big_int r,
  Big_int.zero_big_int)
                                   else (Big_int.sub_big_int
   (Big_int.minus_big_int r) (Big_int.big_int_of_int 1),
  Big_int.sub_big_int (Big_int.abs_big_int l) s))))));;

let rec snd (x1, x2) = x2;;

let rec modulo_integer k l = snd (divmod_integer k l);;

type nat = Nat of Big_int.big_int;;

let rec integer_of_nat (Nat x) = x;;

let rec modulo_nata
  m n = Nat (modulo_integer (integer_of_nat m) (integer_of_nat n));;

let rec fst (x1, x2) = x1;;

let rec divide_integer k l = fst (divmod_integer k l);;

let rec divide_nata
  m n = Nat (divide_integer (integer_of_nat m) (integer_of_nat n));;

let rec max _A a b = (if less_eq _A a b then b else a);;

let ord_integer =
  ({less_eq = Big_int.le_big_int; less = Big_int.lt_big_int} :
    Big_int.big_int ord);;

let rec nat_of_integer k = Nat (max ord_integer Big_int.zero_big_int k);;

let rec bit_cut_integer
  k = (if Big_int.eq_big_int k Big_int.zero_big_int
        then (Big_int.zero_big_int, false)
        else (let (r, s) =
                Big_int.quomod_big_int (Big_int.abs_big_int k)
                  (Big_int.abs_big_int (Big_int.big_int_of_int 2))
                in
               ((if Big_int.lt_big_int Big_int.zero_big_int k then r
                  else Big_int.sub_big_int (Big_int.minus_big_int r) s),
                 Big_int.eq_big_int s (Big_int.big_int_of_int 1))));;

type char = Chara of bool * bool * bool * bool * bool * bool * bool * bool;;

let rec char_of_integer
  k = (let (q0, b0) = bit_cut_integer k in
       let (q1, b1) = bit_cut_integer q0 in
       let (q2, b2) = bit_cut_integer q1 in
       let (q3, b3) = bit_cut_integer q2 in
       let (q4, b4) = bit_cut_integer q3 in
       let (q5, b5) = bit_cut_integer q4 in
       let (q6, b6) = bit_cut_integer q5 in
       let a = bit_cut_integer q6 in
       let (_, aa) = a in
        Chara (b0, b1, b2, b3, b4, b5, b6, aa));;

let rec of_bool _A = function true -> onea _A.one_zero_neq_one
                     | false -> zerob _A.zero_zero_neq_one;;

let one_integera : Big_int.big_int = (Big_int.big_int_of_int 1);;

let zero_integer = ({zerob = Big_int.zero_big_int} : Big_int.big_int zero);;

let one_integer = ({onea = one_integera} : Big_int.big_int one);;

let zero_neq_one_integer =
  ({one_zero_neq_one = one_integer; zero_zero_neq_one = zero_integer} :
    Big_int.big_int zero_neq_one);;

let rec integer_of_char
  (Chara (b0, b1, b2, b3, b4, b5, b6, b7)) =
    Big_int.add_big_int
      (Big_int.mult_big_int
        (Big_int.add_big_int
          (Big_int.mult_big_int
            (Big_int.add_big_int
              (Big_int.mult_big_int
                (Big_int.add_big_int
                  (Big_int.mult_big_int
                    (Big_int.add_big_int
                      (Big_int.mult_big_int
                        (Big_int.add_big_int
                          (Big_int.mult_big_int
                            (Big_int.add_big_int
                              (Big_int.mult_big_int
                                (of_bool zero_neq_one_integer b7)
                                (Big_int.big_int_of_int 2))
                              (of_bool zero_neq_one_integer b6))
                            (Big_int.big_int_of_int 2))
                          (of_bool zero_neq_one_integer b5))
                        (Big_int.big_int_of_int 2))
                      (of_bool zero_neq_one_integer b4))
                    (Big_int.big_int_of_int 2))
                  (of_bool zero_neq_one_integer b3))
                (Big_int.big_int_of_int 2))
              (of_bool zero_neq_one_integer b2))
            (Big_int.big_int_of_int 2))
          (of_bool zero_neq_one_integer b1))
        (Big_int.big_int_of_int 2))
      (of_bool zero_neq_one_integer b0);;

let rec map f x1 = match f, x1 with f, [] -> []
              | f, x21 :: x22 -> f x21 :: map f x22;;

let rec implode
  cs = (let ks = (map integer_of_char
                   cs) in let res = Bytes.create (List.length ks) in let rec imp i = function | [] -> res | k :: ks ->
      let l = Big_int.int_of_big_int k in if 0 <= l && l < 128 then Bytes.set res i (Char.chr l) else failwith "Non-ASCII character in literal"; imp (i + 1) ks in imp 0 ks);;

let rec lit_of_digit
  n = implode
        [char_of_integer
           (Big_int.add_big_int (Big_int.big_int_of_int 48)
             (integer_of_nat n))];;

let rec less_nat
  m n = Big_int.lt_big_int (integer_of_nat m) (integer_of_nat n);;

let rec showsl_lit x = (fun a -> x ^ a);;

let rec showsl_nat
  n = (if less_nat n (nat_of_integer (Big_int.big_int_of_int 10))
        then showsl_lit (lit_of_digit n)
        else comp (showsl_nat
                    (divide_nata n
                      (nat_of_integer (Big_int.big_int_of_int 10))))
               (showsl_lit
                 (lit_of_digit
                   (modulo_nata n
                     (nat_of_integer (Big_int.big_int_of_int 10))))));;

let rec nat k = Nat (max ord_integer Big_int.zero_big_int (integer_of_int k));;

let rec showsl_int
  i = (if less_int i zero_inta
        then comp (showsl_lit "-") (showsl_nat (nat (uminus_inta i)))
        else showsl_nat (nat i));;

let rec showsl_sep
  s sep x2 = match s, sep, x2 with s, sep, [] -> showsl_lit ""
    | s, sep, [x] -> s x
    | s, sep, x :: v :: va ->
        comp (comp (s x) sep) (showsl_sep s sep (v :: va));;

let rec null = function [] -> true
               | x :: xs -> false;;

let rec showsl_list_gen
  showslx e l s r xs =
    (if null xs then showsl_lit e
      else comp (comp (showsl_lit l) (showsl_sep showslx (showsl_lit s) xs))
             (showsl_lit r));;

let rec default_showsl_list sl = showsl_list_gen sl "[]" "[" ", " "]";;

let rec showsl_list_int xs = default_showsl_list showsl_int xs;;

type 'a showl =
  {showsl : 'a -> string -> string; showsl_list : 'a list -> string -> string};;
let showsl _A = _A.showsl;;
let showsl_list _A = _A.showsl_list;;

let showl_int =
  ({showsl = showsl_int; showsl_list = showsl_list_int} : int showl);;

let cEnum_int :
  (int list * (((int -> bool) -> bool) * ((int -> bool) -> bool))) option
  = None;;

type 'a cenum =
  {cEnum :
     ('a list * ((('a -> bool) -> bool) * (('a -> bool) -> bool))) option};;
let cEnum _A = _A.cEnum;;

let cenum_int = ({cEnum = cEnum_int} : int cenum);;

type 'a linordered_ab_semigroup_add =
  {ordered_ab_semigroup_add_linordered_ab_semigroup_add :
     'a ordered_ab_semigroup_add;
    linorder_linordered_ab_semigroup_add : 'a linorder};;

type 'a linordered_cancel_ab_semigroup_add =
  {linordered_ab_semigroup_add_linordered_cancel_ab_semigroup_add :
     'a linordered_ab_semigroup_add;
    ordered_ab_semigroup_add_imp_le_linordered_cancel_ab_semigroup_add :
      'a ordered_ab_semigroup_add_imp_le};;

type 'a linordered_semiring =
  {linordered_cancel_ab_semigroup_add_linordered_semiring :
     'a linordered_cancel_ab_semigroup_add;
    ordered_ab_semigroup_monoid_add_imp_le_linordered_semiring :
      'a ordered_ab_semigroup_monoid_add_imp_le;
    ordered_cancel_semiring_linordered_semiring : 'a ordered_cancel_semiring};;

type 'a linordered_semiring_strict =
  {linordered_semiring_linordered_semiring_strict : 'a linordered_semiring};;

type 'a linordered_semiring_1 =
  {linordered_semiring_linordered_semiring_1 : 'a linordered_semiring;
    semiring_1_linordered_semiring_1 : 'a semiring_1;
    zero_less_one_linordered_semiring_1 : 'a zero_less_one};;

type 'a linordered_semiring_1_strict =
  {linordered_semiring_1_linordered_semiring_1_strict :
     'a linordered_semiring_1;
    linordered_semiring_strict_linordered_semiring_1_strict :
      'a linordered_semiring_strict};;

type 'a ordered_ab_group_add_abs =
  {abs_ordered_ab_group_add_abs : 'a abs;
    ordered_ab_group_add_ordered_ab_group_add_abs : 'a ordered_ab_group_add};;

type 'a linordered_ab_group_add =
  {linordered_cancel_ab_semigroup_add_linordered_ab_group_add :
     'a linordered_cancel_ab_semigroup_add;
    ordered_ab_group_add_linordered_ab_group_add : 'a ordered_ab_group_add};;

type 'a linordered_ring =
  {linordered_ab_group_add_linordered_ring : 'a linordered_ab_group_add;
    ordered_ab_group_add_abs_linordered_ring : 'a ordered_ab_group_add_abs;
    abs_if_linordered_ring : 'a abs_if;
    linordered_semiring_linordered_ring : 'a linordered_semiring;
    ordered_ring_linordered_ring : 'a ordered_ring};;

type 'a linordered_ring_strict =
  {linordered_ring_linordered_ring_strict : 'a linordered_ring;
    linordered_semiring_strict_linordered_ring_strict :
      'a linordered_semiring_strict;
    ring_no_zero_divisors_linordered_ring_strict : 'a ring_no_zero_divisors};;

type 'a ordered_comm_semiring =
  {comm_semiring_0_ordered_comm_semiring : 'a comm_semiring_0;
    ordered_semiring_ordered_comm_semiring : 'a ordered_semiring};;

type 'a ordered_cancel_comm_semiring =
  {comm_semiring_0_cancel_ordered_cancel_comm_semiring :
     'a comm_semiring_0_cancel;
    ordered_cancel_semiring_ordered_cancel_comm_semiring :
      'a ordered_cancel_semiring;
    ordered_comm_semiring_ordered_cancel_comm_semiring :
      'a ordered_comm_semiring};;

type 'a linordered_comm_semiring_strict =
  {linordered_semiring_strict_linordered_comm_semiring_strict :
     'a linordered_semiring_strict;
    ordered_cancel_comm_semiring_linordered_comm_semiring_strict :
      'a ordered_cancel_comm_semiring};;

type 'a linordered_nonzero_semiring =
  {semiring_char_0_linordered_nonzero_semiring : 'a semiring_char_0;
    linorder_linordered_nonzero_semiring : 'a linorder;
    comm_semiring_1_linordered_nonzero_semiring : 'a comm_semiring_1;
    ordered_comm_semiring_linordered_nonzero_semiring :
      'a ordered_comm_semiring;
    zero_less_one_linordered_nonzero_semiring : 'a zero_less_one};;

type 'a linordered_semidom =
  {linordered_comm_semiring_strict_linordered_semidom :
     'a linordered_comm_semiring_strict;
    linordered_nonzero_semiring_linordered_semidom :
      'a linordered_nonzero_semiring;
    semidom_linordered_semidom : 'a semidom};;

type 'a ordered_comm_ring =
  {comm_ring_ordered_comm_ring : 'a comm_ring;
    ordered_cancel_comm_semiring_ordered_comm_ring :
      'a ordered_cancel_comm_semiring;
    ordered_ring_ordered_comm_ring : 'a ordered_ring};;

type 'a ordered_ring_abs =
  {ordered_ab_group_add_abs_ordered_ring_abs : 'a ordered_ab_group_add_abs;
    ordered_ring_ordered_ring_abs : 'a ordered_ring};;

type 'a linordered_idom =
  {ring_char_0_linordered_idom : 'a ring_char_0;
    idom_abs_sgn_linordered_idom : 'a idom_abs_sgn;
    linordered_ring_strict_linordered_idom : 'a linordered_ring_strict;
    linordered_semidom_linordered_idom : 'a linordered_semidom;
    linordered_semiring_1_strict_linordered_idom :
      'a linordered_semiring_1_strict;
    ordered_comm_ring_linordered_idom : 'a ordered_comm_ring;
    ordered_ring_abs_linordered_idom : 'a ordered_ring_abs};;

let linordered_ab_semigroup_add_int =
  ({ordered_ab_semigroup_add_linordered_ab_semigroup_add =
      ordered_ab_semigroup_add_int;
     linorder_linordered_ab_semigroup_add = linorder_int}
    : int linordered_ab_semigroup_add);;

let linordered_cancel_ab_semigroup_add_int =
  ({linordered_ab_semigroup_add_linordered_cancel_ab_semigroup_add =
      linordered_ab_semigroup_add_int;
     ordered_ab_semigroup_add_imp_le_linordered_cancel_ab_semigroup_add =
       ordered_ab_semigroup_add_imp_le_int}
    : int linordered_cancel_ab_semigroup_add);;

let linordered_semiring_int =
  ({linordered_cancel_ab_semigroup_add_linordered_semiring =
      linordered_cancel_ab_semigroup_add_int;
     ordered_ab_semigroup_monoid_add_imp_le_linordered_semiring =
       ordered_ab_semigroup_monoid_add_imp_le_int;
     ordered_cancel_semiring_linordered_semiring = ordered_cancel_semiring_int}
    : int linordered_semiring);;

let linordered_semiring_strict_int =
  ({linordered_semiring_linordered_semiring_strict = linordered_semiring_int} :
    int linordered_semiring_strict);;

let linordered_semiring_1_int =
  ({linordered_semiring_linordered_semiring_1 = linordered_semiring_int;
     semiring_1_linordered_semiring_1 = semiring_1_int;
     zero_less_one_linordered_semiring_1 = zero_less_one_int}
    : int linordered_semiring_1);;

let linordered_semiring_1_strict_int =
  ({linordered_semiring_1_linordered_semiring_1_strict =
      linordered_semiring_1_int;
     linordered_semiring_strict_linordered_semiring_1_strict =
       linordered_semiring_strict_int}
    : int linordered_semiring_1_strict);;

let ordered_ab_group_add_abs_int =
  ({abs_ordered_ab_group_add_abs = abs_int;
     ordered_ab_group_add_ordered_ab_group_add_abs = ordered_ab_group_add_int}
    : int ordered_ab_group_add_abs);;

let linordered_ab_group_add_int =
  ({linordered_cancel_ab_semigroup_add_linordered_ab_group_add =
      linordered_cancel_ab_semigroup_add_int;
     ordered_ab_group_add_linordered_ab_group_add = ordered_ab_group_add_int}
    : int linordered_ab_group_add);;

let linordered_ring_int =
  ({linordered_ab_group_add_linordered_ring = linordered_ab_group_add_int;
     ordered_ab_group_add_abs_linordered_ring = ordered_ab_group_add_abs_int;
     abs_if_linordered_ring = abs_if_int;
     linordered_semiring_linordered_ring = linordered_semiring_int;
     ordered_ring_linordered_ring = ordered_ring_int}
    : int linordered_ring);;

let linordered_ring_strict_int =
  ({linordered_ring_linordered_ring_strict = linordered_ring_int;
     linordered_semiring_strict_linordered_ring_strict =
       linordered_semiring_strict_int;
     ring_no_zero_divisors_linordered_ring_strict = ring_no_zero_divisors_int}
    : int linordered_ring_strict);;

let ordered_comm_semiring_int =
  ({comm_semiring_0_ordered_comm_semiring = comm_semiring_0_int;
     ordered_semiring_ordered_comm_semiring = ordered_semiring_int}
    : int ordered_comm_semiring);;

let ordered_cancel_comm_semiring_int =
  ({comm_semiring_0_cancel_ordered_cancel_comm_semiring =
      comm_semiring_0_cancel_int;
     ordered_cancel_semiring_ordered_cancel_comm_semiring =
       ordered_cancel_semiring_int;
     ordered_comm_semiring_ordered_cancel_comm_semiring =
       ordered_comm_semiring_int}
    : int ordered_cancel_comm_semiring);;

let linordered_comm_semiring_strict_int =
  ({linordered_semiring_strict_linordered_comm_semiring_strict =
      linordered_semiring_strict_int;
     ordered_cancel_comm_semiring_linordered_comm_semiring_strict =
       ordered_cancel_comm_semiring_int}
    : int linordered_comm_semiring_strict);;

let linordered_nonzero_semiring_int =
  ({semiring_char_0_linordered_nonzero_semiring = semiring_char_0_int;
     linorder_linordered_nonzero_semiring = linorder_int;
     comm_semiring_1_linordered_nonzero_semiring = comm_semiring_1_int;
     ordered_comm_semiring_linordered_nonzero_semiring =
       ordered_comm_semiring_int;
     zero_less_one_linordered_nonzero_semiring = zero_less_one_int}
    : int linordered_nonzero_semiring);;

let linordered_semidom_int =
  ({linordered_comm_semiring_strict_linordered_semidom =
      linordered_comm_semiring_strict_int;
     linordered_nonzero_semiring_linordered_semidom =
       linordered_nonzero_semiring_int;
     semidom_linordered_semidom = semidom_int}
    : int linordered_semidom);;

let ordered_comm_ring_int =
  ({comm_ring_ordered_comm_ring = comm_ring_int;
     ordered_cancel_comm_semiring_ordered_comm_ring =
       ordered_cancel_comm_semiring_int;
     ordered_ring_ordered_comm_ring = ordered_ring_int}
    : int ordered_comm_ring);;

let ordered_ring_abs_int =
  ({ordered_ab_group_add_abs_ordered_ring_abs = ordered_ab_group_add_abs_int;
     ordered_ring_ordered_ring_abs = ordered_ring_int}
    : int ordered_ring_abs);;

let linordered_idom_int =
  ({ring_char_0_linordered_idom = ring_char_0_int;
     idom_abs_sgn_linordered_idom = idom_abs_sgn_int;
     linordered_ring_strict_linordered_idom = linordered_ring_strict_int;
     linordered_semidom_linordered_idom = linordered_semidom_int;
     linordered_semiring_1_strict_linordered_idom =
       linordered_semiring_1_strict_int;
     ordered_comm_ring_linordered_idom = ordered_comm_ring_int;
     ordered_ring_abs_linordered_idom = ordered_ring_abs_int}
    : int linordered_idom);;

type 'a non_strict_order = {ord_non_strict_order : 'a ord};;

type 'a ordered_ab_semigroup =
  {ab_semigroup_add_ordered_ab_semigroup : 'a ab_semigroup_add;
    monoid_add_ordered_ab_semigroup : 'a monoid_add;
    non_strict_order_ordered_ab_semigroup : 'a non_strict_order};;

type 'a ordered_semiring_0a =
  {semiring_0_ordered_semiring_0a : 'a semiring_0;
    ordered_ab_semigroup_ordered_semiring_0 : 'a ordered_ab_semigroup};;

type 'a ordered_semiring_1 =
  {semiring_1_ordered_semiring_1 : 'a semiring_1;
    ordered_semiring_0_ordered_semiring_1 : 'a ordered_semiring_0a};;

type 'a poly_carrier =
  {comm_semiring_1_poly_carrier : 'a comm_semiring_1;
    ordered_semiring_1_poly_carrier : 'a ordered_semiring_1};;

let non_strict_order_int =
  ({ord_non_strict_order = ord_int} : int non_strict_order);;

let ordered_ab_semigroup_int =
  ({ab_semigroup_add_ordered_ab_semigroup = ab_semigroup_add_int;
     monoid_add_ordered_ab_semigroup = monoid_add_int;
     non_strict_order_ordered_ab_semigroup = non_strict_order_int}
    : int ordered_ab_semigroup);;

let ordered_semiring_0_inta =
  ({semiring_0_ordered_semiring_0a = semiring_0_int;
     ordered_ab_semigroup_ordered_semiring_0 = ordered_ab_semigroup_int}
    : int ordered_semiring_0a);;

let ordered_semiring_1_int =
  ({semiring_1_ordered_semiring_1 = semiring_1_int;
     ordered_semiring_0_ordered_semiring_1 = ordered_semiring_0_inta}
    : int ordered_semiring_1);;

let poly_carrier_int =
  ({comm_semiring_1_poly_carrier = comm_semiring_1_int;
     ordered_semiring_1_poly_carrier = ordered_semiring_1_int}
    : int poly_carrier);;

type ordera = Eq | Lt | Gt;;

let rec eq _A a b = equal _A a b;;

let rec comparator_of (_A1, _A2)
  x y = (if less _A2.order_linorder.preorder_order.ord_preorder x y then Lt
          else (if eq _A1 x y then Eq else Gt));;

let rec compare_int x = comparator_of (equal_int, linorder_int) x;;

let ccompare_inta : (int -> int -> ordera) option = Some compare_int;;

type 'a ccompare = {ccompare : ('a -> 'a -> ordera) option};;
let ccompare _A = _A.ccompare;;

let ccompare_int = ({ccompare = ccompare_inta} : int ccompare);;

type rat = Frct of (int * int);;

type mini_alg = Abs_mini_alg of (rat * (rat * nat));;

type mini_alg_unique = Abs_mini_alg_unique of mini_alg;;

let zero_rata : rat = Frct (zero_inta, one_inta);;

let zero_nata : nat = Nat Big_int.zero_big_int;;

let rec ma_of_rat xa = Abs_mini_alg (xa, (zero_rata, zero_nata));;

let rec mau_of_rat xa = Abs_mini_alg_unique (ma_of_rat xa);;

type real = Real_of_u of mini_alg_unique;;

let rec ratreal x = comp (fun a -> Real_of_u a) mau_of_rat x;;

let rec of_int a = Frct (a, one_inta);;

let rec real_of_int x = ratreal (of_int x);;

type 'a real_embedding =
  {linordered_idom_real_embedding : 'a linordered_idom; real_of : 'a -> real};;
let real_of _A = _A.real_of;;

let real_embedding_int =
  ({linordered_idom_real_embedding = linordered_idom_int; real_of = real_of_int}
    : int real_embedding);;

type 'a large_ordered_semiring_1 =
  {poly_carrier_large_ordered_semiring_1 : 'a poly_carrier};;

let large_ordered_semiring_1_int =
  ({poly_carrier_large_ordered_semiring_1 = poly_carrier_int} :
    int large_ordered_semiring_1);;

type 'a large_real_ordered_semiring_1 =
  {real_embedding_large_real_ordered_semiring_1 : 'a real_embedding;
    large_ordered_semiring_1_large_real_ordered_semiring_1 :
      'a large_ordered_semiring_1};;

let large_real_ordered_semiring_1_int =
  ({real_embedding_large_real_ordered_semiring_1 = real_embedding_int;
     large_ordered_semiring_1_large_real_ordered_semiring_1 =
       large_ordered_semiring_1_int}
    : int large_real_ordered_semiring_1);;

let rec equal_nata
  m n = Big_int.eq_big_int (integer_of_nat m) (integer_of_nat n);;

let equal_nat = ({equal = equal_nata} : nat equal);;

let rec times_nata
  m n = Nat (Big_int.mult_big_int (integer_of_nat m) (integer_of_nat n));;

let times_nat = ({times = times_nata} : nat times);;

let dvd_nat = ({times_dvd = times_nat} : nat dvd);;

let one_nata : nat = Nat (Big_int.big_int_of_int 1);;

let one_nat = ({onea = one_nata} : nat one);;

let rec plus_nata
  m n = Nat (Big_int.add_big_int (integer_of_nat m) (integer_of_nat n));;

let plus_nat = ({plus = plus_nata} : nat plus);;

let zero_nat = ({zerob = zero_nata} : nat zero);;

let semigroup_add_nat = ({plus_semigroup_add = plus_nat} : nat semigroup_add);;

let numeral_nat =
  ({one_numeral = one_nat; semigroup_add_numeral = semigroup_add_nat} :
    nat numeral);;

let power_nat = ({one_power = one_nat; times_power = times_nat} : nat power);;

let rec minus_nata
  m n = Nat (max ord_integer Big_int.zero_big_int
              (Big_int.sub_big_int (integer_of_nat m) (integer_of_nat n)));;

let minus_nat = ({minus = minus_nata} : nat minus);;

let rec min _A a b = (if less_eq _A a b then a else b);;

let rec less_eq_nat
  m n = Big_int.le_big_int (integer_of_nat m) (integer_of_nat n);;

let ord_nat = ({less_eq = less_eq_nat; less = less_nat} : nat ord);;

let rec inf_nata x = min ord_nat x;;

type 'a inf = {inf : 'a -> 'a -> 'a};;
let inf _A = _A.inf;;

let inf_nat = ({inf = inf_nata} : nat inf);;

let rec sup_nata x = max ord_nat x;;

type 'a sup = {sup : 'a -> 'a -> 'a};;
let sup _A = _A.sup;;

let sup_nat = ({sup = sup_nata} : nat sup);;

type 'a divide = {divide : 'a -> 'a -> 'a};;
let divide _A = _A.divide;;

let divide_nat = ({divide = divide_nata} : nat divide);;

type 'a modulo =
  {divide_modulo : 'a divide; dvd_modulo : 'a dvd; modulo : 'a -> 'a -> 'a};;
let modulo _A = _A.modulo;;

let modulo_nat =
  ({divide_modulo = divide_nat; dvd_modulo = dvd_nat; modulo = modulo_nata} :
    nat modulo);;

let ab_semigroup_add_nat =
  ({semigroup_add_ab_semigroup_add = semigroup_add_nat} :
    nat ab_semigroup_add);;

let monoid_add_nat =
  ({semigroup_add_monoid_add = semigroup_add_nat; zero_monoid_add = zero_nat} :
    nat monoid_add);;

let comm_monoid_add_nat =
  ({ab_semigroup_add_comm_monoid_add = ab_semigroup_add_nat;
     monoid_add_comm_monoid_add = monoid_add_nat}
    : nat comm_monoid_add);;

let mult_zero_nat =
  ({times_mult_zero = times_nat; zero_mult_zero = zero_nat} : nat mult_zero);;

let semigroup_mult_nat =
  ({times_semigroup_mult = times_nat} : nat semigroup_mult);;

let semiring_nat =
  ({ab_semigroup_add_semiring = ab_semigroup_add_nat;
     semigroup_mult_semiring = semigroup_mult_nat}
    : nat semiring);;

let semiring_0_nat =
  ({comm_monoid_add_semiring_0 = comm_monoid_add_nat;
     mult_zero_semiring_0 = mult_zero_nat; semiring_semiring_0 = semiring_nat}
    : nat semiring_0);;

let semiring_no_zero_divisors_nat =
  ({semiring_0_semiring_no_zero_divisors = semiring_0_nat} :
    nat semiring_no_zero_divisors);;

let monoid_mult_nat =
  ({semigroup_mult_monoid_mult = semigroup_mult_nat;
     power_monoid_mult = power_nat}
    : nat monoid_mult);;

let semiring_numeral_nat =
  ({monoid_mult_semiring_numeral = monoid_mult_nat;
     numeral_semiring_numeral = numeral_nat;
     semiring_semiring_numeral = semiring_nat}
    : nat semiring_numeral);;

let zero_neq_one_nat =
  ({one_zero_neq_one = one_nat; zero_zero_neq_one = zero_nat} :
    nat zero_neq_one);;

let semiring_1_nat =
  ({semiring_numeral_semiring_1 = semiring_numeral_nat;
     semiring_0_semiring_1 = semiring_0_nat;
     zero_neq_one_semiring_1 = zero_neq_one_nat}
    : nat semiring_1);;

let semiring_1_no_zero_divisors_nat =
  ({semiring_1_semiring_1_no_zero_divisors = semiring_1_nat;
     semiring_no_zero_divisors_semiring_1_no_zero_divisors =
       semiring_no_zero_divisors_nat}
    : nat semiring_1_no_zero_divisors);;

let cancel_semigroup_add_nat =
  ({semigroup_add_cancel_semigroup_add = semigroup_add_nat} :
    nat cancel_semigroup_add);;

let cancel_ab_semigroup_add_nat =
  ({ab_semigroup_add_cancel_ab_semigroup_add = ab_semigroup_add_nat;
     cancel_semigroup_add_cancel_ab_semigroup_add = cancel_semigroup_add_nat;
     minus_cancel_ab_semigroup_add = minus_nat}
    : nat cancel_ab_semigroup_add);;

let cancel_comm_monoid_add_nat =
  ({cancel_ab_semigroup_add_cancel_comm_monoid_add =
      cancel_ab_semigroup_add_nat;
     comm_monoid_add_cancel_comm_monoid_add = comm_monoid_add_nat}
    : nat cancel_comm_monoid_add);;

let semiring_0_cancel_nat =
  ({cancel_comm_monoid_add_semiring_0_cancel = cancel_comm_monoid_add_nat;
     semiring_0_semiring_0_cancel = semiring_0_nat}
    : nat semiring_0_cancel);;

let ab_semigroup_mult_nat =
  ({semigroup_mult_ab_semigroup_mult = semigroup_mult_nat} :
    nat ab_semigroup_mult);;

let comm_semiring_nat =
  ({ab_semigroup_mult_comm_semiring = ab_semigroup_mult_nat;
     semiring_comm_semiring = semiring_nat}
    : nat comm_semiring);;

let comm_semiring_0_nat =
  ({comm_semiring_comm_semiring_0 = comm_semiring_nat;
     semiring_0_comm_semiring_0 = semiring_0_nat}
    : nat comm_semiring_0);;

let comm_semiring_0_cancel_nat =
  ({comm_semiring_0_comm_semiring_0_cancel = comm_semiring_0_nat;
     semiring_0_cancel_comm_semiring_0_cancel = semiring_0_cancel_nat}
    : nat comm_semiring_0_cancel);;

let semiring_1_cancel_nat =
  ({semiring_0_cancel_semiring_1_cancel = semiring_0_cancel_nat;
     semiring_1_semiring_1_cancel = semiring_1_nat}
    : nat semiring_1_cancel);;

let comm_monoid_mult_nat =
  ({ab_semigroup_mult_comm_monoid_mult = ab_semigroup_mult_nat;
     monoid_mult_comm_monoid_mult = monoid_mult_nat;
     dvd_comm_monoid_mult = dvd_nat}
    : nat comm_monoid_mult);;

let comm_semiring_1_nat =
  ({comm_monoid_mult_comm_semiring_1 = comm_monoid_mult_nat;
     comm_semiring_0_comm_semiring_1 = comm_semiring_0_nat;
     semiring_1_comm_semiring_1 = semiring_1_nat}
    : nat comm_semiring_1);;

let comm_semiring_1_cancel_nat =
  ({comm_semiring_0_cancel_comm_semiring_1_cancel = comm_semiring_0_cancel_nat;
     comm_semiring_1_comm_semiring_1_cancel = comm_semiring_1_nat;
     semiring_1_cancel_comm_semiring_1_cancel = semiring_1_cancel_nat}
    : nat comm_semiring_1_cancel);;

let semidom_nat =
  ({comm_semiring_1_cancel_semidom = comm_semiring_1_cancel_nat;
     semiring_1_no_zero_divisors_semidom = semiring_1_no_zero_divisors_nat}
    : nat semidom);;

let quasi_order_nat = ({ord_quasi_order = ord_nat} : nat quasi_order);;

let weak_order_nat =
  ({quasi_order_weak_order = quasi_order_nat} : nat weak_order);;

let preorder_nat = ({ord_preorder = ord_nat} : nat preorder);;

let order_nat =
  ({preorder_order = preorder_nat; weak_order_order = weak_order_nat} :
    nat order);;

let linorder_nat = ({order_linorder = order_nat} : nat linorder);;

let rec compare_nata x = comparator_of (equal_nat, linorder_nat) x;;

type 'a compare = {compare : 'a -> 'a -> ordera};;
let compare _A = _A.compare;;

let compare_nat = ({compare = compare_nata} : nat compare);;

type 'a quasi_order_sup =
  {sup_quasi_order_sup : 'a sup; quasi_order_quasi_order_sup : 'a quasi_order};;

type 'a quasi_semilattice_sup =
  {quasi_order_sup_quasi_semilattice_sup : 'a quasi_order_sup};;

type 'a semilattice_sup =
  {order_semilattice_sup : 'a order;
    quasi_semilattice_sup_semilattice_sup : 'a quasi_semilattice_sup};;

type 'a semilattice_inf =
  {inf_semilattice_inf : 'a inf; order_semilattice_inf : 'a order};;

type 'a lattice =
  {semilattice_inf_lattice : 'a semilattice_inf;
    semilattice_sup_lattice : 'a semilattice_sup};;

let quasi_order_sup_nat =
  ({sup_quasi_order_sup = sup_nat;
     quasi_order_quasi_order_sup = quasi_order_nat}
    : nat quasi_order_sup);;

let quasi_semilattice_sup_nat =
  ({quasi_order_sup_quasi_semilattice_sup = quasi_order_sup_nat} :
    nat quasi_semilattice_sup);;

let semilattice_sup_nat =
  ({order_semilattice_sup = order_nat;
     quasi_semilattice_sup_semilattice_sup = quasi_semilattice_sup_nat}
    : nat semilattice_sup);;

let semilattice_inf_nat =
  ({inf_semilattice_inf = inf_nat; order_semilattice_inf = order_nat} :
    nat semilattice_inf);;

let lattice_nat =
  ({semilattice_inf_lattice = semilattice_inf_nat;
     semilattice_sup_lattice = semilattice_sup_nat}
    : nat lattice);;

let ceq_nata : (nat -> nat -> bool) option = Some equal_nata;;

let ceq_nat = ({ceq = ceq_nata} : nat ceq);;

let set_impl_nata : (nat, set_impla) phantom = Phantom Set_RBT;;

let set_impl_nat = ({set_impl = set_impl_nata} : nat set_impl);;

let rec showsl_list_nat xs = default_showsl_list showsl_nat xs;;

let showl_nat =
  ({showsl = showsl_nat; showsl_list = showsl_list_nat} : nat showl);;

type 'a semidom_divide =
  {divide_semidom_divide : 'a divide; semidom_semidom_divide : 'a semidom;
    semiring_no_zero_divisors_cancel_semidom_divide :
      'a semiring_no_zero_divisors_cancel};;

let semiring_no_zero_divisors_cancel_nat =
  ({semiring_no_zero_divisors_semiring_no_zero_divisors_cancel =
      semiring_no_zero_divisors_nat}
    : nat semiring_no_zero_divisors_cancel);;

let semidom_divide_nat =
  ({divide_semidom_divide = divide_nat; semidom_semidom_divide = semidom_nat;
     semiring_no_zero_divisors_cancel_semidom_divide =
       semiring_no_zero_divisors_cancel_nat}
    : nat semidom_divide);;

type 'a algebraic_semidom =
  {semidom_divide_algebraic_semidom : 'a semidom_divide};;

type 'a semiring_modulo =
  {comm_semiring_1_cancel_semiring_modulo : 'a comm_semiring_1_cancel;
    modulo_semiring_modulo : 'a modulo};;

type 'a semidom_modulo =
  {algebraic_semidom_semidom_modulo : 'a algebraic_semidom;
    semiring_modulo_semidom_modulo : 'a semiring_modulo};;

let algebraic_semidom_nat =
  ({semidom_divide_algebraic_semidom = semidom_divide_nat} :
    nat algebraic_semidom);;

let semiring_modulo_nat =
  ({comm_semiring_1_cancel_semiring_modulo = comm_semiring_1_cancel_nat;
     modulo_semiring_modulo = modulo_nat}
    : nat semiring_modulo);;

let semidom_modulo_nat =
  ({algebraic_semidom_semidom_modulo = algebraic_semidom_nat;
     semiring_modulo_semidom_modulo = semiring_modulo_nat}
    : nat semidom_modulo);;

let finite_UNIV_nata : (nat, bool) phantom = Phantom false;;

let card_UNIV_nata : (nat, nat) phantom = Phantom zero_nata;;

type 'a finite_UNIV = {finite_UNIV : ('a, bool) phantom};;
let finite_UNIV _A = _A.finite_UNIV;;

type 'a card_UNIV =
  {finite_UNIV_card_UNIV : 'a finite_UNIV; card_UNIVa : ('a, nat) phantom};;
let card_UNIVa _A = _A.card_UNIVa;;

let finite_UNIV_nat = ({finite_UNIV = finite_UNIV_nata} : nat finite_UNIV);;

let card_UNIV_nat =
  ({finite_UNIV_card_UNIV = finite_UNIV_nat; card_UNIVa = card_UNIV_nata} :
    nat card_UNIV);;

let cEnum_nat :
  (nat list * (((nat -> bool) -> bool) * ((nat -> bool) -> bool))) option
  = None;;

let cenum_nat = ({cEnum = cEnum_nat} : nat cenum);;

type 'a compare_order =
  {compare_compare_order : 'a compare; linorder_compare_order : 'a linorder};;

let compare_order_nat =
  ({compare_compare_order = compare_nat; linorder_compare_order = linorder_nat}
    : nat compare_order);;

let ccompare_nata : (nat -> nat -> ordera) option = Some compare_nata;;

let ccompare_nat = ({ccompare = ccompare_nata} : nat ccompare);;

type mapping_impla = Mapping_Choose | Mapping_Assoc_List | Mapping_RBT |
  Mapping_Mapping;;

let mapping_impl_nata : (nat, mapping_impla) phantom = Phantom Mapping_RBT;;

type 'a mapping_impl = {mapping_impl : ('a, mapping_impla) phantom};;
let mapping_impl _A = _A.mapping_impl;;

let mapping_impl_nat = ({mapping_impl = mapping_impl_nata} : nat mapping_impl);;

let rec proper_interval_nat
  no x1 = match no, x1 with no, None -> true
    | None, Some x -> less_nat zero_nata x
    | Some x, Some y -> less_nat one_nata (minus_nata y x);;

let rec cproper_interval_nata x = proper_interval_nat x;;

type 'a cproper_interval =
  {ccompare_cproper_interval : 'a ccompare;
    cproper_interval : 'a option -> 'a option -> bool};;
let cproper_interval _A = _A.cproper_interval;;

let cproper_interval_nat =
  ({ccompare_cproper_interval = ccompare_nat;
     cproper_interval = cproper_interval_nata}
    : nat cproper_interval);;

let rec equal_proda _A _B (x1, x2) (y1, y2) = eq _A x1 y1 && eq _B x2 y2;;

let rec quotient_of (Frct x) = x;;

let rec equal_rata
  a b = equal_proda equal_int equal_int (quotient_of a) (quotient_of b);;

let equal_rat = ({equal = equal_rata} : rat equal);;

let rec divide_int
  k l = Int_of_integer (divide_integer (integer_of_int k) (integer_of_int l));;

let rec gcd_int
  (Int_of_integer x) (Int_of_integer y) =
    Int_of_integer (Big_int.gcd_big_int x y);;

let rec normalize
  p = (if less_int zero_inta (snd p)
        then (let a = gcd_int (fst p) (snd p) in
               (divide_int (fst p) a, divide_int (snd p) a))
        else (if equal_inta (snd p) zero_inta then (zero_inta, one_inta)
               else (let a = uminus_inta (gcd_int (fst p) (snd p)) in
                      (divide_int (fst p) a, divide_int (snd p) a))));;

let rec times_rata
  p q = Frct (let a = quotient_of p in
              let (aa, c) = a in
              let b = quotient_of q in
              let (ba, d) = b in
               normalize (times_inta aa ba, times_inta c d));;

let times_rat = ({times = times_rata} : rat times);;

let dvd_rat = ({times_dvd = times_rat} : rat dvd);;

let rec abs_rata p = Frct (let a = quotient_of p in
                           let (aa, b) = a in
                            (abs_inta aa, b));;

let abs_rat = ({abs = abs_rata} : rat abs);;

let one_rata : rat = Frct (one_inta, one_inta);;

let one_rat = ({onea = one_rata} : rat one);;

let rec sgn_rata p = Frct (sgn_inta (fst (quotient_of p)), one_inta);;

let sgn_rat = ({sgn = sgn_rata} : rat sgn);;

let rec uminus_rata p = Frct (let a = quotient_of p in
                              let (aa, b) = a in
                               (uminus_inta aa, b));;

let rec minus_rata
  p q = Frct (let a = quotient_of p in
              let (aa, c) = a in
              let b = quotient_of q in
              let (ba, d) = b in
               normalize
                 (minus_inta (times_inta aa d) (times_inta ba c),
                   times_inta c d));;

let rec plus_rata
  p q = Frct (let a = quotient_of p in
              let (aa, c) = a in
              let b = quotient_of q in
              let (ba, d) = b in
               normalize
                 (plus_inta (times_inta aa d) (times_inta ba c),
                   times_inta c d));;

let plus_rat = ({plus = plus_rata} : rat plus);;

let semigroup_add_rat = ({plus_semigroup_add = plus_rat} : rat semigroup_add);;

let cancel_semigroup_add_rat =
  ({semigroup_add_cancel_semigroup_add = semigroup_add_rat} :
    rat cancel_semigroup_add);;

let ab_semigroup_add_rat =
  ({semigroup_add_ab_semigroup_add = semigroup_add_rat} :
    rat ab_semigroup_add);;

let minus_rat = ({minus = minus_rata} : rat minus);;

let cancel_ab_semigroup_add_rat =
  ({ab_semigroup_add_cancel_ab_semigroup_add = ab_semigroup_add_rat;
     cancel_semigroup_add_cancel_ab_semigroup_add = cancel_semigroup_add_rat;
     minus_cancel_ab_semigroup_add = minus_rat}
    : rat cancel_ab_semigroup_add);;

let zero_rat = ({zerob = zero_rata} : rat zero);;

let monoid_add_rat =
  ({semigroup_add_monoid_add = semigroup_add_rat; zero_monoid_add = zero_rat} :
    rat monoid_add);;

let comm_monoid_add_rat =
  ({ab_semigroup_add_comm_monoid_add = ab_semigroup_add_rat;
     monoid_add_comm_monoid_add = monoid_add_rat}
    : rat comm_monoid_add);;

let cancel_comm_monoid_add_rat =
  ({cancel_ab_semigroup_add_cancel_comm_monoid_add =
      cancel_ab_semigroup_add_rat;
     comm_monoid_add_cancel_comm_monoid_add = comm_monoid_add_rat}
    : rat cancel_comm_monoid_add);;

let mult_zero_rat =
  ({times_mult_zero = times_rat; zero_mult_zero = zero_rat} : rat mult_zero);;

let semigroup_mult_rat =
  ({times_semigroup_mult = times_rat} : rat semigroup_mult);;

let semiring_rat =
  ({ab_semigroup_add_semiring = ab_semigroup_add_rat;
     semigroup_mult_semiring = semigroup_mult_rat}
    : rat semiring);;

let semiring_0_rat =
  ({comm_monoid_add_semiring_0 = comm_monoid_add_rat;
     mult_zero_semiring_0 = mult_zero_rat; semiring_semiring_0 = semiring_rat}
    : rat semiring_0);;

let semiring_0_cancel_rat =
  ({cancel_comm_monoid_add_semiring_0_cancel = cancel_comm_monoid_add_rat;
     semiring_0_semiring_0_cancel = semiring_0_rat}
    : rat semiring_0_cancel);;

let ab_semigroup_mult_rat =
  ({semigroup_mult_ab_semigroup_mult = semigroup_mult_rat} :
    rat ab_semigroup_mult);;

let comm_semiring_rat =
  ({ab_semigroup_mult_comm_semiring = ab_semigroup_mult_rat;
     semiring_comm_semiring = semiring_rat}
    : rat comm_semiring);;

let comm_semiring_0_rat =
  ({comm_semiring_comm_semiring_0 = comm_semiring_rat;
     semiring_0_comm_semiring_0 = semiring_0_rat}
    : rat comm_semiring_0);;

let comm_semiring_0_cancel_rat =
  ({comm_semiring_0_comm_semiring_0_cancel = comm_semiring_0_rat;
     semiring_0_cancel_comm_semiring_0_cancel = semiring_0_cancel_rat}
    : rat comm_semiring_0_cancel);;

let power_rat = ({one_power = one_rat; times_power = times_rat} : rat power);;

let monoid_mult_rat =
  ({semigroup_mult_monoid_mult = semigroup_mult_rat;
     power_monoid_mult = power_rat}
    : rat monoid_mult);;

let numeral_rat =
  ({one_numeral = one_rat; semigroup_add_numeral = semigroup_add_rat} :
    rat numeral);;

let semiring_numeral_rat =
  ({monoid_mult_semiring_numeral = monoid_mult_rat;
     numeral_semiring_numeral = numeral_rat;
     semiring_semiring_numeral = semiring_rat}
    : rat semiring_numeral);;

let zero_neq_one_rat =
  ({one_zero_neq_one = one_rat; zero_zero_neq_one = zero_rat} :
    rat zero_neq_one);;

let semiring_1_rat =
  ({semiring_numeral_semiring_1 = semiring_numeral_rat;
     semiring_0_semiring_1 = semiring_0_rat;
     zero_neq_one_semiring_1 = zero_neq_one_rat}
    : rat semiring_1);;

let semiring_1_cancel_rat =
  ({semiring_0_cancel_semiring_1_cancel = semiring_0_cancel_rat;
     semiring_1_semiring_1_cancel = semiring_1_rat}
    : rat semiring_1_cancel);;

let comm_monoid_mult_rat =
  ({ab_semigroup_mult_comm_monoid_mult = ab_semigroup_mult_rat;
     monoid_mult_comm_monoid_mult = monoid_mult_rat;
     dvd_comm_monoid_mult = dvd_rat}
    : rat comm_monoid_mult);;

let comm_semiring_1_rat =
  ({comm_monoid_mult_comm_semiring_1 = comm_monoid_mult_rat;
     comm_semiring_0_comm_semiring_1 = comm_semiring_0_rat;
     semiring_1_comm_semiring_1 = semiring_1_rat}
    : rat comm_semiring_1);;

let comm_semiring_1_cancel_rat =
  ({comm_semiring_0_cancel_comm_semiring_1_cancel = comm_semiring_0_cancel_rat;
     comm_semiring_1_comm_semiring_1_cancel = comm_semiring_1_rat;
     semiring_1_cancel_comm_semiring_1_cancel = semiring_1_cancel_rat}
    : rat comm_semiring_1_cancel);;

let comm_semiring_1_cancel_crossproduct_rat =
  ({comm_semiring_1_cancel_comm_semiring_1_cancel_crossproduct =
      comm_semiring_1_cancel_rat}
    : rat comm_semiring_1_cancel_crossproduct);;

let semiring_no_zero_divisors_rat =
  ({semiring_0_semiring_no_zero_divisors = semiring_0_rat} :
    rat semiring_no_zero_divisors);;

let semiring_1_no_zero_divisors_rat =
  ({semiring_1_semiring_1_no_zero_divisors = semiring_1_rat;
     semiring_no_zero_divisors_semiring_1_no_zero_divisors =
       semiring_no_zero_divisors_rat}
    : rat semiring_1_no_zero_divisors);;

let semiring_no_zero_divisors_cancel_rat =
  ({semiring_no_zero_divisors_semiring_no_zero_divisors_cancel =
      semiring_no_zero_divisors_rat}
    : rat semiring_no_zero_divisors_cancel);;

let uminus_rat = ({uminus = uminus_rata} : rat uminus);;

let group_add_rat =
  ({cancel_semigroup_add_group_add = cancel_semigroup_add_rat;
     minus_group_add = minus_rat; monoid_add_group_add = monoid_add_rat;
     uminus_group_add = uminus_rat}
    : rat group_add);;

let ab_group_add_rat =
  ({cancel_comm_monoid_add_ab_group_add = cancel_comm_monoid_add_rat;
     group_add_ab_group_add = group_add_rat}
    : rat ab_group_add);;

let ring_rat =
  ({ab_group_add_ring = ab_group_add_rat;
     semiring_0_cancel_ring = semiring_0_cancel_rat}
    : rat ring);;

let ring_no_zero_divisors_rat =
  ({ring_ring_no_zero_divisors = ring_rat;
     semiring_no_zero_divisors_cancel_ring_no_zero_divisors =
       semiring_no_zero_divisors_cancel_rat}
    : rat ring_no_zero_divisors);;

let neg_numeral_rat =
  ({group_add_neg_numeral = group_add_rat; numeral_neg_numeral = numeral_rat} :
    rat neg_numeral);;

let ring_1_rat =
  ({neg_numeral_ring_1 = neg_numeral_rat; ring_ring_1 = ring_rat;
     semiring_1_cancel_ring_1 = semiring_1_cancel_rat}
    : rat ring_1);;

let ring_1_no_zero_divisors_rat =
  ({ring_1_ring_1_no_zero_divisors = ring_1_rat;
     ring_no_zero_divisors_ring_1_no_zero_divisors = ring_no_zero_divisors_rat;
     semiring_1_no_zero_divisors_ring_1_no_zero_divisors =
       semiring_1_no_zero_divisors_rat}
    : rat ring_1_no_zero_divisors);;

let comm_ring_rat =
  ({comm_semiring_0_cancel_comm_ring = comm_semiring_0_cancel_rat;
     ring_comm_ring = ring_rat}
    : rat comm_ring);;

let comm_ring_1_rat =
  ({comm_ring_comm_ring_1 = comm_ring_rat;
     comm_semiring_1_cancel_comm_ring_1 = comm_semiring_1_cancel_rat;
     ring_1_comm_ring_1 = ring_1_rat}
    : rat comm_ring_1);;

let semidom_rat =
  ({comm_semiring_1_cancel_semidom = comm_semiring_1_cancel_rat;
     semiring_1_no_zero_divisors_semidom = semiring_1_no_zero_divisors_rat}
    : rat semidom);;

let idom_rat =
  ({comm_ring_1_idom = comm_ring_1_rat;
     ring_1_no_zero_divisors_idom = ring_1_no_zero_divisors_rat;
     semidom_idom = semidom_rat;
     comm_semiring_1_cancel_crossproduct_idom =
       comm_semiring_1_cancel_crossproduct_rat}
    : rat idom);;

let rec inverse_rata
  p = Frct (let a = quotient_of p in
            let (aa, b) = a in
             (if equal_inta aa zero_inta then (zero_inta, one_inta)
               else (times_inta (sgn_inta aa) b, abs_inta aa)));;

let rec divide_rata
  p q = Frct (let a = quotient_of p in
              let (aa, c) = a in
              let b = quotient_of q in
              let (ba, d) = b in
               normalize (times_inta aa d, times_inta c ba));;

type 'a inverse = {divide_inverse : 'a divide; inverse : 'a -> 'a};;
let inverse _A = _A.inverse;;

type 'a division_ring =
  {inverse_division_ring : 'a inverse;
    ring_1_no_zero_divisors_division_ring : 'a ring_1_no_zero_divisors};;

type 'a idom_divide =
  {idom_idom_divide : 'a idom; semidom_divide_idom_divide : 'a semidom_divide};;

type 'a field =
  {division_ring_field : 'a division_ring; idom_divide_field : 'a idom_divide};;

let divide_rat = ({divide = divide_rata} : rat divide);;

let inverse_rat =
  ({divide_inverse = divide_rat; inverse = inverse_rata} : rat inverse);;

let division_ring_rat =
  ({inverse_division_ring = inverse_rat;
     ring_1_no_zero_divisors_division_ring = ring_1_no_zero_divisors_rat}
    : rat division_ring);;

let semidom_divide_rat =
  ({divide_semidom_divide = divide_rat; semidom_semidom_divide = semidom_rat;
     semiring_no_zero_divisors_cancel_semidom_divide =
       semiring_no_zero_divisors_cancel_rat}
    : rat semidom_divide);;

let idom_divide_rat =
  ({idom_idom_divide = idom_rat;
     semidom_divide_idom_divide = semidom_divide_rat}
    : rat idom_divide);;

let field_rat =
  ({division_ring_field = division_ring_rat;
     idom_divide_field = idom_divide_rat}
    : rat field);;

let rec less_eq_rat
  p q = (let a = quotient_of p in
         let (aa, c) = a in
         let b = quotient_of q in
         let (ba, d) = b in
          less_eq_int (times_inta aa d) (times_inta c ba));;

let rec less_rat
  p q = (let a = quotient_of p in
         let (aa, c) = a in
         let b = quotient_of q in
         let (ba, d) = b in
          less_int (times_inta aa d) (times_inta c ba));;

let ord_rat = ({less_eq = less_eq_rat; less = less_rat} : rat ord);;

let abs_if_rat =
  ({abs_abs_if = abs_rat; minus_abs_if = minus_rat; uminus_abs_if = uminus_rat;
     zero_abs_if = zero_rat; ord_abs_if = ord_rat}
    : rat abs_if);;

let quasi_order_rat = ({ord_quasi_order = ord_rat} : rat quasi_order);;

let weak_order_rat =
  ({quasi_order_weak_order = quasi_order_rat} : rat weak_order);;

let preorder_rat = ({ord_preorder = ord_rat} : rat preorder);;

let order_rat =
  ({preorder_order = preorder_rat; weak_order_order = weak_order_rat} :
    rat order);;

let linorder_rat = ({order_linorder = order_rat} : rat linorder);;

let rec compare_rata x = comparator_of (equal_rat, linorder_rat) x;;

let compare_rat = ({compare = compare_rata} : rat compare);;

let semiring_char_0_rat =
  ({semiring_1_semiring_char_0 = semiring_1_rat} : rat semiring_char_0);;

let ring_char_0_rat =
  ({semiring_char_0_ring_char_0 = semiring_char_0_rat;
     ring_1_ring_char_0 = ring_1_rat}
    : rat ring_char_0);;

type 'a no_bot = {order_no_bot : 'a order};;

let no_bot_rat = ({order_no_bot = order_rat} : rat no_bot);;

type 'a no_top = {order_no_top : 'a order};;

let no_top_rat = ({order_no_top = order_rat} : rat no_top);;

let ceq_rata : (rat -> rat -> bool) option = Some equal_rata;;

let ceq_rat = ({ceq = ceq_rata} : rat ceq);;

let set_impl_rata : (rat, set_impla) phantom = Phantom Set_RBT;;

let set_impl_rat = ({set_impl = set_impl_rata} : rat set_impl);;

let idom_abs_sgn_rat =
  ({abs_idom_abs_sgn = abs_rat; sgn_idom_abs_sgn = sgn_rat;
     idom_idom_abs_sgn = idom_rat}
    : rat idom_abs_sgn);;

let ordered_ab_semigroup_add_rat =
  ({ab_semigroup_add_ordered_ab_semigroup_add = ab_semigroup_add_rat;
     order_ordered_ab_semigroup_add = order_rat}
    : rat ordered_ab_semigroup_add);;

let strict_ordered_ab_semigroup_add_rat =
  ({ordered_ab_semigroup_add_strict_ordered_ab_semigroup_add =
      ordered_ab_semigroup_add_rat}
    : rat strict_ordered_ab_semigroup_add);;

let ordered_cancel_ab_semigroup_add_rat =
  ({cancel_ab_semigroup_add_ordered_cancel_ab_semigroup_add =
      cancel_ab_semigroup_add_rat;
     strict_ordered_ab_semigroup_add_ordered_cancel_ab_semigroup_add =
       strict_ordered_ab_semigroup_add_rat}
    : rat ordered_cancel_ab_semigroup_add);;

let ordered_semigroup_mult_zero_rat =
  ({semigroup_mult_ordered_semigroup_mult_zero = semigroup_mult_rat;
     order_ordered_semigroup_mult_zero = order_rat;
     mult_zero_ordered_semigroup_mult_zero = mult_zero_rat}
    : rat ordered_semigroup_mult_zero);;

let ordered_comm_monoid_add_rat =
  ({comm_monoid_add_ordered_comm_monoid_add = comm_monoid_add_rat;
     ordered_ab_semigroup_add_ordered_comm_monoid_add =
       ordered_ab_semigroup_add_rat}
    : rat ordered_comm_monoid_add);;

let ordered_semiring_rat =
  ({ordered_comm_monoid_add_ordered_semiring = ordered_comm_monoid_add_rat;
     semiring_ordered_semiring = semiring_rat}
    : rat ordered_semiring);;

let ordered_semiring_0_rat =
  ({ordered_semigroup_mult_zero_ordered_semiring_0 =
      ordered_semigroup_mult_zero_rat;
     ordered_semiring_ordered_semiring_0 = ordered_semiring_rat;
     semiring_0_ordered_semiring_0 = semiring_0_rat}
    : rat ordered_semiring_0);;

let ordered_cancel_semiring_rat =
  ({ordered_cancel_ab_semigroup_add_ordered_cancel_semiring =
      ordered_cancel_ab_semigroup_add_rat;
     ordered_semiring_0_ordered_cancel_semiring = ordered_semiring_0_rat;
     semiring_0_cancel_ordered_cancel_semiring = semiring_0_cancel_rat}
    : rat ordered_cancel_semiring);;

let ordered_ab_semigroup_add_imp_le_rat =
  ({ordered_cancel_ab_semigroup_add_ordered_ab_semigroup_add_imp_le =
      ordered_cancel_ab_semigroup_add_rat}
    : rat ordered_ab_semigroup_add_imp_le);;

let strict_ordered_comm_monoid_add_rat =
  ({comm_monoid_add_strict_ordered_comm_monoid_add = comm_monoid_add_rat;
     strict_ordered_ab_semigroup_add_strict_ordered_comm_monoid_add =
       strict_ordered_ab_semigroup_add_rat}
    : rat strict_ordered_comm_monoid_add);;

let ordered_cancel_comm_monoid_add_rat =
  ({ordered_cancel_ab_semigroup_add_ordered_cancel_comm_monoid_add =
      ordered_cancel_ab_semigroup_add_rat;
     ordered_comm_monoid_add_ordered_cancel_comm_monoid_add =
       ordered_comm_monoid_add_rat;
     strict_ordered_comm_monoid_add_ordered_cancel_comm_monoid_add =
       strict_ordered_comm_monoid_add_rat}
    : rat ordered_cancel_comm_monoid_add);;

let ordered_ab_semigroup_monoid_add_imp_le_rat =
  ({cancel_comm_monoid_add_ordered_ab_semigroup_monoid_add_imp_le =
      cancel_comm_monoid_add_rat;
     ordered_ab_semigroup_add_imp_le_ordered_ab_semigroup_monoid_add_imp_le =
       ordered_ab_semigroup_add_imp_le_rat;
     ordered_cancel_comm_monoid_add_ordered_ab_semigroup_monoid_add_imp_le =
       ordered_cancel_comm_monoid_add_rat}
    : rat ordered_ab_semigroup_monoid_add_imp_le);;

let ordered_ab_group_add_rat =
  ({ab_group_add_ordered_ab_group_add = ab_group_add_rat;
     ordered_ab_semigroup_monoid_add_imp_le_ordered_ab_group_add =
       ordered_ab_semigroup_monoid_add_imp_le_rat}
    : rat ordered_ab_group_add);;

let ordered_ring_rat =
  ({ordered_ab_group_add_ordered_ring = ordered_ab_group_add_rat;
     ordered_cancel_semiring_ordered_ring = ordered_cancel_semiring_rat;
     ring_ordered_ring = ring_rat}
    : rat ordered_ring);;

type 'a field_char_0 =
  {field_field_char_0 : 'a field; ring_char_0_field_char_0 : 'a ring_char_0};;

let field_char_0_rat =
  ({field_field_char_0 = field_rat; ring_char_0_field_char_0 = ring_char_0_rat}
    : rat field_char_0);;

let zero_less_one_rat =
  ({order_zero_less_one = order_rat;
     zero_neq_one_zero_less_one = zero_neq_one_rat}
    : rat zero_less_one);;

let rec showsl_rat
  x = (let (d, n) = quotient_of x in
        (if equal_inta n one_inta then showsl_int d
          else comp (comp (showsl_int d) (showsl_lit "/")) (showsl_int n)));;

let rec showsl_list_rat xs = default_showsl_list showsl_rat xs;;

let showl_rat =
  ({showsl = showsl_rat; showsl_list = showsl_list_rat} : rat showl);;

type 'a field_abs_sgn =
  {field_field_abs_sgn : 'a field;
    idom_abs_sgn_field_abs_sgn : 'a idom_abs_sgn};;

let field_abs_sgn_rat =
  ({field_field_abs_sgn = field_rat;
     idom_abs_sgn_field_abs_sgn = idom_abs_sgn_rat}
    : rat field_abs_sgn);;

let cEnum_rat :
  (rat list * (((rat -> bool) -> bool) * ((rat -> bool) -> bool))) option
  = None;;

let cenum_rat = ({cEnum = cEnum_rat} : rat cenum);;

type 'a dense_order = {order_dense_order : 'a order};;

let dense_order_rat = ({order_dense_order = order_rat} : rat dense_order);;

let linordered_ab_semigroup_add_rat =
  ({ordered_ab_semigroup_add_linordered_ab_semigroup_add =
      ordered_ab_semigroup_add_rat;
     linorder_linordered_ab_semigroup_add = linorder_rat}
    : rat linordered_ab_semigroup_add);;

let linordered_cancel_ab_semigroup_add_rat =
  ({linordered_ab_semigroup_add_linordered_cancel_ab_semigroup_add =
      linordered_ab_semigroup_add_rat;
     ordered_ab_semigroup_add_imp_le_linordered_cancel_ab_semigroup_add =
       ordered_ab_semigroup_add_imp_le_rat}
    : rat linordered_cancel_ab_semigroup_add);;

let linordered_semiring_rat =
  ({linordered_cancel_ab_semigroup_add_linordered_semiring =
      linordered_cancel_ab_semigroup_add_rat;
     ordered_ab_semigroup_monoid_add_imp_le_linordered_semiring =
       ordered_ab_semigroup_monoid_add_imp_le_rat;
     ordered_cancel_semiring_linordered_semiring = ordered_cancel_semiring_rat}
    : rat linordered_semiring);;

let linordered_semiring_strict_rat =
  ({linordered_semiring_linordered_semiring_strict = linordered_semiring_rat} :
    rat linordered_semiring_strict);;

let linordered_semiring_1_rat =
  ({linordered_semiring_linordered_semiring_1 = linordered_semiring_rat;
     semiring_1_linordered_semiring_1 = semiring_1_rat;
     zero_less_one_linordered_semiring_1 = zero_less_one_rat}
    : rat linordered_semiring_1);;

let linordered_semiring_1_strict_rat =
  ({linordered_semiring_1_linordered_semiring_1_strict =
      linordered_semiring_1_rat;
     linordered_semiring_strict_linordered_semiring_1_strict =
       linordered_semiring_strict_rat}
    : rat linordered_semiring_1_strict);;

let ordered_ab_group_add_abs_rat =
  ({abs_ordered_ab_group_add_abs = abs_rat;
     ordered_ab_group_add_ordered_ab_group_add_abs = ordered_ab_group_add_rat}
    : rat ordered_ab_group_add_abs);;

let linordered_ab_group_add_rat =
  ({linordered_cancel_ab_semigroup_add_linordered_ab_group_add =
      linordered_cancel_ab_semigroup_add_rat;
     ordered_ab_group_add_linordered_ab_group_add = ordered_ab_group_add_rat}
    : rat linordered_ab_group_add);;

let linordered_ring_rat =
  ({linordered_ab_group_add_linordered_ring = linordered_ab_group_add_rat;
     ordered_ab_group_add_abs_linordered_ring = ordered_ab_group_add_abs_rat;
     abs_if_linordered_ring = abs_if_rat;
     linordered_semiring_linordered_ring = linordered_semiring_rat;
     ordered_ring_linordered_ring = ordered_ring_rat}
    : rat linordered_ring);;

let linordered_ring_strict_rat =
  ({linordered_ring_linordered_ring_strict = linordered_ring_rat;
     linordered_semiring_strict_linordered_ring_strict =
       linordered_semiring_strict_rat;
     ring_no_zero_divisors_linordered_ring_strict = ring_no_zero_divisors_rat}
    : rat linordered_ring_strict);;

let ordered_comm_semiring_rat =
  ({comm_semiring_0_ordered_comm_semiring = comm_semiring_0_rat;
     ordered_semiring_ordered_comm_semiring = ordered_semiring_rat}
    : rat ordered_comm_semiring);;

let ordered_cancel_comm_semiring_rat =
  ({comm_semiring_0_cancel_ordered_cancel_comm_semiring =
      comm_semiring_0_cancel_rat;
     ordered_cancel_semiring_ordered_cancel_comm_semiring =
       ordered_cancel_semiring_rat;
     ordered_comm_semiring_ordered_cancel_comm_semiring =
       ordered_comm_semiring_rat}
    : rat ordered_cancel_comm_semiring);;

let linordered_comm_semiring_strict_rat =
  ({linordered_semiring_strict_linordered_comm_semiring_strict =
      linordered_semiring_strict_rat;
     ordered_cancel_comm_semiring_linordered_comm_semiring_strict =
       ordered_cancel_comm_semiring_rat}
    : rat linordered_comm_semiring_strict);;

let linordered_nonzero_semiring_rat =
  ({semiring_char_0_linordered_nonzero_semiring = semiring_char_0_rat;
     linorder_linordered_nonzero_semiring = linorder_rat;
     comm_semiring_1_linordered_nonzero_semiring = comm_semiring_1_rat;
     ordered_comm_semiring_linordered_nonzero_semiring =
       ordered_comm_semiring_rat;
     zero_less_one_linordered_nonzero_semiring = zero_less_one_rat}
    : rat linordered_nonzero_semiring);;

let linordered_semidom_rat =
  ({linordered_comm_semiring_strict_linordered_semidom =
      linordered_comm_semiring_strict_rat;
     linordered_nonzero_semiring_linordered_semidom =
       linordered_nonzero_semiring_rat;
     semidom_linordered_semidom = semidom_rat}
    : rat linordered_semidom);;

let ordered_comm_ring_rat =
  ({comm_ring_ordered_comm_ring = comm_ring_rat;
     ordered_cancel_comm_semiring_ordered_comm_ring =
       ordered_cancel_comm_semiring_rat;
     ordered_ring_ordered_comm_ring = ordered_ring_rat}
    : rat ordered_comm_ring);;

let ordered_ring_abs_rat =
  ({ordered_ab_group_add_abs_ordered_ring_abs = ordered_ab_group_add_abs_rat;
     ordered_ring_ordered_ring_abs = ordered_ring_rat}
    : rat ordered_ring_abs);;

let linordered_idom_rat =
  ({ring_char_0_linordered_idom = ring_char_0_rat;
     idom_abs_sgn_linordered_idom = idom_abs_sgn_rat;
     linordered_ring_strict_linordered_idom = linordered_ring_strict_rat;
     linordered_semidom_linordered_idom = linordered_semidom_rat;
     linordered_semiring_1_strict_linordered_idom =
       linordered_semiring_1_strict_rat;
     ordered_comm_ring_linordered_idom = ordered_comm_ring_rat;
     ordered_ring_abs_linordered_idom = ordered_ring_abs_rat}
    : rat linordered_idom);;

let non_strict_order_rat =
  ({ord_non_strict_order = ord_rat} : rat non_strict_order);;

let ordered_ab_semigroup_rat =
  ({ab_semigroup_add_ordered_ab_semigroup = ab_semigroup_add_rat;
     monoid_add_ordered_ab_semigroup = monoid_add_rat;
     non_strict_order_ordered_ab_semigroup = non_strict_order_rat}
    : rat ordered_ab_semigroup);;

let ordered_semiring_0_rata =
  ({semiring_0_ordered_semiring_0a = semiring_0_rat;
     ordered_ab_semigroup_ordered_semiring_0 = ordered_ab_semigroup_rat}
    : rat ordered_semiring_0a);;

let ordered_semiring_1_rat =
  ({semiring_1_ordered_semiring_1 = semiring_1_rat;
     ordered_semiring_0_ordered_semiring_1 = ordered_semiring_0_rata}
    : rat ordered_semiring_1);;

let poly_carrier_rat =
  ({comm_semiring_1_poly_carrier = comm_semiring_1_rat;
     ordered_semiring_1_poly_carrier = ordered_semiring_1_rat}
    : rat poly_carrier);;

type 'a dense_linorder =
  {dense_order_dense_linorder : 'a dense_order;
    linorder_dense_linorder : 'a linorder};;

type 'a unbounded_dense_linorder =
  {dense_linorder_unbounded_dense_linorder : 'a dense_linorder;
    no_bot_unbounded_dense_linorder : 'a no_bot;
    no_top_unbounded_dense_linorder : 'a no_top};;

type 'a linordered_field =
  {field_abs_sgn_linordered_field : 'a field_abs_sgn;
    field_char_0_linordered_field : 'a field_char_0;
    unbounded_dense_linorder_linordered_field : 'a unbounded_dense_linorder;
    linordered_idom_linordered_field : 'a linordered_idom};;

let dense_linorder_rat =
  ({dense_order_dense_linorder = dense_order_rat;
     linorder_dense_linorder = linorder_rat}
    : rat dense_linorder);;

let unbounded_dense_linorder_rat =
  ({dense_linorder_unbounded_dense_linorder = dense_linorder_rat;
     no_bot_unbounded_dense_linorder = no_bot_rat;
     no_top_unbounded_dense_linorder = no_top_rat}
    : rat unbounded_dense_linorder);;

let linordered_field_rat =
  ({field_abs_sgn_linordered_field = field_abs_sgn_rat;
     field_char_0_linordered_field = field_char_0_rat;
     unbounded_dense_linorder_linordered_field = unbounded_dense_linorder_rat;
     linordered_idom_linordered_field = linordered_idom_rat}
    : rat linordered_field);;

let ccompare_rata : (rat -> rat -> ordera) option = Some compare_rata;;

let ccompare_rat = ({ccompare = ccompare_rata} : rat ccompare);;

let rec real_of_rat x = ratreal x;;

let real_embedding_rat =
  ({linordered_idom_real_embedding = linordered_idom_rat; real_of = real_of_rat}
    : rat real_embedding);;

type 'a archimedean_field =
  {linordered_field_archimedean_field : 'a linordered_field};;

type 'a floor_ceiling =
  {archimedean_field_floor_ceiling : 'a archimedean_field;
    large_ordered_semiring_1_floor_ceiling : 'a large_ordered_semiring_1;
    floor : 'a -> int};;
let floor _A = _A.floor;;

let rec floor_rat p = (let a = quotient_of p in
                       let (aa, b) = a in
                        divide_int aa b);;

let archimedean_field_rat =
  ({linordered_field_archimedean_field = linordered_field_rat} :
    rat archimedean_field);;

let large_ordered_semiring_1_rat =
  ({poly_carrier_large_ordered_semiring_1 = poly_carrier_rat} :
    rat large_ordered_semiring_1);;

let floor_ceiling_rat =
  ({archimedean_field_floor_ceiling = archimedean_field_rat;
     large_ordered_semiring_1_floor_ceiling = large_ordered_semiring_1_rat;
     floor = floor_rat}
    : rat floor_ceiling);;

let large_real_ordered_semiring_1_rat =
  ({real_embedding_large_real_ordered_semiring_1 = real_embedding_rat;
     large_ordered_semiring_1_large_real_ordered_semiring_1 =
       large_ordered_semiring_1_rat}
    : rat large_real_ordered_semiring_1);;

let rec equal_order x0 x1 = match x0, x1 with Lt, Gt -> false
                      | Gt, Lt -> false
                      | Eq, Gt -> false
                      | Gt, Eq -> false
                      | Eq, Lt -> false
                      | Lt, Eq -> false
                      | Gt, Gt -> true
                      | Lt, Lt -> true
                      | Eq, Eq -> true;;

type ('a, 'b) generator = Generator of (('b -> bool) * ('b -> 'a * 'b));;

let rec generator (Generator x) = x;;

let rec has_next g = fst (generator g);;

let rec next g = snd (generator g);;

let rec sorted_list_subset_fusion
  less eq g1 g2 s1 s2 =
    (if has_next g1 s1
      then (let (x, s1a) = next g1 s1 in
             has_next g2 s2 &&
               (let (y, s2a) = next g2 s2 in
                 (if eq x y then sorted_list_subset_fusion less eq g1 g2 s1a s2a
                   else less y x &&
                          sorted_list_subset_fusion less eq g1 g2 s1 s2a)))
      else true);;

let rec list_all_fusion
  g p s =
    (if has_next g s
      then (let (x, sa) = next g s in p x && list_all_fusion g p sa)
      else true);;

type color = R | B;;

type ('a, 'b) rbta = Empty |
  Branch of color * ('a, 'b) rbta * 'a * 'b * ('a, 'b) rbta;;

let rec rbt_keys_next
  = function ((k, t) :: kts, Empty) -> (k, (kts, t))
    | (kts, Branch (c, l, k, v, r)) -> rbt_keys_next ((k, r) :: kts, l);;

let rec rbt_has_next = function ([], Empty) -> false
                       | (vb :: vc, va) -> true
                       | (v, Branch (vb, vc, vd, ve, vf)) -> true;;

let rbt_keys_generator :
  ('a, (('a * ('a, 'b) rbta) list * ('a, 'b) rbta)) generator
  = Generator (rbt_has_next, rbt_keys_next);;

let rec lt_of_comp
  acomp x y = (match acomp x y with Eq -> false | Lt -> true | Gt -> false);;

type ('b, 'a) mapping_rbt = Mapping_RBTa of ('b, 'a) rbta;;

type 'a set_dlist = Abs_dlist of 'a list;;

type 'a set = Collect_set of ('a -> bool) | DList_set of 'a set_dlist |
  RBT_set of ('a, unit) mapping_rbt | Set_Monad of 'a list |
  Complement of 'a set;;

let rec list_of_dlist _A (Abs_dlist x) = x;;

let rec list_all p x1 = match p, x1 with p, [] -> true
                   | p, x :: xs -> p x && list_all p xs;;

let rec dlist_all _A x xc = list_all x (list_of_dlist _A xc);;

let rec impl_ofb _B (Mapping_RBTa x) = x;;

let rec rbt_init x = ([], x);;

let rec init _A xa = rbt_init (impl_ofb _A xa);;

let rec filtera
  p x1 = match p, x1 with p, [] -> []
    | p, x :: xs -> (if p x then x :: filtera p xs else filtera p xs);;

let rec collect _A
  p = (match cEnum _A with None -> Collect_set p
        | Some (enum, _) -> Set_Monad (filtera p enum));;

let rec list_member
  equal x1 y = match equal, x1, y with
    equal, x :: xs, y -> equal x y || list_member equal xs y
    | equal, [], y -> false;;

let rec the (Some x2) = x2;;

let rec memberc _A xa = list_member (the (ceq _A)) (list_of_dlist _A xa);;

let rec equal_optiona _A x0 x1 = match x0, x1 with None, Some x2 -> false
                           | Some x2, None -> false
                           | Some x2, Some y2 -> eq _A x2 y2
                           | None, None -> true;;

let rec rbt_comp_lookup
  c x1 k = match c, x1, k with c, Empty, k -> None
    | c, Branch (uu, l, x, y, r), k ->
        (match c k x with Eq -> Some y | Lt -> rbt_comp_lookup c l k
          | Gt -> rbt_comp_lookup c r k);;

let rec lookupd _A xa = rbt_comp_lookup (the (ccompare _A)) (impl_ofb _A xa);;

let rec equal_unita u v = true;;

let equal_unit = ({equal = equal_unita} : unit equal);;

let rec memberb _A t x = equal_optiona equal_unit (lookupd _A t x) (Some ());;

let rec member (_A1, _A2)
  x xa1 = match x, xa1 with
    x, Set_Monad xs ->
      (match ceq _A1
        with None ->
          failwith "member Set_Monad: ceq = None"
            (fun _ -> member (_A1, _A2) x (Set_Monad xs))
        | Some eq -> list_member eq xs x)
    | xa, Complement x -> not (member (_A1, _A2) xa x)
    | x, RBT_set rbt -> memberb _A2 rbt x
    | x, DList_set dxs -> memberc _A1 dxs x
    | x, Collect_set a -> a x;;

let rec subset_eq (_A1, _A2, _A3)
  x0 c = match x0, c with
    RBT_set rbt1, RBT_set rbt2 ->
      (match ccompare _A3
        with None ->
          failwith "subset RBT_set RBT_set: ccompare = None"
            (fun _ -> subset_eq (_A1, _A2, _A3) (RBT_set rbt1) (RBT_set rbt2))
        | Some c ->
          (match ceq _A2
            with None ->
              sorted_list_subset_fusion (lt_of_comp c)
                (fun x y -> equal_order (c x y) Eq) rbt_keys_generator
                rbt_keys_generator (init _A3 rbt1) (init _A3 rbt2)
            | Some eq ->
              sorted_list_subset_fusion (lt_of_comp c) eq rbt_keys_generator
                rbt_keys_generator (init _A3 rbt1) (init _A3 rbt2)))
    | Complement a1, Complement a2 -> subset_eq (_A1, _A2, _A3) a2 a1
    | Collect_set p, Complement a ->
        subset_eq (_A1, _A2, _A3) a (collect _A1 (fun x -> not (p x)))
    | Set_Monad xs, c -> list_all (fun x -> member (_A2, _A3) x c) xs
    | DList_set dxs, c ->
        (match ceq _A2
          with None ->
            failwith "subset DList_set1: ceq = None"
              (fun _ -> subset_eq (_A1, _A2, _A3) (DList_set dxs) c)
          | Some _ -> dlist_all _A2 (fun x -> member (_A2, _A3) x c) dxs)
    | RBT_set rbt, b ->
        (match ccompare _A3
          with None ->
            failwith "subset RBT_set1: ccompare = None"
              (fun _ -> subset_eq (_A1, _A2, _A3) (RBT_set rbt) b)
          | Some _ ->
            list_all_fusion rbt_keys_generator (fun x -> member (_A2, _A3) x b)
              (init _A3 rbt));;

let rec less_eq_set (_A1, _A2, _A3) = subset_eq (_A1, _A2, _A3);;

let rec equal_seta (_A1, _A2, _A3, _A4)
  a b = less_eq_set (_A1, _A2, _A3) a b && less_eq_set (_A1, _A2, _A3) b a;;

let rec equal_set (_A1, _A2, _A3, _A4) =
  ({equal = equal_seta (_A1, _A2, _A3, _A4)} : 'a set equal);;

let rec uminus_set = function Complement b -> b
                     | Collect_set p -> Collect_set (fun x -> not (p x))
                     | a -> Complement a;;

let rec balance
  x0 s t x3 = match x0, s, t, x3 with
    Branch (R, a, w, x, b), s, t, Branch (R, c, y, z, d) ->
      Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, d))
    | Branch (R, Branch (R, a, w, x, b), s, t, c), y, z, Empty ->
        Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, Empty))
    | Branch (R, Branch (R, a, w, x, b), s, t, c), y, z,
        Branch (B, va, vb, vc, vd)
        -> Branch
             (R, Branch (B, a, w, x, b), s, t,
               Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
    | Branch (R, Empty, w, x, Branch (R, b, s, t, c)), y, z, Empty ->
        Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
    | Branch (R, Branch (B, va, vb, vc, vd), w, x, Branch (R, b, s, t, c)), y,
        z, Empty
        -> Branch
             (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
               Branch (B, c, y, z, Empty))
    | Branch (R, Empty, w, x, Branch (R, b, s, t, c)), y, z,
        Branch (B, va, vb, vc, vd)
        -> Branch
             (R, Branch (B, Empty, w, x, b), s, t,
               Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
    | Branch (R, Branch (B, ve, vf, vg, vh), w, x, Branch (R, b, s, t, c)), y,
        z, Branch (B, va, vb, vc, vd)
        -> Branch
             (R, Branch (B, Branch (B, ve, vf, vg, vh), w, x, b), s, t,
               Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
    | Empty, w, x, Branch (R, b, s, t, Branch (R, c, y, z, d)) ->
        Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, d))
    | Branch (B, va, vb, vc, vd), w, x,
        Branch (R, b, s, t, Branch (R, c, y, z, d))
        -> Branch
             (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
               Branch (B, c, y, z, d))
    | Empty, w, x, Branch (R, Branch (R, b, s, t, c), y, z, Empty) ->
        Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
    | Empty, w, x,
        Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, va, vb, vc, vd))
        -> Branch
             (R, Branch (B, Empty, w, x, b), s, t,
               Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
    | Branch (B, va, vb, vc, vd), w, x,
        Branch (R, Branch (R, b, s, t, c), y, z, Empty)
        -> Branch
             (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
               Branch (B, c, y, z, Empty))
    | Branch (B, va, vb, vc, vd), w, x,
        Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, ve, vf, vg, vh))
        -> Branch
             (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
               Branch (B, c, y, z, Branch (B, ve, vf, vg, vh)))
    | Empty, s, t, Empty -> Branch (B, Empty, s, t, Empty)
    | Empty, s, t, Branch (B, va, vb, vc, vd) ->
        Branch (B, Empty, s, t, Branch (B, va, vb, vc, vd))
    | Empty, s, t, Branch (v, Empty, vb, vc, Empty) ->
        Branch (B, Empty, s, t, Branch (v, Empty, vb, vc, Empty))
    | Empty, s, t, Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty) ->
        Branch
          (B, Empty, s, t,
            Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty))
    | Empty, s, t, Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi)) ->
        Branch
          (B, Empty, s, t,
            Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi)))
    | Empty, s, t,
        Branch
          (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi))
        -> Branch
             (B, Empty, s, t,
               Branch
                 (v, Branch (B, ve, vj, vk, vl), vb, vc,
                   Branch (B, vf, vg, vh, vi)))
    | Branch (B, va, vb, vc, vd), s, t, Empty ->
        Branch (B, Branch (B, va, vb, vc, vd), s, t, Empty)
    | Branch (B, va, vb, vc, vd), s, t, Branch (B, ve, vf, vg, vh) ->
        Branch (B, Branch (B, va, vb, vc, vd), s, t, Branch (B, ve, vf, vg, vh))
    | Branch (B, va, vb, vc, vd), s, t, Branch (v, Empty, vf, vg, Empty) ->
        Branch
          (B, Branch (B, va, vb, vc, vd), s, t,
            Branch (v, Empty, vf, vg, Empty))
    | Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty)
        -> Branch
             (B, Branch (B, va, vb, vc, vd), s, t,
               Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty))
    | Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm))
        -> Branch
             (B, Branch (B, va, vb, vc, vd), s, t,
               Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm)))
    | Branch (B, va, vb, vc, vd), s, t,
        Branch
          (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm))
        -> Branch
             (B, Branch (B, va, vb, vc, vd), s, t,
               Branch
                 (v, Branch (B, vi, vn, vo, vp), vf, vg,
                   Branch (B, vj, vk, vl, vm)))
    | Branch (v, Empty, vb, vc, Empty), s, t, Empty ->
        Branch (B, Branch (v, Empty, vb, vc, Empty), s, t, Empty)
    | Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh)), s, t, Empty ->
        Branch
          (B, Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh)), s, t,
            Empty)
    | Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty), s, t, Empty ->
        Branch
          (B, Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty), s, t,
            Empty)
    | Branch
        (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)),
        s, t, Empty
        -> Branch
             (B, Branch
                   (v, Branch (B, vf, vg, vh, vi), vb, vc,
                     Branch (B, ve, vj, vk, vl)),
               s, t, Empty)
    | Branch (v, Empty, vf, vg, Empty), s, t, Branch (B, va, vb, vc, vd) ->
        Branch
          (B, Branch (v, Empty, vf, vg, Empty), s, t,
            Branch (B, va, vb, vc, vd))
    | Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl)), s, t,
        Branch (B, va, vb, vc, vd)
        -> Branch
             (B, Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl)), s, t,
               Branch (B, va, vb, vc, vd))
    | Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty), s, t,
        Branch (B, va, vb, vc, vd)
        -> Branch
             (B, Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty), s, t,
               Branch (B, va, vb, vc, vd))
    | Branch
        (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)),
        s, t, Branch (B, va, vb, vc, vd)
        -> Branch
             (B, Branch
                   (v, Branch (B, vj, vk, vl, vm), vf, vg,
                     Branch (B, vi, vn, vo, vp)),
               s, t, Branch (B, va, vb, vc, vd));;

let rec rbt_comp_ins
  c f k v x4 = match c, f, k, v, x4 with
    c, f, k, v, Empty -> Branch (R, Empty, k, v, Empty)
    | c, f, k, v, Branch (B, l, x, y, r) ->
        (match c k x with Eq -> Branch (B, l, x, f k y v, r)
          | Lt -> balance (rbt_comp_ins c f k v l) x y r
          | Gt -> balance l x y (rbt_comp_ins c f k v r))
    | c, f, k, v, Branch (R, l, x, y, r) ->
        (match c k x with Eq -> Branch (R, l, x, f k y v, r)
          | Lt -> Branch (R, rbt_comp_ins c f k v l, x, y, r)
          | Gt -> Branch (R, l, x, y, rbt_comp_ins c f k v r));;

let rec paint c x1 = match c, x1 with c, Empty -> Empty
                | c, Branch (uu, l, k, v, r) -> Branch (c, l, k, v, r);;

let rec rbt_comp_insert_with_key c f k v t = paint B (rbt_comp_ins c f k v t);;

let rec rbt_comp_insert c = rbt_comp_insert_with_key c (fun _ _ nv -> nv);;

let rec inserte _A
  xc xd xe =
    Mapping_RBTa (rbt_comp_insert (the (ccompare _A)) xc xd (impl_ofb _A xe));;

let rec comp_sunion_with
  c f asa bs = match c, f, asa, bs with
    c, f, (ka, va) :: asa, (k, v) :: bs ->
      (match c k ka with Eq -> (ka, f ka va v) :: comp_sunion_with c f asa bs
        | Lt -> (k, v) :: comp_sunion_with c f ((ka, va) :: asa) bs
        | Gt -> (ka, va) :: comp_sunion_with c f asa ((k, v) :: bs))
    | c, f, [], bs -> bs
    | c, f, asa, [] -> asa;;

type comparea = LT | GT | EQ;;

let rec skip_red = function Branch (R, l, k, v, r) -> l
                   | Empty -> Empty
                   | Branch (B, va, vb, vc, vd) -> Branch (B, va, vb, vc, vd);;

let rec skip_black
  t = (let ta = skip_red t in
        (match ta with Empty -> ta | Branch (R, _, _, _, _) -> ta
          | Branch (B, l, _, _, _) -> l));;

let rec compare_height
  sx s t tx =
    (match (skip_red sx, (skip_red s, (skip_red t, skip_red tx)))
      with (Empty, (Empty, (_, Empty))) -> EQ
      | (Empty, (Empty, (_, Branch (_, _, _, _, _)))) -> LT
      | (Empty, (Branch (_, _, _, _, _), (Empty, _))) -> EQ
      | (Empty, (Branch (_, _, _, _, _), (Branch (_, _, _, _, _), Empty))) -> EQ
      | (Empty,
          (Branch (_, sa, _, _, _),
            (Branch (_, ta, _, _, _), Branch (_, txa, _, _, _))))
        -> compare_height Empty sa ta (skip_black txa)
      | (Branch (_, _, _, _, _), (Empty, (Empty, Empty))) -> GT
      | (Branch (_, _, _, _, _), (Empty, (Empty, Branch (_, _, _, _, _)))) -> LT
      | (Branch (_, _, _, _, _), (Empty, (Branch (_, _, _, _, _), Empty))) -> EQ
      | (Branch (_, _, _, _, _),
          (Empty, (Branch (_, _, _, _, _), Branch (_, _, _, _, _))))
        -> LT
      | (Branch (_, _, _, _, _), (Branch (_, _, _, _, _), (Empty, _))) -> GT
      | (Branch (_, sxa, _, _, _),
          (Branch (_, sa, _, _, _), (Branch (_, ta, _, _, _), Empty)))
        -> compare_height (skip_black sxa) sa ta Empty
      | (Branch (_, sxa, _, _, _),
          (Branch (_, sa, _, _, _),
            (Branch (_, ta, _, _, _), Branch (_, txa, _, _, _))))
        -> compare_height (skip_black sxa) sa ta (skip_black txa));;

let rec suc n = plus_nata n one_nata;;

let rec gen_length n x1 = match n, x1 with n, x :: xs -> gen_length (suc n) xs
                     | n, [] -> n;;

let rec size_list x = gen_length zero_nata x;;

let rec apfst f (x, y) = (f x, y);;

let rec divmod_nat m n = (divide_nata m n, modulo_nata m n);;

let rec rbtreeify_g
  n kvs =
    (if equal_nata n zero_nata || equal_nata n one_nata then (Empty, kvs)
      else (let (na, r) =
              divmod_nat n (nat_of_integer (Big_int.big_int_of_int 2)) in
             (if equal_nata r zero_nata
               then (let (t1, (k, v) :: kvsa) = rbtreeify_g na kvs in
                      apfst (fun a -> Branch (B, t1, k, v, a))
                        (rbtreeify_g na kvsa))
               else (let (t1, (k, v) :: kvsa) = rbtreeify_f na kvs in
                      apfst (fun a -> Branch (B, t1, k, v, a))
                        (rbtreeify_g na kvsa)))))
and rbtreeify_f
  n kvs =
    (if equal_nata n zero_nata then (Empty, kvs)
      else (if equal_nata n one_nata
             then (let (k, v) :: kvsa = kvs in
                    (Branch (R, Empty, k, v, Empty), kvsa))
             else (let (na, r) =
                     divmod_nat n (nat_of_integer (Big_int.big_int_of_int 2)) in
                    (if equal_nata r zero_nata
                      then (let (t1, (k, v) :: kvsa) = rbtreeify_f na kvs in
                             apfst (fun a -> Branch (B, t1, k, v, a))
                               (rbtreeify_g na kvsa))
                      else (let (t1, (k, v) :: kvsa) = rbtreeify_f na kvs in
                             apfst (fun a -> Branch (B, t1, k, v, a))
                               (rbtreeify_f na kvsa))))));;

let rec rbtreeify kvs = fst (rbtreeify_g (suc (size_list kvs)) kvs);;

let rec gen_entries
  kvts x1 = match kvts, x1 with
    kvts, Branch (c, l, k, v, r) -> gen_entries (((k, v), r) :: kvts) l
    | (kv, t) :: kvts, Empty -> kv :: gen_entries kvts t
    | [], Empty -> [];;

let rec entries x = gen_entries [] x;;

let rec folda
  f xa1 x = match f, xa1, x with
    f, Branch (c, lt, k, v, rt), x -> folda f rt (f k v (folda f lt x))
    | f, Empty, x -> x;;

let rec rbt_comp_union_with_key
  c f t1 t2 =
    (match compare_height t1 t1 t2 t2
      with LT -> folda (rbt_comp_insert_with_key c (fun k v w -> f k w v)) t1 t2
      | GT -> folda (rbt_comp_insert_with_key c f) t2 t1
      | EQ -> rbtreeify (comp_sunion_with c f (entries t1) (entries t2)));;

let rec join _A
  xc xd xe =
    Mapping_RBTa
      (rbt_comp_union_with_key (the (ccompare _A)) xc (impl_ofb _A xd)
        (impl_ofb _A xe));;

let rec list_insert
  equal x xs = (if list_member equal xs x then xs else x :: xs);;

let rec insertc _A
  xb xc = Abs_dlist (list_insert (the (ceq _A)) xb (list_of_dlist _A xc));;

let rec fold f x1 s = match f, x1, s with f, x :: xs, s -> fold f xs (f x s)
               | f, [], s -> s;;

let rec foldc _A x xc = fold x (list_of_dlist _A xc);;

let rec uniona _A = foldc _A (insertc _A);;

let rec id x = (fun xa -> xa) x;;

let rec is_none = function Some x -> false
                  | None -> true;;

let rec inter_list _A
  xb xc =
    Mapping_RBTa
      (fold (fun k -> rbt_comp_insert (the (ccompare _A)) k ())
        (filtera
          (fun x ->
            not (is_none
                  (rbt_comp_lookup (the (ccompare _A)) (impl_ofb _A xb) x)))
          xc)
        Empty);;

let rec filterd _A
  xb xc = Mapping_RBTa (rbtreeify (filtera xb (entries (impl_ofb _A xc))));;

let rec comp_sinter_with
  c f uv uu = match c, f, uv, uu with
    c, f, (ka, va) :: asa, (k, v) :: bs ->
      (match c k ka with Eq -> (ka, f ka va v) :: comp_sinter_with c f asa bs
        | Lt -> comp_sinter_with c f ((ka, va) :: asa) bs
        | Gt -> comp_sinter_with c f asa ((k, v) :: bs))
    | c, f, [], uu -> []
    | c, f, uv, [] -> [];;

let rec map_option f x1 = match f, x1 with f, None -> None
                     | f, Some x2 -> Some (f x2);;

let rec map_filter
  f x1 = match f, x1 with f, [] -> []
    | f, x :: xs ->
        (match f x with None -> map_filter f xs
          | Some y -> y :: map_filter f xs);;

let rec rbt_comp_inter_with_key
  c f t1 t2 =
    (match compare_height t1 t1 t2 t2
      with LT ->
        rbtreeify
          (map_filter
            (fun (k, v) ->
              map_option (fun w -> (k, f k v w)) (rbt_comp_lookup c t2 k))
            (entries t1))
      | GT ->
        rbtreeify
          (map_filter
            (fun (k, v) ->
              map_option (fun w -> (k, f k w v)) (rbt_comp_lookup c t1 k))
            (entries t2))
      | EQ -> rbtreeify (comp_sinter_with c f (entries t1) (entries t2)));;

let rec meet _A
  xc xd xe =
    Mapping_RBTa
      (rbt_comp_inter_with_key (the (ccompare _A)) xc (impl_ofb _A xd)
        (impl_ofb _A xe));;

let rec filterc _A xb xc = Abs_dlist (filtera xb (list_of_dlist _A xc));;

let rec inf_seta (_A1, _A2)
  g ga = match g, ga with
    RBT_set rbt1, Set_Monad xs ->
      (match ccompare _A2
        with None ->
          failwith "inter RBT_set Set_Monad: ccompare = None"
            (fun _ -> inf_seta (_A1, _A2) (RBT_set rbt1) (Set_Monad xs))
        | Some _ -> RBT_set (inter_list _A2 rbt1 xs))
    | RBT_set rbt, DList_set dxs ->
        (match ccompare _A2
          with None ->
            failwith "inter RBT_set DList_set: ccompare = None"
              (fun _ -> inf_seta (_A1, _A2) (RBT_set rbt) (DList_set dxs))
          | Some _ ->
            (match ceq _A1
              with None ->
                failwith "inter RBT_set DList_set: ceq = None"
                  (fun _ -> inf_seta (_A1, _A2) (RBT_set rbt) (DList_set dxs))
              | Some _ -> RBT_set (inter_list _A2 rbt (list_of_dlist _A1 dxs))))
    | RBT_set rbt1, RBT_set rbt2 ->
        (match ccompare _A2
          with None ->
            failwith "inter RBT_set RBT_set: ccompare = None"
              (fun _ -> inf_seta (_A1, _A2) (RBT_set rbt1) (RBT_set rbt2))
          | Some _ -> RBT_set (meet _A2 (fun _ _ -> id) rbt1 rbt2))
    | DList_set dxs1, Set_Monad xs ->
        (match ceq _A1
          with None ->
            failwith "inter DList_set Set_Monad: ceq = None"
              (fun _ -> inf_seta (_A1, _A2) (DList_set dxs1) (Set_Monad xs))
          | Some eq -> DList_set (filterc _A1 (list_member eq xs) dxs1))
    | DList_set dxs1, DList_set dxs2 ->
        (match ceq _A1
          with None ->
            failwith "inter DList_set DList_set: ceq = None"
              (fun _ -> inf_seta (_A1, _A2) (DList_set dxs1) (DList_set dxs2))
          | Some _ -> DList_set (filterc _A1 (memberc _A1 dxs2) dxs1))
    | DList_set dxs, RBT_set rbt ->
        (match ccompare _A2
          with None ->
            failwith "inter DList_set RBT_set: ccompare = None"
              (fun _ -> inf_seta (_A1, _A2) (DList_set dxs) (RBT_set rbt))
          | Some _ ->
            (match ceq _A1
              with None ->
                failwith "inter DList_set RBT_set: ceq = None"
                  (fun _ -> inf_seta (_A1, _A2) (DList_set dxs) (RBT_set rbt))
              | Some _ -> RBT_set (inter_list _A2 rbt (list_of_dlist _A1 dxs))))
    | Set_Monad xs1, Set_Monad xs2 ->
        (match ceq _A1
          with None ->
            failwith "inter Set_Monad Set_Monad: ceq = None"
              (fun _ -> inf_seta (_A1, _A2) (Set_Monad xs1) (Set_Monad xs2))
          | Some eq -> Set_Monad (filtera (list_member eq xs2) xs1))
    | Set_Monad xs, DList_set dxs2 ->
        (match ceq _A1
          with None ->
            failwith "inter Set_Monad DList_set: ceq = None"
              (fun _ -> inf_seta (_A1, _A2) (Set_Monad xs) (DList_set dxs2))
          | Some eq -> DList_set (filterc _A1 (list_member eq xs) dxs2))
    | Set_Monad xs, RBT_set rbt1 ->
        (match ccompare _A2
          with None ->
            failwith "inter Set_Monad RBT_set: ccompare = None"
              (fun _ -> inf_seta (_A1, _A2) (RBT_set rbt1) (Set_Monad xs))
          | Some _ -> RBT_set (inter_list _A2 rbt1 xs))
    | Complement ba, Complement b -> Complement (sup_seta (_A1, _A2) ba b)
    | g, RBT_set rbt2 ->
        (match ccompare _A2
          with None ->
            failwith "inter RBT_set2: ccompare = None"
              (fun _ -> inf_seta (_A1, _A2) g (RBT_set rbt2))
          | Some _ ->
            RBT_set
              (filterd _A2 (comp (fun x -> member (_A1, _A2) x g) fst) rbt2))
    | RBT_set rbt1, g ->
        (match ccompare _A2
          with None ->
            failwith "inter RBT_set1: ccompare = None"
              (fun _ -> inf_seta (_A1, _A2) (RBT_set rbt1) g)
          | Some _ ->
            RBT_set
              (filterd _A2 (comp (fun x -> member (_A1, _A2) x g) fst) rbt1))
    | h, DList_set dxs2 ->
        (match ceq _A1
          with None ->
            failwith "inter DList_set2: ceq = None"
              (fun _ -> inf_seta (_A1, _A2) h (DList_set dxs2))
          | Some _ ->
            DList_set (filterc _A1 (fun x -> member (_A1, _A2) x h) dxs2))
    | DList_set dxs1, h ->
        (match ceq _A1
          with None ->
            failwith "inter DList_set1: ceq = None"
              (fun _ -> inf_seta (_A1, _A2) (DList_set dxs1) h)
          | Some _ ->
            DList_set (filterc _A1 (fun x -> member (_A1, _A2) x h) dxs1))
    | i, Set_Monad xs -> Set_Monad (filtera (fun x -> member (_A1, _A2) x i) xs)
    | Set_Monad xs, i -> Set_Monad (filtera (fun x -> member (_A1, _A2) x i) xs)
    | j, Collect_set a -> Collect_set (fun x -> a x && member (_A1, _A2) x j)
    | Collect_set a, j -> Collect_set (fun x -> a x && member (_A1, _A2) x j)
and sup_seta (_A1, _A2)
  ba b = match ba, b with
    ba, Complement b -> Complement (inf_seta (_A1, _A2) (uminus_set ba) b)
    | Complement ba, b -> Complement (inf_seta (_A1, _A2) ba (uminus_set b))
    | b, Collect_set a -> Collect_set (fun x -> a x || member (_A1, _A2) x b)
    | Collect_set a, b -> Collect_set (fun x -> a x || member (_A1, _A2) x b)
    | Set_Monad xs, Set_Monad ys -> Set_Monad (xs @ ys)
    | DList_set dxs1, Set_Monad ws ->
        (match ceq _A1
          with None ->
            failwith "union DList_set Set_Monad: ceq = None"
              (fun _ -> sup_seta (_A1, _A2) (DList_set dxs1) (Set_Monad ws))
          | Some _ -> DList_set (fold (insertc _A1) ws dxs1))
    | Set_Monad ws, DList_set dxs2 ->
        (match ceq _A1
          with None ->
            failwith "union Set_Monad DList_set: ceq = None"
              (fun _ -> sup_seta (_A1, _A2) (Set_Monad ws) (DList_set dxs2))
          | Some _ -> DList_set (fold (insertc _A1) ws dxs2))
    | RBT_set rbt1, Set_Monad zs ->
        (match ccompare _A2
          with None ->
            failwith "union RBT_set Set_Monad: ccompare = None"
              (fun _ -> sup_seta (_A1, _A2) (RBT_set rbt1) (Set_Monad zs))
          | Some _ -> RBT_set (fold (fun k -> inserte _A2 k ()) zs rbt1))
    | Set_Monad zs, RBT_set rbt2 ->
        (match ccompare _A2
          with None ->
            failwith "union Set_Monad RBT_set: ccompare = None"
              (fun _ -> sup_seta (_A1, _A2) (Set_Monad zs) (RBT_set rbt2))
          | Some _ -> RBT_set (fold (fun k -> inserte _A2 k ()) zs rbt2))
    | DList_set dxs1, DList_set dxs2 ->
        (match ceq _A1
          with None ->
            failwith "union DList_set DList_set: ceq = None"
              (fun _ -> sup_seta (_A1, _A2) (DList_set dxs1) (DList_set dxs2))
          | Some _ -> DList_set (uniona _A1 dxs1 dxs2))
    | DList_set dxs, RBT_set rbt ->
        (match ccompare _A2
          with None ->
            failwith "union DList_set RBT_set: ccompare = None"
              (fun _ -> sup_seta (_A1, _A2) (RBT_set rbt) (DList_set dxs))
          | Some _ ->
            (match ceq _A1
              with None ->
                failwith "union DList_set RBT_set: ceq = None"
                  (fun _ -> sup_seta (_A1, _A2) (RBT_set rbt) (DList_set dxs))
              | Some _ ->
                RBT_set (foldc _A1 (fun k -> inserte _A2 k ()) dxs rbt)))
    | RBT_set rbt, DList_set dxs ->
        (match ccompare _A2
          with None ->
            failwith "union RBT_set DList_set: ccompare = None"
              (fun _ -> sup_seta (_A1, _A2) (RBT_set rbt) (DList_set dxs))
          | Some _ ->
            (match ceq _A1
              with None ->
                failwith "union RBT_set DList_set: ceq = None"
                  (fun _ -> sup_seta (_A1, _A2) (RBT_set rbt) (DList_set dxs))
              | Some _ ->
                RBT_set (foldc _A1 (fun k -> inserte _A2 k ()) dxs rbt)))
    | RBT_set rbt1, RBT_set rbt2 ->
        (match ccompare _A2
          with None ->
            failwith "union RBT_set RBT_set: ccompare = None"
              (fun _ -> sup_seta (_A1, _A2) (RBT_set rbt1) (RBT_set rbt2))
          | Some _ -> RBT_set (join _A2 (fun _ _ -> id) rbt1 rbt2));;

let rec inf_set (_A1, _A2) = ({inf = inf_seta (_A1, _A2)} : 'a set inf);;

let rec sup_set (_A1, _A2) = ({sup = sup_seta (_A1, _A2)} : 'a set sup);;

let rec less_set (_A1, _A2, _A3)
  a b = less_eq_set (_A1, _A2, _A3) a b &&
          not (less_eq_set (_A1, _A2, _A3) b a);;

let rec ord_set (_A1, _A2, _A3) =
  ({less_eq = less_eq_set (_A1, _A2, _A3); less = less_set (_A1, _A2, _A3)} :
    'a set ord);;

let rec quasi_order_set (_A1, _A2, _A3) =
  ({ord_quasi_order = (ord_set (_A1, _A2, _A3))} : 'a set quasi_order);;

let rec weak_order_set (_A1, _A2, _A3) =
  ({quasi_order_weak_order = (quasi_order_set (_A1, _A2, _A3))} :
    'a set weak_order);;

let rec preorder_set (_A1, _A2, _A3) =
  ({ord_preorder = (ord_set (_A1, _A2, _A3))} : 'a set preorder);;

let rec order_set (_A1, _A2, _A3) =
  ({preorder_order = (preorder_set (_A1, _A2, _A3));
     weak_order_order = (weak_order_set (_A1, _A2, _A3))}
    : 'a set order);;

let rec quasi_order_sup_set (_A1, _A2, _A3) =
  ({sup_quasi_order_sup = (sup_set (_A2, _A3));
     quasi_order_quasi_order_sup = (quasi_order_set (_A1, _A2, _A3))}
    : 'a set quasi_order_sup);;

let rec quasi_semilattice_sup_set (_A1, _A2, _A3) =
  ({quasi_order_sup_quasi_semilattice_sup =
      (quasi_order_sup_set (_A1, _A2, _A3))}
    : 'a set quasi_semilattice_sup);;

let rec semilattice_sup_set (_A1, _A2, _A3) =
  ({order_semilattice_sup = (order_set (_A1, _A2, _A3));
     quasi_semilattice_sup_semilattice_sup =
       (quasi_semilattice_sup_set (_A1, _A2, _A3))}
    : 'a set semilattice_sup);;

let rec semilattice_inf_set (_A1, _A2, _A3) =
  ({inf_semilattice_inf = (inf_set (_A2, _A3));
     order_semilattice_inf = (order_set (_A1, _A2, _A3))}
    : 'a set semilattice_inf);;

let rec lattice_set (_A1, _A2, _A3) =
  ({semilattice_inf_lattice = (semilattice_inf_set (_A1, _A2, _A3));
     semilattice_sup_lattice = (semilattice_sup_set (_A1, _A2, _A3))}
    : 'a set lattice);;

let rec list_all2_fusion
  p g1 g2 s1 s2 =
    (if has_next g1 s1
      then has_next g2 s2 &&
             (let (x, s1a) = next g1 s1 in
              let (y, s2a) = next g2 s2 in
               p x y && list_all2_fusion p g1 g2 s1a s2a)
      else not (has_next g2 s2));;

let rec set_eq (_A1, _A2, _A3)
  a b = match a, b with
    RBT_set rbt1, RBT_set rbt2 ->
      (match ccompare _A3
        with None ->
          failwith "set_eq RBT_set RBT_set: ccompare = None"
            (fun _ -> set_eq (_A1, _A2, _A3) (RBT_set rbt1) (RBT_set rbt2))
        | Some c ->
          (match ceq _A2
            with None ->
              list_all2_fusion (fun x y -> equal_order (c x y) Eq)
                rbt_keys_generator rbt_keys_generator (init _A3 rbt1)
                (init _A3 rbt2)
            | Some eq ->
              list_all2_fusion eq rbt_keys_generator rbt_keys_generator
                (init _A3 rbt1) (init _A3 rbt2)))
    | Complement a, Complement b -> set_eq (_A1, _A2, _A3) a b
    | a, b ->
        less_eq_set (_A1, _A2, _A3) a b && less_eq_set (_A1, _A2, _A3) b a;;

let rec ceq_seta (_A1, _A2, _A3)
  = (match ceq _A2 with None -> None
      | Some _ -> Some (set_eq (_A1, _A2, _A3)));;

let rec ceq_set (_A1, _A2, _A3) =
  ({ceq = ceq_seta (_A1, _A2, _A3)} : 'a set ceq);;

let set_impl_seta : ('a set, set_impla) phantom = Phantom Set_Choose;;

let set_impl_set = ({set_impl = set_impl_seta} : 'a set set_impl);;

let rec of_phantom (Phantom x) = x;;

let rec emptye _A = Mapping_RBTa Empty;;

let rec emptyc _A = Abs_dlist [];;

let rec set_empty_choose (_A1, _A2)
  = (match ccompare _A2
      with None ->
        (match ceq _A1 with None -> Set_Monad []
          | Some _ -> DList_set (emptyc _A1))
      | Some _ -> RBT_set (emptye _A2));;

let rec set_empty (_A1, _A2)
  = function Set_Choose -> set_empty_choose (_A1, _A2)
    | Set_Monada -> Set_Monad []
    | Set_RBT -> RBT_set (emptye _A2)
    | Set_DList -> DList_set (emptyc _A1)
    | Set_Collect -> Collect_set (fun _ -> false);;

let rec fun_upda equal f aa b a = (if equal aa a then b else f a);;

let rec balance_right
  a k x xa3 = match a, k, x, xa3 with
    a, k, x, Branch (R, b, s, y, c) ->
      Branch (R, a, k, x, Branch (B, b, s, y, c))
    | Branch (B, a, k, x, b), s, y, Empty ->
        balance (Branch (R, a, k, x, b)) s y Empty
    | Branch (B, a, k, x, b), s, y, Branch (B, va, vb, vc, vd) ->
        balance (Branch (R, a, k, x, b)) s y (Branch (B, va, vb, vc, vd))
    | Branch (R, a, k, x, Branch (B, b, s, y, c)), t, z, Empty ->
        Branch (R, balance (paint R a) k x b, s, y, Branch (B, c, t, z, Empty))
    | Branch (R, a, k, x, Branch (B, b, s, y, c)), t, z,
        Branch (B, va, vb, vc, vd)
        -> Branch
             (R, balance (paint R a) k x b, s, y,
               Branch (B, c, t, z, Branch (B, va, vb, vc, vd)))
    | Empty, k, x, Empty -> Empty
    | Branch (R, va, vb, vc, Empty), k, x, Empty -> Empty
    | Branch (R, va, vb, vc, Branch (R, ve, vf, vg, vh)), k, x, Empty -> Empty
    | Empty, k, x, Branch (B, va, vb, vc, vd) -> Empty
    | Branch (R, ve, vf, vg, Empty), k, x, Branch (B, va, vb, vc, vd) -> Empty
    | Branch (R, ve, vf, vg, Branch (R, vi, vj, vk, vl)), k, x,
        Branch (B, va, vb, vc, vd)
        -> Empty;;

let rec balance_left
  x0 s y c = match x0, s, y, c with
    Branch (R, a, k, x, b), s, y, c ->
      Branch (R, Branch (B, a, k, x, b), s, y, c)
    | Empty, k, x, Branch (B, a, s, y, b) ->
        balance Empty k x (Branch (R, a, s, y, b))
    | Branch (B, va, vb, vc, vd), k, x, Branch (B, a, s, y, b) ->
        balance (Branch (B, va, vb, vc, vd)) k x (Branch (R, a, s, y, b))
    | Empty, k, x, Branch (R, Branch (B, a, s, y, b), t, z, c) ->
        Branch (R, Branch (B, Empty, k, x, a), s, y, balance b t z (paint R c))
    | Branch (B, va, vb, vc, vd), k, x,
        Branch (R, Branch (B, a, s, y, b), t, z, c)
        -> Branch
             (R, Branch (B, Branch (B, va, vb, vc, vd), k, x, a), s, y,
               balance b t z (paint R c))
    | Empty, k, x, Empty -> Empty
    | Empty, k, x, Branch (R, Empty, vb, vc, vd) -> Empty
    | Empty, k, x, Branch (R, Branch (R, ve, vf, vg, vh), vb, vc, vd) -> Empty
    | Branch (B, va, vb, vc, vd), k, x, Empty -> Empty
    | Branch (B, va, vb, vc, vd), k, x, Branch (R, Empty, vf, vg, vh) -> Empty
    | Branch (B, va, vb, vc, vd), k, x,
        Branch (R, Branch (R, vi, vj, vk, vl), vf, vg, vh)
        -> Empty;;

let rec combine
  xa0 x = match xa0, x with Empty, x -> x
    | Branch (v, va, vb, vc, vd), Empty -> Branch (v, va, vb, vc, vd)
    | Branch (R, a, k, x, b), Branch (R, c, s, y, d) ->
        (match combine b c
          with Empty -> Branch (R, a, k, x, Branch (R, Empty, s, y, d))
          | Branch (R, b2, t, z, c2) ->
            Branch (R, Branch (R, a, k, x, b2), t, z, Branch (R, c2, s, y, d))
          | Branch (B, b2, t, z, c2) ->
            Branch (R, a, k, x, Branch (R, Branch (B, b2, t, z, c2), s, y, d)))
    | Branch (B, a, k, x, b), Branch (B, c, s, y, d) ->
        (match combine b c
          with Empty -> balance_left a k x (Branch (B, Empty, s, y, d))
          | Branch (R, b2, t, z, c2) ->
            Branch (R, Branch (B, a, k, x, b2), t, z, Branch (B, c2, s, y, d))
          | Branch (B, b2, t, z, c2) ->
            balance_left a k x (Branch (B, Branch (B, b2, t, z, c2), s, y, d)))
    | Branch (B, va, vb, vc, vd), Branch (R, b, k, x, c) ->
        Branch (R, combine (Branch (B, va, vb, vc, vd)) b, k, x, c)
    | Branch (R, a, k, x, b), Branch (B, va, vb, vc, vd) ->
        Branch (R, a, k, x, combine b (Branch (B, va, vb, vc, vd)));;

let rec rbt_comp_del
  c x xa2 = match c, x, xa2 with c, x, Empty -> Empty
    | c, x, Branch (uu, a, y, s, b) ->
        (match c x y with Eq -> combine a b
          | Lt -> rbt_comp_del_from_left c x a y s b
          | Gt -> rbt_comp_del_from_right c x a y s b)
and rbt_comp_del_from_left
  c x xa2 y s b = match c, x, xa2, y, s, b with
    c, x, Branch (B, lt, z, v, rt), y, s, b ->
      balance_left (rbt_comp_del c x (Branch (B, lt, z, v, rt))) y s b
    | c, x, Empty, y, s, b -> Branch (R, rbt_comp_del c x Empty, y, s, b)
    | c, x, Branch (R, va, vb, vc, vd), y, s, b ->
        Branch (R, rbt_comp_del c x (Branch (R, va, vb, vc, vd)), y, s, b)
and rbt_comp_del_from_right
  c x a y s xa5 = match c, x, a, y, s, xa5 with
    c, x, a, y, s, Branch (B, lt, z, v, rt) ->
      balance_right a y s (rbt_comp_del c x (Branch (B, lt, z, v, rt)))
    | c, x, a, y, s, Empty -> Branch (R, a, y, s, rbt_comp_del c x Empty)
    | c, x, a, y, s, Branch (R, va, vb, vc, vd) ->
        Branch (R, a, y, s, rbt_comp_del c x (Branch (R, va, vb, vc, vd)));;

let rec rbt_comp_delete c k t = paint B (rbt_comp_del c k t);;

let rec deletea _A
  xb xc =
    Mapping_RBTa (rbt_comp_delete (the (ccompare _A)) xb (impl_ofb _A xc));;

let rec list_remove1
  equal x xa2 = match equal, x, xa2 with
    equal, x, y :: xs ->
      (if equal x y then xs else y :: list_remove1 equal x xs)
    | equal, x, [] -> [];;

let rec removea _A
  xb xc = Abs_dlist (list_remove1 (the (ceq _A)) xb (list_of_dlist _A xc));;

let rec inserta (_A1, _A2)
  xa x1 = match xa, x1 with
    xa, Complement x -> Complement (remove (_A1, _A2) xa x)
    | x, RBT_set rbt ->
        (match ccompare _A2
          with None ->
            failwith "insert RBT_set: ccompare = None"
              (fun _ -> inserta (_A1, _A2) x (RBT_set rbt))
          | Some _ -> RBT_set (inserte _A2 x () rbt))
    | x, DList_set dxs ->
        (match ceq _A1
          with None ->
            failwith "insert DList_set: ceq = None"
              (fun _ -> inserta (_A1, _A2) x (DList_set dxs))
          | Some _ -> DList_set (insertc _A1 x dxs))
    | x, Set_Monad xs -> Set_Monad (x :: xs)
    | x, Collect_set a ->
        (match ceq _A1
          with None ->
            failwith "insert Collect_set: ceq = None"
              (fun _ -> inserta (_A1, _A2) x (Collect_set a))
          | Some eq -> Collect_set (fun_upda eq a x true))
and remove (_A1, _A2)
  x xa1 = match x, xa1 with
    x, Complement a -> Complement (inserta (_A1, _A2) x a)
    | x, RBT_set rbt ->
        (match ccompare _A2
          with None ->
            failwith "remove RBT_set: ccompare = None"
              (fun _ -> remove (_A1, _A2) x (RBT_set rbt))
          | Some _ -> RBT_set (deletea _A2 x rbt))
    | x, DList_set dxs ->
        (match ceq _A1
          with None ->
            failwith "remove DList_set: ceq = None"
              (fun _ -> remove (_A1, _A2) x (DList_set dxs))
          | Some _ -> DList_set (removea _A1 x dxs))
    | x, Collect_set a ->
        (match ceq _A1
          with None ->
            failwith "remove Collect: ceq = None"
              (fun _ -> remove (_A1, _A2) x (Collect_set a))
          | Some eq -> Collect_set (fun_upda eq a x false));;

let rec foldl f a x2 = match f, a, x2 with f, a, [] -> a
                | f, a, x :: xs -> foldl f (f a x) xs;;

let rec set_aux (_A1, _A2)
  = function Set_Monada -> (fun a -> Set_Monad a)
    | Set_Choose ->
        (match ccompare _A2
          with None ->
            (match ceq _A1 with None -> (fun a -> Set_Monad a)
              | Some _ ->
                foldl (fun s x -> inserta (_A1, _A2) x s)
                  (DList_set (emptyc _A1)))
          | Some _ ->
            foldl (fun s x -> inserta (_A1, _A2) x s) (RBT_set (emptye _A2)))
    | impl ->
        foldl (fun s x -> inserta (_A1, _A2) x s) (set_empty (_A1, _A2) impl);;

let rec set (_A1, _A2, _A3)
  xs = set_aux (_A1, _A2) (of_phantom (set_impl _A3)) xs;;

let rec subseqs
  = function [] -> [[]]
    | x :: xs -> (let xss = subseqs xs in map (fun a -> x :: a) xss @ xss);;

let rec list_ex p x1 = match p, x1 with p, [] -> false
                  | p, x :: xs -> p x || list_ex p xs;;

let rec cEnum_set (_A1, _A2, _A3, _A4)
  = (match cEnum _A1 with None -> None
      | Some (enum_a, (_, _)) ->
        Some (map (set (_A2, _A3, _A4)) (subseqs enum_a),
               ((fun p ->
                  list_all p (map (set (_A2, _A3, _A4)) (subseqs enum_a))),
                 (fun p ->
                   list_ex p (map (set (_A2, _A3, _A4)) (subseqs enum_a))))));;

let rec cenum_set (_A1, _A2, _A3, _A4) =
  ({cEnum = cEnum_set (_A1, _A2, _A3, _A4)} : 'a set cenum);;

let rec finite_UNIV_seta _A = Phantom (of_phantom (finite_UNIV _A));;

let rec finite_UNIV_set _A =
  ({finite_UNIV = finite_UNIV_seta _A} : 'a set finite_UNIV);;

let rec set_less_eq_aux_Compl_fusion
  less proper_interval g1 g2 ao s1 s2 =
    (if has_next g1 s1
      then (if has_next g2 s2
             then (let (x, s1a) = next g1 s1 in
                   let (y, s2a) = next g2 s2 in
                    (if less x y
                      then proper_interval ao (Some x) ||
                             set_less_eq_aux_Compl_fusion less proper_interval
                               g1 g2 (Some x) s1a s2
                      else (if less y x
                             then proper_interval ao (Some y) ||
                                    set_less_eq_aux_Compl_fusion less
                                      proper_interval g1 g2 (Some y) s1 s2a
                             else proper_interval ao (Some y))))
             else true)
      else true);;

let rec compl_set_less_eq_aux_fusion
  less proper_interval g1 g2 ao s1 s2 =
    (if has_next g1 s1
      then (let (x, s1a) = next g1 s1 in
             (if has_next g2 s2
               then (let (y, s2a) = next g2 s2 in
                      (if less x y
                        then not (proper_interval ao (Some x)) &&
                               compl_set_less_eq_aux_fusion less proper_interval
                                 g1 g2 (Some x) s1a s2
                        else (if less y x
                               then not (proper_interval ao (Some y)) &&
                                      compl_set_less_eq_aux_fusion less
proper_interval g1 g2 (Some y) s1 s2a
                               else not (proper_interval ao (Some y)))))
               else not (proper_interval ao (Some x)) &&
                      compl_set_less_eq_aux_fusion less proper_interval g1 g2
                        (Some x) s1a s2))
      else (if has_next g2 s2
             then (let (y, s2a) = next g2 s2 in
                    not (proper_interval ao (Some y)) &&
                      compl_set_less_eq_aux_fusion less proper_interval g1 g2
                        (Some y) s1 s2a)
             else not (proper_interval ao None)));;

let rec set_less_eq_aux_Compl
  less proper_interval ao xs ys = match less, proper_interval, ao, xs, ys with
    less, proper_interval, ao, x :: xs, y :: ys ->
      (if less x y
        then proper_interval ao (Some x) ||
               set_less_eq_aux_Compl less proper_interval (Some x) xs (y :: ys)
        else (if less y x
               then proper_interval ao (Some y) ||
                      set_less_eq_aux_Compl less proper_interval (Some y)
                        (x :: xs) ys
               else proper_interval ao (Some y)))
    | less, proper_interval, ao, xs, [] -> true
    | less, proper_interval, ao, [], ys -> true;;

let rec compl_set_less_eq_aux
  less proper_interval ao x3 x4 = match less, proper_interval, ao, x3, x4 with
    less, proper_interval, ao, x :: xs, y :: ys ->
      (if less x y
        then not (proper_interval ao (Some x)) &&
               compl_set_less_eq_aux less proper_interval (Some x) xs (y :: ys)
        else (if less y x
               then not (proper_interval ao (Some y)) &&
                      compl_set_less_eq_aux less proper_interval (Some y)
                        (x :: xs) ys
               else not (proper_interval ao (Some y))))
    | less, proper_interval, ao, x :: xs, [] ->
        not (proper_interval ao (Some x)) &&
          compl_set_less_eq_aux less proper_interval (Some x) xs []
    | less, proper_interval, ao, [], y :: ys ->
        not (proper_interval ao (Some y)) &&
          compl_set_less_eq_aux less proper_interval (Some y) [] ys
    | less, proper_interval, ao, [], [] -> not (proper_interval ao None);;

let rec lexord_eq_fusion
  less g1 g2 s1 s2 =
    (if has_next g1 s1
      then has_next g2 s2 &&
             (let (x, s1a) = next g1 s1 in
              let (y, s2a) = next g2 s2 in
               less x y ||
                 not (less y x) && lexord_eq_fusion less g1 g2 s1a s2a)
      else true);;

let rec remdups_sorted
  less x1 = match less, x1 with
    less, x :: y :: xs ->
      (if less x y then x :: remdups_sorted less (y :: xs)
        else remdups_sorted less (y :: xs))
    | less, [x] -> [x]
    | less, [] -> [];;

let rec quicksort_acc
  less ac x2 = match less, ac, x2 with
    less, ac, x :: v :: va -> quicksort_part less ac x [] [] [] (v :: va)
    | less, ac, [x] -> x :: ac
    | less, ac, [] -> ac
and quicksort_part
  less ac x lts eqs gts xa6 = match less, ac, x, lts, eqs, gts, xa6 with
    less, ac, x, lts, eqs, gts, z :: zs ->
      (if less x z then quicksort_part less ac x lts eqs (z :: gts) zs
        else (if less z x then quicksort_part less ac x (z :: lts) eqs gts zs
               else quicksort_part less ac x lts (z :: eqs) gts zs))
    | less, ac, x, lts, eqs, gts, [] ->
        quicksort_acc less (eqs @ x :: quicksort_acc less ac gts) lts;;

let rec quicksort less = quicksort_acc less [];;

let rec gen_keys
  kts x1 = match kts, x1 with
    kts, Branch (c, l, k, v, r) -> gen_keys ((k, r) :: kts) l
    | (k, t) :: kts, Empty -> k :: gen_keys kts t
    | [], Empty -> [];;

let rec keys x = gen_keys [] x;;

let rec keysa _A xa = keys (impl_ofb _A xa);;

let rec csorted_list_of_set (_A1, _A2)
  = function
    Set_Monad xs ->
      (match ccompare _A2
        with None ->
          failwith "csorted_list_of_set Set_Monad: ccompare = None"
            (fun _ -> csorted_list_of_set (_A1, _A2) (Set_Monad xs))
        | Some c -> remdups_sorted (lt_of_comp c) (quicksort (lt_of_comp c) xs))
    | DList_set dxs ->
        (match ceq _A1
          with None ->
            failwith "csorted_list_of_set DList_set: ceq = None"
              (fun _ -> csorted_list_of_set (_A1, _A2) (DList_set dxs))
          | Some _ ->
            (match ccompare _A2
              with None ->
                failwith "csorted_list_of_set DList_set: ccompare = None"
                  (fun _ -> csorted_list_of_set (_A1, _A2) (DList_set dxs))
              | Some c -> quicksort (lt_of_comp c) (list_of_dlist _A1 dxs)))
    | RBT_set rbt ->
        (match ccompare _A2
          with None ->
            failwith "csorted_list_of_set RBT_set: ccompare = None"
              (fun _ -> csorted_list_of_set (_A1, _A2) (RBT_set rbt))
          | Some _ -> keysa _A2 rbt);;

let rec bot_set (_A1, _A2, _A3)
  = set_empty (_A1, _A2) (of_phantom (set_impl _A3));;

let rec top_set (_A1, _A2, _A3) = uminus_set (bot_set (_A1, _A2, _A3));;

let rec le_of_comp
  acomp x y = (match acomp x y with Eq -> true | Lt -> true | Gt -> false);;

let rec lexordp_eq
  less xs ys = match less, xs, ys with
    less, x :: xs, y :: ys ->
      less x y || not (less y x) && lexordp_eq less xs ys
    | less, x :: xs, [] -> false
    | less, xs, [] -> null xs
    | less, [], ys -> true;;

let rec finite (_A1, _A2, _A3)
  = function
    Collect_set p ->
      of_phantom (finite_UNIV _A1) ||
        failwith "finite Collect_set"
          (fun _ -> finite (_A1, _A2, _A3) (Collect_set p))
    | Set_Monad xs -> true
    | Complement a ->
        (if of_phantom (finite_UNIV _A1) then true
          else (if finite (_A1, _A2, _A3) a then false
                 else failwith "finite Complement: infinite set"
                        (fun _ -> finite (_A1, _A2, _A3) (Complement a))))
    | RBT_set rbt ->
        (match ccompare _A3
          with None ->
            failwith "finite RBT_set: ccompare = None"
              (fun _ -> finite (_A1, _A2, _A3) (RBT_set rbt))
          | Some _ -> true)
    | DList_set dxs ->
        (match ceq _A2
          with None ->
            failwith "finite DList_set: ceq = None"
              (fun _ -> finite (_A1, _A2, _A3) (DList_set dxs))
          | Some _ -> true);;

let rec set_less_aux_Compl_fusion
  less proper_interval g1 g2 ao s1 s2 =
    (if has_next g1 s1
      then (let (x, s1a) = next g1 s1 in
             (if has_next g2 s2
               then (let (y, s2a) = next g2 s2 in
                      (if less x y
                        then proper_interval ao (Some x) ||
                               set_less_aux_Compl_fusion less proper_interval g1
                                 g2 (Some x) s1a s2
                        else (if less y x
                               then proper_interval ao (Some y) ||
                                      set_less_aux_Compl_fusion less
proper_interval g1 g2 (Some y) s1 s2a
                               else proper_interval ao (Some y))))
               else proper_interval ao (Some x) ||
                      set_less_aux_Compl_fusion less proper_interval g1 g2
                        (Some x) s1a s2))
      else (if has_next g2 s2
             then (let (y, s2a) = next g2 s2 in
                    proper_interval ao (Some y) ||
                      set_less_aux_Compl_fusion less proper_interval g1 g2
                        (Some y) s1 s2a)
             else proper_interval ao None));;

let rec compl_set_less_aux_fusion
  less proper_interval g1 g2 ao s1 s2 =
    has_next g1 s1 &&
      (has_next g2 s2 &&
        (let (x, s1a) = next g1 s1 in
         let (y, s2a) = next g2 s2 in
          (if less x y
            then not (proper_interval ao (Some x)) &&
                   compl_set_less_aux_fusion less proper_interval g1 g2 (Some x)
                     s1a s2
            else (if less y x
                   then not (proper_interval ao (Some y)) &&
                          compl_set_less_aux_fusion less proper_interval g1 g2
                            (Some y) s1 s2a
                   else not (proper_interval ao (Some y))))));;

let rec set_less_aux_Compl
  less proper_interval ao x3 x4 = match less, proper_interval, ao, x3, x4 with
    less, proper_interval, ao, x :: xs, y :: ys ->
      (if less x y
        then proper_interval ao (Some x) ||
               set_less_aux_Compl less proper_interval (Some x) xs (y :: ys)
        else (if less y x
               then proper_interval ao (Some y) ||
                      set_less_aux_Compl less proper_interval (Some y) (x :: xs)
                        ys
               else proper_interval ao (Some y)))
    | less, proper_interval, ao, x :: xs, [] ->
        proper_interval ao (Some x) ||
          set_less_aux_Compl less proper_interval (Some x) xs []
    | less, proper_interval, ao, [], y :: ys ->
        proper_interval ao (Some y) ||
          set_less_aux_Compl less proper_interval (Some y) [] ys
    | less, proper_interval, ao, [], [] -> proper_interval ao None;;

let rec compl_set_less_aux
  less proper_interval ao xs ys = match less, proper_interval, ao, xs, ys with
    less, proper_interval, ao, x :: xs, y :: ys ->
      (if less x y
        then not (proper_interval ao (Some x)) &&
               compl_set_less_aux less proper_interval (Some x) xs (y :: ys)
        else (if less y x
               then not (proper_interval ao (Some y)) &&
                      compl_set_less_aux less proper_interval (Some y) (x :: xs)
                        ys
               else not (proper_interval ao (Some y))))
    | less, proper_interval, ao, xs, [] -> false
    | less, proper_interval, ao, [], ys -> false;;

let rec lexord_fusion
  less g1 g2 s1 s2 =
    (if has_next g1 s1
      then (if has_next g2 s2
             then (let (x, s1a) = next g1 s1 in
                   let (y, s2a) = next g2 s2 in
                    less x y ||
                      not (less y x) && lexord_fusion less g1 g2 s1a s2a)
             else false)
      else has_next g2 s2);;

let rec lexordp
  less xs ys = match less, xs, ys with
    less, x :: xs, y :: ys -> less x y || not (less y x) && lexordp less xs ys
    | less, xs, [] -> false
    | less, [], ys -> not (null ys);;

let rec comp_of_ords
  le lt x y = (if lt x y then Lt else (if le x y then Eq else Gt));;

let rec ccompare_seta (_A1, _A2, _A3, _A4)
  = (match ccompare _A3.ccompare_cproper_interval with None -> None
      | Some _ ->
        Some (comp_of_ords (cless_eq_set (_A1, _A2, _A3, _A4))
               (cless_set (_A1, _A2, _A3, _A4))))
and cless_set (_A1, _A2, _A3, _A4)
  a b = match a, b with
    Complement (RBT_set rbt1), RBT_set rbt2 ->
      (match ccompare _A3.ccompare_cproper_interval
        with None ->
          failwith "cless_set (Complement RBT_set) RBT_set: ccompare = None"
            (fun _ ->
              cless_set (_A1, _A2, _A3, _A4) (Complement (RBT_set rbt1))
                (RBT_set rbt2))
        | Some c ->
          finite (_A1, _A2, _A3.ccompare_cproper_interval)
            (top_set (_A2, _A3.ccompare_cproper_interval, _A4)) &&
            compl_set_less_aux_fusion (lt_of_comp c) (cproper_interval _A3)
              rbt_keys_generator rbt_keys_generator None
              (init _A3.ccompare_cproper_interval rbt1)
              (init _A3.ccompare_cproper_interval rbt2))
    | RBT_set rbt1, Complement (RBT_set rbt2) ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "cless_set RBT_set (Complement RBT_set): ccompare = None"
              (fun _ ->
                cless_set (_A1, _A2, _A3, _A4) (RBT_set rbt1)
                  (Complement (RBT_set rbt2)))
          | Some c ->
            (if finite (_A1, _A2, _A3.ccompare_cproper_interval)
                  (top_set (_A2, _A3.ccompare_cproper_interval, _A4))
              then set_less_aux_Compl_fusion (lt_of_comp c)
                     (cproper_interval _A3) rbt_keys_generator
                     rbt_keys_generator None
                     (init _A3.ccompare_cproper_interval rbt1)
                     (init _A3.ccompare_cproper_interval rbt2)
              else true))
    | RBT_set rbta, RBT_set rbt ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "cless_set RBT_set RBT_set: ccompare = None"
              (fun _ ->
                cless_set (_A1, _A2, _A3, _A4) (RBT_set rbta) (RBT_set rbt))
          | Some c ->
            lexord_fusion (fun x y -> lt_of_comp c y x) rbt_keys_generator
              rbt_keys_generator (init _A3.ccompare_cproper_interval rbta)
              (init _A3.ccompare_cproper_interval rbt))
    | Complement a, Complement b ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "cless_set Complement Complement: ccompare = None"
              (fun _ ->
                cless_set (_A1, _A2, _A3, _A4) (Complement a) (Complement b))
          | Some _ -> lt_of_comp (the (ccompare_seta (_A1, _A2, _A3, _A4))) b a)
    | Complement a, b ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "cless_set Complement1: ccompare = None"
              (fun _ -> cless_set (_A1, _A2, _A3, _A4) (Complement a) b)
          | Some c ->
            (if finite (_A1, _A2, _A3.ccompare_cproper_interval) a &&
                  finite (_A1, _A2, _A3.ccompare_cproper_interval) b
              then finite (_A1, _A2, _A3.ccompare_cproper_interval)
                     (top_set (_A2, _A3.ccompare_cproper_interval, _A4)) &&
                     compl_set_less_aux (lt_of_comp c) (cproper_interval _A3)
                       None
                       (csorted_list_of_set (_A2, _A3.ccompare_cproper_interval)
                         a)
                       (csorted_list_of_set (_A2, _A3.ccompare_cproper_interval)
                         b)
              else failwith "cless_set Complement1: infinite set"
                     (fun _ ->
                       cless_set (_A1, _A2, _A3, _A4) (Complement a) b)))
    | a, Complement b ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "cless_set Complement2: ccompare = None"
              (fun _ -> cless_set (_A1, _A2, _A3, _A4) a (Complement b))
          | Some c ->
            (if finite (_A1, _A2, _A3.ccompare_cproper_interval) a &&
                  finite (_A1, _A2, _A3.ccompare_cproper_interval) b
              then (if finite (_A1, _A2, _A3.ccompare_cproper_interval)
                         (top_set (_A2, _A3.ccompare_cproper_interval, _A4))
                     then set_less_aux_Compl (lt_of_comp c)
                            (cproper_interval _A3) None
                            (csorted_list_of_set
                              (_A2, _A3.ccompare_cproper_interval) a)
                            (csorted_list_of_set
                              (_A2, _A3.ccompare_cproper_interval) b)
                     else true)
              else failwith "cless_set Complement2: infinite set"
                     (fun _ ->
                       cless_set (_A1, _A2, _A3, _A4) a (Complement b))))
    | a, b ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "cless_set: ccompare = None"
              (fun _ -> cless_set (_A1, _A2, _A3, _A4) a b)
          | Some c ->
            (if finite (_A1, _A2, _A3.ccompare_cproper_interval) a &&
                  finite (_A1, _A2, _A3.ccompare_cproper_interval) b
              then lexordp (fun x y -> lt_of_comp c y x)
                     (csorted_list_of_set (_A2, _A3.ccompare_cproper_interval)
                       a)
                     (csorted_list_of_set (_A2, _A3.ccompare_cproper_interval)
                       b)
              else failwith "cless_set: infinite set"
                     (fun _ -> cless_set (_A1, _A2, _A3, _A4) a b)))
and cless_eq_set (_A1, _A2, _A3, _A4)
  a b = match a, b with
    Complement (RBT_set rbt1), RBT_set rbt2 ->
      (match ccompare _A3.ccompare_cproper_interval
        with None ->
          failwith "cless_eq_set (Complement RBT_set) RBT_set: ccompare = None"
            (fun _ ->
              cless_eq_set (_A1, _A2, _A3, _A4) (Complement (RBT_set rbt1))
                (RBT_set rbt2))
        | Some c ->
          finite (_A1, _A2, _A3.ccompare_cproper_interval)
            (top_set (_A2, _A3.ccompare_cproper_interval, _A4)) &&
            compl_set_less_eq_aux_fusion (lt_of_comp c) (cproper_interval _A3)
              rbt_keys_generator rbt_keys_generator None
              (init _A3.ccompare_cproper_interval rbt1)
              (init _A3.ccompare_cproper_interval rbt2))
    | RBT_set rbt1, Complement (RBT_set rbt2) ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith
              "cless_eq_set RBT_set (Complement RBT_set): ccompare = None"
              (fun _ ->
                cless_eq_set (_A1, _A2, _A3, _A4) (RBT_set rbt1)
                  (Complement (RBT_set rbt2)))
          | Some c ->
            (if finite (_A1, _A2, _A3.ccompare_cproper_interval)
                  (top_set (_A2, _A3.ccompare_cproper_interval, _A4))
              then set_less_eq_aux_Compl_fusion (lt_of_comp c)
                     (cproper_interval _A3) rbt_keys_generator
                     rbt_keys_generator None
                     (init _A3.ccompare_cproper_interval rbt1)
                     (init _A3.ccompare_cproper_interval rbt2)
              else true))
    | RBT_set rbta, RBT_set rbt ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "cless_eq_set RBT_set RBT_set: ccompare = None"
              (fun _ ->
                cless_eq_set (_A1, _A2, _A3, _A4) (RBT_set rbta) (RBT_set rbt))
          | Some c ->
            lexord_eq_fusion (fun x y -> lt_of_comp c y x) rbt_keys_generator
              rbt_keys_generator (init _A3.ccompare_cproper_interval rbta)
              (init _A3.ccompare_cproper_interval rbt))
    | Complement a, Complement b ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "cless_eq_set Complement Complement: ccompare = None"
              (fun _ ->
                le_of_comp (the (ccompare_seta (_A1, _A2, _A3, _A4)))
                  (Complement a) (Complement b))
          | Some _ -> cless_eq_set (_A1, _A2, _A3, _A4) b a)
    | Complement a, b ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "cless_eq_set Complement1: ccompare = None"
              (fun _ -> cless_eq_set (_A1, _A2, _A3, _A4) (Complement a) b)
          | Some c ->
            (if finite (_A1, _A2, _A3.ccompare_cproper_interval) a &&
                  finite (_A1, _A2, _A3.ccompare_cproper_interval) b
              then finite (_A1, _A2, _A3.ccompare_cproper_interval)
                     (top_set (_A2, _A3.ccompare_cproper_interval, _A4)) &&
                     compl_set_less_eq_aux (lt_of_comp c) (cproper_interval _A3)
                       None
                       (csorted_list_of_set (_A2, _A3.ccompare_cproper_interval)
                         a)
                       (csorted_list_of_set (_A2, _A3.ccompare_cproper_interval)
                         b)
              else failwith "cless_eq_set Complement1: infinite set"
                     (fun _ ->
                       cless_eq_set (_A1, _A2, _A3, _A4) (Complement a) b)))
    | a, Complement b ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "cless_eq_set Complement2: ccompare = None"
              (fun _ -> cless_eq_set (_A1, _A2, _A3, _A4) a (Complement b))
          | Some c ->
            (if finite (_A1, _A2, _A3.ccompare_cproper_interval) a &&
                  finite (_A1, _A2, _A3.ccompare_cproper_interval) b
              then (if finite (_A1, _A2, _A3.ccompare_cproper_interval)
                         (top_set (_A2, _A3.ccompare_cproper_interval, _A4))
                     then set_less_eq_aux_Compl (lt_of_comp c)
                            (cproper_interval _A3) None
                            (csorted_list_of_set
                              (_A2, _A3.ccompare_cproper_interval) a)
                            (csorted_list_of_set
                              (_A2, _A3.ccompare_cproper_interval) b)
                     else true)
              else failwith "cless_eq_set Complement2: infinite set"
                     (fun _ ->
                       cless_eq_set (_A1, _A2, _A3, _A4) a (Complement b))))
    | a, b ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "cless_eq_set: ccompare = None"
              (fun _ -> cless_eq_set (_A1, _A2, _A3, _A4) a b)
          | Some c ->
            (if finite (_A1, _A2, _A3.ccompare_cproper_interval) a &&
                  finite (_A1, _A2, _A3.ccompare_cproper_interval) b
              then lexordp_eq (fun x y -> lt_of_comp c y x)
                     (csorted_list_of_set (_A2, _A3.ccompare_cproper_interval)
                       a)
                     (csorted_list_of_set (_A2, _A3.ccompare_cproper_interval)
                       b)
              else failwith "cless_eq_set: infinite set"
                     (fun _ -> cless_eq_set (_A1, _A2, _A3, _A4) a b)));;

let rec ccompare_set (_A1, _A2, _A3, _A4) =
  ({ccompare = ccompare_seta (_A1, _A2, _A3, _A4)} : 'a set ccompare);;

let rec fold_fusion
  g f s b =
    (if has_next g s then (let (x, sa) = next g s in fold_fusion g f sa (f x b))
      else b);;

let rec length_last_fusion
  g s = (if has_next g s
          then (let (x, sa) = next g s in
                 fold_fusion g (fun xa (n, _) -> (plus_nata n one_nata, xa)) sa
                   (one_nata, x))
          else (zero_nata, failwith "undefined"));;

let rec gen_length_fusion
  g n s =
    (if has_next g s then gen_length_fusion g (suc n) (snd (next g s)) else n);;

let rec length_fusion g = gen_length_fusion g zero_nata;;

let rec card_UNIV _A = card_UNIVa _A;;

let rec proper_interval_set_Compl_aux_fusion _A
  less proper_interval g1 g2 ao n s1 s2 =
    (if has_next g1 s1
      then (let (x, s1a) = next g1 s1 in
             (if has_next g2 s2
               then (let (y, s2a) = next g2 s2 in
                      (if less x y
                        then proper_interval ao (Some x) ||
                               proper_interval_set_Compl_aux_fusion _A less
                                 proper_interval g1 g2 (Some x)
                                 (plus_nata n one_nata) s1a s2
                        else (if less y x
                               then proper_interval ao (Some y) ||
                                      proper_interval_set_Compl_aux_fusion _A
less proper_interval g1 g2 (Some y) (plus_nata n one_nata) s1 s2a
                               else proper_interval ao (Some x) &&
                                      (let m =
 minus_nata (of_phantom (card_UNIV _A)) n in
not (equal_nata (minus_nata m (length_fusion g2 s2a))
      (nat_of_integer (Big_int.big_int_of_int 2))) ||
  not (equal_nata (minus_nata m (length_fusion g1 s1a))
        (nat_of_integer (Big_int.big_int_of_int 2)))))))
               else (let m = minus_nata (of_phantom (card_UNIV _A)) n in
                     let (len_x, xa) = length_last_fusion g1 s1 in
                      not (equal_nata m len_x) &&
                        (if equal_nata m (plus_nata len_x one_nata)
                          then not (proper_interval (Some xa) None)
                          else true))))
      else (if has_next g2 s2
             then (let (_, _) = next g2 s2 in
                   let m = minus_nata (of_phantom (card_UNIV _A)) n in
                   let (len_y, y) = length_last_fusion g2 s2 in
                    not (equal_nata m len_y) &&
                      (if equal_nata m (plus_nata len_y one_nata)
                        then not (proper_interval (Some y) None) else true))
             else less_nat (plus_nata n one_nata)
                    (of_phantom (card_UNIV _A))));;

let rec proper_interval_Compl_set_aux_fusion
  less proper_interval g1 g2 ao s1 s2 =
    has_next g1 s1 &&
      (has_next g2 s2 &&
        (let (x, s1a) = next g1 s1 in
         let (y, s2a) = next g2 s2 in
          (if less x y
            then not (proper_interval ao (Some x)) &&
                   proper_interval_Compl_set_aux_fusion less proper_interval g1
                     g2 (Some x) s1a s2
            else (if less y x
                   then not (proper_interval ao (Some y)) &&
                          proper_interval_Compl_set_aux_fusion less
                            proper_interval g1 g2 (Some y) s1 s2a
                   else not (proper_interval ao (Some x)) &&
                          (has_next g2 s2a || has_next g1 s1a)))));;

let rec exhaustive_above_fusion
  proper_interval g y s =
    (if has_next g s
      then (let (x, sa) = next g s in
             not (proper_interval (Some y) (Some x)) &&
               exhaustive_above_fusion proper_interval g x sa)
      else not (proper_interval (Some y) None));;

let rec proper_interval_set_aux_fusion
  less proper_interval g1 g2 s1 s2 =
    has_next g2 s2 &&
      (let (y, s2a) = next g2 s2 in
        (if has_next g1 s1
          then (let (x, s1a) = next g1 s1 in
                 (if less x y then false
                   else (if less y x
                          then proper_interval (Some y) (Some x) ||
                                 (has_next g2 s2a ||
                                   not (exhaustive_above_fusion proper_interval
 g1 x s1a))
                          else proper_interval_set_aux_fusion less
                                 proper_interval g1 g2 s1a s2a)))
          else has_next g2 s2a || proper_interval (Some y) None));;

let rec length_last
  = function
    x :: xs ->
      fold (fun xa (n, _) -> (plus_nata n one_nata, xa)) xs (one_nata, x)
    | [] -> (zero_nata, failwith "undefined");;

let rec proper_interval_set_Compl_aux _A
  less proper_interval ao n x4 x5 = match less, proper_interval, ao, n, x4, x5
    with
    less, proper_interval, ao, n, x :: xs, y :: ys ->
      (if less x y
        then proper_interval ao (Some x) ||
               proper_interval_set_Compl_aux _A less proper_interval (Some x)
                 (plus_nata n one_nata) xs (y :: ys)
        else (if less y x
               then proper_interval ao (Some y) ||
                      proper_interval_set_Compl_aux _A less proper_interval
                        (Some y) (plus_nata n one_nata) (x :: xs) ys
               else proper_interval ao (Some x) &&
                      (let m = minus_nata (of_phantom (card_UNIV _A)) n in
                        not (equal_nata (minus_nata m (size_list ys))
                              (nat_of_integer (Big_int.big_int_of_int 2))) ||
                          not (equal_nata (minus_nata m (size_list xs))
                                (nat_of_integer (Big_int.big_int_of_int 2))))))
    | less, proper_interval, ao, n, x :: xs, [] ->
        (let m = minus_nata (of_phantom (card_UNIV _A)) n in
         let (len_x, xa) = length_last (x :: xs) in
          not (equal_nata m len_x) &&
            (if equal_nata m (plus_nata len_x one_nata)
              then not (proper_interval (Some xa) None) else true))
    | less, proper_interval, ao, n, [], y :: ys ->
        (let m = minus_nata (of_phantom (card_UNIV _A)) n in
         let (len_y, ya) = length_last (y :: ys) in
          not (equal_nata m len_y) &&
            (if equal_nata m (plus_nata len_y one_nata)
              then not (proper_interval (Some ya) None) else true))
    | less, proper_interval, ao, n, [], [] ->
        less_nat (plus_nata n one_nata) (of_phantom (card_UNIV _A));;

let rec proper_interval_Compl_set_aux
  less proper_interval ao uu uv = match less, proper_interval, ao, uu, uv with
    less, proper_interval, ao, uu, [] -> false
    | less, proper_interval, ao, [], uv -> false
    | less, proper_interval, ao, x :: xs, y :: ys ->
        (if less x y
          then not (proper_interval ao (Some x)) &&
                 proper_interval_Compl_set_aux less proper_interval (Some x) xs
                   (y :: ys)
          else (if less y x
                 then not (proper_interval ao (Some y)) &&
                        proper_interval_Compl_set_aux less proper_interval
                          (Some y) (x :: xs) ys
                 else not (proper_interval ao (Some x)) &&
                        (if null ys then not (null xs) else true)));;

let rec exhaustive_above
  proper_interval x xa2 = match proper_interval, x, xa2 with
    proper_interval, x, y :: ys ->
      not (proper_interval (Some x) (Some y)) &&
        exhaustive_above proper_interval y ys
    | proper_interval, x, [] -> not (proper_interval (Some x) None);;

let rec proper_interval_set_aux
  less proper_interval xs x3 = match less, proper_interval, xs, x3 with
    less, proper_interval, x :: xs, y :: ys ->
      (if less x y then false
        else (if less y x
               then proper_interval (Some y) (Some x) ||
                      (not (null ys) ||
                        not (exhaustive_above proper_interval x xs))
               else proper_interval_set_aux less proper_interval xs ys))
    | less, proper_interval, [], y :: ys ->
        not (null ys) || proper_interval (Some y) None
    | less, proper_interval, xs, [] -> false;;

let rec exhaustive_fusion
  proper_interval g s =
    has_next g s &&
      (let (x, sa) = next g s in
        not (proper_interval None (Some x)) &&
          exhaustive_above_fusion proper_interval g x sa);;

let rec list_remdups
  equal x1 = match equal, x1 with
    equal, x :: xs ->
      (if list_member equal xs x then list_remdups equal xs
        else x :: list_remdups equal xs)
    | equal, [] -> [];;

let rec lengthb _A xa = size_list (list_of_dlist _A xa);;

let rec card (_A1, _A2, _A3)
  = function
    Complement a ->
      (let aa = card (_A1, _A2, _A3) a in
       let s = of_phantom (card_UNIV _A1) in
        (if less_nat zero_nata s then minus_nata s aa
          else (if finite (_A1.finite_UNIV_card_UNIV, _A2, _A3) a then zero_nata
                 else failwith "card Complement: infinite"
                        (fun _ -> card (_A1, _A2, _A3) (Complement a)))))
    | Set_Monad xs ->
        (match ceq _A2
          with None ->
            failwith "card Set_Monad: ceq = None"
              (fun _ -> card (_A1, _A2, _A3) (Set_Monad xs))
          | Some eq -> size_list (list_remdups eq xs))
    | RBT_set rbt ->
        (match ccompare _A3
          with None ->
            failwith "card RBT_set: ccompare = None"
              (fun _ -> card (_A1, _A2, _A3) (RBT_set rbt))
          | Some _ -> size_list (keysa _A3 rbt))
    | DList_set dxs ->
        (match ceq _A2
          with None ->
            failwith "card DList_set: ceq = None"
              (fun _ -> card (_A1, _A2, _A3) (DList_set dxs))
          | Some _ -> lengthb _A2 dxs);;

let rec is_UNIV (_A1, _A2, _A3)
  = function
    RBT_set rbt ->
      (match ccompare _A3.ccompare_cproper_interval
        with None ->
          failwith "is_UNIV RBT_set: ccompare = None"
            (fun _ -> is_UNIV (_A1, _A2, _A3) (RBT_set rbt))
        | Some _ ->
          of_phantom (finite_UNIV _A1.finite_UNIV_card_UNIV) &&
            exhaustive_fusion (cproper_interval _A3) rbt_keys_generator
              (init _A3.ccompare_cproper_interval rbt))
    | a -> (let aa = of_phantom (card_UNIV _A1) in
            let b = card (_A1, _A2, _A3.ccompare_cproper_interval) a in
             (if less_nat zero_nata aa then equal_nata aa b
               else (if less_nat zero_nata b then false
                      else failwith "is_UNIV called on infinite type and set"
                             (fun _ -> is_UNIV (_A1, _A2, _A3) a))));;

let rec is_emptya _A
  xa = (match impl_ofb _A xa with Empty -> true
         | Branch (_, _, _, _, _) -> false);;

let rec nulla _A xa = null (list_of_dlist _A xa);;

let rec is_empty (_A1, _A2, _A3)
  = function Complement a -> is_UNIV (_A1, _A2, _A3) a
    | RBT_set rbt ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "is_empty RBT_set: ccompare = None"
              (fun _ -> is_empty (_A1, _A2, _A3) (RBT_set rbt))
          | Some _ -> is_emptya _A3.ccompare_cproper_interval rbt)
    | DList_set dxs ->
        (match ceq _A2
          with None ->
            failwith "is_empty DList_set: ceq = None"
              (fun _ -> is_empty (_A1, _A2, _A3) (DList_set dxs))
          | Some _ -> nulla _A2 dxs)
    | Set_Monad xs -> null xs;;

let rec cproper_interval_seta (_A1, _A2, _A3, _A4)
  x0 x1 = match x0, x1 with
    Some (Complement (RBT_set rbt1)), Some (RBT_set rbt2) ->
      (match ccompare _A3.ccompare_cproper_interval
        with None ->
          failwith
            "cproper_interval (Complement RBT_set) RBT_set: ccompare = None"
            (fun _ ->
              cproper_interval_seta (_A1, _A2, _A3, _A4)
                (Some (Complement (RBT_set rbt1))) (Some (RBT_set rbt2)))
        | Some c ->
          finite (_A1.finite_UNIV_card_UNIV, _A2, _A3.ccompare_cproper_interval)
            (top_set (_A2, _A3.ccompare_cproper_interval, _A4)) &&
            proper_interval_Compl_set_aux_fusion (lt_of_comp c)
              (cproper_interval _A3) rbt_keys_generator rbt_keys_generator None
              (init _A3.ccompare_cproper_interval rbt1)
              (init _A3.ccompare_cproper_interval rbt2))
    | Some (RBT_set rbt1), Some (Complement (RBT_set rbt2)) ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith
              "cproper_interval RBT_set (Complement RBT_set): ccompare = None"
              (fun _ ->
                cproper_interval_seta (_A1, _A2, _A3, _A4) (Some (RBT_set rbt1))
                  (Some (Complement (RBT_set rbt2))))
          | Some c ->
            finite
              (_A1.finite_UNIV_card_UNIV, _A2, _A3.ccompare_cproper_interval)
              (top_set (_A2, _A3.ccompare_cproper_interval, _A4)) &&
              proper_interval_set_Compl_aux_fusion _A1 (lt_of_comp c)
                (cproper_interval _A3) rbt_keys_generator rbt_keys_generator
                None zero_nata (init _A3.ccompare_cproper_interval rbt1)
                (init _A3.ccompare_cproper_interval rbt2))
    | Some (RBT_set rbt1), Some (RBT_set rbt2) ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "cproper_interval RBT_set RBT_set: ccompare = None"
              (fun _ ->
                cproper_interval_seta (_A1, _A2, _A3, _A4) (Some (RBT_set rbt1))
                  (Some (RBT_set rbt2)))
          | Some c ->
            finite
              (_A1.finite_UNIV_card_UNIV, _A2, _A3.ccompare_cproper_interval)
              (top_set (_A2, _A3.ccompare_cproper_interval, _A4)) &&
              proper_interval_set_aux_fusion (lt_of_comp c)
                (cproper_interval _A3) rbt_keys_generator rbt_keys_generator
                (init _A3.ccompare_cproper_interval rbt1)
                (init _A3.ccompare_cproper_interval rbt2))
    | Some (Complement a), Some (Complement b) ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "cproper_interval Complement Complement: ccompare = None"
              (fun _ ->
                cproper_interval_seta (_A1, _A2, _A3, _A4) (Some (Complement a))
                  (Some (Complement b)))
          | Some _ ->
            cproper_interval_seta (_A1, _A2, _A3, _A4) (Some b) (Some a))
    | Some (Complement a), Some b ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "cproper_interval Complement1: ccompare = None"
              (fun _ ->
                cproper_interval_seta (_A1, _A2, _A3, _A4) (Some (Complement a))
                  (Some b))
          | Some c ->
            finite
              (_A1.finite_UNIV_card_UNIV, _A2, _A3.ccompare_cproper_interval)
              (top_set (_A2, _A3.ccompare_cproper_interval, _A4)) &&
              proper_interval_Compl_set_aux (lt_of_comp c)
                (cproper_interval _A3) None
                (csorted_list_of_set (_A2, _A3.ccompare_cproper_interval) a)
                (csorted_list_of_set (_A2, _A3.ccompare_cproper_interval) b))
    | Some a, Some (Complement b) ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "cproper_interval Complement2: ccompare = None"
              (fun _ ->
                cproper_interval_seta (_A1, _A2, _A3, _A4) (Some a)
                  (Some (Complement b)))
          | Some c ->
            finite
              (_A1.finite_UNIV_card_UNIV, _A2, _A3.ccompare_cproper_interval)
              (top_set (_A2, _A3.ccompare_cproper_interval, _A4)) &&
              proper_interval_set_Compl_aux _A1 (lt_of_comp c)
                (cproper_interval _A3) None zero_nata
                (csorted_list_of_set (_A2, _A3.ccompare_cproper_interval) a)
                (csorted_list_of_set (_A2, _A3.ccompare_cproper_interval) b))
    | Some a, Some b ->
        (match ccompare _A3.ccompare_cproper_interval
          with None ->
            failwith "cproper_interval: ccompare = None"
              (fun _ ->
                cproper_interval_seta (_A1, _A2, _A3, _A4) (Some a) (Some b))
          | Some c ->
            finite
              (_A1.finite_UNIV_card_UNIV, _A2, _A3.ccompare_cproper_interval)
              (top_set (_A2, _A3.ccompare_cproper_interval, _A4)) &&
              proper_interval_set_aux (lt_of_comp c) (cproper_interval _A3)
                (csorted_list_of_set (_A2, _A3.ccompare_cproper_interval) a)
                (csorted_list_of_set (_A2, _A3.ccompare_cproper_interval) b))
    | Some a, None -> not (is_UNIV (_A1, _A2, _A3) a)
    | None, Some b -> not (is_empty (_A1, _A2, _A3) b)
    | None, None -> true;;

let rec cproper_interval_set (_A1, _A2, _A3, _A4) =
  ({ccompare_cproper_interval =
      (ccompare_set (_A1.finite_UNIV_card_UNIV, _A2, _A3, _A4));
     cproper_interval = cproper_interval_seta (_A1, _A2, _A3, _A4)}
    : 'a set cproper_interval);;

let rec equal_boola p pa = match p, pa with p, true -> p
                      | p, false -> not p
                      | true, p -> p
                      | false, p -> not p;;

let equal_bool = ({equal = equal_boola} : bool equal);;

let rec equal_lista _A
  x0 x1 = match x0, x1 with [], x21 :: x22 -> false
    | x21 :: x22, [] -> false
    | x21 :: x22, y21 :: y22 -> eq _A x21 y21 && equal_lista _A x22 y22
    | [], [] -> true;;

let rec equal_list _A = ({equal = equal_lista _A} : ('a list) equal);;

let one_lista : 'a list = [];;

let one_list = ({onea = one_lista} : ('a list) one);;

let default_lista : 'a list = [];;

type 'a default = {default : 'a};;
let default _A = _A.default;;

let default_list = ({default = default_lista} : ('a list) default);;

let rec times_lista p q = p @ q;;

let times_list = ({times = times_lista} : ('a list) times);;

let power_list =
  ({one_power = one_list; times_power = times_list} : ('a list) power);;

let rec less_eq_list (_A1, _A2)
  x0 xs = match x0, xs with
    x :: xs, y :: ys ->
      less _A2.preorder_order.ord_preorder x y ||
        eq _A1 x y && less_eq_list (_A1, _A2) xs ys
    | [], xs -> true
    | x :: xs, [] -> false;;

let rec less_list (_A1, _A2)
  xs x1 = match xs, x1 with
    x :: xs, y :: ys ->
      less _A2.preorder_order.ord_preorder x y ||
        eq _A1 x y && less_list (_A1, _A2) xs ys
    | [], x :: xs -> true
    | xs, [] -> false;;

let rec ord_list (_A1, _A2) =
  ({less_eq = less_eq_list (_A1, _A2); less = less_list (_A1, _A2)} :
    ('a list) ord);;

let rec comparator_list
  comp_a x1 x2 = match comp_a, x1, x2 with
    comp_a, x :: xa, y :: ya ->
      (match comp_a x y with Eq -> comparator_list comp_a xa ya | Lt -> Lt
        | Gt -> Gt)
    | comp_a, x :: xa, [] -> Gt
    | comp_a, [], y :: ya -> Lt
    | comp_a, [], [] -> Eq;;

let rec compare_lista _A = comparator_list (compare _A);;

let rec compare_list _A = ({compare = compare_lista _A} : ('a list) compare);;

let rec quasi_order_list (_A1, _A2) =
  ({ord_quasi_order = (ord_list (_A1, _A2))} : ('a list) quasi_order);;

let rec weak_order_list (_A1, _A2) =
  ({quasi_order_weak_order = (quasi_order_list (_A1, _A2))} :
    ('a list) weak_order);;

let rec preorder_list (_A1, _A2) =
  ({ord_preorder = (ord_list (_A1, _A2))} : ('a list) preorder);;

let rec order_list (_A1, _A2) =
  ({preorder_order = (preorder_list (_A1, _A2));
     weak_order_order = (weak_order_list (_A1, _A2))}
    : ('a list) order);;

let rec equality_list
  eq_a x1 x2 = match eq_a, x1, x2 with
    eq_a, x :: xa, y :: ya -> eq_a x y && equality_list eq_a xa ya
    | eq_a, x :: xa, [] -> false
    | eq_a, [], y :: ya -> false
    | eq_a, [], [] -> true;;

let rec ceq_lista _A
  = (match ceq _A with None -> None | Some eq_a -> Some (equality_list eq_a));;

let rec ceq_list _A = ({ceq = ceq_lista _A} : ('a list) ceq);;

type 'a infinite = unit;;

let infinite_list = (() : ('a list) infinite);;

let set_impl_lista : (('a list), set_impla) phantom = Phantom Set_Choose;;

let set_impl_list = ({set_impl = set_impl_lista} : ('a list) set_impl);;

let rec linorder_list (_A1, _A2) =
  ({order_linorder = (order_list (_A1, _A2.order_linorder))} :
    ('a list) linorder);;

type 'a countable = unit;;

let rec countable_list _A = (() : ('a list) countable);;

let rec showsl_lista _A xs = showsl_list _A xs;;

let rec showsl_list_list _A xs = default_showsl_list (showsl_lista _A) xs;;

let rec showl_list _A =
  ({showsl = showsl_lista _A; showsl_list = showsl_list_list _A} :
    ('a list) showl);;

let finite_UNIV_lista : (('a list), bool) phantom = Phantom false;;

let card_UNIV_lista : (('a list), nat) phantom = Phantom zero_nata;;

let finite_UNIV_list =
  ({finite_UNIV = finite_UNIV_lista} : ('a list) finite_UNIV);;

let card_UNIV_list =
  ({finite_UNIV_card_UNIV = finite_UNIV_list; card_UNIVa = card_UNIV_lista} :
    ('a list) card_UNIV);;

let cEnum_list :
  (('a list) list *
    ((('a list -> bool) -> bool) * (('a list -> bool) -> bool))) option
  = None;;

let cenum_list = ({cEnum = cEnum_list} : ('a list) cenum);;

let rec compare_order_list (_A1, _A2) =
  ({compare_compare_order = (compare_list _A1.compare_compare_order);
     linorder_compare_order = (linorder_list (_A2, _A1.linorder_compare_order))}
    : ('a list) compare_order);;

let rec ccompare_lista _A
  = (match ccompare _A with None -> None
      | Some comp_a -> Some (comparator_list comp_a));;

let rec ccompare_list _A =
  ({ccompare = ccompare_lista _A} : ('a list) ccompare);;

let mapping_impl_lista : (('a list), mapping_impla) phantom
  = Phantom Mapping_Choose;;

let mapping_impl_list =
  ({mapping_impl = mapping_impl_lista} : ('a list) mapping_impl);;

let rec cproper_interval_lista _A xso yso = failwith "undefined";;

let rec cproper_interval_list _A =
  ({ccompare_cproper_interval = (ccompare_list _A);
     cproper_interval = cproper_interval_lista _A}
    : ('a list) cproper_interval);;

type 'a unit_factor = {unit_factor : 'a -> 'a};;
let unit_factor _A = _A.unit_factor;;

type 'a semidom_divide_unit_factor =
  {semidom_divide_semidom_divide_unit_factor : 'a semidom_divide;
    unit_factor_semidom_divide_unit_factor : 'a unit_factor};;

type 'a normalization_semidom =
  {algebraic_semidom_normalization_semidom : 'a algebraic_semidom;
    semidom_divide_unit_factor_normalization_semidom :
      'a semidom_divide_unit_factor;
    normalizeb : 'a -> 'a};;
let normalizeb _A = _A.normalizeb;;

type 'a euclidean_semiring =
  {semidom_modulo_euclidean_semiring : 'a semidom_modulo;
    euclidean_size : 'a -> nat};;
let euclidean_size _A = _A.euclidean_size;;

type 'a factorial_semiring =
  {normalization_semidom_factorial_semiring : 'a normalization_semidom};;

type 'a normalization_euclidean_semiring =
  {euclidean_semiring_normalization_euclidean_semiring : 'a euclidean_semiring;
    factorial_semiring_normalization_euclidean_semiring :
      'a factorial_semiring};;

let rec gcdc (_A1, _A2)
  a b = (if eq _A2 b
              (zerob
                _A1.euclidean_semiring_normalization_euclidean_semiring.semidom_modulo_euclidean_semiring.algebraic_semidom_semidom_modulo.semidom_divide_algebraic_semidom.semidom_semidom_divide.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
          then normalizeb
                 _A1.factorial_semiring_normalization_euclidean_semiring.normalization_semidom_factorial_semiring
                 a
          else gcdc (_A1, _A2) b
                 (modulo
                   _A1.euclidean_semiring_normalization_euclidean_semiring.semidom_modulo_euclidean_semiring.semiring_modulo_semidom_modulo.modulo_semiring_modulo
                   a b));;

let rec lcmc (_A1, _A2)
  a b = divide
          _A1.euclidean_semiring_normalization_euclidean_semiring.semidom_modulo_euclidean_semiring.semiring_modulo_semidom_modulo.modulo_semiring_modulo.divide_modulo
          (normalizeb
            _A1.factorial_semiring_normalization_euclidean_semiring.normalization_semidom_factorial_semiring
            (times
              _A1.euclidean_semiring_normalization_euclidean_semiring.semidom_modulo_euclidean_semiring.semiring_modulo_semidom_modulo.modulo_semiring_modulo.dvd_modulo.times_dvd
              a b))
          (gcdc (_A1, _A2) a b);;

let rec euclidean_size_field (_A1, _A2)
  x = (if eq _A2 x
            (zerob
              _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
        then zero_nata else one_nata);;

let rec rep_mini_alg_unique (Abs_mini_alg_unique x) = x;;

let rec rep_mini_alg (Abs_mini_alg x) = x;;

let rec numeral _A
  = function
    Bit1 n ->
      (let m = numeral _A n in
        plus _A.semigroup_add_numeral.plus_semigroup_add
          (plus _A.semigroup_add_numeral.plus_semigroup_add m m)
          (onea _A.one_numeral))
    | Bit0 n ->
        (let m = numeral _A n in
          plus _A.semigroup_add_numeral.plus_semigroup_add m m)
    | One -> onea _A.one_numeral;;

let rec of_nat _A
  n = (if equal_nata n zero_nata
        then zerob _A.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
        else (let (m, q) =
                divmod_nat n (nat_of_integer (Big_int.big_int_of_int 2)) in
              let ma =
                times _A.semiring_numeral_semiring_1.monoid_mult_semiring_numeral.power_monoid_mult.times_power
                  (numeral
                    _A.semiring_numeral_semiring_1.numeral_semiring_numeral
                    (Bit0 One))
                  (of_nat _A m)
                in
               (if equal_nata q zero_nata then ma
                 else plus _A.semiring_numeral_semiring_1.numeral_semiring_numeral.semigroup_add_numeral.plus_semigroup_add
                        ma (onea _A.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral))));;

let rec ma_normalize
  x = (let (a, b) = x in
       let (ba, c) = b in
        (if equal_rata ba zero_rata then (a, (zero_rata, zero_nata))
          else (a, (ba, c))));;

let rec ma_inverse
  xa = Abs_mini_alg
         (let (p, (q, b)) = rep_mini_alg xa in
          let d =
            inverse_rata
              (minus_rata (times_rata p p)
                (times_rata (times_rata (of_nat semiring_1_rat b) q) q))
            in
           ma_normalize (times_rata p d, (times_rata (uminus_rata q) d, b)));;

let rec mau_inverse
  xa = Abs_mini_alg_unique (ma_inverse (rep_mini_alg_unique xa));;

let rec inverse_reala (Real_of_u r) = Real_of_u (mau_inverse r);;

let rec ma_uminus
  xa = Abs_mini_alg
         (let (p1, (q1, b1)) = rep_mini_alg xa in
           (uminus_rata p1, (uminus_rata q1, b1)));;

let rec mau_uminus
  xa = Abs_mini_alg_unique (ma_uminus (rep_mini_alg_unique xa));;

let rec uminus_reala (Real_of_u r) = Real_of_u (mau_uminus r);;

let rec ma_compatible
  xa xc =
    (let (_, (q1, b1)) = rep_mini_alg xa in
      (fun (_, (q2, b2)) ->
        equal_rata q1 zero_rata ||
          (equal_rata q2 zero_rata || equal_nata b1 b2)))
      (rep_mini_alg xc);;

let rec mau_compatible
  xa xc = ma_compatible (rep_mini_alg_unique xa) (rep_mini_alg_unique xc);;

let rec ma_times
  xb xc =
    Abs_mini_alg
      ((let (p1, (q1, b1)) = rep_mini_alg xb in
         (fun (p2, (q2, b2)) ->
           (if equal_rata q1 zero_rata
             then ma_normalize (times_rata p1 p2, (times_rata p1 q2, b2))
             else ma_normalize
                    (plus_rata (times_rata p1 p2)
                       (times_rata (times_rata (of_nat semiring_1_rat b2) q1)
                         q2),
                      (plus_rata (times_rata p1 q2) (times_rata q1 p2), b1)))))
        (rep_mini_alg xc));;

let rec mau_times
  xb xc =
    Abs_mini_alg_unique
      (ma_times (rep_mini_alg_unique xb) (rep_mini_alg_unique xc));;

let rec times_reala
  (Real_of_u r1) (Real_of_u r2) =
    (if mau_compatible r1 r2 then Real_of_u (mau_times r1 r2)
      else failwith "different base"
             (fun _ -> times_reala (Real_of_u r1) (Real_of_u r2)));;

let rec divide_reala x y = times_reala x (inverse_reala y);;

let rec ma_plus
  xb xc =
    Abs_mini_alg
      ((let (p1, (q1, b1)) = rep_mini_alg xb in
         (fun (p2, (q2, b2)) ->
           (if equal_rata q1 zero_rata then (plus_rata p1 p2, (q2, b2))
             else ma_normalize (plus_rata p1 p2, (plus_rata q1 q2, b1)))))
        (rep_mini_alg xc));;

let rec mau_plus
  xb xc =
    Abs_mini_alg_unique
      (ma_plus (rep_mini_alg_unique xb) (rep_mini_alg_unique xc));;

let rec plus_reala
  (Real_of_u r1) (Real_of_u r2) =
    (if mau_compatible r1 r2 then Real_of_u (mau_plus r1 r2)
      else failwith "different base"
             (fun _ -> plus_reala (Real_of_u r1) (Real_of_u r2)));;

let rec minus_reala x y = plus_reala x (uminus_reala y);;

let zero_reala : real = Real_of_u (mau_of_rat zero_rata);;

let one_reala : real = Real_of_u (mau_of_rat one_rata);;

let plus_real = ({plus = plus_reala} : real plus);;

let semigroup_add_real =
  ({plus_semigroup_add = plus_real} : real semigroup_add);;

let ab_semigroup_add_real =
  ({semigroup_add_ab_semigroup_add = semigroup_add_real} :
    real ab_semigroup_add);;

let zero_real = ({zerob = zero_reala} : real zero);;

let monoid_add_real =
  ({semigroup_add_monoid_add = semigroup_add_real; zero_monoid_add = zero_real}
    : real monoid_add);;

let comm_monoid_add_real =
  ({ab_semigroup_add_comm_monoid_add = ab_semigroup_add_real;
     monoid_add_comm_monoid_add = monoid_add_real}
    : real comm_monoid_add);;

let times_real = ({times = times_reala} : real times);;

let mult_zero_real =
  ({times_mult_zero = times_real; zero_mult_zero = zero_real} :
    real mult_zero);;

let semigroup_mult_real =
  ({times_semigroup_mult = times_real} : real semigroup_mult);;

let semiring_real =
  ({ab_semigroup_add_semiring = ab_semigroup_add_real;
     semigroup_mult_semiring = semigroup_mult_real}
    : real semiring);;

let semiring_0_real =
  ({comm_monoid_add_semiring_0 = comm_monoid_add_real;
     mult_zero_semiring_0 = mult_zero_real; semiring_semiring_0 = semiring_real}
    : real semiring_0);;

let semiring_no_zero_divisors_real =
  ({semiring_0_semiring_no_zero_divisors = semiring_0_real} :
    real semiring_no_zero_divisors);;

let one_real = ({onea = one_reala} : real one);;

let power_real =
  ({one_power = one_real; times_power = times_real} : real power);;

let monoid_mult_real =
  ({semigroup_mult_monoid_mult = semigroup_mult_real;
     power_monoid_mult = power_real}
    : real monoid_mult);;

let numeral_real =
  ({one_numeral = one_real; semigroup_add_numeral = semigroup_add_real} :
    real numeral);;

let semiring_numeral_real =
  ({monoid_mult_semiring_numeral = monoid_mult_real;
     numeral_semiring_numeral = numeral_real;
     semiring_semiring_numeral = semiring_real}
    : real semiring_numeral);;

let zero_neq_one_real =
  ({one_zero_neq_one = one_real; zero_zero_neq_one = zero_real} :
    real zero_neq_one);;

let semiring_1_real =
  ({semiring_numeral_semiring_1 = semiring_numeral_real;
     semiring_0_semiring_1 = semiring_0_real;
     zero_neq_one_semiring_1 = zero_neq_one_real}
    : real semiring_1);;

let semiring_1_no_zero_divisors_real =
  ({semiring_1_semiring_1_no_zero_divisors = semiring_1_real;
     semiring_no_zero_divisors_semiring_1_no_zero_divisors =
       semiring_no_zero_divisors_real}
    : real semiring_1_no_zero_divisors);;

let semiring_no_zero_divisors_cancel_real =
  ({semiring_no_zero_divisors_semiring_no_zero_divisors_cancel =
      semiring_no_zero_divisors_real}
    : real semiring_no_zero_divisors_cancel);;

let cancel_semigroup_add_real =
  ({semigroup_add_cancel_semigroup_add = semigroup_add_real} :
    real cancel_semigroup_add);;

let minus_real = ({minus = minus_reala} : real minus);;

let cancel_ab_semigroup_add_real =
  ({ab_semigroup_add_cancel_ab_semigroup_add = ab_semigroup_add_real;
     cancel_semigroup_add_cancel_ab_semigroup_add = cancel_semigroup_add_real;
     minus_cancel_ab_semigroup_add = minus_real}
    : real cancel_ab_semigroup_add);;

let cancel_comm_monoid_add_real =
  ({cancel_ab_semigroup_add_cancel_comm_monoid_add =
      cancel_ab_semigroup_add_real;
     comm_monoid_add_cancel_comm_monoid_add = comm_monoid_add_real}
    : real cancel_comm_monoid_add);;

let semiring_0_cancel_real =
  ({cancel_comm_monoid_add_semiring_0_cancel = cancel_comm_monoid_add_real;
     semiring_0_semiring_0_cancel = semiring_0_real}
    : real semiring_0_cancel);;

let uminus_real = ({uminus = uminus_reala} : real uminus);;

let group_add_real =
  ({cancel_semigroup_add_group_add = cancel_semigroup_add_real;
     minus_group_add = minus_real; monoid_add_group_add = monoid_add_real;
     uminus_group_add = uminus_real}
    : real group_add);;

let ab_group_add_real =
  ({cancel_comm_monoid_add_ab_group_add = cancel_comm_monoid_add_real;
     group_add_ab_group_add = group_add_real}
    : real ab_group_add);;

let ring_real =
  ({ab_group_add_ring = ab_group_add_real;
     semiring_0_cancel_ring = semiring_0_cancel_real}
    : real ring);;

let ring_no_zero_divisors_real =
  ({ring_ring_no_zero_divisors = ring_real;
     semiring_no_zero_divisors_cancel_ring_no_zero_divisors =
       semiring_no_zero_divisors_cancel_real}
    : real ring_no_zero_divisors);;

let semiring_1_cancel_real =
  ({semiring_0_cancel_semiring_1_cancel = semiring_0_cancel_real;
     semiring_1_semiring_1_cancel = semiring_1_real}
    : real semiring_1_cancel);;

let neg_numeral_real =
  ({group_add_neg_numeral = group_add_real; numeral_neg_numeral = numeral_real}
    : real neg_numeral);;

let ring_1_real =
  ({neg_numeral_ring_1 = neg_numeral_real; ring_ring_1 = ring_real;
     semiring_1_cancel_ring_1 = semiring_1_cancel_real}
    : real ring_1);;

let ring_1_no_zero_divisors_real =
  ({ring_1_ring_1_no_zero_divisors = ring_1_real;
     ring_no_zero_divisors_ring_1_no_zero_divisors = ring_no_zero_divisors_real;
     semiring_1_no_zero_divisors_ring_1_no_zero_divisors =
       semiring_1_no_zero_divisors_real}
    : real ring_1_no_zero_divisors);;

let divide_real = ({divide = divide_reala} : real divide);;

let inverse_real =
  ({divide_inverse = divide_real; inverse = inverse_reala} : real inverse);;

let division_ring_real =
  ({inverse_division_ring = inverse_real;
     ring_1_no_zero_divisors_division_ring = ring_1_no_zero_divisors_real}
    : real division_ring);;

let ab_semigroup_mult_real =
  ({semigroup_mult_ab_semigroup_mult = semigroup_mult_real} :
    real ab_semigroup_mult);;

let comm_semiring_real =
  ({ab_semigroup_mult_comm_semiring = ab_semigroup_mult_real;
     semiring_comm_semiring = semiring_real}
    : real comm_semiring);;

let comm_semiring_0_real =
  ({comm_semiring_comm_semiring_0 = comm_semiring_real;
     semiring_0_comm_semiring_0 = semiring_0_real}
    : real comm_semiring_0);;

let comm_semiring_0_cancel_real =
  ({comm_semiring_0_comm_semiring_0_cancel = comm_semiring_0_real;
     semiring_0_cancel_comm_semiring_0_cancel = semiring_0_cancel_real}
    : real comm_semiring_0_cancel);;

let dvd_real = ({times_dvd = times_real} : real dvd);;

let comm_monoid_mult_real =
  ({ab_semigroup_mult_comm_monoid_mult = ab_semigroup_mult_real;
     monoid_mult_comm_monoid_mult = monoid_mult_real;
     dvd_comm_monoid_mult = dvd_real}
    : real comm_monoid_mult);;

let comm_semiring_1_real =
  ({comm_monoid_mult_comm_semiring_1 = comm_monoid_mult_real;
     comm_semiring_0_comm_semiring_1 = comm_semiring_0_real;
     semiring_1_comm_semiring_1 = semiring_1_real}
    : real comm_semiring_1);;

let comm_semiring_1_cancel_real =
  ({comm_semiring_0_cancel_comm_semiring_1_cancel = comm_semiring_0_cancel_real;
     comm_semiring_1_comm_semiring_1_cancel = comm_semiring_1_real;
     semiring_1_cancel_comm_semiring_1_cancel = semiring_1_cancel_real}
    : real comm_semiring_1_cancel);;

let semidom_real =
  ({comm_semiring_1_cancel_semidom = comm_semiring_1_cancel_real;
     semiring_1_no_zero_divisors_semidom = semiring_1_no_zero_divisors_real}
    : real semidom);;

let semidom_divide_real =
  ({divide_semidom_divide = divide_real; semidom_semidom_divide = semidom_real;
     semiring_no_zero_divisors_cancel_semidom_divide =
       semiring_no_zero_divisors_cancel_real}
    : real semidom_divide);;

let comm_semiring_1_cancel_crossproduct_real =
  ({comm_semiring_1_cancel_comm_semiring_1_cancel_crossproduct =
      comm_semiring_1_cancel_real}
    : real comm_semiring_1_cancel_crossproduct);;

let comm_ring_real =
  ({comm_semiring_0_cancel_comm_ring = comm_semiring_0_cancel_real;
     ring_comm_ring = ring_real}
    : real comm_ring);;

let comm_ring_1_real =
  ({comm_ring_comm_ring_1 = comm_ring_real;
     comm_semiring_1_cancel_comm_ring_1 = comm_semiring_1_cancel_real;
     ring_1_comm_ring_1 = ring_1_real}
    : real comm_ring_1);;

let idom_real =
  ({comm_ring_1_idom = comm_ring_1_real;
     ring_1_no_zero_divisors_idom = ring_1_no_zero_divisors_real;
     semidom_idom = semidom_real;
     comm_semiring_1_cancel_crossproduct_idom =
       comm_semiring_1_cancel_crossproduct_real}
    : real idom);;

let idom_divide_real =
  ({idom_idom_divide = idom_real;
     semidom_divide_idom_divide = semidom_divide_real}
    : real idom_divide);;

let field_real =
  ({division_ring_field = division_ring_real;
     idom_divide_field = idom_divide_real}
    : real field);;

let rec equal_prod _A _B = ({equal = equal_proda _A _B} : ('a * 'b) equal);;

let rec ma_identity
  xa xc =
    equal_proda equal_rat (equal_prod equal_rat equal_nat) (rep_mini_alg xa)
      (rep_mini_alg xc);;

let rec mau_equal
  xa xc = ma_identity (rep_mini_alg_unique xa) (rep_mini_alg_unique xc);;

let rec equal_reala (Real_of_u r1) (Real_of_u r2) = mau_equal r1 r2;;

let equal_real = ({equal = equal_reala} : real equal);;

let rec euclidean_size_real
  x = euclidean_size_field (field_real, equal_real) x;;

let rec normalize_field (_A1, _A2)
  x = (if eq _A2 x
            (zerob
              _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
        then zerob _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
        else onea _A1.idom_divide_field.idom_idom_divide.comm_ring_1_idom.ring_1_comm_ring_1.neg_numeral_ring_1.numeral_neg_numeral.one_numeral);;

let rec normalize_real x = normalize_field (field_real, equal_real) x;;

let rec unit_factor_field _A x = x;;

let rec unit_factor_reala x = unit_factor_field field_real x;;

let rec mod_field (_A1, _A2)
  x y = (if eq _A2 y
              (zerob
                _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
          then x
          else zerob _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero);;

let rec modulo_reala x = mod_field (field_real, equal_real) x;;

let algebraic_semidom_real =
  ({semidom_divide_algebraic_semidom = semidom_divide_real} :
    real algebraic_semidom);;

let modulo_real =
  ({divide_modulo = divide_real; dvd_modulo = dvd_real; modulo = modulo_reala} :
    real modulo);;

let semiring_modulo_real =
  ({comm_semiring_1_cancel_semiring_modulo = comm_semiring_1_cancel_real;
     modulo_semiring_modulo = modulo_real}
    : real semiring_modulo);;

let semidom_modulo_real =
  ({algebraic_semidom_semidom_modulo = algebraic_semidom_real;
     semiring_modulo_semidom_modulo = semiring_modulo_real}
    : real semidom_modulo);;

let euclidean_semiring_real =
  ({semidom_modulo_euclidean_semiring = semidom_modulo_real;
     euclidean_size = euclidean_size_real}
    : real euclidean_semiring);;

let unit_factor_real = ({unit_factor = unit_factor_reala} : real unit_factor);;

let semidom_divide_unit_factor_real =
  ({semidom_divide_semidom_divide_unit_factor = semidom_divide_real;
     unit_factor_semidom_divide_unit_factor = unit_factor_real}
    : real semidom_divide_unit_factor);;

let normalization_semidom_real =
  ({algebraic_semidom_normalization_semidom = algebraic_semidom_real;
     semidom_divide_unit_factor_normalization_semidom =
       semidom_divide_unit_factor_real;
     normalizeb = normalize_real}
    : real normalization_semidom);;

let factorial_semiring_real =
  ({normalization_semidom_factorial_semiring = normalization_semidom_real} :
    real factorial_semiring);;

let normalization_euclidean_semiring_real =
  ({euclidean_semiring_normalization_euclidean_semiring =
      euclidean_semiring_real;
     factorial_semiring_normalization_euclidean_semiring =
       factorial_semiring_real}
    : real normalization_euclidean_semiring);;

let rec lcm_reala
  x = lcmc (normalization_euclidean_semiring_real, equal_real) x;;

let rec gcd_realc
  x = gcdc (normalization_euclidean_semiring_real, equal_real) x;;

let lcmb _ = failwith
  "Euclidean_Algorithm.normalization_euclidean_semiring_class.Lcm";;

let rec lcm_real x = lcmb x;;

let gcdb _ = failwith
  "Euclidean_Algorithm.normalization_euclidean_semiring_class.Gcd";;

let rec gcd_realb x = gcdb x;;

type 'a gcda =
  {one_gcd : 'a one; zero_gcd : 'a zero; dvd_gcd : 'a dvd;
    gcda : 'a -> 'a -> 'a; lcma : 'a -> 'a -> 'a};;
let gcda _A = _A.gcda;;
let lcma _A = _A.lcma;;

type 'a gcd = {gcd_Gcd : 'a gcda; gcd : 'a set -> 'a; lcm : 'a set -> 'a};;
let gcd _A = _A.gcd;;
let lcm _A = _A.lcm;;

let gcd_reala =
  ({one_gcd = one_real; zero_gcd = zero_real; dvd_gcd = dvd_real;
     gcda = gcd_realc; lcma = lcm_reala}
    : real gcda);;

let gcd_real =
  ({gcd_Gcd = gcd_reala; gcd = gcd_realb; lcm = lcm_real} : real gcd);;

let rec sqrt_int_maina
  x n = (let x2 = times_inta x x in
          (if less_eq_int x2 n then (x, equal_inta x2 n)
            else sqrt_int_maina
                   (divide_int (plus_inta (divide_int n x) x)
                     (Int_of_integer (Big_int.big_int_of_int 2)))
                   n));;

let rec ceiling _A
  x = uminus_inta
        (floor _A
          (uminus
            _A.archimedean_field_floor_ceiling.linordered_field_archimedean_field.linordered_idom_linordered_field.linordered_ring_strict_linordered_idom.linordered_ring_linordered_ring_strict.abs_if_linordered_ring.uminus_abs_if
            x));;

let rec power _A
  a n = (if equal_nata n zero_nata then onea _A.one_power
          else times _A.times_power a (power _A a (minus_nata n one_nata)));;

type proper_base = Abs_proper_base of int;;

let rec into_base
  xa = Abs_proper_base
         (if less_eq_int (Int_of_integer (Big_int.big_int_of_int 2)) xa then xa
           else Int_of_integer (Big_int.big_int_of_int 2));;

let rec rep_proper_base (Abs_proper_base x) = x;;

let rec square_base
  xa = Abs_proper_base (times_inta (rep_proper_base xa) (rep_proper_base xa));;

let rec get_base x = rep_proper_base x;;

let rec log_main
  b x = (if less_int x (get_base b) then (zero_nata, one_inta)
          else (let (z, bz) = log_main (square_base b) x in
                let l = times_nata (nat_of_integer (Big_int.big_int_of_int 2)) z
                  in
                let bz1 = times_inta bz (get_base b) in
                 (if less_int x bz1 then (l, bz) else (suc l, bz1))));;

let rec log_ceiling
  b x = (let (y, by) = log_main (into_base b) x in
          (if equal_inta x by then y else suc y));;

let rec start_value
  n p = power power_int (Int_of_integer (Big_int.big_int_of_int 2))
          (nat (ceiling floor_ceiling_rat
                 (divide_rata
                   (of_nat semiring_1_rat
                     (log_ceiling (Int_of_integer (Big_int.big_int_of_int 2))
                       n))
                   (of_nat semiring_1_rat p))));;

let rec sqrt_int_main
  x = sqrt_int_maina (start_value x (nat_of_integer (Big_int.big_int_of_int 2)))
        x;;

let rec sqrt_int_ceiling_pos
  x = (match sqrt_int_main x with (y, true) -> y
        | (y, false) -> plus_inta y one_inta);;

let rec sqrt_int_floor_pos x = fst (sqrt_int_main x);;

let rec int_of_nat n = Int_of_integer (integer_of_nat n);;

let rec ma_floor
  xa = (let (p, (q, b)) = rep_mini_alg xa in
        let ((z1, n1), (z2, n2)) = (quotient_of p, quotient_of q) in
        let z2n1 = times_inta z2 n1 in
        let z1n2 = times_inta z1 n2 in
        let n12 = times_inta n1 n2 in
        let prod = times_inta (times_inta z2n1 z2n1) (int_of_nat b) in
         divide_int
           (plus_inta z1n2
             (if less_eq_int zero_inta z2n1 then sqrt_int_floor_pos prod
               else uminus_inta (sqrt_int_ceiling_pos prod)))
           n12);;

let rec mau_floor xa = ma_floor (rep_mini_alg_unique xa);;

let rec floor_real (Real_of_u r) = mau_floor r;;

let rec real_lt
  x y = (let fx = floor_real x in
         let fy = floor_real y in
          (if less_int fx fy then true
            else (if less_int fy fx then false
                   else real_lt
                          (times_reala x
                            (ratreal
                              (of_int
                                (Int_of_integer
                                  (Big_int.big_int_of_int 1024)))))
                          (times_reala y
                            (ratreal
                              (of_int
                                (Int_of_integer
                                  (Big_int.big_int_of_int 1024))))))));;

let rec ma_ge_0
  xa = (let (p, (q, b)) = rep_mini_alg xa in
        let bqq = times_rata (times_rata (of_nat semiring_1_rat b) q) q in
        let pp = times_rata p p in
         less_eq_rat zero_rata p && less_eq_rat bqq pp ||
           less_eq_rat zero_rata q && less_eq_rat pp bqq);;

let rec mau_ge_0 xa = ma_ge_0 (rep_mini_alg_unique xa);;

let rec ge_0 (Real_of_u x) = mau_ge_0 x;;

let rec less_real
  x y = match x, y with
    Real_of_u x, Real_of_u y ->
      not (equal_reala (Real_of_u x) (Real_of_u y)) &&
        (if mau_compatible x y
          then ge_0 (minus_reala (Real_of_u y) (Real_of_u x))
          else real_lt (Real_of_u x) (Real_of_u y))
    | x, y -> not (equal_reala x y) && ge_0 (minus_reala y x);;

let rec abs_reala a = (if less_real a zero_reala then uminus_reala a else a);;

let abs_real = ({abs = abs_reala} : real abs);;

let rec sgn_reala
  a = (if equal_reala a zero_reala then zero_reala
        else (if less_real zero_reala a then one_reala
               else uminus_reala one_reala));;

let sgn_real = ({sgn = sgn_reala} : real sgn);;

type 'a semiring_gcd =
  {gcd_semiring_gcd : 'a gcda;
    normalization_semidom_semiring_gcd : 'a normalization_semidom};;

type 'a ring_gcd =
  {semiring_gcd_ring_gcd : 'a semiring_gcd;
    comm_ring_1_ring_gcd : 'a comm_ring_1};;

let semiring_gcd_real =
  ({gcd_semiring_gcd = gcd_reala;
     normalization_semidom_semiring_gcd = normalization_semidom_real}
    : real semiring_gcd);;

let ring_gcd_real =
  ({semiring_gcd_ring_gcd = semiring_gcd_real;
     comm_ring_1_ring_gcd = comm_ring_1_real}
    : real ring_gcd);;

let rec less_eq_real
  x y = match x, y with
    Real_of_u x, Real_of_u y ->
      equal_reala (Real_of_u x) (Real_of_u y) ||
        (if mau_compatible x y
          then ge_0 (minus_reala (Real_of_u y) (Real_of_u x))
          else real_lt (Real_of_u x) (Real_of_u y))
    | x, y -> ge_0 (minus_reala y x);;

let ord_real = ({less_eq = less_eq_real; less = less_real} : real ord);;

let abs_if_real =
  ({abs_abs_if = abs_real; minus_abs_if = minus_real;
     uminus_abs_if = uminus_real; zero_abs_if = zero_real;
     ord_abs_if = ord_real}
    : real abs_if);;

let semiring_char_0_real =
  ({semiring_1_semiring_char_0 = semiring_1_real} : real semiring_char_0);;

let ring_char_0_real =
  ({semiring_char_0_ring_char_0 = semiring_char_0_real;
     ring_1_ring_char_0 = ring_1_real}
    : real ring_char_0);;

let quasi_order_real = ({ord_quasi_order = ord_real} : real quasi_order);;

let weak_order_real =
  ({quasi_order_weak_order = quasi_order_real} : real weak_order);;

let preorder_real = ({ord_preorder = ord_real} : real preorder);;

let order_real =
  ({preorder_order = preorder_real; weak_order_order = weak_order_real} :
    real order);;

type 'a semiring_Gcd =
  {gcd_semiring_Gcd : 'a gcd; semiring_gcd_semiring_Gcd : 'a semiring_gcd};;

let semiring_Gcd_real =
  ({gcd_semiring_Gcd = gcd_real; semiring_gcd_semiring_Gcd = semiring_gcd_real}
    : real semiring_Gcd);;

let no_bot_real = ({order_no_bot = order_real} : real no_bot);;

let no_top_real = ({order_no_top = order_real} : real no_top);;

let ceq_reala : (real -> real -> bool) option = Some equal_reala;;

let ceq_real = ({ceq = ceq_reala} : real ceq);;

let set_impl_reala : (real, set_impla) phantom = Phantom Set_RBT;;

let set_impl_real = ({set_impl = set_impl_reala} : real set_impl);;

let linorder_real = ({order_linorder = order_real} : real linorder);;

let idom_abs_sgn_real =
  ({abs_idom_abs_sgn = abs_real; sgn_idom_abs_sgn = sgn_real;
     idom_idom_abs_sgn = idom_real}
    : real idom_abs_sgn);;

let ordered_ab_semigroup_add_real =
  ({ab_semigroup_add_ordered_ab_semigroup_add = ab_semigroup_add_real;
     order_ordered_ab_semigroup_add = order_real}
    : real ordered_ab_semigroup_add);;

let strict_ordered_ab_semigroup_add_real =
  ({ordered_ab_semigroup_add_strict_ordered_ab_semigroup_add =
      ordered_ab_semigroup_add_real}
    : real strict_ordered_ab_semigroup_add);;

let ordered_cancel_ab_semigroup_add_real =
  ({cancel_ab_semigroup_add_ordered_cancel_ab_semigroup_add =
      cancel_ab_semigroup_add_real;
     strict_ordered_ab_semigroup_add_ordered_cancel_ab_semigroup_add =
       strict_ordered_ab_semigroup_add_real}
    : real ordered_cancel_ab_semigroup_add);;

let ordered_semigroup_mult_zero_real =
  ({semigroup_mult_ordered_semigroup_mult_zero = semigroup_mult_real;
     order_ordered_semigroup_mult_zero = order_real;
     mult_zero_ordered_semigroup_mult_zero = mult_zero_real}
    : real ordered_semigroup_mult_zero);;

let ordered_comm_monoid_add_real =
  ({comm_monoid_add_ordered_comm_monoid_add = comm_monoid_add_real;
     ordered_ab_semigroup_add_ordered_comm_monoid_add =
       ordered_ab_semigroup_add_real}
    : real ordered_comm_monoid_add);;

let ordered_semiring_real =
  ({ordered_comm_monoid_add_ordered_semiring = ordered_comm_monoid_add_real;
     semiring_ordered_semiring = semiring_real}
    : real ordered_semiring);;

let ordered_semiring_0_real =
  ({ordered_semigroup_mult_zero_ordered_semiring_0 =
      ordered_semigroup_mult_zero_real;
     ordered_semiring_ordered_semiring_0 = ordered_semiring_real;
     semiring_0_ordered_semiring_0 = semiring_0_real}
    : real ordered_semiring_0);;

let ordered_cancel_semiring_real =
  ({ordered_cancel_ab_semigroup_add_ordered_cancel_semiring =
      ordered_cancel_ab_semigroup_add_real;
     ordered_semiring_0_ordered_cancel_semiring = ordered_semiring_0_real;
     semiring_0_cancel_ordered_cancel_semiring = semiring_0_cancel_real}
    : real ordered_cancel_semiring);;

let ordered_ab_semigroup_add_imp_le_real =
  ({ordered_cancel_ab_semigroup_add_ordered_ab_semigroup_add_imp_le =
      ordered_cancel_ab_semigroup_add_real}
    : real ordered_ab_semigroup_add_imp_le);;

let strict_ordered_comm_monoid_add_real =
  ({comm_monoid_add_strict_ordered_comm_monoid_add = comm_monoid_add_real;
     strict_ordered_ab_semigroup_add_strict_ordered_comm_monoid_add =
       strict_ordered_ab_semigroup_add_real}
    : real strict_ordered_comm_monoid_add);;

let ordered_cancel_comm_monoid_add_real =
  ({ordered_cancel_ab_semigroup_add_ordered_cancel_comm_monoid_add =
      ordered_cancel_ab_semigroup_add_real;
     ordered_comm_monoid_add_ordered_cancel_comm_monoid_add =
       ordered_comm_monoid_add_real;
     strict_ordered_comm_monoid_add_ordered_cancel_comm_monoid_add =
       strict_ordered_comm_monoid_add_real}
    : real ordered_cancel_comm_monoid_add);;

let ordered_ab_semigroup_monoid_add_imp_le_real =
  ({cancel_comm_monoid_add_ordered_ab_semigroup_monoid_add_imp_le =
      cancel_comm_monoid_add_real;
     ordered_ab_semigroup_add_imp_le_ordered_ab_semigroup_monoid_add_imp_le =
       ordered_ab_semigroup_add_imp_le_real;
     ordered_cancel_comm_monoid_add_ordered_ab_semigroup_monoid_add_imp_le =
       ordered_cancel_comm_monoid_add_real}
    : real ordered_ab_semigroup_monoid_add_imp_le);;

let ordered_ab_group_add_real =
  ({ab_group_add_ordered_ab_group_add = ab_group_add_real;
     ordered_ab_semigroup_monoid_add_imp_le_ordered_ab_group_add =
       ordered_ab_semigroup_monoid_add_imp_le_real}
    : real ordered_ab_group_add);;

let ordered_ring_real =
  ({ordered_ab_group_add_ordered_ring = ordered_ab_group_add_real;
     ordered_cancel_semiring_ordered_ring = ordered_cancel_semiring_real;
     ring_ordered_ring = ring_real}
    : real ordered_ring);;

let field_char_0_real =
  ({field_field_char_0 = field_real;
     ring_char_0_field_char_0 = ring_char_0_real}
    : real field_char_0);;

let zero_less_one_real =
  ({order_zero_less_one = order_real;
     zero_neq_one_zero_less_one = zero_neq_one_real}
    : real zero_less_one);;

let rec string_of_digit
  n = (if equal_nata n zero_nata
        then [Chara (false, false, false, false, true, true, false, false)]
        else (if equal_nata n one_nata
               then [Chara (true, false, false, false, true, true, false,
                             false)]
               else (if equal_nata n (nat_of_integer (Big_int.big_int_of_int 2))
                      then [Chara (false, true, false, false, true, true, false,
                                    false)]
                      else (if equal_nata n
                                 (nat_of_integer (Big_int.big_int_of_int 3))
                             then [Chara (true, true, false, false, true, true,
   false, false)]
                             else (if equal_nata n
(nat_of_integer (Big_int.big_int_of_int 4))
                                    then [Chara
    (false, false, true, false, true, true, false, false)]
                                    else (if equal_nata n
       (nat_of_integer (Big_int.big_int_of_int 5))
   then [Chara (true, false, true, false, true, true, false, false)]
   else (if equal_nata n (nat_of_integer (Big_int.big_int_of_int 6))
          then [Chara (false, true, true, false, true, true, false, false)]
          else (if equal_nata n (nat_of_integer (Big_int.big_int_of_int 7))
                 then [Chara (true, true, true, false, true, true, false,
                               false)]
                 else (if equal_nata n
                            (nat_of_integer (Big_int.big_int_of_int 8))
                        then [Chara (false, false, false, true, true, true,
                                      false, false)]
                        else [Chara (true, false, false, true, true, true,
                                      false, false)])))))))));;

let rec shows_string x = (fun a -> x @ a);;

let rec showsp_nat
  p n = (if less_nat n (nat_of_integer (Big_int.big_int_of_int 10))
          then shows_string (string_of_digit n)
          else comp (showsp_nat p
                      (divide_nata n
                        (nat_of_integer (Big_int.big_int_of_int 10))))
                 (shows_string
                   (string_of_digit
                     (modulo_nata n
                       (nat_of_integer (Big_int.big_int_of_int 10))))));;

let rec showsp_int
  p i = (if less_int i zero_inta
          then comp (shows_string
                      [Chara (true, false, true, true, false, true, false,
                               false)])
                 (showsp_nat p (nat (uminus_inta i)))
          else showsp_nat p (nat i));;

let rec showsp_rat
  p x = (let (d, n) = quotient_of x in
          (if equal_inta n one_inta then showsp_int p d
            else comp (comp (showsp_int p d)
                        (shows_string
                          [Chara (true, true, true, true, false, true, false,
                                   false)]))
                   (showsp_int p n)));;

let rec shows_prec_rat x = showsp_rat x;;

let rec shows_prec_nat x = showsp_nat x;;

type 'a show =
  {shows_prec : nat -> 'a -> char list -> char list;
    shows_list : 'a list -> char list -> char list};;
let shows_prec _A = _A.shows_prec;;
let shows_list _A = _A.shows_list;;

let rec shows_prec_list _A p xs = shows_list _A xs;;

let rec shows_prec_char p c = (fun a -> c :: a);;

let rec shows_list_char cs = shows_string cs;;

let show_char =
  ({shows_prec = shows_prec_char; shows_list = shows_list_char} : char show);;

let rec ma_show_real
  xa = (let (p, (q, b)) = rep_mini_alg xa in
        let sb =
          comp (comp (shows_prec_list show_char zero_nata
                       [Chara (true, true, false, false, true, true, true,
                                false);
                         Chara (true, false, false, false, true, true, true,
                                 false);
                         Chara (false, true, false, false, true, true, true,
                                 false);
                         Chara (false, false, true, false, true, true, true,
                                 false);
                         Chara (false, false, false, true, false, true, false,
                                 false)])
                 (shows_prec_nat zero_nata b))
            (shows_prec_list show_char zero_nata
              [Chara (true, false, false, true, false, true, false, false)])
          in
        let qb =
          (if equal_rata q one_rata then sb
            else (if equal_rata q (uminus_rata one_rata)
                   then comp (shows_prec_list show_char zero_nata
                               [Chara (true, false, true, true, false, true,
false, false)])
                          sb
                   else comp (comp (shows_prec_rat zero_nata q)
                               (shows_prec_list show_char zero_nata
                                 [Chara (false, true, false, true, false, true,
  false, false)]))
                          sb))
          in
         (if equal_rata q zero_rata then shows_prec_rat zero_nata p []
           else (if equal_rata p zero_rata then qb []
                  else (if less_rat q zero_rata
                         then shows_prec_rat zero_nata p (qb [])
                         else shows_prec_rat zero_nata p
                                (shows_prec_list show_char zero_nata
                                  [Chara (true, true, false, true, false, true,
   false, false)]
                                  (qb []))))));;

let rec mau_show_real xa = ma_show_real (rep_mini_alg_unique xa);;

let rec show_real (Real_of_u x) = mau_show_real x;;

let rec showsl_real x = showsl_lit (implode (show_real x));;

let rec showsl_list_real xs = default_showsl_list showsl_real xs;;

let showl_real =
  ({showsl = showsl_real; showsl_list = showsl_list_real} : real showl);;

let field_abs_sgn_real =
  ({field_field_abs_sgn = field_real;
     idom_abs_sgn_field_abs_sgn = idom_abs_sgn_real}
    : real field_abs_sgn);;

let cEnum_real :
  (real list * (((real -> bool) -> bool) * ((real -> bool) -> bool))) option
  = None;;

let cenum_real = ({cEnum = cEnum_real} : real cenum);;

let dense_order_real = ({order_dense_order = order_real} : real dense_order);;

let linordered_ab_semigroup_add_real =
  ({ordered_ab_semigroup_add_linordered_ab_semigroup_add =
      ordered_ab_semigroup_add_real;
     linorder_linordered_ab_semigroup_add = linorder_real}
    : real linordered_ab_semigroup_add);;

let linordered_cancel_ab_semigroup_add_real =
  ({linordered_ab_semigroup_add_linordered_cancel_ab_semigroup_add =
      linordered_ab_semigroup_add_real;
     ordered_ab_semigroup_add_imp_le_linordered_cancel_ab_semigroup_add =
       ordered_ab_semigroup_add_imp_le_real}
    : real linordered_cancel_ab_semigroup_add);;

let linordered_semiring_real =
  ({linordered_cancel_ab_semigroup_add_linordered_semiring =
      linordered_cancel_ab_semigroup_add_real;
     ordered_ab_semigroup_monoid_add_imp_le_linordered_semiring =
       ordered_ab_semigroup_monoid_add_imp_le_real;
     ordered_cancel_semiring_linordered_semiring = ordered_cancel_semiring_real}
    : real linordered_semiring);;

let linordered_semiring_strict_real =
  ({linordered_semiring_linordered_semiring_strict = linordered_semiring_real} :
    real linordered_semiring_strict);;

let linordered_semiring_1_real =
  ({linordered_semiring_linordered_semiring_1 = linordered_semiring_real;
     semiring_1_linordered_semiring_1 = semiring_1_real;
     zero_less_one_linordered_semiring_1 = zero_less_one_real}
    : real linordered_semiring_1);;

let linordered_semiring_1_strict_real =
  ({linordered_semiring_1_linordered_semiring_1_strict =
      linordered_semiring_1_real;
     linordered_semiring_strict_linordered_semiring_1_strict =
       linordered_semiring_strict_real}
    : real linordered_semiring_1_strict);;

let ordered_ab_group_add_abs_real =
  ({abs_ordered_ab_group_add_abs = abs_real;
     ordered_ab_group_add_ordered_ab_group_add_abs = ordered_ab_group_add_real}
    : real ordered_ab_group_add_abs);;

let linordered_ab_group_add_real =
  ({linordered_cancel_ab_semigroup_add_linordered_ab_group_add =
      linordered_cancel_ab_semigroup_add_real;
     ordered_ab_group_add_linordered_ab_group_add = ordered_ab_group_add_real}
    : real linordered_ab_group_add);;

let linordered_ring_real =
  ({linordered_ab_group_add_linordered_ring = linordered_ab_group_add_real;
     ordered_ab_group_add_abs_linordered_ring = ordered_ab_group_add_abs_real;
     abs_if_linordered_ring = abs_if_real;
     linordered_semiring_linordered_ring = linordered_semiring_real;
     ordered_ring_linordered_ring = ordered_ring_real}
    : real linordered_ring);;

let linordered_ring_strict_real =
  ({linordered_ring_linordered_ring_strict = linordered_ring_real;
     linordered_semiring_strict_linordered_ring_strict =
       linordered_semiring_strict_real;
     ring_no_zero_divisors_linordered_ring_strict = ring_no_zero_divisors_real}
    : real linordered_ring_strict);;

let ordered_comm_semiring_real =
  ({comm_semiring_0_ordered_comm_semiring = comm_semiring_0_real;
     ordered_semiring_ordered_comm_semiring = ordered_semiring_real}
    : real ordered_comm_semiring);;

let ordered_cancel_comm_semiring_real =
  ({comm_semiring_0_cancel_ordered_cancel_comm_semiring =
      comm_semiring_0_cancel_real;
     ordered_cancel_semiring_ordered_cancel_comm_semiring =
       ordered_cancel_semiring_real;
     ordered_comm_semiring_ordered_cancel_comm_semiring =
       ordered_comm_semiring_real}
    : real ordered_cancel_comm_semiring);;

let linordered_comm_semiring_strict_real =
  ({linordered_semiring_strict_linordered_comm_semiring_strict =
      linordered_semiring_strict_real;
     ordered_cancel_comm_semiring_linordered_comm_semiring_strict =
       ordered_cancel_comm_semiring_real}
    : real linordered_comm_semiring_strict);;

let linordered_nonzero_semiring_real =
  ({semiring_char_0_linordered_nonzero_semiring = semiring_char_0_real;
     linorder_linordered_nonzero_semiring = linorder_real;
     comm_semiring_1_linordered_nonzero_semiring = comm_semiring_1_real;
     ordered_comm_semiring_linordered_nonzero_semiring =
       ordered_comm_semiring_real;
     zero_less_one_linordered_nonzero_semiring = zero_less_one_real}
    : real linordered_nonzero_semiring);;

let linordered_semidom_real =
  ({linordered_comm_semiring_strict_linordered_semidom =
      linordered_comm_semiring_strict_real;
     linordered_nonzero_semiring_linordered_semidom =
       linordered_nonzero_semiring_real;
     semidom_linordered_semidom = semidom_real}
    : real linordered_semidom);;

let ordered_comm_ring_real =
  ({comm_ring_ordered_comm_ring = comm_ring_real;
     ordered_cancel_comm_semiring_ordered_comm_ring =
       ordered_cancel_comm_semiring_real;
     ordered_ring_ordered_comm_ring = ordered_ring_real}
    : real ordered_comm_ring);;

let ordered_ring_abs_real =
  ({ordered_ab_group_add_abs_ordered_ring_abs = ordered_ab_group_add_abs_real;
     ordered_ring_ordered_ring_abs = ordered_ring_real}
    : real ordered_ring_abs);;

let linordered_idom_real =
  ({ring_char_0_linordered_idom = ring_char_0_real;
     idom_abs_sgn_linordered_idom = idom_abs_sgn_real;
     linordered_ring_strict_linordered_idom = linordered_ring_strict_real;
     linordered_semidom_linordered_idom = linordered_semidom_real;
     linordered_semiring_1_strict_linordered_idom =
       linordered_semiring_1_strict_real;
     ordered_comm_ring_linordered_idom = ordered_comm_ring_real;
     ordered_ring_abs_linordered_idom = ordered_ring_abs_real}
    : real linordered_idom);;

let non_strict_order_real =
  ({ord_non_strict_order = ord_real} : real non_strict_order);;

let ordered_ab_semigroup_real =
  ({ab_semigroup_add_ordered_ab_semigroup = ab_semigroup_add_real;
     monoid_add_ordered_ab_semigroup = monoid_add_real;
     non_strict_order_ordered_ab_semigroup = non_strict_order_real}
    : real ordered_ab_semigroup);;

let ordered_semiring_0_reala =
  ({semiring_0_ordered_semiring_0a = semiring_0_real;
     ordered_ab_semigroup_ordered_semiring_0 = ordered_ab_semigroup_real}
    : real ordered_semiring_0a);;

let ordered_semiring_1_real =
  ({semiring_1_ordered_semiring_1 = semiring_1_real;
     ordered_semiring_0_ordered_semiring_1 = ordered_semiring_0_reala}
    : real ordered_semiring_1);;

let poly_carrier_real =
  ({comm_semiring_1_poly_carrier = comm_semiring_1_real;
     ordered_semiring_1_poly_carrier = ordered_semiring_1_real}
    : real poly_carrier);;

let dense_linorder_real =
  ({dense_order_dense_linorder = dense_order_real;
     linorder_dense_linorder = linorder_real}
    : real dense_linorder);;

let unbounded_dense_linorder_real =
  ({dense_linorder_unbounded_dense_linorder = dense_linorder_real;
     no_bot_unbounded_dense_linorder = no_bot_real;
     no_top_unbounded_dense_linorder = no_top_real}
    : real unbounded_dense_linorder);;

let linordered_field_real =
  ({field_abs_sgn_linordered_field = field_abs_sgn_real;
     field_char_0_linordered_field = field_char_0_real;
     unbounded_dense_linorder_linordered_field = unbounded_dense_linorder_real;
     linordered_idom_linordered_field = linordered_idom_real}
    : real linordered_field);;

let rec dist_reala x y = abs_reala (minus_reala x y);;

type 'a dist = {dist : 'a -> 'a -> real};;
let dist _A = _A.dist;;

let dist_real = ({dist = dist_reala} : real dist);;

let rec norm_reala r = abs_reala r;;

type 'a norm = {norm : 'a -> real};;
let norm _A = _A.norm;;

let norm_real = ({norm = norm_reala} : real norm);;

let open_reala _ = failwith "Real_Vector_Spaces.open_real_inst.open_real";;

type 'a opena = {opena : 'a set -> bool};;
let opena _A = _A.opena;;

let open_real = ({opena = open_reala} : real opena);;

let rec compare_real x = comparator_of (equal_real, linorder_real) x;;

let ccompare_reala : (real -> real -> ordera) option = Some compare_real;;

let ccompare_real = ({ccompare = ccompare_reala} : real ccompare);;

let rec scaleR_reala a x = times_reala a x;;

type 'a scaleR = {scaleR : real -> 'a -> 'a};;
let scaleR _A = _A.scaleR;;

let scaleR_real = ({scaleR = scaleR_reala} : real scaleR);;

type 'a topological_space = {open_topological_space : 'a opena};;

type 'a t0_space = {topological_space_t0_space : 'a topological_space};;

let topological_space_real =
  ({open_topological_space = open_real} : real topological_space);;

let t0_space_real =
  ({topological_space_t0_space = topological_space_real} : real t0_space);;

type 'a t1_space = {t0_space_t1_space : 'a t0_space};;

let t1_space_real = ({t0_space_t1_space = t0_space_real} : real t1_space);;

type 'a t2_space = {t1_space_t2_space : 'a t1_space};;

let t2_space_real = ({t1_space_t2_space = t1_space_real} : real t2_space);;

type 'a t3_space = {t2_space_t3_space : 'a t2_space};;

let t3_space_real = ({t2_space_t3_space = t2_space_real} : real t3_space);;

type 'a t4_space = {t3_space_t4_space : 'a t3_space};;

let t4_space_real = ({t3_space_t4_space = t3_space_real} : real t4_space);;

type 'a topological_monoid_add =
  {monoid_add_topological_monoid_add : 'a monoid_add;
    topological_space_topological_monoid_add : 'a topological_space};;

type 'a topological_group_add =
  {group_add_topological_group_add : 'a group_add;
    topological_monoid_add_topological_group_add : 'a topological_monoid_add};;

let topological_monoid_add_real =
  ({monoid_add_topological_monoid_add = monoid_add_real;
     topological_space_topological_monoid_add = topological_space_real}
    : real topological_monoid_add);;

let topological_group_add_real =
  ({group_add_topological_group_add = group_add_real;
     topological_monoid_add_topological_group_add = topological_monoid_add_real}
    : real topological_group_add);;

type 'a dist_norm =
  {minus_dist_norm : 'a minus; dist_dist_norm : 'a dist;
    norm_dist_norm : 'a norm};;

let dist_norm_real =
  ({minus_dist_norm = minus_real; dist_dist_norm = dist_real;
     norm_dist_norm = norm_real}
    : real dist_norm);;

type 'a filter = Principal of 'a set | Abstract_filter of (unit -> 'a filter);;

let rec uniformity_reala
  () = Abstract_filter
         (fun _ ->
           failwith "uniformity is not executable"
             (fun _ -> uniformity_reala ()));;
let uniformity_reala = uniformity_reala ();;

type 'a uniformity = {uniformity : ('a * 'a) filter};;
let uniformity _A = _A.uniformity;;

let uniformity_real = ({uniformity = uniformity_reala} : real uniformity);;

type 'a real_vector =
  {ab_group_add_real_vector : 'a ab_group_add; scaleR_real_vector : 'a scaleR};;

let real_vector_real =
  ({ab_group_add_real_vector = ab_group_add_real;
     scaleR_real_vector = scaleR_real}
    : real real_vector);;

let rec real_of_real x = x;;

let real_embedding_real =
  ({linordered_idom_real_embedding = linordered_idom_real;
     real_of = real_of_real}
    : real real_embedding);;

let archimedean_field_real =
  ({linordered_field_archimedean_field = linordered_field_real} :
    real archimedean_field);;

let large_ordered_semiring_1_real =
  ({poly_carrier_large_ordered_semiring_1 = poly_carrier_real} :
    real large_ordered_semiring_1);;

let floor_ceiling_real =
  ({archimedean_field_floor_ceiling = archimedean_field_real;
     large_ordered_semiring_1_floor_ceiling = large_ordered_semiring_1_real;
     floor = floor_real}
    : real floor_ceiling);;

type 'a topological_comm_monoid_add =
  {comm_monoid_add_topological_comm_monoid_add : 'a comm_monoid_add;
    topological_monoid_add_topological_comm_monoid_add :
      'a topological_monoid_add};;

type 'a topological_ab_group_add =
  {ab_group_add_topological_ab_group_add : 'a ab_group_add;
    topological_comm_monoid_add_topological_ab_group_add :
      'a topological_comm_monoid_add;
    topological_group_add_topological_ab_group_add : 'a topological_group_add};;

let topological_comm_monoid_add_real =
  ({comm_monoid_add_topological_comm_monoid_add = comm_monoid_add_real;
     topological_monoid_add_topological_comm_monoid_add =
       topological_monoid_add_real}
    : real topological_comm_monoid_add);;

let topological_ab_group_add_real =
  ({ab_group_add_topological_ab_group_add = ab_group_add_real;
     topological_comm_monoid_add_topological_ab_group_add =
       topological_comm_monoid_add_real;
     topological_group_add_topological_ab_group_add =
       topological_group_add_real}
    : real topological_ab_group_add);;

type 'a first_countable_topology =
  {topological_space_first_countable_topology : 'a topological_space};;

type 'a uniformity_dist =
  {dist_uniformity_dist : 'a dist; uniformity_uniformity_dist : 'a uniformity};;

type 'a open_uniformity =
  {topological_space_open_uniformity : 'a topological_space;
    uniformity_open_uniformity : 'a uniformity};;

type 'a uniform_space = {open_uniformity_uniform_space : 'a open_uniformity};;

type 'a metric_space =
  {uniformity_dist_metric_space : 'a uniformity_dist;
    first_countable_topology_metric_space : 'a first_countable_topology;
    t4_space_metric_space : 'a t4_space;
    uniform_space_metric_space : 'a uniform_space};;

let first_countable_topology_real =
  ({topological_space_first_countable_topology = topological_space_real} :
    real first_countable_topology);;

let uniformity_dist_real =
  ({dist_uniformity_dist = dist_real;
     uniformity_uniformity_dist = uniformity_real}
    : real uniformity_dist);;

let open_uniformity_real =
  ({topological_space_open_uniformity = topological_space_real;
     uniformity_open_uniformity = uniformity_real}
    : real open_uniformity);;

let uniform_space_real =
  ({open_uniformity_uniform_space = open_uniformity_real} :
    real uniform_space);;

let metric_space_real =
  ({uniformity_dist_metric_space = uniformity_dist_real;
     first_countable_topology_metric_space = first_countable_topology_real;
     t4_space_metric_space = t4_space_real;
     uniform_space_metric_space = uniform_space_real}
    : real metric_space);;

type 'a sgn_div_norm =
  {sgn_sgn_div_norm : 'a sgn; norm_sgn_div_norm : 'a norm;
    scaleR_sgn_div_norm : 'a scaleR};;

let sgn_div_norm_real =
  ({sgn_sgn_div_norm = sgn_real; norm_sgn_div_norm = norm_real;
     scaleR_sgn_div_norm = scaleR_real}
    : real sgn_div_norm);;

type 'a factorial_semiring_gcd =
  {factorial_semiring_factorial_semiring_gcd : 'a factorial_semiring;
    semiring_Gcd_factorial_semiring_gcd : 'a semiring_Gcd};;

type 'a factorial_ring_gcd =
  {factorial_semiring_gcd_factorial_ring_gcd : 'a factorial_semiring_gcd;
    ring_gcd_factorial_ring_gcd : 'a ring_gcd;
    idom_divide_factorial_ring_gcd : 'a idom_divide};;

let factorial_semiring_gcd_real =
  ({factorial_semiring_factorial_semiring_gcd = factorial_semiring_real;
     semiring_Gcd_factorial_semiring_gcd = semiring_Gcd_real}
    : real factorial_semiring_gcd);;

let factorial_ring_gcd_real =
  ({factorial_semiring_gcd_factorial_ring_gcd = factorial_semiring_gcd_real;
     ring_gcd_factorial_ring_gcd = ring_gcd_real;
     idom_divide_factorial_ring_gcd = idom_divide_real}
    : real factorial_ring_gcd);;

type 'a real_normed_vector =
  {topological_ab_group_add_real_normed_vector : 'a topological_ab_group_add;
    dist_norm_real_normed_vector : 'a dist_norm;
    metric_space_real_normed_vector : 'a metric_space;
    real_vector_real_normed_vector : 'a real_vector;
    sgn_div_norm_real_normed_vector : 'a sgn_div_norm};;

let real_normed_vector_real =
  ({topological_ab_group_add_real_normed_vector = topological_ab_group_add_real;
     dist_norm_real_normed_vector = dist_norm_real;
     metric_space_real_normed_vector = metric_space_real;
     real_vector_real_normed_vector = real_vector_real;
     sgn_div_norm_real_normed_vector = sgn_div_norm_real}
    : real real_normed_vector);;

let large_real_ordered_semiring_1_real =
  ({real_embedding_large_real_ordered_semiring_1 = real_embedding_real;
     large_ordered_semiring_1_large_real_ordered_semiring_1 =
       large_ordered_semiring_1_real}
    : real large_real_ordered_semiring_1);;

type ('a, 'b) term = Var of 'b | Fun of 'a * ('a, 'b) term list;;

let rec equal_term _A _B = ({equal = equal_terma _A _B} : ('a, 'b) term equal)
and equal_terma _A _B
  x0 x1 = match x0, x1 with Var x1, Fun (x21, x22) -> false
    | Fun (x21, x22), Var x1 -> false
    | Fun (x21, x22), Fun (y21, y22) ->
        eq _A x21 y21 && equal_lista (equal_term _A _B) x22 y22
    | Var x1, Var y1 -> eq _B x1 y1;;

let rec comparator_term
  comp_f comp_v x2 x3 = match comp_f, comp_v, x2, x3 with
    comp_f, comp_v, Fun (x, xa), Fun (ya, yb) ->
      (match comp_f x ya
        with Eq -> comparator_list (comparator_term comp_f comp_v) xa yb
        | Lt -> Lt | Gt -> Gt)
    | comp_f, comp_v, Fun (x, xa), Var y -> Gt
    | comp_f, comp_v, Var x, Fun (ya, yb) -> Lt
    | comp_f, comp_v, Var x, Var y -> comp_v x y;;

let rec compare_terma _A _B = comparator_term (compare _A) (compare _B);;

let rec less_eq_term _A _B = le_of_comp (compare_terma _A _B);;

let rec less_term _A _B = lt_of_comp (compare_terma _A _B);;

let rec ord_term _A _B =
  ({less_eq = less_eq_term _A _B; less = less_term _A _B} : ('a, 'b) term ord);;

let rec compare_term _A _B =
  ({compare = compare_terma _A _B} : ('a, 'b) term compare);;

let rec quasi_order_term _A _B =
  ({ord_quasi_order = (ord_term _A _B)} : ('a, 'b) term quasi_order);;

let rec weak_order_term _A _B =
  ({quasi_order_weak_order = (quasi_order_term _A _B)} :
    ('a, 'b) term weak_order);;

let rec preorder_term _A _B =
  ({ord_preorder = (ord_term _A _B)} : ('a, 'b) term preorder);;

let rec order_term _A _B =
  ({preorder_order = (preorder_term _A _B);
     weak_order_order = (weak_order_term _A _B)}
    : ('a, 'b) term order);;

let rec ceq_terma _A _B = Some (equal_terma _A _B);;

let rec ceq_term _A _B = ({ceq = ceq_terma _A _B} : ('a, 'b) term ceq);;

let set_impl_terma : (('a, 'b) term, set_impla) phantom = Phantom Set_RBT;;

let set_impl_term = ({set_impl = set_impl_terma} : ('a, 'b) term set_impl);;

let rec linorder_term _A _B =
  ({order_linorder = (order_term _A _B)} : ('a, 'b) term linorder);;

let rec showsl_term
  funa var x2 = match funa, var, x2 with funa, var, Var x -> var x
    | funa, var, Fun (f, ts) ->
        comp (funa f)
          (showsl_list_gen id "[]" "(" ", " ")"
            (map (showsl_term funa var) ts));;

let rec showsl_terma _A _B t = showsl_term (showsl _A) (showsl _B) t;;

let rec showsl_list_term _A _B
  xs = default_showsl_list (showsl_terma _A _B) xs;;

let rec showl_term _A _B =
  ({showsl = showsl_terma _A _B; showsl_list = showsl_list_term _A _B} :
    ('a, 'b) term showl);;

let cEnum_term :
  (('a, 'b) term list *
    (((('a, 'b) term -> bool) -> bool) *
      ((('a, 'b) term -> bool) -> bool))) option
  = None;;

let cenum_term = ({cEnum = cEnum_term} : ('a, 'b) term cenum);;

let rec compare_order_term _A _B =
  ({compare_compare_order = (compare_term _A _B);
     linorder_compare_order = (linorder_term _A _B)}
    : ('a, 'b) term compare_order);;

let finite_UNIV_terma : (('a, 'b) term, bool) phantom = Phantom false;;

let finite_UNIV_term =
  ({finite_UNIV = finite_UNIV_terma} : ('a, 'b) term finite_UNIV);;

let rec ccompare_terma _A _B = Some (compare_terma _A _B);;

let rec ccompare_term _A _B =
  ({ccompare = ccompare_terma _A _B} : ('a, 'b) term ccompare);;

let rec cproper_interval_terma _A _B = (fun _ _ -> false);;

let rec cproper_interval_term _A _B =
  ({ccompare_cproper_interval = (ccompare_term _A _B);
     cproper_interval = cproper_interval_terma _A _B}
    : ('a, 'b) term cproper_interval);;

type 'a iarray = IArray of 'a list;;

type 'a mat_impl = Abs_mat_impl of (nat * (nat * 'a iarray iarray));;

let rec rep_mat_impl (Abs_mat_impl x) = x;;

let rec lengtha (IArray asa) = integer_of_nat (size_list asa);;

let rec length asa = nat_of_integer (lengtha asa);;

let rec nth
  (x :: xs) n =
    (if equal_nata n zero_nata then x else nth xs (minus_nata n one_nata));;

let rec suba (IArray asa, n) = nth asa (nat_of_integer n);;

let rec sub asa n = suba (asa, integer_of_nat n);;

let rec upt i j = (if less_nat i j then i :: upt (suc i) j else []);;

let rec list_of asa = map (sub asa) (upt zero_nata (length asa));;

let rec equal_iarraya _A asa bs = equal_lista _A (list_of asa) (list_of bs);;

let rec equal_iarray _A = ({equal = equal_iarraya _A} : 'a iarray equal);;

let rec mat_equal_impl _A
  xa xc =
    (let (nr1, (nc1, m1)) = rep_mat_impl xa in
      (fun (nr2, (nc2, m2)) ->
        equal_nata nr1 nr2 &&
          (equal_nata nc1 nc2 && equal_iarraya (equal_iarray _A) m1 m2)))
      (rep_mat_impl xc);;

type 'a mat = Mat_impl of 'a mat_impl;;

let rec equal_mata _A (Mat_impl m1) (Mat_impl m2) = mat_equal_impl _A m1 m2;;

let rec equal_mat _A = ({equal = equal_mata _A} : 'a mat equal);;

let rec ceq_mata _A = Some (equal_mata _A);;

let rec ceq_mat _A = ({ceq = ceq_mata _A} : 'a mat ceq);;

let set_impl_mata : ('a mat, set_impla) phantom = Phantom Set_DList;;

let set_impl_mat = ({set_impl = set_impl_mata} : 'a mat set_impl);;

let rec index_mat_impl
  xa = (let (nr, (_, m)) = rep_mat_impl xa in
         (fun (i, j) ->
           (if less_nat i nr then sub (sub m i) j
             else sub (IArray (nth [] (minus_nata i nr))) j)));;

let rec index_mat (Mat_impl m) ij = index_mat_impl m ij;;

let rec dim_row_impl xa = fst (rep_mat_impl xa);;

let rec dim_row (Mat_impl m) = dim_row_impl m;;

let rec dim_col_impl xa = fst (snd (rep_mat_impl xa));;

let rec dim_col (Mat_impl m) = dim_col_impl m;;

let rec mat_to_list
  a = map (fun i ->
            map (fun j -> index_mat a (i, j)) (upt zero_nata (dim_col a)))
        (upt zero_nata (dim_row a));;

let rec showsl_mat _A
  a = default_showsl_list id (map (showsl_list _A) (mat_to_list a));;

let rec showsl_list_mat _A xs = default_showsl_list (showsl_mat _A) xs;;

let rec showl_mat _A =
  ({showsl = showsl_mat _A; showsl_list = showsl_list_mat _A} : 'a mat showl);;

let cEnum_mat :
  ('a mat list *
    ((('a mat -> bool) -> bool) * (('a mat -> bool) -> bool))) option
  = None;;

let cenum_mat = ({cEnum = cEnum_mat} : 'a mat cenum);;

let ccompare_mata : ('a mat -> 'a mat -> ordera) option = None;;

let ccompare_mat = ({ccompare = ccompare_mata} : 'a mat ccompare);;

let rec equal_chara
  (Chara (x1, x2, x3, x4, x5, x6, x7, x8))
    (Chara (y1, y2, y3, y4, y5, y6, y7, y8)) =
    equal_boola x1 y1 &&
      (equal_boola x2 y2 &&
        (equal_boola x3 y3 &&
          (equal_boola x4 y4 &&
            (equal_boola x5 y5 &&
              (equal_boola x6 y6 &&
                (equal_boola x7 y7 && equal_boola x8 y8))))));;

let equal_char = ({equal = equal_chara} : char equal);;

let rec nat_of_char c = Nat (integer_of_char c);;

let rec less_eq_char c1 c2 = less_eq_nat (nat_of_char c1) (nat_of_char c2);;

let rec less_char c1 c2 = less_nat (nat_of_char c1) (nat_of_char c2);;

let ord_char = ({less_eq = less_eq_char; less = less_char} : char ord);;

let quasi_order_char = ({ord_quasi_order = ord_char} : char quasi_order);;

let weak_order_char =
  ({quasi_order_weak_order = quasi_order_char} : char weak_order);;

let preorder_char = ({ord_preorder = ord_char} : char preorder);;

let order_char =
  ({preorder_order = preorder_char; weak_order_order = weak_order_char} :
    char order);;

let linorder_char = ({order_linorder = order_char} : char linorder);;

let rec compare_chara x = comparator_of (equal_char, linorder_char) x;;

let compare_char = ({compare = compare_chara} : char compare);;

let ceq_chara : (char -> char -> bool) option = Some equal_chara;;

let ceq_char = ({ceq = ceq_chara} : char ceq);;

let countable_char = (() : char countable);;

let rec showsl_list_char cs s = showsl_lit (implode cs) s;;

let rec showsl_char c = showsl_lit (implode [c]);;

let showl_char =
  ({showsl = showsl_char; showsl_list = showsl_list_char} : char showl);;

let compare_order_char =
  ({compare_compare_order = compare_char;
     linorder_compare_order = linorder_char}
    : char compare_order);;

let ccompare_chara : (char -> char -> ordera) option = Some compare_chara;;

let ccompare_char = ({ccompare = ccompare_chara} : char ccompare);;

type 'a hint = Default | Base of 'a | Distribute of nat * 'a hint list |
  Erase of nat * 'a hint | LexStrict of 'a hint list | LexWeak of 'a hint list;;

let rec default_hinta _A = Default;;

let rec default_hint _A = ({default = default_hinta _A} : 'a hint default);;

type 'a atom = Leq of nat * 'a | Geq of nat * 'a;;

let rec equal_atom _A
  x0 x1 = match x0, x1 with Leq (x11, x12), Geq (x21, x22) -> false
    | Geq (x21, x22), Leq (x11, x12) -> false
    | Geq (x21, x22), Geq (y21, y22) -> equal_nata x21 y21 && eq _A x22 y22
    | Leq (x11, x12), Leq (y11, y12) -> equal_nata x11 y11 && eq _A x12 y12;;

let rec ceq_atoma _A = Some (equal_atom _A);;

let rec ceq_atom _A = ({ceq = ceq_atoma _A} : 'a atom ceq);;

let set_impl_atoma : ('a atom, set_impla) phantom = Phantom Set_RBT;;

let set_impl_atom = ({set_impl = set_impl_atoma} : 'a atom set_impl);;

let rec comparator_atom
  comp_a x1 x2 = match comp_a, x1, x2 with
    comp_a, Geq (x, xa), Geq (yb, yc) ->
      (match comparator_of (equal_nat, linorder_nat) x yb
        with Eq -> comp_a xa yc | Lt -> Lt | Gt -> Gt)
    | comp_a, Geq (x, xa), Leq (y, ya) -> Gt
    | comp_a, Leq (x, xa), Geq (yb, yc) -> Lt
    | comp_a, Leq (x, xa), Leq (y, ya) ->
        (match comparator_of (equal_nat, linorder_nat) x y
          with Eq -> comp_a xa ya | Lt -> Lt | Gt -> Gt);;

let rec ccompare_atoma _A
  = (match ccompare _A with None -> None
      | Some comp_a -> Some (comparator_atom comp_a));;

let rec ccompare_atom _A = ({ccompare = ccompare_atoma _A} : 'a atom ccompare);;

type ('a, 'b) sum = Inl of 'a | Inr of 'b;;

let rec equal_suma _A _B x0 x1 = match x0, x1 with Inl x1, Inr x2 -> false
                           | Inr x2, Inl x1 -> false
                           | Inr x2, Inr y2 -> eq _B x2 y2
                           | Inl x1, Inl y1 -> eq _A x1 y1;;

let rec equal_sum _A _B = ({equal = equal_suma _A _B} : ('a, 'b) sum equal);;

let rec showsl_sum _A _B
  = function
    Inl x -> comp (comp (showsl_lit "Inl (") (showsl _A x)) (showsl_lit ")")
    | Inr x -> comp (comp (showsl_lit "Inr (") (showsl _B x)) (showsl_lit ")");;

let rec showsl_list_sum _A _B xs = default_showsl_list (showsl_sum _A _B) xs;;

let rec showl_sum _A _B =
  ({showsl = showsl_sum _A _B; showsl_list = showsl_list_sum _A _B} :
    ('a, 'b) sum showl);;

type 'a trans_var = Pre of 'a | Post of 'a | Intermediate of 'a;;

let rec equal_trans_vara _A
  x0 x1 = match x0, x1 with Post x2, Intermediate x3 -> false
    | Intermediate x3, Post x2 -> false
    | Pre x1, Intermediate x3 -> false
    | Intermediate x3, Pre x1 -> false
    | Pre x1, Post x2 -> false
    | Post x2, Pre x1 -> false
    | Intermediate x3, Intermediate y3 -> eq _A x3 y3
    | Post x2, Post y2 -> eq _A x2 y2
    | Pre x1, Pre y1 -> eq _A x1 y1;;

let rec equal_trans_var _A =
  ({equal = equal_trans_vara _A} : 'a trans_var equal);;

let rec comparator_trans_var
  comp_v x1 x2 = match comp_v, x1, x2 with
    comp_v, Intermediate x, Intermediate yb -> comp_v x yb
    | comp_v, Intermediate x, Post ya -> Gt
    | comp_v, Intermediate x, Pre y -> Gt
    | comp_v, Post x, Intermediate yb -> Lt
    | comp_v, Post x, Post ya -> comp_v x ya
    | comp_v, Post x, Pre y -> Gt
    | comp_v, Pre x, Intermediate yb -> Lt
    | comp_v, Pre x, Post ya -> Lt
    | comp_v, Pre x, Pre y -> comp_v x y;;

let rec compare_trans_var _A = comparator_trans_var (compare _A);;

let rec less_eq_trans_var _A = le_of_comp (compare_trans_var _A);;

let rec less_trans_var _A = lt_of_comp (compare_trans_var _A);;

let rec ord_trans_var _A =
  ({less_eq = less_eq_trans_var _A; less = less_trans_var _A} :
    'a trans_var ord);;

let rec quasi_order_trans_var _A =
  ({ord_quasi_order = (ord_trans_var _A)} : 'a trans_var quasi_order);;

let rec weak_order_trans_var _A =
  ({quasi_order_weak_order = (quasi_order_trans_var _A)} :
    'a trans_var weak_order);;

let rec preorder_trans_var _A =
  ({ord_preorder = (ord_trans_var _A)} : 'a trans_var preorder);;

let rec order_trans_var _A =
  ({preorder_order = (preorder_trans_var _A);
     weak_order_order = (weak_order_trans_var _A)}
    : 'a trans_var order);;

let rec linorder_trans_var _A =
  ({order_linorder = (order_trans_var _A)} : 'a trans_var linorder);;

let rec showsl_trans_var _A
  = function Pre v -> showsl _A v
    | Post v -> comp (showsl _A v) (showsl_lit "\039")
    | Intermediate v -> comp (showsl _A v) (showsl_lit "#");;

let rec showsl_list_trans_var _A
  xs = default_showsl_list (showsl_trans_var _A) xs;;

let rec showl_trans_var _A =
  ({showsl = showsl_trans_var _A; showsl_list = showsl_list_trans_var _A} :
    'a trans_var showl);;

let rec ccompare_trans_vara _A
  = (match ccompare _A with None -> None
      | Some comp_v -> Some (comparator_trans_var comp_v));;

let rec ccompare_trans_var _A =
  ({ccompare = ccompare_trans_vara _A} : 'a trans_var ccompare);;

let mapping_impl_trans_vara : ('a trans_var, mapping_impla) phantom
  = Phantom Mapping_RBT;;

let mapping_impl_trans_var =
  ({mapping_impl = mapping_impl_trans_vara} : 'a trans_var mapping_impl);;

type ('a, 'b) lab = Lab of ('a, 'b) lab * 'b |
  FunLab of ('a, 'b) lab * ('a, 'b) lab list | UnLab of 'a |
  Sharp of ('a, 'b) lab;;

let rec equal_lab _A _B = ({equal = equal_laba _A _B} : ('a, 'b) lab equal)
and equal_laba _A _B
  x0 x1 = match x0, x1 with UnLab x3, Sharp x4 -> false
    | Sharp x4, UnLab x3 -> false
    | FunLab (x21, x22), Sharp x4 -> false
    | Sharp x4, FunLab (x21, x22) -> false
    | FunLab (x21, x22), UnLab x3 -> false
    | UnLab x3, FunLab (x21, x22) -> false
    | Lab (x11, x12), Sharp x4 -> false
    | Sharp x4, Lab (x11, x12) -> false
    | Lab (x11, x12), UnLab x3 -> false
    | UnLab x3, Lab (x11, x12) -> false
    | Lab (x11, x12), FunLab (x21, x22) -> false
    | FunLab (x21, x22), Lab (x11, x12) -> false
    | Sharp x4, Sharp y4 -> equal_laba _A _B x4 y4
    | UnLab x3, UnLab y3 -> eq _A x3 y3
    | FunLab (x21, x22), FunLab (y21, y22) ->
        equal_laba _A _B x21 y21 && equal_lista (equal_lab _A _B) x22 y22
    | Lab (x11, x12), Lab (y11, y12) ->
        equal_laba _A _B x11 y11 && eq _B x12 y12;;

let rec default_laba _A = UnLab (default _A);;

let rec default_lab _A = ({default = default_laba _A} : ('a, 'b) lab default);;

let rec comparator_lab
  comp_f comp_l x2 x3 = match comp_f, comp_l, x2, x3 with
    comp_f, comp_l, Sharp x, Sharp ye -> comparator_lab comp_f comp_l x ye
    | comp_f, comp_l, Sharp x, UnLab yd -> Gt
    | comp_f, comp_l, Sharp x, FunLab (yb, yc) -> Gt
    | comp_f, comp_l, Sharp x, Lab (y, ya) -> Gt
    | comp_f, comp_l, UnLab x, Sharp ye -> Lt
    | comp_f, comp_l, UnLab x, UnLab yd -> comp_f x yd
    | comp_f, comp_l, UnLab x, FunLab (yb, yc) -> Gt
    | comp_f, comp_l, UnLab x, Lab (y, ya) -> Gt
    | comp_f, comp_l, FunLab (x, xa), Sharp ye -> Lt
    | comp_f, comp_l, FunLab (x, xa), UnLab yd -> Lt
    | comp_f, comp_l, FunLab (x, xa), FunLab (yb, yc) ->
        (match comparator_lab comp_f comp_l x yb
          with Eq -> comparator_list (comparator_lab comp_f comp_l) xa yc
          | Lt -> Lt | Gt -> Gt)
    | comp_f, comp_l, FunLab (x, xa), Lab (y, ya) -> Gt
    | comp_f, comp_l, Lab (x, xa), Sharp ye -> Lt
    | comp_f, comp_l, Lab (x, xa), UnLab yd -> Lt
    | comp_f, comp_l, Lab (x, xa), FunLab (yb, yc) -> Lt
    | comp_f, comp_l, Lab (x, xa), Lab (y, ya) ->
        (match comparator_lab comp_f comp_l x y with Eq -> comp_l xa ya
          | Lt -> Lt | Gt -> Gt);;

let rec compare_laba _A _B = comparator_lab (compare _A) (compare _B);;

let rec less_eq_lab _A _B = le_of_comp (compare_laba _A _B);;

let rec less_lab _A _B = lt_of_comp (compare_laba _A _B);;

let rec ord_lab _A _B =
  ({less_eq = less_eq_lab _A _B; less = less_lab _A _B} : ('a, 'b) lab ord);;

let rec compare_lab _A _B =
  ({compare = compare_laba _A _B} : ('a, 'b) lab compare);;

let rec quasi_order_lab _A _B =
  ({ord_quasi_order = (ord_lab _A _B)} : ('a, 'b) lab quasi_order);;

let rec weak_order_lab _A _B =
  ({quasi_order_weak_order = (quasi_order_lab _A _B)} :
    ('a, 'b) lab weak_order);;

let rec preorder_lab _A _B =
  ({ord_preorder = (ord_lab _A _B)} : ('a, 'b) lab preorder);;

let rec order_lab _A _B =
  ({preorder_order = (preorder_lab _A _B);
     weak_order_order = (weak_order_lab _A _B)}
    : ('a, 'b) lab order);;

let rec ceq_laba _A _B = Some (equal_laba _A _B);;

let rec ceq_lab _A _B = ({ceq = ceq_laba _A _B} : ('a, 'b) lab ceq);;

let set_impl_laba : (('a, 'b) lab, set_impla) phantom = Phantom Set_RBT;;

let set_impl_lab = ({set_impl = set_impl_laba} : ('a, 'b) lab set_impl);;

let rec linorder_lab _A _B =
  ({order_linorder = (order_lab _A _B)} : ('a, 'b) lab linorder);;

let rec showsl_lab _A _B
  = function UnLab f -> showsl _A f
    | Lab (f, l) ->
        comp (comp (comp (showsl_lab _A _B f) (showsl_lit "[")) (showsl _B l))
          (showsl_lit "]")
    | Sharp f -> comp (showsl_lab _A _B f) (showsl_lit "#")
    | FunLab (f, l) ->
        comp (showsl_lab _A _B f)
          (default_showsl_list id (map (showsl_lab _A _B) l));;

let rec showsl_list_lab _A _B xs = default_showsl_list (showsl_lab _A _B) xs;;

let rec showl_lab _A _B =
  ({showsl = showsl_lab _A _B; showsl_list = showsl_list_lab _A _B} :
    ('a, 'b) lab showl);;

let finite_UNIV_laba : (('a, 'b) lab, bool) phantom = Phantom false;;

let card_UNIV_laba : (('a, 'b) lab, nat) phantom = Phantom zero_nata;;

let finite_UNIV_lab =
  ({finite_UNIV = finite_UNIV_laba} : ('a, 'b) lab finite_UNIV);;

let card_UNIV_lab =
  ({finite_UNIV_card_UNIV = finite_UNIV_lab; card_UNIVa = card_UNIV_laba} :
    ('a, 'b) lab card_UNIV);;

let cEnum_lab :
  (('a, 'b) lab list *
    (((('a, 'b) lab -> bool) -> bool) *
      ((('a, 'b) lab -> bool) -> bool))) option
  = None;;

let cenum_lab = ({cEnum = cEnum_lab} : ('a, 'b) lab cenum);;

let rec compare_order_lab _A _B =
  ({compare_compare_order = (compare_lab _A _B);
     linorder_compare_order = (linorder_lab _A _B)}
    : ('a, 'b) lab compare_order);;

let rec ccompare_laba _A _B = Some (compare_laba _A _B);;

let rec ccompare_lab _A _B =
  ({ccompare = ccompare_laba _A _B} : ('a, 'b) lab ccompare);;

let mapping_impl_laba : (('a, 'b) lab, mapping_impla) phantom
  = Phantom Mapping_RBT;;

let mapping_impl_lab =
  ({mapping_impl = mapping_impl_laba} : ('a, 'b) lab mapping_impl);;

let rec cproper_interval_laba _A _B = (fun _ _ -> false);;

let rec cproper_interval_lab _A _B =
  ({ccompare_cproper_interval = (ccompare_lab _A _B);
     cproper_interval = cproper_interval_laba _A _B}
    : ('a, 'b) lab cproper_interval);;

let rec equal_option _A = ({equal = equal_optiona _A} : ('a option) equal);;

let rec equality_option
  eq_a x1 x2 = match eq_a, x1, x2 with eq_a, Some x, Some y -> eq_a x y
    | eq_a, Some x, None -> false
    | eq_a, None, Some y -> false
    | eq_a, None, None -> true;;

let rec ceq_optiona _A
  = (match ceq _A with None -> None
      | Some eq_a -> Some (equality_option eq_a));;

let rec ceq_option _A = ({ceq = ceq_optiona _A} : ('a option) ceq);;

let rec set_impl_optiona _A = Phantom (of_phantom (set_impl _A));;

let rec set_impl_option _A =
  ({set_impl = set_impl_optiona _A} : ('a option) set_impl);;

let rec cEnum_option _A
  = (match cEnum _A with None -> None
      | Some (enum_a, (enum_all_a, enum_ex_a)) ->
        Some (None :: map (fun a -> Some a) enum_a,
               ((fun p -> p None && enum_all_a (fun x -> p (Some x))),
                 (fun p -> p None || enum_ex_a (fun x -> p (Some x))))));;

let rec cenum_option _A = ({cEnum = cEnum_option _A} : ('a option) cenum);;

let rec finite_UNIV_optiona _A = Phantom (of_phantom (finite_UNIV _A));;

let rec finite_UNIV_option _A =
  ({finite_UNIV = finite_UNIV_optiona _A} : ('a option) finite_UNIV);;

let rec comparator_option
  comp_a x1 x2 = match comp_a, x1, x2 with comp_a, Some x, Some y -> comp_a x y
    | comp_a, Some x, None -> Gt
    | comp_a, None, Some y -> Lt
    | comp_a, None, None -> Eq;;

let rec ccompare_optiona _A
  = (match ccompare _A with None -> None
      | Some comp_a -> Some (comparator_option comp_a));;

let rec ccompare_option _A =
  ({ccompare = ccompare_optiona _A} : ('a option) ccompare);;

let rec mapping_impl_optiona _A = Phantom (of_phantom (mapping_impl _A));;

let rec mapping_impl_option _A =
  ({mapping_impl = mapping_impl_optiona _A} : ('a option) mapping_impl);;

let rec cproper_interval_optiona _A
  x0 x1 = match x0, x1 with None, None -> true
    | None, Some x -> not (is_none x)
    | Some x, None -> cproper_interval _A x None
    | Some x, Some None -> false
    | Some x, Some (Some y) -> cproper_interval _A x (Some y);;

let rec cproper_interval_option _A =
  ({ccompare_cproper_interval = (ccompare_option _A.ccompare_cproper_interval);
     cproper_interval = cproper_interval_optiona _A}
    : ('a option) cproper_interval);;

type qDelta = QDelta of rat * rat;;

let rec equal_QDeltaa
  (QDelta (x1, x2)) (QDelta (y1, y2)) = equal_rata x1 y1 && equal_rata x2 y2;;

let equal_QDelta = ({equal = equal_QDeltaa} : qDelta equal);;

let one_QDeltaa : qDelta = QDelta (one_rata, zero_rata);;

let one_QDelta = ({onea = one_QDeltaa} : qDelta one);;

let rec qdsnd (QDelta (a, b)) = b;;

let rec qdfst (QDelta (a, b)) = a;;

let rec plus_QDeltaa
  qd1 qd2 =
    QDelta
      (plus_rata (qdfst qd1) (qdfst qd2), plus_rata (qdsnd qd1) (qdsnd qd2));;

let plus_QDelta = ({plus = plus_QDeltaa} : qDelta plus);;

let zero_QDeltaa : qDelta = QDelta (zero_rata, zero_rata);;

let zero_QDelta = ({zerob = zero_QDeltaa} : qDelta zero);;

let rec minus_QDeltaa
  qd1 qd2 =
    QDelta
      (minus_rata (qdfst qd1) (qdfst qd2), minus_rata (qdsnd qd1) (qdsnd qd2));;

let minus_QDelta = ({minus = minus_QDeltaa} : qDelta minus);;

let rec uminus_QDeltaa
  qd = QDelta (uminus_rata (qdfst qd), uminus_rata (qdsnd qd));;

let uminus_QDelta = ({uminus = uminus_QDeltaa} : qDelta uminus);;

let rec less_eq_QDelta
  qd1 qd2 =
    less_rat (qdfst qd1) (qdfst qd2) ||
      equal_rata (qdfst qd1) (qdfst qd2) &&
        less_eq_rat (qdsnd qd1) (qdsnd qd2);;

let rec less_QDelta
  qd1 qd2 =
    less_rat (qdfst qd1) (qdfst qd2) ||
      equal_rata (qdfst qd1) (qdfst qd2) && less_rat (qdsnd qd1) (qdsnd qd2);;

let ord_QDelta = ({less_eq = less_eq_QDelta; less = less_QDelta} : qDelta ord);;

let quasi_order_QDelta = ({ord_quasi_order = ord_QDelta} : qDelta quasi_order);;

let weak_order_QDelta =
  ({quasi_order_weak_order = quasi_order_QDelta} : qDelta weak_order);;

let preorder_QDelta = ({ord_preorder = ord_QDelta} : qDelta preorder);;

let order_QDelta =
  ({preorder_order = preorder_QDelta; weak_order_order = weak_order_QDelta} :
    qDelta order);;

let semigroup_add_QDelta =
  ({plus_semigroup_add = plus_QDelta} : qDelta semigroup_add);;

let cancel_semigroup_add_QDelta =
  ({semigroup_add_cancel_semigroup_add = semigroup_add_QDelta} :
    qDelta cancel_semigroup_add);;

let monoid_add_QDelta =
  ({semigroup_add_monoid_add = semigroup_add_QDelta;
     zero_monoid_add = zero_QDelta}
    : qDelta monoid_add);;

let group_add_QDelta =
  ({cancel_semigroup_add_group_add = cancel_semigroup_add_QDelta;
     minus_group_add = minus_QDelta; monoid_add_group_add = monoid_add_QDelta;
     uminus_group_add = uminus_QDelta}
    : qDelta group_add);;

let linorder_QDelta = ({order_linorder = order_QDelta} : qDelta linorder);;

let ab_semigroup_add_QDelta =
  ({semigroup_add_ab_semigroup_add = semigroup_add_QDelta} :
    qDelta ab_semigroup_add);;

let cancel_ab_semigroup_add_QDelta =
  ({ab_semigroup_add_cancel_ab_semigroup_add = ab_semigroup_add_QDelta;
     cancel_semigroup_add_cancel_ab_semigroup_add = cancel_semigroup_add_QDelta;
     minus_cancel_ab_semigroup_add = minus_QDelta}
    : qDelta cancel_ab_semigroup_add);;

let comm_monoid_add_QDelta =
  ({ab_semigroup_add_comm_monoid_add = ab_semigroup_add_QDelta;
     monoid_add_comm_monoid_add = monoid_add_QDelta}
    : qDelta comm_monoid_add);;

let cancel_comm_monoid_add_QDelta =
  ({cancel_ab_semigroup_add_cancel_comm_monoid_add =
      cancel_ab_semigroup_add_QDelta;
     comm_monoid_add_cancel_comm_monoid_add = comm_monoid_add_QDelta}
    : qDelta cancel_comm_monoid_add);;

let ab_group_add_QDelta =
  ({cancel_comm_monoid_add_ab_group_add = cancel_comm_monoid_add_QDelta;
     group_add_ab_group_add = group_add_QDelta}
    : qDelta ab_group_add);;

let rec scaleRat_QDeltaa
  r qd = QDelta (times_rata r (qdfst qd), times_rata r (qdsnd qd));;

type 'a scaleRat = {scaleRat : rat -> 'a -> 'a};;
let scaleRat _A = _A.scaleRat;;

type 'a rational_vector =
  {ab_group_add_rational_vector : 'a ab_group_add;
    scaleRat_rational_vector : 'a scaleRat};;

type 'a ordered_rational_vector =
  {order_ordered_rational_vector : 'a order;
    rational_vector_ordered_rational_vector : 'a rational_vector};;

type 'a linordered_rational_vector =
  {ordered_ab_semigroup_add_linordered_rational_vector :
     'a ordered_ab_semigroup_add;
    linorder_linordered_rational_vector : 'a linorder;
    ordered_rational_vector_linordered_rational_vector :
      'a ordered_rational_vector};;

type 'a lrv =
  {one_lrv : 'a one;
    linordered_rational_vector_lrv : 'a linordered_rational_vector};;

let scaleRat_QDelta = ({scaleRat = scaleRat_QDeltaa} : qDelta scaleRat);;

let rational_vector_QDelta =
  ({ab_group_add_rational_vector = ab_group_add_QDelta;
     scaleRat_rational_vector = scaleRat_QDelta}
    : qDelta rational_vector);;

let ordered_rational_vector_QDelta =
  ({order_ordered_rational_vector = order_QDelta;
     rational_vector_ordered_rational_vector = rational_vector_QDelta}
    : qDelta ordered_rational_vector);;

let ordered_ab_semigroup_add_QDelta =
  ({ab_semigroup_add_ordered_ab_semigroup_add = ab_semigroup_add_QDelta;
     order_ordered_ab_semigroup_add = order_QDelta}
    : qDelta ordered_ab_semigroup_add);;

let linordered_rational_vector_QDelta =
  ({ordered_ab_semigroup_add_linordered_rational_vector =
      ordered_ab_semigroup_add_QDelta;
     linorder_linordered_rational_vector = linorder_QDelta;
     ordered_rational_vector_linordered_rational_vector =
       ordered_rational_vector_QDelta}
    : qDelta linordered_rational_vector);;

let lrv_QDelta =
  ({one_lrv = one_QDelta;
     linordered_rational_vector_lrv = linordered_rational_vector_QDelta}
    : qDelta lrv);;

let ccompare_QDeltaa : (qDelta -> qDelta -> ordera) option
  = Some (comparator_of (equal_QDelta, linorder_QDelta));;

let ccompare_QDelta = ({ccompare = ccompare_QDeltaa} : qDelta ccompare);;

type 'a formula = Atom of 'a | NegAtom of 'a | Conjunction of 'a formula list |
  Disjunction of 'a formula list;;

let rec equal_formula _A = ({equal = equal_formulaa _A} : 'a formula equal)
and equal_formulaa _A
  x0 x1 = match x0, x1 with Conjunction x3, Disjunction x4 -> false
    | Disjunction x4, Conjunction x3 -> false
    | NegAtom x2, Disjunction x4 -> false
    | Disjunction x4, NegAtom x2 -> false
    | NegAtom x2, Conjunction x3 -> false
    | Conjunction x3, NegAtom x2 -> false
    | Atom x1, Disjunction x4 -> false
    | Disjunction x4, Atom x1 -> false
    | Atom x1, Conjunction x3 -> false
    | Conjunction x3, Atom x1 -> false
    | Atom x1, NegAtom x2 -> false
    | NegAtom x2, Atom x1 -> false
    | Disjunction x4, Disjunction y4 -> equal_lista (equal_formula _A) x4 y4
    | Conjunction x3, Conjunction y3 -> equal_lista (equal_formula _A) x3 y3
    | NegAtom x2, NegAtom y2 -> eq _A x2 y2
    | Atom x1, Atom y1 -> eq _A x1 y1;;

type 'a poly = Poly of 'a list;;

let rec coeffs _A (Poly x) = x;;

let rec equal_polya (_A1, _A2)
  p q = equal_lista _A2 (coeffs _A1 p) (coeffs _A1 q);;

let rec equal_poly (_A1, _A2) =
  ({equal = equal_polya (_A1, _A2)} : 'a poly equal);;

let rec zero_polyb _A = Poly [];;

let rec cCons (_A1, _A2)
  x xs = (if null xs && eq _A2 x (zerob _A1) then [] else x :: xs);;

let rec plus_coeffs (_A1, _A2)
  xs x1 = match xs, x1 with xs, [] -> xs
    | [], v :: va -> v :: va
    | x :: xs, y :: ys ->
        cCons (_A1.monoid_add_comm_monoid_add.zero_monoid_add, _A2)
          (plus _A1.monoid_add_comm_monoid_add.semigroup_add_monoid_add.plus_semigroup_add
            x y)
          (plus_coeffs (_A1, _A2) xs ys);;

let rec plus_polya (_A1, _A2)
  p q = Poly (plus_coeffs (_A1, _A2)
               (coeffs _A1.monoid_add_comm_monoid_add.zero_monoid_add p)
               (coeffs _A1.monoid_add_comm_monoid_add.zero_monoid_add q));;

let rec foldr f x1 = match f, x1 with f, [] -> id
                | f, x :: xs -> comp (f x) (foldr f xs);;

let rec fold_coeffs _A f p = foldr f (coeffs _A p);;

let rec smult (_A1, _A2, _A3)
  a p = Poly (if eq _A1 a
                   (zerob
                     _A2.semiring_0_comm_semiring_0.mult_zero_semiring_0.zero_mult_zero)
               then []
               else map (times
                          _A2.semiring_0_comm_semiring_0.mult_zero_semiring_0.times_mult_zero
                          a)
                      (coeffs
                        _A2.semiring_0_comm_semiring_0.mult_zero_semiring_0.zero_mult_zero
                        p));;

let rec pCons (_A1, _A2) a p = Poly (cCons (_A1, _A2) a (coeffs _A1 p));;

let rec times_polya (_A1, _A2, _A3)
  p q = fold_coeffs
          _A2.semiring_0_comm_semiring_0.mult_zero_semiring_0.zero_mult_zero
          (fun a pa ->
            plus_polya
              (_A2.semiring_0_comm_semiring_0.comm_monoid_add_semiring_0, _A1)
              (smult (_A1, _A2, _A3) a q)
              (pCons
                (_A2.semiring_0_comm_semiring_0.mult_zero_semiring_0.zero_mult_zero,
                  _A1)
                (zerob
                  _A2.semiring_0_comm_semiring_0.mult_zero_semiring_0.zero_mult_zero)
                pa))
          p (zero_polyb
              _A2.semiring_0_comm_semiring_0.mult_zero_semiring_0.zero_mult_zero);;

let rec times_poly (_A1, _A2, _A3) =
  ({times = times_polya (_A1, _A2, _A3)} : 'a poly times);;

let rec dvd_poly (_A1, _A2, _A3) =
  ({times_dvd = (times_poly (_A1, _A2.comm_semiring_0_comm_semiring_1, _A3))} :
    'a poly dvd);;

let rec one_polyb _A
  = Poly [onea _A.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral];;

let rec one_poly _A = ({onea = one_polyb _A} : 'a poly one);;

let rec uminus_polya _A
  p = Poly (map (uminus _A.group_add_ab_group_add.uminus_group_add)
             (coeffs
               _A.group_add_ab_group_add.monoid_add_group_add.zero_monoid_add
               p));;

let rec minus_polya (_A1, _A2)
  p q = plus_polya
          (_A1.cancel_comm_monoid_add_ab_group_add.comm_monoid_add_cancel_comm_monoid_add,
            _A2)
          p (uminus_polya _A1 q);;

let rec plus_poly (_A1, _A2) = ({plus = plus_polya (_A1, _A2)} : 'a poly plus);;

let rec semigroup_add_poly (_A1, _A2) =
  ({plus_semigroup_add = (plus_poly (_A1, _A2))} : 'a poly semigroup_add);;

let rec cancel_semigroup_add_poly (_A1, _A2) =
  ({semigroup_add_cancel_semigroup_add =
      (semigroup_add_poly (_A1.comm_monoid_add_cancel_comm_monoid_add, _A2))}
    : 'a poly cancel_semigroup_add);;

let rec ab_semigroup_add_poly (_A1, _A2) =
  ({semigroup_add_ab_semigroup_add = (semigroup_add_poly (_A1, _A2))} :
    'a poly ab_semigroup_add);;

let rec minus_poly (_A1, _A2) =
  ({minus = minus_polya (_A1, _A2)} : 'a poly minus);;

let rec cancel_ab_semigroup_add_poly (_A1, _A2) =
  ({ab_semigroup_add_cancel_ab_semigroup_add =
      (ab_semigroup_add_poly
        (_A1.cancel_comm_monoid_add_ab_group_add.comm_monoid_add_cancel_comm_monoid_add,
          _A2));
     cancel_semigroup_add_cancel_ab_semigroup_add =
       (cancel_semigroup_add_poly
         (_A1.cancel_comm_monoid_add_ab_group_add, _A2));
     minus_cancel_ab_semigroup_add = (minus_poly (_A1, _A2))}
    : 'a poly cancel_ab_semigroup_add);;

let rec zero_poly _A = ({zerob = zero_polyb _A} : 'a poly zero);;

let rec monoid_add_poly (_A1, _A2) =
  ({semigroup_add_monoid_add = (semigroup_add_poly (_A1, _A2));
     zero_monoid_add =
       (zero_poly _A1.monoid_add_comm_monoid_add.zero_monoid_add)}
    : 'a poly monoid_add);;

let rec comm_monoid_add_poly (_A1, _A2) =
  ({ab_semigroup_add_comm_monoid_add = (ab_semigroup_add_poly (_A1, _A2));
     monoid_add_comm_monoid_add = (monoid_add_poly (_A1, _A2))}
    : 'a poly comm_monoid_add);;

let rec cancel_comm_monoid_add_poly (_A1, _A2) =
  ({cancel_ab_semigroup_add_cancel_comm_monoid_add =
      (cancel_ab_semigroup_add_poly (_A1, _A2));
     comm_monoid_add_cancel_comm_monoid_add =
       (comm_monoid_add_poly
         (_A1.cancel_comm_monoid_add_ab_group_add.comm_monoid_add_cancel_comm_monoid_add,
           _A2))}
    : 'a poly cancel_comm_monoid_add);;

let rec mult_zero_poly (_A1, _A2, _A3) =
  ({times_mult_zero = (times_poly (_A1, _A2, _A3));
     zero_mult_zero =
       (zero_poly
         _A2.semiring_0_comm_semiring_0.mult_zero_semiring_0.zero_mult_zero)}
    : 'a poly mult_zero);;

let rec semigroup_mult_poly (_A1, _A2, _A3) =
  ({times_semigroup_mult = (times_poly (_A1, _A2, _A3))} :
    'a poly semigroup_mult);;

let rec semiring_poly (_A1, _A2, _A3) =
  ({ab_semigroup_add_semiring =
      (ab_semigroup_add_poly
        (_A2.semiring_0_comm_semiring_0.comm_monoid_add_semiring_0, _A1));
     semigroup_mult_semiring = (semigroup_mult_poly (_A1, _A2, _A3))}
    : 'a poly semiring);;

let rec semiring_0_poly (_A1, _A2, _A3) =
  ({comm_monoid_add_semiring_0 =
      (comm_monoid_add_poly
        (_A2.semiring_0_comm_semiring_0.comm_monoid_add_semiring_0, _A1));
     mult_zero_semiring_0 = (mult_zero_poly (_A1, _A2, _A3));
     semiring_semiring_0 = (semiring_poly (_A1, _A2, _A3))}
    : 'a poly semiring_0);;

let rec semiring_0_cancel_poly (_A1, _A2, _A3, _A4) =
  ({cancel_comm_monoid_add_semiring_0_cancel =
      (cancel_comm_monoid_add_poly (_A1, _A2));
     semiring_0_semiring_0_cancel =
       (semiring_0_poly (_A2, _A3.comm_semiring_0_comm_semiring_0_cancel, _A4))}
    : 'a poly semiring_0_cancel);;

let rec ab_semigroup_mult_poly (_A1, _A2, _A3) =
  ({semigroup_mult_ab_semigroup_mult = (semigroup_mult_poly (_A1, _A2, _A3))} :
    'a poly ab_semigroup_mult);;

let rec comm_semiring_poly (_A1, _A2, _A3) =
  ({ab_semigroup_mult_comm_semiring = (ab_semigroup_mult_poly (_A1, _A2, _A3));
     semiring_comm_semiring = (semiring_poly (_A1, _A2, _A3))}
    : 'a poly comm_semiring);;

let rec comm_semiring_0_poly (_A1, _A2, _A3) =
  ({comm_semiring_comm_semiring_0 = (comm_semiring_poly (_A1, _A2, _A3));
     semiring_0_comm_semiring_0 = (semiring_0_poly (_A1, _A2, _A3))}
    : 'a poly comm_semiring_0);;

let rec comm_semiring_0_cancel_poly (_A1, _A2, _A3, _A4) =
  ({comm_semiring_0_comm_semiring_0_cancel =
      (comm_semiring_0_poly
        (_A2, _A3.comm_semiring_0_comm_semiring_0_cancel, _A4));
     semiring_0_cancel_comm_semiring_0_cancel =
       (semiring_0_cancel_poly (_A1, _A2, _A3, _A4))}
    : 'a poly comm_semiring_0_cancel);;

let rec power_poly (_A1, _A2, _A3) =
  ({one_power = (one_poly _A2);
     times_power = (times_poly (_A1, _A2.comm_semiring_0_comm_semiring_1, _A3))}
    : 'a poly power);;

let rec monoid_mult_poly (_A1, _A2, _A3) =
  ({semigroup_mult_monoid_mult =
      (semigroup_mult_poly (_A1, _A2.comm_semiring_0_comm_semiring_1, _A3));
     power_monoid_mult = (power_poly (_A1, _A2, _A3))}
    : 'a poly monoid_mult);;

let rec numeral_poly (_A1, _A2) =
  ({one_numeral = (one_poly _A2);
     semigroup_add_numeral =
       (semigroup_add_poly
         (_A2.semiring_1_comm_semiring_1.semiring_0_semiring_1.comm_monoid_add_semiring_0,
           _A1))}
    : 'a poly numeral);;

let rec semiring_numeral_poly (_A1, _A2, _A3) =
  ({monoid_mult_semiring_numeral = (monoid_mult_poly (_A1, _A2, _A3));
     numeral_semiring_numeral = (numeral_poly (_A1, _A2));
     semiring_semiring_numeral =
       (semiring_poly (_A1, _A2.comm_semiring_0_comm_semiring_1, _A3))}
    : 'a poly semiring_numeral);;

let rec zero_neq_one_poly _A =
  ({one_zero_neq_one = (one_poly _A);
     zero_zero_neq_one =
       (zero_poly
         _A.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)}
    : 'a poly zero_neq_one);;

let rec semiring_1_poly (_A1, _A2, _A3) =
  ({semiring_numeral_semiring_1 = (semiring_numeral_poly (_A1, _A2, _A3));
     semiring_0_semiring_1 =
       (semiring_0_poly (_A1, _A2.comm_semiring_0_comm_semiring_1, _A3));
     zero_neq_one_semiring_1 = (zero_neq_one_poly _A2)}
    : 'a poly semiring_1);;

let rec semiring_1_cancel_poly (_A1, _A2, _A3) =
  ({semiring_0_cancel_semiring_1_cancel =
      (semiring_0_cancel_poly
        (_A2.ring_1_comm_ring_1.ring_ring_1.ab_group_add_ring, _A1,
          _A2.comm_ring_comm_ring_1.comm_semiring_0_cancel_comm_ring, _A3));
     semiring_1_semiring_1_cancel =
       (semiring_1_poly
         (_A1, _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel,
           _A3))}
    : 'a poly semiring_1_cancel);;

let rec comm_monoid_mult_poly (_A1, _A2, _A3) =
  ({ab_semigroup_mult_comm_monoid_mult =
      (ab_semigroup_mult_poly (_A1, _A2.comm_semiring_0_comm_semiring_1, _A3));
     monoid_mult_comm_monoid_mult = (monoid_mult_poly (_A1, _A2, _A3));
     dvd_comm_monoid_mult = (dvd_poly (_A1, _A2, _A3))}
    : 'a poly comm_monoid_mult);;

let rec comm_semiring_1_poly (_A1, _A2, _A3) =
  ({comm_monoid_mult_comm_semiring_1 = (comm_monoid_mult_poly (_A1, _A2, _A3));
     comm_semiring_0_comm_semiring_1 =
       (comm_semiring_0_poly (_A1, _A2.comm_semiring_0_comm_semiring_1, _A3));
     semiring_1_comm_semiring_1 = (semiring_1_poly (_A1, _A2, _A3))}
    : 'a poly comm_semiring_1);;

let rec comm_semiring_1_cancel_poly (_A1, _A2, _A3) =
  ({comm_semiring_0_cancel_comm_semiring_1_cancel =
      (comm_semiring_0_cancel_poly
        (_A2.ring_1_comm_ring_1.ring_ring_1.ab_group_add_ring, _A1,
          _A2.comm_ring_comm_ring_1.comm_semiring_0_cancel_comm_ring, _A3));
     comm_semiring_1_comm_semiring_1_cancel =
       (comm_semiring_1_poly
         (_A1, _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel,
           _A3));
     semiring_1_cancel_comm_semiring_1_cancel =
       (semiring_1_cancel_poly (_A1, _A2, _A3))}
    : 'a poly comm_semiring_1_cancel);;

let rec comm_semiring_1_cancel_crossproduct_poly (_A1, _A2) =
  ({comm_semiring_1_cancel_comm_semiring_1_cancel_crossproduct =
      (comm_semiring_1_cancel_poly
        (_A1, _A2.comm_ring_1_idom,
          _A2.semidom_idom.semiring_1_no_zero_divisors_semidom.semiring_no_zero_divisors_semiring_1_no_zero_divisors))}
    : 'a poly comm_semiring_1_cancel_crossproduct);;

let rec semiring_no_zero_divisors_poly (_A1, _A2, _A3) =
  ({semiring_0_semiring_no_zero_divisors = (semiring_0_poly (_A1, _A2, _A3))} :
    'a poly semiring_no_zero_divisors);;

let rec semiring_1_no_zero_divisors_poly (_A1, _A2, _A3) =
  ({semiring_1_semiring_1_no_zero_divisors =
      (semiring_1_poly
        (_A1, _A2, _A3.semiring_no_zero_divisors_semiring_1_no_zero_divisors));
     semiring_no_zero_divisors_semiring_1_no_zero_divisors =
       (semiring_no_zero_divisors_poly
         (_A1, _A2.comm_semiring_0_comm_semiring_1,
           _A3.semiring_no_zero_divisors_semiring_1_no_zero_divisors))}
    : 'a poly semiring_1_no_zero_divisors);;

let rec semiring_no_zero_divisors_cancel_poly (_A1, _A2) =
  ({semiring_no_zero_divisors_semiring_no_zero_divisors_cancel =
      (semiring_no_zero_divisors_poly
        (_A1, _A2.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.comm_semiring_0_comm_semiring_1,
          _A2.semidom_idom.semiring_1_no_zero_divisors_semidom.semiring_no_zero_divisors_semiring_1_no_zero_divisors))}
    : 'a poly semiring_no_zero_divisors_cancel);;

let rec uminus_poly _A = ({uminus = uminus_polya _A} : 'a poly uminus);;

let rec group_add_poly (_A1, _A2) =
  ({cancel_semigroup_add_group_add =
      (cancel_semigroup_add_poly
        (_A1.cancel_comm_monoid_add_ab_group_add, _A2));
     minus_group_add = (minus_poly (_A1, _A2));
     monoid_add_group_add =
       (monoid_add_poly
         (_A1.cancel_comm_monoid_add_ab_group_add.comm_monoid_add_cancel_comm_monoid_add,
           _A2));
     uminus_group_add = (uminus_poly _A1)}
    : 'a poly group_add);;

let rec ab_group_add_poly (_A1, _A2) =
  ({cancel_comm_monoid_add_ab_group_add =
      (cancel_comm_monoid_add_poly (_A1, _A2));
     group_add_ab_group_add = (group_add_poly (_A1, _A2))}
    : 'a poly ab_group_add);;

let rec ring_poly (_A1, _A2, _A3) =
  ({ab_group_add_ring =
      (ab_group_add_poly (_A2.ring_comm_ring.ab_group_add_ring, _A1));
     semiring_0_cancel_ring =
       (semiring_0_cancel_poly
         (_A2.ring_comm_ring.ab_group_add_ring, _A1,
           _A2.comm_semiring_0_cancel_comm_ring, _A3))}
    : 'a poly ring);;

let rec ring_no_zero_divisors_poly (_A1, _A2) =
  ({ring_ring_no_zero_divisors =
      (ring_poly
        (_A1, _A2.comm_ring_1_idom.comm_ring_comm_ring_1,
          _A2.semidom_idom.semiring_1_no_zero_divisors_semidom.semiring_no_zero_divisors_semiring_1_no_zero_divisors));
     semiring_no_zero_divisors_cancel_ring_no_zero_divisors =
       (semiring_no_zero_divisors_cancel_poly (_A1, _A2))}
    : 'a poly ring_no_zero_divisors);;

let rec neg_numeral_poly (_A1, _A2) =
  ({group_add_neg_numeral =
      (group_add_poly
        (_A2.ring_1_comm_ring_1.ring_ring_1.ab_group_add_ring, _A1));
     numeral_neg_numeral =
       (numeral_poly
         (_A1, _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel))}
    : 'a poly neg_numeral);;

let rec ring_1_poly (_A1, _A2, _A3) =
  ({neg_numeral_ring_1 = (neg_numeral_poly (_A1, _A2));
     ring_ring_1 = (ring_poly (_A1, _A2.comm_ring_comm_ring_1, _A3));
     semiring_1_cancel_ring_1 = (semiring_1_cancel_poly (_A1, _A2, _A3))}
    : 'a poly ring_1);;

let rec ring_1_no_zero_divisors_poly (_A1, _A2) =
  ({ring_1_ring_1_no_zero_divisors =
      (ring_1_poly
        (_A1, _A2.comm_ring_1_idom,
          _A2.semidom_idom.semiring_1_no_zero_divisors_semidom.semiring_no_zero_divisors_semiring_1_no_zero_divisors));
     ring_no_zero_divisors_ring_1_no_zero_divisors =
       (ring_no_zero_divisors_poly (_A1, _A2));
     semiring_1_no_zero_divisors_ring_1_no_zero_divisors =
       (semiring_1_no_zero_divisors_poly
         (_A1, _A2.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel,
           _A2.semidom_idom.semiring_1_no_zero_divisors_semidom))}
    : 'a poly ring_1_no_zero_divisors);;

let rec comm_ring_poly (_A1, _A2, _A3) =
  ({comm_semiring_0_cancel_comm_ring =
      (comm_semiring_0_cancel_poly
        (_A2.ring_comm_ring.ab_group_add_ring, _A1,
          _A2.comm_semiring_0_cancel_comm_ring, _A3));
     ring_comm_ring = (ring_poly (_A1, _A2, _A3))}
    : 'a poly comm_ring);;

let rec comm_ring_1_poly (_A1, _A2, _A3) =
  ({comm_ring_comm_ring_1 =
      (comm_ring_poly (_A1, _A2.comm_ring_comm_ring_1, _A3));
     comm_semiring_1_cancel_comm_ring_1 =
       (comm_semiring_1_cancel_poly (_A1, _A2, _A3));
     ring_1_comm_ring_1 = (ring_1_poly (_A1, _A2, _A3))}
    : 'a poly comm_ring_1);;

let rec semidom_poly (_A1, _A2) =
  ({comm_semiring_1_cancel_semidom =
      (comm_semiring_1_cancel_poly
        (_A1, _A2.comm_ring_1_idom,
          _A2.semidom_idom.semiring_1_no_zero_divisors_semidom.semiring_no_zero_divisors_semiring_1_no_zero_divisors));
     semiring_1_no_zero_divisors_semidom =
       (semiring_1_no_zero_divisors_poly
         (_A1, _A2.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel,
           _A2.semidom_idom.semiring_1_no_zero_divisors_semidom))}
    : 'a poly semidom);;

let rec idom_poly (_A1, _A2) =
  ({comm_ring_1_idom =
      (comm_ring_1_poly
        (_A1, _A2.comm_ring_1_idom,
          _A2.semidom_idom.semiring_1_no_zero_divisors_semidom.semiring_no_zero_divisors_semiring_1_no_zero_divisors));
     ring_1_no_zero_divisors_idom = (ring_1_no_zero_divisors_poly (_A1, _A2));
     semidom_idom = (semidom_poly (_A1, _A2));
     comm_semiring_1_cancel_crossproduct_idom =
       (comm_semiring_1_cancel_crossproduct_poly (_A1, _A2))}
    : 'a poly idom);;

let rec minus_poly_rev_list _A
  xs x1 = match xs, x1 with
    x :: xs, y :: ys ->
      minus _A.minus_group_add x y :: minus_poly_rev_list _A xs ys
    | xs, [] -> xs
    | [], y :: ys -> [];;

let rec tla = function [] -> []
              | x21 :: x22 -> x22;;

let rec hda (x21 :: x22) = x21;;

let rec divide_poly_main_list (_A1, _A2)
  lc q r d n =
    (if equal_nata n zero_nata then q
      else (let cr = hda r in
             (if eq _A1 cr
                   (zerob
                     _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
               then divide_poly_main_list (_A1, _A2) lc
                      (cCons
                        (_A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero,
                          _A1)
                        cr q)
                      (tla r) d (minus_nata n one_nata)
               else (let a =
                       divide
                         _A2.semidom_divide_idom_divide.divide_semidom_divide cr
                         lc
                       in
                     let qq =
                       cCons (_A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero,
                               _A1)
                         a q
                       in
                     let rr =
                       minus_poly_rev_list
                         _A2.idom_idom_divide.comm_ring_1_idom.ring_1_comm_ring_1.neg_numeral_ring_1.group_add_neg_numeral
                         r (map (times
                                  _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd
                                  a)
                             d)
                       in
                      (if eq _A1 (hda rr)
                            (zerob
                              _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
                        then divide_poly_main_list (_A1, _A2) lc qq (tla rr) d
                               (minus_nata n one_nata)
                        else [])))));;

let rec dropWhile p x1 = match p, x1 with p, [] -> []
                    | p, x :: xs -> (if p x then dropWhile p xs else x :: xs);;

let rec rev xs = fold (fun a b -> a :: b) xs [];;

let rec strip_while p = comp (comp rev (dropWhile p)) rev;;

let rec poly_of_list (_A1, _A2)
  asa = Poly (strip_while
               (eq _A2 (zerob _A1.monoid_add_comm_monoid_add.zero_monoid_add))
               asa);;

let rec divide_poly_list (_A1, _A2)
  f g = (let cg =
           coeffs
             _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
             g
           in
          (if null cg then g
            else (let cf =
                    coeffs
                      _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
                      f
                    in
                  let cgr = rev cg in
                   poly_of_list
                     (_A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.comm_monoid_add_semiring_0,
                       _A1)
                     (divide_poly_main_list (_A1, _A2) (hda cgr) [] (rev cf) cgr
                       (minus_nata (plus_nata one_nata (size_list cf))
                         (size_list cg))))));;

let rec divide_polya (_A1, _A2) f g = divide_poly_list (_A1, _A2) f g;;

let rec divide_poly (_A1, _A2) =
  ({divide = divide_polya (_A1, _A2)} : 'a poly divide);;

let rec semidom_divide_poly (_A1, _A2) =
  ({divide_semidom_divide = (divide_poly (_A1, _A2));
     semidom_semidom_divide = (semidom_poly (_A1, _A2.idom_idom_divide));
     semiring_no_zero_divisors_cancel_semidom_divide =
       (semiring_no_zero_divisors_cancel_poly (_A1, _A2.idom_idom_divide))}
    : 'a poly semidom_divide);;

let rec idom_divide_poly (_A1, _A2) =
  ({idom_idom_divide = (idom_poly (_A1, _A2.idom_idom_divide));
     semidom_divide_idom_divide = (semidom_divide_poly (_A1, _A2))}
    : 'a poly idom_divide);;

type hints = Hints of int list | Simplex;;

let default_hintsa : hints = Simplex;;

let default_hints = ({default = default_hintsa} : hints default);;

let rec showsl_hints
  = function
    Hints xs -> comp (showsl_lit "Linear-Combination: ") (showsl_list_int xs)
    | Simplex -> showsl_lit "Simplex";;

let rec showsl_list_hints xs = default_showsl_list showsl_hints xs;;

let showl_hints =
  ({showsl = showsl_hints; showsl_list = showsl_list_hints} : hints showl);;

let rec map_of _A
  x0 k = match x0, k with
    (l, v) :: ps, k -> (if eq _A l k then Some v else map_of _A ps k)
    | [], k -> None;;

let rec count_of _A
  xs x = (match map_of _A xs x with None -> zero_nata | Some n -> n);;

type ('b, 'a) alist = Alist of ('b * 'a) list;;

let rec impl_ofa (Alist x) = x;;

type 'a multiset = Bag of ('a, nat) alist;;

let rec count _A (Bag xs) = count_of _A (impl_ofa xs);;

let rec subseteq_mset _A
  (Bag xs) a =
    list_all (fun (x, n) -> less_eq_nat n (count _A a x)) (impl_ofa xs);;

let rec equal_multiseta _A
  m1 m2 = subseteq_mset _A m1 m2 && subseteq_mset _A m2 m1;;

let rec equal_multiset _A = ({equal = equal_multiseta _A} : 'a multiset equal);;

let rec map_default _A
  k v f x3 = match k, v, f, x3 with k, v, f, [] -> [(k, v)]
    | k, v, f, p :: ps ->
        (if eq _A (fst p) k then (k, f (snd p)) :: ps
          else p :: map_default _A k v f ps);;

let rec join_raw _A
  f xs ys =
    foldr (fun (k, v) -> map_default _A k v (fun va -> f k (va, v))) ys xs;;

let rec joina _A xc xd xe = Alist (join_raw _A xc (impl_ofa xd) (impl_ofa xe));;

let rec plus_multiseta _A
  (Bag xs) (Bag ys) = Bag (joina _A (fun _ (a, b) -> plus_nata a b) xs ys);;

let rec plus_multiset _A = ({plus = plus_multiseta _A} : 'a multiset plus);;

let emptya : ('a, 'b) alist = Alist [];;

let zero_multiseta : 'a multiset = Bag emptya;;

let zero_multiset = ({zerob = zero_multiseta} : 'a multiset zero);;

let rec semigroup_add_multiset _A =
  ({plus_semigroup_add = (plus_multiset _A)} : 'a multiset semigroup_add);;

let rec monoid_add_multiset _A =
  ({semigroup_add_monoid_add = (semigroup_add_multiset _A);
     zero_monoid_add = zero_multiset}
    : 'a multiset monoid_add);;

let rec ab_semigroup_add_multiset _A =
  ({semigroup_add_ab_semigroup_add = (semigroup_add_multiset _A)} :
    'a multiset ab_semigroup_add);;

let rec comm_monoid_add_multiset _A =
  ({ab_semigroup_add_comm_monoid_add = (ab_semigroup_add_multiset _A);
     monoid_add_comm_monoid_add = (monoid_add_multiset _A)}
    : 'a multiset comm_monoid_add);;

type 'a monom = Abs_monom of ('a * nat) list;;

let rec rep_monom _A (Abs_monom x) = x;;

let rec equal_monoma (_A1, _A2)
  xa xc =
    equal_lista (equal_prod _A1 equal_nat) (rep_monom _A2 xa)
      (rep_monom _A2 xc);;

let rec equal_monom (_A1, _A2) =
  ({equal = equal_monoma (_A1, _A2)} : 'a monom equal);;

let rec less_eq_prod _A _B
  (x1, y1) (x2, y2) = less _A x1 x2 || less_eq _A x1 x2 && less_eq _B y1 y2;;

let rec less_prod _A _B
  (x1, y1) (x2, y2) = less _A x1 x2 || less_eq _A x1 x2 && less _B y1 y2;;

let rec ord_prod _A _B =
  ({less_eq = less_eq_prod _A _B; less = less_prod _A _B} : ('a * 'b) ord);;

let rec comparator_prod
  comp_a comp_b (x, xa) (y, ya) =
    (match comp_a x y with Eq -> comp_b xa ya | Lt -> Lt | Gt -> Gt);;

let rec compare_proda _A _B = comparator_prod (compare _A) (compare _B);;

let rec compare_prod _A _B =
  ({compare = compare_proda _A _B} : ('a * 'b) compare);;

let rec quasi_order_prod _A _B =
  ({ord_quasi_order =
      (ord_prod _A.preorder_order.ord_preorder _B.preorder_order.ord_preorder)}
    : ('a * 'b) quasi_order);;

let rec weak_order_prod _A _B =
  ({quasi_order_weak_order = (quasi_order_prod _A _B)} : ('a * 'b) weak_order);;

let rec preorder_prod _A _B =
  ({ord_preorder = (ord_prod _A.ord_preorder _B.ord_preorder)} :
    ('a * 'b) preorder);;

let rec order_prod _A _B =
  ({preorder_order = (preorder_prod _A.preorder_order _B.preorder_order);
     weak_order_order = (weak_order_prod _A _B)}
    : ('a * 'b) order);;

let rec equality_prod eq_a eq_b (x, xa) (y, ya) = eq_a x y && eq_b xa ya;;

let rec ceq_proda _A _B
  = (match ceq _A with None -> None
      | Some eq_a ->
        (match ceq _B with None -> None
          | Some eq_b -> Some (equality_prod eq_a eq_b)));;

let rec ceq_prod _A _B = ({ceq = ceq_proda _A _B} : ('a * 'b) ceq);;

let rec set_impl_choose2
  x y = match x, y with Set_Monada, Set_Monada -> Set_Monada
    | Set_RBT, Set_RBT -> Set_RBT
    | Set_DList, Set_DList -> Set_DList
    | Set_Collect, Set_Collect -> Set_Collect
    | x, y -> Set_Choose;;

let rec set_impl_proda _A _B
  = Phantom
      (set_impl_choose2 (of_phantom (set_impl _A)) (of_phantom (set_impl _B)));;

let rec set_impl_prod _A _B =
  ({set_impl = set_impl_proda _A _B} : ('a * 'b) set_impl);;

let rec linorder_prod _A _B =
  ({order_linorder = (order_prod _A.order_linorder _B.order_linorder)} :
    ('a * 'b) linorder);;

let rec showsl_prod _A _B
  (x, y) =
    comp (comp (comp (comp (showsl_lit "(") (showsl _A x)) (showsl_lit ", "))
           (showsl _B y))
      (showsl_lit ")");;

let rec showsl_list_prod _A _B xs = default_showsl_list (showsl_prod _A _B) xs;;

let rec showl_prod _A _B =
  ({showsl = showsl_prod _A _B; showsl_list = showsl_list_prod _A _B} :
    ('a * 'b) showl);;

let rec finite_UNIV_proda _A _B
  = Phantom (of_phantom (finite_UNIV _A) && of_phantom (finite_UNIV _B));;

let rec card_UNIV_proda _A _B
  = Phantom
      (times_nata (of_phantom (card_UNIVa _A)) (of_phantom (card_UNIVa _B)));;

let rec finite_UNIV_prod _A _B =
  ({finite_UNIV = finite_UNIV_proda _A _B} : ('a * 'b) finite_UNIV);;

let rec card_UNIV_prod _A _B =
  ({finite_UNIV_card_UNIV =
      (finite_UNIV_prod _A.finite_UNIV_card_UNIV _B.finite_UNIV_card_UNIV);
     card_UNIVa = card_UNIV_proda _A _B}
    : ('a * 'b) card_UNIV);;

let rec product x0 uu = match x0, uu with [], uu -> []
                  | x :: xs, ys -> map (fun a -> (x, a)) ys @ product xs ys;;

let rec cEnum_prod _A _B
  = (match cEnum _A with None -> None
      | Some (enum_a, (enum_all_a, enum_ex_a)) ->
        (match cEnum _B with None -> None
          | Some (enum_b, (enum_all_b, enum_ex_b)) ->
            Some (product enum_a enum_b,
                   ((fun p ->
                      enum_all_a (fun x -> enum_all_b (fun y -> p (x, y)))),
                     (fun p ->
                       enum_ex_a (fun x -> enum_ex_b (fun y -> p (x, y))))))));;

let rec cenum_prod _A _B = ({cEnum = cEnum_prod _A _B} : ('a * 'b) cenum);;

let rec compare_order_prod _A _B =
  ({compare_compare_order =
      (compare_prod _A.compare_compare_order _B.compare_compare_order);
     linorder_compare_order =
       (linorder_prod _A.linorder_compare_order _B.linorder_compare_order)}
    : ('a * 'b) compare_order);;

let rec ccompare_proda _A _B
  = (match ccompare _A with None -> None
      | Some comp_a ->
        (match ccompare _B with None -> None
          | Some comp_b -> Some (comparator_prod comp_a comp_b)));;

let rec ccompare_prod _A _B =
  ({ccompare = ccompare_proda _A _B} : ('a * 'b) ccompare);;

let rec mapping_impl_choose2
  x y = match x, y with Mapping_RBT, Mapping_RBT -> Mapping_RBT
    | Mapping_Assoc_List, Mapping_Assoc_List -> Mapping_Assoc_List
    | Mapping_Mapping, Mapping_Mapping -> Mapping_Mapping
    | x, y -> Mapping_Choose;;

let rec mapping_impl_proda _A _B
  = Phantom
      (mapping_impl_choose2 (of_phantom (mapping_impl _A))
        (of_phantom (mapping_impl _B)));;

let rec mapping_impl_prod _A _B =
  ({mapping_impl = mapping_impl_proda _A _B} : ('a * 'b) mapping_impl);;

let rec cproper_interval_proda _A _B
  x0 x1 = match x0, x1 with None, None -> true
    | None, Some (y1, y2) ->
        cproper_interval _A None (Some y1) || cproper_interval _B None (Some y2)
    | Some (x1, x2), None ->
        cproper_interval _A (Some x1) None || cproper_interval _B (Some x2) None
    | Some (x1, x2), Some (y1, y2) ->
        cproper_interval _A (Some x1) (Some y1) ||
          (lt_of_comp (the (ccompare _A.ccompare_cproper_interval)) x1 y1 &&
             (cproper_interval _B (Some x2) None ||
               cproper_interval _B None (Some y2)) ||
            not (lt_of_comp (the (ccompare _A.ccompare_cproper_interval)) y1
                  x1) &&
              cproper_interval _B (Some x2) (Some y2));;

let rec cproper_interval_prod _A _B =
  ({ccompare_cproper_interval =
      (ccompare_prod _A.ccompare_cproper_interval _B.ccompare_cproper_interval);
     cproper_interval = cproper_interval_proda _A _B}
    : ('a * 'b) cproper_interval);;

let rec comparator_unit x y = Eq;;

let rec compare_unita x = comparator_unit x;;

let compare_unit = ({compare = compare_unita} : unit compare);;

type ('a, 'b, 'c, 'd) transition_rule =
  Transition of 'd * 'd * ('a, ('b trans_var * 'c)) term formula;;

let rec equal_transition_rule _A _B _C _D
  (Transition (x1, x2, x3)) (Transition (y1, y2, y3)) =
    eq _D x1 y1 &&
      (eq _D x2 y2 &&
        equal_formulaa (equal_term _A (equal_prod (equal_trans_var _B) _C)) x3
          y3);;

let rec ceq_transition_rulea _A _B _C _D
  = Some (equal_transition_rule _A _B _C _D);;

let rec ceq_transition_rule _A _B _C _D =
  ({ceq = ceq_transition_rulea _A _B _C _D} :
    ('a, 'b, 'c, 'd) transition_rule ceq);;

let set_impl_transition_rulea :
  (('a, 'b, 'c, 'd) transition_rule, set_impla) phantom
  = Phantom Set_RBT;;

let set_impl_transition_rule =
  ({set_impl = set_impl_transition_rulea} :
    ('a, 'b, 'c, 'd) transition_rule set_impl);;

let rec comparator_formula
  comp_a x1 x2 = match comp_a, x1, x2 with
    comp_a, Disjunction x, Disjunction yc ->
      comparator_list (comparator_formula comp_a) x yc
    | comp_a, Disjunction x, Conjunction yb -> Gt
    | comp_a, Disjunction x, NegAtom ya -> Gt
    | comp_a, Disjunction x, Atom y -> Gt
    | comp_a, Conjunction x, Disjunction yc -> Lt
    | comp_a, Conjunction x, Conjunction yb ->
        comparator_list (comparator_formula comp_a) x yb
    | comp_a, Conjunction x, NegAtom ya -> Gt
    | comp_a, Conjunction x, Atom y -> Gt
    | comp_a, NegAtom x, Disjunction yc -> Lt
    | comp_a, NegAtom x, Conjunction yb -> Lt
    | comp_a, NegAtom x, NegAtom ya -> comp_a x ya
    | comp_a, NegAtom x, Atom y -> Gt
    | comp_a, Atom x, Disjunction yc -> Lt
    | comp_a, Atom x, Conjunction yb -> Lt
    | comp_a, Atom x, NegAtom ya -> Lt
    | comp_a, Atom x, Atom y -> comp_a x y;;

let rec comparator_transition_rule
  comp_f comp_v comp_t comp_l (Transition (x, xa, xb)) (Transition (y, ya, yb))
    = (match comp_l x y
        with Eq ->
          (match comp_l xa ya
            with Eq ->
              comparator_formula
                (comparator_term comp_f
                  (comparator_prod (comparator_trans_var comp_v) comp_t))
                xb yb
            | Lt -> Lt | Gt -> Gt)
        | Lt -> Lt | Gt -> Gt);;

let rec ccompare_transition_rulea _A _B _C _D
  = (match ccompare _A with None -> None
      | Some comp_f ->
        (match ccompare _B with None -> None
          | Some comp_v ->
            (match ccompare _C with None -> None
              | Some comp_t ->
                (match ccompare _D with None -> None
                  | Some comp_l ->
                    Some (comparator_transition_rule comp_f comp_v comp_t
                           comp_l)))));;

let rec ccompare_transition_rule _A _B _C _D =
  ({ccompare = ccompare_transition_rulea _A _B _C _D} :
    ('a, 'b, 'c, 'd) transition_rule ccompare);;

type ('a, 'b) gctxt = GCHole | GCFun of 'a * ('a, 'b) gctxt list;;

let rec equal_gctxt _A = ({equal = equal_gctxta _A} : ('a, 'b) gctxt equal)
and equal_gctxta _A
  x0 x1 = match x0, x1 with GCHole, GCFun (x21, x22) -> false
    | GCFun (x21, x22), GCHole -> false
    | GCFun (x21, x22), GCFun (y21, y22) ->
        eq _A x21 y21 && equal_lista (equal_gctxt _A) x22 y22
    | GCHole, GCHole -> true;;

let rec ceq_gctxta _A = Some (equal_gctxta _A);;

let rec ceq_gctxt _A = ({ceq = ceq_gctxta _A} : ('a, 'b) gctxt ceq);;

let set_impl_gctxta : (('a, 'b) gctxt, set_impla) phantom = Phantom Set_RBT;;

let set_impl_gctxt = ({set_impl = set_impl_gctxta} : ('a, 'b) gctxt set_impl);;

let rec comparator_gctxt
  comp_f x1 x2 = match comp_f, x1, x2 with
    comp_f, GCFun (x, xa), GCFun (y, ya) ->
      (match comp_f x y
        with Eq -> comparator_list (comparator_gctxt comp_f) xa ya | Lt -> Lt
        | Gt -> Gt)
    | comp_f, GCFun (x, xa), GCHole -> Gt
    | comp_f, GCHole, GCFun (y, ya) -> Lt
    | comp_f, GCHole, GCHole -> Eq;;

let rec compare_gctxt _A = comparator_gctxt (compare _A);;

let rec ccompare_gctxta _A _B = Some (compare_gctxt _A);;

let rec ccompare_gctxt _A _B =
  ({ccompare = ccompare_gctxta _A _B} : ('a, 'b) gctxt ccompare);;

type ('a, 'b) acterm = AVar of 'b | AFun of 'a * ('a, 'b) acterm list |
  AAC of 'a * ('a, 'b) acterm multiset;;

let rec equal_acterm _A _B =
  ({equal = equal_acterma _A _B} : ('a, 'b) acterm equal)
and equal_acterma _A _B
  x0 x1 = match x0, x1 with AFun (x21, x22), AAC (x31, x32) -> false
    | AAC (x31, x32), AFun (x21, x22) -> false
    | AVar x1, AAC (x31, x32) -> false
    | AAC (x31, x32), AVar x1 -> false
    | AVar x1, AFun (x21, x22) -> false
    | AFun (x21, x22), AVar x1 -> false
    | AAC (x31, x32), AAC (y31, y32) ->
        eq _A x31 y31 && equal_multiseta (equal_acterm _A _B) x32 y32
    | AFun (x21, x22), AFun (y21, y22) ->
        eq _A x21 y21 && equal_lista (equal_acterm _A _B) x22 y22
    | AVar x1, AVar y1 -> eq _B x1 y1;;

type ('a, 'b) ta_rule = TA_rule of 'b * 'a list * 'a;;

let rec equal_ta_rulea _A _B
  (TA_rule (x1, x2, x3)) (TA_rule (y1, y2, y3)) =
    eq _B x1 y1 && (equal_lista _A x2 y2 && eq _A x3 y3);;

let rec equal_ta_rule _A _B =
  ({equal = equal_ta_rulea _A _B} : ('a, 'b) ta_rule equal);;

let rec comparator_ta_rule
  comp_q comp_f (TA_rule (x, xa, xb)) (TA_rule (y, ya, yb)) =
    (match comp_f x y
      with Eq ->
        (match comparator_list comp_q xa ya with Eq -> comp_q xb yb | Lt -> Lt
          | Gt -> Gt)
      | Lt -> Lt | Gt -> Gt);;

let rec compare_ta_rule _A _B = comparator_ta_rule (compare _A) (compare _B);;

let rec less_eq_ta_rule _A _B = le_of_comp (compare_ta_rule _A _B);;

let rec less_ta_rule _A _B = lt_of_comp (compare_ta_rule _A _B);;

let rec ord_ta_rule _A _B =
  ({less_eq = less_eq_ta_rule _A _B; less = less_ta_rule _A _B} :
    ('a, 'b) ta_rule ord);;

let rec quasi_order_ta_rule _A _B =
  ({ord_quasi_order = (ord_ta_rule _A _B)} : ('a, 'b) ta_rule quasi_order);;

let rec weak_order_ta_rule _A _B =
  ({quasi_order_weak_order = (quasi_order_ta_rule _A _B)} :
    ('a, 'b) ta_rule weak_order);;

let rec preorder_ta_rule _A _B =
  ({ord_preorder = (ord_ta_rule _A _B)} : ('a, 'b) ta_rule preorder);;

let rec order_ta_rule _A _B =
  ({preorder_order = (preorder_ta_rule _A _B);
     weak_order_order = (weak_order_ta_rule _A _B)}
    : ('a, 'b) ta_rule order);;

let rec ceq_ta_rulea _A _B = Some (equal_ta_rulea _A _B);;

let rec ceq_ta_rule _A _B =
  ({ceq = ceq_ta_rulea _A _B} : ('a, 'b) ta_rule ceq);;

let set_impl_ta_rulea : (('a, 'b) ta_rule, set_impla) phantom
  = Phantom Set_RBT;;

let set_impl_ta_rule =
  ({set_impl = set_impl_ta_rulea} : ('a, 'b) ta_rule set_impl);;

let rec linorder_ta_rule _A _B =
  ({order_linorder = (order_ta_rule _A _B)} : ('a, 'b) ta_rule linorder);;

let rec showsl_ta_rule _A _B
  (TA_rule (f, qs, q)) =
    comp (comp (comp (showsl _B f) (showsl_list _A qs)) (showsl_lit " -> "))
      (showsl _A q);;

let rec showsl_list_ta_rule _A _B
  xs = default_showsl_list (showsl_ta_rule _A _B) xs;;

let rec showl_ta_rule _A _B =
  ({showsl = showsl_ta_rule _A _B; showsl_list = showsl_list_ta_rule _A _B} :
    ('a, 'b) ta_rule showl);;

let cEnum_ta_rule :
  (('a, 'b) ta_rule list *
    (((('a, 'b) ta_rule -> bool) -> bool) *
      ((('a, 'b) ta_rule -> bool) -> bool))) option
  = None;;

let cenum_ta_rule = ({cEnum = cEnum_ta_rule} : ('a, 'b) ta_rule cenum);;

let finite_UNIV_ta_rulea : (('a, 'b) ta_rule, bool) phantom = Phantom false;;

let finite_UNIV_ta_rule =
  ({finite_UNIV = finite_UNIV_ta_rulea} : ('a, 'b) ta_rule finite_UNIV);;

let rec ccompare_ta_rulea _A _B
  = (match ccompare _A with None -> None
      | Some comp_q ->
        (match ccompare _B with None -> None
          | Some comp_f -> Some (comparator_ta_rule comp_q comp_f)));;

let rec ccompare_ta_rule _A _B =
  ({ccompare = ccompare_ta_rulea _A _B} : ('a, 'b) ta_rule ccompare);;

let rec cproper_interval_ta_rulea _A _B = (fun _ _ -> false);;

let rec cproper_interval_ta_rule _A _B =
  ({ccompare_cproper_interval = (ccompare_ta_rule _A _B);
     cproper_interval = cproper_interval_ta_rulea _A _B}
    : ('a, 'b) ta_rule cproper_interval);;

type ('a, 'b) mctxt = MVar of 'b | MHole | MFun of 'a * ('a, 'b) mctxt list;;

let rec equal_mctxt _A _B =
  ({equal = equal_mctxta _A _B} : ('a, 'b) mctxt equal)
and equal_mctxta _A _B
  x0 x1 = match x0, x1 with MHole, MFun (x31, x32) -> false
    | MFun (x31, x32), MHole -> false
    | MVar x1, MFun (x31, x32) -> false
    | MFun (x31, x32), MVar x1 -> false
    | MVar x1, MHole -> false
    | MHole, MVar x1 -> false
    | MFun (x31, x32), MFun (y31, y32) ->
        eq _A x31 y31 && equal_lista (equal_mctxt _A _B) x32 y32
    | MVar x1, MVar y1 -> eq _B x1 y1
    | MHole, MHole -> true;;

let rec ceq_mctxta _A _B = Some (equal_mctxta _A _B);;

let rec ceq_mctxt _A _B = ({ceq = ceq_mctxta _A _B} : ('a, 'b) mctxt ceq);;

let set_impl_mctxta : (('a, 'b) mctxt, set_impla) phantom = Phantom Set_RBT;;

let set_impl_mctxt = ({set_impl = set_impl_mctxta} : ('a, 'b) mctxt set_impl);;

let rec comparator_mctxt
  comp_f comp_v x2 x3 = match comp_f, comp_v, x2, x3 with
    comp_f, comp_v, MFun (x, xa), MFun (ya, yb) ->
      (match comp_f x ya
        with Eq -> comparator_list (comparator_mctxt comp_f comp_v) xa yb
        | Lt -> Lt | Gt -> Gt)
    | comp_f, comp_v, MFun (x, xa), MHole -> Gt
    | comp_f, comp_v, MFun (x, xa), MVar y -> Gt
    | comp_f, comp_v, MHole, MFun (ya, yb) -> Lt
    | comp_f, comp_v, MHole, MHole -> Eq
    | comp_f, comp_v, MHole, MVar y -> Gt
    | comp_f, comp_v, MVar x, MFun (ya, yb) -> Lt
    | comp_f, comp_v, MVar x, MHole -> Lt
    | comp_f, comp_v, MVar x, MVar y -> comp_v x y;;

let rec compare_mctxt _A _B = comparator_mctxt (compare _A) (compare _B);;

let rec ccompare_mctxta _A _B = Some (compare_mctxt _A _B);;

let rec ccompare_mctxt _A _B =
  ({ccompare = ccompare_mctxta _A _B} : ('a, 'b) mctxt ccompare);;

type arctic = MinInfty | Num_arc of int;;

let rec equal_arctica x0 x1 = match x0, x1 with MinInfty, Num_arc x2 -> false
                        | Num_arc x2, MinInfty -> false
                        | Num_arc x2, Num_arc y2 -> equal_inta x2 y2
                        | MinInfty, MinInfty -> true;;

let equal_arctic = ({equal = equal_arctica} : arctic equal);;

let one_arctica : arctic = Num_arc zero_inta;;

let one_arctic = ({onea = one_arctica} : arctic one);;

let rec plus_arctica x0 y = match x0, y with MinInfty, y -> y
                       | Num_arc v, MinInfty -> Num_arc v
                       | Num_arc x, Num_arc y -> Num_arc (max ord_int x y);;

let plus_arctic = ({plus = plus_arctica} : arctic plus);;

let zero_arctica : arctic = MinInfty;;

let zero_arctic = ({zerob = zero_arctica} : arctic zero);;

let semigroup_add_arctic =
  ({plus_semigroup_add = plus_arctic} : arctic semigroup_add);;

let numeral_arctic =
  ({one_numeral = one_arctic; semigroup_add_numeral = semigroup_add_arctic} :
    arctic numeral);;

let rec times_arctica x0 y = match x0, y with MinInfty, y -> MinInfty
                        | Num_arc v, MinInfty -> MinInfty
                        | Num_arc x, Num_arc y -> Num_arc (plus_inta x y);;

let times_arctic = ({times = times_arctica} : arctic times);;

let power_arctic =
  ({one_power = one_arctic; times_power = times_arctic} : arctic power);;

let rec less_eq_arctic xa0 x = match xa0, x with MinInfty, x -> true
                         | Num_arc uu, MinInfty -> false
                         | Num_arc y, Num_arc x -> less_eq_int y x;;

let rec less_arctic xa0 x = match xa0, x with MinInfty, x -> true
                      | Num_arc uu, MinInfty -> false
                      | Num_arc y, Num_arc x -> less_int y x;;

let ord_arctic = ({less_eq = less_eq_arctic; less = less_arctic} : arctic ord);;

let ab_semigroup_add_arctic =
  ({semigroup_add_ab_semigroup_add = semigroup_add_arctic} :
    arctic ab_semigroup_add);;

let semigroup_mult_arctic =
  ({times_semigroup_mult = times_arctic} : arctic semigroup_mult);;

let semiring_arctic =
  ({ab_semigroup_add_semiring = ab_semigroup_add_arctic;
     semigroup_mult_semiring = semigroup_mult_arctic}
    : arctic semiring);;

let mult_zero_arctic =
  ({times_mult_zero = times_arctic; zero_mult_zero = zero_arctic} :
    arctic mult_zero);;

let monoid_add_arctic =
  ({semigroup_add_monoid_add = semigroup_add_arctic;
     zero_monoid_add = zero_arctic}
    : arctic monoid_add);;

let comm_monoid_add_arctic =
  ({ab_semigroup_add_comm_monoid_add = ab_semigroup_add_arctic;
     monoid_add_comm_monoid_add = monoid_add_arctic}
    : arctic comm_monoid_add);;

let semiring_0_arctic =
  ({comm_monoid_add_semiring_0 = comm_monoid_add_arctic;
     mult_zero_semiring_0 = mult_zero_arctic;
     semiring_semiring_0 = semiring_arctic}
    : arctic semiring_0);;

let monoid_mult_arctic =
  ({semigroup_mult_monoid_mult = semigroup_mult_arctic;
     power_monoid_mult = power_arctic}
    : arctic monoid_mult);;

let semiring_numeral_arctic =
  ({monoid_mult_semiring_numeral = monoid_mult_arctic;
     numeral_semiring_numeral = numeral_arctic;
     semiring_semiring_numeral = semiring_arctic}
    : arctic semiring_numeral);;

let zero_neq_one_arctic =
  ({one_zero_neq_one = one_arctic; zero_zero_neq_one = zero_arctic} :
    arctic zero_neq_one);;

let semiring_1_arctic =
  ({semiring_numeral_semiring_1 = semiring_numeral_arctic;
     semiring_0_semiring_1 = semiring_0_arctic;
     zero_neq_one_semiring_1 = zero_neq_one_arctic}
    : arctic semiring_1);;

let ceq_arctica : (arctic -> arctic -> bool) option = Some equal_arctica;;

let ceq_arctic = ({ceq = ceq_arctica} : arctic ceq);;

let set_impl_arctica : (arctic, set_impla) phantom = Phantom Set_RBT;;

let set_impl_arctic = ({set_impl = set_impl_arctica} : arctic set_impl);;

let rec showsl_arctic = function MinInfty -> showsl_lit "-infinity"
                        | Num_arc i -> showsl_int i;;

let rec showsl_list_arctic xs = default_showsl_list showsl_arctic xs;;

let showl_arctic =
  ({showsl = showsl_arctic; showsl_list = showsl_list_arctic} : arctic showl);;

let cEnum_arctic :
  (arctic list *
    (((arctic -> bool) -> bool) * ((arctic -> bool) -> bool))) option
  = None;;

let cenum_arctic = ({cEnum = cEnum_arctic} : arctic cenum);;

let rec comparator_arctic
  x0 x1 = match x0, x1 with
    Num_arc x, Num_arc y -> comparator_of (equal_int, linorder_int) x y
    | Num_arc x, MinInfty -> Gt
    | MinInfty, Num_arc y -> Lt
    | MinInfty, MinInfty -> Eq;;

let rec compare_arctic x = comparator_arctic x;;

let ccompare_arctica : (arctic -> arctic -> ordera) option
  = Some compare_arctic;;

let ccompare_arctic = ({ccompare = ccompare_arctica} : arctic ccompare);;

let non_strict_order_arctic =
  ({ord_non_strict_order = ord_arctic} : arctic non_strict_order);;

let ordered_ab_semigroup_arctic =
  ({ab_semigroup_add_ordered_ab_semigroup = ab_semigroup_add_arctic;
     monoid_add_ordered_ab_semigroup = monoid_add_arctic;
     non_strict_order_ordered_ab_semigroup = non_strict_order_arctic}
    : arctic ordered_ab_semigroup);;

let ordered_semiring_0_arctic =
  ({semiring_0_ordered_semiring_0a = semiring_0_arctic;
     ordered_ab_semigroup_ordered_semiring_0 = ordered_ab_semigroup_arctic}
    : arctic ordered_semiring_0a);;

let ordered_semiring_1_arctic =
  ({semiring_1_ordered_semiring_1 = semiring_1_arctic;
     ordered_semiring_0_ordered_semiring_1 = ordered_semiring_0_arctic}
    : arctic ordered_semiring_1);;

type 'a filtered = FPair of 'a * nat;;

let rec equal_filtereda _A
  (FPair (x1, x2)) (FPair (y1, y2)) = eq _A x1 y1 && equal_nata x2 y2;;

let rec equal_filtered _A = ({equal = equal_filtereda _A} : 'a filtered equal);;

let rec comparator_filtered
  comp_f (FPair (x, xa)) (FPair (y, ya)) =
    (match comp_f x y with Eq -> comparator_of (equal_nat, linorder_nat) xa ya
      | Lt -> Lt | Gt -> Gt);;

let rec compare_filtereda _A = comparator_filtered (compare _A);;

let rec less_eq_filtered _A = le_of_comp (compare_filtereda _A);;

let rec less_filtered _A = lt_of_comp (compare_filtereda _A);;

let rec ord_filtered _A =
  ({less_eq = less_eq_filtered _A; less = less_filtered _A} : 'a filtered ord);;

let rec compare_filtered _A =
  ({compare = compare_filtereda _A} : 'a filtered compare);;

let rec quasi_order_filtered _A =
  ({ord_quasi_order = (ord_filtered _A)} : 'a filtered quasi_order);;

let rec weak_order_filtered _A =
  ({quasi_order_weak_order = (quasi_order_filtered _A)} :
    'a filtered weak_order);;

let rec preorder_filtered _A =
  ({ord_preorder = (ord_filtered _A)} : 'a filtered preorder);;

let rec order_filtered _A =
  ({preorder_order = (preorder_filtered _A);
     weak_order_order = (weak_order_filtered _A)}
    : 'a filtered order);;

let rec equality_filtered
  eq_f (FPair (x, xa)) (FPair (y, ya)) = eq_f x y && equal_nata xa ya;;

let rec ceq_filtereda _A
  = (match ceq _A with None -> None
      | Some eq_f -> Some (equality_filtered eq_f));;

let rec ceq_filtered _A = ({ceq = ceq_filtereda _A} : 'a filtered ceq);;

let rec set_impl_filtereda _A = Phantom (of_phantom (set_impl _A));;

let rec set_impl_filtered _A =
  ({set_impl = set_impl_filtereda _A} : 'a filtered set_impl);;

let rec linorder_filtered _A =
  ({order_linorder = (order_filtered _A)} : 'a filtered linorder);;

let rec filtered_fun (FPair (f, n)) = f;;

let rec showsl_filtered _A f = showsl _A (filtered_fun f);;

let rec showsl_list_filtered _A
  xs = default_showsl_list (showsl_filtered _A) xs;;

let rec showl_filtered _A =
  ({showsl = showsl_filtered _A; showsl_list = showsl_list_filtered _A} :
    'a filtered showl);;

let rec compare_order_filtered _A =
  ({compare_compare_order = (compare_filtered _A);
     linorder_compare_order = (linorder_filtered _A)}
    : 'a filtered compare_order);;

let rec ccompare_filtereda _A
  = (match ccompare _A with None -> None
      | Some comp_f -> Some (comparator_filtered comp_f));;

let rec ccompare_filtered _A =
  ({ccompare = ccompare_filtereda _A} : 'a filtered ccompare);;

type ('a, 'b) l_poly = LPoly of 'b * ('a * 'b) list;;

let cEnum_l_poly :
  (('a, 'b) l_poly list *
    (((('a, 'b) l_poly -> bool) -> bool) *
      ((('a, 'b) l_poly -> bool) -> bool))) option
  = None;;

let cenum_l_poly = ({cEnum = cEnum_l_poly} : ('a, 'b) l_poly cenum);;

type ('a, 'b) ctxt = Hole |
  More of 'a * ('a, 'b) term list * ('a, 'b) ctxt * ('a, 'b) term list;;

let rec equal_ctxta _A _B
  x0 x1 = match x0, x1 with Hole, More (x21, x22, x23, x24) -> false
    | More (x21, x22, x23, x24), Hole -> false
    | More (x21, x22, x23, x24), More (y21, y22, y23, y24) ->
        eq _A x21 y21 &&
          (equal_lista (equal_term _A _B) x22 y22 &&
            (equal_ctxta _A _B x23 y23 &&
              equal_lista (equal_term _A _B) x24 y24))
    | Hole, Hole -> true;;

let rec equal_ctxt _A _B = ({equal = equal_ctxta _A _B} : ('a, 'b) ctxt equal);;

let rec ceq_ctxta _A _B = Some (equal_ctxta _A _B);;

let rec ceq_ctxt _A _B = ({ceq = ceq_ctxta _A _B} : ('a, 'b) ctxt ceq);;

let set_impl_ctxta : (('a, 'b) ctxt, set_impla) phantom = Phantom Set_RBT;;

let set_impl_ctxt = ({set_impl = set_impl_ctxta} : ('a, 'b) ctxt set_impl);;

let rec comparator_ctxt
  comp_f comp_v x2 x3 = match comp_f, comp_v, x2, x3 with
    comp_f, comp_v, More (x, xa, xb, xc), More (y, ya, yb, yc) ->
      (match comp_f x y
        with Eq ->
          (match comparator_list (comparator_term comp_f comp_v) xa ya
            with Eq ->
              (match comparator_ctxt comp_f comp_v xb yb
                with Eq -> comparator_list (comparator_term comp_f comp_v) xc yc
                | Lt -> Lt | Gt -> Gt)
            | Lt -> Lt | Gt -> Gt)
        | Lt -> Lt | Gt -> Gt)
    | comp_f, comp_v, More (x, xa, xb, xc), Hole -> Gt
    | comp_f, comp_v, Hole, More (y, ya, yb, yc) -> Lt
    | comp_f, comp_v, Hole, Hole -> Eq;;

let rec compare_ctxt _A _B = comparator_ctxt (compare _A) (compare _B);;

let rec ccompare_ctxta _A _B = Some (compare_ctxt _A _B);;

let rec ccompare_ctxt _A _B =
  ({ccompare = ccompare_ctxta _A _B} : ('a, 'b) ctxt ccompare);;

type 'a sharp = Flat of 'a | Sharpa of 'a;;

let rec equal_sharpa _A x0 x1 = match x0, x1 with Flat x1, Sharpa x2 -> false
                          | Sharpa x2, Flat x1 -> false
                          | Sharpa x2, Sharpa y2 -> eq _A x2 y2
                          | Flat x1, Flat y1 -> eq _A x1 y1;;

let rec equal_sharp _A = ({equal = equal_sharpa _A} : 'a sharp equal);;

let rec comparator_sharp
  comp_l x1 x2 = match comp_l, x1, x2 with
    comp_l, Sharpa x, Sharpa ya -> comp_l x ya
    | comp_l, Sharpa x, Flat y -> Gt
    | comp_l, Flat x, Sharpa ya -> Lt
    | comp_l, Flat x, Flat y -> comp_l x y;;

let rec compare_sharpa _A = comparator_sharp (compare _A);;

let rec less_eq_sharp _A = le_of_comp (compare_sharpa _A);;

let rec less_sharp _A = lt_of_comp (compare_sharpa _A);;

let rec ord_sharp _A =
  ({less_eq = less_eq_sharp _A; less = less_sharp _A} : 'a sharp ord);;

let rec compare_sharp _A = ({compare = compare_sharpa _A} : 'a sharp compare);;

let rec quasi_order_sharp _A =
  ({ord_quasi_order = (ord_sharp _A)} : 'a sharp quasi_order);;

let rec weak_order_sharp _A =
  ({quasi_order_weak_order = (quasi_order_sharp _A)} : 'a sharp weak_order);;

let rec preorder_sharp _A =
  ({ord_preorder = (ord_sharp _A)} : 'a sharp preorder);;

let rec order_sharp _A =
  ({preorder_order = (preorder_sharp _A);
     weak_order_order = (weak_order_sharp _A)}
    : 'a sharp order);;

let rec ceq_sharpa _A = Some (equal_sharpa _A);;

let rec ceq_sharp _A = ({ceq = ceq_sharpa _A} : 'a sharp ceq);;

let set_impl_sharpa : ('a sharp, set_impla) phantom = Phantom Set_RBT;;

let set_impl_sharp = ({set_impl = set_impl_sharpa} : 'a sharp set_impl);;

let rec linorder_sharp _A =
  ({order_linorder = (order_sharp _A)} : 'a sharp linorder);;

let rec showsl_sharp _A = function Flat s -> showsl _A s
                          | Sharpa s -> comp (showsl_lit "#") (showsl _A s);;

let rec showsl_list_sharp _A xs = default_showsl_list (showsl_sharp _A) xs;;

let rec showl_sharp _A =
  ({showsl = showsl_sharp _A; showsl_list = showsl_list_sharp _A} :
    'a sharp showl);;

let cEnum_sharp :
  ('a sharp list *
    ((('a sharp -> bool) -> bool) * (('a sharp -> bool) -> bool))) option
  = None;;

let cenum_sharp = ({cEnum = cEnum_sharp} : 'a sharp cenum);;

let rec compare_order_sharp _A =
  ({compare_compare_order = (compare_sharp _A);
     linorder_compare_order = (linorder_sharp _A)}
    : 'a sharp compare_order);;

let rec ccompare_sharpa _A
  = (match ccompare _A with None -> None
      | Some comp_l -> Some (comparator_sharp comp_l));;

let rec ccompare_sharp _A =
  ({ccompare = ccompare_sharpa _A} : 'a sharp ccompare);;

let mapping_impl_sharpa : ('a sharp, mapping_impla) phantom
  = Phantom Mapping_RBT;;

let mapping_impl_sharp =
  ({mapping_impl = mapping_impl_sharpa} : 'a sharp mapping_impl);;

type location = H | A | Ba | Ra;;

let rec equal_locationa x0 x1 = match x0, x1 with Ba, Ra -> false
                          | Ra, Ba -> false
                          | A, Ra -> false
                          | Ra, A -> false
                          | A, Ba -> false
                          | Ba, A -> false
                          | H, Ra -> false
                          | Ra, H -> false
                          | H, Ba -> false
                          | Ba, H -> false
                          | H, A -> false
                          | A, H -> false
                          | Ra, Ra -> true
                          | Ba, Ba -> true
                          | A, A -> true
                          | H, H -> true;;

let equal_location = ({equal = equal_locationa} : location equal);;

let ceq_locationa : (location -> location -> bool) option
  = Some equal_locationa;;

let ceq_location = ({ceq = ceq_locationa} : location ceq);;

let set_impl_locationa : (location, set_impla) phantom = Phantom Set_RBT;;

let set_impl_location = ({set_impl = set_impl_locationa} : location set_impl);;

let rec comparator_location x0 x1 = match x0, x1 with Ra, Ra -> Eq
                              | Ra, Ba -> Gt
                              | Ra, A -> Gt
                              | Ra, H -> Gt
                              | Ba, Ra -> Lt
                              | Ba, Ba -> Eq
                              | Ba, A -> Gt
                              | Ba, H -> Gt
                              | A, Ra -> Lt
                              | A, Ba -> Lt
                              | A, A -> Eq
                              | A, H -> Gt
                              | H, Ra -> Lt
                              | H, Ba -> Lt
                              | H, A -> Lt
                              | H, H -> Eq;;

let rec compare_location x = comparator_location x;;

let ccompare_locationa : (location -> location -> ordera) option
  = Some compare_location;;

let ccompare_location = ({ccompare = ccompare_locationa} : location ccompare);;

type ('a, 'b) scg = Null | Scg of 'a * 'a * ('b * 'b) list * ('b * 'b) list;;

let rec equal_scga _A _B
  x0 x1 = match x0, x1 with Null, Scg (x21, x22, x23, x24) -> false
    | Scg (x21, x22, x23, x24), Null -> false
    | Scg (x21, x22, x23, x24), Scg (y21, y22, y23, y24) ->
        eq _A x21 y21 &&
          (eq _A x22 y22 &&
            (equal_lista (equal_prod _B _B) x23 y23 &&
              equal_lista (equal_prod _B _B) x24 y24))
    | Null, Null -> true;;

let rec equal_scg _A _B = ({equal = equal_scga _A _B} : ('a, 'b) scg equal);;

let rec comparator_scg
  comp_p_p comp_a_p x2 x3 = match comp_p_p, comp_a_p, x2, x3 with
    comp_p_p, comp_a_p, Scg (x, xa, xb, xc), Scg (y, ya, yb, yc) ->
      (match comp_p_p x y
        with Eq ->
          (match comp_p_p xa ya
            with Eq ->
              (match comparator_list (comparator_prod comp_a_p comp_a_p) xb yb
                with Eq ->
                  comparator_list (comparator_prod comp_a_p comp_a_p) xc yc
                | Lt -> Lt | Gt -> Gt)
            | Lt -> Lt | Gt -> Gt)
        | Lt -> Lt | Gt -> Gt)
    | comp_p_p, comp_a_p, Scg (x, xa, xb, xc), Null -> Gt
    | comp_p_p, comp_a_p, Null, Scg (y, ya, yb, yc) -> Lt
    | comp_p_p, comp_a_p, Null, Null -> Eq;;

let rec compare_scga _A _B = comparator_scg (compare _A) (compare _B);;

let rec less_eq_scg _A _B = le_of_comp (compare_scga _A _B);;

let rec less_scg _A _B = lt_of_comp (compare_scga _A _B);;

let rec ord_scg _A _B =
  ({less_eq = less_eq_scg _A _B; less = less_scg _A _B} : ('a, 'b) scg ord);;

let rec compare_scg _A _B =
  ({compare = compare_scga _A _B} : ('a, 'b) scg compare);;

let rec quasi_order_scg _A _B =
  ({ord_quasi_order = (ord_scg _A _B)} : ('a, 'b) scg quasi_order);;

let rec weak_order_scg _A _B =
  ({quasi_order_weak_order = (quasi_order_scg _A _B)} :
    ('a, 'b) scg weak_order);;

let rec preorder_scg _A _B =
  ({ord_preorder = (ord_scg _A _B)} : ('a, 'b) scg preorder);;

let rec order_scg _A _B =
  ({preorder_order = (preorder_scg _A _B);
     weak_order_order = (weak_order_scg _A _B)}
    : ('a, 'b) scg order);;

let rec linorder_scg _A _B =
  ({order_linorder = (order_scg _A _B)} : ('a, 'b) scg linorder);;

let rec compare_order_scg _A _B =
  ({compare_compare_order = (compare_scg _A _B);
     linorder_compare_order = (linorder_scg _A _B)}
    : ('a, 'b) scg compare_order);;

type 'a arctic_delta = MinInfty_delta | Num_arc_delta of 'a;;

let rec equal_arctic_deltaa _A
  x0 x1 = match x0, x1 with MinInfty_delta, Num_arc_delta x2 -> false
    | Num_arc_delta x2, MinInfty_delta -> false
    | Num_arc_delta x2, Num_arc_delta y2 -> eq _A x2 y2
    | MinInfty_delta, MinInfty_delta -> true;;

let rec equal_arctic_delta _A =
  ({equal = equal_arctic_deltaa _A} : 'a arctic_delta equal);;

let rec one_arctic_deltaa _A
  = Num_arc_delta
      (zerob
        _A.linordered_idom_linordered_field.linordered_ring_strict_linordered_idom.linordered_ring_linordered_ring_strict.abs_if_linordered_ring.zero_abs_if);;

let rec one_arctic_delta _A =
  ({onea = one_arctic_deltaa _A} : 'a arctic_delta one);;

let rec plus_arctic_deltaa _A
  x0 y = match x0, y with MinInfty_delta, y -> y
    | Num_arc_delta v, MinInfty_delta -> Num_arc_delta v
    | Num_arc_delta x, Num_arc_delta y ->
        Num_arc_delta
          (max _A.linordered_idom_linordered_field.linordered_ring_strict_linordered_idom.linordered_ring_linordered_ring_strict.abs_if_linordered_ring.ord_abs_if
            x y);;

let rec plus_arctic_delta _A =
  ({plus = plus_arctic_deltaa _A} : 'a arctic_delta plus);;

let rec zero_arctic_deltaa _A = MinInfty_delta;;

let rec zero_arctic_delta _A =
  ({zerob = zero_arctic_deltaa _A} : 'a arctic_delta zero);;

let rec semigroup_add_arctic_delta _A =
  ({plus_semigroup_add = (plus_arctic_delta _A)} :
    'a arctic_delta semigroup_add);;

let rec numeral_arctic_delta _A =
  ({one_numeral = (one_arctic_delta _A);
     semigroup_add_numeral = (semigroup_add_arctic_delta _A)}
    : 'a arctic_delta numeral);;

let rec times_arctic_deltaa _A
  x0 y = match x0, y with MinInfty_delta, y -> MinInfty_delta
    | Num_arc_delta v, MinInfty_delta -> MinInfty_delta
    | Num_arc_delta x, Num_arc_delta y ->
        Num_arc_delta
          (plus _A.field_char_0_linordered_field.ring_char_0_field_char_0.ring_1_ring_char_0.neg_numeral_ring_1.numeral_neg_numeral.semigroup_add_numeral.plus_semigroup_add
            x y);;

let rec times_arctic_delta _A =
  ({times = times_arctic_deltaa _A} : 'a arctic_delta times);;

let rec power_arctic_delta _A =
  ({one_power = (one_arctic_delta _A); times_power = (times_arctic_delta _A)} :
    'a arctic_delta power);;

let rec less_eq_arctic_delta _A
  xa0 x = match xa0, x with MinInfty_delta, x -> true
    | Num_arc_delta uu, MinInfty_delta -> false
    | Num_arc_delta y, Num_arc_delta x -> less_eq _A y x;;

let rec less_arctic_delta _A
  xa0 x = match xa0, x with MinInfty_delta, x -> true
    | Num_arc_delta uu, MinInfty_delta -> false
    | Num_arc_delta y, Num_arc_delta x -> less _A y x;;

let rec ord_arctic_delta _A =
  ({less_eq = less_eq_arctic_delta _A; less = less_arctic_delta _A} :
    'a arctic_delta ord);;

let rec ab_semigroup_add_arctic_delta _A =
  ({semigroup_add_ab_semigroup_add = (semigroup_add_arctic_delta _A)} :
    'a arctic_delta ab_semigroup_add);;

let rec semigroup_mult_arctic_delta _A =
  ({times_semigroup_mult = (times_arctic_delta _A)} :
    'a arctic_delta semigroup_mult);;

let rec semiring_arctic_delta _A =
  ({ab_semigroup_add_semiring = (ab_semigroup_add_arctic_delta _A);
     semigroup_mult_semiring = (semigroup_mult_arctic_delta _A)}
    : 'a arctic_delta semiring);;

let rec mult_zero_arctic_delta _A =
  ({times_mult_zero = (times_arctic_delta _A);
     zero_mult_zero = (zero_arctic_delta _A)}
    : 'a arctic_delta mult_zero);;

let rec monoid_add_arctic_delta _A =
  ({semigroup_add_monoid_add = (semigroup_add_arctic_delta _A);
     zero_monoid_add = (zero_arctic_delta _A)}
    : 'a arctic_delta monoid_add);;

let rec comm_monoid_add_arctic_delta _A =
  ({ab_semigroup_add_comm_monoid_add = (ab_semigroup_add_arctic_delta _A);
     monoid_add_comm_monoid_add = (monoid_add_arctic_delta _A)}
    : 'a arctic_delta comm_monoid_add);;

let rec semiring_0_arctic_delta _A =
  ({comm_monoid_add_semiring_0 = (comm_monoid_add_arctic_delta _A);
     mult_zero_semiring_0 = (mult_zero_arctic_delta _A);
     semiring_semiring_0 = (semiring_arctic_delta _A)}
    : 'a arctic_delta semiring_0);;

let rec monoid_mult_arctic_delta _A =
  ({semigroup_mult_monoid_mult = (semigroup_mult_arctic_delta _A);
     power_monoid_mult = (power_arctic_delta _A)}
    : 'a arctic_delta monoid_mult);;

let rec semiring_numeral_arctic_delta _A =
  ({monoid_mult_semiring_numeral = (monoid_mult_arctic_delta _A);
     numeral_semiring_numeral = (numeral_arctic_delta _A);
     semiring_semiring_numeral = (semiring_arctic_delta _A)}
    : 'a arctic_delta semiring_numeral);;

let rec zero_neq_one_arctic_delta _A =
  ({one_zero_neq_one = (one_arctic_delta _A);
     zero_zero_neq_one = (zero_arctic_delta _A)}
    : 'a arctic_delta zero_neq_one);;

let rec semiring_1_arctic_delta _A =
  ({semiring_numeral_semiring_1 = (semiring_numeral_arctic_delta _A);
     semiring_0_semiring_1 = (semiring_0_arctic_delta _A);
     zero_neq_one_semiring_1 = (zero_neq_one_arctic_delta _A)}
    : 'a arctic_delta semiring_1);;

let rec ceq_arctic_deltaa _A = Some (equal_arctic_deltaa _A);;

let rec ceq_arctic_delta _A =
  ({ceq = ceq_arctic_deltaa _A} : 'a arctic_delta ceq);;

let set_impl_arctic_deltaa : ('a arctic_delta, set_impla) phantom
  = Phantom Set_RBT;;

let set_impl_arctic_delta =
  ({set_impl = set_impl_arctic_deltaa} : 'a arctic_delta set_impl);;

let rec showsl_arctic_delta _A
  = function MinInfty_delta -> showsl_lit "-infinity"
    | Num_arc_delta x -> showsl _A x;;

let rec showsl_list_arctic_delta _A
  xs = default_showsl_list (showsl_arctic_delta _A) xs;;

let rec showl_arctic_delta _A =
  ({showsl = showsl_arctic_delta _A; showsl_list = showsl_list_arctic_delta _A}
    : 'a arctic_delta showl);;

let cEnum_arctic_delta :
  ('a arctic_delta list *
    ((('a arctic_delta -> bool) -> bool) *
      (('a arctic_delta -> bool) -> bool))) option
  = None;;

let cenum_arctic_delta =
  ({cEnum = cEnum_arctic_delta} : 'a arctic_delta cenum);;

let rec comparator_arctic_delta
  comp_a x1 x2 = match comp_a, x1, x2 with
    comp_a, Num_arc_delta x, Num_arc_delta y -> comp_a x y
    | comp_a, Num_arc_delta x, MinInfty_delta -> Gt
    | comp_a, MinInfty_delta, Num_arc_delta y -> Lt
    | comp_a, MinInfty_delta, MinInfty_delta -> Eq;;

let rec compare_arctic_delta _A = comparator_arctic_delta (compare _A);;

let rec ccompare_arctic_deltaa _A = Some (compare_arctic_delta _A);;

let rec ccompare_arctic_delta _A =
  ({ccompare = ccompare_arctic_deltaa _A} : 'a arctic_delta ccompare);;

let rec non_strict_order_arctic_delta _A =
  ({ord_non_strict_order =
      (ord_arctic_delta
        _A.linordered_idom_linordered_field.linordered_ring_strict_linordered_idom.linordered_ring_linordered_ring_strict.abs_if_linordered_ring.ord_abs_if)}
    : 'a arctic_delta non_strict_order);;

let rec ordered_ab_semigroup_arctic_delta _A =
  ({ab_semigroup_add_ordered_ab_semigroup = (ab_semigroup_add_arctic_delta _A);
     monoid_add_ordered_ab_semigroup = (monoid_add_arctic_delta _A);
     non_strict_order_ordered_ab_semigroup = (non_strict_order_arctic_delta _A)}
    : 'a arctic_delta ordered_ab_semigroup);;

let rec ordered_semiring_0_arctic_delta _A =
  ({semiring_0_ordered_semiring_0a = (semiring_0_arctic_delta _A);
     ordered_ab_semigroup_ordered_semiring_0 =
       (ordered_ab_semigroup_arctic_delta _A)}
    : 'a arctic_delta ordered_semiring_0a);;

let rec ordered_semiring_1_arctic_delta _A =
  ({semiring_1_ordered_semiring_1 = (semiring_1_arctic_delta _A);
     ordered_semiring_0_ordered_semiring_1 =
       (ordered_semiring_0_arctic_delta _A)}
    : 'a arctic_delta ordered_semiring_1);;

type ty = BoolT | IntT;;

let rec equal_tya x0 x1 = match x0, x1 with BoolT, IntT -> false
                    | IntT, BoolT -> false
                    | IntT, IntT -> true
                    | BoolT, BoolT -> true;;

let equal_ty = ({equal = equal_tya} : ty equal);;

let ceq_tya : (ty -> ty -> bool) option = Some equal_tya;;

let ceq_ty = ({ceq = ceq_tya} : ty ceq);;

let rec showsl_ty = function BoolT -> showsl_lit "Bool"
                    | IntT -> showsl_lit "Int";;

let rec showsl_list_ty xs = default_showsl_list showsl_ty xs;;

let showl_ty = ({showsl = showsl_ty; showsl_list = showsl_list_ty} : ty showl);;

let rec comparator_ty x0 x1 = match x0, x1 with IntT, IntT -> Eq
                        | IntT, BoolT -> Gt
                        | BoolT, IntT -> Lt
                        | BoolT, BoolT -> Eq;;

let ccompare_tya : (ty -> ty -> ordera) option = Some comparator_ty;;

let ccompare_ty = ({ccompare = ccompare_tya} : ty ccompare);;

let rec rel_option r x1 x2 = match r, x1, x2 with r, None, Some y2 -> false
                     | r, Some y2, None -> false
                     | r, None, None -> true
                     | r, Some x2, Some y2 -> r x2 y2;;

type ('a, 'b) fmap = Fmap_of_list of ('a * 'b) list;;

let rec fmlookup _A (Fmap_of_list m) = map_of _A m;;

type 'a fset = Abs_fset of 'a set;;

let rec fset_of_list (_A1, _A2, _A3) xa = Abs_fset (set (_A1, _A2, _A3) xa);;

let rec fset (Abs_fset x) = x;;

let rec foldb _A x xc = folda (fun a _ -> x a) (impl_ofb _A xc);;

let rec image (_A1, _A2) (_B1, _B2, _B3)
  h x1 = match h, x1 with
    h, RBT_set rbt ->
      (match ccompare _A2
        with None ->
          failwith "image RBT_set: ccompare = None"
            (fun _ -> image (_A1, _A2) (_B1, _B2, _B3) h (RBT_set rbt))
        | Some _ ->
          foldb _A2 (comp (inserta (_B1, _B2)) h) rbt (bot_set (_B1, _B2, _B3)))
    | g, DList_set dxs ->
        (match ceq _A1
          with None ->
            failwith "image DList_set: ceq = None"
              (fun _ -> image (_A1, _A2) (_B1, _B2, _B3) g (DList_set dxs))
          | Some _ ->
            foldc _A1 (comp (inserta (_B1, _B2)) g) dxs
              (bot_set (_B1, _B2, _B3)))
    | f, Complement (Complement b) -> image (_A1, _A2) (_B1, _B2, _B3) f b
    | f, Collect_set a ->
        failwith "image Collect_set"
          (fun _ -> image (_A1, _A2) (_B1, _B2, _B3) f (Collect_set a))
    | f, Set_Monad xs -> Set_Monad (map f xs);;

let rec fimage (_B1, _B2) (_A1, _A2, _A3)
  xb xc = Abs_fset (image (_B1, _B2) (_A1, _A2, _A3) xb (fset xc));;

let rec fmdom (_A1, _A2, _A3) (_B1, _B2, _B3)
  (Fmap_of_list m) =
    fimage ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2)) (_A1, _A2, _A3) fst
      (fset_of_list
        ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2), (set_impl_prod _A3 _B3))
        m);;

let rec rBT_Impl_rbt_all
  p x1 = match p, x1 with
    p, Branch (c, l, k, v, r) ->
      p k v && (rBT_Impl_rbt_all p l && rBT_Impl_rbt_all p r)
    | p, Empty -> true;;

let rec all _A xb xc = rBT_Impl_rbt_all xb (impl_ofb _A xc);;

let rec ball (_A1, _A2)
  x0 p = match x0, p with
    RBT_set rbt, p ->
      (match ccompare _A2
        with None ->
          failwith "Ball RBT_set: ccompare = None"
            (fun _ -> ball (_A1, _A2) (RBT_set rbt) p)
        | Some _ -> all _A2 (fun k _ -> p k) rbt)
    | DList_set dxs, p ->
        (match ceq _A1
          with None ->
            failwith "Ball DList_set: ceq = None"
              (fun _ -> ball (_A1, _A2) (DList_set dxs) p)
          | Some _ -> dlist_all _A1 p dxs)
    | Set_Monad xs, p -> list_all p xs;;

let rec fBall (_A1, _A2) xa = ball (_A1, _A2) (fset xa);;

let rec fmrel (_A1, _A2, _A3) (_B1, _B2, _B3) (_C1, _C2, _C3, _C4)
  r m n =
    fBall (_C1, _C2) (fmdom (_C1, _C2, _C4) (_A1, _A2, _A3) m)
      (fun x -> rel_option r (fmlookup _C3 m x) (fmlookup _C3 n x)) &&
      fBall (_C1, _C2) (fmdom (_C1, _C2, _C4) (_B1, _B2, _B3) n)
        (fun x -> rel_option r (fmlookup _C3 m x) (fmlookup _C3 n x));;

let rec equal_fmap (_A1, _A2, _A3, _A4) (_B1, _B2, _B3, _B4)
  = fmrel (_B1, _B2, _B4) (_B1, _B2, _B4) (_A1, _A2, _A3, _A4) (equal _B3);;

type linear_poly = LinearPoly of (nat, rat) fmap;;

let rec linear_poly_map (LinearPoly x) = x;;

let rec equal_linear_polya
  x y = equal_fmap (ceq_nat, ccompare_nat, equal_nat, set_impl_nat)
          (ceq_rat, ccompare_rat, equal_rat, set_impl_rat) (linear_poly_map x)
          (linear_poly_map y);;

let equal_linear_poly = ({equal = equal_linear_polya} : linear_poly equal);;

type siga = LessF | LeF | SumF of nat | ConstF of int | ProdF of nat | EqF;;

let rec equal_sigc x0 x1 = match x0, x1 with ProdF x5, EqF -> false
                     | EqF, ProdF x5 -> false
                     | ConstF x4, EqF -> false
                     | EqF, ConstF x4 -> false
                     | ConstF x4, ProdF x5 -> false
                     | ProdF x5, ConstF x4 -> false
                     | SumF x3, EqF -> false
                     | EqF, SumF x3 -> false
                     | SumF x3, ProdF x5 -> false
                     | ProdF x5, SumF x3 -> false
                     | SumF x3, ConstF x4 -> false
                     | ConstF x4, SumF x3 -> false
                     | LeF, EqF -> false
                     | EqF, LeF -> false
                     | LeF, ProdF x5 -> false
                     | ProdF x5, LeF -> false
                     | LeF, ConstF x4 -> false
                     | ConstF x4, LeF -> false
                     | LeF, SumF x3 -> false
                     | SumF x3, LeF -> false
                     | LessF, EqF -> false
                     | EqF, LessF -> false
                     | LessF, ProdF x5 -> false
                     | ProdF x5, LessF -> false
                     | LessF, ConstF x4 -> false
                     | ConstF x4, LessF -> false
                     | LessF, SumF x3 -> false
                     | SumF x3, LessF -> false
                     | LessF, LeF -> false
                     | LeF, LessF -> false
                     | ProdF x5, ProdF y5 -> equal_nata x5 y5
                     | ConstF x4, ConstF y4 -> equal_inta x4 y4
                     | SumF x3, SumF y3 -> equal_nata x3 y3
                     | EqF, EqF -> true
                     | LeF, LeF -> true
                     | LessF, LessF -> true;;

let equal_sig = ({equal = equal_sigc} : siga equal);;

let rec showsl_sig = function LessF -> showsl_lit "<"
                     | LeF -> showsl_lit "<="
                     | EqF -> showsl_lit "="
                     | SumF n -> showsl_lit "+"
                     | ProdF n -> showsl_lit "*"
                     | ConstF n -> showsl_int n;;

let rec showsl_list_sig xs = default_showsl_list showsl_sig xs;;

let showl_sig =
  ({showsl = showsl_sig; showsl_list = showsl_list_sig} : siga showl);;

let rec comparator_siga
  x0 x1 = match x0, x1 with EqF, EqF -> Eq
    | EqF, ProdF yb -> Gt
    | EqF, ConstF ya -> Gt
    | EqF, SumF y -> Gt
    | EqF, LeF -> Gt
    | EqF, LessF -> Gt
    | ProdF x, EqF -> Lt
    | ProdF x, ProdF yb -> comparator_of (equal_nat, linorder_nat) x yb
    | ProdF x, ConstF ya -> Gt
    | ProdF x, SumF y -> Gt
    | ProdF x, LeF -> Gt
    | ProdF x, LessF -> Gt
    | ConstF x, EqF -> Lt
    | ConstF x, ProdF yb -> Lt
    | ConstF x, ConstF ya -> comparator_of (equal_int, linorder_int) x ya
    | ConstF x, SumF y -> Gt
    | ConstF x, LeF -> Gt
    | ConstF x, LessF -> Gt
    | SumF x, EqF -> Lt
    | SumF x, ProdF yb -> Lt
    | SumF x, ConstF ya -> Lt
    | SumF x, SumF y -> comparator_of (equal_nat, linorder_nat) x y
    | SumF x, LeF -> Gt
    | SumF x, LessF -> Gt
    | LeF, EqF -> Lt
    | LeF, ProdF yb -> Lt
    | LeF, ConstF ya -> Lt
    | LeF, SumF y -> Lt
    | LeF, LeF -> Eq
    | LeF, LessF -> Gt
    | LessF, EqF -> Lt
    | LessF, ProdF yb -> Lt
    | LessF, ConstF ya -> Lt
    | LessF, SumF y -> Lt
    | LessF, LeF -> Lt
    | LessF, LessF -> Eq;;

let ccompare_siga : (siga -> siga -> ordera) option = Some comparator_siga;;

let ccompare_sig = ({ccompare = ccompare_siga} : siga ccompare);;

type sigb = ConstFa of nat | SumFa | ProdFa | MaxF;;

let rec equal_sigb x0 x1 = match x0, x1 with ProdFa, MaxF -> false
                     | MaxF, ProdFa -> false
                     | SumFa, MaxF -> false
                     | MaxF, SumFa -> false
                     | SumFa, ProdFa -> false
                     | ProdFa, SumFa -> false
                     | ConstFa x1, MaxF -> false
                     | MaxF, ConstFa x1 -> false
                     | ConstFa x1, ProdFa -> false
                     | ProdFa, ConstFa x1 -> false
                     | ConstFa x1, SumFa -> false
                     | SumFa, ConstFa x1 -> false
                     | ConstFa x1, ConstFa y1 -> equal_nata x1 y1
                     | MaxF, MaxF -> true
                     | ProdFa, ProdFa -> true
                     | SumFa, SumFa -> true;;

let equal_siga = ({equal = equal_sigb} : sigb equal);;

let rec comparator_sig
  x0 x1 = match x0, x1 with MaxF, MaxF -> Eq
    | MaxF, ProdFa -> Gt
    | MaxF, SumFa -> Gt
    | MaxF, ConstFa y -> Gt
    | ProdFa, MaxF -> Lt
    | ProdFa, ProdFa -> Eq
    | ProdFa, SumFa -> Gt
    | ProdFa, ConstFa y -> Gt
    | SumFa, MaxF -> Lt
    | SumFa, ProdFa -> Lt
    | SumFa, SumFa -> Eq
    | SumFa, ConstFa y -> Gt
    | ConstFa x, MaxF -> Lt
    | ConstFa x, ProdFa -> Lt
    | ConstFa x, SumFa -> Lt
    | ConstFa x, ConstFa y -> comparator_of (equal_nat, linorder_nat) x y;;

let rec compare_siga x = comparator_sig x;;

let compare_sig = ({compare = compare_siga} : sigb compare);;

type ('a, 'b) cond_constraint =
  CC_cond of bool * (('a, 'b) term * ('a, 'b) term) |
  CC_rewr of ('a, 'b) term * ('a, 'b) term |
  CC_impl of ('a, 'b) cond_constraint list * ('a, 'b) cond_constraint |
  CC_all of 'b * ('a, 'b) cond_constraint;;

let rec equal_cond_constraint _A _B =
  ({equal = equal_cond_constrainta _A _B} : ('a, 'b) cond_constraint equal)
and equal_cond_constrainta _A _B
  x0 x1 = match x0, x1 with CC_impl (x31, x32), CC_all (x41, x42) -> false
    | CC_all (x41, x42), CC_impl (x31, x32) -> false
    | CC_rewr (x21, x22), CC_all (x41, x42) -> false
    | CC_all (x41, x42), CC_rewr (x21, x22) -> false
    | CC_rewr (x21, x22), CC_impl (x31, x32) -> false
    | CC_impl (x31, x32), CC_rewr (x21, x22) -> false
    | CC_cond (x11, x12), CC_all (x41, x42) -> false
    | CC_all (x41, x42), CC_cond (x11, x12) -> false
    | CC_cond (x11, x12), CC_impl (x31, x32) -> false
    | CC_impl (x31, x32), CC_cond (x11, x12) -> false
    | CC_cond (x11, x12), CC_rewr (x21, x22) -> false
    | CC_rewr (x21, x22), CC_cond (x11, x12) -> false
    | CC_all (x41, x42), CC_all (y41, y42) ->
        eq _B x41 y41 && equal_cond_constrainta _A _B x42 y42
    | CC_impl (x31, x32), CC_impl (y31, y32) ->
        equal_lista (equal_cond_constraint _A _B) x31 y31 &&
          equal_cond_constrainta _A _B x32 y32
    | CC_rewr (x21, x22), CC_rewr (y21, y22) ->
        equal_terma _A _B x21 y21 && equal_terma _A _B x22 y22
    | CC_cond (x11, x12), CC_cond (y11, y12) ->
        equal_boola x11 y11 &&
          equal_proda (equal_term _A _B) (equal_term _A _B) x12 y12;;

type 'a itself = Type;;

type ('b, 'a) rbt = RBT of ('b, 'a) rbta;;

type xml = XML of char list * (char list * char list) list * xml list |
  XML_text of char list;;

type 'a vec_impl = Abs_vec_impl of (nat * 'a iarray);;

type 'a vec = Vec_impl of 'a vec_impl;;

type xmldoc = XMLDOC of (char list) list * xml;;

type ('a, 'b, 'c, 'd, 'e) lts_impl =
  Lts_Impl of
    'd list * ('e * ('a, 'b, 'c, 'd) transition_rule) list *
      ('d * ('a, ('b * 'c)) term formula) list;;

type order_tag = Lex | Mul;;

type ('a, 'b) mapping = Assoc_List_Mapping of ('a, 'b) alist |
  RBT_Mapping of ('a, 'b) mapping_rbt | Mapping of ('a -> 'b option);;

type ('a, 'b) state =
  State of
    (nat * linear_poly) list * (nat, ('a * 'b)) mapping *
      (nat, ('a * 'b)) mapping * (nat, 'b) mapping * bool * ('a list) option;;

type 'a istate =
  IState of
    nat * (nat * linear_poly) list * ('a * qDelta atom) list *
      (linear_poly -> nat option);;

type ('a, 'b, 'c, 'd, 'e) lts_ext =
  Lts_ext of
    'd set * ('a, 'b, 'c, 'd) transition_rule set *
      ('d -> ('a, ('b * 'c)) term formula) * 'e;;

type 'a xml_error = TagMismatch of string list | Fatal of 'a;;

type 'a array = Array of 'a list;;

type domain = Natural of nat | Integer | Arctic | Arctic_rat |
  Int_mat of nat * nat | Arctic_mat of nat | Arctic_rat_mat of nat |
  Rational of rat * nat | Rat_mat of nat * nat | Mini_Alg of real * nat |
  Mini_Alg_mat of nat * nat;;

type enat = Enat of nat | Infinity_enat;;

type ('a, 'b) tpoly = PVar of 'a | PNum of 'b | PSum of ('a, 'b) tpoly list |
  PMult of ('a, 'b) tpoly list;;

type ('a, 'b) direction =
  Direction of
    ('b -> 'b -> bool) * (('a, 'b) state -> (nat, ('a * 'b)) mapping) *
      (('a, 'b) state -> (nat, ('a * 'b)) mapping) *
      (('a, 'b) state -> nat -> 'b option) *
      (('a, 'b) state -> nat -> 'b option) * (('a, 'b) state -> nat -> 'a) *
      (('a, 'b) state -> nat -> 'a) *
      (((nat, ('a * 'b)) mapping -> (nat, ('a * 'b)) mapping) ->
        ('a, 'b) state -> ('a, 'b) state) *
      (nat -> 'b -> 'b atom) * (nat -> 'b -> 'b atom);;

type 'a status = Abs_status of ('a * nat -> nat list);;

type ('a, 'b) claim = Yes | No | Terminating | Upperbound of nat |
  Nonterminating | Confluent | Nonconfluent | Completed | Ordered_Completed |
  Anything;;

type 'a reduction_order_input =
  RPO_Input of (('a * nat) * (nat * order_tag)) list |
  KBO_Input of ((('a * nat) * (nat * (nat * (nat list) option))) list * nat);;

type ('a, 'b) tree_automaton =
  Tree_Automaton of 'a list * ('a, 'b) ta_rule list * ('a * 'a) list;;

type ('a, 'b) complexity_measure = Derivational_Complexity of ('a * nat) list |
  Runtime_Complexity of ('a * nat) list * ('a * nat) list;;

type ('a, 'b) equation_literal = Equation of (('a, 'b) term * ('a, 'b) term) |
  Inequality of (('a, 'b) term * ('a, 'b) term);;

type complexity_class = Comp_Poly of nat;;

type ('a, 'b) fp_strategy = Outermost |
  Context_Sensitive of (('a * nat) * nat list) list |
  Forbidden_Patterns of (('a, 'b) ctxt * (('a, 'b) term * location)) list;;

type start_term = Full | Constructor_Based;;

type ('a, 'b) strategy = No_Strategy | Innermost |
  Innermost_Q of ('a, 'b) term list;;

type ('a, 'b) input =
  DP_input of
    bool * (('a, 'b) term * ('a, 'b) term) list * ('a, 'b) strategy *
      (('a, 'b) term * ('a, 'b) term) list
  | Inn_TRS_input of
      ('a, 'b) strategy * (('a, 'b) term * ('a, 'b) term) list *
        (('a, 'b) term * ('a, 'b) term) list * start_term
  | CPX_input of
      ('a, 'b) strategy * (('a, 'b) term * ('a, 'b) term) list *
        (('a, 'b) term * ('a, 'b) term) list * ('a, 'b) complexity_measure *
        complexity_class
  | COMP_input of
      (('a, 'b) term * ('a, 'b) term) list *
        (('a, 'b) term * ('a, 'b) term) list
  | OCOMP_input of
      (('a, 'b) term * ('a, 'b) term) list *
        (('a, 'b) term * ('a, 'b) term) list *
        (('a, 'b) term * ('a, 'b) term) list * 'a reduction_order_input
  | EQ_input of (('a, 'b) term * ('a, 'b) term) list * ('a, 'b) equation_literal
  | FP_TRS_input of ('a, 'b) fp_strategy * (('a, 'b) term * ('a, 'b) term) list
  | CTRS_input of
      ((('a, 'b) term * ('a, 'b) term) *
        (('a, 'b) term * ('a, 'b) term) list) list
  | TA_input of
      ((char list), 'a) tree_automaton * (('a, 'b) term * ('a, 'b) term) list
  | AC_input of (('a, 'b) term * ('a, 'b) term) list * 'a list * 'a list |
  LTS_input of (siga, 'b, ty, (char list), (char list)) lts_impl |
  LTS_safety_input of
    (siga, 'b, ty, (char list), (char list)) lts_impl * (char list) list
  | Unknown_input of char list;;

type ('a, 'b) const_string_sound_proof =
  Const_string_sound_proof of
    'b * ('a * 'a) list * (('a, 'b) term * ('a, 'b) term) list *
      (('a, 'b) term * ('a, 'b) term) list;;

type af_entry = Collapse of nat | AFList of nat list;;

type 'a redtriple_impl = Int_carrier of (('a * nat) * (int * int list)) list |
  Int_nl_carrier of (('a * nat) * (nat monom * int) list) list |
  Rat_carrier of (('a * nat) * (rat * rat list)) list |
  Rat_nl_carrier of rat * (('a * nat) * (nat monom * rat) list) list |
  Real_carrier of (('a * nat) * (real * real list)) list |
  Real_nl_carrier of real * (('a * nat) * (nat monom * real) list) list |
  Arctic_carrier of (('a * nat) * (arctic * arctic list)) list |
  Arctic_rat_carrier of
    (('a * nat) * (rat arctic_delta * rat arctic_delta list)) list
  | Int_mat_carrier of nat * nat * (('a * nat) * (int mat * int mat list)) list
  | Rat_mat_carrier of nat * nat * (('a * nat) * (rat mat * rat mat list)) list
  | Real_mat_carrier of
      nat * nat * (('a * nat) * (real mat * real mat list)) list
  | Arctic_mat_carrier of
      nat * (('a * nat) * (arctic mat * arctic mat list)) list
  | Arctic_rat_mat_carrier of
      nat * (('a * nat) *
              (rat arctic_delta mat * rat arctic_delta mat list)) list
  | RPO of (('a * nat) * (nat * order_tag)) list * (('a * nat) * af_entry) list
  | KBO of
      ((('a * nat) * (nat * (nat * (nat list) option))) list * nat) *
        (('a * nat) * af_entry) list
  | ACKBO of
      ((('a * nat) * (nat * (nat * bool))) list * nat) *
        (('a * nat) * af_entry) list
  | WPO of (('a * nat) * (nat * nat list)) list * 'a redtriple_impl |
  Max_poly of (('a * nat) * (sigb, nat) term) list;;

type ('a, 'b) complex_constant_removal_prf =
  Complex_Constant_Removal_Proof of
    ('a, 'b) term *
      ((('a, 'b) term * ('a, 'b) term) * (('a, 'b) term * ('a, 'b) term)) list;;

type list_order_type = MS_Ext | Max_Ext | Min_Ext | Dms_Ext;;

type 'a root_redtriple_impl =
  SCNP of
    list_order_type * (('a * nat) * (nat * nat) list) list * 'a redtriple_impl;;

type ('a, 'b) cond_constraint_prf = Final |
  Delete_Condition of ('a, 'b) cond_constraint * ('a, 'b) cond_constraint_prf |
  Different_Constructor of ('a, 'b) cond_constraint |
  Same_Constructor of
    ('a, 'b) cond_constraint * ('a, 'b) cond_constraint *
      ('a, 'b) cond_constraint_prf
  | Variable_Equation of
      'b * ('a, 'b) term * ('a, 'b) cond_constraint *
        ('a, 'b) cond_constraint_prf
  | Funarg_Into_Var of
      ('a, 'b) cond_constraint * nat * 'b * ('a, 'b) cond_constraint *
        ('a, 'b) cond_constraint_prf
  | Simplify_Condition of
      ('a, 'b) cond_constraint * ('b * ('a, 'b) term) list *
        ('a, 'b) cond_constraint * ('a, 'b) cond_constraint_prf
  | Induction of
      ('a, 'b) cond_constraint * ('a, 'b) cond_constraint list *
        ((('a, 'b) term * ('a, 'b) term) *
          ((('a, 'b) term * 'b list) list *
            (('a, 'b) cond_constraint * ('a, 'b) cond_constraint_prf))) list;;

type ('a, 'b) cond_red_pair_prf =
  Cond_Red_Pair_Prf of
    'a * (('a, 'b) cond_constraint *
           ((('a, 'b) term * ('a, 'b) term) list *
             ('a, 'b) cond_constraint_prf)) list *
      nat * nat;;

type arithFun = Arg of nat | Const of nat | Sum of arithFun list |
  Max of arithFun list | Min of arithFun list | Prod of arithFun list |
  IfEqual of arithFun * arithFun * arithFun * arithFun;;

type 'a sl_inter = SL_Inter of nat * (('a * nat) * arithFun) list;;

type ('a, 'b) sl_variant = Rootlab of ('a * nat) option |
  Finitelab of 'a sl_inter | QuasiFinitelab of 'a sl_inter * 'b;;

type ('a, 'b, 'c, 'd, 'e, 'f, 'g, 'h) generic_assm_proof =
  SN_assm_proof of
    (bool *
      ((('a, 'b) lab, 'c) term list *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
          ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list))) *
      'd
  | Finite_assm_proof of
      (bool *
        (bool *
          (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
            (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
              ((('a, 'b) lab, 'c) term list *
                (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
                  ((('a, 'b) lab, 'c) term *
                    (('a, 'b) lab, 'c) term) list)))))) *
        'e
  | SN_FP_assm_proof of
      (((('a, 'b) lab, 'c) ctxt * ((('a, 'b) lab, 'c) term * location)) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) *
        'f
  | Not_SN_assm_proof of
      (bool *
        ((('a, 'b) lab, 'c) term list *
          ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list)) *
        'd
  | Infinite_assm_proof of
      (bool *
        (bool *
          (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
            (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
              ((('a, 'b) lab, 'c) term list *
                (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
                  ((('a, 'b) lab, 'c) term *
                    (('a, 'b) lab, 'c) term) list)))))) *
        'e
  | Not_RelSN_assm_proof of
      (bool *
        ((('a, 'b) lab, 'c) term list *
          (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
            ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list))) *
        'f
  | Not_SN_FP_assm_proof of
      (((('a, 'b) lab, 'c) ctxt * ((('a, 'b) lab, 'c) term * location)) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) *
        'g
  | Complexity_assm_proof of
      ((('a, 'b) lab, 'c) term list *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
          (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
            ((('a, 'b) lab, 'c) complexity_measure * complexity_class)))) *
        'd
  | Unknown_assm_proof of char list * 'h;;

type 'a join_info =
  Guided of
    (('a, (char list)) term *
      ((nat list *
         ((('a, (char list)) term * ('a, (char list)) term) *
           ('a, (char list)) term)) list *
        (('a, (char list)) term *
          (nat list *
            ((('a, (char list)) term * ('a, (char list)) term) *
              ('a, (char list)) term)) list))) list
  | Join_NF | Join_BFS of nat;;

type 'a projL = Projection of (('a * nat) * nat) list;;

type 'a ta_relation = Decision_Proc_Old | Decision_Proc | Id_Relation |
  Some_Relation of ('a * 'a) list;;

type boundstype = Roof | Match;;

type ('a, 'b) bounds_info =
  Bounds_Info of
    boundstype * nat * 'b list * ('b, ('a * nat)) tree_automaton *
      'b ta_relation;;

type ('a, 'b, 'c) trs_termination_proof =
  DP_Trans of
    bool * bool * ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
      ('a, 'b, 'c) dp_termination_proof
  | Rule_Removal of
      ('a, 'b) lab redtriple_impl *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) trs_termination_proof
  | String_Reversal of ('a, 'b, 'c) trs_termination_proof |
  Constant_String of
    (('a, 'b) lab, 'c) const_string_sound_proof *
      ('a, 'b, 'c) trs_termination_proof
  | Bounds of (('a, 'b) lab, 'c) bounds_info |
  Uncurry of
    (('a, 'b) lab *
      (((('a, 'b) lab * nat) * ('a, 'b) lab list) list *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
          ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list))) *
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
      ('a, 'b, 'c) trs_termination_proof
  | Semlab of
      (('a, 'b) lab, 'c) sl_variant * (('a, 'b) lab, 'c) term list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) trs_termination_proof
  | R_is_Empty |
  Fcc of
    (('a, 'b) lab, 'c) ctxt list *
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
      ('a, 'b, 'c) trs_termination_proof
  | Split of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) trs_termination_proof * ('a, 'b, 'c) trs_termination_proof
  | Switch_Innermost of
      ('a, 'b) lab join_info * ('a, 'b, 'c) trs_termination_proof
  | Drop_Equality of ('a, 'b, 'c) trs_termination_proof |
  Remove_Nonapplicable_Rules of
    ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
      ('a, 'b, 'c) trs_termination_proof
  | Permuting_AFS of
      ((('a, 'b) lab * nat) * af_entry) list *
        ('a, 'b, 'c) trs_termination_proof
  | Assume_SN of
      (bool *
        ((('a, 'b) lab, 'c) term list *
          (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
            ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list))) *
        ('a, 'b, 'c, ('a, 'b, 'c) trs_termination_proof,
          ('a, 'b, 'c) dp_termination_proof,
          ('a, 'b, 'c) fptrs_termination_proof, unit,
          ('a, 'b, 'c) unknown_proof)
          generic_assm_proof list
and ('a, 'b, 'c) fptrs_termination_proof =
  Assume_FP_SN of
    (((('a, 'b) lab, 'c) ctxt * ((('a, 'b) lab, 'c) term * location)) list *
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) *
      ('a, 'b, 'c, ('a, 'b, 'c) trs_termination_proof,
        ('a, 'b, 'c) dp_termination_proof, ('a, 'b, 'c) fptrs_termination_proof,
        unit, ('a, 'b, 'c) unknown_proof)
        generic_assm_proof list
and ('a, 'b, 'c) dp_termination_proof = P_is_Empty |
  Subterm_Criterion_Proc of
    ('a, 'b) lab projL *
      (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
        (nat list *
          (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
            (('a, 'b) lab, 'c) term)) list) list *
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
      ('a, 'b, 'c) dp_termination_proof
  | Gen_Subterm_Criterion_Proc of
      ((('a, 'b) lab * nat) * nat list) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dp_termination_proof
  | Redpair_Proc of
      (('a, 'b) lab root_redtriple_impl, ('a, 'b) lab redtriple_impl) sum *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dp_termination_proof
  | Redpair_UR_Proc of
      (('a, 'b) lab root_redtriple_impl, ('a, 'b) lab redtriple_impl) sum *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dp_termination_proof
  | Usable_Rules_Proc of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dp_termination_proof
  | Dep_Graph_Proc of
      (('a, 'b, 'c) dp_termination_proof option *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) list
  | Mono_Redpair_Proc of
      ('a, 'b) lab redtriple_impl *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dp_termination_proof
  | Mono_URM_Redpair_Proc of
      ('a, 'b) lab redtriple_impl *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dp_termination_proof
  | Mono_Redpair_UR_Proc of
      ('a, 'b) lab redtriple_impl *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dp_termination_proof
  | Size_Change_Subterm_Proc of
      (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
        ((nat * nat) list * (nat * nat) list)) list
  | Size_Change_Redpair_Proc of
      ('a, 'b) lab redtriple_impl *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) option *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
          ((nat * nat) list * (nat * nat) list)) list
  | Uncurry_Proc of
      nat option *
        (('a, 'b) lab *
          (((('a, 'b) lab * nat) * ('a, 'b) lab list) list *
            (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
              ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list))) *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dp_termination_proof
  | Fcc_Proc of
      ('a, 'b) lab * (('a, 'b) lab, 'c) ctxt list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dp_termination_proof
  | Split_Proc of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dp_termination_proof * ('a, 'b, 'c) dp_termination_proof
  | Semlab_Proc of
      (('a, 'b) lab, 'c) sl_variant *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        (('a, 'b) lab, 'c) term list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dp_termination_proof
  | Switch_Innermost_Proc of
      ('a, 'b) lab join_info * ('a, 'b, 'c) dp_termination_proof
  | Rewriting_Proc of
      (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) option *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) * nat list *
        ('a, 'b, 'c) dp_termination_proof
  | Instantiation_Proc of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dp_termination_proof
  | Forward_Instantiation_Proc of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) option *
        ('a, 'b, 'c) dp_termination_proof
  | Narrowing_Proc of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) * nat list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dp_termination_proof
  | Assume_Finite of
      (bool *
        (bool *
          (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
            (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
              ((('a, 'b) lab, 'c) term list *
                (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
                  ((('a, 'b) lab, 'c) term *
                    (('a, 'b) lab, 'c) term) list)))))) *
        ('a, 'b, 'c, ('a, 'b, 'c) trs_termination_proof,
          ('a, 'b, 'c) dp_termination_proof,
          ('a, 'b, 'c) fptrs_termination_proof, unit,
          ('a, 'b, 'c) unknown_proof)
          generic_assm_proof list
  | Unlab_Proc of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dp_termination_proof
  | Q_Reduction_Proc of
      (('a, 'b) lab, 'c) term list * ('a, 'b, 'c) dp_termination_proof
  | Complex_Constant_Removal_Proc of
      (('a, 'b) lab, 'c) complex_constant_removal_prf *
        ('a, 'b, 'c) dp_termination_proof
  | General_Redpair_Proc of
      ('a, 'b) lab redtriple_impl *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        (('a, 'b) lab, 'c) cond_red_pair_prf *
        ('a, 'b, 'c) dp_termination_proof list
  | To_Trs_Proc of ('a, 'b, 'c) trs_termination_proof
and ('a, 'b, 'c) unknown_proof =
  Assume_Unknown of
    char list *
      ('a, 'b, 'c, ('a, 'b, 'c) trs_termination_proof,
        ('a, 'b, 'c) dp_termination_proof, ('a, 'b, 'c) fptrs_termination_proof,
        unit, ('a, 'b, 'c) unknown_proof)
        generic_assm_proof list;;

type ('a, 'b, 'c) cr_proof =
  SN_WCR of ('a, 'b) lab join_info * ('a, 'b, 'c) trs_termination_proof |
  Weakly_Orthogonal | Strongly_Closed of nat |
  Rule_Labeling of
    (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) * nat) list *
      ((('a, 'b) lab, (char list)) term *
        ((nat list *
           (((('a, 'b) lab, (char list)) term *
              (('a, 'b) lab, (char list)) term) *
             (('a, 'b) lab, (char list)) term)) list *
          ((('a, 'b) lab, (char list)) term *
            (nat list *
              (((('a, 'b) lab, (char list)) term *
                 (('a, 'b) lab, (char list)) term) *
                (('a, 'b) lab, (char list)) term)) list))) list *
      ('a, 'b, 'c) trs_termination_proof option
  | Rule_Labeling_Conv of
      (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) * nat) list *
        ((('a, 'b) lab, (char list)) term *
          ((('a, 'b) lab, (char list)) term *
            ((nat list *
               (((('a, 'b) lab, (char list)) term *
                  (('a, 'b) lab, (char list)) term) *
                 (bool * (('a, 'b) lab, (char list)) term))) list *
              ((nat list *
                 (((('a, 'b) lab, (char list)) term *
                    (('a, 'b) lab, (char list)) term) *
                   (('a, 'b) lab, (char list)) term)) list *
                ((nat list *
                   (((('a, 'b) lab, (char list)) term *
                      (('a, 'b) lab, (char list)) term) *
                     (bool * (('a, 'b) lab, (char list)) term))) list *
                  ((('a, 'b) lab, (char list)) term *
                    ((nat list *
                       (((('a, 'b) lab, (char list)) term *
                          (('a, 'b) lab, (char list)) term) *
                         (bool * (('a, 'b) lab, (char list)) term))) list *
                      ((nat list *
                         (((('a, 'b) lab, (char list)) term *
                            (('a, 'b) lab, (char list)) term) *
                           (('a, 'b) lab, (char list)) term)) list *
                        (nat list *
                          (((('a, 'b) lab, (char list)) term *
                             (('a, 'b) lab, (char list)) term) *
                            (bool *
                              (('a, 'b) lab, (char list))
                                term))) list)))))))) list *
        (nat * ('a, 'b, 'c) trs_termination_proof) option
  | Redundant_Rules of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list * nat *
        ((nat list *
           (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
             (bool * (('a, 'b) lab, 'c) term))) list) list *
        ('a, 'b, 'c) cr_proof
  | Parallel_Closed of nat option |
  Critical_Pair_Closing_System of
    ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
      ('a, 'b, 'c) trs_termination_proof * nat
  | Persistent_Decomposition of
      (('a, 'b) lab * ((char list) list * char list)) list *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
          ('a, 'b, 'c) cr_proof) list;;

type comp_res = LESS | EQUAL | GREATER;;

type ('a, 'b) multimap =
  Abs_multimap of (('b -> 'a) * (('a, ('b list)) mapping * 'b list));;

type ('a, 'b) ring_ext = Ring_ext of 'a * ('a -> 'a -> 'a) * 'b;;

type constrainta = LTa of linear_poly * rat | GTa of linear_poly * rat |
  LEQ of linear_poly * rat | GEQ of linear_poly * rat | EQa of linear_poly * rat
  | LTPP of linear_poly * linear_poly | GTPP of linear_poly * linear_poly |
  LEQPP of linear_poly * linear_poly | GEQPP of linear_poly * linear_poly |
  EQPP of linear_poly * linear_poly;;

type ('a, 'b) sum_bot = Sumbot of ('a, 'b) sum;;

type 'a afs = Abs_afs of (('a * nat -> af_entry) * ('a * nat) set);;

type ('a, 'b, 'c) non_join_info = Diff_NFs |
  Tcap_Non_Unif of (('a, 'b) term -> ('a, 'b) term -> 'b -> ('a, 'b) term) |
  Tree_Aut_Intersect_Empty of
    ('c, 'a) tree_automaton * 'c ta_relation * ('c, 'a) tree_automaton *
      'c ta_relation
  | Finite_Model_Gt of ('a, 'b) sl_variant |
  Reduction_Pair_Gt of 'a redtriple_impl |
  Usable_Rules_Reach_NJ of ('a, 'b, 'c) non_join_info |
  Usable_Rules_Reach_Unif_NJ of
    (((('a, 'b) term * ('a, 'b) term) list),
      ((('a, 'b) term * ('a, 'b) term) list))
      sum *
      ('a, 'b, 'c) non_join_info
  | Argument_Filter_NJ of
      (('a * nat) * af_entry) list * ('a, 'b, 'c) non_join_info
  | Grounding of ('b * ('a, 'b) term) list * ('a, 'b, 'c) non_join_info |
  Subterm_NJ of nat list * ('a, 'b, 'c) non_join_info;;

type ('a, 'b, 'c, 'd) ncr_proof = SN_NWCR of ('a, 'b, 'c) trs_termination_proof
  | Non_Join of
      (('a, 'b) lab, 'c) term *
        (nat list *
          (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
            (('a, 'b) lab, 'c) term)) list *
        (nat list *
          (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
            (('a, 'b) lab, 'c) term)) list *
        (('a, 'b) lab, 'c, 'd) non_join_info
  | NCR_Disj_Subtrs of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c, 'd) ncr_proof
  | NCR_Redundant_Rules of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list * nat *
        ('a, 'b, 'c, 'd) ncr_proof
  | NCR_Persistent_Decomposition of
      (('a, 'b) lab * ((char list) list * char list)) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c, 'd) ncr_proof;;

type ('a, 'b) rule_removal_nonterm_reltrs_prf =
  Rule_removal_nonterm_reltrs_prf of
    ((('a, 'b) term * ('a, 'b) term) list) option *
      ((('a, 'b) term * ('a, 'b) term) list) option;;

type ('a, 'b, 'c) dp_trans_nontermination_tt_prf =
  DP_trans_nontermination_tt_prf of
    ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list;;

type ('a, 'b) const_string_complete_proof =
  Const_string_complete_proof of
    'b * ('a * 'a) list * (('a, 'b) term * ('a, 'b) term) list;;

type ('a, 'b) rule_removal_nonterm_trs_prf =
  Rule_removal_nonterm_trs_prf of (('a, 'b) term * ('a, 'b) term) list;;

type ('a, 'b) q_increase_nonterm_trs_prf =
  Q_increase_nonterm_trs_prf of ('a, 'b) term list;;

type ('a, 'b) dp_q_reduction_nonterm_prf =
  DP_q_reduction_nonterm_prf of ('a, 'b) term list;;

type ('a, 'b) instantiation_complete_proc_prf =
  Instantiation_complete_proc_prf of (('a, 'b) term * ('a, 'b) term) list;;

type ('a, 'b) rule_removal_nonterm_dp_prf =
  Rule_removal_nonterm_dp_prf of
    ((('a, 'b) term * ('a, 'b) term) list) option *
      ((('a, 'b) term * ('a, 'b) term) list) option;;

type ('a, 'b) q_increase_nonterm_dp_prf =
  Q_increase_nonterm_dp_prf of ('a, 'b) term list;;

type ('a, 'b) rewriting_complete_proc_prf =
  Rewriting_complete_proc_prf of
    ((('a, 'b) term * ('a, 'b) term) list) option *
      (('a, 'b) term * ('a, 'b) term) * (('a, 'b) term * ('a, 'b) term) *
      (('a, 'b) term * ('a, 'b) term) * (('a, 'b) term * ('a, 'b) term) *
      nat list;;

type ('a, 'b) narrowing_complete_proc_prf =
  Narrowing_complete_proc_prf of
    (('a, 'b) term * ('a, 'b) term) * nat list *
      (('a, 'b) term * ('a, 'b) term) list;;

type pat_rule_pos = Pat_Base | Pat_Pump | Pat_Close;;

type ('a, 'b) pat_eqv_prf = Pat_Dom_Renaming of ('b * ('a, 'b) term) list |
  Pat_Irrelevant of ('b * ('a, 'b) term) list * ('b * ('a, 'b) term) list |
  Pat_Simplify of ('b * ('a, 'b) term) list * ('b * ('a, 'b) term) list;;

type ('a, 'b) pat_rule_prf =
  Pat_OrigRule of (('a, 'b) term * ('a, 'b) term) * bool |
  Pat_InitPump of
    ('a, 'b) pat_rule_prf * ('b * ('a, 'b) term) list *
      ('b * ('a, 'b) term) list
  | Pat_InitPumpCtxt of
      ('a, 'b) pat_rule_prf * ('b * ('a, 'b) term) list * nat list * 'b
  | Pat_Equiv of ('a, 'b) pat_rule_prf * bool * ('a, 'b) pat_eqv_prf |
  Pat_Narrow of ('a, 'b) pat_rule_prf * ('a, 'b) pat_rule_prf * nat list |
  Pat_Inst of ('a, 'b) pat_rule_prf * ('b * ('a, 'b) term) list * pat_rule_pos |
  Pat_Rewr of
    ('a, 'b) pat_rule_prf *
      (('a, 'b) term *
        (nat list * ((('a, 'b) term * ('a, 'b) term) * ('a, 'b) term)) list) *
      pat_rule_pos * 'b
  | Pat_Exp_Sigma of ('a, 'b) pat_rule_prf * nat;;

type ('a, 'b) non_loop_prf =
  Non_loop_prf of
    ('a, 'b) pat_rule_prf * ('b * ('a, 'b) term) list *
      ('b * ('a, 'b) term) list * nat * nat * nat list;;

type ('a, 'b) dp_loop_prf =
  DP_loop_prf of
    ('a, 'b) term *
      (nat list *
        ((('a, 'b) term * ('a, 'b) term) * (bool * ('a, 'b) term))) list *
      ('b * ('a, 'b) term) list * ('a, 'b) ctxt;;

type ('a, 'b) not_wn_ta_prf =
  Not_wn_ta_prf of ('b, 'a) tree_automaton * 'b ta_relation;;

type 'a dp_proof_step = OC1 of ('a list * 'a list) * bool |
  OC2 of
    ('a list * 'a list) * ('a list * 'a list) * ('a list * 'a list) * 'a list *
      'a list * 'a list
  | OC2p of
      ('a list * 'a list) * ('a list * 'a list) * ('a list * 'a list) *
        'a list * 'a list * 'a list
  | OC3 of
      ('a list * 'a list) * ('a list * 'a list) * ('a list * 'a list) *
        'a list * 'a list
  | OC3p of
      ('a list * 'a list) * ('a list * 'a list) * ('a list * 'a list) *
        'a list * 'a list
  | OCDP1 of
      (('a list * ((nat * (nat * 'a list)) * 'a list)) *
        ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        ('a list * 'a list)
  | OCDP2 of
      (('a list * ((nat * (nat * 'a list)) * 'a list)) *
        ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        ('a list * 'a list)
  | WPEQ of
      (('a list * ((nat * (nat * 'a list)) * 'a list)) *
        ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        (('a list * ((nat * (nat * 'a list)) * 'a list)) *
          ('a list * ((nat * (nat * 'a list)) * 'a list)))
  | Lift of
      (('a list * ((nat * (nat * 'a list)) * 'a list)) *
        ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        (('a list * ((nat * (nat * 'a list)) * 'a list)) *
          ('a list * ((nat * (nat * 'a list)) * 'a list)))
  | DPOC1_1 of
      (('a list * ((nat * (nat * 'a list)) * 'a list)) *
        ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        (('a list * ((nat * (nat * 'a list)) * 'a list)) *
          ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        ('a list * 'a list) * 'a list * 'a list
  | DPOC1_2 of
      (('a list * ((nat * (nat * 'a list)) * 'a list)) *
        ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        (('a list * ((nat * (nat * 'a list)) * 'a list)) *
          ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        ('a list * 'a list) * 'a list * 'a list * 'a list
  | DPOC2 of
      (('a list * ((nat * (nat * 'a list)) * 'a list)) *
        ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        (('a list * ((nat * (nat * 'a list)) * 'a list)) *
          ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        ('a list * 'a list) * 'a list * 'a list
  | DPOC3_1 of
      (('a list * ((nat * (nat * 'a list)) * 'a list)) *
        ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        (('a list * ((nat * (nat * 'a list)) * 'a list)) *
          ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        ('a list * 'a list) * 'a list * 'a list
  | DPOC3_2 of
      (('a list * ((nat * (nat * 'a list)) * 'a list)) *
        ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        (('a list * ((nat * (nat * 'a list)) * 'a list)) *
          ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        ('a list * 'a list) * 'a list * 'a list * 'a list
  | DPDP1_1 of
      (('a list * ((nat * (nat * 'a list)) * 'a list)) *
        ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        (('a list * ((nat * (nat * 'a list)) * 'a list)) *
          ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        (('a list * ((nat * (nat * 'a list)) * 'a list)) *
          ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        'a list * 'a list
  | DPDP1_2 of
      (('a list * ((nat * (nat * 'a list)) * 'a list)) *
        ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        (('a list * ((nat * (nat * 'a list)) * 'a list)) *
          ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        (('a list * ((nat * (nat * 'a list)) * 'a list)) *
          ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        'a list * 'a list
  | DPDP2_1 of
      (('a list * ((nat * (nat * 'a list)) * 'a list)) *
        ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        (('a list * ((nat * (nat * 'a list)) * 'a list)) *
          ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        (('a list * ((nat * (nat * 'a list)) * 'a list)) *
          ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        'a list * 'a list
  | DPDP2_2 of
      (('a list * ((nat * (nat * 'a list)) * 'a list)) *
        ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        (('a list * ((nat * (nat * 'a list)) * 'a list)) *
          ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        (('a list * ((nat * (nat * 'a list)) * 'a list)) *
          ('a list * ((nat * (nat * 'a list)) * 'a list))) *
        'a list * 'a list;;

type 'a non_loop_srs_proof =
  SE_OC of ('a list * 'a list) * 'a list * 'a list * 'a dp_proof_step list |
  SE_DP of
    (('a list * ((nat * (nat * 'a list)) * 'a list)) *
      ('a list * ((nat * (nat * 'a list)) * 'a list))) *
      'a list * 'a list * 'a dp_proof_step list;;

type ('a, 'b, 'c) uncurry_nt_proof =
  Uncurry_nt_proof of
    (('a, 'b) lab *
      (((('a, 'b) lab * nat) * ('a, 'b) lab list) list *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
          ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list))) *
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list;;

type ('a, 'b) trs_loop_prf =
  TRS_loop_prf of
    ('a, 'b) term *
      (nat list * ((('a, 'b) term * ('a, 'b) term) * ('a, 'b) term)) list *
      ('b * ('a, 'b) term) list * ('a, 'b) ctxt;;

type ('a, 'b) rel_trs_loop_prf =
  Rel_trs_loop_prf of
    ('a, 'b) term *
      (nat list *
        ((('a, 'b) term * ('a, 'b) term) * (bool * ('a, 'b) term))) list *
      ('b * ('a, 'b) term) list * ('a, 'b) ctxt;;

type ('a, 'b, 'c) reltrs_nontermination_proof =
  Rel_Loop of (('a, 'b) lab, 'c) rel_trs_loop_prf |
  Rel_TRS_String_Reversal of ('a, 'b, 'c) reltrs_nontermination_proof |
  Rel_Not_Well_Formed |
  Rel_Rule_Removal of
    (('a, 'b) lab, 'c) rule_removal_nonterm_reltrs_prf *
      ('a, 'b, 'c) reltrs_nontermination_proof
  | Rel_R_Not_SN of ('a, 'b, 'c) trs_nontermination_proof |
  Rel_TRS_Assume_Not_SN of
    (bool *
      ((('a, 'b) lab, 'c) term list *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
          ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list))) *
      ('a, 'b, 'c, ('a, 'b, 'c) trs_nontermination_proof,
        ('a, 'b, 'c) dp_nontermination_proof,
        ('a, 'b, 'c) reltrs_nontermination_proof,
        ('a, 'b, 'c) trs_nontermination_proof, ('a, 'b, 'c) neg_unknown_proof)
        generic_assm_proof list
and ('a, 'b, 'c) trs_nontermination_proof =
  TRS_Loop of (('a, 'b) lab, 'c) trs_loop_prf | TRS_Not_Well_Formed |
  TRS_Rule_Removal of
    (('a, 'b) lab, 'c) rule_removal_nonterm_trs_prf *
      ('a, 'b, 'c) trs_nontermination_proof
  | TRS_String_Reversal of ('a, 'b, 'c) trs_nontermination_proof |
  TRS_Constant_String of
    (('a, 'b) lab, 'c) const_string_complete_proof *
      ('a, 'b, 'c) trs_nontermination_proof
  | TRS_DP_Trans of
      ('a, 'b, 'c) dp_trans_nontermination_tt_prf *
        ('a, 'b, 'c) dp_nontermination_proof
  | TRS_Termination_Switch of
      ('a, 'b) lab join_info * ('a, 'b, 'c) trs_nontermination_proof
  | TRS_Nonloop of (('a, 'b) lab, 'c) non_loop_prf |
  TRS_Nonloop_SRS of ('a, 'b) lab non_loop_srs_proof |
  TRS_Q_Increase of
    (('a, 'b) lab, 'c) q_increase_nonterm_trs_prf *
      ('a, 'b, 'c) trs_nontermination_proof
  | TRS_Uncurry of
      ('a, 'b, 'c) uncurry_nt_proof * ('a, 'b, 'c) trs_nontermination_proof
  | TRS_Not_WN_Tree_Automaton of (('a, 'b) lab, 'c) not_wn_ta_prf |
  TRS_Assume_Not_SN of
    (bool *
      ((('a, 'b) lab, 'c) term list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list)) *
      ('a, 'b, 'c, ('a, 'b, 'c) trs_nontermination_proof,
        ('a, 'b, 'c) dp_nontermination_proof,
        ('a, 'b, 'c) reltrs_nontermination_proof,
        ('a, 'b, 'c) trs_nontermination_proof, ('a, 'b, 'c) neg_unknown_proof)
        generic_assm_proof list
  | FPTRS_Assume_Not_SN of
      (((('a, 'b) lab, 'c) ctxt * ((('a, 'b) lab, 'c) term * location)) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) *
        ('a, 'b, 'c, ('a, 'b, 'c) trs_nontermination_proof,
          ('a, 'b, 'c) dp_nontermination_proof,
          ('a, 'b, 'c) reltrs_nontermination_proof,
          ('a, 'b, 'c) trs_nontermination_proof, ('a, 'b, 'c) neg_unknown_proof)
          generic_assm_proof list
and ('a, 'b, 'c) dp_nontermination_proof =
  DP_Loop of (('a, 'b) lab, 'c) dp_loop_prf |
  DP_Nonloop of (('a, 'b) lab, 'c) non_loop_prf |
  DP_Rule_Removal of
    (('a, 'b) lab, 'c) rule_removal_nonterm_dp_prf *
      ('a, 'b, 'c) dp_nontermination_proof
  | DP_Q_Increase of
      (('a, 'b) lab, 'c) q_increase_nonterm_dp_prf *
        ('a, 'b, 'c) dp_nontermination_proof
  | DP_Q_Reduction of
      (('a, 'b) lab, 'c) dp_q_reduction_nonterm_prf *
        ('a, 'b, 'c) dp_nontermination_proof
  | DP_Termination_Switch of
      ('a, 'b) lab join_info * ('a, 'b, 'c) dp_nontermination_proof
  | DP_Instantiation of
      (('a, 'b) lab, 'c) instantiation_complete_proc_prf *
        ('a, 'b, 'c) dp_nontermination_proof
  | DP_Rewriting of
      (('a, 'b) lab, 'c) rewriting_complete_proc_prf *
        ('a, 'b, 'c) dp_nontermination_proof
  | DP_Narrowing of
      (('a, 'b) lab, 'c) narrowing_complete_proc_prf *
        ('a, 'b, 'c) dp_nontermination_proof
  | DP_Assume_Infinite of
      (bool *
        (bool *
          (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
            (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
              ((('a, 'b) lab, 'c) term list *
                (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
                  ((('a, 'b) lab, 'c) term *
                    (('a, 'b) lab, 'c) term) list)))))) *
        ('a, 'b, 'c, ('a, 'b, 'c) trs_nontermination_proof,
          ('a, 'b, 'c) dp_nontermination_proof,
          ('a, 'b, 'c) reltrs_nontermination_proof,
          ('a, 'b, 'c) trs_nontermination_proof, ('a, 'b, 'c) neg_unknown_proof)
          generic_assm_proof list
and ('a, 'b, 'c) neg_unknown_proof =
  Assume_NT_Unknown of
    char list *
      ('a, 'b, 'c, ('a, 'b, 'c) trs_nontermination_proof,
        ('a, 'b, 'c) dp_nontermination_proof,
        ('a, 'b, 'c) reltrs_nontermination_proof,
        ('a, 'b, 'c) trs_nontermination_proof, ('a, 'b, 'c) neg_unknown_proof)
        generic_assm_proof list;;

type ('a, 'b) oc_irule =
  OC_Deduce of ('a, 'b) term * ('a, 'b) term * ('a, 'b) term |
  OC_Orientl of ('a, 'b) term * ('a, 'b) term |
  OC_Orientr of ('a, 'b) term * ('a, 'b) term | OC_Delete of ('a, 'b) term |
  OC_Compose of ('a, 'b) term * ('a, 'b) term * ('a, 'b) term |
  OC_Simplifyl of ('a, 'b) term * ('a, 'b) term * ('a, 'b) term |
  OC_Simplifyr of ('a, 'b) term * ('a, 'b) term * ('a, 'b) term |
  OC_Collapse of ('a, 'b) term * ('a, 'b) term * ('a, 'b) term;;

type ('a, 'b, 'c) ordered_completion_proof =
  OKB of (('a, 'b) lab, 'c) oc_irule list;;

type ('a, 'b, 'c) quasi_reductive_proof =
  Unravel of
    ((((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
       ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) *
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) list *
      ('a, 'b, 'c) trs_termination_proof;;

type ('a, 'b) nonreachability_proof = Nonreachable_Tcap | Nonreachable_Gtcap |
  Nonreachable_ETAC of
    ('a * nat) list * 'a * 'a * (('a, 'b) term, 'a) tree_automaton
  | Nonreachable_Subst_Approx of
      (('a, 'b) term * ('a, 'b) term) list * ('a, 'b) nonreachability_proof
  | Nonreachable_Reverse of ('a, 'b) nonreachability_proof;;

type ('a, 'b) nonjoinability_proof = Nonjoinable_Tcap |
  Nonjoinable_Ground_NF of ('a, 'b) nonreachability_proof;;

type ('a, 'b) infeasibility_proof =
  Infeasible_Compound_Conditions of 'a * ('a, 'b) nonreachability_proof |
  Infeasible_Equation of
    ('a, 'b) term * ('a, 'b) term * ('a, 'b) nonreachability_proof
  | Infeasible_Subset of
      (('a, 'b) term * ('a, 'b) term) list * ('a, 'b) infeasibility_proof
  | Infeasible_Rhss_Equal of
      ('a, 'b) term * ('a, 'b) term * ('a, 'b) term *
        ('a, 'b) nonjoinability_proof
  | Infeasible_Trans of
      ('a, 'b) term * ('a, 'b) term * ('a, 'b) term *
        ('a, 'b) nonreachability_proof;;

type ('a, 'b, 'c) ccr_transformation =
  Inline_Conditions_CCRT of
    (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) list *
      ((((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
         ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) list
  | Infeasible_Rule_Removal_CCRT of
      ((((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
         ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) *
        (('a, 'b) lab, 'c) infeasibility_proof) list;;

type ('a, 'b) cstep_proof =
  Cstep_step of
    ((('a, 'b) term * ('a, 'b) term) * (('a, 'b) term * ('a, 'b) term) list) *
      nat list * ('b -> ('a, 'b) term) * ('a, 'b) term * ('a, 'b) term *
      (('a, 'b) cstep_proof list) list;;

type ('a, 'b, 'c, 'd) conditional_ncr_proof =
  Unconditional_CNCR of ('a, 'b, 'c, 'd) ncr_proof |
  Transformation_CNCR of
    ('a, 'b, 'c) ccr_transformation * ('a, 'b, 'c, 'd) conditional_ncr_proof
  | Non_Join_CNCR of
      (('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term *
        (('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) cstep_proof list *
        (('a, 'b) lab, 'c) cstep_proof list *
        (('a, 'b) lab, 'c, 'd) non_join_info;;

type ('a, 'b, 'c) completion_proof =
  SN_WCR_Eq of
    ('a, 'b) lab join_info * ('a, 'b, 'c) trs_termination_proof *
      (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
        (nat list *
          (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
            (bool * (('a, 'b) lab, 'c) term))) list) list *
      ((((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
         (nat list *
           (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
             (bool * (('a, 'b) lab, 'c) term))) list) list) option;;

type ('a, 'b, 'c) equational_disproof =
  Completion_and_Normalization_Different of
    ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
      ('a, 'b, 'c) completion_proof
  | Convertible_Instance of
      (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
        (nat list *
          (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
            (bool * (('a, 'b) lab, 'c) term))) list) list;;

type ('a, 'b) ao_infeasibility_proof =
  AO_Infeasibility_Proof of ('a, 'b) infeasibility_proof |
  AO_Lhss_Equal of
    ('a, 'b) term * ('a, 'b) term * ('a, 'b) term *
      ('a, 'b) nonjoinability_proof;;

type ('a, 'b) context_joinable_proof =
  Contextual_Join of
    ('a, 'b) term * ('a, 'b) cstep_proof list * ('a, 'b) cstep_proof list;;

type ('a, 'b) unfeasible_proof =
  UnfeasibleOverlap of
    ('a, 'b) term * ('a, 'b) term * ('a, 'b) term * ('a, 'b) cstep_proof list *
      ('a, 'b) cstep_proof list *
      ((('a, 'b) term * ('a, 'b) term) * (('a, 'b) term * ('a, 'b) term) list) *
      ((('a, 'b) term * ('a, 'b) term) * (('a, 'b) term * ('a, 'b) term) list);;

type ('a, 'b, 'c) conditional_cr_proof =
  Unconditional_CR of ('a, 'b, 'c) cr_proof |
  Unravel_CR of
    ((((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
       ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) *
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) list *
      ('a, 'b, 'c) cr_proof
  | Transformation_CR of
      ('a, 'b, 'c) ccr_transformation * ('a, 'b, 'c) conditional_cr_proof
  | Almost_Orthogonal_CR |
  Almost_Orthogonal_Modulo_Infeasibility_CR of
    (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
      (('a, 'b) lab, 'c) infeasibility_proof) list
  | Almost_Orthogonal_Modulo_Infeasibility_CRa of
      (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
          (('a, 'b) lab, 'c) ao_infeasibility_proof)) list
  | AL94_CR of
      ('a, 'b, 'c) quasi_reductive_proof *
        ((('a, 'b) lab, 'c) term *
          ((('a, 'b) lab, 'c) term *
            (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
              (('a, 'b) lab, 'c) context_joinable_proof))) list *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
          (('a, 'b) lab, 'c) infeasibility_proof) list *
        (('c -> (('a, 'b) lab, 'c) term) *
          (('a, 'b) lab, 'c) unfeasible_proof) list;;

type ('a, 'b) ac_dependency_pairs_proof =
  AC_dependency_pairs_proof of
    (('a, 'b) term * ('a, 'b) term) list *
      (('a, 'b) term * ('a, 'b) term) list *
      (('a, 'b) term * ('a, 'b) term) list *
      (('a, 'b) term * ('a, 'b) term) list;;

type ('a, 'b) ac_dp_termination_proof = AC_P_is_Empty |
  AC_Subterm_Proc of
    (('a * nat) * nat list) list * (('a, 'b) term * ('a, 'b) term) list *
      ('a, 'b) ac_dp_termination_proof
  | AC_Redpair_UR_Proc of
      'a redtriple_impl * (('a, 'b) term * ('a, 'b) term) list *
        (('a, 'b) term * ('a, 'b) term) list * ('a, 'b) ac_dp_termination_proof
  | AC_Mono_Redpair_UR_Proc of
      'a redtriple_impl * (('a, 'b) term * ('a, 'b) term) list *
        (('a, 'b) term * ('a, 'b) term) list *
        (('a, 'b) term * ('a, 'b) term) list * ('a, 'b) ac_dp_termination_proof
  | AC_Dep_Graph_Proc of
      (('a, 'b) ac_dp_termination_proof option *
        (('a, 'b) term * ('a, 'b) term) list) list;;

type ('a, 'b, 'c) ac_termination_proof =
  AC_DP_Trans of
    (('a, 'b) lab, 'c) ac_dependency_pairs_proof *
      (('a, 'b) lab, 'c) ac_dp_termination_proof *
      (('a, 'b) lab, 'c) ac_dp_termination_proof
  | AC_DP_Trans_Single of
      (('a, 'b) lab, 'c) ac_dependency_pairs_proof *
        (('a, 'b) lab, 'c) ac_dp_termination_proof
  | AC_Rule_Removal of
      ('a, 'b) lab redtriple_impl *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) ac_termination_proof
  | AC_R_is_Empty;;

type ('a, 'b, 'c, 'd) fresh_variable_addition_info =
  Fresh_Variable_Addition_Info of
    'b * 'c * ('d * ('a, ('b trans_var * 'c)) term formula) list;;

type ('a, 'b, 'c, 'd, 'e) transition_removal_info =
  Transition_removal_info of
    ('c sharp -> 'a) * 'd list * 'b * 'a * ('d -> 'e hint);;

type ('a, 'b, 'c, 'd, 'e) location_addition_info =
  Location_Addition_Info of 'd * 'd * 'e * ('a, 'b, 'c, 'd) transition_rule;;

type ('a, 'b, 'c) art_edge_impl = Cover_Edge of 'a * 'c |
  Children_Edge of ('b * ('a * 'c)) list;;

type ('a, 'b, 'c, 'd, 'e, 'f, 'g) art_node_impl =
  Art_Node of
    'e * ('a, ('b * 'c)) term formula * 'd * ('e, 'f, 'g) art_edge_impl;;

type ('a, 'b, 'c, 'd, 'e, 'f, 'g, 'h) art_impl_ext =
  Art_impl_ext of
    'e list * ('a, 'b, 'c, 'd, 'e, 'f, 'g) art_node_impl list * 'h;;

type ('a, 'b, 'c, 'd, 'e, 'f, 'g) invariant_proof =
  Impact of
    ('d * ('a, ('b * 'c)) term formula) list *
      ('a, 'b, 'c, 'd, 'e, 'f, 'g hint, unit) art_impl_ext;;

type ('a, 'b, 'c, 'd, 'e, 'f) cooperation_proof = Triviala |
  Invariants_Update of
    ('a, 'b, 'c, 'd sharp, (char list), 'e, 'f) invariant_proof *
      ('a, 'b, 'c, 'd, 'e, 'f) cooperation_proof
  | Location_Addition of
      ('a, 'b, 'c, 'd sharp, 'e) location_addition_info *
        ('a, 'b, 'c, 'd, 'e, 'f) cooperation_proof
  | Fresh_Variable_Addition of
      ('a, 'b, 'c, 'e) fresh_variable_addition_info *
        ('a, 'b, 'c, 'd, 'e, 'f) cooperation_proof
  | Transition_Removal of
      ((('a, ('b * 'c)) term list), 'c, 'd, 'e, 'f) transition_removal_info *
        ('a, 'b, 'c, 'd, 'e, 'f) cooperation_proof
  | Scc_Decomp of
      ('d sharp list * ('a, 'b, 'c, 'd, 'e, 'f) cooperation_proof) list
  | Cut_Transition_Split of
      ('e list * ('a, 'b, 'c, 'd, 'e, 'f) cooperation_proof) list;;

type ('a, 'b, 'c, 'd, 'e, 'f) termination_proof = Trivialb |
  Via_Cooperation of
    (('e sharp * ('a, 'b, 'c, 'd sharp) transition_rule) list *
      ('a, 'b, 'c, 'd, 'e sharp, 'f) cooperation_proof) list
  | Invariants_Update_LTS of
      ('a, 'b, 'c, 'd, (char list), 'e, 'f) invariant_proof *
        ('a, 'b, 'c, 'd, 'e, 'f) termination_proof;;

type ('a, 'b) eq_proof = Refl of ('a, 'b) term | Sym of ('a, 'b) eq_proof |
  Trans of ('a, 'b) eq_proof * ('a, 'b) eq_proof |
  Assm of (('a, 'b) term * ('a, 'b) term) * ('b -> ('a, 'b) term) |
  Cong of 'a * ('a, 'b) eq_proof list;;

type ('a, 'b, 'c) equational_proof =
  Equational_Proof_Tree of (('a, 'b) lab, 'c) eq_proof |
  Completion_and_Normalization of
    ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
      ('a, 'b, 'c) completion_proof
  | Conversion of
      (nat list *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
          (bool * (('a, 'b) lab, 'c) term))) list
  | Conversion_With_History of
      (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
        (nat list *
          (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
            (bool * (('a, 'b) lab, 'c) term))) list) list;;

type ('a, 'b) dt_transformation_info =
  DT_Transformation_Info of
    ((('a, 'b) term * ('a, 'b) term) * (('a, 'b) term * ('a, 'b) term)) list *
      ((('a, 'b) term * ('a, 'b) term) * (('a, 'b) term * ('a, 'b) term)) list *
      ('a, 'b) term list;;

type ('a, 'b) wdp_trans_info =
  WDP_Trans_Info of
    ('a * nat) set *
      ((('a, 'b) term * ('a, 'b) term) * (('a, 'b) term * ('a, 'b) term)) list *
      ((('a, 'b) term * ('a, 'b) term) * (('a, 'b) term * ('a, 'b) term)) list *
      ('a, 'b) term list;;

type ('a, 'b, 'c) complexity_proof =
  Rule_Shift_Complexity of
    ('a, 'b) lab redtriple_impl *
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
      (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) option *
      ('a, 'b, 'c) complexity_proof
  | RisEmpty_Complexity |
  Remove_Nonapplicable_Rules_Complexity of
    ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
      ('a, 'b, 'c) complexity_proof
  | Matchbounds_Complexity of (('a, 'b) lab, 'c) bounds_info |
  Matchbounds_Rel_Complexity of
    (('a, 'b) lab, 'c) bounds_info *
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
      ('a, 'b, 'c) complexity_proof
  | DT_Transformation of
      (('a, 'b) lab, 'c) dt_transformation_info * ('a, 'b, 'c) complexity_proof
  | WDP_Transformation of
      (('a, 'b) lab, 'c) wdp_trans_info * ('a, 'b, 'c) complexity_proof
  | Complexity_Assumption of
      ((('a, 'b) lab, 'c) term list *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
          (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
            ((('a, 'b) lab, 'c) complexity_measure * complexity_class)))) *
        ('a, 'b, 'c, ('a, 'b, 'c) complexity_proof, unit, unit, unit, unit)
          generic_assm_proof list
  | Usable_Rules_Complexity of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) complexity_proof
  | Split_Complexity of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) complexity_proof * ('a, 'b, 'c) complexity_proof;;

type ('a, 'b, 'c, 'd, 'e, 'f, 'g) safety_proof = Trivial |
  Invariant_Assertion of
    ('a, 'b, 'c, 'd, 'e, 'f, 'g) invariant_proof *
      ('a, 'b, 'c, 'd, 'e, 'f, 'g) safety_proof;;

type ('a, 'b, 'c) proof =
  TRS_Termination_Proof of ('a, 'b, 'c) trs_termination_proof |
  Complexity_Proof of ('a, 'b, 'c) complexity_proof |
  DP_Termination_Proof of ('a, 'b, 'c) dp_termination_proof |
  DP_Nontermination_Proof of ('a, 'b, 'c) dp_nontermination_proof |
  TRS_Nontermination_Proof of ('a, 'b, 'c) trs_nontermination_proof |
  FP_Termination_Proof of ('a, 'b, 'c) fptrs_termination_proof |
  Relative_TRS_Nontermination_Proof of ('a, 'b, 'c) reltrs_nontermination_proof
  | TRS_Confluence_Proof of ('a, 'b, 'c) cr_proof |
  TRS_Non_Confluence_Proof of ('a, 'b, 'c, 'c) ncr_proof |
  Completion_Proof of ('a, 'b, 'c) completion_proof |
  Ordered_Completion_Proof of ('a, 'b, 'c) ordered_completion_proof |
  Equational_Proof of ('a, 'b, 'c) equational_proof |
  Equational_Disproof of ('a, 'b, 'c) equational_disproof |
  Quasi_Reductive_Proof of ('a, 'b, 'c) quasi_reductive_proof |
  Conditional_CR_Proof of ('a, 'b, 'c) conditional_cr_proof |
  Conditional_Non_CR_Proof of ('a, 'b, 'c, 'c) conditional_ncr_proof |
  Tree_Automata_Closed_Proof of (char list) ta_relation |
  AC_Termination_Proof of ('a, 'b, 'c) ac_termination_proof |
  LTS_Termination_Proof of
    (siga, 'c, ty, (char list), (char list), hints) termination_proof
  | LTS_Safety_Proof of
      (siga, 'c, ty, (char list), (char list), (char list), hints) safety_proof
  | Unknown_Proof of ('a, 'b, 'c) unknown_proof |
  Unknown_Disproof of ('a, 'b, 'c) neg_unknown_proof;;

type 'a linearity = Non_Linear | Onea | Variable of 'a;;

type 'a ns_constraint = LEQ_ns of linear_poly * 'a |
  GEQ_ns of linear_poly * 'a;;

type ('b, 'a) comp_fun_idem = Abs_comp_fun_idem of ('b -> 'a -> 'a);;

type ('a, 'b) monoid_ext = Monoid_ext of ('a -> 'a -> 'a) * 'a * 'b;;

type ('a, 'b, 'c) ta_ext =
  Ta_ext of 'a set * ('a, 'b) ta_rule set * ('a * 'a) set * 'c;;

type dependance = Ignore | Increase | Decrease | Wild;;

type 'a semilattice_set = Abs_semilattice_set of ('a -> 'a -> 'a);;

type 'a interpretation = Int_linear_poly of (('a * nat) * (int * int list)) |
  Rat_linear_poly of (('a * nat) * (rat * rat list)) |
  Arctic_linear_poly of (('a * nat) * (arctic * arctic list)) |
  Arctic_rat_linear_poly of
    (('a * nat) * (rat arctic_delta * rat arctic_delta list))
  | Real_linear_poly of (('a * nat) * (real * real list)) |
  Int_matrix of (('a * nat) * (int mat * int mat list)) |
  Rat_matrix of (('a * nat) * (rat mat * rat mat list)) |
  Arctic_matrix of (('a * nat) * (arctic mat * arctic mat list)) |
  Arctic_rat_matrix of
    (('a * nat) * (rat arctic_delta mat * rat arctic_delta mat list))
  | Real_matrix of (('a * nat) * (real mat * real mat list)) |
  Int_non_linear_poly of (('a * nat) * (nat monom * int) list) |
  Rat_non_linear_poly of (('a * nat) * (nat monom * rat) list) |
  Real_non_linear_poly of (('a * nat) * (nat monom * real) list);;

type relation_kind = Strict_TRS | Weak_TRS of nat option;;

type cert_result = Certified | Unsupported of string | Error of string;;

type ('a, 'b) subst_incr =
  Abs_subst_incr of
    (('b -> ('a, 'b) term) * ('b set * (('a, 'b) term -> 'b list)));;

type ('a, 'b, 'c, 'd, 'e) art_edge = Cover of 'e |
  Children of (('a, 'b, 'c, 'd) transition_rule * 'e) list;;

type ('a, 'b, 'c, 'd) memory_ext =
  Memory_ext of
    (unit -> 'a) * ('a -> 'b -> 'c option) * ('a -> 'b * 'c -> 'a) * 'd;;

type ('a, 'b) ta_rule_impl =
  TA_rule_impl of 'b * 'a list * 'a * ('a, unit) rbt;;

type ('a, 'b) ta_impl =
  TA_Impl of
    ('a, unit) rbt * (('b * nat), (('a, 'b) ta_rule_impl list)) rbt * 'a list *
      ('a, unit) rbt * ('a * 'a) list * ('a -> ('a, unit) rbt) *
      ('a -> ('a, unit) rbt);;

type ('b, 'a) tp =
  TP of (bool *
          (('b, 'a) term list *
            (bool *
              ((('b, 'a) term * ('b, 'a) term) list *
                ((('b, 'a) term * ('b, 'a) term) list *
                  ((('b * nat), ((bool * (('b, 'a) term * ('b, 'a) term)) list))
                     rbt *
                    (('b, 'a) term -> bool)))))));;

type ('a, 'b, 'c, 'd, 'e, 'f) art_ext =
  Art_ext of
    'e list * 'e list * ('e -> ('a, 'b, 'c, 'd, 'e) art_edge) * ('e -> 'd) *
      ('e -> ('a, ('b * 'c)) term formula) * 'f;;

type ('b, 'a) ta_code =
  Abs_ta_code of
    (('b, unit) rbta *
      ((('b, 'a) ta_rule, unit) rbta *
        (('b * 'b) list *
          (('b, unit) rbta *
            ((('a * nat), (('b list * 'b), unit) rbta) rbta *
              (bool *
                (('b -> ('b, unit) rbta) * ('b -> ('b, unit) rbta))))))));;

type ('a, 'b) c_constraint =
  Conditional_C of
    bool * (('a, 'b) term * ('a, 'b) term) * (('a, 'b) term * ('a, 'b) term)
  | Unconditional_C of bool * (('a, 'b) term * ('a, 'b) term);;

type ('b, 'a) dpp =
  DPP of
    (bool *
      (bool *
        ((('b, 'a) term * ('b, 'a) term) list *
          ((('b, 'a) term * ('b, 'a) term) list *
            (('b, 'a) term list *
              (bool *
                (bool *
                  ((('b, 'a) term * ('b, 'a) term) list *
                    ((('b, 'a) term * ('b, 'a) term) list *
                      ((('b * nat),
                         ((bool * (('b, 'a) term * ('b, 'a) term)) list))
                         rbt *
                        ((('b * nat),
                           ((bool * (('b, 'a) term * ('b, 'a) term)) list))
                           rbt *
                          (bool * (('b, 'a) term -> bool)))))))))))));;

type 'a poly_constraint = Poly_Ge of ('a monom * int) list |
  Poly_Eq of ('a monom * int) list;;

type ('a, 'b, 'c, 'd, 'e) hinter_ext =
  Hinter_ext of
    'c list * ('b -> 'd list) * ('c -> 'a) * ('c -> ('a list) option) * 'e;;

type ('b, 'a) ac_dpp =
  AC_DPP of
    ((('b, 'a) term * ('b, 'a) term) list *
      ((('b, 'a) term * ('b, 'a) term) list *
        ((('b, 'a) term * ('b, 'a) term) list *
          ((('b, 'a) term * ('b, 'a) term) list *
            ((('b, 'a) term * ('b, 'a) term) list *
              (('b * nat), ((unit * (('b, 'a) term * ('b, 'a) term)) list))
                rbt)))));;

type ('a, 'b, 'c, 'd) gen_g_impl_ext = Gen_g_impl_ext of 'a * 'b * 'c * 'd;;

type ('a, 'b, 'c) redord_ext =
  Redord_ext of
    ((string -> string), unit) sum * (('a, 'b) term -> ('a, 'b) term -> bool) *
      'a * 'c;;

type condition_type = Bound | Strict | Non_Strict;;

type ('a, 'b, 'c) redtriple_ext =
  Redtriple_ext of
    ((string -> string), unit) sum *
      (('a, 'b) term * ('a, 'b) term -> ((string -> string), unit) sum) *
      (('a, 'b) term * ('a, 'b) term -> ((string -> string), unit) sum) *
      (('a, 'b) term * ('a, 'b) term -> ((string -> string), unit) sum) *
      ('a * nat -> nat set) * ('a * nat -> nat set) *
      ((('a, 'b) term * ('a, 'b) term) list -> ((string -> string), unit) sum) *
      (string -> string) * (('a * nat) list) option *
      (('a, 'b) complexity_measure ->
        complexity_class -> ((string -> string), unit) sum) *
      'c;;

type ('a, 'b) fp_loop_prf =
  FP_loop_prf of
    ('a, 'b) ctxt * ('b * ('a, 'b) term) list * ('a, 'b) term *
      (nat list * ((('a, 'b) term * ('a, 'b) term) * ('a, 'b) term)) list;;

type ('a, 'b, 'c, 'd, 'e) sl_ops_ext =
  Sl_ops_ext of
    ('a -> 'b list -> 'c) * ('a -> nat -> 'c -> bool) * ('a -> 'b list -> 'b) *
      'b list * 'b *
      ((('a, 'd) term * ('a, 'd) term) list -> ((string -> string), unit) sum) *
      ('a -> 'b list -> 'c) * ('a -> nat -> 'c -> bool) * ('c -> 'c list) *
      ('a -> nat -> 'c list) * 'e;;

type ('a, 'b, 'c, 'd) slm_ops_ext =
  Slm_ops_ext of
    ('a -> 'b list -> 'c) * ('a -> 'b list -> 'b) * 'b list * 'b *
      ('a -> 'b list -> 'c) * 'd;;

type ('a, 'b, 'c, 'd) tp_ops_ext =
  Tp_ops_ext of
    ('a ->
      bool *
        (('b, 'c) term set *
          ((('b, 'c) term * ('b, 'c) term) set *
            (('b, 'c) term * ('b, 'c) term) set))) *
      ('a -> ('b, 'c) term list) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) * ('a -> bool) *
      ('a -> ('b, 'c) term -> bool) * ('a -> bool) *
      ('a -> 'b * nat -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a ->
        (('b, 'c) term * ('b, 'c) term) list ->
          (('b, 'c) term * ('b, 'c) term) list -> 'a) *
      ('a ->
        (('b, 'c) term * ('b, 'c) term) list ->
          (('b, 'c) term * ('b, 'c) term) list *
            (('b, 'c) term * ('b, 'c) term) list) *
      (bool ->
        ('b, 'c) term list ->
          (('b, 'c) term * ('b, 'c) term) list ->
            (('b, 'c) term * ('b, 'c) term) list -> 'a) *
      ('a -> bool) * 'd;;

type ('a, 'b) partial_object_ext = Partial_object_ext of 'a set * 'b;;

type 'a cut_transition_split_info =
  Cut_Transition_Split_Info of ('a list) list;;

type ('a, 'b, 'c, 'd) dpp_ops_ext =
  Dpp_ops_ext of
    ('a ->
      bool *
        (bool *
          ((('b, 'c) term * ('b, 'c) term) set *
            ((('b, 'c) term * ('b, 'c) term) set *
              (('b, 'c) term set *
                ((('b, 'c) term * ('b, 'c) term) set *
                  (('b, 'c) term * ('b, 'c) term) set)))))) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a -> ('b, 'c) term list) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) * ('a -> bool) *
      ('a -> bool) * ('a -> bool) * ('a -> ('b, 'c) term -> bool) *
      ('a -> bool) * ('a -> 'b * nat -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a -> 'b * nat -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list -> 'a) *
      ('a ->
        ('b, 'c) term * ('b, 'c) term ->
          (('b, 'c) term * ('b, 'c) term) list -> 'a) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list -> 'a) *
      ('a ->
        (('b, 'c) term * ('b, 'c) term) list ->
          (('b, 'c) term * ('b, 'c) term) list -> 'a) *
      ('a ->
        (('b, 'c) term * ('b, 'c) term) list ->
          (('b, 'c) term * ('b, 'c) term) list -> 'a) *
      ('a ->
        (('b, 'c) term * ('b, 'c) term) list ->
          (('b, 'c) term * ('b, 'c) term) list *
            (('b, 'c) term * ('b, 'c) term) list) *
      ('a ->
        (('b, 'c) term * ('b, 'c) term) list ->
          (('b, 'c) term * ('b, 'c) term) list *
            (('b, 'c) term * ('b, 'c) term) list) *
      (bool ->
        bool ->
          (('b, 'c) term * ('b, 'c) term) list ->
            (('b, 'c) term * ('b, 'c) term) list ->
              ('b, 'c) term list ->
                (('b, 'c) term * ('b, 'c) term) list ->
                  (('b, 'c) term * ('b, 'c) term) list -> 'a) *
      ('a -> bool) * ('a -> bool) * ('a -> bool) * 'd;;

type ('a, 'b, 'c) root_redtriple_ext =
  Root_redtriple_ext of
    ((string -> string), unit) sum *
      (('a, 'b) term * ('a, 'b) term -> ((string -> string), unit) sum) *
      (('a, 'b) term * ('a, 'b) term -> ((string -> string), unit) sum) *
      (('a, 'b) term * ('a, 'b) term -> ((string -> string), unit) sum) *
      ('a * nat -> nat set) * ('a * nat -> nat set) * (string -> string) * 'c;;

type ('a, 'b, 'c) non_inf_order_ext =
  Non_inf_order_ext of
    ((string -> string), unit) sum *
      (('a, 'b) term * ('a, 'b) term -> ((string -> string), unit) sum) *
      (('a, 'b) c_constraint -> ((string -> string), unit) sum) *
      ('a * nat -> nat -> dependance) * (string -> string) * 'c;;

type ('a, 'b, 'c, 'd) ac_tp_ops_ext =
  Ac_tp_ops_ext of
    ('a -> (('b, 'c) term * ('b, 'c) term) set * ('b set * 'b set)) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) * ('a -> 'b list) *
      ('a -> 'b list) *
      ((('b, 'c) term * ('b, 'c) term) list -> 'b list -> 'b list -> 'a) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list -> 'a) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) * 'd;;

type ('a, 'b) ordered_semiring_ext =
  Ordered_semiring_ext of
    ('a -> 'a -> bool) * ('a -> 'a -> bool) * ('a -> 'a -> 'a) * 'b;;

type ('a, 'b, 'c, 'd) ac_dpp_ops_ext =
  Ac_dpp_ops_ext of
    ('a ->
      (('b, 'c) term * ('b, 'c) term) set *
        ((('b, 'c) term * ('b, 'c) term) set *
          ((('b, 'c) term * ('b, 'c) term) set *
            ((('b, 'c) term * ('b, 'c) term) set *
              (('b, 'c) term * ('b, 'c) term) set)))) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list) *
      ((('b, 'c) term * ('b, 'c) term) list ->
        (('b, 'c) term * ('b, 'c) term) list ->
          (('b, 'c) term * ('b, 'c) term) list ->
            (('b, 'c) term * ('b, 'c) term) list ->
              (('b, 'c) term * ('b, 'c) term) list -> 'a) *
      ('a -> 'b * nat -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a -> 'b * nat -> (('b, 'c) term * ('b, 'c) term) list) *
      ('a ->
        (('b, 'c) term * ('b, 'c) term) list ->
          (('b, 'c) term * ('b, 'c) term) list -> 'a) *
      ('a -> bool) * ('a -> bool) *
      ('a -> (('b, 'c) term * ('b, 'c) term) list -> 'a) * 'd;;

type ('a, 'b) lpoly_order_semiring_ext =
  Lpoly_order_semiring_ext of
    bool * 'a * ('a -> bool) * ('a -> bool) * ('a -> nat) *
      ('a -> nat -> ((string -> string), unit) sum) * (string -> string) * 'b;;

type 'a x_f_status_option_x_x_nat_list_nat_x_f_prod_fun_option =
  Abs_x_f_status_option_x_x_nat_list_nat_x_f_prod_fun_option of
    ('a * nat -> nat list) option;;

type 'a x_compare_order_x_x_f_afs_option_x_x_nat_x_compare_order_x_x_f_prod_set_af_entry_nat_x_compare_order_x_x_f_prod_fun_prod_option
  = Abs_x_compare_order_x_x_f_afs_option_x_x_nat_x_compare_order_x_x_f_prod_set_af_entry_nat_x_compare_order_x_x_f_prod_fun_prod_option
      of (('a * nat -> af_entry) * ('a * nat) set) option;;

let rec dom _A m = collect _A (fun a -> not (is_none (m a)));;

let rec dlist_ex _A x xc = list_ex x (list_of_dlist _A xc);;

let rec rBT_Impl_rbt_ex
  p x1 = match p, x1 with
    p, Branch (c, l, k, v, r) ->
      p k v || (rBT_Impl_rbt_ex p l || rBT_Impl_rbt_ex p r)
    | p, Empty -> false;;

let rec ex _A xb xc = rBT_Impl_rbt_ex xb (impl_ofb _A xc);;

let rec bex (_A1, _A2)
  x0 p = match x0, p with
    RBT_set rbt, p ->
      (match ccompare _A2
        with None ->
          failwith "Bex RBT_set: ccompare = None"
            (fun _ -> bex (_A1, _A2) (RBT_set rbt) p)
        | Some _ -> ex _A2 (fun k _ -> p k) rbt)
    | DList_set dxs, p ->
        (match ceq _A1
          with None ->
            failwith "Bex DList_set: ceq = None"
              (fun _ -> bex (_A1, _A2) (DList_set dxs) p)
          | Some _ -> dlist_ex _A1 p dxs)
    | Set_Monad xs, p -> list_ex p xs;;

let rec zip xs ys = match xs, ys with x :: xs, y :: ys -> (x, y) :: zip xs ys
              | xs, [] -> []
              | [], ys -> [];;

let rec skol (_A1, _A2)
  v x1 = match v, x1 with
    v, Var x -> (if member (_A1, _A2) x v then Fun (Inr x, []) else Var x)
    | v, Fun (f, ts) -> Fun (Inl f, map (skol (_A1, _A2) v) ts);;

let rec drop
  n x1 = match n, x1 with n, [] -> []
    | n, x :: xs ->
        (if equal_nata n zero_nata then x :: xs
          else drop (minus_nata n one_nata) xs);;

let rec find uu x1 = match uu, x1 with uu, [] -> None
               | p, x :: xs -> (if p x then Some x else find p xs);;

let rec last (x :: xs) = (if null xs then x else last xs);;

let rec maps f x1 = match f, x1 with f, [] -> []
               | f, x :: xs -> f x @ maps f xs;;

let rec take
  n x1 = match n, x1 with n, [] -> []
    | n, x :: xs ->
        (if equal_nata n zero_nata then []
          else x :: take (minus_nata n one_nata) xs);;

let rec empty _A = RBT Empty;;

let rec bind x0 f = match x0, f with None, f -> None
               | Some x, f -> f x;;

let rec zip_option
  x0 x1 = match x0, x1 with [], [] -> Some []
    | x :: xs, y :: ys ->
        bind (zip_option xs ys) (fun zs -> Some ((x, y) :: zs))
    | x :: xs, [] -> None
    | [], y :: ys -> None;;

let rec guard b = (if b then Some () else None);;

let rec match_lista _A
  = function
    (GCFun (f, ss), Var x) :: ps ->
      bind (match_lista _A ps) (fun psa -> Some ((GCFun (f, ss), x) :: psa))
    | (GCFun (f, ss), Fun (g, ts)) :: ps ->
        bind (guard (eq _A f g))
          (fun _ ->
            bind (zip_option ss ts) (fun psa -> match_lista _A (psa @ ps)))
    | (GCHole, t) :: ps -> match_lista _A ps
    | [] -> Some [];;

let rec merge_lists _A
  x0 x1 = match x0, x1 with c :: cs, [] -> None
    | [], d :: ds -> None
    | GCFun (f, ss) :: cs, GCFun (g, ts) :: ds ->
        bind (guard (eq _A f g))
          (fun _ ->
            bind (merge_lists _A ss ts)
              (fun us ->
                bind (merge_lists _A cs ds)
                  (fun es -> Some (GCFun (f, us) :: es))))
    | c :: cs, GCHole :: ds ->
        bind (merge_lists _A cs ds) (fun es -> Some (c :: es))
    | GCHole :: cs, d :: ds ->
        bind (merge_lists _A cs ds) (fun es -> Some (d :: es))
    | [], [] -> Some [];;

let rec mergea _A
  c d = bind (merge_lists _A [c] [d]) (fun es -> Some (nth es zero_nata));;

let rec merge_var _A _B
  x c xa2 = match x, c, xa2 with
    x, c, (d, y) :: ps ->
      (if eq _A x y then bind (mergea _B c d) (fun e -> merge_var _A _B x e ps)
        else bind (merge_var _A _B x c ps)
               (fun (b, psa) -> Some (b, (d, y) :: psa)))
    | x, c, [] -> Some ((c, x), []);;

let rec merge_all _A _B
  = function [] -> Some []
    | (c, x) :: ps ->
        bind (merge_var _B _A x c ps)
          (fun (cx, psa) ->
            bind (merge_all _A _B psa) (fun psb -> Some (cx :: psb)));;

let rec matchd _A _B (c, t) = bind (match_lista _A [(c, t)]) (merge_all _A _B);;

let rec matchc _A _B c t = not (is_none (matchd _A _B (c, t)));;

let rec tcap (_A1, _A2) (_B1, _B2)
  uu x1 = match uu, x1 with uu, Var uv -> GCHole
    | r, Fun (f, ts) ->
        (let h = GCFun (f, map (tcap (_A1, _A2) (_B1, _B2) r) ts) in
          (if bex ((ceq_prod (ceq_term _A2 _B2) (ceq_term _A2 _B2)),
                    (ccompare_prod (ccompare_term _A1 _B1)
                      (ccompare_term _A1 _B1)))
                r (fun ra -> matchc _A2 _B2 h (fst ra))
            then GCHole else h));;

let rec root = function Var x -> None
               | Fun (f, ts) -> Some (f, size_list ts);;

let rec transitions_impl (Lts_Impl (x1, x2, x3)) = x2;;

let rec initiala (Lts_Impl (x1, x2, x3)) = x1;;

let rec assertion_impl (Lts_Impl (x1, x2, x3)) = x3;;

let rec lookupa _A xa = map_of _A (impl_ofa xa);;

let rec lookupb (_A1, _A2) = function RBT_Mapping t -> lookupd _A1 t
                             | Assoc_List_Mapping al -> lookupa _A2 al;;

let rec lookup_default (_B1, _B2)
  d m k = (match lookupb (_B1, _B2) m k with None -> d | Some v -> v);;

let rec update _A
  k v x2 = match k, v, x2 with k, v, [] -> [(k, v)]
    | k, v, p :: ps ->
        (if eq _A (fst p) k then (k, v) :: ps else p :: update _A k v ps);;

let rec updatea _A xc xd xe = Alist (update _A xc xd (impl_ofa xe));;

let rec fun_upd _A f a b = (fun x -> (if eq _A x a then b else f x));;

let rec updateb (_A1, _A2)
  k v x2 = match k, v, x2 with
    k, v, RBT_Mapping t ->
      (match ccompare _A1
        with None ->
          failwith "update RBT_Mapping: ccompare = None"
            (fun _ -> updateb (_A1, _A2) k v (RBT_Mapping t))
        | Some _ -> RBT_Mapping (inserte _A1 k v t))
    | k, v, Assoc_List_Mapping al -> Assoc_List_Mapping (updatea _A2 k v al)
    | k, v, Mapping m -> Mapping (fun_upd _A2 m k (Some v));;

let rec mapping_empty_choose _A
  = (match ccompare _A with None -> Assoc_List_Mapping emptya
      | Some _ -> RBT_Mapping (emptye _A));;

let rec mapping_empty _A = function Mapping_RBT -> RBT_Mapping (emptye _A)
                           | Mapping_Assoc_List -> Assoc_List_Mapping emptya
                           | Mapping_Mapping -> Mapping (fun _ -> None)
                           | Mapping_Choose -> mapping_empty_choose _A;;

let rec emptyb (_A1, _A2) = mapping_empty _A1 (of_phantom (mapping_impl _A2));;

let rec of_alist (_A1, _A2, _A3)
  xs = foldr (fun (a, b) -> updateb (_A1, _A2) a b) xs (emptyb (_A1, _A3));;

let rec map_of_default (_B1, _B2, _B3)
  d xs = lookup_default (_B1, _B2) d (of_alist (_B1, _B2, _B3) xs);;

let rec assertion_of (_D1, _D2, _D3)
  pi = map_of_default (_D1, _D2, _D3) (Conjunction []) (assertion_impl pi);;

let rec lts_of (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2, _D3, _D4, _D5)
  (_E1, _E2, _E3)
  pi = Lts_ext
         (set (_D1, _D2, _D5) (initiala pi),
           image ((ceq_prod _E1 (ceq_transition_rule _A2 _B2 _C2 _D3)),
                   (ccompare_prod _E2
                     (ccompare_transition_rule _A1 _B1 _C1 _D2)))
             ((ceq_transition_rule _A2 _B2 _C2 _D3),
               (ccompare_transition_rule _A1 _B1 _C1 _D2),
               set_impl_transition_rule)
             snd (set ((ceq_prod _E1 (ceq_transition_rule _A2 _B2 _C2 _D3)),
                        (ccompare_prod _E2
                          (ccompare_transition_rule _A1 _B1 _C1 _D2)),
                        (set_impl_prod _E3 set_impl_transition_rule))
                   (transitions_impl pi)),
           assertion_of (_D2, _D3, _D4) pi, ());;

let rec source (Transition (l, uu, uv)) = l;;

let rec target (Transition (uu, r, uv)) = r;;

let rec membera _A x0 y = match x0, y with [], y -> false
                     | x :: xs, y -> eq _A x y || membera _A xs y;;

let rec insertb _A x xs = (if membera _A xs x then xs else x :: xs);;

let rec union _A = fold (insertb _A);;

let rec tabulate
  (n, f) =
    IArray (map (comp f integer_of_nat) (upt zero_nata (nat_of_integer n)));;

let rec of_fun f n = tabulate (integer_of_nat n, comp f nat_of_integer);;

let rec vec_of_fun xb xc = Abs_vec_impl (xb, of_fun xc xb);;

let rec vec n f = Vec_impl (vec_of_fun n f);;

let rec col a j = vec (dim_row a) (fun i -> index_mat a (i, j));;

let rec mat_of_fun
  xc xd xe =
    Abs_mat_impl
      (xc, (xd, of_fun (fun i -> of_fun (fun j -> xe (i, j)) xd) xc));;

let rec mat nr nc f = Mat_impl (mat_of_fun nr nc f);;

let rec row a i = vec (dim_col a) (fun j -> index_mat a (i, j));;

let rec funpow
  n f = (if equal_nata n zero_nata then id
          else comp f (funpow (minus_nata n one_nata) f));;

let rec vala qd delta = plus_rata (qdfst qd) (times_rata delta (qdsnd qd));;

let rec rbt_del _A
  x xa1 = match x, xa1 with
    x, Branch (c, a, y, s, b) ->
      (match compare _A.compare_compare_order x y with Eq -> combine a b
        | Lt -> rbt_del_from_left _A x a y s b
        | Gt -> rbt_del_from_right _A x a y s b)
    | x, Empty -> Empty
and rbt_del_from_left _A
  x xa1 y s b = match x, xa1, y, s, b with
    x, Branch (B, lt, z, v, rt), y, s, b ->
      balance_left (rbt_del _A x (Branch (B, lt, z, v, rt))) y s b
    | x, Empty, y, s, b -> Branch (R, rbt_del _A x Empty, y, s, b)
    | x, Branch (R, va, vb, vc, vd), y, s, b ->
        Branch (R, rbt_del _A x (Branch (R, va, vb, vc, vd)), y, s, b)
and rbt_del_from_right _A
  x a y s xa4 = match x, a, y, s, xa4 with
    x, a, y, s, Branch (B, lt, z, v, rt) ->
      balance_right a y s (rbt_del _A x (Branch (B, lt, z, v, rt)))
    | x, a, y, s, Empty -> Branch (R, a, y, s, rbt_del _A x Empty)
    | x, a, y, s, Branch (R, va, vb, vc, vd) ->
        Branch (R, a, y, s, rbt_del _A x (Branch (R, va, vb, vc, vd)));;

let rec rbt_delete _A k t = paint B (rbt_del _A k t);;

let rec impl_of _B (RBT x) = x;;

let rec delete _A
  xb xc = RBT (rbt_delete _A xb (impl_of _A.linorder_compare_order xc));;

let rec rbt_ins _A
  f k v x3 = match f, k, v, x3 with
    f, k, v, Branch (R, l, x, y, r) ->
      (match compare _A.compare_compare_order k x
        with Eq -> Branch (R, l, x, f k y v, r)
        | Lt -> Branch (R, rbt_ins _A f k v l, x, y, r)
        | Gt -> Branch (R, l, x, y, rbt_ins _A f k v r))
    | f, k, v, Branch (B, l, x, y, r) ->
        (match compare _A.compare_compare_order k x
          with Eq -> Branch (B, l, x, f k y v, r)
          | Lt -> balance (rbt_ins _A f k v l) x y r
          | Gt -> balance l x y (rbt_ins _A f k v r))
    | f, k, v, Empty -> Branch (R, Empty, k, v, Empty);;

let rec rbt_insert_with_key _A f k v t = paint B (rbt_ins _A f k v t);;

let rec rbt_insert _A = rbt_insert_with_key _A (fun _ _ nv -> nv);;

let rec insert _A
  xc xd xe = RBT (rbt_insert _A xc xd (impl_of _A.linorder_compare_order xe));;

let rec rbt_lookup _A
  x0 k = match x0, k with
    Branch (uu, l, x, y, r), k ->
      (match compare _A.compare_compare_order k x with Eq -> Some y
        | Lt -> rbt_lookup _A l k | Gt -> rbt_lookup _A r k)
    | Empty, k -> None;;

let rec lookup _A x = rbt_lookup _A (impl_of _A.linorder_compare_order x);;

let rec proja
  (Fun (f, ts)) i =
    (if less_nat i (size_list ts) then nth ts i else Fun (f, ts));;

let rec filter (_A1, _A2) p a = inf_seta (_A1, _A2) a (Collect_set p);;

let rec subst _A x t = fun_upd _A (fun a -> Var a) x t;;

let rec debug i t x = x;;

let rec merge _A qs ps = foldr (fun (a, b) -> update _A a b) ps qs;;

let rec concat xss = foldr (fun a b -> a @ b) xss [];;

let rec lterms
  pi = (fun (Fun (f, ts)) ->
         map (fun (i, a) -> (proja (Fun (f, ts)) i, a))
           (pi (f, size_list ts)));;

let rec lhs (l, r) = l;;

let rec rhs (l, r) = r;;

let rec insert_vars_term _B
  x0 xs = match x0, xs with Var x, xs -> insertb _B x xs
    | Fun (f, ts), xs -> foldr (insert_vars_term _B) ts xs;;

let rec is_Var = function Var x1 -> true
                 | Fun (x21, x22) -> false;;

let rec wf_rule _B
  r = not (is_Var (fst r)) &&
        list_all (membera _B (insert_vars_term _B (fst r) []))
          (insert_vars_term _B (snd r) []);;

let rec fun_of _A vec x = the (map_of _A vec x);;

let letters : char list
  = [Chara (true, false, false, false, false, true, true, false);
      Chara (false, true, false, false, false, true, true, false);
      Chara (true, true, false, false, false, true, true, false);
      Chara (false, false, true, false, false, true, true, false);
      Chara (true, false, true, false, false, true, true, false);
      Chara (false, true, true, false, false, true, true, false);
      Chara (true, true, true, false, false, true, true, false);
      Chara (false, false, false, true, false, true, true, false);
      Chara (true, false, false, true, false, true, true, false);
      Chara (false, true, false, true, false, true, true, false);
      Chara (true, true, false, true, false, true, true, false);
      Chara (false, false, true, true, false, true, true, false);
      Chara (true, false, true, true, false, true, true, false);
      Chara (false, true, true, true, false, true, true, false);
      Chara (true, true, true, true, false, true, true, false);
      Chara (false, false, false, false, true, true, true, false);
      Chara (true, false, false, false, true, true, true, false);
      Chara (false, true, false, false, true, true, true, false);
      Chara (true, true, false, false, true, true, true, false);
      Chara (false, false, true, false, true, true, true, false);
      Chara (true, false, true, false, true, true, true, false);
      Chara (false, true, true, false, true, true, true, false);
      Chara (true, true, true, false, true, true, true, false);
      Chara (false, false, false, true, true, true, true, false);
      Chara (true, false, false, true, true, true, true, false);
      Chara (false, true, false, true, true, true, true, false);
      Chara (true, false, false, false, false, false, true, false);
      Chara (false, true, false, false, false, false, true, false);
      Chara (true, true, false, false, false, false, true, false);
      Chara (false, false, true, false, false, false, true, false);
      Chara (true, false, true, false, false, false, true, false);
      Chara (false, true, true, false, false, false, true, false);
      Chara (true, true, true, false, false, false, true, false);
      Chara (false, false, false, true, false, false, true, false);
      Chara (true, false, false, true, false, false, true, false);
      Chara (false, true, false, true, false, false, true, false);
      Chara (true, true, false, true, false, false, true, false);
      Chara (false, false, true, true, false, false, true, false);
      Chara (true, false, true, true, false, false, true, false);
      Chara (false, true, true, true, false, false, true, false);
      Chara (true, true, true, true, false, false, true, false);
      Chara (false, false, false, false, true, false, true, false);
      Chara (true, false, false, false, true, false, true, false);
      Chara (false, true, false, false, true, false, true, false);
      Chara (true, true, false, false, true, false, true, false);
      Chara (false, false, true, false, true, false, true, false);
      Chara (true, false, true, false, true, false, true, false);
      Chara (false, true, true, false, true, false, true, false);
      Chara (true, true, true, false, true, false, true, false);
      Chara (false, false, false, true, true, false, true, false);
      Chara (true, false, false, true, true, false, true, false);
      Chara (false, true, false, true, true, false, true, false);
      Chara (true, true, true, true, true, false, true, false);
      Chara (false, false, false, false, true, true, false, false);
      Chara (true, false, false, false, true, true, false, false);
      Chara (false, true, false, false, true, true, false, false);
      Chara (true, true, false, false, true, true, false, false);
      Chara (false, false, true, false, true, true, false, false);
      Chara (true, false, true, false, true, true, false, false);
      Chara (false, true, true, false, true, true, false, false);
      Chara (true, true, true, false, true, true, false, false);
      Chara (false, false, false, true, true, true, false, false);
      Chara (true, false, false, true, true, true, false, false);
      Chara (false, true, true, false, false, true, false, false);
      Chara (true, true, false, true, true, true, false, false);
      Chara (false, true, false, true, true, true, false, false);
      Chara (true, false, true, true, false, true, false, false)];;

let rec bind2 (Sumbot a) f g = (match a with Inl aa -> f aa | Inr aa -> g aa);;

let rec hd _A xa = hda (list_of_dlist _A xa);;

let rec tl _A xa = Abs_dlist (tla (list_of_dlist _A xa));;

let rec foldli
  x0 c f sigma = match x0, c, f, sigma with [], c, f, sigma -> sigma
    | x :: xs, c, f, sigma ->
        (if c sigma then foldli xs c f (f x sigma) else sigma);;

let rec replicate
  n x = (if equal_nata n zero_nata then []
          else x :: replicate (minus_nata n one_nata) x);;

let rec scf_list
  scf xs =
    maps (fun (x, i) -> replicate (scf i) x)
      (zip xs (upt zero_nata (size_list xs)));;

let rec scf_term
  scf x1 = match scf, x1 with scf, Var x -> Var x
    | scf, Fun (f, ts) ->
        Fun (f, scf_list (scf (f, size_list ts)) (map (scf_term scf) ts));;

let rec butlast = function [] -> []
                  | x :: xs -> (if null xs then [] else x :: butlast xs);;

let rec extract
  p x1 = match p, x1 with
    p, x :: xs ->
      (if p x then Some ([], (x, xs))
        else (match extract p xs with None -> None
               | Some (ys, (y, zs)) -> Some (x :: ys, (y, zs))))
    | p, [] -> None;;

let rec productb _A _B
  dxs1 dxs2 =
    Abs_dlist
      (foldc _A (fun a -> foldc _B (fun c -> (fun b -> (a, c) :: b)) dxs2) dxs1
        []);;

let rec rbt_product
  f rbt1 rbt2 =
    rbtreeify
      (rev (folda
             (fun a b ->
               folda (fun c d -> (fun e -> ((a, c), f a b c d) :: e)) rbt2)
             rbt1 []));;

let rec productd _A _B
  xc xd xe = Mapping_RBTa (rbt_product xc (impl_ofb _A xd) (impl_ofb _B xe));;

let rec producta _A _B
  rbt1 rbt2 = productd _A _B (fun _ _ _ _ -> ()) rbt1 rbt2;;

let rec productc (_A1, _A2, _A3) (_B1, _B2, _B3)
  a2 b2 = match a2, b2 with
    RBT_set rbt1, RBT_set rbt2 ->
      (match ccompare _A2
        with None ->
          failwith "product RBT_set RBT_set: ccompare1 = None"
            (fun _ ->
              productc (_A1, _A2, _A3) (_B1, _B2, _B3) (RBT_set rbt1)
                (RBT_set rbt2))
        | Some _ ->
          (match ccompare _B2
            with None ->
              failwith "product RBT_set RBT_set: ccompare2 = None"
                (fun _ ->
                  productc (_A1, _A2, _A3) (_B1, _B2, _B3) (RBT_set rbt1)
                    (RBT_set rbt2))
            | Some _ -> RBT_set (producta _A2 _B2 rbt1 rbt2)))
    | a2, RBT_set rbt2 ->
        (match ccompare _B2
          with None ->
            failwith "product RBT_set: ccompare2 = None"
              (fun _ ->
                productc (_A1, _A2, _A3) (_B1, _B2, _B3) a2 (RBT_set rbt2))
          | Some _ ->
            foldb _B2
              (fun y ->
                sup_seta ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2))
                  (image (_A1, _A2)
                    ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2),
                      (set_impl_prod _A3 _B3))
                    (fun x -> (x, y)) a2))
              rbt2
              (bot_set
                ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2),
                  (set_impl_prod _A3 _B3))))
    | RBT_set rbt1, b2 ->
        (match ccompare _A2
          with None ->
            failwith "product RBT_set: ccompare1 = None"
              (fun _ ->
                productc (_A1, _A2, _A3) (_B1, _B2, _B3) (RBT_set rbt1) b2)
          | Some _ ->
            foldb _A2
              (fun x ->
                sup_seta ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2))
                  (image (_B1, _B2)
                    ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2),
                      (set_impl_prod _A3 _B3))
                    (fun a -> (x, a)) b2))
              rbt1
              (bot_set
                ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2),
                  (set_impl_prod _A3 _B3))))
    | DList_set dxs, DList_set dys ->
        (match ceq _A1
          with None ->
            failwith "product DList_set DList_set: ceq1 = None"
              (fun _ ->
                productc (_A1, _A2, _A3) (_B1, _B2, _B3) (DList_set dxs)
                  (DList_set dys))
          | Some _ ->
            (match ceq _B1
              with None ->
                failwith "product DList_set DList_set: ceq2 = None"
                  (fun _ ->
                    productc (_A1, _A2, _A3) (_B1, _B2, _B3) (DList_set dxs)
                      (DList_set dys))
              | Some _ -> DList_set (productb _A1 _B1 dxs dys)))
    | a1, DList_set dys ->
        (match ceq _B1
          with None ->
            failwith "product DList_set2: ceq = None"
              (fun _ ->
                productc (_A1, _A2, _A3) (_B1, _B2, _B3) a1 (DList_set dys))
          | Some _ ->
            foldc _B1
              (fun y ->
                sup_seta ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2))
                  (image (_A1, _A2)
                    ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2),
                      (set_impl_prod _A3 _B3))
                    (fun x -> (x, y)) a1))
              dys (bot_set
                    ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2),
                      (set_impl_prod _A3 _B3))))
    | DList_set dxs, b1 ->
        (match ceq _A1
          with None ->
            failwith "product DList_set1: ceq = None"
              (fun _ ->
                productc (_A1, _A2, _A3) (_B1, _B2, _B3) (DList_set dxs) b1)
          | Some _ ->
            foldc _A1
              (fun x ->
                sup_seta ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2))
                  (image (_B1, _B2)
                    ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2),
                      (set_impl_prod _A3 _B3))
                    (fun a -> (x, a)) b1))
              dxs (bot_set
                    ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2),
                      (set_impl_prod _A3 _B3))))
    | Set_Monad xs, Set_Monad ys ->
        Set_Monad
          (fold (fun x -> fold (fun y -> (fun a -> (x, y) :: a)) ys) xs [])
    | a, b ->
        Collect_set
          (fun (x, y) -> member (_A1, _A2) x a && member (_B1, _B2) y b);;

let rec set_Cons (_A1, _A2, _A3)
  a xs =
    image ((ceq_prod _A1 (ceq_list _A1)),
            (ccompare_prod _A2 (ccompare_list _A2)))
      ((ceq_list _A1), (ccompare_list _A2), set_impl_list)
      (fun (aa, b) -> aa :: b)
      (productc (_A1, _A2, _A3)
        ((ceq_list _A1), (ccompare_list _A2), set_impl_list)
        (inf_seta (_A1, _A2) (image (_A1, _A2) (_A1, _A2, _A3) (fun x -> x) a)
          (top_set (_A1, _A2, _A3)))
        (inf_seta ((ceq_list _A1), (ccompare_list _A2))
          (top_set ((ceq_list _A1), (ccompare_list _A2), set_impl_list))
          (image ((ceq_list _A1), (ccompare_list _A2))
            ((ceq_list _A1), (ccompare_list _A2), set_impl_list)
            (fun xsa -> xsa) xs)));;

let rec listset (_A1, _A2, _A3)
  = function
    [] -> inserta ((ceq_list _A1), (ccompare_list _A2)) []
            (set_empty ((ceq_list _A1), (ccompare_list _A2))
              (of_phantom set_impl_lista))
    | a :: asa -> set_Cons (_A1, _A2, _A3) a (listset (_A1, _A2, _A3) asa);;

let rec remdups _A
  = function [] -> []
    | x :: xs ->
        (if membera _A xs x then remdups _A xs else x :: remdups _A xs);;

let rec l2m_lookup _A _B
  x0 uu = match x0, uu with [], uu -> None
    | (a, kvs) :: reca, (k1, k2) ->
        (match eq _A k1 a with true -> map_of _B kvs k2
          | false -> l2m_lookup _A _B reca (k1, k2));;

let rec lm_store_acc _A
  x0 kv accu = match x0, kv, accu with [], kv, accu -> kv :: accu
    | (a, uu) :: reca, (k, v), accu ->
        (match eq _A k a with true -> accu
          | false -> lm_store_acc _A reca (k, v) accu);;

let rec lm_store _A m kv = lm_store_acc _A m kv m;;

let rec l2m_store _A _B
  x0 x1 = match x0, x1 with [], ((k1, k2), v) -> [(k1, [(k2, v)])]
    | (a, kvs) :: reca, ((k1, k2), v) ->
        (match eq _A k1 a with true -> (a, lm_store _B kvs (k2, v)) :: reca
          | false -> (a, kvs) :: l2m_store _A _B reca ((k1, k2), v));;

let rec l2m _A _B
  = Memory_ext ((fun _ -> []), l2m_lookup _A _B, l2m_store _A _B, ());;

let rec aux _A _B
  key ma v m =
    (match key v with None -> m
      | Some k ->
        (match lookup _B ma k with None -> m
          | Some ws ->
            (if membera _A ws v
              then (match lookup _B m k with None -> insert _B k [v] m
                     | Some vs -> insert _B k (insertb _A v vs) m)
              else m)));;

let rec ma_coeff xa = fst (snd (rep_mini_alg xa));;

let rec mau_coeff xa = ma_coeff (rep_mini_alg_unique xa);;

let rec root_int_maina
  pm ipm ip x n =
    (let xpm = power power_int x pm in
     let xp = times_inta xpm x in
      (if less_eq_int xp n then (x, equal_inta xp n)
        else root_int_maina pm ipm ip
               (divide_int (plus_inta (divide_int n xpm) (times_inta x ipm)) ip)
               n));;

let rec root_int_main
  p n = (if equal_nata p zero_nata then (one_inta, equal_inta n one_inta)
          else (let pm = minus_nata p one_nata in
                 root_int_maina pm (int_of_nat pm) (int_of_nat p)
                   (start_value n p) n));;

let rec root_int_floor_pos
  p x = (if equal_nata p zero_nata then zero_inta
          else fst (root_int_main p x));;

let rec root_nat_floor p x = root_int_floor_pos p (int_of_nat x);;

let rec sqrt_int
  x = (if less_int x zero_inta then []
        else (match sqrt_int_main x
               with (y, true) ->
                 (if equal_inta y zero_inta then [zero_inta]
                   else [y; uminus_inta y])
               | (_, false) -> []));;

let rec sqrt_nat x = map nat (take one_nata (sqrt_int (int_of_nat x)));;

let rec dvd (_A1, _A2)
  a b = eq _A1
          (modulo _A2.semiring_modulo_semidom_modulo.modulo_semiring_modulo b a)
          (zerob
            _A2.algebraic_semidom_semidom_modulo.semidom_divide_algebraic_semidom.semidom_semidom_divide.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero);;

let rec prime_product_factor_main
  factor_sq factor_pr limit n i =
    (if less_eq_nat i limit &&
          less_eq_nat (nat_of_integer (Big_int.big_int_of_int 2)) i
      then (if dvd (equal_nat, semidom_modulo_nat) i n
             then (let na = divide_nata n i in
                    (if dvd (equal_nat, semidom_modulo_nat) i na
                      then (let nb = divide_nata na i in
                             prime_product_factor_main (times_nata factor_sq i)
                               factor_pr
                               (nat (root_nat_floor
                                      (nat_of_integer
(Big_int.big_int_of_int 3))
                                      nb))
                               nb i)
                      else (match sqrt_nat na
                             with [] ->
                               prime_product_factor_main factor_sq
                                 (times_nata factor_pr i)
                                 (nat (root_nat_floor
(nat_of_integer (Big_int.big_int_of_int 3)) na))
                                 na (suc i)
                             | sn :: _ ->
                               (times_nata factor_sq sn,
                                 times_nata factor_pr i))))
             else prime_product_factor_main factor_sq factor_pr limit n (suc i))
      else (factor_sq, times_nata factor_pr n));;

let rec prime_product_factor
  n = (match sqrt_nat n
        with [] ->
          prime_product_factor_main one_nata one_nata
            (nat (root_nat_floor (nat_of_integer (Big_int.big_int_of_int 3)) n))
            n (nat_of_integer (Big_int.big_int_of_int 2))
        | s :: _ -> (s, one_nata));;

let rec ma_sqrt
  xa = Abs_mini_alg
         (let (p, (_, _)) = rep_mini_alg xa in
          let a = quotient_of p in
          let (aa, b) = a in
          let aaa = abs_inta (times_inta aa b) in
           (match sqrt_int aaa
             with [] -> (zero_rata, (inverse_rata (of_int b), nat aaa))
             | s :: _ ->
               (divide_rata (of_int s) (of_int b), (zero_rata, zero_nata))));;

let rec ma_rat xa = fst (rep_mini_alg xa);;

let rec mau_sqrt
  xa = Abs_mini_alg_unique
         (let a = quotient_of (ma_rat (rep_mini_alg_unique xa)) in
          let (aa, b) = a in
          let (sq, fact) =
            prime_product_factor (nat (times_inta (abs_inta aa) b)) in
          let ma =
            ma_of_rat
              (divide_rata
                (times_rata (of_int (sgn_inta aa)) (of_nat semiring_1_rat sq))
                (of_int b))
            in
           ma_times ma (ma_sqrt (ma_of_rat (of_nat semiring_1_rat fact))));;

let rec sqrt
  (Real_of_u r) =
    (if equal_rata (mau_coeff r) zero_rata then Real_of_u (mau_sqrt r)
      else failwith "cannot represent sqrt of irrational number"
             (fun _ -> sqrt (Real_of_u r)));;

let rec these (_A1, _A2, _A3)
  a = image ((ceq_option _A1), (ccompare_option _A2)) (_A1, _A2, _A3) the
        (filter ((ceq_option _A1), (ccompare_option _A2))
          (fun x -> not (is_none x)) a);;

let rec key = function (uu, (Fun (f, ts), uv)) -> Some (f, size_list ts)
              | (uw, (Var ux, uy)) -> None;;

let rec gelb _A
  lt c b = (match b with None -> true | Some ba -> lt ba c || eq _A ba c);;

let rec geub _A
  lt c b = (match b with None -> false | Some ba -> lt ba c || eq _A ba c);;

let rec gtlb lt c b = (match b with None -> true | Some ba -> lt ba c);;

let rec leub _A
  lt c b = (match b with None -> true | Some ba -> lt c ba || eq _A c ba);;

let rec ltlb lt c b = (match b with None -> false | Some a -> lt c a);;

let rec ltub lt c b = (match b with None -> true | Some a -> lt c a);;

let rec poly = function LEQ_ns (p, a) -> p
               | GEQ_ns (p, a) -> p;;

let rec comp_fun_idem_apply (Abs_comp_fun_idem x) = x;;

let rec set_fold_cfi (_A1, _A2)
  f b x2 = match f, b, x2 with
    f, b, RBT_set rbt ->
      (match ccompare _A2
        with None ->
          failwith "set_fold_cfi RBT_set: ccompare = None"
            (fun _ -> set_fold_cfi (_A1, _A2) f b (RBT_set rbt))
        | Some _ -> foldb _A2 (comp_fun_idem_apply f) rbt b)
    | f, b, DList_set dxs ->
        (match ceq _A1
          with None ->
            failwith "set_fold_cfi DList_set: ceq = None"
              (fun _ -> set_fold_cfi (_A1, _A2) f b (DList_set dxs))
          | Some _ -> foldc _A1 (comp_fun_idem_apply f) dxs b)
    | f, b, Set_Monad xs -> fold (comp_fun_idem_apply f) xs b
    | f, b, Collect_set p ->
        failwith "set_fold_cfi not supported on Collect_set"
          (fun _ -> set_fold_cfi (_A1, _A2) f b (Collect_set p))
    | f, b, Complement a ->
        failwith "set_fold_cfi not supported on Complement"
          (fun _ -> set_fold_cfi (_A1, _A2) f b (Complement a));;

let rec sup_cfi _A
  = Abs_comp_fun_idem
      (sup _A.semilattice_sup_lattice.quasi_semilattice_sup_semilattice_sup.quasi_order_sup_quasi_semilattice_sup.sup_quasi_order_sup);;

let rec sup_setb (_A1, _A2, _A3, _A4, _A5)
  a = (if finite
            ((finite_UNIV_set _A1),
              (ceq_set (_A2, _A3, _A4.ccompare_cproper_interval)),
              (ccompare_set (_A1, _A3, _A4, _A5)))
            a
        then set_fold_cfi
               ((ceq_set (_A2, _A3, _A4.ccompare_cproper_interval)),
                 (ccompare_set (_A1, _A3, _A4, _A5)))
               (sup_cfi (lattice_set (_A2, _A3, _A4.ccompare_cproper_interval)))
               (bot_set (_A3, _A4.ccompare_cproper_interval, _A5)) a
        else failwith "Sup: infinite"
               (fun _ -> sup_setb (_A1, _A2, _A3, _A4, _A5) a));;

let rec vars_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  = function
    Var x1 ->
      inserta (_B3, _B4.ccompare_cproper_interval) x1
        (bot_set (_B3, _B4.ccompare_cproper_interval, _B7))
    | Fun (x21, x22) ->
        sup_setb (_B1, _B2, _B3, _B4, _B7)
          (image ((ceq_term _A2 _B6), (ccompare_term _A1 _B5))
            ((ceq_set (_B2, _B3, _B4.ccompare_cproper_interval)),
              (ccompare_set (_B1, _B3, _B4, _B7)), set_impl_set)
            (vars_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7))
            (set ((ceq_term _A2 _B6), (ccompare_term _A1 _B5), set_impl_term)
              x22));;

let rec vars_rule (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  r = sup_seta (_B3, _B4.ccompare_cproper_interval)
        (vars_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7) (fst r))
        (vars_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7) (snd r));;

let rec vars_trs (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  r = sup_setb (_B1, _B2, _B3, _B4, _B7)
        (image
          ((ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6)),
            (ccompare_prod (ccompare_term _A1 _B5) (ccompare_term _A1 _B5)))
          ((ceq_set (_B2, _B3, _B4.ccompare_cproper_interval)),
            (ccompare_set (_B1, _B3, _B4, _B7)), set_impl_set)
          (vars_rule (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)) r);;

let rec explode
  s = map char_of_integer
        (let s = s in let rec exp i l = if i < 0 then l else exp (i - 1) (let k = Char.code (Bytes.get s i) in
      if k < 128 then Big_int.big_int_of_int k :: l else failwith "Non-ASCII character in literal") in exp (Bytes.length s - 1) []);;

let rec left x = Sumbot (Inl x);;

let rec xml_error
  str x =
    (let (xmls, (_, (_, (_, pos)))) = x in
     let next =
       (match xmls with [] -> "tag close"
         | XML (tag, _, _) :: _ -> ("<" ^ implode tag) ^ ">"
         | XML_text stra :: _ -> ("text element \034" ^ implode stra) ^ "\034")
       in
      left (Fatal
             ((((("parse error on " ^ next) ^ " at ") ^
                 default_showsl_list showsl_lit pos "") ^
                ":\092") ^
               str)));;

let rec mismatch
  tag x =
    (match x
      with (_, (_, (true, (cands, _)))) -> left (TagMismatch (tag :: cands))
      | (_, (_, (false, (cands, _)))) ->
        xml_error
          ("expecting " ^ default_showsl_list showsl_lit (tag :: cands) "") x);;

let rec xml_do
  tag p x =
    (match x
      with (XML (nam, atts, xmls), (_, (_, (_, pos)))) ->
        (if equal_lista equal_char nam (explode tag)
          then p (xmls, (atts, (false, ([], tag :: pos))))
          else mismatch tag ([fst x], snd x))
      | (XML_text _, _) -> mismatch tag ([fst x], snd x));;

let rec right x = Sumbot (Inr x);;

let rec xml_or
  p1 p2 x =
    (let (x1, (atts, (flag, (cands, rest)))) = x in
      bind2 (p1 (x1, (atts, (true, (cands, rest)))))
        (fun a ->
          (match a
            with TagMismatch cands1 -> p2 (x1, (atts, (flag, (cands1, rest))))
            | Fatal aa -> left (Fatal aa)))
        right);;

let rec filterb xb xc = Alist (filtera xb (impl_ofa xc));;

let rec distinct _A = function [] -> true
                      | x :: xs -> not (membera _A xs x) && distinct _A xs;;

let rec add_mset _A
  x (Bag xs) =
    Bag (joina _A (fun _ (a, b) -> plus_nata a b) (updatea _A x one_nata emptya)
          xs);;

let rec mset _A = function a :: x -> add_mset _A a (mset _A x)
                  | [] -> zero_multiseta;;

let rec add
  (Partial_object_ext
    (carrier, Monoid_ext (mult, one, Ring_ext (zero, add, more))))
    = add;;

let rec lvars t = set (ceq_nat, ccompare_nat, set_impl_nat) (map lhs t);;

let rec fmmap f (Fmap_of_list m) = Fmap_of_list (map (apsnd f) m);;

let fmempty : ('a, 'b) fmap = Fmap_of_list [];;

let rec scale
  r lp = (if equal_rata r zero_rata then fmempty else fmmap (times_rata r) lp);;

let rec scaleRat_linear_poly r p = LinearPoly (scale r (linear_poly_map p));;

let rec uminus_linear_poly lp = scaleRat_linear_poly (uminus_rata one_rata) lp;;

let rec get_var_coeff
  lp v = (match fmlookup equal_nat lp v with None -> zero_rata | Some c -> c);;

let rec sequences _B
  key x1 = match key, x1 with
    key, a :: b :: xs ->
      (if less _B.order_linorder.preorder_order.ord_preorder (key b) (key a)
        then desca _B key b [a] xs else asc _B key b (fun ba -> a :: ba) xs)
    | key, [] -> [[]]
    | key, [v] -> [[v]]
and asc _B
  key a f x3 = match key, a, f, x3 with
    key, a, f, b :: bs ->
      (if not (less _B.order_linorder.preorder_order.ord_preorder (key b)
                (key a))
        then asc _B key b (comp f (fun ba -> a :: ba)) bs
        else f [a] :: sequences _B key (b :: bs))
    | key, a, f, [] -> f [a] :: sequences _B key []
and desca _B
  key a asa x3 = match key, a, asa, x3 with
    key, a, asa, b :: bs ->
      (if less _B.order_linorder.preorder_order.ord_preorder (key b) (key a)
        then desca _B key b (a :: asa) bs
        else (a :: asa) :: sequences _B key (b :: bs))
    | key, a, asa, [] -> (a :: asa) :: sequences _B key [];;

let rec mergec _B
  key x1 bs = match key, x1, bs with
    key, a :: asa, b :: bs ->
      (if less _B.order_linorder.preorder_order.ord_preorder (key b) (key a)
        then b :: mergec _B key (a :: asa) bs
        else a :: mergec _B key asa (b :: bs))
    | key, [], bs -> bs
    | key, v :: va, [] -> v :: va;;

let rec merge_pairs _B
  key x1 = match key, x1 with
    key, a :: b :: xs -> mergec _B key a b :: merge_pairs _B key xs
    | key, [] -> []
    | key, [v] -> [v];;

let rec merge_alla _B
  key x1 = match key, x1 with key, [] -> []
    | key, [x] -> x
    | key, v :: vb :: vc ->
        merge_alla _B key (merge_pairs _B key (v :: vb :: vc));;

let rec sort_key _B key = comp (merge_alla _B key) (sequences _B key);;

let rec sorted_list_of_set (_A1, _A2, _A3, _A4)
  = function
    RBT_set rbt ->
      (match ccompare _A2
        with None ->
          failwith "sorted_list_of_set RBT_set: ccompare = None"
            (fun _ -> sorted_list_of_set (_A1, _A2, _A3, _A4) (RBT_set rbt))
        | Some _ -> sort_key _A4 (fun x -> x) (keysa _A2 rbt))
    | DList_set dxs ->
        (match ceq _A1
          with None ->
            failwith "sorted_list_of_set DList_set: ceq = None"
              (fun _ -> sorted_list_of_set (_A1, _A2, _A3, _A4) (DList_set dxs))
          | Some _ -> sort_key _A4 (fun x -> x) (list_of_dlist _A1 dxs))
    | Set_Monad xs -> sort_key _A4 (fun x -> x) (remdups _A3 xs);;

let rec ordered_keys (_A1, _A2, _A3, _A4, _A5) (_B1, _B2, _B3)
  m = sorted_list_of_set (_A1, _A2, _A3, _A4)
        (fset (fmdom (_A1, _A2, _A5) (_B1, _B2, _B3) m));;

let rec fmfilter
  p (Fmap_of_list m) = Fmap_of_list (filtera (fun (k, _) -> p k) m);;

let rec fmdrop _A a = fmfilter (fun aa -> not (eq _A aa a));;

let rec fmadd _A
  (Fmap_of_list m) (Fmap_of_list n) = Fmap_of_list (merge _A m n);;

let rec fmupd _A k v m = fmadd _A m (Fmap_of_list [(k, v)]);;

let rec set_var_coeff
  v c lp =
    (if equal_rata c zero_rata then fmdrop equal_nat v lp
      else fmupd equal_nat v c lp);;

let rec add_monom
  c v lp = set_var_coeff v (plus_rata c (get_var_coeff lp v)) lp;;

let rec adda
  lp1 lp2 =
    foldl (fun lp v -> add_monom (get_var_coeff lp1 v) v lp) lp2
      (ordered_keys
        (ceq_nat, ccompare_nat, equal_nat, linorder_nat, set_impl_nat)
        (ceq_rat, ccompare_rat, set_impl_rat) lp1);;

let rec plus_linear_poly
  p1 p2 = LinearPoly (adda (linear_poly_map p1) (linear_poly_map p2));;

let rec minus_linear_poly
  lp1 lp2 = plus_linear_poly lp1 (uminus_linear_poly lp2);;

let rec constraint_to_qdelta_constraint
  = function LTa (l, r) -> [LEQ_ns (l, QDelta (r, uminus_rata one_rata))]
    | GTa (l, r) -> [GEQ_ns (l, QDelta (r, one_rata))]
    | LEQ (l, r) -> [LEQ_ns (l, QDelta (r, zero_rata))]
    | GEQ (l, r) -> [GEQ_ns (l, QDelta (r, zero_rata))]
    | EQa (l, r) ->
        [LEQ_ns (l, QDelta (r, zero_rata)); GEQ_ns (l, QDelta (r, zero_rata))]
    | LTPP (l1, l2) ->
        [LEQ_ns
           (minus_linear_poly l1 l2, QDelta (zero_rata, uminus_rata one_rata))]
    | GTPP (l1, l2) ->
        [GEQ_ns (minus_linear_poly l1 l2, QDelta (zero_rata, one_rata))]
    | LEQPP (l1, l2) -> [LEQ_ns (minus_linear_poly l1 l2, zero_QDeltaa)]
    | GEQPP (l1, l2) -> [GEQ_ns (minus_linear_poly l1 l2, zero_QDeltaa)]
    | EQPP (l1, l2) ->
        [LEQ_ns (minus_linear_poly l1 l2, zero_QDeltaa);
          GEQ_ns (minus_linear_poly l1 l2, zero_QDeltaa)];;

let rec i_constraint_to_qdelta_constraint
  (i, c) = map (fun a -> (i, a)) (constraint_to_qdelta_constraint c);;

let rec to_ns l = maps i_constraint_to_qdelta_constraint l;;

let rec funas_term (_A1, _A2, _A3)
  = function
    Var uu ->
      bot_set
        ((ceq_prod _A1 ceq_nat), (ccompare_prod _A2 ccompare_nat),
          (set_impl_prod _A3 set_impl_nat))
    | Fun (f, ts) ->
        sup_seta ((ceq_prod _A1 ceq_nat), (ccompare_prod _A2 ccompare_nat))
          (inserta ((ceq_prod _A1 ceq_nat), (ccompare_prod _A2 ccompare_nat))
            (f, size_list ts)
            (bot_set
              ((ceq_prod _A1 ceq_nat), (ccompare_prod _A2 ccompare_nat),
                (set_impl_prod _A3 set_impl_nat))))
          (foldr
            (comp (sup_seta
                    ((ceq_prod _A1 ceq_nat), (ccompare_prod _A2 ccompare_nat)))
              (funas_term (_A1, _A2, _A3)))
            ts (bot_set
                 ((ceq_prod _A1 ceq_nat), (ccompare_prod _A2 ccompare_nat),
                   (set_impl_prod _A3 set_impl_nat))));;

let rec funas_rule (_A1, _A2, _A3)
  r = sup_seta ((ceq_prod _A1 ceq_nat), (ccompare_prod _A2 ccompare_nat))
        (funas_term (_A1, _A2, _A3) (fst r))
        (funas_term (_A1, _A2, _A3) (snd r));;

let rec funas_trs (_A1, _A2, _A3, _A4, _A5, _A6, _A7) (_B1, _B2)
  r = sup_setb
        ((finite_UNIV_prod _A1 finite_UNIV_nat), (cenum_prod _A2 cenum_nat),
          (ceq_prod _A3 ceq_nat),
          (cproper_interval_prod _A4 cproper_interval_nat),
          (set_impl_prod _A7 set_impl_nat))
        (image
          ((ceq_prod (ceq_term _A6 _B2) (ceq_term _A6 _B2)),
            (ccompare_prod (ccompare_term _A5 _B1) (ccompare_term _A5 _B1)))
          ((ceq_set
             ((cenum_prod _A2 cenum_nat), (ceq_prod _A3 ceq_nat),
               (cproper_interval_prod _A4
                 cproper_interval_nat).ccompare_cproper_interval)),
            (ccompare_set
              ((finite_UNIV_prod _A1 finite_UNIV_nat), (ceq_prod _A3 ceq_nat),
                (cproper_interval_prod _A4 cproper_interval_nat),
                (set_impl_prod _A7 set_impl_nat))),
            set_impl_set)
          (funas_rule (_A3, _A4.ccompare_cproper_interval, _A7)) r);;

let rec unapp _A
  a x1 = match a, x1 with a, Var x -> (Var x, [])
    | a, Fun (f, ss) ->
        (if eq _A f a &&
              equal_nata (size_list ss)
                (nat_of_integer (Big_int.big_int_of_int 2))
          then (let (r, ts) = unapp _A a (nth ss zero_nata) in
                 (r, ts @ [nth ss one_nata]))
          else (Fun (f, ss), []));;

let rec is_letter
  c = less_eq_char (Chara (true, false, false, false, false, true, true, false))
        c &&
        less_eq_char c
          (Chara (false, true, false, true, true, true, true, false)) ||
        (less_eq_char
           (Chara (true, false, false, false, false, false, true, false)) c &&
           less_eq_char c
             (Chara (false, true, false, true, true, false, true, false)) ||
          (less_eq_char
             (Chara (false, false, false, false, true, true, false, false)) c &&
             less_eq_char c
               (Chara (true, false, false, true, true, true, false, false)) ||
            membera equal_char
              [Chara (true, true, true, true, true, false, true, false);
                Chara (false, true, true, false, false, true, false, false);
                Chara (true, true, false, true, true, true, false, false);
                Chara (false, true, false, true, true, true, false, false);
                Chara (true, false, true, true, false, true, false, false)]
              c));;

let rec update_tokens f ts = Inr (ts, f ts);;

let rec remove_comments_aux_acc
  a uu x2 = match a, uu, x2 with
    a, false, c :: cs ->
      (if equal_chara c
            (Chara (false, false, true, true, true, true, false, false)) &&
            equal_lista equal_char
              (take (nat_of_integer (Big_int.big_int_of_int 3)) cs)
              [Chara (true, false, false, false, false, true, false, false);
                Chara (true, false, true, true, false, true, false, false);
                Chara (true, false, true, true, false, true, false, false)]
        then remove_comments_aux_acc a true (tla cs)
        else remove_comments_aux_acc (c :: a) false cs)
    | a, true, c :: cs ->
        (if equal_chara c
              (Chara (true, false, true, true, false, true, false, false)) &&
              equal_lista equal_char
                (take (nat_of_integer (Big_int.big_int_of_int 2)) cs)
                [Chara (true, false, true, true, false, true, false, false);
                  Chara (false, true, true, true, true, true, false, false)]
          then remove_comments_aux_acc a false
                 (drop (nat_of_integer (Big_int.big_int_of_int 2)) cs)
          else remove_comments_aux_acc a true cs)
    | a, uu, [] -> a;;

let rec remove_comments_aux
  uu x1 = match uu, x1 with
    false, c :: cs ->
      (if equal_chara c
            (Chara (false, false, true, true, true, true, false, false)) &&
            equal_lista equal_char
              (take (nat_of_integer (Big_int.big_int_of_int 3)) cs)
              [Chara (true, false, false, false, false, true, false, false);
                Chara (true, false, true, true, false, true, false, false);
                Chara (true, false, true, true, false, true, false, false)]
        then remove_comments_aux true (tla cs)
        else c :: remove_comments_aux false cs)
    | true, c :: cs ->
        (if equal_chara c
              (Chara (true, false, true, true, false, true, false, false)) &&
              equal_lista equal_char
                (take (nat_of_integer (Big_int.big_int_of_int 2)) cs)
                [Chara (true, false, true, true, false, true, false, false);
                  Chara (false, true, true, true, true, true, false, false)]
          then remove_comments_aux false
                 (drop (nat_of_integer (Big_int.big_int_of_int 2)) cs)
          else remove_comments_aux true cs)
    | uu, [] -> [];;

let rec remove_comments
  x = (if false then comp rev (remove_comments_aux_acc [] false)
        else remove_comments_aux false)
        x;;

let rec returna x = (fun ts -> Inr (x, ts));;

let rec bindb m f = (match m with Inl a -> Inl a | Inr a -> f a);;

let rec bindc m f ts = bindb (m ts) (fun (a, b) -> f a b);;

let rec shows_quote
  s = comp (comp (shows_prec_char zero_nata
                   (Chara (true, true, true, false, false, true, false, false)))
             s)
        (shows_prec_char zero_nata
          (Chara (true, true, true, false, false, true, false, false)));;

let rec scan_upto
  enda x1 = match enda, x1 with
    enda, t :: ts ->
      (if equal_lista equal_char (map snd (zip enda (t :: ts))) enda
        then Inr (enda, drop (size_list enda) (t :: ts))
        else bindb (scan_upto enda ts) (fun (res, tsa) -> Inr (t :: res, tsa)))
    | enda, [] ->
        Inl ([Chara (false, false, true, false, false, true, true, false);
               Chara (true, false, false, true, false, true, true, false);
               Chara (false, false, true, false, false, true, true, false);
               Chara (false, false, false, false, false, true, false, false);
               Chara (false, true, true, true, false, true, true, false);
               Chara (true, true, true, true, false, true, true, false);
               Chara (false, false, true, false, true, true, true, false);
               Chara (false, false, false, false, false, true, false, false);
               Chara (false, true, true, false, false, true, true, false);
               Chara (true, false, false, true, false, true, true, false);
               Chara (false, true, true, true, false, true, true, false);
               Chara (false, false, true, false, false, true, true, false);
               Chara (false, false, false, false, false, true, false, false);
               Chara (true, false, true, false, false, true, true, false);
               Chara (false, true, true, true, false, true, true, false);
               Chara (false, false, true, false, false, true, true, false);
               Chara (true, false, true, true, false, true, false, false);
               Chara (true, false, true, true, false, true, true, false);
               Chara (true, false, false, false, false, true, true, false);
               Chara (false, true, false, false, true, true, true, false);
               Chara (true, true, false, true, false, true, true, false);
               Chara (true, false, true, false, false, true, true, false);
               Chara (false, true, false, false, true, true, true, false);
               Chara (false, false, false, false, false, true, false, false)] @
              shows_quote (shows_prec_list show_char zero_nata enda) []);;

let rec trim
  x = dropWhile
        (membera equal_char
          [Chara (false, false, false, false, false, true, false, false);
            Chara (false, true, false, true, false, false, false, false);
            Chara (true, false, false, true, false, false, false, false);
            Chara (true, false, true, true, false, false, false, false)])
        x;;

let rec many
  p x1 = match p, x1 with
    p, t :: ts ->
      (if p t then bindb (many p ts) (fun (rs, tsa) -> Inr (t :: rs, tsa))
        else Inr ([], t :: ts))
    | p, [] -> Inr ([], []);;

let rec parse_header
  ts = (if equal_lista equal_char
             (take (nat_of_integer (Big_int.big_int_of_int 2)) (trim ts))
             [Chara (false, false, true, true, true, true, false, false);
               Chara (true, true, true, true, true, true, false, false)]
         then bindc (scan_upto
                      [Chara (true, true, true, true, true, true, false, false);
                        Chara (false, true, true, true, true, true, false,
                                false)])
                (fun h -> bindc parse_header (fun hs -> returna (h :: hs))) ts
         else bindc (bindc
                      (many (membera equal_char
                              [Chara (false, false, false, false, false, true,
                                       false, false);
                                Chara (false, true, false, true, false, false,
false, false);
                                Chara (true, false, false, true, false, false,
false, false);
                                Chara (true, false, true, true, false, false,
false, false)]))
                      (fun _ -> returna ()))
                (fun _ -> returna []) ts);;

let rec err_expecting _A
  msg ts =
    Inl ([Chara (true, false, true, false, false, true, true, false);
           Chara (false, false, false, true, true, true, true, false);
           Chara (false, false, false, false, true, true, true, false);
           Chara (true, false, true, false, false, true, true, false);
           Chara (true, true, false, false, false, true, true, false);
           Chara (false, false, true, false, true, true, true, false);
           Chara (true, false, false, true, false, true, true, false);
           Chara (false, true, true, true, false, true, true, false);
           Chara (true, true, true, false, false, true, true, false);
           Chara (false, false, false, false, false, true, false, false)] @
          msg @ [Chara (false, false, true, true, false, true, false, false);
                  Chara (false, false, false, false, false, true, false, false);
                  Chara (false, true, false, false, false, true, true, false);
                  Chara (true, false, true, false, true, true, true, false);
                  Chara (false, false, true, false, true, true, true, false);
                  Chara (false, false, false, false, false, true, false, false);
                  Chara (false, true, true, false, false, true, true, false);
                  Chara (true, true, true, true, false, true, true, false);
                  Chara (true, false, true, false, true, true, true, false);
                  Chara (false, true, true, true, false, true, true, false);
                  Chara (false, false, true, false, false, true, true, false);
                  Chara (false, true, false, true, true, true, false, false);
                  Chara (false, false, false, false, false, true, false,
                          false)] @
                  shows_quote
                    (shows_prec_list _A zero_nata
                      (take (nat_of_integer (Big_int.big_int_of_int 30)) ts))
                    []);;

let rec eoi _A
  = function [] -> Inr ((), [])
    | v :: va ->
        err_expecting _A
          [Chara (true, false, true, false, false, true, true, false);
            Chara (false, true, true, true, false, true, true, false);
            Chara (false, false, true, false, false, true, true, false);
            Chara (false, false, false, false, false, true, false, false);
            Chara (true, true, true, true, false, true, true, false);
            Chara (false, true, true, false, false, true, true, false);
            Chara (false, false, false, false, false, true, false, false);
            Chara (true, false, false, true, false, true, true, false);
            Chara (false, true, true, true, false, true, true, false);
            Chara (false, false, false, false, true, true, true, false);
            Chara (true, false, true, false, true, true, true, false);
            Chara (false, false, true, false, true, true, true, false)]
          (v :: va);;

let rec exactly_aux
  s i x2 xs = match s, i, x2, xs with
    s, i, x :: xs, y :: ys ->
      (if equal_chara x y then exactly_aux s i xs ys
        else err_expecting show_char
               ([Chara (false, true, false, false, false, true, false, false)] @
                 s @ [Chara (false, true, false, false, false, true, false,
                              false)])
               i)
    | s, i, [], xs -> Inr (s, trim xs)
    | s, i, x :: xs, [] ->
        err_expecting show_char
          ([Chara (false, true, false, false, false, true, false, false)] @
            s @ [Chara (false, true, false, false, false, true, false, false)])
          i;;

let rec exactly s x = exactly_aux s x s x;;

let rec parse_attribute_value
  x = bindc (exactly
              [Chara (false, true, false, false, false, true, false, false)])
        (fun _ ->
          bindc (many (fun y ->
                        not (equal_chara
                              (Chara
                                (false, true, false, false, false, true, false,
                                  false))
                              y)))
            (fun v ->
              bindc (exactly
                      [Chara (false, true, false, false, false, true, false,
                               false)])
                (fun _ -> returna v)))
        x;;

let rec errora e = (fun _ -> Inl e);;

let rec parse_name
  s = bindc (many is_letter)
        (fun n ->
          bindc (bindc
                  (many (membera equal_char
                          [Chara (false, false, false, false, false, true,
                                   false, false);
                            Chara (false, true, false, true, false, false,
                                    false, false);
                            Chara (true, false, false, true, false, false,
                                    false, false);
                            Chara (true, false, true, true, false, false, false,
                                    false)]))
                  (fun _ -> returna ()))
            (fun _ ->
              (if null n
                then errora
                       ([Chara (true, false, true, false, false, true, true,
                                 false);
                          Chara (false, false, false, true, true, true, true,
                                  false);
                          Chara (false, false, false, false, true, true, true,
                                  false);
                          Chara (true, false, true, false, false, true, true,
                                  false);
                          Chara (true, true, false, false, false, true, true,
                                  false);
                          Chara (false, false, true, false, true, true, true,
                                  false);
                          Chara (true, false, true, false, false, true, true,
                                  false);
                          Chara (false, false, true, false, false, true, true,
                                  false);
                          Chara (false, false, false, false, false, true, false,
                                  false);
                          Chara (false, false, true, true, false, true, true,
                                  false);
                          Chara (true, false, true, false, false, true, true,
                                  false);
                          Chara (false, false, true, false, true, true, true,
                                  false);
                          Chara (false, false, true, false, true, true, true,
                                  false);
                          Chara (true, false, true, false, false, true, true,
                                  false);
                          Chara (false, true, false, false, true, true, true,
                                  false);
                          Chara (false, false, false, false, false, true, false,
                                  false)] @
                         letters @
                           [Chara (false, false, false, false, false, true,
                                    false, false);
                             Chara (false, true, false, false, false, true,
                                     true, false);
                             Chara (true, false, true, false, true, true, true,
                                     false);
                             Chara (false, false, true, false, true, true, true,
                                     false);
                             Chara (false, false, false, false, false, true,
                                     false, false);
                             Chara (false, true, true, false, false, true, true,
                                     false);
                             Chara (true, false, false, true, false, true, true,
                                     false);
                             Chara (false, true, false, false, true, true, true,
                                     false);
                             Chara (true, true, false, false, true, true, true,
                                     false);
                             Chara (false, false, true, false, true, true, true,
                                     false);
                             Chara (false, false, false, false, false, true,
                                     false, false);
                             Chara (true, true, false, false, true, true, true,
                                     false);
                             Chara (true, false, false, true, true, true, true,
                                     false);
                             Chara (true, false, true, true, false, true, true,
                                     false);
                             Chara (false, true, false, false, false, true,
                                     true, false);
                             Chara (true, true, true, true, false, true, true,
                                     false);
                             Chara (false, false, true, true, false, true, true,
                                     false);
                             Chara (false, false, false, false, false, true,
                                     false, false);
                             Chara (true, false, false, true, false, true, true,
                                     false);
                             Chara (true, true, false, false, true, true, true,
                                     false);
                             Chara (false, false, false, false, false, true,
                                     false, false);
                             Chara (false, true, false, false, false, true,
                                     false, false)] @
                             take one_nata s @
                               [Chara (false, true, false, false, false, true,
false, false)])
                else returna n)))
        s;;

let rec parse_attributes
  = function [] -> Inr ([], [])
    | c :: s ->
        (if membera equal_char
              [Chara (true, true, true, true, false, true, false, false);
                Chara (false, true, true, true, true, true, false, false)]
              c
          then Inr ([], c :: s)
          else bindc parse_name
                 (fun k ->
                   bindc (exactly
                           [Chara (true, false, true, true, true, true, false,
                                    false)])
                     (fun _ ->
                       bindc parse_attribute_value
                         (fun v ->
                           bindc parse_attributes
                             (fun atts -> returna ((k, v) :: atts)))))
                 (c :: s));;

let rec oneof_closed
  = function
    x :: xs ->
      (if equal_chara x
            (Chara (false, true, true, true, true, true, false, false))
        then Inr ([Chara (false, true, true, true, true, true, false, false)],
                   trim xs)
        else (if equal_chara x
                   (Chara
                     (true, true, true, true, false, true, false, false)) &&
                   (match xs with [] -> false
                     | y :: _ ->
                       equal_chara y
                         (Chara
                           (false, true, true, true, true, true, false, false)))
               then Inr ([Chara (true, true, true, true, false, true, false,
                                  false);
                           Chara (false, true, true, true, true, true, false,
                                   false)],
                          trim (tla xs))
               else err_expecting show_char
                      [Chara (true, true, true, true, false, true, true, false);
                        Chara (false, true, true, true, false, true, true,
                                false);
                        Chara (true, false, true, false, false, true, true,
                                false);
                        Chara (false, false, false, false, false, true, false,
                                false);
                        Chara (true, true, true, true, false, true, true,
                                false);
                        Chara (false, true, true, false, false, true, true,
                                false);
                        Chara (false, false, false, false, false, true, false,
                                false);
                        Chara (true, true, false, true, true, false, true,
                                false);
                        Chara (true, true, true, true, false, true, false,
                                false);
                        Chara (false, true, true, true, true, true, false,
                                false);
                        Chara (false, false, true, true, false, true, false,
                                false);
                        Chara (false, false, false, false, false, true, false,
                                false);
                        Chara (false, true, true, true, true, true, false,
                                false);
                        Chara (true, false, true, true, true, false, true,
                                false)]
                      (x :: xs)))
    | [] -> err_expecting show_char
              [Chara (true, true, true, true, false, true, true, false);
                Chara (false, true, true, true, false, true, true, false);
                Chara (true, false, true, false, false, true, true, false);
                Chara (false, false, false, false, false, true, false, false);
                Chara (true, true, true, true, false, true, true, false);
                Chara (false, true, true, false, false, true, true, false);
                Chara (false, false, false, false, false, true, false, false);
                Chara (true, true, false, true, true, false, true, false);
                Chara (true, true, true, true, false, true, false, false);
                Chara (false, true, true, true, true, true, false, false);
                Chara (false, false, true, true, false, true, false, false);
                Chara (false, false, false, false, false, true, false, false);
                Chara (false, true, true, true, true, true, false, false);
                Chara (true, false, true, true, true, false, true, false)]
              [];;

let rec parse_text
  x = bindc (many (fun y ->
                    not (equal_chara
                          (Chara
                            (false, false, true, true, true, true, false,
                              false))
                          y)))
        (fun ts ->
          (let text = trim ts in
            (if null text then returna None
              else returna (Some (rev (trim (rev text)))))))
        x;;

let rec parse_nodes
  ts = (if null ts ||
             equal_lista equal_char
               (take (nat_of_integer (Big_int.big_int_of_int 2)) ts)
               [Chara (false, false, true, true, true, true, false, false);
                 Chara (true, true, true, true, false, true, false, false)]
         then returna [] ts
         else (if not (equal_chara (hda ts)
                        (Chara
                          (false, false, true, true, true, true, false, false)))
                then bindc parse_text
                       (fun t ->
                         bindc parse_nodes
                           (fun ns -> returna (XML_text (the t) :: ns)))
                       ts
                else bindc (exactly
                             [Chara (false, false, true, true, true, true,
                                      false, false)])
                       (fun _ ->
                         bindc parse_name
                           (fun n ->
                             bindc parse_attributes
                               (fun atts ->
                                 bindc oneof_closed
                                   (fun e ->
                                     (if equal_lista equal_char e
   [Chara (true, true, true, true, false, true, false, false);
     Chara (false, true, true, true, true, true, false, false)]
                                       then bindc parse_nodes
      (fun cs -> returna (XML (n, atts, []) :: cs))
                                       else bindc parse_nodes
      (fun cs ->
        bindc (exactly
                [Chara (false, false, true, true, true, true, false, false);
                  Chara (true, true, true, true, false, true, false, false)])
          (fun _ ->
            bindc (exactly n)
              (fun _ ->
                bindc (exactly
                        [Chara (false, true, true, true, true, true, false,
                                 false)])
                  (fun _ ->
                    bindc parse_nodes
                      (fun ns -> returna (XML (n, atts, cs) :: ns)))))))))))
                       ts));;

let rec parse_node
  x = bindc (exactly
              [Chara (false, false, true, true, true, true, false, false)])
        (fun _ ->
          bindc parse_name
            (fun n ->
              bindc parse_attributes
                (fun atts ->
                  bindc oneof_closed
                    (fun e ->
                      (if equal_lista equal_char e
                            [Chara (true, true, true, true, false, true, false,
                                     false);
                              Chara (false, true, true, true, true, true, false,
                                      false)]
                        then returna (XML (n, atts, []))
                        else bindc parse_nodes
                               (fun cs ->
                                 bindc (exactly
 [Chara (false, false, true, true, true, true, false, false);
   Chara (true, true, true, true, false, true, false, false)])
                                   (fun _ ->
                                     bindc (exactly n)
                                       (fun _ ->
 bindc (exactly [Chara (false, true, true, true, true, true, false, false)])
   (fun _ -> returna (XML (n, atts, cs)))))))))))
        x;;

let rec parse_doc
  x = bindc (update_tokens remove_comments)
        (fun _ ->
          bindc parse_header
            (fun h ->
              bindc parse_node
                (fun xml ->
                  bindc (eoi show_char) (fun _ -> returna (XMLDOC (h, xml))))))
        x;;

let rec pick_up _A
  rest key x2 = match rest, key, x2 with rest, key, [] -> None
    | rest, key, (l, r) :: s ->
        (if eq _A key l then Some (r, rest @ s)
          else pick_up _A ((l, r) :: rest) key s);;

let rec nat_of_digit
  x = (if equal_chara x
            (Chara (false, false, false, false, true, true, false, false))
        then Some zero_nata
        else (if equal_chara x
                   (Chara (true, false, false, false, true, true, false, false))
               then Some one_nata
               else (if equal_chara x
                          (Chara
                            (false, true, false, false, true, true, false,
                              false))
                      then Some (nat_of_integer (Big_int.big_int_of_int 2))
                      else (if equal_chara x
                                 (Chara
                                   (true, true, false, false, true, true, false,
                                     false))
                             then Some (nat_of_integer
 (Big_int.big_int_of_int 3))
                             else (if equal_chara x
(Chara (false, false, true, false, true, true, false, false))
                                    then Some (nat_of_integer
        (Big_int.big_int_of_int 4))
                                    else (if equal_chara x
       (Chara (true, false, true, false, true, true, false, false))
   then Some (nat_of_integer (Big_int.big_int_of_int 5))
   else (if equal_chara x
              (Chara (false, true, true, false, true, true, false, false))
          then Some (nat_of_integer (Big_int.big_int_of_int 6))
          else (if equal_chara x
                     (Chara (true, true, true, false, true, true, false, false))
                 then Some (nat_of_integer (Big_int.big_int_of_int 7))
                 else (if equal_chara x
                            (Chara
                              (false, false, false, true, true, true, false,
                                false))
                        then Some (nat_of_integer (Big_int.big_int_of_int 8))
                        else (if equal_chara x
                                   (Chara
                                     (true, false, false, true, true, true,
                                       false, false))
                               then Some (nat_of_integer
   (Big_int.big_int_of_int 9))
                               else None))))))))));;

let rec nat_of_string_aux
  n x1 = match n, x1 with n, [] -> Some n
    | n, d :: s ->
        bind (nat_of_digit d)
          (fun m ->
            nat_of_string_aux
              (plus_nata
                (times_nata (nat_of_integer (Big_int.big_int_of_int 10)) n) m)
              s);;

let rec nat_of_string
  s = (match (if null s then None else nat_of_string_aux zero_nata s)
        with None ->
          Inl (("cannot convert \034" ^ implode s) ^ "\034 to a number")
        | Some a -> Inr a);;

let rec safe_head = function [] -> None
                    | x :: xs -> Some x;;

let rec int_of_string
  s = (if equal_optiona equal_char (safe_head s)
            (Some (Chara (true, false, true, true, false, true, false, false)))
        then bindb (nat_of_string (tla s))
               (fun n -> Inr (uminus_inta (int_of_nat n)))
        else bindb (nat_of_string s) (fun n -> Inr (int_of_nat n)));;

let rec xml_take_int
  p xs =
    (match xs with ([], _) -> xml_error "expecting an integer" xs
      | (XML (_, _, _) :: _, _) -> xml_error "expecting an integer" xs
      | (XML_text text :: xmls, s) ->
        (match int_of_string text with Inl x -> xml_error x xs
          | Inr n -> p n (xmls, s)));;

let rec xml_return
  v x = (match x with ([], _) -> right v
          | (_ :: _, _) -> xml_error "expecting tag close" x);;

let rec xml_int tag = xml_do tag (xml_take_int xml_return);;

let rec xml_take_nat
  p xs =
    (match xs with ([], _) -> xml_error "expecting a number" xs
      | (XML (_, _, _) :: _, _) -> xml_error "expecting a number" xs
      | (XML_text text :: xmls, s) ->
        (match nat_of_string text with Inl x -> xml_error x xs
          | Inr n -> p n (xmls, s)));;

let rec xml_nat tag = xml_do tag (xml_take_nat xml_return);;

let zero_enat : enat = Enat zero_nata;;

let rec minus_enat x0 n = match x0, n with Enat a, Infinity_enat -> zero_enat
                     | Infinity_enat, n -> Infinity_enat
                     | Enat a, Enat b -> Enat (minus_nata a b);;

let rec equal_enat x0 x1 = match x0, x1 with Enat nat, Infinity_enat -> false
                     | Infinity_enat, Enat nat -> false
                     | Enat nata, Enat nat -> equal_nata nata nat
                     | Infinity_enat, Infinity_enat -> true;;

let one_enat : enat = Enat one_nata;;

let rec xml_take_many_sub
  acc minOccurs maxOccurs p1 p2 x5 = match acc, minOccurs, maxOccurs, p1, p2, x5
    with
    acc, minOccurs, maxOccurs, p1, p2, ([], (atts, (allow, rest))) ->
      (if equal_nata minOccurs zero_nata
        then p2 (rev acc) ([], (atts, (allow, rest)))
        else bind2 (p1 (XML ([], [], []), (atts, (false, rest)))) left
               (fun _ -> left (Fatal "unexpected")))
    | acc, minOccurs, maxOccurs, p1, p2,
        (xml :: xmls, (atts, (allow, (cands, rest))))
        -> (if equal_enat maxOccurs zero_enat
             then p2 (rev acc) (xml :: xmls, (atts, (allow, (cands, rest))))
             else bind2 (p1 (xml, (atts,
                                    (equal_nata minOccurs zero_nata,
                                      (cands, rest)))))
                    (fun e ->
                      (match e
                        with TagMismatch _ ->
                          p2 (rev acc)
                            (xml :: xmls, (atts, (allow, (cands, rest))))
                        | Fatal _ -> left e))
                    (fun a ->
                      xml_take_many_sub (a :: acc)
                        (minus_nata minOccurs one_nata)
                        (minus_enat maxOccurs one_enat) p1 p2
                        (xmls, (atts, (false, ([], rest))))));;

let rec position
  x = xml_do "position"
        (xml_take_nat (fun n -> xml_return (minus_nata n one_nata))) x;;

let rec xml_change
  p f x = bind2 (p x) left (fun a -> (let (_, rest) = x in f a ([], rest)));;

let rec xml_take
  p1 p2 x =
    (match x
      with ([], rest) ->
        bind2 (p1 (XML ([], [], []), rest)) left
          (fun _ -> left (Fatal "unexpected"))
      | (xa :: xs, rest) ->
        (let (atts, (flag, (cands, resta))) = rest in
          bind2 (p1 (xa, (atts, (flag, (cands, resta))))) left
            (fun a -> p2 a (xs, (atts, (false, ([], resta)))))));;

let rec afs _A
  xml2name =
    xml_do "argumentFilter"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "argumentFilterEntry"
          (xml_take xml2name
            (fun name ->
              xml_take (xml_nat "arity")
                (fun arity ->
                  xml_take
                    (xml_or
                      (xml_change (xml_nat "collapsing")
                        (fun n ->
                          xml_return (Collapse (minus_nata n one_nata))))
                      (xml_do "nonCollapsing"
                        (xml_take_many_sub [] zero_nata Infinity_enat position
                          (fun ls -> xml_return (AFList ls)))))
                    (fun main -> xml_return ((name, arity), main))))))
        xml_return);;

let rec scg_position x = xml_nat "position" x;;

let rec xml_take_default
  a p1 p2 xs =
    (match xs with ([], _) -> p2 a xs
      | (xml :: xmls, (atts, (allow, (cands, rest)))) ->
        bind2 (p1 (xml, (atts, (true, (cands, rest)))))
          (fun e ->
            (match e
              with TagMismatch cands1 ->
                p2 a (xml :: xmls, (atts, (allow, (cands1, rest))))
              | Fatal _ -> left e))
          (fun aa -> p2 aa (xmls, (atts, (false, ([], rest))))));;

let rec xml_take_text
  p xs =
    (match xs with ([], _) -> xml_error "expecting a text" xs
      | (XML (_, _, _) :: _, _) -> xml_error "expecting a text" xs
      | (XML_text text :: xmls, s) -> p text (xmls, s));;

let rec xml_text tag = xml_do tag (xml_take_text xml_return);;

let rec plain_var x = xml_text "var" x;;

let rec var x = xml_change plain_var (comp xml_return (fun a -> Var a)) x;;

let rec term _A
  xml2name x =
    xml_or var
      (xml_do "funapp"
        (xml_take xml2name
          (fun name ->
            xml_take_many_sub [] zero_nata Infinity_enat
              (xml_or (term _A xml2name)
                (xml_do "arg" (xml_take (term _A xml2name) xml_return)))
              (fun args -> xml_return (Fun (name, args))))))
      x;;

let rec conditions _A
  xml2name tag =
    xml_do tag
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "condition"
          (xml_take (xml_do "lhs" (xml_take (term _A xml2name) xml_return))
            (fun l ->
              xml_take (xml_do "rhs" (xml_take (term _A xml2name) xml_return))
                (fun r -> xml_return (l, r)))))
        xml_return);;

let rec crule _A
  xml2name tag =
    xml_do tag
      (xml_take (xml_do "lhs" (xml_take (term _A xml2name) xml_return))
        (fun l ->
          xml_take (xml_do "rhs" (xml_take (term _A xml2name) xml_return))
            (fun r ->
              xml_take_default [] (conditions _A xml2name "conditions")
                (fun conds -> xml_return ((l, r), conds)))));;

let rec rule _A
  xml2name =
    xml_change (crule _A xml2name "rule")
      (fun (lr, conds) ->
        (if null conds then xml_return lr
          else xml_error "conditional rule is not allowed here"));;

let rec bool_of_string
  s = (if equal_lista equal_char s
            [Chara (false, false, true, false, true, true, true, false);
              Chara (false, true, false, false, true, true, true, false);
              Chara (true, false, true, false, true, true, true, false);
              Chara (true, false, true, false, false, true, true, false)]
        then Inr true
        else (if equal_lista equal_char s
                   [Chara (false, true, true, false, false, true, true, false);
                     Chara (true, false, false, false, false, true, true,
                             false);
                     Chara (false, false, true, true, false, true, true, false);
                     Chara (true, true, false, false, true, true, true, false);
                     Chara (true, false, true, false, false, true, true, false)]
               then Inr false
               else Inl (("cannot convert \034" ^ implode s) ^
                          "\034 into Boolean")));;

let rec xml_bool
  tag x =
    bind2 (xml_text tag x) left
      (fun str ->
        (match bool_of_string str with Inl err -> xml_error err ([fst x], snd x)
          | Inr a -> right a));;

let rec scg _A
  xml2name =
    xml_do "sizeChangeGraph"
      (xml_take (rule _A xml2name)
        (fun lr ->
          xml_take_many_sub [] zero_nata Infinity_enat
            (xml_do "edge"
              (xml_take scg_position
                (fun p ->
                  xml_take (xml_bool "strict")
                    (fun s ->
                      xml_take scg_position
                        (fun q -> xml_return (s, (p, q)))))))
            (fun edges ->
              xml_return
                (lr, (map_filter
                        (fun x -> (if fst x then Some (snd x) else None)) edges,
                       map_filter
                         (fun x -> (if not (fst x) then Some (snd x) else None))
                         edges)))));;

let rec gt1 x = map (fun (s, t) -> (root s, root t)) x;;

let rec map_term
  f1 f2 x2 = match f1, f2, x2 with f1, f2, Var x1 -> Var (f2 x1)
    | f1, f2, Fun (x21, x22) -> Fun (f1 x21, map (map_term f1 f2) x22);;

let rec class_to_term_intern
  iv i x2 = match iv, i, x2 with
    iv, i, Fun (f, ts) ->
      (let (k, ss) =
         foldr (fun t (j, ss) ->
                 (let (k, s) = class_to_term_intern iv j t in (k, s :: ss)))
           ts (i, [])
         in
        (k, Fun (f, ss)))
    | iv, i, Var (Inl uu) -> (plus_nata i one_nata, Var (iv i))
    | iv, i, Var (Inr x) -> (i, Var x);;

let rec class_to_term
  c t = snd (class_to_term_intern (fun i -> c :: shows_prec_nat zero_nata i [])
              zero_nata t);;

let rec subst_apply_term
  x0 alpha = match x0, alpha with Var x, alpha -> alpha x
    | Fun (f, ss), alpha ->
        Fun (f, map (fun s -> subst_apply_term s alpha) ss);;

let rec subst_compose sigma tau = (fun x -> subst_apply_term (sigma x) tau);;

let rec subst_of _A
  ss = foldr (fun (x, t) sigma -> subst_compose sigma (subst _A x t)) ss
         (fun a -> Var a);;

let rec contains_var_term _A
  x xa1 = match x, xa1 with x, Var y -> eq _A x y
    | x, Fun (uu, ts) -> list_ex (contains_var_term _A x) ts;;

let rec subst_list
  sigma ys =
    map (fun p ->
          (subst_apply_term (fst p) sigma, subst_apply_term (snd p) sigma))
      ys;;

let rec decompose _A
  s t = (match (s, t) with (Var _, _) -> None | (Fun (_, _), Var _) -> None
          | (Fun (f, ss), Fun (g, ts)) ->
            (if eq _A f g then zip_option ss ts else None));;

let rec unify _A _B
  x0 bs = match x0, bs with [], bs -> Some bs
    | (Fun (f, ss), Fun (g, ts)) :: e, bs ->
        (match decompose _A (Fun (f, ss)) (Fun (g, ts)) with None -> None
          | Some us -> unify _A _B (us @ e) bs)
    | (Var x, t) :: e, bs ->
        (if equal_terma _A _B t (Var x) then unify _A _B e bs
          else (if contains_var_term _B x t then None
                 else unify _A _B (subst_list (subst _B x t) e) ((x, t) :: bs)))
    | (Fun (v, va), Var x) :: e, bs ->
        (if contains_var_term _B x (Fun (v, va)) then None
          else unify _A _B (subst_list (subst _B x (Fun (v, va))) e)
                 ((x, Fun (v, va)) :: bs));;

let rec mgu _A _B
  s t = (match unify _A _B [(s, t)] [] with None -> None
          | Some res -> Some (subst_of _B res));;

let rec mgu_class _A
  cs t =
    mgu _A (equal_list equal_char)
      (class_to_term (Chara (false, true, false, true, true, true, true, false))
        cs)
      (map_term (fun x -> x)
        (fun a ->
          Chara (true, false, false, true, true, true, true, false) :: a)
        t);;

let rec enumerate
  n x1 = match n, x1 with n, x :: xs -> (n, x) :: enumerate (suc n) xs
    | n, [] -> [];;

let rec partition
  p x1 = match p, x1 with p, [] -> ([], [])
    | p, x :: xs ->
        (let (yes, no) = partition p xs in
          (if p x then (x :: yes, no) else (yes, x :: no)));;

let rec removeAll _A
  x xa1 = match x, xa1 with x, [] -> []
    | x, y :: xs ->
        (if eq _A x y then removeAll _A x xs else y :: removeAll _A x xs);;

let rec takeWhile p x1 = match p, x1 with p, [] -> []
                    | p, x :: xs -> (if p x then x :: takeWhile p xs else []);;

let rec match_term_list_code _A (_B1, _B2) _C
  x0 sigma = match x0, sigma with (Fun (f, ss), Var x) :: p, sigma -> None
    | (Fun (f, ss), Fun (g, ts)) :: p, sigma ->
        (match decompose _A (Fun (f, ss)) (Fun (g, ts)) with None -> None
          | Some us -> match_term_list_code _A (_B1, _B2) _C (us @ p) sigma)
    | (Var x, t) :: p, sigma ->
        (if is_none (lookupb (_B1, _B2) sigma x) ||
              equal_optiona (equal_term _A _C) (lookupb (_B1, _B2) sigma x)
                (Some t)
          then match_term_list_code _A (_B1, _B2) _C p
                 (updateb (_B1, _B2) x t sigma)
          else None)
    | [], sigma -> Some sigma;;

let rec subst_of_map
  d sigma x = (match sigma x with None -> d x | Some t -> t);;

let rec match_list (_A1, _A2, _A3) _B _C
  d p = map_option (comp (subst_of_map d) (lookupb (_A1, _A2)))
          (match_term_list_code _B (_A1, _A2) _C p (emptyb (_A1, _A3)));;

let rec matcha _A (_B1, _B2, _B3)
  t l = match_list (_B1, _B2, _B3) _A _B2 (fun a -> Var a) [(l, t)];;

let rec rep_vec_impl (Abs_vec_impl x) = x;;

let rec dim_vec_impl xa = fst (rep_vec_impl xa);;

let rec dim_vec (Vec_impl v) = dim_vec_impl v;;

let rec map_mat
  f a = mat (dim_row a) (dim_col a) (fun ij -> f (index_mat a ij));;

let rec one_mat (_A1, _A2)
  n = mat n n (fun (i, j) -> (if equal_nata i j then onea _A1 else zerob _A2));;

let rec sum_list _A
  xs = foldr (plus _A.semigroup_add_monoid_add.plus_semigroup_add) xs
         (zerob _A.zero_monoid_add);;

let rec vec_index_impl xa = (let (_, a) = rep_vec_impl xa in sub a);;

let rec vec_index (Vec_impl v) i = vec_index_impl v i;;

let rec scalar_prod _A
  v w = sum_list _A.comm_monoid_add_semiring_0.monoid_add_comm_monoid_add
          (map (fun i ->
                 times _A.mult_zero_semiring_0.times_mult_zero (vec_index v i)
                   (vec_index w i))
            (upt zero_nata (dim_vec w)));;

let rec times_mat _A
  a b = mat (dim_row a) (dim_col b)
          (fun (i, j) -> scalar_prod _A (row a i) (col b j));;

let rec pow_mat _A
  a k = (if equal_nata k zero_nata
          then one_mat
                 (_A.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral,
                   _A.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
                 (dim_row a)
          else times_mat _A.semiring_0_semiring_1
                 (pow_mat _A a (minus_nata k one_nata)) a);;

let rec rBT_Impl_fold1
  f x1 = match f, x1 with
    f, Branch (ca, Branch (c, l, ka, va, ra), k, v, r) ->
      folda (fun kb _ -> f kb) r
        (f k (rBT_Impl_fold1 f (Branch (c, l, ka, va, ra))))
    | f, Branch (c, Empty, k, v, r) -> folda (fun ka _ -> f ka) r k
    | f, Empty -> failwith "undefined";;

let rec fold1 _A x xc = rBT_Impl_fold1 x (impl_ofb _A xc);;

let rec imagea (_A1, _A2) (_B1, _B2, _B3)
  x c = match x, c with
    RBT_set rbt, c ->
      (match ccompare _A2
        with None ->
          failwith "Image RBT_set: ccompare1 = None"
            (fun _ -> imagea (_A1, _A2) (_B1, _B2, _B3) (RBT_set rbt) c)
        | Some _ ->
          (match ccompare _B2
            with None ->
              failwith "Image RBT_set: ccompare2 = None"
                (fun _ -> imagea (_A1, _A2) (_B1, _B2, _B3) (RBT_set rbt) c)
            | Some _ ->
              foldb (ccompare_prod _A2 _B2)
                (fun (x, y) acc ->
                  (if member (_A1, _A2) x c then inserta (_B1, _B2) y acc
                    else acc))
                rbt (bot_set (_B1, _B2, _B3))))
    | DList_set dxs, b ->
        (match ceq _A1
          with None ->
            failwith "Image DList_set: ceq1 = None"
              (fun _ -> imagea (_A1, _A2) (_B1, _B2, _B3) (DList_set dxs) b)
          | Some _ ->
            (match ceq _B1
              with None ->
                failwith "Image DList_set: ceq2 = None"
                  (fun _ -> imagea (_A1, _A2) (_B1, _B2, _B3) (DList_set dxs) b)
              | Some _ ->
                foldc (ceq_prod _A1 _B1)
                  (fun (x, y) acc ->
                    (if member (_A1, _A2) x b then inserta (_B1, _B2) y acc
                      else acc))
                  dxs (bot_set (_B1, _B2, _B3))))
    | Set_Monad rxs, a ->
        Set_Monad
          (fold (fun (x, y) rest ->
                  (if member (_A1, _A2) x a then y :: rest else rest))
            rxs [])
    | x, y ->
        image ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2)) (_B1, _B2, _B3) snd
          (filter ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2))
            (fun (xa, _) -> member (_A1, _A2) xa y) x);;

let rec zero
  (Partial_object_ext
    (carrier, Monoid_ext (mult, one, Ring_ext (zero, add, more))))
    = zero;;

let rec b_i_l (State (x1, x2, x3, x4, x5, x6)) = x2;;

let rec indexl
  s = comp (comp fst the) (lookupb (ccompare_nat, equal_nat) (b_i_l s));;

let rec b_i_u (State (x1, x2, x3, x4, x5, x6)) = x3;;

let rec indexu
  s = comp (comp fst the) (lookupb (ccompare_nat, equal_nat) (b_i_u s));;

let rec args = function Var x1 -> []
               | Fun (x21, x22) -> x22;;

let rec aarity sm f n = minus_nata (size_list (sm f n)) one_nata;;

let rec list2position = function [] -> []
                        | n :: ns -> n :: list2position ns;;

let rec pos
  x = xml_do "positionInTerm"
        (xml_take_many_sub [] zero_nata Infinity_enat position
          (fun lst -> xml_return (list2position lst)))
        x;;

let rec xml_leaf tag ret = xml_do tag (xml_return ret);;

let rec map_entry _A
  k f x2 = match k, f, x2 with k, f, [] -> []
    | k, f, p :: ps ->
        (if eq _A (fst p) k then (k, f (snd p)) :: ps
          else p :: map_entry _A k f ps);;

let rec ctxt _A
  xml2name x =
    xml_or (xml_do "box" (xml_return Hole))
      (xml_do "funContext"
        (xml_take xml2name
          (fun name ->
            xml_take
              (xml_do "before"
                (xml_take_many_sub [] zero_nata Infinity_enat (term _A xml2name)
                  xml_return))
              (fun left ->
                xml_take (ctxt _A xml2name)
                  (fun mid ->
                    xml_take
                      (xml_do "after"
                        (xml_take_many_sub [] zero_nata Infinity_enat
                          (term _A xml2name) xml_return))
                      (fun right ->
                        xml_return (More (name, left, mid, right))))))))
      x;;

let rec xml_take_optional
  p1 p2 xs =
    (match xs with ([], _) -> p2 None xs
      | (xml :: xmls, (atts, (allow, (cands, rest)))) ->
        bind2 (p1 (xml, (atts, (true, (cands, rest)))))
          (fun e ->
            (match e
              with TagMismatch cands1 ->
                p2 None (xml :: xmls, (atts, (allow, (cands1, rest))))
              | Fatal _ -> left e))
          (fun a -> p2 (Some a) (xmls, (atts, (false, ([], rest))))));;

let rec relstep _A
  xml2name =
    xml_do "rewriteStep"
      (xml_take pos
        (fun p ->
          xml_take (rule _A xml2name)
            (fun r ->
              xml_take_optional (xml_leaf "relative" ())
                (fun rel ->
                  xml_take (term _A xml2name)
                    (fun t -> xml_return (p, (r, (is_none rel, t))))))));;

let rec relsteps _A
  xml2name =
    xml_do "rewriteSequence"
      (xml_take
        (xml_do "startTerm"
          (xml_take (term _A xml2name) (fun x -> xml_return (id x))))
        (fun start ->
          xml_take_many_sub [] zero_nata Infinity_enat (relstep _A xml2name)
            (fun steps -> xml_return (start, steps))));;

let rec substa _A
  xml2name =
    xml_do "substitution"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "substEntry"
          (xml_take (xml_do "var" (xml_take_text xml_return))
            (fun var ->
              xml_take (term _A xml2name) (fun trm -> xml_return (var, trm)))))
        xml_return);;

let rec loop _A
  xml2name =
    xml_do "loop"
      (xml_take (relsteps _A xml2name)
        (fun (s, rseq) ->
          xml_take (substa _A xml2name)
            (fun sigma ->
              xml_take (ctxt _A xml2name)
                (fun c -> xml_return (s, (rseq, (sigma, c)))))));;

let rec proj _A
  xml2name =
    xml_change (afs _A xml2name)
      (fun afl ->
        xml_return
          (Projection
            (map (fun (fa, a) ->
                   (match a with Collapse aa -> (fa, aa)
                     | AFList _ -> (fa, snd fa)))
              afl)));;

let rec proj_term
  p x1 = match p, x1 with p, Var x -> Var x
    | p, Fun (f, ts) ->
        (let n = size_list ts in
         let i = p (f, n) in
          (if less_nat i n then nth ts i else Fun (f, ts)));;

let rec start_term (_A1, _A2)
  xml2name t =
    xml_do "startTerm"
      (xml_take (term _A2 xml2name)
        (fun s ->
          (if equal_terma _A1 (equal_list equal_char) s t then xml_return t
            else xml_error "<startTerm> does not match lhs")));;

let rec rstep _A
  xml2name =
    xml_do "rewriteStep"
      (xml_take pos
        (fun p ->
          xml_take (rule _A xml2name)
            (fun r ->
              xml_take (term _A xml2name) (fun t -> xml_return (p, (r, t))))));;

let rec rseq (_A1, _A2)
  xml2name pi r =
    xml_do "rewriteSequence"
      (xml_take (start_term (_A1, _A2) xml2name (proj_term pi (fst r)))
        (fun _ ->
          xml_take_many_sub [] zero_nata Infinity_enat (rstep _A2 xml2name)
            (fun rseq -> xml_return (r, rseq))));;

let rec trivial_mute_fun _A
  x y = (x, (y, onea _A.ring_1_comm_ring_1.neg_numeral_ring_1.numeral_neg_numeral.one_numeral));;

let rec fst_sel_fun x = fst (hda x);;

let rec prod_list _A
  xs = foldr (times _A.power_monoid_mult.times_power) xs
         (onea _A.power_monoid_mult.one_power);;

let rec list_update
  x0 i y = match x0, i, y with [], i, y -> []
    | x :: xs, i, y ->
        (if equal_nata i zero_nata then y :: xs
          else x :: list_update xs (minus_nata i one_nata) y);;

let rec mat_swaprows_impl
  xc xd xe =
    Abs_mat_impl
      (let (nr, a) = rep_mat_impl xe in
       let (nc, aa) = a in
        (if less_nat xc nr && less_nat xd nr
          then (let ai = sub aa xc in
                let aj = sub aa xd in
                let arows = list_of aa in
                let ab = IArray (list_update (list_update arows xc aj) xd ai) in
                 (nr, (nc, ab)))
          else (nr, (nc, aa))));;

let rec mat_swaprows
  k l (Mat_impl a) =
    (let nr = dim_row_impl a in
      (if less_nat l nr && less_nat k nr then Mat_impl (mat_swaprows_impl k l a)
        else failwith "index out of bounds in mat_swaprows"
               (fun _ -> mat_swaprows k l (Mat_impl a))));;

let rec find_non0 (_A1, _A2)
  sel_fun l a =
    (let is = upt (suc l) (dim_row a) in
     let ais =
       filtera
         (fun (_, ail) ->
           not (eq _A1 ail
                 (zerob
                   _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)))
         (map (fun i -> (i, index_mat a (i, l))) is)
       in
      (match ais with [] -> None | _ :: _ -> Some (sel_fun ais)));;

let rec mat_multrow_gen_impl
  xc xd xe xf =
    Abs_mat_impl
      (let (nr, a) = rep_mat_impl xf in
       let (nc, aa) = a in
       let ak = sub aa xd in
       let arows = list_of aa in
       let aka = IArray (map (xc xe) (list_of ak)) in
       let ab = IArray (list_update arows xd aka) in
        (nr, (nc, ab)));;

let rec mat_multrow_gen
  mul k aa (Mat_impl a) = Mat_impl (mat_multrow_gen_impl mul k aa a);;

let rec mat_addrow_gen_impl
  xd xe xf xh xi xj =
    Abs_mat_impl
      (let (nr, a) = rep_mat_impl xj in
       let (nc, aa) = a in
        (if less_nat xi nr
          then (let ak = sub aa xh in
                let al = sub aa xi in
                let aka =
                  of_fun (fun i -> xd (xe xf (sub al i)) (sub ak i))
                    (min ord_nat (length ak) (length al))
                  in
                let ab =
                  of_fun (fun i -> (if equal_nata i xh then aka else sub aa i))
                    (length aa)
                  in
                 (nr, (nc, ab)))
          else (nr, (nc, aa))));;

let rec mat_addrow_gen
  ad mul aa k l (Mat_impl a) =
    (if less_nat l (dim_row_impl a)
      then Mat_impl (mat_addrow_gen_impl ad mul aa k l a)
      else failwith "index out of bounds in mat_addrow"
             (fun _ -> mat_addrow_gen ad mul aa k l (Mat_impl a)));;

let rec mute (_A1, _A2)
  mf a_ll k l (r, a) =
    (let p = index_mat a (k, l) in
      (if eq _A1 p
            (zerob
              _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
        then (r, a)
        else (let (q, (pa, _)) = mf a_ll p in
               (times _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd
                  r q,
                 mat_addrow_gen
                   (plus _A2.idom_idom_divide.comm_ring_1_idom.ring_1_comm_ring_1.neg_numeral_ring_1.numeral_neg_numeral.semigroup_add_numeral.plus_semigroup_add)
                   (times
                     _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd)
                   (uminus
                     _A2.idom_idom_divide.comm_ring_1_idom.ring_1_comm_ring_1.neg_numeral_ring_1.group_add_neg_numeral.uminus_group_add
                     pa)
                   k l (mat_multrow_gen
                         (times
                           _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd)
                         k q a)))));;

let rec sub1 (_A1, _A2)
  mf q k l rA =
    (if equal_nata k zero_nata then rA
      else mute (_A1, _A2) mf q (plus_nata l (suc (minus_nata k one_nata))) l
             (sub1 (_A1, _A2) mf q (minus_nata k one_nata) l rA));;

let rec sub2 (_A1, _A2)
  sel_fun mf d l (r, a) =
    (match find_non0 (_A1, _A2) sel_fun l a with None -> (r, a)
      | Some m ->
        (let aa = mat_swaprows m l a in
          sub1 (_A1, _A2) mf (index_mat aa (l, l)) (minus_nata d (suc l)) l
            (uminus
               _A2.idom_idom_divide.comm_ring_1_idom.ring_1_comm_ring_1.neg_numeral_ring_1.group_add_neg_numeral.uminus_group_add
               r,
              aa)));;

let rec sub3 (_A1, _A2)
  sel_fun mf d l rA =
    (if equal_nata l zero_nata then rA
      else sub2 (_A1, _A2) sel_fun mf d (minus_nata l one_nata)
             (sub3 (_A1, _A2) sel_fun mf d (minus_nata l one_nata) rA));;

let rec triangulize (_A1, _A2)
  sel_fun mf a =
    sub3 (_A1, _A2) sel_fun mf (dim_row a) (dim_row a)
      (onea _A2.idom_idom_divide.comm_ring_1_idom.ring_1_comm_ring_1.neg_numeral_ring_1.numeral_neg_numeral.one_numeral,
        a);;

let rec diag_mat
  a = map (fun i -> index_mat a (i, i)) (upt zero_nata (dim_row a));;

let rec det_code (_A1, _A2)
  sel_fun mf a =
    (if equal_nata (dim_row a) (dim_col a)
      then (let (m, aa) = triangulize (_A1, _A2) sel_fun mf a in
             divide _A2.semidom_divide_idom_divide.divide_semidom_divide
               (prod_list
                 _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.monoid_mult_semiring_numeral
                 (diag_mat aa))
               m)
      else zerob _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero);;

let rec det (_A1, _A2)
  a = det_code (_A1, _A2) fst_sel_fun
        (trivial_mute_fun _A2.idom_idom_divide.comm_ring_1_idom) a;;

let rec form_or
  x0 psi = match x0, psi with Disjunction [], psi -> psi
    | Atom v, Disjunction [] -> Atom v
    | NegAtom v, Disjunction [] -> NegAtom v
    | Conjunction v, Disjunction [] -> Conjunction v
    | Disjunction (va :: vb), Disjunction [] -> Disjunction (va :: vb)
    | Disjunction (v :: va), Disjunction (vb :: vc) ->
        Disjunction ((v :: va) @ vb :: vc)
    | Disjunction (v :: va), Atom vb -> Disjunction ((v :: va) @ [Atom vb])
    | Disjunction (v :: va), NegAtom vb ->
        Disjunction ((v :: va) @ [NegAtom vb])
    | Disjunction (v :: va), Conjunction vb ->
        Disjunction ((v :: va) @ [Conjunction vb])
    | Atom v, Disjunction (va :: vb) -> Disjunction (Atom v :: va :: vb)
    | NegAtom v, Disjunction (va :: vb) -> Disjunction (NegAtom v :: va :: vb)
    | Conjunction v, Disjunction (va :: vb) ->
        Disjunction (Conjunction v :: va :: vb)
    | Atom v, Atom va -> Disjunction [Atom v; Atom va]
    | Atom v, NegAtom va -> Disjunction [Atom v; NegAtom va]
    | Atom v, Conjunction va -> Disjunction [Atom v; Conjunction va]
    | NegAtom v, Atom va -> Disjunction [NegAtom v; Atom va]
    | NegAtom v, NegAtom va -> Disjunction [NegAtom v; NegAtom va]
    | NegAtom v, Conjunction va -> Disjunction [NegAtom v; Conjunction va]
    | Conjunction v, Atom va -> Disjunction [Conjunction v; Atom va]
    | Conjunction v, NegAtom va -> Disjunction [Conjunction v; NegAtom va]
    | Conjunction v, Conjunction va ->
        Disjunction [Conjunction v; Conjunction va];;

let rec cnf_form_or
  phi psi = match phi, psi with
    Conjunction phi_s, Conjunction psi_s ->
      Conjunction (maps (fun phi -> map (form_or phi) psi_s) phi_s)
    | Atom v, psi -> form_or (Atom v) psi
    | NegAtom v, psi -> form_or (NegAtom v) psi
    | Disjunction v, psi -> form_or (Disjunction v) psi
    | phi, Atom v -> form_or phi (Atom v)
    | phi, NegAtom v -> form_or phi (NegAtom v)
    | phi, Disjunction v -> form_or phi (Disjunction v);;

let rec form_cnf_ex = function [] -> Conjunction [Disjunction []]
                      | phi :: phi_s -> cnf_form_or phi (form_cnf_ex phi_s);;

let rec form_and
  x0 psi = match x0, psi with Conjunction [], psi -> psi
    | Atom v, Conjunction [] -> Atom v
    | NegAtom v, Conjunction [] -> NegAtom v
    | Conjunction (va :: vb), Conjunction [] -> Conjunction (va :: vb)
    | Disjunction v, Conjunction [] -> Disjunction v
    | Conjunction (v :: va), Conjunction (vb :: vc) ->
        Conjunction ((v :: va) @ vb :: vc)
    | Conjunction (v :: va), Atom vb -> Conjunction ((v :: va) @ [Atom vb])
    | Conjunction (v :: va), NegAtom vb ->
        Conjunction ((v :: va) @ [NegAtom vb])
    | Conjunction (v :: va), Disjunction vb ->
        Conjunction ((v :: va) @ [Disjunction vb])
    | Atom v, Conjunction (va :: vb) -> Conjunction (Atom v :: va :: vb)
    | NegAtom v, Conjunction (va :: vb) -> Conjunction (NegAtom v :: va :: vb)
    | Disjunction v, Conjunction (va :: vb) ->
        Conjunction (Disjunction v :: va :: vb)
    | Atom v, Atom va -> Conjunction [Atom v; Atom va]
    | Atom v, NegAtom va -> Conjunction [Atom v; NegAtom va]
    | Atom v, Disjunction va -> Conjunction [Atom v; Disjunction va]
    | NegAtom v, Atom va -> Conjunction [NegAtom v; Atom va]
    | NegAtom v, NegAtom va -> Conjunction [NegAtom v; NegAtom va]
    | NegAtom v, Disjunction va -> Conjunction [NegAtom v; Disjunction va]
    | Disjunction v, Atom va -> Conjunction [Disjunction v; Atom va]
    | Disjunction v, NegAtom va -> Conjunction [Disjunction v; NegAtom va]
    | Disjunction v, Disjunction va ->
        Conjunction [Disjunction v; Disjunction va];;

let rec form_all = function [] -> Conjunction []
                   | phi :: phi_s -> form_and phi (form_all phi_s);;

let rec flatten = function Conjunction phi_s -> form_all (map flatten phi_s)
                  | Disjunction phi_s -> form_cnf_ex (map flatten phi_s)
                  | Atom v -> Conjunction [Disjunction [Atom v]]
                  | NegAtom v -> Conjunction [Disjunction [NegAtom v]];;

let rec is_Atom = function Atom uu -> true
                  | NegAtom v -> false
                  | Conjunction v -> false
                  | Disjunction v -> false;;

let rec gen_set emp ins l = fold ins l emp;;

let rec initial
  (Lts_ext (initial, transition_rules, assertion, more)) = initial;;

let rec untrans_var = function Pre x -> x
                      | Post x -> x
                      | Intermediate x -> x;;

let rec label f n x2 = match f, n, x2 with f, n, Inl l -> Lab (f, l)
                | f, n, Inr l -> FunLab (f, l);;

let rec unlab = function Lab (f, l) -> f
                | FunLab (f, l) -> f
                | UnLab v -> UnLab v
                | Sharp v -> Sharp v;;

let rec plus_mat _A
  a b = mat (dim_row b) (dim_col b)
          (fun ij -> plus _A (index_mat a ij) (index_mat b ij));;

let rec carrier_mat
  nr nc =
    Collect_set
      (fun a -> equal_nata (dim_row a) nr && equal_nata (dim_col a) nc);;

let rec zero_mat _A nr nc = mat nr nc (fun _ -> zerob _A);;

let rec ring_mat _A
  ty n b =
    Partial_object_ext
      (carrier_mat n n,
        Monoid_ext
          (times_mat _A.semiring_0_semiring_1,
            one_mat
              (_A.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral,
                _A.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
              n,
            Ring_ext
              (zero_mat
                 _A.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero n
                 n,
                plus_mat
                  _A.semiring_numeral_semiring_1.numeral_semiring_numeral.semigroup_add_numeral.plus_semigroup_add,
                b)));;

let rec emptyd (_B1, _B2) xa = Abs_multimap (xa, (emptyb (_B1, _B2), []));;

let rec rm_iterateoi
  x0 c f sigma = match x0, c, f, sigma with Empty, c, f, sigma -> sigma
    | Branch (col, l, k, v, r), c, f, sigma ->
        (if c sigma
          then (let sigmaa = rm_iterateoi l c f sigma in
                 (if c sigmaa then rm_iterateoi r c f (f (k, v) sigmaa)
                   else sigmaa))
          else sigma);;

let rec iteratei_bmap_op_list_it_rm_basic_ops _A
  s = rm_iterateoi (impl_of _A s);;

let rec g_to_list_rm_basic_ops _A
  m = iteratei_bmap_op_list_it_rm_basic_ops _A m (fun _ -> true)
        (fun a b -> a :: b) [];;

let rec values _A m = maps snd (g_to_list_rm_basic_ops _A m);;

let rec polya _A
  p = fold_coeffs
        _A.semiring_0_comm_semiring_0.mult_zero_semiring_0.zero_mult_zero
        (fun a f x ->
          plus _A.semiring_0_comm_semiring_0.comm_monoid_add_semiring_0.monoid_add_comm_monoid_add.semigroup_add_monoid_add.plus_semigroup_add
            a (times
                _A.semiring_0_comm_semiring_0.mult_zero_semiring_0.times_mult_zero
                x (f x)))
        p (fun _ ->
            zerob _A.semiring_0_comm_semiring_0.mult_zero_semiring_0.zero_mult_zero);;

let rec boundsl
  s = comp (map_option snd) (lookupb (ccompare_nat, equal_nat) (b_i_l s));;

let rec boundsu
  s = comp (map_option snd) (lookupb (ccompare_nat, equal_nat) (b_i_u s));;

let rec vars_list
  lp = ordered_keys
         (ceq_nat, ccompare_nat, equal_nat, linorder_nat, set_impl_nat)
         (ceq_rat, ccompare_rat, set_impl_rat) (linear_poly_map lp);;

let rec valuate _A
  lp vala =
    (let lpm = linear_poly_map lp in
      sum_list
        _A.ab_group_add_rational_vector.group_add_ab_group_add.monoid_add_group_add
        (map (fun x ->
               scaleRat _A.scaleRat_rational_vector
                 (the (fmlookup equal_nat lpm x)) (vala x))
          (vars_list lp)));;

let rec delta_0
  qd1 qd2 =
    (let c1 = qdfst qd1 in
     let c2 = qdfst qd2 in
     let k1 = qdsnd qd1 in
     let k2 = qdsnd qd2 in
      (if less_rat c1 c2 && less_rat k2 k1
        then divide_rata (minus_rata c2 c1) (minus_rata k1 k2) else one_rata));;

let rec delta_0_val
  x0 vl = match x0, vl with
    LEQ_ns (lll, rrr), vl -> delta_0 (valuate rational_vector_QDelta lll vl) rrr
    | GEQ_ns (lll, rrr), vl ->
        delta_0 rrr (valuate rational_vector_QDelta lll vl);;

let rec delta_0_val_min
  x0 vl = match x0, vl with [], vl -> one_rata
    | h :: t, vl -> min ord_rat (delta_0_val_min t vl) (delta_0_val h vl);;

let rec tabulatea (_A1, _A2, _A3)
  xs f = fold (fun k -> updateb (_A1, _A2) k (f k)) xs (emptyb (_A1, _A3));;

let rec map2fun _A
  v = (fun x ->
        (match lookupb (ccompare_nat, equal_nat) v x with None -> zerob _A
          | Some y -> y));;

let rec from_ns
  vl cs =
    (let delta = delta_0_val_min cs (map2fun zero_QDelta vl) in
      tabulatea (ccompare_nat, equal_nat, mapping_impl_nat)
        (remdups equal_nat (maps vars_list (map poly cs)))
        (fun var -> vala (map2fun zero_QDelta vl var) delta));;

let rec uBI_upd _B (Direction (x1, x2, x3, x4, x5, x6, x7, x8, x9, x10)) = x8;;

let rec update_B_I (_A1, _A2)
  field_update i x c s = field_update (updateb (_A1, _A2) x (i, c)) s;;

let rec lt _B (Direction (x1, x2, x3, x4, x5, x6, x7, x8, x9, x10)) = x1;;

let rec ub _B (Direction (x1, x2, x3, x4, x5, x6, x7, x8, x9, x10)) = x5;;

let rec li _B (Direction (x1, x2, x3, x4, x5, x6, x7, x8, x9, x10)) = x6;;

let rec lb _B (Direction (x1, x2, x3, x4, x5, x6, x7, x8, x9, x10)) = x4;;

let rec v_update
  v (State (t, bil, biu, v_old, u, uc)) = State (t, bil, biu, v, u, uc);;

let rec v (State (x1, x2, x3, x4, x5, x6)) = x4;;

let rec t (State (x1, x2, x3, x4, x5, x6)) = x1;;

let rec coeffa lp = get_var_coeff (linear_poly_map lp);;

let rec rhs_eq_val (_A1, _A2, _A3, _A4)
  v x_i c e =
    (let x_j = lhs e in
     let a_i_j = coeffa (rhs e) x_i in
      plus _A2 (map2fun _A3 v x_j)
        (scaleRat _A4 a_i_j (minus _A1 c (map2fun _A3 v x_i))));;

let rec update_code _A
  x c s =
    v_update
      (updateb (ccompare_nat, equal_nat) x c
        (foldl
          (fun va e ->
            updateb (ccompare_nat, equal_nat) (lhs e)
              (rhs_eq_val
                (_A.linordered_rational_vector_lrv.ordered_rational_vector_linordered_rational_vector.rational_vector_ordered_rational_vector.ab_group_add_rational_vector.group_add_ab_group_add.minus_group_add,
                  _A.linordered_rational_vector_lrv.ordered_rational_vector_linordered_rational_vector.rational_vector_ordered_rational_vector.ab_group_add_rational_vector.group_add_ab_group_add.monoid_add_group_add.semigroup_add_monoid_add.plus_semigroup_add,
                  _A.linordered_rational_vector_lrv.ordered_rational_vector_linordered_rational_vector.rational_vector_ordered_rational_vector.ab_group_add_rational_vector.group_add_ab_group_add.monoid_add_group_add.zero_monoid_add,
                  _A.linordered_rational_vector_lrv.ordered_rational_vector_linordered_rational_vector.rational_vector_ordered_rational_vector.scaleRat_rational_vector)
                (v s) x c e)
              va)
          (v s) (t s)))
      s;;

let rec set_unsat _A
  i (State (t, bil, biu, v, u, uc)) =
    State (t, bil, biu, v, true, Some (remdups _A i));;

let rec assert_bound_codea _A (_B1, _B2)
  dir i x c s =
    (if geub _B1
          (lt _B2.linordered_rational_vector_lrv.linorder_linordered_rational_vector
            dir)
          c (ub _B2.linordered_rational_vector_lrv.linorder_linordered_rational_vector
              dir s x)
      then s
      else (let sa =
              update_B_I (ccompare_nat, equal_nat)
                (uBI_upd
                  _B2.linordered_rational_vector_lrv.linorder_linordered_rational_vector
                  dir)
                i x c s
              in
             (if ltlb (lt _B2.linordered_rational_vector_lrv.linorder_linordered_rational_vector
                        dir)
                   c (lb _B2.linordered_rational_vector_lrv.linorder_linordered_rational_vector
                       dir s x)
               then set_unsat _A
                      [i; li _B2.linordered_rational_vector_lrv.linorder_linordered_rational_vector
                            dir s x]
                      sa
               else (if not (member (ceq_nat, ccompare_nat) x (lvars (t sa))) &&
                          lt _B2.linordered_rational_vector_lrv.linorder_linordered_rational_vector
                            dir c
                            (map2fun
                              _B2.linordered_rational_vector_lrv.ordered_rational_vector_linordered_rational_vector.rational_vector_ordered_rational_vector.ab_group_add_rational_vector.group_add_ab_group_add.monoid_add_group_add.zero_monoid_add
                              (v s) x)
                      then update_code _B2 x c sa else sa))));;

let rec b_i_u_update
  up (State (t, bil, biu, v, u, uc)) = State (t, bil, up biu, v, u, uc);;

let rec positive _B
  = Direction
      (less _B.order_linorder.preorder_order.ord_preorder, b_i_l, b_i_u,
        boundsl, boundsu, indexl, indexu, b_i_u_update, (fun a b -> Leq (a, b)),
        (fun a b -> Geq (a, b)));;

let rec b_i_l_update
  up (State (t, bil, biu, v, u, uc)) = State (t, up bil, biu, v, u, uc);;

let rec negative _B
  = Direction
      ((fun x y -> less _B.order_linorder.preorder_order.ord_preorder y x),
        b_i_u, b_i_l, boundsu, boundsl, indexu, indexl, b_i_l_update,
        (fun a b -> Geq (a, b)), (fun a b -> Leq (a, b)));;

let rec assert_bound_code _A (_B1, _B2)
  x0 s = match x0, s with
    (i, Leq (x, c)), s ->
      assert_bound_codea _A (_B1, _B2)
        (positive
          _B2.linordered_rational_vector_lrv.linorder_linordered_rational_vector)
        i x c s
    | (i, Geq (x, c)), s ->
        assert_bound_codea _A (_B1, _B2)
          (negative
            _B2.linordered_rational_vector_lrv.linorder_linordered_rational_vector)
          i x c s;;

let rec u (State (x1, x2, x3, x4, x5, x6)) = x5;;

let rec assert_bound_loop_code _A (_B1, _B2)
  ats s =
    foldl (fun sa a ->
            (if u sa then sa else assert_bound_code _A (_B1, _B2) a sa))
      s ats;;

let rec init_state _B
  t = State (t, mapping_empty ccompare_nat (of_phantom mapping_impl_nata),
              mapping_empty ccompare_nat (of_phantom mapping_impl_nata),
              tabulatea (ccompare_nat, equal_nat, mapping_impl_nat)
                (remdups equal_nat (map lhs t @ maps (comp vars_list rhs) t))
                (fun _ -> zerob _B),
              false, None);;

let rec min_satisfying _A
  p l = (let xs = filtera p l in
          (if null xs then None
            else Some (foldl (min _A.order_linorder.preorder_order.ord_preorder)
                        (hda xs) (tla xs))));;

let rec le_ubound (_A1, _A2)
  c b = leub _A1 (less _A2.order_linorder.preorder_order.ord_preorder) c b;;

let rec ge_lbound (_A1, _A2)
  c b = gelb _A1 (less _A2.order_linorder.preorder_order.ord_preorder) c b;;

let rec in_bounds (_B1, _B2)
  x v (lb, ub) =
    ge_lbound (_B1, _B2) (v x) (lb x) && le_ubound (_B1, _B2) (v x) (ub x);;

let rec min_lvar_not_in_bounds (_B1, _B2, _B3)
  s = min_satisfying linorder_nat
        (fun x ->
          not (in_bounds (_B2, _B3) x (map2fun _B1 (v s))
                (boundsl s, boundsu s)))
        (map lhs (t s));;

let rec vara x = LinearPoly (set_var_coeff x one_rata fmempty);;

let rec subst_var
  v lpa lp =
    minus_linear_poly
      (plus_linear_poly lp (scaleRat_linear_poly (coeffa lp v) lpa))
      (scaleRat_linear_poly (coeffa lp v) (vara v));;

let rec subst_var_eq_code v lp eq = (lhs eq, subst_var v lp (rhs eq));;

let rec eq_idx_for_lvar_aux
  xa0 x i = match xa0, x, i with [], x, i -> i
    | eq :: t, x, i ->
        (if equal_nata (lhs eq) x then i
          else eq_idx_for_lvar_aux t x (plus_nata i one_nata));;

let rec eq_idx_for_lvar t x = eq_idx_for_lvar_aux t x zero_nata;;

let rec eq_for_lvar_code t v = nth t (eq_idx_for_lvar t v);;

let rec pivot_eq
  e y = (let cy = coeffa (rhs e) y in
          (y, plus_linear_poly
                (scaleRat_linear_poly (divide_rata (uminus_rata one_rata) cy)
                  (minus_linear_poly (rhs e)
                    (scaleRat_linear_poly cy (vara y))))
                (scaleRat_linear_poly (divide_rata one_rata cy)
                  (vara (lhs e)))));;

let rec pivot_tableau_code
  x_i x_j t =
    (let eq = eq_for_lvar_code t x_i in
     let eqa = pivot_eq eq x_j in
      map (fun e ->
            (if equal_nata (lhs e) (lhs eq) then eqa
              else subst_var_eq_code x_j (rhs eqa) e))
        t);;

let rec t_update
  t (State (t_old, bil, biu, v, u, uc)) = State (t, bil, biu, v, u, uc);;

let rec pivot_code _B
  x_i x_j s = t_update (pivot_tableau_code x_i x_j (t s)) s;;

let rec pivot_and_update_code _A
  x_i x_j c s = update_code _A x_i c (pivot_code _A x_i x_j s);;

let rec ui _B (Direction (x1, x2, x3, x4, x5, x6, x7, x8, x9, x10)) = x7;;

let rec unsat_indices _A _B
  dir s vs eq =
    (let r = rhs eq in
     let lia = li _B dir s in
     let uia = ui _B dir s in
      remdups _A
        (lia (lhs eq) ::
          map (fun x ->
                (if less_rat (coeffa r x) zero_rata then lia x else uia x))
            vs));;

let rec min_rvar_incdec_eq _A _B
  dir s eq =
    (let rvars = vars_list (rhs eq) in
      (match
        min_satisfying linorder_nat
          (fun x ->
            less_rat zero_rata (coeffa (rhs eq) x) &&
              ltub (lt _B.linordered_rational_vector_lrv.linorder_linordered_rational_vector
                     dir)
                (map2fun
                  _B.linordered_rational_vector_lrv.ordered_rational_vector_linordered_rational_vector.rational_vector_ordered_rational_vector.ab_group_add_rational_vector.group_add_ab_group_add.monoid_add_group_add.zero_monoid_add
                  (v s) x)
                (ub _B.linordered_rational_vector_lrv.linorder_linordered_rational_vector
                  dir s x) ||
              less_rat (coeffa (rhs eq) x) zero_rata &&
                gtlb (lt _B.linordered_rational_vector_lrv.linorder_linordered_rational_vector
                       dir)
                  (map2fun
                    _B.linordered_rational_vector_lrv.ordered_rational_vector_linordered_rational_vector.rational_vector_ordered_rational_vector.ab_group_add_rational_vector.group_add_ab_group_add.monoid_add_group_add.zero_monoid_add
                    (v s) x)
                  (lb _B.linordered_rational_vector_lrv.linorder_linordered_rational_vector
                    dir s x))
          rvars
        with None ->
          Inl (unsat_indices _A
                _B.linordered_rational_vector_lrv.linorder_linordered_rational_vector
                dir s rvars eq)
        | Some a -> Inr a));;

let rec check_codea _A _B
  dir x_i s =
    (let l_i =
       the (lb _B.linordered_rational_vector_lrv.linorder_linordered_rational_vector
             dir s x_i)
       in
      (match min_rvar_incdec_eq _A _B dir s (eq_for_lvar_code (t s) x_i)
        with Inl i -> set_unsat _A i s
        | Inr x_j -> pivot_and_update_code _B x_i x_j l_i s));;

let rec lt_lbound _A
  c b = ltlb (less _A.order_linorder.preorder_order.ord_preorder) c b;;

let rec check_code _A (_B1, _B2)
  s = (if u s then s
        else (match
               min_lvar_not_in_bounds
                 (_B2.linordered_rational_vector_lrv.ordered_rational_vector_linordered_rational_vector.rational_vector_ordered_rational_vector.ab_group_add_rational_vector.group_add_ab_group_add.monoid_add_group_add.zero_monoid_add,
                   _B1,
                   _B2.linordered_rational_vector_lrv.linorder_linordered_rational_vector)
                 s
               with None -> s
               | Some x_i ->
                 (let dir =
                    (if lt_lbound
                          _B2.linordered_rational_vector_lrv.linorder_linordered_rational_vector
                          (map2fun
                            _B2.linordered_rational_vector_lrv.ordered_rational_vector_linordered_rational_vector.rational_vector_ordered_rational_vector.ab_group_add_rational_vector.group_add_ab_group_add.monoid_add_group_add.zero_monoid_add
                            (v s) x_i)
                          (boundsl s x_i)
                      then positive
                             _B2.linordered_rational_vector_lrv.linorder_linordered_rational_vector
                      else negative
                             _B2.linordered_rational_vector_lrv.linorder_linordered_rational_vector)
                    in
                   check_code _A (_B1, _B2) (check_codea _A _B2 dir x_i s))));;

let rec assert_all_state_code _A (_B1, _B2)
  t ats =
    check_code _A (_B1, _B2)
      (assert_bound_loop_code _A (_B1, _B2) ats
        (init_state
          _B2.linordered_rational_vector_lrv.ordered_rational_vector_linordered_rational_vector.rational_vector_ordered_rational_vector.ab_group_add_rational_vector.group_add_ab_group_add.monoid_add_group_add.zero_monoid_add
          t));;

let rec u_c (State (x1, x2, x3, x4, x5, x6)) = x6;;

let rec assert_all_code _A (_B1, _B2)
  t asa =
    (let s = assert_all_state_code _A (_B1, _B2) t asa in
      (if u s then Inl (the (u_c s)) else Inr (v s)));;

let rec pivot_tableau_eq
  t1 eq t2 x =
    (let eqa = pivot_eq eq x in
     let m = map (subst_var_eq_code x (rhs eqa)) in
      (m t1, (eqa, m t2)));;

let rec preprocess_opt _A
  x t1 xa2 = match x, t1, xa2 with x, t1, [] -> (t1, id)
    | xa, t1, (x, p) :: t2 ->
        (if not (member (ceq_nat, ccompare_nat) x xa)
          then (let (t, tv) = preprocess_opt _A xa t1 t2 in
                 (t, comp (fun v ->
                            updateb (ccompare_nat, equal_nat) x
                              (valuate
                                _A.linordered_rational_vector_lrv.ordered_rational_vector_linordered_rational_vector.rational_vector_ordered_rational_vector
                                p (map2fun
                                    _A.linordered_rational_vector_lrv.ordered_rational_vector_linordered_rational_vector.rational_vector_ordered_rational_vector.ab_group_add_rational_vector.group_add_ab_group_add.monoid_add_group_add.zero_monoid_add
                                    v))
                              v)
                       tv))
          else (match
                 find (fun xb -> not (member (ceq_nat, ccompare_nat) xb xa))
                   (vars_list p)
                 with None -> preprocess_opt _A xa ((x, p) :: t1) t2
                 | Some y ->
                   (let (tt1, ((z, q), tt2)) = pivot_tableau_eq t1 (x, p) t2 y
                      in
                    let (t, tv) = preprocess_opt _A xa tt1 tt2 in
                     (t, comp (fun v ->
                                updateb (ccompare_nat, equal_nat) z
                                  (valuate
                                    _A.linordered_rational_vector_lrv.ordered_rational_vector_linordered_rational_vector.rational_vector_ordered_rational_vector
                                    q (map2fun
_A.linordered_rational_vector_lrv.ordered_rational_vector_linordered_rational_vector.rational_vector_ordered_rational_vector.ab_group_add_rational_vector.group_add_ab_group_add.monoid_add_group_add.zero_monoid_add
v))
                                  v)
                           tv))));;

let rec atom_var = function Leq (var, a) -> var
                   | Geq (var, a) -> var;;

let rec preprocess_part_2 (_A1, _A2, _A3) (_B1, _B2) _C
  asa t =
    preprocess_opt _C
      (image ((ceq_atom _B2), (ccompare_atom _B1))
        (ceq_nat, ccompare_nat, set_impl_nat) atom_var
        (image
          ((ceq_prod _A1 (ceq_atom _B2)),
            (ccompare_prod _A2 (ccompare_atom _B1)))
          ((ceq_atom _B2), (ccompare_atom _B1), set_impl_atom) snd
          (set ((ceq_prod _A1 (ceq_atom _B2)),
                 (ccompare_prod _A2 (ccompare_atom _B1)),
                 (set_impl_prod _A3 set_impl_atom))
            asa)))
      [] t;;

let rec max_var
  lp = (let vl = vars_list lp in
         (if null vl then zero_nata
           else foldl (max ord_nat) (hda vl) (tla vl)));;

let rec start_fresh_variable
  = function [] -> zero_nata
    | (i, h) :: t ->
        max ord_nat (plus_nata (max_var (poly h)) one_nata)
          (start_fresh_variable t);;

let rec tableau (IState (x1, x2, x3, x4)) = x2;;

let rec atoms (IState (x1, x2, x3, x4)) = x3;;

let rec is_monom l = equal_nata (size_list (vars_list l)) one_nata;;

let rec monom_var l = max_var l;;

let rec monom_coeff l = coeffa l (monom_var l);;

let rec monom_to_atom
  = function
    LEQ_ns (l, r) ->
      (if less_rat (monom_coeff l) zero_rata
        then Geq (monom_var l,
                   scaleRat_QDeltaa (inverse_rata (monom_coeff l)) r)
        else Leq (monom_var l,
                   scaleRat_QDeltaa (inverse_rata (monom_coeff l)) r))
    | GEQ_ns (l, r) ->
        (if less_rat (monom_coeff l) zero_rata
          then Leq (monom_var l,
                     scaleRat_QDeltaa (inverse_rata (monom_coeff l)) r)
          else Geq (monom_var l,
                     scaleRat_QDeltaa (inverse_rata (monom_coeff l)) r));;

let rec qdelta_constraint_to_atom
  x0 v = match x0, v with
    LEQ_ns (l, r), v ->
      (if is_monom l then monom_to_atom (LEQ_ns (l, r)) else Leq (v, r))
    | GEQ_ns (l, r), v ->
        (if is_monom l then monom_to_atom (GEQ_ns (l, r)) else Geq (v, r));;

let rec firstFreshVariable (IState (x1, x2, x3, x4)) = x1;;

let rec poly_Mapping (IState (x1, x2, x3, x4)) = x4;;

let rec linear_poly_to_eq p v = (v, p);;

let rec preprocessa
  x0 v = match x0, v with [], v -> IState (v, [], [], (fun _ -> None))
    | (i, h) :: t, v ->
        (let s = preprocessa t v in
         let p = poly h in
         let is_monom_h = is_monom p in
         let va = firstFreshVariable s in
         let ta = tableau s in
         let a = atoms s in
         let m = poly_Mapping s in
          (if is_monom_h
            then IState (va, ta, (i, qdelta_constraint_to_atom h va) :: a, m)
            else (match m p
                   with None ->
                     IState
                       (plus_nata va one_nata, linear_poly_to_eq p va :: ta,
                         (i, qdelta_constraint_to_atom h va) :: a,
                         fun_upd equal_linear_poly m p (Some va))
                   | Some vaa ->
                     IState
                       (va, ta, (i, qdelta_constraint_to_atom h vaa) :: a,
                         m))));;

let rec preprocess_part_1
  l = (let start = start_fresh_variable l in
       let is = preprocessa l start in
        (tableau is, atoms is));;

let rec preprocess (_A1, _A2, _A3)
  l = (let (t, asa) = preprocess_part_1 l in
       let (ta, tv) =
         preprocess_part_2 (_A1, _A2, _A3) (ccompare_QDelta, equal_QDelta)
           lrv_QDelta asa t
         in
        (ta, (asa, tv)));;

let rec solve_exec_ns_code (_A1, _A2, _A3, _A4)
  s = (let (t, (asa, trans_v)) = preprocess (_A1, _A2, _A4) s in
        (match assert_all_code _A3 (equal_QDelta, lrv_QDelta) t asa
          with Inl a -> Inl a | Inr v -> Inr (trans_v v)));;

let rec solve_exec_code (_A1, _A2, _A3, _A4)
  cs = (let csa = to_ns cs in
         (match solve_exec_ns_code (_A1, _A2, _A3, _A4) csa with Inl a -> Inl a
           | Inr v -> Inr (from_ns v (map snd csa))));;

let rec simplex_index (_A1, _A2, _A3, _A4)
  = solve_exec_code (_A1, _A2, _A3, _A4);;

let rec simplex
  cs = simplex_index (ceq_nat, ccompare_nat, equal_nat, set_impl_nat)
         (zip (upt zero_nata (size_list cs)) cs);;

let rec binda
  (Sumbot a) f = (match a with Inl b -> Sumbot (Inl b) | Inr aa -> f aa);;

let rec tcapI _A _B
  uu x1 = match uu, x1 with uu, Var uv -> GCHole
    | r, Fun (f, ts) ->
        (let h = GCFun (f, map (tcapI _A _B r) ts) in
          (if list_ex (fun ra -> matchc _A _B h (fst ra)) r then GCHole
            else h));;

let rec ground = function Var x -> false
                 | Fun (f, ts) -> list_all ground ts;;

let rec hvf_top _A
  a n x2 = match a, n, x2 with
    a, n, Fun (f, ts) ->
      (if eq _A f a && equal_nata (size_list ts) n then not (is_Var (hda ts))
        else true)
    | a, n, Var uu -> false;;

let rec status (Abs_status x) = x;;

let rec af_wpo
  pi sigma f =
    sup_seta (ceq_nat, ccompare_nat)
      (set (ceq_nat, ccompare_nat, set_impl_nat) (status sigma f)) (pi f);;

let rec lex_ext_unbounded
  f x1 x2 = match f, x1, x2 with f, [], [] -> (false, true)
    | f, uu :: uv, [] -> (true, true)
    | f, [], uw :: ux -> (false, false)
    | f, a :: asa, b :: bs ->
        (match f a b with (true, _) -> (true, true)
          | (false, true) -> lex_ext_unbounded f asa bs
          | (false, false) -> (false, false));;

let rec wpo_ub _B
  pr prl cS cNS sigma s t =
    (match s
      with Var x ->
        (false,
          (match t with Var a -> eq _B x a
            | Fun (g, ts) ->
              cNS s t &&
                (null (status sigma (g, size_list ts)) &&
                  prl (g, size_list ts))))
      | Fun (f, ss) ->
        (if cS s t then (true, true)
          else (let ff = (f, size_list ss) in
                let sf = status sigma ff in
                 (if cNS s t
                   then (if list_ex
                              (fun i ->
                                snd (wpo_ub _B pr prl cS cNS sigma (nth ss i)
                                      t))
                              sf
                          then (true, true)
                          else (match t with Var _ -> (false, false)
                                 | Fun (g, ts) ->
                                   (let gg = (g, size_list ts) in
                                    let sg = status sigma gg in
                                    let (prs, prns) = pr ff gg in
                                     (if prns &&
   list_all (fun j -> fst (wpo_ub _B pr prl cS cNS sigma s (nth ts j))) sg
                                       then (if prs then (true, true)
      else lex_ext_unbounded (wpo_ub _B pr prl cS cNS sigma) (map (nth ss) sf)
             (map (nth ts) sg))
                                       else (false, false)))))
                   else (false, false)))));;

let rec xml_error_to_string
  = function Fatal e -> explode ("Fatal: " ^ e)
    | TagMismatch e ->
        explode ("tag mismatch: " ^ default_showsl_list showsl_lit e "");;

let rec parse_xml
  p xml =
    bind2 (xml_take p xml_return ([xml], ([], (false, ([], [])))))
      (comp left xml_error_to_string) right;;

let rec mk_subst _A
  d xts = (fun x -> (match map_of _A xts x with None -> d x | Some t -> t));;

let rec cstep (_A1, _A2) (_B1, _B2)
  xml2name x =
    xml_do "conditionalRewriteStep"
      (xml_take (crule (showl_lab _A2 _B2) xml2name "rule")
        (fun a ->
          xml_take pos
            (fun b ->
              xml_take (substa (showl_lab _A2 _B2) xml2name)
                (fun c ->
                  xml_take
                    (xml_do "source"
                      (xml_take (term (showl_lab _A2 _B2) xml2name)
                        (fun xa -> xml_return (id xa))))
                    (fun d ->
                      xml_take
                        (xml_do "target"
                          (xml_take (term (showl_lab _A2 _B2) xml2name)
                            (fun xa -> xml_return (id xa))))
                        (fun e ->
                          xml_take
                            (xml_do "conditions"
                              (xml_take_many_sub [] zero_nata Infinity_enat
                                (csteps (_A1, _A2) (_B1, _B2) xml2name)
                                (fun aa -> xml_return (id aa))))
                            (fun f ->
                              xml_return
                                (Cstep_step
                                  (a, b,
                                    mk_subst (equal_list equal_char)
                                      (fun aa -> Var aa) c,
                                    d, e, f)))))))))
      x
and csteps (_A1, _A2) (_B1, _B2)
  xml2name x =
    xml_do "conditionalRewritingSequence"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (cstep (_A1, _A2) (_B1, _B2) xml2name) (fun a -> xml_return (id a)))
      x;;

let rec estep _A
  xml2name =
    xml_do "equationStep"
      (xml_take pos
        (fun p ->
          xml_take (rule _A xml2name)
            (fun r ->
              xml_take
                (xml_or (xml_leaf "leftRight" true)
                  (xml_leaf "rightLeft" false))
                (fun b ->
                  xml_take (term _A xml2name)
                    (fun t -> xml_return (p, (r, (b, t))))))));;

let rec state x = xml_text "state" x;;

let rec isOK m = (match m with Inl _ -> false | Inr _ -> true);;

let rec mapM
  f x1 = match f, x1 with f, [] -> Inr []
    | f, x :: xs ->
        bindb (f x) (fun y -> bindb (mapM f xs) (fun ys -> Inr (y :: ys)));;

let rec form_not = function Atom a -> NegAtom a
                   | NegAtom a -> Atom a
                   | Conjunction phi_s -> Disjunction (map form_not phi_s)
                   | Disjunction phi_s -> Conjunction (map form_not phi_s);;

let rec get_Atom = function Atom a -> a
                   | NegAtom a -> a;;

let rec simplify
  = function
    Disjunction (phi :: phi_s) ->
      form_or (simplify phi) (simplify (Disjunction phi_s))
    | Conjunction [phi] -> simplify phi
    | Conjunction (phi :: v :: va) ->
        form_and (simplify phi) (simplify (Conjunction (v :: va)))
    | Atom v -> Atom v
    | NegAtom v -> NegAtom v
    | Conjunction [] -> Conjunction []
    | Disjunction [] -> Disjunction [];;

let rec gen_ball it m p = it m (fun x -> x) (fun x _ -> p x) true;;

let rec gen_balla it s p = it s (fun x -> x) (fun x _ -> p x) true;;

let rec gen_pick
  it s =
    the (it s (fun a -> (match a with None -> true | Some _ -> false))
           (fun x _ -> Some x)
          None);;

let rec one (Partial_object_ext (carrier, Monoid_ext (mult, one, more))) = one;;

let rec remdups_adj _A
  = function [] -> []
    | [x] -> [x]
    | x :: y :: xs ->
        (if eq _A x y then remdups_adj _A (x :: xs)
          else x :: remdups_adj _A (y :: xs));;

let rec base (f, h) = f;;

let rec lift h f = (f, h);;

let rec add_vars_term x0 xs = match x0, xs with Var x, xs -> x :: xs
                        | Fun (uu, ts), xs -> foldr add_vars_term ts xs;;

let rec vars_term_list t = add_vars_term t [];;

let rec roof (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  (l, r) =
    (let xs = vars_term_list r in
      (fun t ->
        (let xt = vars_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7) t in
          list_all (fun x -> member (_B3, _B4.ccompare_cproper_interval) x xt)
            xs)));;

let rec matches _A _B (_C1, _C2, _C3)
  t p = (match match_list (_C1, _C2, _C3) _A _B (fun _ -> t) [(p, t)]
          with None -> false | Some _ -> true);;

let rec smult_mat _A aa a = map_mat (times _A aa) a;;

let rec rep_multimap (Abs_multimap x) = x;;

let rec option_list_to_list = function None -> []
                              | Some asa -> asa;;

let rec insertd (_B1, _B2)
  xb xc =
    Abs_multimap
      (let (f, (m, all)) = rep_multimap xc in
       let k = f xb in
       let old = option_list_to_list (lookupb (_B1, _B2) m k) in
       let newa = xb :: old in
        (f, (updateb (_B1, _B2) k newa m, xb :: all)));;

let rec lookupc (_A1, _A2)
  xa = (let (_, (m, _)) = rep_multimap xa in
         (fun k -> option_list_to_list (lookupb (_A1, _A2) m k)));;

let rec valuesa xa = (let (_, (_, all)) = rep_multimap xa in all);;

let rec subtract_entries_raw _A _B
  xs ys = foldr (fun (k, v) -> map_entry _A k (fun va -> minus _B va v)) ys xs;;

let rec subtract_entries _A _B
  xb xc = Alist (subtract_entries_raw _A _B (impl_ofa xb) (impl_ofa xc));;

let rec minus_multiset _A
  (Bag xs) (Bag ys) = Bag (subtract_entries _A minus_nat xs ys);;

let rec inf_subset_mset _A a b = minus_multiset _A a (minus_multiset _A a b);;

let rec fold_impl
  fn e x2 = match fn, e, x2 with
    fn, e, (a, n) :: ms -> fold_impl fn (fn a n e) ms
    | fn, e, [] -> e;;

let rec foldd f e al = fold_impl f e (impl_ofa al);;

let rec set_mset (_A1, _A2, _A3)
  (Bag ms) =
    foldd (fun a n ->
            (if equal_nata n zero_nata then (fun m -> m)
              else inserta (_A1, _A2) a))
      (bot_set (_A1, _A2, _A3)) ms;;

let rec multeqp (_A1, _A2, _A3, _A4)
  p n m =
    (let z = inf_subset_mset _A3 m n in
     let x = minus_multiset _A3 m z in
     let y = minus_multiset _A3 n z in
      ball (_A1, _A2) (set_mset (_A1, _A2, _A4) y)
        (fun ya -> bex (_A1, _A2) (set_mset (_A1, _A2, _A4) x) (p ya)));;

let rec is_Inr = function Inr uu -> true
                 | Inl uv -> false;;

let rec one_monom _A = Abs_monom [];;

let rec poly_split (_A1, _A2) _B
  m p = (match extract (fun (n, _) -> eq (equal_monom (_A1, _A2)) m n) p
          with None -> (zerob _B, p)
          | Some (p1, ((_, c), p2)) -> (c, p1 @ p2));;

let rec max_v (_A1, _A2)
  v i fs =
    max _A2 v
      (foldr
        (fun f m ->
          max _A2 m
            (fst (poly_split (equal_nat, linorder_nat) _A1
                   (one_monom linorder_nat) (i f))))
        fs (zerob _A1));;

let rec is_zero _A p = null (coeffs _A p);;

let rec order (_A1, _A2)
  a p = (if is_zero
              _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
              p
          then failwith "order of polynomial 0 undefined"
                 (fun _ -> order (_A1, _A2) a p)
          else (if not (eq _A1
                         (polya
                           _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.comm_semiring_0_comm_semiring_1
                           p a)
                         (zerob
                           _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero))
                 then zero_nata
                 else suc (order (_A1, _A2) a
                            (divide_polya (_A1, _A2) p
                              (pCons
                                (_A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero,
                                  _A1)
                                (uminus
                                  _A2.idom_idom_divide.comm_ring_1_idom.ring_1_comm_ring_1.neg_numeral_ring_1.group_add_neg_numeral.uminus_group_add
                                  a)
                                (pCons
                                  (_A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero,
                                    _A1)
                                  (onea _A2.idom_idom_divide.comm_ring_1_idom.ring_1_comm_ring_1.neg_numeral_ring_1.numeral_neg_numeral.one_numeral)
                                  (zero_polyb
                                    _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)))))));;

let rec relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
  x0 x1 = match x0, x1 with
    Set_Monad xs6, DList_set dxs4 ->
      (match ceq _B1
        with None ->
          failwith "relcomp Set_Monad DList_set: ceq1 = None"
            (fun _ ->
              relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3) (Set_Monad xs6)
                (DList_set dxs4))
        | Some eq ->
          (match ceq _C1
            with None ->
              failwith "relcomp Set_Monad DList_set: ceq2 = None"
                (fun _ ->
                  relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                    (Set_Monad xs6) (DList_set dxs4))
            | Some _ ->
              fold (fun (x, y) ->
                     foldc (ceq_prod _B1 _C1)
                       (fun (ya, z) a ->
                         (if eq y ya
                           then inserta
                                  ((ceq_prod _A1 _C1), (ccompare_prod _A2 _C2))
                                  (x, z) a
                           else a))
                       dxs4)
                xs6 (bot_set
                      ((ceq_prod _A1 _C1), (ccompare_prod _A2 _C2),
                        (set_impl_prod _A3 _C3)))))
    | DList_set dxs3, Set_Monad xs5 ->
        (match ceq _A1
          with None ->
            failwith "relcomp DList_set Set_Monad: ceq1 = None"
              (fun _ ->
                relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                  (DList_set dxs3) (Set_Monad xs5))
          | Some _ ->
            (match ceq _B1
              with None ->
                failwith "relcomp DList_set Set_Monad: ceq2 = None"
                  (fun _ ->
                    relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                      (DList_set dxs3) (Set_Monad xs5))
              | Some eq ->
                foldc (ceq_prod _A1 _B1)
                  (fun (x, y) ->
                    fold (fun (ya, z) a ->
                           (if eq y ya
                             then inserta
                                    ((ceq_prod _A1 _C1),
                                      (ccompare_prod _A2 _C2))
                                    (x, z) a
                             else a))
                      xs5)
                  dxs3
                  (bot_set
                    ((ceq_prod _A1 _C1), (ccompare_prod _A2 _C2),
                      (set_impl_prod _A3 _C3)))))
    | Set_Monad xs4, RBT_set rbt5 ->
        (match ccompare _C2
          with None ->
            failwith "relcomp Set_Monad RBT_set: ccompare1 = None"
              (fun _ ->
                relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                  (Set_Monad xs4) (RBT_set rbt5))
          | Some _ ->
            (match ccompare _B2
              with None ->
                failwith "relcomp Set_Monad RBT_set: ccompare2 = None"
                  (fun _ ->
                    relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                      (Set_Monad xs4) (RBT_set rbt5))
              | Some c_b ->
                fold (fun (x, y) ->
                       foldb (ccompare_prod _B2 _C2)
                         (fun (ya, z) a ->
                           (if not (equal_order (c_b y ya) Eq) then a
                             else inserta
                                    ((ceq_prod _A1 _C1),
                                      (ccompare_prod _A2 _C2))
                                    (x, z) a))
                         rbt5)
                  xs4 (bot_set
                        ((ceq_prod _A1 _C1), (ccompare_prod _A2 _C2),
                          (set_impl_prod _A3 _C3)))))
    | RBT_set rbt1, Set_Monad xs3 ->
        (match ccompare _A2
          with None ->
            failwith "relcomp RBT_set Set_Monad: ccompare1 = None"
              (fun _ ->
                relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                  (RBT_set rbt1) (Set_Monad xs3))
          | Some _ ->
            (match ccompare _B2
              with None ->
                failwith "relcomp RBT_set Set_Monad: ccompare2 = None"
                  (fun _ ->
                    relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                      (RBT_set rbt1) (Set_Monad xs3))
              | Some c_b ->
                foldb (ccompare_prod _A2 _B2)
                  (fun (x, y) ->
                    fold (fun (ya, z) a ->
                           (if not (equal_order (c_b y ya) Eq) then a
                             else inserta
                                    ((ceq_prod _A1 _C1),
                                      (ccompare_prod _A2 _C2))
                                    (x, z) a))
                      xs3)
                  rbt1
                  (bot_set
                    ((ceq_prod _A1 _C1), (ccompare_prod _A2 _C2),
                      (set_impl_prod _A3 _C3)))))
    | Set_Monad xs1, Set_Monad xs2 ->
        (match ceq _B1
          with None ->
            failwith "relcomp Set_Monad Set_Monad: ceq = None"
              (fun _ ->
                relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                  (Set_Monad xs1) (Set_Monad xs2))
          | Some eq ->
            fold (fun (x, y) ->
                   fold (fun (ya, z) a ->
                          (if eq y ya
                            then inserta
                                   ((ceq_prod _A1 _C1), (ccompare_prod _A2 _C2))
                                   (x, z) a
                            else a))
                     xs2)
              xs1 (bot_set
                    ((ceq_prod _A1 _C1), (ccompare_prod _A2 _C2),
                      (set_impl_prod _A3 _C3))))
    | DList_set dxs3, DList_set dxs4 ->
        (match ceq _A1
          with None ->
            failwith "relcomp DList_set DList_set: ceq1 = None"
              (fun _ ->
                relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                  (DList_set dxs3) (DList_set dxs4))
          | Some _ ->
            (match ceq _B1
              with None ->
                failwith "relcomp DList_set DList_set: ceq2 = None"
                  (fun _ ->
                    relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                      (DList_set dxs3) (DList_set dxs4))
              | Some eq ->
                (match ceq _C1
                  with None ->
                    failwith "relcomp DList_set DList_set: ceq3 = None"
                      (fun _ ->
                        relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                          (DList_set dxs3) (DList_set dxs4))
                  | Some _ ->
                    foldc (ceq_prod _A1 _B1)
                      (fun (x, y) ->
                        foldc (ceq_prod _B1 _C1)
                          (fun (ya, z) a ->
                            (if eq y ya
                              then inserta
                                     ((ceq_prod _A1 _C1),
                                       (ccompare_prod _A2 _C2))
                                     (x, z) a
                              else a))
                          dxs4)
                      dxs3
                      (bot_set
                        ((ceq_prod _A1 _C1), (ccompare_prod _A2 _C2),
                          (set_impl_prod _A3 _C3))))))
    | DList_set dxs2, RBT_set rbt4 ->
        (match ceq _A1
          with None ->
            failwith "relcomp DList_set RBT_set: ceq1 = None"
              (fun _ ->
                relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                  (DList_set dxs2) (RBT_set rbt4))
          | Some _ ->
            (match ccompare _B2
              with None ->
                failwith "relcomp DList_set RBT_set: ceq2 = None"
                  (fun _ ->
                    relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                      (DList_set dxs2) (RBT_set rbt4))
              | Some _ ->
                (match ceq _B1
                  with None ->
                    failwith "relcomp DList_set RBT_set: ccompare2 = None"
                      (fun _ ->
                        relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                          (DList_set dxs2) (RBT_set rbt4))
                  | Some eq ->
                    (match ccompare _C2
                      with None ->
                        failwith "relcomp DList_set RBT_set: ccompare3 = None"
                          (fun _ ->
                            relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                              (DList_set dxs2) (RBT_set rbt4))
                      | Some _ ->
                        foldc (ceq_prod _A1 _B1)
                          (fun (x, y) ->
                            foldb (ccompare_prod _B2 _C2)
                              (fun (ya, z) a ->
                                (if eq y ya
                                  then inserta
 ((ceq_prod _A1 _C1), (ccompare_prod _A2 _C2)) (x, z) a
                                  else a))
                              rbt4)
                          dxs2
                          (bot_set
                            ((ceq_prod _A1 _C1), (ccompare_prod _A2 _C2),
                              (set_impl_prod _A3 _C3)))))))
    | RBT_set rbt3, DList_set dxs1 ->
        (match ccompare _A2
          with None ->
            failwith "relcomp RBT_set DList_set: ccompare1 = None"
              (fun _ ->
                relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                  (RBT_set rbt3) (DList_set dxs1))
          | Some _ ->
            (match ccompare _B2
              with None ->
                failwith "relcomp RBT_set DList_set: ccompare2 = None"
                  (fun _ ->
                    relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                      (RBT_set rbt3) (DList_set dxs1))
              | Some _ ->
                (match ceq _B1
                  with None ->
                    failwith "relcomp RBT_set DList_set: ceq2 = None"
                      (fun _ ->
                        relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                          (RBT_set rbt3) (DList_set dxs1))
                  | Some eq ->
                    (match ceq _C1
                      with None ->
                        failwith "relcomp RBT_set DList_set: ceq3 = None"
                          (fun _ ->
                            relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                              (RBT_set rbt3) (DList_set dxs1))
                      | Some _ ->
                        foldb (ccompare_prod _A2 _B2)
                          (fun (x, y) ->
                            foldc (ceq_prod _B1 _C1)
                              (fun (ya, z) a ->
                                (if eq y ya
                                  then inserta
 ((ceq_prod _A1 _C1), (ccompare_prod _A2 _C2)) (x, z) a
                                  else a))
                              dxs1)
                          rbt3
                          (bot_set
                            ((ceq_prod _A1 _C1), (ccompare_prod _A2 _C2),
                              (set_impl_prod _A3 _C3)))))))
    | RBT_set rbt1, RBT_set rbt2 ->
        (match ccompare _A2
          with None ->
            failwith "relcomp RBT_set RBT_set: ccompare1 = None"
              (fun _ ->
                relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                  (RBT_set rbt1) (RBT_set rbt2))
          | Some _ ->
            (match ccompare _B2
              with None ->
                failwith "relcomp RBT_set RBT_set: ccompare2 = None"
                  (fun _ ->
                    relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                      (RBT_set rbt1) (RBT_set rbt2))
              | Some c_b ->
                (match ccompare _C2
                  with None ->
                    failwith "relcomp RBT_set RBT_set: ccompare3 = None"
                      (fun _ ->
                        relcomp (_A1, _A2, _A3) (_B1, _B2) (_C1, _C2, _C3)
                          (RBT_set rbt1) (RBT_set rbt2))
                  | Some _ ->
                    foldb (ccompare_prod _A2 _B2)
                      (fun (x, y) ->
                        foldb (ccompare_prod _B2 _C2)
                          (fun (ya, z) a ->
                            (if not (equal_order (c_b y ya) Eq) then a
                              else inserta
                                     ((ceq_prod _A1 _C1),
                                       (ccompare_prod _A2 _C2))
                                     (x, z) a))
                          rbt2)
                      rbt1
                      (bot_set
                        ((ceq_prod _A1 _C1), (ccompare_prod _A2 _C2),
                          (set_impl_prod _A3 _C3))))));;

let rec max_sls _A
  = Abs_semilattice_set (max _A.order_linorder.preorder_order.ord_preorder);;

let rec error x = Sumbot (Inl x);;

let rec tcapRM2 _A _B
  uu x1 = match uu, x1 with uu, Var uv -> GCHole
    | rm, Fun (f, ts) ->
        (let h = GCFun (f, map (tcapRM2 _A _B rm) ts) in
         let n = size_list ts in
          (if list_ex (fun r -> matchc _A _B h (fst r)) (rm (f, n)) then GCHole
            else h));;

let rec tcapRM _A _B
  nlv rm = (if nlv then tcapRM2 _A _B rm else (fun _ -> GCHole));;

let rec funposs_list
  = function Var x -> []
    | Fun (f, ss) ->
        [] :: maps (fun (i, a) -> map (fun aa -> i :: aa) a)
                (zip (upt zero_nata (size_list ss)) (map funposs_list ss));;

let rec ctxt_apply_term
  x0 s = match x0, s with Hole, s -> s
    | More (f, ss1, c, ss2), s -> Fun (f, ss1 @ ctxt_apply_term c s :: ss2);;

let rec ctxt_of_pos_term
  x0 t = match x0, t with [], t -> Hole
    | i :: ps, Fun (f, ts) ->
        More (f, take i ts, ctxt_of_pos_term ps (nth ts i), drop (suc i) ts);;

let rec poss_list
  = function Var x -> [[]]
    | Fun (f, ss) ->
        [] :: maps (fun (i, a) -> map (fun aa -> i :: aa) a)
                (zip (upt zero_nata (size_list ss)) (map poss_list ss));;

let rec subt_at s x1 = match s, x1 with s, [] -> s
                  | Fun (f, ss), i :: p -> subt_at (nth ss i) p;;

let rec rewrite _A (_B1, _B2, _B3)
  r s = maps (fun (l, ra) ->
               maps (fun p ->
                      (match matcha _A (_B1, _B2, _B3) (subt_at s p) l
                        with None -> []
                        | Some sigma ->
                          [ctxt_apply_term (ctxt_of_pos_term p s)
                             (subst_apply_term ra sigma)]))
                 (poss_list s))
          r;;

let rec hvf_term _A
  a t = (match unapp _A a t with (Var _, ts) -> null ts
          | (Fun (_, us), ts) -> list_all (hvf_term _A a) (us @ ts));;

let rec max_list = function [] -> zero_nata
                   | x :: xs -> max ord_nat x (max_list xs);;

let rec rules _A
  xml2name =
    xml_do "rules"
      (xml_take_many_sub [] zero_nata Infinity_enat (rule _A xml2name)
        xml_return);;

let rec string2xml
  str = (match
          bindc (update_tokens remove_comments)
            (fun _ ->
              bindc parse_node
                (fun xml -> bindc (eoi show_char) (fun _ -> returna xml)))
            str
          with Inl a -> Inl a | Inr (xml, _) -> Inr xml);;

let rec filter_mset p (Bag xs) = Bag (filterb (comp p fst) xs);;

let rec filter_fun
  t p f =
    filter_mset
      (fun a ->
        (match a with Var _ -> false | Fun (g, ts) -> p (g, size_list ts) f))
      t;;

let rec stringa (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "string"
      (xml_take_many_sub [] zero_nata Infinity_enat xml2name
        (fun a -> xml_return (id a)));;

let rec oc_srs (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "overlapClosureSRS"
      (xml_take (stringa (_A1, _A2) (_B1, _B2) xml2name)
        (fun a ->
          xml_take (stringa (_A1, _A2) (_B1, _B2) xml2name)
            (fun b -> xml_return (a, b))));;

let rec rsteps _A
  xml2name =
    xml_do "rewriteSequence"
      (xml_take
        (xml_do "startTerm"
          (xml_take (term _A xml2name) (fun x -> xml_return (id x))))
        (fun start ->
          xml_take_many_sub [] zero_nata Infinity_enat (rstep _A xml2name)
            (fun steps -> xml_return (start, steps))));;

let rec check b e = (if b then Inr () else Inl e);;

let rec or_ok x0 b = match x0, b with Inl a, b -> b
                | Inr a, b -> Inr a;;

let rec degree (Comp_Poly d) = d;;

let rec gen_image
  it1 emp2 ins2 f s1 = it1 s1 (fun _ -> true) (fun x -> ins2 (f x)) emp2;;

let rec catch_error m f = (match m with Inl a -> f a | Inr a -> Inr a);;

let rec check_no_edge
  ss g m n =
    check (not (g (m, n)))
      (comp (comp (comp (showsl_lit "edge from ") (ss m)) (showsl_lit " to "))
        (ss n));;

let rec forallM
  f x1 = match f, x1 with f, [] -> Inr ()
    | f, x :: xs ->
        bindb (catch_error (f x) (fun xa -> Inl (x, xa)))
          (fun _ -> forallM f xs);;

let rec check_edges
  ss g c d =
    catch_error
      (forallM
        (fun n ->
          catch_error (forallM (check_no_edge ss g n) d) (fun x -> Inl (snd x)))
        c)
      (fun x -> Inl (snd x));;

let rec mult
  (Partial_object_ext (carrier, Monoid_ext (mult, one, more))) = mult;;

let rec comp2eq
  cmp a b =
    (match cmp a b with LESS -> false | EQUAL -> true | GREATER -> false);;

let rec prec_ext _A
  prwm =
    (fun f g ->
      (match prwm f with None -> (false, eq _A f g)
        | Some pf ->
          (match prwm g with None -> (true, true)
            | Some pg ->
              (less_nat (fst pg) (fst pf), less_eq_nat (fst pg) (fst pf)))));;

let rec assertion
  (Lts_ext (initial, transition_rules, assertion, more)) = assertion;;

let rec map_of_total (_A1, _A2, _A3)
  err xys =
    (let m = of_alist (_A1, _A2, _A3) xys in
      (fun x ->
        (match lookupb (_A1, _A2) m x
          with None -> failwith (err x "") (fun _ -> the None)
          | Some y -> y)));;

let rec transition_of (_E1, _E2, _E3, _E4)
  pi = map_of_total (_E1, _E2, _E3)
         (fun a ->
           comp (comp (comp (showsl_lit "access to non-existing transition ")
                        (showsl _E4 a))
                  (showsl_lit "\092available transitions:\092"))
             (showsl_list _E4 (map fst (transitions_impl pi))))
         (transitions_impl pi);;

let rec matchb x = (fun _ _ -> true) x;;

let rec span
  p x1 = match p, x1 with
    p, x :: xs ->
      (if p x then (let a = span p xs in
                    let (ys, aa) = a in
                     (x :: ys, aa))
        else ([], x :: xs))
    | uu, [] -> ([], []);;

let rec fresh_strings_list
  name offset used n =
    take n
      (filtera
        (fun s ->
          not (membera (equal_list equal_char)
                (remdups (equal_list equal_char) used) s))
        (map (fun i -> name @ shows_prec_nat zero_nata (plus_nata i offset) [])
          (upt zero_nata
            (plus_nata n
              (size_list (remdups (equal_list equal_char) used))))));;

let rec fresh_string
  pre = (fun s -> hda (fresh_strings_list pre one_nata s one_nata));;

let rec mapMa
  f x1 = match f, x1 with f, [] -> Some []
    | f, x :: xs ->
        bind (f x) (fun y -> bind (mapMa f xs) (fun ys -> Some (y :: ys)));;

let rec degreea _A p = minus_nata (size_list (coeffs _A p)) one_nata;;

let rec pderiv_coeffs_code (_A1, _A2, _A3)
  f x1 = match f, x1 with
    f, x :: xs ->
      cCons (_A2.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero,
              _A1)
        (times
          _A2.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd f
          x)
        (pderiv_coeffs_code (_A1, _A2, _A3)
          (plus _A2.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.semigroup_add_numeral.plus_semigroup_add
            f (onea _A2.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral))
          xs)
    | f, [] -> [];;

let rec pderiv_coeffs (_A1, _A2, _A3)
  xs = pderiv_coeffs_code (_A1, _A2, _A3)
         (onea _A2.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral)
         (tla xs);;

let rec pderiv (_A1, _A2, _A3)
  p = Poly (pderiv_coeffs (_A1, _A2, _A3)
             (coeffs
               _A2.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
               p));;

let rec less_eq_pos
  x0 p = match x0, p with
    i :: q1, j :: q2 -> equal_nata i j && less_eq_pos q1 q2
    | i :: q1, [] -> false
    | [], p -> true;;

let rec less_pos p q = less_eq_pos p q && not (equal_lista equal_nat p q);;

let rec equal_order_tag x0 x1 = match x0, x1 with Lex, Mul -> false
                          | Mul, Lex -> false
                          | Mul, Mul -> true
                          | Lex, Lex -> true;;

let rec any_nstri_efficient_m
  f x1 m = match f, x1, m with f, [], m -> (false, m)
    | f, a :: asa, m ->
        (let aa = f a m in
         let (ab, b) = aa in
          (let (_, ns) = ab in
            (fun ma ->
              (if ns then (true, ma) else any_nstri_efficient_m f asa ma)))
            b);;

let rec all_stri_efficient_m
  f x1 m = match f, x1, m with f, [], m -> (true, m)
    | f, a :: asa, m ->
        (let aa = f a m in
         let (ab, b) = aa in
          (let (s, _) = ab in
            (fun ma ->
              (if s then all_stri_efficient_m f asa ma else (false, ma))))
            b);;

let rec filter_not_stri_efficient_m_aux
  f x1 bs m = match f, x1, bs, m with f, [], bs, m -> (bs, m)
    | f, a :: asa, bs, m ->
        (let b = f a m in
         let (ba, c) = b in
          (let (s, _) = ba in
            (fun ma ->
              (if s then filter_not_stri_efficient_m_aux f asa bs ma
                else filter_not_stri_efficient_m_aux f asa (a :: bs) ma)))
            c);;

let rec filter_not_stri_efficient_m
  f asa m =
    (let a = filter_not_stri_efficient_m_aux f asa [] m in
     let (ys, aa) = a in
      (rev ys, aa));;

let rec mul_ext_efficient_m
  f x1 x2 m = match f, x1, x2, m with f, [], [], m -> ((false, true), m)
    | f, [], v :: va, m -> ((false, false), m)
    | f, v :: va, [], m -> ((true, true), m)
    | f, v :: va, y :: ys, m -> mul_ex_dom_efficient_m f (v :: va) [] y ys m
and mul_ex_dom_efficient_m
  f x1 xs y ys m = match f, x1, xs, y, ys, m with
    f, [], xs, y, ys, m -> ((false, false), m)
    | f, x :: xsa, xs, y, ys, m ->
        (let a = f x y m in
         let (aa, b) = a in
          (let (s, ns) = aa in
            (fun ma ->
              (if s then (let (ysa, m2) =
                            filter_not_stri_efficient_m (f x) ys ma in
                          let ab = mul_ext_efficient_m f (xsa @ xs) ysa m2 in
                          let (ac, ba) = ab in
                           (let (_, nsa) = ac in
                             (fun m3 ->
                               (if nsa then ((true, true), m3)
                                 else mul_ex_dom_efficient_m f xsa (x :: xs) y
ys m3)))
                             ba)
                else (if ns
                       then (let ab = mul_ext_efficient_m f (xsa @ xs) ys ma in
                             let (ac, ba) = ab in
                              (let (s2, ns2) = ac in
                                (fun m4 ->
                                  (if s2 then ((true, true), m4)
                                    else (let ad =
    mul_ex_dom_efficient_m f xsa (x :: xs) y ys m4 in
  let (ae, bb) = ad in
   (let (s3, ns3) = ae in
     (fun m5 -> (if s3 then ((true, true), m5) else ((false, ns2 || ns3), m5))))
     bb))))
                                ba)
                       else mul_ex_dom_efficient_m f xsa (x :: xs) y ys ma))))
            b);;

let rec lex_ext_efficient_m
  f x1 x2 m = match f, x1, x2, m with f, [], [], m -> ((false, true), m)
    | f, uu :: uv, [], m -> ((true, true), m)
    | f, [], uw :: ux, m -> ((false, false), m)
    | f, a :: asa, b :: bs, m ->
        (let aa = f a b m in
         let (ab, ba) = aa in
          (let (s, ns) = ab in
            (fun ma ->
              (if s then ((true, true), ma)
                else (if ns then lex_ext_efficient_m f asa bs ma
                       else ((false, false), ma)))))
            ba);;

let rec lookupe (Memory_ext (empty, lookup, store, more)) = lookup;;

let rec store (Memory_ext (empty, lookup, store, more)) = store;;

let rec diag_l (a, b) = ((a, a), b);;

let rec rpo_efficient_m _C
  model pr tag s t m =
    (match lookupe model m (s, t)
      with None ->
        (let (res, res_m) =
           (match (s, t) with (Var x, Var y) -> ((false, eq _C x y), m)
             | (Var _, Fun (g, ts)) ->
               ((false, null ts && snd pr (g, zero_nata)), m)
             | (Fun (_, ss), Var y) ->
               diag_l
                 (any_nstri_efficient_m
                   (fun sa -> rpo_efficient_m _C model pr tag sa (Var y)) ss m)
             | (Fun (f, ss), Fun (g, ts)) ->
               (match
                 any_nstri_efficient_m
                   (fun sa -> rpo_efficient_m _C model pr tag sa (Fun (g, ts)))
                   ss m
                 with (true, ma) -> ((true, true), ma)
                 | (false, ma) ->
                   (let maa = size_list ss in
                    let n = size_list ts in
                     (match fst pr (f, maa) (g, n)
                       with (prs, true) ->
                         (match
                           all_stri_efficient_m
                             (rpo_efficient_m _C model pr tag (Fun (f, ss))) ts
                             ma
                           with (true, mb) ->
                             (if prs then ((true, true), mb)
                               else (let cf = tag (f, maa) in
                                      (if equal_order_tag cf (tag (g, n))
then (if equal_order_tag cf Mul
       then mul_ext_efficient_m (rpo_efficient_m _C model pr tag) ss ts mb
       else lex_ext_efficient_m (rpo_efficient_m _C model pr tag) ss ts mb)
else (let b = equal_nata n zero_nata in
       ((not (equal_nata maa zero_nata) && b, b), mb)))))
                           | (false, mb) -> ((false, false), mb))
                       | (_, false) -> ((false, false), ma)))))
           in
          (res, store model res_m ((s, t), res)))
      | Some v -> (v, m));;

let rec emptyf (Memory_ext (empty, lookup, store, more)) = empty;;

let rec efficient_rpo_2 _A _B
  pr tag p q =
    fst (rpo_efficient_m _B (l2m (equal_term _A _B) (equal_term _A _B)) pr tag p
          q (emptyf (l2m (equal_term _A _B) (equal_term _A _B)) ()));;

let rec rpo_unbounded_impl _A _B = efficient_rpo_2 _A _B;;

let rec rpo_unbounded _A _B = rpo_unbounded_impl _A _B;;

let rec converse (_A1, _A2, _A3) (_B1, _B2, _B3)
  r = image ((ceq_prod _A1 _B1), (ccompare_prod _A2 _B2))
        ((ceq_prod _B1 _A1), (ccompare_prod _B2 _A2), (set_impl_prod _B3 _A3))
        (fun (x, y) -> (y, x)) r;;

let rec return x = Sumbot (Inr x);;

let rec the_Var (Var x1) = x1;;

let rec in_poss
  x0 uu = match x0, uu with [], uu -> true
    | i :: p, Fun (f, ts) -> less_nat i (size_list ts) && in_poss p (nth ts i)
    | i :: p, Var uv -> false;;

let rec varposs
  = function
    Var x ->
      inserta ((ceq_list ceq_nat), (ccompare_list ccompare_nat)) []
        (set_empty ((ceq_list ceq_nat), (ccompare_list ccompare_nat))
          (of_phantom set_impl_lista))
    | Fun (f, ts) ->
        sup_setb
          (finite_UNIV_list, cenum_list, (ceq_list ceq_nat),
            (cproper_interval_list ccompare_nat), set_impl_list)
          (image (ceq_nat, ccompare_nat)
            ((ceq_set
               (cenum_list, (ceq_list ceq_nat),
                 (cproper_interval_list
                   ccompare_nat).ccompare_cproper_interval)),
              (ccompare_set
                (finite_UNIV_list, (ceq_list ceq_nat),
                  (cproper_interval_list ccompare_nat), set_impl_list)),
              set_impl_set)
            (fun i ->
              image ((ceq_list ceq_nat), (ccompare_list ccompare_nat))
                ((ceq_list ceq_nat), (ccompare_list ccompare_nat),
                  set_impl_list)
                (fun a -> i :: a) (varposs (nth ts i)))
            (set (ceq_nat, ccompare_nat, set_impl_nat)
              (upt zero_nata (size_list ts))));;

let rec add_funs_term x0 fs = match x0, fs with Var uu, fs -> fs
                        | Fun (f, ts), fs -> f :: foldr add_funs_term ts fs;;

let rec add_funs_rule r fs = add_funs_term (fst r) (add_funs_term (snd r) fs);;

let rec funs_trs_list trs = foldr add_funs_rule trs [];;

let rec instance_rule _A _B (_C1, _C2, _C3)
  lr st =
    not (is_none
          (match_list (_C1, _C2, _C3) _A _B (fun _ -> fst lr)
            [(fst st, fst lr); (snd st, snd lr)]));;

let rec map_funs_rule
  fg lr =
    (map_term fg (fun x -> x) (fst lr), map_term fg (fun x -> x) (snd lr));;

let rec add_vars_rule r xs = add_vars_term (fst r) (add_vars_term (snd r) xs);;

let rec vars_trs_list trs = foldr add_vars_rule trs [];;

let rec elem_list_to_rm _B
  key x1 = match key, x1 with
    key, d :: ds ->
      (let rm = elem_list_to_rm _B key ds in
       let k = key d in
        (match lookup _B rm k with None -> insert _B k [d] rm
          | Some data -> insert _B k (d :: data) rm))
    | key, [] -> empty _B.linorder_compare_order;;

let rec fun_of_map m d a = (match m a with None -> d | Some b -> b);;

let rec term_map _A
  ts = fun_of_map
         (lookup (compare_order_prod _A compare_order_nat)
           (elem_list_to_rm (compare_order_prod _A compare_order_nat)
             (comp the root) ts))
         [];;

let rec label_depth = function UnLab uu -> zero_nata
                      | Lab (f, uv) -> suc (label_depth f)
                      | FunLab (f, uw) -> suc (label_depth f)
                      | Sharp f -> suc (label_depth f);;

let rec gen_label
  f n = (if equal_nata n zero_nata then f
          else FunLab (gen_label f (minus_nata n one_nata), []));;

let rec fmap _A _B
  a nn sml =
    (let m = suc (max_list (map label_depth (a :: maps snd sml))) in
      (fun f n ->
        (if equal_proda (equal_lab _A _B) equal_nat (f, n) (a, nn) then a
          else gen_label f m)));;

let rec enum_vectors
  c x1 = match c, x1 with c, [] -> [[]]
    | c, x :: xs ->
        (let a = enum_vectors c xs in
          maps (fun vec -> map (fun ca -> (x, ca) :: vec) c) a);;

let rec replace_impl _A
  a bs m =
    (if membera _A m a then bs @ filtera (fun b -> not (eq _A b a)) m else m);;

let rec doc_of_string s = bindb (parse_doc s) (fun (doc, _) -> Inr doc);;

let rec special_map
  x = map_of (equal_list equal_char)
        [([Chara (true, false, false, false, true, true, true, false);
            Chara (true, false, true, false, true, true, true, false);
            Chara (true, true, true, true, false, true, true, false);
            Chara (false, false, true, false, true, true, true, false)],
           [Chara (false, true, false, false, false, true, false, false)]);
          ([Chara (true, true, false, false, false, true, false, false);
             Chara (true, true, false, false, true, true, false, false);
             Chara (false, false, true, false, true, true, false, false)],
            [Chara (false, true, false, false, false, true, false, false)]);
          ([Chara (true, false, false, false, false, true, true, false);
             Chara (true, false, true, true, false, true, true, false);
             Chara (false, false, false, false, true, true, true, false)],
            [Chara (false, true, true, false, false, true, false, false)]);
          ([Chara (true, true, false, false, false, true, false, false);
             Chara (true, true, false, false, true, true, false, false);
             Chara (false, false, false, true, true, true, false, false)],
            [Chara (false, true, true, false, false, true, false, false)]);
          ([Chara (true, false, false, false, false, true, true, false);
             Chara (false, false, false, false, true, true, true, false);
             Chara (true, true, true, true, false, true, true, false);
             Chara (true, true, false, false, true, true, true, false)],
            [Chara (true, true, true, false, false, true, false, false)]);
          ([Chara (true, true, false, false, false, true, false, false);
             Chara (true, true, false, false, true, true, false, false);
             Chara (true, false, false, true, true, true, false, false)],
            [Chara (true, true, true, false, false, true, false, false)]);
          ([Chara (false, false, true, true, false, true, true, false);
             Chara (false, false, true, false, true, true, true, false)],
            [Chara (false, false, true, true, true, true, false, false)]);
          ([Chara (true, true, false, false, false, true, false, false);
             Chara (false, true, true, false, true, true, false, false);
             Chara (false, false, false, false, true, true, false, false)],
            [Chara (false, false, true, true, true, true, false, false)]);
          ([Chara (true, true, true, false, false, true, true, false);
             Chara (false, false, true, false, true, true, true, false)],
            [Chara (false, true, true, true, true, true, false, false)]);
          ([Chara (true, true, false, false, false, true, false, false);
             Chara (false, true, true, false, true, true, false, false);
             Chara (false, true, false, false, true, true, false, false)],
            [Chara (false, true, true, true, true, true, false, false)])]
        x;;

let rec pat_eqv_prf _A
  xml2name =
    (let sub = substa _A xml2name in
      xml_do "patternEquivalence"
        (xml_take
          (xml_or
            (xml_do "domainRenaming"
              (xml_take sub (fun n -> xml_return (Pat_Dom_Renaming n))))
            (xml_or
              (xml_do "irrelevant"
                (xml_take sub
                  (fun n ->
                    xml_take sub
                      (fun n2 -> xml_return (Pat_Irrelevant (n, n2))))))
              (xml_do "simplification"
                (xml_take sub
                  (fun n ->
                    xml_take sub
                      (fun n2 -> xml_return (Pat_Simplify (n, n2))))))))
          (fun x -> xml_return (id x))));;

let rec pat_term _A
  xml2name =
    xml_do "patternTerm"
      (xml_take (term _A xml2name)
        (fun t ->
          xml_take (substa _A xml2name)
            (fun s1 ->
              xml_take (substa _A xml2name)
                (fun s2 -> xml_return (t, (s1, s2))))));;

let rec pat_rule_prf _A
  xml2name x =
    (let pat = pat_rule_prf _A xml2name in
     let sub = substa _A xml2name in
     let pt = pat_term _A xml2name in
     let var = xml_text "var" in
      xml_do "patternRule"
        (xml_take pt
          (fun _ ->
            xml_take pt
              (fun _ ->
                xml_take
                  (xml_or
                    (xml_do "originalRule"
                      (xml_take (rule _A xml2name)
                        (fun r ->
                          xml_take (xml_bool "isPair")
                            (fun b -> xml_return (Pat_OrigRule (r, b))))))
                    (xml_or
                      (xml_do "initialPumping"
                        (xml_take pat
                          (fun a ->
                            xml_take sub
                              (fun b ->
                                xml_take sub
                                  (fun c ->
                                    xml_return (Pat_InitPump (a, b, c)))))))
                      (xml_or
                        (xml_do "initialPumpingContext"
                          (xml_take pat
                            (fun a ->
                              xml_take sub
                                (fun b ->
                                  xml_take pos
                                    (fun c ->
                                      xml_take var
(fun d -> xml_return (Pat_InitPumpCtxt (a, b, c, d))))))))
                        (xml_or
                          (xml_do "equivalence"
                            (xml_take pat
                              (fun a ->
                                xml_take
                                  (xml_or (xml_leaf "left" true)
                                    (xml_leaf "right" false))
                                  (fun b ->
                                    xml_take (pat_eqv_prf _A xml2name)
                                      (fun c ->
xml_return (Pat_Equiv (a, b, c)))))))
                          (xml_or
                            (xml_do "narrowing"
                              (xml_take pat
                                (fun a ->
                                  xml_take pat
                                    (fun b ->
                                      xml_take pos
(fun c -> xml_return (Pat_Narrow (a, b, c)))))))
                            (xml_or
                              (xml_do "instantiation"
                                (xml_take pat
                                  (fun a ->
                                    xml_take sub
                                      (fun b ->
xml_take
  (xml_or (xml_leaf "base" Pat_Base)
    (xml_or (xml_leaf "pumping" Pat_Pump) (xml_leaf "closing" Pat_Close)))
  (fun c -> xml_return (Pat_Inst (a, b, c)))))))
                              (xml_or
                                (xml_do "instantiationPumping"
                                  (xml_take pat
                                    (fun a ->
                                      xml_take (xml_nat "power")
(fun b -> xml_return (Pat_Exp_Sigma (a, b))))))
                                (xml_do "rewriting"
                                  (xml_take pat
                                    (fun a ->
                                      xml_take (rsteps _A xml2name)
(fun b ->
  xml_take
    (xml_or (xml_leaf "base" (Pat_Base, []))
      (xml_or
        (xml_do "pumping" (xml_take var (fun v -> xml_return (Pat_Pump, v))))
        (xml_do "closing" (xml_take var (fun v -> xml_return (Pat_Close, v))))))
    (fun (po, va) -> xml_return (Pat_Rewr (a, b, po, va))))))))))))))
                  xml_return)))
        x);;

let rec nonloop (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "nonLoop"
      (xml_take (pat_rule_prf (showl_lab _A2 _B2) xml2name)
        (fun a ->
          xml_take (substa (showl_lab _A2 _B2) xml2name)
            (fun b ->
              xml_take (substa (showl_lab _A2 _B2) xml2name)
                (fun c ->
                  xml_take (xml_nat "natural")
                    (fun d ->
                      xml_take (xml_nat "natural")
                        (fun e ->
                          xml_take pos
                            (fun f ->
                              xml_return
                                (Non_loop_prf (a, b, c, d, e, f)))))))));;

let rec xml_rat
  x = xml_or (xml_change (xml_int "integer") (comp xml_return of_int))
        (xml_do "rational"
          (xml_take (xml_int "numerator")
            (fun a ->
              xml_take (xml_int "denominator")
                (fun b -> xml_return (divide_rata (of_int a) (of_int b))))))
        x;;

let rec eq_set (_A1, _A2, _A3, _A4) = set_eq (_A2, _A3, _A4);;

let rec choice
  = function [] -> Inl []
    | x :: xs ->
        catch_error x
          (fun e -> catch_error (choice xs) (fun xa -> Inl (e :: xa)));;

let rec firstM
  f x1 = match f, x1 with f, [] -> Inl []
    | f, x :: xs ->
        catch_error (bindb (f x) (fun _ -> Inr x))
          (fun e -> catch_error (firstM f xs) (fun xa -> Inl (e :: xa)));;

let rec gen_filter
  it1 emp2 ins2 p s1 =
    it1 s1 (fun _ -> true) (fun x s -> (if p x then ins2 x s else s)) emp2;;

let rec gt_term _A _C
  nlv ne gt_fun rm s t =
    ne && (let root1 = root s in
           let root2 = root t in
            (is_Var s ||
              (is_Var t ||
                (gt_fun None None ||
                  (gt_fun root1 root2 ||
                    (gt_fun root1 None || gt_fun None root2))))) &&
              (if nlv
                then (match root1 with None -> true
                       | Some fn ->
                         list_ex
                           (fun r ->
                             matchc _A _C (tcapRM _A _C nlv rm s) (fst r))
                           (rm fn))
                else true));;

let rec rd_impl _A
  gt x1 = match gt, x1 with
    gt, (Fun (f, ss), Fun (g, ts)) ->
      (if eq _A f g &&
            (equal_nata (size_list ss) (size_list ts) &&
              not (gt (Fun (f, ss)) (Fun (g, ts))))
        then maps (rd_impl _A gt) (zip ss ts) else [(Fun (f, ss), Fun (g, ts))])
    | uu, (Var v, t) -> [(Var v, t)]
    | uu, (s, Var v) -> [(s, Var v)];;

let rec nodes_lts_impl _D
  pi = remdups _D
         (map (comp source snd) (transitions_impl pi) @
           map (comp target snd) (transitions_impl pi));;

let rec rec_list f1 f2 x2 = match f1, f2, x2 with f1, f2, [] -> f1
                   | f1, f2, x21 :: x22 -> f2 x21 x22 (rec_list f1 f2 x22);;

let rec product_lists
  = function [] -> [[]]
    | xs :: xss ->
        maps (fun x -> map (fun a -> x :: a) (product_lists xss)) xs;;

let rec height (f, h) = h;;

let rec mat_of_cols
  n cs = mat n (size_list cs) (fun (i, j) -> vec_index (nth cs j) i);;

let rec vec_of_list_impl xa = Abs_vec_impl (size_list xa, IArray xa);;

let rec vec_of_list v = Vec_impl (vec_of_list_impl v);;

let rec showsl_literal s = showsl_lit s;;

let rec showsl_monom_list (_A1, _A2)
  = function
    [(x, p)] ->
      (if equal_nata p one_nata then showsl _A2 x
        else comp (comp (showsl _A2 x) (showsl_lit "^")) (showsl_nat p))
    | (x, p) :: v :: va ->
        comp (comp (if equal_nata p one_nata then showsl _A2 x
                     else comp (comp (showsl _A2 x) (showsl_literal "^"))
                            (showsl_nat p))
               (showsl_literal "*"))
          (showsl_monom_list (_A1, _A2) (v :: va))
    | [] -> showsl_literal "1";;

let rec showsl_monom (_A1, _A2)
  xa = showsl_monom_list (_A1, _A2) (rep_monom _A1 xa);;

let rec showsl_poly (_A1, _A2, _A3) (_B1, _B2, _B3)
  = function [] -> showsl_lit "0"
    | (m, c) :: p ->
        comp (if eq _B2 c (onea _B1) then showsl_monom (_A2, _A3) m
               else (if eq (equal_monom (_A1, _A2)) m (one_monom _A2)
                      then showsl _B3 c
                      else comp (comp (showsl _B3 c) (showsl_lit "*"))
                             (showsl_monom (_A2, _A3) m)))
          (if null p then id
            else comp (showsl_lit " + ")
                   (showsl_poly (_A1, _A2, _A3) (_B1, _B2, _B3) p));;

let rec check_poly_ge (_A1, _A2) _B
  x0 q = match x0, q with
    [], q ->
      list_all
        (fun (_, d) ->
          less_eq
            _B.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order
            d (zerob
                _B.semiring_0_ordered_semiring_0a.mult_zero_semiring_0.zero_mult_zero))
        q
    | (m, c) :: p, q ->
        (match extract (fun nd -> eq (equal_monom (_A1, _A2)) (fst nd) m) q
          with None ->
            less_eq
              _B.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order
              (zerob
                _B.semiring_0_ordered_semiring_0a.mult_zero_semiring_0.zero_mult_zero)
              c &&
              check_poly_ge (_A1, _A2) _B p q
          | Some (q1, ((_, d), q2)) ->
            less_eq
              _B.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order
              d c &&
              check_poly_ge (_A1, _A2) _B p (q1 @ q2));;

let rec check_poly_gt _A (_B1, _B2)
  gt p q =
    (let (a1, p1) =
       poly_split (_B1, _B2)
         _A.semiring_0_ordered_semiring_0a.mult_zero_semiring_0.zero_mult_zero
         (one_monom _B2) p
       in
     let (b1, q1) =
       poly_split (_B1, _B2)
         _A.semiring_0_ordered_semiring_0a.mult_zero_semiring_0.zero_mult_zero
         (one_monom _B2) q
       in
      gt a1 b1 && check_poly_ge (_B1, _B2) _A p1 q1);;

let rec monom_mult_list (_A1, _A2)
  x0 n = match x0, n with [], n -> n
    | (x, p) :: m, n ->
        (match n with [] -> (x, p) :: m
          | (y, q) :: na ->
            (if eq _A1 x y
              then (x, plus_nata p q) :: monom_mult_list (_A1, _A2) m na
              else (if less _A2.order_linorder.preorder_order.ord_preorder x y
                     then (x, p) :: monom_mult_list (_A1, _A2) m n
                     else (y, q) ::
                            monom_mult_list (_A1, _A2) ((x, p) :: m) na)));;

let rec times_monom (_A1, _A2)
  xb xc =
    Abs_monom
      (monom_mult_list (_A1, _A2) (rep_monom _A2 xb) (rep_monom _A2 xc));;

let rec monom_mult_poly (_A1, _A2) (_B1, _B2)
  uu x1 = match uu, x1 with uu, [] -> []
    | (ma, c), (m, d) :: p ->
        (if eq _B1 (times _B2.mult_zero_semiring_0.times_mult_zero c d)
              (zerob _B2.mult_zero_semiring_0.zero_mult_zero)
          then monom_mult_poly (_A1, _A2) (_B1, _B2) (ma, c) p
          else (times_monom (_A1, _A2) ma m,
                 times _B2.mult_zero_semiring_0.times_mult_zero c d) ::
                 monom_mult_poly (_A1, _A2) (_B1, _B2) (ma, c) p);;

let rec poly_add (_A1, _A2) (_B1, _B2)
  x0 q = match x0, q with [], q -> q
    | (m, c) :: p, q ->
        (match extract (fun mc -> eq (equal_monom (_A1, _A2)) (fst mc) m) q
          with None -> (m, c) :: poly_add (_A1, _A2) (_B1, _B2) p q
          | Some (q1, ((_, d), q2)) ->
            (if eq _B1
                  (plus _B2.comm_monoid_add_semiring_0.monoid_add_comm_monoid_add.semigroup_add_monoid_add.plus_semigroup_add
                    c d)
                  (zerob _B2.mult_zero_semiring_0.zero_mult_zero)
              then poly_add (_A1, _A2) (_B1, _B2) p (q1 @ q2)
              else (m, plus _B2.comm_monoid_add_semiring_0.monoid_add_comm_monoid_add.semigroup_add_monoid_add.plus_semigroup_add
                         c d) ::
                     poly_add (_A1, _A2) (_B1, _B2) p (q1 @ q2)));;

let rec poly_mult (_A1, _A2) (_B1, _B2)
  x0 q = match x0, q with [], q -> []
    | mc :: p, q ->
        poly_add (_A1, _A2) (_B1, _B2)
          (monom_mult_poly (_A1, _A2) (_B1, _B2) mc q)
          (poly_mult (_A1, _A2) (_B1, _B2) p q);;

let rec one_polya _A _B
  = [(one_monom _A,
       onea _B.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral)];;

let rec poly_power (_A1, _A2) (_B1, _B2)
  uu n =
    (if equal_nata n zero_nata then one_polya _A2 _B2.semiring_1_comm_semiring_1
      else poly_mult (_A1, _A2)
             (_B1, _B2.semiring_1_comm_semiring_1.semiring_0_semiring_1) uu
             (poly_power (_A1, _A2) (_B1, _B2) uu (minus_nata n one_nata)));;

let rec monom_list_subst (_B1, _B2) (_C1, _C2)
  sigma x1 = match sigma, x1 with
    sigma, [] -> one_polya _B2 _C2.semiring_1_comm_semiring_1
    | sigma, (x, p) :: m ->
        poly_mult (_B1, _B2)
          (_C1, _C2.semiring_1_comm_semiring_1.semiring_0_semiring_1)
          (poly_power (_B1, _B2) (_C1, _C2) (sigma x) p)
          (monom_list_subst (_B1, _B2) (_C1, _C2) sigma m);;

let rec monom_list _A x = rep_monom _A x;;

let rec monom_subst _A (_B1, _B2) (_C1, _C2)
  sigma m = monom_list_subst (_B1, _B2) (_C1, _C2) sigma (monom_list _A m);;

let zero_polya : ('a monom * 'b) list = [];;

let rec poly_subst _A (_B1, _B2) (_C1, _C2)
  sigma x1 = match sigma, x1 with sigma, [] -> zero_polya
    | sigma, (m, c) :: p ->
        poly_add (_B1, _B2)
          (_C1, _C2.semiring_1_comm_semiring_1.semiring_0_semiring_1)
          (poly_mult (_B1, _B2)
            (_C1, _C2.semiring_1_comm_semiring_1.semiring_0_semiring_1)
            [(one_monom _B2, c)] (monom_subst _A (_B1, _B2) (_C1, _C2) sigma m))
          (poly_subst _A (_B1, _B2) (_C1, _C2) sigma p);;

let rec var_monom _A xa = Abs_monom [(xa, one_nata)];;

let rec eval_term (_B1, _B2) (_C1, _C2)
  uu x1 = match uu, x1 with
    uu, Var x ->
      [(var_monom _C2 x,
         onea _B2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral)]
    | i, Fun (f, ts) ->
        (let ps = map (eval_term (_B1, _B2) (_C1, _C2) i) ts in
         let n = size_list ts in
          poly_subst linorder_nat (_C1, _C2)
            (_B1, _B2.comm_semiring_1_poly_carrier)
            (fun ia -> (if less_nat ia n then nth ps ia else zero_polya))
            (i (f, n)));;

let rec check_s (_A1, _A2, _A3) _B (_C1, _C2, _C3)
  gt i =
    (fun (s, t) ->
      (let p = eval_term (_A1, _A2) (_C1, _C2) i s in
       let q = eval_term (_A1, _A2) (_C1, _C2) i t in
        check (check_poly_gt
                _A2.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1
                (_C1, _C2) gt p q)
          (comp (comp (comp (comp (comp (comp
  (comp (showsl_literal "could not ensure ") (showsl_terma _B _C3 s))
  (showsl_literal " > "))
                                    (showsl_terma _B _C3 t))
                              (showsl_literal " since we\092could not ensure "))
                        (showsl_poly (_C1, _C2, _C3)
                          (_A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral,
                            _A1, _A3)
                          p))
                  (showsl_literal " > "))
            (showsl_poly (_C1, _C2, _C3)
              (_A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral,
                _A1, _A3)
              q))));;

let rec content _A
  p = fold_coeffs _A.gcd_semiring_gcd.zero_gcd (gcda _A.gcd_semiring_gcd) p
        (zerob _A.gcd_semiring_gcd.zero_gcd);;

let rec curry f a b = f (a, b);;

let rec scnp_af_to_af
  pia pi =
    (fun (f, n) ->
      (let is = map fst (pia (f, n)) in
        (if list_ex (less_eq_nat n) is
          then sup_seta (ceq_nat, ccompare_nat) (pi (f, n))
                 (set (ceq_nat, ccompare_nat, set_impl_nat) is)
          else set (ceq_nat, ccompare_nat, set_impl_nat) is)));;

let rec semilattice_set_apply (Abs_semilattice_set x) = x;;

let rec set_fold1 (_A1, _A2, _A3)
  f x1 = match f, x1 with
    f, RBT_set rbt ->
      (match ccompare _A2
        with None ->
          failwith "set_fold1 RBT_set: ccompare = None"
            (fun _ -> set_fold1 (_A1, _A2, _A3) f (RBT_set rbt))
        | Some _ ->
          (if is_emptya _A2 rbt
            then failwith "set_fold1 RBT_set: empty set"
                   (fun _ -> set_fold1 (_A1, _A2, _A3) f (RBT_set rbt))
            else fold1 _A2 (semilattice_set_apply f) rbt))
    | f, DList_set dxs ->
        (match ceq _A1
          with None ->
            failwith "set_fold1 DList_set: ceq = None"
              (fun _ -> set_fold1 (_A1, _A2, _A3) f (DList_set dxs))
          | Some _ ->
            (if nulla _A1 dxs
              then failwith "set_fold1 DList_set: empty set"
                     (fun _ -> set_fold1 (_A1, _A2, _A3) f (DList_set dxs))
              else foldc _A1 (semilattice_set_apply f) (tl _A1 dxs)
                     (hd _A1 dxs)))
    | f, Set_Monad (x :: xs) -> fold (semilattice_set_apply f) xs x
    | f, Collect_set p ->
        failwith "set_fold1: Collect_set"
          (fun _ -> set_fold1 (_A1, _A2, _A3) f (Collect_set p))
    | f, Complement a ->
        failwith "set_fold1: Complement"
          (fun _ -> set_fold1 (_A1, _A2, _A3) f (Complement a));;

let rec projr (Inr x2) = x2;;

let rec rec_term
  f1 f2 x2 = match f1, f2, x2 with f1, f2, Var x1 -> f1 x1
    | f1, f2, Fun (x21, x22) ->
        f2 x21 (map (fun term -> (term, rec_term f1 f2 term)) x22);;

let rec mk_subst_domain _A _B
  sigma =
    (let tau = mk_subst _A (fun a -> Var a) sigma in
      filtera (fun (x, t) -> not (equal_terma _B _A (Var x) t))
        (map (fun x -> (x, tau x)) (remdups _A (map fst sigma))));;

let rec subst_eq (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2)
  sigma tau =
    (let sigmaa = mk_subst_domain _A5 _B2 sigma in
     let taua = mk_subst_domain _A5 _B2 tau in
      set_eq
        ((cenum_prod _A1 cenum_term), (ceq_prod _A2 (ceq_term _B2 _A5)),
          (ccompare_prod _A3 (ccompare_term _B1 _A4)))
        (set ((ceq_prod _A2 (ceq_term _B2 _A5)),
               (ccompare_prod _A3 (ccompare_term _B1 _A4)),
               (set_impl_prod _A6 set_impl_term))
          sigmaa)
        (set ((ceq_prod _A2 (ceq_term _B2 _A5)),
               (ccompare_prod _A3 (ccompare_term _B1 _A4)),
               (set_impl_prod _A6 set_impl_term))
          taua));;

let rec hole_pos = function Hole -> []
                   | More (f, ss, d, ts) -> size_list ss :: hole_pos d;;

let rec full_af
  fn = set (ceq_nat, ccompare_nat, set_impl_nat) (upt zero_nata (snd fn));;

let rec add_funas_term
  x0 fs = match x0, fs with Var uu, fs -> fs
    | Fun (f, ts), fs -> (f, size_list ts) :: foldr add_funas_term ts fs;;

let rec add_funas_rule
  r fs = add_funas_term (fst r) (add_funas_term (snd r) fs);;

let rec funas_trs_list trs = foldr add_funas_rule trs [];;

let rec funs_rule_list r = add_funs_rule r [];;

let rec vars_rule_list r = add_vars_rule r [];;

let rec supteq_list = function Var x -> [Var x]
                      | Fun (f, ts) -> Fun (f, ts) :: maps supteq_list ts;;

let rec is_NF_main (_A1, _A2) (_B1, _B2, _B3)
  var_cond r_empty m =
    (if var_cond then (fun _ -> false)
      else (if r_empty then (fun _ -> true)
             else (fun t ->
                    list_all
                      (fun u ->
                        (if not (is_Var u)
                          then list_all
                                 (fun l ->
                                   not (matches _A2 _B2 (_B1, _B2, _B3) u l))
                                 (m (the (root u)))
                          else true))
                      (supteq_list t))));;

let rec is_NF_trs (_A1, _A2) (_B1, _B2, _B3)
  r = is_NF_main (_A1, _A2) (_B1, _B2, _B3)
        (list_ex (fun ra -> is_Var (fst ra)) r) (null r)
        (term_map _A1 (map fst r));;

let rec apply_args a t x2 = match a, t, x2 with a, t, [] -> t
                     | a, t, s :: ss -> apply_args a (Fun (a, [t; s])) ss;;

let rec get_symbol sm f n i = nth (sm f n) i;;

let rec strategy_of_string
  str = (if equal_lista equal_char str
              [Chara (false, true, true, false, false, false, true, false);
                Chara (true, false, true, false, true, false, true, false);
                Chara (false, false, true, true, false, false, true, false);
                Chara (false, false, true, true, false, false, true, false)]
          then xml_return (Inl No_Strategy)
          else (if equal_lista equal_char str
                     [Chara (true, false, false, true, false, false, true,
                              false);
                       Chara (false, true, true, true, false, false, true,
                               false);
                       Chara (false, true, true, true, false, false, true,
                               false);
                       Chara (true, false, true, false, false, false, true,
                               false);
                       Chara (false, true, false, false, true, false, true,
                               false);
                       Chara (true, false, true, true, false, false, true,
                               false);
                       Chara (true, true, true, true, false, false, true,
                               false);
                       Chara (true, true, false, false, true, false, true,
                               false);
                       Chara (false, false, true, false, true, false, true,
                               false)]
                 then xml_return (Inl Innermost)
                 else (if equal_lista equal_char str
                            [Chara (true, true, true, true, false, false, true,
                                     false);
                              Chara (true, false, true, false, true, false,
                                      true, false);
                              Chara (false, false, true, false, true, false,
                                      true, false);
                              Chara (true, false, true, false, false, false,
                                      true, false);
                              Chara (false, true, false, false, true, false,
                                      true, false);
                              Chara (true, false, true, true, false, false,
                                      true, false);
                              Chara (true, true, true, true, false, false, true,
                                      false);
                              Chara (true, true, false, false, true, false,
                                      true, false);
                              Chara (false, false, true, false, true, false,
                                      true, false)]
                        then xml_return (Inr Outermost)
                        else xml_error
                               (("unknown strategy \034" ^ implode str) ^
                                 "\034"))));;

let rec signature_to_AC _A
  asa cs x2 = match asa, cs, x2 with asa, cs, [] -> (asa, cs)
    | asa, cs, fnac :: siga ->
        (let (f, a) = fnac in
         let (_, aa) = a in
         let (ab, c) = aa in
          signature_to_AC _A (if ab then f :: asa else asa)
            (if c then f :: cs else cs) siga);;

let rec xml_take_attribute
  att p xs =
    (let (xmls, (atts, (allow, (cands, pos)))) = xs in
      (match pick_up (equal_list equal_char) [] (explode att) atts
        with None -> xml_error (("attribute \034" ^ att) ^ "\034 not found") xs
        | Some (v, rest) -> p v (xmls, (rest, (allow, (cands, pos))))));;

let rec signature _A
  xml2name =
    xml_do "signature"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "funcsym"
          (xml_take xml2name
            (fun name ->
              xml_take (xml_nat "arity")
                (fun arity ->
                  xml_take_optional (xml_text "theory")
                    (fun a ->
                      (match a
                        with None -> xml_return (name, (arity, (false, false)))
                        | Some th ->
                          (if equal_lista equal_char th
                                [Chara (true, false, false, false, false, false,
 true, false)]
                            then (if equal_nata arity
                                       (nat_of_integer
 (Big_int.big_int_of_int 2))
                                   then xml_return
  (name, (arity, (true, false)))
                                   else xml_error
  (((("wrong arity " ^ showsl_nat arity "") ^ " for associative symbol \034") ^
     showsl _A name "") ^
    "\034"))
                            else (if equal_lista equal_char th
                                       [Chara
  (true, false, false, false, false, false, true, false);
 Chara (true, true, false, false, false, false, true, false)]
                                   then (if equal_nata arity
      (nat_of_integer (Big_int.big_int_of_int 2))
  then xml_return (name, (arity, (true, true)))
  else xml_error
         (((("wrong arity " ^ showsl_nat arity "") ^ " for AC symbol \034") ^
            showsl _A name "") ^
           "\034"))
                                   else (if equal_lista equal_char th
      [Chara (true, true, false, false, false, false, true, false)]
  then (if equal_nata arity (nat_of_integer (Big_int.big_int_of_int 2))
         then xml_return (name, (arity, (false, true)))
         else xml_error
                (((("wrong arity " ^ showsl_nat arity "") ^
                    " for commutative symbol \034") ^
                   showsl _A name "") ^
                  "\034"))
  else xml_error (("unknown theory \034" ^ implode th) ^ "\034"))))))))))
        xml_return);;

let rec problem _A
  xml2name =
    xml_do "problem"
      (xml_take_attribute "type"
        (fun _ ->
          xml_take
            (xml_do "trs"
              (xml_take
                (xml_do "rules"
                  (xml_take_many_sub [] zero_nata Infinity_enat
                    (crule _A xml2name "rule")
                    (fun crules ->
                      xml_take_default []
                        (xml_do "relrules"
                          (xml_take_many_sub [] zero_nata Infinity_enat
                            (rule _A xml2name) xml_return))
                        (fun rel -> xml_return (crules, rel)))))
                (fun (crules, rel) ->
                  xml_take (signature _A xml2name)
                    (fun siga ->
                      (let a = signature_to_AC _A [] [] siga in
                       let (aa, c) = a in
                        xml_take_optional
                          (xml_do "comment"
                            (xml_take_many_sub [] zero_nata Infinity_enat
                              (fun x -> right (fst x))
                              (fun _ -> xml_return ())))
                          (fun _ ->
                            xml_take_optional
                              (xml_do "conditiontype"
                                (xml_take_many_sub [] zero_nata Infinity_enat
                                  (fun x -> right (fst x))
                                  (fun _ -> xml_return ())))
                              (fun _ ->
                                xml_return (crules, (rel, (aa, c))))))))))
            (fun (crules, a) ->
              (let (rel, aa) = a in
               let (ab, c) = aa in
                xml_take (xml_change (xml_text "strategy") strategy_of_string)
                  (fun strat ->
                    xml_take_default Full
                      (xml_do "startterm"
                        (xml_take
                          (xml_or
                            (xml_do "constructor-based"
                              (xml_return Constructor_Based))
                            (xml_do "full" (xml_return Full)))
                          xml_return))
                      (fun start ->
                        xml_take_optional
                          (xml_do "status"
                            (xml_take_many_sub [] zero_nata Infinity_enat
                              (fun x -> right (fst x))
                              (fun _ -> xml_return ())))
                          (fun _ ->
                            xml_take_optional
                              (xml_do "metainformation"
                                (xml_take_many_sub [] zero_nata Infinity_enat
                                  (fun x -> right (fst x))
                                  (fun _ -> xml_return ())))
                              (fun _ ->
                                (let relative = not (null rel) in
                                 let (conditional, r) =
                                   (if list_all (fun (_, ac) -> null ac) crules
                                     then (false, map fst crules)
                                     else (true, []))
                                   in
                                 let equational = not (null ab) || not (null c)
                                   in
                                  (match
                                    (conditional,
                                      (equational, (relative, (strat, start))))
                                    with (true, (true, _)) ->
                                      xml_error "equational conditional system"
                                    | (true, (false, (true, _))) ->
                                      xml_error "relative conditional system"
                                    | (true,
(false, (false, (Inl No_Strategy, Full))))
                                      -> xml_return (CTRS_input crules)
                                    | (true,
(false, (false, (Inl No_Strategy, Constructor_Based))))
                                      -> xml_error "unsupported combination"
                                    | (true,
(false, (false, (Inl Innermost, _))))
                                      -> xml_error "unsupported combination"
                                    | (true,
(false, (false, (Inl (Innermost_Q _), _))))
                                      -> xml_error "unsupported combination"
                                    | (true, (false, (false, (Inr _, _)))) ->
                                      xml_error "unsupported combination"
                                    | (false, (true, (true, _))) ->
                                      xml_error "relative equational system"
                                    | (false,
(true, (false, (Inl No_Strategy, Full))))
                                      -> xml_return (AC_input (r, ab, c))
                                    | (false,
(true, (false, (Inl No_Strategy, Constructor_Based))))
                                      -> xml_error "unsupported combination"
                                    | (false,
(true, (false, (Inl Innermost, _))))
                                      -> xml_error "unsupported combination"
                                    | (false,
(true, (false, (Inl (Innermost_Q _), _))))
                                      -> xml_error "unsupported combination"
                                    | (false, (true, (false, (Inr _, _)))) ->
                                      xml_error "unsupported combination"
                                    | (false, (false, (true, (Inl istrat, _))))
                                      -> xml_return
   (Inn_TRS_input (istrat, r, rel, start))
                                    | (false, (false, (true, (Inr _, _)))) ->
                                      xml_error "unsupported combination"
                                    | (false, (false, (false, (Inl istrat, _))))
                                      -> xml_return
   (Inn_TRS_input (istrat, r, rel, start))
                                    | (false,
(false, (false, (Inr fpstrat, Full))))
                                      -> xml_return (FP_TRS_input (fpstrat, r))
                                    | (false,
(false, (false, (Inr _, Constructor_Based))))
                                      -> xml_error
   "unsupported combination"))))))))));;

let rec symbols _A
  xml2name tagname =
    xml_do tagname
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "symbol"
          (xml_take xml2name
            (fun a -> xml_take (xml_nat "arity") (fun b -> xml_return (a, b)))))
        xml_return);;

let rec map_xml_text
  f x1 = match f, x1 with
    f, XML (t, asa, cs) -> XML (t, asa, map (map_xml_text f) cs)
    | f, XML_text txt -> XML_text (f txt);;

let rec orig_term
  m x1 = match m, x1 with m, Var x -> Var x
    | m, Fun (f, []) -> (match m f with None -> Fun (f, []) | Some a -> Var a)
    | m, Fun (f, v :: va) -> Fun (f, map (orig_term m) (v :: va));;

let rec prec_exta
  prwm =
    (fun f g ->
      (match prwm f with None -> false
        | Some pf ->
          (match prwm g with None -> true
            | Some pg -> less_nat (fst pg) (fst pf))));;

let rec rep_afs (Abs_afs x) = x;;

let rec afsa xa = fst (rep_afs xa);;

let rec sum_lpoly c x1 ys = match c, x1, ys with c, [], ys -> ys
                    | c, v :: va, [] -> v :: va
                    | c, x :: xs, y :: ys -> add c x y :: sum_lpoly c xs ys;;

let rec lpoly_of _A
  c x1 = match c, x1 with c, PNum i -> return (i, [])
    | c, PVar x -> return (zero c, replicate x (zero c) @ [one c])
    | c, PSum [] -> return (zero c, [])
    | c, PSum (p :: ps) ->
        binda (lpoly_of _A c p)
          (fun (cp, ncp) ->
            binda (lpoly_of _A c (PSum ps))
              (fun (cq, ncq) -> return (add c cp cq, sum_lpoly c ncp ncq)))
    | c, PMult [] -> return (one c, [])
    | c, PMult (p :: ps) ->
        binda (lpoly_of _A c p)
          (fun (cp, ncp) ->
            binda (lpoly_of _A c (PMult ps))
              (fun (cq, ncq) ->
                (if list_all (eq _A (zero c)) ncp
                  then return (mult c cp cq, map (mult c cp) ncq)
                  else (if list_all (eq _A (zero c)) ncq
                         then return
                                (mult c cp cq, map (fun x -> mult c x cq) ncp)
                         else error "cannot transform non-linear polynomial to linear polynomial"))));;

let rec renaming _A
  xml2name =
    xml_do "renaming"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "renamingEntry"
          (xml_take xml2name
            (fun a -> xml_take xml2name (fun b -> xml_return (a, b)))))
        xml_return);;

let rec xml_real
  x = xml_or (xml_change xml_rat (comp xml_return ratreal))
        (xml_do "algebraic"
          (xml_take xml_rat
            (fun a ->
              xml_take xml_rat
                (fun b ->
                  xml_take xml_rat
                    (fun c ->
                      xml_return
                        (plus_reala (ratreal a)
                          (times_reala (ratreal b) (sqrt (ratreal c)))))))))
        x;;

let rec subset (_A1, _A2, _A3, _A4) = subset_eq (_A2, _A3, _A4);;

let rec char_poly_matrix (_A1, _A2, _A3)
  a = plus_mat
        (plus_poly
          (_A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.comm_monoid_add_semiring_0,
            _A1))
        (smult_mat
          (times_poly
            (_A1, _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.comm_semiring_0_comm_semiring_1,
              _A3))
          (pCons
            (_A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero,
              _A1)
            (zerob
              _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
            (pCons
              (_A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero,
                _A1)
              (onea _A2.ring_1_comm_ring_1.neg_numeral_ring_1.numeral_neg_numeral.one_numeral)
              (zero_polyb
                _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)))
          (one_mat
            ((one_poly
               _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel),
              (zero_poly
                _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero))
            (dim_row a)))
        (map_mat
          (fun aa ->
            pCons (_A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero,
                    _A1)
              (uminus
                _A2.ring_1_comm_ring_1.neg_numeral_ring_1.group_add_neg_numeral.uminus_group_add
                aa)
              (zero_polyb
                _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero))
          a);;

let rec char_poly (_A1, _A2)
  a = det ((equal_poly
             (_A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero,
               _A1)),
            (idom_divide_poly (_A1, _A2)))
        (char_poly_matrix
          (_A1, _A2.idom_idom_divide.comm_ring_1_idom,
            _A2.idom_idom_divide.semidom_idom.semiring_1_no_zero_divisors_semidom.semiring_no_zero_divisors_semiring_1_no_zero_divisors)
          a);;

let rec existsM
  f x1 = match f, x1 with f, [] -> Inl []
    | f, x :: xs ->
        catch_error (f x)
          (fun e -> catch_error (existsM f xs) (fun xa -> Inl (e :: xa)));;

let rec is_neg_atom = function NegAtom uu -> true
                      | Atom v -> false
                      | Conjunction v -> false
                      | Disjunction v -> false;;

let rec showsl_hint _A
  = function Default -> showsl_lit " auto"
    | Base raw -> comp (showsl_lit " ") (showsl _A raw)
    | Distribute (i, hints) ->
        comp (comp (showsl_lit " Distribute ") (showsl_nat i))
          (showsl_list_gen id " []" " [" ", " "]" (map (showsl_hint _A) hints))
    | LexStrict hints ->
        comp (showsl_lit " LexStrict")
          (showsl_list_gen id " []" " [" ", " "]" (map (showsl_hint _A) hints))
    | LexWeak hints ->
        comp (showsl_lit " LexWeak")
          (showsl_list_gen id " []" " [" ", " "]" (map (showsl_hint _A) hints))
    | Erase (i, hint) ->
        comp (comp (comp (comp (showsl_lit " Erase ") (showsl_nat i))
                     (showsl_lit " ["))
               (showsl_hint _A hint))
          (showsl_lit " ]");;

let rec gen_isEmpty ball m = ball m (fun _ -> false);;

let rec gen_isEmptya ball s = ball s (fun _ -> false);;

let rec nonreach _A
  gt s t =
    (match (s, t) with (Var _, _) -> false | (Fun (_, _), Var _) -> false
      | (Fun (f, ss), Fun (g, ts)) ->
        not (equal_proda _A equal_nat (f, size_list ss) (g, size_list ts)) &&
          not (gt s t));;

let rec sum_mset _A
  (Bag ms) =
    foldd (fun a n ->
            funpow n
              (plus _A.monoid_add_comm_monoid_add.semigroup_add_monoid_add.plus_semigroup_add
                a))
      (zerob _A.monoid_add_comm_monoid_add.zero_monoid_add) ms;;

let rec vars_term_ms _B
  = function Var x -> add_mset _B x zero_multiseta
    | Fun (f, ts) ->
        sum_mset (comm_monoid_add_multiset _B)
          (mset (equal_multiset _B) (map (vars_term_ms _B) ts));;

let rec weight
  w w0 scf x3 = match w, w0, scf, x3 with
    w, w0, scf, Fun (f, ts) ->
      (let n = size_list ts in
       let scff = scf (f, n) in
        plus_nata (w (f, n))
          (sum_list monoid_add_nat
            (map (fun (ti, i) -> times_nata (weight w w0 scf ti) (scff i))
              (zip ts (upt zero_nata n)))))
    | w, w0, scf, Var x -> w0;;

let rec kbo_impl _B
  w w0 prc least scf s t =
    (let wt = weight w w0 scf t in
     let ws = weight w w0 scf s in
      (if subseteq_mset _B (vars_term_ms _B (scf_term scf t))
            (vars_term_ms _B (scf_term scf s)) &&
            less_eq_nat wt ws
        then (if less_nat wt ws then (true, true)
               else (match s
                      with Var _ ->
                        (false,
                          (match t with Var _ -> true
                            | Fun (g, ts) -> null ts && least g))
                      | Fun (f, ss) ->
                        (match t with Var _ -> (true, true)
                          | Fun (g, ts) ->
                            (let p = prc (f, size_list ss) (g, size_list ts) in
                              (if fst p then (true, true)
                                else (if snd p
                                       then lex_ext_unbounded
      (kbo_impl _B w w0 prc least scf) ss ts
                                       else (false, false)))))))
        else (false, false)));;

let rec kbo_strict _A (_B1, _B2)
  pr w w0 least scf =
    (fun (s, t) ->
      check (fst (kbo_impl _B1 w w0 pr least scf s t))
        (comp (comp (comp (comp (showsl_literal "could not orient ")
                            (showsl_terma _A _B2 s))
                      (showsl_literal " >KBO "))
                (showsl_terma _A _B2 t))
          (showsl_literal "\092")));;

let rec trans_id trans_id_parser = trans_id_parser "transitionId";;

let rec list_all2
  p xs ys = match p, xs, ys with
    p, x :: xs, y :: ys -> p x y && list_all2 p xs ys
    | p, xs, [] -> null xs
    | p, [], ys -> null ys;;

let rec elements_mat (_A1, _A2, _A3)
  a = set (_A1, _A2, _A3)
        (maps (fun i ->
                map (fun j -> index_mat a (i, j)) (upt zero_nata (dim_col a)))
          (upt zero_nata (dim_row a)));;

let rec single_alist_entry xb xc = Alist [(xb, xc)];;

let rec image_mset _B
  f (Bag ms) =
    foldd (fun a n -> plus_multiseta _B (Bag (single_alist_entry (f a) n)))
      zero_multiseta ms;;

let rec eval i alpha x2 = match i, alpha, x2 with i, alpha, Var x -> alpha x
               | i, alpha, Fun (f, ts) -> i f (map (eval i alpha) ts);;

let rec monom_vars_list _A xa = map fst (rep_monom _A xa);;

let rec poly_vars_list (_A1, _A2)
  p = remdups _A1 (maps (comp (monom_vars_list _A2) fst) p);;

let rec concat_lists
  = function [] -> [[]]
    | asa :: xs ->
        maps (fun vec -> map (fun a -> a :: vec) asa) (concat_lists xs);;

let rec poly_of (_A1, _A2) (_B1, _B2)
  = function
    PNum i ->
      (if eq _B1 i
            (zerob
              _B2.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
        then [] else [(one_monom _A2, i)])
    | PVar x ->
        [(var_monom _A2 x,
           onea _B2.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral)]
    | PSum [] -> zero_polya
    | PSum (p :: ps) ->
        poly_add (_A1, _A2)
          (_B1, _B2.semiring_1_comm_semiring_1.semiring_0_semiring_1)
          (poly_of (_A1, _A2) (_B1, _B2) p)
          (poly_of (_A1, _A2) (_B1, _B2) (PSum ps))
    | PMult [] -> one_polya _A2 _B2.semiring_1_comm_semiring_1
    | PMult (p :: ps) ->
        poly_mult (_A1, _A2)
          (_B1, _B2.semiring_1_comm_semiring_1.semiring_0_semiring_1)
          (poly_of (_A1, _A2) (_B1, _B2) p)
          (poly_of (_A1, _A2) (_B1, _B2) (PMult ps));;

let rec square_possibilities (_A1, _A2) (_B1, _B2)
  sqrt p =
    (let roots =
       map (fun x ->
             map (fun a -> (x, a))
               (sqrt (fst (poly_split (_B1, _B2)
                            _A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
                            (times_monom (_B1, _B2) (var_monom _B2 x)
                              (var_monom _B2 x))
                            p))))
         (poly_vars_list (_B1, _B2) p)
       in
     let choices =
       (if membera (equal_list (equal_prod _B1 _A1)) roots [] then []
         else concat_lists roots)
       in
     let polys =
       map (fun xas ->
             poly_of (_B1, _B2) (_A1, _A2.comm_semiring_1_poly_carrier)
               (PSum (map (fun (x, a) -> PMult [PVar x; PNum a]) xas)))
         choices
       in
      polys);;

let rec check_poly_eq (_A1, _A2) (_B1, _B2)
  x0 q = match x0, q with [], q -> null q
    | (m, c) :: p, q ->
        (match extract (fun nd -> eq (equal_monom (_A1, _A2)) (fst nd) m) q
          with None -> false
          | Some (q1, ((_, d), q2)) ->
            eq _B1 c d && check_poly_eq (_A1, _A2) (_B1, _B2) p (q1 @ q2));;

let rec monom_list_degree xps = sum_list monoid_add_nat (map snd xps);;

let rec monom_degree _A xa = monom_list_degree (rep_monom _A xa);;

let rec poly_degree _A p = max_list (map (fun (m, _) -> monom_degree _A m) p);;

let rec check_quadratic (_A1, _A2)
  sqrt p =
    bindb (check
            (equal_nata (poly_degree linorder_nat p)
              (nat_of_integer (Big_int.big_int_of_int 2)))
            (showsl_literal "not quadratic"))
      (fun _ ->
        (let polys =
           square_possibilities (_A1, _A2) (equal_nat, linorder_nat) sqrt p in
          check (list_ex
                  (fun q ->
                    check_poly_eq (equal_nat, linorder_nat)
                      (_A1, _A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_0_semiring_1)
                      (poly_mult (equal_nat, linorder_nat)
                        (_A1, _A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_0_semiring_1)
                        q q)
                      p)
                  polys)
            (showsl_literal "could not find quadratic polynomial")));;

let rec check_quadratic_ge_const (_A1, _A2) (_C1, _C2)
  sq i st =
    (let (s, t) = st in
      bindb (check (not (is_Var s))
              (showsl_literal "require non-variables as arguments"))
        (fun _ ->
          (let pt = eval_term (_A1, _A2) (_C1, _C2) i t in
           let (c, p0) =
             poly_split (_C1, _C2)
               _A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
               (one_monom _C2) pt
             in
            bindb (check
                    (equal_lista (equal_prod (equal_monom (_C1, _C2)) _A1) p0
                      zero_polya)
                    (showsl_literal "rhs must evaluate to constant"))
              (fun _ ->
                (let ps = i (the (root s)) in
                 let (d, psx) =
                   poly_split (equal_nat, linorder_nat)
                     _A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
                     (one_monom linorder_nat) ps
                   in
                  bindb (check
                          (less_eq
                            _A2.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order
                            c d)
                          (showsl_literal "problem in comparing constants"))
                    (fun _ -> check_quadratic (_A1, _A2) sq psx))))));;

let rec check_ns _A (_B1, _B2, _B3) (_C1, _C2, _C3)
  i = (fun (s, t) ->
        (let p = eval_term (_B1, _B2) (_C1, _C2) i s in
         let q = eval_term (_B1, _B2) (_C1, _C2) i t in
          check (check_poly_ge (_C1, _C2)
                  _B2.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1
                  p q)
            (comp (comp (comp (comp (comp (comp
    (comp (showsl_literal "could not ensure ") (showsl_terma _A _C3 s))
    (showsl_literal " >= "))
                                      (showsl_terma _A _C3 t))
                                (showsl_literal
                                  " since we\092could not ensure "))
                          (showsl_poly (_C1, _C2, _C3)
                            (_B2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral,
                              _B1, _B3)
                            p))
                    (showsl_literal " >= "))
              (showsl_poly (_C1, _C2, _C3)
                (_B2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral,
                  _B1, _B3)
                q))));;

let rec check_cc (_A1, _A2, _A3) _B (_C1, _C2, _C3)
  sq gt i x3 = match sq, gt, i, x3 with
    sq, gt, i, Unconditional_C (false, st) ->
      (if isOK (check_quadratic_ge_const (_A1, _A2) (_C1, _C2) sq i st)
        then Inr () else check_ns _B (_A1, _A2, _A3) (_C1, _C2, _C3) i st)
    | sq, gt, i, Unconditional_C (true, st) ->
        check_s (_A1, _A2, _A3) _B (_C1, _C2, _C3) gt i st
    | sq, gt, i, Conditional_C (true, (u, v), (s, t)) ->
        (let ss = eval_term (_A1, _A2) (_C1, _C2) i s in
         let tt = eval_term (_A1, _A2) (_C1, _C2) i t in
         let uu = eval_term (_A1, _A2) (_C1, _C2) i u in
         let vv = eval_term (_A1, _A2) (_C1, _C2) i v in
          (if check_poly_gt
                _A2.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1
                (_C1, _C2) gt ss tt
            then Inr ()
            else check (check_poly_ge (_C1, _C2)
                         _A2.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1
                         (poly_add (_C1, _C2)
                           (_A1, _A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_0_semiring_1)
                           ss vv)
                         (poly_add (_C1, _C2)
                           (_A1, _A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_0_semiring_1)
                           tt uu))
                   (comp (comp (comp (comp (comp
     (comp (comp (showsl_literal "could not ensure ") (showsl_terma _B _C3 u))
       (showsl_literal " > "))
     (showsl_terma _B _C3 v))
                                       (showsl_literal " ==> "))
                                 (showsl_terma _B _C3 s))
                           (showsl_literal " > "))
                     (showsl_terma _B _C3 t))))
    | sq, gt, i, Conditional_C (false, (u, v), (s, t)) ->
        (if isOK (check_quadratic_ge_const (_A1, _A2) (_C1, _C2) sq i (s, t))
          then Inr ()
          else (let ss = eval_term (_A1, _A2) (_C1, _C2) i s in
                let tt = eval_term (_A1, _A2) (_C1, _C2) i t in
                let uu = eval_term (_A1, _A2) (_C1, _C2) i u in
                let vv = eval_term (_A1, _A2) (_C1, _C2) i v in
                 (if check_poly_ge (_C1, _C2)
                       _A2.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1
                       ss tt
                   then Inr ()
                   else check (check_poly_ge (_C1, _C2)
                                _A2.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1
                                (poly_add (_C1, _C2)
                                  (_A1, _A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_0_semiring_1)
                                  ss vv)
                                (poly_add (_C1, _C2)
                                  (_A1, _A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_0_semiring_1)
                                  tt uu))
                          (comp (comp (comp
(comp (comp (comp (comp (showsl_literal "could not ensure ")
                    (showsl_terma _B _C3 u))
              (showsl_literal " >= "))
        (showsl_terma _B _C3 v))
  (showsl_literal " ==> "))
(showsl_terma _B _C3 s))
                                  (showsl_literal " >= "))
                            (showsl_terma _B _C3 t)))));;

let rec map_poly _B (_A1, _A2)
  f p = Poly (strip_while (eq _A2 (zerob _A1)) (map f (coeffs _B p)));;

let rec is_left_of
  x0 q = match x0, q with [], q -> false
    | i :: p, q ->
        (match q with [] -> false
          | j :: qa ->
            (if less_nat i j then true
              else (if less_nat j i then false else is_left_of p qa)));;

let rec showsl_rpo_repr _A
  prs = comp (comp (comp (comp (showsl_literal
                                 "RPO with the following precedence\092")
                           (foldr
                             (fun (a, b) ->
                               (let (f, n) = a in
                                 (fun (pr, _) ->
                                   comp (comp
  (comp (comp (comp (comp (showsl_literal "precedence(") (showsl _A f))
                (showsl_literal "["))
          (showsl_nat n))
    (showsl_literal "]) = "))
  (showsl_nat pr))
                                     (showsl_literal "\092")))
                                 b)
                             prs))
                     (showsl_literal
                       "\092precedence(_) = 0\092and the following status\092"))
               (foldr
                 (fun (a, b) ->
                   (let (f, n) = a in
                     (fun (_, s) ->
                       comp (comp (comp (comp
  (comp (comp (showsl_literal "status(") (showsl _A f)) (showsl_literal "["))
  (showsl_nat n))
                                    (showsl_literal "]) = "))
                              (showsl_literal
                                (match s with Lex -> "lex" | Mul -> "mul")))
                         (showsl_literal "\092")))
                     b)
                 prs))
          (showsl_literal "\092status(_) = lex\092");;

let rec rules_with _B
  p m = map_filter (fun x -> (if comp p fst x then Some (snd x) else None))
          (values (linorder_prod _B linorder_nat) m);;

let rec scnp_desc _A
  af mu =
    comp (comp (comp (comp (showsl_lit "SCNP-version with mu = ") mu)
                 (showsl_lit " and the level mapping defined by\092"))
           (showsl_sep
             (fun (a, b) ->
               (let (f, n) = a in
                 (fun asa ->
                   comp (comp (comp (showsl_lit "pi(") (showsl _A f))
                          (showsl_lit ") = "))
                     (default_showsl_list
                       (fun (p, l) ->
                         comp (comp (comp (comp (showsl_lit "(")
    (if less_nat p n then showsl_nat (suc p) else showsl_lit "epsilon"))
                                      (showsl_lit ","))
                                (showsl_nat l))
                           (showsl_lit ")"))
                       asa)))
                 b)
             (showsl_literal "\092") af))
      (showsl_literal "\092");;

let rec split_rulesb
  (Tp_ops_ext
    (qreltrs, q, r, rw, rules, q_empty, is_QNF, nFQ_subset_NF_rules, rules_map,
      delete_R_Rw, split_rules, mk, nfs, more))
    = split_rules;;

let rec delete_R_Rwb
  (Tp_ops_ext
    (qreltrs, q, r, rw, rules, q_empty, is_QNF, nFQ_subset_NF_rules, rules_map,
      delete_R_Rw, split_rules, mk, nfs, more))
    = delete_R_Rw;;

let rec nfsb
  (Tp_ops_ext
    (qreltrs, q, r, rw, rules, q_empty, is_QNF, nFQ_subset_NF_rules, rules_map,
      delete_R_Rw, split_rules, mk, nfs, more))
    = nfs;;

let rec mkc
  (Tp_ops_ext
    (qreltrs, q, r, rw, rules, q_empty, is_QNF, nFQ_subset_NF_rules, rules_map,
      delete_R_Rw, split_rules, mk, nfs, more))
    = mk;;

let rec qb
  (Tp_ops_ext
    (qreltrs, q, r, rw, rules, q_empty, is_QNF, nFQ_subset_NF_rules, rules_map,
      delete_R_Rw, split_rules, mk, nfs, more))
    = q;;

let rec split_tt _B _C
  i tp r_remove = (let (r, rw) = split_rulesb i tp r_remove in
                   let tp1 = mkc i (nfsb i tp) (qb i tp) r rw in
                   let a = delete_R_Rwb i tp r_remove r_remove in
                    (tp1, a));;

let rec mod_poly_one_main_list (_A1, _A2)
  r d n =
    (if equal_nata n zero_nata then r
      else (let a = hda r in
            let rr =
              tla (if eq _A1 a
                        (zerob
                          _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
                    then r
                    else minus_poly_rev_list
                           _A2.ring_1_comm_ring_1.neg_numeral_ring_1.group_add_neg_numeral
                           r (map (times
                                    _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd
                                    a)
                               d))
              in
             mod_poly_one_main_list (_A1, _A2) rr d (minus_nata n one_nata)));;

let rec modulo_poly (_A1, _A2)
  f g = (let cg =
           coeffs
             _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
             g
           in
          (if null cg then f
            else (let cf =
                    coeffs
                      _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
                      f
                    in
                  let ilc =
                    inverse _A1.division_ring_field.inverse_division_ring
                      (last cg)
                    in
                  let ch =
                    map (times
                          _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd
                          ilc)
                      cg
                    in
                  let r =
                    mod_poly_one_main_list
                      (_A2, _A1.idom_divide_field.idom_idom_divide.comm_ring_1_idom)
                      (rev cf) (rev ch)
                      (minus_nata (plus_nata one_nata (size_list cf))
                        (size_list cg))
                    in
                   poly_of_list
                     (_A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.comm_monoid_add_semiring_0,
                       _A2)
                     (rev r))));;

let rec sturm_aux
  p q = (if equal_nata (degreea zero_real q) zero_nata then [p; q]
          else p :: sturm_aux q
                      (uminus_polya ab_group_add_real
                        (modulo_poly (field_real, equal_real) p q)));;

let rec sturm
  p = sturm_aux p
        (pderiv
          (equal_real, comm_semiring_1_real, semiring_no_zero_divisors_real)
          p);;

let rec supt_impl _A _B
  x0 t = match x0, t with Var x, t -> false
    | Fun (f, ss), t ->
        membera (equal_term _A _B) ss t ||
          list_ex (fun s -> supt_impl _A _B s t) ss;;

let rec supt_list = function Var x -> []
                    | Fun (f, ts) -> maps supteq_list ts;;

let rec af_inter pi mu f = inf_seta (ceq_nat, ccompare_nat) (pi f) (mu f);;

let rec empty_af
  fn = set_empty (ceq_nat, ccompare_nat) (of_phantom set_impl_nata);;

let rec r_sym (TA_rule (f, qs, q)) = (f, size_list qs);;

let rec showsl_rulea
  funa var arr (l, r) =
    comp (comp (showsl_term funa var l) (showsl_literal arr))
      (showsl_term funa var r);;

let rec showsl_rulesa
  funa var arr trs =
    comp (showsl_list_gen (showsl_rulea funa var arr) "" "" "\092" "" trs)
      (showsl_literal "\092");;

let rec showsl_trsa
  funa var name arr r =
    comp (comp (showsl_literal name) (showsl_literal "\092\092"))
      (showsl_rulesa funa var arr r);;

let rec showsl_trs _A _B
  = showsl_trsa (showsl _A) (showsl _B) "rewrite system:" " -> ";;

let rec aarity_term _A
  a sm t =
    (match unapp _A a t with (Var _, _) -> None
      | (Fun (f, ss), ts) ->
        Some (minus_nata (aarity sm f (size_list ss)) (size_list ts)));;

let rec map_funs_term_wa
  fg x1 = match fg, x1 with fg, Var x -> Var x
    | fg, Fun (f, ts) ->
        Fun (fg (f, size_list ts), map (map_funs_term_wa fg) ts);;

let rec uncurry_top _A
  a n sm x3 = match a, n, sm, x3 with
    a, n, sm, Fun (f, ts) ->
      (let mt =
         map (map_funs_term_wa (fun (fa, na) -> get_symbol sm fa na zero_nata))
         in
       let t = hda ts in
        (if eq _A f a &&
              (equal_nata (size_list ts) n &&
                (not (is_Var t) &&
                  (let (h, m) = the (root t) in
                    not (equal_nata (aarity sm h m) zero_nata))))
          then (let Fun (g, ss) = t in
                 Fun (get_symbol sm g (size_list ss) one_nata,
                       mt (ss @ tla ts)))
          else Fun ((let (fa, na) = (f, size_list ts) in
                      get_symbol sm fa na zero_nata),
                     mt ts)))
    | a, n, sm, Var x -> Var x;;

let rec innermostLhss _A
  xml2name =
    xml_do "innermostLhss"
      (xml_take_many_sub [] zero_nata Infinity_enat (term _A xml2name)
        xml_return);;

let rec strategy _A
  xml2name =
    xml_do "strategy"
      (xml_take
        (xml_or (xml_leaf "innermost" Innermost)
          (xml_change (innermostLhss _A xml2name)
            (comp xml_return (fun a -> Innermost_Q a))))
        xml_return);;

let rec extract_special
  acc x1 = match acc, x1 with acc, [] -> None
    | acc, x :: xs ->
        (if equal_chara x
              (Chara (true, true, false, true, true, true, false, false))
          then map_option (fun s -> (s, xs)) (special_map (rev acc))
          else extract_special (x :: acc) xs);;

let rec normalize_special
  = function [] -> []
    | x :: xs ->
        (if equal_chara x
              (Chara (false, true, true, false, false, true, false, false))
          then (match extract_special [] xs
                 with None ->
                   [Chara (false, true, true, false, false, true, false,
                            false)] @
                     normalize_special xs
                 | Some (spec, ys) -> spec @ normalize_special ys)
          else x :: normalize_special xs);;

let rec parse_xmlfile
  p str =
    (match doc_of_string str with Inl a -> left a
      | Inr (XMLDOC (_, xml)) ->
        parse_xml p (map_xml_text normalize_special xml));;

let rec aABin _A
  f x1 u = match f, x1, u with
    f, AFun (g, [s; t]), u ->
      (if eq _A f g then AFun (f, [s; aABin _A f t u])
        else AFun (f, [AFun (g, [s; t]); u]))
    | f, AVar v, t -> AFun (f, [AVar v; t])
    | f, AFun (v, []), t -> AFun (f, [AFun (v, []); t])
    | f, AFun (v, [vb]), t -> AFun (f, [AFun (v, [vb]); t])
    | f, AFun (v, vb :: vd :: vf :: vg), t ->
        AFun (f, [AFun (v, vb :: vd :: vf :: vg); t])
    | f, AAC (v, va), t -> AFun (f, [AAC (v, va); t]);;

let rec actop _A _B
  f x1 = match f, x1 with
    f, Fun (g, [s; t]) ->
      (if eq _A f g
        then plus_multiseta (equal_term _A _B) (actop _A _B f s)
               (actop _A _B f t)
        else add_mset (equal_term _A _B) (Fun (g, [s; t])) zero_multiseta)
    | f, Var v -> add_mset (equal_term _A _B) (Var v) zero_multiseta
    | f, Fun (v, []) -> add_mset (equal_term _A _B) (Fun (v, [])) zero_multiseta
    | f, Fun (v, [vb]) ->
        add_mset (equal_term _A _B) (Fun (v, [vb])) zero_multiseta
    | f, Fun (v, vb :: vd :: vf :: vg) ->
        add_mset (equal_term _A _B) (Fun (v, vb :: vd :: vf :: vg))
          zero_multiseta;;

let rec aocnf (_A1, _A2, _A3) _B
  f_A f_C x2 = match f_A, f_C, x2 with
    f_A, f_C, Fun (f, s :: t :: u :: us) ->
      AFun (f, aocnf (_A1, _A2, _A3) _B f_A f_C s ::
                 aocnf (_A1, _A2, _A3) _B f_A f_C t ::
                   aocnf (_A1, _A2, _A3) _B f_A f_C u ::
                     map (aocnf (_A1, _A2, _A3) _B f_A f_C) us)
    | f_A, f_C, Fun (f, [t]) -> AFun (f, [aocnf (_A1, _A2, _A3) _B f_A f_C t])
    | f_A, f_C, Fun (f, []) -> AFun (f, [])
    | f_A, f_C, Fun (f, [s; t]) ->
        (let a = member (_A1, _A2) f f_A in
         let c = member (_A1, _A2) f f_C in
          (if a && c
            then AAC (f, image_mset (equal_acterm _A3 _B)
                           (aocnf (_A1, _A2, _A3) _B f_A f_C)
                           (actop _A3 _B f (Fun (f, [s; t]))))
            else (if a then aABin _A3 f (aocnf (_A1, _A2, _A3) _B f_A f_C s)
                              (aocnf (_A1, _A2, _A3) _B f_A f_C t)
                   else (if c then AAC (f,
 add_mset (equal_acterm _A3 _B) (aocnf (_A1, _A2, _A3) _B f_A f_C s)
   (add_mset (equal_acterm _A3 _B) (aocnf (_A1, _A2, _A3) _B f_A f_C t)
     zero_multiseta))
                          else AFun (f, [aocnf (_A1, _A2, _A3) _B f_A f_C s;
  aocnf (_A1, _A2, _A3) _B f_A f_C t])))))
    | f_A, f_C, Var x -> AVar x;;

let rec cstep_trg (Cstep_step (x1, x2, x3, x4, x5, x6)) = x5;;

let rec cstep_src (Cstep_step (x1, x2, x3, x4, x5, x6)) = x4;;

let rec showsl_eqa _A _B = showsl_rulea (showsl _A) (showsl _B) " ->* ";;

let rec showsl_conditions _A _B
  = showsl_sep (showsl_eqa _A _B) (showsl_lit ", ");;

let rec showsl_rule _A _B = showsl_rulea (showsl _A) (showsl _B) " -> ";;

let rec showsl_crule _A _B _C _D
  cr = comp (showsl_rule _A _B (fst cr))
         (if null (snd cr) then id
           else comp (showsl_lit " | ") (showsl_conditions _C _D (snd cr)));;

let rec match_rules _A (_B1, _B2, _B3)
  rs_1 rs_2 =
    bind (zip_option (map fst rs_2 @ map snd rs_2)
           (map fst rs_1 @ map snd rs_1))
      (match_list (_B1, _B2, _B3) _A _B2 (fun a -> Var a));;

let rec check_crule_variants (_A1, _A2) (_B1, _B2, _B3, _B4)
  ra r =
    (let rs = fst ra :: snd ra in
     let rsa = fst r :: snd r in
      check (not (is_none (match_rules _A1 (_B1, _B2, _B3) rs rsa)) &&
              not (is_none (match_rules _A1 (_B1, _B2, _B3) rsa rs)))
        (comp (comp (comp (showsl_crule _A2 _B4 _A2 _B4 ra)
                      (showsl_lit " and "))
                (showsl_crule _A2 _B4 _A2 _B4 r))
          (showsl_lit " are not variants of each other\092")));;

let rec showsl_crules _A _B _C _D
  ctrs =
    comp (showsl_list_gen (showsl_crule _A _B _C _D) "" "" "\092" "" ctrs)
      (showsl_literal "\092");;

let rec showsl_ctrs _A _B _C _D
  r = comp (showsl_lit "CTRS:\092\092") (showsl_crules _A _B _C _D r);;

let rec check_variant_in_ctrs (_A1, _A2) (_B1, _B2, _B3, _B4)
  ra r =
    catch_error
      (catch_error
        (existsM (check_crule_variants (_A1, _A2) (_B1, _B2, _B3, _B4) r) ra)
        (fun x -> Inl (showsl_sep id id x)))
      (fun _ ->
        Inl (comp (comp (comp (showsl_lit "rule ")
                          (showsl_crule _A2 _B4 _A2 _B4 r))
                    (showsl_lit " is not a variant of any rule in:\092"))
              (showsl_ctrs _A2 _B4 _A2 _B4 ra)));;

let rec check_csteps (_A1, _A2) (_B1, _B2, _B3, _B4)
  r s t x3 = match r, s, t, x3 with
    r, s, t, [] ->
      check (equal_terma _A1 _B2 s t)
        (comp (comp (comp (comp (showsl_lit
                                  "empty rewrite sequence but source ")
                            (showsl_terma _A2 _B4 s))
                      (showsl_lit " and target "))
                (showsl_terma _A2 _B4 t))
          (showsl_lit " differ"))
    | r, s, t, [p] ->
        bindb (check (equal_terma _A1 _B2 (cstep_src p) s)
                (comp (comp (showsl_terma _A2 _B4 (cstep_src p))
                        (showsl_lit " does not match the source "))
                  (showsl_terma _A2 _B4 s)))
          (fun _ ->
            bindb (check (equal_terma _A1 _B2 (cstep_trg p) t)
                    (comp (comp (showsl_terma _A2 _B4 (cstep_trg p))
                            (showsl_lit " does not match the target "))
                      (showsl_terma _A2 _B4 t)))
              (fun _ -> check_cstep (_A1, _A2) (_B1, _B2, _B3, _B4) r p))
    | r, s, t, p :: v :: va ->
        bindb (check (equal_terma _A1 _B2 (cstep_src p) s)
                (comp (comp (showsl_terma _A2 _B4 (cstep_src p))
                        (showsl_lit " does not match the source "))
                  (showsl_terma _A2 _B4 s)))
          (fun _ ->
            bindb (check_cstep (_A1, _A2) (_B1, _B2, _B3, _B4) r p)
              (fun _ ->
                check_csteps (_A1, _A2) (_B1, _B2, _B3, _B4) r (cstep_trg p) t
                  (v :: va)))
and check_cstep (_A1, _A2) (_B1, _B2, _B3, _B4)
  ra (Cstep_step (((l, r), cs), p, sigma, s, t, pss)) =
    bindb (check_variant_in_ctrs (_A1, _A2) (_B1, _B2, _B3, _B4) ra
            ((l, r), cs))
      (fun _ ->
        bindb (check (equal_nata (size_list pss) (size_list cs))
                (showsl_lit
                  "mismatch between number of conditions and number of rewrite sequences"))
          (fun _ ->
            bindb (check
                    (equal_terma _A1 _B2 s
                      (ctxt_apply_term (ctxt_of_pos_term p s)
                        (subst_apply_term l sigma)))
                    (comp (comp (showsl_terma _A2 _B4 s)
                            (showsl_lit " does not contain an instance of "))
                      (showsl_terma _A2 _B4 l)))
              (fun _ ->
                bindb (check
                        (equal_terma _A1 _B2 t
                          (ctxt_apply_term (ctxt_of_pos_term p s)
                            (subst_apply_term r sigma)))
                        (comp (comp (showsl_terma _A2 _B4 t)
                                (showsl_lit
                                  " does not contain an instance of "))
                          (showsl_terma _A2 _B4 r)))
                  (fun _ ->
                    catch_error
                      (forallM
                        (fun i ->
                          check_csteps (_A1, _A2) (_B1, _B2, _B3, _B4) ra
                            (subst_apply_term (fst (nth cs i)) sigma)
                            (subst_apply_term (snd (nth cs i)) sigma)
                            (nth pss i))
                        (upt zero_nata (size_list cs)))
                      (fun x -> Inl (snd x))))));;

let rec map_funs_crule
  f r = ((map_term f (fun x -> x) (fst (fst r)),
           map_term f (fun x -> x) (snd (fst r))),
          map (map_funs_rule f) (snd r));;

let rec skol_crule (_A1, _A2) (_B1, _B2, _B3, _B4)
  ra v r =
    (if member
          ((ceq_prod (ceq_prod (ceq_term _A2 _B4) (ceq_term _A2 _B4))
             (ceq_list (ceq_prod (ceq_term _A2 _B4) (ceq_term _A2 _B4)))),
            (ccompare_prod
              (ccompare_prod (ccompare_term _A1 _B3) (ccompare_term _A1 _B3))
              (ccompare_list
                (ccompare_prod (ccompare_term _A1 _B3)
                  (ccompare_term _A1 _B3)))))
          r ra
      then map_funs_crule (fun a -> Inl a) r
      else ((skol (_B1, _B2) v (fst (fst r)), skol (_B1, _B2) v (snd (fst r))),
             map (fun (s, t) -> (skol (_B1, _B2) v s, skol (_B1, _B2) v t))
               (snd r)));;

let rec skol_cstep_proof (_A1, _A2) (_B1, _B2, _B3, _B4)
  r v (Cstep_step (rho, p, sigma, s, t, pss)) =
    Cstep_step
      (skol_crule (_A1, _A2) (_B1, _B2, _B3, _B4) r v rho, p,
        comp (skol (_B1, _B2) v) sigma, skol (_B1, _B2) v s,
        skol (_B1, _B2) v t,
        map (map (skol_cstep_proof (_A1, _A2) (_B1, _B2, _B3, _B4) r v)) pss);;

let rec map_funs_crules f r = map (map_funs_crule f) r;;

let rec rules2crules rs = map (fun r -> (r, [])) rs;;

let rec skol_rules (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  cs = (let v =
          vars_trs (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
            (set ((ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6)),
                   (ccompare_prod (ccompare_term _A1 _B5)
                     (ccompare_term _A1 _B5)),
                   (set_impl_prod set_impl_term set_impl_term))
              cs)
          in
         map (fun (l, r) ->
               (skol (_B3, _B4.ccompare_cproper_interval) v l,
                 skol (_B3, _B4.ccompare_cproper_interval) v r))
           cs);;

let rec check_context_joinablea (_A1, _A2, _A3)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9)
  (Contextual_Join (u, ps, qs)) r s t cs =
    catch_error
      (let c = skol_rules (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B8) cs in
       let v =
         vars_trs (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B8)
           (set ((ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6)),
                  (ccompare_prod (ccompare_term _A1 _B5)
                    (ccompare_term _A1 _B5)),
                  (set_impl_prod set_impl_term set_impl_term))
             cs)
         in
       let sa = skol (_B3, _B4.ccompare_cproper_interval) v s in
       let ta = skol (_B3, _B4.ccompare_cproper_interval) v t in
       let ua = skol (_B3, _B4.ccompare_cproper_interval) v u in
       let psa =
         map (skol_cstep_proof (_A1, _A2)
               (_B3, _B4.ccompare_cproper_interval, _B5, _B6)
               (set ((ceq_prod (ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6))
                       (ceq_list
                         (ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6)))),
                      (ccompare_prod
                        (ccompare_prod (ccompare_term _A1 _B5)
                          (ccompare_term _A1 _B5))
                        (ccompare_list
                          (ccompare_prod (ccompare_term _A1 _B5)
                            (ccompare_term _A1 _B5)))),
                      (set_impl_prod (set_impl_prod set_impl_term set_impl_term)
                        set_impl_list))
                 r)
               v)
           ps
         in
       let qsa =
         map (skol_cstep_proof (_A1, _A2)
               (_B3, _B4.ccompare_cproper_interval, _B5, _B6)
               (set ((ceq_prod (ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6))
                       (ceq_list
                         (ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6)))),
                      (ccompare_prod
                        (ccompare_prod (ccompare_term _A1 _B5)
                          (ccompare_term _A1 _B5))
                        (ccompare_list
                          (ccompare_prod (ccompare_term _A1 _B5)
                            (ccompare_term _A1 _B5)))),
                      (set_impl_prod (set_impl_prod set_impl_term set_impl_term)
                        set_impl_list))
                 r)
               v)
           qs
         in
       let ra = map_funs_crules (fun a -> Inl a) r @ rules2crules c in
        bindb (check_csteps ((equal_sum _A2 _B6), (showl_sum _A3 _B9))
                (_B4.ccompare_cproper_interval, _B6, _B7, _B9) ra sa ua psa)
          (fun _ ->
            check_csteps ((equal_sum _A2 _B6), (showl_sum _A3 _B9))
              (_B4.ccompare_cproper_interval, _B6, _B7, _B9) ra ta ua qsa))
      (fun x ->
        Inl (comp (comp (comp (comp (showsl_terma _A3 _B9 s)
                                (showsl_lit " and "))
                          (showsl_terma _A3 _B9 t))
                    (showsl_lit " are not context-joinable\092"))
              x));;

let rec check_context_joinable (_A1, _A2, _A3)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9, _B10)
  cj r s t cs =
    catch_error
      (existsM
        (fun (sa, (ta, (csa, p))) ->
          bindb (check
                  (not (is_none
                         (match_rules _A2
                           (_B4.ccompare_cproper_interval, _B6, _B7)
                           ((s, t) :: cs) ((sa, ta) :: csa))))
                  id)
            (fun _ ->
              bindb (check
                      (not (is_none
                             (match_rules _A2
                               (_B4.ccompare_cproper_interval, _B6, _B7)
                               ((sa, ta) :: csa) ((s, t) :: cs))))
                      id)
                (fun _ ->
                  check_context_joinablea (_A1.compare_compare_order, _A2, _A3)
                    (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B9, _B10) p r sa ta
                    csa)))
        cj)
      (fun x -> Inl (showsl_sep id (showsl_literal "\092") x));;

let rec showsl_pos
  x = showsl_list_gen (fun i -> showsl_nat (suc i)) "empty" "" "." "" x;;

let rec showsl_coverlap _A _B _C _D _E _F _G _H
  rho_1 rho_2 p =
    comp (comp (comp (comp (comp (showsl_lit "overlap of conditional rules ")
                             (showsl_crule _A _B _C _D rho_1))
                       (showsl_lit " and "))
                 (showsl_crule _E _F _G _H rho_2))
           (showsl_lit " at position "))
      (showsl_pos p);;

let rec mgu_var_disjoint_generic _B _D
  vu wu s t =
    (match mgu _D _B (map_term (fun x -> x) vu s) (map_term (fun x -> x) wu t)
      with None -> None | Some gamma -> Some (comp gamma vu, comp gamma wu));;

let rec check_overlap (_A1, _A2, _A3, _A4, _A5) (_B1, _B2)
  xvar yvar check_context_joinable check_infeasible check_unfeasible r rho_1
    rho_2 p =
    catch_error
      (match
        mgu_var_disjoint_generic _A2 _B1 xvar yvar (subt_at (fst (fst rho_1)) p)
          (fst (fst rho_2))
        with None -> Inr ()
        | Some (sigma_1, sigma_2) ->
          (let cs =
             subst_list sigma_1 (snd rho_1) @ subst_list sigma_2 (snd rho_2) in
           let s = subst_apply_term (snd (fst rho_1)) sigma_1 in
           let t =
             ctxt_apply_term
               (ctxt_of_pos_term p (subst_apply_term (fst (fst rho_1)) sigma_1))
               (subst_apply_term (snd (fst rho_2)) sigma_2)
             in
            catch_error
              (choice
                [catch_error
                   (bindb (check (null p) id)
                     (fun _ ->
                       check_crule_variants (_B1, _B2) (_A1, _A2, _A3, _A5)
                         rho_1 rho_2))
                   (fun x ->
                     Inl (comp (showsl_lit "the overlap is critical\092") x));
                  catch_error (check_context_joinable r s t cs)
                    (fun x ->
                      Inl (comp (showsl_lit
                                  "could not be shown to be context-joinable\092")
                            x));
                  catch_error (check_infeasible r cs)
                    (fun x ->
                      Inl (comp (showsl_lit
                                  "could not be shown to be infeasible\092")
                            x));
                  catch_error (check_unfeasible r (fst (fst rho_1)) sigma_1 cs)
                    (fun x ->
                      Inl (comp (showsl_lit
                                  "could not be shown to be unfeasible\092")
                            x))])
              (fun x -> Inl (showsl_sep id (showsl_literal "\092") x))))
      (fun x ->
        Inl (comp (comp (showsl_coverlap _B2 _A5 _B2 _A5 _B2 _A5 _B2 _A5 rho_1
                          rho_2 p)
                    (showsl_lit ":\092\092"))
              x));;

let rec rule2 (UnfeasibleOverlap (x1, x2, x3, x4, x5, x6, x7)) = x7;;

let rec rule1 (UnfeasibleOverlap (x1, x2, x3, x4, x5, x6, x7)) = x6;;

let rec check_airr (_A1, _A2)
  r t = catch_error
          (catch_error
            (forallM
              (fun cr ->
                catch_error
                  (forallM
                    (fun p ->
                      check (is_none
                              (mgu_var_disjoint_generic (equal_list equal_char)
                                _A1 (fun a ->
                                      Chara
(false, false, false, true, true, true, true, false) ::
a)
                                (fun a ->
                                  Chara (true, false, false, true, true, true,
  true, false) ::
                                    a)
                                (subt_at t p) (fst (fst cr))))
                        (comp (comp (comp (comp
    (comp (showsl_lit "the term ") (showsl_terma _A2 (showl_list showl_char) t))
    (showsl_lit " is unifiable with the left-hand side of rule "))
                                      (showsl_crule _A2 (showl_list showl_char)
_A2 (showl_list showl_char) cr))
                                (showsl_lit " at position "))
                          (showsl_pos p)))
                    (funposs_list t))
                  (fun x -> Inl (snd x)))
              r)
            (fun x -> Inl (snd x)))
          (fun x ->
            Inl (comp (comp (comp (showsl_lit "the term ")
                              (showsl_terma _A2 (showl_list showl_char) t))
                        (showsl_lit " is not absolutely irreducible"))
                  x));;

let rec check_unfeasiblea (_A1, _A2, _A3)
  (UnfeasibleOverlap (t, u, v, ps, qs, rho_1, rho_2)) r l mu cs =
    catch_error
      (let c =
         skol_rules (_A1, _A2)
           (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
             (cproper_interval_list ccompare_char), (compare_list compare_char),
             (equal_list equal_char), set_impl_list)
           cs
         in
       let va =
         vars_trs (_A1, _A2)
           (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
             (cproper_interval_list ccompare_char), (compare_list compare_char),
             (equal_list equal_char), set_impl_list)
           (set ((ceq_prod (ceq_term _A2 (equal_list equal_char))
                   (ceq_term _A2 (equal_list equal_char))),
                  (ccompare_prod (ccompare_term _A1 (compare_list compare_char))
                    (ccompare_term _A1 (compare_list compare_char))),
                  (set_impl_prod set_impl_term set_impl_term))
             cs)
         in
       let ta = skol ((ceq_list ceq_char), (ccompare_list ccompare_char)) va t
         in
       let ua = skol ((ceq_list ceq_char), (ccompare_list ccompare_char)) va u
         in
       let vb = skol ((ceq_list ceq_char), (ccompare_list ccompare_char)) va v
         in
       let psa =
         map (skol_cstep_proof (_A1, _A2)
               ((ceq_list ceq_char), (ccompare_list ccompare_char),
                 (compare_list compare_char), (equal_list equal_char))
               (set ((ceq_prod
                       (ceq_prod (ceq_term _A2 (equal_list equal_char))
                         (ceq_term _A2 (equal_list equal_char)))
                       (ceq_list
                         (ceq_prod (ceq_term _A2 (equal_list equal_char))
                           (ceq_term _A2 (equal_list equal_char))))),
                      (ccompare_prod
                        (ccompare_prod
                          (ccompare_term _A1 (compare_list compare_char))
                          (ccompare_term _A1 (compare_list compare_char)))
                        (ccompare_list
                          (ccompare_prod
                            (ccompare_term _A1 (compare_list compare_char))
                            (ccompare_term _A1 (compare_list compare_char))))),
                      (set_impl_prod (set_impl_prod set_impl_term set_impl_term)
                        set_impl_list))
                 r)
               va)
           ps
         in
       let qsa =
         map (skol_cstep_proof (_A1, _A2)
               ((ceq_list ceq_char), (ccompare_list ccompare_char),
                 (compare_list compare_char), (equal_list equal_char))
               (set ((ceq_prod
                       (ceq_prod (ceq_term _A2 (equal_list equal_char))
                         (ceq_term _A2 (equal_list equal_char)))
                       (ceq_list
                         (ceq_prod (ceq_term _A2 (equal_list equal_char))
                           (ceq_term _A2 (equal_list equal_char))))),
                      (ccompare_prod
                        (ccompare_prod
                          (ccompare_term _A1 (compare_list compare_char))
                          (ccompare_term _A1 (compare_list compare_char)))
                        (ccompare_list
                          (ccompare_prod
                            (ccompare_term _A1 (compare_list compare_char))
                            (ccompare_term _A1 (compare_list compare_char))))),
                      (set_impl_prod (set_impl_prod set_impl_term set_impl_term)
                        set_impl_list))
                 r)
               va)
           qs
         in
       let ra = map_funs_crules (fun a -> Inl a) r @ rules2crules c in
        bindb (check
                (equal_terma _A2 (equal_list equal_char) l (fst (fst rho_1)))
                id)
          (fun _ ->
            bindb (check
                    (list_all
                      (fun (a, b) ->
                        (let (la, _) = a in (fun _ -> not (is_Var la))) b)
                      r)
                    (showsl_lit "variable left-hand side"))
              (fun _ ->
                bindb (check_variant_in_ctrs (_A2, _A3)
                        ((ccompare_list ccompare_char), (equal_list equal_char),
                          mapping_impl_list, (showl_list showl_char))
                        r rho_1)
                  (fun _ ->
                    bindb (check_variant_in_ctrs (_A2, _A3)
                            ((ccompare_list ccompare_char),
                              (equal_list equal_char), mapping_impl_list,
                              (showl_list showl_char))
                            r rho_2)
                      (fun _ ->
                        bindb (check
                                (equal_terma _A2 (equal_list equal_char)
                                   (subst_apply_term l mu)
                                   (subst_apply_term (fst (fst rho_2)) mu) ||
                                  supt_impl _A2 (equal_list equal_char)
                                    (subst_apply_term l mu)
                                    (subst_apply_term (fst (fst rho_2)) mu))
                                id)
                          (fun _ ->
                            bindb (check
                                    (member
                                      ((ceq_term _A2 (equal_list equal_char)),
(ccompare_term _A1 (compare_list compare_char)))
                                      t (image
  ((ceq_prod (ceq_term _A2 (equal_list equal_char))
     (ceq_term _A2 (equal_list equal_char))),
    (ccompare_prod (ccompare_term _A1 (compare_list compare_char))
      (ccompare_term _A1 (compare_list compare_char))))
  ((ceq_term _A2 (equal_list equal_char)),
    (ccompare_term _A1 (compare_list compare_char)), set_impl_term)
  fst (set ((ceq_prod (ceq_term _A2 (equal_list equal_char))
              (ceq_term _A2 (equal_list equal_char))),
             (ccompare_prod (ccompare_term _A1 (compare_list compare_char))
               (ccompare_term _A1 (compare_list compare_char))),
             (set_impl_prod set_impl_term set_impl_term))
        cs)))
                                    id)
                              (fun _ ->
                                bindb (check
(equal_lista
  (equal_prod (equal_term _A2 (equal_list equal_char))
    (equal_term _A2 (equal_list equal_char)))
  cs (subst_list mu (snd rho_1 @ snd rho_2)))
id)
                                  (fun _ ->
                                    bindb (check_csteps
    ((equal_sum _A2 (equal_list equal_char)),
      (showl_sum _A3 (showl_list showl_char)))
    ((ccompare_list ccompare_char), (equal_list equal_char), mapping_impl_list,
      (showl_list showl_char))
    ra ta ua psa)
                                      (fun _ ->
bindb (check_csteps
        ((equal_sum _A2 (equal_list equal_char)),
          (showl_sum _A3 (showl_list showl_char)))
        ((ccompare_list ccompare_char), (equal_list equal_char),
          mapping_impl_list, (showl_list showl_char))
        ra ta vb qsa)
  (fun _ ->
    bindb (check_airr (_A2, _A3) r u)
      (fun _ ->
        bindb (check_airr (_A2, _A3) r v)
          (fun _ ->
            check (is_none (mgu _A2 (equal_list equal_char) u v))
              (comp (comp (comp (showsl_terma _A3 (showl_list showl_char) u)
                            (showsl_lit " and "))
                      (showsl_terma _A3 (showl_list showl_char) v))
                (showsl_lit " are unifiable"))))))))))))))
      (fun x ->
        Inl (comp (comp (comp (showsl_lit "conditions ")
                          (showsl_conditions _A3 (showl_list showl_char) cs))
                    (showsl_lit " are not unfeasible\092"))
              x));;

let rec check_unfeasible (_A1, _A2, _A3)
  css r l mu cs =
    catch_error
      (existsM
        (fun (mua, uo) ->
          (let cs_1 = snd (rule1 uo) in
           let cs_2 = snd (rule2 uo) in
           let csa = subst_list mua (cs_1 @ cs_2) in
           let la = fst (fst (rule1 uo)) in
            bindb (check (equal_nata (size_list csa) (size_list cs))
                    (showsl_lit "lengths differ"))
              (fun _ ->
                bindb (check
                        (not (is_none
                               (match_rules _A2
                                 ((ccompare_list ccompare_char),
                                   (equal_list equal_char), mapping_impl_list)
                                 ((subst_apply_term l mu,
                                    subst_apply_term l mu) ::
                                   cs)
                                 ((subst_apply_term la mua,
                                    subst_apply_term la mua) ::
                                   csa))) &&
                          not (is_none
                                (match_rules _A2
                                  ((ccompare_list ccompare_char),
                                    (equal_list equal_char), mapping_impl_list)
                                  ((subst_apply_term la mua,
                                     subst_apply_term la mua) ::
                                    csa)
                                  ((subst_apply_term l mu,
                                     subst_apply_term l mu) ::
                                    cs))))
                        id)
                  (fun _ ->
                    check_unfeasiblea (_A1.compare_compare_order, _A2, _A3) uo r
                      la mua csa))))
        css)
      (fun x -> Inl (showsl_sep id (showsl_literal "\092") x));;

let rec ta_rules_implb (Tree_Automaton (x1, x2, x3)) = x2;;

let rec check_subseteq _A
  xs ys =
    catch_error
      (forallM (fun x -> (if membera _A ys x then Inr () else Inl x)) xs)
      (fun x -> Inl (snd x));;

let rec check_rules_subseteq (_A1, _A2) (_B1, _B2)
  rs a =
    catch_error (check_subseteq (equal_ta_rule _A1 _B1) rs (ta_rules_implb a))
      (fun x ->
        Inl (comp (comp (showsl_lit "rule ") (showsl_ta_rule _A2 _B2 x))
              (showsl_lit " is missing")));;

let rec ta_rules (Ta_ext (ta_final, ta_rules, ta_eps, more)) = ta_rules;;

let rec ta_final (Ta_ext (ta_final, ta_rules, ta_eps, more)) = ta_final;;

let rec ta_syms (_A1, _A2) (_B1, _B2, _B3, _B4)
  ta = image ((ceq_ta_rule _A2 _B3), (ccompare_ta_rule _A1 _B2))
         ((ceq_prod _B1 ceq_nat), (ccompare_prod _B2 ccompare_nat),
           (set_impl_prod _B4 set_impl_nat))
         r_sym (ta_rules ta);;

let rec ta_inter_eps_empty (_A1, _A2, _A3, _A4) (_B1, _B2, _B3, _B4)
  (_C1, _C2, _C3, _C4)
  ta t =
    Ta_ext
      (productc (_A1, _A2, _A4) (_C1, _C2, _C4) (ta_final ta) (ta_final t),
        image ((ceq_prod (ceq_ta_rule _A3 _B3) (ceq_ta_rule _C3 _B3)),
                (ccompare_prod (ccompare_ta_rule _A2 _B2)
                  (ccompare_ta_rule _C2 _B2)))
          ((ceq_ta_rule (equal_prod _A3 _C3) _B3),
            (ccompare_ta_rule (ccompare_prod _A2 _C2) _B2), set_impl_ta_rule)
          (fun (TA_rule (f, ps, p), TA_rule (_, qs, q)) ->
            TA_rule (f, zip ps qs, (p, q)))
          (sup_setb
            ((finite_UNIV_prod finite_UNIV_ta_rule finite_UNIV_ta_rule),
              (cenum_prod cenum_ta_rule cenum_ta_rule),
              (ceq_prod (ceq_ta_rule _A3 _B3) (ceq_ta_rule _C3 _B3)),
              (cproper_interval_prod (cproper_interval_ta_rule _A2 _B2)
                (cproper_interval_ta_rule _C2 _B2)),
              (set_impl_prod set_impl_ta_rule set_impl_ta_rule))
            (image ((ceq_prod _B1 ceq_nat), (ccompare_prod _B2 ccompare_nat))
              ((ceq_set
                 ((cenum_prod cenum_ta_rule cenum_ta_rule),
                   (ceq_prod (ceq_ta_rule _A3 _B3) (ceq_ta_rule _C3 _B3)),
                   (cproper_interval_prod (cproper_interval_ta_rule _A2 _B2)
                     (cproper_interval_ta_rule _C2
                       _B2)).ccompare_cproper_interval)),
                (ccompare_set
                  ((finite_UNIV_prod finite_UNIV_ta_rule finite_UNIV_ta_rule),
                    (ceq_prod (ceq_ta_rule _A3 _B3) (ceq_ta_rule _C3 _B3)),
                    (cproper_interval_prod (cproper_interval_ta_rule _A2 _B2)
                      (cproper_interval_ta_rule _C2 _B2)),
                    (set_impl_prod set_impl_ta_rule set_impl_ta_rule))),
                set_impl_set)
              (fun f ->
                productc
                  ((ceq_ta_rule _A3 _B3), (ccompare_ta_rule _A2 _B2),
                    set_impl_ta_rule)
                  ((ceq_ta_rule _C3 _B3), (ccompare_ta_rule _C2 _B2),
                    set_impl_ta_rule)
                  (filter ((ceq_ta_rule _A3 _B3), (ccompare_ta_rule _A2 _B2))
                    (fun r -> equal_proda _B3 equal_nat (r_sym r) f)
                    (ta_rules ta))
                  (filter ((ceq_ta_rule _C3 _B3), (ccompare_ta_rule _C2 _B2))
                    (fun r -> equal_proda _B3 equal_nat (r_sym r) f)
                    (ta_rules t)))
              (ta_syms (_A2, _A3) (_B1, _B2, _B3, _B4) ta))),
        bot_set
          ((ceq_prod (ceq_prod _A1 _C1) (ceq_prod _A1 _C1)),
            (ccompare_prod (ccompare_prod _A2 _C2) (ccompare_prod _A2 _C2)),
            (set_impl_prod (set_impl_prod _A4 _C4) (set_impl_prod _A4 _C4))),
        ());;

let rec add_rule_states _A
  rs ss =
    fold (fun r ssa ->
           (let TA_rule (_, qs, q) = r in
             insertb _A q (fold (insertb _A) qs ssa)))
      rs ss;;

let rec sig_rules (_A1, _A2, _A3) (_B1, _B2)
  f c = image ((ceq_prod _A1 ceq_nat), (ccompare_prod _A2 ccompare_nat))
          ((ceq_ta_rule _B2 _A3), (ccompare_ta_rule _B1 _A2), set_impl_ta_rule)
          (fun (fa, n) -> TA_rule (fa, replicate n c, c)) f;;

let rec star c x1 = match c, x1 with c, Var x -> Fun (c, [])
               | c, Fun (f, ts) -> Fun (f, map (star c) ts);;

let rec ground_instances_rules (_A1, _A2, _A3, _A4) (_B1, _B2)
  f c x2 = match f, c, x2 with
    f, c, Var x ->
      sig_rules (_A1, _A2, _A4) ((ccompare_term _A3 _B1), (equal_term _A4 _B2))
        f (Fun (c, []))
    | fa, c, Fun (f, ts) ->
        sup_seta
          ((ceq_ta_rule (equal_term _A4 _B2) _A4),
            (ccompare_ta_rule (ccompare_term _A3 _B1) _A2))
          (inserta
            ((ceq_ta_rule (equal_term _A4 _B2) _A4),
              (ccompare_ta_rule (ccompare_term _A3 _B1) _A2))
            (TA_rule (f, map (star c) ts, star c (Fun (f, ts))))
            (set_empty
              ((ceq_ta_rule (equal_term _A4 _B2) _A4),
                (ccompare_ta_rule (ccompare_term _A3 _B1) _A2))
              (of_phantom set_impl_ta_rulea)))
          (sup_setb
            (finite_UNIV_ta_rule, cenum_ta_rule,
              (ceq_ta_rule (equal_term _A4 _B2) _A4),
              (cproper_interval_ta_rule (ccompare_term _A3 _B1) _A2),
              set_impl_ta_rule)
            (image ((ceq_term _A4 _B2), (ccompare_term _A3 _B1))
              ((ceq_set
                 (cenum_ta_rule, (ceq_ta_rule (equal_term _A4 _B2) _A4),
                   (cproper_interval_ta_rule (ccompare_term _A3 _B1)
                     _A2).ccompare_cproper_interval)),
                (ccompare_set
                  (finite_UNIV_ta_rule, (ceq_ta_rule (equal_term _A4 _B2) _A4),
                    (cproper_interval_ta_rule (ccompare_term _A3 _B1) _A2),
                    set_impl_ta_rule)),
                set_impl_set)
              (ground_instances_rules (_A1, _A2, _A3, _A4) (_B1, _B2) fa c)
              (set ((ceq_term _A4 _B2), (ccompare_term _A3 _B1), set_impl_term)
                ts)));;

let rec ground_instances_ta (_A1, _A2, _A3, _A4) (_B1, _B2)
  f c t =
    Ta_ext
      (inserta ((ceq_term _A4 _B2), (ccompare_term _A3 _B1)) (star c t)
         (set_empty ((ceq_term _A4 _B2), (ccompare_term _A3 _B1))
           (of_phantom set_impl_terma)),
        ground_instances_rules (_A1, _A2, _A3, _A4) (_B1, _B2) f c t,
        set_empty
          ((ceq_prod (ceq_term _A4 _B2) (ceq_term _A4 _B2)),
            (ccompare_prod (ccompare_term _A3 _B1) (ccompare_term _A3 _B1)))
          (of_phantom (set_impl_proda set_impl_term set_impl_term)),
        ());;

let rec sig_rules_list
  f c = map (fun (fa, n) -> TA_rule (fa, replicate n c, c)) f;;

let rec gi_rules_list
  f c x2 = match f, c, x2 with f, c, Var x -> sig_rules_list f (Fun (c, []))
    | fa, c, Fun (f, ts) ->
        TA_rule (f, map (star c) ts, star c (Fun (f, ts))) ::
          maps (gi_rules_list fa c) ts;;

let rec growing_rule (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  (l, r) =
    ball (_B3, _B4.ccompare_cproper_interval)
      (vars_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7) r)
      (fun x ->
        ball ((ceq_list ceq_nat), (ccompare_list ccompare_nat)) (varposs l)
          (fun p ->
            (if equal_terma _A2 _B6 (Var x) (subt_at l p)
              then less_eq_nat (size_list p) one_nata else true)));;

let rec growing (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  r = ball ((ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6)),
             (ccompare_prod (ccompare_term _A1 _B5) (ccompare_term _A1 _B5)))
        r (growing_rule (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7));;

let rec check_growing (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  r = check (growing (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
              (set ((ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6)),
                     (ccompare_prod (ccompare_term _A1 _B5)
                       (ccompare_term _A1 _B5)),
                     (set_impl_prod set_impl_term set_impl_term))
                r))
        (showsl_lit "TRS is not growing");;

let rec combs
  x0 ys = match x0, ys with [], ys -> [[]]
    | x :: xs, ys ->
        maps (fun l -> map (fun y -> (x, y) :: l) ys) (combs xs ys);;

let rec state_substs v q = combs v q;;

let rec lhss_impl _A r = remdups _A (map fst r);;

let rec mp_ta_rules _A _B
  r f c = maps (gi_rules_list f c) (lhss_impl (equal_term _A _B) r);;

let rec r_lhs_states (TA_rule (x1, x2, x3)) = x2;;

let rec r_root (TA_rule (x1, x2, x3)) = x1;;

let rec r_rhs (TA_rule (x1, x2, x3)) = x3;;

let rec all_interval_nat
  p i j = less_eq_nat j i || p i && all_interval_nat p (suc i) j;;

let rec reachable_states (_A1, _A2, _A3, _A4) (_B1, _B2)
  delta x1 = match delta, x1 with
    delta, Fun (Inr q, []) -> inserta (_A1, _A2) q (bot_set (_A1, _A2, _A4))
    | delta, Fun (Inl f, ts) ->
        image ((ceq_ta_rule _A3 _B2), (ccompare_ta_rule _A2 _B1))
          (_A1, _A2, _A4) r_rhs
          (filter ((ceq_ta_rule _A3 _B2), (ccompare_ta_rule _A2 _B1))
            (fun r ->
              eq _B2 (r_root r) f &&
                (equal_nata (size_list (r_lhs_states r)) (size_list ts) &&
                  all_interval_nat
                    (fun i ->
                      member (_A1, _A2) (nth (r_lhs_states r) i)
                        (reachable_states (_A1, _A2, _A3, _A4) (_B1, _B2) delta
                          (nth ts i)))
                    zero_nata (size_list ts)))
            delta)
    | delta, Var v -> bot_set (_A1, _A2, _A4)
    | delta, Fun (Inr va, vb :: vc) -> bot_set (_A1, _A2, _A4);;

let rec qi _B
  c t g x3 = match c, t, g, x3 with
    c, t, g, Var x -> (if contains_var_term _B x t then g x else star c (Var x))
    | c, t, g, Fun (f, ts) -> star c (Fun (f, ts));;

let rec inf_step (_A1, _A2, _A3) (_B1, _B2)
  c r s delta =
    foldr (comp (sup_seta
                  ((ceq_ta_rule (equal_term _A3 _B2) _A3),
                    (ccompare_ta_rule (ccompare_term _A2 _B1) _A1)))
            (fun (a, b) ->
              (let (l, ra) = a in
                (fun theta ->
                  (let Fun (f, ls) = l in
                    image ((ceq_term _A3 _B2), (ccompare_term _A2 _B1))
                      ((ceq_ta_rule (equal_term _A3 _B2) _A3),
                        (ccompare_ta_rule (ccompare_term _A2 _B1) _A1),
                        set_impl_ta_rule)
                      (fun aa ->
                        TA_rule
                          (f, map (qi _B2 c ra (fun_of _B2 theta)) ls, aa))
                      (reachable_states
                        ((ceq_term _A3 _B2), (ccompare_term _A2 _B1),
                          (equal_term _A3 _B2), set_impl_term)
                        (_A1, _A3) delta
                        (subst_apply_term
                          (map_term (fun aa -> Inl aa) (fun x -> x) ra)
                          (comp (fun fa -> Fun (Inr fa, []))
                            (fun_of _B2 theta)))))))
                b))
      (product r s)
      (set_empty
        ((ceq_ta_rule (equal_term _A3 _B2) _A3),
          (ccompare_ta_rule (ccompare_term _A2 _B1) _A1))
        (of_phantom set_impl_ta_rulea));;

let rec funas_ta_rule (_B1, _B2, _B3)
  r = inserta ((ceq_prod _B1 ceq_nat), (ccompare_prod _B2 ccompare_nat))
        (r_root r, size_list (r_lhs_states r))
        (bot_set
          ((ceq_prod _B1 ceq_nat), (ccompare_prod _B2 ccompare_nat),
            (set_impl_prod _B3 set_impl_nat)));;

let rec funas_ta (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6)
  a = sup_setb
        ((finite_UNIV_prod _B1 finite_UNIV_nat), (cenum_prod _B2 cenum_nat),
          (ceq_prod _B3 ceq_nat),
          (cproper_interval_prod _B4 cproper_interval_nat),
          (set_impl_prod _B6 set_impl_nat))
        (image
          ((ceq_ta_rule _A2 _B5),
            (ccompare_ta_rule _A1 _B4.ccompare_cproper_interval))
          ((ceq_set
             ((cenum_prod _B2 cenum_nat), (ceq_prod _B3 ceq_nat),
               (cproper_interval_prod _B4
                 cproper_interval_nat).ccompare_cproper_interval)),
            (ccompare_set
              ((finite_UNIV_prod _B1 finite_UNIV_nat), (ceq_prod _B3 ceq_nat),
                (cproper_interval_prod _B4 cproper_interval_nat),
                (set_impl_prod _B6 set_impl_nat))),
            set_impl_set)
          (funas_ta_rule (_B3, _B4.ccompare_cproper_interval, _B6))
          (ta_rules a));;

let rec check_varcond_no_Var_lhs _A _B
  = (fun xs ->
      catch_error
        (forallM
          (fun rule ->
            check (not (is_Var (fst rule)))
              (comp (comp (showsl_literal "variable left-hand side in rule ")
                      (showsl_rule _A _B rule))
                (showsl_literal "\092")))
          xs)
        (fun x -> Inl (snd x)));;

let rec ta_of_ta (_A1, _A2, _A3, _A4) (_B1, _B2)
  (Tree_Automaton (fin, rules, eps)) =
    Ta_ext
      (set (_A1, _A2, _A4) fin,
        set ((ceq_ta_rule _A3 _B2), (ccompare_ta_rule _A2 _B1),
              set_impl_ta_rule)
          rules,
        set ((ceq_prod _A1 _A1), (ccompare_prod _A2 _A2),
              (set_impl_prod _A4 _A4))
          eps,
        ());;

let rec insert_funas_term _A
  x0 fs = match x0, fs with Var x, fs -> fs
    | Fun (f, ts), fs ->
        insertb (equal_prod _A equal_nat) (f, size_list ts)
          (foldr (insert_funas_term _A) ts fs);;

let rec insert_funas_rule _A
  r fs = insert_funas_term _A (fst r) (insert_funas_term _A (snd r) fs);;

let rec insert_funas_trs _A trs = foldr (insert_funas_rule _A) trs;;

let rec linear_term_impl (_A1, _A2)
  xs x1 = match xs, x1 with
    xs, Var x ->
      (if member (_A1, _A2) x xs then None else Some (inserta (_A1, _A2) x xs))
    | xs, Fun (uu, []) -> Some xs
    | xs, Fun (f, t :: ts) ->
        (match linear_term_impl (_A1, _A2) xs t with None -> None
          | Some ys -> linear_term_impl (_A1, _A2) ys (Fun (f, ts)));;

let rec linear_term (_B1, _B2, _B3)
  t = not (is_none (linear_term_impl (_B1, _B2) (bot_set (_B1, _B2, _B3)) t));;

let rec check_linear_trs _A (_B1, _B2, _B3, _B4)
  r = catch_error
        (catch_error
          (forallM
            (fun x ->
              (if (let (l, ra) = x in
                    linear_term (_B1, _B2, _B3) l &&
                      linear_term (_B1, _B2, _B3) ra)
                then Inr () else Inl x))
            r)
          (fun x -> Inl (snd x)))
        (fun _ ->
          Inl (comp (showsl_trs _A _B4 r)
                (showsl_literal "\092is not linear\092")));;

let rec ta_eps (Ta_ext (ta_final, ta_rules, ta_eps, more)) = ta_eps;;

let rec reduced_TA (_A1, _A2) (_B1, _B2, _B3, _B4)
  f ta q =
    Ta_ext
      (bot_set (_B1, _B2, _B4),
        sup_seta ((ceq_ta_rule _B3 _A2), (ccompare_ta_rule _B2 _A1))
          (image ((ceq_ta_rule _B3 _A2), (ccompare_ta_rule _B2 _A1))
            ((ceq_ta_rule _B3 _A2), (ccompare_ta_rule _B2 _A1),
              set_impl_ta_rule)
            (fun (TA_rule (fa, qs, a)) ->
              TA_rule
                (fa, filtera (fun qa -> not (member (_B1, _B2) qa q)) qs, a))
            (filter ((ceq_ta_rule _B3 _A2), (ccompare_ta_rule _B2 _A1))
              (fun r -> not (member (_B1, _B2) (r_rhs r) q)) (ta_rules ta)))
          (image ((ceq_prod _B1 _B1), (ccompare_prod _B2 _B2))
            ((ceq_ta_rule _B3 _A2), (ccompare_ta_rule _B2 _A1),
              set_impl_ta_rule)
            (fun p -> TA_rule (f, [], snd p))
            (filter ((ceq_prod _B1 _B1), (ccompare_prod _B2 _B2))
              (fun p ->
                member (_B1, _B2) (fst p) q &&
                  not (member (_B1, _B2) (snd p) q))
              (ta_eps ta))),
        filter ((ceq_prod _B1 _B1), (ccompare_prod _B2 _B2))
          (fun p ->
            not (member (_B1, _B2) (fst p) q) &&
              not (member (_B1, _B2) (snd p) q))
          (ta_eps ta),
        ());;

let rec new_reach (_A1, _A2, _A3, _A4) (_B1, _B2)
  ta = image ((ceq_ta_rule _A3 _B2), (ccompare_ta_rule _A2 _B1)) (_A1, _A2, _A4)
         r_rhs
         (filter ((ceq_ta_rule _A3 _B2), (ccompare_ta_rule _A2 _B1))
           (fun r -> null (r_lhs_states r)) (ta_rules ta));;

let rec ta_reachable (_A1, _A2, _A3, _A4, _A5) (_B1, _B2, _B3)
  ta = (let q = new_reach (_A2, _A3, _A4, _A5) (_B1, _B3) ta in
         (if less_eq_set (_A1, _A2, _A3) q (bot_set (_A2, _A3, _A5))
           then bot_set (_A2, _A3, _A5)
           else sup_seta (_A2, _A3) q
                  (ta_reachable (_A1, _A2, _A3, _A4, _A5) (_B1, _B2, _B3)
                    (reduced_TA (_B1, _B3) (_A2, _A3, _A4, _A5) (default _B2) ta
                      q))));;

let rec ta_empty (_A1, _A2, _A3, _A4, _A5) (_B1, _B2, _B3)
  ta = less_eq_set (_A1, _A2, _A3)
         (inf_seta (_A2, _A3)
           (ta_reachable (_A1, _A2, _A3, _A4, _A5) (_B1, _B2, _B3) ta)
           (ta_final ta))
         (bot_set (_A2, _A3, _A5));;

let rec check_etac_nonreachable (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  f aa c a r s t =
    (let fa =
       set ((ceq_prod _A3 ceq_nat),
             (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat),
             (set_impl_prod _A8 set_impl_nat))
         f
       in
      bindb (check
              (member
                ((ceq_prod _A3 ceq_nat),
                  (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat))
                (aa, zero_nata) fa)
              (comp (comp (showsl_lit "constant ") (showsl _A9 aa))
                (showsl_lit " is not in signature")))
        (fun _ ->
          bindb (check
                  (not (member
                         ((ceq_prod _A3 ceq_nat),
                           (ccompare_prod _A4.ccompare_cproper_interval
                             ccompare_nat))
                         (c, zero_nata) fa))
                  (showsl_lit "star-symbol is not fresh w.r.t. signature"))
            (fun _ ->
              bindb (catch_error
                      (check_subseteq (equal_prod _A7 equal_nat)
                        (insert_funas_term _A7 s []) f)
                      (fun _ -> Inl (showsl_lit "lhs violates signature")))
                (fun _ ->
                  bindb (catch_error
                          (check_subseteq (equal_prod _A7 equal_nat)
                            (insert_funas_term _A7 t []) f)
                          (fun _ -> Inl (showsl_lit "rhs violates signature")))
                    (fun _ ->
                      (let fs = insert_funas_trs _A7 r [] in
                        bindb (catch_error
                                (check_subseteq (equal_prod _A7 equal_nat) fs f)
                                (fun _ ->
                                  Inl (showsl_lit "TRS violates signature")))
                          (fun _ ->
                            bindb (check_varcond_no_Var_lhs _A9 _B8 r)
                              (fun _ ->
                                bindb (check_linear_trs _A9
(_B3, _B4.ccompare_cproper_interval, _B7, _B8) r)
                                  (fun _ ->
                                    bindb (check_growing (_A5, _A7)
    (_B1, _B2, _B3, _B4, _B5, _B6, _B7) r)
                                      (fun _ ->
(let aaa =
   ta_of_ta
     ((ceq_term _A7 _B6), (ccompare_term _A5 _B5), (equal_term _A7 _B6),
       set_impl_term)
     (_A4.ccompare_cproper_interval, _A7) a
   in
  bindb (check
          (set_eq
            ((cenum_prod cenum_term cenum_term),
              (ceq_prod (ceq_term _A7 _B6) (ceq_term _A7 _B6)),
              (ccompare_prod (ccompare_term _A5 _B5) (ccompare_term _A5 _B5)))
            (ta_eps aaa)
            (set_empty
              ((ceq_prod (ceq_term _A7 _B6) (ceq_term _A7 _B6)),
                (ccompare_prod (ccompare_term _A5 _B5) (ccompare_term _A5 _B5)))
              (of_phantom (set_impl_proda set_impl_term set_impl_term))))
          (showsl_lit "no epsilon transitions allowed"))
    (fun _ ->
      bindb (check
              (member ((ceq_term _A7 _B6), (ccompare_term _A5 _B5)) (star c t)
                (ta_final aaa))
              (comp (comp (showsl_lit "final state for ")
                      (showsl_terma _A9 _B8 t))
                (showsl_lit " is missing")))
        (fun _ ->
          bindb (check
                  (less_eq_set
                    ((cenum_prod _A2 cenum_nat), (ceq_prod _A3 ceq_nat),
                      (ccompare_prod _A4.ccompare_cproper_interval
                        ccompare_nat))
                    (funas_ta ((ccompare_term _A5 _B5), (equal_term _A7 _B6))
                      (_A1, _A2, _A3, _A4, _A7, _A8) aaa)
                    fa)
                  (showsl_lit
                    "the given automaton does not respect the signature"))
            (fun _ ->
              (let ts = gi_rules_list f c t in
               let ms = mp_ta_rules _A7 _B6 r f c in
                bindb (check_rules_subseteq
                        ((equal_term _A7 _B6), (showl_term _A9 _B8)) (_A7, _A9)
                        ts a)
                  (fun _ ->
                    bindb (check_rules_subseteq
                            ((equal_term _A7 _B6), (showl_term _A9 _B8))
                            (_A7, _A9) ms a)
                      (fun _ ->
                        (let q =
                           add_rule_states (equal_term _A7 _B6) ts
                             (add_rule_states (equal_term _A7 _B6) ms [])
                           in
                         let ss =
                           state_substs
                             (remdups _B6 (maps (comp vars_term_list snd) r)) q
                           in
                         let d =
                           set ((ceq_ta_rule (equal_term _A7 _B6) _A7),
                                 (ccompare_ta_rule (ccompare_term _A5 _B5)
                                   _A4.ccompare_cproper_interval),
                                 set_impl_ta_rule)
                             (ta_rules_implb a)
                           in
                         let da =
                           inf_step (_A4.ccompare_cproper_interval, _A5, _A7)
                             (_B5, _B6) c r ss d
                           in
                          bindb (check
                                  (less_eq_set
                                    (cenum_ta_rule,
                                      (ceq_ta_rule (equal_term _A7 _B6) _A7),
                                      (ccompare_ta_rule (ccompare_term _A5 _B5)
_A4.ccompare_cproper_interval))
                                    da d)
                                  (showsl_lit
                                    "the given tree automaton is not closed under completion rules"))
                            (fun _ ->
                              check (ta_empty
                                      ((cenum_prod cenum_term cenum_term),
(ceq_prod (ceq_term _A7 _B6) (ceq_term _A7 _B6)),
(ccompare_prod (ccompare_term _A5 _B5) (ccompare_term _A5 _B5)),
(equal_prod (equal_term _A7 _B6) (equal_term _A7 _B6)),
(set_impl_prod set_impl_term set_impl_term))
                                      (_A4.ccompare_cproper_interval, _A6, _A7)
                                      (ta_inter_eps_empty
((ceq_term _A7 _B6), (ccompare_term _A5 _B5), (equal_term _A7 _B6),
  set_impl_term)
(_A3, _A4.ccompare_cproper_interval, _A7, _A8)
((ceq_term _A7 _B6), (ccompare_term _A5 _B5), (equal_term _A7 _B6),
  set_impl_term)
aaa (ground_instances_ta (_A3, _A4.ccompare_cproper_interval, _A5, _A7)
      (_B5, _B6) fa c s)))
                                (showsl_lit
                                  "the given tree automaton does not certify non-reachability"))))))))))))))))))));;

let rec is_instance_rule _A (_B1, _B2, _B3)
  ra r =
    (match
      match_list (_B1, _B2, _B3) _A _B2 (fun a -> Var a)
        [(fst r, fst ra); (snd r, snd ra)]
      with None -> false | Some _ -> true);;

let rec check_subst_overapproximation (_A1, _A2) (_B1, _B2, _B3, _B4)
  ra r =
    catch_error
      (catch_error
        (forallM
          (fun raa ->
            catch_error
              (existsM
                (fun rb ->
                  check (is_instance_rule _A1 (_B1, _B2, _B3) raa rb) id)
                r)
              (fun _ ->
                Inl (comp (comp (showsl_lit "growing rule for ")
                            (showsl_rule _A2 _B4 raa))
                      (showsl_lit " is missing\092"))))
          ra)
        (fun x -> Inl (snd x)))
      (fun x ->
        Inl (comp (comp (comp (comp (showsl_trs _A2 _B4 r)
                                (showsl_lit
                                  "\092is not an overapproximation of\092"))
                          (showsl_trs _A2 _B4 ra))
                    (showsl_literal "\092"))
              x));;

let rec group_key _B
  f x1 = match f, x1 with f, [] -> []
    | f, x :: xs ->
        (x :: takeWhile (fun y -> eq _B (f x) (f y)) xs) ::
          group_key _B f (dropWhile (fun y -> eq _B (f x) (f y)) xs);;

let rec nonlinear_var_nonreach (_A1, _A2, _A3, _A4, _A5) (_B1, _B2) (_C1, _C2)
  f gt_fun xs =
    (let xs1 = filtera (comp is_Var fst) xs in
     let xs2 = sort_key (linorder_term _B1 _C1) fst xs1 in
     let a = group_key (equal_term _B2 _C2) fst xs2 in
      list_ex
        (fun xts ->
          less_nat one_nata (size_list xts) &&
            list_all
              (fun fa ->
                not (gt_fun (Some fa) None) &&
                  list_ex
                    (fun (_, v) ->
                      not (is_Var v) &&
                        (less_eq_set
                           ((cenum_prod _A1 cenum_nat), (ceq_prod _A2 ceq_nat),
                             (ccompare_prod _A3 ccompare_nat))
                           (funas_term (_A2, _A3, _A5) v)
                           (set ((ceq_prod _A2 ceq_nat),
                                  (ccompare_prod _A3 ccompare_nat),
                                  (set_impl_prod _A5 set_impl_nat))
                             f) &&
                          (not (equal_optiona (equal_prod _A4 equal_nat)
                                 (root v) (Some fa)) &&
                            not (gt_fun (Some fa) (root v)))))
                    xts)
              f)
        a);;

let rec nonreachable_gtcapRM (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2)
  fs nlv ne gt_fun rm s t =
    (let gt = gt_term _A5 _B2 nlv ne gt_fun rm in
     let rs = rd_impl _A5 gt (s, t) in
      list_ex (fun (a, b) -> nonreach _A5 gt a b) rs ||
        nlv &&
          (ne &&
            nonlinear_var_nonreach (_A1, _A2, _A3, _A5, _A6) (_A4, _A5)
              (_B1, _B2) fs gt_fun rs));;

let rec insert_value _A _B
  key v m =
    (match key v with None -> m
      | Some k ->
        (match lookup _B m k with None -> insert _B k [v] m
          | Some vs -> insert _B k (insertb _A v vs) m));;

let rec insert_values _A _B
  uu x1 m = match uu, x1, m with uu, [], m -> m
    | key, v :: vs, m ->
        insert_value _A _B key v (insert_values _A _B key vs m);;

let rec insert_rules _A (_B1, _B2) _C
  a rs =
    insert_values
      (equal_prod _A (equal_prod (equal_term _B2 _C) (equal_term _B2 _C)))
      (compare_order_prod _B1 compare_order_nat) key
      (map (fun b -> (a, b)) rs);;

let rec rule_map (_A1, _A2) _B
  r fn =
    (match
      lookup (compare_order_prod _A1 compare_order_nat)
        (insert_rules equal_unit (_A1, _A2) _B () r
          (empty (linorder_prod _A1.linorder_compare_order linorder_nat)))
        fn
      with None -> [] | Some a -> map snd a);;

let rec relpow_impl
  succ un memb newa have m =
    (if equal_nata m zero_nata then un newa have
      else (if null newa then have
             else (let maybe = succ newa in
                   let havea = un newa have in
                   let newb = filtera (fun n -> not (memb n havea)) maybe in
                    relpow_impl succ un memb newb havea
                      (minus_nata m one_nata))));;

let rec trancl_impl
  gen_succ un memb emp rel =
    (let succ = gen_succ rel in
     let n = size_list rel in
      (fun asa -> relpow_impl succ un memb (succ asa) emp n));;

let rec trancl_list_impl _A
  = trancl_impl
      (fun r asa ->
        remdups _A
          (map_filter
            (fun x ->
              (if (let (a, _) = x in membera _A asa a) then Some (snd x)
                else None))
            r))
      (fun xs ys -> filtera (fun x -> not (membera _A ys x)) xs @ ys)
      (fun x xs -> membera _A xs x) [];;

let rec memo_list_trancl _A
  r = (let tr = trancl_list_impl _A r in
       let rm = map (fun a -> (a, tr [a])) (comp (remdups _A) (map fst) r) in
        (fun a -> (match map_of _A rm a with None -> [] | Some asa -> asa)));;

let rec mk_gt_fun _A
  rs = (let in_trancl =
          memo_list_trancl (equal_option (equal_prod _A equal_nat)) (gt1 rs) in
         (fun f ->
           membera (equal_option (equal_prod _A equal_nat)) (in_trancl f)));;

let rec check_nonreachable (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9)
  r s t x3 = match r, s, t, x3 with
    r, s, t, Nonreachable_Tcap ->
      check (not (matchc _A7 _B6 (tcapI _A7 _B6 r s) t))
        (showsl_lit "could not show nonreachability via tcap")
    | r, s, t, Nonreachable_Gtcap ->
        (let nlv = list_all (fun lr -> not (is_Var (fst lr))) r in
         let fs = funas_trs_list r in
          check (not (matchc _A7 _B6 (tcapI _A7 _B6 r s) t) ||
                  nonreachable_gtcapRM
                    (_A2, _A3, _A4.ccompare_cproper_interval,
                      _A5.compare_compare_order, _A7, _A8)
                    (_B5.compare_compare_order, _B6) fs nlv (not (null r))
                    (mk_gt_fun _A7 r) (rule_map (_A5, _A7) _B6 r) s t)
            (showsl_lit "could not show nonreachability via generalized tcap"))
    | r, s, t, Nonreachable_ETAC (f, aa, c, a) ->
        check_etac_nonreachable
          (_A1, _A2, _A3, _A4, _A5.compare_compare_order, _A6, _A7, _A8, _A9)
          (_B1, _B2, _B3, _B4, _B5.compare_compare_order, _B6, _B8, _B9) f aa c
          a r s t
    | ra, s, t, Nonreachable_Subst_Approx (r, p) ->
        bindb (check_subst_overapproximation (_A7, _A9)
                (_B4.ccompare_cproper_interval, _B6, _B7, _B9) ra r)
          (fun _ ->
            check_nonreachable (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
              (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9) r s t p)
    | r, s, t, Nonreachable_Reverse p ->
        check_nonreachable (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
          (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9)
          (map (fun (x, y) -> (y, x)) r) t s p;;

let rec unifiable _A s t = not (is_none (mergea _A s t));;

let rec check_nonjoinable (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9)
  r s t x3 = match r, s, t, x3 with
    r, s, t, Nonjoinable_Tcap ->
      check (not (unifiable _A7 (tcapI _A7 _B6 r s) (tcapI _A7 _B6 r t)))
        (showsl_lit "could not show nonjoinability via tcap")
    | r, s, t, Nonjoinable_Ground_NF p ->
        (if is_NF_trs (_A5, _A7) (_B4.ccompare_cproper_interval, _B6, _B7) r
              s &&
              ground s
          then check_nonreachable (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
                 (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9) r t s p
          else (if is_NF_trs (_A5, _A7)
                     (_B4.ccompare_cproper_interval, _B6, _B7) r t &&
                     ground t
                 then check_nonreachable
                        (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
                        (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9) r s t p
                 else Inl (showsl_lit "non NF")));;

let rec check_infeasibleb (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9)
  r cs x2 = match r, cs, x2 with
    r, cs, Infeasible_Compound_Conditions (f, p) ->
      check_nonreachable (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
        (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9) (map fst r)
        (Fun (f, map fst cs)) (Fun (f, map snd cs)) p
    | r, cs, Infeasible_Equation (s, t, p) ->
        bindb (check
                (membera (equal_prod (equal_term _A7 _B6) (equal_term _A7 _B6))
                  cs (s, t))
                (comp (comp (showsl_lit "equation ")
                        (showsl_eqa _A9 _B9 (s, t)))
                  (showsl_lit " is not in list of conditions\092")))
          (fun _ ->
            check_nonreachable (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
              (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9) (map fst r) s t p)
    | r, csa, Infeasible_Subset (cs, p) ->
        bindb (catch_error
                (check_subseteq
                  (equal_prod (equal_term _A7 _B6) (equal_term _A7 _B6)) cs csa)
                (fun x ->
                  Inl (comp (comp (showsl_lit "equation ")
                              (showsl_eqa _A9 _B9 x))
                        (showsl_lit " is not in list of conditions\092"))))
          (fun _ ->
            check_infeasibleb (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
              (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9) r cs p)
    | r, cs, Infeasible_Rhss_Equal (s, t, u, p) ->
        bindb (check
                (membera (equal_prod (equal_term _A7 _B6) (equal_term _A7 _B6))
                  cs (s, u))
                (comp (comp (showsl_lit "equation ")
                        (showsl_eqa _A9 _B9 (s, u)))
                  (showsl_lit " is not in list of conditions\092")))
          (fun _ ->
            bindb (check
                    (membera
                      (equal_prod (equal_term _A7 _B6) (equal_term _A7 _B6)) cs
                      (t, u))
                    (comp (comp (showsl_lit "equation ")
                            (showsl_eqa _A9 _B9 (t, u)))
                      (showsl_lit " is not in list of conditions\092")))
              (fun _ ->
                check_nonjoinable (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
                  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9) (map fst r) s t
                  p))
    | r, cs, Infeasible_Trans (s, t, u, p) ->
        bindb (check
                (membera (equal_prod (equal_term _A7 _B6) (equal_term _A7 _B6))
                  cs (s, t))
                (comp (comp (showsl_lit "equation ")
                        (showsl_eqa _A9 _B9 (s, t)))
                  (showsl_lit " is not in list of conditions\092")))
          (fun _ ->
            bindb (check
                    (membera
                      (equal_prod (equal_term _A7 _B6) (equal_term _A7 _B6)) cs
                      (t, u))
                    (comp (comp (showsl_lit "equation ")
                            (showsl_eqa _A9 _B9 (t, u)))
                      (showsl_lit " is not in list of conditions\092")))
              (fun _ ->
                check_nonreachable (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
                  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9) (map fst r) s u
                  p));;

let rec check_infeasible (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9)
  css r cs =
    catch_error
      (existsM
        (fun (csa, p) ->
          bindb (check (equal_nata (size_list csa) (size_list cs))
                  (showsl_lit "lengths differ"))
            (fun _ ->
              bindb (check
                      (not (is_none
                             (match_rules _A7
                               (_B4.ccompare_cproper_interval, _B6, _B7) cs
                               csa)) &&
                        not (is_none
                              (match_rules _A7
                                (_B4.ccompare_cproper_interval, _B6, _B7) csa
                                cs)))
                      id)
                (fun _ ->
                  check_infeasibleb
                    (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
                    (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9) r csa p)))
        css)
      (fun x -> Inl (showsl_sep id (showsl_literal "\092") x));;

let rec check_CCPs (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
  cj css uo r =
    catch_error
      (forallM
        (fun rho_1 ->
          (let l_1 = fst (fst rho_1) in
            catch_error
              (forallM
                (fun rho_2 ->
                  catch_error
                    (forallM
                      (check_overlap
                        ((ccompare_list ccompare_char), (equal_list equal_char),
                          mapping_impl_list, infinite_list,
                          (showl_list showl_char))
                        (_A7, _A9)
                        (fun a ->
                          Chara (false, false, false, true, true, true, true,
                                  false) ::
                            a)
                        (fun a ->
                          Chara (true, false, false, true, true, true, true,
                                  false) ::
                            a)
                        (check_context_joinable (_A5, _A7, _A9)
                          (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
                            (cproper_interval_list ccompare_char),
                            (compare_list compare_char),
                            (equal_list equal_char), mapping_impl_list,
                            infinite_list, set_impl_list,
                            (showl_list showl_char))
                          cj)
                        (check_infeasible
                          (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
                          (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
                            (cproper_interval_list ccompare_char),
                            (compare_order_list
                              (compare_order_char, equal_char)),
                            (equal_list equal_char), mapping_impl_list,
                            set_impl_list, (showl_list showl_char))
                          css)
                        (check_unfeasible (_A5, _A7, _A9) uo) r rho_1 rho_2)
                      (funposs_list l_1))
                    (fun x -> Inl (snd x)))
                r)
              (fun x -> Inl (snd x))))
        r)
      (fun x -> Inl (snd x));;

let rec check_type3 _A (_B1, _B2)
  r = catch_error
        (catch_error
          (forallM
            (fun cr ->
              catch_error
                (check_subseteq _B1 (vars_term_list (snd (fst cr)))
                  (vars_term_list (fst (fst cr)) @ vars_trs_list (snd cr)))
                (fun x ->
                  Inl (comp (comp (comp (comp (showsl_lit "variable ")
  (showsl _B2 x))
                                    (showsl_lit
                                      " occurs only in right-hand side of rule "))
                              (showsl_crule _A _B2 _A _B2 cr))
                        (showsl_literal "\092"))))
            r)
          (fun x -> Inl (snd x)))
        (fun x -> Inl (comp (showsl_lit "the CTRS is not of type 3\092") x));;

let rec x_impl
  cr i =
    concat
      (vars_term_list (fst (fst cr)) ::
        map (comp vars_term_list snd) (take i (snd cr)));;

let rec check_dctrs _A (_B1, _B2)
  r = catch_error
        (catch_error
          (forallM
            (fun cr ->
              catch_error
                (forallM
                  (fun i ->
                    catch_error
                      (check_subseteq _B1
                        (vars_term_list (fst (nth (snd cr) i))) (x_impl cr i))
                      (fun x ->
                        Inl (comp (comp (comp
  (comp (comp (comp (showsl_lit "variable ") (showsl _B2 x))
          (showsl_lit " in condition "))
    (showsl_rule _A _B2 (nth (snd cr) i)))
  (showsl_lit " of rule "))
                                    (showsl_crule _A _B2 _A _B2 cr))
                              (showsl_lit "violates DCTRS condition\092"))))
                  (upt zero_nata (size_list (snd cr))))
                (fun x -> Inl (snd x)))
            r)
          (fun x -> Inl (snd x)))
        (fun x ->
          Inl (comp (showsl_lit "the CTRS is not deterministic\092") x));;

let rec check_wf_ctrs _A (_B1, _B2)
  r = catch_error
        (bindb (check_varcond_no_Var_lhs _A _B2 (map fst r))
          (fun _ ->
            bindb (check_dctrs _A (_B1, _B2) r)
              (fun _ -> check_type3 _A (_B1, _B2) r)))
        (fun x -> Inl (comp (showsl_lit "the CTRS is not well-formed\092") x));;

let rec check_adtrs (_A1, _A2)
  r = catch_error
        (catch_error
          (forallM
            (fun cr ->
              catch_error
                (forallM
                  (fun i ->
                    (let a = snd (nth (snd cr) i) in check_airr (_A1, _A2) r a))
                  (upt zero_nata (size_list (snd cr))))
                (fun x -> Inl (snd x)))
            r)
          (fun x -> Inl (snd x)))
        (fun x ->
          Inl (comp (showsl_lit "the CTRS is not absolutely deterministic\092")
                x));;

let rec check_al94 (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
  cj css uo r =
    catch_error
      (bindb
        (check_wf_ctrs _A9 ((equal_list equal_char), (showl_list showl_char)) r)
        (fun _ ->
          bindb (check_adtrs (_A7, _A9) r)
            (fun _ ->
              check_CCPs (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9) cj css uo
                r)))
      (fun x ->
        Inl (comp (showsl_lit "Avenhaus & Loria-Saenz 1994 does not apply\092")
              x));;

let rec orig_crule
  m r = ((orig_term m (fst (fst r)), orig_term m (snd (fst r))),
          map (fun (s, t) -> (orig_term m s, orig_term m t)) (snd r));;

let rec orig_cstep
  m (Cstep_step (rho, p, sigma, s, t, css)) =
    Cstep_step
      (orig_crule m rho, p, comp (orig_term m) sigma, orig_term m s,
        orig_term m t, map (map (orig_cstep m)) css);;

let rec arith_fun
  xml = xml_do "arithFunction"
          (xml_take
            (xml_or
              (xml_change (xml_nat "natural") (fun n -> xml_return (Const n)))
              (xml_or
                (xml_change (xml_nat "variable")
                  (fun n -> xml_return (Arg (minus_nata n one_nata))))
                (xml_or
                  (xml_do "sum"
                    (xml_take_many_sub [] zero_nata Infinity_enat arith_fun
                      (fun asa -> xml_return (Sum asa))))
                  (xml_or
                    (xml_do "product"
                      (xml_take_many_sub [] zero_nata Infinity_enat arith_fun
                        (fun asa -> xml_return (Prod asa))))
                    (xml_or
                      (xml_do "max"
                        (xml_take_many_sub [] zero_nata Infinity_enat arith_fun
                          (fun asa -> xml_return (Max asa))))
                      (xml_or
                        (xml_do "min"
                          (xml_take_many_sub [] zero_nata Infinity_enat
                            arith_fun (fun a -> xml_return (Min a))))
                        (xml_do "ifEqual"
                          (xml_take arith_fun
                            (fun a ->
                              xml_take arith_fun
                                (fun b ->
                                  xml_take arith_fun
                                    (fun c ->
                                      xml_take arith_fun
(fun d -> xml_return (IfEqual (a, b, c, d))))))))))))))
            xml_return)
          xml;;

let rec int_coeff x = xml_int "integer" x;;

let rec vec_coeff
  xml2coeff =
    xml_do "vector"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "coefficient" (xml_take xml2coeff (fun x -> xml_return (id x))))
        (fun a -> xml_return (vec_of_list a)));;

let rec mat_coeff
  n ze xml2coeff =
    xml_or
      (xml_do "matrix"
        (xml_take_many_sub [] zero_nata Infinity_enat (vec_coeff xml2coeff)
          (fun a -> xml_return (mat_of_cols n a))))
      (xml_change (vec_coeff xml2coeff)
        (fun v ->
          xml_return
            (mat n n
              (fun (i, j) ->
                (if equal_nata j zero_nata then vec_index v i else ze)))));;

let rec closed_criterion
  x = xml_do "criterion"
        (xml_take
          (xml_or (xml_leaf "compatibility" Id_Relation)
            (xml_or
              (xml_do "stateCompatibility"
                (xml_take
                  (xml_do "relation"
                    (xml_take_many_sub [] zero_nata Infinity_enat
                      (xml_do "entry"
                        (xml_take state
                          (fun a ->
                            xml_take state (fun b -> xml_return (a, b)))))
                      xml_return))
                  (fun xa -> xml_return (Some_Relation xa))))
              (xml_or (xml_leaf "decisionProcedure" Decision_Proc)
                (xml_leaf "decisionProcedureOld" Decision_Proc_Old))))
          xml_return)
        x;;

let rec final_states
  x = xml_do "finalStates"
        (xml_take_many_sub [] zero_nata Infinity_enat state xml_return) x;;

let rec transition
  xml2lhs =
    xml_do "transition"
      (xml_take xml2lhs
        (fun a ->
          xml_take (xml_do "rhs" (xml_take state (fun x -> xml_return (id x))))
            (fun b ->
              xml_return
                (match a with Inl (f, qs) -> Inl (TA_rule (f, qs, b))
                  | Inr q -> Inr (q, b)))));;

let rec transitions
  xml2lhs =
    xml_do "transitions"
      (xml_take_many_sub [] zero_nata Infinity_enat (transition xml2lhs)
        (fun rls ->
          (let (rules, eps) =
             partition (fun a -> (match a with Inl _ -> true | Inr _ -> false))
               rls
             in
           let ruls = map (fun (Inl r) -> r) rules in
           let ep = map (fun (Inr e) -> e) eps in
            xml_return (ruls, ep))));;

let rec tree_automaton
  xml2lhs =
    xml_do "treeAutomaton"
      (xml_take final_states
        (fun a ->
          xml_take (transitions xml2lhs)
            (fun b ->
              xml_return (let (ba, c) = b in Tree_Automaton (a, ba, c)))));;

let rec ta_normal_lhs
  xml2name =
    xml_do "lhs"
      (xml_or (xml_take state (fun a -> xml_return (Inr a)))
        (xml_take xml2name
          (fun a ->
            xml_take_many_sub [] zero_nata Infinity_enat state
              (fun b -> xml_return (Inl (a, b))))));;

let rec not_wn_ta (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "notWNTreeAutomaton"
      (xml_take (tree_automaton (ta_normal_lhs xml2name))
        (fun a ->
          xml_take closed_criterion
            (fun b -> xml_return (Not_wn_ta_prf (a, b)))));;

let rec precedence_weight_ac _A
  xml2name =
    xml_do "precedenceWeight"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "precedenceWeightEntry"
          (xml_take xml2name
            (fun f ->
              xml_take (xml_nat "arity")
                (fun a ->
                  xml_take (xml_nat "precedence")
                    (fun p ->
                      xml_take (xml_nat "weight")
                        (fun w ->
                          xml_take_default false (xml_bool "isAC")
                            (fun e -> xml_return ((f, a), (p, (w, e))))))))))
        (fun ret -> xml_return (fun a -> (ret, a))));;

let rec ac_knuth_bendix_order _A
  xml2name =
    xml_do "ACKBO"
      (xml_take (xml_nat "w0")
        (fun w0 ->
          xml_take (precedence_weight_ac _A xml2name)
            (fun prw ->
              xml_take_default [] (afs _A xml2name)
                (fun afsa -> xml_return (ACKBO (prw w0, afsa))))));;

let rec precedence_weight _A
  xml2name =
    xml_do "precedenceWeight"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "precedenceWeightEntry"
          (xml_take xml2name
            (fun f ->
              xml_take (xml_nat "arity")
                (fun a ->
                  xml_take (xml_nat "precedence")
                    (fun p ->
                      xml_take (xml_nat "weight")
                        (fun w ->
                          xml_take_optional
                            (xml_do "subtermCoefficientEntries"
                              (xml_take_many_sub [] zero_nata Infinity_enat
                                (xml_nat "entry") xml_return))
                            (fun e -> xml_return ((f, a), (p, (w, e))))))))))
        (fun ret -> xml_return (fun a -> (ret, a))));;

let rec knuth_bendix_order _A
  xml2name =
    xml_do "knuthBendixOrder"
      (xml_take (xml_nat "w0")
        (fun w0 ->
          xml_take (precedence_weight _A xml2name)
            (fun prw ->
              xml_take_default [] (afs _A xml2name)
                (fun afsa -> xml_return (KBO (prw w0, afsa))))));;

let rec arctic_rat_linear_poly (Arctic_rat_linear_poly x4) = x4;;

let rec real_non_linear_poly (Real_non_linear_poly x13) = x13;;

let rec rat_non_linear_poly (Rat_non_linear_poly x12) = x12;;

let rec int_non_linear_poly (Int_non_linear_poly x11) = x11;;

let rec arctic_linear_poly (Arctic_linear_poly x3) = x3;;

let rec arctic_rat_matrix (Arctic_rat_matrix x9) = x9;;

let rec real_linear_poly (Real_linear_poly x5) = x5;;

let rec rat_linear_poly (Rat_linear_poly x2) = x2;;

let rec int_linear_poly (Int_linear_poly x1) = x1;;

let rec arctic_matrix (Arctic_matrix x8) = x8;;

let rec real_matrix (Real_matrix x10) = x10;;

let rec rat_matrix (Rat_matrix x7) = x7;;

let rec int_matrix (Int_matrix x6) = x6;;

let rec class_semiring (_A1, _A2, _A3, _A4, _A5, _A6, _A7)
  uu b =
    Partial_object_ext
      (top_set (_A1, _A2, _A7),
        Monoid_ext (times _A5, onea _A3, Ring_ext (zerob _A6, plus _A4, b)));;

let rec real_domain
  x = xml_do "algebraicNumbers"
        (xml_take
          (xml_do "delta" (xml_take xml_real (fun xa -> xml_return (id xa))))
          (fun xa -> xml_return (id xa)))
        x;;

let rec rat_domain
  x = xml_do "rationals"
        (xml_take
          (xml_do "delta" (xml_take xml_rat (fun xa -> xml_return (id xa))))
          (fun xa -> xml_return (id xa)))
        x;;

let rec basic_domain
  x = xml_or (xml_leaf "naturals" (fun a -> Natural a))
        (xml_or (xml_leaf "integers" (fun _ -> Integer))
          (xml_or
            (xml_change rat_domain
              (comp xml_return (fun a b -> Rational (a, b))))
            (xml_or
              (xml_change real_domain
                (comp xml_return (fun a b -> Mini_Alg (a, b))))
              (xml_do "arctic"
                (xml_take
                  (xml_do "domain"
                    (xml_take
                      (xml_or (xml_leaf "naturals" (fun _ -> Arctic))
                        (xml_or (xml_leaf "integers" (fun _ -> Arctic))
                          (xml_change rat_domain
                            (fun _ -> xml_return (fun _ -> Arctic_rat)))))
                      (fun xa -> xml_return (id xa))))
                  (fun xa -> xml_return (id xa)))))))
        x;;

let rec interpretation_type
  x = xml_do "type"
        (xml_take
          (xml_or
            (xml_do "matrixInterpretation"
              (xml_take
                (xml_do "domain"
                  (xml_take basic_domain (fun xa -> xml_return (id xa))))
                (fun a ->
                  xml_take (xml_nat "dimension")
                    (fun b ->
                      xml_take (xml_nat "strictDimension")
                        (fun c ->
                          xml_return
                            (match a zero_nata with Natural _ -> Int_mat (b, c)
                              | Integer -> Int_mat (b, c)
                              | Arctic -> Arctic_mat b
                              | Arctic_rat -> Arctic_rat_mat b
                              | Rational (_, _) -> Rat_mat (b, c)
                              | Mini_Alg (_, _) -> Mini_Alg_mat (b, c)))))))
            (xml_do "polynomial"
              (xml_take
                (xml_do "domain"
                  (xml_take
                    (xml_or basic_domain
                      (xml_do "matrices"
                        (xml_take (xml_nat "dimension")
                          (fun a ->
                            xml_take (xml_nat "strictDimension")
                              (fun b ->
                                xml_take
                                  (xml_do "domain"
                                    (xml_take basic_domain
                                      (fun xa -> xml_return (id xa))))
                                  (fun c ->
                                    xml_return
                                      (fun d ->
(match c d with Natural _ -> Int_mat (a, b) | Integer -> Int_mat (a, b)
  | Arctic -> Arctic_mat a | Arctic_rat -> Arctic_rat_mat a
  | Rational (_, _) -> Rat_mat (a, b)
  | Mini_Alg (_, _) -> Mini_Alg_mat (a, b)))))))))
                    (fun xa -> xml_return (id xa))))
                (fun a ->
                  xml_take (xml_nat "degree") (fun b -> xml_return (a b))))))
          (fun xa -> xml_return (id xa)))
        x;;

let rec arctic_rat_coeff
  x = xml_or (xml_change xml_rat (comp xml_return (fun a -> Num_arc_delta a)))
        (xml_leaf "minusInfinity" MinInfty_delta) x;;

let rec arctic_coeff
  x = xml_or (xml_change int_coeff (comp xml_return (fun a -> Num_arc a)))
        (xml_leaf "minusInfinity" MinInfty) x;;

let rec xml_changea f p x = bind2 (f p) (fun e -> xml_error e x) right;;

let rec polynomial
  xml2coeff x =
    xml_do "polynomial"
      (xml_take
        (xml_or
          (xml_do "coefficient"
            (xml_take xml2coeff (fun xa -> xml_return (PNum xa))))
          (xml_or
            (xml_change (xml_nat "variable")
              (fun n -> xml_return (PVar (minus_nata n one_nata))))
            (xml_or
              (xml_do "sum"
                (xml_take_many_sub [] zero_nata Infinity_enat
                  (polynomial xml2coeff) (fun a -> xml_return (PSum a))))
              (xml_do "product"
                (xml_take_many_sub [] zero_nata Infinity_enat
                  (polynomial xml2coeff) (fun a -> xml_return (PMult a)))))))
        (fun xa -> xml_return (id xa)))
      x;;

let rec fit_length
  c n uu =
    (if equal_nata n zero_nata then []
      else (match uu with [] -> replicate (suc (minus_nata n one_nata)) (zero c)
             | b :: bs -> b :: fit_length c (minus_nata n one_nata) bs));;

let rec interpretation
  bi xml2name =
    xml_do "interpretation"
      (xml_take interpretation_type
        (fun typea ->
          xml_take_many_sub [] zero_nata Infinity_enat
            (xml_do "interpret"
              (xml_take xml2name
                (fun f ->
                  xml_take (xml_nat "arity")
                    (fun a ->
                      (match typea
                        with Natural deg ->
                          (if less_eq_nat deg one_nata && not bi
                            then (let c =
                                    class_semiring
                                      (ceq_int, ccompare_int, one_int, plus_int,
times_int, zero_int, set_impl_int)
                                      Type ()
                                    in
                                   xml_take
                                     (xml_change (polynomial int_coeff)
                                       (xml_changea (lpoly_of equal_int c)))
                                     (fun poly ->
                                       xml_return
 (Int_linear_poly ((f, a), (fst poly, fit_length c a (snd poly))))))
                            else xml_take
                                   (xml_change (polynomial int_coeff)
                                     (comp xml_return
                                       (poly_of (equal_nat, linorder_nat)
 (equal_int, comm_semiring_1_int))))
                                   (fun poly ->
                                     xml_return
                                       (Int_non_linear_poly ((f, a), poly))))
                        | Integer ->
                          xml_take
                            (xml_change (polynomial int_coeff)
                              (comp xml_return
                                (poly_of (equal_nat, linorder_nat)
                                  (equal_int, comm_semiring_1_int))))
                            (fun poly ->
                              xml_return (Int_non_linear_poly ((f, a), poly)))
                        | Arctic ->
                          (let c =
                             class_semiring
                               (ceq_arctic, ccompare_arctic, one_arctic,
                                 plus_arctic, times_arctic, zero_arctic,
                                 set_impl_arctic)
                               Type ()
                             in
                            xml_take
                              (xml_change (polynomial arctic_coeff)
                                (xml_changea (lpoly_of equal_arctic c)))
                              (fun poly ->
                                xml_return
                                  (Arctic_linear_poly
                                    ((f, a),
                                      (fst poly, fit_length c a (snd poly))))))
                        | Arctic_rat ->
                          (let c =
                             class_semiring
                               ((ceq_arctic_delta equal_rat),
                                 (ccompare_arctic_delta compare_rat),
                                 (one_arctic_delta linordered_field_rat),
                                 (plus_arctic_delta linordered_field_rat),
                                 (times_arctic_delta linordered_field_rat),
                                 (zero_arctic_delta linordered_field_rat),
                                 set_impl_arctic_delta)
                               Type ()
                             in
                            xml_take
                              (xml_change (polynomial arctic_rat_coeff)
                                (xml_changea
                                  (lpoly_of (equal_arctic_delta equal_rat) c)))
                              (fun poly ->
                                xml_return
                                  (Arctic_rat_linear_poly
                                    ((f, a),
                                      (fst poly, fit_length c a (snd poly))))))
                        | Int_mat (n, _) ->
                          (let c = ring_mat semiring_1_int Type n () in
                            xml_take
                              (xml_change
                                (polynomial (mat_coeff n zero_inta int_coeff))
                                (xml_changea
                                  (lpoly_of (equal_mat equal_int) c)))
                              (fun poly ->
                                xml_return
                                  (Int_matrix
                                    ((f, a),
                                      (fst poly, fit_length c a (snd poly))))))
                        | Arctic_mat n ->
                          (let c = ring_mat semiring_1_arctic Type n () in
                            xml_take
                              (xml_change
                                (polynomial
                                  (mat_coeff n zero_arctica arctic_coeff))
                                (xml_changea
                                  (lpoly_of (equal_mat equal_arctic) c)))
                              (fun poly ->
                                xml_return
                                  (Arctic_matrix
                                    ((f, a),
                                      (fst poly, fit_length c a (snd poly))))))
                        | Arctic_rat_mat n ->
                          (let c =
                             ring_mat
                               (semiring_1_arctic_delta linordered_field_rat)
                               Type n ()
                             in
                            xml_take
                              (xml_change
                                (polynomial
                                  (mat_coeff n
                                    (zero_arctic_deltaa linordered_field_rat)
                                    arctic_rat_coeff))
                                (xml_changea
                                  (lpoly_of
                                    (equal_mat (equal_arctic_delta equal_rat))
                                    c)))
                              (fun poly ->
                                xml_return
                                  (Arctic_rat_matrix
                                    ((f, a),
                                      (fst poly, fit_length c a (snd poly))))))
                        | Rational (_, deg) ->
                          (if less_eq_nat deg one_nata && not bi
                            then (let c =
                                    class_semiring
                                      (ceq_rat, ccompare_rat, one_rat, plus_rat,
times_rat, zero_rat, set_impl_rat)
                                      Type ()
                                    in
                                   xml_take
                                     (xml_change (polynomial xml_rat)
                                       (xml_changea (lpoly_of equal_rat c)))
                                     (fun poly ->
                                       xml_return
 (Rat_linear_poly ((f, a), (fst poly, fit_length c a (snd poly))))))
                            else xml_take
                                   (xml_change (polynomial xml_rat)
                                     (comp xml_return
                                       (poly_of (equal_nat, linorder_nat)
 (equal_rat, comm_semiring_1_rat))))
                                   (fun poly ->
                                     xml_return
                                       (Rat_non_linear_poly ((f, a), poly))))
                        | Rat_mat (n, _) ->
                          (let c = ring_mat semiring_1_rat Type n () in
                            xml_take
                              (xml_change
                                (polynomial (mat_coeff n zero_rata xml_rat))
                                (xml_changea
                                  (lpoly_of (equal_mat equal_rat) c)))
                              (fun poly ->
                                xml_return
                                  (Rat_matrix
                                    ((f, a),
                                      (fst poly, fit_length c a (snd poly))))))
                        | Mini_Alg (_, deg) ->
                          (if less_eq_nat deg one_nata && not bi
                            then (let c =
                                    class_semiring
                                      (ceq_real, ccompare_real, one_real,
plus_real, times_real, zero_real, set_impl_real)
                                      Type ()
                                    in
                                   xml_take
                                     (xml_change (polynomial xml_real)
                                       (xml_changea (lpoly_of equal_real c)))
                                     (fun poly ->
                                       xml_return
 (Real_linear_poly ((f, a), (fst poly, fit_length c a (snd poly))))))
                            else xml_take
                                   (xml_change (polynomial xml_real)
                                     (comp xml_return
                                       (poly_of (equal_nat, linorder_nat)
 (equal_real, comm_semiring_1_real))))
                                   (fun poly ->
                                     xml_return
                                       (Real_non_linear_poly ((f, a), poly))))
                        | Mini_Alg_mat (n, _) ->
                          (let c = ring_mat semiring_1_real Type n () in
                            xml_take
                              (xml_change
                                (polynomial (mat_coeff n zero_reala xml_real))
                                (xml_changea
                                  (lpoly_of (equal_mat equal_real) c)))
                              (fun poly ->
                                xml_return
                                  (Real_matrix
                                    ((f, a),
                                      (fst poly,
fit_length c a (snd poly)))))))))))
            (fun pi ->
              xml_return
                (match typea
                  with Natural deg ->
                    (if less_eq_nat deg one_nata && not bi
                      then Int_carrier (map int_linear_poly pi)
                      else Int_nl_carrier (map int_non_linear_poly pi))
                  | Integer -> Int_nl_carrier (map int_non_linear_poly pi)
                  | Arctic -> Arctic_carrier (map arctic_linear_poly pi)
                  | Arctic_rat ->
                    Arctic_rat_carrier (map arctic_rat_linear_poly pi)
                  | Int_mat (n, sd) ->
                    Int_mat_carrier (n, sd, map int_matrix pi)
                  | Arctic_mat n -> Arctic_mat_carrier (n, map arctic_matrix pi)
                  | Arctic_rat_mat n ->
                    Arctic_rat_mat_carrier (n, map arctic_rat_matrix pi)
                  | Rational (d, deg) ->
                    (if less_eq_nat deg one_nata && not bi
                      then Rat_carrier (map rat_linear_poly pi)
                      else Rat_nl_carrier (d, map rat_non_linear_poly pi))
                  | Rat_mat (n, sd) ->
                    Rat_mat_carrier (n, sd, map rat_matrix pi)
                  | Mini_Alg (d, deg) ->
                    (if less_eq_nat deg one_nata && not bi
                      then Real_carrier (map real_linear_poly pi)
                      else Real_nl_carrier (d, map real_non_linear_poly pi))
                  | Mini_Alg_mat (n, sd) ->
                    Real_mat_carrier (n, sd, map real_matrix pi)))));;

let rec wpo_params _A
  xml2name =
    xml_do "precedenceStatus"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "precedenceStatusEntry"
          (xml_take xml2name
            (fun f ->
              xml_take (xml_nat "arity")
                (fun a ->
                  xml_take (xml_nat "precedence")
                    (fun p ->
                      xml_take
                        (xml_do "status"
                          (xml_take_many_sub [] zero_nata Infinity_enat position
                            xml_return))
                        (fun s -> xml_return ((f, a), (p, s))))))))
        xml_return);;

let rec status_precedence _A
  xml2name =
    xml_do "statusPrecedence"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "statusPrecedenceEntry"
          (xml_take xml2name
            (fun f ->
              xml_take (xml_nat "arity")
                (fun a ->
                  xml_take (xml_nat "precedence")
                    (fun p ->
                      xml_take
                        (xml_or (xml_leaf "lex" Lex) (xml_leaf "mul" Mul))
                        (fun s -> xml_return ((f, a), (p, s))))))))
        xml_return);;

let rec path_order _A
  xml2name =
    xml_do "pathOrder"
      (xml_take (status_precedence _A xml2name)
        (fun prec_tau ->
          xml_take_default [] (afs _A xml2name)
            (fun af -> xml_return (RPO (prec_tau, af)))));;

let rec redtriple _A
  xml2name bi x =
    xml_or (path_order _A xml2name)
      (xml_or (knuth_bendix_order _A xml2name)
        (xml_or (ac_knuth_bendix_order _A xml2name)
          (xml_or
            (xml_do "weightedPathOrder"
              (xml_take (wpo_params _A xml2name)
                (fun a ->
                  xml_take (redtriple _A xml2name bi)
                    (fun b -> xml_return (WPO (a, b))))))
            (interpretation bi xml2name))))
      x;;

let rec joinable_critical_pairs _A
  xml2name =
    (let rew = rsteps _A xml2name in
      xml_do "joinableCriticalPairs"
        (xml_take_many_sub [] zero_nata Infinity_enat
          (xml_do "joinableCriticalPair"
            (xml_take rew
              (fun (s, sseq) ->
                xml_take rew
                  (fun (t, tseq) -> xml_return (s, (sseq, (t, tseq)))))))
          xml_return));;

let rec wcr_proof _A
  xml2name =
    xml_do "wcrProof"
      (xml_take
        (xml_or
          (xml_change (joinable_critical_pairs _A xml2name)
            (comp xml_return (fun a -> Guided a)))
          (xml_or
            (xml_change (xml_nat "joinableCriticalPairsBFS")
              (comp xml_return (fun a -> Join_BFS a)))
            (xml_leaf "joinableCriticalPairsAuto" Join_NF)))
        (fun x -> xml_return (id x)));;

let rec exp_parsera
  v xml =
    xml_or
      (xml_do "product"
        (xml_take_many_sub [] zero_nata Infinity_enat (exp_parsera v)
          (fun exps -> xml_return (Fun (ProdF (size_list exps), exps)))))
      (xml_or
        (xml_do "sum"
          (xml_take_many_sub [] zero_nata Infinity_enat (exp_parsera v)
            (fun exps -> xml_return (Fun (SumF (size_list exps), exps)))))
        (xml_or
          (xml_change (xml_int "constant")
            (fun i -> xml_return (Fun (ConstF i, []))))
          (xml_change v (fun x -> xml_return (Var (x, IntT))))))
      xml;;

let rec bexp_parser
  v = xml_or
        (xml_do "leq"
          (xml_take (exp_parsera v)
            (fun l ->
              xml_take (exp_parsera v)
                (fun r -> xml_return (Fun (LeF, [l; r]))))))
        (xml_or
          (xml_do "less"
            (xml_take (exp_parsera v)
              (fun l ->
                xml_take (exp_parsera v)
                  (fun r -> xml_return (Fun (LessF, [l; r]))))))
          (xml_do "eq"
            (xml_take (exp_parsera v)
              (fun l ->
                xml_take (exp_parsera v)
                  (fun r -> xml_return (Fun (EqF, [l; r])))))));;

let rec variable_parser x = xml_text "variableId" x;;

let rec trans_var_parser
  x = xml_or
        (xml_do "post"
          (xml_take variable_parser (fun v -> xml_return (Post v))))
        (xml_or
          (xml_change (xml_text "aux")
            (comp xml_return (fun a -> Intermediate a)))
          (xml_change variable_parser (comp xml_return (fun a -> Pre a))))
        x;;

let rec tatom_parsera
  x = xml_change (bexp_parser trans_var_parser)
        (comp xml_return (fun a -> Atom a)) x;;

let rec formula_parser
  atom_parser x =
    xml_or
      (xml_do "disjunction"
        (xml_take_many_sub [] zero_nata Infinity_enat
          (formula_parser atom_parser) (fun fs -> xml_return (Disjunction fs))))
      (xml_or
        (xml_do "conjunction"
          (xml_take_many_sub [] zero_nata Infinity_enat
            (formula_parser atom_parser)
            (fun fs -> xml_return (Conjunction fs))))
        atom_parser)
      x;;

let rec transition_parser
  location_parser trans_parser tatom_parser =
    xml_do "transition"
      (xml_take trans_parser
        (fun tr ->
          xml_take (xml_do "source" (xml_take location_parser xml_return))
            (fun l ->
              xml_take (xml_do "target" (xml_take location_parser xml_return))
                (fun r ->
                  xml_take
                    (xml_do "formula"
                      (xml_take (formula_parser tatom_parser) xml_return))
                    (fun phi -> xml_return (tr, Transition (l, r, phi)))))));;

let rec lts_parser
  location_parser trans_parser tatom_parser tag =
    xml_do tag
      (xml_take
        (xml_do "initial"
          (xml_take_many_sub [] one_nata Infinity_enat location_parser
            xml_return))
        (fun i ->
          xml_take_many_sub [] zero_nata Infinity_enat
            (transition_parser location_parser trans_parser tatom_parser)
            (fun t -> xml_return (Lts_Impl (i, t, [])))));;

let rec safety_input_parser
  location_parser trans_parser tatom_parser =
    xml_do "ltsSafetyInput"
      (xml_take (lts_parser location_parser trans_parser tatom_parser "lts")
        (fun lts ->
          xml_take
            (xml_do "error"
              (xml_take_many_sub [] one_nata Infinity_enat location_parser
                xml_return))
            (fun err -> xml_return (lts, err))));;

let rec location_parser location_id_parser = location_id_parser "locationId";;

let rec lts_safety_input_parser
  x = safety_input_parser (location_parser xml_text) (trans_id xml_text)
        tatom_parsera x;;

let rec precedence_weighta _A
  xml2name =
    xml_do "precedenceWeight"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "precedenceWeightEntry"
          (xml_take xml2name
            (fun f ->
              xml_take (xml_nat "arity")
                (fun a ->
                  xml_take (xml_nat "precedence")
                    (fun p ->
                      xml_take (xml_nat "weight")
                        (fun w ->
                          xml_take_optional
                            (xml_do "subtermCoefficientEntries"
                              (xml_take_many_sub [] zero_nata Infinity_enat
                                (xml_nat "entry") xml_return))
                            (fun e -> xml_return ((f, a), (p, (w, e))))))))))
        (fun ret -> xml_return (fun a -> (ret, a))));;

let rec kbo_input _A
  xml2name =
    xml_do "knuthBendixOrder"
      (xml_take (xml_nat "w0")
        (fun w0 ->
          xml_take (precedence_weighta _A xml2name)
            (fun prw -> xml_return (KBO_Input (prw w0)))));;

let rec xml2ordered_completion_input _A
  xml2name =
    xml_do "orderedCompletionInput"
      (xml_take (xml_do "equations" (xml_take (rules _A xml2name) xml_return))
        (fun es0 ->
          xml_take (xml_do "trs" (xml_take (rules _A xml2name) xml_return))
            (fun rs ->
              xml_take
                (xml_do "equations" (xml_take (rules _A xml2name) xml_return))
                (fun es ->
                  xml_take
                    (xml_do "reductionOrder"
                      (xml_take (kbo_input _A xml2name) xml_return))
                    (fun ro -> xml_return (OCOMP_input (es0, es, rs, ro)))))));;

let rec lts_input_parser
  x = lts_parser (location_parser xml_text) (trans_id xml_text) tatom_parsera
        "lts" x;;

let rec xml2equational_input _A
  xml2name =
    xml_do "equationalReasoningInput"
      (xml_take (xml_do "equations" (xml_take (rules _A xml2name) xml_return))
        (fun eqs ->
          xml_take
            (xml_or
              (xml_do "equation"
                (xml_take (term _A xml2name)
                  (fun a ->
                    xml_take (term _A xml2name)
                      (fun b -> xml_return (Equation (a, b))))))
              (xml_do "inequality"
                (xml_take (term _A xml2name)
                  (fun a ->
                    xml_take (term _A xml2name)
                      (fun b -> xml_return (Inequality (a, b)))))))
            (fun goal -> xml_return (EQ_input (eqs, goal)))));;

let rec xml2completion_input _A
  xml2name =
    xml_do "completionInput"
      (xml_take (xml_do "equations" (xml_take (rules _A xml2name) xml_return))
        (fun a ->
          xml_take (xml_do "trs" (xml_take (rules _A xml2name) xml_return))
            (fun b -> xml_return (COMP_input (a, b)))));;

let rec cPFsignature _A xml2name = symbols _A xml2name "signature";;

let rec complexity_measure _A
  xml2name =
    xml_or
      (xml_do "derivationalComplexity"
        (xml_take (cPFsignature _A xml2name)
          (fun a -> xml_return (Derivational_Complexity a))))
      (xml_do "runtimeComplexity"
        (xml_take (cPFsignature _A xml2name)
          (fun a ->
            xml_take (cPFsignature _A xml2name)
              (fun b -> xml_return (Runtime_Complexity (a, b))))));;

let rec complexity_class
  x = xml_change (xml_nat "polynomial") (comp xml_return (fun a -> Comp_Poly a))
        x;;

let rec forbidden_pattern _A
  xml2name =
    xml_do "forbiddenPattern"
      (xml_take (term _A xml2name)
        (fun t ->
          xml_take pos
            (fun p ->
              xml_take
                (xml_or (xml_leaf "here" H)
                  (xml_or (xml_leaf "above" A)
                    (xml_or (xml_leaf "below" Ba) (xml_leaf "below" Ra))))
                (fun l ->
                  (if in_poss p t
                    then xml_return (ctxt_of_pos_term p t, (subt_at t p, l))
                    else xml_error "position does not exist in term")))));;

let rec forbidden_patterns _A
  xml2name =
    xml_do "forbiddenPatterns"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (forbidden_pattern _A xml2name) xml_return);;

let rec replacement_map _A
  xml2name =
    xml_do "contextSensitive"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "replacementMapEntry"
          (xml_take xml2name
            (fun f ->
              xml_take (xml_nat "arity")
                (fun a ->
                  xml_take_many_sub [] zero_nata Infinity_enat position
                    (fun is -> xml_return ((f, a), is))))))
        xml_return);;

let rec inn_fp_strategy _A
  xml2name =
    xml_do "strategy"
      (xml_take
        (xml_or (xml_leaf "innermost" (Inl Innermost))
          (xml_or (xml_leaf "outermost" (Inr Outermost))
            (xml_or
              (xml_change (replacement_map _A xml2name)
                (fun p -> xml_return (Inr (Context_Sensitive p))))
              (xml_or
                (xml_change (forbidden_patterns _A xml2name)
                  (fun p -> xml_return (Inr (Forbidden_Patterns p))))
                (xml_change (innermostLhss _A xml2name)
                  (fun q -> xml_return (Inl (Innermost_Q q))))))))
        xml_return);;

let rec xml2_trs_input _A
  xml2name =
    xml_do "trsInput"
      (xml_take (xml_do "trs" (xml_take (rules _A xml2name) xml_return))
        (fun r ->
          xml_take_default (Inl No_Strategy) (inn_fp_strategy _A xml2name)
            (fun str ->
              xml_take_default []
                (xml_do "relativeRules"
                  (xml_take (rules _A xml2name) xml_return))
                (fun rel ->
                  (match str
                    with Inl istrat ->
                      xml_return (Inn_TRS_input (istrat, r, rel, Full))
                    | Inr fpstrat ->
                      (if null rel then xml_return (FP_TRS_input (fpstrat, r))
                        else xml_error
                               "the combination of relative rules with strategies is only supported for innermost"))))));;

let rec xml2complexity_input _A
  xml2name =
    xml_do "complexityInput"
      (xml_take
        (xml_change (xml2_trs_input _A xml2name)
          (fun a ->
            (match a
              with DP_input (_, _, _, _) ->
                xml_error "require innermost trs at this point"
              | Inn_TRS_input (q, r, s, start) ->
                xml_return (q, (r, (s, start)))
              | CPX_input (_, _, _, _, _) ->
                xml_error "require innermost trs at this point"
              | COMP_input (_, _) ->
                xml_error "require innermost trs at this point"
              | OCOMP_input (_, _, _, _) ->
                xml_error "require innermost trs at this point"
              | EQ_input (_, _) ->
                xml_error "require innermost trs at this point"
              | FP_TRS_input (_, _) ->
                xml_error "require innermost trs at this point"
              | CTRS_input _ -> xml_error "require innermost trs at this point"
              | TA_input (_, _) ->
                xml_error "require innermost trs at this point"
              | AC_input (_, _, _) ->
                xml_error "require innermost trs at this point"
              | LTS_input _ -> xml_error "require innermost trs at this point"
              | LTS_safety_input (_, _) ->
                xml_error "require innermost trs at this point"
              | Unknown_input _ ->
                xml_error "require innermost trs at this point")))
        (fun (q, (r, (s, _))) ->
          xml_take (complexity_measure _A xml2name)
            (fun cm ->
              xml_take complexity_class
                (fun cc -> xml_return (CPX_input (q, r, s, cm, cc))))));;

let rec xml2unknown_input x = xml_text "unknownInput" x;;

let rec symbols_no_arity _A
  xml2name tagname =
    xml_do tagname
      (xml_take_many_sub [] zero_nata Infinity_enat xml2name xml_return);;

let rec xml2ac_tp_input _A
  xml2name =
    xml_do "acRewriteSystem"
      (xml_take
        (xml_do "trs"
          (xml_take (rules _A xml2name) (fun x -> xml_return (id x))))
        (fun a ->
          xml_take (symbols_no_arity _A xml2name "Asymbols")
            (fun b ->
              xml_take (symbols_no_arity _A xml2name "Csymbols")
                (fun c -> xml_return (AC_input (a, b, c))))));;

let rec xml2ctrs_input _A
  xml2name =
    xml_do "ctrsInput"
      (xml_take
        (xml_do "rules"
          (xml_take_many_sub [] zero_nata Infinity_enat
            (crule _A xml2name "rule") xml_return))
        (fun a -> xml_return (CTRS_input a)));;

let rec xml2dp_input _A
  xml2name =
    xml_do "dpInput"
      (xml_take
        (xml_do "trs"
          (xml_take (rules _A xml2name) (fun x -> xml_return (id x))))
        (fun r ->
          xml_take
            (xml_do "dps"
              (xml_take (rules _A xml2name) (fun x -> xml_return (id x))))
            (fun p ->
              xml_take_default No_Strategy (strategy _A xml2name)
                (fun s ->
                  xml_take (xml_bool "minimal")
                    (fun m -> xml_return (DP_input (m, p, s, r)))))));;

let rec xml2input _A
  xml2name =
    xml_do "input"
      (xml_take
        (xml_or (xml2dp_input _A xml2name)
          (xml_or (xml2completion_input _A xml2name)
            (xml_or (xml2ordered_completion_input _A xml2name)
              (xml_or (xml2_trs_input _A xml2name)
                (xml_or (xml2equational_input _A xml2name)
                  (xml_or (xml2complexity_input _A xml2name)
                    (xml_or (xml2ctrs_input _A xml2name)
                      (xml_or
                        (xml_do "treeAutomatonProblem"
                          (xml_take (tree_automaton (ta_normal_lhs xml2name))
                            (fun a ->
                              xml_take
                                (xml_do "trs"
                                  (xml_take (rules _A xml2name)
                                    (fun x -> xml_return (id x))))
                                (fun b -> xml_return (TA_input (a, b))))))
                        (xml_or (xml2ac_tp_input _A xml2name)
                          (xml_or
                            (xml_change lts_input_parser
                              (comp xml_return (fun a -> LTS_input a)))
                            (xml_or
                              (xml_change lts_safety_input_parser
                                (fun (x, y) ->
                                  xml_return (LTS_safety_input (x, y))))
                              (xml_change xml2unknown_input
                                (comp xml_return
                                  (fun a -> Unknown_input a))))))))))))))
        (fun x -> xml_return (id x)));;

let rec bind2a (Sumbot a) f g = (match a with Inl aa -> f aa | Inr aa -> g aa);;

let rec missing
  s x = comp (comp (comp (comp (showsl_literal "the ") (showsl_literal s))
                     (showsl_literal " "))
               x)
          (showsl_literal " is missing");;

let rec toomuch
  s x = comp (comp (comp (showsl_lit "superfluous ") (showsl_lit s))
               (showsl_lit " "))
          x;;

let rec array_get (Array a) n = nth a n;;

let rec array_set (Array aa) n a = Array (list_update aa n a);;

let rec as_length x = snd x;;

let rec array_shrink
  (Array a) sz =
    (if less_nat (size_list a) sz then failwith "undefined"
      else Array (take sz a));;

let rec array_length (Array a) = size_list a;;

let rec as_shrink
  s = (let a = s in
       let (aa, n) = a in
       let ab =
         (if less_eq_nat
               (times_nata (nat_of_integer (Big_int.big_int_of_int 128)) n)
               (array_length aa) &&
               less_nat (nat_of_integer (Big_int.big_int_of_int 4)) n
           then array_shrink aa n else aa)
         in
        (ab, n));;

let rec as_take
  m s = (let a = s in
         let (aa, n) = a in
          (if less_nat m n then as_shrink (aa, m) else (aa, n)));;

let rec as_top s = (let a = s in
                    let (aa, n) = a in
                     array_get aa (minus_nata n one_nata));;

let rec as_pop s = (let a = s in
                    let (aa, n) = a in
                     as_shrink (aa, minus_nata n one_nata));;

let rec as_get s i = (let a = s in
                      let (aa, _) = a in
                       array_get aa i);;

let rec whilea b c s = (if b s then whilea b c (c s) else s);;

let rec pop_tr _A
  s = (let (a, (aa, (ab, bb))) = s in
       let x = minus_nata (as_length aa) one_nata in
       let xa =
         (let (_, bc) =
            whilea
              (fun (xe, _) ->
                less_nat xe
                  (if equal_nata (plus_nata x one_nata) (as_length aa)
                    then as_length a else as_get aa (plus_nata x one_nata)))
              (fun (ac, bc) ->
                (suc ac, rbt_insert _A (as_get a ac) (uminus_inta one_inta) bc))
              (as_get aa x, ab)
            in
           bc)
         in
       let xb = as_take (as_top aa) a in
       let xc = as_pop aa in
        (xb, (xc, (xa, bb))));;

let rec gen_disjoint
  it1 mem2 s1 s2 = it1 s1 (fun x -> x) (fun x _ -> not (mem2 x s2)) true;;

let rec is_NF_subset is_Q_nf q = list_all (fun qa -> not (is_Q_nf qa)) q;;

let rec icap_impl_gen _A
  nf isQnf ls s sx x5 = match nf, isQnf, ls, s, sx, x5 with
    nf, isQnf, ls, s, sx, Var x ->
      (if nf && sx x then Var (Inr x) else Var (Inl ()))
    | nf, isQnf, ls, s, sx, Fun (f, ts) ->
        (let t = Fun (f, map (icap_impl_gen _A nf isQnf ls s sx) ts) in
          (if list_ex
                (fun l ->
                  (match mgu_class _A t l with None -> false
                    | Some mu ->
                      list_all
                        (fun u ->
                          isQnf (subst_apply_term
                                  (map_term (fun x -> x)
                                    (fun a ->
                                      Chara
(true, false, false, true, true, true, true, false) ::
a)
                                    u)
                                  mu))
                        (args l) &&
                        list_all (fun u -> isQnf (subst_apply_term u mu)) s))
                ls
            then Var (Inl ()) else t));;

let rec ins_rm_basic_ops _A x s = insert _A x () s;;

let rec g_from_list_aux_dflt_basic_oops_rm_basic_ops _A
  accs x1 = match accs, x1 with
    accs, x :: l ->
      g_from_list_aux_dflt_basic_oops_rm_basic_ops _A
        (ins_rm_basic_ops _A x accs) l
    | y, [] -> y;;

let rec empty_rm_basic_ops _A = (fun _ -> empty _A);;

let rec g_from_list_dflt_basic_oops_rm_basic_ops _A
  l = g_from_list_aux_dflt_basic_oops_rm_basic_ops _A
        (empty_rm_basic_ops _A.linorder_compare_order ()) l;;

let rec memb_rm_basic_ops _A x s = not (is_none (lookup _A s x));;

let rec ceta_set_of _A
  ps = (let tree = g_from_list_dflt_basic_oops_rm_basic_ops _A ps in
         (fun a -> memb_rm_basic_ops _A a tree));;

let rec icap_impl _A
  isnf r =
    (let ls = map fst r in
     let nf = is_NF_subset isnf ls in
     let ic = icap_impl_gen _A nf isnf ls in
      (fun s ->
        (let sa =
           map (map_term (fun x -> x)
                 (fun a ->
                   Chara (false, false, false, true, true, true, true, false) ::
                     a))
             s
           in
         let sx =
           ceta_set_of (compare_order_list (compare_order_char, equal_char))
             (maps vars_term_list sa)
           in
          (fun t ->
            ic sa sx
              (map_term (fun x -> x)
                (fun a ->
                  Chara (false, false, false, true, true, true, true, false) ::
                    a)
                t)))));;

let rec rep_subst_incr (Abs_subst_incr x) = x;;

let rec si_W xa = snd (snd (rep_subst_incr xa));;

let rec kbo_nstrict _A (_B1, _B2)
  pr w w0 least scf =
    (fun (s, t) ->
      check (snd (kbo_impl _B1 w w0 pr least scf s t))
        (comp (comp (comp (comp (showsl_literal "could not orient ")
                            (showsl_terma _A _B2 s))
                      (showsl_literal " >=KBO "))
                (showsl_terma _A _B2 t))
          (showsl_literal "\092")));;

let rec succ_transitions _D
  (Lts_Impl (i, ts, lc)) l =
    maps (fun (_, tau) -> (if eq _D (source tau) l then [tau] else [])) ts;;

let rec default_I _A
  def n =
    (one_monom linorder_nat, def) ::
      map (fun i ->
            (var_monom linorder_nat i,
              onea _A.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral))
        (upt zero_nata n);;

let rec proper_prefix_list
  = function [] -> []
    | i :: p -> [] :: map (fun a -> i :: a) (proper_prefix_list p);;

let rec prefix_list p = p :: proper_prefix_list p;;

let rec scnp_arity af = max_list (map (fun (_, a) -> size_list a) af);;

let rec check_supt (_A1, _A2) (_B1, _B2)
  s t = check (supt_impl _A1 _B1 s t)
          (comp (comp (showsl_terma _A2 _B2 t)
                  (showsl_literal " is not a proper subterm of "))
            (showsl_terma _A2 _B2 s));;

let rec sharp_term shp x1 = match shp, x1 with shp, Var x -> Var x
                     | shp, Fun (f, ss) -> Fun (shp f, ss);;

let rec compute_trancl (_A1, _A2, _A3, _A4)
  a r = (let b = imagea (_A2, _A3) (_A2, _A3, _A4) r a in
          (if less_eq_set (_A1, _A2, _A3) b (bot_set (_A2, _A3, _A4))
            then bot_set (_A2, _A3, _A4)
            else sup_seta (_A2, _A3) b
                   (compute_trancl (_A1, _A2, _A3, _A4) b
                     (filter ((ceq_prod _A2 _A2), (ccompare_prod _A3 _A3))
                       (fun ab ->
                         not (member (_A2, _A3) (fst ab) a) &&
                           not (member (_A2, _A3) (snd ab) b))
                       r))));;

let rec ta_res (_A1, _A2, _A3, _A4, _A5) (_B1, _B2)
  ta x1 = match ta, x1 with
    ta, Fun (f, ts) ->
      (let qs = map (ta_res (_A1, _A2, _A3, _A4, _A5) (_B1, _B2) ta) ts in
       let g = (f, size_list ts) in
        sup_seta (_A2, _A3)
          (image ((ceq_ta_rule _A4 _B2), (ccompare_ta_rule _A3 _B1))
            (_A2, _A3, _A5) r_rhs
            (filter ((ceq_ta_rule _A4 _B2), (ccompare_ta_rule _A3 _B1))
              (fun r ->
                equal_proda _B2 equal_nat (r_sym r) g &&
                  list_all (fun qq -> member (_A2, _A3) (snd qq) (fst qq))
                    (zip qs (r_lhs_states r)))
              (ta_rules ta)))
          (compute_trancl (_A1, _A2, _A3, _A5)
            (image ((ceq_ta_rule _A4 _B2), (ccompare_ta_rule _A3 _B1))
              (_A2, _A3, _A5) r_rhs
              (filter ((ceq_ta_rule _A4 _B2), (ccompare_ta_rule _A3 _B1))
                (fun r ->
                  equal_proda _B2 equal_nat (r_sym r) g &&
                    list_all (fun qq -> member (_A2, _A3) (snd qq) (fst qq))
                      (zip qs (r_lhs_states r)))
                (ta_rules ta)))
            (ta_eps ta)))
    | ta, Var q ->
        sup_seta (_A2, _A3) (inserta (_A2, _A3) q (bot_set (_A2, _A3, _A5)))
          (compute_trancl (_A1, _A2, _A3, _A5)
            (inserta (_A2, _A3) q (bot_set (_A2, _A3, _A5))) (ta_eps ta));;

let rec eq_rule_mod_vars _A (_B1, _B2, _B3)
  lr st =
    instance_rule _A _B2 (_B1, _B2, _B3) lr st &&
      instance_rule _A _B2 (_B1, _B2, _B3) st lr;;

let rec is_NF_terms (_A1, _A2) (_B1, _B2, _B3)
  q = is_NF_main (_A1, _A2) (_B1, _B2, _B3) (list_ex is_Var q) (null q)
        (term_map _A1 q);;

let rec generate_var
  i = Chara (false, false, false, true, true, true, true, false) ::
        shows_prec_nat zero_nata i [];;

let rec uncurry_term _A
  a sm t =
    (match unapp _A a t
      with (Var x, ts) -> apply_args a (Var x) (map (uncurry_term _A a sm) ts)
      | (Fun (f, ss), ts) ->
        (let n = size_list ss in
         let uss = map (uncurry_term _A a sm) ss in
         let uts = map (uncurry_term _A a sm) ts in
         let aa = aarity sm f n in
         let m = min ord_nat (size_list ts) aa in
         let fm = get_symbol sm f n m in
          apply_args a (Fun (fm, uss @ take m uts)) (drop m uts)));;

let rec xml2claim
  x = xml_or
        (xml_do "terminating"
          (xml_take_optional
            (xml_do "upperbound"
              (xml_take
                (xml_do "polynomial"
                  (xml_take_attribute "degree"
                    (fun deg_s ->
                      (match nat_of_string deg_s with Inl a -> xml_error a
                        | Inr deg -> xml_return (Upperbound deg)))))
                xml_return))
            (fun a ->
              (match a with None -> xml_return Terminating
                | Some aa -> xml_return aa))))
        (xml_or (xml_leaf "nonterminating" Nonterminating)
          (xml_or (xml_leaf "confluent" Confluent)
            (xml_leaf "nonconfluent" Nonconfluent)))
        x;;

let rec conversion _A
  xml2name =
    xml_do "conversion"
      (xml_take
        (xml_do "startTerm"
          (xml_take (term _A xml2name) (fun x -> xml_return (id x))))
        (fun a ->
          xml_take_many_sub [] zero_nata Infinity_enat (estep _A xml2name)
            (fun b -> xml_return (a, b))));;

let rec plain_name
  x = xml_or (xml_change (xml_text "name") (comp xml_return (fun a -> UnLab a)))
        (xml_or
          (xml_do "sharp"
            (xml_take plain_name (fun xa -> xml_return (Sharp xa))))
          (xml_do "labeledSymbol"
            (xml_take plain_name
              (fun a ->
                xml_take
                  (xml_or
                    (xml_do "numberLabel"
                      (xml_take_many_sub [] zero_nata Infinity_enat
                        (xml_nat "number") (fun aa -> xml_return (Inl aa))))
                    (xml_do "symbolLabel"
                      (xml_take_many_sub [] zero_nata Infinity_enat plain_name
                        (fun aa -> xml_return (Inr aa)))))
                  (fun b ->
                    xml_return
                      (match b with Inl ba -> Lab (a, ba)
                        | Inr ba -> FunLab (a, ba)))))))
        x;;

let rec level_mapping _A
  xml2name =
    xml_do "levelMapping"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "levelMappingEntry"
          (xml_take xml2name
            (fun f ->
              xml_take (xml_nat "arity")
                (fun a ->
                  xml_take_many_sub [] zero_nata Infinity_enat
                    (xml_do "positionLevelEntry"
                      (xml_take (xml_nat "position")
                        (fun aa ->
                          xml_take (xml_nat "level")
                            (fun b -> xml_return (aa, b)))))
                    (fun ps ->
                      xml_return
                        ((f, a),
                          map (fun (p, b) ->
                                ((if equal_nata p zero_nata then a
                                   else minus_nata p one_nata),
                                  b))
                            ps))))))
        xml_return);;

let rec redtriplea _A
  xml2name bi =
    xml_do "redPair"
      (xml_take
        (xml_or
          (xml_do "scnp"
            (xml_take
              (xml_do "status"
                (xml_take
                  (xml_or (xml_leaf "ms" MS_Ext)
                    (xml_or (xml_leaf "min" Min_Ext)
                      (xml_or (xml_leaf "dms" Dms_Ext)
                        (xml_leaf "max" Max_Ext))))
                  xml_return))
              (fun a ->
                xml_take (level_mapping _A xml2name)
                  (fun b ->
                    xml_take
                      (xml_do "redPair"
                        (xml_take (redtriple _A xml2name false)
                          (fun x -> xml_return (id x))))
                      (fun c -> xml_return (Inl (SCNP (a, b, c))))))))
          (xml_change (redtriple _A xml2name bi)
            (comp xml_return (fun a -> Inr a))))
        xml_return);;

let rec rule_pairs _A
  xml2name s p =
    xml_do s
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do p
          (xml_take (rule _A xml2name)
            (fun a ->
              xml_take (rule _A xml2name) (fun b -> xml_return (a, b)))))
        xml_return);;

let rec sl_variant
  xml2name =
    xml_do "model"
      (xml_take
        (xml_or
          (xml_do "rootLabeling"
            (xml_take_optional xml2name
              (fun fo ->
                xml_return
                  (Rootlab
                    (match fo with None -> None
                      | Some f -> Some (f, one_nata))))))
          (xml_do "finiteModel"
            (xml_take
              (xml_change (xml_nat "carrierSize")
                (fun n -> xml_return (minus_nata n one_nata)))
              (fun n ->
                xml_take_optional
                  (xml_do "tupleOrder"
                    (xml_take (xml_leaf "pointWise" ())
                      (fun x -> xml_return (id x))))
                  (fun toa ->
                    xml_take_many_sub [] zero_nata Infinity_enat
                      (xml_do "interpret"
                        (xml_take xml2name
                          (fun a ->
                            xml_take (xml_nat "arity")
                              (fun b ->
                                xml_take arith_fun
                                  (fun c -> xml_return ((a, b), c))))))
                      (fun inter ->
                        xml_return
                          (match toa
                            with None -> Finitelab (SL_Inter (n, inter))
                            | Some _ ->
                              QuasiFinitelab (SL_Inter (n, inter), []))))))))
        (fun x -> xml_return (id x)));;

let rec array_grow (Array a) inc x = Array (a @ replicate inc x);;

let rec diff_by_label (_A1, _A2)
  pairs l = filtera (fun v -> not (member (_A1, _A2) (fst v) l)) pairs;;

let rec formula_pos_parser
  x = xml_or (xml_do "conclusion" (xml_return zero_nata))
        (xml_or
          (xml_do "assertion"
            (xml_return (nat_of_integer (Big_int.big_int_of_int 2))))
          (xml_or
            (xml_do "transition"
              (xml_return (nat_of_integer (Big_int.big_int_of_int 3))))
            (xml_do "targetAssertion"
              (xml_return (nat_of_integer (Big_int.big_int_of_int 4))))))
        x;;

let rec hints_parser _A
  hint_parser xml =
    xml_or (xml_do "auto" (xml_return (default_hinta _A)))
      (xml_or
        (xml_do "distribute"
          (xml_take formula_pos_parser
            (fun pos ->
              xml_take_many_sub [] zero_nata Infinity_enat
                (hints_parser _A hint_parser)
                (fun hints -> xml_return (Distribute (pos, hints))))))
        (xml_or
          (xml_do "erase"
            (xml_take formula_pos_parser
              (fun pos ->
                xml_take (hints_parser _A hint_parser)
                  (fun hint -> xml_return (Erase (pos, hint))))))
          (xml_or
            (xml_do "lexWeak"
              (xml_take_many_sub [] zero_nata Infinity_enat
                (hints_parser _A hint_parser)
                (fun hints -> xml_return (LexWeak hints))))
            (xml_or
              (xml_do "lexStrict"
                (xml_take_many_sub [] zero_nata Infinity_enat
                  (hints_parser _A hint_parser)
                  (fun hints -> xml_return (LexStrict hints))))
              (xml_change hint_parser (comp xml_return (fun a -> Base a)))))))
      xml;;

let rec art_node_parser _G
  art_node_id_parser location_parser trans_parser atom_parser hint_parser =
    xml_do "node"
      (xml_take_default false (xml_do "initial" (xml_return true))
        (fun init ->
          xml_take (art_node_id_parser "nodeId")
            (fun nodeId ->
              xml_take
                (xml_do "invariant"
                  (xml_take (formula_parser atom_parser) xml_return))
                (fun invariant ->
                  xml_take
                    (xml_do "location" (xml_take location_parser xml_return))
                    (fun location ->
                      xml_take
                        (xml_or
                          (xml_do "children"
                            (xml_take_many_sub [] zero_nata Infinity_enat
                              (xml_do "child"
                                (xml_take trans_parser
                                  (fun tr ->
                                    xml_take (art_node_id_parser "nodeId")
                                      (fun n ->
xml_take_default (default_hinta _G)
  (xml_do "hints" (xml_take (hints_parser _G hint_parser) xml_return))
  (fun h -> xml_return (tr, (n, h)))))))
                              (fun chs -> xml_return (Children_Edge chs))))
                          (xml_do "coverEdge"
                            (xml_take (art_node_id_parser "nodeId")
                              (fun n ->
                                xml_take_default (default_hinta _G)
                                  (xml_do "hints"
                                    (xml_take (hints_parser _G hint_parser)
                                      xml_return))
                                  (fun h -> xml_return (Cover_Edge (n, h)))))))
                        (fun edges ->
                          xml_return
                            (Art_Node (nodeId, invariant, location, edges),
                              (if init then [nodeId] else []))))))));;

let rec art_parser _G
  art_node_id_parser location_parser trans_parser atom_parser hint_parser =
    xml_do "impact"
      (xml_take_default []
        (xml_change (art_node_id_parser "initial") (fun x -> xml_return [x]))
        (fun init ->
          xml_take
            (xml_do "nodes"
              (xml_take_many_sub [] zero_nata Infinity_enat
                (art_node_parser _G art_node_id_parser location_parser
                  trans_parser atom_parser hint_parser)
                xml_return))
            (fun pairs ->
              (let nodes = map fst pairs in
               let inits = init @ maps snd pairs in
                xml_return (Art_impl_ext (inits, nodes, ()))))));;

let zeroa : (nat, rat) fmap = fmempty;;

let rec check_prop_rstep_rule (_A1, _A2) (_B1, _B2, _B3, _B4)
  nfs pa p rule s t =
    bindb (check (in_poss p s)
            (comp (comp (comp (showsl_pos p)
                          (showsl_literal " is not a position of "))
                    (showsl_terma _A2 _B4 s))
              (showsl_literal "\092")))
      (fun _ ->
        bindb (check (in_poss p t)
                (comp (comp (comp (showsl_pos p)
                              (showsl_literal " is not a position of "))
                        (showsl_terma _A2 _B4 t))
                  (showsl_literal "\092")))
          (fun _ ->
            (let c = ctxt_of_pos_term p s in
             let d = ctxt_of_pos_term p t in
             let u = subt_at s p in
             let v = subt_at t p in
              (match
                match_list (_B1, _B2, _B3) _A1 _B2 (fun a -> Var a)
                  [(fst rule, u); (snd rule, v)]
                with None ->
                  Inl (comp (comp (comp (comp
  (comp (comp (showsl_literal "the term ") (showsl_terma _A2 _B4 t))
    (showsl_literal " does not result from a proper application of rule\092"))
  (showsl_rule _A2 _B4 rule))
                                    (showsl_literal " at position "))
                              (showsl_pos p))
                        (showsl_literal "\092"))
                | Some tau ->
                  bindb (catch_error
                          (forallM pa
                            (args u @
                              (if nfs then map tau (vars_rule_list rule)
                                else [])))
                          (fun x -> Inl (snd x)))
                    (fun _ ->
                      check (equal_ctxta _A1 _B2 c d)
                        (comp (comp (comp (comp
    (comp (comp (showsl_literal "the term ") (showsl_terma _A2 _B4 t))
      (showsl_literal " does not result from a proper application of rule\092"))
    (showsl_rule _A2 _B4 rule))
                                      (showsl_literal " at position "))
                                (showsl_pos p))
                          (showsl_literal "\092")))))));;

let rec check_prop_rstep (_A1, _A2) (_B1, _B2, _B3, _B4)
  nfs pa r p rule s t =
    check (list_ex
            (fun ra ->
              eq_rule_mod_vars _A1 (_B1, _B2, _B3) rule ra &&
                isOK (check_prop_rstep_rule (_A1, _A2) (_B1, _B2, _B3, _B4) nfs
                       pa p ra s t))
            r)
      (comp (comp (comp (comp (comp (comp (comp
    (comp (showsl_literal "the step from ") (showsl_terma _A2 _B4 s))
    (showsl_literal " to "))
                                      (showsl_terma _A2 _B4 t))
                                (showsl_literal " via rule "))
                          (showsl_rule _A2 _B4 rule))
                    (showsl_literal " at position "))
              (showsl_pos p))
        (showsl_literal " is problematic\092"));;

let rec check_qrstep (_A1, _A2) (_B1, _B2, _B3, _B4)
  nf nfs =
    check_prop_rstep (_A1, _A2) (_B1, _B2, _B3, _B4) nfs
      (fun t ->
        check (nf t)
          (comp (showsl_terma _A2 _B4 t)
            (showsl_literal " is not in Q-normal form")));;

let rec check_rqrstep (_A1, _A2) (_B1, _B2, _B3, _B4)
  nf nfs r rule s t =
    check_qrstep (_A1, _A2) (_B1, _B2, _B3, _B4) nf nfs r [] rule s t;;

let rec check_qsteps (_A1, _A2) (_B1, _B2, _B3, _B4)
  nf nfs p r x4 s u = match nf, nfs, p, r, x4, s, u with
    nf, nfs, p, r, [], s, u ->
      check (equal_terma _A1 _B2 s u)
        (comp (comp (comp (comp (showsl_literal
                                  "the last term of the rewrite sequence\092")
                            (showsl_terma _A2 _B4 s))
                      (showsl_literal
                        "\092does not correspond to the goal term\092"))
                (showsl_terma _A2 _B4 u))
          (showsl_literal "\092"))
    | nf, nfs, p, ra, (uu, (r, (true, t))) :: prts, s, u ->
        bindb (check_rqrstep (_A1, _A2) (_B1, _B2, _B3, _B4) nf nfs p r s t)
          (fun _ ->
            check_qsteps (_A1, _A2) (_B1, _B2, _B3, _B4) nf nfs p ra prts t u)
    | nf, nfs, pa, ra, (p, (r, (false, t))) :: prts, s, u ->
        bindb (check_qrstep (_A1, _A2) (_B1, _B2, _B3, _B4) nf nfs ra p r s t)
          (fun _ ->
            check_qsteps (_A1, _A2) (_B1, _B2, _B3, _B4) nf nfs pa ra prts t
              u);;

let rec check_qrsteps (_A1, _A2) (_B1, _B2, _B3, _B4)
  nf nfs r prts s u =
    check_qsteps (_A1, _A2) (_B1, _B2, _B3, _B4) nf nfs [] r
      (map (fun (p, (ra, t)) -> (p, (ra, (false, t)))) prts) s u;;

let rec check_qrstep_subst (_A1, _A2) (_B1, _B2, _B3, _B4)
  cni nfs =
    (let main =
       check_prop_rstep (_A1, _A2) (_B1, _B2, _B3, _B4) nfs
         (fun t ->
           catch_error (cni t)
             (fun _ ->
               Inl (comp (showsl_terma _A2 _B4 t)
                     (showsl_lit
                       " mu ^^ i is not in Q-normal form for all i"))))
       in
      (fun r p ra s t ->
        bindb (check (not (is_Var (fst ra)))
                (showsl_lit "loop check requires lhss to be non-variable"))
          (fun _ -> main r p ra s t)));;

let rec check_rqrstep_subst (_A1, _A2) (_B1, _B2, _B3, _B4)
  cni nfs =
    (fun r -> check_qrstep_subst (_A1, _A2) (_B1, _B2, _B3, _B4) cni nfs r []);;

let rec check_qsteps_subst (_A1, _A2) (_B1, _B2, _B3, _B4)
  cni nfs p r x4 s u = match cni, nfs, p, r, x4, s, u with
    cni, nfs, p, r, [], s, u ->
      check (equal_terma _A1 _B2 s u)
        (comp (comp (comp (comp (showsl_lit
                                  "the last term of the rewrite sequence\092")
                            (showsl_terma _A2 _B4 s))
                      (showsl_lit
                        "\092does not correspond to the goal term\092"))
                (showsl_terma _A2 _B4 u))
          (showsl_literal "\092"))
    | cni, nfs, p, ra, (uu, (r, (true, t))) :: prts, s, u ->
        bindb (check_rqrstep_subst (_A1, _A2) (_B1, _B2, _B3, _B4) cni nfs p r s
                t)
          (fun _ ->
            check_qsteps_subst (_A1, _A2) (_B1, _B2, _B3, _B4) cni nfs p ra prts
              t u)
    | cni, nfs, pa, ra, (p, (r, (false, t))) :: prts, s, u ->
        bindb (check_qrstep_subst (_A1, _A2) (_B1, _B2, _B3, _B4) cni nfs ra p r
                s t)
          (fun _ ->
            check_qsteps_subst (_A1, _A2) (_B1, _B2, _B3, _B4) cni nfs pa ra
              prts t u);;

let rec check_qrsteps_subst (_A1, _A2) (_B1, _B2, _B3, _B4)
  cni nfs r prts s u =
    check_qsteps_subst (_A1, _A2) (_B1, _B2, _B3, _B4) cni nfs [] r
      (map (fun (p, (ra, t)) -> (p, (ra, (false, t)))) prts) s u;;

let rec si_subst xa = fst (rep_subst_incr xa);;

let rec match_prob_of_rp_impl _A _B
  mu x1 = match mu, x1 with mu, (t, Var x) -> [(t, Var x)]
    | mu, (t, Fun (v, va)) ->
        (let sterms =
           remdups (equal_term _A _B) (t :: map (si_subst mu) (si_W mu t)) in
         let uterms =
           maps (comp (filtera (fun ta -> not (is_Var ta))) supteq_list) sterms
           in
          map (fun u -> (u, Fun (v, va))) (remdups (equal_term _A _B) uterms));;

let rec si_v_incr xa = fst (snd (rep_subst_incr xa));;

let rec simplify_mp _A (_B1, _B2)
  mu_incr x1 solved = match mu_incr, x1, solved with
    mu_incr, (Var x, Fun (f, ls)) :: mp, solved ->
      bind (guard (member (_B1, _B2) x (si_v_incr mu_incr)))
        (fun _ ->
          (let m =
             map (fun (s, a) -> (subst_apply_term s (si_subst mu_incr), a)) in
            bind (simplify_mp _A (_B1, _B2) mu_incr
                   (m ((Var x, Fun (f, ls)) :: mp)) (m solved))
              (fun (smp, i) -> Some (smp, suc i))))
    | mu_incr, (Fun (g, ts), Fun (f, ls)) :: mp, solved ->
        bind (guard (eq _A f g))
          (fun _ ->
            bind (zip_option ts ls)
              (fun pairs ->
                simplify_mp _A (_B1, _B2) mu_incr (pairs @ mp) solved))
    | mu_incr, (s, Var x) :: mp, solved ->
        simplify_mp _A (_B1, _B2) mu_incr mp ((s, Var x) :: solved)
    | mu_incr, [], solved -> Some (solved, zero_nata);;

let rec conflicts _A _B
  mu x1 = match mu, x1 with
    mu, (Fun (f, ts), (Fun (g, ss), n)) ->
      (if eq _B f g && equal_nata (size_list ts) (size_list ss)
        then maps (fun (s, t) -> conflicts _A _B mu (s, (t, n))) (zip ts ss)
        else [(Fun (f, ts), (Fun (g, ss), n))])
    | mu, (Var x, (Fun (g, ss), n)) -> [(Var x, (Fun (g, ss), n))]
    | mu, (Fun (f, ts), (Var y, n)) ->
        (if equal_nata n zero_nata then [(Var y, (Fun (f, ts), zero_nata))]
          else conflicts _A _B mu (Fun (f, ts), (mu y, minus_nata n one_nata)))
    | mu, (Var x, (Var y, n)) ->
        (if equal_nata n zero_nata
          then (if eq _A x y then [] else [(Var x, (Var y, zero_nata))])
          else conflicts _A _B mu (Var x, (mu y, minus_nata n one_nata)));;

let rec ident_solvea (_A1, _A2) (_B1, _B2)
  mu_incr cps st =
    (let cp = conflicts _B2 _A2 (si_subst mu_incr) st in
      (if list_ex (fun (u, (_, _)) -> not (is_Var u)) cp then None
        else (if list_ex
                   (fun (u, (v, _)) ->
                     member
                       ((ceq_prod (ceq_term _A2 _B2) (ceq_term _A2 _B2)),
                         (ccompare_prod (ccompare_term _A1 _B1)
                           (ccompare_term _A1 _B1)))
                       (u, v)
                       (image
                         ((ceq_prod (ceq_term _A2 _B2)
                            (ceq_prod (ceq_term _A2 _B2) ceq_nat)),
                           (ccompare_prod (ccompare_term _A1 _B1)
                             (ccompare_prod (ccompare_term _A1 _B1)
                               ccompare_nat)))
                         ((ceq_prod (ceq_term _A2 _B2) (ceq_term _A2 _B2)),
                           (ccompare_prod (ccompare_term _A1 _B1)
                             (ccompare_term _A1 _B1)),
                           (set_impl_prod set_impl_term set_impl_term))
                         (fun (ua, (va, _)) -> (ua, va)) cps))
                   cp
               then None
               else bind (mapMa
                           (fun (u, (v, m)) ->
                             ident_solvea (_A1, _A2) (_B1, _B2) mu_incr
                               (inserta
                                 ((ceq_prod (ceq_term _A2 _B2)
                                    (ceq_prod (ceq_term _A2 _B2) ceq_nat)),
                                   (ccompare_prod (ccompare_term _A1 _B1)
                                     (ccompare_prod (ccompare_term _A1 _B1)
                                       ccompare_nat)))
                                 (u, (v, m)) cps)
                               (subst_apply_term u (si_subst mu_incr),
                                 (v, suc m)))
                           (conflicts _B2 _A2 (si_subst mu_incr) st))
                      (fun is -> Some (max_list (map suc is))))));;

let rec ident_solve (_A1, _A2) (_B1, _B2)
  mu_incr =
    (fun (s, t) ->
      ident_solvea (_A1, _A2) (_B1, _B2) mu_incr
        (set_empty
          ((ceq_prod (ceq_term _A2 _B2) (ceq_prod (ceq_term _A2 _B2) ceq_nat)),
            (ccompare_prod (ccompare_term _A1 _B1)
              (ccompare_prod (ccompare_term _A1 _B1) ccompare_nat)))
          (of_phantom
            (set_impl_proda set_impl_term
              (set_impl_prod set_impl_term set_impl_nat))))
        (s, (t, zero_nata)));;

let rec ident_decision (_A1, _A2) (_B1, _B2)
  sigma ip = not (is_none (ident_solve (_A1, _A2) (_B1, _B2) sigma ip));;

let rec ident_prob_of_smp _A _B
  = function [] -> []
    | (t, l) :: other ->
        map_filter
          (fun x ->
            (if (let (_, s) = x in equal_terma _A _B s l)
              then Some (let (s, _) = x in (t, s)) else None))
          other @
          ident_prob_of_smp _A _B other;;

let rec gmatch_decision (_A1, _A2) (_B1, _B2, _B3, _B4)
  sigma mp =
    not (is_none
          (bind (simplify_mp _A2 (_B1, _B2) sigma mp [])
            (fun (smp, _) ->
              guard (list_all (ident_decision (_A1, _A2) (_B3, _B4) sigma)
                      (ident_prob_of_smp _A2 _B4 smp)))));;

let rec match_decision (_A1, _A2) (_B1, _B2, _B3, _B4)
  mu mp = gmatch_decision (_A1, _A2) (_B1, _B2, _B3, _B4) mu [mp];;

let rec redex_decision (_A1, _A2) (_B1, _B2, _B3, _B4)
  mu rp =
    list_ex (match_decision (_A1, _A2) (_B1, _B2, _B3, _B4) mu)
      (match_prob_of_rp_impl _A2 _B4 mu rp);;

let rec new_as p bs asa = filtera (fun a -> list_ex (fun b -> p b a) bs) asa;;

let rec list_diff _A
  x0 ys = match x0, ys with [], ys -> []
    | x :: xs, ys ->
        (let zs = list_diff _A xs ys in
          (if membera _A ys x then zs else x :: zs));;

let rec the_set_impl_main _A _B
  p q remain have bs =
    (let newa = new_as p bs remain in
      (if null newa then have
        else the_set_impl_main _A _B p q (list_diff _B remain newa)
               (newa @ have) (remdups _A (maps q newa))));;

let rec the_set_impl _A _B r p q bs = the_set_impl_main _B _A p q r [] bs;;

let rec inductive_set_impl _A _B = the_set_impl _A _B;;

let rec v_incr_impl _A _B
  mu = inductive_set_impl _A _A (map fst mu) (eq _A)
         (fun x ->
           maps (fun (y, t) ->
                  (if equal_terma _B _A t (Var x) then [y] else []))
             mu)
         (maps (fun (y, t) -> (if not (is_Var t) then [y] else [])) mu);;

let rec rtrancl_impl
  gen_succ un memb emp rel =
    (let succ = gen_succ rel in
     let n = size_list rel in
      (fun asa -> relpow_impl succ un memb asa emp n));;

let rec rtrancl_list_impl _A
  = rtrancl_impl
      (fun r asa ->
        remdups _A
          (map_filter
            (fun x ->
              (if (let (a, _) = x in membera _A asa a) then Some (snd x)
                else None))
            r))
      (fun xs ys -> filtera (fun x -> not (membera _A ys x)) xs @ ys)
      (fun x xs -> membera _A xs x) [];;

let rec w_impl _A
  d = (let filt = filtera (fun (x, y) -> not (eq _A x y)) in
       let xvs = maps (fun (x, t) -> map (fun a -> (x, a)) (vars_term_list t)) d
         in
       let rel = filt xvs in
       let rtran = rtrancl_list_impl _A rel in
        (fun t -> rtran (vars_term_list t)));;

let rec subst_incr (_A1, _A2, _A3, _A4) _B
  xa = Abs_subst_incr
         (let dom = mk_subst_domain _A3 _B xa in
           (mk_subst _A3 (fun a -> Var a) xa,
             (set (_A1, _A2, _A4) (v_incr_impl _A3 _B dom), w_impl _A3 dom)));;

let rec redex_rps_decision (_A1, _A2, _A3, _A4, _A5) (_B1, _B2)
  mu = (let mua = subst_incr (_A1, _A2, _A4, _A5) _B2 mu in
        let main = redex_decision (_B1, _B2) (_A1, _A2, _A3, _A4) mua in
         (fun xs ->
           catch_error (forallM (fun tl -> check (not (main tl)) tl) xs)
             (fun x -> Inl (snd x))));;

let rec check_NF_iteration (_A1, _A2, _A3, _A4, _A5) (_B1, _B2)
  mu = (let dec = redex_rps_decision (_A1, _A2, _A3, _A4, _A5) (_B1, _B2) mu in
         (fun q t -> dec (map (fun a -> (t, a)) q)));;

let rec check_loop (_A1, _A2, _A3) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  q nfs s rseq sigma c r =
    bindb (check (not (null rseq))
            (showsl_lit "rewrite sequence must be non-empty"))
      (fun _ ->
        (if null q
          then check_qrsteps (_A2, _A3) (_B2, _B4, _B5, _B7) (fun _ -> true) nfs
                 r rseq s
                 (ctxt_apply_term c
                   (subst_apply_term s (mk_subst _B4 (fun a -> Var a) sigma)))
          else check_qrsteps_subst (_A2, _A3) (_B2, _B4, _B5, _B7)
                 (check_NF_iteration (_B1, _B2, _B3, _B4, _B6) (_A1, _A2) sigma
                   q)
                 nfs r rseq s
                 (ctxt_apply_term c
                   (subst_apply_term s
                     (mk_subst _B4 (fun a -> Var a) sigma)))));;

let rec min_list _A
  = function [x] -> x
    | x :: v :: va ->
        min _A.order_linorder.preorder_order.ord_preorder x
          (min_list _A (v :: va));;

let rec nth_default
  dflt xs n = (if less_nat n (size_list xs) then nth xs n else dflt);;

let rec delete_value _A _B
  key v m =
    (match key v with None -> m
      | Some k ->
        (match lookup _B m k with None -> m
          | Some vs -> insert _B k (removeAll _A v vs) m));;

let rec check_ge_v _A
  v p = (match p with [] -> true
          | [(m, c)] ->
            equal_monoma (equal_nat, linorder_nat) m (one_monom linorder_nat) &&
              less_eq
                _A.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order
                c v
          | (_, _) :: _ :: _ -> false);;

let rec check_poly_weak_anti_mono_discrete (_A1, _A2) (_B1, _B2)
  p v = check_poly_ge (_A1, _A2)
          _B2.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1
          p (poly_subst _A2 (_A1, _A2) (_B1, _B2.comm_semiring_1_poly_carrier)
              (fun w ->
                poly_of (_A1, _A2) (_B1, _B2.comm_semiring_1_poly_carrier)
                  (if eq _A1 w v
                    then PSum [PNum (onea _B2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral);
                                PVar v]
                    else PVar w))
              p);;

let rec monom_vars (_A1, _A2, _A3, _A4)
  m = set (_A1, _A2, _A4) (monom_vars_list _A3 m);;

let rec check_poly_weak_anti_mono (_A1, _A2, _A3, _A4) _B
  p v = list_all
          (fun (m, c) ->
            less_eq
              _B.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order
              c (zerob
                  _B.semiring_0_ordered_semiring_0a.mult_zero_semiring_0.zero_mult_zero) ||
              not (member (_A1, _A2) v (monom_vars (_A1, _A2, _A3, _A4) m)))
          p;;

let rec check_poly_weak_anti_mono_smart (_A1, _A2, _A3, _A4, _A5) (_B1, _B2)
  discrete =
    (if discrete then check_poly_weak_anti_mono_discrete (_A3, _A4) (_B1, _B2)
      else check_poly_weak_anti_mono (_A1, _A2, _A4, _A5)
             _B2.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1);;

let rec check_poly_weak_mono_discrete (_A1, _A2) (_B1, _B2)
  p v = check_poly_ge (_A1, _A2)
          _B2.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1
          (poly_subst _A2 (_A1, _A2) (_B1, _B2.comm_semiring_1_poly_carrier)
            (fun w ->
              poly_of (_A1, _A2) (_B1, _B2.comm_semiring_1_poly_carrier)
                (if eq _A1 w v
                  then PSum [PNum (onea _B2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral);
                              PVar v]
                  else PVar w))
            p)
          p;;

let rec check_poly_weak_mono (_A1, _A2, _A3, _A4) _B
  p v = list_all
          (fun (m, c) ->
            less_eq
              _B.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order
              (zerob
                _B.semiring_0_ordered_semiring_0a.mult_zero_semiring_0.zero_mult_zero)
              c ||
              not (member (_A1, _A2) v (monom_vars (_A1, _A2, _A3, _A4) m)))
          p;;

let rec check_poly_weak_mono_smart (_A1, _A2, _A3, _A4, _A5) (_B1, _B2)
  discrete =
    (if discrete then check_poly_weak_mono_discrete (_A3, _A4) (_B1, _B2)
      else check_poly_weak_mono (_A1, _A2, _A4, _A5)
             _B2.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1);;

let rec fun_of_map_fun m d a = (match m a with None -> d a | Some b -> b);;

let rec g_list_to_map_rm_basic_ops _A
  l = foldl (fun m (k, v) -> insert _A k v m) (empty _A.linorder_compare_order)
        (rev l);;

let rec ceta_map_of _A ps = lookup _A (g_list_to_map_rm_basic_ops _A ps);;

let rec poly_inter_list_to_inter _A _B
  def i =
    fun_of_map_fun (ceta_map_of (compare_order_prod _B compare_order_nat) i)
      (fun fn -> default_I _A def (snd fn));;

let rec fun_of_map_funa
  m d f a = (match m a with None -> d a | Some aa -> f aa);;

let rec create_dep (_A1, _A2) (_B1, _B2)
  discrete def i =
    (let fs = remdups (equal_prod _B2 equal_nat) (map fst i) in
     let ii = poly_inter_list_to_inter _A2 _B1 def i in
     let fsres =
       map (fun fn ->
             (let p = ii fn in
              let vars = poly_vars_list (equal_nat, linorder_nat) p in
              let is = upt zero_nata (snd fn) in
              let a =
                map (fun ia ->
                      (if membera equal_nat vars ia
                        then (if check_poly_weak_mono_smart
                                   (ceq_nat, ccompare_nat, equal_nat,
                                     linorder_nat, set_impl_nat)
                                   (_A1, _A2) discrete p ia
                               then Increase
                               else (if check_poly_weak_anti_mono_smart
  (ceq_nat, ccompare_nat, equal_nat, linorder_nat, set_impl_nat) (_A1, _A2)
  discrete p ia
                                      then Decrease else Wild))
                        else Ignore))
                  is
                in
               (fn, a)))
         fs
       in
     let iii =
       fun_of_map_funa
         (ceta_map_of (compare_order_prod _B1 compare_order_nat) fsres)
         (fun _ _ -> Increase) nth
       in
      iii);;

let rec coeff _A p = nth_default (zerob _A) (coeffs _A p);;

let rec pseudo_mod_main_list (_A1, _A2)
  lc r d n =
    (if equal_nata n zero_nata then r
      else (let rr =
              map (times
                    _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd
                    lc)
                r
              in
            let a = hda r in
            let rrr =
              tla (if eq _A1 a
                        (zerob
                          _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
                    then rr
                    else minus_poly_rev_list
                           _A2.ring_1_comm_ring_1.neg_numeral_ring_1.group_add_neg_numeral
                           rr (map (times
                                     _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd
                                     a)
                                d))
              in
             pseudo_mod_main_list (_A1, _A2) lc rrr d
               (minus_nata n one_nata)));;

let rec pseudo_mod_list (_A1, _A2)
  p q = (if null q then p
          else (let rq = rev q in
                let a =
                  pseudo_mod_main_list (_A1, _A2) (hda rq) (rev p) rq
                    (minus_nata (plus_nata one_nata (size_list p))
                      (size_list q))
                  in
                 rev a));;

let rec pseudo_mod (_A1, _A2, _A3)
  f g = poly_of_list
          (_A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.comm_monoid_add_semiring_0,
            _A1)
          (pseudo_mod_list (_A1, _A2)
            (coeffs
              _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
              f)
            (coeffs
              _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
              g));;

let rec eval_monom_list _A _B
  alpha x1 = match alpha, x1 with
    alpha, [] ->
      onea _B.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral
    | alpha, (x, p) :: m ->
        times _B.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd
          (eval_monom_list _A _B alpha m)
          (power
            _B.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.monoid_mult_semiring_numeral.power_monoid_mult
            (alpha x) p);;

let rec eval_monom _A _B x xc = eval_monom_list _A _B x (rep_monom _A xc);;

let rec eval_poly _A _B
  alpha x1 = match alpha, x1 with
    alpha, [] ->
      zerob _B.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
    | alpha, mc :: p ->
        plus _B.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.semigroup_add_numeral.plus_semigroup_add
          (times
            _B.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd
            (eval_monom _A _B alpha (fst mc)) (snd mc))
          (eval_poly _A _B alpha p);;

let rec poly_vars (_A1, _A2, _A3, _A4)
  p = set (_A1, _A2, _A4) (maps (comp (monom_vars_list _A3) fst) p);;

let rec map_prod f g (a, b) = (f a, g b);;

let rec enfc_q _A
  isQnf isRnf r q s x5 = match isQnf, isRnf, r, q, s, x5 with
    isQnf, isRnf, r, q, s, Var x -> true
    | isQnf, isRnf, r, q, s, Fun (f, ts) ->
        list_all
          (fun qa ->
            (match mgu_class _A (Fun (f, map (icap_impl _A isQnf r s) ts)) qa
              with None -> true
              | Some mu ->
                not (list_all
                       (fun u ->
                         isQnf (subst_apply_term
                                 (map_term (fun x -> x)
                                   (fun a ->
                                     Chara (false, false, false, true, true,
     true, true, false) ::
                                       a)
                                   u)
                                 mu))
                       s &&
                      isRnf (subst_apply_term
                              (map_term (fun x -> x)
                                (fun a ->
                                  Chara (true, false, false, true, true, true,
  true, false) ::
                                    a)
                                qa)
                              mu))))
          q;;

let rec delete_values _A _B
  uu x1 m = match uu, x1, m with uu, [], m -> m
    | key, v :: vs, m ->
        delete_value _A _B key v (delete_values _A _B key vs m);;

let rec delete_rules _A (_B1, _B2) _C
  a rs =
    delete_values
      (equal_prod _A (equal_prod (equal_term _B2 _C) (equal_term _B2 _C)))
      (compare_order_prod _B1 compare_order_nat) key
      (map (fun b -> (a, b)) rs);;

let rec split_rulesc
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = split_rules;;

let rec split_pairsa
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = split_pairs;;

let rec delete_R_Rwc
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = delete_R_Rw;;

let rec delete_P_Pwa
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = delete_P_Pw;;

let rec minimal
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = minimal;;

let rec nfsc
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = nfs;;

let rec mkd
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = mk;;

let rec qc
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = q;;

let rec split_proc _B _C
  i d p_remove r_remove =
    (let (p, pw) = split_pairsa i d p_remove in
     let (r, rw) = split_rulesc i d r_remove in
     let dpp1 = mkd i (nfsc i d) (minimal i d) p pw (qc i d) r rw in
     let a =
       delete_R_Rwc i (delete_P_Pwa i d p_remove p_remove) r_remove r_remove in
      (dpp1, a));;

let rec rep_x_f_status_option_x_x_nat_list_nat_x_f_prod_fun_option
  (Abs_x_f_status_option_x_x_nat_list_nat_x_f_prod_fun_option x) = x;;

let rec sel21
  xa = Abs_status
         (match rep_x_f_status_option_x_x_nat_list_nat_x_f_prod_fun_option xa
           with None -> status (failwith "undefined") | Some x2 -> x2);;

let rec dis1
  xa = (match rep_x_f_status_option_x_x_nat_list_nat_x_f_prod_fun_option xa
         with None -> true | Some _ -> false);;

let rec rep_isom x = (if dis1 x then None else Some (sel21 x));;

let rec status_of_aux _A
  xa = Abs_x_f_status_option_x_x_nat_list_nat_x_f_prod_fun_option
         (if list_all
               (fun fidx ->
                 list_all (fun i -> less_nat i (snd (fst fidx))) (snd fidx))
               xa
           then Some (fun_of_map_fun (map_of (equal_prod _A equal_nat) xa)
                       (fun (_, a) -> upt zero_nata a))
           else None);;

let rec status_of _A x = rep_isom (status_of_aux _A x);;

let rec r_states (_A1, _A2, _A3)
  = (fun ta_rule ->
      inserta (_A1, _A2) (r_rhs ta_rule)
        (set (_A1, _A2, _A3) (r_lhs_states ta_rule)));;

let rec ta_states (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2)
  ta = sup_seta (_A3, _A4.ccompare_cproper_interval)
         (sup_seta (_A3, _A4.ccompare_cproper_interval)
           (sup_setb (_A1, _A2, _A3, _A4, _A6)
             (image
               ((ceq_ta_rule _A5 _B2),
                 (ccompare_ta_rule _A4.ccompare_cproper_interval _B1))
               ((ceq_set (_A2, _A3, _A4.ccompare_cproper_interval)),
                 (ccompare_set (_A1, _A3, _A4, _A6)), set_impl_set)
               (r_states (_A3, _A4.ccompare_cproper_interval, _A6))
               (ta_rules ta)))
           (sup_setb (_A1, _A2, _A3, _A4, _A6)
             (image
               ((ceq_prod _A3 _A3),
                 (ccompare_prod _A4.ccompare_cproper_interval
                   _A4.ccompare_cproper_interval))
               ((ceq_set (_A2, _A3, _A4.ccompare_cproper_interval)),
                 (ccompare_set (_A1, _A3, _A4, _A6)), set_impl_set)
               (fun (q, qa) ->
                 inserta (_A3, _A4.ccompare_cproper_interval) q
                   (inserta (_A3, _A4.ccompare_cproper_interval) qa
                     (bot_set (_A3, _A4.ccompare_cproper_interval, _A6))))
               (ta_eps ta))))
         (ta_final ta);;

let rec prod_ta (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2, _B3, _B4)
  (_C1, _C2, _C3, _C4, _C5, _C6)
  tA1 tA2 f =
    Ta_ext
      (f, image ((ceq_prod (ceq_ta_rule _A5 _B3) (ceq_ta_rule _C5 _B3)),
                  (ccompare_prod
                    (ccompare_ta_rule _A4.ccompare_cproper_interval _B2)
                    (ccompare_ta_rule _C4.ccompare_cproper_interval _B2)))
            ((ceq_ta_rule (equal_prod _A5 _C5) _B3),
              (ccompare_ta_rule
                (ccompare_prod _A4.ccompare_cproper_interval
                  _C4.ccompare_cproper_interval)
                _B2),
              set_impl_ta_rule)
            (fun (TA_rule (fa, qs1, q1), TA_rule (_, qs2, q2)) ->
              TA_rule (fa, zip qs1 qs2, (q1, q2)))
            (sup_setb
              ((finite_UNIV_prod finite_UNIV_ta_rule finite_UNIV_ta_rule),
                (cenum_prod cenum_ta_rule cenum_ta_rule),
                (ceq_prod (ceq_ta_rule _A5 _B3) (ceq_ta_rule _C5 _B3)),
                (cproper_interval_prod
                  (cproper_interval_ta_rule _A4.ccompare_cproper_interval _B2)
                  (cproper_interval_ta_rule _C4.ccompare_cproper_interval _B2)),
                (set_impl_prod set_impl_ta_rule set_impl_ta_rule))
              (image ((ceq_prod _B1 ceq_nat), (ccompare_prod _B2 ccompare_nat))
                ((ceq_set
                   ((cenum_prod cenum_ta_rule cenum_ta_rule),
                     (ceq_prod (ceq_ta_rule _A5 _B3) (ceq_ta_rule _C5 _B3)),
                     (cproper_interval_prod
                       (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                         _B2)
                       (cproper_interval_ta_rule _C4.ccompare_cproper_interval
                         _B2)).ccompare_cproper_interval)),
                  (ccompare_set
                    ((finite_UNIV_prod finite_UNIV_ta_rule finite_UNIV_ta_rule),
                      (ceq_prod (ceq_ta_rule _A5 _B3) (ceq_ta_rule _C5 _B3)),
                      (cproper_interval_prod
                        (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                          _B2)
                        (cproper_interval_ta_rule _C4.ccompare_cproper_interval
                          _B2)),
                      (set_impl_prod set_impl_ta_rule set_impl_ta_rule))),
                  set_impl_set)
                (fun fa ->
                  productc
                    ((ceq_ta_rule _A5 _B3),
                      (ccompare_ta_rule _A4.ccompare_cproper_interval _B2),
                      set_impl_ta_rule)
                    ((ceq_ta_rule _C5 _B3),
                      (ccompare_ta_rule _C4.ccompare_cproper_interval _B2),
                      set_impl_ta_rule)
                    (filter
                      ((ceq_ta_rule _A5 _B3),
                        (ccompare_ta_rule _A4.ccompare_cproper_interval _B2))
                      (fun r -> equal_proda _B3 equal_nat (r_sym r) fa)
                      (ta_rules tA1))
                    (filter
                      ((ceq_ta_rule _C5 _B3),
                        (ccompare_ta_rule _C4.ccompare_cproper_interval _B2))
                      (fun r -> equal_proda _B3 equal_nat (r_sym r) fa)
                      (ta_rules tA2)))
                (ta_syms (_A4.ccompare_cproper_interval, _A5)
                  (_B1, _B2, _B3, _B4) tA1))),
        sup_seta
          ((ceq_prod (ceq_prod _A3 _C3) (ceq_prod _A3 _C3)),
            (ccompare_prod
              (ccompare_prod _A4.ccompare_cproper_interval
                _C4.ccompare_cproper_interval)
              (ccompare_prod _A4.ccompare_cproper_interval
                _C4.ccompare_cproper_interval)))
          (image
            ((ceq_prod (ceq_prod _A3 _A3) _C3),
              (ccompare_prod
                (ccompare_prod _A4.ccompare_cproper_interval
                  _A4.ccompare_cproper_interval)
                _C4.ccompare_cproper_interval))
            ((ceq_prod (ceq_prod _A3 _C3) (ceq_prod _A3 _C3)),
              (ccompare_prod
                (ccompare_prod _A4.ccompare_cproper_interval
                  _C4.ccompare_cproper_interval)
                (ccompare_prod _A4.ccompare_cproper_interval
                  _C4.ccompare_cproper_interval)),
              (set_impl_prod (set_impl_prod _A6 _C6) (set_impl_prod _A6 _C6)))
            (fun (a, b) -> (let (q, qa) = a in (fun p -> ((q, p), (qa, p)))) b)
            (productc
              ((ceq_prod _A3 _A3),
                (ccompare_prod _A4.ccompare_cproper_interval
                  _A4.ccompare_cproper_interval),
                (set_impl_prod _A6 _A6))
              (_C3, _C4.ccompare_cproper_interval, _C6) (ta_eps tA1)
              (ta_states (_C1, _C2, _C3, _C4, _C5, _C6) (_B2, _B3) tA2)))
          (image
            ((ceq_prod _A3 (ceq_prod _C3 _C3)),
              (ccompare_prod _A4.ccompare_cproper_interval
                (ccompare_prod _C4.ccompare_cproper_interval
                  _C4.ccompare_cproper_interval)))
            ((ceq_prod (ceq_prod _A3 _C3) (ceq_prod _A3 _C3)),
              (ccompare_prod
                (ccompare_prod _A4.ccompare_cproper_interval
                  _C4.ccompare_cproper_interval)
                (ccompare_prod _A4.ccompare_cproper_interval
                  _C4.ccompare_cproper_interval)),
              (set_impl_prod (set_impl_prod _A6 _C6) (set_impl_prod _A6 _C6)))
            (fun (p, (q, qa)) -> ((p, q), (p, qa)))
            (productc (_A3, _A4.ccompare_cproper_interval, _A6)
              ((ceq_prod _C3 _C3),
                (ccompare_prod _C4.ccompare_cproper_interval
                  _C4.ccompare_cproper_interval),
                (set_impl_prod _C6 _C6))
              (ta_states (_A1, _A2, _A3, _A4, _A5, _A6) (_B2, _B3) tA1)
              (ta_eps tA2))),
        ());;

let rec productive_relation (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2)
  ta = sup_seta
         ((ceq_prod _A3 _A3),
           (ccompare_prod _A4.ccompare_cproper_interval
             _A4.ccompare_cproper_interval))
         (image
           ((ceq_prod _A3 _A3),
             (ccompare_prod _A4.ccompare_cproper_interval
               _A4.ccompare_cproper_interval))
           ((ceq_prod _A3 _A3),
             (ccompare_prod _A4.ccompare_cproper_interval
               _A4.ccompare_cproper_interval),
             (set_impl_prod _A6 _A6))
           (fun (a, b) -> (b, a)) (ta_eps ta))
         (sup_setb
           ((finite_UNIV_prod _A1 _A1), (cenum_prod _A2 _A2),
             (ceq_prod _A3 _A3), (cproper_interval_prod _A4 _A4),
             (set_impl_prod _A6 _A6))
           (image
             ((ceq_ta_rule _A5 _B2),
               (ccompare_ta_rule _A4.ccompare_cproper_interval _B1))
             ((ceq_set
                ((cenum_prod _A2 _A2), (ceq_prod _A3 _A3),
                  (cproper_interval_prod _A4 _A4).ccompare_cproper_interval)),
               (ccompare_set
                 ((finite_UNIV_prod _A1 _A1), (ceq_prod _A3 _A3),
                   (cproper_interval_prod _A4 _A4), (set_impl_prod _A6 _A6))),
               set_impl_set)
             (fun r ->
               image (_A3, _A4.ccompare_cproper_interval)
                 ((ceq_prod _A3 _A3),
                   (ccompare_prod _A4.ccompare_cproper_interval
                     _A4.ccompare_cproper_interval),
                   (set_impl_prod _A6 _A6))
                 (fun a -> (r_rhs r, a))
                 (set (_A3, _A4.ccompare_cproper_interval, _A6)
                   (r_lhs_states r)))
             (ta_rules ta)));;

let rec ta_productive (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2)
  ta = sup_seta (_A3, _A4.ccompare_cproper_interval) (ta_final ta)
         (compute_trancl (_A2, _A3, _A4.ccompare_cproper_interval, _A6)
           (ta_final ta)
           (productive_relation (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2) ta));;

let rec ta_restrict (_A1, _A2, _A3, _A4, _A5) (_B1, _B2)
  ta q =
    Ta_ext
      (inf_seta (_A2, _A3) (ta_final ta) q,
        filter ((ceq_ta_rule _A4 _B2), (ccompare_ta_rule _A3 _B1))
          (fun r -> less_eq_set (_A1, _A2, _A3) (r_states (_A2, _A3, _A5) r) q)
          (ta_rules ta),
        inf_seta ((ceq_prod _A2 _A2), (ccompare_prod _A3 _A3)) (ta_eps ta)
          (productc (_A2, _A3, _A5) (_A2, _A3, _A5) q q),
        ());;

let rec trim_ta (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2, _B3)
  ta = ta_restrict (_A2, _A3, _A4.ccompare_cproper_interval, _A5, _A6)
         (_B1, _B3)
         (ta_restrict (_A2, _A3, _A4.ccompare_cproper_interval, _A5, _A6)
           (_B1, _B3) ta
           (ta_reachable (_A2, _A3, _A4.ccompare_cproper_interval, _A5, _A6)
             (_B1, _B2, _B3) ta))
         (ta_productive (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B3)
           (ta_restrict (_A2, _A3, _A4.ccompare_cproper_interval, _A5, _A6)
             (_B1, _B3) ta
             (ta_reachable (_A2, _A3, _A4.ccompare_cproper_interval, _A5, _A6)
               (_B1, _B2, _B3) ta)));;

let rec check_varcond_subset _A (_B1, _B2)
  r = catch_error
        (forallM
          (fun rule ->
            catch_error
              (check_subseteq _B1 (insert_vars_term _B1 (snd rule) [])
                (insert_vars_term _B1 (fst rule) []))
              (fun x ->
                Inl (comp (comp (comp (comp (showsl_literal "free variable ")
(showsl _B2 x))
                                  (showsl_literal
                                    " in right-hand side of rule "))
                            (showsl_rule _A _B2 rule))
                      (showsl_literal "\092"))))
          r)
        (fun x -> Inl (snd x));;

let rec check_wf_trs _A (_B1, _B2)
  r = catch_error
        (bindb (check_varcond_no_Var_lhs _A _B2 r)
          (fun _ -> check_varcond_subset _A (_B1, _B2) r))
        (fun x ->
          Inl (comp (showsl_literal "the TRS is not well-formed\092") x));;

let rec defined_list
  r = maps (fun (l, _) -> (if not (is_Var l) then [the (root l)] else [])) r;;

let rec showsl_rules _A _B = showsl_rulesa (showsl _A) (showsl _B) " -> ";;

let rec generate_f_xs
  f n = Fun (f, map (fun i -> Var (generate_var i)) (upt zero_nata n));;

let rec num_symbs
  = function Var x -> one_nata
    | Fun (f, ts) ->
        plus_nata (sum_list monoid_add_nat (map num_symbs ts)) one_nata;;

let rec
  rep_x_compare_order_x_x_f_afs_option_x_x_nat_x_compare_order_x_x_f_prod_set_af_entry_nat_x_compare_order_x_x_f_prod_fun_prod_option
  _A (Abs_x_compare_order_x_x_f_afs_option_x_x_nat_x_compare_order_x_x_f_prod_set_af_entry_nat_x_compare_order_x_x_f_prod_fun_prod_option
       x)
       = x;;

let rec sel21a _A
  xa = Abs_afs
         (match
           rep_x_compare_order_x_x_f_afs_option_x_x_nat_x_compare_order_x_x_f_prod_set_af_entry_nat_x_compare_order_x_x_f_prod_fun_prod_option
             _A xa
           with None -> rep_afs (failwith "undefined") | Some x2 -> x2);;

let rec dis1a _A
  xa = (match
         rep_x_compare_order_x_x_f_afs_option_x_x_nat_x_compare_order_x_x_f_prod_set_af_entry_nat_x_compare_order_x_x_f_prod_fun_prod_option
           _A xa
         with None -> true | Some _ -> false);;

let rec rep_isoma _A x = (if dis1a _A x then None else Some (sel21a _A x));;

let rec default_af_entry n = AFList (upt zero_nata n);;

let rec wf_af_entry n x1 = match n, x1 with n, Collapse i -> less_nat i n
                      | n, AFList is -> list_all (fun i -> less_nat i n) is;;

let rec afs_of_aux (_A1, _A2, _A3, _A4)
  xa = Abs_x_compare_order_x_x_f_afs_option_x_x_nat_x_compare_order_x_x_f_prod_set_af_entry_nat_x_compare_order_x_x_f_prod_fun_prod_option
         (if list_all (fun ((_, n), e) -> wf_af_entry n e) xa
           then Some (fun_of_map_fun
                        (ceta_map_of (compare_order_prod _A3 compare_order_nat)
                          xa)
                        (fun fn -> default_af_entry (snd fn)),
                       set ((ceq_prod _A1 ceq_nat),
                             (ccompare_prod _A2 ccompare_nat),
                             (set_impl_prod _A4 set_impl_nat))
                         (map fst xa))
           else None);;

let rec afs_of (_A1, _A2, _A3, _A4)
  x = rep_isoma _A3 (afs_of_aux (_A1, _A2, _A3, _A4) x);;

let rec ta_bounds_lhs
  xml2name =
    xml_do "lhs"
      (xml_or (xml_take state (fun a -> xml_return (Inr a)))
        (xml_take xml2name
          (fun f ->
            xml_take (xml_nat "height")
              (fun h ->
                xml_take_many_sub [] zero_nata Infinity_enat state
                  (fun qs -> xml_return (Inl ((f, h), qs)))))));;

let rec bounds_bound x = xml_nat "bound" x;;

let rec bounds_type
  x = xml_do "type"
        (xml_take (xml_or (xml_leaf "roof" Roof) (xml_leaf "match" Match))
          (fun xa -> xml_return (id xa)))
        x;;

let rec bounds_info
  xml2name =
    xml_do "bounds"
      (xml_take bounds_type
        (fun a ->
          xml_take bounds_bound
            (fun b ->
              xml_take final_states
                (fun c ->
                  xml_take (tree_automaton (ta_bounds_lhs xml2name))
                    (fun d ->
                      xml_take_default Id_Relation closed_criterion
                        (fun e ->
                          xml_return (Bounds_Info (a, b, c, d, e))))))));;

let rec multiset_af _A
  xml2name =
    xml_do "multisetArgumentFilter"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "multisetArgumentFilterEntry"
          (xml_take xml2name
            (fun f ->
              xml_take (xml_nat "arity")
                (fun a ->
                  xml_take
                    (xml_do "status"
                      (xml_take_many_sub [] zero_nata Infinity_enat position
                        xml_return))
                    (fun p -> xml_return ((f, a), p))))))
        xml_return);;

let rec word_pattern (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "wordPattern"
      (xml_take (stringa (_A1, _A2) (_B1, _B2) xml2name)
        (fun l ->
          xml_take (stringa (_A1, _A2) (_B1, _B2) xml2name)
            (fun m ->
              xml_take (xml_nat "factor")
                (fun f ->
                  xml_take (xml_nat "constant")
                    (fun c ->
                      xml_take (stringa (_A1, _A2) (_B1, _B2) xml2name)
                        (fun r -> xml_return (l, ((f, (c, m)), r))))))));;

let rec derivation_pattern (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "derivationPattern"
      (xml_take (word_pattern (_A1, _A2) (_B1, _B2) xml2name)
        (fun a ->
          xml_take (word_pattern (_A1, _A2) (_B1, _B2) xml2name)
            (fun b -> xml_return (a, b))));;

let rec derivation_pattern_proof (_A1, _A2) (_B1, _B2)
  xml2name =
    (let oc = oc_srs (_A1, _A2) (_B1, _B2) xml2name in
     let dp = derivation_pattern (_A1, _A2) (_B1, _B2) xml2name in
     let s = stringa (_A1, _A2) (_B1, _B2) xml2name in
      xml_do "derivationPatternProof"
        (xml_take
          (xml_or
            (xml_do "OC1"
              (xml_take oc
                (fun a ->
                  xml_take (xml_bool "isPair")
                    (fun b -> xml_return (OC1 (a, b))))))
            (xml_or
              (xml_do "OC2"
                (xml_take oc
                  (fun a ->
                    xml_take oc
                      (fun b ->
                        xml_take oc
                          (fun c ->
                            xml_take s
                              (fun d ->
                                xml_take s
                                  (fun e ->
                                    xml_take s
                                      (fun f ->
xml_return (OC2 (a, b, c, d, e, f))))))))))
              (xml_or
                (xml_do "OC2prime"
                  (xml_take oc
                    (fun a ->
                      xml_take oc
                        (fun b ->
                          xml_take oc
                            (fun c ->
                              xml_take s
                                (fun d ->
                                  xml_take s
                                    (fun e ->
                                      xml_take s
(fun f -> xml_return (OC2p (a, b, c, d, e, f))))))))))
                (xml_or
                  (xml_do "OC3"
                    (xml_take oc
                      (fun a ->
                        xml_take oc
                          (fun b ->
                            xml_take oc
                              (fun c ->
                                xml_take s
                                  (fun d ->
                                    xml_take s
                                      (fun e ->
xml_return (OC3 (a, b, c, d, e)))))))))
                  (xml_or
                    (xml_do "OC3prime"
                      (xml_take oc
                        (fun a ->
                          xml_take oc
                            (fun b ->
                              xml_take oc
                                (fun c ->
                                  xml_take s
                                    (fun d ->
                                      xml_take s
(fun e -> xml_return (OC3p (a, b, c, d, e)))))))))
                    (xml_or
                      (xml_do "OCintoDP1"
                        (xml_take dp
                          (fun a ->
                            xml_take oc (fun b -> xml_return (OCDP1 (a, b))))))
                      (xml_or
                        (xml_do "OCintoDP2"
                          (xml_take dp
                            (fun a ->
                              xml_take oc
                                (fun b -> xml_return (OCDP2 (a, b))))))
                        (xml_or
                          (xml_do "equivalent"
                            (xml_take dp
                              (fun a ->
                                xml_take dp
                                  (fun b -> xml_return (WPEQ (a, b))))))
                          (xml_or
                            (xml_do "lift"
                              (xml_take dp
                                (fun a ->
                                  xml_take dp
                                    (fun b -> xml_return (Lift (a, b))))))
                            (xml_or
                              (xml_do "DP_OC_1_1"
                                (xml_take dp
                                  (fun a ->
                                    xml_take dp
                                      (fun b ->
xml_take oc
  (fun c ->
    xml_take s
      (fun d -> xml_take s (fun e -> xml_return (DPOC1_1 (a, b, c, d, e)))))))))
                              (xml_or
                                (xml_do "DP_OC_1_2"
                                  (xml_take dp
                                    (fun a ->
                                      xml_take dp
(fun b ->
  xml_take oc
    (fun c ->
      xml_take s
        (fun d ->
          xml_take s
            (fun e ->
              xml_take s
                (fun f -> xml_return (DPOC1_2 (a, b, c, d, e, f))))))))))
                                (xml_or
                                  (xml_do "DP_OC_2"
                                    (xml_take dp
                                      (fun a ->
xml_take dp
  (fun b ->
    xml_take oc
      (fun c ->
        xml_take s
          (fun d ->
            xml_take s (fun e -> xml_return (DPOC2 (a, b, c, d, e)))))))))
                                  (xml_or
                                    (xml_do "DP_OC_3_1"
                                      (xml_take dp
(fun a ->
  xml_take dp
    (fun b ->
      xml_take oc
        (fun c ->
          xml_take s
            (fun d ->
              xml_take s (fun e -> xml_return (DPOC3_1 (a, b, c, d, e)))))))))
                                    (xml_or
                                      (xml_do "DP_OC_3_2"
(xml_take dp
  (fun a ->
    xml_take dp
      (fun b ->
        xml_take oc
          (fun c ->
            xml_take s
              (fun d ->
                xml_take s
                  (fun e ->
                    xml_take s
                      (fun f -> xml_return (DPOC3_2 (a, b, c, d, e, f))))))))))
                                      (xml_or
(xml_do "DP_DP_1_1"
  (xml_take dp
    (fun a ->
      xml_take dp
        (fun b ->
          xml_take dp
            (fun c ->
              xml_take s
                (fun d ->
                  xml_take s
                    (fun e -> xml_return (DPDP1_1 (a, b, c, d, e)))))))))
(xml_or
  (xml_do "DP_DP_1_2"
    (xml_take dp
      (fun a ->
        xml_take dp
          (fun b ->
            xml_take dp
              (fun c ->
                xml_take s
                  (fun d ->
                    xml_take s
                      (fun e -> xml_return (DPDP1_2 (a, b, c, d, e)))))))))
  (xml_or
    (xml_do "DP_DP_2_1"
      (xml_take dp
        (fun a ->
          xml_take dp
            (fun b ->
              xml_take dp
                (fun c ->
                  xml_take s
                    (fun d ->
                      xml_take s
                        (fun e -> xml_return (DPDP2_1 (a, b, c, d, e)))))))))
    (xml_do "DP_DP_2_2"
      (xml_take dp
        (fun a ->
          xml_take dp
            (fun b ->
              xml_take dp
                (fun c ->
                  xml_take s
                    (fun d ->
                      xml_take s
                        (fun e ->
                          xml_return
                            (DPDP2_2 (a, b, c, d, e))))))))))))))))))))))))))
          (fun x -> xml_return (id x))));;

let rec nonloop_srs_reason (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_or
      (xml_do "selfEmbeddingOC"
        (xml_take (stringa (_A1, _A2) (_B1, _B2) xml2name)
          (fun a ->
            xml_take (stringa (_A1, _A2) (_B1, _B2) xml2name)
              (fun b ->
                xml_take (stringa (_A1, _A2) (_B1, _B2) xml2name)
                  (fun c ->
                    xml_return (fun d -> SE_OC ((b, a @ b @ c), a, c, d)))))))
      (xml_do "selfEmbeddingDP"
        (xml_take (derivation_pattern (_A1, _A2) (_B1, _B2) xml2name)
          (fun a ->
            xml_take (stringa (_A1, _A2) (_B1, _B2) xml2name)
              (fun b ->
                xml_take (stringa (_A1, _A2) (_B1, _B2) xml2name)
                  (fun c -> xml_return (fun d -> SE_DP (a, b, c, d)))))));;

let rec nonloop_srs (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "nonterminatingSRS"
      (xml_take
        (xml_do "derivationPatterns"
          (xml_take_many_sub [] zero_nata Infinity_enat
            (derivation_pattern_proof (_A1, _A2) (_B1, _B2) xml2name)
            (fun a -> xml_return (id a))))
        (fun a ->
          xml_take (nonloop_srs_reason (_A1, _A2) (_B1, _B2) xml2name)
            (fun b -> xml_return (b a))));;

let rec evala
  i x1 alpha = match i, x1, alpha with
    i, Fun (f, ss), alpha -> i f (map (fun s -> evala i s alpha) ss)
    | i, Var x, alpha -> alpha x;;

let rec nFQ_subset_NF_rulesb
  (Tp_ops_ext
    (qreltrs, q, r, rw, rules, q_empty, is_QNF, nFQ_subset_NF_rules, rules_map,
      delete_R_Rw, split_rules, mk, nfs, more))
    = nFQ_subset_NF_rules;;

let rec is_QNFb
  (Tp_ops_ext
    (qreltrs, q, r, rw, rules, q_empty, is_QNF, nFQ_subset_NF_rules, rules_map,
      delete_R_Rw, split_rules, mk, nfs, more))
    = is_QNF;;

let rec rulesd
  (Tp_ops_ext
    (qreltrs, q, r, rw, rules, q_empty, is_QNF, nFQ_subset_NF_rules, rules_map,
      delete_R_Rw, split_rules, mk, nfs, more))
    = rules;;

let rec icap_impl_tp _B
  i d = (let qr = nFQ_subset_NF_rulesb i d in
         let qnf = is_QNFb i d in
         let r = rulesd i d in
         let ic = icap_impl_gen _B qr qnf (map fst r) in
          (fun s ->
            (let a =
               ceta_set_of (compare_order_list (compare_order_char, equal_char))
                 (maps vars_term_list s)
               in
              ic s a)));;

let rec has_type _B
  type_of_fun x1 ty = match type_of_fun, x1, ty with
    type_of_fun, Fun (f, es), ty ->
      eq _B ty (snd (type_of_fun f)) &&
        (equal_nata (size_list es) (size_list (fst (type_of_fun f))) &&
          all_interval_nat
            (fun i ->
              has_type _B type_of_fun (nth es i) (nth (fst (type_of_fun f)) i))
            zero_nata (size_list es))
    | type_of_fun, Var v, ty -> eq _B (snd v) ty;;

let rec is_bool (_B1, _B2, _B3)
  type_of_fun bool_types e =
    bex (_B1, _B2) bool_types (has_type _B3 type_of_fun e) && not (is_Var e);;

let rec formula
  atom x1 = match atom, x1 with
    atom, Disjunction phi_s -> list_all (formula atom) phi_s
    | atom, Conjunction phi_s -> list_all (formula atom) phi_s
    | atom, NegAtom a -> atom a
    | atom, Atom a -> atom a;;

let rec transition_rule (_B1, _B2, _B3)
  type_of_fun bool_types (Transition (l, r, phi)) =
    formula (is_bool (_B1, _B2, _B3) type_of_fun bool_types) phi;;

let rec check_lts_impl (_B1, _B2, _B3)
  type_of_fun bool_types pi =
    (let Lts_Impl (_, ti, lc) = pi in
      bindb (catch_error
              (forallM
                (fun (_, t) ->
                  check (transition_rule (_B1, _B2, _B3) type_of_fun bool_types
                          t)
                    (showsl_lit "ill-formed transition in LTS"))
                ti)
              (fun x -> Inl (snd x)))
        (fun _ ->
          catch_error
            (forallM
              (fun (_, f) ->
                check (formula (is_bool (_B1, _B2, _B3) type_of_fun bool_types)
                        f)
                  (showsl_lit "ill-formed location condition in LTS"))
              lc)
            (fun x -> Inl (snd x))));;

let rec label_decomp = function Lab (f, l) -> (f, Inl l)
                       | FunLab (f, l) -> (f, Inr l);;

let rec list_all_rec p x1 = match p, x1 with p, [] -> true
                       | p, a :: l -> p a && list_all_rec p l;;

let rec poly_const (_A1, _A2) _B
  a = (if eq _A2 a (zerob _A1) then [] else [(one_monom _B, a)]);;

let rec poly_minus (_A1, _A2) (_B1, _B2)
  f g = poly_add (_A1, _A2)
          (_B1, _B2.semiring_1_cancel_ring_1.semiring_1_semiring_1_cancel.semiring_0_semiring_1)
          f (monom_mult_poly (_A1, _A2)
              (_B1, _B2.semiring_1_cancel_ring_1.semiring_1_semiring_1_cancel.semiring_0_semiring_1)
              (one_monom _A2,
                uminus
                  _B2.neg_numeral_ring_1.group_add_neg_numeral.uminus_group_add
                  (onea _B2.neg_numeral_ring_1.numeral_neg_numeral.one_numeral))
              g);;

let rec remove_prefix _A
  x0 ys = match x0, ys with [], ys -> Some ys
    | x :: xs, y :: ys -> (if eq _A x y then remove_prefix _A xs ys else None)
    | v :: va, [] -> None;;

let rec remove_suffix _A
  p q = (match remove_prefix _A (rev p) (rev q) with None -> None
          | Some r -> Some (rev r));;

let rec swap p = (snd p, fst p);;

let rec set_ext
  s_ns =
    (fun asa bs ->
      (not (null asa) &&
         list_all (fun b -> list_ex (fun a -> fst (s_ns a b)) asa) bs,
        list_all (fun b -> list_ex (fun a -> snd (s_ns a b)) asa) bs));;

let rec check_no_var _A _B
  t = check (not (is_Var t)) (showsl_literal "variable found\092");;

let rec check_supteq (_A1, _A2) (_B1, _B2)
  s t = check (equal_terma _A1 _B1 s t || supt_impl _A1 _B1 s t)
          (comp (comp (showsl_terma _A2 _B2 t)
                  (showsl_literal " is not a subterm of "))
            (showsl_terma _A2 _B2 s));;

let rec showsl_ctxt
  funa var x2 = match funa, var, x2 with funa, var, Hole -> showsl_lit "[]"
    | funa, var, More (f, ss1, d, ss2) ->
        comp (comp (comp (comp (funa f) (showsl_literal "("))
                     (showsl_list_gen (showsl_term funa var) "" "" ", " ", "
                       ss1))
               (showsl_ctxt funa var d))
          (showsl_list_gen (showsl_term funa var) ")" ", " ", " ")" ss2);;

let rec funs_term_ms _A
  = function Var x -> zero_multiseta
    | Fun (f, ts) ->
        plus_multiseta _A (add_mset _A f zero_multiseta)
          (sum_mset (comm_monoid_add_multiset _A)
            (mset (equal_multiset _A) (map (funs_term_ms _A) ts)));;

let rec subt_at_ctxt c x1 = match c, x1 with c, [] -> c
                       | More (f, bef, c, aft), i :: p -> subt_at_ctxt c p;;

let rec ta_match (_A1, _A2, _A3, _A4, _A5) (_B1, _B2) (_C1, _C2)
  ta qsig x2 q = match ta, qsig, x2, q with
    ta, qsig, Fun (f, ts), q ->
      (let n = size_list ts in
       let rls =
         filter ((ceq_ta_rule _A4 _B2), (ccompare_ta_rule _A3 _B1))
           (fun rule ->
             equal_proda _B2 equal_nat (r_sym rule) (f, n) &&
               bex (_A2, _A3)
                 (sup_seta (_A2, _A3)
                   (inserta (_A2, _A3) (r_rhs rule) (bot_set (_A2, _A3, _A5)))
                   (compute_trancl (_A1, _A2, _A3, _A5)
                     (inserta (_A2, _A3) (r_rhs rule) (bot_set (_A2, _A3, _A5)))
                     (ta_eps ta)))
                 (fun qa -> member (_A2, _A3) qa q))
           (ta_rules ta)
         in
        sup_setb
          (finite_UNIV_list, cenum_list, (ceq_list (ceq_prod _C1 _A2)),
            (cproper_interval_list (ccompare_prod _C2 _A3)), set_impl_list)
          (image ((ceq_ta_rule _A4 _B2), (ccompare_ta_rule _A3 _B1))
            ((ceq_set
               (cenum_list, (ceq_list (ceq_prod _C1 _A2)),
                 (cproper_interval_list
                   (ccompare_prod _C2 _A3)).ccompare_cproper_interval)),
              (ccompare_set
                (finite_UNIV_list, (ceq_list (ceq_prod _C1 _A2)),
                  (cproper_interval_list (ccompare_prod _C2 _A3)),
                  set_impl_list)),
              set_impl_set)
            (fun (TA_rule (_, qs, _)) ->
              image ((ceq_list (ceq_list (ceq_prod _C1 _A2))),
                      (ccompare_list (ccompare_list (ccompare_prod _C2 _A3))))
                ((ceq_list (ceq_prod _C1 _A2)),
                  (ccompare_list (ccompare_prod _C2 _A3)), set_impl_list)
                concat
                (listset
                  ((ceq_list (ceq_prod _C1 _A2)),
                    (ccompare_list (ccompare_prod _C2 _A3)), set_impl_list)
                  (map (fun (tsi, qsi) ->
                         ta_match (_A1, _A2, _A3, _A4, _A5) (_B1, _B2)
                           (_C1, _C2) ta qsig tsi
                           (inserta (_A2, _A3) qsi (bot_set (_A2, _A3, _A5))))
                    (zip ts qs))))
            rls))
    | ta, qsig, Var x, q ->
        image (_A2, _A3)
          ((ceq_list (ceq_prod _C1 _A2)),
            (ccompare_list (ccompare_prod _C2 _A3)), set_impl_list)
          (fun qa -> [(x, qa)])
          (filter (_A2, _A3) (fun qa -> member (_A2, _A3) qa qsig)
            (sup_seta (_A2, _A3) q
              (compute_trancl (_A1, _A2, _A3, _A5) q
                (converse (_A2, _A3, _A5) (_A2, _A3, _A5) (ta_eps ta)))));;

let rec generate_listset (_A1, _A2, _A3)
  n s = (if equal_nata n zero_nata
          then inserta ((ceq_list _A1), (ccompare_list _A2)) []
                 (set_empty ((ceq_list _A1), (ccompare_list _A2))
                   (of_phantom set_impl_lista))
          else set_Cons (_A1, _A2, _A3) s
                 (generate_listset (_A1, _A2, _A3) (minus_nata n one_nata) s));;

let rec minus_set (_A1, _A2) a b = inf_seta (_A1, _A2) a (uminus_set b);;

let rec mergeb _A
  xa0 x = match xa0, x with
    Fun (fa, tsa), Fun (f, ts) ->
      bind (guard (eq _A fa f))
        (fun _ ->
          bind (guard (equal_nata (size_list tsa) (size_list ts)))
            (fun _ ->
              bind (mapMa (fun (a, b) -> mergeb _A a b) (zip tsa ts))
                (fun tsb -> Some (Fun (fa, tsb)))))
    | Var (), x -> Some x
    | Fun (v, va), Var () -> Some (Fun (v, va));;

let rec merge_cl (_A1, _A2)
  s = (let newa =
         minus_set ((ceq_term _A2 equal_unit), (ccompare_term _A1 compare_unit))
           (these
             ((ceq_term _A2 equal_unit), (ccompare_term _A1 compare_unit),
               set_impl_term)
             (image
               ((ceq_prod (ceq_term _A2 equal_unit) (ceq_term _A2 equal_unit)),
                 (ccompare_prod (ccompare_term _A1 compare_unit)
                   (ccompare_term _A1 compare_unit)))
               ((ceq_option (ceq_term _A2 equal_unit)),
                 (ccompare_option (ccompare_term _A1 compare_unit)),
                 (set_impl_option set_impl_term))
               (fun (a, b) -> mergeb _A2 a b)
               (productc
                 ((ceq_term _A2 equal_unit), (ccompare_term _A1 compare_unit),
                   set_impl_term)
                 ((ceq_term _A2 equal_unit), (ccompare_term _A1 compare_unit),
                   set_impl_term)
                 s s)))
           s
         in
        (if set_eq
              (cenum_term, (ceq_term _A2 equal_unit),
                (ccompare_term _A1 compare_unit))
              newa
              (set_empty
                ((ceq_term _A2 equal_unit), (ccompare_term _A1 compare_unit))
                (of_phantom set_impl_terma))
          then s
          else merge_cl (_A1, _A2)
                 (sup_seta
                   ((ceq_term _A2 equal_unit), (ccompare_term _A1 compare_unit))
                   s newa)));;

let rec subt_merge_cl (_A1, _A2) (_B1, _B2)
  s = merge_cl (_A1, _A2)
        (sup_seta ((ceq_term _A2 equal_unit), (ccompare_term _A1 compare_unit))
          (inserta ((ceq_term _A2 equal_unit), (ccompare_term _A1 compare_unit))
            (Var ())
            (set_empty
              ((ceq_term _A2 equal_unit), (ccompare_term _A1 compare_unit))
              (of_phantom set_impl_terma)))
          (sup_setb
            (finite_UNIV_term, cenum_term, (ceq_term _A2 equal_unit),
              (cproper_interval_term _A1 compare_unit), set_impl_term)
            (image ((ceq_term _A2 equal_unit), (ccompare_term _A1 compare_unit))
              ((ceq_set
                 (cenum_term, (ceq_term _A2 equal_unit),
                   (cproper_interval_term _A1
                     compare_unit).ccompare_cproper_interval)),
                (ccompare_set
                  (finite_UNIV_term, (ceq_term _A2 equal_unit),
                    (cproper_interval_term _A1 compare_unit), set_impl_term)),
                set_impl_set)
              (fun t ->
                set ((ceq_term _A2 equal_unit),
                      (ccompare_term _A1 compare_unit), set_impl_term)
                  (supt_list t))
              (image ((ceq_term _A2 _B2), (ccompare_term _A1 _B1))
                ((ceq_term _A2 equal_unit), (ccompare_term _A1 compare_unit),
                  set_impl_term)
                (map_term (fun x -> x) (fun _ -> ())) s))));;

let rec maxa (_A1, _A2, _A3, _A4)
  a = set_fold1 (_A1, _A2, _A3) (max_sls _A4) a;;

let rec matche _A
  xa0 x = match xa0, x with Var (), x -> true
    | Fun (fa, tsa), Fun (f, ts) -> eq _A fa f && list_all2 (matche _A) tsa ts
    | Fun (v, va), Var vb -> false;;

let rec funs_term_list t = add_funs_term t [];;

let rec shrinks (_A1, _A2)
  ta t =
    (let s =
       filter ((ceq_term _A2 equal_unit), (ccompare_term _A1 compare_unit))
         (fun s -> matche _A2 s ta) t
       in
     let max =
       maxa (ceq_nat, ccompare_nat, lattice_nat, linorder_nat)
         (image ((ceq_term _A2 equal_unit), (ccompare_term _A1 compare_unit))
           (ceq_nat, ccompare_nat, set_impl_nat) (comp size_list funs_term_list)
           s)
       in
      filter ((ceq_term _A2 equal_unit), (ccompare_term _A1 compare_unit))
        (fun sa -> equal_nata (comp size_list funs_term_list sa) max) s);;

let rec nf_rules_states_impl (_A1, _A2, _A3, _A4, _A5, _A6, _A7) (_B1, _B2)
  t siga =
    (let mcl = subt_merge_cl (_A5, _A6) (_B1, _B2) t in
     let states =
       filter ((ceq_term _A6 equal_unit), (ccompare_term _A5 compare_unit))
         (fun q ->
           ball ((ceq_term _A6 _B2), (ccompare_term _A5 _B1)) t
             (fun ta ->
               not (matche _A6 (map_term (fun x -> x) (fun _ -> ()) ta) q)))
         mcl
       in
     let lhss =
       sup_setb
         ((finite_UNIV_prod _A1 finite_UNIV_list), (cenum_prod _A2 cenum_list),
           (ceq_prod _A3 (ceq_list (ceq_term _A6 equal_unit))),
           (cproper_interval_prod _A4
             (cproper_interval_list (ccompare_term _A5 compare_unit))),
           (set_impl_prod _A7 set_impl_list))
         (image
           ((ceq_prod _A3 ceq_nat),
             (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat))
           ((ceq_set
              ((cenum_prod _A2 cenum_list),
                (ceq_prod _A3 (ceq_list (ceq_term _A6 equal_unit))),
                (cproper_interval_prod _A4
                  (cproper_interval_list
                    (ccompare_term _A5
                      compare_unit))).ccompare_cproper_interval)),
             (ccompare_set
               ((finite_UNIV_prod _A1 finite_UNIV_list),
                 (ceq_prod _A3 (ceq_list (ceq_term _A6 equal_unit))),
                 (cproper_interval_prod _A4
                   (cproper_interval_list (ccompare_term _A5 compare_unit))),
                 (set_impl_prod _A7 set_impl_list))),
             set_impl_set)
           (fun (f, n) ->
             image ((ceq_list (ceq_term _A6 equal_unit)),
                     (ccompare_list (ccompare_term _A5 compare_unit)))
               ((ceq_prod _A3 (ceq_list (ceq_term _A6 equal_unit))),
                 (ccompare_prod _A4.ccompare_cproper_interval
                   (ccompare_list (ccompare_term _A5 compare_unit))),
                 (set_impl_prod _A7 set_impl_list))
               (fun a -> (f, a))
               (generate_listset
                 ((ceq_term _A6 equal_unit), (ccompare_term _A5 compare_unit),
                   set_impl_term)
                 n states))
           siga)
       in
     let flhss =
       filter
         ((ceq_prod _A3 (ceq_list (ceq_term _A6 equal_unit))),
           (ccompare_prod _A4.ccompare_cproper_interval
             (ccompare_list (ccompare_term _A5 compare_unit))))
         (fun q ->
           ball ((ceq_term _A6 _B2), (ccompare_term _A5 _B1)) t
             (fun ta ->
               not (matche _A6 (map_term (fun x -> x) (fun _ -> ()) ta)
                     (Fun (fst q, snd q)))))
         lhss
       in
     let rules =
       image ((ceq_prod _A3 (ceq_list (ceq_term _A6 equal_unit))),
               (ccompare_prod _A4.ccompare_cproper_interval
                 (ccompare_list (ccompare_term _A5 compare_unit))))
         ((ceq_set
            (cenum_ta_rule, (ceq_ta_rule (equal_term _A6 equal_unit) _A6),
              (cproper_interval_ta_rule (ccompare_term _A5 compare_unit)
                _A4.ccompare_cproper_interval).ccompare_cproper_interval)),
           (ccompare_set
             (finite_UNIV_ta_rule,
               (ceq_ta_rule (equal_term _A6 equal_unit) _A6),
               (cproper_interval_ta_rule (ccompare_term _A5 compare_unit)
                 _A4.ccompare_cproper_interval),
               set_impl_ta_rule)),
           set_impl_set)
         (fun (f, qs) ->
           image ((ceq_term _A6 equal_unit), (ccompare_term _A5 compare_unit))
             ((ceq_ta_rule (equal_term _A6 equal_unit) _A6),
               (ccompare_ta_rule (ccompare_term _A5 compare_unit)
                 _A4.ccompare_cproper_interval),
               set_impl_ta_rule)
             (fun a -> TA_rule (f, qs, a))
             (shrinks (_A5, _A6) (Fun (f, qs)) mcl))
         flhss
       in
      (sup_setb
         (finite_UNIV_ta_rule, cenum_ta_rule,
           (ceq_ta_rule (equal_term _A6 equal_unit) _A6),
           (cproper_interval_ta_rule (ccompare_term _A5 compare_unit)
             _A4.ccompare_cproper_interval),
           set_impl_ta_rule)
         rules,
        states));;

let rec ta_nf (_A1, _A2, _A3, _A4, _A5, _A6, _A7) (_B1, _B2)
  t siga =
    (let (rules, states) =
       nf_rules_states_impl (_A1, _A2, _A3, _A4, _A5, _A6, _A7) (_B1, _B2) t
         siga
       in
      Ta_ext
        (states, rules,
          set_empty
            ((ceq_prod (ceq_term _A6 equal_unit) (ceq_term _A6 equal_unit)),
              (ccompare_prod (ccompare_term _A5 compare_unit)
                (ccompare_term _A5 compare_unit)))
            (of_phantom (set_impl_proda set_impl_term set_impl_term)),
          ()));;

let rec compute_NF
  f a = (match f a with None -> Some a | Some aa -> compute_NF f aa);;

let rec first_rewrite _A (_B1, _B2, _B3)
  r s = (match rewrite _A (_B1, _B2, _B3) r s with [] -> None
          | t :: _ -> Some t);;

let rec compute_rstep_NF _A (_B1, _B2, _B3)
  r s = compute_NF (first_rewrite _A (_B1, _B2, _B3) r) s;;

let rec check_join_NF (_A1, _A2) (_B1, _B2, _B3, _B4)
  r s t =
    (match
      (compute_rstep_NF _A1 (_B1, _B2, _B3) r s,
        compute_rstep_NF _A1 (_B1, _B2, _B3) r t)
      with (None, _) ->
        Inl (showsl_literal "strange error in normal form computation")
      | (Some _, None) ->
        Inl (showsl_literal "strange error in normal form computation")
      | (Some sa, Some ta) ->
        check (equal_terma _A1 _B2 sa ta)
          (comp (comp (comp (comp (comp (comp
  (comp (showsl_literal "the normal form ") (showsl_terma _A2 _B4 sa))
  (showsl_literal " of "))
                                    (showsl_terma _A2 _B4 s))
                              (showsl_literal
                                " differs from\092the normal form "))
                        (showsl_terma _A2 _B4 ta))
                  (showsl_literal " of "))
            (showsl_terma _A2 _B4 t)));;

let rec in_rstep_impl (_A1, _A2) (_B1, _B2, _B3, _B4)
  s t r =
    list_ex
      (fun p ->
        in_poss p t &&
          (equal_ctxta _A2 _B3 (ctxt_of_pos_term p s) (ctxt_of_pos_term p t) &&
            (let sp = subt_at s p in
             let tp = subt_at t p in
              bex ((ceq_prod (ceq_term _A2 _B3) (ceq_term _A2 _B3)),
                    (ccompare_prod (ccompare_term _A1 _B2)
                      (ccompare_term _A1 _B2)))
                r (fun lr ->
                    not (is_none
                          (match_list (_B1, _B3, _B4) _A2 _B3 (fun a -> Var a)
                            [(fst lr, sp); (snd lr, tp)]))))))
      (poss_list s);;

let rec reverse_rules rs = map swap rs;;

let rec wf_rules_impl _B r = filtera (wf_rule _B) r;;

let rec check_pairwise
  c x1 = match c, x1 with c, [] -> Inr ()
    | c, x :: xs ->
        bindb (catch_error (forallM (c x) xs) (fun xa -> Inl (snd xa)))
          (fun _ -> check_pairwise c xs);;

let rec check_disjoint _A
  xs ys =
    catch_error
      (forallM (fun x -> (if not (membera _A ys x) then Inr () else Inl x)) xs)
      (fun x -> Inl (snd x));;

let rec check_partition _A xss = check_pairwise (check_disjoint _A) xss;;

let rec check_inj (_A1, _A2) (_B1, _B2)
  a nn sml =
    (let symbols =
       map (fun (aa, b) ->
             (let (_, n) = aa in
               (fun fs ->
                 map (fun (g, i) ->
                       (g, plus_nata n (times_nata i (minus_nata nn one_nata))))
                   (zip fs (upt zero_nata (size_list fs)))))
               b)
         sml
       in
     let fsymbols = concat symbols in
      bindb (catch_error
              (check_partition (equal_prod (equal_lab _A1 _B1) equal_nat)
                symbols)
              (fun x ->
                Inl (comp (comp (showsl_lit "symbol ")
                            (showsl_prod (showl_lab _A2 _B2) showl_nat x))
                      (showsl_lit " occurs twice)"))))
        (fun _ ->
          bindb (check
                  (not (membera (equal_prod (equal_lab _A1 _B1) equal_nat)
                         fsymbols (a, nn)))
                  (comp (comp (showsl_lit "application symbol")
                          (showsl_lab _A2 _B2 a))
                    (showsl_lit " must not occur as new symbol")))
            (fun _ ->
              bindb (check
                      (not (membera (equal_prod (equal_lab _A1 _B1) equal_nat)
                             (map fst sml) (a, nn)))
                      (comp (comp (showsl_lit "application symbol")
                              (showsl_lab _A2 _B2 a))
                        (showsl_lit " must not be uncurried")))
                (fun _ ->
                  (if less_eq_nat nn one_nata
                    then catch_error
                           (forallM
                             (check_pairwise
                               (fun gn1 gn2 ->
                                 check (not
 (equal_proda (equal_lab _A1 _B1) equal_nat gn1 gn2))
                                   (comp (comp (showsl_lit "symbol ")
   (showsl_prod (showl_lab _A2 _B2) showl_nat gn1))
                                     (showsl_lit " occurs twice"))))
                             symbols)
                           (fun x -> Inl (snd x))
                    else Inr ())))));;

let rec not_ws_ns
  (Redtriple_ext
    (valid, s, ns, nst, af, mono_af, mono, desc, not_ws_ns, cpx, more))
    = not_ws_ns;;

let rec no_complexity_check
  x = (fun _ _ -> Inl (showsl_literal "complexity analysis unsupported")) x;;

let rec valid
  (Redtriple_ext
    (valid, s, ns, nst, af, mono_af, mono, desc, not_ws_ns, cpx, more))
    = valid;;

let rec desc
  (Redtriple_ext
    (valid, s, ns, nst, af, mono_af, mono, desc, not_ws_ns, cpx, more))
    = desc;;

let rec check_status_ws_info _A
  sigma cns x2 = match sigma, cns, x2 with
    sigma, cns, None ->
      Inl (showsl_lit "missing weak-subterm status of base reduction pair")
    | sigma, cns, Some fs ->
        catch_error
          (forallM
            (fun f ->
              catch_error
                (forallM
                  (fun i ->
                    catch_error
                      (cns (Fun (fst f,
                                  map (fun ia ->
Var (shows_prec_nat zero_nata ia []))
                                    (upt zero_nata (snd f))),
                             Var (shows_prec_nat zero_nata i [])))
                      (fun _ ->
                        Inl (comp (comp (comp
  (showsl_lit "according to weak-mon. information of order, argument #")
  (showsl_nat (suc i)))
                                    (showsl_lit
                                      " must not occur in status of "))
                              (showsl_prod _A showl_nat f))))
                  (status sigma f))
                (fun x -> Inl (snd x)))
            fs)
          (fun x -> Inl (snd x));;

let rec ns
  (Redtriple_ext
    (valid, s, ns, nst, af, mono_af, mono, desc, not_ws_ns, cpx, more))
    = ns;;

let rec af
  (Redtriple_ext
    (valid, s, ns, nst, af, mono_af, mono, desc, not_ws_ns, cpx, more))
    = af;;

let rec s
  (Redtriple_ext
    (valid, s, ns, nst, af, mono_af, mono, desc, not_ws_ns, cpx, more))
    = s;;

let rec showsl_wpo_params _A
  params =
    comp (showsl_lit "status and precedence\092")
      (showsl_sep
        (fun (f, (p, s)) ->
          comp (comp (comp (comp (comp (showsl_lit "symbol ")
                                   (showsl_prod _A showl_nat f))
                             (showsl_lit ": precedence = "))
                       (showsl_nat p))
                 (showsl_lit "   status = "))
            (showsl_list_nat s))
        (showsl_literal "\092") params);;

let rec faulty_redtriple
  uu uv err desc =
    Redtriple_ext
      (Inl err, (fun _ -> Inr ()), (fun _ -> Inr ()), (fun _ -> Inr ()),
        full_af, empty_af, (fun _ -> Inr ()), desc, None, no_complexity_check,
        ());;

let rec prl_nat pr = (fun f -> equal_nata (pr f) zero_nata);;

let rec prc_nat
  pr = (fun f g -> (let pf = pr f in
                    let pg = pr g in
                     (less_nat pg pf, less_eq_nat pg pf)));;

let rec wpo_redtriple (_A1, _A2, _A3)
  rt params =
    (let stat = map (fun (f, ps) -> (f, snd ps)) params in
     let pr =
       fun_of_map_funa
         (ceta_map_of (compare_order_prod _A1 compare_order_nat) params)
         (fun _ -> zero_nata) fst
       in
     let desca =
       comp (comp (comp (showsl_lit "WPO with ") (showsl_wpo_params _A3 params))
              (showsl_lit "\092over the following reduction pair:\092"))
         (desc rt)
       in
      (match status_of _A2 stat
        with None ->
          faulty_redtriple Type Type
            (showsl_lit "problem with indices in status of WPO!") desca
        | Some sigma ->
          (let sa = (fun sa t -> isOK (s rt (sa, t))) in
           let nsa = (fun sb t -> isOK (ns rt (sb, t))) in
           let wpo =
             wpo_ub (equal_list equal_char) (prc_nat pr) (prl_nat pr) sa nsa
               sigma
             in
           let wpo_s =
             (fun (sb, t) ->
               check (fst (wpo sb t))
                 (comp (comp (comp (showsl_terma _A3 (showl_list showl_char) sb)
                               (showsl_lit " >wpo "))
                         (showsl_terma _A3 (showl_list showl_char) t))
                   (showsl_lit " could not be ensured")))
             in
           let wpo_ns =
             (fun (sb, t) ->
               check (snd (wpo sb t))
                 (comp (comp (comp (showsl_terma _A3 (showl_list showl_char) sb)
                               (showsl_lit " >=wpo "))
                         (showsl_terma _A3 (showl_list showl_char) t))
                   (showsl_lit " could not be ensured")))
             in
            Redtriple_ext
              (bindb (valid rt)
                 (fun _ ->
                   check_status_ws_info _A3 sigma (ns rt) (not_ws_ns rt)),
                wpo_s, wpo_ns, wpo_ns, af_wpo (af rt) sigma, empty_af,
                (fun _ ->
                  catch_error
                    (forallM
                      (fun (a, b) ->
                        (let (f, n) = a in
                          (fun idx ->
                            check (eq_set
                                    (card_UNIV_nat, cenum_nat, ceq_nat,
                                      ccompare_nat)
                                    (set (ceq_nat, ccompare_nat, set_impl_nat)
                                      idx)
                                    (set (ceq_nat, ccompare_nat, set_impl_nat)
                                      (upt zero_nata n)))
                              (comp (comp (comp
    (showsl_lit "for monotonicity, status must be complete, but status of ")
    (showsl_prod _A3 showl_nat (f, n)))
                                      (showsl_lit " is "))
                                (showsl_lista showl_nat (map suc idx)))))
                          b)
                      stat)
                    (fun x -> Inl (snd x))),
                desca, Some (map fst stat), no_complexity_check, ()))));;

let rec parse_xml_string
  p str = (match string2xml str with Inl a -> left a | Inr a -> parse_xml p a);;

let rec parse_claim f = parse_xml_string xml2claim;;

let rec size_multiset (Bag ms) = foldd (fun _ -> plus_nata) zero_nata ms;;

let rec lex_ext
  f n ss ts =
    (let lts = size_list ts in
      (if equal_nata (size_list ss) lts || less_eq_nat lts n
        then lex_ext_unbounded f ss ts else (false, false)));;

let rec ass_list_to_single_list
  = function [] -> []
    | (x, n) :: xs -> replicate n x @ ass_list_to_single_list xs;;

let rec or2 a b = (fst a || fst b, snd a || snd b);;

let rec mul_ext_impl
  sns x1 x2 = match sns, x1, x2 with sns, [], [] -> (false, true)
    | sns, [], v :: va -> (false, false)
    | sns, v :: va, [] -> (true, true)
    | sns, v :: va, y :: ys -> mul_ex_dom sns (v :: va) [] y ys
and mul_ex_dom
  sns x1 xs y ys = match sns, x1, xs, y, ys with
    sns, [], xs, y, ys -> (false, false)
    | sns, x :: xsa, xs, y, ys ->
        (match sns x y
          with (true, _) ->
            (if snd (mul_ext_impl sns (xsa @ xs)
                      (filtera (fun ya -> not (fst (sns x ya))) ys))
              then (true, true) else mul_ex_dom sns xsa (x :: xs) y ys)
          | (false, true) ->
            or2 (mul_ext_impl sns (xsa @ xs) ys)
              (mul_ex_dom sns xsa (x :: xs) y ys)
          | (false, false) -> mul_ex_dom sns xsa (x :: xs) y ys);;

let rec mul_ext x = mul_ext_impl x;;

let rec smulextp
  f (Bag xs) (Bag ys) =
    fst (mul_ext f (ass_list_to_single_list (impl_ofa xs))
          (ass_list_to_single_list (impl_ofa ys)));;

let rec mulextp
  f (Bag xs) (Bag ys) =
    mul_ext f (ass_list_to_single_list (impl_ofa xs))
      (ass_list_to_single_list (impl_ofa ys));;

let rec weighta
  w w0 x2 = match w, w0, x2 with
    w, w0, Fun (f, ts) ->
      plus_nata (w (f, size_list ts))
        (sum_list monoid_add_nat (map (weighta w w0) ts))
    | w, w0, Var x -> w0;;

let rec ackbo_impl (_A1, _A2, _A3) _B
  w w0 pr_strict ac s t =
    (if subseteq_mset _B (vars_term_ms _B t) (vars_term_ms _B s) &&
          less_eq_nat (weighta w w0 t) (weighta w w0 s)
      then (if less_nat (weighta w w0 t) (weighta w w0 s) then true
             else (match s with Var _ -> false
                    | Fun (f, ss) ->
                      (match t with Var _ -> true
                        | Fun (g, ts) ->
                          (if pr_strict (f, size_list ss) (g, size_list ts)
                            then true
                            else (if equal_proda _A3 equal_nat (f, size_list ss)
                                       (g, size_list ts)
                                   then (if not (member (_A1, _A2) f ac) ||
      not (equal_nata (size_list ss)
            (nat_of_integer (Big_int.big_int_of_int 2)))
  then fst (lex_ext
             (fun x y ->
               (ackbo_impl (_A1, _A2, _A3) _B w w0 pr_strict ac x y,
                 equal_acterma _A3 _B (aocnf (_A1, _A2, _A3) _B ac ac x)
                   (aocnf (_A1, _A2, _A3) _B ac ac y)))
             (size_list ss) ss ts)
  else (let (sa, ta) =
          (actop _A3 _B f (Fun (f, ss)), actop _A3 _B f (Fun (g, ts))) in
         (match
           mulextp
             (fun tb u ->
               (ackbo_impl (_A1, _A2, _A3) _B w w0 pr_strict ac tb u,
                 equal_acterma _A3 _B (aocnf (_A1, _A2, _A3) _B ac ac tb)
                   (aocnf (_A1, _A2, _A3) _B ac ac u)))
             (filter_fun sa (fun x y -> not (pr_strict y x)) (f, size_list ss))
             (minus_multiset (equal_term _A3 _B)
               (plus_multiseta (equal_term _A3 _B)
                 (filter_fun ta (fun x y -> not (pr_strict y x))
                   (f, size_list ss))
                 (filter_mset is_Var ta))
               (filter_mset is_Var sa))
           with (true, _) -> true
           | (false, ns) ->
             (if ns && less_nat (size_multiset ta) (size_multiset sa) then true
               else (if ns && equal_nata (size_multiset sa) (size_multiset ta)
                      then smulextp
                             (fun tb u ->
                               (ackbo_impl (_A1, _A2, _A3) _B w w0 pr_strict ac
                                  tb u,
                                 equal_acterma _A3 _B
                                   (aocnf (_A1, _A2, _A3) _B ac ac tb)
                                   (aocnf (_A1, _A2, _A3) _B ac ac u)))
                             (filter_fun sa (fun x y -> pr_strict y x)
                               (f, size_list ss))
                             (filter_fun ta (fun x y -> pr_strict y x)
                               (f, size_list ss))
                      else false)))))
                                   else false)))))
      else false);;

let rec ackbo_strict (_A1, _A2, _A3, _A4) (_B1, _B2)
  pr w w0 acset =
    (fun (s, t) ->
      check (ackbo_impl (_A1, _A2, _A3) _B1 w w0 pr acset s t)
        (comp (comp (comp (comp (showsl_lit "could not orient ")
                            (showsl_terma _A4 _B2 s))
                      (showsl_lit " >ACKBO "))
                (showsl_terma _A4 _B2 t))
          (showsl_literal "\092")));;

let rec apply_af_entry
  uu x1 ts = match uu, x1, ts with uu, Collapse i, ts -> nth ts i
    | f, AFList is, ts -> Fun (f, map (nth ts) is);;

let rec afs_term
  pi x1 = match pi, x1 with
    pi, Fun (f, ts) ->
      (let l = size_list ts in
        apply_af_entry (FPair (f, l)) (afsa pi (f, l)) (map (afs_term pi) ts))
    | pi, Var x -> Var x;;

let rec afs_rule pi lr = (afs_term pi (fst lr), afs_term pi (snd lr));;

let rec af_rule pi = (fun t -> map_funs_rule filtered_fun (afs_rule pi t));;

let rec af_term pi t = map_term filtered_fun (fun x -> x) (afs_term pi t);;

let rec nat_or_empty
  = function ([], s) -> right None
    | ([XML_text txt], s) ->
        (match int_of_string txt with Inl err -> xml_error err ([], s)
          | Inr n -> right (Some (nat n)));;

let rec uncurry_info _A
  xml2name =
    xml_do "uncurryInformation"
      (xml_take xml2name
        (fun a ->
          xml_take
            (xml_do "uncurriedSymbols"
              (xml_take_many_sub [] zero_nata Infinity_enat
                (xml_do "uncurriedSymbolEntry"
                  (xml_take xml2name
                    (fun f ->
                      xml_take (xml_nat "arity")
                        (fun n ->
                          xml_take_many_sub [] zero_nata Infinity_enat xml2name
                            (fun fs -> xml_return ((f, n), fs))))))
                xml_return))
            (fun sml ->
              xml_take
                (xml_do "uncurryRules"
                  (xml_take (rules _A xml2name) (fun x -> xml_return (id x))))
                (fun u ->
                  xml_take
                    (xml_do "etaRules"
                      (xml_take (rules _A xml2name)
                        (fun x -> xml_return (id x))))
                    (fun e -> xml_return (a, (sml, (u, e))))))));;

let rec convertible_critical_peaks _A
  xml2name =
    (let rew = rsteps _A xml2name in
     let conv = conversion _A xml2name in
      xml_do "convertibleCriticalPeaks"
        (xml_take_many_sub [] zero_nata Infinity_enat
          (xml_do "convertibleCriticalPeak"
            (xml_take
              (xml_do "source"
                (xml_take (term _A xml2name) (fun x -> xml_return (id x))))
              (fun s ->
                xml_take
                  (xml_do "conversionLeft"
                    (xml_take conv
                      (fun (sa, c1) ->
                        xml_take rew
                          (fun (_, r) ->
                            xml_take conv
                              (fun (_, c2) ->
                                xml_return (sa, (c1, (r, c2))))))))
                  (fun (s1, (cl1, (sl, cl2))) ->
                    xml_take
                      (xml_do "conversionRight"
                        (xml_take conv
                          (fun (sa, c1) ->
                            xml_take rew
                              (fun (_, r) ->
                                xml_take conv
                                  (fun (_, c2) ->
                                    xml_return (sa, (c1, (r, c2))))))))
                      (fun (s2, (cr1, (sr, cr2))) ->
                        xml_return
                          (s, (s1, (cl1, (sl,
   (cl2, (s2, (cr1, (sr, cr2)))))))))))))
          xml_return));;

let rec ordering_constraint_proof _A
  xml2name bi =
    xml_do "orderingConstraintProof"
      (xml_take
        (xml_do "redPair"
          (xml_take (redtriple _A xml2name bi) (fun x -> xml_return (id x))))
        (fun x -> xml_return (id x)));;

let rec ordering_constraint_proofa _A
  xml2name bi =
    xml_do "orderingConstraintProof"
      (xml_take (redtriplea _A xml2name bi) (fun x -> xml_return (id x)));;

let rec create_proj _A
  (Projection p) =
    (let i = ceta_map_of (compare_order_prod _A compare_order_nat) p in
      (fun f -> (match i f with None -> zero_nata | Some n -> n)));;

let rec xml2cond_constraint _A
  xml2name x =
    xml_do "conditionalConstraint"
      (xml_take
        (xml_or
          (xml_do "all"
            (xml_take (xml_text "var")
              (fun a ->
                xml_take (xml2cond_constraint _A xml2name)
                  (fun b -> xml_return (CC_all (a, b))))))
          (xml_or
            (xml_do "implication"
              (xml_take (xml2cond_constraint _A xml2name)
                (fun c ->
                  xml_take_many_sub [] zero_nata Infinity_enat
                    (xml2cond_constraint _A xml2name)
                    (fun cs ->
                      (let ccs = c :: cs in
                        xml_return
                          (CC_impl (take (size_list cs) ccs, last ccs)))))))
            (xml_do "constraint"
              (xml_take (term _A xml2name)
                (fun s ->
                  xml_take
                    (xml_or (xml_leaf "rewrite" None)
                      (xml_or (xml_leaf "strict" (Some true))
                        (xml_leaf "nonStrict" (Some false))))
                    (fun rel ->
                      xml_take (term _A xml2name)
                        (fun t ->
                          xml_return
                            (match rel with None -> CC_rewr (s, t)
                              | Some stri -> CC_cond (stri, (s, t))))))))))
        (fun xa -> xml_return (id xa)))
      x;;

let rec xml2cond_constraint_prf _A
  xml2name x =
    (let cc = xml2cond_constraint _A xml2name in
      xml_do "conditionalConstraintProof"
        (xml_take
          (xml_or (xml_leaf "final" Final)
            (xml_or
              (xml_do "differentConstructor"
                (xml_take cc (fun a -> xml_return (Different_Constructor a))))
              (xml_or
                (xml_do "sameConstructor"
                  (xml_take cc
                    (fun a ->
                      xml_take cc
                        (fun b ->
                          xml_take (xml2cond_constraint_prf _A xml2name)
                            (fun c ->
                              xml_return (Same_Constructor (a, b, c)))))))
                (xml_or
                  (xml_do "variableEquation"
                    (xml_take (xml_text "var")
                      (fun a ->
                        xml_take (term _A xml2name)
                          (fun b ->
                            xml_take cc
                              (fun c ->
                                xml_take (xml2cond_constraint_prf _A xml2name)
                                  (fun d ->
                                    xml_return
                                      (Variable_Equation (a, b, c, d))))))))
                  (xml_or
                    (xml_do "funargIntoVar"
                      (xml_take cc
                        (fun a ->
                          xml_take position
                            (fun b ->
                              xml_take (xml_text "var")
                                (fun c ->
                                  xml_take cc
                                    (fun d ->
                                      xml_take
(xml2cond_constraint_prf _A xml2name)
(fun e -> xml_return (Funarg_Into_Var (a, b, c, d, e)))))))))
                    (xml_or
                      (xml_do "simplifyCondition"
                        (xml_take cc
                          (fun a ->
                            xml_take (substa _A xml2name)
                              (fun b ->
                                xml_take cc
                                  (fun c ->
                                    xml_take
                                      (xml2cond_constraint_prf _A xml2name)
                                      (fun d ->
xml_return (Simplify_Condition (a, b, c, d))))))))
                      (xml_or
                        (xml_do "induction"
                          (xml_take cc
                            (fun a ->
                              xml_take
                                (xml_do "conjuncts"
                                  (xml_take_many_sub [] zero_nata Infinity_enat
                                    cc xml_return))
                                (fun b ->
                                  xml_take
                                    (xml_do "ruleConstraintProofs"
                                      (xml_take_many_sub [] zero_nata
Infinity_enat
(xml_do "ruleConstraintProof"
  (xml_take (rule _A xml2name)
    (fun lr ->
      xml_take
        (xml_do "subtermVarEntries"
          (xml_take_many_sub [] zero_nata Infinity_enat
            (xml_do "subtermVarEntry"
              (xml_take (term _A xml2name)
                (fun aa ->
                  xml_take_many_sub [] zero_nata Infinity_enat (xml_text "var")
                    (fun ba -> xml_return (aa, ba)))))
            xml_return))
        (fun rys ->
          xml_take cc
            (fun cca ->
              xml_take (xml2cond_constraint_prf _A xml2name)
                (fun p -> xml_return (lr, (rys, (cca, p)))))))))
xml_return))
                                    (fun c ->
                                      xml_return (Induction (a, b, c)))))))
                        (xml_do "deleteCondition"
                          (xml_take cc
                            (fun a ->
                              xml_take (xml2cond_constraint_prf _A xml2name)
                                (fun b ->
                                  xml_return
                                    (Delete_Condition (a, b)))))))))))))
          (fun xa -> xml_return (id xa)))
        x);;

let rec xml2cond_red_pair_proof _A
  xml2name =
    xml_do "condRedPairProof"
      (xml_take xml2name
        (fun c ->
          xml_take (xml_nat "before")
            (fun b ->
              xml_take (xml_nat "after")
                (fun a ->
                  xml_take
                    (xml_do "conditions"
                      (xml_take_many_sub [] zero_nata Infinity_enat
                        (xml_do "condition"
                          (xml_take (xml2cond_constraint _A xml2name)
                            (fun ca ->
                              xml_take
                                (xml_do "dpSequence"
                                  (xml_take (rules _A xml2name)
                                    (fun x -> xml_return (id x))))
                                (fun s ->
                                  xml_take (xml2cond_constraint_prf _A xml2name)
                                    (fun p -> xml_return (ca, (s, p)))))))
                        xml_return))
                    (fun ccs ->
                      xml_return (Cond_Red_Pair_Prf (c, ccs, b, a)))))));;

let rec projected_rseq (_A1, _A2)
  xml2name pi =
    xml_do "projectedRewriteSequence"
      (xml_take (rule _A2 xml2name)
        (fun r -> xml_take (rseq (_A1, _A2) xml2name pi r) xml_return));;

let default_nfs_nt_dp : bool = false;;

let rec strategy_to_Q _A _B
  x0 uu = match x0, uu with No_Strategy, uu -> []
    | Innermost, r -> remdups (equal_term _A _B) (map fst r)
    | Innermost_Q q, uv -> q;;

let default_nfs_dp : bool = true;;

let rec xml2dp_inputa (_A1, _A2)
  xml2name termination =
    xml_change (xml2dp_input _A2 xml2name)
      (fun (DP_input (m, p, q, r)) ->
        xml_return
          ((if termination then default_nfs_dp else default_nfs_nt_dp),
            (m, (p, ([], (strategy_to_Q _A1 (equal_list equal_char) q r,
                           ([], r)))))));;

let rec flat_contexts _A
  xml2name =
    xml_do "flatContexts"
      (xml_take_many_sub [] zero_nata Infinity_enat (ctxt _A xml2name)
        xml_return);;

let rec equal_start_term
  x0 x1 = match x0, x1 with Full, Constructor_Based -> false
    | Constructor_Based, Full -> false
    | Constructor_Based, Constructor_Based -> true
    | Full, Full -> true;;

let rec o_to_fp_term t = (Hole, (t, Ba));;

let rec o_to_fp_impl x = map o_to_fp_term x;;

let rec mu_to_fp_impl _A
  mu = (let fs = remdups (equal_prod _A equal_nat) (map fst mu) in
        let m = map_of (equal_prod _A equal_nat) mu in
        let a =
          maps (fun f ->
                 (let xs =
                    map (fun a -> Var a)
                      (fresh_strings_list
                        [Chara (false, false, false, true, true, true, true,
                                 false)]
                        zero_nata [] (snd f))
                    in
                   map (fun i -> (f, (xs, i)))
                     (maps (fun i ->
                             (if not (membera equal_nat (the (m f)) i) then [i]
                               else []))
                       (upt zero_nata (snd f)))))
            fs
          in
         maps (fun (aa, b) ->
                (let (f, _) = aa in
                  (fun (xs, i) ->
                    map (fun loc ->
                          (ctxt_of_pos_term [i] (Fun (f, xs)), (nth xs i, loc)))
                      [Ba; H]))
                  b)
           a);;

let rec strategy_to_fp _A x0 r = match x0, r with Forbidden_Patterns p, r -> p
                            | Outermost, r -> o_to_fp_impl (map fst r)
                            | Context_Sensitive mu, r -> mu_to_fp_impl _A mu;;

let default_nfs_trs : bool = false;;

let rec xml2inn_fp_trs_assm (_A1, _A2)
  xml2name =
    xml_change (xml2_trs_input _A2 xml2name)
      (fun a ->
        (match a with DP_input (_, _, _, _) -> xml_error "trs as input required"
          | Inn_TRS_input (inn, r, s, start) ->
            (if not (equal_start_term start Full)
              then xml_error "start term is not allowed here"
              else xml_return
                     (Inl (default_nfs_trs,
                            (strategy_to_Q _A1 (equal_list equal_char) inn r,
                              (r, s)))))
          | CPX_input (_, _, _, _, _) -> xml_error "trs as input required"
          | COMP_input (_, _) -> xml_error "trs as input required"
          | OCOMP_input (_, _, _, _) -> xml_error "trs as input required"
          | EQ_input (_, _) -> xml_error "trs as input required"
          | FP_TRS_input (fp, r) ->
            xml_return (Inr (strategy_to_fp _A1 fp r, r))
          | CTRS_input _ -> xml_error "trs as input required"
          | TA_input (_, _) -> xml_error "trs as input required"
          | AC_input (_, _, _) -> xml_error "trs as input required"
          | LTS_input _ -> xml_error "trs as input required"
          | LTS_safety_input (_, _) -> xml_error "trs as input required"
          | Unknown_input _ -> xml_error "trs as input required"));;

let rec xml2inn_trs_assm (_A1, _A2)
  xml2name =
    xml_change (xml2inn_fp_trs_assm (_A1, _A2) xml2name)
      (fun a ->
        (match a with Inl aa -> xml_return aa
          | Inr _ ->
            xml_error "innermost (relative) TRS expected at this point"));;

let rec xml2trs_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name x =
    xml_or
      (xml_do "trsTerminationProof"
        (xml_take
          (xml2trs_termination_proof_inner (_A1, _A2, _A3) (_B1, _B2, _B3)
            xml2name)
          (fun xa -> xml_return (id xa))))
      (xml_do "relativeTerminationProof"
        (xml_take
          (xml2trs_termination_proof_inner (_A1, _A2, _A3) (_B1, _B2, _B3)
            xml2name)
          (fun xa -> xml_return (id xa))))
      x
and xml2dp_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name x =
    xml_do "dpProof"
      (xml_take
        (xml_or (xml_leaf "pIsEmpty" P_is_Empty)
          (xml_or
            (xml_do "depGraphProc"
              (xml_take_many_sub [] zero_nata Infinity_enat
                (xml_do "component"
                  (xml_take
                    (xml_do "dps"
                      (xml_take (rules (showl_lab _A3 _B3) xml2name)
                        (fun xa -> xml_return (id xa))))
                    (fun dps ->
                      xml_take_optional (xml_bool "realScc")
                        (fun _ ->
                          xml_take_optional
                            (xml_do "arcs"
                              (xml_take_many_sub [] zero_nata Infinity_enat
                                (fun xa -> right (fst xa))
                                (fun _ -> xml_return ())))
                            (fun _ ->
                              xml_take_optional
                                (xml2dp_termination_proof (_A1, _A2, _A3)
                                  (_B1, _B2, _B3) xml2name)
                                (fun prfOpt -> xml_return (prfOpt, dps)))))))
                (fun a -> xml_return (Dep_Graph_Proc a))))
            (xml_or
              (xml_do "redPairProc"
                (xml_take
                  (ordering_constraint_proofa (showl_lab _A3 _B3) xml2name
                    false)
                  (fun a ->
                    xml_take
                      (xml_do "dps"
                        (xml_take (rules (showl_lab _A3 _B3) xml2name)
                          (fun xa -> xml_return (id xa))))
                      (fun b ->
                        xml_take
                          (xml2dp_termination_proof (_A1, _A2, _A3)
                            (_B1, _B2, _B3) xml2name)
                          (fun c -> xml_return (Redpair_Proc (a, b, c)))))))
              (xml_or
                (xml_do "usableRulesProc"
                  (xml_take
                    (xml_do "usableRules"
                      (xml_take (rules (showl_lab _A3 _B3) xml2name)
                        (fun xa -> xml_return (id xa))))
                    (fun a ->
                      xml_take
                        (xml2dp_termination_proof (_A1, _A2, _A3)
                          (_B1, _B2, _B3) xml2name)
                        (fun b -> xml_return (Usable_Rules_Proc (a, b))))))
                (xml_or
                  (xml_do "innermostLhssRemovalProc"
                    (xml_take (innermostLhss (showl_lab _A3 _B3) xml2name)
                      (fun a ->
                        xml_take
                          (xml2dp_termination_proof (_A1, _A2, _A3)
                            (_B1, _B2, _B3) xml2name)
                          (fun b -> xml_return (Q_Reduction_Proc (a, b))))))
                  (xml_or
                    (xml_do "rewritingProc"
                      (xml_take (rule (showl_lab _A3 _B3) xml2name)
                        (fun (s, t) ->
                          xml_take (rstep (showl_lab _A3 _B3) xml2name)
                            (fun (p, (lr, ta)) ->
                              xml_take_default (s, ta)
                                (rule (showl_lab _A3 _B3) xml2name)
                                (fun st ->
                                  xml_take_optional
                                    (xml_do "usableRules"
                                      (xml_take
(rules (showl_lab _A3 _B3) xml2name) (fun xa -> xml_return (id xa))))
                                    (fun ur ->
                                      xml_take
(xml2dp_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
(fun prof ->
  xml_return (Rewriting_Proc (ur, (s, t), (s, ta), st, lr, p, prof)))))))))
                    (xml_or
                      (xml_do "narrowingProc"
                        (xml_take (rule (showl_lab _A3 _B3) xml2name)
                          (fun a ->
                            xml_take pos
                              (fun b ->
                                xml_take
                                  (xml_do "narrowings"
                                    (xml_take
                                      (rules (showl_lab _A3 _B3) xml2name)
                                      (fun xa -> xml_return (id xa))))
                                  (fun c ->
                                    xml_take
                                      (xml2dp_termination_proof (_A1, _A2, _A3)
(_B1, _B2, _B3) xml2name)
                                      (fun d ->
xml_return (Narrowing_Proc (a, b, c, d))))))))
                      (xml_or
                        (xml_do "instantiationProc"
                          (xml_take (rule (showl_lab _A3 _B3) xml2name)
                            (fun a ->
                              xml_take
                                (xml_do "instantiations"
                                  (xml_take (rules (showl_lab _A3 _B3) xml2name)
                                    (fun xa -> xml_return (id xa))))
                                (fun b ->
                                  xml_take
                                    (xml2dp_termination_proof (_A1, _A2, _A3)
                                      (_B1, _B2, _B3) xml2name)
                                    (fun c ->
                                      xml_return
(Instantiation_Proc (a, b, c)))))))
                        (xml_or
                          (xml_do "forwardInstantiationProc"
                            (xml_take (rule (showl_lab _A3 _B3) xml2name)
                              (fun a ->
                                xml_take
                                  (xml_do "instantiations"
                                    (xml_take
                                      (rules (showl_lab _A3 _B3) xml2name)
                                      (fun xa -> xml_return (id xa))))
                                  (fun b ->
                                    xml_take_optional
                                      (xml_do "usableRules"
(xml_take (rules (showl_lab _A3 _B3) xml2name) (fun xa -> xml_return (id xa))))
                                      (fun c ->
xml_take (xml2dp_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
  (fun d -> xml_return (Forward_Instantiation_Proc (a, b, c, d))))))))
                          (xml_or
                            (xml_do "semlabProc"
                              (xml_take (sl_variant xml2name)
                                (fun sli ->
                                  xml_take
                                    (xml_do "dps"
                                      (xml_take
(rules (showl_lab _A3 _B3) xml2name) (fun xa -> xml_return (id xa))))
                                    (fun lp ->
                                      xml_take
(xml_do "trs"
  (xml_take (rules (showl_lab _A3 _B3) xml2name)
    (fun xa -> xml_return (id xa))))
(fun lr ->
  xml_take_default [] (innermostLhss (showl_lab _A3 _B3) xml2name)
    (fun lq ->
      xml_take
        (xml2dp_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
        (fun p -> xml_return (Semlab_Proc (sli, lp, lq, lr, p)))))))))
                            (xml_or
                              (xml_do "subtermProc"
                                (xml_take
                                  (xml_or
                                    (xml_change
                                      (proj (showl_lab _A3 _B3) xml2name)
                                      (comp xml_return (fun a -> Inl a)))
                                    (xml_change
                                      (multiset_af (showl_lab _A3 _B3) xml2name)
                                      (comp xml_return (fun a -> Inr a))))
                                  (fun pi_mpi ->
                                    xml_take_many_sub [] zero_nata Infinity_enat
                                      (projected_rseq
((equal_lab _A2 _B2), (showl_lab _A3 _B3)) xml2name
(create_proj
  (compare_order_lab _A1.compare_compare_order _B1.compare_compare_order)
  (match pi_mpi with Inl lpi -> lpi | Inr _ -> Projection [])))
                                      (fun seq ->
xml_take
  (xml_do "dps"
    (xml_take (rules (showl_lab _A3 _B3) xml2name)
      (fun xa -> xml_return (id xa))))
  (fun dps ->
    xml_take (xml2dp_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
      (fun prf ->
        xml_return
          (match pi_mpi
            with Inl lpi -> Subterm_Criterion_Proc (lpi, seq, dps, prf)
            | Inr mpi -> Gen_Subterm_Criterion_Proc (mpi, dps, prf))))))))
                              (xml_or
                                (xml_do "redPairUrProc"
                                  (xml_take
                                    (ordering_constraint_proofa
                                      (showl_lab _A3 _B3) xml2name false)
                                    (fun a ->
                                      xml_take
(xml_do "dps"
  (xml_take (rules (showl_lab _A3 _B3) xml2name)
    (fun xa -> xml_return (id xa))))
(fun b ->
  xml_take
    (xml_do "usableRules"
      (xml_take (rules (showl_lab _A3 _B3) xml2name)
        (fun xa -> xml_return (id xa))))
    (fun c ->
      xml_take
        (xml2dp_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
        (fun d -> xml_return (Redpair_UR_Proc (a, b, c, d))))))))
                                (xml_or
                                  (xml_do "monoRedPairUrProc"
                                    (xml_take
                                      (ordering_constraint_proof
(showl_lab _A3 _B3) xml2name false)
                                      (fun a ->
xml_take
  (xml_do "dps"
    (xml_take (rules (showl_lab _A3 _B3) xml2name)
      (fun xa -> xml_return (id xa))))
  (fun b ->
    xml_take
      (xml_do "trs"
        (xml_take (rules (showl_lab _A3 _B3) xml2name)
          (fun xa -> xml_return (id xa))))
      (fun c ->
        xml_take
          (xml_do "usableRules"
            (xml_take (rules (showl_lab _A3 _B3) xml2name)
              (fun xa -> xml_return (id xa))))
          (fun d ->
            xml_take
              (xml2dp_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
                xml2name)
              (fun e -> xml_return (Mono_Redpair_UR_Proc (a, b, c, d, e)))))))))
                                  (xml_or
                                    (xml_do "monoRedPairProc"
                                      (xml_take
(ordering_constraint_proof (showl_lab _A3 _B3) xml2name false)
(fun a ->
  xml_take
    (xml_do "dps"
      (xml_take (rules (showl_lab _A3 _B3) xml2name)
        (fun xa -> xml_return (id xa))))
    (fun b ->
      xml_take
        (xml_do "trs"
          (xml_take (rules (showl_lab _A3 _B3) xml2name)
            (fun xa -> xml_return (id xa))))
        (fun c ->
          xml_take
            (xml2dp_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
            (fun d -> xml_return (Mono_Redpair_Proc (a, b, c, d))))))))
                                    (xml_or
                                      (xml_do "innermostMonoRedPairProc"
(xml_take (ordering_constraint_proof (showl_lab _A3 _B3) xml2name false)
  (fun a ->
    xml_take
      (xml_do "deleted"
        (xml_take
          (xml_do "dps"
            (xml_take (rules (showl_lab _A3 _B3) xml2name)
              (fun xa -> xml_return (id xa))))
          (fun aa ->
            xml_take
              (xml_do "trs"
                (xml_take (rules (showl_lab _A3 _B3) xml2name)
                  (fun xa -> xml_return (id xa))))
              (fun b -> xml_return (aa, b)))))
      (fun b ->
        xml_take
          (xml2dp_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
          (fun c ->
            xml_return
              ((let (ba, ca) = b in
                 (fun d -> Mono_URM_Redpair_Proc (a, ba, ca, d)))
                c))))))
                                      (xml_or
(xml_do "uncurryProc"
  (xml_take_optional (xml_nat "applicativeTop")
    (fun a ->
      xml_take (uncurry_info (showl_lab _A3 _B3) xml2name)
        (fun b ->
          xml_take
            (xml_do "dps"
              (xml_take (rules (showl_lab _A3 _B3) xml2name)
                (fun xa -> xml_return (id xa))))
            (fun c ->
              xml_take
                (xml_do "trs"
                  (xml_take (rules (showl_lab _A3 _B3) xml2name)
                    (fun xa -> xml_return (id xa))))
                (fun d ->
                  xml_take
                    (xml2dp_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
                      xml2name)
                    (fun e -> xml_return (Uncurry_Proc (a, b, c, d, e)))))))))
(xml_or
  (xml_do "flatContextClosureProc"
    (xml_take
      (xml_do "freshSymbol" (xml_take xml2name (fun xa -> xml_return (id xa))))
      (fun a ->
        xml_take (flat_contexts (showl_lab _A3 _B3) xml2name)
          (fun b ->
            xml_take
              (xml_do "dps"
                (xml_take (rules (showl_lab _A3 _B3) xml2name)
                  (fun xa -> xml_return (id xa))))
              (fun c ->
                xml_take
                  (xml_do "trs"
                    (xml_take (rules (showl_lab _A3 _B3) xml2name)
                      (fun xa -> xml_return (id xa))))
                  (fun d ->
                    xml_take
                      (xml2dp_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
                        xml2name)
                      (fun e -> xml_return (Fcc_Proc (a, b, c, d, e)))))))))
  (xml_or
    (xml_do "switchInnermostProc"
      (xml_take (wcr_proof (showl_lab _A3 _B3) xml2name)
        (fun a ->
          xml_take
            (xml2dp_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
            (fun b -> xml_return (Switch_Innermost_Proc (a, b))))))
    (xml_or
      (xml_do "splitProc"
        (xml_take
          (xml_do "dps"
            (xml_take (rules (showl_lab _A3 _B3) xml2name)
              (fun xa -> xml_return (id xa))))
          (fun a ->
            xml_take
              (xml_do "trs"
                (xml_take (rules (showl_lab _A3 _B3) xml2name)
                  (fun xa -> xml_return (id xa))))
              (fun b ->
                xml_take
                  (xml2dp_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
                    xml2name)
                  (fun c ->
                    xml_take
                      (xml2dp_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
                        xml2name)
                      (fun d -> xml_return (Split_Proc (a, b, c, d))))))))
      (xml_or
        (xml_do "finitenessAssumption"
          (xml_take
            (xml2dp_inputa ((equal_lab _A2 _B2), (showl_lab _A3 _B3)) xml2name
              true)
            (fun xa -> xml_return (Assume_Finite (xa, [])))))
        (xml_or
          (xml_do "unknownProof"
            (xml_take (xml_text "description")
              (fun _ ->
                xml_take
                  (xml2dp_inputa ((equal_lab _A2 _B2), (showl_lab _A3 _B3))
                    xml2name true)
                  (fun b ->
                    xml_take_many_sub [] zero_nata Infinity_enat
                      (fun xa -> right (fst xa))
                      (fun _ -> xml_return (Assume_Finite (b, [])))))))
          (xml_or
            (xml_do "switchToTRS"
              (xml_take
                (xml2trs_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
                  xml2name)
                (fun xa -> xml_return (To_Trs_Proc xa))))
            (xml_or
              (xml_do "unlabProc"
                (xml_take
                  (xml_do "dps"
                    (xml_take (rules (showl_lab _A3 _B3) xml2name)
                      (fun xa -> xml_return (id xa))))
                  (fun a ->
                    xml_take
                      (xml_do "trs"
                        (xml_take (rules (showl_lab _A3 _B3) xml2name)
                          (fun xa -> xml_return (id xa))))
                      (fun b ->
                        xml_take
                          (xml2dp_termination_proof (_A1, _A2, _A3)
                            (_B1, _B2, _B3) xml2name)
                          (fun c -> xml_return (Unlab_Proc (a, b, c)))))))
              (xml_or
                (xml_do "generalRedPairProc"
                  (xml_take
                    (ordering_constraint_proof (showl_lab _A3 _B3) xml2name
                      true)
                    (fun rp ->
                      xml_take
                        (xml_do "strict"
                          (xml_take (rules (showl_lab _A3 _B3) xml2name)
                            (fun xa -> xml_return (id xa))))
                        (fun s ->
                          xml_take
                            (xml_do "bound"
                              (xml_take (rules (showl_lab _A3 _B3) xml2name)
                                (fun xa -> xml_return (id xa))))
                            (fun b ->
                              xml_take
                                (xml2cond_red_pair_proof (showl_lab _A3 _B3)
                                  xml2name)
                                (fun c ->
                                  xml_take
                                    (xml2dp_termination_proof (_A1, _A2, _A3)
                                      (_B1, _B2, _B3) xml2name)
                                    (fun ps ->
                                      xml_take_optional
(xml2dp_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
(fun pbo ->
  xml_return
    (General_Redpair_Proc
      (rp, s, b, c,
        (match pbo with None -> [ps] | Some pb -> [ps; pb])))))))))))
                (xml_or
                  (xml_do "complexConstantRemovalProc"
                    (xml_take (term (showl_lab _A3 _B3) xml2name)
                      (fun a ->
                        xml_take
                          (xml_do "ruleMap"
                            (xml_take_many_sub [] zero_nata Infinity_enat
                              (xml_do "ruleMapEntry"
                                (xml_take (rule (showl_lab _A3 _B3) xml2name)
                                  (fun aa ->
                                    xml_take (rule (showl_lab _A3 _B3) xml2name)
                                      (fun b -> xml_return (aa, b)))))
                              (fun aa -> xml_return (id aa))))
                          (fun b ->
                            xml_take
                              (xml2dp_termination_proof (_A1, _A2, _A3)
                                (_B1, _B2, _B3) xml2name)
                              (fun c ->
                                xml_return
                                  (Complex_Constant_Removal_Proc
                                    (Complex_Constant_Removal_Proof (a, b),
                                      c)))))))
                  (xml_do "sizeChangeProc"
                    (xml_take
                      (xml_or (xml_leaf "subtermCriterion" None)
                        (xml_do "reductionPair"
                          (xml_take
                            (ordering_constraint_proof (showl_lab _A3 _B3)
                              xml2name false)
                            (fun redp ->
                              xml_take_optional
                                (xml_do "usableRules"
                                  (xml_take (rules (showl_lab _A3 _B3) xml2name)
                                    (fun xa -> xml_return (id xa))))
                                (fun ur -> xml_return (Some (redp, ur)))))))
                      (fun version ->
                        xml_take_many_sub [] zero_nata Infinity_enat
                          (scg (showl_lab _A3 _B3) xml2name)
                          (fun b ->
                            (match version
                              with None ->
                                xml_return (Size_Change_Subterm_Proc b)
                              | Some redp_ur ->
                                xml_return
                                  (Size_Change_Redpair_Proc
                                    (fst redp_ur, snd redp_ur,
                                      b))))))))))))))))))))))))))))))))
        (fun xa -> xml_return (id xa)))
      x
and xml2trs_termination_proof_inner (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name x =
    xml_or (xml_leaf "rIsEmpty" R_is_Empty)
      (xml_or
        (xml_do "sIsEmpty"
          (xml_take
            (xml2trs_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
            xml_return))
        (xml_or
          (xml_do "semlab"
            (xml_take (sl_variant xml2name)
              (fun sli ->
                xml_take
                  (xml_do "trs"
                    (xml_take (rules (showl_lab _A3 _B3) xml2name)
                      (fun xa -> xml_return (id xa))))
                  (fun lr ->
                    xml_take_default []
                      (xml_do "trs"
                        (xml_take (rules (showl_lab _A3 _B3) xml2name)
                          (fun xa -> xml_return (id xa))))
                      (fun lrw ->
                        xml_take_default []
                          (innermostLhss (showl_lab _A3 _B3) xml2name)
                          (fun lq ->
                            xml_take
                              (xml2trs_termination_proof (_A1, _A2, _A3)
                                (_B1, _B2, _B3) xml2name)
                              (fun p ->
                                xml_return
                                  (Semlab (sli, lq, lr @ lrw, p)))))))))
          (xml_or
            (xml_do "split"
              (xml_take
                (xml_do "trs"
                  (xml_take (rules (showl_lab _A3 _B3) xml2name)
                    (fun xa -> xml_return (id xa))))
                (fun a ->
                  xml_take
                    (xml2trs_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
                      xml2name)
                    (fun b ->
                      xml_take
                        (xml2trs_termination_proof (_A1, _A2, _A3)
                          (_B1, _B2, _B3) xml2name)
                        (fun c -> xml_return (Split (a, b, c)))))))
            (xml_or
              (xml_do "dpTrans"
                (xml_take
                  (xml_do "dps"
                    (xml_take (rules (showl_lab _A3 _B3) xml2name)
                      (fun xa -> xml_return (id xa))))
                  (fun a ->
                    xml_take (xml_bool "markedSymbols")
                      (fun _ ->
                        xml_take
                          (xml2dp_termination_proof (_A1, _A2, _A3)
                            (_B1, _B2, _B3) xml2name)
                          (fun c ->
                            xml_return
                              (DP_Trans (default_nfs_dp, true, a, c)))))))
              (xml_or
                (xml_do "ruleRemoval"
                  (xml_take
                    (ordering_constraint_proof (showl_lab _A3 _B3) xml2name
                      false)
                    (fun ord ->
                      xml_take
                        (xml_do "trs"
                          (xml_take (rules (showl_lab _A3 _B3) xml2name)
                            (fun xa -> xml_return (id xa))))
                        (fun r ->
                          xml_take_default []
                            (xml_do "trs"
                              (xml_take (rules (showl_lab _A3 _B3) xml2name)
                                (fun xa -> xml_return (id xa))))
                            (fun s ->
                              xml_take
                                (xml2trs_termination_proof (_A1, _A2, _A3)
                                  (_B1, _B2, _B3) xml2name)
                                (fun p ->
                                  xml_return
                                    (Rule_Removal (ord, r @ s, p))))))))
                (xml_or
                  (xml_change (bounds_info xml2name)
                    (comp xml_return (fun a -> Bounds a)))
                  (xml_or
                    (xml_do "stringReversal"
                      (xml_take
                        (xml_do "trs"
                          (xml_take (rules (showl_lab _A3 _B3) xml2name)
                            (fun xa -> xml_return (id xa))))
                        (fun _ ->
                          xml_take_optional
                            (xml_do "trs"
                              (xml_take (rules (showl_lab _A3 _B3) xml2name)
                                (fun xa -> xml_return (id xa))))
                            (fun _ ->
                              xml_take
                                (xml2trs_termination_proof (_A1, _A2, _A3)
                                  (_B1, _B2, _B3) xml2name)
                                (fun prf ->
                                  xml_return (String_Reversal prf))))))
                    (xml_or
                      (xml_do "equalityRemoval"
                        (xml_take
                          (xml2trs_termination_proof (_A1, _A2, _A3)
                            (_B1, _B2, _B3) xml2name)
                          (fun xa -> xml_return (Drop_Equality xa))))
                      (xml_or
                        (xml_do "constantToUnary"
                          (xml_take plain_var
                            (fun v ->
                              xml_take (renaming (showl_lab _A3 _B3) xml2name)
                                (fun ren ->
                                  xml_take
                                    (xml_do "trs"
                                      (xml_take
(rules (showl_lab _A3 _B3) xml2name) (fun xa -> xml_return (id xa))))
                                    (fun s ->
                                      xml_take_default []
(xml_do "trs"
  (xml_take (rules (showl_lab _A3 _B3) xml2name)
    (fun xa -> xml_return (id xa))))
(fun sw ->
  xml_take (xml2trs_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
    (fun p ->
      xml_return
        (Constant_String (Const_string_sound_proof (v, ren, s, sw), p)))))))))
                        (xml_or
                          (xml_do "removeNonApplicableRules"
                            (xml_take
                              (xml_do "trs"
                                (xml_take (rules (showl_lab _A3 _B3) xml2name)
                                  (fun xa -> xml_return (id xa))))
                              (fun a ->
                                xml_take
                                  (xml2trs_termination_proof (_A1, _A2, _A3)
                                    (_B1, _B2, _B3) xml2name)
                                  (fun b ->
                                    xml_return
                                      (Remove_Nonapplicable_Rules (a, b))))))
                          (xml_or
                            (xml_do "uncurry"
                              (xml_take
                                (uncurry_info (showl_lab _A3 _B3) xml2name)
                                (fun i ->
                                  xml_take
                                    (xml_do "trs"
                                      (xml_take
(rules (showl_lab _A3 _B3) xml2name) (fun xa -> xml_return (id xa))))
                                    (fun r ->
                                      xml_take_default []
(xml_do "trs"
  (xml_take (rules (showl_lab _A3 _B3) xml2name)
    (fun xa -> xml_return (id xa))))
(fun s ->
  xml_take (xml2trs_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
    (fun p -> xml_return (Uncurry (i, r @ s, p))))))))
                            (xml_or
                              (xml_do "flatContextClosure"
                                (xml_take
                                  (flat_contexts (showl_lab _A3 _B3) xml2name)
                                  (fun i ->
                                    xml_take
                                      (xml_do "trs"
(xml_take (rules (showl_lab _A3 _B3) xml2name) (fun xa -> xml_return (id xa))))
                                      (fun r ->
xml_take_default []
  (xml_do "trs"
    (xml_take (rules (showl_lab _A3 _B3) xml2name)
      (fun xa -> xml_return (id xa))))
  (fun s ->
    xml_take
      (xml2trs_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
      (fun p -> xml_return (Fcc (i, r @ s, p))))))))
                              (xml_or
                                (xml_do "switchInnermost"
                                  (xml_take
                                    (wcr_proof (showl_lab _A3 _B3) xml2name)
                                    (fun a ->
                                      xml_take
(xml2trs_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
(fun b -> xml_return (Switch_Innermost (a, b))))))
                                (xml_or
                                  (xml_do "permutingArgumentFilter"
                                    (xml_take (afs (showl_lab _A3 _B3) xml2name)
                                      (fun a ->
xml_take (xml2trs_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
  (fun b -> xml_return (Permuting_AFS (a, b))))))
                                  (xml_or
                                    (xml_do "terminationAssumption"
                                      (xml_take
(xml2inn_trs_assm ((equal_lab _A2 _B2), (showl_lab _A3 _B3)) xml2name)
(fun xa -> xml_return (Assume_SN (xa, [])))))
                                    (xml_or
                                      (xml_do "relativeTerminationAssumption"
(xml_take
  (xml_do "trsInput"
    (xml_take
      (xml_do "trs"
        (xml_take (rules (showl_lab _A3 _B3) xml2name)
          (fun xa -> xml_return (id xa))))
      (fun r ->
        xml_take_default []
          (xml_do "relativeRules"
            (xml_take (rules (showl_lab _A3 _B3) xml2name)
              (fun xa -> xml_return (id xa))))
          (fun rw ->
            xml_return (Assume_SN ((default_nfs_trs, ([], (r, rw))), []))))))
  xml_return))
                                      (xml_do "unknownProof"
(xml_take (xml_text "description")
  (fun _ ->
    xml_take
      (xml2inn_trs_assm ((equal_lab _A2 _B2), (showl_lab _A3 _B3)) xml2name)
      (fun b ->
        xml_take_many_sub [] zero_nata Infinity_enat (fun xa -> right (fst xa))
          (fun _ -> xml_return (Assume_SN (b, [])))))))))))))))))))))))
      x;;

let rec rule_labeling_function (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "ruleLabelingFunction"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "ruleLabelingFunctionEntry"
          (xml_take (rule (showl_lab _A2 _B2) xml2name)
            (fun a -> xml_take (xml_nat "label") (fun b -> xml_return (a, b)))))
        (fun a -> xml_return (id a)));;

let rec add_source_lab_proof
  x0 uv prf = match x0, uv, prf with
    Rule_Labeling (rl, js, uu), uv, prf -> Rule_Labeling (rl, js, Some prf)
    | Rule_Labeling_Conv (rl, cs, uw), Some n, prf ->
        Rule_Labeling_Conv (rl, cs, Some (n, prf));;

let rec xml2ms_signature (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "manySortedSignature"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "manySortedFunction"
          (xml_take xml2name
            (fun a ->
              xml_take
                (xml_do "args"
                  (xml_take_many_sub [] zero_nata Infinity_enat
                    (xml_text "sort") (fun aa -> xml_return (id aa))))
                (fun b ->
                  xml_take
                    (xml_do "result"
                      (xml_take (xml_text "sort") (fun x -> xml_return (id x))))
                    (fun c -> xml_return (a, (b, c)))))))
        (fun a -> xml_return (id a)));;

let rec xml2cr_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name x =
    xml_do "crProof"
      (xml_take
        (xml_or
          (xml_do "wcrAndSN"
            (xml_take (wcr_proof (showl_lab _A3 _B3) xml2name)
              (fun a ->
                xml_take
                  (xml2trs_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
                    xml2name)
                  (fun b -> xml_return (SN_WCR (a, b))))))
          (xml_or (xml_leaf "orthogonal" Weakly_Orthogonal)
            (xml_or
              (xml_change (xml_nat "stronglyClosed")
                (comp xml_return (fun a -> Strongly_Closed a)))
              (xml_or
                (xml_change (xml_do "parallelClosed" nat_or_empty)
                  (comp xml_return (fun a -> Parallel_Closed a)))
                (xml_or
                  (xml_do "criticalPairClosingSystem"
                    (xml_take
                      (xml_do "trs"
                        (xml_take (rules (showl_lab _A3 _B3) xml2name)
                          (fun xa -> xml_return (id xa))))
                      (fun a ->
                        xml_take
                          (xml2trs_termination_proof (_A1, _A2, _A3)
                            (_B1, _B2, _B3) xml2name)
                          (fun b ->
                            xml_take (xml_nat "nrSteps")
                              (fun c ->
                                xml_return
                                  (Critical_Pair_Closing_System (a, b, c)))))))
                  (xml_or
                    (xml_do "ruleLabeling"
                      (xml_take
                        (rule_labeling_function (_A1, _A3) (_B1, _B3) xml2name)
                        (fun a ->
                          xml_take
                            (joinable_critical_pairs (showl_lab _A3 _B3)
                              xml2name)
                            (fun b ->
                              xml_return (Rule_Labeling (a, b, None))))))
                    (xml_or
                      (xml_do "decreasingDiagrams"
                        (xml_take_optional
                          (xml2trs_termination_proof (_A1, _A2, _A3)
                            (_B1, _B2, _B3) xml2name)
                          (fun a ->
                            xml_take
                              (xml_or
                                (xml_do "ruleLabeling"
                                  (xml_take
                                    (rule_labeling_function (_A1, _A3)
                                      (_B1, _B3) xml2name)
                                    (fun rl ->
                                      xml_take
(joinable_critical_pairs (showl_lab _A3 _B3) xml2name)
(fun js ->
  (match a with None -> xml_return (Rule_Labeling (rl, js, None))
    | Some prf ->
      xml_return
        (add_source_lab_proof (Rule_Labeling (rl, js, None)) None prf))))))
                                (xml_do "ruleLabelingConv"
                                  (xml_take
                                    (rule_labeling_function (_A1, _A3)
                                      (_B1, _B3) xml2name)
                                    (fun rl ->
                                      xml_take
(convertible_critical_peaks (showl_lab _A3 _B3) xml2name)
(fun cs ->
  (match a with None -> xml_return (Rule_Labeling_Conv (rl, cs, None))
    | Some prf ->
      xml_take (xml_nat "nrSteps")
        (fun n ->
          xml_return
            (add_source_lab_proof (Rule_Labeling_Conv (rl, cs, None)) (Some n)
              prf))))))))
                              xml_return)))
                      (xml_or
                        (xml_do "redundantRules"
                          (xml_take
                            (xml_do "trs"
                              (xml_take (rules (showl_lab _A3 _B3) xml2name)
                                (fun xa -> xml_return (id xa))))
                            (fun trs ->
                              xml_take (xml_nat "nrSteps")
                                (fun n ->
                                  xml_take_default []
                                    (xml_do "conversions"
                                      (xml_take_many_sub [] zero_nata
Infinity_enat (conversion (showl_lab _A3 _B3) xml2name)
(fun a -> xml_return (map snd a))))
                                    (fun cs ->
                                      xml_take
(xml2cr_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
(fun prf -> xml_return (Redundant_Rules (trs, n, cs, prf))))))))
                        (xml_do "persistentDecomposition"
                          (xml_take
                            (xml2ms_signature (_A1, _A3) (_B1, _B3) xml2name)
                            (fun siga ->
                              xml_take_many_sub [] zero_nata Infinity_enat
                                (xml_do "component"
                                  (xml_take
                                    (xml_do "trs"
                                      (xml_take
(rules (showl_lab _A3 _B3) xml2name) (fun xa -> xml_return (id xa))))
                                    (fun a ->
                                      xml_take
(xml2cr_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
(fun b -> xml_return (a, b)))))
                                (fun subs ->
                                  xml_return
                                    (Persistent_Decomposition
                                      (siga, subs))))))))))))))
        (fun xa -> xml_return (id xa)))
      x;;

let rec xml2eq_proof _A
  xml2name x =
    xml_or
      (xml_do "refl"
        (xml_take (term _A xml2name) (fun xa -> xml_return (Refl xa))))
      (xml_or
        (xml_do "sym"
          (xml_take (xml2eq_proof _A xml2name) (fun xa -> xml_return (Sym xa))))
        (xml_or
          (xml_do "trans"
            (xml_take (xml2eq_proof _A xml2name)
              (fun a ->
                xml_take (xml2eq_proof _A xml2name)
                  (fun b -> xml_return (Trans (a, b))))))
          (xml_or
            (xml_do "assm"
              (xml_take (rule _A xml2name)
                (fun a ->
                  xml_take (substa _A xml2name)
                    (fun b ->
                      xml_return
                        (Assm (a, mk_subst (equal_list equal_char)
                                    (fun aa -> Var aa) b))))))
            (xml_do "cong"
              (xml_take xml2name
                (fun a ->
                  xml_take_many_sub [] zero_nata Infinity_enat
                    (xml2eq_proof _A xml2name)
                    (fun b -> xml_return (Cong (a, b)))))))))
      x;;

let rec mk_cpx (_B1, _B2) (_C1, _C2)
  i (q, (s, (w, (cm, cc)))) = (mkc i false q s w, (cm, cc));;

let rec mk_tp i (nfs, (q, (r, rw))) = mkc i nfs q r rw;;

let rec in_vars_formula _A
  x xa1 = match x, xa1 with x, Atom a -> membera _A (vars_term_list a) x
    | x, NegAtom a -> membera _A (vars_term_list a) x
    | x, Conjunction phi_s -> list_ex (in_vars_formula _A x) phi_s
    | x, Disjunction phi_s -> list_ex (in_vars_formula _A x) phi_s;;

let rec trivial_checker _A
  lits =
    (let (asa, nas) = partition is_Atom lits in
     let pos = map get_Atom asa in
     let neg = map get_Atom nas in
      list_ex (membera _A neg) pos);;

let rec trivial_formula
  = function Disjunction phi_s -> list_ex trivial_formula phi_s
    | Conjunction phi_s -> list_all trivial_formula phi_s
    | Atom v -> false
    | NegAtom v -> false;;

let rec find_max_nat
  n uu =
    (if equal_nata n zero_nata then zero_nata
      else (if uu (minus_nata n one_nata) then minus_nata n one_nata
             else find_max_nat (minus_nata n one_nata) uu));;

let rec idx_of_tr _A
  s v = (let (_, (aa, (ab, _))) = v in
         let x = (let Some i = rbt_lookup _A ab s in
                  let true = less_eq_int zero_inta i in
                   nat i)
           in
         let xa =
           find_max_nat (as_length aa) (fun j -> less_eq_nat (as_get aa j) x) in
          xa);;

let rec gi_E (Gen_g_impl_ext (gi_V, gi_E, gi_V0, more)) = gi_E;;

let rec stat_newnode x = (fun _ -> ()) x;;

let rec is_Nil a = (match a with [] -> true | _ :: _ -> false);;

let rec as_push
  s x = (let a = s in
         let (aa, n) = a in
         let ab =
           (if equal_nata n (array_length aa)
             then array_grow aa
                    (max ord_nat (nat_of_integer (Big_int.big_int_of_int 4))
                      (times_nata (nat_of_integer (Big_int.big_int_of_int 2))
                        n))
                    x
             else aa)
           in
         let ac = array_set ab n x in
          (ac, plus_nata n one_nata));;

let rec push_code _A
  g_impl =
    (fun x (xa, (xb, (xc, xd))) ->
      (let _ = stat_newnode () in
       let y_a = as_length xa in
       let y_b = as_push xa x in
       let y_c = as_push xb y_a in
       let y_d = rbt_insert _A x (int_of_nat y_a) xc in
       let y_e =
         (if is_Nil (gi_E g_impl x) then xd
           else as_push xd (y_a, gi_E g_impl x))
         in
        (y_b, (y_c, (y_d, y_e)))));;

let rec nFQ_subset_NF_rulesc
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = nFQ_subset_NF_rules;;

let rec is_QNFc
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = is_QNF;;

let rec rulese
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = rules;;

let rec icap_impl_dpp _B
  i d = (let qr = nFQ_subset_NF_rulesc i d in
         let qnf = is_QNFc i d in
         let r = rulese i d in
         let ic = icap_impl_gen _B qr qnf (map fst r) in
          (fun s ->
            (let a =
               ceta_set_of (compare_order_list (compare_order_char, equal_char))
                 (maps vars_term_list s)
               in
              ic s a)));;

let rec as_set s i x = (let a = s in
                        let (aa, b) = a in
                         (array_set aa i x, b));;

let rec shows_kbo_repr _A
  (prs, w0) =
    comp (comp (comp (comp (comp (comp (comp
 (comp (comp (comp (showsl_literal
                     "KBO with the following precedence and weight function\092")
               (foldr
                 (fun (a, b) ->
                   (let (f, n) = a in
                     (fun (pr, (_, _)) ->
                       comp (comp (comp (comp
  (comp (comp (showsl_literal "precedence(") (showsl _A f))
    (showsl_literal "["))
  (showsl_nat n))
                                    (showsl_literal "]) = "))
                              (showsl_nat pr))
                         (showsl_literal "\092")))
                     b)
                 prs))
         (showsl_literal
           "\092precedence(_) = 0\092and the following weight\092"))
   (foldr
     (fun (a, b) ->
       (let (f, n) = a in
         (fun (_, (w, _)) ->
           comp (comp (comp (comp (comp (comp (showsl_literal "weight(")
  (showsl _A f))
                                    (showsl_literal "["))
                              (showsl_nat n))
                        (showsl_literal "]) = "))
                  (showsl_nat w))
             (showsl_literal "\092")))
         b)
     prs))
 (showsl_literal "\092weight(_) = "))
                                   (showsl_nat (suc w0)))
                             (showsl_literal "\092w0 = "))
                       (showsl_nat w0))
                 (showsl_literal
                   "\092and the following subterm coefficient functions\092"))
           (foldr
             (fun (a, b) ->
               (let (f, n) = a in
                 (fun (_, (_, scf)) ->
                   comp (comp (comp (comp (comp
    (comp (showsl_literal "scf(") (showsl _A f)) (showsl_literal "["))
                                      (showsl_nat n))
                                (showsl_literal "]) = "))
                          (if is_none scf then showsl_literal "all 1"
                            else showsl_list_nat (the scf)))
                     (showsl_literal "\092")))
                 b)
             prs))
      (showsl_literal "\092scf(_) = all 1\092");;

let rec roots_of_cm = function Derivational_Complexity f -> f
                      | Runtime_Complexity (c, d) -> d;;

let rec sym_collect
  p x1 = match p, x1 with p, Var x -> []
    | p, Fun (f, ts) ->
        (if p (Fun (f, ts)) then [f] else []) @ maps (sym_collect p) ts;;

let rec upper_triangular (_A1, _A2)
  a = all_interval_nat
        (fun i ->
          all_interval_nat (fun j -> eq _A2 (index_mat a (i, j)) (zerob _A1))
            zero_nata i)
        zero_nata (dim_row a);;

let rec list_inter _A
  x0 bs = match x0, bs with [], bs -> []
    | a :: asa, bs ->
        (if membera _A bs a then a :: list_inter _A asa bs
          else list_inter _A asa bs);;

let rec list_union _A
  x0 ys = match x0, ys with [], ys -> ys
    | x :: xs, ys ->
        (let zs = list_union _A xs ys in
          (if membera _A zs x then zs else x :: zs));;

let rec pat_of
  = function
    OC1 (rl, uu) ->
      (([], ((zero_nata, (zero_nata, [])), fst rl)),
        ([], ((zero_nata, (zero_nata, [])), snd rl)))
    | OC2 (rl, uv, uw, ux, uy, uz) ->
        (([], ((zero_nata, (zero_nata, [])), fst rl)),
          ([], ((zero_nata, (zero_nata, [])), snd rl)))
    | OC2p (rl, va, vb, vc, vd, ve) ->
        (([], ((zero_nata, (zero_nata, [])), fst rl)),
          ([], ((zero_nata, (zero_nata, [])), snd rl)))
    | OC3 (rl, vf, vg, vh, vi) ->
        (([], ((zero_nata, (zero_nata, [])), fst rl)),
          ([], ((zero_nata, (zero_nata, [])), snd rl)))
    | OC3p (rl, vj, vk, vl, vm) ->
        (([], ((zero_nata, (zero_nata, [])), fst rl)),
          ([], ((zero_nata, (zero_nata, [])), snd rl)))
    | OCDP1 (p, vn) -> p
    | OCDP2 (p, vo) -> p
    | WPEQ (p, vp) -> p
    | Lift (p, vq) -> p
    | DPOC1_1 (p, vr, vs, vt, vu) -> p
    | DPOC1_2 (p, vv, vw, vx, vy, vz) -> p
    | DPOC2 (p, wa, wb, wc, wd) -> p
    | DPOC3_1 (p, we, wf, wg, wh) -> p
    | DPOC3_2 (p, wi, wj, wk, wl, wm) -> p
    | DPDP1_1 (p, wn, wo, wp, wq) -> p
    | DPDP1_2 (p, wr, ws, wt, wu) -> p
    | DPDP2_1 (p, wv, ww, wx, wy) -> p
    | DPDP2_2 (p, wz, xa, xb, xc) -> p;;

let rec mk_rtrancl_list_main
  subsumes r todo fin =
    (match todo with [] -> fin
      | a :: tod ->
        (if list_ex (fun b -> subsumes b a) fin
          then mk_rtrancl_list_main subsumes r tod fin
          else mk_rtrancl_list_main subsumes r (r a @ tod) (a :: fin)));;

let rec mk_rtrancl_list
  subsumes r init = mk_rtrancl_list_main subsumes r init [];;

let rec ins_dj_rm_basic_ops _A x s = insert _A x () s;;

let rec mk_rtrancl_set_main _A
  r todo fin =
    (match todo with [] -> fin
      | a :: tod ->
        (if memb_rm_basic_ops _A a fin then mk_rtrancl_set_main _A r tod fin
          else mk_rtrancl_set_main _A r (r a @ tod)
                 (ins_dj_rm_basic_ops _A a fin)));;

let rec mk_rtrancl_set _A
  r init =
    mk_rtrancl_set_main _A r init
      (empty_rm_basic_ops _A.linorder_compare_order ());;

let rec add_index s i = comp (comp s (showsl_lit ".")) (showsl_nat i);;

let rec subst_compose_impl _A _B
  sigma tau =
    (let sigmaa = mk_subst_domain _A _B sigma in
     let taua = mk_subst_domain _A _B tau in
     let d_sigma = map fst sigmaa in
      map (fun (x, t) ->
            (x, subst_apply_term t (mk_subst _A (fun a -> Var a) taua)))
        sigmaa @
        filtera (fun (x, _) -> not (membera _A d_sigma x)) taua);;

let rec commutes_impl (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2)
  sigma mu =
    subst_eq (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2)
      (subst_compose_impl _A5 _B2 sigma mu)
      (subst_compose_impl _A5 _B2 mu sigma);;

let rec mk_subst_case _A _B
  xs sigma tau = subst_compose_impl _A _B (map (fun x -> (x, sigma x)) xs) tau;;

let rec ta_matcha (_A1, _A2, _A3, _A4, _A5) (_B1, _B2) (_C1, _C2)
  ta q t = ta_match (_A1, _A2, _A3, _A4, _A5) (_B1, _B2) (_C1, _C2) ta q t q;;

let rec adapt_vars (Fun (f, ts)) = Fun (f, map adapt_vars ts);;

let rec ta_member (_A1, _A2) (_C1, _C2, _C3, _C4, _C5)
  t ta =
    ground t &&
      not (less_eq_set (_C1, _C2, _C3)
            (inf_seta (_C2, _C3) (ta_final ta)
              (ta_res (_C1, _C2, _C3, _C4, _C5) (_A1, _A2) ta (adapt_vars t)))
            (bot_set (_C2, _C3, _C5)));;

let rec rhs_eps_cl_memo (_A1, _A2) (_B1, _B2, _B3, _B4, _B5) (_C1, _C2)
  memo rules =
    sup_setb (_B1, _B2, _B3, _B4, _B5)
      (image ((ceq_ta_rule _A2 _C2), (ccompare_ta_rule _A1 _C1))
        ((ceq_set (_B2, _B3, _B4.ccompare_cproper_interval)),
          (ccompare_set (_B1, _B3, _B4, _B5)), set_impl_set)
        (comp memo r_rhs) rules);;

let rec list_intera (_A1, _A2, _A3)
  = function [] -> top_set (_A1, _A2, _A3)
    | [x] -> x
    | x :: v :: va ->
        inf_seta (_A1, _A2) x (list_intera (_A1, _A2, _A3) (v :: va));;

let rec ps_states_cons_impl (_A1, _A2, _A3, _A4, _A5, _A6)
  (_B1, _B2, _B3, _B4, _B5) (_C1, _C2) (_D1, _D2)
  meps parts rules q =
    (let lhs_nth_in_Q =
       (fun (n, rs) ->
         map (fun i ->
               minus_set
                 ((ceq_set
                    (cenum_ta_rule, (ceq_ta_rule _A5 _D2),
                      (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                        _D1).ccompare_cproper_interval)),
                   (ccompare_set
                     (finite_UNIV_ta_rule, (ceq_ta_rule _A5 _D2),
                       (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                         _D1),
                       set_impl_ta_rule)))
                 (image
                   ((ceq_set (_A2, _A3, _A4.ccompare_cproper_interval)),
                     (ccompare_set (_A1, _A3, _A4, _A6)))
                   ((ceq_set
                      (cenum_ta_rule, (ceq_ta_rule _A5 _D2),
                        (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                          _D1).ccompare_cproper_interval)),
                     (ccompare_set
                       (finite_UNIV_ta_rule, (ceq_ta_rule _A5 _D2),
                         (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                           _D1),
                         set_impl_ta_rule)),
                     set_impl_set)
                   (fun p ->
                     filter
                       ((ceq_ta_rule _A5 _D2),
                         (ccompare_ta_rule _A4.ccompare_cproper_interval _D1))
                       (fun r ->
                         member (_A3, _A4.ccompare_cproper_interval)
                           (nth (r_lhs_states r) i) p)
                       rs)
                   q)
                 (inserta
                   ((ceq_set
                      (cenum_ta_rule, (ceq_ta_rule _A5 _D2),
                        (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                          _D1).ccompare_cproper_interval)),
                     (ccompare_set
                       (finite_UNIV_ta_rule, (ceq_ta_rule _A5 _D2),
                         (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                           _D1),
                         set_impl_ta_rule)))
                   (set_empty
                     ((ceq_ta_rule _A5 _D2),
                       (ccompare_ta_rule _A4.ccompare_cproper_interval _D1))
                     (of_phantom set_impl_ta_rulea))
                   (set_empty
                     ((ceq_set
                        (cenum_ta_rule, (ceq_ta_rule _A5 _D2),
                          (cproper_interval_ta_rule
                            _A4.ccompare_cproper_interval
                            _D1).ccompare_cproper_interval)),
                       (ccompare_set
                         (finite_UNIV_ta_rule, (ceq_ta_rule _A5 _D2),
                           (cproper_interval_ta_rule
                             _A4.ccompare_cproper_interval _D1),
                           set_impl_ta_rule)))
                     (of_phantom set_impl_seta))))
           (upt zero_nata n))
       in
      sup_setb
        ((finite_UNIV_set _B1.finite_UNIV_card_UNIV),
          (cenum_set (_B2, _B3, _B4.ccompare_cproper_interval, _B5)),
          (ceq_set (_B2, _B3, _B4.ccompare_cproper_interval)),
          (cproper_interval_set (_B1, _B3, _B4, _B5)), set_impl_set)
        (image
          ((ceq_prod _C1
             (ceq_prod ceq_nat
               (ceq_set
                 (cenum_ta_rule, (ceq_ta_rule _A5 _D2),
                   (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                     _D1).ccompare_cproper_interval)))),
            (ccompare_prod _C2
              (ccompare_prod ccompare_nat
                (ccompare_set
                  (finite_UNIV_ta_rule, (ceq_ta_rule _A5 _D2),
                    (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                      _D1),
                    set_impl_ta_rule)))))
          ((ceq_set
             ((cenum_set (_B2, _B3, _B4.ccompare_cproper_interval, _B5)),
               (ceq_set (_B2, _B3, _B4.ccompare_cproper_interval)),
               (cproper_interval_set
                 (_B1, _B3, _B4, _B5)).ccompare_cproper_interval)),
            (ccompare_set
              ((finite_UNIV_set _B1.finite_UNIV_card_UNIV),
                (ceq_set (_B2, _B3, _B4.ccompare_cproper_interval)),
                (cproper_interval_set (_B1, _B3, _B4, _B5)), set_impl_set)),
            set_impl_set)
          (fun (_, nrs) ->
            minus_set
              ((ceq_set (_B2, _B3, _B4.ccompare_cproper_interval)),
                (ccompare_set (_B1.finite_UNIV_card_UNIV, _B3, _B4, _B5)))
              (image
                ((ceq_list
                   (ceq_set
                     (cenum_ta_rule, (ceq_ta_rule _A5 _D2),
                       (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                         _D1).ccompare_cproper_interval))),
                  (ccompare_list
                    (ccompare_set
                      (finite_UNIV_ta_rule, (ceq_ta_rule _A5 _D2),
                        (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                          _D1),
                        set_impl_ta_rule))))
                ((ceq_set (_B2, _B3, _B4.ccompare_cproper_interval)),
                  (ccompare_set (_B1.finite_UNIV_card_UNIV, _B3, _B4, _B5)),
                  set_impl_set)
                (fun rs ->
                  rhs_eps_cl_memo (_A4.ccompare_cproper_interval, _A5)
                    (_B1.finite_UNIV_card_UNIV, _B2, _B3, _B4, _B5) (_D1, _D2)
                    meps
                    (list_intera
                      ((ceq_ta_rule _A5 _D2),
                        (ccompare_ta_rule _A4.ccompare_cproper_interval _D1),
                        set_impl_ta_rule)
                      rs))
                (listset
                  ((ceq_set
                     (cenum_ta_rule, (ceq_ta_rule _A5 _D2),
                       (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                         _D1).ccompare_cproper_interval)),
                    (ccompare_set
                      (finite_UNIV_ta_rule, (ceq_ta_rule _A5 _D2),
                        (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                          _D1),
                        set_impl_ta_rule)),
                    set_impl_set)
                  (lhs_nth_in_Q nrs)))
              (inserta
                ((ceq_set (_B2, _B3, _B4.ccompare_cproper_interval)),
                  (ccompare_set (_B1.finite_UNIV_card_UNIV, _B3, _B4, _B5)))
                (bot_set (_B3, _B4.ccompare_cproper_interval, _B5))
                (set_empty
                  ((ceq_set (_B2, _B3, _B4.ccompare_cproper_interval)),
                    (ccompare_set (_B1.finite_UNIV_card_UNIV, _B3, _B4, _B5)))
                  (of_phantom set_impl_seta))))
          parts));;

let rec ps_states_nil_impl (_A1, _A2) (_B1, _B2, _B3, _B4, _B5)
  (_C1, _C2, _C3, _C4)
  meps rules =
    (let rsz =
       filter ((ceq_ta_rule _A2 _C3), (ccompare_ta_rule _A1 _C2))
         (fun r -> null (r_lhs_states r)) rules
       in
      image (_C1, _C2)
        ((ceq_set (_B2, _B3, _B4.ccompare_cproper_interval)),
          (ccompare_set (_B1, _B3, _B4, _B5)), set_impl_set)
        (fun f ->
          rhs_eps_cl_memo (_A1, _A2) (_B1, _B2, _B3, _B4, _B5) (_C2, _C3) meps
            (filter ((ceq_ta_rule _A2 _C3), (ccompare_ta_rule _A1 _C2))
              (fun r -> eq _C3 (r_root r) f) rsz))
        (image ((ceq_ta_rule _A2 _C3), (ccompare_ta_rule _A1 _C2))
          (_C1, _C2, _C4) r_root rsz));;

let rec ps_rules_cons_impl (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2, _B3)
  (_C1, _C2)
  meps parts rules q =
    (let lhs_nth =
       (fun (n, rs) ->
         map (fun i ->
               filter
                 ((ceq_prod (ceq_set (_A2, _A3, _A4.ccompare_cproper_interval))
                    (ceq_set
                      (cenum_ta_rule, (ceq_ta_rule _A5 _C2),
                        (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                          _C1).ccompare_cproper_interval))),
                   (ccompare_prod
                     (ccompare_set (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6))
                     (ccompare_set
                       (finite_UNIV_ta_rule, (ceq_ta_rule _A5 _C2),
                         (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                           _C1),
                         set_impl_ta_rule))))
                 (fun x ->
                   not (set_eq
                         (cenum_ta_rule, (ceq_ta_rule _A5 _C2),
                           (ccompare_ta_rule _A4.ccompare_cproper_interval _C1))
                         (snd x)
                         (set_empty
                           ((ceq_ta_rule _A5 _C2),
                             (ccompare_ta_rule _A4.ccompare_cproper_interval
                               _C1))
                           (of_phantom set_impl_ta_rulea))))
                 (image
                   ((ceq_set (_A2, _A3, _A4.ccompare_cproper_interval)),
                     (ccompare_set (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6)))
                   ((ceq_prod
                      (ceq_set (_A2, _A3, _A4.ccompare_cproper_interval))
                      (ceq_set
                        (cenum_ta_rule, (ceq_ta_rule _A5 _C2),
                          (cproper_interval_ta_rule
                            _A4.ccompare_cproper_interval
                            _C1).ccompare_cproper_interval))),
                     (ccompare_prod
                       (ccompare_set (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6))
                       (ccompare_set
                         (finite_UNIV_ta_rule, (ceq_ta_rule _A5 _C2),
                           (cproper_interval_ta_rule
                             _A4.ccompare_cproper_interval _C1),
                           set_impl_ta_rule))),
                     (set_impl_prod set_impl_set set_impl_set))
                   (fun p ->
                     (p, filter
                           ((ceq_ta_rule _A5 _C2),
                             (ccompare_ta_rule _A4.ccompare_cproper_interval
                               _C1))
                           (fun r ->
                             member (_A3, _A4.ccompare_cproper_interval)
                               (nth (r_lhs_states r) i) p)
                           rs))
                   q))
           (upt zero_nata n))
       in
      sup_setb
        (finite_UNIV_ta_rule, cenum_ta_rule,
          (ceq_ta_rule
            (equal_set (_A2, _A3, _A4.ccompare_cproper_interval, _A5)) _B3),
          (cproper_interval_ta_rule
            (ccompare_set (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6)) _B2),
          set_impl_ta_rule)
        (image
          ((ceq_prod _B1
             (ceq_prod ceq_nat
               (ceq_set
                 (cenum_ta_rule, (ceq_ta_rule _A5 _C2),
                   (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                     _C1).ccompare_cproper_interval)))),
            (ccompare_prod _B2
              (ccompare_prod ccompare_nat
                (ccompare_set
                  (finite_UNIV_ta_rule, (ceq_ta_rule _A5 _C2),
                    (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                      _C1),
                    set_impl_ta_rule)))))
          ((ceq_set
             (cenum_ta_rule,
               (ceq_ta_rule
                 (equal_set (_A2, _A3, _A4.ccompare_cproper_interval, _A5))
                 _B3),
               (cproper_interval_ta_rule
                 (ccompare_set (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6))
                 _B2).ccompare_cproper_interval)),
            (ccompare_set
              (finite_UNIV_ta_rule,
                (ceq_ta_rule
                  (equal_set (_A2, _A3, _A4.ccompare_cproper_interval, _A5))
                  _B3),
                (cproper_interval_ta_rule
                  (ccompare_set (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6))
                  _B2),
                set_impl_ta_rule)),
            set_impl_set)
          (fun (f, nrs) ->
            filter
              ((ceq_ta_rule
                 (equal_set (_A2, _A3, _A4.ccompare_cproper_interval, _A5))
                 _B3),
                (ccompare_ta_rule
                  (ccompare_set (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6))
                  _B2))
              (fun r -> not (is_empty (_A1, _A3, _A4) (r_rhs r)))
              (image
                ((ceq_list
                   (ceq_prod (ceq_set (_A2, _A3, _A4.ccompare_cproper_interval))
                     (ceq_set
                       (cenum_ta_rule, (ceq_ta_rule _A5 _C2),
                         (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                           _C1).ccompare_cproper_interval)))),
                  (ccompare_list
                    (ccompare_prod
                      (ccompare_set (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6))
                      (ccompare_set
                        (finite_UNIV_ta_rule, (ceq_ta_rule _A5 _C2),
                          (cproper_interval_ta_rule
                            _A4.ccompare_cproper_interval _C1),
                          set_impl_ta_rule)))))
                ((ceq_ta_rule
                   (equal_set (_A2, _A3, _A4.ccompare_cproper_interval, _A5))
                   _B3),
                  (ccompare_ta_rule
                    (ccompare_set (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6))
                    _B2),
                  set_impl_ta_rule)
                (fun rs ->
                  TA_rule
                    (f, map fst rs,
                      rhs_eps_cl_memo (_A4.ccompare_cproper_interval, _A5)
                        (_A1.finite_UNIV_card_UNIV, _A2, _A3, _A4, _A6)
                        (_C1, _C2) meps
                        (list_intera
                          ((ceq_ta_rule _A5 _C2),
                            (ccompare_ta_rule _A4.ccompare_cproper_interval
                              _C1),
                            set_impl_ta_rule)
                          (map snd rs))))
                (listset
                  ((ceq_prod (ceq_set (_A2, _A3, _A4.ccompare_cproper_interval))
                     (ceq_set
                       (cenum_ta_rule, (ceq_ta_rule _A5 _C2),
                         (cproper_interval_ta_rule _A4.ccompare_cproper_interval
                           _C1).ccompare_cproper_interval))),
                    (ccompare_prod
                      (ccompare_set (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6))
                      (ccompare_set
                        (finite_UNIV_ta_rule, (ceq_ta_rule _A5 _C2),
                          (cproper_interval_ta_rule
                            _A4.ccompare_cproper_interval _C1),
                          set_impl_ta_rule))),
                    (set_impl_prod set_impl_set set_impl_set))
                  (lhs_nth nrs))))
          parts));;

let rec ps_rules_nil_impl (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6)
  (_C1, _C2, _C3, _C4)
  meps rules =
    (let rsz =
       filter ((ceq_ta_rule _A2 _C3), (ccompare_ta_rule _A1 _C2))
         (fun r -> null (r_lhs_states r)) rules
       in
      image (_C1, _C2)
        ((ceq_ta_rule (equal_set (_B2, _B3, _B4.ccompare_cproper_interval, _B5))
           _C3),
          (ccompare_ta_rule (ccompare_set (_B1, _B3, _B4, _B6)) _C2),
          set_impl_ta_rule)
        (fun f ->
          TA_rule
            (f, [],
              rhs_eps_cl_memo (_A1, _A2) (_B1, _B2, _B3, _B4, _B6) (_C2, _C3)
                meps
                (filter ((ceq_ta_rule _A2 _C3), (ccompare_ta_rule _A1 _C2))
                  (fun r -> eq _C3 (r_root r) f) rsz)))
        (image ((ceq_ta_rule _A2 _C3), (ccompare_ta_rule _A1 _C2))
          (_C1, _C2, _C4) r_root rsz));;

let rec memo_list_rtrancl_set (_A1, _A2, _A3, _A4)
  r = (let tr = rtrancl_list_impl _A3 r in
       let rm =
         map_of _A3
           (map (fun a -> (a, set (_A1, _A2, _A4) (tr [a])))
             (comp (remdups _A3) (map fst) r))
         in
        (fun a ->
          (match rm a
            with None -> inserta (_A1, _A2) a (bot_set (_A1, _A2, _A4))
            | Some asa -> asa)));;

let rec memo_rtrancl (_A1, _A2, _A3, _A4)
  = function
    RBT_set rbt ->
      (match ccompare _A2
        with None ->
          failwith "memo_rtrancl RBT_set: ccompare = None"
            (fun _ -> memo_rtrancl (_A1, _A2, _A3, _A4) (RBT_set rbt))
        | Some _ ->
          memo_list_rtrancl_set (_A1, _A2, _A3, _A4)
            (keysa (ccompare_prod _A2 _A2) rbt))
    | DList_set dxs ->
        (match ceq _A1
          with None ->
            failwith "memo_rtrancl DList_set: ceq = None"
              (fun _ -> memo_rtrancl (_A1, _A2, _A3, _A4) (DList_set dxs))
          | Some _ ->
            memo_list_rtrancl_set (_A1, _A2, _A3, _A4)
              (list_of_dlist (ceq_prod _A1 _A1) dxs))
    | Set_Monad xs -> memo_list_rtrancl_set (_A1, _A2, _A3, _A4) xs;;

let rec sym_parts (_A1, _A2) (_B1, _B2, _B3, _B4)
  rules =
    image ((ceq_prod _B1 ceq_nat), (ccompare_prod _B2 ccompare_nat))
      ((ceq_prod _B1
         (ceq_prod ceq_nat
           (ceq_set
             (cenum_ta_rule, (ceq_ta_rule _A2 _B3),
               (cproper_interval_ta_rule _A1 _B2).ccompare_cproper_interval)))),
        (ccompare_prod _B2
          (ccompare_prod ccompare_nat
            (ccompare_set
              (finite_UNIV_ta_rule, (ceq_ta_rule _A2 _B3),
                (cproper_interval_ta_rule _A1 _B2), set_impl_ta_rule)))),
        (set_impl_prod _B4 (set_impl_prod set_impl_nat set_impl_set)))
      (fun (f, n) ->
        (f, (n, filter ((ceq_ta_rule _A2 _B3), (ccompare_ta_rule _A1 _B2))
                  (fun r -> equal_proda _B3 equal_nat (r_sym r) (f, n)) rules)))
      (image ((ceq_ta_rule _A2 _B3), (ccompare_ta_rule _A1 _B2))
        ((ceq_prod _B1 ceq_nat), (ccompare_prod _B2 ccompare_nat),
          (set_impl_prod _B4 set_impl_nat))
        r_sym rules);;

let rec ps_ta (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2, _B3, _B4)
  ta = (let eps = ta_eps ta in
        let rules = ta_rules ta in
         (if finite
               ((finite_UNIV_prod _A1.finite_UNIV_card_UNIV
                  _A1.finite_UNIV_card_UNIV),
                 (ceq_prod _A3 _A3),
                 (ccompare_prod _A4.ccompare_cproper_interval
                   _A4.ccompare_cproper_interval))
               eps &&
               finite
                 (finite_UNIV_ta_rule, (ceq_ta_rule _A5 _B3),
                   (ccompare_ta_rule _A4.ccompare_cproper_interval _B2))
                 rules
           then (let meps =
                   memo_rtrancl (_A3, _A4.ccompare_cproper_interval, _A5, _A6)
                     eps
                   in
                 let parts =
                   sym_parts (_A4.ccompare_cproper_interval, _A5)
                     (_B1, _B2, _B3, _B4)
                     (filter
                       ((ceq_ta_rule _A5 _B3),
                         (ccompare_ta_rule _A4.ccompare_cproper_interval _B2))
                       (fun r -> not (null (r_lhs_states r))) rules)
                   in
                 let qinit =
                   ps_states_nil_impl (_A4.ccompare_cproper_interval, _A5)
                     (_A1.finite_UNIV_card_UNIV, _A2, _A3, _A4, _A6)
                     (_B1, _B2, _B3, _B4) meps rules
                   in
                 let q =
                   fst (whilea
                         (fun (qold, qnew) ->
                           not (less_eq_set
                                 ((cenum_set
                                    (_A2, _A3, _A4.ccompare_cproper_interval,
                                      _A6)),
                                   (ceq_set
                                     (_A2, _A3, _A4.ccompare_cproper_interval)),
                                   (ccompare_set
                                     (_A1.finite_UNIV_card_UNIV, _A3, _A4,
                                       _A6)))
                                 qnew qold))
                         (fun (_, qnew) ->
                           (qnew,
                             sup_seta
                               ((ceq_set
                                  (_A2, _A3, _A4.ccompare_cproper_interval)),
                                 (ccompare_set
                                   (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6)))
                               qinit
                               (ps_states_cons_impl
                                 (_A1.finite_UNIV_card_UNIV, _A2, _A3, _A4, _A5,
                                   _A6)
                                 (_A1, _A2, _A3, _A4, _A6) (_B1, _B2) (_B2, _B3)
                                 meps parts rules qnew)))
                         (set_empty
                            ((ceq_set
                               (_A2, _A3, _A4.ccompare_cproper_interval)),
                              (ccompare_set
                                (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6)))
                            (of_phantom set_impl_seta),
                           qinit))
                   in
                 let final =
                   filter
                     ((ceq_set (_A2, _A3, _A4.ccompare_cproper_interval)),
                       (ccompare_set
                         (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6)))
                     (fun qa ->
                       not (is_empty (_A1, _A3, _A4)
                             (inf_seta (_A3, _A4.ccompare_cproper_interval) qa
                               (ta_final ta))))
                     q
                   in
                 let rulesa =
                   sup_seta
                     ((ceq_ta_rule
                        (equal_set
                          (_A2, _A3, _A4.ccompare_cproper_interval, _A5))
                        _B3),
                       (ccompare_ta_rule
                         (ccompare_set
                           (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6))
                         _B2))
                     (ps_rules_nil_impl (_A4.ccompare_cproper_interval, _A5)
                       (_A1.finite_UNIV_card_UNIV, _A2, _A3, _A4, _A5, _A6)
                       (_B1, _B2, _B3, _B4) meps rules)
                     (ps_rules_cons_impl (_A1, _A2, _A3, _A4, _A5, _A6)
                       (_B1, _B2, _B3) (_B2, _B3) meps parts rules q)
                   in
                  Ta_ext
                    (final, rulesa,
                      set_empty
                        ((ceq_prod
                           (ceq_set (_A2, _A3, _A4.ccompare_cproper_interval))
                           (ceq_set (_A2, _A3, _A4.ccompare_cproper_interval))),
                          (ccompare_prod
                            (ccompare_set
                              (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6))
                            (ccompare_set
                              (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6))))
                        (of_phantom (set_impl_proda set_impl_set set_impl_set)),
                      ()))
           else ps_ta (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2, _B3, _B4) ta));;

let rec add_funas_args_term t fs = foldr add_funas_term (args t) fs;;

let rec add_funas_args_rule
  r fs = add_funas_args_term (fst r) (add_funas_args_term (snd r) fs);;

let rec funas_args_trs_list trs = foldr add_funas_args_rule trs [];;

let rec rwc
  (Tp_ops_ext
    (qreltrs, q, r, rw, rules, q_empty, is_QNF, nFQ_subset_NF_rules, rules_map,
      delete_R_Rw, split_rules, mk, nfs, more))
    = rw;;

let rec rc
  (Tp_ops_ext
    (qreltrs, q, r, rw, rules, q_empty, is_QNF, nFQ_subset_NF_rules, rules_map,
      delete_R_Rw, split_rules, mk, nfs, more))
    = r;;

let rec uncurry_of_sig_list
  a sml sm =
    maps (fun (b, c) ->
           (let (f, n) = b in
             (fun _ ->
               (let g = get_symbol sm f n in
                 map (fun i ->
                       (Fun (a, [generate_f_xs (g i) (plus_nata n i);
                                  Var (generate_var (plus_nata n i))]),
                         generate_f_xs (g (suc i)) (plus_nata n (suc i))))
                   (upt zero_nata (aarity sm f n)))))
             c)
      sml;;

let rec sig_list_to_sig_map _A
  a sml fmap =
    (let fm = fmap sml in
      (fun f n ->
        (match map_of (equal_prod _A equal_nat) sml (f, n) with None -> [fm f n]
          | Some xs -> (if null xs then [fm f n] else xs))));;

let rec uncurry_eta_split _A (_B1, _B2, _B3)
  eboth rtest =
    (let test =
       (fun (l, r) ->
         less_nat zero_nata (size_list (args l)) &&
           (less_nat zero_nata (size_list (args r)) &&
             list_ex
               (eq_rule_mod_vars _A (_B1, _B2, _B3)
                 (hda (args l), hda (args r)))
               rtest))
       in
      partition test eboth);;

let rec insert_vars_rule _B
  r xs = insert_vars_term _B (fst r) (insert_vars_term _B (snd r) xs);;

let rec eta_closed_rules (_A1, _A2) (_B1, _B2, _B3, _B4)
  a sm ra r =
    catch_error
      (forallM
        (fun (l, raa) ->
          (match aarity_term _A1 a sm l with None -> Inr ()
            | Some ab ->
              (if equal_nata ab zero_nata then Inr ()
                else check (list_ex
                             (fun (lll, rrr) ->
                               (match (lll, rrr) with (Var _, _) -> false
                                 | (Fun (_, []), _) -> false
                                 | (Fun (_, [_]), _) -> false
                                 | (Fun (_, [_; Var _]), Var _) -> false
                                 | (Fun (_, [_; Var _]), Fun (_, [])) -> false
                                 | (Fun (_, [_; Var _]), Fun (_, [_])) -> false
                                 | (Fun (f, [ll; Var x]), Fun (g, [rr; Var y]))
                                   -> eq _A1 f a &&
(eq _A1 g a &&
  (eq _B2 x y &&
    (not (membera _B2 (insert_vars_rule _B2 (ll, rr) []) x) &&
      instance_rule _A1 _B2 (_B1, _B2, _B3) (l, raa) (ll, rr))))
                                 | (Fun (_, [_; Var _]),
                                     Fun (_, _ :: Var _ :: _ :: _))
                                   -> false
                                 | (Fun (_, [_; Var _]),
                                     Fun (_, _ :: Fun (_, _) :: _))
                                   -> false
                                 | (Fun (_, _ :: Var _ :: _ :: _), _) -> false
                                 | (Fun (_, _ :: Fun (_, _) :: _), _) -> false))
                             r)
                       (comp (comp (showsl_lit "eta expansion of ")
                               (showsl_rule _A2 _B4 (l, raa)))
                         (showsl_lit " missing")))))
        ra)
      (fun x -> Inl (snd x));;

let rec uncurry_rules _A
  a sm = map (fun (l, r) -> (uncurry_term _A a sm l, uncurry_term _A a sm r));;

let rec check_CS_subseteq _A (_B1, _B2, _B3)
  r s = catch_error
          (forallM
            (fun (l, ra) ->
              check (list_ex (instance_rule _A _B2 (_B1, _B2, _B3) (l, ra)) s)
                (l, ra))
            r)
          (fun x -> Inl (snd x));;

let rec uncurry_tt (_B1, _B2)
  i info r tp =
    (let (a, (sml, (u, eb))) = info in
     let ra = rc i tp in
     let rw = rwc i tp in
     let (e, ew) =
       uncurry_eta_split _B1
         ((ccompare_list ccompare_char), (equal_list equal_char),
           mapping_impl_list)
         eb ra
       in
     let r_eta = e @ ra in
     let rw_eta = ew @ rw in
     let rb_eta = r_eta @ rw_eta in
     let fmap = (fun _ f _ -> f) in
     let sm = sig_list_to_sig_map _B1 a sml fmap in
     let uR = uncurry_rules _B1 a sm r_eta in
     let uRw = uncurry_rules _B1 a sm rw_eta in
      (match
        (let s = uncurry_of_sig_list a sml sm in
          bindb (catch_error
                  (forallM
                    (fun (l, _) ->
                      check (hvf_term _B1 a l)
                        (comp (comp (showsl_lit "head variable in lhs ")
                                (showsl_terma _B2 (showl_list showl_char) l))
                          (showsl_lit " not allowed")))
                    r_eta)
                  (fun x -> Inl (snd x)))
            (fun _ ->
              bindb (catch_error
                      (forallM
                        (fun (l, _) ->
                          check (hvf_term _B1 a l)
                            (comp (comp (showsl_lit "head variable in lhs ")
                                    (showsl_terma _B2 (showl_list showl_char)
                                      l))
                              (showsl_lit " not allowed")))
                        rw_eta)
                      (fun x -> Inl (snd x)))
                (fun _ ->
                  bindb (eta_closed_rules (_B1, _B2)
                          ((ccompare_list ccompare_char),
                            (equal_list equal_char), mapping_impl_list,
                            (showl_list showl_char))
                          a sm r_eta r_eta)
                    (fun _ ->
                      bindb (eta_closed_rules (_B1, _B2)
                              ((ccompare_list ccompare_char),
                                (equal_list equal_char), mapping_impl_list,
                                (showl_list showl_char))
                              a sm rb_eta rb_eta)
                        (fun _ ->
                          bindb (catch_error
                                  (forallM
                                    (fun (l, rb) ->
                                      check (not (is_Var l))
(comp (showsl_lit "lhs must not be a variable in rule ")
  (showsl_rule _B2 (showl_list showl_char) (l, rb))))
                                    rw_eta)
                                  (fun x -> Inl (snd x)))
                            (fun _ ->
                              bindb (catch_error
                                      (check_subseteq
(equal_prod (equal_term _B1 (equal_list equal_char))
  (equal_term _B1 (equal_list equal_char)))
uR r)
                                      (fun x ->
Inl (comp (comp (showsl_lit "uncurried rule ")
            (showsl_rule _B2 (showl_list showl_char) x))
      (showsl_lit " is missing"))))
                                (fun _ ->
                                  bindb (catch_error
  (check_subseteq
    (equal_prod (equal_term _B1 (equal_list equal_char))
      (equal_term _B1 (equal_list equal_char)))
    uRw r)
  (fun x ->
    Inl (comp (comp (showsl_lit "uncurried rule ")
                (showsl_rule _B2 (showl_list showl_char) x))
          (showsl_lit " is missing"))))
                                    (fun _ ->
                                      bindb
(catch_error
  (check_CS_subseteq _B1
    ((ccompare_list ccompare_char), (equal_list equal_char), mapping_impl_list)
    s u)
  (fun x ->
    Inl (comp (comp (showsl_lit "uncurry rule ")
                (showsl_rule _B2 (showl_list showl_char) x))
          (showsl_lit " is missing"))))
(fun _ ->
  catch_error
    (check_subseteq
      (equal_prod (equal_term _B1 (equal_list equal_char))
        (equal_term _B1 (equal_list equal_char)))
      u r)
    (fun x ->
      Inl (comp (comp (showsl_lit "uncurry rule ")
                  (showsl_rule _B2 (showl_list showl_char) x))
            (showsl_lit " is missing in new TRS"))))))))))))
        with Inl aa -> Inl aa
        | Inr _ -> Inr (mkc i (nfsb i tp) [] uR (uRw @ u))));;

let rec check_prefix_equivalent (_A1, _A2) (_B1, _B2)
  rhoa rho n =
    catch_error
      (bindb
        (check (less_nat n (size_list (snd rhoa)))
          (comp (comp (comp (showsl_lit "There are fewer than ") (showsl_nat n))
                  (showsl_lit " conditions in "))
            (showsl_crule _A2 _B2 _A2 _B2 rhoa)))
        (fun _ ->
          bindb (check (less_nat n (size_list (snd rho)))
                  (comp (comp (comp (showsl_lit "There are fewer than ")
                                (showsl_nat n))
                          (showsl_lit " conditions in "))
                    (showsl_crule _A2 _B2 _A2 _B2 rho)))
            (fun _ ->
              bindb (check
                      (equal_terma _A1 _B1 (fst (fst rhoa)) (fst (fst rho)))
                      (showsl_lit "Left-hand sides are different."))
                (fun _ ->
                  bindb (catch_error
                          (forallM
                            (fun i ->
                              check (equal_terma _A1 _B1
                                      (snd (nth (snd rhoa) i))
                                      (snd (nth (snd rho) i)))
                                (showsl_lit
                                  "Rhs of conditions are different\092"))
                            (upt zero_nata n))
                          (fun x -> Inl (snd x)))
                    (fun _ ->
                      catch_error
                        (forallM
                          (fun i ->
                            check (equal_terma _A1 _B1 (fst (nth (snd rhoa) i))
                                    (fst (nth (snd rho) i)))
                              (showsl_lit
                                "Lhs of conditions are different\092"))
                          (upt zero_nata (suc n)))
                        (fun x -> Inl (snd x)))))))
      (fun x ->
        Inl (comp (comp (comp (comp (comp (comp
    (comp (showsl_lit "Rules") (showsl_crule _A2 _B2 _A2 _B2 rhoa))
    (showsl_lit " and "))
                                      (showsl_crule _A2 _B2 _A2 _B2 rho))
                                (showsl_lit " are not "))
                          (showsl_nat n))
                    (showsl_lit " equivalent.\092"))
              x));;

let rec check_f (_A1, _A2) (_B1, _B2)
  cr j f crs u =
    catch_error
      (forallM
        (fun cra ->
          catch_error
            (forallM
              (fun i ->
                (match u cra i with Hole -> Inr ()
                  | More (g, _, _, _) ->
                    (if eq _A1 f g
                      then catch_error
                             (bindb
                               (check (equal_nata i j)
                                 (showsl_lit
                                   "Same symbol occurs at different levels\092"))
                               (fun _ ->
                                 bindb (catch_error
 (forallM
   (fun k ->
     check (equal_ctxta _A1 _B1 (u cra k) (u cr k))
       (showsl_lit "Contexts are different\092"))
   (upt zero_nata (suc j)))
 (fun x -> Inl (snd x)))
                                   (fun _ ->
                                     check_prefix_equivalent (_A1, _A2)
                                       (_B1, _B2) cra cr j)))
                             (fun x ->
                               Inl (comp (comp
   (comp (comp (comp (showsl_lit "Rules") (showsl_crule _A2 _B2 _A2 _B2 cra))
           (showsl_lit " and "))
     (showsl_crule _A2 _B2 _A2 _B2 cr))
   (showsl_lit " share a symbol.\092"))
                                     x))
                      else Inr ())))
              (upt zero_nata (size_list (snd cra))))
            (fun x -> Inl (snd x)))
        crs)
      (fun x -> Inl (snd x));;

let rec ackbo_nstrict (_A1, _A2, _A3, _A4) (_B1, _B2)
  pr w w0 acset =
    (fun (s, t) ->
      check (ackbo_impl (_A1, _A2, _A3) _B1 w w0 pr acset s t ||
              equal_acterma _A3 _B1 (aocnf (_A1, _A2, _A3) _B1 acset acset s)
                (aocnf (_A1, _A2, _A3) _B1 acset acset t))
        (comp (comp (comp (comp (showsl_lit "could not orient ")
                            (showsl_terma _A4 _B2 s))
                      (showsl_lit " >=ACKBO "))
                (showsl_terma _A4 _B2 t))
          (showsl_literal "\092")));;

let rec af_rules pi r = map (af_rule pi) r;;

let rec afs_syms xa = snd (rep_afs xa);;

let rec mono_af_entry
  n x1 = match n, x1 with n, Collapse i -> less_eq_nat n one_nata
    | n, AFList ids -> all_interval_nat (membera equal_nat ids) zero_nata n;;

let rec mono_afs (_A1, _A2)
  pi = ball ((ceq_prod _A1 ceq_nat), (ccompare_prod _A2 ccompare_nat))
         (afs_syms pi) (fun (f, n) -> mono_af_entry n (afsa pi (f, n)));;

let rec compatible_ta (_A1, _A2) (_B1, _B2)
  xml2name tag =
    xml_do tag
      (xml_take (tree_automaton (ta_normal_lhs xml2name))
        (fun a ->
          xml_take_default Id_Relation closed_criterion
            (fun b -> xml_return (a, b))));;

let rec xml2const_map (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name =
    xml_do "constMap"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "entry"
          (xml_take
            (xml_do "symbol" (xml_take xml2name (fun x -> xml_return (id x))))
            (fun a ->
              xml_take (xml_text "const") (fun b -> xml_return (a, b)))))
        (fun a -> xml_return (map_of (equal_lab _A2 _B2) a)));;

let rec default_grd_fun
  s t = (let f = funs_rule_list (s, t) in
         let m =
           fold (fun fa m ->
                  (match fa with Lab (_, _) -> m | FunLab (_, _) -> m
                    | UnLab _ -> m | Sharp (Lab (_, _)) -> m
                    | Sharp (FunLab (_, _)) -> m
                    | Sharp (UnLab g) -> max ord_nat (size_list g) m
                    | Sharp (Sharp _) -> m))
             f zero_nata
           in
         let suffix =
           replicate (suc m)
             (Chara (true, false, false, false, false, true, true, false))
           in
          (fun x -> Fun (Sharp (UnLab (x @ suffix)), [])));;

let rec xml2non_join_info
  xml2name x =
    xml_or (xml_leaf "distinctNormalForms" Diff_NFs)
      (xml_or (xml_leaf "capNotUnif" (Tcap_Non_Unif default_grd_fun))
        (xml_or
          (xml_do "subterm"
            (xml_take pos
              (fun a ->
                xml_take (xml2non_join_info xml2name)
                  (fun b -> xml_return (Subterm_NJ (a, b))))))
          (xml_or
            (xml_do "grounding"
              (xml_take
                (substa
                  (showl_lab (showl_list showl_char) (showl_list showl_nat))
                  xml2name)
                (fun a ->
                  xml_take (xml2non_join_info xml2name)
                    (fun b -> xml_return (Grounding (a, b))))))
            (xml_or
              (xml_do "emptyTreeAutomataIntersection"
                (xml_take
                  (compatible_ta
                    ((compare_order_list (compare_order_char, equal_char)),
                      (showl_list showl_char))
                    ((compare_order_list (compare_order_nat, equal_nat)),
                      (showl_list showl_nat))
                    xml2name "firstAutomaton")
                  (fun a ->
                    xml_take
                      (compatible_ta
                        ((compare_order_list (compare_order_char, equal_char)),
                          (showl_list showl_char))
                        ((compare_order_list (compare_order_nat, equal_nat)),
                          (showl_list showl_nat))
                        xml2name "secondAutomaton")
                      (fun b ->
                        xml_return
                          ((let (ta1, rel1) = a in
                             (fun (aa, ba) ->
                               Tree_Aut_Intersect_Empty (ta1, rel1, aa, ba)))
                            b)))))
              (xml_or
                (xml_do "differentInterpretation"
                  (xml_take (sl_variant xml2name)
                    (fun xa -> xml_return (Finite_Model_Gt xa))))
                (xml_or
                  (xml_do "strictDecrease"
                    (xml_take
                      (ordering_constraint_proof
                        (showl_lab (showl_list showl_char)
                          (showl_list showl_nat))
                        xml2name false)
                      (fun xa -> xml_return (Reduction_Pair_Gt xa))))
                  (xml_or
                    (xml_do "argumentFilterNonJoin"
                      (xml_take
                        (afs (showl_lab (showl_list showl_char)
                               (showl_list showl_nat))
                          xml2name)
                        (fun a ->
                          xml_take (xml2non_join_info xml2name)
                            (fun b -> xml_return (Argument_Filter_NJ (a, b))))))
                    (xml_do "usableRulesNonJoin"
                      (xml_take_optional
                        (xml_or (xml_leaf "left" true) (xml_leaf "right" false))
                        (fun a ->
                          (match a
                            with None ->
                              xml_take (xml2non_join_info xml2name)
                                (fun b -> xml_return (Usable_Rules_Reach_NJ b))
                            | Some left ->
                              xml_take
                                (xml_do "usableRules"
                                  (xml_take
                                    (rules
                                      (showl_lab (showl_list showl_char)
(showl_list showl_nat))
                                      xml2name)
                                    (fun xa -> xml_return (id xa))))
                                (fun u ->
                                  xml_take (xml2non_join_info xml2name)
                                    (fun p ->
                                      xml_return
(Usable_Rules_Reach_Unif_NJ ((if left then Inl u else Inr u), p)))))))))))))))
      x;;

let rec xml2ncr_proof
  xml2name x =
    (let rew =
       rsteps (showl_lab (showl_list showl_char) (showl_list showl_nat))
         xml2name
       in
      xml_do "crDisproof"
        (xml_take
          (xml_or
            (xml_do "nonWcrAndSN"
              (xml_take (fun xa -> right (fst xa))
                (fun _ ->
                  xml_take
                    (xml2trs_termination_proof
                      ((compare_order_list (compare_order_char, equal_char)),
                        (equal_list equal_char), (showl_list showl_char))
                      ((compare_order_list (compare_order_nat, equal_nat)),
                        (equal_list equal_nat), (showl_list showl_nat))
                      xml2name)
                    (fun prf -> xml_return (SN_NWCR prf)))))
            (xml_or
              (xml_do "nonJoinableFork"
                (xml_take rew
                  (fun a ->
                    xml_take rew
                      (fun b ->
                        xml_take (xml2non_join_info xml2name)
                          (fun c ->
                            xml_return
                              ((let (s, seq1) = a in
                                 (fun (_, aa) ->
                                   (fun ba -> Non_Join (s, seq1, aa, ba))))
                                 b
                                c))))))
              (xml_or
                (xml_do "modularityDisjoint"
                  (xml_take
                    (xml_do "trs"
                      (xml_take
                        (rules
                          (showl_lab (showl_list showl_char)
                            (showl_list showl_nat))
                          xml2name)
                        (fun xa -> xml_return (id xa))))
                    (fun a ->
                      xml_take (xml2ncr_proof xml2name)
                        (fun b -> xml_return (NCR_Disj_Subtrs (a, b))))))
                (xml_or
                  (xml_do "redundantRules"
                    (xml_take
                      (xml_do "trs"
                        (xml_take
                          (rules
                            (showl_lab (showl_list showl_char)
                              (showl_list showl_nat))
                            xml2name)
                          (fun xa -> xml_return (id xa))))
                      (fun a ->
                        xml_take (xml_nat "nrSteps")
                          (fun b ->
                            xml_take (xml2ncr_proof xml2name)
                              (fun c ->
                                xml_return (NCR_Redundant_Rules (a, b, c)))))))
                  (xml_do "persistentDecomposition"
                    (xml_take
                      (xml2ms_signature
                        ((compare_order_list (compare_order_char, equal_char)),
                          (showl_list showl_char))
                        ((compare_order_list (compare_order_nat, equal_nat)),
                          (showl_list showl_nat))
                        xml2name)
                      (fun a ->
                        xml_take
                          (xml_do "component"
                            (xml_take
                              (xml_do "trs"
                                (xml_take
                                  (rules
                                    (showl_lab (showl_list showl_char)
                                      (showl_list showl_nat))
                                    xml2name)
                                  (fun xa -> xml_return (id xa))))
                              (fun aa ->
                                xml_take (xml2ncr_proof xml2name)
                                  (fun b -> xml_return (aa, b)))))
                          (fun b ->
                            xml_return
                              (let (ba, c) = b in
                                NCR_Persistent_Decomposition (a, ba, c))))))))))
          (fun xa -> xml_return (id xa)))
        x);;

let rec xml2state_map (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "stateMap"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "entry"
          (xml_take state
            (fun a ->
              xml_take (term (showl_lab _A2 _B2) xml2name)
                (fun b -> xml_return (a, b)))))
        (fun a ->
          xml_return (fun x -> the (map_of (equal_list equal_char) a x))));;

let rec is_root_step _A (_B1, _B2, _B3)
  r s t =
    list_ex
      (fun (l, ra) ->
        (match matcha _A (_B1, _B2, _B3) s l with None -> false
          | Some sigma -> equal_terma _A _B2 (subst_apply_term ra sigma) t))
      r;;

let rec is_par_rstep _A (_B1, _B2, _B3)
  r x1 x2 = match r, x1, x2 with r, Var x, Var y -> eq _B2 x y
    | r, Fun (f, ss), Fun (g, ts) ->
        equal_terma _A _B2 (Fun (f, ss)) (Fun (g, ts)) ||
          (is_root_step _A (_B1, _B2, _B3) r (Fun (f, ss)) (Fun (g, ts)) ||
            (if eq _A f g && equal_nata (size_list ss) (size_list ts)
              then list_all2 (is_par_rstep _A (_B1, _B2, _B3) r) ss ts
              else false))
    | r, Fun (v, va), Var vb ->
        is_root_step _A (_B1, _B2, _B3) r (Fun (v, va)) (Var vb)
    | r, Var vb, Fun (v, va) ->
        is_root_step _A (_B1, _B2, _B3) r (Var vb) (Fun (v, va));;

let rec reachable_terms _A (_B1, _B2, _B3)
  r s n =
    (if equal_nata n zero_nata then [s]
      else (let ts =
              reachable_terms _A (_B1, _B2, _B3) r s (minus_nata n one_nata) in
             remdups (equal_term _A _B2)
               (ts @ maps (rewrite _A (_B1, _B2, _B3) r) ts)));;

let rec is_critical_pair_closing_cp _A (_B1, _B2, _B3)
  c n x2 = match c, n, x2 with
    c, n, (false, (s, t)) ->
      not (is_none
            (find (is_par_rstep _A (_B1, _B2, _B3) c t)
              (reachable_terms _A (_B1, _B2, _B3) c s n)))
    | c, n, (true, (s, t)) ->
        not (is_none
              (find (membera (equal_term _A _B2)
                      (reachable_terms _A (_B1, _B2, _B3) c s n))
                (reachable_terms _A (_B1, _B2, _B3) c t n)));;

let rec subst_apply_ctxt
  x0 sigma = match x0, sigma with Hole, sigma -> Hole
    | More (f, ss1, d, ss2), sigma ->
        More (f, map (fun t -> subst_apply_term t sigma) ss1,
               subst_apply_ctxt d sigma,
               map (fun t -> subst_apply_term t sigma) ss2);;

let rec critical_pairs_impl _A
  p r = maps (fun (l, ra) ->
               maps (fun pa ->
                      (let c = ctxt_of_pos_term pa l in
                       let la = subt_at l pa in
                       let b = equal_ctxta _A (equal_list equal_char) c Hole in
                        (if is_Var la then []
                          else maps (fun (laa, rb) ->
                                      (match
mgu_var_disjoint_generic (equal_list equal_char) _A
  (fun a -> Chara (false, false, false, true, true, true, true, false) :: a)
  (fun a -> Chara (true, false, false, true, true, true, true, false) :: a) la
  laa
with None -> []
| Some (sigma, tau) ->
  [(b, (subst_apply_term ra sigma,
         ctxt_apply_term (subst_apply_ctxt c sigma)
           (subst_apply_term rb tau)))]))
                                 r)))
                 (poss_list l))
          p;;

let rec check_left_linear_trs _A (_B1, _B2, _B3, _B4)
  trs = catch_error
          (catch_error
            (forallM
              (fun x ->
                (if linear_term (_B1, _B2, _B3) (fst x) then Inr () else Inl x))
              trs)
            (fun x -> Inl (snd x)))
          (fun _ ->
            Inl (comp (showsl_trs _A _B4 trs)
                  (showsl_literal "\092is not left-linear\092")));;

let rec check_critical_pair_closing (_A1, _A2)
  r c n =
    catch_error
      (bindb
        (check_left_linear_trs _A2
          ((ceq_list ceq_char), (ccompare_list ccompare_char), set_impl_list,
            (showl_list showl_char))
          r)
        (fun _ ->
          bindb (catch_error
                  (check_subseteq
                    (equal_prod (equal_term _A1 (equal_list equal_char))
                      (equal_term _A1 (equal_list equal_char)))
                    c r)
                  (fun _ -> Inl (showsl_lit "C not a subsystem of R")))
            (fun _ ->
              catch_error
                (forallM
                  (fun (b, (s, t)) ->
                    check (is_critical_pair_closing_cp _A1
                            ((ccompare_list ccompare_char),
                              (equal_list equal_char), mapping_impl_list)
                            c n (b, (s, t)))
                      (comp (comp (comp (comp
  (comp (comp (showsl_lit "the critical pair ")
          (showsl_terma _A2 (showl_list showl_char) s))
    (showsl_lit " <- . -> "))
  (showsl_terma _A2 (showl_list showl_char) t))
                                    (showsl_lit " is not closed within "))
                              (showsl_nat n))
                        (showsl_lit " steps.")))
                  (critical_pairs_impl _A1 r r))
                (fun x -> Inl (snd x)))))
      (fun x ->
        Inl (comp (comp x
                    (showsl_lit
                      "\092hence the following TRS is not critical pair closing\092"))
              (showsl_trs _A2 (showl_list showl_char) r)));;

let rec check_strongly_closed (_A1, _A2)
  r n = catch_error
          (bindb
            (check_linear_trs _A2
              ((ceq_list ceq_char), (ccompare_list ccompare_char),
                set_impl_list, (showl_list showl_char))
              r)
            (fun _ ->
              catch_error
                (forallM
                  (fun (_, (s, t)) ->
                    check (not (is_none
                                 (find (membera
 (equal_term _A1 (equal_list equal_char))
 (reachable_terms _A1
   ((ccompare_list ccompare_char), (equal_list equal_char), mapping_impl_list) r
   s n))
                                   (reachable_terms _A1
                                     ((ccompare_list ccompare_char),
                                       (equal_list equal_char),
                                       mapping_impl_list)
                                     r t (suc zero_nata)))) &&
                            not (is_none
                                  (find (membera
  (equal_term _A1 (equal_list equal_char))
  (reachable_terms _A1
    ((ccompare_list ccompare_char), (equal_list equal_char), mapping_impl_list)
    r t n))
                                    (reachable_terms _A1
                                      ((ccompare_list ccompare_char),
(equal_list equal_char), mapping_impl_list)
                                      r s (suc zero_nata)))))
                      (comp (comp (comp (comp
  (comp (comp (showsl_lit "the critical pair ")
          (showsl_terma _A2 (showl_list showl_char) s))
    (showsl_lit " <- . -> "))
  (showsl_terma _A2 (showl_list showl_char) t))
                                    (showsl_lit
                                      " is not strongly closed within "))
                              (showsl_nat n))
                        (showsl_lit " steps.")))
                  (critical_pairs_impl _A1 r r))
                (fun x -> Inl (snd x))))
          (fun x ->
            Inl (comp (comp x
                        (showsl_lit
                          "\092hence the following TRS is not strongly closed\092"))
                  (showsl_trs _A2 (showl_list showl_char) r)));;

let rec root_rewrite _A (_B1, _B2, _B3)
  r s = maps (fun (l, ra) ->
               (match matcha _A (_B1, _B2, _B3) s l with None -> []
                 | Some sigma -> [subst_apply_term ra sigma]))
          r;;

let rec parallel_rewrite _A (_B1, _B2, _B3)
  r x1 = match r, x1 with r, Var x -> [Var x]
    | r, Fun (f, ss) ->
        remdups (equal_term _A _B2)
          (root_rewrite _A (_B1, _B2, _B3) r (Fun (f, ss)) @
            map (fun a -> Fun (f, a))
              (product_lists
                (map (parallel_rewrite _A (_B1, _B2, _B3) r) ss)));;

let rec is_parallel_closed_cp _A (_B1, _B2, _B3)
  r uu x2 = match r, uu, x2 with
    r, uu, (false, (s, t)) -> is_par_rstep _A (_B1, _B2, _B3) r t s
    | r, Some n, (true, (s, t)) ->
        not (is_none
              (find (is_par_rstep _A (_B1, _B2, _B3) r t)
                (reachable_terms _A (_B1, _B2, _B3) r s n)))
    | r, None, (true, (s, t)) ->
        not (is_none
              (find (is_par_rstep _A (_B1, _B2, _B3) r t)
                (parallel_rewrite _A (_B1, _B2, _B3) r s)));;

let rec showsl_option _A
  = function None -> showsl_lit "None"
    | Some x ->
        comp (comp (showsl_lit "Some (") (showsl _A x)) (showsl_lit ")");;

let rec check_parallel_closed (_A1, _A2)
  r n = catch_error
          (bindb
            (check_left_linear_trs _A2
              ((ceq_list ceq_char), (ccompare_list ccompare_char),
                set_impl_list, (showl_list showl_char))
              r)
            (fun _ ->
              catch_error
                (forallM
                  (fun (b, (s, t)) ->
                    check (is_parallel_closed_cp _A1
                            ((ccompare_list ccompare_char),
                              (equal_list equal_char), mapping_impl_list)
                            r n (b, (s, t)))
                      (comp (comp (comp (comp
  (comp (comp (showsl_lit "the critical pair ")
          (showsl_terma _A2 (showl_list showl_char) s))
    (showsl_lit " <- . -> "))
  (showsl_terma _A2 (showl_list showl_char) t))
                                    (showsl_lit
                                      " is not (almost) parallel closed within "))
                              (showsl_option showl_nat n))
                        (showsl_lit " steps.")))
                  (critical_pairs_impl _A1 r r))
                (fun x -> Inl (snd x))))
          (fun x ->
            Inl (comp (comp x
                        (showsl_lit
                          "\092hence the following TRS is not (almost) parallel closed\092"))
                  (showsl_trs _A2 (showl_list showl_char) r)));;

let rec applicable_rule_impl isNF = (fun (l, _) -> list_all isNF (args l));;

let rec check_non_applicable_rules
  isNF r =
    catch_error
      (forallM
        (fun x -> (if not (applicable_rule_impl isNF x) then Inr () else Inl x))
        r)
      (fun x -> Inl (snd x));;

let rec rule_match_impl_aux _A
  nfq s fts l =
    (match
      mgu _A (equal_list equal_char) fts
        (map_term (fun x -> x)
          (fun a ->
            Chara (true, false, false, true, true, true, true, false) :: a)
          l)
      with None -> false
      | Some mu ->
        list_all
          (fun u ->
            nfq (subst_apply_term
                  (map_term (fun x -> x)
                    (fun a ->
                      Chara (true, false, false, true, true, true, true,
                              false) ::
                        a)
                    u)
                  mu))
          (args l) &&
          list_all (fun u -> nfq (subst_apply_term u mu)) s);;

let rec is_ur_closed_term_af_impl (_A1, _A2) (_C1, _C2)
  nfq e_cap pi r u s x6 = match nfq, e_cap, pi, r, u, s, x6 with
    nfq, e_cap, pi, r, u, s, Fun (f, ts) ->
      (let n = size_list ts in
       let pi_f = pi (f, n) in
        list_all
          (fun (i, t) ->
            (if member (ceq_nat, ccompare_nat) i pi_f
              then is_ur_closed_term_af_impl (_A1, _A2) (_C1, _C2) nfq e_cap pi
                     r u s t
              else true))
          (zip (upt zero_nata n) ts) &&
          (let fts =
             class_to_term
               (Chara (false, true, false, true, true, true, true, false))
               (Fun (f, map e_cap ts))
             in
            list_all
              (fun (l, ra) ->
                member
                  ((ceq_prod (ceq_term _A2 (equal_list equal_char)) _C1),
                    (ccompare_prod
                      (ccompare_term _A1 (compare_list compare_char)) _C2))
                  (l, ra) u ||
                  not (rule_match_impl_aux _A2 nfq s fts l))
              r))
    | nfq, e_cap, pi, r, u, s, Var x -> true;;

let rec is_ur_closed_af_impl_dpp_mv (_B1, _B2, _B3)
  i d pi u =
    (let ic = icap_impl_dpp _B2 i d in
     let qnf = is_QNFc i d in
     let r = rulese i d in
     let urc =
       (fun s ->
         is_ur_closed_term_af_impl (_B1.compare_compare_order, _B2)
           ((ceq_term _B2 (equal_list equal_char)),
             (ccompare_term _B1.compare_compare_order
               (compare_list compare_char)))
           qnf (ic s) pi r
           (set ((ceq_prod (ceq_term _B2 (equal_list equal_char))
                   (ceq_term _B2 (equal_list equal_char))),
                  (ccompare_prod
                    (ccompare_term _B1.compare_compare_order
                      (compare_list compare_char))
                    (ccompare_term _B1.compare_compare_order
                      (compare_list compare_char))),
                  (set_impl_prod set_impl_term set_impl_term))
             u))
       in
      (fun s ->
        (let sa =
           map (map_term (fun x -> x)
                 (fun a ->
                   Chara (false, false, false, true, true, true, true, false) ::
                     a))
             s
           in
          (fun t ->
            urc sa sa
              (map_term (fun x -> x)
                (fun a ->
                  Chara (false, false, false, true, true, true, true, false) ::
                    a)
                t)))));;

let rec wwf_rulesa
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = wwf_rules;;

let rec rules_mapc
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = rules_map;;

let rec q_emptyc
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = q_empty;;

let rec matchCapRMBelow _A _B
  rm l (Fun (f, ts)) = matchc _A _B (GCFun (f, map (tcapRM2 _A _B rm) ts)) l;;

let rec forallM_index_aux
  p i x2 = match p, i, x2 with p, i, [] -> Inr ()
    | p, i, x :: xs ->
        bindb (catch_error (p x i) (fun xa -> Inl ((x, i), xa)))
          (fun _ -> forallM_index_aux p (suc i) xs);;

let rec forallM_index p xs = forallM_index_aux p zero_nata xs;;

let rec check_ur_closed_term_rm_af (_A1, _A2) (_B1, _B2)
  uu uv uw x3 = match uu, uv, uw, x3 with uu, uv, uw, Var x -> Inr ()
    | rm, ur, pi, Fun (f, ts) ->
        (let n = size_list ts in
         let pia = pi (f, n) in
          bindb (catch_error
                  (forallM_index
                    (fun t i ->
                      (if member (ceq_nat, ccompare_nat) i pia
                        then check_ur_closed_term_rm_af (_A1, _A2) (_B1, _B2) rm
                               ur pi t
                        else Inr ()))
                    ts)
                  (fun x -> Inl (snd x)))
            (fun _ ->
              catch_error
                (forallM
                  (fun lr ->
                    check (membera
                             (equal_prod (equal_term _A1 _B1)
                               (equal_term _A1 _B1))
                             ur lr ||
                            not (matchCapRMBelow _A1 _B1 rm (fst lr)
                                  (Fun (f, ts))))
                      (comp (comp (comp (comp (showsl_lit "due to the subterm ")
  (showsl_terma _A2 _B2 (Fun (f, ts))))
                                    (showsl_lit " of some usable rhs, rule "))
                              (showsl_rule _A2 _B2 lr))
                        (showsl_lit " should be usable.")))
                  (rm (f, n)))
                (fun x -> Inl (snd x))));;

let rec check_ur_P_closed_rm_af (_A1, _A2) (_B1, _B2)
  rm ur pi p =
    bindb (catch_error
            (catch_error
              (forallM
                (fun lr ->
                  check_ur_closed_term_rm_af (_A1, _A2) (_B1, _B2) rm ur pi
                    (snd lr))
                ur)
              (fun x -> Inl (snd x)))
            (fun x ->
              Inl (comp (showsl_lit
                          "error when checking closure properties of rhs of usable rules\092")
                    x)))
      (fun _ ->
        catch_error
          (catch_error
            (forallM
              (fun st ->
                check_ur_closed_term_rm_af (_A1, _A2) (_B1, _B2) rm ur pi
                  (snd st))
              p)
            (fun x -> Inl (snd x)))
          (fun x ->
            Inl (comp (showsl_lit
                        "error when checking closure properties of rhs of DPs\092")
                  x)));;

let rec smart_usable_rules_checker_impl (_B1, _B2, _B3)
  i d pi u_opt sts =
    (let nfs = nfsc i d in
     let m = minimal i d in
     let wwf = wwf_rulesa i d in
     let qempty = q_emptyc i d in
      (match u_opt with None -> Inr (rulese i d)
        | Some u ->
          (if nFQ_subset_NF_rulesc i d &&
                ((nfs ||
                   isOK (check_varcond_subset _B3
                          ((equal_list equal_char), (showl_list showl_char))
                          sts)) &&
                  (nfs || (m || wwf)))
            then (let urc = is_ur_closed_af_impl_dpp_mv (_B1, _B2, _B3) i d pi u
                    in
                  let check_urc =
                    (fun s t ->
                      check (urc s t)
                        (comp (comp (showsl_lit "term ")
                                (showsl_terma _B3 (showl_list showl_char) t))
                          (showsl_lit " is not closed under usable rules")))
                    in
                   bindb (catch_error
                           (forallM (fun (s, a) -> check_urc [s] a) sts)
                           (fun x -> Inl (snd x)))
                     (fun _ ->
                       bindb (catch_error
                               (forallM (fun (l, a) -> check_urc (args l) a) u)
                               (fun x -> Inl (snd x)))
                         (fun _ -> Inr u)))
            else bindb (check (m && (if nfs then qempty || wwf else true))
                         (showsl_lit "minimality and well formedness required"))
                   (fun _ ->
                     bindb (catch_error
                             (forallM
                               (fun (l, _) ->
                                 check (not (is_Var l))
                                   (showsl_lit "variables as lhss not allowed"))
                               (rulese i d))
                             (fun x -> Inl (snd x)))
                       (fun _ ->
                         (let rm = rules_mapc i d in
                           bindb (check_ur_P_closed_rm_af (_B2, _B3)
                                   ((equal_list equal_char),
                                     (showl_list showl_char))
                                   rm u pi sts)
                             (fun _ -> Inr u)))))));;

let rec root_aft_to_entry
  s t pi =
    (let rt = the (root t) in
     let pi_t = pi rt in
     let ts = args t in
      map_filter
        (fun x ->
          (if member (ceq_nat, ccompare_nat) x pi_t then Some (s, nth ts x)
            else None))
        (upt zero_nata (snd rt)));;

let rec check_no_defined_root _A _B
  isdef t =
    check (not (isdef (the (root t))))
      (comp (comp (showsl_literal "the root of ") (showsl_terma _A _B t))
        (showsl_literal " is defined"));;

let rec pairsb
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = pairs;;

let rec validb
  (Root_redtriple_ext (valid, s, ns, nst, af, aft, desc, more)) = valid;;

let rec descb
  (Root_redtriple_ext (valid, s, ns, nst, af, aft, desc, more)) = desc;;

let rec nsta
  (Root_redtriple_ext (valid, s, ns, nst, af, aft, desc, more)) = nst;;

let rec aft
  (Root_redtriple_ext (valid, s, ns, nst, af, aft, desc, more)) = aft;;

let rec nsa (Root_redtriple_ext (valid, s, ns, nst, af, aft, desc, more)) = ns;;

let rec afa (Root_redtriple_ext (valid, s, ns, nst, af, aft, desc, more)) = af;;

let rec sa (Root_redtriple_ext (valid, s, ns, nst, af, aft, desc, more)) = s;;

let rec generic_ur_af_root_redtriple_proc (_B1, _B2, _B3)
  i rp u_opt premove dpp =
    (match
      catch_error
        (bindb (validb rp)
          (fun _ ->
            (let (ps, pns) = split_pairsa i dpp premove in
             let p = pairsb i dpp in
             let pi = afa rp in
             let pia = aft rp in
             let is_def = (fun fn -> not (null (rules_mapc i dpp fn))) in
              bindb (catch_error
                      (forallM
                        (fun (l, r) ->
                          bindb (check_no_var _B3 (showl_list showl_char) l)
                            (fun _ ->
                              bindb (check_no_var _B3 (showl_list showl_char) r)
                                (fun _ ->
                                  check_no_defined_root _B3
                                    (showl_list showl_char) is_def r)))
                        p)
                      (fun x -> Inl (snd x)))
                (fun _ ->
                  bindb (catch_error
                          (forallM
                            (fun (l, _) ->
                              check_no_var _B3 (showl_list showl_char) l)
                            (rulese i dpp))
                          (fun x -> Inl (snd x)))
                    (fun _ ->
                      bindb (smart_usable_rules_checker_impl (_B1, _B2, _B3) i
                              dpp pi u_opt
                              (maps (fun (s, t) -> root_aft_to_entry s t pia)
                                p))
                        (fun u ->
                          bindb (catch_error
                                  (catch_error (forallM (nsa rp) u)
                                    (fun x -> Inl (snd x)))
                                  (fun x ->
                                    Inl (comp
  (showsl_lit "problem when orienting (usable) rules\092") x)))
                            (fun _ ->
                              bindb (catch_error
                                      (catch_error (forallM (nsta rp) pns)
(fun x -> Inl (snd x)))
                                      (fun x ->
Inl (comp (showsl_lit "problem when orienting DPs\092") x)))
                                (fun _ ->
                                  catch_error
                                    (catch_error (forallM (sa rp) ps)
                                      (fun x -> Inl (snd x)))
                                    (fun x ->
                                      Inl
(comp (showsl_lit "problem when orienting DPs\092") x))))))))))
        (fun x ->
          Inl (comp (comp (comp (showsl_lit
                                  "could not apply the generic root reduction pair processor with the following\092")
                            (descb rp))
                      (showsl_literal "\092"))
                x))
      with Inl a -> Inl a | Inr _ -> Inr (delete_P_Pwa i dpp premove premove));;

let rec nst
  (Redtriple_ext
    (valid, s, ns, nst, af, mono_af, mono, desc, not_ws_ns, cpx, more))
    = nst;;

let rec generic_ur_af_redtriple_proc (_B1, _B2, _B3)
  i rp u_opt premove dpp =
    (match
      catch_error
        (bindb (valid rp)
          (fun _ ->
            (let (ps, pns) = split_pairsa i dpp premove in
             let p = pairsb i dpp in
              bindb (smart_usable_rules_checker_impl (_B1, _B2, _B3) i dpp
                      (af rp) u_opt p)
                (fun u ->
                  bindb (catch_error
                          (catch_error (forallM (ns rp) u)
                            (fun x -> Inl (snd x)))
                          (fun x ->
                            Inl (comp (showsl_lit
"problem when orienting (usable) rules\092")
                                  x)))
                    (fun _ ->
                      bindb (catch_error
                              (catch_error (forallM (nst rp) pns)
                                (fun x -> Inl (snd x)))
                              (fun x ->
                                Inl (comp (showsl_lit
    "problem when orienting DPs\092")
                                      x)))
                        (fun _ ->
                          catch_error
                            (catch_error (forallM (s rp) ps)
                              (fun x -> Inl (snd x)))
                            (fun x ->
                              Inl (comp (showsl_lit
  "problem when orienting DPs\092")
                                    x))))))))
        (fun x ->
          Inl (comp (comp (comp (showsl_lit
                                  "could not apply the generic reduction pair processor with the following\092")
                            (desc rp))
                      (showsl_literal "\092"))
                x))
      with Inl a -> Inl a | Inr _ -> Inr (delete_P_Pwa i dpp premove premove));;

let rec is_ur_closed_term_impl (_A1, _A2) (_C1, _C2)
  nfq e_cap r u s x5 = match nfq, e_cap, r, u, s, x5 with
    nfq, e_cap, r, u, s, Fun (f, ts) ->
      list_all (is_ur_closed_term_impl (_A1, _A2) (_C1, _C2) nfq e_cap r u s)
        ts &&
        (let fts =
           class_to_term
             (Chara (false, true, false, true, true, true, true, false))
             (Fun (f, map e_cap ts))
           in
          list_all
            (fun (l, ra) ->
              member
                ((ceq_prod (ceq_term _A2 (equal_list equal_char)) _C1),
                  (ccompare_prod (ccompare_term _A1 (compare_list compare_char))
                    _C2))
                (l, ra) u ||
                not (rule_match_impl_aux _A2 nfq s fts l))
            r)
    | nfq, e_cap, r, u, s, Var x -> true;;

let rec is_ur_closed_impl_dpp_mv (_B1, _B2, _B3)
  i d u =
    (let ic = icap_impl_dpp _B2 i d in
     let qnf = is_QNFc i d in
     let r = rulese i d in
     let urc =
       (fun s ->
         is_ur_closed_term_impl (_B1.compare_compare_order, _B2)
           ((ceq_term _B2 (equal_list equal_char)),
             (ccompare_term _B1.compare_compare_order
               (compare_list compare_char)))
           qnf (ic s) r
           (set ((ceq_prod (ceq_term _B2 (equal_list equal_char))
                   (ceq_term _B2 (equal_list equal_char))),
                  (ccompare_prod
                    (ccompare_term _B1.compare_compare_order
                      (compare_list compare_char))
                    (ccompare_term _B1.compare_compare_order
                      (compare_list compare_char))),
                  (set_impl_prod set_impl_term set_impl_term))
             u))
       in
      (fun s ->
        (let sa =
           map (map_term (fun x -> x)
                 (fun a ->
                   Chara (false, false, false, true, true, true, true, false) ::
                     a))
             s
           in
          (fun t ->
            urc sa sa
              (map_term (fun x -> x)
                (fun a ->
                  Chara (false, false, false, true, true, true, true, false) ::
                    a)
                t)))));;

let rec intersect_rulesb
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = intersect_rules;;

let rec usable_rules_proc (_B1, _B2, _B3)
  i u dpp =
    (match
      bindb (check (nFQ_subset_NF_rulesc i dpp)
              (showsl_lit "innermost rewriting required"))
        (fun _ ->
          bindb (check (nfsc i dpp || (minimal i dpp || wwf_rulesa i dpp))
                  (showsl_lit
                    "normal form subst, minimality or well-formedness required"))
            (fun _ ->
              (let p = pairsb i dpp in
               let urc = is_ur_closed_impl_dpp_mv (_B1, _B2, _B3) i dpp u in
               let check_urc =
                 (fun s t ->
                   check (urc s t)
                     (comp (comp (showsl_lit "term ")
                             (showsl_terma _B3 (showl_list showl_char) t))
                       (showsl_lit " is not closed under usable rules")))
                 in
               let nfs = nfsc i dpp in
                bindb (catch_error
                        (forallM
                          (fun (l, r) ->
                            bindb (if nfs then Inr ()
                                    else catch_error
   (check_subseteq (equal_list equal_char) (vars_term_list r)
     (vars_term_list l))
   (fun _ -> Inl (showsl_lit "variable condition in P violated")))
                              (fun _ -> check_urc [l] r))
                          p)
                        (fun x -> Inl (snd x)))
                  (fun _ ->
                    catch_error (forallM (fun (l, a) -> check_urc (args l) a) u)
                      (fun x -> Inl (snd x))))))
      with Inl a -> Inl a | Inr _ -> Inr (intersect_rulesb i dpp u));;

let rec mono
  (Redtriple_ext
    (valid, s, ns, nst, af, mono_af, mono, desc, not_ws_ns, cpx, more))
    = mono;;

let rec mono_inn_usable_rules_ce_proc (_B1, _B2, _B3, _B4, _B5, _B6)
  i rp premove rremove ur dpp =
    (match
      catch_error
        (bindb (usable_rules_proc (_B3, _B4, _B6) i ur dpp)
          (fun _ ->
            (let p = pairsb i dpp in
             let us =
               foldr (comp (sup_seta
                             ((ceq_prod _B1 ceq_nat),
                               (ccompare_prod _B2 ccompare_nat)))
                       (comp (funas_term (_B1, _B2, _B5)) snd))
                 (p @ ur)
                 (bot_set
                   ((ceq_prod _B1 ceq_nat), (ccompare_prod _B2 ccompare_nat),
                     (set_impl_prod _B5 set_impl_nat)))
               in
             let filt =
               (fun lr ->
                 ball ((ceq_prod _B1 ceq_nat), (ccompare_prod _B2 ccompare_nat))
                   (funas_term (_B1, _B2, _B5) (fst lr))
                   (fun f ->
                     member
                       ((ceq_prod _B1 ceq_nat),
                         (ccompare_prod _B2 ccompare_nat))
                       f us))
               in
             let (pms, pns) = split_pairsa i dpp premove in
             let (ps, pnwf) = partition filt pms in
             let (urms, urns) =
               partition
                 (membera
                   (equal_prod (equal_term _B4 (equal_list equal_char))
                     (equal_term _B4 (equal_list equal_char)))
                   rremove)
                 ur
               in
             let (urs, urnwf) = partition filt urms in
              bindb (valid rp)
                (fun _ ->
                  bindb (mono rp (ps @ urs @ urns @ urnwf @ pns @ pnwf))
                    (fun _ ->
                      bindb (catch_error
                              (catch_error (forallM (ns rp) (urns @ urnwf))
                                (fun x -> Inl (snd x)))
                              (fun x ->
                                Inl (comp (showsl_lit
    "problem when orienting usable rules\092")
                                      x)))
                        (fun _ ->
                          bindb (catch_error
                                  (catch_error (forallM (s rp) urs)
                                    (fun x -> Inl (snd x)))
                                  (fun x ->
                                    Inl (comp
  (showsl_lit "problem when orienting usable rules\092") x)))
                            (fun _ ->
                              bindb (catch_error
                                      (catch_error
(forallM (ns rp) (pns @ pnwf)) (fun x -> Inl (snd x)))
                                      (fun x ->
Inl (comp (showsl_lit "problem when orienting DPs\092") x)))
                                (fun _ ->
                                  catch_error
                                    (catch_error (forallM (s rp) ps)
                                      (fun x -> Inl (snd x)))
                                    (fun x ->
                                      Inl
(comp (showsl_lit "problem when orienting DPs\092") x))))))))))
        (fun x ->
          Inl (comp (comp (comp (showsl_lit
                                  "could not apply the innermost usable rules reduction pair processor with the following\092")
                            (desc rp))
                      (showsl_literal "\092"))
                x))
      with Inl a -> Inl a
      | Inr _ ->
        Inr (delete_R_Rwc i (delete_P_Pwa i dpp premove premove) rremove
              rremove));;

let rec mono_ur_redpair_proc (_B1, _B2, _B3, _B4, _B5) (_C1, _C2)
  i rp premove rremove ur dpp =
    (match
      catch_error
        (bindb (check (minimal i dpp) (showsl_lit "minimality required"))
          (fun _ ->
            bindb (check
                    (if nfsc i dpp
                      then (if not (q_emptyc i dpp) then wwf_rulesa i dpp
                             else true)
                      else true)
                    (showsl_lit "well formedness required"))
              (fun _ ->
                (let p = pairsb i dpp in
                 let us =
                   foldr (comp (sup_seta
                                 ((ceq_prod _B1 ceq_nat),
                                   (ccompare_prod _B2 ccompare_nat)))
                           (comp (funas_term (_B1, _B2, _B4)) snd))
                     (p @ ur)
                     (bot_set
                       ((ceq_prod _B1 ceq_nat),
                         (ccompare_prod _B2 ccompare_nat),
                         (set_impl_prod _B4 set_impl_nat)))
                   in
                 let filt =
                   (fun lr ->
                     ball ((ceq_prod _B1 ceq_nat),
                            (ccompare_prod _B2 ccompare_nat))
                       (funas_term (_B1, _B2, _B4) (fst lr))
                       (fun f ->
                         member
                           ((ceq_prod _B1 ceq_nat),
                             (ccompare_prod _B2 ccompare_nat))
                           f us))
                   in
                 let (pms, pns) = split_pairsa i dpp premove in
                 let (ps, pnwf) = partition filt pms in
                 let (urms, urns) =
                   partition
                     (membera
                       (equal_prod (equal_term _B3 _C1) (equal_term _B3 _C1))
                       rremove)
                     ur
                   in
                 let (urs, urnwf) = partition filt urms in
                 let rm = rules_mapc i dpp in
                  bindb (valid rp)
                    (fun _ ->
                      bindb (mono rp (ps @ urs @ urns @ urnwf @ pns @ pnwf))
                        (fun _ ->
                          bindb (check_ur_P_closed_rm_af (_B3, _B5) (_C1, _C2)
                                  rm ur full_af p)
                            (fun _ ->
                              bindb (catch_error
                                      (forallM
(fun (l, _) ->
  check (not (is_Var l)) (showsl_lit "variables as lhss not allowed"))
(rulese i dpp))
                                      (fun x -> Inl (snd x)))
                                (fun _ ->
                                  bindb (catch_error
  (catch_error (forallM (ns rp) (urns @ urnwf)) (fun x -> Inl (snd x)))
  (fun x ->
    Inl (comp (showsl_lit "problem when orienting usable rules\092") x)))
                                    (fun _ ->
                                      bindb
(catch_error (catch_error (forallM (s rp) urs) (fun x -> Inl (snd x)))
  (fun x ->
    Inl (comp (showsl_lit "problem when orienting usable rules\092") x)))
(fun _ ->
  bindb (catch_error
          (catch_error (forallM (ns rp) (pns @ pnwf)) (fun x -> Inl (snd x)))
          (fun x -> Inl (comp (showsl_lit "problem when orienting DPs\092") x)))
    (fun _ ->
      catch_error (catch_error (forallM (s rp) ps) (fun x -> Inl (snd x)))
        (fun x ->
          Inl (comp (showsl_lit "problem when orienting DPs\092") x)))))))))))))
        (fun x ->
          Inl (comp (comp (comp (showsl_lit
                                  "could not apply the monotonic reduction pair processor with the following\092")
                            (desc rp))
                      (showsl_literal "\092"))
                x))
      with Inl a -> Inl a
      | Inr _ ->
        Inr (delete_R_Rwc i (delete_P_Pwa i dpp premove premove) rremove
              rremove));;

let rec generic_mono_ur_redpair_proc (_B1, _B2, _B3, _B4, _B5, _B6)
  i rp premove rremove ur dpp =
    (if nFQ_subset_NF_rulesc i dpp
      then mono_inn_usable_rules_ce_proc (_B1, _B2, _B3, _B4, _B5, _B6) i rp
             premove rremove ur dpp
      else bindb (check (minimal i dpp)
                   (showsl_lit
                     "minimality or innermost required for mon. red. pair proc. with usable rules"))
             (fun _ ->
               mono_ur_redpair_proc (_B1, _B2, _B4, _B5, _B6)
                 ((equal_list equal_char), (showl_list showl_char)) i rp premove
                 rremove ur dpp));;

let rec compat_roota _A
  uu x1 = match uu, x1 with uu, None -> false
    | Var uv, Some v -> false
    | Fun (va, vb), Some v ->
        equal_optiona (equal_prod _A equal_nat) (root (Fun (va, vb))) (Some v);;

let rec usable_rules_gen2 _A
  r (fn, b) =
    (let rr =
       maps (fun (l, ra) -> (if compat_roota _A l fn then [(l, ra)] else [])) r
       in
      (if b then rr else map (fun (l, ra) -> (ra, l)) rr));;

let rec rel_dep_prod uu x1 = match uu, x1 with uu, Ignore -> []
                       | b, Increase -> [b]
                       | b, Decrease -> [not b]
                       | uv, Wild -> [true; false];;

let rec compat_root _A
  uu x1 = match uu, x1 with uu, Var uv -> false
    | Var uw, Fun (v, va) -> false
    | Fun (vb, vc), Fun (v, va) ->
        equal_optiona (equal_prod _A equal_nat) (root (Fun (vb, vc)))
          (root (Fun (v, va)));;

let rec usable_rules_gen1 _A
  pi r x2 = match pi, r, x2 with pi, r, (Var uu, uv) -> []
    | pi, r, (Fun (f, ts), b) ->
        (let n = size_list ts in
          maps (fun (l, ra) ->
                 (if compat_root _A l (Fun (f, ts)) then [(ra, b)] else []))
            r @
            maps (fun i ->
                   map (fun a -> (nth ts i, a)) (rel_dep_prod b (pi (f, n) i)))
              (upt zero_nata n));;

let rec usable_rules_gen _A _B
  pi r p =
    maps (usable_rules_gen2 _A r)
      (remdups (equal_prod (equal_option (equal_prod _A equal_nat)) equal_bool)
        (map (fun (t, a) -> (root t, a))
          (mk_rtrancl_list (equal_proda (equal_term _A _B) equal_bool)
            (usable_rules_gen1 _A pi r) (map (fun (_, t) -> (t, true)) p))));;

let rec generate_lists
  n xs = concat_lists (map (fun _ -> xs) (upt zero_nata n));;

let rec initial_conditions_gen_impl
  pa bef_len aft_len p st =
    (let pairs = (fun n -> generate_lists n p) in
     let a =
       maps (fun bef -> map (fun aft -> bef @ st :: aft) (pairs aft_len))
         (pairs bef_len)
       in
      filtera
        (fun bef_st_aft ->
          all_interval_nat
            (fun i -> pa (nth bef_st_aft i) (nth bef_st_aft (suc i))) zero_nata
            (plus_nata bef_len aft_len))
        a);;

let rec is_partition_impl (_A1, _A2, _A3, _A4)
  = function [] -> Some (bot_set (_A2, _A3.ccompare_cproper_interval, _A4))
    | asa :: rest ->
        bind (is_partition_impl (_A1, _A2, _A3, _A4) rest)
          (fun all ->
            (if is_empty (_A1, _A2, _A3)
                  (inf_seta (_A2, _A3.ccompare_cproper_interval) asa all)
              then Some (sup_seta (_A2, _A3.ccompare_cproper_interval) all asa)
              else None));;

let rec is_partition (_A1, _A2, _A3, _A4)
  asa = not (is_none (is_partition_impl (_A1, _A2, _A3, _A4) asa));;

let rec disjoint_variant (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  sts uvs =
    equal_nata (size_list sts) (size_list uvs) &&
      (all_interval_nat
         (fun i ->
           eq_rule_mod_vars _A2 (_B4.ccompare_cproper_interval, _B6, _B7)
             (nth sts i) (nth uvs i))
         zero_nata (size_list sts) &&
        is_partition (_B1, _B3, _B4, _B8)
          (map (vars_rule (_A1, _A2)
                 (_B1.finite_UNIV_card_UNIV, _B2, _B3, _B4, _B5, _B6, _B8))
            uvs));;

let rec condition_of
  c x1 st = match c, x1, st with
    c, Bound, (s, uu) -> CC_cond (false, (s, Fun (c, [])))
    | c, Strict, st -> CC_cond (true, st)
    | c, Non_Strict, st -> CC_cond (false, st);;

let rec constraint_of
  c ctype uvs bef =
    CC_impl
      (map (fun i -> CC_rewr (snd (nth uvs i), fst (nth uvs (suc i))))
         (upt zero_nata (minus_nata (size_list uvs) one_nata)),
        condition_of c ctype (nth uvs bef));;

let rec the_set_impl_main_lazy (_A1, _A2, _A3) _B
  q gen_as have_asa have_as bs =
    (let new_as =
       maps (fun b ->
              maps (fun a ->
                     (if not (member (_A1, _A2) a have_as) then [a] else []))
                (gen_as b))
         bs
       in
      (if null new_as then have_asa
        else the_set_impl_main_lazy (_A1, _A2, _A3) _B q gen_as
               (new_as @ have_asa)
               (sup_seta (_A1, _A2) (set (_A1, _A2, _A3) new_as) have_as)
               (remdups _B (maps q new_as))));;

let rec the_set_impl_lazy (_A1, _A2, _A3) _B
  q gen_as bs =
    the_set_impl_main_lazy (_A1, _A2, _A3) _B q gen_as []
      (bot_set (_A1, _A2, _A3)) bs;;

let rec inductive_set_impl_lazy _A (_B1, _B2, _B3)
  q p = the_set_impl_lazy (_B1, _B2, _B3) _A p q;;

let rec rule_match_impl _A
  nfq e_cap s f ts l =
    (match mgu_class _A (Fun (f, map e_cap ts)) l with None -> false
      | Some mu ->
        list_all
          (fun u ->
            nfq (subst_apply_term
                  (map_term (fun x -> x)
                    (fun a ->
                      Chara (true, false, false, true, true, true, true,
                              false) ::
                        a)
                    u)
                  mu))
          (args l) &&
          list_all (fun u -> nfq (subst_apply_term u mu)) s);;

let rec ur_term_impl _A _C
  nfq e_cap r pi s x5 = match nfq, e_cap, r, pi, s, x5 with
    nfq, e_cap, r, pi, s, Var x -> []
    | nfq, e_cap, r, pi, s, Fun (f, ts) ->
        (let n = size_list ts in
         let reca = map (ur_term_impl _A _C nfq e_cap r pi s) ts in
          remdups (equal_prod (equal_term _A (equal_list equal_char)) _C)
            (maps (fun (i, urs) ->
                    (if member (ceq_nat, ccompare_nat) i (pi (f, n)) then urs
                      else []))
               (zip (upt zero_nata n) reca) @
              filtera (fun (l, _) -> rule_match_impl _A nfq e_cap s f ts l)
                r));;

let rec precompute_fun _A
  f asa = fun_of_map_fun (ceta_map_of _A (map (fun a -> (a, f a)) asa)) f;;

let rec usable_rules_calc_impl (_A1, _A2)
  nfq e_cap r =
    (let urt =
       (fun (s, t) ->
         (let sa =
            map (map_term (fun x -> x)
                  (fun a ->
                    Chara (false, false, false, true, true, true, true,
                            false) ::
                      a))
              s
            in
           ur_term_impl _A2 (equal_term _A2 (equal_list equal_char)) nfq
             (e_cap sa) r full_af sa
             (map_term (fun x -> x)
               (fun a ->
                 Chara (false, false, false, true, true, true, true, false) ::
                   a)
               t)))
       in
     let urules = map (fun (l, a) -> (args l, a)) r in
     let ufun =
       precompute_fun
         (compare_order_prod
           (compare_order_list
             ((compare_order_term _A1.compare_compare_order
                (compare_list compare_char)),
               (equal_term _A2 (equal_list equal_char))))
           (compare_order_term _A1.compare_compare_order
             (compare_list compare_char)))
         urt urules
       in
      inductive_set_impl_lazy
        (equal_prod (equal_list (equal_term _A2 (equal_list equal_char)))
          (equal_term _A2 (equal_list equal_char)))
        ((ceq_prod (ceq_term _A2 (equal_list equal_char))
           (ceq_term _A2 (equal_list equal_char))),
          (ccompare_prod
            (ccompare_term _A1.compare_compare_order
              (compare_list compare_char))
            (ccompare_term _A1.compare_compare_order
              (compare_list compare_char))),
          (set_impl_prod set_impl_term set_impl_term))
        ufun (fun (l, ra) -> [(args l, ra)]));;

let rec ur_calc_singleton (_A1, _A2)
  nfq e_cap r st = usable_rules_calc_impl (_A1, _A2) nfq e_cap r [st];;

let rec inn_usable_rules_pair (_B1, _B2)
  i d = (let inn = nFQ_subset_NF_rulesc i d in
         let r = rulese i d in
         let qnf = is_QNFc i d in
         let ic = icap_impl_dpp _B2 i d in
         let calc = ur_calc_singleton (_B1, _B2) qnf ic r in
         let nfs = nfsc i d in
         let wwf = wwf_rulesa i d in
         let m = minimal i d in
          (fun (s, t) ->
            (if inn &&
                  ((nfs ||
                     subset
                       (card_UNIV_list, cenum_list, (ceq_list ceq_char),
                         (ccompare_list ccompare_char))
                       (vars_term (_B1.compare_compare_order, _B2)
                         (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
                           (cproper_interval_list ccompare_char),
                           (compare_list compare_char), (equal_list equal_char),
                           set_impl_list)
                         t)
                       (vars_term (_B1.compare_compare_order, _B2)
                         (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
                           (cproper_interval_list ccompare_char),
                           (compare_list compare_char), (equal_list equal_char),
                           set_impl_list)
                         s)) &&
                    (nfs || (m || wwf)))
              then calc ([s], t) else r)));;

let rec icap_impl_dpp_mv _B
  i d = (let qr = nFQ_subset_NF_rulesc i d in
         let qnf = is_QNFc i d in
         let r = rulese i d in
         let ic = icap_impl_gen _B qr qnf (map fst r) in
          (fun s ->
            (let sa =
               map (map_term (fun x -> x)
                     (fun a ->
                       Chara (false, false, false, true, true, true, true,
                               false) ::
                         a))
                 s
               in
             let sx =
               ceta_set_of (compare_order_list (compare_order_char, equal_char))
                 (maps vars_term_list sa)
               in
              (fun t ->
                ic sa sx
                  (map_term (fun x -> x)
                    (fun a ->
                      Chara (false, false, false, true, true, true, true,
                              false) ::
                        a)
                    t)))));;

let rec is_iedg_edge_dpp (_B1, _B2, _B3)
  i d = (let qnf = is_QNFc i d in
         let ic = icap_impl_dpp_mv _B2 i d in
         let _ = rulese i d in
         let urules = inn_usable_rules_pair (_B1, _B2) i d in
          (fun (s, t) ->
            (let cst = ic [s] t in
             let urls = reverse_rules (urules (s, t)) in
             let ica =
               icap_impl _B2
                 (is_NF_terms (_B1, _B2)
                   ((ccompare_list ccompare_char), (equal_list equal_char),
                     mapping_impl_list)
                   [])
                 urls
               in
              (fun u ->
                (match mgu_class _B2 cst u with None -> false
                  | Some mu ->
                    qnf (subst_apply_term
                          (map_term (fun x -> x)
                            (fun a ->
                              Chara (false, false, false, true, true, true,
                                      true, false) ::
                                a)
                            s)
                          mu) &&
                      qnf (subst_apply_term
                            (map_term (fun x -> x)
                              (fun a ->
                                Chara (true, false, false, true, true, true,
true, false) ::
                                  a)
                              u)
                            mu)) &&
                  (let cu = ica [] u in
                    (match mgu_class _B2 cu t with None -> false
                      | Some mu ->
                        qnf (subst_apply_term
                              (map_term (fun x -> x)
                                (fun a ->
                                  Chara (true, false, false, true, true, true,
  true, false) ::
                                    a)
                                s)
                              mu)))))));;

let rec deep_normalize_cca
  = function CC_impl ([], c) -> deep_normalize_cca c
    | CC_impl (v :: va, c) ->
        CC_impl (map deep_normalize_cca (v :: va), deep_normalize_cca c)
    | CC_cond (s, c) -> CC_cond (s, c)
    | CC_all (s, c) -> CC_all (s, deep_normalize_cca c)
    | CC_rewr (s, c) -> CC_rewr (s, c);;

let rec vars_cc_list _B
  = function CC_cond (ct, (s, t)) -> vars_term_list s @ vars_term_list t
    | CC_rewr (s, t) -> vars_term_list s @ vars_term_list t
    | CC_impl (c1, c2) -> maps (vars_cc_list _B) c1 @ vars_cc_list _B c2
    | CC_all (x, c) ->
        maps (fun y -> (if not (eq _B y x) then [y] else []))
          (vars_cc_list _B c);;

let rec cc_subst_apply _A
  fresh x1 sigma = match fresh, x1, sigma with
    fresh, CC_all (x, c), (sigma, vs) ->
      (let y = fresh (vs @ vars_cc_list _A (CC_all (x, c))) in
        CC_all
          (y, cc_subst_apply _A fresh c (fun_upd _A sigma x (Var y), y :: vs)))
    | fresh, CC_impl (c1, c2), sigma ->
        CC_impl
          (map (fun c -> cc_subst_apply _A fresh c sigma) c1,
            cc_subst_apply _A fresh c2 sigma)
    | fresh, CC_rewr (s, t), (sigma, uv) ->
        CC_rewr (subst_apply_term s sigma, subst_apply_term t sigma)
    | fresh, CC_cond (ct, (s, t)), (sigma, uu) ->
        CC_cond (ct, (subst_apply_term s sigma, subst_apply_term t sigma));;

let rec normalize_alpha _A
  fresh c = cc_subst_apply _A fresh c ((fun a -> Var a), []);;

let rec deep_normalize_cc _A
  fresh c = normalize_alpha _A fresh (deep_normalize_cca c);;

let rec check_subsumesa _A _B
  c x1 = match c, x1 with
    CC_impl (cs, c), CC_impl (ds, d) ->
      check_subsumesa _A _B c d &&
        list_all (fun ca -> list_ex (fun da -> check_subsumesa _A _B da ca) ds)
          cs
    | CC_cond (v, va), CC_impl (ds, d) ->
        check_subsumesa _A _B (CC_cond (v, va)) d
    | CC_rewr (v, va), CC_impl (ds, d) ->
        check_subsumesa _A _B (CC_rewr (v, va)) d
    | CC_all (v, va), CC_impl (ds, d) ->
        check_subsumesa _A _B (CC_all (v, va)) d
    | CC_all (x, c), CC_all (y, d) -> eq _B x y && check_subsumesa _A _B c d
    | CC_cond (v, va), CC_cond (vb, vc) ->
        equal_cond_constrainta _A _B (CC_cond (v, va)) (CC_cond (vb, vc))
    | CC_cond (v, va), CC_rewr (vb, vc) ->
        equal_cond_constrainta _A _B (CC_cond (v, va)) (CC_rewr (vb, vc))
    | CC_cond (v, va), CC_all (vb, vc) ->
        equal_cond_constrainta _A _B (CC_cond (v, va)) (CC_all (vb, vc))
    | CC_rewr (v, va), CC_cond (vb, vc) ->
        equal_cond_constrainta _A _B (CC_rewr (v, va)) (CC_cond (vb, vc))
    | CC_rewr (v, va), CC_rewr (vb, vc) ->
        equal_cond_constrainta _A _B (CC_rewr (v, va)) (CC_rewr (vb, vc))
    | CC_rewr (v, va), CC_all (vb, vc) ->
        equal_cond_constrainta _A _B (CC_rewr (v, va)) (CC_all (vb, vc))
    | CC_all (v, va), CC_cond (vb, vc) ->
        equal_cond_constrainta _A _B (CC_all (v, va)) (CC_cond (vb, vc))
    | CC_all (v, va), CC_rewr (vb, vc) ->
        equal_cond_constrainta _A _B (CC_all (v, va)) (CC_rewr (vb, vc))
    | c, CC_cond (v, va) -> equal_cond_constrainta _A _B c (CC_cond (v, va))
    | c, CC_rewr (v, va) -> equal_cond_constrainta _A _B c (CC_rewr (v, va))
    | CC_impl (vb, vc), CC_all (v, va) ->
        equal_cond_constrainta _A _B (CC_impl (vb, vc)) (CC_all (v, va));;

let rec check_subsumes _A _B
  fresh c d =
    (let n = deep_normalize_cc _A fresh in check_subsumesa _B _A (n c) (n d));;

let rec check_constraint_present (_B1, _B2, _B3)
  i dpp constant p bef aft ccs =
    (let edg = is_iedg_edge_dpp (_B1, _B2, _B3) i dpp in
     let init_conds =
       initial_conditions_gen_impl (fun st uv -> edg st (fst uv)) bef aft p in
      (fun ct st ->
        catch_error
          (forallM
            (fun sts ->
              check (list_ex
                      (fun (c, uvs) ->
                        disjoint_variant (_B1.compare_compare_order, _B2)
                          (card_UNIV_list, cenum_list, (ceq_list ceq_char),
                            (cproper_interval_list ccompare_char),
                            (compare_list compare_char),
                            (equal_list equal_char), mapping_impl_list,
                            set_impl_list)
                          sts uvs &&
                          check_subsumes (equal_list equal_char) _B2
                            (fresh_string
                              [Chara (false, false, false, true, true, true,
                                       true, false);
                                Chara (false, false, false, true, true, true,
true, false)])
                            c (constraint_of constant ct uvs bef))
                      ccs)
                (comp (comp (comp (showsl_lit "did not find ")
                              (showsl_lit
                                (match ct with Bound -> "bound"
                                  | Strict -> "strict"
                                  | Non_Strict -> "non-strict")))
                        (showsl_lit " constraint for sequence "))
                  (showsl_rules _B3 (showl_list showl_char) sts)))
            (init_conds st))
          (fun x -> Inl (snd x))));;

let rec check_weakly_orthogonal (_A1, _A2)
  r = catch_error
        (bindb
          (check_left_linear_trs _A2
            ((ceq_list ceq_char), (ccompare_list ccompare_char), set_impl_list,
              (showl_list showl_char))
            r)
          (fun _ ->
            bindb (catch_error
                    (forallM
                      (fun (l, _) ->
                        check (not (is_Var l))
                          (showsl_lit
                            "the TRS has variables as left-hand sides"))
                      r)
                    (fun x -> Inl (snd x)))
              (fun _ ->
                catch_error
                  (forallM
                    (fun (_, (s, t)) ->
                      check (equal_terma _A1 (equal_list equal_char) s t)
                        (comp (comp (comp (showsl_lit
    "there is a non-trivial critical pair: ")
                                      (showsl_terma _A2 (showl_list showl_char)
s))
                                (showsl_lit " <- . -> "))
                          (showsl_terma _A2 (showl_list showl_char) t)))
                    (critical_pairs_impl _A1 r r))
                  (fun x -> Inl (snd x)))))
        (fun x ->
          Inl (comp (comp x
                      (showsl_lit
                        "\092hence, the following TRS is not weakly orthogonal\092"))
                (showsl_trs _A2 (showl_list showl_char) r)));;

let rec validc (Non_inf_order_ext (valid, ns, cc, af, desc, more)) = valid;;

let rec descc (Non_inf_order_ext (valid, ns, cc, af, desc, more)) = desc;;

let rec nsb (Non_inf_order_ext (valid, ns, cc, af, desc, more)) = ns;;

let rec cc (Non_inf_order_ext (valid, ns, cc, af, desc, more)) = cc;;

let rec afb (Non_inf_order_ext (valid, ns, cc, af, desc, more)) = af;;

let rec range_vars_impl _A _B
  sigma =
    (let a = mk_subst_domain _A _B sigma in maps (comp vars_term_list snd) a);;

let rec cc_bound x0 c = match x0, c with [], c -> c
                   | x :: xs, c -> CC_all (x, cc_bound xs c);;

let rec cc_ih_prems _A _B
  fresh f q xs phi psi rs_ys_list =
    map (fun (r, ys) ->
          (let rs = args r in
           let mu = mk_subst _A (fun a -> Var a) (zip xs rs) in
           let vs = range_vars_impl _A _B (zip xs rs) in
           let mua = (fun c -> cc_subst_apply _A fresh c (mu, vs)) in
           let a =
             CC_impl
               (CC_rewr (r, subst_apply_term q mu) :: map mua phi, mua psi)
             in
            cc_bound ys a))
      rs_ys_list;;

let rec cc_rule_constraint _A _B
  fresh f ls r q xs phi psi rs_ys_list =
    (let sigma = mk_subst _A (fun a -> Var a) (zip xs ls) in
     let vs = range_vars_impl _A _B (zip xs ls) in
     let rew = CC_rewr (r, subst_apply_term q sigma) in
     let phi_sig = map (fun c -> cc_subst_apply _A fresh c (sigma, vs)) phi in
     let psi_sig = cc_subst_apply _A fresh psi (sigma, vs) in
     let ihs = cc_ih_prems _A _B fresh f q xs phi psi rs_ys_list in
      CC_impl (rew :: phi_sig @ ihs, psi_sig));;

let rec cc_unbound
  = function CC_all (x, c) -> (let (xs, a) = cc_unbound c in (x :: xs, a))
    | CC_cond (v, va) -> ([], CC_cond (v, va))
    | CC_rewr (v, va) -> ([], CC_rewr (v, va))
    | CC_impl (v, va) -> ([], CC_impl (v, va));;

let rec showsl_cc_aux _A _B
  b x1 = match b, x1 with
    b, CC_rewr (s, t) ->
      comp (comp (showsl_terma _A _B s) (showsl_lit " = "))
        (showsl_terma _A _B t)
    | b, CC_cond (stri, (s, t)) ->
        comp (comp (showsl_terma _A _B s)
               (showsl_lit (if stri then " > " else " >= ")))
          (showsl_terma _A _B t)
    | b, CC_all (x, c) ->
        (let s =
           comp (comp (comp (showsl_lit "ALL ") (showsl _B x))
                  (showsl_lit ". "))
             (showsl_cc_aux _A _B false c)
           in
          (if b then comp (comp (showsl_lit "(") s) (showsl_lit ")") else s))
    | b, CC_impl (cs, c2) ->
        comp (comp (comp (comp (showsl_lit "(")
                           (showsl_list_gen id "True" "" " and " ""
                             (map (showsl_cc_aux _A _B true) cs)))
                     (showsl_lit " => "))
               (showsl_cc_aux _A _B true c2))
          (showsl_lit ")");;

let rec showsl_cc _A _B = showsl_cc_aux _A _B false;;

let rec funas_term_list
  = function Var uu -> []
    | Fun (f, ts) -> (f, size_list ts) :: maps funas_term_list ts;;

let rec funas_args_term_list t = maps funas_term_list (args t);;

let rec check_rys (_A1, _A2) (_B1, _B2)
  d rt r rys =
    (let (ra, ys) = rys in
      bindb (check (equal_optiona (equal_prod _A1 equal_nat) (root ra) rt)
              (comp (comp (comp (showsl_lit "root of ")
                            (showsl_terma _A2 _B2 ra))
                      (showsl_lit " is not "))
                (showsl_prod _A2 showl_nat (the rt))))
        (fun _ ->
          bindb (check (equal_terma _A1 _B1 r ra || supt_impl _A1 _B1 r ra)
                  (comp (comp (showsl_terma _A2 _B2 ra)
                          (showsl_lit " is not a subterm of "))
                    (showsl_terma _A2 _B2 r)))
            (fun _ ->
              bindb (catch_error
                      (forallM
                        (fun f ->
                          check (not (d f))
                            (comp (comp (comp
  (comp (showsl_lit "the defined symbol ") (showsl_prod _A2 showl_nat f))
  (showsl_lit " occurs in the subterm "))
                                    (showsl_terma _A2 _B2 ra))
                              (showsl_lit " of the rhs")))
                        (funas_args_term_list ra))
                      (fun x -> Inl (snd x)))
                (fun _ ->
                  catch_error (check_disjoint _B1 ys (vars_term_list r))
                    (fun x ->
                      Inl (comp (comp (showsl _B2 x) (showsl_lit " occurs in "))
                            (showsl_terma _A2 _B2 r)))))));;

let rec prems_of = function CC_impl (c1, c2) -> c1
                   | CC_cond (v, va) -> []
                   | CC_rewr (v, va) -> []
                   | CC_all (v, va) -> [];;

let rec concl_of = function CC_impl (c1, c2) -> c2
                   | CC_cond (v, va) -> CC_cond (v, va)
                   | CC_rewr (v, va) -> CC_rewr (v, va)
                   | CC_all (v, va) -> CC_all (v, va);;

let rec normalize_cc c = CC_impl (prems_of c, concl_of c);;

let rec check_cc_prf (_A1, _A2, _A3, _A4, _A5)
  r d f m_ortho cc x5 = match r, d, f, m_ortho, cc, x5 with
    r, d, f, m_ortho, cc, Final ->
      (match normalize_cc cc
        with CC_cond (_, _) ->
          Inl (comp (showsl_lit
                      "problem in final constraint: it is neither a condition nor an implification of two conditions, but it is\092")
                (showsl_cc _A5 (showl_list showl_char) cc))
        | CC_rewr (_, _) ->
          Inl (comp (showsl_lit
                      "problem in final constraint: it is neither a condition nor an implification of two conditions, but it is\092")
                (showsl_cc _A5 (showl_list showl_char) cc))
        | CC_impl ([], CC_cond (stri, st)) -> Inr [Unconditional_C (stri, st)]
        | CC_impl ([], CC_rewr (_, _)) ->
          Inl (comp (showsl_lit
                      "problem in final constraint: it is neither a condition nor an implification of two conditions, but it is\092")
                (showsl_cc _A5 (showl_list showl_char) cc))
        | CC_impl ([], CC_impl (_, _)) ->
          Inl (comp (showsl_lit
                      "problem in final constraint: it is neither a condition nor an implification of two conditions, but it is\092")
                (showsl_cc _A5 (showl_list showl_char) cc))
        | CC_impl ([], CC_all (_, _)) ->
          Inl (comp (showsl_lit
                      "problem in final constraint: it is neither a condition nor an implification of two conditions, but it is\092")
                (showsl_cc _A5 (showl_list showl_char) cc))
        | CC_impl ([CC_cond (stri, uv)], CC_cond (stria, st)) ->
          (if equal_boola stri stria then Inr [Conditional_C (stri, uv, st)]
            else Inl (comp (showsl_lit
                             "problem in final constraint: different relations for finalizing ")
                       (showsl_cc _A5 (showl_list showl_char) cc)))
        | CC_impl ([CC_cond (_, _)], CC_rewr (_, _)) ->
          Inl (comp (showsl_lit
                      "problem in final constraint: it is neither a condition nor an implification of two conditions, but it is\092")
                (showsl_cc _A5 (showl_list showl_char) cc))
        | CC_impl ([CC_cond (_, _)], CC_impl (_, _)) ->
          Inl (comp (showsl_lit
                      "problem in final constraint: it is neither a condition nor an implification of two conditions, but it is\092")
                (showsl_cc _A5 (showl_list showl_char) cc))
        | CC_impl ([CC_cond (_, _)], CC_all (_, _)) ->
          Inl (comp (showsl_lit
                      "problem in final constraint: it is neither a condition nor an implification of two conditions, but it is\092")
                (showsl_cc _A5 (showl_list showl_char) cc))
        | CC_impl (CC_cond (_, _) :: _ :: _, _) ->
          Inl (comp (showsl_lit
                      "problem in final constraint: it is neither a condition nor an implification of two conditions, but it is\092")
                (showsl_cc _A5 (showl_list showl_char) cc))
        | CC_impl (CC_rewr (_, _) :: _, _) ->
          Inl (comp (showsl_lit
                      "problem in final constraint: it is neither a condition nor an implification of two conditions, but it is\092")
                (showsl_cc _A5 (showl_list showl_char) cc))
        | CC_impl (CC_impl (_, _) :: _, _) ->
          Inl (comp (showsl_lit
                      "problem in final constraint: it is neither a condition nor an implification of two conditions, but it is\092")
                (showsl_cc _A5 (showl_list showl_char) cc))
        | CC_impl (CC_all (_, _) :: _, _) ->
          Inl (comp (showsl_lit
                      "problem in final constraint: it is neither a condition nor an implification of two conditions, but it is\092")
                (showsl_cc _A5 (showl_list showl_char) cc))
        | CC_all (_, _) ->
          Inl (comp (showsl_lit
                      "problem in final constraint: it is neither a condition nor an implification of two conditions, but it is\092")
                (showsl_cc _A5 (showl_list showl_char) cc)))
    | r, da, f, m_ortho, c, Delete_Condition (d, prf) ->
        bindb (check
                (check_subsumes (equal_list equal_char) _A3
                  (fresh_string
                    [Chara (false, false, false, true, true, true, true, false);
                      Chara (false, false, false, true, true, true, true,
                              false)])
                  d c)
                (comp (comp (comp (showsl_lit
                                    "problem in delete conditions when switching from\092")
                              (showsl_cc _A5 (showl_list showl_char) c))
                        (showsl_lit " to\092"))
                  (showsl_cc _A5 (showl_list showl_char) d)))
          (fun _ -> check_cc_prf (_A1, _A2, _A3, _A4, _A5) r da f m_ortho d prf)
    | r, da, f, m_ortho, c, Different_Constructor d ->
        (let CC_impl (cs, _) = normalize_cc c in
          catch_error
            (bindb
              (check
                (membera (equal_cond_constraint _A3 (equal_list equal_char)) cs
                  d)
                (comp (comp (showsl_cc _A5 (showl_list showl_char) d)
                        (showsl_lit "\092is not a premise of "))
                  (showsl_cc _A5 (showl_list showl_char) c)))
              (fun _ ->
                (match d
                  with CC_cond (_, _) ->
                    Inl (comp (showsl_cc _A5 (showl_list showl_char) d)
                          (showsl_lit
                            " is not a rewrite condition of the correct shape"))
                  | CC_rewr (Var _, _) ->
                    Inl (comp (showsl_cc _A5 (showl_list showl_char) d)
                          (showsl_lit
                            " is not a rewrite condition of the correct shape"))
                  | CC_rewr (Fun (_, _), Var _) ->
                    Inl (comp (showsl_cc _A5 (showl_list showl_char) d)
                          (showsl_lit
                            " is not a rewrite condition of the correct shape"))
                  | CC_rewr (Fun (fa, ss), Fun (g, ts)) ->
                    bindb (check (not (da (fa, size_list ss)))
                            (comp (showsl _A5 fa) (showsl_lit " is defined")))
                      (fun _ ->
                        bindb (check
                                (not (equal_proda _A3 equal_nat
                                       (fa, size_list ss) (g, size_list ts)))
                                (comp (comp (showsl_lit "the root ")
(showsl _A5 fa))
                                  (showsl_lit " is identical on both sides")))
                          (fun _ -> Inr []))
                  | CC_impl (_, _) ->
                    Inl (comp (showsl_cc _A5 (showl_list showl_char) d)
                          (showsl_lit
                            " is not a rewrite condition of the correct shape"))
                  | CC_all (_, _) ->
                    Inl (comp (showsl_cc _A5 (showl_list showl_char) d)
                          (showsl_lit
                            " is not a rewrite condition of the correct shape")))))
            (fun x ->
              Inl (comp (comp (comp (comp (comp
    (showsl_lit "problem in Different Constructor with rewrite condition ")
    (showsl_cc _A5 (showl_list showl_char) d))
                                      (showsl_lit
"\092on input constraint\092"))
                                (showsl_cc _A5 (showl_list showl_char) c))
                          (showsl_literal "\092"))
                    x)))
    | r, da, f, m_ortho, ca, Same_Constructor (d, c, p) ->
        (let CC_impl (cs, con) = normalize_cc ca in
          bindb (catch_error
                  (bindb
                    (check
                      (membera
                        (equal_cond_constraint _A3 (equal_list equal_char)) cs
                        d)
                      (comp (comp (showsl_cc _A5 (showl_list showl_char) d)
                              (showsl_lit "\092is not a premise of "))
                        (showsl_cc _A5 (showl_list showl_char) ca)))
                    (fun _ ->
                      (match d
                        with CC_cond (_, _) ->
                          Inl (comp (showsl_cc _A5 (showl_list showl_char) d)
                                (showsl_lit
                                  " is not a rewrite condition of the correct shape"))
                        | CC_rewr (Var _, _) ->
                          Inl (comp (showsl_cc _A5 (showl_list showl_char) d)
                                (showsl_lit
                                  " is not a rewrite condition of the correct shape"))
                        | CC_rewr (Fun (_, _), Var _) ->
                          Inl (comp (showsl_cc _A5 (showl_list showl_char) d)
                                (showsl_lit
                                  " is not a rewrite condition of the correct shape"))
                        | CC_rewr (Fun (fa, ss), Fun (g, ts)) ->
                          bindb (check (not (da (fa, size_list ss)))
                                  (comp (showsl _A5 fa)
                                    (showsl_lit " is defined")))
                            (fun _ ->
                              bindb (check
                                      (equal_proda _A3 equal_nat
(fa, size_list ss) (g, size_list ts))
                                      (comp
(comp (comp (showsl _A5 fa) (showsl_lit " and ")) (showsl _A5 g))
(showsl_lit " are not identical")))
                                (fun _ ->
                                  (let ds =
                                     cs @ map (fun (a, b) -> CC_rewr (a, b))
    (zip ss ts)
                                     in
                                   let db = CC_impl (ds, con) in
                                    check (check_subsumes
    (equal_list equal_char) _A3
    (fresh_string
      [Chara (false, false, false, true, true, true, true, false);
        Chara (false, false, false, true, true, true, true, false)])
    c db)
                                      (comp
(comp (comp (showsl_lit "new constraint is ")
        (showsl_cc _A5 (showl_list showl_char) c))
  (showsl_lit "\092but expected was "))
(showsl_cc _A5 (showl_list showl_char) db)))))
                        | CC_impl (_, _) ->
                          Inl (comp (showsl_cc _A5 (showl_list showl_char) d)
                                (showsl_lit
                                  " is not a rewrite condition of the correct shape"))
                        | CC_all (_, _) ->
                          Inl (comp (showsl_cc _A5 (showl_list showl_char) d)
                                (showsl_lit
                                  " is not a rewrite condition of the correct shape")))))
                  (fun x ->
                    Inl (comp (comp (comp (comp
    (comp (comp (comp (showsl_lit
                        "problem in Same Constructor with rewrite condition ")
                  (showsl_cc _A5 (showl_list showl_char) d))
            (showsl_lit "\092 when switching from\092"))
      (showsl_cc _A5 (showl_list showl_char) ca))
    (showsl_lit " to\092"))
                                      (showsl_cc _A5 (showl_list showl_char) c))
                                (showsl_literal "\092"))
                          x)))
            (fun _ ->
              check_cc_prf (_A1, _A2, _A3, _A4, _A5) r da f m_ortho c p))
    | r, da, f, m_ortho, c, Variable_Equation (x, t, d, p) ->
        (let CC_impl (cs, _) = normalize_cc c in
          bindb (catch_error
                  (bindb
                    (check
                      (membera
                         (equal_cond_constraint _A3 (equal_list equal_char)) cs
                         (CC_rewr (Var x, t)) ||
                        membera
                          (equal_cond_constraint _A3 (equal_list equal_char)) cs
                          (CC_rewr (t, Var x)) &&
                          ball ((ceq_prod _A1 ceq_nat),
                                 (ccompare_prod _A2 ccompare_nat))
                            (funas_term (_A1, _A2, _A4) t)
                            (fun fa -> not (da fa)))
                      (comp (comp (comp (showsl_lit "could not find ")
                                    (showsl_cc _A5 (showl_list showl_char)
                                      (CC_rewr (Var x, t))))
                              (showsl_lit " or reversed as a premise of\092"))
                        (showsl_cc _A5 (showl_list showl_char) c)))
                    (fun _ ->
                      (let ca =
                         cc_subst_apply (equal_list equal_char)
                           (fresh_string
                             [Chara (false, false, false, true, true, true,
                                      true, false);
                               Chara (false, false, false, true, true, true,
                                       true, false)])
                           c (fun_upd (equal_list equal_char) (fun a -> Var a) x
                                t,
                               vars_term_list t)
                         in
                        check (check_subsumes (equal_list equal_char) _A3
                                (fresh_string
                                  [Chara (false, false, false, true, true, true,
   true, false);
                                    Chara (false, false, false, true, true,
    true, true, false)])
                                d ca)
                          (comp (comp (comp (showsl_lit "new constraint is ")
(showsl_cc _A5 (showl_list showl_char) d))
                                  (showsl_lit "\092but expected was "))
                            (showsl_cc _A5 (showl_list showl_char) ca)))))
                  (fun xa ->
                    Inl (comp (comp (comp (comp
    (comp (comp (comp (comp (comp (showsl_lit
                                    "problem in Variable Equation with substitution ")
                              (showsl_lista showl_char x))
                        (showsl_lit "/"))
                  (showsl_terma _A5 (showl_list showl_char) t))
            (showsl_lit " to switch from\092"))
      (showsl_cc _A5 (showl_list showl_char) c))
    (showsl_lit "\092to\092"))
                                      (showsl_cc _A5 (showl_list showl_char) d))
                                (showsl_literal "\092"))
                          xa)))
            (fun _ ->
              check_cc_prf (_A1, _A2, _A3, _A4, _A5) r da f m_ortho d p))
    | r, da, f, m_ortho, ca, Funarg_Into_Var (c, i, x, d, p) ->
        (let CC_impl (cs, con) = normalize_cc ca in
          bindb (catch_error
                  (bindb
                    (check
                      (membera
                        (equal_cond_constraint _A3 (equal_list equal_char)) cs
                        c)
                      (comp (comp (showsl_cc _A5 (showl_list showl_char) c)
                              (showsl_lit "\092is not a premise of "))
                        (showsl_cc _A5 (showl_list showl_char) ca)))
                    (fun _ ->
                      bindb (check
                              (not (membera (equal_list equal_char)
                                     (vars_cc_list (equal_list equal_char) ca)
                                     x))
                              (comp (comp (showsl_lit "variable ")
                                      (showsl_lista showl_char x))
                                (showsl_lit " is not fresh")))
                        (fun _ ->
                          (match c
                            with CC_cond (_, _) ->
                              Inl (comp (showsl_cc _A5 (showl_list showl_char)
  c)
                                    (showsl_lit
                                      " is not a rewrite condition of the correct shape"))
                            | CC_rewr (Var _, _) ->
                              Inl (comp (showsl_cc _A5 (showl_list showl_char)
  c)
                                    (showsl_lit
                                      " is not a rewrite condition of the correct shape"))
                            | CC_rewr (Fun (fa, ss), q) ->
                              bindb (check (less_nat i (size_list ss))
                                      (showsl_lit "invalid position"))
                                (fun _ ->
                                  (let (bef, (pa, aft)) =
                                     (take i ss, (nth ss i, drop (suc i) ss)) in
                                    bindb (catch_error
    (check_subseteq (equal_prod _A3 equal_nat) (funas_term_list pa) f)
    (fun xa ->
      Inl (comp (comp (showsl_lit "function symbol ")
                  (showsl_prod _A5 showl_nat xa))
            (showsl_lit " is not allowed in argument"))))
                                      (fun _ ->
(let px = CC_rewr (pa, Var x) in
 let fq = CC_rewr (Fun (fa, bef @ Var x :: aft), q) in
 let ds = px :: fq :: cs in
 let daa = CC_impl (ds, con) in
  check (check_subsumes (equal_list equal_char) _A3
          (fresh_string
            [Chara (false, false, false, true, true, true, true, false);
              Chara (false, false, false, true, true, true, true, false)])
          d daa)
    (comp (comp (comp (showsl_lit "new constraint is ")
                  (showsl_cc _A5 (showl_list showl_char) d))
            (showsl_lit "\092but expected was "))
      (showsl_cc _A5 (showl_list showl_char) daa))))))
                            | CC_impl (_, _) ->
                              Inl (comp (showsl_cc _A5 (showl_list showl_char)
  c)
                                    (showsl_lit
                                      " is not a rewrite condition of the correct shape"))
                            | CC_all (_, _) ->
                              Inl (comp (showsl_cc _A5 (showl_list showl_char)
  c)
                                    (showsl_lit
                                      " is not a rewrite condition of the correct shape"))))))
                  (fun xa ->
                    Inl (comp (comp (comp (comp
    (comp (comp (comp (comp (comp (comp (comp
  (showsl_lit "problem in introducing fresh variable ")
  (showsl_lista showl_char x))
                                    (showsl_lit " on "))
                              (showsl_nat (suc i)))
                        (showsl_lit "-th argument of lhs of "))
                  (showsl_cc _A5 (showl_list showl_char) c))
            (showsl_lit " to switch from "))
      (showsl_cc _A5 (showl_list showl_char) ca))
    (showsl_lit "to\092"))
                                      (showsl_cc _A5 (showl_list showl_char) d))
                                (showsl_literal "\092"))
                          xa)))
            (fun _ ->
              check_cc_prf (_A1, _A2, _A3, _A4, _A5) r da f m_ortho d p))
    | r, da, f, m_ortho, c, Simplify_Condition (bc, sigma, d, p) ->
        (let CC_impl (cs, psi) = normalize_cc c in
          bindb (catch_error
                  (bindb
                    (check
                      (membera
                        (equal_cond_constraint _A3 (equal_list equal_char)) cs
                        bc)
                      (comp (comp (showsl_cc _A5 (showl_list showl_char) bc)
                              (showsl_lit "\092is not a premise of "))
                        (showsl_cc _A5 (showl_list showl_char) c)))
                    (fun _ ->
                      (let (ys, cc) = cc_unbound bc in
                       let (phi, psia) =
                         (let CC_impl (a, b) = normalize_cc cc in (a, b)) in
                       let dom_ran =
                         mk_subst_domain (equal_list equal_char) _A3 sigma in
                        bindb (catch_error
                                (check_subseteq (equal_list equal_char)
                                  (map fst dom_ran) ys)
                                (fun x ->
                                  Inl (comp (showsl_lista showl_char x)
(showsl_lit " is in the domain of sigma, but not a bound variable "))))
                          (fun _ ->
                            bindb (catch_error
                                    (forallM
                                      (fun fn ->
bindb (check (not (da fn))
        (comp (comp (showsl_lit "symbol ") (showsl_prod _A5 showl_nat fn))
          (showsl_lit " is not allowed in range of sigma, as it is defined")))
  (fun _ ->
    check (membera (equal_prod _A3 equal_nat) f fn)
      (comp (comp (showsl_lit "symbol ") (showsl_prod _A5 showl_nat fn))
        (showsl_lit " is not allowed in range of sigma, as it is not in F"))))
                                      (maps
(fun x_t -> funas_term_list (snd x_t)) dom_ran))
                                    (fun x -> Inl (snd x)))
                              (fun _ ->
                                (let vs =
                                   remdups (equal_list equal_char)
                                     (maps (fun x_t -> vars_term_list (snd x_t))
                                       dom_ran)
                                   in
                                 let sigmaa =
                                   (fun ca ->
                                     cc_subst_apply (equal_list equal_char)
                                       (fresh_string
 [Chara (false, false, false, true, true, true, true, false);
   Chara (false, false, false, true, true, true, true, false)])
                                       ca (mk_subst (equal_list equal_char)
     (fun a -> Var a) sigma,
    vs))
                                   in
                                  bindb (catch_error
  (forallM
    (fun ca ->
      check (list_ex
              (fun caa ->
                check_subsumes (equal_list equal_char) _A3
                  (fresh_string
                    [Chara (false, false, false, true, true, true, true, false);
                      Chara (false, false, false, true, true, true, true,
                              false)])
                  caa (sigmaa ca))
              cs)
        (comp (showsl_cc _A5 (showl_list showl_char) (sigmaa ca))
          (showsl_lit
            "\092is not contained as premise of the input implication")))
    phi)
  (fun x -> Inl (snd x)))
                                    (fun _ ->
                                      (let daa =
 CC_impl (sigmaa psia :: cs, psi) in
check (check_subsumes (equal_list equal_char) _A3
        (fresh_string
          [Chara (false, false, false, true, true, true, true, false);
            Chara (false, false, false, true, true, true, true, false)])
        d daa)
  (comp (comp (comp (showsl_lit "new constraint is ")
                (showsl_cc _A5 (showl_list showl_char) d))
          (showsl_lit "\092but expected was "))
    (showsl_cc _A5 (showl_list showl_char) daa))))))))))
                  (fun x ->
                    Inl (comp (comp (comp (comp
    (comp (comp (comp (comp (comp (showsl_lit
                                    "problem in Simplify Condition with substitution ")
                              (showsl_lista
                                (showl_prod (showl_list showl_char)
                                  (showl_term _A5 (showl_list showl_char)))
                                sigma))
                        (showsl_lit " on IH\092"))
                  (showsl_cc _A5 (showl_list showl_char) bc))
            (showsl_lit "\092to switch from\092"))
      (showsl_cc _A5 (showl_list showl_char) c))
    (showsl_lit "\092to\092"))
                                      (showsl_cc _A5 (showl_list showl_char) d))
                                (showsl_literal "\092"))
                          x)))
            (fun _ ->
              check_cc_prf (_A1, _A2, _A3, _A4, _A5) r da f m_ortho d p))
    | r, da, f, m_ortho, c, Induction (d, ccs, ihs) ->
        (let CC_impl (cs, ca) = normalize_cc c in
          bindb (catch_error
                  (bindb
                    (check m_ortho (showsl_lit "CR or minimality required"))
                    (fun _ ->
                      bindb (catch_error
                              (forallM
                                (fun cc ->
                                  check (membera
  (equal_cond_constraint _A3 (equal_list equal_char)) cs cc)
                                    (comp (comp
    (showsl_cc _A5 (showl_list showl_char) cc)
    (showsl_lit "\092is not a premise of "))
                                      (showsl_cc _A5 (showl_list showl_char)
c)))
                                (d :: ccs))
                              (fun x -> Inl (snd x)))
                        (fun _ ->
                          (match d
                            with CC_cond (_, _) ->
                              Inl (comp (showsl_cc _A5 (showl_list showl_char)
  d)
                                    (showsl_lit
                                      " is not a rewrite condition of the correct shape"))
                            | CC_rewr (Var _, _) ->
                              Inl (comp (showsl_cc _A5 (showl_list showl_char)
  d)
                                    (showsl_lit
                                      " is not a rewrite condition of the correct shape"))
                            | CC_rewr (Fun (fa, xs), q) ->
                              (let csa =
                                 vars_cc_list (equal_list equal_char)
                                   (CC_impl
                                     (CC_rewr (Fun (fa, xs), q) :: ccs, ca))
                                 in
                                bindb (check
(list_all is_Var xs && distinct (equal_list equal_char) (map the_Var xs))
(comp (comp (showsl_lit "arguments of ")
        (showsl_terma _A5 (showl_list showl_char) (Fun (fa, xs))))
  (showsl_lit " are not different variables")))
                                  (fun _ ->
                                    (let xsa = map the_Var xs in
                                     let rt = root (Fun (fa, xs)) in
                                      bindb
(check (is_none (mgu _A3 (equal_list equal_char) (Fun (fa, xs)) q))
  (showsl_lit "lhs and rhs unify"))
(fun _ ->
  bindb (catch_error
          (forallM
            (fun lr ->
              check (if equal_optiona (equal_prod _A3 equal_nat) (root (fst lr))
                          rt
                      then list_ex
                             (fun lra ->
                               eq_rule_mod_vars _A3
                                 ((ccompare_list ccompare_char),
                                   (equal_list equal_char), mapping_impl_list)
                                 lr lra &&
                                 isOK (check_disjoint (equal_list equal_char)
csa (vars_rule_list lra)))
                             (map (fun (ra, _) -> ra) ihs)
                      else true)
                (comp (showsl_lit
                        "could not find variable renamed version of rule ")
                  (showsl_rule _A5 (showl_list showl_char) lr)))
            r)
          (fun x -> Inl (snd x)))
    (fun _ ->
      catch_error
        (forallM
          (fun (a, b) ->
            (let (l, ra) = a in
              (fun (rys, (cc, _)) ->
                catch_error
                  (let cca =
                     cc_rule_constraint (equal_list equal_char) _A3
                       (fresh_string
                         [Chara (false, false, false, true, true, true, true,
                                  false);
                           Chara (false, false, false, true, true, true, true,
                                   false)])
                       fa (args l) ra q xsa ccs ca rys
                     in
                    bindb (catch_error
                            (forallM
                              (check_rys (_A3, _A5)
                                ((equal_list equal_char),
                                  (showl_list showl_char))
                                da rt ra)
                              rys)
                            (fun x -> Inl (snd x)))
                      (fun _ ->
                        check (check_subsumes (equal_list equal_char) _A3
                                (fresh_string
                                  [Chara (false, false, false, true, true, true,
   true, false);
                                    Chara (false, false, false, true, true,
    true, true, false)])
                                cc cca)
                          (comp (comp (comp (showsl_lit "new constraint is ")
(showsl_cc _A5 (showl_list showl_char) cc))
                                  (showsl_lit "\092but expected was "))
                            (showsl_cc _A5 (showl_list showl_char) cca))))
                  (fun x ->
                    Inl (comp (comp (comp (showsl_lit
    "problem in constraint for rule ")
                                      (showsl_rule _A5 (showl_list showl_char)
(l, ra)))
                                (showsl_literal "\092"))
                          x))))
              b)
          ihs)
        (fun x -> Inl (snd x)))))))
                            | CC_impl (_, _) ->
                              Inl (comp (showsl_cc _A5 (showl_list showl_char)
  d)
                                    (showsl_lit
                                      " is not a rewrite condition of the correct shape"))
                            | CC_all (_, _) ->
                              Inl (comp (showsl_cc _A5 (showl_list showl_char)
  d)
                                    (showsl_lit
                                      " is not a rewrite condition of the correct shape"))))))
                  (fun x ->
                    Inl (comp (comp (comp (comp
    (comp (comp (showsl_lit "problem in Induction rule with rewrite condition ")
            (showsl_cc _A5 (showl_list showl_char) d))
      (showsl_lit " to switch from\092"))
    (showsl_cc _A5 (showl_list showl_char) c))
                                      (showsl_lit "\092to\092"))
                                (showsl_list_gen
                                  (fun (_, (_, (cb, _))) ->
                                    comp (showsl_cc _A5 (showl_list showl_char)
   cb)
                                      (showsl_literal "\092"))
                                  "" "" "" "" ihs))
                          x)))
            (fun _ ->
              bindb (mapM (fun (_, a) ->
                            (let (_, aa) = a in
                             let (ab, b) = aa in
                              check_cc_prf (_A1, _A2, _A3, _A4, _A5) r da f
                                m_ortho ab b))
                      ihs)
                (fun fcss -> Inr (concat fcss))));;

let rec check_cc_prfs (_A1, _A2, _A3, _A4, _A5)
  r d f m_ortho x4 = match r, d, f, m_ortho, x4 with
    r, d, f, m_ortho, [] -> Inr []
    | r, d, f, m_ortho, (c, (uu, prf)) :: cpfs ->
        bindb (check_cc_prf (_A1, _A2, _A3, _A4, _A5) r d f m_ortho c prf)
          (fun l1 ->
            bindb (check_cc_prfs (_A1, _A2, _A3, _A4, _A5) r d f m_ortho cpfs)
              (fun l2 -> Inr (l1 @ l2)));;

let rec conditional_general_reduction_pair_proc (_B1, _B2, _B3, _B4, _B5, _B6)
  i grp pstrict pbound prof merge dpp =
    (let Cond_Red_Pair_Prf (c, ccs, bef, aft) = prof in
     let p = pairsb i dpp in
     let r = rulese i dpp in
     let f =
       remdups (equal_prod _B4 equal_nat)
         (funas_trs_list r @
           funas_args_trs_list p @ maps funas_term_list (qc i dpp))
       in
     let rp = grp f in
      (match
        catch_error
          (bindb (validc rp)
            (fun _ ->
              bindb (check (wwf_rulesa i dpp)
                      (showsl_lit "require well-formedness of TRS"))
                (fun _ ->
                  (let is_def = (fun fn -> not (null (rules_mapc i dpp fn))) in
                    bindb (check_varcond_subset _B6
                            ((equal_list equal_char), (showl_list showl_char))
                            p)
                      (fun _ ->
                        bindb (catch_error
                                (forallM
                                  (fun (l, ra) ->
                                    bindb (check_no_var _B6
    (showl_list showl_char) l)
                                      (fun _ ->
bindb (check_no_var _B6 (showl_list showl_char) ra)
  (fun _ -> check_no_defined_root _B6 (showl_list showl_char) is_def ra)))
                                  p)
                                (fun x -> Inl (snd x)))
                          (fun _ ->
                            (let ccsa =
                               map (fun (ca, (uvs, _)) -> (ca, uvs)) ccs in
                             let check_present =
                               check_constraint_present (_B3, _B4, _B6) i dpp c
                                 p bef aft ccsa
                               in
                             let (ps, pns) = split_pairsa i dpp pstrict in
                             let (pb, _) = split_pairsa i dpp pbound in
                             let pi = afb rp in
                             let us =
                               usable_rules_gen _B4 (equal_list equal_char) pi r
                                 p
                               in
                              bindb (catch_error
                                      (forallM (check_present Strict) ps)
                                      (fun x -> Inl (snd x)))
                                (fun _ ->
                                  bindb (catch_error
  (forallM (check_present Non_Strict) pns) (fun x -> Inl (snd x)))
                                    (fun _ ->
                                      bindb
(catch_error (forallM (check_present Bound) pb) (fun x -> Inl (snd x)))
(fun _ ->
  bindb (check (nFQ_subset_NF_rulesc i dpp) (showsl_lit "innermost required"))
    (fun _ ->
      bindb (catch_error
              (catch_error (forallM (nsb rp) us) (fun x -> Inl (snd x)))
              (fun x ->
                Inl (comp (showsl_lit "problem when orienting usable rules\092")
                      x)))
        (fun _ ->
          (let m = minimal i dpp in
           let ortho = isOK (check_weakly_orthogonal (_B4, _B6) r) in
            bindb (catch_error
                    (check_cc_prfs (_B1, _B2, _B4, _B5, _B6) r
                      (fun fn -> not (null (rules_mapc i dpp fn))) f
                      (m || ortho) ccs)
                    (fun x ->
                      Inl (comp (showsl_lit
                                  "problem when simplifying conditional constraints\092")
                            x)))
              (fun fcs ->
                catch_error
                  (catch_error (forallM (cc rp) fcs) (fun x -> Inl (snd x)))
                  (fun x ->
                    Inl (comp (showsl_lit
                                "problem when orienting final (conditional) constraints for pairs\092")
                          x))))))))))))))))
          (fun x ->
            Inl (comp (comp (comp (showsl_lit
                                    "could not apply the bounded increase processor with the following\092")
                              (descc rp))
                        (showsl_literal "\092"))
                  x))
        with Inl a -> Inl a
        | Inr _ ->
          Inr (if merge
                then [delete_P_Pwa i dpp
                        (list_inter
                          (equal_prod (equal_term _B4 (equal_list equal_char))
                            (equal_term _B4 (equal_list equal_char)))
                          pstrict pbound)
                        (list_inter
                          (equal_prod (equal_term _B4 (equal_list equal_char))
                            (equal_term _B4 (equal_list equal_char)))
                          pstrict pbound)]
                else [delete_P_Pwa i dpp pstrict pstrict;
                       delete_P_Pwa i dpp pbound pbound])));;

let rec rules_no_left_vara
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = rules_no_left_var;;

let rec critical_pairs_top_impl _A
  p r = maps (fun (l, ra) ->
               (if is_Var l then []
                 else maps (fun (la, rb) ->
                             (match
                               mgu_var_disjoint_generic (equal_list equal_char)
                                 _A (fun a ->
                                      Chara
(false, false, false, true, true, true, true, false) ::
a)
                                 (fun a ->
                                   Chara (true, false, false, true, true, true,
   true, false) ::
                                     a)
                                 l la
                               with None -> []
                               | Some (sigma, tau) ->
                                 [(subst_apply_term ra sigma,
                                    subst_apply_term rb tau)]))
                        r))
          p;;

let rec showsl_crit_pair _A _B
  lr = comp (comp (comp (comp (showsl_lit "(") (showsl_terma _A _B (fst lr)))
                    (showsl_lit ", "))
              (showsl_terma _A _B (snd lr)))
         (showsl_lit ")");;

let rec check_critical_pairs_innermost (_A1, _A2)
  r = catch_error
        (forallM
          (fun (l, ra) ->
            check (equal_terma _A1 (equal_list equal_char) l ra)
              (comp (showsl_lit "there is a non-trivial critical pair ")
                (showsl_crit_pair _A2 (showl_list showl_char) (l, ra))))
          (critical_pairs_top_impl _A1 r r))
        (fun x -> Inl (snd x));;

let rec extract_fresh_var
  sts = (match
          (match sts with [] -> None | (_, (Var _, _)) :: _ -> None
            | (_, (Fun (_, ss), _)) :: _ ->
              (if null ss then None
                else (match last ss with Var a -> Some a | Fun (_, _) -> None)))
          with None ->
            Inl (showsl_lit
                  "could not extract fresh variable (as last argument from some lhs of new pairs)")
          | Some a -> Inr a);;

let rec extract_ren _A
  ps_ps =
    bindb (check
            (list_all
              (fun (a, b) ->
                (let (s, t) = a in
                  (fun (sa, ta) ->
                    not (is_Var s) &&
                      (not (is_Var t) && (not (is_Var sa) && not (is_Var ta)))))
                  b)
              ps_ps)
            (showsl_lit "all lhss and rhss of pairs must be non-variables"))
      (fun _ ->
        (let rt = (fun t -> the (root t)) in
         let pair = (fun s sa -> (rt s, fst (rt sa))) in
         let pairs =
           (fun (st, sta) -> [pair (fst st) (fst sta); pair (snd st) (snd sta)])
           in
         let ren =
           map_of (equal_prod _A equal_nat)
             (remdups (equal_prod (equal_prod _A equal_nat) _A)
               (maps pairs ps_ps))
           in
         let a = (fun fn -> (match ren fn with None -> fst fn | Some f -> f)) in
          Inr a));;

let rec check_drop (_A1, _A2) (_B1, _B2)
  x c ren st_st =
    (let ((s, t), (sa, ta)) = st_st in
     let Fun (f, ss) = s in
     let Fun (g, ts) = t in
      bindb (check
              (equal_terma _B1 _A1 sa
                (Fun (ren (f, size_list ss), ss @ [Var x])))
              (comp (comp (comp (showsl_lit "could not relate ")
                            (showsl_terma _B2 _A2 s))
                      (showsl_lit " with "))
                (showsl_terma _B2 _A2 sa)))
        (fun _ ->
          (let tsa = args ta in
           let tsaa = take (minus_nata (size_list tsa) one_nata) tsa in
            check (equal_terma _B1 _A1 ta
                     (Fun (ren (g, size_list ts), tsaa @ [Var x])) &&
                    equal_lista (equal_term _B1 _A1) ts
                      (map (fun tb -> subst_apply_term tb (subst _A1 x c))
                        tsaa))
              (comp (comp (comp (showsl_lit "could not relate ")
                            (showsl_terma _B2 _A2 t))
                      (showsl_lit " with "))
                (showsl_terma _B2 _A2 ta)))));;

let rec rwd
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = rw;;

let rec pwb
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = pw;;

let rec rd
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = r;;

let rec pb
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = p;;

let rec complex_constant_removal_proc (_B1, _B2, _B3)
  i (Complex_Constant_Removal_Proof (c, ps)) dpp =
    catch_error
      (let p = pb i dpp in
       let pw = pwb i dpp in
       let r = rwd i dpp in
       let q = qc i dpp in
       let pairs = pairsb i dpp in
        bindb (extract_fresh_var ps)
          (fun x ->
            bindb (extract_ren _B2 ps)
              (fun ren ->
                (let is_def = (fun fn -> not (null (rules_mapc i dpp fn))) in
                 let rQs =
                   remdups (equal_option (equal_prod _B2 equal_nat))
                     (map root q)
                   in
                  bindb (catch_error
                          (forallM
                            (fun (s, t) ->
                              bindb (check_no_var _B3 (showl_list showl_char) s)
                                (fun _ ->
                                  bindb (check_no_var _B3
  (showl_list showl_char) t)
                                    (fun _ ->
                                      bindb
(check_no_defined_root _B3 (showl_list showl_char) is_def t)
(fun _ ->
  bindb (check (not (membera (equal_list equal_char) (vars_rule_list (s, t)) x))
          (comp (comp (showsl_lista showl_char x)
                  (showsl_lit " is not fresh for pair "))
            (showsl_rule _B3 (showl_list showl_char) (s, t))))
    (fun _ ->
      (let f = the (root s) in
       let fa = (ren f, suc (snd f)) in
        bindb (check
                (not (membera (equal_option (equal_prod _B2 equal_nat)) rQs
                       (Some fa)))
                (showsl_lit "renaming delivers defined symbol of Q"))
          (fun _ ->
            check (not (is_def fa))
              (showsl_lit "renaming delivers defined symbol of R"))))))))
                            pairs)
                          (fun xa -> Inl (snd xa)))
                    (fun _ ->
                      (let pps =
                         filtera
                           (fun st_st ->
                             membera
                               (equal_prod
                                 (equal_term _B2 (equal_list equal_char))
                                 (equal_term _B2 (equal_list equal_char)))
                               p (fst st_st))
                           ps
                         in
                       let pwps =
                         filtera
                           (fun st_st ->
                             membera
                               (equal_prod
                                 (equal_term _B2 (equal_list equal_char))
                                 (equal_term _B2 (equal_list equal_char)))
                               pw (fst st_st))
                           ps
                         in
                        bindb (catch_error
                                (forallM
                                  (fun st ->
                                    check (membera
    (equal_prod (equal_term _B2 (equal_list equal_char))
      (equal_term _B2 (equal_list equal_char)))
    (map fst pps) st)
                                      (comp
(showsl_lit "could not find entry for pair ")
(showsl_rule _B3 (showl_list showl_char) st)))
                                  p)
                                (fun xa -> Inl (snd xa)))
                          (fun _ ->
                            bindb (catch_error
                                    (forallM
                                      (fun st ->
check (membera
        (equal_prod (equal_term _B2 (equal_list equal_char))
          (equal_term _B2 (equal_list equal_char)))
        (map fst pwps) st)
  (comp (showsl_lit "could not find entry for pair ")
    (showsl_rule _B3 (showl_list showl_char) st)))
                                      pw)
                                    (fun xa -> Inl (snd xa)))
                              (fun _ ->
                                bindb (check (ground c)
(comp (comp (showsl_lit "the term ")
        (showsl_terma _B3 (showl_list showl_char) c))
  (showsl_lit " is not ground")))
                                  (fun _ ->
                                    bindb (check (nFQ_subset_NF_rulesc i dpp)
    (showsl_lit "innermost required"))
                                      (fun _ ->
bindb (check (null (rd i dpp)) (showsl_lit "strict rules not allowed"))
  (fun _ ->
    bindb (check (rules_no_left_vara i dpp)
            (showsl_lit "rules may not have variables as lhss"))
      (fun _ ->
        bindb (if is_NF_trs (_B1, _B2)
                    ((ccompare_list ccompare_char), (equal_list equal_char),
                      mapping_impl_list)
                    r c
                then Inr ()
                else catch_error (check_critical_pairs_innermost (_B2, _B3) r)
                       (fun xa ->
                         Inl (comp (showsl_lit
                                     "could not ensure confluence\092")
                               xa)))
          (fun _ ->
            bindb (catch_error
                    (forallM
                      (fun st_st ->
                        catch_error
                          (check_drop
                            ((equal_list equal_char), (showl_list showl_char))
                            (_B2, _B3) x c ren st_st)
                          (fun xa ->
                            Inl (comp (comp
(comp (comp (comp (showsl_lit "problem in finding correspondence between rule ")
              (showsl_rule _B3 (showl_list showl_char) (fst st_st)))
        (showsl_lit " and rule "))
  (showsl_rule _B3 (showl_list showl_char) (snd st_st)))
(showsl_literal "\092"))
                                  xa)))
                      ps)
                    (fun xa -> Inl (snd xa)))
              (fun _ ->
                Inr (mkd i (nfsc i dpp) (minimal i dpp) (map snd pps)
                      (map snd pwps) q [] r)))))))))))))))
      (fun x ->
        Inl (comp (showsl_lit "problem in complex constant removal proc:\092")
              x));;

let rec fun_of_default _A
  m d = (let mm = map_of _A m in
          (fun i -> (match mm i with None -> d | Some e -> e)));;

let rec get_arg
  t n = (if equal_nata n zero_nata then t
          else nth (args t) (minus_nata n one_nata));;

let rec iteratei_bset_op_list_it_dflt_basic_oops_rm_basic_ops _A
  s = (fun c f -> rm_iterateoi (impl_of _A s) c (comp f fst));;

let rec g_ball_dflt_basic_oops_rm_basic_ops _A
  s p = iteratei_bset_op_list_it_dflt_basic_oops_rm_basic_ops _A s (fun c -> c)
          (fun x _ -> p x) true;;

let rec subtract_list_sorted (_A1, _A2)
  x0 ys = match x0, ys with
    x :: xs, y :: ys ->
      (if eq _A1 x y then subtract_list_sorted (_A1, _A2) xs (y :: ys)
        else (if less _A2.order_linorder.preorder_order.ord_preorder x y
               then x :: subtract_list_sorted (_A1, _A2) xs (y :: ys)
               else subtract_list_sorted (_A1, _A2) (x :: xs) ys))
    | [], ys -> []
    | v :: va, [] -> v :: va;;

let rec remdups_sort (_A1, _A2)
  xs = remdups_adj _A1 (sort_key _A2 (fun x -> x) xs);;

let rec compa (_A1, _A2)
  esa es =
    remdups_sort ((equal_prod _A1 _A1), (linorder_prod _A2 _A2))
      (maps (fun (x, y) ->
              maps (fun (ya, z) -> (if eq _A1 y ya then [(x, z)] else [])) es)
        esa);;

let rec scg_comp (_B1, _B2)
  conn x1 g = match conn, x1, g with
    conn, Scg (pa, qa, stra, wka), Scg (p, q, str, wk) ->
      (if not (conn qa p) then Null
        else (let strs =
                remdups_sort ((equal_prod _B1 _B1), (linorder_prod _B2 _B2))
                  (compa (_B1, _B2) stra str @
                    compa (_B1, _B2) stra wk @ compa (_B1, _B2) wka str)
                in
              let a =
                subtract_list_sorted
                  ((equal_prod _B1 _B1), (linorder_prod _B2 _B2))
                  (remdups_sort ((equal_prod _B1 _B1), (linorder_prod _B2 _B2))
                    (compa (_B1, _B2) wka wk))
                  strs
                in
               Scg (pa, q, strs, a)))
    | conn, Null, g -> Null
    | conn, Scg (v, va, vb, vc), Null -> Null;;

let rec generate_scgs _A (_B1, _B2)
  conn base g =
    filtera (fun ga -> not (equal_scga _A _B1 ga Null))
      (map (scg_comp (_B1, _B2) conn g) base);;

let rec subsumes _A _B
  g x1 = match g, x1 with
    Scg (pa, qa, stra, wka), Scg (p, q, str, wk) ->
      eq _A pa p &&
        (eq _A qa q &&
          (list_all (membera (equal_prod _B _B) str) stra &&
            list_all (membera (equal_prod _B _B) (str @ wk)) wka))
    | g, Null -> true
    | Null, Scg (v, va, vb, vc) -> false;;

let rec in_situ _B
  = function Null -> true
    | Scg (p, q, str, wk) -> list_ex (fun (a, b) -> eq _B a b) str;;

let rec union_list_sorted (_A1, _A2)
  x0 ys = match x0, ys with
    x :: xs, y :: ys ->
      (if eq _A1 x y then x :: union_list_sorted (_A1, _A2) xs ys
        else (if less _A2 x y
               then x :: union_list_sorted (_A1, _A2) xs (y :: ys)
               else y :: union_list_sorted (_A1, _A2) (x :: xs) ys))
    | [], ys -> ys
    | v :: va, [] -> v :: va;;

let rec combinea (_B1, _B2)
  x0 s = match x0, s with
    Scg (pa, qa, stra, wka), Scg (p, q, str, wk) ->
      Scg (pa, qa,
            union_list_sorted
              ((equal_prod _B1 _B1),
                (ord_prod _B2.order_linorder.preorder_order.ord_preorder
                  _B2.order_linorder.preorder_order.ord_preorder))
              stra str,
            union_list_sorted
              ((equal_prod _B1 _B1),
                (ord_prod _B2.order_linorder.preorder_order.ord_preorder
                  _B2.order_linorder.preorder_order.ord_preorder))
              wka wk)
    | Null, s -> Null
    | Scg (v, va, vb, vc), Null -> Null;;

let rec sagiv _A (_B1, _B2)
  conn g =
    (if in_situ _B1 g then true
      else (let gg = scg_comp (_B1, _B2) conn g g in
             (if subsumes _A _B1 gg g then false
               else sagiv _A (_B1, _B2) conn (combinea (_B1, _B2) g gg))));;

let rec check_SCT (_A1, _A2) (_B1, _B2)
  conn gs =
    g_ball_dflt_basic_oops_rm_basic_ops
      (linorder_scg _A1.compare_compare_order _B1.compare_compare_order)
      (mk_rtrancl_set
        (compare_order_scg _A1.compare_compare_order _B1.compare_compare_order)
        (generate_scgs _A2 (_B2, _B1.linorder_compare_order) conn gs) gs)
      (sagiv _A2 (_B2, _B1.linorder_compare_order) conn);;

let rec sct_subterm_precise_proc (_B1, _B2, _B3)
  i gs dpp =
    catch_error
      (let p = pairsb i dpp in
       let is_def = (fun fn -> not (null (rules_mapc i dpp fn))) in
       let eidg = is_iedg_edge_dpp (_B1, _B2, _B3) i dpp in
        bindb (catch_error
                (check_subseteq
                  (equal_prod (equal_term _B2 (equal_list equal_char))
                    (equal_term _B2 (equal_list equal_char)))
                  p (map fst gs))
                (fun x ->
                  Inl (comp (showsl_lit
                              "there is no size-change graph for the pair ")
                        (showsl_rule _B3 (showl_list showl_char) x))))
          (fun _ ->
            (let gGs =
               filtera
                 (fun g ->
                   membera
                     (equal_prod (equal_term _B2 (equal_list equal_char))
                       (equal_term _B2 (equal_list equal_char)))
                     p (fst g))
                 gs
               in
              bindb (check (minimal i dpp || nFQ_subset_NF_rulesc i dpp)
                      (showsl_lit "minimality or innermost required"))
                (fun _ ->
                  bindb (catch_error
                          (forallM
                            (fun (l, _) ->
                              check_no_var _B3 (showl_list showl_char) l)
                            (rulese i dpp))
                          (fun x -> Inl (snd x)))
                    (fun _ ->
                      bindb (catch_error
                              (forallM
                                (fun (a, b) ->
                                  (let (s, t) = a in
                                    (fun (stri, nstri) ->
                                      catch_error
(bindb (check_no_var _B3 (showl_list showl_char) s)
  (fun _ ->
    bindb (check_no_var _B3 (showl_list showl_char) t)
      (fun _ ->
        bindb (check_no_defined_root _B3 (showl_list showl_char) is_def t)
          (fun _ ->
            (let m = size_list (args t) in
             let n = size_list (args s) in
              bindb (catch_error
                      (forallM
                        (fun (ia, j) ->
                          check (less_eq_nat ia n &&
                                  (less_eq_nat j m &&
                                    isOK (check_supt (_B2, _B3)
   ((equal_list equal_char), (showl_list showl_char)) (get_arg s ia)
   (get_arg t j))))
                            (comp (comp (comp
  (comp (showsl_lit "problem with edge ") (showsl_nat ia)) (showsl_lit " |> "))
                                    (showsl_nat j))
                              (showsl_literal "\092")))
                        stri)
                      (fun x -> Inl (snd x)))
                (fun _ ->
                  catch_error
                    (forallM
                      (fun (ia, j) ->
                        check (less_eq_nat ia n &&
                                (less_eq_nat j m &&
                                  isOK (check_supteq (_B2, _B3)
 ((equal_list equal_char), (showl_list showl_char)) (get_arg s ia)
 (get_arg t j))))
                          (comp (comp (comp
(comp (showsl_lit "problem with edge ") (showsl_nat ia)) (showsl_lit " |>= "))
                                  (showsl_nat j))
                            (showsl_literal "\092")))
                      nstri)
                    (fun x -> Inl (snd x))))))))
(fun x ->
  Inl (comp (comp (comp (showsl_lit "problem with pair ")
                    (showsl_rule _B3 (showl_list showl_char) (s, t)))
              (showsl_literal "\092"))
        x))))
                                    b)
                                gGs)
                              (fun x -> Inl (snd x)))
                        (fun _ ->
                          (let n = size_list p in
                           let nums = upt zero_nata n in
                           let numPs = zip p nums in
                           let num_of =
                             fun_of_default
                               (equal_prod
                                 (equal_term _B2 (equal_list equal_char))
                                 (equal_term _B2 (equal_list equal_char)))
                               numPs n
                             in
                            check (check_SCT
                                    ((compare_order_prod compare_order_nat
                                       (compare_order_list
 (compare_order_nat, equal_nat))),
                                      (equal_prod equal_nat
(equal_list equal_nat)))
                                    (compare_order_nat, equal_nat)
                                    (fun (_, succs) (uv, _) ->
                                      membera equal_nat succs uv)
                                    (map (fun (st, (stri, nstri)) ->
   (let eidg_st = eidg st in
    let ia = num_of st in
    let e =
      (ia, map_filter
             (fun x ->
               (if comp (comp eidg_st fst) fst x then Some (snd x) else None))
             numPs)
      in
     Scg (e, e,
           remdups_sort
             ((equal_prod equal_nat equal_nat),
               (linorder_prod linorder_nat linorder_nat))
             stri,
           remdups_sort
             ((equal_prod equal_nat equal_nat),
               (linorder_prod linorder_nat linorder_nat))
             nstri)))
                                      gGs))
                              (showsl_lit
                                "size-change analysis failed\092"))))))))
      (fun x ->
        Inl (comp (showsl_lit
                    "could not apply the size-change processor based on the subterm-relation\092")
              x));;

let rec sct_subterm_approx_proc (_B1, _B2, _B3)
  i gs dpp =
    catch_error
      (let p = pairsb i dpp in
       let is_def = (fun fn -> not (null (rules_mapc i dpp fn))) in
        bindb (catch_error
                (check_subseteq
                  (equal_prod (equal_term _B2 (equal_list equal_char))
                    (equal_term _B2 (equal_list equal_char)))
                  p (map fst gs))
                (fun x ->
                  Inl (comp (showsl_lit
                              "there is no size-change graph for the pair ")
                        (showsl_rule _B3 (showl_list showl_char) x))))
          (fun _ ->
            (let gGs =
               filtera
                 (fun g ->
                   membera
                     (equal_prod (equal_term _B2 (equal_list equal_char))
                       (equal_term _B2 (equal_list equal_char)))
                     p (fst g))
                 gs
               in
              bindb (check (minimal i dpp || nFQ_subset_NF_rulesc i dpp)
                      (showsl_lit "minimality or innermost required"))
                (fun _ ->
                  bindb (catch_error
                          (forallM
                            (fun (l, _) ->
                              check_no_var _B3 (showl_list showl_char) l)
                            (rulese i dpp))
                          (fun x -> Inl (snd x)))
                    (fun _ ->
                      bindb (catch_error
                              (forallM
                                (fun (a, b) ->
                                  (let (s, t) = a in
                                    (fun (stri, nstri) ->
                                      catch_error
(bindb (check_no_var _B3 (showl_list showl_char) s)
  (fun _ ->
    bindb (check_no_var _B3 (showl_list showl_char) t)
      (fun _ ->
        bindb (check_no_defined_root _B3 (showl_list showl_char) is_def t)
          (fun _ ->
            (let m = size_list (args t) in
             let n = size_list (args s) in
              bindb (catch_error
                      (forallM
                        (fun (ia, j) ->
                          check (less_eq_nat ia n &&
                                  (less_eq_nat j m &&
                                    isOK (check_supt (_B2, _B3)
   ((equal_list equal_char), (showl_list showl_char)) (get_arg s ia)
   (get_arg t j))))
                            (comp (comp (comp
  (comp (showsl_lit "problem with edge ") (showsl_nat ia)) (showsl_lit " |> "))
                                    (showsl_nat j))
                              (showsl_literal "\092")))
                        stri)
                      (fun x -> Inl (snd x)))
                (fun _ ->
                  catch_error
                    (forallM
                      (fun (ia, j) ->
                        check (less_eq_nat ia n &&
                                (less_eq_nat j m &&
                                  isOK (check_supteq (_B2, _B3)
 ((equal_list equal_char), (showl_list showl_char)) (get_arg s ia)
 (get_arg t j))))
                          (comp (comp (comp
(comp (showsl_lit "problem with edge ") (showsl_nat ia)) (showsl_lit " |>= "))
                                  (showsl_nat j))
                            (showsl_literal "\092")))
                      nstri)
                    (fun x -> Inl (snd x))))))))
(fun x ->
  Inl (comp (comp (comp (showsl_lit "problem with pair ")
                    (showsl_rule _B3 (showl_list showl_char) (s, t)))
              (showsl_literal "\092"))
        x))))
                                    b)
                                gGs)
                              (fun x -> Inl (snd x)))
                        (fun _ ->
                          check (check_SCT
                                  ((compare_order_prod
                                     (compare_order_prod _B1 compare_order_nat)
                                     (compare_order_prod _B1
                                       compare_order_nat)),
                                    (equal_prod (equal_prod _B2 equal_nat)
                                      (equal_prod _B2 equal_nat)))
                                  (compare_order_nat, equal_nat)
                                  (fun (_, g) (h, _) ->
                                    equal_proda _B2 equal_nat g h)
                                  (remdups
                                    (equal_scg
                                      (equal_prod (equal_prod _B2 equal_nat)
(equal_prod _B2 equal_nat))
                                      equal_nat)
                                    (map (fun (st, (stri, nstri)) ->
   (let e = (the (root (fst st)), the (root (snd st))) in
     Scg (e, e,
           remdups_sort
             ((equal_prod equal_nat equal_nat),
               (linorder_prod linorder_nat linorder_nat))
             stri,
           remdups_sort
             ((equal_prod equal_nat equal_nat),
               (linorder_prod linorder_nat linorder_nat))
             nstri)))
                                      gGs)))
                            (showsl_lit "size-change analysis failed\092")))))))
      (fun x ->
        Inl (comp (showsl_lit
                    "could not apply the size-change processor based on the subterm-relation\092")
              x));;

let rec sct_subterm_proc (_B1, _B2, _B3)
  i gs dpp =
    (if isOK (sct_subterm_approx_proc (_B1, _B2, _B3) i gs dpp) then Inr ()
      else sct_subterm_precise_proc (_B1, _B2, _B3) i gs dpp);;

let rec sct_entry_to_sts
  s t stri nstri =
    (let a = remdups equal_nat (map snd (stri @ nstri)) in
      map (fun j -> (s, get_arg t j)) a);;

let rec check_sct_entry _A _B
  is_def sa nst s t stri nstri =
    catch_error
      (bindb (check_no_var _A _B s)
        (fun _ ->
          bindb (check_no_var _A _B t)
            (fun _ ->
              bindb (check_no_defined_root _A _B is_def t)
                (fun _ ->
                  (let m = size_list (args t) in
                   let n = size_list (args s) in
                    bindb (catch_error
                            (forallM
                              (fun i ->
                                check (less_eq_nat i n)
                                  (comp (comp (showsl_lit "left-index to large")
  (showsl_nat i))
                                    (showsl_literal "\092")))
                              (remdups equal_nat (map fst (stri @ nstri))))
                            (fun x -> Inl (snd x)))
                      (fun _ ->
                        bindb (catch_error
                                (forallM
                                  (fun j ->
                                    check (less_eq_nat j m)
                                      (comp
(showsl_lit "right-index to large or argument violates usable-rules condition")
(showsl_nat j)))
                                  (remdups equal_nat (map snd (stri @ nstri))))
                                (fun x -> Inl (snd x)))
                          (fun _ ->
                            (let _ = args s in
                             let _ = args t in
                              bindb (catch_error
                                      (forallM
(fun (i, j) ->
  check (isOK (sa (get_arg s i, get_arg t j)))
    (comp (comp (comp (showsl_lit "problem with edge ") (showsl_nat i))
            (showsl_lit " -S-> "))
      (showsl_nat j)))
stri)
                                      (fun x -> Inl (snd x)))
                                (fun _ ->
                                  catch_error
                                    (forallM
                                      (fun (i, j) ->
check (isOK (nst (get_arg s i, get_arg t j)))
  (comp (comp (comp (showsl_lit "problem with edge ") (showsl_nat i))
          (showsl_lit " -NS-> "))
    (showsl_nat j)))
                                      nstri)
                                    (fun x -> Inl (snd x)))))))))))
      (fun x ->
        Inl (comp (comp (comp (showsl_lit "problems with DP ")
                          (showsl_rule _A _B (s, t)))
                    (showsl_literal "\092"))
              x));;

let rec sct_ur_af_proc (_B1, _B2, _B3)
  i rp gs u_opt dpp =
    catch_error
      (bindb (valid rp)
        (fun _ ->
          (let is_def = (fun fn -> not (null (rules_mapc i dpp fn))) in
           let pi = af rp in
           let sa = s rp in
           let nsa = ns rp in
           let nsta = nst rp in
           let p = pairsb i dpp in
           let gGs =
             filtera
               (fun g ->
                 membera
                   (equal_prod (equal_term _B2 (equal_list equal_char))
                     (equal_term _B2 (equal_list equal_char)))
                   p (fst g))
               gs
             in
            bindb (catch_error
                    (forallM
                      (fun (l, _) -> check_no_var _B3 (showl_list showl_char) l)
                      (rulese i dpp))
                    (fun x -> Inl (snd x)))
              (fun _ ->
                bindb (catch_error
                        (forallM
                          (fun (a, b) ->
                            (let (saa, t) = a in
                              (fun (aa, ba) ->
                                check_sct_entry _B3 (showl_list showl_char)
                                  is_def sa nsta saa t aa ba))
                              b)
                          gGs)
                        (fun x -> Inl (snd x)))
                  (fun _ ->
                    (let sts =
                       maps (fun (a, b) ->
                              (let (sb, t) = a in
                                (fun (aa, ba) -> sct_entry_to_sts sb t aa ba))
                                b)
                         gGs
                       in
                      bindb (smart_usable_rules_checker_impl (_B1, _B2, _B3) i
                              dpp pi u_opt sts)
                        (fun u ->
                          bindb (catch_error
                                  (catch_error (forallM nsa u)
                                    (fun x -> Inl (snd x)))
                                  (fun x ->
                                    Inl (comp
  (showsl_lit "problem when orienting usable rules\092") x)))
                            (fun _ ->
                              (let eidg = is_iedg_edge_dpp (_B1, _B2, _B3) i dpp
                                 in
                                bindb (catch_error
(check_subseteq
  (equal_prod (equal_term _B2 (equal_list equal_char))
    (equal_term _B2 (equal_list equal_char)))
  p (map fst gs))
(fun x ->
  Inl (comp (showsl_lit "there is no size-change graph for DP ")
        (showsl_rule _B3 (showl_list showl_char) x))))
                                  (fun _ ->
                                    (let n = size_list p in
                                     let nums = upt zero_nata n in
                                     let numPs = zip p nums in
                                     let num_of =
                                       fun_of_default
 (equal_prod (equal_term _B2 (equal_list equal_char))
   (equal_term _B2 (equal_list equal_char)))
 numPs n
                                       in
                                      check
(check_SCT
  ((compare_order_prod compare_order_nat
     (compare_order_list (compare_order_nat, equal_nat))),
    (equal_prod equal_nat (equal_list equal_nat)))
  (compare_order_nat, equal_nat)
  (fun (_, succs) (uv, _) -> membera equal_nat succs uv)
  (map (fun (st, (stri, nstri)) ->
         (let eidg_st = eidg st in
          let ia = num_of st in
          let e =
            (ia, map_filter
                   (fun x ->
                     (if comp (comp eidg_st fst) fst x then Some (snd x)
                       else None))
                   numPs)
            in
           Scg (e, e, stri, nstri)))
    gGs))
(showsl_lit "size-change analysis failed\092"))))))))))))
      (fun x ->
        Inl (comp (comp (comp (showsl_lit
                                "could not apply the size-change processor with the following\092")
                          (desc rp))
                    (showsl_lit "\092for the following reason\092"))
              x));;

let rec proj_terma (_A1, _A2, _A3) _B
  proj f x2 = match proj, f, x2 with
    proj, f, Var x -> add_mset (equal_term _A3 _B) (Var x) zero_multiseta
    | proj, fa, Fun (f, ts) ->
        (if member ((ceq_prod _A1 ceq_nat), (ccompare_prod _A2 ccompare_nat))
              (f, size_list ts) fa
          then sum_mset (comm_monoid_add_multiset (equal_term _A3 _B))
                 (mset (equal_multiset (equal_term _A3 _B))
                   (map (fun i ->
                          proj_terma (_A1, _A2, _A3) _B proj fa (nth ts i))
                     (status proj (f, size_list ts))))
          else add_mset (equal_term _A3 _B) (Fun (f, ts)) zero_multiseta);;

let rec weak_supt_mul (_A1, _A2, _A3, _A4) (_B1, _B2)
  = (fun proj f s t ->
      multeqp
        ((ceq_term _A4 _B2), (ccompare_term _A3 _B1), (equal_term _A4 _B2),
          set_impl_term)
        (fun x y -> supt_impl _A4 _B2 y x)
        (proj_terma (_A1, _A2, _A4) _B2 proj f t)
        (proj_terma (_A1, _A2, _A4) _B2 proj f s));;

let rec check_supteqproj_pred (_A1, _A2, _A3, _A4, _A5) (_B1, _B2, _B3)
  pi f lr =
    check (let (a, b) = lr in
            weak_supt_mul (_A1, _A2, _A3, _A4) (_B1, _B2) pi f a b)
      (comp (comp (showsl_lit "could not orient rule ")
              (showsl_rule _A5 _B3 lr))
        (showsl_lit " by supteq^mul"));;

let rec strict_supt_mul (_A1, _A2, _A3, _A4) (_B1, _B2)
  = (fun proj f s t ->
      multeqp
        ((ceq_term _A4 _B2), (ccompare_term _A3 _B1), (equal_term _A4 _B2),
          set_impl_term)
        (fun x y -> supt_impl _A4 _B2 y x)
        (proj_terma (_A1, _A2, _A4) _B2 proj f t)
        (proj_terma (_A1, _A2, _A4) _B2 proj f s) &&
        not (equal_multiseta (equal_term _A4 _B2)
              (proj_terma (_A1, _A2, _A4) _B2 proj f s)
              (proj_terma (_A1, _A2, _A4) _B2 proj f t)));;

let rec check_suptproj_pred (_A1, _A2, _A3, _A4, _A5) (_B1, _B2, _B3)
  pi f lr =
    check (let (a, b) = lr in
            strict_supt_mul (_A1, _A2, _A3, _A4) (_B1, _B2) pi f a b)
      (comp (comp (showsl_lit "could not orient rule ")
              (showsl_rule _A5 _B3 lr))
        (showsl_lit " by supt^mul"));;

let rec generalized_subterm_proc (_B1, _B2, _B3, _B4, _B5, _B6) (_C1, _C2, _C3)
  i pi p_remove dpp =
    (match
      catch_error
        (let p = pairsb i dpp in
         let r = rulese i dpp in
         let f = map fst pi in
         let ff =
           set ((ceq_prod _B1 ceq_nat), (ccompare_prod _B2 ccompare_nat),
                 (set_impl_prod _B5 set_impl_nat))
             f
           in
         let pi_opt = status_of _B4 pi in
          bindb (check (null (qc i dpp))
                  (showsl_lit
                    "currently generalized subterm criterion does not support strategies"))
            (fun _ ->
              bindb (check (minimal i dpp) (showsl_lit "minimality required"))
                (fun _ ->
                  bindb (check (not (is_none pi_opt))
                          (showsl_lit
                            "argument filter lists invalid positions"))
                    (fun _ ->
                      (let pia = the pi_opt in
                       let premove =
                         set ((ceq_prod (ceq_term _B4 _C2) (ceq_term _B4 _C2)),
                               (ccompare_prod (ccompare_term _B3 _C1)
                                 (ccompare_term _B3 _C1)),
                               (set_impl_prod set_impl_term set_impl_term))
                           p_remove
                         in
                       let (ps, pns) =
                         partition
                           (fun lr ->
                             member
                               ((ceq_prod (ceq_term _B4 _C2)
                                  (ceq_term _B4 _C2)),
                                 (ccompare_prod (ccompare_term _B3 _C1)
                                   (ccompare_term _B3 _C1)))
                               lr premove)
                           p
                         in
                        bindb (catch_error
                                (forallM
                                  (fun fa ->
                                    check (not (null (status pia fa)))
                                      (comp
(comp (showsl_lit "status of symbol ") (showsl_prod _B6 showl_nat fa))
(showsl_lit " in F must be non-empty")))
                                  f)
                                (fun x -> Inl (snd x)))
                          (fun _ ->
                            bindb (catch_error
                                    (forallM
                                      (fun (l, _) ->
check (not (is_Var l)) (showsl_lit "variables as lhss not allowed"))
                                      r)
                                    (fun x -> Inl (snd x)))
                              (fun _ ->
                                bindb (catch_error
(catch_error
  (forallM
    (check_supteqproj_pred (_B1, _B2, _B3, _B4, _B6) (_C1, _C2, _C3) pia ff)
    (filtera
      (fun lr ->
        member ((ceq_prod _B1 ceq_nat), (ccompare_prod _B2 ccompare_nat))
          (the (root (fst lr))) ff)
      r))
  (fun x -> Inl (snd x)))
(fun x ->
  Inl (comp (showsl_lit "problem when orienting rules with root in F\092") x)))
                                  (fun _ ->
                                    bindb (catch_error
    (catch_error
      (forallM
        (check_supteqproj_pred (_B1, _B2, _B3, _B4, _B6) (_C1, _C2, _C3) pia ff)
        pns)
      (fun x -> Inl (snd x)))
    (fun x -> Inl (comp (showsl_lit "problem when orienting DPs\092") x)))
                                      (fun _ ->
catch_error
  (catch_error
    (forallM
      (check_suptproj_pred (_B1, _B2, _B3, _B4, _B6) (_C1, _C2, _C3) pia ff) ps)
    (fun x -> Inl (snd x)))
  (fun x -> Inl (comp (showsl_lit "problem when orienting DPs\092") x)))))))))))
        (fun x ->
          Inl (comp (showsl_lit "could not apply the subterm processor\092") x))
      with Inl a -> Inl a
      | Inr _ -> Inr (delete_P_Pwa i dpp p_remove p_remove));;

let rec simplifya _A
  = function Var x -> Var x
    | Fun (ConstFa n, ss) -> Fun (ConstFa n, [])
    | Fun (SumFa, ss) ->
        (let ssa =
           filtera
             (fun s ->
               not (equal_terma equal_siga _A s (Fun (ConstFa zero_nata, []))))
             (map (simplifya _A) ss)
           in
          (match ssa with [] -> Fun (ConstFa zero_nata, []) | [s] -> s
            | _ :: _ :: _ -> Fun (SumFa, ssa)))
    | Fun (ProdFa, ss) ->
        (let ssa =
           filtera
             (fun s ->
               not (equal_terma equal_siga _A s (Fun (ConstFa one_nata, []))))
             (map (simplifya _A) ss)
           in
          (if membera (equal_term equal_siga _A) ssa
                (Fun (ConstFa zero_nata, []))
            then Fun (ConstFa zero_nata, [])
            else (match ssa with [] -> Fun (ConstFa one_nata, []) | [s] -> s
                   | _ :: _ :: _ -> Fun (ProdFa, ssa))))
    | Fun (MaxF, ss) ->
        (let ssa =
           filtera
             (fun s ->
               not (equal_terma equal_siga _A s (Fun (ConstFa zero_nata, []))))
             (map (simplifya _A) ss)
           in
          (match ssa with [] -> Fun (ConstFa zero_nata, []) | [s] -> s
            | _ :: _ :: _ -> Fun (MaxF, ssa)));;

let rec simplified_alist _A
  alist = map (map_prod id (simplifya equal_nat)) alist;;

let rec iA_exp_to_tpoly
  = function Var (a, ty) -> PVar a
    | Fun (SumF uu, asa) -> PSum (map iA_exp_to_tpoly asa)
    | Fun (ConstF a, []) -> PNum a
    | Fun (ProdF uv, asa) -> PMult (map iA_exp_to_tpoly asa);;

let rec iA_exp_to_poly (_A1, _A2)
  = comp (poly_of (_A1, _A2) (equal_int, comm_semiring_1_int)) iA_exp_to_tpoly;;

let rec showsl_IA_exp (_A1, _A2, _A3)
  = function
    Fun (LessF, [s; t]) ->
      comp (comp (showsl_IA_exp (_A1, _A2, _A3) s) (showsl_lit " < "))
        (showsl_IA_exp (_A1, _A2, _A3) t)
    | Fun (LeF, [s; t]) ->
        comp (comp (showsl_IA_exp (_A1, _A2, _A3) s) (showsl_lit " <= "))
          (showsl_IA_exp (_A1, _A2, _A3) t)
    | Fun (EqF, [s; t]) ->
        comp (comp (showsl_IA_exp (_A1, _A2, _A3) s) (showsl_lit " = "))
          (showsl_IA_exp (_A1, _A2, _A3) t)
    | Var v ->
        showsl_poly (_A1, _A2, _A3) (one_int, equal_int, showl_int)
          (iA_exp_to_poly (_A1, _A2) (Var v))
    | Fun (LeF, []) ->
        showsl_poly (_A1, _A2, _A3) (one_int, equal_int, showl_int)
          (iA_exp_to_poly (_A1, _A2) (Fun (LeF, [])))
    | Fun (LeF, [v]) ->
        showsl_poly (_A1, _A2, _A3) (one_int, equal_int, showl_int)
          (iA_exp_to_poly (_A1, _A2) (Fun (LeF, [v])))
    | Fun (LeF, v :: vc :: ve :: vf) ->
        showsl_poly (_A1, _A2, _A3) (one_int, equal_int, showl_int)
          (iA_exp_to_poly (_A1, _A2) (Fun (LeF, v :: vc :: ve :: vf)))
    | Fun (SumF vb, va) ->
        showsl_poly (_A1, _A2, _A3) (one_int, equal_int, showl_int)
          (iA_exp_to_poly (_A1, _A2) (Fun (SumF vb, va)))
    | Fun (ConstF vb, va) ->
        showsl_poly (_A1, _A2, _A3) (one_int, equal_int, showl_int)
          (iA_exp_to_poly (_A1, _A2) (Fun (ConstF vb, va)))
    | Fun (ProdF vb, va) ->
        showsl_poly (_A1, _A2, _A3) (one_int, equal_int, showl_int)
          (iA_exp_to_poly (_A1, _A2) (Fun (ProdF vb, va)))
    | Fun (EqF, []) ->
        showsl_poly (_A1, _A2, _A3) (one_int, equal_int, showl_int)
          (iA_exp_to_poly (_A1, _A2) (Fun (EqF, [])))
    | Fun (EqF, [v]) ->
        showsl_poly (_A1, _A2, _A3) (one_int, equal_int, showl_int)
          (iA_exp_to_poly (_A1, _A2) (Fun (EqF, [v])))
    | Fun (EqF, v :: vc :: ve :: vf) ->
        showsl_poly (_A1, _A2, _A3) (one_int, equal_int, showl_int)
          (iA_exp_to_poly (_A1, _A2) (Fun (EqF, v :: vc :: ve :: vf)))
    | Fun (v, []) ->
        showsl_poly (_A1, _A2, _A3) (one_int, equal_int, showl_int)
          (iA_exp_to_poly (_A1, _A2) (Fun (v, [])))
    | Fun (v, [vb]) ->
        showsl_poly (_A1, _A2, _A3) (one_int, equal_int, showl_int)
          (iA_exp_to_poly (_A1, _A2) (Fun (v, [vb])))
    | Fun (v, vb :: vd :: vf :: vg) ->
        showsl_poly (_A1, _A2, _A3) (one_int, equal_int, showl_int)
          (iA_exp_to_poly (_A1, _A2) (Fun (v, vb :: vd :: vf :: vg)));;

let rec to_IA
  = function Var x -> [Var (x, IntT)]
    | Fun (f, ss) ->
        (match f with ConstFa n -> [Fun (ConstF (int_of_nat n), [])]
          | SumFa ->
            (if null ss then [Fun (SumF zero_nata, [])]
              else map (fun a -> Fun (SumF (size_list ss), a))
                     (product_lists (map to_IA ss)))
          | ProdFa ->
            (if null ss then [Fun (ProdF zero_nata, [])]
              else map (fun a -> Fun (ProdF (size_list ss), a))
                     (product_lists (map to_IA ss)))
          | MaxF ->
            (if null ss then [Fun (ConstF zero_inta, [])] else maps to_IA ss));;

let rec le_via_IA
  s t = Conjunction
          (map (fun sa ->
                 Disjunction
                   (map (fun a -> Atom a)
                     (map (fun ta -> Fun (LeF, [sa; ta])) (to_IA t))))
            (to_IA s));;

let rec is_or_and_shape
  = function Disjunction (va :: vc :: ve :: vf) -> false
    | Disjunction (va :: Disjunction ve :: vd) -> false
    | Disjunction (va :: Conjunction (vf :: vh :: vj :: vk) :: vd) -> false
    | Disjunction (va :: Conjunction [vf] :: vd) -> false
    | Disjunction (va :: Conjunction [] :: vd) -> false
    | Disjunction (va :: NegAtom ve :: vd) -> false
    | Disjunction (va :: Atom ve :: vd) -> false
    | Disjunction [va] -> false
    | Disjunction [] -> false
    | Conjunction v -> false
    | NegAtom v -> false
    | Atom v -> false
    | Disjunction [phi_1; Conjunction [phi_2; phi_3]] -> true;;

let rec showsl_formula
  showsl_atom x1 = match showsl_atom, x1 with
    showsl_atom, Disjunction fs ->
      (let a = map (showsl_formula showsl_atom) fs in
        showsl_list_gen id "False" "Disj[" ", " "]" a)
    | showsl_atom, Conjunction fs ->
        (let a = map (showsl_formula showsl_atom) fs in
          showsl_list_gen id "True" "Conj[" ", " "]" a)
    | showsl_atom, NegAtom a ->
        comp (comp (showsl_literal "! (") (showsl_atom a)) (showsl_literal ")")
    | showsl_atom, Atom a -> showsl_atom a;;

let rec remove_Atom _B
  negate_atom x1 = match negate_atom, x1 with
    negate_atom, Atom phi -> form_not (negate_atom phi)
    | negate_atom, NegAtom phi -> NegAtom phi
    | negate_atom, Disjunction phi_s ->
        Disjunction (map (remove_Atom _B negate_atom) phi_s)
    | negate_atom, Conjunction phi_s ->
        Conjunction (map (remove_Atom _B negate_atom) phi_s);;

let rec trivial_clause_checker _A
  f = (let Disjunction a = f in trivial_checker _A a);;

let rec check_valid_formula _A (_B1, _B2) _C (_D1, _D2)
  showsl_atom logic_checker negate_atom phi =
    catch_error
      (let Conjunction phi_s = flatten phi in
        catch_error
          (forallM
            (fun phia ->
              catch_error
                (check
                  (trivial_clause_checker (equal_term _A (equal_prod _B1 _C))
                    phia)
                  "trivial clause checker failed")
                (fun _ ->
                  (let Conjunction phi_sa =
                     flatten (remove_Atom _B2 negate_atom phia) in
                    catch_error (forallM (logic_checker (default _D1)) phi_sa)
                      (fun x -> Inl (snd x)))))
            phi_s)
          (fun x -> Inl (snd x)))
      (fun x ->
        Inl (comp (comp (comp (showsl_lit
                                "problem in checking validity of formula ")
                          (showsl_formula showsl_atom phi))
                    (showsl_literal "\092"))
              x));;

let rec is_disj_shape
  uu uv = match uu, uv with uu, Disjunction (va :: vc :: ve :: vf) -> false
    | uu, Disjunction (va :: Disjunction ve :: vd) -> false
    | uu, Disjunction (va :: Conjunction (vf :: vh :: vj :: vk) :: vd) -> false
    | uu, Disjunction (va :: Conjunction [vf] :: vd) -> false
    | uu, Disjunction (va :: Conjunction [] :: vd) -> false
    | uu, Disjunction (va :: NegAtom ve :: vd) -> false
    | uu, Disjunction (va :: Atom ve :: vd) -> false
    | uu, Disjunction [va] -> false
    | uu, Disjunction (Disjunction vc :: vb) -> false
    | uu, Disjunction (Conjunction (vd :: vf :: vh :: vi) :: vb) -> false
    | uu, Disjunction (Conjunction [vd] :: vb) -> false
    | uu, Disjunction (Conjunction [] :: vb) -> false
    | uu, Disjunction (NegAtom vc :: vb) -> false
    | uu, Disjunction (Atom vc :: vb) -> false
    | uu, Disjunction [] -> false
    | uu, Conjunction v -> false
    | uu, NegAtom v -> false
    | uu, Atom v -> false
    | [], uv -> false
    | hint1 :: hints,
        Disjunction [Conjunction [phi_1; phi_2]; Conjunction [phi_3; phi_4]]
        -> true;;

let rec is_conj_shape uu uv = match uu, uv with uu, Disjunction v -> false
                        | uu, Conjunction (va :: vc :: ve :: vf) -> false
                        | uu, Conjunction [va] -> false
                        | uu, Conjunction [] -> false
                        | uu, NegAtom v -> false
                        | uu, Atom v -> false
                        | v :: vb :: vd :: ve, uv -> false
                        | [v], uv -> false
                        | [], uv -> false
                        | [hint1; hint2], Conjunction [phi_1; phi_2] -> true;;

let rec is_neg_atom_clause = function NegAtom a -> false
                             | Atom a -> false
                             | Conjunction xs -> false
                             | Disjunction ls -> list_all is_neg_atom ls;;

let rec check_formula _A (_B1, _B2) _C (_D1, _D2)
  showsl_atom logic_checker negate_atom x3 psi = match
    showsl_atom, logic_checker, negate_atom, x3, psi with
    showsl_atom, logic_checker, negate_atom, LexStrict hints, psi ->
      (match psi
        with Atom _ ->
          Inl (comp (showsl_literal "LexStrict hint applied on ")
                (showsl_formula showsl_atom psi))
        | NegAtom _ ->
          Inl (comp (showsl_literal "LexStrict hint applied on ")
                (showsl_formula showsl_atom psi))
        | Conjunction _ ->
          Inl (comp (showsl_literal "LexStrict hint applied on ")
                (showsl_formula showsl_atom psi))
        | Disjunction a ->
          (match a
            with [] ->
              Inl (comp (showsl_literal "LexStrict hint applied on ")
                    (showsl_formula showsl_atom psi))
            | aa :: b ->
              check_formula_lex _A (_B1, _B2) _C (_D1, _D2) showsl_atom
                logic_checker negate_atom hints aa b))
    | showsl_atom, logic_checker, negate_atom, LexWeak hints, psi ->
        (match psi
          with Atom _ ->
            Inl (comp (showsl_literal "LexWeak hint applied on ")
                  (showsl_formula showsl_atom psi))
          | NegAtom _ ->
            Inl (comp (showsl_literal "LexWeak hint applied on ")
                  (showsl_formula showsl_atom psi))
          | Conjunction _ ->
            Inl (comp (showsl_literal "LexWeak hint applied on ")
                  (showsl_formula showsl_atom psi))
          | Disjunction a ->
            (match a
              with [] ->
                Inl (comp (showsl_literal "LexWeak hint applied on ")
                      (showsl_formula showsl_atom psi))
              | aa :: b ->
                check_formula_lex_weak _A (_B1, _B2) _C (_D1, _D2) showsl_atom
                  logic_checker negate_atom hints aa b))
    | showsl_atom, logic_checker, negate_atom, Distribute (n, hints), phi ->
        (match phi
          with Atom _ ->
            Inl (comp (showsl_literal "Distribute hint for non Disjunction ")
                  (showsl_formula showsl_atom phi))
          | NegAtom _ ->
            Inl (comp (showsl_literal "Distribute hint for non Disjunction ")
                  (showsl_formula showsl_atom phi))
          | Conjunction _ ->
            Inl (comp (showsl_literal "Distribute hint for non Disjunction ")
                  (showsl_formula showsl_atom phi))
          | Disjunction phi_s ->
            (let l = size_list phi_s in
              bindb (check (less_nat n l)
                      (comp (comp (comp (showsl_literal
  "distribute hint at position ")
                                    (showsl_nat n))
                              (showsl_literal " while goal is length "))
                        (showsl_nat l)))
                (fun _ ->
                  (let pre = take n phi_s in
                   let post = drop (suc n) phi_s in
                    (match nth phi_s n
                      with Atom _ ->
                        Inl (comp (comp (comp
  (showsl_literal "Distribute hint in: \092") (showsl_formula showsl_atom phi))
                                    (showsl_literal
                                      "\092 at non-Conjunction position:\092"))
                              (showsl_formula showsl_atom (nth phi_s n)))
                      | NegAtom _ ->
                        Inl (comp (comp (comp
  (showsl_literal "Distribute hint in: \092") (showsl_formula showsl_atom phi))
                                    (showsl_literal
                                      "\092 at non-Conjunction position:\092"))
                              (showsl_formula showsl_atom (nth phi_s n)))
                      | Conjunction a ->
                        check_formula_dist _A (_B1, _B2) _C (_D1, _D2)
                          showsl_atom logic_checker negate_atom hints pre post a
                      | Disjunction _ ->
                        Inl (comp (comp (comp
  (showsl_literal "Distribute hint in: \092") (showsl_formula showsl_atom phi))
                                    (showsl_literal
                                      "\092 at non-Conjunction position:\092"))
                              (showsl_formula showsl_atom (nth phi_s n))))))))
    | showsl_atom, logic_checker, negate_atom, Erase (n, hint), phi ->
        (match phi
          with Atom _ -> Inl (showsl_literal "Erase hint to non-Disjunction")
          | NegAtom _ -> Inl (showsl_literal "Erase hint to non-Disjunction")
          | Conjunction _ ->
            Inl (showsl_literal "Erase hint to non-Disjunction")
          | Disjunction phi_s ->
            (let l = size_list phi_s in
              bindb (check (less_nat n l)
                      (comp (comp (comp (showsl_literal
  "erase hint at position ")
                                    (showsl_nat n))
                              (showsl_literal " while goal is length "))
                        (showsl_nat l)))
                (fun _ ->
                  (let pre = take n phi_s in
                   let post = drop (suc n) phi_s in
                    check_formula _A (_B1, _B2) _C (_D1, _D2) showsl_atom
                      logic_checker negate_atom hint
                      (Disjunction (pre @ Disjunction [] :: post))))))
    | showsl_atom, logic_checker, negate_atom, Base h, phi ->
        (let psi = remove_Atom _B2 negate_atom (simplify phi) in
          bindb (check (is_neg_atom_clause psi)
                  (comp (showsl_literal "base hint given to ")
                    (showsl_formula showsl_atom psi)))
            (fun _ ->
              catch_error (logic_checker h psi)
                (fun x ->
                  Inl (comp (comp (comp (showsl_literal "problem in checking ")
                                    (showsl_formula showsl_atom phi))
                              (showsl_literal "\092"))
                        x))))
    | showsl_atom, logic_checker, negate_atom, Default, phi ->
        check_valid_formula _A (_B1, _B2) _C (_D1, _D2) showsl_atom
          logic_checker negate_atom phi
and check_formula_lex _A (_B1, _B2) _C (_D1, _D2)
  showsl_atom logic_checker negate_atom hints phi phi_s =
    (if is_conj_shape hints phi
      then (let [hint1; hint2] = hints in
            let Conjunction [phi_1; phi_2] = phi in
             bindb (check_formula _A (_B1, _B2) _C (_D1, _D2) showsl_atom
                     logic_checker negate_atom hint1
                     (Disjunction (phi_1 :: phi_s)))
               (fun _ ->
                 check_formula _A (_B1, _B2) _C (_D1, _D2) showsl_atom
                   logic_checker negate_atom hint2
                   (Disjunction (phi_2 :: phi_s))))
      else (if is_disj_shape hints phi
             then (let hint1 :: hints2 = hints in
                   let Disjunction
                         [Conjunction [phi_1; phi_2];
                           Conjunction [phi_3; phi_4]]
                     = phi in
                    (match hints2
                      with [] ->
                        bindb (check_formula _A (_B1, _B2) _C (_D1, _D2)
                                showsl_atom logic_checker negate_atom hint1
                                (Disjunction (phi_3 :: phi_s)))
                          (fun _ ->
                            check_formula_lex _A (_B1, _B2) _C (_D1, _D2)
                              showsl_atom logic_checker negate_atom hints2 phi_4
                              phi_s)
                      | [hint2] ->
                        bindb (check_formula _A (_B1, _B2) _C (_D1, _D2)
                                showsl_atom logic_checker negate_atom hint1
                                (Disjunction (phi_1 :: phi_s)))
                          (fun _ ->
                            check_formula _A (_B1, _B2) _C (_D1, _D2)
                              showsl_atom logic_checker negate_atom hint2
                              (Disjunction (phi_2 :: phi_s)))
                      | _ :: _ :: _ ->
                        bindb (check_formula _A (_B1, _B2) _C (_D1, _D2)
                                showsl_atom logic_checker negate_atom hint1
                                (Disjunction (phi_3 :: phi_s)))
                          (fun _ ->
                            check_formula_lex _A (_B1, _B2) _C (_D1, _D2)
                              showsl_atom logic_checker negate_atom hints2 phi_4
                              phi_s)))
             else Inl (comp (comp (comp (showsl_literal
  "LexStrict hint application error: ")
                                    (showsl_hint _D2 (LexStrict hints)))
                              (showsl_literal "\092applied on "))
                        (showsl_formula showsl_atom phi))))
and check_formula_dist _A (_B1, _B2) _C (_D1, _D2)
  showsl_atom logic_checker negate_atom x3 uv uw x6 = match
    showsl_atom, logic_checker, negate_atom, x3, uv, uw, x6 with
    showsl_atom, logic_checker, negate_atom, [], uv, uw, v :: va ->
      Inl (showsl_literal "Length mismatch in Distribute hints")
    | showsl_atom, logic_checker, negate_atom, v :: va, uv, uw, [] ->
        Inl (showsl_literal "Length mismatch in Distribute hints")
    | showsl_atom, logic_checker, negate_atom, h :: hs, pre, post, phi :: phi_s
        -> bindb (check_formula _A (_B1, _B2) _C (_D1, _D2) showsl_atom
                   logic_checker negate_atom h
                   (Disjunction (pre @ phi :: post)))
             (fun _ ->
               check_formula_dist _A (_B1, _B2) _C (_D1, _D2) showsl_atom
                 logic_checker negate_atom hs pre post phi_s)
    | showsl_atom, logic_checker, negate_atom, [], pre, post, [] -> Inr ()
and check_formula_lex_weak _A (_B1, _B2) _C (_D1, _D2)
  showsl_atom logic_checker negate_atom x3 phi phi_s = match
    showsl_atom, logic_checker, negate_atom, x3, phi, phi_s with
    showsl_atom, logic_checker, negate_atom, hint :: hints, phi, phi_s ->
      (if is_or_and_shape phi
        then (let Disjunction [_; Conjunction [phi_2; phi_3]] = phi in
               bindb (check_formula _A (_B1, _B2) _C (_D1, _D2) showsl_atom
                       logic_checker negate_atom hint
                       (Disjunction (phi_2 :: phi_s)))
                 (fun _ ->
                   check_formula_lex_weak _A (_B1, _B2) _C (_D1, _D2)
                     showsl_atom logic_checker negate_atom hints phi_3 phi_s))
        else (match hints
               with [] ->
                 check_formula _A (_B1, _B2) _C (_D1, _D2) showsl_atom
                   logic_checker negate_atom hint (Disjunction (phi :: phi_s))
               | _ :: _ ->
                 Inl (showsl_literal "LexWeak hint application error")))
    | showsl_atom, logic_checker, negate_atom, [], phi, phi_s ->
        check (trivial_formula phi)
          (comp (showsl_literal "LexWeak base case error: ")
            (showsl_formula showsl_atom phi));;

let rec ea (_A1, _A2)
  alist default_fun =
    curry (fun_of_map_fun (map_of (equal_prod _A1 equal_nat) alist)
            (fun (_, n) ->
              Fun (default_fun n, map (fun a -> Var a) (upt zero_nata n))));;

let rec iA_exp_to_poly_constraint (_A1, _A2)
  = function
    Fun (LeF, [a; b]) ->
      Poly_Ge
        (poly_minus (_A1, _A2) (equal_int, ring_1_int)
          (iA_exp_to_poly (_A1, _A2) b) (iA_exp_to_poly (_A1, _A2) a))
    | Fun (EqF, [a; b]) ->
        Poly_Eq
          (poly_minus (_A1, _A2) (equal_int, ring_1_int)
            (iA_exp_to_poly (_A1, _A2) b) (iA_exp_to_poly (_A1, _A2) a))
    | Fun (LessF, [a; b]) ->
        Poly_Ge
          (poly_minus (_A1, _A2) (equal_int, ring_1_int)
            (poly_minus (_A1, _A2) (equal_int, ring_1_int)
              (iA_exp_to_poly (_A1, _A2) b) (iA_exp_to_poly (_A1, _A2) a))
            (one_polya _A2 semiring_1_int));;

let rec showsl_poly_constraint (_A1, _A2, _A3)
  = function
    Poly_Ge p ->
      comp (showsl_poly (_A1, _A2, _A3) (one_int, equal_int, showl_int) p)
        (showsl_lit " >= 0")
    | Poly_Eq p ->
        comp (showsl_poly (_A1, _A2, _A3) (one_int, equal_int, showl_int) p)
          (showsl_lit " = 0");;

let rec translate_atom (Atom e) = e;;

let rec translate_atoms x = map translate_atom x;;

let rec translate_conj (Conjunction phi_s) = translate_atoms phi_s;;

let rec poly_is_negative_constant (_A1, _A2, _A3)
  f = catch_error
        (bindb
          (check (null (poly_vars_list (_A1, _A2) f))
            (showsl_lit "polynomial is not a constant"))
          (fun _ ->
            check (less_int
                    (eval_poly _A2 comm_semiring_1_int (fun _ -> zero_inta) f)
                    zero_inta)
              (showsl_lit "polynomial is not a negative constant")))
        (fun x ->
          Inl (comp (comp (comp (showsl_lit "could not that ")
                            (showsl_poly (_A1, _A2, _A3)
                              (one_int, equal_int, showl_int) f))
                      (showsl_lit " is a negative constant\092"))
                x));;

let rec vars_poly_constraint_list (_A1, _A2)
  = function Poly_Ge p -> poly_vars_list (_A1, _A2) p
    | Poly_Eq p -> poly_vars_list (_A1, _A2) p;;

let zero_linear_poly : linear_poly = LinearPoly zeroa;;

let rec lp_monom
  c x = LinearPoly
          (if equal_rata c zero_rata then fmempty
            else fmupd equal_nat x c fmempty);;

let rec monom_list_linearity
  = function [] -> Onea
    | [(x, n)] -> (if equal_nata n one_nata then Variable x else Non_Linear)
    | v :: vb :: vc -> Non_Linear;;

let rec monom_linearity _A xa = monom_list_linearity (rep_monom _A xa);;

let rec ipoly_to_linear_poly _A
  rho x1 = match rho, x1 with rho, [] -> Some (zero_linear_poly, zero_inta)
    | rho, (mon, c) :: rest ->
        bind (ipoly_to_linear_poly _A rho rest)
          (fun (p, d) ->
            (match monom_linearity _A mon with Non_Linear -> None
              | Onea -> Some (p, plus_inta c d)
              | Variable x ->
                Some (plus_linear_poly (lp_monom (of_int c) (rho x)) p, d)));;

let rec to_simplex_constraint _A
  rho x1 = match rho, x1 with
    rho, Poly_Ge p ->
      (match ipoly_to_linear_poly _A rho p with None -> []
        | Some (q, c) -> [GEQ (q, of_int (uminus_inta c))])
    | rho, Poly_Eq p ->
        (match ipoly_to_linear_poly _A rho p with None -> []
          | Some (q, c) -> [EQa (q, of_int (uminus_inta c))]);;

let rec unsat_via_simplex (_A1, _A2, _A3, _A4)
  les = (let vs = remdups _A2 (maps (vars_poly_constraint_list (_A2, _A4)) les)
           in
         let ren_map =
           of_alist (_A1, _A2, _A3) (zip vs (upt zero_nata (size_list vs))) in
         let ren_fun =
           (fun v ->
             (match lookupb (_A1, _A2) ren_map v with None -> zero_nata
               | Some n -> n))
           in
         let cs = maps (to_simplex_constraint _A4 ren_fun) les in
          (match simplex cs with Inl _ -> true | Inr _ -> false));;

let rec apply_hint (_A1, _A2)
  uu x1 = match uu, x1 with
    n :: ns, Poly_Ge a :: asa ->
      poly_add (_A1, _A2) (equal_int, semiring_0_int)
        (poly_mult (_A1, _A2) (equal_int, semiring_0_int)
          (poly_const (zero_int, equal_int) _A2 (abs_inta n)) a)
        (apply_hint (_A1, _A2) ns asa)
    | n :: ns, Poly_Eq a :: asa ->
        poly_add (_A1, _A2) (equal_int, semiring_0_int)
          (poly_mult (_A1, _A2) (equal_int, semiring_0_int)
            (poly_const (zero_int, equal_int) _A2 n) a)
          (apply_hint (_A1, _A2) ns asa)
    | [], Poly_Ge a :: asa ->
        poly_add (_A1, _A2) (equal_int, semiring_0_int) a
          (apply_hint (_A1, _A2) [] asa)
    | [], Poly_Eq a :: asa ->
        poly_add (_A1, _A2) (equal_int, semiring_0_int) a
          (apply_hint (_A1, _A2) [] asa)
    | uu, [] -> poly_const (zero_int, equal_int) _A2 zero_inta;;

let rec unsat_checker (_A1, _A2, _A3, _A4, _A5)
  hints cnjs =
    catch_error
      (match hints
        with Hints coeffs ->
          or_ok (poly_is_negative_constant (_A2, _A4, _A5)
                  (apply_hint (_A2, _A4) (zero_inta :: coeffs) cnjs))
            (poly_is_negative_constant (_A2, _A4, _A5)
              (apply_hint (_A2, _A4) (one_inta :: coeffs) cnjs))
        | Simplex ->
          check (unsat_via_simplex (_A1, _A2, _A3, _A4) cnjs)
            (showsl_lit
              "could not use simplex algorithm to prove unsatisfiability"))
      (fun x ->
        Inl (comp (comp (comp (comp (comp (showsl_lit
    "The linear inequalities\092  ")
                                      (showsl_sep
(showsl_poly_constraint (_A2, _A4, _A5)) (showsl_lit "\092  ") cnjs))
                                (showsl_lit
                                  "\092cannot be proved unsatisfiable via hints\092  "))
                          (showsl_hints hints))
                    (showsl_literal "\092"))
              x));;

let rec check_clause (_A1, _A2, _A3, _A4, _A5)
  hints phi =
    (let es =
       map (iA_exp_to_poly_constraint (_A2, _A4))
         (translate_conj (form_not phi))
       in
      catch_error (unsat_checker (_A1, _A2, _A3, _A4, _A5) hints es)
        (fun x ->
          Inl (comp (comp (comp (showsl_lit
                                  "Could not prove unsatisfiability of IA conjunction\092")
                            (showsl_list_gen
                              (showsl_poly_constraint (_A2, _A4, _A5)) "False"
                              "" " && " "" es))
                      (showsl_literal "\092"))
                x)));;

let rec negate
  = function Fun (LessF, [a; b]) -> Atom (Fun (LeF, [b; a]))
    | Fun (LeF, [a; b]) -> Atom (Fun (LessF, [b; a]))
    | Fun (EqF, [a; b]) ->
        Disjunction [Atom (Fun (LessF, [a; b])); Atom (Fun (LessF, [b; a]))]
    | Var v -> Conjunction []
    | Fun (LeF, []) -> Conjunction []
    | Fun (LeF, [v]) -> Conjunction []
    | Fun (LeF, v :: vc :: ve :: vf) -> Conjunction []
    | Fun (SumF vb, va) -> Conjunction []
    | Fun (ConstF vb, va) -> Conjunction []
    | Fun (ProdF vb, va) -> Conjunction []
    | Fun (EqF, []) -> Conjunction []
    | Fun (EqF, [v]) -> Conjunction []
    | Fun (EqF, v :: vc :: ve :: vf) -> Conjunction []
    | Fun (v, []) -> Conjunction []
    | Fun (v, [vb]) -> Conjunction []
    | Fun (v, vb :: vd :: vf :: vg) -> Conjunction [];;

let rec check_less_eq_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5)
  alist h s t =
    check_formula equal_sig (_B2, _B5) equal_ty (default_hints, showl_hints)
      (showsl_IA_exp (_B2, _B4, _B5)) (check_clause (_B1, _B2, _B3, _B4, _B5))
      negate h
      (le_via_IA
        (evala
          (fun f ss ->
            subst_apply_term
              (ea (_A1, _A2) (simplified_alist _A2 alist) (fun _ -> MaxF) f
                (size_list ss))
              (nth ss))
          s (fun a -> Var a))
        (evala
          (fun f ss ->
            subst_apply_term
              (ea (_A1, _A2) (simplified_alist _A2 alist) (fun _ -> MaxF) f
                (size_list ss))
              (nth ss))
          t (fun a -> Var a)));;

let rec less_via_IA
  s t = Conjunction
          (map (fun sa ->
                 Disjunction
                   (map (fun a -> Atom a)
                     (map (fun ta -> Fun (LessF, [sa; ta])) (to_IA t))))
            (to_IA s));;

let rec check_less_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5)
  alist h s t =
    check_formula equal_sig (_B2, _B5) equal_ty (default_hints, showl_hints)
      (showsl_IA_exp (_B2, _B4, _B5)) (check_clause (_B1, _B2, _B3, _B4, _B5))
      negate h
      (less_via_IA
        (evala
          (fun f ss ->
            subst_apply_term
              (ea (_A1, _A2) (simplified_alist _A2 alist) (fun _ -> MaxF) f
                (size_list ss))
              (nth ss))
          s (fun a -> Var a))
        (evala
          (fun f ss ->
            subst_apply_term
              (ea (_A1, _A2) (simplified_alist _A2 alist) (fun _ -> MaxF) f
                (size_list ss))
              (nth ss))
          t (fun a -> Var a)));;

let rec check_encoding _A (_B1, _B2)
  alist =
    catch_error
      (forallM
        (fun (a, b) ->
          (let (f, n) = a in
            (fun t ->
              check (subset (card_UNIV_nat, cenum_nat, ceq_nat, ccompare_nat)
                      (vars_term (_B1, _B2)
                        (finite_UNIV_nat, cenum_nat, ceq_nat,
                          cproper_interval_nat, compare_nat, equal_nat,
                          set_impl_nat)
                        t)
                      (set (ceq_nat, ccompare_nat, set_impl_nat)
                        (upt zero_nata n)))
                (comp (comp (comp (comp (showsl_lit "interpretation of ")
                                    (showsl _A f))
                              (showsl_lit " arity "))
                        (showsl_nat n))
                  (showsl_lit " has extra parameter"))))
            b)
        alist)
      (fun x -> Inl (snd x));;

let rec constant_positions
  e f n =
    filtera (fun i -> not (contains_var_term equal_nat i (e f n)))
      (upt zero_nata n);;

let rec create_max_poly_redtriple (_A1, _A2) (_B1, _B2, _B3, _B4, _B5)
  alist =
    Redtriple_ext
      (check_encoding _A2 (compare_sig, equal_siga)
         (simplified_alist _A2 alist),
        (fun (s, t) ->
          check_less_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5) alist
            (default_hinta default_hints) t s),
        (fun (s, t) ->
          check_less_eq_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5) alist
            (default_hinta default_hints) t s),
        (fun (s, t) ->
          check_less_eq_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5) alist
            (default_hinta default_hints) t s),
        (fun (f, n) ->
          minus_set (ceq_nat, ccompare_nat)
            (set (ceq_nat, ccompare_nat, set_impl_nat) (upt zero_nata n))
            (set (ceq_nat, ccompare_nat, set_impl_nat)
              (constant_positions
                (ea (_A1, _A2) (simplified_alist _A2 alist) (fun _ -> MaxF)) f
                n))),
        empty_af,
        (fun _ ->
          Inl (showsl_lit "monotonicity of max-poly is not yet supported")),
        showsl_lit "a max polynomial interpretation",
        Some (map fst (simplified_alist _A2 alist)), no_complexity_check, ());;

let rec filter_prec_weight_ac_repr
  pi (prw, w0) =
    (let fprw =
       filtera
         (fun (fn, _) ->
           (match pi fn with Collapse _ -> false | AFList _ -> true))
         prw
       in
     let mprw =
       map (fun (a, b) ->
             (let (f, n) = a in
               (fun aa ->
                 ((FPair (f, n),
                    (match pi (f, n) with Collapse _ -> zero_nata
                      | AFList ab -> size_list ab)),
                   aa)))
               b)
         fprw
       in
      (mprw, w0));;

let rec filter_prec_weight_repr
  pi (prw, w0) =
    (let fprw =
       filtera
         (fun (fn, _) ->
           (match pi fn with Collapse _ -> false | AFList _ -> true))
         prw
       in
     let mprw =
       map (fun (a, b) ->
             (let (f, n) = a in
               (fun aa ->
                 ((FPair (f, n),
                    (match pi (f, n) with Collapse _ -> zero_nata
                      | AFList ab -> size_list ab)),
                   aa)))
               b)
         fprw
       in
      (mprw, w0));;

let rec prec_repr_to_status _A
  prs = (let m = ceta_map_of (compare_order_prod _A compare_order_nat) prs in
          (fun (FPair (f, a), _) ->
            (match m (f, a) with None -> Lex | Some aa -> snd aa)));;

let rec prec_repr_to_pr _A
  prs = (let m = ceta_map_of (compare_order_prod _A compare_order_nat) prs in
          (fun (FPair (f, a), _) ->
            (match m (f, a) with None -> zero_nata | Some aa -> fst aa)));;

let rec plus_single_mono
  (Partial_object_ext
    (carrier,
      Monoid_ext
        (mult, one,
          Ring_ext
            (zero, add,
              Ordered_semiring_ext
                (geq, gt, max,
                  Lpoly_order_semiring_ext
                    (plus_single_mono, default, arcpos, checkmono, bound,
                      check_complexity, description, more))))))
    = plus_single_mono;;

let rec check_complexity
  (Partial_object_ext
    (carrier,
      Monoid_ext
        (mult, one,
          Ring_ext
            (zero, add,
              Ordered_semiring_ext
                (geq, gt, max,
                  Lpoly_order_semiring_ext
                    (plus_single_mono, default, arcpos, checkmono, bound,
                      check_complexity, description, more))))))
    = check_complexity;;

let rec maxb
  (Partial_object_ext
    (carrier,
      Monoid_ext
        (mult, one,
          Ring_ext (zero, add, Ordered_semiring_ext (geq, gt, max, more)))))
    = max;;

let rec poly_c_max_inter_bcoeff_strict
  r f pi = foldr (maxb r) (maps (fun fn -> snd (pi fn)) f) (zero r);;

let rec geq
  (Partial_object_ext
    (carrier,
      Monoid_ext
        (mult, one,
          Ring_ext (zero, add, Ordered_semiring_ext (geq, gt, max, more)))))
    = geq;;

let rec poly_c_max_inter_bcoeff
  r f pi =
    foldr (maxb r)
      (maps (fun fn -> filtera (fun b -> not (geq r (one r) b)) (snd (pi fn)))
        f)
      (zero r);;

let rec convert_lpoly_complexity _A
  r pi cm (Comp_Poly deg) =
    (let f =
       (match cm with Derivational_Complexity f -> f
         | Runtime_Complexity (c, _) -> c)
       in
     let bc = poly_c_max_inter_bcoeff r f pi in
     let bca = poly_c_max_inter_bcoeff_strict r f pi in
      bindb (check (less_nat zero_nata deg || eq _A bca (zero r))
              (showsl_literal
                "constant complexity not fully supported for linear (poly/matrix)-interpretations"))
        (fun _ -> check_complexity r bc (minus_nata deg one_nata)));;

let rec check_poly_mono_npsm (_A1, _A2) (_C1, _C2)
  r f pi =
    bindb (catch_error
            (forallM
              (fun (a, b) ->
                (let (fa, n) = a in
                  (fun (c, cs) ->
                    catch_error
                      (bindb
                        (check
                          (if equal_nata n (suc zero_nata)
                            then eq _A1 c (zero r) else true)
                          (comp (comp (showsl_literal "constant part ")
                                  (showsl _A2 c))
                            (showsl_literal " must be 0\092")))
                        (fun _ ->
                          bindb (check (equal_nata n (size_list cs))
                                  (showsl_literal
                                    "the arity is not the same as the number of arguments\092"))
                            (fun _ ->
                              check (less_eq_nat n (suc zero_nata))
                                (showsl_literal
                                  "symbol has arity larger than 1\092"))))
                      (fun x ->
                        Inl (comp (comp (comp
  (comp (comp (showsl_literal
                "problem with monotonicity due to interpretation of ")
          (showsl _C2 fa))
    (showsl_literal "/"))
  (showsl_nat n))
                                    (showsl_literal "\092"))
                              x))))
                  b)
              pi)
            (fun x -> Inl (snd x)))
      (fun _ ->
        catch_error (check_subseteq (equal_prod _C1 equal_nat) f (map fst pi))
          (fun x ->
            Inl (comp (comp (showsl_literal "unknown interpretation for ")
                        (showsl_prod _C2 showl_nat x))
                  (showsl_literal "\092"))));;

let rec arcpos
  (Partial_object_ext
    (carrier,
      Monoid_ext
        (mult, one,
          Ring_ext
            (zero, add,
              Ordered_semiring_ext
                (geq, gt, max,
                  Lpoly_order_semiring_ext
                    (plus_single_mono, default, arcpos, checkmono, bound,
                      check_complexity, description, more))))))
    = arcpos;;

let rec carrier (Partial_object_ext (carrier, more)) = carrier;;

let rec check_lpoly_coeffs (_A1, _A2, _A3) _C
  r i = catch_error
          (forallM
            (fun (a, b) ->
              (let (f, n) = a in
                (fun (c, cs) ->
                  catch_error
                    (bindb
                      (check (member (_A1, _A2) c (carrier r))
                        (comp (comp (showsl_literal "constant part ")
                                (showsl _A3 c))
                          (showsl_literal " is not well-formed\092")))
                      (fun _ ->
                        bindb (check (less_eq_nat (size_list cs) n)
                                (comp (showsl_literal
"number of coefficients exceeds arity of symbol ")
                                  (showsl _C f)))
                          (fun _ ->
                            bindb (check (arcpos r c || list_ex (arcpos r) cs)
                                    (showsl_literal
                                      "could not find positive entry which is required for arctic interpretations\092"))
                              (fun _ ->
                                catch_error
                                  (forallM
                                    (fun aa ->
                                      check
(geq r aa (zero r) && member (_A1, _A2) aa (carrier r))
(comp (comp (showsl_literal "coefficient ") (showsl _A3 aa))
  (showsl_literal " is not allowed\092")))
                                    cs)
                                  (fun x -> Inl (snd x))))))
                    (fun x ->
                      Inl (comp (comp (comp
(comp (comp (showsl_literal "problem with interpretation of ") (showsl _C f))
  (showsl_literal "/"))
(showsl_nat n))
                                  (showsl_literal "\092"))
                            x))))
                b)
            i)
          (fun x -> Inl (snd x));;

let rec description
  (Partial_object_ext
    (carrier,
      Monoid_ext
        (mult, one,
          Ring_ext
            (zero, add,
              Ordered_semiring_ext
                (geq, gt, max,
                  Lpoly_order_semiring_ext
                    (plus_single_mono, default, arcpos, checkmono, bound,
                      check_complexity, description, more))))))
    = description;;

let rec defaulta
  (Partial_object_ext
    (carrier,
      Monoid_ext
        (mult, one,
          Ring_ext
            (zero, add,
              Ordered_semiring_ext
                (geq, gt, max,
                  Lpoly_order_semiring_ext
                    (plus_single_mono, default, arcpos, checkmono, bound,
                      check_complexity, description, more))))))
    = default;;

let rec to_lpoly_inter _C
  r i = fun_of_map_fun (ceta_map_of (compare_order_prod _C compare_order_nat) i)
          (fun fn -> (defaulta r, replicate (snd fn) (one r)));;

let rec showl_pvars (_A1, _A2) _C
  r x1 = match r, x1 with r, [] -> []
    | r, (x, c) :: vas ->
        (if eq _A1 c (one r) then id else showsl _A2 c) (showsl _C x "") ::
          showl_pvars (_A1, _A2) _C r vas;;

let rec showsl_lpoly (_A1, _A2) _C
  r (LPoly (c, cs)) =
    (match showl_pvars (_A1, _A2) _C r cs with [] -> showsl _A2 c
      | a :: lista ->
        comp (if eq _A1 c (zero r) then id
               else comp (showsl _A2 c) (showsl_literal " + "))
          (showsl_list_gen showsl_literal "" "" " + " "" (a :: lista)));;

let rec add_var _A _C
  r x a xa3 = match r, x, a, xa3 with r, x, a, [] -> [(x, a)]
    | r, x, a, (y, b) :: vas ->
        (if eq _C x y
          then (let s = add r a b in
                 (if eq _A s (zero r) then vas else (x, s) :: vas))
          else (y, b) :: add_var _A _C r x a vas);;

let rec sum_pvars _A _C
  r x1 vbs = match r, x1, vbs with r, [], vbs -> vbs
    | r, (x, a) :: vas, vbs ->
        (if eq _A a (zero r) then sum_pvars _A _C r vas vbs
          else sum_pvars _A _C r vas (add_var _A _C r x a vbs));;

let rec sum_lpolya _A _C
  r (LPoly (a, vas)) (LPoly (b, vbs)) =
    LPoly (add r a b, sum_pvars _A _C r vas vbs);;

let rec mul_pvars _A
  r a x2 = match r, a, x2 with r, a, [] -> []
    | r, a, (x, b) :: vas ->
        (let p = mult r a b in
         let res = mul_pvars _A r a vas in
          (if eq _A p (zero r) then res else (x, p) :: res));;

let rec mul_lpoly _A
  r a (LPoly (b, vas)) = LPoly (mult r a b, mul_pvars _A r a vas);;

let rec list_prod r x1 = match r, x1 with r, [] -> one r
                    | r, x :: xs -> mult r x (list_prod r xs);;

let rec wf_pvars (_A1, _A2, _A3, _A4)
  r vas =
    less_eq_set (_A1, _A2, _A3) (set (_A2, _A3, _A4) (map snd vas))
      (carrier r);;

let rec wf_lpoly (_A1, _A2, _A3, _A4)
  r (LPoly (a, vas)) =
    member (_A2, _A3) a (carrier r) && wf_pvars (_A1, _A2, _A3, _A4) r vas;;

let rec pleftI (_A1, _A2, _A3, _A4, _A5) _D
  r pi x2 = match r, pi, x2 with r, pi, Var x -> LPoly (zero r, [(x, one r)])
    | r, pi, Fun (f, ts) ->
        (let (c, asa) = pi (f, size_list ts) in
          (match
            sum_lpolya _A4 _D r (LPoly (c, []))
              (list_prod
                (Partial_object_ext
                  (collect cenum_l_poly (wf_lpoly (_A1, _A2, _A3, _A5) r),
                    Monoid_ext (sum_lpolya _A4 _D r, LPoly (zero r, []), ())))
                (map (fun at ->
                       mul_lpoly _A4 r (fst at)
                         (pleftI (_A1, _A2, _A3, _A4, _A5) _D r pi (snd at)))
                  (zip asa ts)))
            with LPoly (d, []) -> LPoly (maxb r (zero r) d, [])
            | LPoly (d, ab :: lista) -> LPoly (d, ab :: lista)));;

let rec create_lpoly_repr (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2, _B3)
  c i = (let pi = to_lpoly_inter _B1 c i in
          comp (comp (comp (showsl_literal "polynomial interpretation over ")
                       (description c))
                 (showsl_literal "\092"))
            (showsl_sep
              (fun (f, n) ->
                (let t =
                   Fun (f, map (fun a -> Var a)
                             (fresh_strings_list
                               [Chara (false, false, false, true, true, true,
true, false);
                                 Chara (true, true, true, true, true, false,
 true, false)]
                               one_nata [] n))
                   in
                  comp (comp (comp (showsl_literal "Pol(")
                               (showsl_terma _B3 (showl_list showl_char) t))
                         (showsl_literal ") = "))
                    (showsl_lpoly (_A4, _A6) (showl_list showl_char) c
                      (pleftI (_A1, _A2, _A3, _A4, _A5) (equal_list equal_char)
                        c pi t))))
              (showsl_literal "\092")
              (remdups (equal_prod _B2 equal_nat) (map fst i))));;

let rec checkmono
  (Partial_object_ext
    (carrier,
      Monoid_ext
        (mult, one,
          Ring_ext
            (zero, add,
              Ordered_semiring_ext
                (geq, gt, max,
                  Lpoly_order_semiring_ext
                    (plus_single_mono, default, arcpos, checkmono, bound,
                      check_complexity, description, more))))))
    = checkmono;;

let rec check_poly_mono _A _C
  r = (fun xs ->
        catch_error
          (forallM
            (fun (a, b) ->
              (let (f, n) = a in
                (fun (c, cs) ->
                  catch_error
                    (bindb
                      (check (geq r c (zero r))
                        (comp (comp (comp (comp
    (showsl_literal "constant part ") (showsl _A c))
                                      (showsl_literal " must be at least "))
                                (showsl _A (zero r)))
                          (showsl_literal "\092")))
                      (fun _ ->
                        bindb (check (less_eq_nat n (size_list cs))
                                (showsl_literal
                                  "the last argument is ignored\092"))
                          (fun _ ->
                            catch_error
                              (forallM
                                (fun d ->
                                  check (checkmono r d)
                                    (comp (comp (showsl_literal "coefficient ")
    (showsl _A d))
                                      (showsl_literal " is not allowed\092")))
                                cs)
                              (fun x -> Inl (snd x)))))
                    (fun x ->
                      Inl (comp (comp (comp
(comp (comp (showsl_literal
              "problem with monotonicity due to interpretation of ")
        (showsl _C f))
  (showsl_literal "/"))
(showsl_nat n))
                                  (showsl_literal "\092"))
                            x))))
                b)
            xs)
          (fun x -> Inl (snd x)));;

let rec create_mono_af _A _C
  r i = (if plus_single_mono r
          then fun_of_map_funa
                 (ceta_map_of (compare_order_prod _C compare_order_nat) i)
                 (fun (_, n) ->
                   set (ceq_nat, ccompare_nat, set_impl_nat) (upt zero_nata n))
                 (fun (c, coeffs) ->
                   set (ceq_nat, ccompare_nat, set_impl_nat)
                     (if geq r c (zero r)
                       then maps (fun (ca, ia) ->
                                   (if eq _A ca (one r) || checkmono r ca
                                     then [ia] else []))
                              (zip coeffs (upt zero_nata (size_list coeffs)))
                       else []))
          else empty_af);;

let rec lookup_rest _A
  x xa1 = match x, xa1 with x, [] -> None
    | x, (y, c) :: ycs ->
        (if eq _A x y then Some (c, ycs)
          else (match lookup_rest _A x ycs with None -> None
                 | Some (d, yccs) -> Some (d, (y, c) :: yccs)));;

let rec check_pvars _A _C
  r rel vas x3 = match r, rel, vas, x3 with
    r, rel, vas, [] ->
      catch_error
        (forallM (fun va -> check (rel (snd va) (zero r)) (fst va)) vas)
        (fun x -> Inl (snd x))
    | r, rel, vas, (x, b) :: vbs ->
        (let a =
           (match lookup_rest _C x vas with None -> (zero r, vas)
             | Some a -> (let (aa, ba) = a in (aa, ba)))
           in
         let (aa, vasa) = a in
          bindb (check (rel aa b) x)
            (fun _ -> check_pvars _A _C r rel vasa vbs));;

let rec check_lpoly_ns (_A1, _A2) (_C1, _C2)
  r (LPoly (a, vas)) (LPoly (b, vbs)) =
    catch_error
      (bindb
        (check (geq r a b)
          (showsl_literal "problem when comparing constant parts"))
        (fun _ ->
          catch_error (check_pvars _A2 _C1 r (geq r) vas vbs)
            (fun x ->
              Inl (comp (showsl_literal
                          "problem when comparing coefficients of variable ")
                    (showsl _C2 x)))))
      (fun x ->
        Inl (comp (comp (comp (comp (comp (showsl_literal
    "problem when comparing ")
                                      (showsl_lpoly (_A1, _A2) _C2 r
(LPoly (a, vas))))
                                (showsl_literal " >= "))
                          (showsl_lpoly (_A1, _A2) _C2 r (LPoly (b, vbs))))
                    (showsl_literal "\092"))
              x));;

let rec prightI (_A1, _A2, _A3, _A4, _A5) _D
  r pi x2 = match r, pi, x2 with r, pi, Var x -> LPoly (zero r, [(x, one r)])
    | r, pi, Fun (f, ts) ->
        (let (c, asa) = pi (f, size_list ts) in
         let LPoly (d, a) =
           sum_lpolya _A4 _D r (LPoly (c, []))
             (list_prod
               (Partial_object_ext
                 (collect cenum_l_poly (wf_lpoly (_A1, _A2, _A3, _A5) r),
                   Monoid_ext (sum_lpolya _A4 _D r, LPoly (zero r, []), ())))
               (map (fun at ->
                      mul_lpoly _A4 r (fst at)
                        (prightI (_A1, _A2, _A3, _A4, _A5) _D r pi (snd at)))
                 (zip asa ts)))
           in
          LPoly (maxb r (zero r) d, a));;

let rec check_polo_ns (_A1, _A2, _A3, _A4, _A5, _A6) _C (_D1, _D2)
  r pi (s, t) =
    (let left = pleftI (_A1, _A2, _A3, _A4, _A5) _D1 r pi s in
     let right = prightI (_A1, _A2, _A3, _A4, _A5) _D1 r pi t in
      catch_error (check_lpoly_ns (_A4, _A6) (_D1, _D2) r left right)
        (fun x ->
          Inl (comp (comp (comp (comp (comp (showsl_literal "could not ensure ")
(showsl_terma _C _D2 s))
                                  (showsl_literal " >= "))
                            (showsl_terma _C _D2 t))
                      (showsl_literal "\092"))
                x)));;

let rec gt
  (Partial_object_ext
    (carrier,
      Monoid_ext
        (mult, one,
          Ring_ext (zero, add, Ordered_semiring_ext (geq, gt, max, more)))))
    = gt;;

let rec check_lpoly_s (_A1, _A2) (_C1, _C2)
  r (LPoly (a, vas)) (LPoly (b, vbs)) =
    catch_error
      (bindb
        (check (gt r a b)
          (showsl_literal "problem when comparing constant part"))
        (fun _ ->
          catch_error
            (check_pvars _A2 _C1 r (if plus_single_mono r then geq r else gt r)
              vas vbs)
            (fun x ->
              Inl (comp (showsl_literal
                          "problem when comparing coefficients of variable ")
                    (showsl _C2 x)))))
      (fun x ->
        Inl (comp (comp (comp (comp (comp (showsl_literal
    "problem when comparing ")
                                      (showsl_lpoly (_A1, _A2) _C2 r
(LPoly (a, vas))))
                                (showsl_literal " > "))
                          (showsl_lpoly (_A1, _A2) _C2 r (LPoly (b, vbs))))
                    (showsl_literal "\092"))
              x));;

let rec check_polo_s (_A1, _A2, _A3, _A4, _A5, _A6) _C (_D1, _D2)
  r pi (s, t) =
    (let left = pleftI (_A1, _A2, _A3, _A4, _A5) _D1 r pi s in
     let right = prightI (_A1, _A2, _A3, _A4, _A5) _D1 r pi t in
      catch_error (check_lpoly_s (_A4, _A6) (_D1, _D2) r left right)
        (fun x ->
          Inl (comp (comp (comp (comp (comp (showsl_literal "could not ensure ")
(showsl_terma _C _D2 s))
                                  (showsl_literal " > "))
                            (showsl_terma _C _D2 t))
                      (showsl_literal "\092"))
                x)));;

let rec create_af _A _C
  r i = fun_of_map_funa
          (ceta_map_of (compare_order_prod _C compare_order_nat) i)
          (fun (_, n) ->
            set (ceq_nat, ccompare_nat, set_impl_nat) (upt zero_nata n))
          (fun (_, coeffs) ->
            set (ceq_nat, ccompare_nat, set_impl_nat)
              (maps (fun (c, ia) ->
                      (if not (eq _A c (zero r)) then [ia] else []))
                (zip coeffs (upt zero_nata (size_list coeffs)))));;

let rec create_poly_redtriple (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2, _B3)
  (_C1, _C2)
  c cI i =
    (let pi = to_lpoly_inter _B1 c i in
     let ns = check_polo_ns (_A1, _A2, _A3, _A4, _A5, _A6) _B3 (_C1, _C2) c pi
       in
      Redtriple_ext
        (bindb cI (fun _ -> check_lpoly_coeffs (_A2, _A3, _A6) _B3 c i),
          check_polo_s (_A1, _A2, _A3, _A4, _A5, _A6) _B3 (_C1, _C2) c pi, ns,
          ns, create_af _A4 _B1 c i, create_mono_af _A4 _B1 c i,
          (fun s_ns_nst ->
            (if plus_single_mono c then check_poly_mono _A6 _B3 c i
              else check_poly_mono_npsm (_A4, _A6) (_B2, _B3) c
                     (funas_trs_list s_ns_nst) i)),
          create_lpoly_repr (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2, _B3) c i,
          Some (map fst i),
          (if plus_single_mono c then convert_lpoly_complexity _A4 c pi
            else no_complexity_check),
          ()));;

let rec weak_gt_arctic_delta _A
  uu x1 = match uu, x1 with uu, MinInfty_delta -> true
    | MinInfty_delta, Num_arc_delta uv -> false
    | Num_arc_delta x, Num_arc_delta y ->
        less _A.archimedean_field_floor_ceiling.linordered_field_archimedean_field.linordered_idom_linordered_field.linordered_ring_strict_linordered_idom.linordered_ring_linordered_ring_strict.abs_if_linordered_ring.ord_abs_if
          y x;;

let rec check_poly_strict_mono_discrete (_A1, _A2) (_B1, _B2)
  gt p v =
    check_poly_gt
      _A2.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1
      (_B1, _B2) gt
      (poly_subst _B2 (_B1, _B2) (_A1, _A2.comm_semiring_1_poly_carrier)
        (fun w ->
          poly_of (_B1, _B2) (_A1, _A2.comm_semiring_1_poly_carrier)
            (if eq _B1 w v
              then PSum [PNum (onea _A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral);
                          PVar v]
              else PVar w))
        p)
      p;;

let rec univariate_power_list _A
  x xa1 = match x, xa1 with x, [(y, n)] -> (if eq _A x y then Some n else None)
    | uu, [] -> None
    | uu, v :: vb :: vc -> None;;

let rec univariate_power (_A1, _A2)
  x xc = univariate_power_list _A1 x (rep_monom _A2 xc);;

let rec check_monom_strict_mono (_A1, _A2)
  pm m v =
    (match univariate_power (_A1, _A2) v m with None -> false
      | Some p -> pm || equal_nata p one_nata);;

let rec check_poly_strict_mono (_A1, _A2) _B
  pm p v =
    list_ex
      (fun (m, c) ->
        less_eq
          _B.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order
          (onea _B.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral)
          c &&
          check_monom_strict_mono (_A1, _A2) pm m v)
      p;;

let rec check_poly_strict_mono_smart (_A1, _A2) (_B1, _B2)
  discrete pm gt p v =
    (if discrete
      then check_poly_strict_mono_discrete (_A1, _A2) (_B1, _B2) gt p v
      else check_poly_strict_mono (_B1, _B2) _A2 pm p v);;

let rec check_poly_weak_mono_all _B
  p = list_all
        (fun (_, a) ->
          less_eq
            _B.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order
            (zerob
              _B.semiring_0_ordered_semiring_0a.mult_zero_semiring_0.zero_mult_zero)
            a)
        p;;

let rec check_poly_weak_mono_and_pos (_A1, _A2) (_B1, _B2)
  discrete p =
    (if discrete
      then list_all (check_poly_weak_mono_discrete (_A1, _A2) (_B1, _B2) p)
             (poly_vars_list (_A1, _A2) p) &&
             less_eq
               _B2.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order
               (zerob
                 _B2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
               (eval_poly _A2 _B2.comm_semiring_1_poly_carrier
                 (fun _ ->
                   zerob _B2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
                 p)
      else check_poly_weak_mono_all
             _B2.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1
             p);;

let rec poly_inter_to_mono_af (_A1, _A2) _B
  discrete power_mono gt i =
    fun_of_map_fun
      (ceta_map_of (compare_order_prod _B compare_order_nat)
        (map (fun (a, b) ->
               (let (f, n) = a in
                 (fun e ->
                   ((f, n),
                     set (ceq_nat, ccompare_nat, set_impl_nat)
                       (filtera
                         (fun ia ->
                           check_poly_weak_mono_and_pos
                             (equal_nat, linorder_nat) (_A1, _A2) discrete e &&
                             check_poly_strict_mono_smart (_A1, _A2)
                               (equal_nat, linorder_nat) discrete power_mono gt
                               e ia)
                         (upt zero_nata n)))))
                 b)
          i))
      (fun fn ->
        set (ceq_nat, ccompare_nat, set_impl_nat) (upt zero_nata (snd fn)));;

let rec check_poly_inter_list _A (_B1, _B2)
  discrete i =
    bindb (check (distinct (equal_prod _A equal_nat) (map fst i))
            (Inl (showsl_literal "some symbol has two interpretations")))
      (fun _ ->
        catch_error
          (catch_error
            (forallM
              (fun x ->
                (if (let (_, a) = x in
                      check_poly_weak_mono_and_pos (equal_nat, linorder_nat)
                        (_B1, _B2) discrete a)
                  then Inr () else Inl x))
              i)
            (fun x -> Inl (snd x)))
          (fun x ->
            Inl (let (a, b) = x in
                  (let (f, _) = a in (fun p -> Inr (f, p))) b)));;

let rec less_eq_complexity_class x y = less_eq_nat (degree x) (degree y);;

let rec strongly_linear _A
  x p v =
    (if equal_nata x zero_nata then check_ge_v _A v p
      else (let a =
              poly_split (equal_nat, linorder_nat)
                _A.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
                (var_monom linorder_nat (minus_nata x one_nata)) p
              in
            let (aa, pa) = a in
             less_eq
               _A.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order
               aa (onea _A.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral) &&
               strongly_linear _A (minus_nata x one_nata) pa v));;

let rec sl_complexity_sig_check _B
  i v f =
    catch_error
      (forallM
        (fun (fa, n) -> check (strongly_linear _B n (i (fa, n)) v) (fa, n)) f)
      (fun x -> Inl (snd x));;

let rec sl_complexity_check _A _B
  v i f =
    (let w =
       max_v (_A.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero,
               _A.ordered_semiring_1_poly_carrier.ordered_semiring_0_ordered_semiring_1.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order)
         v i f
       in
      catch_error (sl_complexity_sig_check _A i w f)
        (fun x ->
          Inl (let (fa, _) = x in
                comp (comp (showsl_literal "symbol ") (showsl _B fa))
                  (showsl_literal
                    " does not possess a strongly linear interpretation"))));;

let rec nl_complexity_check _A _B
  i x1 cc = match i, x1, cc with
    i, Derivational_Complexity f, cc ->
      bindb (sl_complexity_check _B _A
              (zerob
                _B.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
              i f)
        (fun _ ->
          check (less_eq_complexity_class (Comp_Poly one_nata) cc)
            (showsl_literal
              "cannot deduce constant complexity for derivational complexity"))
    | i, Runtime_Complexity (c, d), Comp_Poly deg ->
        bindb (sl_complexity_check _B _A
                (onea _B.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral)
                i c)
          (fun _ ->
            catch_error
              (forallM
                (fun f ->
                  check (less_eq_nat (poly_degree linorder_nat (i f)) deg)
                    (comp (comp (showsl_literal "degree of interpretation for ")
                            (showsl_prod _A showl_nat f))
                      (showsl_literal " exceeds bound ")))
                d)
              (fun x -> Inl (snd x)));;

let rec poly_inter_to_af _A
  i = fun_of_map_fun
        (ceta_map_of (compare_order_prod _A compare_order_nat)
          (map (fun (fn, e) ->
                 (fn, poly_vars
                        (ceq_nat, ccompare_nat, linorder_nat, set_impl_nat) e))
            i))
        (fun fn ->
          set (ceq_nat, ccompare_nat, set_impl_nat) (upt zero_nata (snd fn)));;

let rec create_nlpoly_redtriple (_A1, _A2, _A3) (_B1, _B2, _B3) (_C1, _C2, _C3)
  cI def gt power_mono discrete i =
    (let j = poly_inter_list_to_inter _A2 _B1 def i in
     let x =
       poly_subst linorder_nat
         ((equal_list equal_char), (linorder_list (equal_char, linorder_char)))
         (_A1, _A2.comm_semiring_1_poly_carrier)
         (fun n ->
           poly_of
             ((equal_list equal_char),
               (linorder_list (equal_char, linorder_char)))
             (_A1, _A2.comm_semiring_1_poly_carrier)
             (PVar ([Chara (false, false, false, true, true, true, true, false);
                      Chara (true, true, true, true, true, false, true,
                              false)] @
                     shows_prec_nat zero_nata n [])))
       in
      Redtriple_ext
        (bindb cI
           (fun _ ->
             catch_error (check_poly_inter_list _B2 (_A1, _A2) discrete i)
               (fun xa ->
                 Inl (match xa with Inl a -> id a
                       | Inr (f, p) ->
                         comp (comp (comp (comp
    (showsl_literal "interpretation ")
    (showsl_poly
      ((equal_list equal_char), (linorder_list (equal_char, linorder_char)),
        (showl_list showl_char))
      (_A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral,
        _A1, _A3)
      (x p)))
                                      (showsl_literal " of "))
                                (showsl _B3 f))
                           (showsl_literal " invalid ")))),
          check_s (_A1, _A2, _A3) _B3 (_C1, _C2, _C3) gt j,
          check_ns _B3 (_A1, _A2, _A3) (_C1, _C2, _C3) j,
          check_ns _B3 (_A1, _A2, _A3) (_C1, _C2, _C3) j,
          poly_inter_to_af _B1 i,
          poly_inter_to_mono_af (_A1, _A2) _B1 discrete power_mono gt i,
          (fun _ ->
            catch_error
              (catch_error
                (forallM
                  (fun xa ->
                    (if (let (a, b) = xa in
                          (let (_, n) = a in
                            (fun p ->
                              all_interval_nat
                                (check_poly_strict_mono_smart (_A1, _A2)
                                  (equal_nat, linorder_nat) discrete power_mono
                                  gt p)
                                zero_nata n))
                            b)
                      then Inr () else Inl xa))
                  i)
                (fun xa -> Inl (snd xa)))
              (fun xa ->
                Inl (let (a, b) = xa in
                      (let (f, _) = a in
                        (fun p ->
                          comp (comp (comp (showsl_literal
     "could not ensure monotonicty of ")
                                       (showsl_poly
 ((equal_list equal_char), (linorder_list (equal_char, linorder_char)),
   (showl_list showl_char))
 (_A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral,
   _A1, _A3)
 (x p)))
                                 (showsl_literal " as interpretation of "))
                            (showsl _B3 f)))
                        b))),
          comp (showsl_literal "polynomial interpretation\092")
            (showsl_sep
              (fun (a, b) ->
                (let (f, n) = a in
                  (fun p ->
                    comp (comp (comp (comp (comp (showsl_literal "Pol(")
     (showsl _B3 f))
                                       (showsl_literal "/"))
                                 (showsl_nat n))
                           (showsl_literal ") = "))
                      (showsl_poly
                        ((equal_list equal_char),
                          (linorder_list (equal_char, linorder_char)),
                          (showl_list showl_char))
                        (_A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral,
                          _A1, _A3)
                        (x p))))
                  b)
              (showsl_literal "\092") i),
          Some (map fst i), nl_complexity_check _B3 _A2 j, ()));;

let rec pos_arctic_delta _A
  = function MinInfty_delta -> false
    | Num_arc_delta n ->
        less_eq
          _A.archimedean_field_floor_ceiling.linordered_field_archimedean_field.linordered_idom_linordered_field.linordered_ring_strict_linordered_idom.linordered_ring_linordered_ring_strict.abs_if_linordered_ring.ord_abs_if
          (zerob
            _A.archimedean_field_floor_ceiling.linordered_field_archimedean_field.linordered_idom_linordered_field.linordered_ring_strict_linordered_idom.linordered_ring_linordered_ring_strict.abs_if_linordered_ring.zero_abs_if)
          n;;

let rec class_ordered_semiring (_A1, _A2, _A3, _A4)
  a gt b =
    class_semiring
      (_A1, _A2,
        _A3.semiring_1_ordered_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral,
        _A3.semiring_1_ordered_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.semigroup_add_numeral.plus_semigroup_add,
        _A3.semiring_1_ordered_semiring_1.semiring_numeral_semiring_1.monoid_mult_semiring_numeral.power_monoid_mult.times_power,
        _A3.semiring_1_ordered_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero,
        _A4)
      a (Ordered_semiring_ext
          ((fun x y ->
             less_eq
               _A3.ordered_semiring_0_ordered_semiring_1.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order
               y x),
            gt, max _A3.ordered_semiring_0_ordered_semiring_1.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order,
            b));;

let rec class_arc_complexity
  a deg = Inl (showsl_literal "complexity for arctic semirings not supported");;

let rec class_arc_lpoly_order (_A1, _A2, _A3, _A4)
  def apos gtt =
    class_ordered_semiring (_A1, _A2, _A3, _A4) Type gtt
      (Lpoly_order_semiring_ext
        (false, def, apos, (fun _ -> false), (fun _ -> zero_nata),
          class_arc_complexity,
          showsl_lit "polynomial interpretation over arctic semiring", ()));;

let rec prec_weight_ac_repr_to_prec_weight_funs (_A1, _A2, _A3, _A4)
  prw_w0 =
    (let (prw, w0) = prw_w0 in
     let prwm = ceta_map_of (compare_order_prod _A3 compare_order_nat) prw in
     let w_fun = fun_of_map_funa prwm (fun _ -> suc w0) (comp fst snd) in
     let p_fun = prec_exta prwm in
     let acset =
       set (_A1, _A2, _A4)
         (map_filter
           (fun x ->
             (if (let (a, b) = x in
                   (let (_, _) = a in (fun (_, (_, ac)) -> ac)) b)
               then Some (comp fst fst x) else None))
           prw)
       in
      (p_fun, (w_fun, (w0, acset))));;

let rec prec_weight_ac_repr_to_prec_weight (_A1, _A2, _A3, _A4, _A5, _A6)
  prw_w0 =
    (let (p_fun, (w_fun, (_, acset))) =
       prec_weight_ac_repr_to_prec_weight_funs (_A1, _A2, _A3, _A5) prw_w0 in
     let (prw, w0) = prw_w0 in
     let fs = map fst prw in
     let cw_okay =
       catch_error
         (forallM
           (fun fn ->
             check (if equal_nata (snd fn) zero_nata
                     then less_eq_nat w0 (w_fun fn) else true)
               (comp (comp (showsl_lit "weight of constant ")
                       (showsl _A6 (fst fn)))
                 (showsl_lit " must be at least w0")))
           fs)
         (fun x -> Inl (snd x))
       in
     let adm =
       catch_error
         (forallM
           (fun fn ->
             check (if equal_nata (snd fn) one_nata
                     then (if equal_nata (w_fun fn) zero_nata
                            then list_all
                                   (fun x ->
                                     p_fun fn x ||
                                       equal_proda _A4 equal_nat x fn)
                                   fs
                            else true)
                     else true)
               (comp (comp (showsl_lit "unary symbol ") (showsl _A6 (fst fn)))
                 (showsl_lit
                   " with weight 0 does not have maximal precedence")))
           (map fst prw))
         (fun x -> Inl (snd x))
       in
     let irr =
       catch_error
         (forallM
           (fun fn ->
             check (not (p_fun fn fn))
               (comp (comp (showsl_lit "function symbol ")
                       (showsl _A6 (fst fn)))
                 (showsl_lit " violates irreflexibity")))
           fs)
         (fun x -> Inl (snd x))
       in
     let ok =
       bindb (check (less_nat zero_nata w0)
               (showsl_lit "w0 must be larger than 0"))
         (fun _ -> bindb adm (fun _ -> bindb cw_okay (fun _ -> irr)))
       in
      (ok, (p_fun, (w_fun, (w0, acset)))));;

let rec showsl_ackbo_repr _A
  (prs, w0) =
    comp (comp (comp (comp (comp (comp (comp
 (comp (comp (comp (comp (comp (showsl_lit
                                 "ACKBO with the following precedence and weight function\092")
                           (foldr
                             (fun (a, b) ->
                               (let (f, n) = a in
                                 (fun (pr, (_, _)) ->
                                   comp (comp
  (comp (comp (comp (comp (showsl_lit "precedence(") (showsl _A f))
                (showsl_lit "["))
          (showsl_nat n))
    (showsl_lit "]) = "))
  (showsl_nat pr))
                                     (showsl_literal "\092")))
                                 b)
                             prs))
                     (showsl_lit
                       "\092precedence(_) = 0\092and the following weight\092"))
               (foldr
                 (fun (a, b) ->
                   (let (f, n) = a in
                     (fun (_, (w, _)) ->
                       comp (comp (comp (comp
  (comp (comp (showsl_lit "weight(") (showsl _A f)) (showsl_lit "["))
  (showsl_nat n))
                                    (showsl_lit "]) = "))
                              (showsl_nat w))
                         (showsl_literal "\092")))
                     b)
                 prs))
         (showsl_lit "\092weight(_) = "))
   (showsl_nat (suc w0)))
 (showsl_lit "\092w0 = "))
                                   (showsl_nat w0))
                             (showsl_lit
                               "\092and the following are the ac functions\092"))
                       (foldr
                         (fun (a, b) ->
                           (let (f, n) = a in
                             (fun (_, (_, _)) ->
                               comp (comp (comp
    (comp (showsl_lit "ac(") (showsl _A f)) (showsl_lit "["))
                                      (showsl_nat n))
                                 (showsl_lit "]) =\092")))
                             b)
                         (filtera
                           (fun (a, b) ->
                             (let (_, _) = a in (fun (_, (_, ac)) -> ac)) b)
                           prs)))
                 (showsl_lit
                   "\092and the following are the non ac functions\092"))
           (foldr
             (fun (a, b) ->
               (let (f, n) = a in
                 (fun (_, (_, _)) ->
                   comp (comp (comp (comp (showsl_lit "not ac(") (showsl _A f))
                                (showsl_lit "["))
                          (showsl_nat n))
                     (showsl_lit "]) =\092")))
                 b)
             (filtera
               (fun (a, b) ->
                 (let (_, _) = a in
                   (fun (_, aa) -> (let (_, ab) = aa in not ab)))
                   b)
               prs)))
      (showsl_literal "\092");;

let rec create_ACKBO_redtriple _A (_B1, _B2, _B3, _B4, _B5, _B6) (_C1, _C2)
  f_to_g pr =
    (let (ch, (p, (w, (w0, ac)))) =
       prec_weight_ac_repr_to_prec_weight (_B1, _B2, _B3, _B4, _B5, _B6)
         (f_to_g pr)
       in
     let ns = ackbo_nstrict (_B1, _B2, _B4, _B6) (_C1, _C2) p w w0 ac in
     let s = ackbo_strict (_B1, _B2, _B4, _B6) (_C1, _C2) p w w0 ac in
      Redtriple_ext
        (ch, s, ns, ns, full_af, full_af, (fun _ -> Inr ()),
          showsl_ackbo_repr _A pr, Some [], no_complexity_check, ()));;

let rec mat_max _A
  a b = mat (dim_row a) (dim_col a)
          (fun ij -> max _A (index_mat a ij) (index_mat b ij));;

let rec mat_ge _A
  a b = all_interval_nat
          (fun i ->
            all_interval_nat
              (fun j -> less_eq _A (index_mat b (i, j)) (index_mat a (i, j)))
              zero_nata (dim_col a))
          zero_nata (dim_row a);;

let rec mat_comp_all
  r a b =
    all_interval_nat
      (fun i ->
        all_interval_nat (fun j -> r (index_mat a (i, j)) (index_mat b (i, j)))
          zero_nata (dim_col a))
      zero_nata (dim_row a);;

let rec mat_both_ordered_semiring _A
  n gt b =
    ring_mat _A.semiring_1_ordered_semiring_1 Type n
      (Ordered_semiring_ext
        (mat_ge
           _A.ordered_semiring_0_ordered_semiring_1.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order,
          mat_comp_all gt,
          mat_max
            _A.ordered_semiring_0_ordered_semiring_1.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order,
          b));;

let rec mat_default _A
  d n = mat n n (fun (i, j) -> (if equal_nata i j then d else zerob _A));;

let rec mat_arc_complexity
  m deg = Inl (showsl_lit "complexity for arctic matrices not supported");;

let rec mat_arc_posI ap a = ap (index_mat a (zero_nata, zero_nata));;

let rec mat_arc_lpoly_order _A
  n def apos gtt =
    mat_both_ordered_semiring _A n gtt
      (Lpoly_order_semiring_ext
        (false,
          mat_default
            _A.semiring_1_ordered_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
            def n,
          mat_arc_posI apos, (fun _ -> false), (fun _ -> zero_nata),
          mat_arc_complexity, showsl_lit "arctic matrix interpretation", ()));;

let rec check_arc_dimension (_A1, _A2)
  n = check (less _A2 (zerob _A1) n)
        (showsl_lit "dimension must be at least 1");;

let rec class_complexity _A
  a deg =
    check (less_eq
            _A.ordered_semiring_0_ordered_semiring_1.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order
            a (onea _A.semiring_1_ordered_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral))
      (showsl_literal "value is larger than 1");;

let rec class_lpoly_order (_A1, _A2, _A3, _A4)
  def mon gtt =
    class_ordered_semiring (_A1, _A2, _A3, _A4) Type gtt
      (Lpoly_order_semiring_ext
        (true, def, (fun _ -> true), mon, (fun _ -> zero_nata),
          class_complexity _A3, showsl_literal "polynomial interpretation",
          ()));;

let rec prec_weight_repr_to_prec_weight_funs (_A1, _A2)
  prw_w0 =
    (let (prw, w0) = prw_w0 in
     let prwm = ceta_map_of (compare_order_prod _A1 compare_order_nat) prw in
     let w_fun = fun_of_map_funa prwm (fun _ -> suc w0) (comp fst snd) in
     let p_fun = prec_ext (equal_prod _A2 equal_nat) prwm in
     let scf_fun = fun_of_map_funa prwm (fun _ -> None) (comp snd snd) in
     let fs = map fst prw in
     let cs =
       filtera
         (fun fn -> equal_nata (snd fn) zero_nata && equal_nata (w_fun fn) w0)
         fs
       in
     let lcs =
       map_filter
         (fun x ->
           (if list_all (fun c -> snd (p_fun c x)) cs then Some (fst x)
             else None))
         cs
       in
      (p_fun, (w_fun, (w0, (lcs, scf_fun)))));;

let rec scf_repr_to_scf
  scf fn i = (match scf fn with None -> one_nata | Some xs -> nth xs i);;

let rec check_scf_entry _A
  fn x1 = match fn, x1 with fn, None -> Inr ()
    | (f, n), Some es ->
        catch_error
          (bindb
            (check (equal_nata (size_list es) n)
              (comp (showsl_literal "nr of entries should be ") (showsl_nat n)))
            (fun _ ->
              check (list_all (less_nat zero_nata) es)
                (showsl_literal "all entries must be non-zero")))
          (fun x ->
            Inl (comp (comp (comp (comp (showsl_literal
  "problem with subterm coefficients for ")
                                    (showsl_prod _A showl_nat (f, n)))
                              (showsl_literal ": "))
                        x)
                  (showsl_literal "\092")));;

let rec prec_weight_repr_to_prec_weight (_A1, _A2, _A3)
  prw_w0 =
    (let (p_fun, (w_fun, (_, (lcs, scf_fun)))) =
       prec_weight_repr_to_prec_weight_funs (_A1, _A2) prw_w0 in
     let (prw, w0) = prw_w0 in
     let fs = map fst prw in
     let cw_okay =
       catch_error
         (forallM
           (fun fn ->
             check (if equal_nata (snd fn) zero_nata
                     then less_eq_nat w0 (w_fun fn) else true)
               (comp (comp (showsl_literal "weight of constant ")
                       (showsl _A3 (fst fn)))
                 (showsl_literal " must be at least w0")))
           (map fst prw))
         (fun x -> Inl (snd x))
       in
     let adm =
       catch_error
         (forallM
           (fun fn ->
             check (if equal_nata (snd fn) one_nata
                     then (if equal_nata (w_fun fn) zero_nata
                            then list_all (comp snd (p_fun fn)) fs else true)
                     else true)
               (comp (comp (showsl_literal "unary symbol ")
                       (showsl _A3 (fst fn)))
                 (showsl_literal
                   " with weight 0 does not have maximal precedence")))
           (map fst prw))
         (fun x -> Inl (snd x))
       in
     let scf_ok =
       catch_error
         (forallM (fun fn -> check_scf_entry _A3 fn (scf_fun fn)) (map fst prw))
         (fun x -> Inl (snd x))
       in
     let ok =
       bindb (check (less_nat zero_nata w0)
               (showsl_literal "w0 must be larger than 0"))
         (fun _ -> bindb adm (fun _ -> bindb cw_okay (fun _ -> scf_ok)))
       in
      (ok, (p_fun, (w_fun, (w0, (lcs, scf_repr_to_scf scf_fun))))));;

let rec create_KBO_redtriple _A (_B1, _B2, _B3) (_C1, _C2)
  f_to_g pr =
    (let (ch, (p, (w, (w0, (lcs, scf))))) =
       prec_weight_repr_to_prec_weight (_B1, _B2, _B3) (f_to_g pr) in
     let ns = kbo_nstrict _B3 (_C1, _C2) p w w0 (membera _B2 lcs) scf in
     let s = kbo_strict _B3 (_C1, _C2) p w w0 (membera _B2 lcs) scf in
      Redtriple_ext
        (ch, s, ns, ns, full_af, full_af, (fun _ -> Inr ()),
          shows_kbo_repr _A pr, Some [], no_complexity_check, ()));;

let rec check_dimensions
  n sd c =
    bindb c
      (fun _ ->
        check (less_eq_nat sd n && less_nat zero_nata sd)
          (showsl_literal
            "strict dimension must be at least 1 and less than total dimension"));;

let rec fpair_f (FPair (x1, x2)) = x1;;

let rec check_mono_afs (_A1, _A2)
  pi = check (mono_afs (_A1, _A2) pi)
         (showsl_literal "argument filter is not monotone");;

let rec showsl_afs _A
  af = foldr (fun (a, b) ->
               (let (f, n) = a in
                 (fun e ->
                   comp (comp (comp (comp (comp
    (comp (showsl_literal "pi(") (showsl _A f)) (showsl_literal "/"))
                                      (showsl_nat n))
                                (showsl_literal ") = "))
                          (match e with Collapse i -> showsl_nat (suc i)
                            | AFList ids -> showsl_list_nat (map suc ids)))
                     (showsl_literal "\092")))
                 b)
         af;;

let rec afs_to_af _A
  pi fn =
    (match afsa pi fn
      with Collapse j ->
        inserta (ceq_nat, ccompare_nat) j
          (set_empty (ceq_nat, ccompare_nat) (of_phantom set_impl_nata))
      | AFList a -> set (ceq_nat, ccompare_nat, set_impl_nat) a);;

let rec afs_check _A _B
  r pi g lr =
    (let pl = afs_term pi (fst lr) in
     let pr = afs_term pi (snd lr) in
      catch_error (g (pl, pr))
        (fun x ->
          Inl (comp (comp (comp (comp (comp
(comp (comp (comp (comp (comp (comp (comp (comp
    (comp (comp (showsl_literal "could not orient ")
            (showsl_terma _A _B (fst lr)))
      (showsl_literal " "))
    r)
                                      (showsl_literal " "))
                                (showsl_terma _A _B (snd lr)))
                          (showsl_literal "\092pi( "))
                    (showsl_terma _A _B (fst lr)))
              (showsl_literal " ) = "))
        (showsl_terma (showl_filtered _A) _B pl))
  (showsl_literal "\092pi( "))
(showsl_terma _A _B (snd lr)))
                                  (showsl_literal " ) = "))
                            (showsl_terma (showl_filtered _A) _B pr))
                      (showsl_literal "\092"))
                x)));;

let rec af_redtriple (_A1, _A2, _A3, _A4, _A5) _B
  pi rp =
    (let afso = afs_of (_A1, _A2, _A3, _A4) pi in
     let afs = the afso in
     let af = afs_to_af _A3 afs in
      Redtriple_ext
        (bindb (check (not (is_none afso))
                 (showsl_literal "invalid positions in argument filter"))
           (fun _ -> valid rp),
          afs_check _A5 _B (showsl_literal ">") afs (s rp),
          afs_check _A5 _B (showsl_literal ">=") afs (ns rp),
          afs_check _A5 _B (showsl_literal ">=") afs (nst rp), af, empty_af,
          (fun cs ->
            bindb (check_mono_afs (_A1, _A2) afs)
              (fun _ -> mono rp (map (afs_rule afs) cs))),
          comp (comp (comp (showsl_literal "Argument Filter:\092")
                       (showsl_afs _A5 pi))
                 (showsl_literal "\092"))
            (desc rp),
          map_option
            (fun fs -> map fst pi @ map (fun (f, a) -> (fpair_f f, a)) fs)
            (not_ws_ns rp),
          no_complexity_check, ()));;

let rec pos_arctic = function MinInfty -> false
                     | Num_arc n -> less_eq_int zero_inta n;;

let rec delta_mono _A
  x = less_eq
        _A.archimedean_field_floor_ceiling.linordered_field_archimedean_field.linordered_idom_linordered_field.linordered_ring_strict_linordered_idom.linordered_ring_linordered_ring_strict.abs_if_linordered_ring.ord_abs_if
        (onea _A.archimedean_field_floor_ceiling.linordered_field_archimedean_field.field_char_0_linordered_field.ring_char_0_field_char_0.ring_1_ring_char_0.neg_numeral_ring_1.numeral_neg_numeral.one_numeral)
        x;;

let rec mat_gt _A
  gt sd a b =
    mat_ge _A a b &&
      not (all_interval_nat
            (comp not
              (fun i ->
                not (all_interval_nat
                      (comp not
                        (fun j -> gt (index_mat a (i, j)) (index_mat b (i, j))))
                      zero_nata sd)))
            zero_nata sd);;

let rec mat_ordered_semiring _A
  n sd gt b =
    ring_mat _A.semiring_1_ordered_semiring_1 Type n
      (Ordered_semiring_ext
        (mat_ge
           _A.ordered_semiring_0_ordered_semiring_1.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order,
          mat_gt
            _A.ordered_semiring_0_ordered_semiring_1.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order
            gt sd,
          mat_max
            _A.ordered_semiring_0_ordered_semiring_1.ordered_ab_semigroup_ordered_semiring_0.non_strict_order_ordered_ab_semigroup.ord_non_strict_order,
          b));;

let rec normalize_poly (_A1, _A2, _A3)
  p = divide_polya (_A1, _A2) p
        (pCons
          (_A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero,
            _A1)
          (unit_factor _A3.unit_factor_semidom_divide_unit_factor
            (coeff
              _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
              p (degreea
                  _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
                  p)))
          (zero_polyb
            _A2.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero));;

let rec primitive_part (_A1, _A2)
  p = map_poly _A1.gcd_semiring_gcd.zero_gcd
        (_A1.gcd_semiring_gcd.zero_gcd, _A2)
        (fun x ->
          divide
            _A1.normalization_semidom_semiring_gcd.algebraic_semidom_normalization_semidom.semidom_divide_algebraic_semidom.divide_semidom_divide
            x (content _A1 p))
        p;;

let rec gcd_poly_code_aux (_A1, _A2)
  p q = (if is_zero
              _A1.factorial_semiring_gcd_factorial_ring_gcd.semiring_Gcd_factorial_semiring_gcd.gcd_semiring_Gcd.gcd_Gcd.zero_gcd
              q
          then normalize_poly
                 (_A2, _A1.idom_divide_factorial_ring_gcd,
                   _A1.ring_gcd_factorial_ring_gcd.semiring_gcd_ring_gcd.normalization_semidom_semiring_gcd.semidom_divide_unit_factor_normalization_semidom)
                 p
          else gcd_poly_code_aux (_A1, _A2) q
                 (primitive_part
                   (_A1.ring_gcd_factorial_ring_gcd.semiring_gcd_ring_gcd, _A2)
                   (pseudo_mod
                     (_A2, _A1.idom_divide_factorial_ring_gcd.idom_idom_divide.comm_ring_1_idom,
                       _A1.idom_divide_factorial_ring_gcd.idom_idom_divide.semidom_idom.semiring_1_no_zero_divisors_semidom)
                     p q)));;

let rec gcd_poly_code (_A1, _A2)
  p q = (if is_zero
              _A1.factorial_semiring_gcd_factorial_ring_gcd.semiring_Gcd_factorial_semiring_gcd.gcd_semiring_Gcd.gcd_Gcd.zero_gcd
              p
          then normalize_poly
                 (_A2, _A1.idom_divide_factorial_ring_gcd,
                   _A1.ring_gcd_factorial_ring_gcd.semiring_gcd_ring_gcd.normalization_semidom_semiring_gcd.semidom_divide_unit_factor_normalization_semidom)
                 q
          else (if is_zero
                     _A1.factorial_semiring_gcd_factorial_ring_gcd.semiring_Gcd_factorial_semiring_gcd.gcd_semiring_Gcd.gcd_Gcd.zero_gcd
                     q
                 then normalize_poly
                        (_A2, _A1.idom_divide_factorial_ring_gcd,
                          _A1.ring_gcd_factorial_ring_gcd.semiring_gcd_ring_gcd.normalization_semidom_semiring_gcd.semidom_divide_unit_factor_normalization_semidom)
                        p
                 else smult (_A2, _A1.idom_divide_factorial_ring_gcd.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.comm_semiring_0_comm_semiring_1,
                              _A1.idom_divide_factorial_ring_gcd.idom_idom_divide.semidom_idom.semiring_1_no_zero_divisors_semidom.semiring_no_zero_divisors_semiring_1_no_zero_divisors)
                        (gcda _A1.factorial_semiring_gcd_factorial_ring_gcd.semiring_Gcd_factorial_semiring_gcd.gcd_semiring_Gcd.gcd_Gcd
                          (content
                            _A1.ring_gcd_factorial_ring_gcd.semiring_gcd_ring_gcd
                            p)
                          (content
                            _A1.ring_gcd_factorial_ring_gcd.semiring_gcd_ring_gcd
                            q))
                        (gcd_poly_code_aux (_A1, _A2)
                          (primitive_part
                            (_A1.ring_gcd_factorial_ring_gcd.semiring_gcd_ring_gcd,
                              _A2)
                            p)
                          (primitive_part
                            (_A1.ring_gcd_factorial_ring_gcd.semiring_gcd_ring_gcd,
                              _A2)
                            q))));;

let rec gcd_poly (_A1, _A2) p q = gcd_poly_code (_A1, _A2) p q;;

let rec divmod_poly_one_main_list (_A1, _A2)
  q r d n =
    (if equal_nata n zero_nata then (q, r)
      else (let a = hda r in
            let qqq =
              cCons (_A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero,
                      _A1)
                a q
              in
            let rr =
              tla (if eq _A1 a
                        (zerob
                          _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
                    then r
                    else minus_poly_rev_list
                           _A2.ring_1_comm_ring_1.neg_numeral_ring_1.group_add_neg_numeral
                           r (map (times
                                    _A2.comm_semiring_1_cancel_comm_ring_1.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd
                                    a)
                               d))
              in
             divmod_poly_one_main_list (_A1, _A2) qqq rr d
               (minus_nata n one_nata)));;

let rec div_field_poly_impl (_A1, _A2)
  f g = (let cg =
           coeffs
             _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
             g
           in
          (if null cg
            then zero_polyb
                   _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
            else (let cf =
                    coeffs
                      _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
                      f
                    in
                  let ilc =
                    inverse _A1.division_ring_field.inverse_division_ring
                      (last cg)
                    in
                  let ch =
                    map (times
                          _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd
                          ilc)
                      cg
                    in
                  let q =
                    fst (divmod_poly_one_main_list
                          (_A2, _A1.idom_divide_field.idom_idom_divide.comm_ring_1_idom)
                          [] (rev cf) (rev ch)
                          (minus_nata (plus_nata one_nata (size_list cf))
                            (size_list cg)))
                    in
                   poly_of_list
                     (_A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.comm_monoid_add_semiring_0,
                       _A2)
                     (map (times
                            _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd
                            ilc)
                       q))));;

let rec sturm_squarefree
  p = sturm (div_field_poly_impl (field_real, equal_real) p
              (gcd_poly (factorial_ring_gcd_real, equal_real) p
                (pderiv
                  (equal_real, comm_semiring_1_real,
                    semiring_no_zero_divisors_real)
                  p)));;

let rec poly_inf _A
  p = sgn _A.sgn_div_norm_real_normed_vector.sgn_sgn_div_norm
        (coeff
          _A.real_vector_real_normed_vector.ab_group_add_real_vector.group_add_ab_group_add.monoid_add_group_add.zero_monoid_add
          p (degreea
              _A.real_vector_real_normed_vector.ab_group_add_real_vector.group_add_ab_group_add.monoid_add_group_add.zero_monoid_add
              p));;

let rec sign_changes_inf (_A1, _A2)
  ps = minus_nata
         (size_list
           (remdups_adj _A1
             (filtera
               (fun x ->
                 not (eq _A1 x
                       (zerob
                         _A2.real_vector_real_normed_vector.ab_group_add_real_vector.group_add_ab_group_add.monoid_add_group_add.zero_monoid_add)))
               (map (poly_inf _A2) ps))))
         one_nata;;

let rec sign_changes
  ps x =
    minus_nata
      (size_list
        (remdups_adj equal_real
          (filtera (fun xa -> not (equal_reala xa zero_reala))
            (map (fun p -> sgn_reala (polya comm_semiring_0_real p x)) ps))))
      one_nata;;

let rec count_roots_above
  p a = (let q =
           pderiv
             (equal_real, comm_semiring_1_real, semiring_no_zero_divisors_real)
             p
           in
          (if equal_polya (zero_real, equal_real) p (zero_polyb zero_real)
            then zero_nata
            else (if not (equal_reala (polya comm_semiring_0_real p a)
                           zero_reala) ||
                       not (equal_reala (polya comm_semiring_0_real q a)
                             zero_reala)
                   then (let ps = sturm p in
                          minus_nata (sign_changes ps a)
                            (sign_changes_inf
                              (equal_real, real_normed_vector_real) ps))
                   else (let ps = sturm_squarefree p in
                          minus_nata (sign_changes ps a)
                            (sign_changes_inf
                              (equal_real, real_normed_vector_real) ps)))));;

let rec minus_mat _A
  a b = mat (dim_row b) (dim_col b)
          (fun ij -> minus _A (index_mat a ij) (index_mat b ij));;

let rec pivot_positions_main_gen _A
  zero a nr nc i j =
    (if less_nat i nr
      then (if less_nat j nc
             then (if eq _A (index_mat a (i, j)) zero
                    then pivot_positions_main_gen _A zero a nr nc i (suc j)
                    else (i, j) ::
                           pivot_positions_main_gen _A zero a nr nc (suc i)
                             (suc j))
             else [])
      else []);;

let rec pivot_positions_gen _A
  zer a =
    pivot_positions_main_gen _A zer a (dim_row a) (dim_col a) zero_nata
      zero_nata;;

let rec eliminate_entries_gen
  minus times v a i j =
    mat (dim_row a) (dim_col a)
      (fun (ia, ja) ->
        (if not (equal_nata ia i)
          then minus (index_mat a (ia, ja)) (times (v ia) (index_mat a (i, ja)))
          else index_mat a (ia, ja)));;

let rec gauss_jordan_main (_A1, _A2)
  a b i j =
    (let nr = dim_row a in
     let nc = dim_col a in
      (if less_nat i nr && less_nat j nc
        then (let aij = index_mat a (i, j) in
               (if eq _A2 aij
                     (zerob
                       _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
                 then (match
                        maps (fun ia ->
                               (if not (eq _A2 (index_mat a (ia, j))
 (zerob
   _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero))
                                 then [ia] else []))
                          (upt (suc i) nr)
                        with [] -> gauss_jordan_main (_A1, _A2) a b i (suc j)
                        | ia :: _ ->
                          gauss_jordan_main (_A1, _A2) (mat_swaprows i ia a)
                            (mat_swaprows i ia b) i j)
                 else (if eq _A2 aij
                            (onea _A1.idom_divide_field.idom_idom_divide.comm_ring_1_idom.ring_1_comm_ring_1.neg_numeral_ring_1.numeral_neg_numeral.one_numeral)
                        then (let v = (fun ia -> index_mat a (ia, j)) in
                               gauss_jordan_main (_A1, _A2)
                                 (eliminate_entries_gen
                                   (minus
                                     _A1.idom_divide_field.idom_idom_divide.comm_ring_1_idom.ring_1_comm_ring_1.neg_numeral_ring_1.group_add_neg_numeral.minus_group_add)
                                   (times
                                     _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd)
                                   v a i j)
                                 (eliminate_entries_gen
                                   (minus
                                     _A1.idom_divide_field.idom_idom_divide.comm_ring_1_idom.ring_1_comm_ring_1.neg_numeral_ring_1.group_add_neg_numeral.minus_group_add)
                                   (times
                                     _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd)
                                   v b i j)
                                 (suc i) (suc j))
                        else (let iaij =
                                inverse
                                  _A1.division_ring_field.inverse_division_ring
                                  aij
                                in
                              let aa =
                                mat_multrow_gen
                                  (times
                                    _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd)
                                  i iaij a
                                in
                              let ba =
                                mat_multrow_gen
                                  (times
                                    _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd)
                                  i iaij b
                                in
                              let v = (fun ia -> index_mat aa (ia, j)) in
                               gauss_jordan_main (_A1, _A2)
                                 (eliminate_entries_gen
                                   (minus
                                     _A1.idom_divide_field.idom_idom_divide.comm_ring_1_idom.ring_1_comm_ring_1.neg_numeral_ring_1.group_add_neg_numeral.minus_group_add)
                                   (times
                                     _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd)
                                   v aa i j)
                                 (eliminate_entries_gen
                                   (minus
                                     _A1.idom_divide_field.idom_idom_divide.comm_ring_1_idom.ring_1_comm_ring_1.neg_numeral_ring_1.group_add_neg_numeral.minus_group_add)
                                   (times
                                     _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.comm_monoid_mult_comm_semiring_1.dvd_comm_monoid_mult.times_dvd)
                                   v ba i j)
                                 (suc i) (suc j)))))
        else (a, b)));;

let rec gauss_jordan (_A1, _A2)
  a b = gauss_jordan_main (_A1, _A2) a b zero_nata zero_nata;;

let rec gauss_jordan_single (_A1, _A2)
  a = fst (gauss_jordan (_A1, _A2) a
            (zero_mat
              _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero
              (dim_row a) zero_nata));;

let rec kernel_dim (_A1, _A2)
  a = minus_nata (dim_col a)
        (size_list
          (pivot_positions_gen _A2
            (zerob
              _A1.idom_divide_field.idom_idom_divide.semidom_idom.comm_semiring_1_cancel_semidom.comm_semiring_1_comm_semiring_1_cancel.semiring_1_comm_semiring_1.semiring_0_semiring_1.mult_zero_semiring_0.zero_mult_zero)
            (gauss_jordan_single (_A1, _A2) a)));;

let rec check_matrix_complexity
  a cp d =
    equal_nata (count_roots_above cp one_reala) zero_nata &&
      (if equal_reala (polya comm_semiring_0_real cp one_reala) zero_reala
        then (let ord = order (equal_real, idom_divide_real) one_reala cp in
               (if less_nat (plus_nata d one_nata) ord
                 then equal_nata
                        (kernel_dim (field_real, equal_real)
                          (pow_mat semiring_1_real
                            (minus_mat minus_real a
                              (one_mat (one_real, zero_real) (dim_row a)))
                            (plus_nata d one_nata)))
                        ord
                 else true))
        else true);;

let rec nonneg_mat (_A1, _A2, _A3, _A4)
  a = ball (_A1, _A2) (elements_mat (_A1, _A2, _A4) a)
        (less_eq
          _A3.linordered_ring_strict_linordered_idom.linordered_ring_linordered_ring_strict.abs_if_linordered_ring.ord_abs_if
          (zerob
            _A3.linordered_ring_strict_linordered_idom.linordered_ring_linordered_ring_strict.abs_if_linordered_ring.zero_abs_if));;

let rec complexity_via_perron_frobenius
  d cp a =
    nonneg_mat (ceq_real, ccompare_real, linordered_idom_real, set_impl_real)
      a &&
      check_matrix_complexity a cp (minus_nata d one_nata);;

let rec count_ones_check
  diag d =
    list_all
      (fun a ->
        (let aa = abs_reala a in
          less_eq_real aa one_reala &&
            (if equal_reala aa one_reala
              then less_eq_nat (size_list (filtera (equal_reala a) diag)) d
              else true)))
      diag;;

let rec combined_growth_check_real_mat
  d a = catch_error
          (if upper_triangular (zero_real, equal_real) a &&
                count_ones_check (diag_mat a) d
            then Inr ()
            else (if complexity_via_perron_frobenius d
                       (char_poly (equal_real, idom_divide_real) a) a
                   then Inr ()
                   else Inl (if nonneg_mat
                                  (ceq_real, ccompare_real,
                                    linordered_idom_real, set_impl_real)
                                  a
                              then showsl_lit
                                     "matrix does not have intended growth rate"
                              else showsl_lit
                                     "only non-negative matrices supported")))
          (fun x ->
            Inl (comp (comp (comp (comp (comp
  (showsl_lit "could not deduce that ") (showsl_mat showl_real a))
                                    (showsl_literal " in O(n^"))
                              (showsl_nat (minus_nata d one_nata)))
                        (showsl_lit ")\092"))
                  x));;

let rec mat_estimate_complexity_jb _A
  d a = (let b =
           map_mat (real_of _A.real_embedding_large_real_ordered_semiring_1) a
           in
          bindb (check (equal_nata (dim_row a) (dim_col a))
                  (comp (showsl_lit "expected square matrix but got \092")
                    (showsl_mat showl_real b)))
            (fun _ -> combined_growth_check_real_mat d b));;

let rec mat_complexity _A n m d = mat_estimate_complexity_jb _A (suc d) m;;

let rec mat_mono
  mon sd a =
    all_interval_nat
      (fun j ->
        not (all_interval_nat (comp not (fun i -> mon (index_mat a (i, j))))
              zero_nata sd))
      zero_nata sd;;

let rec mat_lpoly_order _A
  n sd def mon gtt =
    mat_ordered_semiring
      _A.large_ordered_semiring_1_large_real_ordered_semiring_1.poly_carrier_large_ordered_semiring_1.ordered_semiring_1_poly_carrier
      n sd gtt
      (Lpoly_order_semiring_ext
        (true,
          mat_default
            _A.real_embedding_large_real_ordered_semiring_1.linordered_idom_real_embedding.linordered_ring_strict_linordered_idom.linordered_ring_linordered_ring_strict.abs_if_linordered_ring.zero_abs_if
            def n,
          (fun _ -> true), mat_mono mon sd, (fun _ -> zero_nata),
          mat_complexity _A n, showsl_lit "matrix interpretation", ()));;

let rec int_mono x = less_eq_int one_inta x;;

let rec delta_gt _A
  delta =
    (fun x y ->
      less_eq
        _A.archimedean_field_floor_ceiling.linordered_field_archimedean_field.linordered_idom_linordered_field.linordered_ring_strict_linordered_idom.linordered_ring_linordered_ring_strict.abs_if_linordered_ring.ord_abs_if
        delta
        (minus
          _A.archimedean_field_floor_ceiling.linordered_field_archimedean_field.linordered_idom_linordered_field.linordered_ring_strict_linordered_idom.linordered_ring_linordered_ring_strict.abs_if_linordered_ring.minus_abs_if
          x y));;

let rec check_def_pos (_A1, _A2)
  d = check (less _A2 (zerob _A1) d)
        (showsl_lit "default value must be positive");;

let rec rpo_nstrict_unbounded (_A1, _A2) (_B1, _B2)
  pr c =
    (fun (s, t) ->
      check (snd (rpo_unbounded _A1 _B1 pr c s t))
        (comp (comp (comp (comp (showsl_literal "could not orient ")
                            (showsl_terma _A2 _B2 s))
                      (showsl_literal " >=RPO "))
                (showsl_terma _A2 _B2 t))
          (showsl_literal "\092")));;

let rec rpo_strict_unbounded (_A1, _A2) (_B1, _B2)
  pr c =
    (fun (s, t) ->
      check (fst (rpo_unbounded _A1 _B1 pr c s t))
        (comp (comp (comp (comp (showsl_literal "could not orient ")
                            (showsl_terma _A2 _B2 s))
                      (showsl_literal " >RPO "))
                (showsl_terma _A2 _B2 t))
          (showsl_literal "\092")));;

let rec create_RPO_redtriple _A (_B1, _B2) (_C1, _C2)
  prec_repr_to_pr pr =
    (let (p, tau) = prec_repr_to_pr pr in
     let ns =
       rpo_nstrict_unbounded (_B1, _B2) (_C1, _C2) (prc_nat p, prl_nat p) tau in
      Redtriple_ext
        (Inr (),
          rpo_strict_unbounded (_B1, _B2) (_C1, _C2) (prc_nat p, prl_nat p) tau,
          ns, ns, full_af, full_af, (fun _ -> Inr ()), showsl_rpo_repr _A pr,
          Some [], no_complexity_check, ()));;

let rec get_redtriple (_A1, _A2, _A3, _A4, _A5, _A6)
  = function
    Int_carrier i ->
      create_poly_redtriple
        (cenum_int, ceq_int, ccompare_int, equal_int, set_impl_int, showl_int)
        (_A3, _A4, _A6) ((equal_list equal_char), (showl_list showl_char))
        (class_lpoly_order
          (ceq_int, ccompare_int, ordered_semiring_1_int, set_impl_int) one_inta
          int_mono (fun x y -> less_int y x))
        (Inr ()) i
    | Int_nl_carrier i ->
        create_nlpoly_redtriple (equal_int, poly_carrier_int, showl_int)
          (_A3, _A4, _A6)
          ((equal_list equal_char), (linorder_list (equal_char, linorder_char)),
            (showl_list showl_char))
          (Inr ()) one_inta (fun x y -> less_int y x) true true i
    | Rat_carrier i ->
        create_poly_redtriple
          (cenum_rat, ceq_rat, ccompare_rat, equal_rat, set_impl_rat, showl_rat)
          (_A3, _A4, _A6) ((equal_list equal_char), (showl_list showl_char))
          (class_lpoly_order
            (ceq_rat, ccompare_rat, ordered_semiring_1_rat, set_impl_rat)
            one_rata (delta_mono floor_ceiling_rat) (fun x y -> less_rat y x))
          (check_def_pos (zero_rat, ord_rat) one_rata) i
    | Rat_nl_carrier (d, i) ->
        create_nlpoly_redtriple (equal_rat, poly_carrier_rat, showl_rat)
          (_A3, _A4, _A6)
          ((equal_list equal_char), (linorder_list (equal_char, linorder_char)),
            (showl_list showl_char))
          (check_def_pos (zero_rat, ord_rat) d) d (delta_gt floor_ceiling_rat d)
          (less_eq_rat one_rata d) false i
    | Real_carrier i ->
        create_poly_redtriple
          (cenum_real, ceq_real, ccompare_real, equal_real, set_impl_real,
            showl_real)
          (_A3, _A4, _A6) ((equal_list equal_char), (showl_list showl_char))
          (class_lpoly_order
            (ceq_real, ccompare_real, ordered_semiring_1_real, set_impl_real)
            one_reala (delta_mono floor_ceiling_real)
            (fun x y -> less_real y x))
          (check_def_pos (zero_real, ord_real) one_reala) i
    | Real_nl_carrier (d, i) ->
        create_nlpoly_redtriple (equal_real, poly_carrier_real, showl_real)
          (_A3, _A4, _A6)
          ((equal_list equal_char), (linorder_list (equal_char, linorder_char)),
            (showl_list showl_char))
          (check_def_pos (zero_real, ord_real) d) d
          (delta_gt floor_ceiling_real d) (less_eq_real one_reala d) false i
    | Arctic_carrier i ->
        create_poly_redtriple
          (cenum_arctic, ceq_arctic, ccompare_arctic, equal_arctic,
            set_impl_arctic, showl_arctic)
          (_A3, _A4, _A6) ((equal_list equal_char), (showl_list showl_char))
          (class_arc_lpoly_order
            (ceq_arctic, ccompare_arctic, ordered_semiring_1_arctic,
              set_impl_arctic)
            one_arctica pos_arctic (fun x y -> less_arctic y x))
          (Inr ()) i
    | Arctic_rat_carrier i ->
        create_poly_redtriple
          (cenum_arctic_delta, (ceq_arctic_delta equal_rat),
            (ccompare_arctic_delta compare_rat), (equal_arctic_delta equal_rat),
            set_impl_arctic_delta, (showl_arctic_delta showl_rat))
          (_A3, _A4, _A6) ((equal_list equal_char), (showl_list showl_char))
          (class_arc_lpoly_order
            ((ceq_arctic_delta equal_rat), (ccompare_arctic_delta compare_rat),
              (ordered_semiring_1_arctic_delta linordered_field_rat),
              set_impl_arctic_delta)
            (one_arctic_deltaa linordered_field_rat)
            (pos_arctic_delta floor_ceiling_rat)
            (weak_gt_arctic_delta floor_ceiling_rat))
          (Inr ()) i
    | Int_mat_carrier (n, sd, i) ->
        create_poly_redtriple
          (cenum_mat, (ceq_mat equal_int), ccompare_mat, (equal_mat equal_int),
            set_impl_mat, (showl_mat showl_int))
          (_A3, _A4, _A6) ((equal_list equal_char), (showl_list showl_char))
          (mat_lpoly_order large_real_ordered_semiring_1_int n sd one_inta
            int_mono (fun x y -> less_int y x))
          (check_dimensions n sd (Inr ())) i
    | Rat_mat_carrier (n, sd, i) ->
        create_poly_redtriple
          (cenum_mat, (ceq_mat equal_rat), ccompare_mat, (equal_mat equal_rat),
            set_impl_mat, (showl_mat showl_rat))
          (_A3, _A4, _A6) ((equal_list equal_char), (showl_list showl_char))
          (mat_lpoly_order large_real_ordered_semiring_1_rat n sd one_rata
            (delta_mono floor_ceiling_rat) (fun x y -> less_rat y x))
          (check_dimensions n sd (check_def_pos (zero_rat, ord_rat) one_rata)) i
    | Real_mat_carrier (n, sd, i) ->
        create_poly_redtriple
          (cenum_mat, (ceq_mat equal_real), ccompare_mat,
            (equal_mat equal_real), set_impl_mat, (showl_mat showl_real))
          (_A3, _A4, _A6) ((equal_list equal_char), (showl_list showl_char))
          (mat_lpoly_order large_real_ordered_semiring_1_real n sd one_reala
            (delta_mono floor_ceiling_real) (fun x y -> less_real y x))
          (check_dimensions n sd
            (check_def_pos (zero_real, ord_real) one_reala))
          i
    | Arctic_mat_carrier (n, i) ->
        create_poly_redtriple
          (cenum_mat, (ceq_mat equal_arctic), ccompare_mat,
            (equal_mat equal_arctic), set_impl_mat, (showl_mat showl_arctic))
          (_A3, _A4, _A6) ((equal_list equal_char), (showl_list showl_char))
          (mat_arc_lpoly_order ordered_semiring_1_arctic n one_arctica
            pos_arctic (fun x y -> less_arctic y x))
          (check_arc_dimension (zero_nat, ord_nat) n) i
    | Arctic_rat_mat_carrier (n, i) ->
        create_poly_redtriple
          (cenum_mat, (ceq_mat (equal_arctic_delta equal_rat)), ccompare_mat,
            (equal_mat (equal_arctic_delta equal_rat)), set_impl_mat,
            (showl_mat (showl_arctic_delta showl_rat)))
          (_A3, _A4, _A6) ((equal_list equal_char), (showl_list showl_char))
          (mat_arc_lpoly_order
            (ordered_semiring_1_arctic_delta linordered_field_rat) n
            (one_arctic_deltaa linordered_field_rat)
            (pos_arctic_delta floor_ceiling_rat)
            (weak_gt_arctic_delta floor_ceiling_rat))
          (check_arc_dimension (zero_nat, ord_nat) n) i
    | RPO (prec_tau, pi) ->
        af_redtriple (_A1, _A2, _A3, _A5, _A6) (showl_list showl_char) pi
          (create_RPO_redtriple _A6 ((equal_filtered _A4), (showl_filtered _A6))
            ((equal_list equal_char), (showl_list showl_char))
            (fun pr -> (prec_repr_to_pr _A3 pr, prec_repr_to_status _A3 pr))
            prec_tau)
    | KBO (precw, pi) ->
        af_redtriple (_A1, _A2, _A3, _A5, _A6) (showl_list showl_char) pi
          (create_KBO_redtriple _A6
            ((compare_order_filtered _A3.compare_compare_order),
              (equal_filtered _A4), (showl_filtered _A6))
            ((equal_list equal_char), (showl_list showl_char))
            (filter_prec_weight_repr
              (fun_of_map_fun
                (ceta_map_of (compare_order_prod _A3 compare_order_nat) pi)
                (fun fn -> default_af_entry (snd fn))))
            precw)
    | ACKBO (precw, pi) ->
        af_redtriple (_A1, _A2, _A3, _A5, _A6) (showl_list showl_char) pi
          (create_ACKBO_redtriple _A6
            ((ceq_filtered _A1), (ccompare_filtered _A2),
              (compare_order_filtered _A3.compare_compare_order),
              (equal_filtered _A4), (set_impl_filtered _A5),
              (showl_filtered _A6))
            ((equal_list equal_char), (showl_list showl_char))
            (filter_prec_weight_ac_repr
              (fun_of_map_fun
                (ceta_map_of (compare_order_prod _A3 compare_order_nat) pi)
                (fun fn -> default_af_entry (snd fn))))
            precw)
    | WPO (params, rp) ->
        wpo_redtriple (_A3, _A4, _A6)
          (get_redtriple (_A1, _A2, _A3, _A4, _A5, _A6) rp) params
    | Max_poly alist ->
        create_max_poly_redtriple (_A4, _A6)
          ((ccompare_list ccompare_char), (equal_list equal_char),
            mapping_impl_list, (linorder_list (equal_char, linorder_char)),
            (showl_list showl_char))
          alist;;

let rec list_ext_name = function MS_Ext -> "MS"
                        | Dms_Ext -> "DMS"
                        | Min_Ext -> "MIN"
                        | Max_Ext -> "MAX";;

let rec dms_preprocess
  p = map (filtera (fun (_, a) -> (let (aa, b) = a in aa || b))) p;;

let rec dms_simplify
  stri is p = (if list_ex (fun i -> null (nth p i)) is then [[]] else p);;

let rec dms_decide_singletons
  stri n p =
    all_interval_nat
      (fun i ->
        (let (j, (s, ns)) = nth p i in
          s && not (membera
                     (equal_prod equal_nat (equal_prod equal_bool equal_bool))
                     (drop (suc i) p) (j, (false, true))) ||
            ns && not (membera equal_nat (map fst (drop (suc i) p)) j)))
      zero_nata (size_list p) &&
      (if stri
        then not (all_interval_nat
                   (comp not
                     (fun j ->
                       not (membera
                             (equal_prod equal_nat
                               (equal_prod equal_bool equal_bool))
                             p (j, (false, true)))))
                   zero_nata n)
        else true);;

let rec dms_select
  stri p =
    snd (hda (sort_key linorder_nat fst
               (filtera (fun (l, _) -> less_nat one_nata l)
                 (zip (map size_list p) (upt zero_nata (size_list p))))));;

let rec dms_solve_or_select
  stri n p =
    (if list_all (fun jsns -> less_eq_nat (size_list jsns) one_nata) p
      then Inl (if membera
                     (equal_list
                       (equal_prod equal_nat
                         (equal_prod equal_bool equal_bool)))
                     p []
                 then false else dms_decide_singletons stri n (map hda p))
      else Inr (dms_select stri p));;

let rec dms_solve
  stri n p =
    (match dms_solve_or_select stri n p with Inl res -> res
      | Inr k ->
        (let ksns = nth p k in
          dms_solve stri n
            (dms_simplify stri [k] (list_update p k [hda ksns])) ||
            dms_solve stri n
              (dms_simplify stri [k] (list_update p k (tla ksns)))));;

let rec dms_bool_ex_idx_impl
  stri n p =
    dms_solve stri n
      (dms_simplify stri (upt zero_nata (size_list p)) (dms_preprocess p));;

let rec dms_bool_ex_idx x = dms_bool_ex_idx_impl x;;

let rec dms_convert
  f asa bs =
    (let jbs = zip (upt zero_nata (size_list bs)) bs in
      map (fun a -> map (fun (j, b) -> (j, f a b)) jbs) asa);;

let rec dms_order_ext
  n f asa bs =
    (let p = dms_convert f asa bs in
     let lts = size_list bs in
     let len = less_eq_nat lts n || equal_nata (size_list asa) lts in
      (len && dms_bool_ex_idx true lts p, len && dms_bool_ex_idx false lts p));;

let rec min_set_ext
  s_ns =
    (fun asa bs ->
      (not (null bs) &&
         list_all (fun a -> list_ex (fun b -> fst (s_ns a b)) bs) asa,
        list_all (fun a -> list_ex (fun b -> snd (s_ns a b)) bs) asa));;

let rec list_ext uu x1 = match uu, x1 with uu, MS_Ext -> mul_ext
                   | uv, Max_Ext -> set_ext
                   | uw, Min_Ext -> min_set_ext
                   | n, Dms_Ext -> dms_order_ext n;;

let rec label_s_ns_impl _A _B
  cS cNS s t =
    (let (sa, i) = s in
     let (ta, j) = t in
      (if isOK (cS (sa, ta)) then (true, true)
        else (if isOK (cNS (sa, ta)) then (less_nat j i, less_eq_nat j i)
               else (false, false))));;

let rec nST_label_mul_impl _A _B
  list_ext af cS cNS st =
    (match st
      with (Var _, _) ->
        Inl (comp (comp (showsl_literal "roots of ") (showsl_rule _A _B st))
              (showsl_literal " must be non-variable"))
      | (Fun (_, _), Var _) ->
        Inl (comp (comp (showsl_literal "roots of ") (showsl_rule _A _B st))
              (showsl_literal " must be non-variable"))
      | (Fun (f, ss), Fun (g, ts)) ->
        check (snd (list_ext (label_s_ns_impl _A _B cS cNS)
                      (lterms af (Fun (f, ss)))
                     (lterms af (Fun (g, ts)))))
          (comp (comp (comp (comp (comp (comp
  (showsl_literal "cannot orient pair ") (showsl_rule _A _B st))
                                    (showsl_literal " weakly:\092"))
                              (showsl_list_prod (showl_term _A _B) showl_nat
                                (lterms af (Fun (f, ss)))))
                        (showsl_literal " >=mu "))
                  (showsl_list_prod (showl_term _A _B) showl_nat
                    (lterms af (Fun (g, ts)))))
            (showsl_literal " could not be ensured")));;

let rec s_label_mul_impl _A _B
  list_ext af cS cNS st =
    (match st
      with (Var _, _) ->
        Inl (comp (comp (showsl_lit "roots of ") (showsl_rule _A _B st))
              (showsl_lit " must be non-variable"))
      | (Fun (_, _), Var _) ->
        Inl (comp (comp (showsl_lit "roots of ") (showsl_rule _A _B st))
              (showsl_lit " must be non-variable"))
      | (Fun (f, ss), Fun (g, ts)) ->
        check (fst (list_ext (label_s_ns_impl _A _B cS cNS)
                      (lterms af (Fun (f, ss)))
                     (lterms af (Fun (g, ts)))))
          (comp (comp (comp (comp (comp (comp (showsl_lit "cannot orient pair ")
  (showsl_rule _A _B st))
                                    (showsl_lit " strictly:\092"))
                              (showsl_list_prod (showl_term _A _B) showl_nat
                                (lterms af (Fun (f, ss)))))
                        (showsl_lit " >mu "))
                  (showsl_list_prod (showl_term _A _B) showl_nat
                    (lterms af (Fun (g, ts)))))
            (showsl_lit " could not be ensured")));;

let rec generate_scnp_rp (_A1, _A2) _B
  list_ext list_ext_name afa rti x =
    (let rt = rti x in
     let afaa =
       fun_of_map (ceta_map_of (compare_order_prod _A1 compare_order_nat) afa)
         []
       in
     let pi = af rt in
     let cS = s rt in
     let cNS = ns rt in
      Root_redtriple_ext
        (valid rt, s_label_mul_impl _A2 _B list_ext afaa cS cNS, ns rt,
          nST_label_mul_impl _A2 _B list_ext afaa cS cNS, pi,
          scnp_af_to_af afaa pi,
          comp (scnp_desc _A2 afa list_ext_name) (desc rt), ()));;

let rec get_root_redtriple (_A1, _A2, _A3, _A4, _A5, _A6)
  (SCNP (typea, af, rp)) =
    generate_scnp_rp (_A3, _A6) (showl_list showl_char)
      (list_ext (scnp_arity af) typea) (showsl_lit (list_ext_name typea)) af
      (get_redtriple (_A1, _A2, _A3, _A4, _A5, _A6)) rp;;

let rec faulty_non_inf_order _B _C
  s f = Non_inf_order_ext
          (Inl (showsl_lit s), (fun _ -> Inr ()), (fun _ -> Inr ()),
            (fun _ _ -> Wild), id, ());;

let rec sqrt_real
  x = (if less_eq_real zero_reala x
        then (let y = sqrt x in remdups equal_real [y; uminus_reala y])
        else []);;

let rec check_non_inf_poly_inter_list _A (_B1, _B2)
  discrete f i =
    catch_error
      (catch_error
        (forallM
          (fun x ->
            (if (let (_, a) = x in
                  check_poly_weak_mono_and_pos (equal_nat, linorder_nat)
                    (_B1, _B2) discrete a)
              then Inr () else Inl x))
          (filtera (fun (fn, _) -> membera (equal_prod _A equal_nat) f fn) i))
        (fun x -> Inl (snd x)))
      (fun x ->
        Inl (let (a, b) = x in (let (fa, _) = a in (fun aa -> (fa, aa))) b));;

let rec create_nlpoly_non_inf_order (_A1, _A2, _A3) (_B1, _B2, _B3)
  (_C1, _C2, _C3)
  cI def gt power_mono discrete sqrt i f =
    (let j = poly_inter_list_to_inter _A2 _B1 def i in
     let x =
       poly_subst linorder_nat
         ((equal_list equal_char), (linorder_list (equal_char, linorder_char)))
         (_A1, _A2.comm_semiring_1_poly_carrier)
         (fun n ->
           poly_of
             ((equal_list equal_char),
               (linorder_list (equal_char, linorder_char)))
             (_A1, _A2.comm_semiring_1_poly_carrier)
             (PVar ([Chara (false, false, false, true, true, true, true, false);
                      Chara (true, true, true, true, true, false, true,
                              false)] @
                     shows_prec_nat zero_nata n [])))
       in
      Non_inf_order_ext
        (bindb cI
           (fun _ ->
             catch_error
               (check_non_inf_poly_inter_list _B2 (_A1, _A2) discrete f i)
               (fun xa ->
                 Inl (let (fa, p) = xa in
                       comp (comp (comp (comp (showsl_literal "interpretation ")
  (showsl_poly
    ((equal_list equal_char), (linorder_list (equal_char, linorder_char)),
      (showl_list showl_char))
    (_A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral,
      _A1, _A3)
    (x p)))
                                    (showsl_literal " of "))
                              (showsl _B3 fa))
                         (showsl_literal " invalid ")))),
          check_ns _B3 (_A1, _A2, _A3) (_C1, _C2, _C3) j,
          check_cc (_A1, _A2, _A3) _B3 (_C1, _C2, _C3) sqrt gt j,
          create_dep (_A1, _A2) (_B1, _B2) discrete def i,
          comp (showsl_literal "polynomial interpretation\092")
            (showsl_sep
              (fun (a, b) ->
                (let (fa, n) = a in
                  (fun p ->
                    comp (comp (comp (comp (comp (showsl_literal "Pol(")
     (showsl _B3 fa))
                                       (showsl_literal "/"))
                                 (showsl_nat n))
                           (showsl_literal ") = "))
                      (showsl_poly
                        ((equal_list equal_char),
                          (linorder_list (equal_char, linorder_char)),
                          (showl_list showl_char))
                        (_A2.comm_semiring_1_poly_carrier.semiring_1_comm_semiring_1.semiring_numeral_semiring_1.numeral_semiring_numeral.one_numeral,
                          _A1, _A3)
                        (x p))))
                  b)
              (showsl_literal "\092") i),
          ()));;

let rec sqrt_rat
  x = (let (z, n) = quotient_of x in
        (match sqrt_int n with [] -> []
          | sn :: _ ->
            map (fun sz -> divide_rata (of_int sz) (of_int sn)) (sqrt_int z)));;

let rec get_non_inf_order (_A1, _A2, _A3) (_B1, _B2, _B3)
  = function
    Int_nl_carrier i ->
      create_nlpoly_non_inf_order (equal_int, poly_carrier_int, showl_int)
        (_A1, _A2, _A3) (_B1, _B2, _B3) (Inr ()) one_inta
        (fun x y -> less_int y x) true true sqrt_int i
    | Rat_nl_carrier (d, i) ->
        create_nlpoly_non_inf_order (equal_rat, poly_carrier_rat, showl_rat)
          (_A1, _A2, _A3) (_B1, _B2, _B3) (check_def_pos (zero_rat, ord_rat) d)
          d (delta_gt floor_ceiling_rat d) (less_eq_rat one_rata d) false
          sqrt_rat i
    | Real_nl_carrier (d, i) ->
        create_nlpoly_non_inf_order (equal_real, poly_carrier_real, showl_real)
          (_A1, _A2, _A3) (_B1, _B2, _B3)
          (check_def_pos (zero_real, ord_real) d) d
          (delta_gt floor_ceiling_real d) (less_eq_real one_reala d) false
          sqrt_real i
    | Int_carrier v ->
        faulty_non_inf_order _A3 _B3
          "only integers, rationals and reals are supported for non-inf orders"
    | Rat_carrier v ->
        faulty_non_inf_order _A3 _B3
          "only integers, rationals and reals are supported for non-inf orders"
    | Real_carrier v ->
        faulty_non_inf_order _A3 _B3
          "only integers, rationals and reals are supported for non-inf orders"
    | Arctic_carrier v ->
        faulty_non_inf_order _A3 _B3
          "only integers, rationals and reals are supported for non-inf orders"
    | Arctic_rat_carrier v ->
        faulty_non_inf_order _A3 _B3
          "only integers, rationals and reals are supported for non-inf orders"
    | Int_mat_carrier (v, va, vb) ->
        faulty_non_inf_order _A3 _B3
          "only integers, rationals and reals are supported for non-inf orders"
    | Rat_mat_carrier (v, va, vb) ->
        faulty_non_inf_order _A3 _B3
          "only integers, rationals and reals are supported for non-inf orders"
    | Real_mat_carrier (v, va, vb) ->
        faulty_non_inf_order _A3 _B3
          "only integers, rationals and reals are supported for non-inf orders"
    | Arctic_mat_carrier (v, va) ->
        faulty_non_inf_order _A3 _B3
          "only integers, rationals and reals are supported for non-inf orders"
    | Arctic_rat_mat_carrier (v, va) ->
        faulty_non_inf_order _A3 _B3
          "only integers, rationals and reals are supported for non-inf orders"
    | RPO (v, va) ->
        faulty_non_inf_order _A3 _B3
          "only integers, rationals and reals are supported for non-inf orders"
    | KBO (v, va) ->
        faulty_non_inf_order _A3 _B3
          "only integers, rationals and reals are supported for non-inf orders"
    | ACKBO (v, va) ->
        faulty_non_inf_order _A3 _B3
          "only integers, rationals and reals are supported for non-inf orders"
    | WPO (v, va) ->
        faulty_non_inf_order _A3 _B3
          "only integers, rationals and reals are supported for non-inf orders"
    | Max_poly v ->
        faulty_non_inf_order _A3 _B3
          "only integers, rationals and reals are supported for non-inf orders";;

let rec all_terms_impl (_A1, _A2)
  rr initt =
    remdups
      (equal_prod (equal_list (equal_term _A2 (equal_list equal_char)))
        (equal_term _A2 (equal_list equal_char)))
      (map (fun (ss, (t, _)) -> (ss, t)) initt @
        map (fun (l, a) -> (args l, a)) rr);;

let rec all_subterms_impl (_A1, _A2)
  rr initt =
    remdups
      (equal_prod (equal_list (equal_term _A2 (equal_list equal_char)))
        (equal_term _A2 (equal_list equal_char)))
      (maps (fun (ss, s) -> map (fun a -> (ss, a)) (supteq_list s))
        (all_terms_impl (_A1, _A2) rr initt));;

let rec everything_impl (_A1, _A2)
  rr initt =
    map (fun a -> Inl a)
      (maps (fun (ss, t) ->
              map (fun lr -> (ss, (t, lr)))
                (remdups
                  (equal_prod (equal_term _A2 (equal_list equal_char))
                    (equal_term _A2 (equal_list equal_char)))
                  (map (comp snd snd) initt)))
        (all_subterms_impl (_A1, _A2) rr initt)) @
      remdups
        (equal_sum
          (equal_prod (equal_list (equal_term _A2 (equal_list equal_char)))
            (equal_prod (equal_term _A2 (equal_list equal_char))
              (equal_prod (equal_term _A2 (equal_list equal_char))
                (equal_term _A2 (equal_list equal_char)))))
          (equal_prod (equal_prod _A2 equal_nat) equal_nat))
        (map (fun a -> Inr a)
          (maps (fun t ->
                  (if not (is_Var t)
                    then maps (fun (f, ts) ->
                                map (fun a -> ((f, size_list ts), a))
                                  (upt zero_nata (size_list ts)))
                           (let Fun (f, ts) = t in [(f, ts)])
                    else []))
            (remdups (equal_term _A2 (equal_list equal_char))
              (map snd (all_subterms_impl (_A1, _A2) rr initt)))));;

let rec generate_impl (_A1, _A2) _B _C
  rr nfq e_cap uu x4 = match rr, nfq, e_cap, uu, x4 with
    rr, nfq, e_cap, uu, Inr v -> []
    | rr, nfq, e_cap, uu, Inl (va, (Var ve, vd)) -> []
    | rr, nfq, e_cap, uu, Inl (ss, (Fun (f, ts), (l, r))) ->
        maps (fun i ->
               (if membera (equal_prod _B _C) (uu ss (nth ts i)) (l, r)
                 then map (fun u -> u)
                        [Inl (ss, (nth ts i, (l, r)));
                          Inr ((f, size_list ts), i)]
                 else []))
          (upt zero_nata (size_list ts)) @
          maps (fun (la, ra) ->
                 maps (fun mss ->
                        (if rule_match_impl _A2 nfq (e_cap mss) mss f
                              (map (map_term (fun x -> x)
                                     (fun a ->
                                       Chara
 (false, false, false, true, true, true, true, false) ::
 a))
                                ts)
                              la
                          then (if membera (equal_prod _B _C) (uu (args la) ra)
                                     (l, r)
                                 then [Inl (args la, (ra, (l, r)))] else [])
                          else []))
                   [map (map_term (fun x -> x)
                          (fun a ->
                            Chara (false, false, false, true, true, true, true,
                                    false) ::
                              a))
                      ss])
            rr;;

let rec mu_approx_impl (_A1, _A2)
  rr initt u_impl nfq e_cap =
    (let uu =
       precompute_fun
         (compare_order_prod
           (compare_order_list
             ((compare_order_term _A1.compare_compare_order
                (compare_list compare_char)),
               (equal_term _A2 (equal_list equal_char))))
           (compare_order_term _A1.compare_compare_order
             (compare_list compare_char)))
         (fun (a, b) -> u_impl a b) (all_subterms_impl (_A1, _A2) rr initt)
       in
     let uua = (fun s t -> uu (s, t)) in
     let fis =
       remdups (equal_prod (equal_prod _A2 equal_nat) equal_nat)
         (maps (fun entry ->
                 map (fun fi -> fi)
                   (match entry with Inl _ -> [] | Inr fi -> [fi]))
           (inductive_set_impl
             (equal_sum
               (equal_prod (equal_list (equal_term _A2 (equal_list equal_char)))
                 (equal_prod (equal_term _A2 (equal_list equal_char))
                   (equal_prod (equal_term _A2 (equal_list equal_char))
                     (equal_term _A2 (equal_list equal_char)))))
               (equal_prod (equal_prod _A2 equal_nat) equal_nat))
             (equal_sum
               (equal_prod (equal_list (equal_term _A2 (equal_list equal_char)))
                 (equal_prod (equal_term _A2 (equal_list equal_char))
                   (equal_prod (equal_term _A2 (equal_list equal_char))
                     (equal_term _A2 (equal_list equal_char)))))
               (equal_prod (equal_prod _A2 equal_nat) equal_nat))
             (everything_impl (_A1, _A2) rr initt)
             (equal_suma
               (equal_prod (equal_list (equal_term _A2 (equal_list equal_char)))
                 (equal_prod (equal_term _A2 (equal_list equal_char))
                   (equal_prod (equal_term _A2 (equal_list equal_char))
                     (equal_term _A2 (equal_list equal_char)))))
               (equal_prod (equal_prod _A2 equal_nat) equal_nat))
             (generate_impl (_A1, _A2) (equal_term _A2 (equal_list equal_char))
               (equal_term _A2 (equal_list equal_char)) rr nfq e_cap uua)
             (map (fun a -> Inl a) initt)))
       in
     let fs = remdups (equal_prod _A2 equal_nat) (map fst fis) in
     let mu =
       (fun f ->
         set (ceq_nat, ccompare_nat, set_impl_nat)
           (map_filter
             (fun x ->
               (if (let (g, _) = x in equal_proda _A2 equal_nat g f)
                 then Some (snd x) else None))
             fis))
       in
      (fs, (precompute_fun (compare_order_prod _A1 compare_order_nat) mu fs,
             "innermost URM wrt. specific rules")));;

let rec inn_usable_rules_wf (_A1, _A2)
  nfq e_cap r nfs =
    (fun (ss, t) ->
      (if nfs ||
            list_all
              (fun x ->
                list_ex (contains_var_term (equal_list equal_char) x) ss)
              (remdups (equal_list equal_char) (vars_term_list t))
        then ur_calc_singleton (_A1, _A2) nfq e_cap r (ss, t) else r));;

let rec inn_usable_rules_wf_dpp (_B1, _B2)
  i d nfs =
    inn_usable_rules_wf (_B1, _B2) (is_QNFc i d) (icap_impl_dpp _B2 i d)
      (rulese i d) nfs;;

let rec get_innermost_strict_repl_map_dpp (_B1, _B2)
  i d s =
    (let r = rulese i d in
     let p = pairsb i d in
     let isNF = is_QNFc i d in
     let u = inn_usable_rules_wf_dpp (_B1, _B2) i d true in
     let a = icap_impl_dpp _B2 i d in
      mu_approx_impl (_B1, _B2) r
        (maps (fun (sa, t) -> map (fun lr -> ([sa], (t, lr))) s) p)
        (fun ss t -> u (ss, t)) isNF a);;

let rec showsl_position_set
  f s = showsl_list_nat
          (maps (fun i ->
                  (if member (ceq_nat, ccompare_nat) i s then [suc i] else []))
            (upt zero_nata (snd f)));;

let rec mono_af
  (Redtriple_ext
    (valid, s, ns, nst, af, mono_af, mono, desc, not_ws_ns, cpx, more))
    = mono_af;;

let rec mono_urm_redpair_proc (_B1, _B2, _B3)
  i rp premove rremove dpp =
    (match
      catch_error
        (let (ps, pns) = split_pairsa i dpp premove in
         let (rs, rns) = split_rulesc i dpp rremove in
         let r = rulese i dpp in
         let _ = qc i dpp in
         let p = pairsb i dpp in
          bindb (check_wf_trs _B3
                  ((equal_list equal_char), (showl_list showl_char)) p)
            (fun _ ->
              bindb (check_wf_trs _B3
                      ((equal_list equal_char), (showl_list showl_char)) r)
                (fun _ ->
                  bindb (check (nFQ_subset_NF_rulesc i dpp)
                          (showsl_lit "innermost required"))
                    (fun _ ->
                      (let (fs, (mu, info)) =
                         get_innermost_strict_repl_map_dpp (_B1, _B2) i dpp rs
                         in
                        bindb (valid rp)
                          (fun _ ->
                            (let mua = mono_af rp in
                              bindb (catch_error
                                      (catch_error
(forallM
  (fun f ->
    check (subset (card_UNIV_nat, cenum_nat, ceq_nat, ccompare_nat) (mu f)
            (mua f))
      (comp (comp (comp (comp (comp (showsl_lit
                                      "error in monotonicity: strict order for ")
                                (showsl_prod _B3 showl_nat f))
                          (showsl_lit " ensures monotonicity in positions "))
                    (showsl_position_set f (mua f)))
              (showsl_lit "\092but usable replacement map is "))
        (showsl_position_set f (mu f))))
  fs)
(fun x -> Inl (snd x)))
                                      (fun x ->
Inl (comp (comp (comp (comp (comp x
                              (showsl_lit
                                "\092the computed usable replacement map ("))
                        (showsl_literal info))
                  (showsl_lit ") is\092"))
            (showsl_sep
              (fun f ->
                comp (comp (comp (showsl_lit "mu(")
                             (showsl_prod _B3 showl_nat f))
                       (showsl_lit ") = "))
                  (showsl_position_set f (mu f)))
              (showsl_literal "\092") fs))
      (showsl_lit "\092and mu(f) = {} for all other symbols f"))))
                                (fun _ ->
                                  bindb (catch_error
  (catch_error (forallM (ns rp) rns) (fun x -> Inl (snd x)))
  (fun x -> Inl (comp (showsl_lit "problem when orienting TRS\092") x)))
                                    (fun _ ->
                                      bindb
(catch_error (catch_error (forallM (s rp) rs) (fun x -> Inl (snd x)))
  (fun x -> Inl (comp (showsl_lit "problem when orienting TRS\092") x)))
(fun _ ->
  bindb (catch_error (catch_error (forallM (ns rp) pns) (fun x -> Inl (snd x)))
          (fun x -> Inl (comp (showsl_lit "problem when orienting DPs\092") x)))
    (fun _ ->
      catch_error (catch_error (forallM (s rp) ps) (fun x -> Inl (snd x)))
        (fun x ->
          Inl (comp (showsl_lit "problem when orienting DPs\092") x)))))))))))))
        (fun x ->
          Inl (comp (comp (comp (showsl_lit
                                  "could not apply the reduction pair processor with usable repl. maps and the following\092")
                            (desc rp))
                      (showsl_literal "\092"))
                x))
      with Inl a -> Inl a
      | Inr _ ->
        Inr (delete_R_Rwc i (delete_P_Pwa i dpp premove premove) rremove
              rremove));;

let rec check_strict_one_rstep (_A1, _A2) (_B1, _B2, _B3, _B4)
  ra rseqm p r =
    (let s = proj_term p (fst r) in
     let t = proj_term p (snd r) in
      (match rseqm r with None -> check_supt (_A1, _A2) (_B2, _B4) s t
        | Some [] ->
          Inl (showsl_lit "more than a single rewrite step is not allowed")
        | Some [(pos, (rule, u))] ->
          bindb (check_qrstep (_A1, _A2) (_B1, _B2, _B3, _B4) (fun _ -> true)
                  false ra pos rule s u)
            (fun _ -> check_supteq (_A1, _A2) (_B2, _B4) u t)
        | Some ((_, (_, _)) :: _ :: _) ->
          Inl (showsl_lit "more than a single rewrite step is not allowed")));;

let rec check_rsteps (_A1, _A2) (_B1, _B2, _B3, _B4)
  = check_qrsteps (_A1, _A2) (_B1, _B2, _B3, _B4) (fun _ -> true) false;;

let rec rseq_last s steps = last (s :: map (fun (_, (_, sa)) -> sa) steps);;

let rec check_rsteps_last (_A1, _A2) (_B1, _B2, _B3, _B4)
  = (fun r s steps ->
      check_rsteps (_A1, _A2) (_B1, _B2, _B3, _B4) r steps s
        (rseq_last s steps));;

let rec check_strict_rstep (_A1, _A2) (_B1, _B2, _B3, _B4)
  ra rseqm p r =
    (let s = proj_term p (fst r) in
     let t = proj_term p (snd r) in
      (match rseqm r with None -> check_supt (_A1, _A2) (_B2, _B4) s t
        | Some rseq ->
          (if equal_nata (size_list rseq) zero_nata
            then check_supt (_A1, _A2) (_B2, _B4) s t
            else bindb (check_rsteps_last (_A1, _A2) (_B1, _B2, _B3, _B4) ra s
                         rseq)
                   (fun _ ->
                     check_supteq (_A1, _A2) (_B2, _B4) (rseq_last s rseq)
                       t))));;

let rec create_rseq_map _A _B
  rseqs =
    ceta_map_of
      (compare_order_prod
        (compare_order_term _A.compare_compare_order _B.compare_compare_order)
        (compare_order_term _A.compare_compare_order _B.compare_compare_order))
      rseqs;;

let rec check_weak (_A1, _A2) (_B1, _B2)
  p r = catch_error
          (check
            (equal_terma _A1 _B1 (proj_term p (fst r)) (proj_term p (snd r)))
            (showsl_lit
              "the projected lhs is not equal to the projected rhs\092"))
          (fun x ->
            Inl (comp (comp (comp (comp (comp
  (comp (comp (showsl_lit "Could not orient rule ") (showsl_rule _A2 _B2 r))
    (showsl_lit ", since\092"))
  (showsl_terma _A2 _B2 (proj_term p (fst r))))
                                    (showsl_lit " != "))
                              (showsl_terma _A2 _B2 (proj_term p (snd r))))
                        (showsl_literal "\092"))
                  x));;

let rec subterm_criterion_proc (_B1, _B2, _B3) (_C1, _C2, _C3, _C4, _C5)
  i pL rseqmL prm dpp =
    (match
      (let p = create_proj _B1 pL in
       let rseqm = create_rseq_map _B1 _C2 rseqmL in
       let pa = pairsb i dpp in
       let _ = nfsc i dpp in
       let r = rulese i dpp in
       let pb = snd (split_pairsa i dpp prm) in
       let wfR = wf_rules_impl _C3 r in
        bindb (catch_error
                (forallM
                  (fun (l, ra) ->
                    bindb (check_no_var _B3 _C5 l)
                      (fun _ ->
                        bindb (check_no_var _B3 _C5 ra)
                          (fun _ ->
                            check_no_defined_root _B3 _C5
                              (fun fn -> not (null (rules_mapc i dpp fn))) ra)))
                  pa)
                (fun x -> Inl (snd x)))
          (fun _ ->
            bindb (check (minimal i dpp || nFQ_subset_NF_rulesc i dpp)
                    (showsl_lit "minimality or innermost required"))
              (fun _ ->
                bindb (catch_error
                        (forallM (fun (l, _) -> check_no_var _B3 _C5 l) r)
                        (fun x -> Inl (snd x)))
                  (fun _ ->
                    bindb (if q_emptyc i dpp
                            then catch_error
                                   (forallM
                                     (check_strict_rstep (_B2, _B3)
                                       (_C1, _C3, _C4, _C5) r rseqm p)
                                     prm)
                                   (fun x -> Inl (snd x))
                            else catch_error
                                   (forallM
                                     (check_strict_one_rstep (_B2, _B3)
                                       (_C1, _C3, _C4, _C5) wfR rseqm p)
                                     prm)
                                   (fun x -> Inl (snd x)))
                      (fun _ ->
                        catch_error
                          (forallM (check_weak (_B2, _B3) (_C3, _C5) p) pb)
                          (fun x -> Inl (snd x)))))))
      with Inl a -> Inl a | Inr _ -> Inr (delete_P_Pwa i dpp prm prm));;

let rec replace_paira
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = replace_pair;;

let rec forward_instantiation_proc (_B1, _B2, _B3)
  i st sts u_opt dpp =
    (match
      (let isnf = is_QNFc i dpp in
       let (s, t) = st in
       let iedg = is_iedg_edge_dpp (_B1, _B2, _B3) i dpp (s, t) in
       let sy =
         map_term (fun x -> x)
           (fun a ->
             Chara (true, false, false, true, true, true, true, false) :: a)
           s
         in
       let ty =
         map_term (fun x -> x)
           (fun a ->
             Chara (true, false, false, true, true, true, true, false) :: a)
           t
         in
       let u = (match u_opt with None -> rulese i dpp | Some u -> u) in
        bindb (if is_none u_opt then Inr ()
                else (let urc = is_ur_closed_impl_dpp_mv (_B1, _B2, _B3) i dpp u
                        in
                      let check_urc =
                        (fun sa ta ->
                          check (urc sa ta)
                            (comp (comp (showsl_lit "term ")
                                    (showsl_terma _B3 (showl_list showl_char)
                                      ta))
                              (showsl_lit " is not closed under usable rules")))
                        in
                       bindb (check (nfsc i dpp || minimal i dpp)
                               (showsl_lit
                                 "minimality or normal subst required"))
                         (fun _ ->
                           bindb (check (nFQ_subset_NF_rulesc i dpp)
                                   (showsl_lit "innermost rewriting required"))
                             (fun _ ->
                               bindb (catch_error
                                       (forallM
 (fun (l, a) -> check_urc (args l) a) u)
                                       (fun x -> Inl (snd x)))
                                 (fun _ ->
                                   bindb (check_urc [s] t)
                                     (fun _ ->
                                       (if nfsc i dpp then Inr ()
 else catch_error
        (check_subseteq (equal_list equal_char) (vars_term_list t)
          (vars_term_list s))
        (fun _ -> Inl (showsl_lit "variable condition in pair violated")))))))))
          (fun _ ->
            (let ur = map (fun (l, r) -> (r, l)) u in
             let ic =
               icap_impl _B2
                 (is_NF_terms (_B1, _B2)
                   ((ccompare_list ccompare_char), (equal_list equal_char),
                     mapping_impl_list)
                   [])
                 ur []
               in
              catch_error
                (forallM
                  (fun (ua, v) ->
                    (match mgu_class _B2 (ic ua) t with None -> Inr ()
                      | Some mu ->
                        check (not (isnf (subst_apply_term sy mu)) ||
                                (not (isnf (subst_apply_term
     (map_term (fun x -> x)
       (fun a ->
         Chara (false, false, false, true, true, true, true, false) :: a)
       ua)
     mu)) ||
                                  list_ex
                                    (fun sta ->
                                      instance_rule _B2 (equal_list equal_char)
((ccompare_list ccompare_char), (equal_list equal_char), mapping_impl_list) sta
st &&
instance_rule _B2 (equal_list equal_char)
  ((ccompare_list ccompare_char), (equal_list equal_char), mapping_impl_list)
  (subst_apply_term sy mu, subst_apply_term ty mu) sta)
                                    sts))
                          (comp (comp (comp
(showsl_lit "could not find instance of pair ")
(showsl_rule _B3 (showl_list showl_char)
  (subst_apply_term sy mu, subst_apply_term ty mu)))
                                  (showsl_lit "\092which resulted from DP "))
                            (showsl_rule _B3 (showl_list showl_char) (ua, v)))))
                  (filtera (fun (ua, _) -> iedg ua) (pairsb i dpp)))
                (fun x -> Inl (snd x)))))
      with Inl a -> Inl a | Inr _ -> Inr (replace_paira i dpp st sts));;

let rec check_join (_A1, _A2) (_B1, _B2, _B3, _B4)
  r s sseq t tseq =
    (let chk = check_rsteps_last (_A1, _A2) (_B1, _B2, _B3, _B4) r in
      bindb (chk s sseq)
        (fun _ ->
          bindb (chk t tseq)
            (fun _ ->
              (let u = rseq_last s sseq in
               let v = rseq_last t tseq in
                check (equal_terma _A1 _B2 u v)
                  (comp (comp (comp (showsl_lit
                                      "the rewrite sequences end in different terms ")
                                (showsl_terma _A2 _B4 u))
                          (showsl_lit " and "))
                    (showsl_terma _A2 _B4 v))))));;

let rec check_critical_pairs_guided (_A1, _A2) (_B1, _B2, _B3, _B4)
  r cp joins =
    bindb (catch_error
            (forallM
              (fun (s, a) ->
                (let (sseq, aa) = a in
                 let (ab, b) = aa in
                  check_join (_A1, _A2) (_B1, _B2, _B3, _B4) r s sseq ab b))
              joins)
            (fun x -> Inl (snd x)))
      (fun _ ->
        (let cpa = map (fun (s, (_, (t, _))) -> (s, t)) joins in
          catch_error
            (forallM
              (fun (_, (s, t)) ->
                check (equal_terma _A1 _B2 s t ||
                        list_ex
                          (fun st ->
                            instance_rule _A1 _B2 (_B1, _B2, _B3) (s, t) st ||
                              instance_rule _A1 _B2 (_B1, _B2, _B3) (t, s) st)
                          cpa)
                  (comp (comp (showsl_lit "critical pair ")
                          (showsl_crit_pair _A2 _B4 (s, t)))
                    (showsl_lit " missing")))
              cp)
            (fun x -> Inl (snd x))));;

let rec iterative_join_search_main _A (_B1, _B2, _B3)
  r s t i n =
    (if less_eq_nat i n
      then not (null (list_inter (equal_term _A _B2)
                       (reachable_terms _A (_B1, _B2, _B3) r s i)
                       (reachable_terms _A (_B1, _B2, _B3) r t i))) ||
             iterative_join_search_main _A (_B1, _B2, _B3) r s t (suc i) n
      else false);;

let rec iterative_join_search _A (_B1, _B2, _B3)
  r s t n = iterative_join_search_main _A (_B1, _B2, _B3) r s t zero_nata n;;

let rec check_join_BFS_limit (_A1, _A2) (_B1, _B2, _B3, _B4)
  n r s t =
    check (iterative_join_search _A1 (_B1, _B2, _B3) r s t n)
      (comp (comp (comp (comp (comp (comp (showsl_literal
    "could not find a joining sequence of length at most ")
                                      (showsl_nat n))
                                (showsl_literal " for the terms "))
                          (showsl_terma _A2 _B4 s))
                    (showsl_literal " and "))
              (showsl_terma _A2 _B4 t))
        (showsl_literal "\092"));;

let rec check_critical_pairs_BFS (_A1, _A2) (_B1, _B2, _B3, _B4)
  n r cp =
    catch_error
      (forallM
        (fun (_, (s, t)) ->
          (if equal_terma _A1 _B2 s t then Inr ()
            else catch_error
                   (check_join_BFS_limit (_A1, _A2) (_B1, _B2, _B3, _B4) n r s
                     t)
                   (fun x ->
                     Inl (comp (comp (comp (showsl_lit
     "problem when joining critical pair ")
                                       (showsl_crit_pair _A2 _B4 (s, t)))
                                 (showsl_literal "\092"))
                           x))))
        cp)
      (fun x -> Inl (snd x));;

let rec check_critical_pairs_NF (_A1, _A2) (_B1, _B2, _B3, _B4)
  r cp =
    catch_error
      (forallM
        (fun (_, (s, t)) ->
          (if equal_terma _A1 _B2 s t then Inr ()
            else catch_error
                   (check_join_NF (_A1, _A2) (_B1, _B2, _B3, _B4) r s t)
                   (fun x ->
                     Inl (comp (comp (comp (showsl_lit
     "problem when joining critical pair ")
                                       (showsl_crit_pair _A2 _B4 (s, t)))
                                 (showsl_literal "\092"))
                           x))))
        cp)
      (fun x -> Inl (snd x));;

let rec check_critical_pairs (_A1, _A2)
  r cp join_info =
    (match join_info
      with Guided a ->
        check_critical_pairs_guided (_A1, _A2)
          ((ccompare_list ccompare_char), (equal_list equal_char),
            mapping_impl_list, (showl_list showl_char))
          r cp a
      | Join_NF ->
        check_critical_pairs_NF (_A1, _A2)
          ((ccompare_list ccompare_char), (equal_list equal_char),
            mapping_impl_list, (showl_list showl_char))
          r cp
      | Join_BFS n ->
        check_critical_pairs_BFS (_A1, _A2)
          ((ccompare_list ccompare_char), (equal_list equal_char),
            mapping_impl_list, (showl_list showl_char))
          n r cp);;

let rec switch_innermost_proc (_B1, _B2)
  i joins_i dpp =
    (let r = rwd i dpp in
     let p = pb i dpp in
     let pw = pwb i dpp in
     let nfs = nfsc i dpp in
      (match
        bindb (catch_error
                (forallM
                  (fun (l, _) ->
                    check (not (is_Var l))
                      (showsl_lit "left variables in R forbidden"))
                  (if nfs then r else []))
                (fun x -> Inl (snd x)))
          (fun _ ->
            bindb (check (minimal i dpp) (showsl_lit "minimality required"))
              (fun _ ->
                bindb (check (null (qc i dpp))
                        (showsl_lit "non-empty Q not yet supported"))
                  (fun _ ->
                    bindb (check (null (rd i dpp))
                            (showsl_lit "strict rules not allowed"))
                      (fun _ ->
                        bindb (check (null (critical_pairs_impl _B1 (p @ pw) r))
                                (showsl_lit
                                  "overlaps between P and R not allowed"))
                          (fun _ ->
                            check_critical_pairs (_B1, _B2) r
                              (critical_pairs_impl _B1 r r) joins_i)))))
        with Inl a -> Inl a
        | Inr _ -> Inr (mkd i nfs true p pw (map fst r) [] r)));;

let rec mono_redpair_proc _B _C
  i rp premove rremove dpp =
    (match
      catch_error
        (let (ps, pns) = split_pairsa i dpp premove in
         let (rs, rns) = split_rulesc i dpp rremove in
          bindb (valid rp)
            (fun _ ->
              bindb (catch_error (mono rp (ps @ rs @ pns @ rns))
                      (fun x ->
                        Inl (comp (showsl_lit
                                    "problem with monotonicity of strict order\092")
                              x)))
                (fun _ ->
                  bindb (catch_error
                          (catch_error (forallM (ns rp) rns)
                            (fun x -> Inl (snd x)))
                          (fun x ->
                            Inl (comp (showsl_lit
"problem when orienting TRS\092")
                                  x)))
                    (fun _ ->
                      bindb (catch_error
                              (catch_error (forallM (s rp) rs)
                                (fun x -> Inl (snd x)))
                              (fun x ->
                                Inl (comp (showsl_lit
    "problem when orienting TRS\092")
                                      x)))
                        (fun _ ->
                          bindb (catch_error
                                  (catch_error (forallM (ns rp) pns)
                                    (fun x -> Inl (snd x)))
                                  (fun x ->
                                    Inl (comp
  (showsl_lit "problem when orienting DPs\092") x)))
                            (fun _ ->
                              catch_error
                                (catch_error (forallM (s rp) ps)
                                  (fun x -> Inl (snd x)))
                                (fun x ->
                                  Inl (comp
(showsl_lit "problem when orienting DPs\092") x))))))))
        (fun x ->
          Inl (comp (comp (comp (showsl_lit
                                  "could not apply the reduction pair processor with the following\092")
                            (desc rp))
                      (showsl_literal "\092"))
                x))
      with Inl a -> Inl a
      | Inr _ ->
        Inr (delete_R_Rwc i (delete_P_Pwa i dpp premove premove) rremove
              rremove));;

let rec mk_dpp i (nfs, (m, (p, (pw, (q, (r, rw)))))) = mkd i nfs m p pw q r rw;;

let rec check_assma (_B1, _B2, _B3)
  j i tp_check dp_check fptp_check unk_check x6 = match
    j, i, tp_check, dp_check, fptp_check, unk_check, x6 with
    j, i, tp_check, dp_check, fptp_check, unk_check, SN_assm_proof (t, prf) ->
      tp_check (mk_tp j t) prf
    | j, i, tp_check, dp_check, fptp_check, unk_check,
        Finite_assm_proof (t, prf)
        -> dp_check (mk_dpp i t) prf
    | j, i, tp_check, dp_check, fptp_check, unk_check,
        Unknown_assm_proof (t, prf)
        -> unk_check t prf
    | j, i, tp_check, dp_check, fptp_check, unk_check, SN_FP_assm_proof (t, prf)
        -> fptp_check t prf
    | j, i, uu, uv, uw, ux, Not_SN_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for non-termination assumptions in termination proof")
    | j, i, uu, uv, uw, ux, Infinite_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for non-termination assumptions in termination proof")
    | j, i, uu, uv, uw, ux, Not_RelSN_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for non-termination assumptions in termination proof")
    | j, i, uu, uv, uw, ux, Not_SN_FP_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for non-termination assumptions in termination proof")
    | j, i, uu, uv, uw, ux, Complexity_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for non-termination assumptions in termination proof");;

let rec quasi_splitter (_A1, _A2) (_C1, _C2)
  ld lAll uRw =
    (let unlab = (fun lf -> fst (ld lf)) in
     let la = map (fun r -> (r, map_funs_rule unlab r)) lAll in
     let (d, nD) =
       partition
         (fun (r, ur) ->
           equal_terma _A2 _C2 (fst ur) (snd ur) &&
             not (equal_terma _A2 _C2 (fst r) (snd r)))
         la
       in
     let (rw, r) =
       partition
         (fun (_, ur) ->
           member
             ((ceq_prod (ceq_term _A2 _C2) (ceq_term _A2 _C2)),
               (ccompare_prod (ccompare_term _A1 _C1) (ccompare_term _A1 _C1)))
             ur uRw)
         nD
       in
      (map fst r, (map fst rw, map fst d)));;

let rec model_splitter (_A1, _A2) (_C1, _C2)
  ld lAll uRw =
    (let unlab = (fun lf -> fst (ld lf)) in
     let la = map (fun r -> (r, map_funs_rule unlab r)) lAll in
     let (rw, r) =
       partition
         (fun (_, ur) ->
           member
             ((ceq_prod (ceq_term _A2 _C2) (ceq_term _A2 _C2)),
               (ccompare_prod (ccompare_term _A1 _C1) (ccompare_term _A1 _C1)))
             ur uRw)
         la
       in
      (map fst r, (map fst rw, [])));;

let rec check_sl_Q (_A1, _A2) (_C1, _C2, _C3, _C4)
  ld lQ q =
    (let u = (fun l -> fst (ld l)) in
      catch_error
        (forallM
          (fun lq ->
            check (let mlq = map_term u (fun x -> x) lq in
                    list_ex
                      (fun qa ->
                        matches _A1 _C2 (_C1, _C2, _C3) mlq qa &&
                          matches _A1 _C2 (_C1, _C2, _C3) qa mlq)
                      q)
              (comp (comp (showsl_lit "unlabeling ") (showsl_terma _A2 _C4 lq))
                (showsl_lit " yields a term not in Q")))
          lQ)
        (fun x -> Inl (snd x)));;

let rec sem_lab_quasi_root_proc (_A1, _A2, _A3)
  (_D1, _D2, _D3, _D4, _D5, _D6, _D7)
  ld i valid check_decra check_decr check_lhss_more check_lab_all
    check_lab_all_trs check_model_lab lPAll lQ lRAll dpp =
    (let r = rd i dpp in
     let rw = rwd i dpp in
     let pw = pwb i dpp in
     let p = pb i dpp in
     let nfs = nfsc i dpp in
     let m = minimal i dpp in
     let (lP, (lPw, _)) =
       model_splitter (_A1, _A2) (_D3, _D4) ld lPAll
         (set ((ceq_prod (ceq_term _A2 _D4) (ceq_term _A2 _D4)),
                (ccompare_prod (ccompare_term _A1 _D3) (ccompare_term _A1 _D3)),
                (set_impl_prod set_impl_term set_impl_term))
           pw)
       in
     let (lR, (lRw, d)) =
       quasi_splitter (_A1, _A2) (_D3, _D4) ld lRAll
         (set ((ceq_prod (ceq_term _A2 _D4) (ceq_term _A2 _D4)),
                (ccompare_prod (ccompare_term _A1 _D3) (ccompare_term _A1 _D3)),
                (set_impl_prod set_impl_term set_impl_term))
           rw)
       in
     let qempty = q_emptyc i dpp in
      (match
        bindb valid
          (fun _ ->
            bindb (check
                    (if nfs then (if not qempty then wwf_rulesa i dpp else true)
                      else true)
                    (showsl_lit "well formedness required"))
              (fun _ ->
                bindb (catch_error
                        (forallM
                          (fun (l, ra) ->
                            bindb (check_no_var _A3 _D7 l)
                              (fun _ ->
                                bindb (check_no_var _A3 _D7 ra)
                                  (fun _ ->
                                    check_no_defined_root _A3 _D7
                                      (fun fn ->
not (null (rules_mapc i dpp fn)))
                                      ra)))
                          (pairsb i dpp))
                        (fun x -> Inl (snd x)))
                  (fun _ ->
                    bindb (catch_error
                            (forallM (fun (l, _) -> check_no_var _A3 _D7 l)
                              (rulese i dpp))
                            (fun x -> Inl (snd x)))
                      (fun _ ->
                        (let q = qc i dpp in
                          bindb (if nfs && not qempty
                                  then check_wf_trs _A3 (_D4, _D7) d
                                  else Inr ())
                            (fun _ ->
                              bindb (check_decra d)
                                (fun _ ->
                                  bindb (check_decr d)
                                    (fun _ ->
                                      bindb
(catch_error
  (forallM
    (fun qa ->
      check (linear_term (_D1, _D2, _D6) qa)
        (showsl_lit "Q must not contain non-linear terms"))
    q)
  (fun x -> Inl (snd x)))
(fun _ ->
  catch_error
    (bindb (check_lhss_more lQ q)
      (fun _ ->
        bindb (check_sl_Q (_A2, _A3) (_D2, _D4, _D5, _D7) ld lQ q)
          (fun _ ->
            bindb (check_lab_all
                     (set ((ceq_prod (ceq_term _A2 _D4) (ceq_term _A2 _D4)),
                            (ccompare_prod (ccompare_term _A1 _D3)
                              (ccompare_term _A1 _D3)),
                            (set_impl_prod set_impl_term set_impl_term))
                       lP)
                    p)
              (fun _ ->
                bindb (check_lab_all
                         (set ((ceq_prod (ceq_term _A2 _D4) (ceq_term _A2 _D4)),
                                (ccompare_prod (ccompare_term _A1 _D3)
                                  (ccompare_term _A1 _D3)),
                                (set_impl_prod set_impl_term set_impl_term))
                           lPw)
                        pw)
                  (fun _ ->
                    bindb (check_model_lab
                             (set ((ceq_prod (ceq_term _A2 _D4)
                                     (ceq_term _A2 _D4)),
                                    (ccompare_prod (ccompare_term _A1 _D3)
                                      (ccompare_term _A1 _D3)),
                                    (set_impl_prod set_impl_term set_impl_term))
                               lR)
                            r)
                      (fun _ ->
                        bindb (check_model_lab
                                 (set ((ceq_prod (ceq_term _A2 _D4)
 (ceq_term _A2 _D4)),
(ccompare_prod (ccompare_term _A1 _D3) (ccompare_term _A1 _D3)),
(set_impl_prod set_impl_term set_impl_term))
                                   lRw)
                                rw)
                          (fun _ ->
                            bindb (check_lab_all_trs lR r)
                              (fun _ -> check_lab_all_trs lRw rw))))))))
    (fun x -> Inl (comp (showsl_lit "problem during labeling:\092") x)))))))))))
        with Inl a -> Inl a
        | Inr _ -> Inr (mkd i nfs m lP lPw lQ lR (lRw @ d))));;

let rec eval_lab
  i l lc alpha x4 = match i, l, lc, alpha, x4 with
    i, l, lc, alpha, Var x -> (alpha x, Var x)
    | i, l, lc, alpha, Fun (f, ts) ->
        (let clts = map (eval_lab i l lc alpha) ts in
         let cs = map fst clts in
         let c = i f cs in
         let lts = map snd clts in
          (c, Fun (lc f (size_list ts) (l f cs), lts)));;

let rec check_sl_rule_ass _A _B (_D1, _D2, _D3) (_E1, _E2, _E3)
  mc i la lc cge lR alpha (l, r) =
    (let cl_ll = eval_lab i la lc alpha l in
     let cr_lr = eval_lab i la lc alpha r in
      bindb (check (if mc then cge (fst cl_ll) (fst cr_lr) else true)
              (comp (comp (comp (comp (comp (showsl_lit "rule ")
(showsl_rule _A _E3 (l, r)))
                                  (showsl_lit
                                    " violates the model condition, [lhs] = "))
                            (showsl _B (fst cl_ll)))
                      (showsl_lit ", [rhs] = "))
                (showsl _B (fst cr_lr))))
        (fun _ ->
          check (member
                  ((ceq_prod (ceq_term _D2 _E2) (ceq_term _D2 _E2)),
                    (ccompare_prod (ccompare_term _D1 _E1)
                      (ccompare_term _D1 _E1)))
                  (snd cl_ll, snd cr_lr) lR)
            (comp (comp (showsl_lit "labeled rule ")
                    (showsl_rule _D3 _E3 (snd cl_ll, snd cr_lr)))
              (showsl_lit " missing"))));;

let rec check_sl_rule _A _B (_D1, _D2, _D3) (_E1, _E2, _E3)
  i l lc c cge mc lR lr =
    catch_error
      (forallM
        (fun alpha ->
          check_sl_rule_ass _A _B (_D1, _D2, _D3) (_E1, _E2, _E3) mc i l lc cge
            lR alpha lr)
        (map (fun_of _E2) (enum_vectors c (insert_vars_rule _E2 lr []))))
      (fun x -> Inl (snd x));;

let rec check_sl_model_lab_trs_set (_A1, _A2, _A3) _B (_D1, _D2, _D3)
  i l c cge labl lR r =
    catch_error
      (forallM
        (check_sl_rule _A3 _B (_A1, _A2, _A3) (_D1, _D2, _D3) i l labl c cge
          true lR)
        r)
      (fun x -> Inl (snd x));;

let rec check_sl_model_lab_trs (_A1, _A2, _A3) _B (_D1, _D2, _D3)
  i l c cge labl lR r =
    check_sl_model_lab_trs_set (_A1, _A2, _A3) _B (_D1, _D2, _D3) i l c cge labl
      lR r;;

let rec check_NF_vars_subset _A (_B1, _B2, _B3)
  qa q =
    catch_error
      (forallM
        (fun qaa -> check (list_ex (matches _A _B2 (_B1, _B2, _B3) qaa) q) qaa)
        qa)
      (fun x -> Inl (snd x));;

let rec flatten_term_enum
  = function Var x -> [Var x]
    | Fun (fs, ts) ->
        (let lts = map flatten_term_enum ts in
         let ss = concat_lists lts in
          maps (fun f -> map (fun a -> Fun (f, a)) ss) fs);;

let rec lab_lhss_more_impl
  lc lS_gen q =
    (let f_all = (fun (f, n) -> map (lc f n) (lS_gen f n)) in
      maps (fun qa -> flatten_term_enum (map_funs_term_wa f_all qa)) q);;

let rec check_sl_lab_lhss_more (_A1, _A2) (_C1, _C2, _C3, _C4)
  lc lS_gen lQ q =
    catch_error
      (check_NF_vars_subset _A1 (_C1, _C2, _C3) (lab_lhss_more_impl lc lS_gen q)
        lQ)
      (fun x ->
        Inl (comp (showsl_terma _A2 _C4 x)
              (showsl_lit " is missing in labeled Q")));;

let rec sl_check_decr
  (Sl_ops_ext
    (sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS, sl_lgen,
      sl_LS_gen, more))
    = sl_check_decr;;

let rec lab_root
  i la l lc alpha x5 = match i, la, l, lc, alpha, x5 with
    i, la, l, lc, alpha, Fun (f, ts) ->
      (let clts = map (eval_lab i la lc alpha) ts in
       let cs = map fst clts in
       let a = map snd clts in
        Fun (lc f (size_list ts) (l f cs), a))
    | uu, uv, uw, ux, uy, Var x -> Var x;;

let rec check_sl_rule_all_ass _A _B (_D1, _D2, _D3) (_E1, _E2, _E3)
  i lb la lc gen_smaller lR alpha x7 = match
    i, lb, la, lc, gen_smaller, lR, alpha, x7 with
    i, lb, la, lc, gen_smaller, lR, alpha, (l, Fun (f, ts)) ->
      (let ll = lab_root i lb la lc alpha l in
       let clts = map (eval_lab i lb lc alpha) ts in
       let lts = map snd clts in
       let ld = la f (map fst clts) in
       let n = size_list ts in
       let small = gen_smaller ld in
        catch_error
          (forallM
            (fun le ->
              check (member
                      ((ceq_prod (ceq_term _D2 _E2) (ceq_term _D2 _E2)),
                        (ccompare_prod (ccompare_term _D1 _E1)
                          (ccompare_term _D1 _E1)))
                      (ll, Fun (lc f n le, lts)) lR)
                (comp (comp (showsl_lit "labeled rule ")
                        (showsl_rule _D3 _E3 (ll, Fun (lc f n le, lts))))
                  (showsl_lit " missing")))
            small)
          (fun x -> Inl (snd x)))
    | i, lb, la, lc, gen_smaller, lR, alpha, (l, Var x) ->
        (let ll = lab_root i lb la lc alpha l in
         let lr = lab_root i lb la lc alpha (Var x) in
          check (member
                  ((ceq_prod (ceq_term _D2 _E2) (ceq_term _D2 _E2)),
                    (ccompare_prod (ccompare_term _D1 _E1)
                      (ccompare_term _D1 _E1)))
                  (ll, lr) lR)
            (comp (comp (showsl_lit "labeled rule ")
                    (showsl_rule _D3 _E3 (ll, lr)))
              (showsl_lit " missing")));;

let rec check_sl_rule_all _A _B (_D1, _D2, _D3) (_E1, _E2, _E3)
  i la l lc c gen_smaller lR lr =
    catch_error
      (forallM
        (fun alpha ->
          check_sl_rule_all_ass _A _B (_D1, _D2, _D3) (_E1, _E2, _E3) i la l lc
            gen_smaller lR alpha lr)
        (map (fun_of _E2) (enum_vectors c (insert_vars_rule _E2 lr []))))
      (fun x -> Inl (snd x));;

let rec check_sl_lab_all_trs _A _B (_D1, _D2, _D3) (_E1, _E2, _E3)
  i la l c gen labl lP p =
    catch_error
      (forallM
        (check_sl_rule_all _A _B (_D1, _D2, _D3) (_E1, _E2, _E3) i la l labl c
          gen lP)
        p)
      (fun x -> Inl (snd x));;

let rec check_wf_sym_F_all (_C1, _C2)
  lc ld ls =
    (fun (lf, n) ->
      (let (f, l) = ld lf in
        check (ls f n l && eq _C1 lf (lc f n l))
          (comp (comp (showsl_lit "labeled symbol ") (showsl _C2 lf))
            (showsl_lit " not allowed"))));;

let rec check_wf_terms_F_all (_C1, _C2)
  lc ld ls lt =
    (let lfs = insert_funas_term _C1 lt [] in
      catch_error (forallM (check_wf_sym_F_all (_C1, _C2) lc ld ls) lfs)
        (fun x -> Inl (snd x)));;

let rec check_Lab_all_trs (_A1, _A2) (_C1, _C2)
  lc ld ls lR r =
    catch_error
      (forallM
        (fun (l, ra) ->
          catch_error
            (bindb (check_wf_terms_F_all (_A1, _A2) lc ld ls ra)
              (fun _ ->
                check (membera
                        (equal_prod (equal_term _A1 _C1) (equal_term _A1 _C1)) r
                        (map_funs_rule (fun lf -> fst (ld lf)) (l, ra)))
                  (showsl_lit
                    "unlabeling of the rule does not yield original rule")))
            (fun x ->
              Inl (comp (comp (comp (showsl_lit "problem with labeled rule")
                                (showsl_rule _A2 _C2 (l, ra)))
                          (showsl_literal "\092"))
                    x)))
        lR)
      (fun x -> Inl (snd x));;

let rec sl_LS_gen
  (Sl_ops_ext
    (sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS, sl_lgen,
      sl_LS_gen, more))
    = sl_LS_gen;;

let rec sl_lgen
  (Sl_ops_ext
    (sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS, sl_lgen,
      sl_LS_gen, more))
    = sl_lgen;;

let rec lge_to_lgr _B
  lge ls f n =
    (let lSfn = ls f n in
     let lgefn = lge f n in
      (fun l la -> not (eq _B l la) && (lSfn l && (lSfn la && lgefn l la))));;

let rec check_sl_decr_rule _A _B _C _D
  lc ld ls lge x4 = match lc, ld, ls, lge, x4 with
    lc, ld, ls, lge, (Fun (lf, ts), Fun (lg, us)) ->
      (let (f, l1) = ld lf in
       let (g, l2) = ld lg in
       let n = size_list ts in
        eq _A f g &&
          (equal_lista (equal_term _C _D) ts us &&
            (eq _C lf (lc f n l1) &&
              (eq _C lg (lc f n l2) &&
                (ls f n l1 && (ls f n l2 && lge_to_lgr _B lge ls f n l1 l2))))))
    | uu, uv, uw, ux, (Var vb, va) -> false
    | uu, uv, uw, ux, (v, Var vb) -> false;;

let rec check_sl_decr _A _B (_C1, _C2) (_D1, _D2)
  lc ld ls lge d =
    catch_error
      (forallM
        (fun lr ->
          check (check_sl_decr_rule _A _B _C1 _D1 lc ld ls lge lr)
            (comp (showsl_rule _C2 _D2 lr)
              (showsl_lit " is not a decreasing rule")))
        d)
      (fun x -> Inl (snd x));;

let rec sl_La
  (Sl_ops_ext
    (sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS, sl_lgen,
      sl_LS_gen, more))
    = sl_L;;

let rec sl_LS
  (Sl_ops_ext
    (sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS, sl_lgen,
      sl_LS_gen, more))
    = sl_LSa;;

let rec sl_L
  (Sl_ops_ext
    (sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS, sl_lgen,
      sl_LS_gen, more))
    = sl_La;;

let rec sl_I
  (Sl_ops_ext
    (sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS, sl_lgen,
      sl_LS_gen, more))
    = sl_I;;

let rec sl_C
  (Sl_ops_ext
    (sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS, sl_lgen,
      sl_LS_gen, more))
    = sl_C;;

let rec insert_funas_args_term _A
  t fs = foldr (insert_funas_term _A) (args t) fs;;

let rec insert_funas_args_rule _A
  r fs =
    insert_funas_args_term _A (fst r) (insert_funas_args_term _A (snd r) fs);;

let rec insert_funas_args_trs _A trs = foldr (insert_funas_args_rule _A) trs;;

let rec option_to_list = function Some a -> [a]
                         | None -> [];;

let rec insert_roots_rule _A
  r fs =
    foldr (insertb (equal_prod _A equal_nat))
      (option_to_list (root (fst r)) @ option_to_list (root (snd r))) fs;;

let rec insert_roots_trs _A trs = foldr (insert_roots_rule _A) trs;;

let rec sem_lab_fin_quasi_root_proc (_A1, _A2, _A3) _B _C
  (_E1, _E2, _E3, _E4, _E5, _E6, _E7)
  lc ld cge lge i gen lPAll lQ lRAll dp =
    (let pairs = pairsb i dp in
      bindb (gen (list_union (equal_prod _A2 equal_nat)
                   (insert_funas_trs _A2 (rulese i dp) [])
                   (insert_funas_args_trs _A2 pairs []))
              (insert_roots_trs _A2 pairs []))
        (fun ops ->
          (let check_d = sl_check_decr ops in
           let check_da =
             check_sl_decr _A2 _B (_A2, _A3) (_E4, _E7) lc ld (sl_LS ops) lge in
           let check_q =
             check_sl_lab_lhss_more (_A2, _A3) (_E2, _E4, _E5, _E7) lc
               (sl_LS_gen ops)
             in
           let check_ml =
             check_sl_model_lab_trs (_A1, _A2, _A3) _C (_E3, _E4, _E7)
               (sl_I ops) (sl_L ops) (sl_C ops) cge lc
             in
           let check_l =
             check_sl_lab_all_trs _A3 _C (_A1, _A2, _A3) (_E3, _E4, _E7)
               (sl_I ops) (sl_L ops) (sl_La ops) (sl_C ops) (sl_lgen ops) lc
             in
           let check_la =
             check_Lab_all_trs (_A2, _A3) (_E4, _E7) lc ld (sl_LS ops) in
            sem_lab_quasi_root_proc (_A1, _A2, _A3)
              (_E1, _E2, _E3, _E4, _E5, _E6, _E7) ld i (Inr ()) check_d check_da
              check_q check_l check_la check_ml lPAll lQ lRAll dp)));;

let rec check_sl_rule_root _A _B (_D1, _D2, _D3) (_E1, _E2, _E3)
  i la l lc c lR lr =
    catch_error
      (forallM
        (fun alpha ->
          (let laa = lab_root i la l lc alpha in
           let lb = laa (fst lr) in
           let r = laa (snd lr) in
            check (member
                    ((ceq_prod (ceq_term _D2 _E2) (ceq_term _D2 _E2)),
                      (ccompare_prod (ccompare_term _D1 _E1)
                        (ccompare_term _D1 _E1)))
                    (lb, r) lR)
              (comp (comp (showsl_lit "labeled rule ")
                      (showsl_rule _D3 _E3 (lb, r)))
                (showsl_lit " is missing"))))
        (map (fun_of _E2) (enum_vectors c (insert_vars_rule _E2 lr []))))
      (fun x -> Inl (snd x));;

let rec check_sl_lab_root_trs _A _B (_D1, _D2, _D3) (_E1, _E2, _E3)
  i la l c labl lP p =
    catch_error
      (forallM
        (check_sl_rule_root _A _B (_D1, _D2, _D3) (_E1, _E2, _E3) i la l labl c
          lP)
        p)
      (fun x -> Inl (snd x));;

let rec sem_lab_root_proc (_A1, _A2, _A3) (_D1, _D2, _D3, _D4, _D5)
  ld i valid check_Q check_laba check_lab check_model_lab lPAll lQ lRAll dpp =
    (let r = rd i dpp in
     let rw = rwd i dpp in
     let pw = pwb i dpp in
     let p = pb i dpp in
     let nfs = nfsc i dpp in
     let m = minimal i dpp in
     let (lP, (lPw, _)) =
       model_splitter (_A1, _A2) (_D2, _D3) ld lPAll
         (set ((ceq_prod (ceq_term _A2 _D3) (ceq_term _A2 _D3)),
                (ccompare_prod (ccompare_term _A1 _D2) (ccompare_term _A1 _D2)),
                (set_impl_prod set_impl_term set_impl_term))
           pw)
       in
     let (lR, (lRw, _)) =
       model_splitter (_A1, _A2) (_D2, _D3) ld lRAll
         (set ((ceq_prod (ceq_term _A2 _D3) (ceq_term _A2 _D3)),
                (ccompare_prod (ccompare_term _A1 _D2) (ccompare_term _A1 _D2)),
                (set_impl_prod set_impl_term set_impl_term))
           rw)
       in
      (match
        bindb valid
          (fun _ ->
            bindb (catch_error
                    (forallM
                      (fun (l, ra) ->
                        bindb (check_no_var _A3 _D5 l)
                          (fun _ ->
                            bindb (check_no_var _A3 _D5 ra)
                              (fun _ ->
                                check_no_defined_root _A3 _D5
                                  (fun fn -> not (null (rules_mapc i dpp fn)))
                                  ra)))
                      (pairsb i dpp))
                    (fun x -> Inl (snd x)))
              (fun _ ->
                bindb (catch_error
                        (forallM (fun (l, _) -> check_no_var _A3 _D5 l)
                          (rulese i dpp))
                        (fun x -> Inl (snd x)))
                  (fun _ ->
                    (let q = qc i dpp in
                      catch_error
                        (bindb
                          (check
                            (if nfs
                              then (if not (q_emptyc i dpp)
                                     then wwf_rulesa i dpp else true)
                              else true)
                            (showsl_lit "well formedness required"))
                          (fun _ ->
                            bindb (check_Q lQ q)
                              (fun _ ->
                                bindb (check_sl_Q (_A2, _A3)
(_D1, _D3, _D4, _D5) ld lQ q)
                                  (fun _ ->
                                    bindb (check_laba
     (set ((ceq_prod (ceq_term _A2 _D3) (ceq_term _A2 _D3)),
            (ccompare_prod (ccompare_term _A1 _D2) (ccompare_term _A1 _D2)),
            (set_impl_prod set_impl_term set_impl_term))
       lP)
    p)
                                      (fun _ ->
bindb (check_laba
         (set ((ceq_prod (ceq_term _A2 _D3) (ceq_term _A2 _D3)),
                (ccompare_prod (ccompare_term _A1 _D2) (ccompare_term _A1 _D2)),
                (set_impl_prod set_impl_term set_impl_term))
           lPw)
        pw)
  (fun _ ->
    bindb (check_model_lab
             (set ((ceq_prod (ceq_term _A2 _D3) (ceq_term _A2 _D3)),
                    (ccompare_prod (ccompare_term _A1 _D2)
                      (ccompare_term _A1 _D2)),
                    (set_impl_prod set_impl_term set_impl_term))
               lR)
            r)
      (fun _ ->
        bindb (check_model_lab
                 (set ((ceq_prod (ceq_term _A2 _D3) (ceq_term _A2 _D3)),
                        (ccompare_prod (ccompare_term _A1 _D2)
                          (ccompare_term _A1 _D2)),
                        (set_impl_prod set_impl_term set_impl_term))
                   lRw)
                rw)
          (fun _ -> bindb (check_lab lR r) (fun _ -> check_lab lRw rw)))))))))
                        (fun x ->
                          Inl (comp (showsl_lit "problem during labeling:\092")
                                x))))))
        with Inl a -> Inl a | Inr _ -> Inr (mkd i nfs m lP lPw lQ lR lRw)));;

let rec lab_rule_ass
  i l lc alpha rule =
    (snd (eval_lab i l lc alpha (fst rule)),
      snd (eval_lab i l lc alpha (snd rule)));;

let rec lab_rule_list _E
  i l lc c lr =
    map (fun alpha -> lab_rule_ass i l lc alpha lr)
      (map (fun_of _E) (enum_vectors c (insert_vars_rule _E lr [])));;

let rec lab_trs_list _E i l lc c r = maps (lab_rule_list _E i l lc c) r;;

let rec check_sl_lab (_A1, _A2) (_D1, _D2)
  i l lc c lR r =
    bindb (check (not (null c)) (showsl_lit "carrier must be non-empty"))
      (fun _ ->
        catch_error
          (check_subseteq (equal_prod (equal_term _A1 _D1) (equal_term _A1 _D1))
            lR (lab_trs_list _D1 i l lc c r))
          (fun x ->
            Inl (comp (comp (showsl_lit "labeled rule ")
                        (showsl_rule _A2 _D2 x))
                  (showsl_lit " is not allowed"))));;

let rec lab_lhs_list _E
  i l lc c t =
    map (fun alpha -> snd (eval_lab i l lc alpha t))
      (map (fun_of _E) (enum_vectors c (insert_vars_term _E t [])));;

let rec lab_lhss_list _E i l lc c q = maps (lab_lhs_list _E i l lc c) q;;

let rec check_sl_Qa (_A1, _A2) (_D1, _D2, _D3, _D4)
  i l lc c lQ q =
    bindb (check (not (null c)) (showsl_lit "carrier must be non-empty"))
      (fun _ ->
        catch_error
          (check_NF_vars_subset _A1 (_D1, _D2, _D3)
            (lab_lhss_list _D2 i l lc c q) lQ)
          (fun x ->
            Inl (comp (comp (showsl_lit "labeled term ")
                        (showsl_terma _A2 _D4 x))
                  (showsl_lit " is missing"))));;

let rec sem_lab_fin_root_proc (_A1, _A2, _A3) (_D1, _D2, _D3, _D4, _D5)
  (_E1, _E2)
  lc ld i gen lPAll lQ lRAll dp =
    (let pairs = pairsb i dp in
      bindb (gen (list_union (equal_prod _A2 equal_nat)
                   (insert_funas_trs _A2 (rulese i dp) [])
                   (insert_funas_args_trs _A2 pairs []))
              (insert_roots_trs _A2 pairs []))
        (fun ops ->
          (let check_q =
             check_sl_Qa (_A2, _A3) (_D1, _D3, _D4, _D5) (sl_I ops) (sl_L ops)
               lc (sl_C ops)
             in
           let check_ml =
             check_sl_model_lab_trs (_A1, _A2, _A3) _E2 (_D2, _D3, _D5)
               (sl_I ops) (sl_L ops) (sl_C ops) (eq _E1) lc
             in
           let check_l =
             check_sl_lab_root_trs _A3 _E2 (_A1, _A2, _A3) (_D2, _D3, _D5)
               (sl_I ops) (sl_L ops) (sl_La ops) (sl_C ops) lc
             in
           let check_la =
             check_sl_lab (_A2, _A3) (_D3, _D5) (sl_I ops) (sl_L ops) lc
               (sl_C ops)
             in
            sem_lab_root_proc (_A1, _A2, _A3) (_D1, _D2, _D3, _D4, _D5) ld i
              (Inr ()) check_q check_l check_la check_ml lPAll lQ lRAll dp)));;

let rec slm_La
  (Slm_ops_ext (slm_La, slm_I, slm_C, slm_c, slm_L, more)) = slm_L;;

let rec slm_c (Slm_ops_ext (slm_La, slm_I, slm_C, slm_c, slm_L, more)) = slm_c;;

let rec slm_L
  (Slm_ops_ext (slm_La, slm_I, slm_C, slm_c, slm_L, more)) = slm_La;;

let rec slm_I (Slm_ops_ext (slm_La, slm_I, slm_C, slm_c, slm_L, more)) = slm_I;;

let rec slm_C (Slm_ops_ext (slm_La, slm_I, slm_C, slm_c, slm_L, more)) = slm_C;;

let rec slm_to_sl
  ops = Sl_ops_ext
          (slm_L ops, (fun _ _ _ -> true), slm_I ops, slm_C ops, slm_c ops,
            (fun _ -> Inr ()), slm_La ops, (fun _ _ _ -> true), (fun l -> [l]),
            (fun _ _ -> []), ());;

let rec slm_gen_to_sl_gen
  gen = (fun f g -> bindb (gen f g) (fun ops -> Inr (slm_to_sl ops)));;

let rec check_sl_lab_trs_set (_A1, _A2, _A3) _B (_D1, _D2, _D3)
  i l c cge labl lP p =
    catch_error
      (forallM
        (check_sl_rule _A3 _B (_A1, _A2, _A3) (_D1, _D2, _D3) i l labl c cge
          false lP)
        p)
      (fun x -> Inl (snd x));;

let rec check_sl_lab_trs (_A1, _A2, _A3) _B (_D1, _D2, _D3)
  i l c cge labl lP p =
    check_sl_lab_trs_set (_A1, _A2, _A3) _B (_D1, _D2, _D3) i l c cge labl lP
      p;;

let rec sem_lab_proc (_A1, _A2, _A3) (_D1, _D2, _D3, _D4, _D5)
  ld i valid check_Q check_laba check_lab check_model_lab lPAll lQ lRAll dpp =
    (let r = rd i dpp in
     let rw = rwd i dpp in
     let pw = pwb i dpp in
     let p = pb i dpp in
     let nfs = nfsc i dpp in
     let m = minimal i dpp in
     let (lP, (lPw, _)) =
       model_splitter (_A1, _A2) (_D2, _D3) ld lPAll
         (set ((ceq_prod (ceq_term _A2 _D3) (ceq_term _A2 _D3)),
                (ccompare_prod (ccompare_term _A1 _D2) (ccompare_term _A1 _D2)),
                (set_impl_prod set_impl_term set_impl_term))
           pw)
       in
     let (lR, (lRw, _)) =
       model_splitter (_A1, _A2) (_D2, _D3) ld lRAll
         (set ((ceq_prod (ceq_term _A2 _D3) (ceq_term _A2 _D3)),
                (ccompare_prod (ccompare_term _A1 _D2) (ccompare_term _A1 _D2)),
                (set_impl_prod set_impl_term set_impl_term))
           rw)
       in
      (match
        bindb valid
          (fun _ ->
            (let q = qc i dpp in
              catch_error
                (bindb
                  (check
                    (if nfs
                      then (if not (q_emptyc i dpp) then wwf_rulesa i dpp
                             else true)
                      else true)
                    (showsl_lit "well formedness required"))
                  (fun _ ->
                    bindb (check_Q lQ q)
                      (fun _ ->
                        bindb (check_sl_Q (_A2, _A3) (_D1, _D3, _D4, _D5) ld lQ
                                q)
                          (fun _ ->
                            bindb (check_laba
                                     (set ((ceq_prod (ceq_term _A2 _D3)
     (ceq_term _A2 _D3)),
    (ccompare_prod (ccompare_term _A1 _D2) (ccompare_term _A1 _D2)),
    (set_impl_prod set_impl_term set_impl_term))
                                       lP)
                                    p)
                              (fun _ ->
                                bindb (check_laba
 (set ((ceq_prod (ceq_term _A2 _D3) (ceq_term _A2 _D3)),
        (ccompare_prod (ccompare_term _A1 _D2) (ccompare_term _A1 _D2)),
        (set_impl_prod set_impl_term set_impl_term))
   lPw)
pw)
                                  (fun _ ->
                                    bindb (check_model_lab
     (set ((ceq_prod (ceq_term _A2 _D3) (ceq_term _A2 _D3)),
            (ccompare_prod (ccompare_term _A1 _D2) (ccompare_term _A1 _D2)),
            (set_impl_prod set_impl_term set_impl_term))
       lR)
    r)
                                      (fun _ ->
bindb (check_model_lab
         (set ((ceq_prod (ceq_term _A2 _D3) (ceq_term _A2 _D3)),
                (ccompare_prod (ccompare_term _A1 _D2) (ccompare_term _A1 _D2)),
                (set_impl_prod set_impl_term set_impl_term))
           lRw)
        rw)
  (fun _ -> bindb (check_lab lR r) (fun _ -> check_lab lRw rw)))))))))
                (fun x ->
                  Inl (comp (showsl_lit "problem during labeling:\092") x))))
        with Inl a -> Inl a | Inr _ -> Inr (mkd i nfs m lP lPw lQ lR lRw)));;

let rec sem_lab_fin_proc (_A1, _A2, _A3) (_D1, _D2, _D3, _D4, _D5) (_E1, _E2)
  lc ld i gen lPAll lQ lRAll dp =
    bindb (gen (list_union (equal_prod _A2 equal_nat)
                 (insert_funas_trs _A2 (rulese i dp) [])
                 (insert_funas_args_trs _A2 (pairsb i dp) []))
            [])
      (fun ops ->
        (let check_q =
           check_sl_Qa (_A2, _A3) (_D1, _D3, _D4, _D5) (sl_I ops) (sl_L ops) lc
             (sl_C ops)
           in
         let check_ml =
           check_sl_model_lab_trs (_A1, _A2, _A3) _E2 (_D2, _D3, _D5) (sl_I ops)
             (sl_L ops) (sl_C ops) (eq _E1) lc
           in
         let check_l =
           check_sl_lab_trs (_A1, _A2, _A3) _E2 (_D2, _D3, _D5) (sl_I ops)
             (sl_L ops) (sl_C ops) (eq _E1) lc
           in
         let check_la =
           check_sl_lab (_A2, _A3) (_D3, _D5) (sl_I ops) (sl_L ops) lc
             (sl_C ops)
           in
          sem_lab_proc (_A1, _A2, _A3) (_D1, _D2, _D3, _D4, _D5) ld i (Inr ())
            check_q check_l check_la check_ml lPAll lQ lRAll dp));;

let rec get_largest_element (SL_Inter (n, uu)) = n;;

let rec take_default
  def x1 uu = match def, x1, uu with def, [], uu -> def
    | uv, x :: xs, i ->
        (if equal_nata i zero_nata then x
          else take_default uv xs (minus_nata i one_nata));;

let rec eval_arithFun c nats f = modulo_nata (eval_arithFun_unbound c nats f) c
and eval_arithFun_unbound
  c nats x2 = match c, nats, x2 with
    c, nats, Arg i -> take_default zero_nata nats i
    | c, nats, Const n -> n
    | c, nats, Sum [] -> zero_nata
    | c, nats, Sum (f :: fs) ->
        plus_nata (eval_arithFun c nats f) (eval_arithFun c nats (Sum fs))
    | c, nats, Prod [] -> one_nata
    | c, nats, Prod (f :: fs) ->
        times_nata (eval_arithFun c nats f) (eval_arithFun c nats (Prod fs))
    | c, nats, Max [f] -> eval_arithFun c nats f
    | c, nats, Max (f :: v :: va) ->
        max ord_nat (eval_arithFun c nats f)
          (eval_arithFun c nats (Max (v :: va)))
    | c, nats, Min [f] -> eval_arithFun c nats f
    | c, nats, Min (f :: v :: va) ->
        min ord_nat (eval_arithFun c nats f)
          (eval_arithFun c nats (Min (v :: va)))
    | c, nats, IfEqual (f1, f2, ft, fe) ->
        (if equal_nata (eval_arithFun c nats f1) (eval_arithFun c nats f2)
          then eval_arithFun c nats ft else eval_arithFun c nats fe);;

let rec sl_inter_to_inter _A
  (SL_Inter (c, ls)) fl cs =
    (match map_of (equal_prod _A equal_nat) ls (fl, size_list cs)
      with None -> zero_nata | Some a -> eval_arithFun (suc c) cs a);;

let rec sli_to_slm _A
  sli = (let c = get_largest_element sli in
          Slm_ops_ext
            ((fun _ -> (fun a -> Inl a)),
              sl_inter_to_inter (equal_lab _A (equal_list equal_nat)) sli,
              upt zero_nata (suc c), c, (fun _ -> (fun a -> Inl a)), ()));;

let rec enum_vectors_nat
  c n = (if equal_nata n zero_nata then [[]]
          else (let a = enum_vectors_nat c (minus_nata n one_nata) in
                 maps (fun vec -> map (fun ca -> ca :: vec) c) a));;

let rec qmodel_check_interpretation
  f n c =
    (let ca = upt zero_nata (suc c) in
     let css = enum_vectors_nat ca n in
      catch_error
        (forallM
          (fun cs ->
            catch_error
              (forallM
                (fun i ->
                  catch_error
                    (forallM
                      (fun l ->
                        check (less_eq_nat (eval_arithFun (suc c) cs f)
                                (eval_arithFun (suc c) (list_update cs i l) f))
                          (comp (comp (showsl_lit "not monotone in ")
                                  (showsl_nat (suc i)))
                            (showsl_lit ". argument")))
                      (upt (nth cs i) (suc c)))
                    (fun x -> Inl (snd x)))
                (upt zero_nata n))
              (fun x -> Inl (snd x)))
          css)
        (fun x -> Inl (snd x)));;

let rec qmodel_check_valid _A
  (SL_Inter (c, ls)) =
    catch_error
      (forallM
        (fun (a, b) ->
          (let (f, n) = a in
            (fun g ->
              catch_error (qmodel_check_interpretation g n c)
                (fun x ->
                  Inl (comp (comp (comp (showsl_lit
  "problem in weak-monotonicity of interpretation of ")
                                    (showsl _A f))
                              (showsl_literal "\092"))
                        x))))
            b)
        ls)
      (fun x -> Inl (snd x));;

let rec check_decr_present_aux_1 _A (_B1, _B2, _B3)
  r v f1 f2 n =
    (let vs = map (fun na -> Var (v, na)) (upt zero_nata n) in
     let rule = (Fun (f1, vs), Fun (f2, vs)) in
      check (not (is_none
                   (find (instance_rule _A (equal_prod _B2 equal_nat)
                           ((ccompare_prod _B1 ccompare_nat),
                             (equal_prod _B2 equal_nat),
                             (mapping_impl_prod _B3 mapping_impl_nat))
                           rule)
                     r)))
        rule);;

let rec check_decr_present_aux_2 _A (_B1, _B2, _B3)
  r v req =
    (let add_nats = map_term (fun x -> x) (fun va -> (va, zero_nata)) in
     let ra = map (fun (l, ra) -> (add_nats l, add_nats ra)) r in
      catch_error
        (forallM
          (fun (f1, a) ->
            (let (aa, b) = a in
              check_decr_present_aux_1 _A (_B1, _B2, _B3) ra v f1 aa b))
          req)
        (fun x -> Inl (snd x)));;

let rec check_decr_present _A (_B1, _B2, _B3)
  siga l v c r =
    (let ca = upt zero_nata (suc c) in
     let ls =
       (fun (f, n) ->
         maps (fun cs ->
                maps (fun i ->
                       (let ci = nth cs i in
                         (if less_nat ci c
                           then [(l f (list_update cs i (suc ci)), (l f cs, n))]
                           else [])))
                  (upt zero_nata n))
           (enum_vectors_nat ca n))
       in
      check_decr_present_aux_2 _A (_B1, _B2, _B3) r v (maps ls siga));;

let rec qmodel_check_decr (_A1, _A2) (_B1, _B2, _B3)
  siga v c =
    (fun lR ->
      catch_error
        (check_decr_present (equal_lab _A1 (equal_list equal_nat))
          (_B1, _B2, _B3) siga (fun a b -> Lab (a, b)) v c lR)
        (fun x ->
          Inl (let display =
                 map_term (fun xa -> xa)
                   (fun (_, n) ->
                     comp (shows_prec_list show_char zero_nata
                            [Chara (false, false, false, true, true, true, true,
                                     false)])
                       (shows_prec_nat zero_nata n) [])
                 in
                comp (comp (showsl_lit "decreasing rule ")
                       (showsl_rule (showl_lab _A2 (showl_list showl_nat))
                         (showl_list showl_char)
                         (display (fst x), display (snd x))))
                  (showsl_lit " missing"))));;

let rec qmodel_LS_gen _A
  siga ls =
    (fun f n ->
      (if membera (equal_prod _A equal_nat) siga (f, n)
        then map (fun a -> Inl a) (enum_vectors_nat ls n) else [Inl []]));;

let rec pointwise_lgen
  ns = concat_lists (map (fun n -> upt zero_nata (suc n)) ns);;

let rec qmodel_lgen
  l = (match l with Inl ns -> map (fun a -> Inl a) (pointwise_lgen ns)
        | Inr _ -> []);;

let rec qmodel_LSa
  x = (fun _ _ a -> (match a with Inl _ -> true | Inr _ -> false)) x;;

let rec qmodel_LS _A _B _C
  siga ls =
    (fun f n ->
      membera (equal_sum (equal_list _B) _C) (qmodel_LS_gen _A siga ls f n));;

let rec qmodel_L _A _B
  siga =
    (fun f cs ->
      (if membera (equal_prod (equal_lab _A _B) equal_nat) siga
            (f, size_list cs)
        then Inl cs else Inl []));;

let rec qsli_to_sl_unsafe (_A1, _A2, _A3) (_B1, _B2)
  v f g sli =
    (let c = get_largest_element sli in
     let ca = upt zero_nata (suc c) in
      Sl_ops_ext
        (qmodel_L _B1 (equal_list equal_nat) f,
          qmodel_LS (equal_lab _B1 (equal_list equal_nat)) equal_nat
            (equal_list (equal_lab _B1 (equal_list equal_nat))) f ca,
          sl_inter_to_inter (equal_lab _B1 (equal_list equal_nat)) sli, ca, c,
          qmodel_check_decr (_B1, _B2) (_A1, _A2, _A3) f v c,
          qmodel_L _B1 (equal_list equal_nat) g, qmodel_LSa, qmodel_lgen,
          qmodel_LS_gen (equal_lab _B1 (equal_list equal_nat)) f ca, ()));;

let rec qsli_to_sl (_A1, _A2, _A3) (_B1, _B2)
  v f g sli =
    bindb (qmodel_check_valid (showl_lab _B2 (showl_list showl_nat)) sli)
      (fun _ -> Inr (qsli_to_sl_unsafe (_A1, _A2, _A3) (_B1, _B2) v f g sli));;

let rec pointwise_ext
  f x1 x2 = match f, x1, x2 with f, [], [] -> (false, true)
    | f, uu :: uv, [] -> (false, false)
    | f, [], uw :: ux -> (false, false)
    | f, a :: asa, b :: bs ->
        (match f a b
          with (stri, true) ->
            (match pointwise_ext f asa bs
              with (strir, true) -> (stri || strir, true)
              | (_, false) -> (false, false))
          | (_, false) -> (false, false));;

let rec qmodel_lge
  f n = (fun l r ->
          (match (l, r)
            with (Inl cs1, Inl cs2) ->
              snd (pointwise_ext (fun x y -> (less_nat y x, less_eq_nat y x))
                    cs1 cs2)
            | (Inl _, Inr _) -> false | (Inr _, _) -> false));;

let rec qmodel_cge x = (fun xa y -> less_eq_nat y xa) x;;

let rec rl_slm _A _B
  delt_opt pre_fs g =
    (let fs =
       (if is_none delt_opt then pre_fs
         else filtera
                (fun f ->
                  not (equal_proda (equal_lab _A _B) equal_nat f
                        (the delt_opt)))
                pre_fs)
       in
      bindb (check (not (null fs))
              (showsl_lit
                "root-labeling requires at least one function symbol in the signature\092"))
        (fun _ ->
          (let f = fst (hda fs) in
            Inr (Slm_ops_ext
                  ((fun _ -> (fun a -> Inr a)),
                    (fun ga cs ->
                      (if membera (equal_prod (equal_lab _A _B) equal_nat) fs
                            (ga, size_list cs)
                        then ga else f)),
                    map fst fs, f,
                    (if is_none delt_opt then (fun _ -> (fun a -> Inr a))
                      else (fun _ gs ->
                             Inr (replicate (size_list gs)
                                   (fst (the delt_opt))))),
                    ())))));;

let rec semlab_fin_proc (_B1, _B2, _B3) (_C1, _C2, _C3, _C4, _C5, _C6, _C7)
  j x1 = match j, x1 with
    j, Rootlab None ->
      sem_lab_fin_proc
        ((compare_lab _B1 (compare_list compare_nat)),
          (equal_lab _B2 (equal_list equal_nat)),
          (showl_lab _B3 (showl_list showl_nat)))
        (_C2, _C3, _C4, _C5, _C7)
        ((equal_lab _B2 (equal_list equal_nat)),
          (showl_lab _B3 (showl_list showl_nat)))
        label label_decomp j
        (slm_gen_to_sl_gen (rl_slm _B2 (equal_list equal_nat) None))
    | j, Rootlab (Some d) ->
        sem_lab_fin_root_proc
          ((compare_lab _B1 (compare_list compare_nat)),
            (equal_lab _B2 (equal_list equal_nat)),
            (showl_lab _B3 (showl_list showl_nat)))
          (_C2, _C3, _C4, _C5, _C7)
          ((equal_lab _B2 (equal_list equal_nat)),
            (showl_lab _B3 (showl_list showl_nat)))
          label label_decomp j
          (slm_gen_to_sl_gen (rl_slm _B2 (equal_list equal_nat) (Some d)))
    | j, Finitelab sli ->
        sem_lab_fin_proc
          ((compare_lab _B1 (compare_list compare_nat)),
            (equal_lab _B2 (equal_list equal_nat)),
            (showl_lab _B3 (showl_list showl_nat)))
          (_C2, _C3, _C4, _C5, _C7) (equal_nat, showl_nat) label label_decomp j
          (slm_gen_to_sl_gen (fun _ _ -> Inr (sli_to_slm _B2 sli)))
    | j, QuasiFinitelab (sli, v) ->
        sem_lab_fin_quasi_root_proc
          ((compare_lab _B1 (compare_list compare_nat)),
            (equal_lab _B2 (equal_list equal_nat)),
            (showl_lab _B3 (showl_list showl_nat)))
          (equal_sum (equal_list equal_nat)
            (equal_list (equal_lab _B2 (equal_list equal_nat))))
          showl_nat (_C1, _C2, _C3, _C4, _C5, _C6, _C7) label label_decomp
          qmodel_cge qmodel_lge j
          (fun f g -> qsli_to_sl (_C2, _C4, _C5) (_B2, _B3) v f g sli);;

let rec partition_rules _A _B
  cs r =
    partition
      (fun lr ->
        list_ex
          (fun (u, v) ->
            list_ex
              (fun c ->
                equal_proda (equal_term _A _B) (equal_term _A _B) lr
                  (ctxt_apply_term c u, ctxt_apply_term c v))
              (Hole :: cs))
          r);;

let rec strip_unary_root _A
  f x1 = match f, x1 with
    f, Fun (g, [t]) -> (if eq _A f g then t else Fun (g, [t]))
    | f, Var v -> Var v
    | f, Fun (v, []) -> Fun (v, [])
    | f, Fun (v, vb :: vd :: ve) -> Fun (v, vb :: vd :: ve);;

let rec has_unary_root _A f x1 = match f, x1 with f, Fun (g, [t]) -> eq _A f g
                            | f, Var v -> false
                            | f, Fun (v, []) -> false
                            | f, Fun (v, vb :: vd :: ve) -> false;;

let rec unblock_term _A
  f x1 = match f, x1 with
    f, Fun (g, ts) ->
      (if list_all (has_unary_root _A f) ts
        then Fun (g, map (strip_unary_root _A f) ts) else Fun (g, ts))
    | f, Var v -> Var v;;

let rec unblock_rule _A
  f r = (unblock_term _A f (fst r), unblock_term _A f (snd r));;

let rec partition_pairs _A _B
  f p = partition
          (fun r ->
            membera (equal_prod (equal_term _A _B) (equal_term _A _B)) p
              (unblock_rule _A f r));;

let rec check_no_defined_root_defined (_A1, _A2) _B
  f t = check (not (membera (equal_prod _A1 equal_nat) f (the (root t))))
          (comp (comp (showsl_lit "the root of ") (showsl_terma _A2 _B t))
            (showsl_lit " is defined"));;

let rec block_term
  f x1 = match f, x1 with f, Var x -> Var x
    | f, Fun (g, ts) -> Fun (g, map (fun t -> Fun (f, [t])) ts);;

let rec block_rule f r = (block_term f (fst r), block_term f (snd r));;

let rec check_superset_of_blocked (_A1, _A2) (_B1, _B2)
  f pa p =
    catch_error
      (catch_error
        (catch_error
          (forallM
            (fun x ->
              (if membera (equal_prod (equal_term _A1 _B1) (equal_term _A1 _B1))
                    pa (block_rule f x)
                then Inr () else Inl x))
            p)
          (fun x -> Inl (snd x)))
        (fun x ->
          Inl (comp (comp (showsl_lit "the rule ")
                      (showsl_rule _A2 _B2 (block_rule f x)))
                (showsl_lit " is missing\092"))))
      (fun x ->
        Inl (comp (comp (comp (comp (showsl_trs _A2 _B2 p)
                                (showsl_lit "is not a subset of"))
                          (showsl_trs _A2 _B2 pa))
                    x)
              (showsl_literal "\092")));;

let rec hole_at _A
  n i f x3 = match n, i, f, x3 with
    n, i, f, More (g, ss1, Hole, ss2) ->
      eq _A g f &&
        (equal_nata (size_list ss1) i &&
          equal_nata (size_list ss2) (minus_nata (minus_nata n i) one_nata))
    | n, i, f, Hole -> false
    | n, i, f, More (v, va, More (vd, ve, vf, vg), vc) -> false;;

let rec check_flat_ctxt_complete (_A1, _A2) _B
  fcs fa =
    check (all_interval_nat
            (fun i -> list_ex (hole_at _A1 (snd fa) i (fst fa)) fcs) zero_nata
            (snd fa))
      (showsl_lit "the list of flat contexts is incomplete\092");;

let rec check_rule_reflecting (_A1, _A2) (_B1, _B2)
  fcs rs rule =
    check (list_ex
            (fun (l, r) ->
              list_ex
                (fun c ->
                  equal_terma _A1 _B1 (fst rule) (ctxt_apply_term c l) &&
                    equal_terma _A1 _B1 (snd rule) (ctxt_apply_term c r))
                (Hole :: fcs))
            rs)
      (comp (comp (showsl_lit "the rule ") (showsl_rule _A2 _B2 rule))
        (showsl_lit
          " is neither contained in the original set of rules nor obtained by applying a flat context\092"));;

let rec check_rule_preserving (_A1, _A2) (_B1, _B2, _B3, _B4)
  fcs rs rule =
    check (list_ex (instance_rule _A1 _B2 (_B1, _B2, _B3) rule) rs ||
            list_all
              (fun c ->
                list_ex
                  (instance_rule _A1 _B2 (_B1, _B2, _B3)
                    (ctxt_apply_term c (fst rule),
                      ctxt_apply_term c (snd rule)))
                  rs)
              fcs)
      (comp (comp (showsl_lit "the rule ") (showsl_rule _A2 _B4 rule))
        (showsl_lit
          " is neither contained in the resulting set of rules nor closed under all flat contexts\092"));;

let rec is_flat_ctxt_list _A _B
  vs fas x2 = match vs, fas, x2 with
    vs, fas, More (f, ss1, Hole, ss2) ->
      (let ss = ss1 @ ss2 in
        membera (equal_prod _B equal_nat) fas (f, suc (size_list ss)) &&
          (list_all is_Var ss &&
            (distinct (equal_term _B _A) ss &&
              null (list_inter _A (map the_Var ss) vs))))
    | vs, fas, Hole -> false
    | vs, fas, More (v, va, More (vd, ve, vf, vg), vc) -> false;;

let rec showsl_ctxta _A _B t = showsl_ctxt (showsl _A) (showsl _B) t;;

let rec check_is_flat_ctxt (_A1, _A2) (_B1, _B2)
  vs fas c =
    check (is_flat_ctxt_list _A1 _B1 vs fas c)
      (comp (showsl_ctxta _B2 _A2 c)
        (showsl_lit " is not a flat context\092"));;

let rec check_flat_ctxt (_A1, _A2) (_B1, _B2)
  vs c =
    (match c
      with Hole ->
        Inl (comp (showsl_ctxta _B2 _A2 c)
              (showsl_lit " is not a flat context\092"))
      | More (_, ss1, Hole, ss2) ->
        (let ss = ss1 @ ss2 in
          bindb (check (distinct (equal_term _B1 _A1) ss)
                  (comp (showsl_ctxta _B2 _A2 c)
                    (showsl_lit " contains duplicate variables\092")))
            (fun _ ->
              bindb (check (list_all is_Var ss)
                      (comp (showsl_ctxta _B2 _A2 c)
                        (showsl_lit
                          " is not flat, i.e., has depth greater than one\092")))
                (fun _ ->
                  check (list_all (fun t -> not (membera _A1 vs (the_Var t)))
                          (ss1 @ ss2))
                    (comp (showsl_ctxta _B2 _A2 c)
                      (showsl_lit
                        " has to contain only fresh variables\092")))))
      | More (_, _, More (_, _, _, _), _) ->
        Inl (comp (showsl_ctxta _B2 _A2 c)
              (showsl_lit " is not a flat context\092")));;

let rec showsl_terms _A
  name ts =
    comp (comp (showsl_lit name) (showsl_literal "\092"))
      (showsl_list_gen (showsl _A) "" "" "\092" "" ts);;

let rec showsl_dpp _B _C
  i d = (let m = minimal i d in
         let nfs = nfsc i d in
         let p = pb i d in
         let pw = pwb i d in
         let r = rd i d in
         let rw = rwd i d in
         let q = qc i d in
          comp (comp (comp (comp (comp (comp
 (showsl_trsa (showsl _B) (showsl _C) "pairs:" " -> " p)
 (if null pw then id
   else showsl_trsa (showsl _B) (showsl _C) "weak pairs:" " ->= " pw))
                                   (if null r then id
                                     else showsl_trsa (showsl _B) (showsl _C)
    "strict rules:" " ->! " r))
                             (showsl_trsa (showsl _B) (showsl _C) "rules:"
                               " -> " rw))
                       (if null q then id
                         else showsl_terms (showl_term _B _C) "Q-component:" q))
                 (if m then showsl_lit "\092(minimal)" else id))
            (if nfs && not (null q)
              then showsl_lit "\092(normal form substitutions)" else id));;

let rec fcc_proc_cond (_B1, _B2) (_C1, _C2, _C3, _C4)
  i f fcs p pw r rw dpp =
    (let pa = pb i dpp in
     let pwa = pwb i dpp in
     let ra = rd i dpp in
     let rwa = rwd i dpp in
     let nfs = nfsc i dpp in
     let m = minimal i dpp in
     let new_dpp = mkd i nfs m p pw [] r rw in
      (match
        catch_error
          (let pba =
             list_union (equal_prod (equal_term _B1 _C2) (equal_term _B1 _C2))
               pa pwa
             in
           let rb =
             list_union (equal_prod (equal_term _B1 _C2) (equal_term _B1 _C2))
               ra rwa
             in
           let rba =
             list_union (equal_prod (equal_term _B1 _C2) (equal_term _B1 _C2)) r
               rw
             in
           let fa = (f, one_nata) in
           let cf = More (f, [], Hole, []) in
           let fcsa = cf :: fcs in
           let vs = vars_trs_list rb in
           let fs =
             list_union (equal_prod _B1 equal_nat) (funas_trs_list rb)
               (funas_args_trs_list pba)
             in
           let fas = fa :: fs in
           let ds = defined_list rb in
            bindb (check (not (membera (equal_prod _B1 equal_nat) ds fa))
                    (comp (showsl _B2 f) (showsl_lit "is not fresh\092")))
              (fun _ ->
                bindb (check_wf_trs _B2 (_C2, _C4) rb)
                  (fun _ ->
                    bindb (catch_error
                            (forallM
                              (fun rc ->
                                bindb (check_no_var _B2 _C4 (fst rc))
                                  (fun _ ->
                                    bindb (check_no_var _B2 _C4 (snd rc))
                                      (fun _ ->
check_no_defined_root_defined (_B1, _B2) _C4 ds (snd rc))))
                              pba)
                            (fun x -> Inl (snd x)))
                      (fun _ ->
                        bindb (catch_error
                                (forallM
                                  (check_flat_ctxt (_C2, _C4) (_B1, _B2) vs)
                                  fcsa)
                                (fun x -> Inl (snd x)))
                          (fun _ ->
                            bindb (catch_error
                                    (forallM
                                      (check_is_flat_ctxt (_C2, _C4) (_B1, _B2)
vs fas)
                                      fcsa)
                                    (fun x -> Inl (snd x)))
                              (fun _ ->
                                bindb (catch_error
(forallM (check_flat_ctxt_complete (_B1, _B2) _C4 fcsa) fas)
(fun x -> Inl (snd x)))
                                  (fun _ ->
                                    bindb (catch_error
    (forallM (check_rule_preserving (_B1, _B2) (_C1, _C2, _C3, _C4) fcsa r) ra)
    (fun x -> Inl (snd x)))
                                      (fun _ ->
bindb (catch_error
        (forallM
          (check_rule_preserving (_B1, _B2) (_C1, _C2, _C3, _C4) fcsa rba) rwa)
        (fun x -> Inl (snd x)))
  (fun _ ->
    bindb (catch_error
            (forallM (check_rule_reflecting (_B1, _B2) (_C2, _C4) fcsa rb) rba)
            (fun x -> Inl (snd x)))
      (fun _ ->
        bindb (check_superset_of_blocked (_B1, _B2) (_C2, _C4) f p pa)
          (fun _ ->
            check_superset_of_blocked (_B1, _B2) (_C2, _C4) f pw pwa)))))))))))
          (fun x ->
            Inl (comp (comp (comp (comp (comp
  (showsl_lit
    "problem when checking flat context closure conditions to switch from\092")
  (showsl_dpp _B2 _C4 i dpp))
                                    (showsl_lit "\092to the DP problem\092"))
                              (showsl_dpp _B2 _C4 i new_dpp))
                        (showsl_literal "\092"))
                  x))
        with Inl a -> Inl a | Inr _ -> Inr new_dpp));;

let rec g_to_list_dflt_basic_oops_rm_basic_ops _A
  s = iteratei_bset_op_list_it_dflt_basic_oops_rm_basic_ops _A s (fun _ -> true)
        (fun a b -> a :: b) [];;

let rec delete_rm_basic_ops _A x s = delete _A x s;;

let rec ceta_list_diff _A
  xs ys =
    g_to_list_dflt_basic_oops_rm_basic_ops _A.linorder_compare_order
      (foldl (fun a b -> delete_rm_basic_ops _A b a)
        (g_from_list_dflt_basic_oops_rm_basic_ops _A xs) ys);;

let rec fcc_split_proc (_B1, _B2, _B3) (_C1, _C2, _C3, _C4, _C5, _C6, _C7)
  i f fcs pba rb ps rs dpp =
    (let p = pb i dpp in
     let pw = pwb i dpp in
     let r = rd i dpp in
     let rw = rwd i dpp in
     let q = qc i dpp in
     let nfs = nfsc i dpp in
     let m = minimal i dpp in
     let pbaa =
       list_union (equal_prod (equal_term _B2 _C4) (equal_term _B2 _C4)) p pw in
     let rba = rw in
     let pns =
       ceta_list_diff
         (compare_order_prod
           (compare_order_term _B1.compare_compare_order
             _C3.compare_compare_order)
           (compare_order_term _B1.compare_compare_order
             _C3.compare_compare_order))
         pbaa ps
       in
     let rns =
       ceta_list_diff
         (compare_order_prod
           (compare_order_term _B1.compare_compare_order
             _C3.compare_compare_order)
           (compare_order_term _B1.compare_compare_order
             _C3.compare_compare_order))
         rba rs
       in
     let (pa, pwa) = partition_pairs _B2 _C4 f ps pba in
     let (ra, rwa) =
       partition_rules _B2 _C4 (More (f, [], Hole, []) :: fcs) rs rb in
     let two =
       mkd i nfs m
         (ceta_list_diff
           (compare_order_prod
             (compare_order_term _B1.compare_compare_order
               _C3.compare_compare_order)
             (compare_order_term _B1.compare_compare_order
               _C3.compare_compare_order))
           p ps)
         (ceta_list_diff
           (compare_order_prod
             (compare_order_term _B1.compare_compare_order
               _C3.compare_compare_order)
             (compare_order_term _B1.compare_compare_order
               _C3.compare_compare_order))
           pw ps)
         [] [] rns
       in
     let dpp_mid = mkd i nfs m ps pns [] rs rns in
      bindb (bindb
              (catch_error
                (check_subseteq
                  (equal_prod (equal_term _B2 _C4) (equal_term _B2 _C4)) ps
                  pbaa)
                (fun x ->
                  Inl (comp (comp (showsl_lit "pair ") (showsl_rule _B3 _C7 x))
                        (showsl_lit " should be deleted but is not present"))))
              (fun _ ->
                bindb (catch_error
                        (check_subseteq
                          (equal_prod (equal_term _B2 _C4) (equal_term _B2 _C4))
                          rs rba)
                        (fun x ->
                          Inl (comp (comp (showsl_lit "rule ")
                                      (showsl_rule _B3 _C7 x))
                                (showsl_lit
                                  " should be deleted but is not present"))))
                  (fun _ ->
                    bindb (check (null q) (showsl_lit "Q is not empty\092"))
                      (fun _ ->
                        bindb (check (null r)
                                (showsl_lit "strict rules not allowed\092"))
                          (fun _ ->
                            bindb (check_left_linear_trs _B3
                                    (_C1, _C2, _C6, _C7) rw)
                              (fun _ ->
                                fcc_proc_cond (_B2, _B3) (_C2, _C4, _C5, _C7) i
                                  f fcs pa pwa ra rwa dpp_mid))))))
        (fun one -> Inr (one, two)));;

let rec instantiation_proc (_B1, _B2, _B3)
  i st sts dpp =
    (match
      (let ic = icap_impl_dpp_mv _B2 i dpp in
       let isnf = is_QNFc i dpp in
       let (s, t) = st in
       let sy =
         map_term (fun x -> x)
           (fun a ->
             Chara (true, false, false, true, true, true, true, false) :: a)
           s
         in
       let ty =
         map_term (fun x -> x)
           (fun a ->
             Chara (true, false, false, true, true, true, true, false) :: a)
           t
         in
       let iedg = is_iedg_edge_dpp (_B1, _B2, _B3) i dpp in
        catch_error
          (forallM
            (fun (u, v) ->
              (match mgu_class _B2 (ic [u] v) s with None -> Inr ()
                | Some mu ->
                  check (not (isnf (subst_apply_term sy mu)) ||
                          (not (isnf (subst_apply_term
                                       (map_term (fun x -> x)
 (fun a -> Chara (false, false, false, true, true, true, true, false) :: a) u)
                                       mu)) ||
                            list_ex
                              (fun sta ->
                                instance_rule _B2 (equal_list equal_char)
                                  ((ccompare_list ccompare_char),
                                    (equal_list equal_char), mapping_impl_list)
                                  sta st &&
                                  instance_rule _B2 (equal_list equal_char)
                                    ((ccompare_list ccompare_char),
                                      (equal_list equal_char),
                                      mapping_impl_list)
                                    (subst_apply_term sy mu,
                                      subst_apply_term ty mu)
                                    sta)
                              sts))
                    (comp (comp (comp (showsl_lit
"could not find instance of pair ")
                                  (showsl_rule _B3 (showl_list showl_char)
                                    (subst_apply_term sy mu,
                                      subst_apply_term ty mu)))
                            (showsl_lit "\092which resulted from DP "))
                      (showsl_rule _B3 (showl_list showl_char) (u, v)))))
            (filtera (fun (u, v) -> iedg (u, v) s) (pairsb i dpp)))
          (fun x -> Inl (snd x)))
      with Inl a -> Inl a | Inr _ -> Inr (replace_paira i dpp st sts));;

let rec intersect_pairsb
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = intersect_pairs;;

let rec graph_approx_rt_sym_main (_A1, _A2)
  m x1 = match m, x1 with
    m, GCFun (f, ts) ->
      lookupc
        ((ccompare_option (ccompare_prod _A1 ccompare_nat)),
          (equal_option (equal_prod _A2 equal_nat)))
        m None @
        lookupc
          ((ccompare_option (ccompare_prod _A1 ccompare_nat)),
            (equal_option (equal_prod _A2 equal_nat)))
          m (Some (f, size_list ts))
    | m, GCHole -> valuesa m;;

let rec graph_approx_rt_sym (_A1, _A2)
  m (uu, (uv, (ct, uw))) = graph_approx_rt_sym_main (_A1, _A2) m ct;;

let rec rules_non_collapsinga
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = rules_non_collapsing;;

let rec reverse_rules_mapb
  (Dpp_ops_ext
    (dpp, p, pw, pairs, q, r, rw, rules, q_empty, rules_no_left_var,
      rules_non_collapsing, is_QNF, nFQ_subset_NF_rules, rules_map,
      reverse_rules_map, intersect_pairs, replace_pair, intersect_rules,
      delete_P_Pw, delete_R_Rw, split_pairs, split_rules, mk, minimal, nfs,
      wwf_rules, more))
    = reverse_rules_map;;

let rec reverse_tcapRM_dpp _B _C
  i dpp =
    tcapRM _B _C (rules_non_collapsinga i dpp) (reverse_rules_mapb i dpp);;

let rec tcapRM_dpp _B _C
  i dpp = tcapRM _B _C (rules_no_left_vara i dpp) (rules_mapc i dpp);;

let rec check_no_back_edges
  ss candidates add_index g i x5 = match ss, candidates, add_index, g, i, x5
    with ss, candidates, add_index, g, i, [] -> Inr ()
    | ss, candidates, add_index, g, i, asa :: cs ->
        bindb (catch_error
                (forallM
                  (fun a ->
                    catch_error
                      (forallM (check_no_edge ss g a) (candidates i a))
                      (fun x -> Inl (snd x)))
                  asa)
                (fun x -> Inl (snd x)))
          (fun _ ->
            check_no_back_edges ss candidates add_index g
              (foldr add_index asa i) cs);;

let rec check_graph_decomp
  ss empty_index candidates add_index g rcs =
    bindb (check_no_back_edges ss candidates add_index g empty_index
            (map snd rcs))
      (fun _ ->
        catch_error
          (forallM (fun c -> check_edges ss g c c)
            (map_filter (fun x -> (if not (fst x) then Some (snd x) else None))
              rcs))
          (fun x -> Inl (snd x)));;

let rec check_dep_graph_proc (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  i dpp dps =
    (let c = tcapRM_dpp _B5 (equal_list equal_char) i dpp in
     let rc = reverse_tcapRM_dpp _B5 (equal_list equal_char) i dpp in
     let iedg = is_iedg_edge_dpp (_B4, _B5, _B8) i dpp in
     let p = pairsb i dpp in
     let r = rulese i dpp in
     let f = funas_trs_list r in
     let gt_fun = mk_gt_fun _B5 r in
     let rm = rules_mapc i dpp in
     let nlv = rules_no_left_vara i dpp in
      bindb (catch_error
              (check_subseteq
                (equal_prod (equal_term _B5 (equal_list equal_char))
                  (equal_term _B5 (equal_list equal_char)))
                p (maps snd dps))
              (fun x ->
                Inl (comp (comp (showsl_lit "Dependency Pair ")
                            (showsl_rule _B8 (showl_list showl_char) x))
                      (showsl_lit " is missing in decomposition\092"))))
        (fun _ ->
          catch_error
            (check_graph_decomp
              (comp (showsl_prod (showl_term _B8 (showl_list showl_char))
                      (showl_term _B8 (showl_list showl_char)))
                fst)
              (emptyd
                ((ccompare_option (ccompare_prod _B3 ccompare_nat)),
                  (mapping_impl_option
                    (mapping_impl_prod _B6 mapping_impl_nat)))
                (comp (comp root fst) fst))
              (graph_approx_rt_sym (_B3, _B5))
              (insertd
                ((ccompare_option (ccompare_prod _B3 ccompare_nat)),
                  (equal_option (equal_prod _B5 equal_nat))))
              (fun (a, b) ->
                (let (aa, ba) = a in
                  (let (_, t) = aa in
                    (fun (_, (ct, ict)) (ab, bb) ->
                      (let (u, _) = ab in
                        (fun (cu, (_, _)) ->
                          matchc _B5 (equal_list equal_char) ct u &&
                            (matchc _B5 (equal_list equal_char) cu t &&
                              (ict u &&
                                not (nonreachable_gtcapRM
                                      (_B1, _B2, _B3, _B4.compare_compare_order,
_B5, _B7)
                                      ((compare_list compare_char),
(equal_list equal_char))
                                      f nlv (not (null r)) gt_fun rm t u)))))
                        bb))
                    ba)
                  b)
              (map (fun (real, cs) ->
                     (not (is_none real),
                       map (fun (s, t) -> ((s, t), (rc s, (c t, iedg (s, t)))))
                         cs))
                dps))
            (fun x ->
              Inl (comp (comp (showsl_lit
                                "our estimation (EDG*** + IEDG***) could not show that you have a valid decomposition ")
                          (showsl_lit "due to the following reason\092"))
                    x))));;

let rec dep_graph_proc (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  i d dps =
    (match check_dep_graph_proc (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) i d dps
      with Inl a -> Inl a
      | Inr _ ->
        Inr (map_filter
              (fun x ->
                (if not (is_none (fst x))
                  then Some (the (fst x), intersect_pairsb i d (snd x))
                  else None))
              dps));;

let rec check_NF_terms_subset
  is_Q_nf =
    (fun q ->
      catch_error
        (forallM (fun x -> (if not (is_Q_nf x) then Inr () else Inl x)) q)
        (fun x -> Inl (snd x)));;

let rec check_NF_terms_eq (_A1, _A2) (_B1, _B2, _B3)
  qa q =
    bindb (check_NF_terms_subset (is_NF_terms (_A1, _A2) (_B1, _B2, _B3) qa) q)
      (fun _ ->
        check_NF_terms_subset (is_NF_terms (_A1, _A2) (_B1, _B2, _B3) q) qa);;

let rec check_dpp_subsumes (_B1, _B2, _B3) (_C1, _C2, _C3) (_D1, _D2, _D3, _D4)
  i (nfs, (m, (p, (pw, (q, (r, rw)))))) d =
    catch_error
      (let pa = pb i d in
       let pwa = pwb i d in
       let qa = qc i d in
       let ra = rd i d in
       let rwa = rwd i d in
       let nfsa = nfsc i d in
       let ma = minimal i d in
       let pba = p @ pw in
       let rb = ra @ rwa in
       let rba = r @ rw in
        bindb (check (equal_boola ma m)
                (showsl_lit "incompatible minimality flags"))
          (fun _ ->
            bindb (check (equal_boola nfsa nfs)
                    (showsl_lit
                      "incompatible substitutions-in-normal-form flags"))
              (fun _ ->
                bindb (catch_error
                        (check_subseteq
                          (equal_prod (equal_term (equal_lab _B2 _C2) _D2)
                            (equal_term (equal_lab _B2 _C2) _D2))
                          pa p)
                        (fun x ->
                          Inl (toomuch "pair"
                                (showsl_rule (showl_lab _B3 _C3) _D4 x))))
                  (fun _ ->
                    bindb (catch_error
                            (check_subseteq
                              (equal_prod (equal_term (equal_lab _B2 _C2) _D2)
                                (equal_term (equal_lab _B2 _C2) _D2))
                              pwa pba)
                            (fun x ->
                              Inl (toomuch "weak pair"
                                    (showsl_rule (showl_lab _B3 _C3) _D4 x))))
                      (fun _ ->
                        bindb (catch_error
                                (check_NF_terms_eq
                                  ((compare_order_lab _B1.compare_compare_order
                                     _C1.compare_compare_order),
                                    (equal_lab _B2 _C2))
                                  (_D1, _D2, _D3) qa q)
                                (fun x ->
                                  Inl (comp
(showsl_lit "NF(Q) differs due to term ")
(showsl_terma (showl_lab _B3 _C3) _D4 x))))
                          (fun _ ->
                            bindb (catch_error
                                    (check_subseteq
                                      (equal_prod
(equal_term (equal_lab _B2 _C2) _D2) (equal_term (equal_lab _B2 _C2) _D2))
                                      ra r)
                                    (fun x ->
                                      Inl
(toomuch "strict rule" (showsl_rule (showl_lab _B3 _C3) _D4 x))))
                              (fun _ ->
                                bindb (catch_error
(check_subseteq
  (equal_prod (equal_term (equal_lab _B2 _C2) _D2)
    (equal_term (equal_lab _B2 _C2) _D2))
  rb rba)
(fun x ->
  Inl (toomuch "strict/weak rule" (showsl_rule (showl_lab _B3 _C3) _D4 x))))
                                  (fun _ ->
                                    bindb (catch_error
    (check_subseteq
      (equal_prod (equal_term (equal_lab _B2 _C2) _D2)
        (equal_term (equal_lab _B2 _C2) _D2))
      rba rb)
    (fun x ->
      Inl (missing "strict/weak rule" (showsl_rule (showl_lab _B3 _C3) _D4 x))))
                                      (fun _ -> Inr ())))))))))
      (fun x ->
        Inl (comp (comp (comp (comp (comp (comp
    (showsl_lit "finiteness of the problem\092")
    (showsl_dpp (showl_lab _B3 _C3) _D4 i d))
                                      (showsl_lit
"\092may not be concluded from assuming finiteness of the problem\092"))
                                (showsl_dpp (showl_lab _B3 _C3) _D4 i
                                  (mkd i nfs m p pw q r rw)))
                          (showsl_literal "\092"))
                    x)
              (showsl_literal "\092")));;

let rec fcc_proc (_B1, _B2) (_C1, _C2, _C3, _C4, _C5, _C6)
  i f fcs pba rw dpp =
    (let p = pb i dpp in
     let q = qc i dpp in
     let r = rd i dpp in
     let (pa, pw) = partition_pairs _B1 _C3 f p pba in
      bindb (check (null q) (showsl_lit "Q is not empty"))
        (fun _ ->
          bindb (check (null r) (showsl_lit "strict rules not allowed"))
            (fun _ ->
              bindb (check_left_linear_trs _B2 (_C1, _C2, _C5, _C6) (rwd i dpp))
                (fun _ ->
                  fcc_proc_cond (_B1, _B2) (_C2, _C3, _C4, _C6) i f fcs pa pw []
                    rw dpp))));;

let rec q_reduction_proc_non_min _B _C
  i q dpp =
    (match
      catch_error
        (bindb
          (catch_error (check_NF_terms_subset (is_QNFc i dpp) q)
            (fun x ->
              Inl (comp (comp (showsl_lit "the term ") (showsl_terma _B _C x))
                    (showsl_lit " is not allowed in Q\039 "))))
          (fun _ -> Inr ()))
        (fun x ->
          Inl (comp (comp (comp (comp (comp
(showsl_lit "problem when reducing Q in the DP problem\092")
(showsl_dpp _B _C i dpp))
                                  (showsl_lit "\092to the set\092"))
                            (showsl_terms (showl_term _B _C) "Q\039:" q))
                      (showsl_literal "\092"))
                x))
      with Inl a -> Inl a
      | Inr _ ->
        Inr (mkd i (nfsc i dpp) false (pb i dpp) (pwb i dpp) q (rd i dpp)
              (rwd i dpp)));;

let rec q_reduction_proc_min_inn (_B1, _B2, _B3) (_C1, _C2, _C3, _C4)
  i q dpp =
    (let pba = pairsb i dpp in
     let rb = rulese i dpp in
     let f = map (fun a -> Some a) (funas_trs_list (pba @ rb)) in
     let qa = qc i dpp in
     let isnf = is_QNFc i dpp in
     let qq = filtera (fun qb -> list_all isnf (args qb)) qa in
     let rQ =
       filtera
         (fun qb ->
           membera (equal_option (equal_prod _B2 equal_nat)) f (root qb))
         qq
       in
      (match
        catch_error
          (bindb
            (check (nFQ_subset_NF_rulesc i dpp)
              (showsl_lit "innermost rewriting required"))
            (fun _ ->
              bindb (check_wf_trs _B3 (_C2, _C4) rb)
                (fun _ ->
                  bindb (catch_error
                          (check_NF_terms_subset
                            (is_NF_terms (_B1, _B2) (_C1, _C2, _C3) q) rQ)
                          (fun x ->
                            Inl (comp (comp (showsl_lit "the term ")
(showsl_terma _B3 _C4 x))
                                  (showsl_lit " is missing in Q\039 "))))
                    (fun _ ->
                      bindb (catch_error
                              (check_NF_terms_subset
                                (is_NF_terms (_B1, _B2) (_C1, _C2, _C3) qq) q)
                              (fun x ->
                                Inl (comp (comp (showsl_lit "the term ")
    (showsl_terma _B3 _C4 x))
                                      (showsl_lit
" is not allowed in Q\039 "))))
                        (fun _ ->
                          (if nfsc i dpp then Inr ()
                            else check_varcond_subset _B3 (_C2, _C4) pba))))))
          (fun x ->
            Inl (comp (comp (comp (comp (comp
  (showsl_lit "problem when reducing Q in the DP problem\092")
  (showsl_dpp _B3 _C4 i dpp))
                                    (showsl_lit "\092to the set\092"))
                              (showsl_terms (showl_term _B3 _C4) "Q\039:" q))
                        (showsl_literal "\092"))
                  x))
        with Inl a -> Inl a
        | Inr _ ->
          Inr (mkd i (nfsc i dpp) (minimal i dpp) (pb i dpp) (pwb i dpp) q
                (rd i dpp) (rwd i dpp))));;

let rec q_reduction_proc (_B1, _B2, _B3) (_C1, _C2, _C3, _C4)
  i q dpp =
    (match q_reduction_proc_min_inn (_B1, _B2, _B3) (_C1, _C2, _C3, _C4) i q dpp
      with Inl _ -> q_reduction_proc_non_min _B3 _C4 i q dpp | Inr a -> Inr a);;

let rec get_fcc_option
  = function Fcc_Proc (f, fcs, pb, rb, prf) -> Some (f, (fcs, (pb, (rb, prf))))
    | P_is_Empty -> None
    | Subterm_Criterion_Proc (v, va, vb, vc) -> None
    | Gen_Subterm_Criterion_Proc (v, va, vb) -> None
    | Redpair_Proc (v, va, vb) -> None
    | Redpair_UR_Proc (v, va, vb, vc) -> None
    | Usable_Rules_Proc (v, va) -> None
    | Dep_Graph_Proc v -> None
    | Mono_Redpair_Proc (v, va, vb, vc) -> None
    | Mono_URM_Redpair_Proc (v, va, vb, vc) -> None
    | Mono_Redpair_UR_Proc (v, va, vb, vc, vd) -> None
    | Size_Change_Subterm_Proc v -> None
    | Size_Change_Redpair_Proc (v, va, vb) -> None
    | Uncurry_Proc (v, va, vb, vc, vd) -> None
    | Split_Proc (v, va, vb, vc) -> None
    | Semlab_Proc (v, va, vb, vc, vd) -> None
    | Switch_Innermost_Proc (v, va) -> None
    | Rewriting_Proc (v, va, vb, vc, vd, ve, vf) -> None
    | Instantiation_Proc (v, va, vb) -> None
    | Forward_Instantiation_Proc (v, va, vb, vc) -> None
    | Narrowing_Proc (v, va, vb, vc) -> None
    | Assume_Finite (v, va) -> None
    | Unlab_Proc (v, va, vb) -> None
    | Q_Reduction_Proc (v, va) -> None
    | Complex_Constant_Removal_Proc (v, va) -> None
    | General_Redpair_Proc (v, va, vb, vc, vd) -> None
    | To_Trs_Proc v -> None;;

let rec uncurry_of_top_sig_list
  a m sml sm =
    maps (fun (b, c) ->
           (let (f, n) = b in
             (fun _ ->
               (let g = get_symbol sm f n in
                 map (fun i ->
                       (Fun (a, generate_f_xs (g i) (plus_nata n i) ::
                                  map (fun ia -> Var (generate_var ia))
                                    (upt (plus_nata n i)
                                      (plus_nata (plus_nata n i)
(minus_nata m one_nata)))),
                         generate_f_xs (g (suc i))
                           (plus_nata (plus_nata n i) (minus_nata m one_nata))))
                   (upt zero_nata (aarity sm f n)))))
             c)
      sml;;

let rec eta_closed_top_rules (_A1, _A2) (_B1, _B2, _B3, _B4)
  a n sm r p =
    catch_error
      (forallM
        (fun (l, ra) ->
          (match l with Var _ -> Inr ()
            | Fun (ff, ls) ->
              check (equal_nata (aarity sm ff (size_list ls)) zero_nata ||
                      list_ex
                        (fun (lll, rrr) ->
                          (match (lll, rrr) with (Var _, _) -> false
                            | (Fun (_, []), _) -> false
                            | (Fun (_, _ :: _), Var _) -> false
                            | (Fun (_, _ :: _), Fun (_, [])) -> false
                            | (Fun (f, ll :: yy), Fun (g, rr :: zz)) ->
                              eq _A1 f a &&
                                (eq _A1 g a &&
                                  (equal_lista (equal_term _A1 _B2) zz yy &&
                                    (equal_nata (size_list yy)
                                       (minus_nata n one_nata) &&
                                      (distinct (equal_term _A1 _B2) yy &&
(list_all is_Var yy &&
  (null (list_inter _B2 (map the_Var yy) (insert_vars_rule _B2 (ll, rr) [])) &&
    instance_rule _A1 _B2 (_B1, _B2, _B3) (l, ra) (ll, rr)))))))))
                        p)
                (comp (comp (showsl_lit "eta expansion of ")
                        (showsl_rule _A2 _B4 (l, ra)))
                  (showsl_lit " missing"))))
        r)
      (fun x -> Inl (snd x));;

let rec uncurry_top_rules _A
  a n sm =
    map (fun (l, r) -> (uncurry_top _A a n sm l, uncurry_top _A a n sm r));;

let rec map_funs_rules_wa
  fg r =
    map (fun (l, ra) -> (map_funs_term_wa fg l, map_funs_term_wa fg ra)) r;;

let rec uncurry_top_proc (_B1, _B2)
  i info n fmap check_inj p r dpp =
    (let (a, (sml, (u, eb))) = info in
     let pa = pb i dpp in
     let pw = pwb i dpp in
     let ra = rd i dpp in
     let rw = rwd i dpp in
     let nfs = nfsc i dpp in
     let m = minimal i dpp in
     let (e, ew) =
       uncurry_eta_split _B1
         ((ccompare_list ccompare_char), (equal_list equal_char),
           mapping_impl_list)
         eb ra
       in
     let sm = sig_list_to_sig_map _B1 a sml fmap in
     let p_eta = e @ pa in
     let pw_eta = ew @ pw in
     let uP = uncurry_top_rules _B1 a n sm p_eta in
     let uPw = uncurry_top_rules _B1 a n sm pw_eta in
     let uR = map_funs_rules_wa (fun (f, na) -> get_symbol sm f na zero_nata) ra
       in
     let uRw =
       map_funs_rules_wa (fun (f, na) -> get_symbol sm f na zero_nata) rw in
      (match
        bindb (check (null (qc i dpp))
                (showsl_lit "strategy currently unsupported"))
          (fun _ ->
            bindb (check (not (equal_nata n zero_nata))
                    (showsl_lit
                      "the arity of the uncurried symbol must be at least 1"))
              (fun _ ->
                bindb (check_inj a n sml)
                  (fun _ ->
                    (let pba = pairsb i dpp in
                     let is_def = (fun fn -> not (null (rules_mapc i dpp fn)))
                       in
                     let rm = rules_mapc i dpp in
                      bindb (catch_error
                              (forallM
                                (fun (l, _) ->
                                  check (not (is_Var l))
                                    (showsl_lit
                                      "lhs as variable is not allowed"))
                                (ra @ rw))
                              (fun x -> Inl (snd x)))
                        (fun _ ->
                          bindb (catch_error
                                  (forallM
                                    (fun (l, rb) ->
                                      bindb
(check (hvf_top _B1 a n l)
  (comp (comp (showsl_lit "head variable in lhs ")
          (showsl_terma _B2 (showl_list showl_char) l))
    (showsl_lit " not allowed")))
(fun _ -> check_no_var _B2 (showl_list showl_char) rb))
                                    (pw_eta @ p_eta))
                                  (fun x -> Inl (snd x)))
                            (fun _ ->
                              bindb (catch_error
                                      (forallM
(fun (_, rb) ->
  check (not (is_def (the (root rb))))
    (comp (comp (showsl_lit "root of ")
            (showsl_terma _B2 (showl_list showl_char) rb))
      (showsl_lit " must not be defined")))
pba)
                                      (fun x -> Inl (snd x)))
                                (fun _ ->
                                  bindb (check (not (is_def (a, n)))
  (comp (comp (showsl_lit "application symbol ") (showsl _B2 a))
    (showsl_lit " must not be defined in R")))
                                    (fun _ ->
                                      bindb
(if list_ex
      (fun (_, rb) ->
        equal_proda _B1 equal_nat (the (root rb)) (a, n) &&
          equal_gctxta _B1
            (tcapRM2 _B1 (equal_list equal_char) rm (hda (args rb))) GCHole)
      pba
  then bindb (catch_error
               (check_CS_subseteq _B1
                 ((ccompare_list ccompare_char), (equal_list equal_char),
                   mapping_impl_list)
                 (uncurry_of_top_sig_list a n sml sm) u)
               (fun x ->
                 Inl (comp (comp (comp (showsl_lit "uncurrying pair ")
                                   (showsl_rule _B2 (showl_list showl_char) x))
                             (showsl_lit " is missing in\092"))
                       (showsl_rules _B2 (showl_list showl_char) u))))
         (fun _ ->
           bindb (eta_closed_top_rules (_B1, _B2)
                   ((ccompare_list ccompare_char), (equal_list equal_char),
                     mapping_impl_list, (showl_list showl_char))
                   a n sm ra p_eta)
             (fun _ ->
               eta_closed_top_rules (_B1, _B2)
                 ((ccompare_list ccompare_char), (equal_list equal_char),
                   mapping_impl_list, (showl_list showl_char))
                 a n sm rw pw_eta))
  else Inr ())
(fun _ ->
  bindb (catch_error
          (check_subseteq
            (equal_prod (equal_term _B1 (equal_list equal_char))
              (equal_term _B1 (equal_list equal_char)))
            uP p)
          (fun x ->
            Inl (comp (comp (showsl_lit "uncurried pair ")
                        (showsl_rule _B2 (showl_list showl_char) x))
                  (showsl_lit " is missing"))))
    (fun _ ->
      bindb (catch_error
              (check_subseteq
                (equal_prod (equal_term _B1 (equal_list equal_char))
                  (equal_term _B1 (equal_list equal_char)))
                uPw p)
              (fun x ->
                Inl (comp (comp (showsl_lit "uncurried pair ")
                            (showsl_rule _B2 (showl_list showl_char) x))
                      (showsl_lit " is missing"))))
        (fun _ ->
          bindb (catch_error
                  (check_subseteq
                    (equal_prod (equal_term _B1 (equal_list equal_char))
                      (equal_term _B1 (equal_list equal_char)))
                    u p)
                  (fun x ->
                    Inl (comp (comp (showsl_lit "uncurrying pair ")
                                (showsl_rule _B2 (showl_list showl_char) x))
                          (showsl_lit " is missing in new pairs"))))
            (fun _ ->
              bindb (catch_error
                      (check_subseteq
                        (equal_prod (equal_term _B1 (equal_list equal_char))
                          (equal_term _B1 (equal_list equal_char)))
                        uR r)
                      (fun x ->
                        Inl (comp (comp (showsl_lit "rule ")
                                    (showsl_rule _B2 (showl_list showl_char) x))
                              (showsl_lit " is missing in new rules"))))
                (fun _ ->
                  catch_error
                    (check_subseteq
                      (equal_prod (equal_term _B1 (equal_list equal_char))
                        (equal_term _B1 (equal_list equal_char)))
                      uRw r)
                    (fun x ->
                      Inl (comp (comp (showsl_lit "rule ")
                                  (showsl_rule _B2 (showl_list showl_char) x))
                            (showsl_lit
                              " is missing in new rules"))))))))))))))))
        with Inl aa -> Inl aa
        | Inr _ -> Inr (mkd i nfs m uP (uPw @ u) [] uR uRw)));;

let rec only_eta_rules (_A1, _A2) (_B1, _B2, _B3, _B4)
  e r_eta =
    catch_error
      (forallM
        (fun (l, r) ->
          check (match (l, r) with (Var _, _) -> false
                  | (Fun (_, _), Var _) -> false
                  | (Fun (f, ls), Fun (g, rs)) ->
                    eq _A1 f g &&
                      (equal_nata (size_list ls)
                         (nat_of_integer (Big_int.big_int_of_int 2)) &&
                        (equal_nata (size_list rs)
                           (nat_of_integer (Big_int.big_int_of_int 2)) &&
                          (equal_terma _A1 _B2 (nth ls one_nata)
                             (nth rs one_nata) &&
                            list_ex
                              (fun (la, ra) ->
                                instance_rule _A1 _B2 (_B1, _B2, _B3)
                                  (hda ls, hda rs) (la, ra))
                              r_eta))))
            (comp (comp (showsl_lit "rule ") (showsl_rule _A2 _B4 (l, r)))
              (showsl_lit " is not an (eta-expanded) original rule")))
        e)
      (fun x -> Inl (snd x));;

let rec uncurry_proc (_B1, _B2)
  i info fmap check_inj p r dpp =
    (let (a, (sml, (u, eb))) = info in
     let pa = pb i dpp in
     let pw = pwb i dpp in
     let ra = rd i dpp in
     let rw = rwd i dpp in
     let nfs = nfsc i dpp in
     let m = minimal i dpp in
     let (e, ew) =
       uncurry_eta_split _B1
         ((ccompare_list ccompare_char), (equal_list equal_char),
           mapping_impl_list)
         eb ra
       in
     let sm = sig_list_to_sig_map _B1 a sml fmap in
     let uP = uncurry_rules _B1 a sm pa in
     let uPw = uncurry_rules _B1 a sm pw in
     let r_eta = e @ ra in
     let rw_eta = ew @ rw in
     let uR = uncurry_rules _B1 a sm r_eta in
     let uRw = uncurry_rules _B1 a sm rw_eta in
      (match
        (let s = uncurry_of_sig_list a sml sm in
          bindb (check (null (qc i dpp))
                  (showsl_lit "strategy not supported for uncurrying"))
            (fun _ ->
              bindb (only_eta_rules (_B1, _B2)
                      ((ccompare_list ccompare_char), (equal_list equal_char),
                        mapping_impl_list, (showl_list showl_char))
                      e r_eta)
                (fun _ ->
                  bindb (only_eta_rules (_B1, _B2)
                          ((ccompare_list ccompare_char),
                            (equal_list equal_char), mapping_impl_list,
                            (showl_list showl_char))
                          ew rw_eta)
                    (fun _ ->
                      bindb (check_inj a
                               (nat_of_integer (Big_int.big_int_of_int 2))
                              sml)
                        (fun _ ->
                          bindb (catch_error
                                  (forallM
                                    (fun (l, _) ->
                                      check (not (is_Var l))
(showsl_lit "lhs as variable is not allowed"))
                                    (ra @ rw))
                                  (fun x -> Inl (snd x)))
                            (fun _ ->
                              bindb (catch_error
                                      (forallM
(fun (l, _) ->
  check (hvf_term _B1 a l)
    (comp (comp (showsl_lit "head variable in lhs ")
            (showsl_terma _B2 (showl_list showl_char) l))
      (showsl_lit " not allowed")))
pa)
                                      (fun x -> Inl (snd x)))
                                (fun _ ->
                                  bindb (catch_error
  (forallM
    (fun (l, _) ->
      check (hvf_term _B1 a l)
        (comp (comp (showsl_lit "head variable in lhs ")
                (showsl_terma _B2 (showl_list showl_char) l))
          (showsl_lit " not allowed")))
    pw)
  (fun x -> Inl (snd x)))
                                    (fun _ ->
                                      bindb
(catch_error
  (forallM
    (fun (l, _) ->
      check (hvf_term _B1 a l)
        (comp (comp (showsl_lit "head variable in lhs ")
                (showsl_terma _B2 (showl_list showl_char) l))
          (showsl_lit " not allowed")))
    r_eta)
  (fun x -> Inl (snd x)))
(fun _ ->
  bindb (catch_error
          (forallM
            (fun (l, _) ->
              check (hvf_term _B1 a l)
                (comp (comp (showsl_lit "head variable in lhs ")
                        (showsl_terma _B2 (showl_list showl_char) l))
                  (showsl_lit " not allowed")))
            rw_eta)
          (fun x -> Inl (snd x)))
    (fun _ ->
      bindb (eta_closed_rules (_B1, _B2)
              ((ccompare_list ccompare_char), (equal_list equal_char),
                mapping_impl_list, (showl_list showl_char))
              a sm r_eta r_eta)
        (fun _ ->
          bindb (eta_closed_rules (_B1, _B2)
                  ((ccompare_list ccompare_char), (equal_list equal_char),
                    mapping_impl_list, (showl_list showl_char))
                  a sm rw_eta rw_eta)
            (fun _ ->
              bindb (catch_error
                      (check_subseteq
                        (equal_prod (equal_term _B1 (equal_list equal_char))
                          (equal_term _B1 (equal_list equal_char)))
                        uP p)
                      (fun x ->
                        Inl (comp (comp (showsl_lit "uncurried pair ")
                                    (showsl_rule _B2 (showl_list showl_char) x))
                              (showsl_lit " is missing"))))
                (fun _ ->
                  bindb (catch_error
                          (check_subseteq
                            (equal_prod (equal_term _B1 (equal_list equal_char))
                              (equal_term _B1 (equal_list equal_char)))
                            uPw p)
                          (fun x ->
                            Inl (comp (comp (showsl_lit "uncurried pair ")
(showsl_rule _B2 (showl_list showl_char) x))
                                  (showsl_lit " is missing"))))
                    (fun _ ->
                      bindb (catch_error
                              (check_subseteq
                                (equal_prod
                                  (equal_term _B1 (equal_list equal_char))
                                  (equal_term _B1 (equal_list equal_char)))
                                uR r)
                              (fun x ->
                                Inl (comp (comp (showsl_lit "uncurried rule ")
    (showsl_rule _B2 (showl_list showl_char) x))
                                      (showsl_lit " is missing"))))
                        (fun _ ->
                          bindb (catch_error
                                  (check_subseteq
                                    (equal_prod
                                      (equal_term _B1 (equal_list equal_char))
                                      (equal_term _B1 (equal_list equal_char)))
                                    uRw r)
                                  (fun x ->
                                    Inl (comp
  (comp (showsl_lit "uncurried rule ")
    (showsl_rule _B2 (showl_list showl_char) x))
  (showsl_lit " is missing"))))
                            (fun _ ->
                              bindb (catch_error
                                      (check_CS_subseteq _B1
((ccompare_list ccompare_char), (equal_list equal_char), mapping_impl_list) s u)
                                      (fun x ->
Inl (comp (comp (showsl_lit "uncurry rule ")
            (showsl_rule _B2 (showl_list showl_char) x))
      (showsl_lit " is missing"))))
                                (fun _ ->
                                  bindb (catch_error
  (check_CS_subseteq _B1
    ((ccompare_list ccompare_char), (equal_list equal_char), mapping_impl_list)
    u s)
  (fun x ->
    Inl (comp (comp (showsl_lit "rule ")
                (showsl_rule _B2 (showl_list showl_char) x))
          (showsl_lit " is not an uncurry rule"))))
                                    (fun _ ->
                                      catch_error
(check_subseteq
  (equal_prod (equal_term _B1 (equal_list equal_char))
    (equal_term _B1 (equal_list equal_char)))
  u r)
(fun x ->
  Inl (comp (comp (showsl_lit "uncurry rule ")
              (showsl_rule _B2 (showl_list showl_char) x))
        (showsl_lit " is missing in new TRS")))))))))))))))))))))
        with Inl aa -> Inl aa
        | Inr _ -> Inr (mkd i nfs m uP uPw [] uR (uRw @ u))));;

let rec uncurry_proc_both (_B1, _B2) (_C1, _C2)
  i x1 x2 = match i, x1, x2 with
    i, None, (a, (sml, (u, eb))) ->
      uncurry_proc ((equal_lab _B1 _C1), (showl_lab _B2 _C2)) i
        (a, (sml, (u, eb)))
        (fmap _B1 _C1 a (nat_of_integer (Big_int.big_int_of_int 2)))
        (check_inj (_B1, _B2) (_C1, _C2))
    | i, Some n, (a, (sml, (u, eb))) ->
        uncurry_top_proc ((equal_lab _B1 _C1), (showl_lab _B2 _C2)) i
          (a, (sml, (u, eb))) n (fmap _B1 _C1 a n)
          (check_inj (_B1, _B2) (_C1, _C2));;

let rec check_rewrite_common_preconditions (_B1, _B2, _B3)
  i u_opt st ss ts t lr p sound dpp =
    (let r = rulese i dpp in
     let s = fst st in
     let ta = snd st in
     let tp = subt_at ta p in
     let u =
       (match u_opt
         with None ->
           maps (fun tb -> inn_usable_rules_pair (_B1, _B2) i dpp (s, tb)) ts
         | Some u -> u)
       in
      bindb (catch_error
              (check_subseteq
                (equal_prod (equal_term _B2 (equal_list equal_char))
                  (equal_term _B2 (equal_list equal_char)))
                u r)
              (fun x ->
                Inl (comp (showsl_rule _B3 (showl_list showl_char) x)
                      (showsl_lit " is not a rule of the rewrite system "))))
        (fun _ ->
          (let urc = is_ur_closed_impl_dpp_mv (_B1, _B2, _B3) i dpp u in
           let check_urc =
             (fun sa tb ->
               check (urc sa tb)
                 (comp (comp (showsl_lit "term ")
                         (showsl_terma _B3 (showl_list showl_char) tb))
                   (showsl_lit " is not closed under usable rules")))
             in
           let nfs = nfsc i dpp in
            bindb (catch_error
                    (forallM
                      (fun (l, _) ->
                        check (not (is_Var l))
                          (showsl_lit "lhss must not be variables"))
                      u)
                    (fun x -> Inl (snd x)))
              (fun _ ->
                bindb (check (wf_rule (equal_list equal_char) lr)
                        (comp (showsl_rule _B3 (showl_list showl_char) lr)
                          (showsl_lit " is not a well formed rule")))
                  (fun _ ->
                    bindb (if nfs && sound then Inr ()
                            else catch_error
                                   (check_subseteq (equal_list equal_char)
                                     (vars_term_list tp) (vars_term_list s))
                                   (fun _ ->
                                     Inl (showsl_lit
   "variable condition in pair violated")))
                      (fun _ ->
                        bindb (catch_error (forallM (check_urc ss) ts)
                                (fun x -> Inl (snd x)))
                          (fun _ ->
                            bindb (catch_error
                                    (forallM
                                      (fun (l, a) -> check_urc (args l) a) u)
                                    (fun x -> Inl (snd x)))
                              (fun _ ->
                                bindb (catch_error
(check_critical_pairs_innermost (_B2, _B3) u)
(fun x ->
  Inl (comp (showsl_lit "problem in showing UNF of usable rules\092") x)))
                                  (fun _ ->
                                    catch_error
                                      (forallM
(fun (_, (sa, tb)) ->
  check (equal_terma _B2 (equal_list equal_char) sa tb)
    (showsl_lit
      "non-trivial critical pair between rule to rewrite and usable rules"))
(critical_pairs_impl _B2 [lr] u))
                                      (fun x -> Inl (snd x)))))))))));;

let rec check_prop_rstepa (_A1, _A2) (_B1, _B2, _B3, _B4)
  nfs pa r p rule s t =
    bindb (check
            (membera (equal_prod (equal_term _A1 _B2) (equal_term _A1 _B2)) r
              rule)
            (comp (comp (comp (showsl_rule _A2 _B4 rule)
                          (showsl_literal " is not a rule of\092"))
                    (showsl_trs _A2 _B4 r))
              (showsl_literal "\092")))
      (fun _ ->
        check_prop_rstep_rule (_A1, _A2) (_B1, _B2, _B3, _B4) nfs pa p rule s
          t);;

let rec check_rstep (_A1, _A2) (_B1, _B2, _B3, _B4)
  = check_prop_rstepa (_A1, _A2) (_B1, _B2, _B3, _B4) false (fun _ -> Inr ());;

let rec rewriting_proc (_B1, _B2, _B3)
  i u stb sta st lr p dpp =
    (match
      (let s = fst stb in
       let t = snd sta in
        bindb (check_rstep (_B2, _B3)
                ((ccompare_list ccompare_char), (equal_list equal_char),
                  mapping_impl_list, (showl_list showl_char))
                (rulese i dpp) p lr (snd stb) t)
          (fun _ ->
            bindb (check (nFQ_subset_NF_rulesc i dpp)
                    (showsl_lit "innermost rewriting required"))
              (fun _ ->
                bindb (check_rewrite_common_preconditions (_B1, _B2, _B3) i u
                        stb [s] [subt_at (snd stb) p] t lr p true dpp)
                  (fun _ ->
                    bindb (check
                            (eq_rule_mod_vars _B2
                              ((ccompare_list ccompare_char),
                                (equal_list equal_char), mapping_impl_list)
                              sta st)
                            (comp (comp (comp (showsl_lit "the rule ")
  (showsl_rule _B3 (showl_list showl_char) sta))
                                    (showsl_lit
                                      " is not a renamed variant of "))
                              (showsl_rule _B3 (showl_list showl_char) st)))
                      (fun _ ->
                        bindb (check
                                (equal_terma _B2 (equal_list equal_char) s
                                  (fst sta))
                                (showsl_lit
                                  "left-hand sides of old and new pair differ"))
                          (fun _ ->
                            bindb (check
                                    (membera
                                       (equal_prod
 (equal_term _B2 (equal_list equal_char))
 (equal_term _B2 (equal_list equal_char)))
                                       (pb i dpp) stb ||
                                      null (rd i dpp))
                                    (showsl_lit
                                      "strict DP or no strict rules required"))
                              (fun _ ->
                                check (nfsc i dpp || wwf_rulesa i dpp)
                                  (showsl_lit
                                    "well-formed rules or normal subst. required"))))))))
      with Inl a -> Inl a | Inr _ -> Inr (replace_paira i dpp stb [st]));;

let rec nF_subst_impl
  nf nfs r sigma =
    (if nfs then list_all (fun x -> nf (sigma x)) (vars_rule_list r)
      else true);;

let rec qnarrows_impl _A
  isnf nfs r t =
    maps (fun p ->
           (let tp = subt_at t p in
             (if not (is_Var tp)
               then maps (fun (l, ra) ->
                           maps (fun (mu_1, mu_2) ->
                                  (if nF_subst_impl isnf nfs (l, ra) mu_2
                                    then (if list_all isnf
       (args (subst_apply_term l mu_2))
   then [(ctxt_apply_term (ctxt_of_pos_term p (subst_apply_term t mu_1))
            (subst_apply_term ra mu_2),
           mu_1)]
   else [])
                                    else []))
                             (option_to_list
                               (mgu_var_disjoint_generic (equal_list equal_char)
                                 _A (fun a ->
                                      Chara
(false, false, false, true, true, true, true, false) ::
a)
                                 (fun a ->
                                   Chara (true, false, false, true, true, true,
   true, false) ::
                                     a)
                                 tp l)))
                      r
               else [])))
      (poss_list t);;

let rec narrowing_proc (_B1, _B2, _B3)
  i st p sts dpp =
    (match
      (let (s, t) = st in
       let q = qc i dpp in
        bindb (check
                (nFQ_subset_NF_rulesc i dpp ||
                  null q &&
                    linear_term
                      ((ceq_list ceq_char), (ccompare_list ccompare_char),
                        set_impl_list)
                      t)
                (showsl_lit
                  "innermost or full rewriting required (and linearity of t in full rewriting case)"))
          (fun _ ->
            (let ic = icap_impl_dpp_mv _B2 i dpp in
             let isnf = is_QNFc i dpp in
             let pairs = pairsb i dpp in
              bindb (check (membera (equal_list equal_nat) (poss_list t) p)
                      (comp (showsl_lit "position not contained in ")
                        (showsl_terma _B3 (showl_list showl_char) t)))
                (fun _ ->
                  (let tp = subt_at t p in
                   let nftp = isnf tp in
                    bindb (check
                            (membera (equal_list equal_nat)
                               (poss_list (ic [s] t)) p ||
                              not nftp)
                            (showsl_lit
                              "neither is position contained in capped term of t, nor is t|_p not in Q-normal form"))
                      (fun _ ->
                        (let nfs = nfsc i dpp in
                         let narrows =
                           qnarrows_impl _B2 isnf nfs (rulese i dpp) tp in
                         let stsa =
                           filtera (fun (smu, _) -> isnf smu)
                             (map (fun (ta, mu) ->
                                    (subst_apply_term s mu,
                                      ctxt_apply_term
(ctxt_of_pos_term p (subst_apply_term t mu)) ta))
                               narrows)
                           in
                          bindb (catch_error
                                  (forallM
                                    (fun newa ->
                                      check
(list_ex
  (fun sta ->
    instance_rule _B2 (equal_list equal_char)
      ((ccompare_list ccompare_char), (equal_list equal_char),
        mapping_impl_list)
      newa sta &&
      (not nfs || (null q || wf_rule (equal_list equal_char) sta)))
  sts)
(comp (showsl_lit "could not find narrowed pair ")
  (showsl_rule _B3 (showl_list showl_char) newa)))
                                    stsa)
                                  (fun x -> Inl (snd x)))
                            (fun _ ->
                              (let iedg =
                                 is_iedg_edge_dpp (_B1, _B2, _B3) i dpp (s, t)
                                 in
                                bindb (catch_error
(check_subseteq (equal_list equal_char) (vars_term_list tp) (vars_term_list s))
(fun x ->
  Inl (comp (comp (showsl_lit "variable ") (showsl_lista showl_char x))
        (showsl_lit " only occurs on rhs of pair"))))
                                  (fun _ ->
                                    bindb (check
    (membera
       (equal_prod (equal_term _B2 (equal_list equal_char))
         (equal_term _B2 (equal_list equal_char)))
       (pb i dpp) st ||
      null (rd i dpp))
    (showsl_lit "strict DP or no strict rules required"))
                                      (fun _ ->
(if nftp
  then catch_error
         (forallM
           (fun (u, v) ->
             bindb (check (membera (equal_list equal_nat) (poss_list u) p)
                     (comp (showsl_lit "position not contained in lhs of pair ")
                       (showsl_rule _B3 (showl_list showl_char) (u, v))))
               (fun _ ->
                 (match
                   mgu_var_disjoint_generic (equal_list equal_char) _B2
                     (fun a ->
                       Chara (false, false, false, true, true, true, true,
                               false) ::
                         a)
                     (fun a ->
                       Chara (true, false, false, true, true, true, true,
                               false) ::
                         a)
                     tp (subt_at u p)
                   with None -> Inr ()
                   | Some (mu_1, mu_2) ->
                     check (not (isnf (subst_apply_term s mu_1)) ||
                             not (isnf (subst_apply_term u mu_2)))
                       (comp (showsl_lit
                               "t |_ p and u |_ p unify and satisfy variable condition for pair (u,v) = ")
                         (showsl_rule _B3 (showl_list showl_char) (u, v))))))
           (filtera (fun (u, _) -> iedg u) pairs))
         (fun x -> Inl (snd x))
  else Inr ()))))))))))))
      with Inl a -> Inl a | Inr _ -> Inr (replace_paira i dpp st sts));;

let rec extract_renamings _A
  old_new =
    (fun_of_map_fun (map_of _A old_new) id,
      fun_of_map_fun (map_of _A (map swap old_new)) id);;

let rec extract_components _A
  mu old_new =
    (let (d, da) = extract_renamings _A old_new in
     let c =
       map_filter
         (fun x ->
           (if (let (_, a) = x in equal_nata a zero_nata) then Some (fst x)
             else None))
         mu
       in
     let nu = map d c in
      (d, (da, nu)));;

let rec check_components (_A1, _A2)
  mu ddNU =
    (let (d, (da, nu)) = ddNU in
      bindb (catch_error
              (forallM
                (fun f ->
                  bindb (check
                          (not (membera (equal_prod _A1 equal_nat) mu
                                 (f, one_nata)))
                          (comp (comp (showsl_lit "new unary symbol ")
                                  (showsl _A2 f))
                            (showsl_lit " clashes with old symbol")))
                    (fun _ ->
                      bindb (check (eq _A1 (d (da f)) f)
                              (comp (showsl_lit
                                      "problem with bijection for renaming of ")
                                (showsl _A2 f)))
                        (fun _ ->
                          check (membera (equal_prod _A1 equal_nat) mu
                                  (da f, zero_nata))
                            (comp (showsl_lit
                                    "problem with inverse renaming of ")
                              (showsl _A2 f)))))
                nu)
              (fun x -> Inl (snd x)))
        (fun _ ->
          catch_error
            (forallM
              (fun (f, n) ->
                bindb (check (less_eq_nat n one_nata)
                        (comp (showsl_lit "arity > 1 for symbol ")
                          (showsl _A2 f)))
                  (fun _ ->
                    check (if equal_nata n zero_nata
                            then membera _A1 nu (d f) && eq _A1 (da (d f)) f
                            else true)
                      (comp (showsl_lit
                              "problem with bijection for renaming of constant ")
                        (showsl _A2 f))))
              mu)
            (fun x -> Inl (snd x))));;

let rec str
  d x xa2 = match d, x, xa2 with
    d, x, Fun (f, v :: vb :: vc) -> Fun (d f, [Var x])
    | d, x, Fun (f, []) -> Fun (d f, [Var x])
    | d, x, Fun (f, [t]) -> Fun (f, [str d x t])
    | d, uu, Var x -> Var x;;

let rec choose_var x l = hda (vars_term_list l @ [x]);;

let rec check_to_srs_sound (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8)
  (_B1, _B2, _B3)
  v old_new r s rw sw =
    (let mu = funas_trs_list (r @ rw) in
     let (d, (da, nu)) = extract_components _B2 mu old_new in
      bindb (check_components (_B2, _B3) mu (d, (da, nu)))
        (fun _ ->
          bindb (check_varcond_subset _B3 (_A6, _A8) r)
            (fun _ ->
              bindb (check_varcond_subset _B3 (_A6, _A8) rw)
                (fun _ ->
                  (let checka =
                     (fun ra sa ->
                       catch_error
                         (forallM
                           (fun (l, rb) ->
                             (let y = choose_var v l in
                              let stra = str d y in
                              let slr = (stra l, stra rb) in
                               check (less_eq_set
(_A2, _A3, _A4.ccompare_cproper_interval)
(vars_term (_B1, _B2) (_A1, _A2, _A3, _A4, _A5, _A6, _A7) l)
(inserta (_A3, _A4.ccompare_cproper_interval) y
  (bot_set (_A3, _A4.ccompare_cproper_interval, _A7))) &&
                                       membera
 (equal_prod (equal_term _B2 _A6) (equal_term _B2 _A6)) sa slr)
                                 (comp (showsl_lit "problem with new rule ")
                                   (showsl_rule _B3 _A8 slr))))
                           ra)
                         (fun x -> Inl (snd x)))
                     in
                    bindb (checka r s) (fun _ -> checka rw sw))))));;

let rec const_to_string_sound_tt (_A1, _A2, _A3)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  (Const_string_sound_proof (v, old_new, s, sw)) i tp =
    bindb (check_to_srs_sound (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
            (_A1, _A2, _A3) v old_new (rc i tp) s (rwc i tp) sw)
      (fun _ -> Inr (mkc i false [] s sw));;

let rec dP_list _A _B
  shp r d_list =
    maps (fun lr ->
           (let l = fst lr in
            let s = sharp_term shp l in
             map_filter
               (fun x ->
                 (if not (supt_impl _A _B l x) &&
                       (not (is_Var x) &&
                         membera (equal_prod _A equal_nat) d_list
                           (the (root x)))
                   then Some (s, sharp_term shp x) else None))
               (supteq_list (snd lr))))
      r;;

let rec dependency_pairs_tt (_A1, _A2, _A3, _A4) (_C1, _C2, _C3, _C4, _C5, _C6)
  shp i j tp nfs m p =
    (let r = rulesd i tp in
     let q = qb i tp in
     let iQ = is_QNFb i tp in
     let u = filtera (applicable_rule_impl iQ) r in
      (match
        catch_error
          (bindb
            (if isOK (check_wf_trs _A4 (_C3, _C6) u) then Inr ()
              else check (nfs &&
                           (nfsb i tp &&
                             (nFQ_subset_NF_rulesb i tp &&
                               list_all (fun l -> not (is_Var l)) (map fst r))))
                     (showsl_lit
                       "neither is the TRS well-formed, nor is the restriction to innermost with normal form substitutions present"))
            (fun _ ->
              bindb (catch_error (forallM (check_no_var _A4 _C6) q)
                      (fun x -> Inl (snd x)))
                (fun _ ->
                  (let qr = map (fun (Fun (f, ss)) -> (f, size_list ss)) q in
                   let d = defined_list u in
                    bindb (catch_error
                            (forallM
                              (fun (f, n) ->
                                check (not
(membera (equal_prod _A2 equal_nat) d (shp f, n)))
                                  (comp (comp
  (comp (showsl_lit "sharping ") (showsl _A4 f))
  (showsl_lit " yields the defined symbol "))
                                    (showsl _A4 (shp f))))
                              d)
                            (fun x -> Inl (snd x)))
                      (fun _ ->
                        bindb (catch_error
                                (forallM
                                  (fun (f, n) ->
                                    check (not
    (membera (equal_prod _A2 equal_nat) qr (shp f, n)))
                                      (comp
(comp (comp (comp (showsl_lit "sharping ") (showsl _A4 f))
        (showsl_lit " yields the symbol "))
  (showsl _A4 (shp f)))
(showsl_lit " which is a root of Q")))
                                  d)
                                (fun x -> Inl (snd x)))
                          (fun _ ->
                            (let pa =
                               set ((ceq_prod (ceq_term _A2 _C3)
                                      (ceq_term _A2 _C3)),
                                     (ccompare_prod (ccompare_term _A1 _C2)
                                       (ccompare_term _A1 _C2)),
                                     (set_impl_prod set_impl_term
                                       set_impl_term))
                                 p
                               in
                              catch_error
                                (catch_error
                                  (forallM
                                    (fun x ->
                                      (if member
    ((ceq_prod (ceq_term _A2 _C3) (ceq_term _A2 _C3)),
      (ccompare_prod (ccompare_term _A1 _C2) (ccompare_term _A1 _C2)))
    x pa ||
    list_ex (eq_rule_mod_vars _A2 (_C1, _C3, _C4) x) p
then Inr () else Inl x))
                                    (dP_list _A2 _C3 shp u d))
                                  (fun x -> Inl (snd x)))
                                (fun x ->
                                  Inl (comp
(comp (showsl_lit "the DP ") (showsl_rule _A4 _C6 x))
(showsl_lit " does not appear in the DP problem\092"))))))))))
          (fun x ->
            Inl (comp (showsl_lit
                        "the DP-transformation is not applied correctly.\092")
                  x))
        with Inl a -> Inl a | Inr _ -> Inr (mkd j nfs m p [] q [] r)));;

let rec switch_innermost_tt (_B1, _B2)
  i joins_i trs =
    (let r = rulesd i trs in
      (match
        (let cp = critical_pairs_impl _B1 r r in
          bindb (catch_error
                  (forallM
                    (fun (b, _) -> check b (showsl_lit "rules are not overlay"))
                    cp)
                  (fun x -> Inl (snd x)))
            (fun _ ->
              bindb (check_critical_pairs (_B1, _B2) r cp joins_i)
                (fun _ ->
                  check_wf_trs _B2
                    ((equal_list equal_char), (showl_list showl_char)) r)))
        with Inl a -> Inl a | Inr _ -> Inr (mkc i true (map fst r) r [])));;

let rec rule_removal_tt _B _C
  i rp rremove trs =
    (match
      catch_error
        (let (rs, rns) = split_rulesb i trs rremove in
          bindb (valid rp)
            (fun _ ->
              bindb (catch_error (mono rp (rs @ rns))
                      (fun x ->
                        Inl (comp (showsl_lit
                                    "problem with monotonicity of strict order\092")
                              x)))
                (fun _ ->
                  bindb (catch_error
                          (catch_error (forallM (ns rp) rns)
                            (fun x -> Inl (snd x)))
                          (fun x ->
                            Inl (comp (showsl_lit
"problem when orienting TRS\092")
                                  x)))
                    (fun _ ->
                      catch_error
                        (catch_error (forallM (s rp) rs) (fun x -> Inl (snd x)))
                        (fun x ->
                          Inl (comp (showsl_lit
                                      "problem when orienting TRS\092")
                                x))))))
        (fun x ->
          Inl (comp (comp (comp (showsl_lit
                                  "could not apply the reduction pair processor with the following\092")
                            (desc rp))
                      (showsl_literal "\092"))
                x))
      with Inl a -> Inl a | Inr _ -> Inr (delete_R_Rwb i trs rremove rremove));;

let rec unary_term = function Var x -> true
                     | Fun (f, [t]) -> unary_term t
                     | Fun (v, []) -> false
                     | Fun (v, vb :: vd :: ve) -> false;;

let rec check_unary_signature _A _B
  r = catch_error
        (catch_error
          (forallM
            (fun x ->
              (if (let (l, ra) = x in unary_term l && unary_term ra) then Inr ()
                else Inl x))
            r)
          (fun x -> Inl (snd x)))
        (fun x ->
          Inl (comp (comp (showsl_lit "the rule\092") (showsl_rule _A _B x))
                (showsl_lit
                  "\092violates the condition that all function symbols\092have to be unary")));;

let rec rev_term c x1 = match c, x1 with c, Var x -> ctxt_apply_term c (Var x)
                   | c, Fun (f, [t]) -> rev_term (More (f, [], c, [])) t;;

let rec rev_rule (l, r) = (rev_term Hole l, rev_term Hole r);;

let rec string_reversal_tt _B _C
  i trs =
    (let rs = rulesd i trs in
     let r = rc i trs in
     let s = rwc i trs in
      (match check_unary_signature _B _C rs with Inl a -> Inl a
        | Inr _ ->
          Inr (mkc i default_nfs_trs [] (map rev_rule r) (map rev_rule s))));;

let rec q_emptyb
  (Tp_ops_ext
    (qreltrs, q, r, rw, rules, q_empty, is_QNF, nFQ_subset_NF_rules, rules_map,
      delete_R_Rw, split_rules, mk, nfs, more))
    = q_empty;;

let rec sem_lab_rel_tt (_A1, _A2, _A3) (_B1, _B2, _B3, _B4, _B5)
  splitter ld i valid check_decr check_model_lab lQ lAll tp =
    (let r = rc i tp in
     let rw = rwc i tp in
     let nfs = nfsb i tp in
     let (lR, (lRw, d)) =
       splitter lAll
         (set ((ceq_prod (ceq_term _A2 _B3) (ceq_term _A2 _B3)),
                (ccompare_prod (ccompare_term _A1 _B2) (ccompare_term _A1 _B2)),
                (set_impl_prod set_impl_term set_impl_term))
           rw)
       in
      (match
        bindb valid
          (fun _ ->
            (let q = qb i tp in
              catch_error
                (bindb
                  (if nfs && not (q_emptyb i tp)
                    then check_wf_trs _A3 (_B3, _B5) d else Inr ())
                  (fun _ ->
                    bindb (check_decr d)
                      (fun _ ->
                        bindb (check_sl_Q (_A2, _A3) (_B1, _B3, _B4, _B5) ld lQ
                                q)
                          (fun _ ->
                            bindb (check_model_lab
                                     (set ((ceq_prod (ceq_term _A2 _B3)
     (ceq_term _A2 _B3)),
    (ccompare_prod (ccompare_term _A1 _B2) (ccompare_term _A1 _B2)),
    (set_impl_prod set_impl_term set_impl_term))
                                       lR)
                                    r)
                              (fun _ ->
                                check_model_lab
                                  (set ((ceq_prod (ceq_term _A2 _B3)
  (ceq_term _A2 _B3)),
 (ccompare_prod (ccompare_term _A1 _B2) (ccompare_term _A1 _B2)),
 (set_impl_prod set_impl_term set_impl_term))
                                    lRw)
                                  rw)))))
                (fun x ->
                  Inl (comp (showsl_lit "problem with labeled TRS:\092") x))))
        with Inl a -> Inl a | Inr _ -> Inr (mkc i nfs lQ lR (lRw @ d))));;

let rec sem_lab_fin_tt (_A1, _A2, _A3) (_B1, _B2, _B3, _B4, _B5) _D
  splitter lc ld cge i gen lQ lAll tp =
    bindb (gen (insert_funas_trs _A2 (rulesd i tp) []) [])
      (fun ops ->
        (let check_ml =
           check_sl_model_lab_trs (_A1, _A2, _A3) _D (_B2, _B3, _B5) (sl_I ops)
             (sl_L ops) (sl_C ops) cge lc
           in
         let check_d = sl_check_decr ops in
          sem_lab_rel_tt (_A1, _A2, _A3) (_B1, _B2, _B3, _B4, _B5) splitter ld i
            (Inr ()) check_d check_ml lQ lAll tp));;

let rec semlab_fin_tt (_B1, _B2, _B3) (_C1, _C2, _C3, _C4, _C5)
  j x1 = match j, x1 with
    j, Rootlab uu ->
      sem_lab_fin_tt
        ((compare_lab _B1 (compare_list compare_nat)),
          (equal_lab _B2 (equal_list equal_nat)),
          (showl_lab _B3 (showl_list showl_nat)))
        (_C1, _C2, _C3, _C4, _C5) (showl_lab _B3 (showl_list showl_nat))
        (model_splitter
          ((compare_lab _B1 (compare_list compare_nat)),
            (equal_lab _B2 (equal_list equal_nat)))
          (_C2, _C3) label_decomp)
        label label_decomp (equal_laba _B2 (equal_list equal_nat)) j
        (slm_gen_to_sl_gen (rl_slm _B2 (equal_list equal_nat) None))
    | j, Finitelab sli ->
        sem_lab_fin_tt
          ((compare_lab _B1 (compare_list compare_nat)),
            (equal_lab _B2 (equal_list equal_nat)),
            (showl_lab _B3 (showl_list showl_nat)))
          (_C1, _C2, _C3, _C4, _C5) showl_nat
          (model_splitter
            ((compare_lab _B1 (compare_list compare_nat)),
              (equal_lab _B2 (equal_list equal_nat)))
            (_C2, _C3) label_decomp)
          label label_decomp equal_nata j
          (slm_gen_to_sl_gen (fun _ _ -> Inr (sli_to_slm _B2 sli)))
    | j, QuasiFinitelab (sli, v) ->
        sem_lab_fin_tt
          ((compare_lab _B1 (compare_list compare_nat)),
            (equal_lab _B2 (equal_list equal_nat)),
            (showl_lab _B3 (showl_list showl_nat)))
          (_C1, _C2, _C3, _C4, _C5) showl_nat
          (quasi_splitter
            ((compare_lab _B1 (compare_list compare_nat)),
              (equal_lab _B2 (equal_list equal_nat)))
            (_C2, _C3) label_decomp)
          label label_decomp qmodel_cge j
          (fun f g -> qsli_to_sl (_C1, _C3, _C4) (_B2, _B3) v f g sli);;

let rec permutation_afs (_A1, _A2)
  pi = ball ((ceq_prod _A1 ceq_nat), (ccompare_prod _A2 ccompare_nat))
         (afs_syms pi)
         (fun (f, n) ->
           (match afsa pi (f, n) with Collapse _ -> false
             | AFList xs ->
               eq_set (card_UNIV_nat, cenum_nat, ceq_nat, ccompare_nat)
                 (set (ceq_nat, ccompare_nat, set_impl_nat) xs)
                 (set (ceq_nat, ccompare_nat, set_impl_nat)
                   (upt zero_nata n)) &&
                 distinct equal_nat xs));;

let rec argument_filter_tt (_B1, _B2, _B3, _B4)
  i pi tp =
    (match afs_of (_B1, _B2, _B3, _B4) pi
      with None -> Inl (showsl_lit "invalid argument filter")
      | Some af ->
        bindb (check (permutation_afs (_B1, _B2) af)
                (showsl_lit "argument filter is not a permutation"))
          (fun _ ->
            (let pia = af_rules af in
              Inr (mkc i default_nfs_trs [] (pia (rc i tp))
                    (pia (rwc i tp))))));;

let rec check_wwf_qtrs _A (_B1, _B2)
  nf r =
    catch_error
      (catch_error
        (forallM
          (fun ra ->
            (if applicable_rule_impl nf ra
              then catch_error
                     (bindb
                       (check (not (is_Var (fst ra)))
                         (showsl_literal "variable left-hand side in"))
                       (fun _ ->
                         catch_error
                           (check_subseteq _B1 (vars_term_list (snd ra))
                             (vars_term_list (fst ra)))
                           (fun x ->
                             Inl (comp (comp (showsl_literal "free variable ")
 (showsl _B2 x))
                                   (showsl_literal " in right-hand side of")))))
                     (fun x ->
                       Inl (comp (comp (comp x (showsl_literal " rule "))
                                   (showsl_rule _A _B2 ra))
                             (showsl_literal "\092")))
              else Inr ()))
          r)
        (fun x -> Inl (snd x)))
      (fun x ->
        Inl (comp (showsl_literal "the Q-TRS is not weakly well-formed\092")
              x));;

let rec check_compatible_nfs _A (_B1, _B2)
  nfs1 nf1 r1 nfs2 q2 =
    equal_boola nfs1 nfs2 ||
      (null q2 || isOK (check_wwf_qtrs _A (_B1, _B2) nf1 r1));;

let rec showsl_tp _B _C
  i t = (let nfs = nfsb i t in
         let r = rc i t in
         let rw = rwc i t in
         let q = qb i t in
          comp (comp (comp (showsl_trsa (showsl _B) (showsl _C) "rules:" " -> "
                             r)
                       (if null rw then id
                         else showsl_trsa (showsl _B) (showsl _C)
                                "relative rules:" " ->= " rw))
                 (if null q then id
                   else showsl_terms (showl_term _B _C) "Q-component:" q))
            (if nfs
              then showsl_lit
                     "substitutions are assumed to be in normal form\092"
              else id));;

let rec check_tp_subsumes (_B1, _B2) (_C1, _C2) (_D1, _D2)
  i (nfs, (q, (r, rw))) tp =
    catch_error
      (let nfsa = nfsb i tp in
       let qa = is_QNFb i tp in
       let ra = rc i tp in
       let rwa = rwc i tp in
       let rb = r @ rw in
       let nf1 = is_QNFb i tp in
        bindb (check
                (check_compatible_nfs (showl_lab _B2 _C2) (_D1, _D2) nfsa nf1
                  (ra @ rwa) nfs q)
                (showsl_lit "incompatible substitutions-in-normal-form flags"))
          (fun _ ->
            bindb (catch_error (check_NF_terms_subset qa q)
                    (fun x ->
                      Inl (comp (showsl_lit
                                  "problem with innermost strategy due to term ")
                            (showsl_terma (showl_lab _B2 _C2) _D2 x))))
              (fun _ ->
                bindb (catch_error
                        (check_subseteq
                          (equal_prod (equal_term (equal_lab _B1 _C1) _D1)
                            (equal_term (equal_lab _B1 _C1) _D1))
                          ra r)
                        (fun x ->
                          Inl (toomuch "rule"
                                (showsl_rule (showl_lab _B2 _C2) _D2 x))))
                  (fun _ ->
                    bindb (catch_error
                            (check_subseteq
                              (equal_prod (equal_term (equal_lab _B1 _C1) _D1)
                                (equal_term (equal_lab _B1 _C1) _D1))
                              rwa rb)
                            (fun x ->
                              Inl (toomuch "relative rule"
                                    (showsl_rule (showl_lab _B2 _C2) _D2 x))))
                      (fun _ -> Inr ())))))
      (fun x ->
        Inl (comp (comp (comp (comp (comp (comp
    (showsl_lit "termination of the problem\092")
    (showsl_tp (showl_lab _B2 _C2) _D2 i tp))
                                      (showsl_lit
"\092may not be concluded from assuming termination of the problem\092"))
                                (showsl_tp (showl_lab _B2 _C2) _D2 i
                                  (mkc i nfs q r rw)))
                          (showsl_literal "\092"))
                    x)
              (showsl_literal "\092")));;

let rec fcc_tt (_B1, _B2) (_C1, _C2, _C3, _C4)
  i fcs cRb tp =
    (let r = rc i tp in
     let rw = rwc i tp in
     let nfs = nfsb i tp in
     let rb = r @ rw in
     let (cR, cRw) = partition_rules _B1 _C2 fcs r cRb in
     let _ = qb i tp in
     let vs = vars_trs_list rb in
     let fas = funas_trs_list rb in
      (match
        bindb (check (not (null fcs))
                (showsl_lit
                  "at least one flat context is required for flat context closure\092"))
          (fun _ ->
            bindb (catch_error
                    (forallM (check_flat_ctxt (_C2, _C4) (_B1, _B2) vs) fcs)
                    (fun x -> Inl (snd x)))
              (fun _ ->
                bindb (catch_error
                        (forallM
                          (check_is_flat_ctxt (_C2, _C4) (_B1, _B2) vs fas) fcs)
                        (fun x -> Inl (snd x)))
                  (fun _ ->
                    bindb (catch_error
                            (forallM
                              (check_flat_ctxt_complete (_B1, _B2) _C4 fcs) fas)
                            (fun x -> Inl (snd x)))
                      (fun _ ->
                        bindb (catch_error
                                (forallM
                                  (check_rule_preserving (_B1, _B2)
                                    (_C1, _C2, _C3, _C4) fcs cR)
                                  r)
                                (fun x -> Inl (snd x)))
                          (fun _ ->
                            catch_error
                              (forallM
                                (check_rule_preserving (_B1, _B2)
                                  (_C1, _C2, _C3, _C4) fcs cRb)
                                rw)
                              (fun x -> Inl (snd x)))))))
        with Inl a -> Inl a | Inr _ -> Inr (mkc i nfs [] cR cRw)));;

let rec g_isEmpty_dflt_basic_oops_rm_basic_ops _A
  s = iteratei_bset_op_list_it_dflt_basic_oops_rm_basic_ops _A s (fun c -> c)
        (fun _ _ -> false) true;;

let rec ta_rhs_states_set (TA_Impl (x1, x2, x3, x4, x5, x6, x7)) = x4;;

let rec ta_rules_impl (TA_Impl (x1, x2, x3, x4, x5, x6, x7)) = x2;;

let rec r_lhs_states_impl (TA_rule_impl (f, qsa, q, qs)) = qsa;;

let rec rm_set_lookup _A
  rm = (fun a -> (match lookup _A rm a with None -> [] | Some rules -> rules));;

let rec g_union_dflt_basic_oops_rm_basic_ops _A
  s1 s2 =
    iteratei_bset_op_list_it_dflt_basic_oops_rm_basic_ops
      _A.linorder_compare_order s1 (fun _ -> true) (ins_rm_basic_ops _A) s2;;

let rec rs_Union _A
  = foldl (g_union_dflt_basic_oops_rm_basic_ops _A)
      (empty_rm_basic_ops _A.linorder_compare_order ());;

let rec rqss_impl (TA_rule_impl (f, qsa, q, qs)) = qs;;

let rec ta_res_impl_all _A _B
  q ta x2 = match q, ta, x2 with q, ta, Var uu -> q
    | q, ta, Fun (f, ts) ->
        (let reca = map (ta_res_impl_all _A _B q ta) ts in
         let n = size_list ts in
         let rules =
           rm_set_lookup (compare_order_prod _B compare_order_nat) ta (f, n) in
         let arules =
           filtera
             (fun rule ->
               (let qs = r_lhs_states_impl rule in
                 all_interval_nat
                   (fun i -> memb_rm_basic_ops _A (nth qs i) (nth reca i))
                   zero_nata n))
             rules
           in
         let a = map rqss_impl arules in
          rs_Union _A a);;

let rec rule_state_compatible_heuristic _A _B _C
  ta l =
    g_isEmpty_dflt_basic_oops_rm_basic_ops _A.linorder_compare_order
      (ta_res_impl_all _A _B (ta_rhs_states_set ta) (ta_rules_impl ta) l);;

let rec check_state_raise_consistent (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2)
  ta rel =
    (let rels =
       set ((ceq_prod _A1 _A1), (ccompare_prod _A2 _A2),
             (set_impl_prod _A5 _A5))
         rel
       in
     let rls = ta_rules_implb ta in
      catch_error
        (forallM
          (fun r1 ->
            (let TA_rule ((f1, i1), qs1, q1) = r1 in
              catch_error
                (forallM
                  (fun r2 ->
                    (let TA_rule ((f2, i2), qs2, q2) = r2 in
                      (if eq _B1 f1 f2 &&
                            (less_nat i1 i2 && equal_lista _A3 qs1 qs2)
                        then check (member
                                     ((ceq_prod _A1 _A1),
                                       (ccompare_prod _A2 _A2))
                                     (q1, q2) rels)
                               (comp (comp (comp
     (comp (comp (comp (comp (showsl_lit
                               "problem with raise consistency because of automaton-rules\092 ")
                         (showsl_ta_rule _A6 (showl_prod _B2 showl_nat) r1))
                   (showsl_literal "\092"))
             (showsl_ta_rule _A6 (showl_prod _B2 showl_nat) r2))
       (showsl_literal "\092"))
     (showsl _A6 q1))
                                       (showsl_lit " is not >>^* "))
                                 (showsl _A6 q2))
                        else Inr ())))
                  rls)
                (fun x -> Inl (snd x))))
          rls)
        (fun x -> Inl (snd x)));;

let rec ta_epsrs_impl (TA_Impl (x1, x2, x3, x4, x5, x6, x7)) = x7;;

let rec ta_epss_impl (TA_Impl (x1, x2, x3, x4, x5, x6, x7)) = x6;;

let rec g_inter_dflt_basic_oops_rm_basic_ops _A
  s1 s2 =
    iteratei_bset_op_list_it_dflt_basic_oops_rm_basic_ops
      _A.linorder_compare_order s1 (fun _ -> true)
      (fun x s ->
        (if memb_rm_basic_ops _A x s2 then ins_dj_rm_basic_ops _A x s else s))
      (empty_rm_basic_ops _A.linorder_compare_order ());;

let rec ta_match_impl _A (_B1, _B2) (_C1, _C2)
  ta qsig eps x3 q = match ta, qsig, eps, x3, q with
    ta, qsig, eps, Var x, q ->
      g_from_list_dflt_basic_oops_rm_basic_ops
        (compare_order_list
          ((compare_order_prod _C1 _B1), (equal_prod _C2 _B2)))
        (map (fun qa -> [(x, qa)])
          (g_to_list_dflt_basic_oops_rm_basic_ops _B1.linorder_compare_order
            (g_inter_dflt_basic_oops_rm_basic_ops _B1 (rs_Union _B1 (map eps q))
              qsig)))
    | ta, qsig, eps, Fun (f, ts), q ->
        (let n = size_list ts in
         let rules =
           rm_set_lookup (compare_order_prod _A compare_order_nat) ta (f, n) in
         let ep = rs_Union _B1 (map eps q) in
         let fa =
           (fun rule ->
             g_from_list_dflt_basic_oops_rm_basic_ops
               (compare_order_list
                 ((compare_order_prod _C1 _B1), (equal_prod _C2 _B2)))
               (let TA_rule_impl (_, qs, qa, _) = rule in
                 (if memb_rm_basic_ops _B1 qa ep
                   then (let reca =
                           map (fun (tsi, qsi) ->
                                 g_to_list_dflt_basic_oops_rm_basic_ops
                                   (linorder_list
                                     ((equal_prod _C2 _B2),
                                       (linorder_prod _C1.linorder_compare_order
 _B1.linorder_compare_order)))
                                   (ta_match_impl _A (_B1, _B2) (_C1, _C2) ta
                                     qsig eps tsi [qsi]))
                             (zip ts qs)
                           in
                          map concat (concat_lists reca))
                   else [])))
           in
          rs_Union
            (compare_order_list
              ((compare_order_prod _C1 _B1), (equal_prod _C2 _B2)))
            (map fa rules));;

let rec ta_match_impla _A (_B1, _B2) (_C1, _C2)
  ta qsig eps rhs t = ta_match_impl _A (_B1, _B2) (_C1, _C2) ta qsig eps t rhs;;

let rec ta_res_impl _A _B
  ta eps x2 = match ta, eps, x2 with ta, eps, Var q -> eps q
    | ta, eps, Fun (f, ts) ->
        (let reca = map (ta_res_impl _A _B ta eps) ts in
         let n = size_list ts in
         let rules =
           rm_set_lookup (compare_order_prod _A compare_order_nat) ta (f, n) in
         let arules =
           filtera
             (fun rule ->
               (let qs = r_lhs_states_impl rule in
                 all_interval_nat
                   (fun i -> memb_rm_basic_ops _B (nth qs i) (nth reca i))
                   zero_nata n))
             rules
           in
         let a = map rqss_impl arules in
          rs_Union _B a);;

let rec rule_state_compatible_eff_list (_A1, _A2) _B (_C1, _C2)
  ta rel (l, r) =
    (let rm = ta_rules_impl ta in
     let eps = ta_epss_impl ta in
     let epsa = ta_epsrs_impl ta in
     let ta_res = ta_res_impl _B _A1 rm eps in
     let rhs_rbt = ta_rhs_states_set ta in
     let rhs =
       g_to_list_dflt_basic_oops_rm_basic_ops _A1.linorder_compare_order rhs_rbt
       in
      catch_error
        (forallM
          (fun sigma ->
            (let sigmaa = fun_of _C2 sigma in
             let l_sigma = map_term (fun x -> x) sigmaa l in
             let r_sigma = map_term (fun x -> x) sigmaa r in
             let qsl = ta_res l_sigma in
             let qsr = ta_res r_sigma in
              (match rel qsl qsr with None -> Inr ()
                | Some q -> Inl ((l_sigma, r_sigma), q))))
          (g_to_list_dflt_basic_oops_rm_basic_ops
            (linorder_list
              ((equal_prod _C2 _A2),
                (linorder_prod _C1.linorder_compare_order
                  _A1.linorder_compare_order)))
            (ta_match_impla _B (_A1, _A2) (_C1, _C2) rm rhs_rbt epsa rhs l)))
        (fun x -> Inl (snd x)));;

let rec state_compatible_eff_list (_A1, _A2) _B (_C1, _C2)
  ta rel r =
    (let check = rule_state_compatible_eff_list (_A1, _A2) _B (_C1, _C2) ta rel
       in
      catch_error
        (forallM (fun lr -> catch_error (check lr) (fun x -> Inl (lr, x))) r)
        (fun x -> Inl (snd x)));;

let rec ta_final_impl (TA_Impl (x1, x2, x3, x4, x5, x6, x7)) = x1;;

let rec flatten_term_enum_filter
  f x1 = match f, x1 with
    f, Var x -> (let tx = Var x in (if f tx then [tx] else []))
    | f, Fun (fs, ts) ->
        (let lts = map (flatten_term_enum_filter f) ts in
          (if list_ex null lts then []
            else (let ss = concat_lists lts in
                   filtera f
                     (maps (fun fa -> map (fun a -> Fun (fa, a)) ss) fs))));;

let rec inverse_base_term_filter
  filt l c =
    (let hs = upt zero_nata (suc c) in
      flatten_term_enum_filter filt
        (map_term (fun f -> map (fun h -> lift h f) hs) (fun x -> x) l));;

let rec compute_height _A _B
  x0 bl br = match x0, bl, br with
    Weak_TRS None, bl, br ->
      (if less_eq_nat (size_multiset (funs_term_ms _A br))
            (size_multiset (funs_term_ms _A bl))
        then (fun l x ->
               (if equal_terma (equal_prod _A equal_nat) _B
                     (map_term (lift x) (fun xa -> xa) bl) l
                 then x else suc x))
        else (fun _ -> suc))
    | Weak_TRS (Some c), bl, br ->
        (if less_eq_nat (size_multiset (funs_term_ms _A br))
              (size_multiset (funs_term_ms _A bl))
          then (fun l x ->
                 (if equal_terma (equal_prod _A equal_nat) _B
                       (map_term (lift x) (fun xa -> xa) bl) l
                   then min ord_nat c x else min ord_nat c (suc x)))
          else (fun _ x -> min ord_nat c (suc x)))
    | Strict_TRS, bl, br -> (fun _ -> suc);;

let rec cover_bound_list_filter _A _B
  filt ff gg c r =
    maps (fun (l, ra) ->
           (let ch = compute_height _A _B gg l ra in
            let ee = ff (l, ra) in
             map (fun la ->
                   (la, map_term
                          (lift (ch la
                                  (min_list linorder_nat
                                    (map height
                                      (sym_collect
(fun t -> ee (map_term base (fun x -> x) t)) la)))))
                          (fun x -> x) ra))
               (inverse_base_term_filter filt l c)))
      r;;

let rec g_bex_dflt_basic_oops_rm_basic_ops _A
  s p = iteratei_bset_op_list_it_dflt_basic_oops_rm_basic_ops _A s not
          (fun x _ -> p x) false;;

let rec ta_contains_aux_impl _A (_B1, _B2, _B3, _B4)
  f qs ta q =
    (let _ = ta_final_impl ta in
     let look =
       rm_set_lookup (compare_order_prod _A compare_order_nat)
         (ta_rules_impl ta)
       in
     let _ = ta_epss_impl ta in
      catch_error
        (forallM
          (fun (fa, n) ->
            (let rules = look (fa, n) in
              catch_error
                (catch_error
                  (forallM
                    (fun x ->
                      (if list_ex
                            (fun rule ->
                              equal_lista _B3 (r_lhs_states_impl rule) x &&
                                (let qq = rqss_impl rule in
                                  g_bex_dflt_basic_oops_rm_basic_ops _B4 qq
                                    (fun qa -> member (_B1, _B2) qa q)))
                            rules
                        then Inr () else Inl x))
                    (concat_lists (replicate n qs)))
                  (fun x -> Inl (snd x)))
                (fun x -> Inl (Fun (fa, map (fun a -> Var a) x)))))
          f)
        (fun x -> Inl (snd x)));;

let rec alpha_rm_basic_ops (_A1, _A2) s = dom _A1 (lookup _A2 s);;

let rec ta_contains_impl _A (_B1, _B2, _B3, _B4, _B5, _B6)
  f g ta qs =
    bindb (ta_contains_aux_impl _A (_B2, _B3, _B5, _B4.linorder_compare_order) f
            qs ta (set (_B2, _B3, _B6) qs))
      (fun _ ->
        ta_contains_aux_impl _A (_B2, _B3, _B5, _B4.linorder_compare_order) g qs
          ta (alpha_rm_basic_ops (_B1, _B4) (ta_final_impl ta)));;

let rec check_coherent_rule (_A1, _A2, _A3) (_B1, _B2)
  iter rel rules (TA_rule (f, qs, q)) =
    catch_error
      (forallM
        (fun i ->
          (let qi = nth qs i in
            catch_error
              (forallM
                (fun qia ->
                  (let qsa = list_update qs i qia in
                    check (not (null (filtera
                                       (fun (TA_rule (g, qsb, qa)) ->
 eq _B1 f g &&
   (equal_lista _A2 qsa qsb &&
     memb_rm_basic_ops (compare_order_prod _A1 _A1) (q, qa) rel))
                                       rules)))
                      (comp (comp (comp (comp
  (comp (comp (comp (comp (comp (comp (showsl_lit "rule ") (showsl _B2 f))
                            (showsl_lit "("))
                      (showsl_list _A3 qs))
                (showsl_lit ") -> "))
          (showsl _A3 q))
    (showsl_lit " with "))
  (showsl_nat (suc i)))
                                    (showsl_lit ". argument decreased to "))
                              (showsl _A3 qia))
                        (showsl_lit " has no counterpart"))))
                (iter qi))
              (fun x -> Inl (snd x))))
        (upt zero_nata (size_list qs)))
      (fun x -> Inl (snd x));;

let rec ntrancl (_A1, _A2, _A3)
  n r = (if equal_nata n zero_nata then r
          else (let ra = ntrancl (_A1, _A2, _A3) (minus_nata n one_nata) r in
                 sup_seta ((ceq_prod _A1 _A1), (ccompare_prod _A2 _A2)) ra
                   (relcomp (_A1, _A2, _A3) (_A1, _A2) (_A1, _A2, _A3) ra r)));;

let rec trancl (_A1, _A2, _A3, _A4)
  a = (if finite
            ((finite_UNIV_prod _A1.finite_UNIV_card_UNIV
               _A1.finite_UNIV_card_UNIV),
              (ceq_prod _A2 _A2), (ccompare_prod _A3 _A3))
            a
        then ntrancl (_A2, _A3, _A4)
               (minus_nata
                 (card ((card_UNIV_prod _A1 _A1), (ceq_prod _A2 _A2),
                         (ccompare_prod _A3 _A3))
                   a)
                 one_nata)
               a
        else failwith "trancl: infinite set"
               (fun _ -> trancl (_A1, _A2, _A3, _A4) a));;

let rec check_coherent (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8) (_B1, _B2)
  uu x1 = match uu, x1 with
    Tree_Automaton (fin, rules, eps), Some_Relation rel ->
      (let iter =
         (fun q ->
           map_filter
             (fun x ->
               (if (let (a, _) = x in eq _A6 a q) then Some (snd x) else None))
             rel)
         in
       let rs =
         g_from_list_dflt_basic_oops_rm_basic_ops (compare_order_prod _A5 _A5)
           rel
         in
       let ep =
         set ((ceq_prod _A3 _A3), (ccompare_prod _A4 _A4),
               (set_impl_prod _A7 _A7))
           eps
         in
       let rell =
         converse (_A3, _A4, _A7) (_A3, _A4, _A7)
           (set ((ceq_prod _A3 _A3), (ccompare_prod _A4 _A4),
                  (set_impl_prod _A7 _A7))
             rel)
         in
        bindb (catch_error (check_subseteq _A6 (maps iter fin) fin)
                (fun x ->
                  Inl (comp (showsl _A8 x)
                        (showsl_lit
                          " is in relation to a final state, but not a final state itself"))))
          (fun _ ->
            bindb (catch_error
                    (forallM
                      (check_coherent_rule (_A5, _A6, _A8) (_B1, _B2) iter rs
                        rules)
                      rules)
                    (fun x -> Inl (snd x)))
              (fun _ ->
                check (less_eq_set
                        ((cenum_prod _A2 _A2), (ceq_prod _A3 _A3),
                          (ccompare_prod _A4 _A4))
                        (relcomp (_A3, _A4, _A7) (_A3, _A4) (_A3, _A4, _A7) rell
                          ep)
                        (sup_seta ((ceq_prod _A3 _A3), (ccompare_prod _A4 _A4))
                          (relcomp (_A3, _A4, _A7) (_A3, _A4) (_A3, _A4, _A7)
                            (trancl (_A1, _A3, _A4, _A7) ep) rell)
                          rell))
                  (showsl_lit "problem in coherence of epsilon rules"))))
    | uu, Decision_Proc_Old -> Inr ()
    | uu, Decision_Proc -> Inr ()
    | uu, Id_Relation -> Inr ();;

let rec rtrancl_rbt_impl _A
  = rtrancl_impl
      (fun r ->
        (let rm = elem_list_to_rm _A fst r in
          (fun asa ->
            g_to_list_dflt_basic_oops_rm_basic_ops _A.linorder_compare_order
              (rs_Union _A
                (map (fun a ->
                       g_from_list_dflt_basic_oops_rm_basic_ops _A
                         (map snd (rm_set_lookup _A rm a)))
                  asa)))))
      (fun asa bs ->
        g_union_dflt_basic_oops_rm_basic_ops _A bs
          (g_from_list_dflt_basic_oops_rm_basic_ops _A asa))
      (memb_rm_basic_ops _A) (empty_rm_basic_ops _A.linorder_compare_order ());;

let rec memo_rbt_rtrancl _A
  r = (let tr = rtrancl_rbt_impl _A r in
       let rm =
         g_list_to_map_rm_basic_ops _A
           (map (fun a -> (a, tr [a]))
             (comp (comp (g_to_list_dflt_basic_oops_rm_basic_ops
                           _A.linorder_compare_order)
                     (g_from_list_dflt_basic_oops_rm_basic_ops _A))
               (map fst) r))
         in
        (fun a ->
          (match lookup _A rm a
            with None -> g_from_list_dflt_basic_oops_rm_basic_ops _A [a]
            | Some asa -> asa)));;

let rec conv_ta_rule eps (TA_rule (f, qs, q)) = TA_rule_impl (f, qs, q, eps q);;

let rec r_sym_impl (TA_rule_impl (f, qsa, q, qs)) = (f, size_list qsa);;

let rec generate_ta _A _B
  (Tree_Automaton (fin, rules, eps)) =
    (let ep = memo_rbt_rtrancl _A eps in
     let epr = memo_rbt_rtrancl _A (map (fun (q, qa) -> (qa, q)) eps) in
     let rqs_rs = rs_Union _A (map (fun rule -> ep (r_rhs rule)) rules) in
     let rrules = map (conv_ta_rule ep) rules in
      TA_Impl
        (g_from_list_dflt_basic_oops_rm_basic_ops _A fin,
          elem_list_to_rm (compare_order_prod _B compare_order_nat) r_sym_impl
            rrules,
          g_to_list_dflt_basic_oops_rm_basic_ops _A.linorder_compare_order
            rqs_rs,
          rqs_rs, eps, ep, epr));;

let rec generate_ta_cond (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8)
  (_B1, _B2, _B3)
  ta rel =
    bindb (catch_error
            (check_coherent (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8) (_B2, _B3)
              ta rel)
            (fun x ->
              Inl (comp (showsl_lit
                          "automaton is not coherent w.r.t. relation\092")
                    x)))
      (fun _ -> Inr (generate_ta _A5 _B1 ta));;

let rec relation_as_list
  = function Some_Relation rel -> Inr rel
    | Id_Relation -> Inr []
    | Decision_Proc ->
        Inl (showsl_lit
              "decision procedure not available for non-left linear TRSs")
    | Decision_Proc_Old ->
        Inl (showsl_lit
              "decision procedure not available for non-left linear TRSs");;

let rec check_ta_bounded (_B1, _B2)
  ta c =
    catch_error
      (catch_error
        (forallM
          (fun x ->
            (if (let (f, _) = x in less_eq_nat (height f) c) then Inr ()
              else Inl x))
          (map fst
            (g_to_list_rm_basic_ops
              (linorder_prod (linorder_prod _B1 linorder_nat) linorder_nat)
              (ta_rules_impl ta))))
        (fun x -> Inl (snd x)))
      (fun x ->
        Inl (let (f, _) = x in
              comp (comp (showsl_prod _B2 showl_nat f)
                     (showsl_lit
                       " is symbol in TA with height larger than c = "))
                (showsl_nat c)));;

let rec bounds_condition _A (_B1, _B2)
  x0 uu = match x0, uu with Roof, uu -> Inr ()
    | Match, r ->
        catch_error
          (catch_error
            (forallM
              (fun x ->
                (if (let (l, ra) = x in
                      subseteq_mset _B1 (vars_term_ms _B1 ra)
                        (vars_term_ms _B1 l))
                  then Inr () else Inl x))
              r)
            (fun x -> Inl (snd x)))
          (fun x ->
            Inl (let (l, ra) = x in
                  comp (comp (showsl_lit "rule ") (showsl_rule _A _B2 (l, ra)))
                    (showsl_lit " is duplicating")));;

let rec non_collapsing_impl
  r = list_all (comp (fun t -> not (is_Var t)) snd) r;;

let rec construct_c_opt c r = (if non_collapsing_impl r then Some c else None);;

let rec boundstype_fun (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  = function Roof -> roof (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
    | Match -> matchb;;

let rec iteratei_set_op_list_it_rs_ops _A
  s = (fun c f -> rm_iterateoi (impl_of _A s) c (comp f fst));;

let rec rs_subset _A
  asa bs =
    iteratei_set_op_list_it_rs_ops _A.linorder_compare_order asa
      (fun a -> (match a with None -> true | Some _ -> false))
      (fun a _ -> (if memb_rm_basic_ops _A a bs then None else Some a)) None;;

let rec rel_checker (_A1, _A2)
  = function
    Some_Relation rel ->
      (fun lhs rhs ->
        (let rlist =
           g_to_list_dflt_basic_oops_rm_basic_ops _A1.linorder_compare_order rhs
           in
          (match
            catch_error
              (forallM
                (fun l ->
                  check (list_ex
                          (fun r -> membera (equal_prod _A2 _A2) rel (l, r))
                          rlist)
                    l)
                (g_to_list_dflt_basic_oops_rm_basic_ops
                  _A1.linorder_compare_order lhs))
              (fun x -> Inl (snd x))
            with Inl a -> Some a | Inr _ -> None)))
    | Id_Relation -> rs_subset _A1
    | Decision_Proc -> rs_subset _A1
    | Decision_Proc_Old -> rs_subset _A1;;

let rec check_det _A _B
  (Tree_Automaton (fin, rules, eps)) =
    catch_error
      (bindb (check (null eps) (showsl_lit "epsilon transitions not allowed"))
        (fun _ ->
          check (distinct (equal_prod _B (equal_list _A))
                  (map (fun (TA_rule (f, qs, _)) -> (f, qs))
                    (remdups (equal_ta_rule _A _B) rules)))
            (showsl_lit "some lhs occurs twice")))
      (fun x ->
        Inl (comp (showsl_lit
                    "problem when ensuring determinism of automata\092")
              x));;

let rec check_bounds_generic (_A1, _A2, _A3)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8)
  (Bounds_Info (typea, c, qfin, preTA, rel)) r s f g =
    (let c_opt = construct_c_opt c r in
     let rs = r @ s in
      bindb (generate_ta_cond (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
              ((compare_order_prod _A1 compare_order_nat),
                (equal_prod _A2 equal_nat), (showl_prod _A3 showl_nat))
              preTA rel)
        (fun ta ->
          (let rell = rel_checker (_B5, _B6) rel in
            bindb (check_wf_trs _A3 (_C6, _C8) rs)
              (fun _ ->
                bindb (check
                        (less_eq_set (_B2, _B3, _B4) (set (_B3, _B4, _B7) qfin)
                          (alpha_rm_basic_ops (_B2, _B5) (ta_final_impl ta)))
                        (showsl_lit
                          "explicitly mentioned final states must be final"))
                  (fun _ ->
                    bindb (if isOK (check_left_linear_trs _A3
                                     (_C3, _C4.ccompare_cproper_interval, _C7,
                                       _C8)
                                     rs)
                            then Inr ()
                            else bindb (catch_error
 (check_det _B6 (equal_prod _A2 equal_nat) preTA)
 (fun x ->
   Inl (comp (showsl_lit
               "for non left-linear TRS we require det. automaton\092")
         x)))
                                   (fun _ ->
                                     bindb (relation_as_list rel)
                                       (check_state_raise_consistent
 (_B3, _B4, _B6, _B5.linorder_compare_order, _B7, _B8) (_A2, _A3) preTA)))
                      (fun _ ->
                        bindb (bounds_condition _A3 (_C6, _C8) typea rs)
                          (fun _ ->
                            bindb (check_ta_bounded
                                    (_A1.linorder_compare_order, _A3) ta c)
                              (fun _ ->
                                bindb (check
(list_ex (fun q -> memb_rm_basic_ops _B5 q (ta_rhs_states_set ta)) qfin)
(showsl_lit "did not find mentioned final state in TA"))
                                  (fun _ ->
                                    bindb (catch_error
    (ta_contains_impl (compare_order_prod _A1 compare_order_nat)
      (_B2, _B3, _B4, _B5, _B6, _B7)
      (map (fun (fa, a) -> (lift zero_nata fa, a)) f)
      (map (fun (fa, a) -> (lift zero_nata fa, a)) g) ta qfin)
    (fun x ->
      Inl (comp (showsl_lit
                  "it could not be guaranteed that lift0(T(Sigma)) is accepted by TA\092")
            (comp (comp (showsl_lit "there is no transition from ")
                    (showsl_terma (showl_prod _A3 showl_nat) _B8 x))
              (showsl_lit " to a final state")))))
                                      (fun _ ->
bindb (catch_error
        (state_compatible_eff_list (_B5, _B6)
          (compare_order_prod _A1 compare_order_nat) (_C5, _C6) ta rell
          (cover_bound_list_filter _A2 _C6
            (fun l ->
              not (rule_state_compatible_heuristic _B5
                    (compare_order_prod _A1 compare_order_nat)
                    _C5.linorder_compare_order ta l))
            (boundstype_fun (_A1.compare_compare_order, _A2)
              (_C1, _C2, _C3, _C4, _C5.compare_compare_order, _C6, _C7) typea)
            Strict_TRS c r))
        (fun x ->
          Inl (let (lr, (lr_rhs, q)) = x in
                comp (comp (comp (comp (comp
 (comp (comp (showsl_lit "TA is not compatible with TRS\092")
         (showsl_lit "for rule "))
   (showsl_rule (showl_prod _A3 showl_nat) _C8 lr))
 (showsl_lit "\092which is instantiated by states to "))
                                   (showsl_rule (showl_prod _A3 showl_nat) _B8
                                     lr_rhs))
                             (showsl_lit "\092the state "))
                       (showsl _B8 q))
                  (showsl_lit " is only reachable from the lhs\092"))))
  (fun _ ->
    catch_error
      (state_compatible_eff_list (_B5, _B6)
        (compare_order_prod _A1 compare_order_nat) (_C5, _C6) ta rell
        (cover_bound_list_filter _A2 _C6
          (fun l ->
            not (rule_state_compatible_heuristic _B5
                  (compare_order_prod _A1 compare_order_nat)
                  _C5.linorder_compare_order ta l))
          matchb (Weak_TRS c_opt) c s))
      (fun x ->
        Inl (let (lr, (lr_rhs, q)) = x in
              comp (comp (comp (comp (comp (comp
     (comp (showsl_lit "TA is not compatible with relative TRS\092")
       (showsl_lit "for rule "))
     (showsl_rule (showl_prod _A3 showl_nat) _C8 lr))
                                       (showsl_lit
 "\092which is instantiated by states to "))
                                 (showsl_rule (showl_prod _A3 showl_nat) _B8
                                   lr_rhs))
                           (showsl_lit "\092the state "))
                     (showsl _B8 q))
                (showsl_lit
                  " is only reachable from the lhs\092"))))))))))))));;

let rec bounds_tt (_B1, _B2, _B3) (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8)
  (_D1, _D2, _D3, _D4, _D5, _D6, _D7, _D8)
  i info tp =
    (let r = rulesd i tp in
     let f = funas_trs_list r in
      check_bounds_generic (_B1, _B2, _B3)
        (_D1, _D2, _D3, _D4, _D5, _D6, _D7, _D8)
        (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8) info r [] f f);;

let rec check_trs_termination_proof_main (_B1, _B2, _B3, _B4)
  j ia assms i tp x5 = match j, ia, assms, i, tp, x5 with
    j, ia, assms, i, tp, R_is_Empty ->
      debug i "R is empty"
        (if null (rc j tp) then Inr ()
          else Inl (comp (comp i
                           (showsl_lit
                             ": R is not empty in the following termination-problem \092"))
                     (showsl_tp (showl_lab _B4 (showl_list showl_nat))
                       (showl_list showl_char) j tp)))
    | j, ia, assms, i, tp, Rule_Removal (redp, rR, prf) ->
        debug i "Rule Removal"
          (let r = rulesd j tp in
           let rr =
             ceta_list_diff
               (compare_order_prod
                 (compare_order_term
                   (compare_lab _B1.compare_compare_order
                     (compare_list compare_nat))
                   (compare_list compare_char))
                 (compare_order_term
                   (compare_lab _B1.compare_compare_order
                     (compare_list compare_nat))
                   (compare_list compare_char)))
               r rR
             in
            bindb (catch_error
                    (rule_removal_tt (showl_lab _B4 (showl_list showl_nat))
                      (showl_list showl_char) j
                      (get_redtriple
                        ((ceq_lab _B3 (equal_list equal_nat)),
                          (ccompare_lab _B1.compare_compare_order
                            (compare_list compare_nat)),
                          (compare_order_lab _B1.compare_compare_order
                            (compare_list compare_nat)),
                          (equal_lab _B3 (equal_list equal_nat)), set_impl_lab,
                          (showl_lab _B4 (showl_list showl_nat)))
                        redp)
                      rr tp)
                    (fun x ->
                      Inl (comp (comp (comp
(comp (comp (comp i
              (showsl_lit
                ": error when applying the rule removal technique on \092"))
        (showsl_tp (showl_lab _B4 (showl_list showl_nat))
          (showl_list showl_char) j tp))
  (showsl_lit "\092 to remove the rules \092"))
(showsl_rules (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char)
  rr))
                                  (showsl_literal "\092"))
                            x)))
              (fun tpa ->
                catch_error
                  (check_trs_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                    assms (add_index i one_nata) tpa prf)
                  (fun x ->
                    Inl (comp (comp i
                                (showsl_lit
                                  ": error below the rule removal technique\092"))
                          x))))
    | j, ia, assms, i, tp, DP_Trans (nfs, m, p, prf) ->
        debug i "DP trans"
          (bindb
            (catch_error
              (dependency_pairs_tt
                ((compare_lab _B1.compare_compare_order
                   (compare_list compare_nat)),
                  (equal_lab _B3 (equal_list equal_nat)),
                  (linorder_lab _B1.compare_compare_order
                    (compare_list compare_nat)),
                  (showl_lab _B4 (showl_list showl_nat)))
                ((ccompare_list ccompare_char), (compare_list compare_char),
                  (equal_list equal_char), mapping_impl_list,
                  (linorder_list (equal_char, linorder_char)),
                  (showl_list showl_char))
                (fun a -> Sharp a) j ia tp nfs m p)
              (fun x ->
                Inl (comp (comp (comp (comp
(comp (comp i (showsl_lit ": error when switching from the TRS\092"))
  (showsl_tp (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char) j
    tp))
(showsl_lit "\092to the initial DP problem with pairs \092"))
                                  (showsl_rules
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) p))
                            (showsl_literal "\092"))
                      x)))
            (fun dpp ->
              catch_error
                (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia assms
                  (add_index i one_nata) dpp prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below switch to dependency pairs\092"))
                        x))))
    | j, ia, assms, i, tp, String_Reversal prf ->
        debug i "String Reversal"
          (bindb
            (catch_error
              (string_reversal_tt (showl_lab _B4 (showl_list showl_nat))
                (showl_list showl_char) j tp)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit ": error when applying string reversal on \092"))
                                  (showsl_tp
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) j tp))
                            (showsl_literal "\092"))
                      x)))
            (fun tpa ->
              catch_error
                (check_trs_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                  assms (add_index i one_nata) tpa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the string reversal technique\092"))
                        x))))
    | j, ia, assms, i, tp, Constant_String (p, prf) ->
        debug i "Constant to Unary"
          (bindb
            (catch_error
              (const_to_string_sound_tt
                ((compare_lab _B1.compare_compare_order
                   (compare_list compare_nat)),
                  (equal_lab _B3 (equal_list equal_nat)),
                  (showl_lab _B4 (showl_list showl_nat)))
                (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
                  (cproper_interval_list ccompare_char),
                  (compare_list compare_char), (equal_list equal_char),
                  set_impl_list, (showl_list showl_char))
                p j tp)
              (fun x ->
                Inl (comp (comp (comp (comp
(comp i (showsl_lit ": error when turning constants into unary symbols on "))
(showsl_literal "\092"))
                                  (showsl_tp
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) j tp))
                            (showsl_literal "\092"))
                      x)))
            (fun tpa ->
              catch_error
                (check_trs_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                  assms (add_index i one_nata) tpa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the constant to unary technique\092"))
                        x))))
    | j, ia, assms, i, tp, Semlab (sli, lQ, lAll, prf) ->
        debug i "Semlab"
          (bindb
            (catch_error
              (semlab_fin_tt (_B1.compare_compare_order, _B3, _B4)
                ((ccompare_list ccompare_char), (compare_list compare_char),
                  (equal_list equal_char), mapping_impl_list,
                  (showl_list showl_char))
                j sli lQ lAll tp)
              (fun x ->
                Inl (comp (comp (comp (comp
(comp i (showsl_lit ": error when applying semantic labelling on "))
(showsl_literal "\092"))
                                  (showsl_tp
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) j tp))
                            (showsl_literal "\092"))
                      x)))
            (fun tpa ->
              catch_error
                (check_trs_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                  assms (add_index i one_nata) tpa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the sem.lab technique\092"))
                        x))))
    | j, ia, assms, i, tp, Bounds info ->
        debug i "Bounds"
          (catch_error
            (bounds_tt
              ((compare_order_lab _B1.compare_compare_order
                 (compare_list compare_nat)),
                (equal_lab _B3 (equal_list equal_nat)),
                (showl_lab _B4 (showl_list showl_nat)))
              (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
                (cproper_interval_list ccompare_char),
                (compare_order_list (compare_order_char, equal_char)),
                (equal_list equal_char), set_impl_list, (showl_list showl_char))
              (card_UNIV_list, cenum_list, (ceq_list ceq_char),
                (ccompare_list ccompare_char),
                (compare_order_list (compare_order_char, equal_char)),
                (equal_list equal_char), set_impl_list, (showl_list showl_char))
              j info tp)
            (fun x ->
              Inl (comp (comp (comp (comp i
                                      (showsl_lit
": error when applying bounds on the termination problem \092"))
                                (showsl_tp
                                  (showl_lab _B4 (showl_list showl_nat))
                                  (showl_list showl_char) j tp))
                          (showsl_literal "\092"))
                    x)))
    | j, ia, assms, i, tp, Uncurry (u_info, rR, prf) ->
        debug i "Uncurry"
          (bindb
            (catch_error
              (uncurry_tt
                ((equal_lab _B3 (equal_list equal_nat)),
                  (showl_lab _B4 (showl_list showl_nat)))
                j u_info rR tp)
              (fun x ->
                Inl (comp (comp i
                            (showsl_lit
                              ": error when applying currying transformation\092"))
                      x)))
            (fun tpa ->
              catch_error
                (check_trs_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                  assms (add_index i one_nata) tpa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the uncurrying technique\092"))
                        x))))
    | j, ia, assms, i, tp, Fcc (fcs, r, prf) ->
        debug i "Fcc"
          (bindb
            (catch_error
              (fcc_tt
                ((equal_lab _B3 (equal_list equal_nat)),
                  (showl_lab _B4 (showl_list showl_nat)))
                ((ccompare_list ccompare_char), (equal_list equal_char),
                  mapping_impl_list, (showl_list showl_char))
                j fcs r tp)
              (fun x ->
                Inl (comp (comp i
                            (showsl_lit
                              ": error when applying flat context closure\092"))
                      x)))
            (fun tpa ->
              catch_error
                (check_trs_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                  assms (add_index i one_nata) tpa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below flat context closure\092"))
                        x))))
    | j, ia, assms, i, tp, Split (rrem, prf1, prf2) ->
        debug i "Split"
          (let (tp1, tp2) =
             split_tt (showl_lab _B4 (showl_list showl_nat))
               (showl_list showl_char) j tp rrem
             in
            bindb (catch_error
                    (check_trs_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                      assms (add_index i one_nata) tp1 prf1)
                    (fun x ->
                      Inl (comp (comp i
                                  (showsl_lit
                                    ": error below the splitting\092"))
                            x)))
              (fun _ ->
                catch_error
                  (check_trs_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                    assms
                    (add_index i (nat_of_integer (Big_int.big_int_of_int 2)))
                    tp2 prf2)
                  (fun x ->
                    Inl (comp (comp i
                                (showsl_lit ": error below the splitting\092"))
                          x))))
    | j, ia, assms, i, tp, Switch_Innermost (joins, prf) ->
        debug i "Switch Innermost"
          (bindb
            (catch_error
              (switch_innermost_tt
                ((equal_lab _B3 (equal_list equal_nat)),
                  (showl_lab _B4 (showl_list showl_nat)))
                j joins tp)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit ": error when switching to innermost on \092"))
                                  (showsl_tp
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) j tp))
                            (showsl_literal "\092"))
                      x)))
            (fun tpa ->
              catch_error
                (check_trs_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                  assms (add_index i one_nata) tpa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the switch to innermost\092"))
                        x))))
    | j, ia, assms, i, tp, Drop_Equality prf ->
        debug i "Drop Equality"
          (let tpa =
             mkc j (nfsb j tp) (qb j tp) (rc j tp)
               (filtera
                 (fun (l, r) ->
                   not (equal_terma (equal_lab _B3 (equal_list equal_nat))
                         (equal_list equal_char) l r))
                 (rwc j tp))
             in
            catch_error
              (check_trs_termination_proof_main (_B1, _B2, _B3, _B4) j ia assms
                (add_index i one_nata) tpa prf)
              (fun x ->
                Inl (comp (comp i
                            (showsl_lit
                              ": error below dropping equality rules\092"))
                      x)))
    | j, ia, assms, i, tp, Remove_Nonapplicable_Rules (r, prf) ->
        debug i "Removing non-applicable rules"
          (let _ = rc j tp in
            bindb (catch_error (check_non_applicable_rules (is_QNFb j tp) r)
                    (fun x ->
                      Inl (comp (comp i
                                  (showsl_lit
                                    ": error when removing non-applicable rules\092"))
                            (comp (showsl_rule
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) x)
                              (showsl_lit " is applicable")))))
              (fun _ ->
                (let tpa = delete_R_Rwb j tp r r in
                  catch_error
                    (check_trs_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                      assms (add_index i one_nata) tpa prf)
                    (fun x ->
                      Inl (comp (comp i
                                  (showsl_lit
                                    ": error below the non-applicable rules removal\092"))
                            x)))))
    | j, ia, assms, i, tp, Permuting_AFS (pi, prf) ->
        debug i "Permuting some rules"
          (bindb
            (catch_error
              (argument_filter_tt
                ((ceq_lab _B3 (equal_list equal_nat)),
                  (ccompare_lab _B1.compare_compare_order
                    (compare_list compare_nat)),
                  (compare_order_lab _B1.compare_compare_order
                    (compare_list compare_nat)),
                  set_impl_lab)
                j pi tp)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit ": error when permuting arguments on \092"))
                                  (showsl_tp
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) j tp))
                            (showsl_literal "\092"))
                      x)))
            (fun tpa ->
              catch_error
                (check_trs_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                  assms (add_index i one_nata) tpa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the permutation of arguments\092"))
                        x))))
    | j, ia, assms, i, tpa, Assume_SN (tp, ass) ->
        debug i "Termination Assumption or Unknown Proof"
          (if assms
            then bindb (catch_error
                         (check_tp_subsumes (_B3, _B4)
                           ((equal_list equal_nat), (showl_list showl_nat))
                           ((equal_list equal_char), (showl_list showl_char)) j
                           tp tpa)
                         (fun x ->
                           Inl (comp (comp (comp i
     (showsl_lit ": error in termination assumption or unknown proof\092"))
                                       x)
                                 (showsl_literal "\092"))))
                   (fun _ ->
                     catch_error
                       (forallM_index
                         (fun asa ja ->
                           check_assma (_B1, _B2, _B4) j ia
                             (check_trs_termination_proof_main
                               (_B1, _B2, _B3, _B4) j ia assms
                               (add_index i (suc ja)))
                             (check_dp_termination_proof_main
                               (_B1, _B2, _B3, _B4) j ia assms
                               (add_index i (suc ja)))
                             (check_fptrs_termination_proof_main
                               (_B1, _B2, _B3, _B4) j ia assms
                               (add_index i (suc ja)))
                             (check_unknown_proof_main (_B1, _B2, _B3, _B4) j ia
                               assms (add_index i (suc ja)))
                             asa)
                         ass)
                       (fun x -> Inl (snd x)))
            else Inl (comp i
                       (showsl_lit
                         ": the proof contains a termination assumption or unknown proof\092")))
and check_fptrs_termination_proof_main (_B1, _B2, _B3, _B4)
  j ia assms i tpa (Assume_FP_SN (tp, ass)) =
    debug i "Outermost Termination Assumption or Unknown Proof"
      (if assms
        then bindb (check
                     (equal_proda
                       (equal_list
                         (equal_prod
                           (equal_ctxt (equal_lab _B3 (equal_list equal_nat))
                             (equal_list equal_char))
                           (equal_prod
                             (equal_term (equal_lab _B3 (equal_list equal_nat))
                               (equal_list equal_char))
                             equal_location)))
                       (equal_list
                         (equal_prod
                           (equal_term (equal_lab _B3 (equal_list equal_nat))
                             (equal_list equal_char))
                           (equal_term (equal_lab _B3 (equal_list equal_nat))
                             (equal_list equal_char))))
                       tpa tp)
                     (showsl_lit
                       "outermost assumption does not match current goal"))
               (fun _ ->
                 catch_error
                   (forallM_index
                     (fun asa ja ->
                       check_assma (_B1, _B2, _B4) j ia
                         (check_trs_termination_proof_main (_B1, _B2, _B3, _B4)
                           j ia assms (add_index i (suc ja)))
                         (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j
                           ia assms (add_index i (suc ja)))
                         (check_fptrs_termination_proof_main
                           (_B1, _B2, _B3, _B4) j ia assms
                           (add_index i (suc ja)))
                         (check_unknown_proof_main (_B1, _B2, _B3, _B4) j ia
                           assms (add_index i (suc ja)))
                         asa)
                     ass)
                   (fun x -> Inl (snd x)))
        else Inl (comp i
                   (showsl_lit
                     ": the proof contains a termination assumption or unknown proof\092")))
and check_dp_termination_proof_main (_B1, _B2, _B3, _B4)
  j ia assms i dpp x5 = match j, ia, assms, i, dpp, x5 with
    j, ia, assms, i, dpp, P_is_Empty ->
      debug i "P is empty"
        (if null (pb ia dpp) && (null (pwb ia dpp) || null (rd ia dpp))
          then Inr ()
          else Inl (comp (comp i
                           (showsl_lit
                             ": P is not empty in the following DP-problem\092"))
                     (showsl_dpp (showl_lab _B4 (showl_list showl_nat))
                       (showl_list showl_char) ia dpp)))
    | j, ia, assms, i, dpp, Subterm_Criterion_Proc (p, rseq, rP, prf) ->
        debug i "Subterm_Criterion_Proc"
          (let pa = pairsb ia dpp in
           let pr =
             ceta_list_diff
               (compare_order_prod
                 (compare_order_term
                   (compare_lab _B1.compare_compare_order
                     (compare_list compare_nat))
                   (compare_list compare_char))
                 (compare_order_term
                   (compare_lab _B1.compare_compare_order
                     (compare_list compare_nat))
                   (compare_list compare_char)))
               pa rP
             in
            bindb (catch_error
                    (subterm_criterion_proc
                      ((compare_order_lab _B1.compare_compare_order
                         (compare_list compare_nat)),
                        (equal_lab _B3 (equal_list equal_nat)),
                        (showl_lab _B4 (showl_list showl_nat)))
                      ((ccompare_list ccompare_char),
                        (compare_order_list (compare_order_char, equal_char)),
                        (equal_list equal_char), mapping_impl_list,
                        (showl_list showl_char))
                      ia p rseq pr dpp)
                    (fun x ->
                      Inl (comp (comp (comp
(comp (comp (comp i
              (showsl_lit
                ": error when applying the subterm criterion to the DP problem\092"))
        (showsl_dpp (showl_lab _B4 (showl_list showl_nat))
          (showl_list showl_char) ia dpp))
  (showsl_lit "\092and trying to remove the pairs\092"))
(showsl_rules (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char)
  pr))
                                  (showsl_literal "\092"))
                            x)))
              (fun dppa ->
                catch_error
                  (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                    assms (add_index i one_nata) dppa prf)
                  (fun x ->
                    Inl (comp (comp i
                                (showsl_lit
                                  ": error below the subterm criterion\092"))
                          x))))
    | j, ia, assms, i, dpp, Gen_Subterm_Criterion_Proc (p, pr, prf) ->
        debug i "Gen_Subterm_Criterion_Proc"
          (bindb
            (catch_error
              (generalized_subterm_proc
                ((ceq_lab _B3 (equal_list equal_nat)),
                  (ccompare_lab _B1.compare_compare_order
                    (compare_list compare_nat)),
                  (compare_lab _B1.compare_compare_order
                    (compare_list compare_nat)),
                  (equal_lab _B3 (equal_list equal_nat)), set_impl_lab,
                  (showl_lab _B4 (showl_list showl_nat)))
                ((compare_list compare_char), (equal_list equal_char),
                  (showl_list showl_char))
                ia p pr dpp)
              (fun x ->
                Inl (comp (comp (comp (comp
(comp (comp i
        (showsl_lit
          ": error when applying the generalized subterm criterion to the DP problem\092"))
  (showsl_dpp (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char) ia
    dpp))
(showsl_lit "\092and trying to remove the pairs\092"))
                                  (showsl_rules
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) pr))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia assms
                  (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the generalized subterm criterion\092"))
                        x))))
    | j, ia, assms, i, dpp, Redpair_Proc (redp, rP, prf) ->
        debug i "Redpair_Proc"
          (let p = pairsb ia dpp in
           let pr =
             ceta_list_diff
               (compare_order_prod
                 (compare_order_term
                   (compare_lab _B1.compare_compare_order
                     (compare_list compare_nat))
                   (compare_list compare_char))
                 (compare_order_term
                   (compare_lab _B1.compare_compare_order
                     (compare_list compare_nat))
                   (compare_list compare_char)))
               p rP
             in
           let proc =
             (match redp
               with Inl rrp ->
                 generic_ur_af_root_redtriple_proc
                   ((compare_order_lab _B1.compare_compare_order
                      (compare_list compare_nat)),
                     (equal_lab _B3 (equal_list equal_nat)),
                     (showl_lab _B4 (showl_list showl_nat)))
                   ia (get_root_redtriple
                        ((ceq_lab _B3 (equal_list equal_nat)),
                          (ccompare_lab _B1.compare_compare_order
                            (compare_list compare_nat)),
                          (compare_order_lab _B1.compare_compare_order
                            (compare_list compare_nat)),
                          (equal_lab _B3 (equal_list equal_nat)), set_impl_lab,
                          (showl_lab _B4 (showl_list showl_nat)))
                        rrp)
                   None
               | Inr rp ->
                 generic_ur_af_redtriple_proc
                   ((compare_order_lab _B1.compare_compare_order
                      (compare_list compare_nat)),
                     (equal_lab _B3 (equal_list equal_nat)),
                     (showl_lab _B4 (showl_list showl_nat)))
                   ia (get_redtriple
                        ((ceq_lab _B3 (equal_list equal_nat)),
                          (ccompare_lab _B1.compare_compare_order
                            (compare_list compare_nat)),
                          (compare_order_lab _B1.compare_compare_order
                            (compare_list compare_nat)),
                          (equal_lab _B3 (equal_list equal_nat)), set_impl_lab,
                          (showl_lab _B4 (showl_list showl_nat)))
                        rp)
                   None)
             in
            bindb (catch_error (proc pr dpp)
                    (fun x ->
                      Inl (comp (comp (comp
(comp (comp (comp i
              (showsl_lit
                ": error when applying the reduction pair processor to remove from the DP problem\092"))
        (showsl_dpp (showl_lab _B4 (showl_list showl_nat))
          (showl_list showl_char) ia dpp))
  (showsl_lit "\092 the pairs\092"))
(showsl_rules (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char)
  pr))
                                  (showsl_literal "\092"))
                            x)))
              (fun dppa ->
                catch_error
                  (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                    assms (add_index i one_nata) dppa prf)
                  (fun x ->
                    Inl (comp (comp i
                                (showsl_lit
                                  ": error below the reduction pair processor\092"))
                          x))))
    | j, ia, assms, i, dpp, Usable_Rules_Proc (u, prf) ->
        debug i "Usable_Rules_Proc"
          (bindb
            (catch_error
              (usable_rules_proc
                ((compare_order_lab _B1.compare_compare_order
                   (compare_list compare_nat)),
                  (equal_lab _B3 (equal_list equal_nat)),
                  (showl_lab _B4 (showl_list showl_nat)))
                ia u dpp)
              (fun x ->
                Inl (comp (comp (comp (comp
(comp (comp i
        (showsl_lit
          ": error when applying the usable rules processor to restrict the DP problem\092"))
  (showsl_dpp (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char) ia
    dpp))
(showsl_lit "\092to the usable rules\092"))
                                  (showsl_rules
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) u))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia assms
                  (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the usable rules processor\092"))
                        x))))
    | j, ia, assms, i, dpp, Q_Reduction_Proc (q, prf) ->
        debug i "Q_Reduction_Proc"
          (bindb
            (catch_error
              (q_reduction_proc
                ((compare_order_lab _B1.compare_compare_order
                   (compare_list compare_nat)),
                  (equal_lab _B3 (equal_list equal_nat)),
                  (showl_lab _B4 (showl_list showl_nat)))
                ((ccompare_list ccompare_char), (equal_list equal_char),
                  mapping_impl_list, (showl_list showl_char))
                ia q dpp)
              (fun x ->
                Inl (comp (comp i
                            (showsl_lit
                              ": error when applying the Q-reduction processor\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia assms
                  (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the Q-reduction processor\092"))
                        x))))
    | j, ia, assms, i, dpp, Mono_Redpair_Proc (redp, rP, rR, prf) ->
        debug i "Mono_Redpair_Proc"
          (let p = pairsb ia dpp in
           let pr =
             ceta_list_diff
               (compare_order_prod
                 (compare_order_term
                   (compare_lab _B1.compare_compare_order
                     (compare_list compare_nat))
                   (compare_list compare_char))
                 (compare_order_term
                   (compare_lab _B1.compare_compare_order
                     (compare_list compare_nat))
                   (compare_list compare_char)))
               p rP
             in
           let r = rulese ia dpp in
           let rr =
             ceta_list_diff
               (compare_order_prod
                 (compare_order_term
                   (compare_lab _B1.compare_compare_order
                     (compare_list compare_nat))
                   (compare_list compare_char))
                 (compare_order_term
                   (compare_lab _B1.compare_compare_order
                     (compare_list compare_nat))
                   (compare_list compare_char)))
               r rR
             in
            bindb (catch_error
                    (mono_redpair_proc (showl_lab _B4 (showl_list showl_nat))
                      (showl_list showl_char) ia
                      (get_redtriple
                        ((ceq_lab _B3 (equal_list equal_nat)),
                          (ccompare_lab _B1.compare_compare_order
                            (compare_list compare_nat)),
                          (compare_order_lab _B1.compare_compare_order
                            (compare_list compare_nat)),
                          (equal_lab _B3 (equal_list equal_nat)), set_impl_lab,
                          (showl_lab _B4 (showl_list showl_nat)))
                        redp)
                      pr rr dpp)
                    (fun x ->
                      Inl (comp (comp (comp
(comp (comp (comp (comp (comp i
                          (showsl_lit
                            ": error when applying the mono reduction pair processor to remove from the DP problem\092"))
                    (showsl_dpp (showl_lab _B4 (showl_list showl_nat))
                      (showl_list showl_char) ia dpp))
              (showsl_lit "\092 the pairs\092"))
        (showsl_rules (showl_lab _B4 (showl_list showl_nat))
          (showl_list showl_char) pr))
  (showsl_lit "\092 and the rules\092"))
(showsl_rules (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char)
  rr))
                                  (showsl_literal "\092"))
                            x)))
              (fun dppa ->
                catch_error
                  (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                    assms (add_index i one_nata) dppa prf)
                  (fun x ->
                    Inl (comp (comp i
                                (showsl_lit
                                  ": error below the mono reduction pair processor\092"))
                          x))))
    | j, ia, assms, i, dpp, Mono_URM_Redpair_Proc (redp, pr, rr, prf) ->
        debug i "Mono_URM_Redpair_Proc"
          (bindb
            (catch_error
              (mono_urm_redpair_proc
                ((compare_order_lab _B1.compare_compare_order
                   (compare_list compare_nat)),
                  (equal_lab _B3 (equal_list equal_nat)),
                  (showl_lab _B4 (showl_list showl_nat)))
                ia (get_redtriple
                     ((ceq_lab _B3 (equal_list equal_nat)),
                       (ccompare_lab _B1.compare_compare_order
                         (compare_list compare_nat)),
                       (compare_order_lab _B1.compare_compare_order
                         (compare_list compare_nat)),
                       (equal_lab _B3 (equal_list equal_nat)), set_impl_lab,
                       (showl_lab _B4 (showl_list showl_nat)))
                     redp)
                pr rr dpp)
              (fun x ->
                Inl (comp (comp (comp (comp
(comp (comp (comp (comp i
                    (showsl_lit
                      ": error when applying the mono reduction pair processor with usable repl. map to remove from the DP problem\092"))
              (showsl_dpp (showl_lab _B4 (showl_list showl_nat))
                (showl_list showl_char) ia dpp))
        (showsl_lit "\092 the pairs\092"))
  (showsl_rules (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char)
    pr))
(showsl_lit "\092 and the rules\092"))
                                  (showsl_rules
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) rr))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia assms
                  (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the mono reduction pair processor\092"))
                        x))))
    | j, ia, assms, i, dpp, Dep_Graph_Proc edpts ->
        debug i "Dep_Graph_Proc"
          (bindb
            (catch_error
              (dep_graph_proc
                (cenum_lab, (ceq_lab _B3 (equal_list equal_nat)),
                  (ccompare_lab _B1.compare_compare_order
                    (compare_list compare_nat)),
                  (compare_order_lab _B1.compare_compare_order
                    (compare_list compare_nat)),
                  (equal_lab _B3 (equal_list equal_nat)), mapping_impl_lab,
                  set_impl_lab, (showl_lab _B4 (showl_list showl_nat)))
                ia dpp edpts)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit
  ": error while trying to perform Sctxt_closure-decomposition  on\092"))
                                  (showsl_dpp
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) ia dpp))
                            (showsl_literal "\092"))
                      x)))
            (fun pdpps ->
              catch_error
                (catch_error
                  (forallM_index
                    (fun (prof, dppa) ja ->
                      check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                        assms (add_index i (suc ja)) dppa prof)
                    pdpps)
                  (fun x -> Inl (snd x)))
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the dependency graph processor\092"))
                        x))))
    | j, ia, assms, i, dpp, Redpair_UR_Proc (redp, rP, ur, prf) ->
        debug i "Redpair_UR_Proc"
          (let p = pairsb ia dpp in
           let pr =
             ceta_list_diff
               (compare_order_prod
                 (compare_order_term
                   (compare_lab _B1.compare_compare_order
                     (compare_list compare_nat))
                   (compare_list compare_char))
                 (compare_order_term
                   (compare_lab _B1.compare_compare_order
                     (compare_list compare_nat))
                   (compare_list compare_char)))
               p rP
             in
           let proc =
             (match redp
               with Inl rrp ->
                 generic_ur_af_root_redtriple_proc
                   ((compare_order_lab _B1.compare_compare_order
                      (compare_list compare_nat)),
                     (equal_lab _B3 (equal_list equal_nat)),
                     (showl_lab _B4 (showl_list showl_nat)))
                   ia (get_root_redtriple
                        ((ceq_lab _B3 (equal_list equal_nat)),
                          (ccompare_lab _B1.compare_compare_order
                            (compare_list compare_nat)),
                          (compare_order_lab _B1.compare_compare_order
                            (compare_list compare_nat)),
                          (equal_lab _B3 (equal_list equal_nat)), set_impl_lab,
                          (showl_lab _B4 (showl_list showl_nat)))
                        rrp)
                   (Some ur)
               | Inr rp ->
                 generic_ur_af_redtriple_proc
                   ((compare_order_lab _B1.compare_compare_order
                      (compare_list compare_nat)),
                     (equal_lab _B3 (equal_list equal_nat)),
                     (showl_lab _B4 (showl_list showl_nat)))
                   ia (get_redtriple
                        ((ceq_lab _B3 (equal_list equal_nat)),
                          (ccompare_lab _B1.compare_compare_order
                            (compare_list compare_nat)),
                          (compare_order_lab _B1.compare_compare_order
                            (compare_list compare_nat)),
                          (equal_lab _B3 (equal_list equal_nat)), set_impl_lab,
                          (showl_lab _B4 (showl_list showl_nat)))
                        rp)
                   (Some ur))
             in
            bindb (catch_error (proc pr dpp)
                    (fun x ->
                      Inl (comp (comp (comp
(comp (comp (comp i
              (showsl_lit
                ": error when applying the reduction pair processor with usable rules to remove from the DP problem\092"))
        (showsl_dpp (showl_lab _B4 (showl_list showl_nat))
          (showl_list showl_char) ia dpp))
  (showsl_lit "\092 the pairs\092"))
(showsl_rules (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char)
  pr))
                                  (showsl_literal "\092"))
                            x)))
              (fun dppa ->
                catch_error
                  (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                    assms (add_index i one_nata) dppa prf)
                  (fun x ->
                    Inl (comp (comp i
                                (showsl_lit
                                  ": error below the reduction pair processor\092"))
                          x))))
    | j, ia, assms, i, dpp, Mono_Redpair_UR_Proc (redp, rP, rR, ur, prf) ->
        debug i "Mono_Redpair_UR_Proc"
          (let p = pairsb ia dpp in
           let pr =
             ceta_list_diff
               (compare_order_prod
                 (compare_order_term
                   (compare_lab _B1.compare_compare_order
                     (compare_list compare_nat))
                   (compare_list compare_char))
                 (compare_order_term
                   (compare_lab _B1.compare_compare_order
                     (compare_list compare_nat))
                   (compare_list compare_char)))
               p rP
             in
           let r = rulese ia dpp in
           let rr =
             ceta_list_diff
               (compare_order_prod
                 (compare_order_term
                   (compare_lab _B1.compare_compare_order
                     (compare_list compare_nat))
                   (compare_list compare_char))
                 (compare_order_term
                   (compare_lab _B1.compare_compare_order
                     (compare_list compare_nat))
                   (compare_list compare_char)))
               r rR
             in
            bindb (catch_error
                    (generic_mono_ur_redpair_proc
                      ((ceq_lab _B3 (equal_list equal_nat)),
                        (ccompare_lab _B1.compare_compare_order
                          (compare_list compare_nat)),
                        (compare_order_lab _B1.compare_compare_order
                          (compare_list compare_nat)),
                        (equal_lab _B3 (equal_list equal_nat)), set_impl_lab,
                        (showl_lab _B4 (showl_list showl_nat)))
                      ia (get_redtriple
                           ((ceq_lab _B3 (equal_list equal_nat)),
                             (ccompare_lab _B1.compare_compare_order
                               (compare_list compare_nat)),
                             (compare_order_lab _B1.compare_compare_order
                               (compare_list compare_nat)),
                             (equal_lab _B3 (equal_list equal_nat)),
                             set_impl_lab,
                             (showl_lab _B4 (showl_list showl_nat)))
                           redp)
                      pr rr ur dpp)
                    (fun x ->
                      Inl (comp (comp (comp
(comp (comp (comp (comp (comp i
                          (showsl_lit
                            ": error when applying the mono reduction pair processor with usable rules to remove from the DP problem\092"))
                    (showsl_dpp (showl_lab _B4 (showl_list showl_nat))
                      (showl_list showl_char) ia dpp))
              (showsl_lit "\092 the pairs\092"))
        (showsl_rules (showl_lab _B4 (showl_list showl_nat))
          (showl_list showl_char) pr))
  (showsl_lit "\092 and the rules\092"))
(showsl_rules (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char)
  rr))
                                  (showsl_literal "\092"))
                            x)))
              (fun dppa ->
                catch_error
                  (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                    assms (add_index i one_nata) dppa prf)
                  (fun x ->
                    Inl (comp (comp i
                                (showsl_lit
                                  ": error below the mono reduction pair processor with usable rules\092"))
                          x))))
    | j, ia, assms, i, dpp, Uncurry_Proc (mode, u_info, p, r, prf) ->
        debug i "Uncurry_Proc"
          (bindb
            (catch_error
              (uncurry_proc_both (_B3, _B4)
                ((equal_list equal_nat), (showl_list showl_nat)) ia mode u_info
                p r dpp)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit
  ": error when applying the uncurrying processor on the DP problem\092"))
                                  (showsl_dpp
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) ia dpp))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia assms
                  (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the uncurrying processor\092"))
                        x))))
    | j, ia, assms, i, dpp, Size_Change_Subterm_Proc graphs ->
        debug i "Size_Change_Subterm_Proc"
          (catch_error
            (sct_subterm_proc
              ((compare_order_lab _B1.compare_compare_order
                 (compare_list compare_nat)),
                (equal_lab _B3 (equal_list equal_nat)),
                (showl_lab _B4 (showl_list showl_nat)))
              ia graphs dpp)
            (fun x ->
              Inl (comp (comp (comp (comp i
                                      (showsl_lit
": error when applying the size-change (subterm) processor on the DP problem\092"))
                                (showsl_dpp
                                  (showl_lab _B4 (showl_list showl_nat))
                                  (showl_list showl_char) ia dpp))
                          (showsl_literal "\092"))
                    x)))
    | j, ia, assms, i, dpp, Size_Change_Redpair_Proc (redp, u_opt, graphs) ->
        debug i "Size_Change_Redpair_Proc"
          (catch_error
            (sct_ur_af_proc
              ((compare_order_lab _B1.compare_compare_order
                 (compare_list compare_nat)),
                (equal_lab _B3 (equal_list equal_nat)),
                (showl_lab _B4 (showl_list showl_nat)))
              ia (get_redtriple
                   ((ceq_lab _B3 (equal_list equal_nat)),
                     (ccompare_lab _B1.compare_compare_order
                       (compare_list compare_nat)),
                     (compare_order_lab _B1.compare_compare_order
                       (compare_list compare_nat)),
                     (equal_lab _B3 (equal_list equal_nat)), set_impl_lab,
                     (showl_lab _B4 (showl_list showl_nat)))
                   redp)
              graphs u_opt dpp)
            (fun x ->
              Inl (comp (comp (comp (comp i
                                      (showsl_lit
": error when applying the size-change (redpair) processor on the DP problem\092"))
                                (showsl_dpp
                                  (showl_lab _B4 (showl_list showl_nat))
                                  (showl_list showl_char) ia dpp))
                          (showsl_literal "\092"))
                    x)))
    | j, ia, assms, i, dpp, Fcc_Proc (f, fcs, pb, r, prf) ->
        debug i "Fcc_Proc"
          (bindb
            (catch_error
              (fcc_proc
                ((equal_lab _B3 (equal_list equal_nat)),
                  (showl_lab _B4 (showl_list showl_nat)))
                ((ceq_list ceq_char), (ccompare_list ccompare_char),
                  (equal_list equal_char), mapping_impl_list, set_impl_list,
                  (showl_list showl_char))
                ia f fcs pb r dpp)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit
  ": error when applying the flat context closure processor on the DP problem\092"))
                                  (showsl_dpp
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) ia dpp))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia assms
                  (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the flat context closure processor\092"))
                        x))))
    | j, ia, assms, i, dpp, Split_Proc (prem, rrem, prf1, prf2) ->
        debug i "Split_Proc"
          (match get_fcc_option prf1
            with None ->
              (let (dpp1, dpp2) =
                 split_proc (showl_lab _B4 (showl_list showl_nat))
                   (showl_list showl_char) ia dpp prem rrem
                 in
                bindb (catch_error
                        (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j
                          ia assms (add_index i one_nata) dpp1 prf1)
                        (fun x ->
                          Inl (comp (comp i
                                      (showsl_lit
": error below the split processor\092"))
                                x)))
                  (fun _ ->
                    catch_error
                      (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia
                        assms
                        (add_index i
                          (nat_of_integer (Big_int.big_int_of_int 2)))
                        dpp2 prf2)
                      (fun x ->
                        Inl (comp (comp i
                                    (showsl_lit
                                      ": error below the split processor\092"))
                              x))))
            | Some (f, (fcs, (pb, (rb, prf1a)))) ->
              debug i "Split_ProcFcc"
                (bindb
                  (fcc_split_proc
                    ((compare_order_lab _B1.compare_compare_order
                       (compare_list compare_nat)),
                      (equal_lab _B3 (equal_list equal_nat)),
                      (showl_lab _B4 (showl_list showl_nat)))
                    ((ceq_list ceq_char), (ccompare_list ccompare_char),
                      (compare_order_list (compare_order_char, equal_char)),
                      (equal_list equal_char), mapping_impl_list, set_impl_list,
                      (showl_list showl_char))
                    ia f fcs pb rb prem rrem dpp)
                  (fun (dpp1, dpp2) ->
                    bindb (catch_error
                            (check_dp_termination_proof_main
                              (_B1, _B2, _B3, _B4) j ia assms
                              (add_index (add_index i one_nata) one_nata) dpp1
                              prf1a)
                            (fun x ->
                              Inl (comp (comp i
  (showsl_lit ": error below the split and fcc processor\092"))
                                    x)))
                      (fun _ ->
                        catch_error
                          (check_dp_termination_proof_main (_B1, _B2, _B3, _B4)
                            j ia assms
                            (add_index i
                              (nat_of_integer (Big_int.big_int_of_int 2)))
                            dpp2 prf2)
                          (fun x ->
                            Inl (comp (comp i
(showsl_lit ": error below the split processor\092"))
                                  x))))))
    | j, ia, assms, i, dpp, Semlab_Proc (sli, lP, lQ, lR, prf) ->
        debug i "Semlab_Proc"
          (bindb
            (catch_error
              (semlab_fin_proc (_B1.compare_compare_order, _B3, _B4)
                ((ceq_list ceq_char), (ccompare_list ccompare_char),
                  (compare_list compare_char), (equal_list equal_char),
                  mapping_impl_list, set_impl_list, (showl_list showl_char))
                ia sli lP lQ lR dpp)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit ": error when applying the semlab processor on the DP problem\092"))
                                  (showsl_dpp
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) ia dpp))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia assms
                  (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the semlab processor\092"))
                        x))))
    | j, ia, assms, i, dpp, Switch_Innermost_Proc (joins, prf) ->
        debug i "Switch_Innermost_Proc"
          (bindb
            (catch_error
              (switch_innermost_proc
                ((equal_lab _B3 (equal_list equal_nat)),
                  (showl_lab _B4 (showl_list showl_nat)))
                ia joins dpp)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit
  ": error when applying the processor to switch to innermost on the DP problem\092"))
                                  (showsl_dpp
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) ia dpp))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia assms
                  (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the switch to innermost processor\092"))
                        x))))
    | j, ia, assms, i, dppa, Assume_Finite (dpp, ass) ->
        debug i "Finiteness Assumption or Unknown Proof"
          (if assms
            then bindb (catch_error
                         (check_dpp_subsumes (_B1, _B3, _B4)
                           ((compare_order_list (compare_order_nat, equal_nat)),
                             (equal_list equal_nat), (showl_list showl_nat))
                           ((ccompare_list ccompare_char),
                             (equal_list equal_char), mapping_impl_list,
                             (showl_list showl_char))
                           ia dpp dppa)
                         (fun x ->
                           Inl (comp (comp (comp i
     (showsl_lit ": error in finiteness assumption or unknown proof\092"))
                                       x)
                                 (showsl_literal "\092"))))
                   (fun _ ->
                     catch_error
                       (catch_error
                         (forallM_index
                           (fun asa ja ->
                             check_assma (_B1, _B2, _B4) j ia
                               (check_trs_termination_proof_main
                                 (_B1, _B2, _B3, _B4) j ia assms
                                 (add_index i (suc ja)))
                               (check_dp_termination_proof_main
                                 (_B1, _B2, _B3, _B4) j ia assms
                                 (add_index i (suc ja)))
                               (check_fptrs_termination_proof_main
                                 (_B1, _B2, _B3, _B4) j ia assms
                                 (add_index i (suc ja)))
                               (check_unknown_proof_main (_B1, _B2, _B3, _B4) j
                                 ia assms (add_index i (suc ja)))
                               asa)
                           ass)
                         (fun x -> Inl (snd x)))
                       (fun x ->
                         Inl (comp (comp i
                                     (showsl_lit
                                       ": error below unknown proof\092"))
                               x)))
            else Inl (comp i
                       (showsl_lit
                         ": the proof contains a finiteness assumption or unknown proof\092")))
    | j, ia, assms, i, dpp, Rewriting_Proc (u_opt, stb, sta, st, lr, p, prf) ->
        debug i "Rewriting_Proc"
          (bindb
            (catch_error
              (rewriting_proc
                ((compare_order_lab _B1.compare_compare_order
                   (compare_list compare_nat)),
                  (equal_lab _B3 (equal_list equal_nat)),
                  (showl_lab _B4 (showl_list showl_nat)))
                ia u_opt stb sta st lr p dpp)
              (fun x ->
                Inl (comp (comp (comp (comp
(comp (comp i
        (showsl_lit
          ": error when applying the rewriting processor to rewrite the pair \092"))
  (showsl_rule (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char)
    stb))
(showsl_lit "\092 to the pair \092"))
                                  (showsl_rule
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) st))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia assms
                  (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the rewriting processor\092"))
                        x))))
    | j, ia, assms, i, dpp, Narrowing_Proc (st, p, sts, prf) ->
        debug i "Narrowing_Proc"
          (bindb
            (catch_error
              (narrowing_proc
                ((compare_order_lab _B1.compare_compare_order
                   (compare_list compare_nat)),
                  (equal_lab _B3 (equal_list equal_nat)),
                  (showl_lab _B4 (showl_list showl_nat)))
                ia st p sts dpp)
              (fun x ->
                Inl (comp (comp (comp (comp
(comp (comp i
        (showsl_lit
          ": error when applying the narrowing processor to narrow the pair \092"))
  (showsl_rule (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char)
    st))
(showsl_lit "\092 to the pairs \092"))
                                  (showsl_trs
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) sts))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia assms
                  (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the narrowing processor\092"))
                        x))))
    | j, ia, assms, i, dpp, Instantiation_Proc (st, sts, prf) ->
        debug i "Instantiation_Proc"
          (bindb
            (catch_error
              (instantiation_proc
                ((compare_order_lab _B1.compare_compare_order
                   (compare_list compare_nat)),
                  (equal_lab _B3 (equal_list equal_nat)),
                  (showl_lab _B4 (showl_list showl_nat)))
                ia st sts dpp)
              (fun x ->
                Inl (comp (comp (comp (comp
(comp (comp i
        (showsl_lit
          ": error when applying the instantiation processor to instantiate the pair \092"))
  (showsl_rule (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char)
    st))
(showsl_lit "\092 to the pairs \092"))
                                  (showsl_trs
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) sts))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia assms
                  (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the instantiation processor\092"))
                        x))))
    | j, ia, assms, i, dpp, Forward_Instantiation_Proc (st, sts, u_opt, prf) ->
        debug i "Forward_Instantiation_Proc"
          (bindb
            (catch_error
              (forward_instantiation_proc
                ((compare_order_lab _B1.compare_compare_order
                   (compare_list compare_nat)),
                  (equal_lab _B3 (equal_list equal_nat)),
                  (showl_lab _B4 (showl_list showl_nat)))
                ia st sts u_opt dpp)
              (fun x ->
                Inl (comp (comp (comp (comp
(comp (comp i
        (showsl_lit
          ": error when applying the forward_instantiation processor to instantiate the pair \092"))
  (showsl_rule (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char)
    st))
(showsl_lit "\092 to the pairs \092"))
                                  (showsl_trs
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) sts))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia assms
                  (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the forward instantiation processor\092"))
                        x))))
    | j, ia, assms, i, dpp, Unlab_Proc (p, r, prf) ->
        debug i "Unlab_Proc"
          (Inl (showsl_lit "unlabeling processor not supported"))
    | j, ia, assms, i, dpp, General_Redpair_Proc (rp, ps, pb, prof, prfs) ->
        debug i "General_Redpair_ProcProc"
          (let n = size_list prfs in
            bindb (check (less_nat zero_nata n)
                    (showsl_lit "at least one subproof is required"))
              (fun _ ->
                (let merge = equal_nata n one_nata in
                  bindb (catch_error
                          (conditional_general_reduction_pair_proc
                            ((ceq_lab _B3 (equal_list equal_nat)),
                              (ccompare_lab _B1.compare_compare_order
                                (compare_list compare_nat)),
                              (compare_order_lab _B1.compare_compare_order
                                (compare_list compare_nat)),
                              (equal_lab _B3 (equal_list equal_nat)),
                              set_impl_lab,
                              (showl_lab _B4 (showl_list showl_nat)))
                            ia (get_non_inf_order
                                 ((compare_order_lab _B1.compare_compare_order
                                    (compare_list compare_nat)),
                                   (equal_lab _B3 (equal_list equal_nat)),
                                   (showl_lab _B4 (showl_list showl_nat)))
                                 ((equal_list equal_char),
                                   (linorder_list (equal_char, linorder_char)),
                                   (showl_list showl_char))
                                 rp)
                            ps pb prof merge dpp)
                          (fun x ->
                            Inl (comp (comp
(comp (comp i
        (showsl_lit
          ": error when applying the generic reduction pair processor to the DP problem\092"))
  (showsl_dpp (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char) ia
    dpp))
(showsl_literal "\092"))
                                  x)))
                    (fun dpps ->
                      bindb (catch_error
                              (check_dp_termination_proof_main
                                (_B1, _B2, _B3, _B4) j ia assms
                                (add_index i one_nata) (nth dpps zero_nata)
                                (nth prfs zero_nata))
                              (fun x ->
                                Inl (comp (comp i
    (showsl_lit ": error below the generic reduction pair processor\092"))
                                      x)))
                        (fun _ ->
                          (if merge then Inr ()
                            else catch_error
                                   (check_dp_termination_proof_main
                                     (_B1, _B2, _B3, _B4) j ia assms
                                     (add_index i
                                       (nat_of_integer
 (Big_int.big_int_of_int 2)))
                                     (nth dpps one_nata) (nth prfs one_nata))
                                   (fun x ->
                                     Inl (comp
   (comp i
     (showsl_lit ": error below the generic reduction pair processor\092"))
   x))))))))
    | j, ia, assms, i, dpp, Complex_Constant_Removal_Proc (p, prf) ->
        debug i "Complex_Constant_Removal_Proc"
          (bindb
            (catch_error
              (complex_constant_removal_proc
                ((compare_order_lab _B1.compare_compare_order
                   (compare_list compare_nat)),
                  (equal_lab _B3 (equal_list equal_nat)),
                  (showl_lab _B4 (showl_list showl_nat)))
                ia p dpp)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit
  ": error when applying the complex constant removal processor to the DP problem\092"))
                                  (showsl_dpp
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) ia dpp))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j ia assms
                  (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the complex constant removal processor\092"))
                        x))))
    | j, ia, assms, i, dpp, To_Trs_Proc prf ->
        debug i "To_Trs_Proc"
          (catch_error
            (check_trs_termination_proof_main (_B1, _B2, _B3, _B4) j ia assms
              (add_index i one_nata)
              (mk_tp j
                (nfsc ia dpp,
                  (qc ia dpp,
                    (pb ia dpp @ rd ia dpp, pwb ia dpp @ rwd ia dpp))))
              prf)
            (fun x ->
              Inl (comp (comp i
                          (showsl_lit ": error below the To-Trs processor\092"))
                    x)))
and check_unknown_proof_main (_B1, _B2, _B3, _B4)
  j ia assms i tpa (Assume_Unknown (tp, ass)) =
    debug i "Unknown Proof"
      (if assms
        then bindb (catch_error
                     (check (equal_lista equal_char tpa tp)
                       (comp (comp (comp (comp
   (showsl_lit "unknown problems are not identical: \092")
   (showsl_lista showl_char tpa))
                                     (showsl_literal "\092"))
                               (showsl_lit " vs \092"))
                         (showsl_lista showl_char tp)))
                     (fun x ->
                       Inl (comp (comp (comp i
 (showsl_lit ": error in termination assumption or unknown proof\092"))
                                   x)
                             (showsl_literal "\092"))))
               (fun _ ->
                 catch_error
                   (forallM_index
                     (fun asa ja ->
                       check_assma (_B1, _B2, _B4) j ia
                         (check_trs_termination_proof_main (_B1, _B2, _B3, _B4)
                           j ia assms (add_index i (suc ja)))
                         (check_dp_termination_proof_main (_B1, _B2, _B3, _B4) j
                           ia assms (add_index i (suc ja)))
                         (check_fptrs_termination_proof_main
                           (_B1, _B2, _B3, _B4) j ia assms
                           (add_index i (suc ja)))
                         (check_unknown_proof_main (_B1, _B2, _B3, _B4) j ia
                           assms (add_index i (suc ja)))
                         asa)
                     ass)
                   (fun x -> Inl (snd x)))
        else Inl (comp i
                   (showsl_lit ": the proof contains an unknown proof\092")));;

let rec map_assm_proof
  tp dpp fptp unk x4 = match tp, dpp, fptp, unk, x4 with
    tp, dpp, fptp, unk, SN_assm_proof (r, p) -> SN_assm_proof (r, tp p)
    | tp, dpp, fptp, unk, Finite_assm_proof (d, p) ->
        Finite_assm_proof (d, dpp p)
    | tp, dpp, fptp, unk, SN_FP_assm_proof (r, p) ->
        SN_FP_assm_proof (r, fptp p)
    | tp, dpp, fptp, unk, Unknown_assm_proof (u, p) ->
        Unknown_assm_proof (u, unk p)
    | tp, dpp, fptp, unk, Not_SN_assm_proof (v, va) -> Not_SN_assm_proof (v, va)
    | tp, dpp, fptp, unk, Infinite_assm_proof (v, va) ->
        Infinite_assm_proof (v, va)
    | tp, dpp, fptp, unk, Not_RelSN_assm_proof (v, va) ->
        Not_RelSN_assm_proof (v, va)
    | tp, dpp, fptp, unk, Not_SN_FP_assm_proof (v, va) ->
        Not_SN_FP_assm_proof (v, va)
    | tp, dpp, fptp, unk, Complexity_assm_proof (v, va) ->
        Complexity_assm_proof (v, va);;

let rec no_decr _A _B _C
  = filtera
      (fun (l, r) ->
        not (not (equal_terma (equal_lab _A _B) _C l r) &&
              equal_terma (equal_lab _A _B) _C (map_term unlab (fun x -> x) l)
                (map_term unlab (fun x -> x) r)));;

let rec unlab_of_trs _A _B _C
  r = map (map_funs_rule unlab) (no_decr _A _B _C r);;

let rec updatec p_to_p (p, info) = (p_to_p p, info);;

let rec unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2)
  = function
    DP_Trans (a1, a2, a3, p) ->
      DP_Trans
        (a1, a2, a3, fst (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p))
    | Rule_Removal (a1, a2, p) ->
        Rule_Removal
          (a1, a2, unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | String_Reversal p ->
        String_Reversal (unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Constant_String (a1, p) ->
        Constant_String
          (a1, unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Bounds a -> Bounds a
    | Uncurry (a1, a2, p) ->
        Uncurry (a1, a2, unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Semlab (a1, a2, a3, p) ->
        Semlab
          (a1, a2, a3, unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | R_is_Empty -> R_is_Empty
    | Fcc (a1, a2, p) ->
        Fcc (a1, a2, unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Split (a1, p, q) ->
        Split (a1, unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) p,
                unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) q)
    | Switch_Innermost (a1, p) ->
        Switch_Innermost
          (a1, unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Drop_Equality p ->
        Drop_Equality (unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Remove_Nonapplicable_Rules (a1, p) ->
        Remove_Nonapplicable_Rules
          (a1, unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Permuting_AFS (a1, p) ->
        Permuting_AFS
          (a1, unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Assume_SN (a, p) ->
        Assume_SN
          (a, map (map_assm_proof
                    (unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2))
                    (comp fst
                      (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2)))
                    (unlab_to_split_otrs (_A1, _A2) (_B1, _B2) (_C1, _C2))
                    (unlab_to_split_unknown (_A1, _A2) (_B1, _B2) (_C1, _C2)))
                p)
and unlab_to_split_unknown (_A1, _A2) (_B1, _B2) (_C1, _C2)
  (Assume_Unknown (a1, p)) =
    Assume_Unknown
      (a1, map (map_assm_proof
                 (unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2))
                 (comp fst (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2)))
                 (unlab_to_split_otrs (_A1, _A2) (_B1, _B2) (_C1, _C2))
                 (unlab_to_split_unknown (_A1, _A2) (_B1, _B2) (_C1, _C2)))
             p)
and unlab_to_split_otrs (_A1, _A2) (_B1, _B2) (_C1, _C2)
  (Assume_FP_SN (a, p)) =
    Assume_FP_SN
      (a, map (map_assm_proof
                (unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2))
                (comp fst (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2)))
                (unlab_to_split_otrs (_A1, _A2) (_B1, _B2) (_C1, _C2))
                (unlab_to_split_unknown (_A1, _A2) (_B1, _B2) (_C1, _C2)))
            p)
and unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2)
  = function P_is_Empty -> (P_is_Empty, [])
    | Dep_Graph_Proc ps ->
        (Dep_Graph_Proc
           (map (fun (po, a) ->
                  (map_option
                     (comp fst
                       (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2)))
                     po,
                    a))
             ps),
          [])
    | Subterm_Criterion_Proc (a1, a2, a3, p) ->
        updatec (fun a -> Subterm_Criterion_Proc (a1, a2, a3, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Gen_Subterm_Criterion_Proc (a1, a2, p) ->
        updatec (fun a -> Gen_Subterm_Criterion_Proc (a1, a2, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Redpair_Proc (a1, a2, p) ->
        updatec (fun a -> Redpair_Proc (a1, a2, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Redpair_UR_Proc (a1, a2, a3, p) ->
        updatec (fun a -> Redpair_UR_Proc (a1, a2, a3, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Usable_Rules_Proc (a1, p) ->
        updatec (fun a -> Usable_Rules_Proc (a1, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Q_Reduction_Proc (a1, p) ->
        updatec (fun a -> Q_Reduction_Proc (a1, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Mono_Redpair_Proc (a1, a2, a3, p) ->
        updatec (fun a -> Mono_Redpair_Proc (a1, a2, a3, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Mono_URM_Redpair_Proc (a1, a2, a3, p) ->
        updatec (fun a -> Mono_URM_Redpair_Proc (a1, a2, a3, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Mono_Redpair_UR_Proc (a1, a2, a3, a4, p) ->
        updatec (fun a -> Mono_Redpair_UR_Proc (a1, a2, a3, a4, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Size_Change_Subterm_Proc a1 -> (Size_Change_Subterm_Proc a1, [])
    | Size_Change_Redpair_Proc (a1, a2, a3) ->
        (Size_Change_Redpair_Proc (a1, a2, a3), [])
    | Uncurry_Proc (a1, a2, a3, a4, p) ->
        updatec (fun a -> Uncurry_Proc (a1, a2, a3, a4, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Fcc_Proc (a1, a2, a3, a4, p) ->
        updatec (fun a -> Fcc_Proc (a1, a2, a3, a4, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Switch_Innermost_Proc (a1, p) ->
        updatec (fun a -> Switch_Innermost_Proc (a1, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Rewriting_Proc (a1, a2, a3, a4, a5, a6, p) ->
        updatec (fun a -> Rewriting_Proc (a1, a2, a3, a4, a5, a6, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Narrowing_Proc (a1, a2, a3, p) ->
        updatec (fun a -> Narrowing_Proc (a1, a2, a3, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Instantiation_Proc (a1, a2, p) ->
        updatec (fun a -> Instantiation_Proc (a1, a2, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Forward_Instantiation_Proc (a1, a2, a3, p) ->
        updatec (fun a -> Forward_Instantiation_Proc (a1, a2, a3, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p)
    | Assume_Finite (d, p) ->
        (Assume_Finite
           (d, map (map_assm_proof
                     (unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2))
                     (comp fst
                       (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2)))
                     (unlab_to_split_otrs (_A1, _A2) (_B1, _B2) (_C1, _C2))
                     (unlab_to_split_unknown (_A1, _A2) (_B1, _B2) (_C1, _C2)))
                 p),
          [])
    | To_Trs_Proc p ->
        (To_Trs_Proc (unlab_to_split_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) p),
          [])
    | Unlab_Proc (pa, r, p) ->
        (let (pb, lista) = unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p
           in
          (P_is_Empty, (pa, (r, pb)) :: lista))
    | Split_Proc (a1, a2, p, q) ->
        (Split_Proc
           (a1, a2, fst (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p),
             fst (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) q)),
          [])
    | Semlab_Proc (a1, lP, a2, lR, p) ->
        (let (pa, lista) = unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p
           in
         let sl = Semlab_Proc (a1, lP, a2, lR, pa) in
          (match lista with [] -> (sl, lista)
            | (pb, (r, prof)) :: listb ->
              (let ulP = unlab_of_trs _A2 _B2 _C2 lP in
               let ulR = unlab_of_trs _A2 _B2 _C2 lR in
               let pr =
                 ceta_list_diff
                   (compare_order_prod
                     (compare_order_term
                       (compare_lab _A1.compare_compare_order
                         _B1.compare_compare_order)
                       _C1.compare_compare_order)
                     (compare_order_term
                       (compare_lab _A1.compare_compare_order
                         _B1.compare_compare_order)
                       _C1.compare_compare_order))
                   ulP pb
                 in
               let rr =
                 ceta_list_diff
                   (compare_order_prod
                     (compare_order_term
                       (compare_lab _A1.compare_compare_order
                         _B1.compare_compare_order)
                       _C1.compare_compare_order)
                     (compare_order_term
                       (compare_lab _A1.compare_compare_order
                         _B1.compare_compare_order)
                       _C1.compare_compare_order))
                   ulR r
                 in
                (Split_Proc (pr, rr, sl, prof), listb))))
    | General_Redpair_Proc (a1, a2, a3, a4, ps) ->
        (General_Redpair_Proc (a1, a2, a3, a4, ps), [])
    | Complex_Constant_Removal_Proc (a1, p) ->
        updatec (fun a -> Complex_Constant_Removal_Proc (a1, a))
          (unlab_to_split_dp (_A1, _A2) (_B1, _B2) (_C1, _C2) p);;

let rec check_trs_termination_proof (_B1, _B2, _B3, _B4)
  ia j a i tp prf =
    check_trs_termination_proof_main (_B1, _B2, _B3, _B4) ia j a i tp
      (unlab_to_split_trs (_B1, _B3)
        ((compare_order_list (compare_order_nat, equal_nat)),
          (equal_list equal_nat))
        ((compare_order_list (compare_order_char, equal_char)),
          (equal_list equal_char))
        prf);;

let rec rule_lab_repr_to_lab _A _B
  ps = fun_of_map
         (ceta_map_of
           (compare_order_prod
             (compare_order_term _A.compare_compare_order
               _B.compare_compare_order)
             (compare_order_term _A.compare_compare_order
               _B.compare_compare_order))
           ps)
         zero_nata;;

let rec critical_peaks_impl _A
  p r = maps (fun (l, ra) ->
               maps (fun pa ->
                      (let c = ctxt_of_pos_term pa l in
                       let la = subt_at l pa in
                       let b = equal_ctxta _A (equal_list equal_char) c Hole in
                        (if is_Var la then []
                          else maps (fun (lb, rb) ->
                                      (match
mgu_var_disjoint_generic (equal_list equal_char) _A
  (fun a -> Chara (false, false, false, true, true, true, true, false) :: a)
  (fun a -> Chara (true, false, false, true, true, true, true, false) :: a) la
  lb
with None -> []
| Some (sigma, tau) ->
  [(b, ((subst_apply_term l sigma,
          ((l, ra), ([], (sigma, (true, subst_apply_term ra sigma))))),
         (subst_apply_term l sigma,
           ((lb, rb),
             (pa, (tau, (true,
                          ctxt_apply_term (subst_apply_ctxt c sigma)
                            (subst_apply_term rb tau))))))))]))
                                 r)))
                 (poss_list l))
          p;;

let rec check_rstepsa (_A1, _A2) (_B1, _B2, _B3, _B4)
  r x1 s u = match r, x1, s, u with
    r, [], s, u ->
      check (equal_terma _A1 _B2 s u)
        (comp (comp (comp (comp (showsl_literal
                                  "the last term of the rewrite sequence\092")
                            (showsl_terma _A2 _B4 s))
                      (showsl_literal
                        "\092does not correspond to the goal term\092"))
                (showsl_terma _A2 _B4 u))
          (showsl_literal "\092"))
    | r, (p, (lr, t)) :: rs, s, u ->
        bindb (check_rstep (_A1, _A2) (_B1, _B2, _B3, _B4) r p lr s t)
          (fun _ -> check_rstepsa (_A1, _A2) (_B1, _B2, _B3, _B4) r rs t u);;

let rec showsl_eqs _A _B
  = showsl_trsa (showsl _A) (showsl _B) "equational system:" " = ";;

let rec showsl_eq _A _B = showsl_rulea (showsl _A) (showsl _B) " = ";;

let rec check_estep (_A1, _A2) (_B1, _B2, _B3, _B4)
  e p rule l_to_r s t =
    bindb (check
            (membera (equal_prod (equal_term _A1 _B2) (equal_term _A1 _B2)) e
              rule)
            (comp (comp (comp (comp (showsl_rule _A2 _B4 rule)
                                (showsl_lit "is not an equation of"))
                          (showsl_literal "\092"))
                    (showsl_eqs _A2 _B4 e))
              (showsl_literal "\092")))
      (fun _ ->
        bindb (check (in_poss p s)
                (comp (comp (comp (showsl_lista showl_nat p)
                              (showsl_lit " is not a position of "))
                        (showsl_terma _A2 _B4 s))
                  (showsl_literal "\092")))
          (fun _ ->
            bindb (check (in_poss p t)
                    (comp (comp (comp (showsl_lista showl_nat p)
                                  (showsl_lit " is not a position of "))
                            (showsl_terma _A2 _B4 t))
                      (showsl_literal "\092")))
              (fun _ ->
                (let c = ctxt_of_pos_term p s in
                 let d = ctxt_of_pos_term p t in
                 let u = subt_at s p in
                 let v = subt_at t p in
                 let rrule = (if l_to_r then rule else (snd rule, fst rule)) in
                 let err =
                   comp (comp (comp (comp (comp
    (comp (comp (comp (comp (showsl_lit "the term ") (showsl_terma _A2 _B4 t))
                  (showsl_lit
                    " does not result from a proper application of term "))
            (showsl_terma _A2 _B4 s))
      (showsl_lit " using equation "))
    (showsl_literal "\092"))
                                      (showsl_eq _A2 _B4 rrule))
                                (showsl_lit " at position "))
                          (showsl_lista showl_nat p))
                     (showsl_literal "\092")
                   in
                  (match
                    match_list (_B1, _B2, _B3) _A1 _B2 (fun a -> Var a)
                      [(fst rrule, u); (snd rrule, v)]
                    with None -> Inl err
                    | Some _ -> check (equal_ctxta _A1 _B2 c d) err)))));;

let rec check_conversion (_A1, _A2) (_B1, _B2, _B3, _B4)
  e x1 s u = match e, x1, s, u with
    e, [], s, u ->
      check (equal_terma _A1 _B2 s u)
        (comp (comp (comp (comp (comp (comp
(comp (showsl_lit "the last term of the conversion ") (showsl_literal "\092"))
(showsl_terma _A2 _B4 s))
                                  (showsl_literal "\092"))
                            (showsl_lit "does not correspond to the goal term"))
                      (showsl_literal "\092"))
                (showsl_terma _A2 _B4 u))
          (showsl_literal "\092"))
    | e, (p, (r, (l_to_r, t))) :: c, s, u ->
        bindb (check_estep (_A1, _A2) (_B1, _B2, _B3, _B4) e p r l_to_r s t)
          (fun _ -> check_conversion (_A1, _A2) (_B1, _B2, _B3, _B4) e c t u);;

let rec check_ELD_1_nat
  beta alpha sigma_1 sigma_2 sigma_3 =
    bindb (catch_error
            (forallM
              (fun x ->
                check (less_nat x beta)
                  (comp (comp (comp (comp (showsl_lit
    "the labels are not decreasing: ")
                                      (showsl_nat x))
                                (showsl_lit " is not smaller "))
                          (showsl_nat beta))
                    (showsl_literal "\092")))
              sigma_1)
            (fun x -> Inl (snd x)))
      (fun _ ->
        bindb (catch_error
                (forallM
                  (fun x ->
                    check (less_eq_nat x alpha)
                      (comp (comp (comp (comp
  (showsl_lit "the labels are not decreasing: ") (showsl_nat x))
                                    (showsl_lit " is not smaller equal "))
                              (showsl_nat alpha))
                        (showsl_literal "\092")))
                  sigma_2)
                (fun x -> Inl (snd x)))
          (fun _ ->
            bindb (check (less_eq_nat (size_list sigma_2) one_nata)
                    (showsl_lit
                      "  the length of the middle sequence is greater 1\092"))
              (fun _ ->
                catch_error
                  (forallM
                    (fun x ->
                      check (less_nat x alpha || less_nat x beta)
                        (comp (comp (comp (comp
    (comp (comp (showsl_lit "the labels are not decreasing: ") (showsl_nat x))
      (showsl_lit " is not smaller "))
    (showsl_nat alpha))
                                      (showsl_lit " or smaller "))
                                (showsl_nat beta))
                          (showsl_literal "\092")))
                    sigma_3)
                  (fun x -> Inl (snd x)))));;

let rec rule_labeling i (s, (rl, (p, (sigma, t)))) = i rl;;

let rec eseq_to_step_list _A (_B1, _B2, _B3)
  s x1 = match s, x1 with s, [] -> []
    | s, (p, (lr, (b, t))) :: steps ->
        (let u = subt_at s p in
         let v = subt_at t p in
          (if b then (let sigma =
                        (let Some tau =
                           match_list (_B1, _B2, _B3) _A _B2 (fun a -> Var a)
                             [(fst lr, u); (snd lr, v)]
                           in
                          tau)
                        in
                       (s, (lr, (p, (sigma, (b, t))))) ::
                         eseq_to_step_list _A (_B1, _B2, _B3) t steps)
            else (let sigma =
                    (let Some tau =
                       match_list (_B1, _B2, _B3) _A _B2 (fun a -> Var a)
                         [(snd lr, u); (fst lr, v)]
                       in
                      tau)
                    in
                   (t, (lr, (p, (sigma, (b, s))))) ::
                     eseq_to_step_list _A (_B1, _B2, _B3) t steps)));;

let rec eseq_to_ddconv _A (_B1, _B2, _B3)
  (s, eseq) = (s, eseq_to_step_list _A (_B1, _B2, _B3) s eseq);;

let rec cpeak_instance _A (_B1, _B2, _B3)
  sa cpa s cp =
    not (is_none
          (match_list (_B1, _B2, _B3) _A _B2 (fun _ -> sa)
            [(sa, s); (fst cpa, fst cp); (snd cpa, snd cp)]));;

let rec rseq_to_step_list _A (_B1, _B2, _B3)
  s x1 = match s, x1 with s, [] -> []
    | s, (p, (lr, t)) :: steps ->
        (let u = subt_at s p in
         let v = subt_at t p in
         let sigma =
           (let Some tau =
              match_list (_B1, _B2, _B3) _A _B2 (fun a -> Var a)
                [(fst lr, u); (snd lr, v)]
              in
             tau)
           in
          (s, (lr, (p, (sigma, (true, t))))) ::
            rseq_to_step_list _A (_B1, _B2, _B3) t steps);;

let rec rseq_to_ddseq _A (_B1, _B2, _B3)
  (s, rseq) = (s, rseq_to_step_list _A (_B1, _B2, _B3) s rseq);;

let rec get_source s = fst s;;

let rec eseq_last
  s steps = last (s :: map (fun (_, (_, (_, sa))) -> sa) steps);;

let rec check_cpeak_eldc (_A1, _A2) (_B1, _B2, _B3, _B4)
  r p s cp cl_1 sl cl_2 cr_1 sr cr_2 lab n =
    (let ((s1, (r1, (_, (_, (true, t1))))), (_, (r2, (_, (_, (true, t2)))))) = p
       in
      bindb (check (cpeak_instance _A1 (_B1, _B2, _B3) s cp s1 (t1, t2)) id)
        (fun _ ->
          (let u_1 = eseq_last (fst cp) cl_1 in
           let v_1 = eseq_last (snd cp) cr_1 in
           let u_2 = rseq_last u_1 sl in
           let v_2 = rseq_last v_1 sr in
           let u_3 = eseq_last u_2 cl_2 in
           let v_3 = eseq_last v_2 cr_2 in
            bindb (check_conversion (_A1, _A2) (_B1, _B2, _B3, _B4) r cl_1
                    (fst cp) u_1)
              (fun _ ->
                bindb (check_rstepsa (_A1, _A2) (_B1, _B2, _B3, _B4) r sl u_1
                        u_2)
                  (fun _ ->
                    bindb (check_conversion (_A1, _A2) (_B1, _B2, _B3, _B4) r
                            cl_2 u_2 u_3)
                      (fun _ ->
                        bindb (check_conversion (_A1, _A2) (_B1, _B2, _B3, _B4)
                                r cr_1 (snd cp) v_1)
                          (fun _ ->
                            bindb (check_rstepsa (_A1, _A2) (_B1, _B2, _B3, _B4)
                                    r sr v_1 v_2)
                              (fun _ ->
                                bindb (check_conversion (_A1, _A2)
(_B1, _B2, _B3, _B4) r cr_2 v_2 v_3)
                                  (fun _ ->
                                    bindb (check (equal_terma _A1 _B2 u_3 v_3)
    (comp (comp (comp (comp (showsl_lit
                              "the conversions end in different terms: ")
                        (showsl_terma _A2 _B4 u_3))
                  (showsl_lit " and "))
            (showsl_terma _A2 _B4 v_3))
      (showsl_literal "\092")))
                                      (fun _ ->
(let alpha = lab r1 in
 let beta = lab r2 in
 let cl_1a = snd (eseq_to_ddconv _A1 (_B1, _B2, _B3) (fst cp, cl_1)) in
 let sla = snd (rseq_to_ddseq _A1 (_B1, _B2, _B3) (u_1, sl)) in
 let cl_2a = snd (eseq_to_ddconv _A1 (_B1, _B2, _B3) (u_2, cl_2)) in
 let cr_1a = snd (eseq_to_ddconv _A1 (_B1, _B2, _B3) (snd cp, cr_1)) in
 let sra = snd (rseq_to_ddseq _A1 (_B1, _B2, _B3) (v_1, sr)) in
 let cr_2a = snd (eseq_to_ddconv _A1 (_B1, _B2, _B3) (v_2, cr_2)) in
 let tau_1 = map (rule_labeling lab) cl_1a in
 let tau_2 = map (rule_labeling lab) sla in
 let tau_3 = map (rule_labeling lab) cl_2a in
 let sigma_1 = map (rule_labeling lab) cr_1a in
 let sigma_2 = map (rule_labeling lab) sra in
 let sigma_3 = map (rule_labeling lab) cr_2a in
  bindb (check_ELD_1_nat alpha beta tau_1 tau_2 tau_3)
    (fun _ ->
      bindb (check_ELD_1_nat beta alpha sigma_1 sigma_2 sigma_3)
        (fun _ ->
          (match n with None -> Inr ()
            | Some na ->
              (let check_reachable =
                 (fun t ->
                   check (membera (equal_term _A1 _B2)
                           (reachable_terms _A1 (_B1, _B2, _B3) r s na) t)
                     (comp (comp (comp (comp
 (comp (comp (showsl_lit "the fan property is violated: ")
         (showsl_terma _A2 _B4 t))
   (showsl_lit " is not reachable from "))
 (showsl_terma _A2 _B4 s))
                                   (showsl_lit " in "))
                             (showsl_nat na))
                       (showsl_lit " steps\092")))
                 in
                bindb (catch_error
                        (forallM check_reachable (map get_source cl_1a))
                        (fun x -> Inl (snd x)))
                  (fun _ ->
                    bindb (catch_error
                            (forallM check_reachable (map get_source sla))
                            (fun x -> Inl (snd x)))
                      (fun _ ->
                        bindb (catch_error
                                (forallM check_reachable (map get_source cl_2a))
                                (fun x -> Inl (snd x)))
                          (fun _ ->
                            bindb (catch_error
                                    (forallM check_reachable
                                      (map get_source cr_1a))
                                    (fun x -> Inl (snd x)))
                              (fun _ ->
                                bindb (catch_error
(forallM check_reachable (map get_source sra)) (fun x -> Inl (snd x)))
                                  (fun _ ->
                                    catch_error
                                      (forallM check_reachable
(map get_source cr_2a))
                                      (fun x ->
Inl (snd x))))))))))))))))))))));;

let rec check_rule_labeling_eldc (_A1, _A2, _A3)
  r lab cs n =
    (let cps = critical_peaks_impl _A2 r r in
     let l =
       rule_lab_repr_to_lab _A1
         (compare_order_list (compare_order_char, equal_char)) lab
       in
     let convs =
       cs @ map (fun (s, (u, (cl_1, (sl, (cl_2, (v, (cr_1, (sr, cr_2)))))))) ->
                  (s, (v, (cr_1, (sr, (cr_2, (u, (cl_1, (sl, cl_2)))))))))
              cs
       in
      catch_error
        (forallM
          (fun (_, a) ->
            (let (aa, b) = a in
              (let (s1, (r1, (p1, (sigma_1, (_, t1))))) = aa in
                (fun (s2, (r2, (p2, (sigma_2, (truea, t2))))) ->
                  catch_error
                    (check (equal_terma _A2 (equal_list equal_char) t1 t2)
                      (showsl_lit " pair non-trivial "))
                    (fun _ ->
                      catch_error
                        (existsM
                          (fun (s, (u, (cl_1,
 (sl, (cl_2, (v, (cr_1, (sr, cr_2))))))))
                            -> check_cpeak_eldc (_A2, _A3)
                                 ((ccompare_list ccompare_char),
                                   (equal_list equal_char), mapping_impl_list,
                                   (showl_list showl_char))
                                 r ((s1, (r1, (p1, (sigma_1, (truea, t1))))),
                                     (s2, (r2, (p2, (sigma_2, (truea, t2))))))
                                 s (u, v) cl_1 sl cl_2 cr_1 sr cr_2 l n)
                          convs)
                        (fun x ->
                          Inl (comp (comp (comp
    (comp (comp (comp (comp (comp (comp (showsl_literal "\092")
                                    (showsl_lit "the critical peak "))
                              (showsl_terma _A3 (showl_list showl_char) t1))
                        (showsl_lit " <- "))
                  (showsl_terma _A3 (showl_list showl_char) s1))
            (showsl_lit " -> "))
      (showsl_terma _A3 (showl_list showl_char) t2))
    (showsl_lit " could not be joined decreasingly:"))
                                      (showsl_literal "\092"))
                                (showsl_sep id id x))))))
                b))
          cps)
        (fun x -> Inl (snd x)));;

let rec split_seq
  alpha beta ss =
    (let (ssa, r) = span (fun n -> less_nat n alpha) ss in
      (match r with [] -> (ssa, ([], []))
        | h :: t ->
          (if less_eq_nat h beta then (ssa, ([h], t)) else (ssa, ([], r)))));;

let rec check_cpeak_eld (_A1, _A2) (_B1, _B2, _B3, _B4)
  r p cp j1 j2 lab =
    (let ((_, (r1, (_, (_, (true, t1))))), (_, (r2, (_, (_, (true, t2)))))) = p
       in
      bindb (check (instance_rule _A1 _B2 (_B1, _B2, _B3) (t1, t2) cp) id)
        (fun _ ->
          (let u = rseq_last (fst cp) j1 in
           let v = rseq_last (snd cp) j2 in
            bindb (check_rstepsa (_A1, _A2) (_B1, _B2, _B3, _B4) r j1 (fst cp)
                    u)
              (fun _ ->
                bindb (check_rstepsa (_A1, _A2) (_B1, _B2, _B3, _B4) r j2
                        (snd cp) v)
                  (fun _ ->
                    bindb (check (equal_terma _A1 _B2 u v)
                            (comp (comp (comp
  (comp (showsl_lit "the rewrite sequences end in different terms: ")
    (showsl_terma _A2 _B4 u))
  (showsl_lit " and "))
                                    (showsl_terma _A2 _B4 v))
                              (showsl_literal "\092")))
                      (fun _ ->
                        (let alpha = lab r1 in
                         let beta = lab r2 in
                         let tau =
                           map (rule_labeling lab)
                             (snd (rseq_to_ddseq _A1 (_B1, _B2, _B3) (t1, j1)))
                           in
                         let sigma =
                           map (rule_labeling lab)
                             (snd (rseq_to_ddseq _A1 (_B1, _B2, _B3) (t2, j2)))
                           in
                         let (tau_1, (tau_2, tau_3)) = split_seq alpha beta tau
                           in
                         let (sigma_1, (sigma_2, sigma_3)) =
                           split_seq beta alpha sigma in
                          bindb (check_ELD_1_nat alpha beta tau_1 tau_2 tau_3)
                            (fun _ ->
                              check_ELD_1_nat beta alpha sigma_1 sigma_2
                                sigma_3))))))));;

let rec check_rule_labeling_eld (_A1, _A2, _A3)
  r lab js =
    (let cps = critical_peaks_impl _A2 r r in
     let l =
       rule_lab_repr_to_lab _A1
         (compare_order_list (compare_order_char, equal_char)) lab
       in
     let joins = js @ map (fun (u, (j1, (v, j2))) -> (v, (j2, (u, j1)))) js in
      catch_error
        (forallM
          (fun (_, a) ->
            (let (aa, b) = a in
              (let (s1, (r1, (p1, (sigma_1, (_, t1))))) = aa in
                (fun (s2, (r2, (p2, (sigma_2, (truea, t2))))) ->
                  catch_error
                    (check (equal_terma _A2 (equal_list equal_char) t1 t2)
                      (showsl_lit " pair non-trivial "))
                    (fun _ ->
                      catch_error
                        (existsM
                          (fun (u, (j1, (v, j2))) ->
                            check_cpeak_eld (_A2, _A3)
                              ((ccompare_list ccompare_char),
                                (equal_list equal_char), mapping_impl_list,
                                (showl_list showl_char))
                              r ((s1, (r1, (p1, (sigma_1, (truea, t1))))),
                                  (s2, (r2, (p2, (sigma_2, (truea, t2))))))
                              (u, v) j1 j2 l)
                          joins)
                        (fun x ->
                          Inl (comp (comp (comp
    (showsl_lit "\092the critical peak ")
    (showsl_rulea (showsl _A3) (showsl_lista showl_char) " <- . -> " (t1, t2)))
                                      (showsl_lit
" could not be joined decreasingly:\092"))
                                (showsl_sep id id x))))))
                b))
          cps)
        (fun x -> Inl (snd x)));;

let rec check_estepa (_A1, _A2) (_B1, _B2, _B3, _B4)
  e p rule l_to_r s t =
    bindb (check (list_ex (eq_rule_mod_vars _A1 (_B1, _B2, _B3) rule) e)
            (comp (comp (comp (showsl_rule _A2 _B4 rule)
                          (showsl_lit " is not an equation of\092"))
                    (showsl_eqs _A2 _B4 e))
              (showsl_literal "\092")))
      (fun _ ->
        bindb (check (in_poss p s)
                (comp (comp (comp (showsl_pos p)
                              (showsl_lit " is not a position of "))
                        (showsl_terma _A2 _B4 s))
                  (showsl_literal "\092")))
          (fun _ ->
            bindb (check (in_poss p t)
                    (comp (comp (comp (showsl_pos p)
                                  (showsl_lit " is not a position of "))
                            (showsl_terma _A2 _B4 t))
                      (showsl_literal "\092")))
              (fun _ ->
                (let c = ctxt_of_pos_term p s in
                 let d = ctxt_of_pos_term p t in
                 let u = subt_at s p in
                 let v = subt_at t p in
                 let rrule = (if l_to_r then rule else (snd rule, fst rule)) in
                 let err =
                   comp (comp (comp (comp (comp
    (comp (comp (comp (showsl_lit "the term ") (showsl_terma _A2 _B4 t))
            (showsl_lit " does not result from a proper application of term "))
      (showsl_terma _A2 _B4 s))
    (showsl_lit " using equation\092"))
                                      (showsl_eq _A2 _B4 rrule))
                                (showsl_lit " at position "))
                          (showsl_pos p))
                     (showsl_literal "\092")
                   in
                  (match
                    match_list (_B1, _B2, _B3) _A1 _B2 (fun a -> Var a)
                      [(fst rrule, u); (snd rrule, v)]
                    with None -> Inl err
                    | Some _ -> check (equal_ctxta _A1 _B2 c d) err)))));;

let rec check_conversiona (_A1, _A2) (_B1, _B2, _B3, _B4)
  e x1 s u = match e, x1, s, u with
    e, [], s, u ->
      check (equal_terma _A1 _B2 s u)
        (comp (comp (comp (comp (showsl_lit
                                  "the last term of the conversion\092")
                            (showsl_terma _A2 _B4 s))
                      (showsl_lit
                        "\092does not correspond to the goal term\092"))
                (showsl_terma _A2 _B4 u))
          (showsl_literal "\092"))
    | e, (p, (r, (l_to_r, t))) :: c, s, u ->
        bindb (check_estepa (_A1, _A2) (_B1, _B2, _B3, _B4) e p r l_to_r s t)
          (fun _ -> check_conversiona (_A1, _A2) (_B1, _B2, _B3, _B4) e c t u);;

let rec check_redundant_rules (_A1, _A2) (_B1, _B2, _B3, _B4)
  ra r n convs =
    (let s =
       list_diff (equal_prod (equal_term _A1 _B2) (equal_term _A1 _B2)) r ra in
     let t =
       list_diff (equal_prod (equal_term _A1 _B2) (equal_term _A1 _B2)) ra r in
      bindb (catch_error
              (forallM
                (fun (l, rb) ->
                  check (membera (equal_term _A1 _B2)
                          (reachable_terms _A1 (_B1, _B2, _B3) ra l n) rb)
                    (comp (showsl_lit "could not simulate rule ")
                      (showsl_rule _A2 _B4 (l, rb))))
                s)
              (fun x -> Inl (snd x)))
        (fun _ ->
          catch_error
            (forallM
              (fun (l, raa) ->
                catch_error
                  (existsM
                    (fun conv ->
                      check_conversiona (_A1, _A2) (_B1, _B2, _B3, _B4) r conv l
                        raa)
                    convs)
                  (fun _ ->
                    check_join_BFS_limit (_A1, _A2) (_B1, _B2, _B3, _B4) n r l
                      raa))
              t)
            (fun x -> Inl (snd x))));;

let rec sig_is_clean (_A1, _A2) _B
  siga =
    distinct (equal_prod _A1 equal_nat)
      (map (fun (f, (tys, _)) -> (f, size_list tys)) siga);;

let rec annotate_term _A _B _C
  sigF alpha x2 = match sigF, alpha, x2 with
    sigF, alpha, Var x -> Inr (Var (x, alpha))
    | sigF, alpha, Fun (f, ts) ->
        (match sigF (f, size_list ts)
          with None ->
            Inl (comp (showsl_lit
                        "persistent decomposition: no signature for symbol ")
                  (showsl _A f))
          | Some (tys, ty) ->
            (if not (eq _B alpha ty)
              then Inl (comp (comp (comp (showsl_lit
   "persistent decomposition: ")
                                     (showsl _A f))
                               (showsl_lit " has wrong type in "))
                         (showsl_terma _A _C (Fun (f, ts))))
              else bindb (mapM (fun (a, b) -> annotate_term _A _B _C sigF a b)
                           (zip tys ts))
                     (fun tsa -> Inr (Fun (f, tsa)))));;

let rec annotate_terma _A _B _C
  sigF t =
    (match root t with None -> Inl id
      | Some fn ->
        (match sigF fn
          with None ->
            Inl (comp (showsl_lit
                        "persistent decomposition: no signature for symbol ")
                  (showsl_prod _A showl_nat fn))
          | Some (_, alpha) ->
            bindb (annotate_term _A _B _C sigF alpha t)
              (fun ta -> Inr (alpha, ta))));;

let rec check_variants_rule (_A1, _A2) (_B1, _B2, _B3, _B4)
  ra r =
    check (not (is_none (match_rules _A1 (_B1, _B2, _B3) [ra] [r])) &&
            not (is_none (match_rules _A1 (_B1, _B2, _B3) [r] [ra])))
      (comp (comp (comp (showsl_rule _A2 _B4 ra) (showsl_literal " and "))
              (showsl_rule _A2 _B4 r))
        (showsl_literal " are not variants of each other\092"));;

let rec check_rule (_A1, _A2) (_B1, _B2, _B3, _B4) (_C1, _C2, _C3, _C4)
  sigF rl =
    (let (l, r) = rl in
      bindb (annotate_terma _A2 _B2 _C4 sigF l)
        (fun (alpha, la) ->
          bindb (annotate_term _A2 _B2 _C4 sigF alpha r)
            (fun ra ->
              bindb (catch_error
                      (check_variants_rule (_A1, _A2)
                        ((ccompare_prod _C1 _B1), (equal_prod _C2 _B2),
                          (mapping_impl_prod _C3 _B3), (showl_prod _C4 _B4))
                        (map_term (fun x -> x) (fun x -> (x, alpha)) l,
                          map_term (fun x -> x) (fun x -> (x, alpha)) r)
                        (la, ra))
                      (fun _ ->
                        Inl (comp (showsl_lit
                                    "persistent decomposition: inconsistent types of variables in rule ")
                              (showsl_rule _A2 _C4 (l, r)))))
                (fun _ -> Inr (la, ra)))));;

let rec mk_sigF _A
  siga (f, a) =
    map_option snd
      (find (fun (fa, (tys, _)) -> eq _A f fa && equal_nata a (size_list tys))
        siga);;

let rec check_persistence1 (_A1, _A2) (_B1, _B2, _B3, _B4) (_C1, _C2, _C3, _C4)
  (_D1, _D2, _D3, _D4)
  siga r rs =
    bindb (check (sig_is_clean (_A1, _A2) _B4 siga)
            (showsl_lit
              "persistent decomposition: duplicate function symbol in signature"))
      (fun _ ->
        bindb (check_wf_trs _A2 (_C2, _C4) r)
          (fun _ ->
            bindb (mapM (check_wf_trs _A2 (_D2, _D4)) rs)
              (fun _ ->
                (let sigF = mk_sigF _A1 siga in
                  bindb (mapM (check_rule (_A1, _A2) (_B1, _B2, _B3, _B4)
                                (_C1, _C2, _C3, _C4) sigF)
                          r)
                    (fun ra ->
                      bindb (mapM (mapM (check_rule (_A1, _A2)
  (_B1, _B2, _B3, _B4) (_D1, _D2, _D3, _D4) sigF))
                              rs)
                        (fun rsa -> Inr (ra, rsa)))))));;

let rec memo_list_rtrancl _A
  r = (let tr = rtrancl_list_impl _A r in
       let rm = map (fun a -> (a, tr [a])) (comp (remdups _A) (map fst) r) in
        (fun a -> (match map_of _A rm a with None -> [a] | Some asa -> asa)));;

let rec sigF_arcs_code _A _B
  siga = maps (fun (_, (tys, ty)) -> map (fun a -> (ty, a)) tys) siga;;

let rec needed_types_code _A (_B1, _B2)
  siga = memo_list_rtrancl _B1 (sigF_arcs_code _A _B2 siga);;

let rec maximal_types_loop _A
  nt x1 = match nt, x1 with nt, [] -> []
    | nt, beta :: tys ->
        (let mt = maximal_types_loop _A nt tys in
         let nt_beta = nt beta in
          (if list_ex (fun alpha -> membera _A (nt alpha) beta) mt then mt
            else beta ::
                   filtera (fun alpha -> not (membera _A nt_beta alpha)) mt));;

let rec maximal_types _A (_B1, _B2)
  siga nt = maximal_types_loop _B1 nt (map (comp snd snd) siga);;

let rec check_variant_in_trs (_A1, _A2) (_B1, _B2, _B3, _B4)
  ra r =
    catch_error
      (catch_error
        (existsM (check_variants_rule (_A1, _A2) (_B1, _B2, _B3, _B4) r) ra)
        (fun x -> Inl (showsl_sep id id x)))
      (fun _ ->
        Inl (comp (comp (showsl_rule _A2 _B4 r)
                    (showsl_literal " is not a variant of any rule in "))
              (showsl_trs _A2 _B4 ra)));;

let rec check_variants_trs (_A1, _A2) (_B1, _B2, _B3, _B4)
  ra r =
    catch_error
      (forallM (check_variant_in_trs (_A1, _A2) (_B1, _B2, _B3, _B4) r) ra)
      (fun x -> Inl (snd x));;

let rec check_litsim_trs (_A1, _A2) (_B1, _B2, _B3, _B4)
  ra r =
    catch_error
      (bindb (check_variants_trs (_A1, _A2) (_B1, _B2, _B3, _B4) ra r)
        (fun _ -> check_variants_trs (_A1, _A2) (_B1, _B2, _B3, _B4) r ra))
      (fun x ->
        Inl (comp (comp (comp (showsl_trs _A2 _B4 ra)
                          (showsl_lit "\092is not literally similar to "))
                    (showsl_trs _A2 _B4 r))
              x));;

let rec type_of_rule sigF r = snd (the (sigF (the (root (fst r)))));;

let rec check_persistence_cr (_A1, _A2) (_B1, _B2, _B3, _B4)
  (_C1, _C2, _C3, _C4)
  siga r rs =
    bindb (check_persistence1 (_A1, _A2) (_B1, _B2, _B3, _B4)
            (_C1, _C2, _C3, _C4) (_C1, _C2, _C3, _C4) siga r rs)
      (fun (ra, rsa) ->
        (let sigF = mk_sigF _A1 siga in
         let needed_types = needed_types_code _A2 (_B2, _B4) siga in
         let types = maximal_types _A2 (_B2, _B4) siga needed_types in
          catch_error
            (forallM
              (fun ty ->
                (let s =
                   filtera
                     (fun rb ->
                       membera _B2 (needed_types ty) (type_of_rule sigF rb))
                     ra
                   in
                  catch_error
                    (existsM
                      (check_litsim_trs (_A1, _A2)
                        ((ccompare_prod _C1 _B1), (equal_prod _C2 _B2),
                          (mapping_impl_prod _C3 _B3), (showl_prod _C4 _B4))
                        s)
                      ([] :: rsa))
                    (fun _ ->
                      Inl (comp (comp (comp
(comp (showsl_lit "persistent decomposition: missing system induced by sort ")
  (showsl _B4 ty))
(showsl_lit ":"))
                                  (showsl_literal "\092"))
                            (showsl_trs _A2 _B4
                              (map (fun (l, rb) ->
                                     (map_term (fun x -> x) snd l,
                                       map_term (fun x -> x) snd rb))
                                s))))))
              types)
            (fun x -> Inl (snd x))));;

let rec check_cr_proof (_B1, _B2, _B3, _B4)
  a ia i j r x5 = match a, ia, i, j, r, x5 with
    a, ia, i, j, r, SN_WCR (joins_i, prf) ->
      debug ia "SN_WCR"
        (let tp = mkc i false [] r [] in
          bindb (catch_error
                  (check_trs_termination_proof (_B1, _B2, _B3, _B4) i j a
                    (add_index ia one_nata) tp prf)
                  (fun x ->
                    Inl (comp (comp ia
                                (showsl_lit
                                  ": error below strong normalization + wcr\092"))
                          x)))
            (fun _ ->
              catch_error
                (check_critical_pairs
                  ((equal_lab _B3 (equal_list equal_nat)),
                    (showl_lab _B4 (showl_list showl_nat)))
                  r (critical_pairs_impl (equal_lab _B3 (equal_list equal_nat))
                      r r)
                  joins_i)
                (fun x ->
                  Inl (comp (comp (comp (comp ia
  (showsl_lit ": error when proving local confluence of "))
                                    (showsl_tp
                                      (showl_lab _B4 (showl_list showl_nat))
                                      (showl_list showl_char) i tp))
                              (showsl_literal "\092"))
                        x))))
    | a, ia, i, j, r, Weakly_Orthogonal ->
        debug ia "Weakly Orthogonal"
          (catch_error
            (check_weakly_orthogonal
              ((equal_lab _B3 (equal_list equal_nat)),
                (showl_lab _B4 (showl_list showl_nat)))
              r)
            (fun x ->
              Inl (comp (comp (comp ia
                                (showsl_lit
                                  ": error in checking weakly orthogonality of the "))
                          (showsl_trs (showl_lab _B4 (showl_list showl_nat))
                            (showl_list showl_char) r))
                    x)))
    | a, ia, i, j, r, Strongly_Closed n ->
        debug ia "Strongly Closed"
          (catch_error
            (check_strongly_closed
              ((equal_lab _B3 (equal_list equal_nat)),
                (showl_lab _B4 (showl_list showl_nat)))
              r n)
            (fun x ->
              Inl (comp (comp (comp ia
                                (showsl_lit
                                  ": error in checking strong closedness for the "))
                          (showsl_trs (showl_lab _B4 (showl_list showl_nat))
                            (showl_list showl_char) r))
                    x)))
    | a, ia, i, j, r, Rule_Labeling (rl, joins, prf) ->
        debug ia "Rule Labeling"
          (bindb
            (match prf
              with None ->
                check_linear_trs (showl_lab _B4 (showl_list showl_nat))
                  ((ceq_list ceq_char), (ccompare_list ccompare_char),
                    set_impl_list, (showl_list showl_char))
                  r
              | Some prfa ->
                bindb (check_left_linear_trs
                        (showl_lab _B4 (showl_list showl_nat))
                        ((ceq_list ceq_char), (ccompare_list ccompare_char),
                          set_impl_list, (showl_list showl_char))
                        r)
                  (fun _ ->
                    (let (rnd, rd) =
                       partition
                         (fun lr ->
                           linear_term
                             ((ceq_list ceq_char),
                               (ccompare_list ccompare_char), set_impl_list)
                             (snd lr))
                         r
                       in
                     let tp = mkc i false [] rd rnd in
                      catch_error
                        (check_trs_termination_proof (_B1, _B2, _B3, _B4) i j a
                          (add_index ia one_nata) tp prfa)
                        (fun x ->
                          Inl (comp (comp ia
                                      (showsl_lit
": error below relative termination for rule labeling\092"))
                                x)))))
            (fun _ ->
              catch_error
                (check_rule_labeling_eld
                  ((compare_order_lab _B1.compare_compare_order
                     (compare_list compare_nat)),
                    (equal_lab _B3 (equal_list equal_nat)),
                    (showl_lab _B4 (showl_list showl_nat)))
                  r rl joins)
                (fun x ->
                  Inl (comp (comp (comp ia
                                    (showsl_lit
                                      ": error in checking decreasingness of CPs using rule labeling for the "))
                              (showsl_trs (showl_lab _B4 (showl_list showl_nat))
                                (showl_list showl_char) r))
                        x))))
    | a, ia, i, j, r, Rule_Labeling_Conv (rl, convs, nprf) ->
        debug ia "Rule Labeling"
          (match nprf
            with None ->
              bindb (check_linear_trs (showl_lab _B4 (showl_list showl_nat))
                      ((ceq_list ceq_char), (ccompare_list ccompare_char),
                        set_impl_list, (showl_list showl_char))
                      r)
                (fun _ ->
                  catch_error
                    (check_rule_labeling_eldc
                      ((compare_order_lab _B1.compare_compare_order
                         (compare_list compare_nat)),
                        (equal_lab _B3 (equal_list equal_nat)),
                        (showl_lab _B4 (showl_list showl_nat)))
                      r rl convs None)
                    (fun x ->
                      Inl (comp (comp (comp ia
(showsl_lit
  ": error in checking decreasingness of CPs using rule labeling for the "))
                                  (showsl_trs
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) r))
                            x)))
            | Some (n, prf) ->
              bindb (check_left_linear_trs
                      (showl_lab _B4 (showl_list showl_nat))
                      ((ceq_list ceq_char), (ccompare_list ccompare_char),
                        set_impl_list, (showl_list showl_char))
                      r)
                (fun _ ->
                  (let (rnd, rd) =
                     partition
                       (fun lr ->
                         linear_term
                           ((ceq_list ceq_char), (ccompare_list ccompare_char),
                             set_impl_list)
                           (snd lr))
                       r
                     in
                   let tp = mkc i false [] rd rnd in
                    bindb (catch_error
                            (check_trs_termination_proof (_B1, _B2, _B3, _B4) i
                              j a (add_index ia one_nata) tp prf)
                            (fun x ->
                              Inl (comp (comp ia
  (showsl_lit ": error below relative termination for rule labeling\092"))
                                    x)))
                      (fun _ ->
                        catch_error
                          (check_rule_labeling_eldc
                            ((compare_order_lab _B1.compare_compare_order
                               (compare_list compare_nat)),
                              (equal_lab _B3 (equal_list equal_nat)),
                              (showl_lab _B4 (showl_list showl_nat)))
                            r rl convs (Some n))
                          (fun x ->
                            Inl (comp (comp
(comp ia
  (showsl_lit
    ": error in checking decreasingness of CPs using rule labeling for the "))
(showsl_trs (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char) r))
                                  x))))))
    | a, ia, i, j, r, Redundant_Rules (rs, n, convs, prf) ->
        debug ia "Redundant Rules"
          (bindb
            (catch_error
              (check_cr_proof (_B1, _B2, _B3, _B4) a (add_index ia one_nata) i j
                rs prf)
              (fun x ->
                Inl (comp (comp (comp ia
                                  (showsl_lit
                                    ": error below confluence of modified TRS\092"))
                            (showsl_trs (showl_lab _B4 (showl_list showl_nat))
                              (showl_list showl_char) rs))
                      x)))
            (fun _ ->
              catch_error
                (check_redundant_rules
                  ((equal_lab _B3 (equal_list equal_nat)),
                    (showl_lab _B4 (showl_list showl_nat)))
                  ((ccompare_list ccompare_char), (equal_list equal_char),
                    mapping_impl_list, (showl_list showl_char))
                  r rs n convs)
                (fun x ->
                  Inl (comp (comp (comp (comp
  (comp ia
    (showsl_lit ": error in checking redundant rules transformation of the "))
  (showsl_trs (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char) r))
                                    (showsl_lit "transformed to the "))
                              (showsl_trs (showl_lab _B4 (showl_list showl_nat))
                                (showl_list showl_char) rs))
                        x))))
    | a, ia, i, j, r, Parallel_Closed n ->
        debug ia "Parallel Closed"
          (catch_error
            (check_parallel_closed
              ((equal_lab _B3 (equal_list equal_nat)),
                (showl_lab _B4 (showl_list showl_nat)))
              r n)
            (fun x ->
              Inl (comp (comp (comp ia
                                (showsl_lit
                                  ": error in checking parallel closedness for the "))
                          (showsl_trs (showl_lab _B4 (showl_list showl_nat))
                            (showl_list showl_char) r))
                    x)))
    | a, ia, i, j, r, Critical_Pair_Closing_System (c, prf, n) ->
        debug ia "Critical-Pair-Closing System"
          (let tp = mkc i false [] c [] in
            bindb (catch_error
                    (check_trs_termination_proof (_B1, _B2, _B3, _B4) i j a
                      (add_index ia one_nata) tp prf)
                    (fun x ->
                      Inl (comp (comp ia
                                  (showsl_lit
                                    ": error below strong normalization of CPCS\092"))
                            x)))
              (fun _ ->
                catch_error
                  (check_critical_pair_closing
                    ((equal_lab _B3 (equal_list equal_nat)),
                      (showl_lab _B4 (showl_list showl_nat)))
                    r c n)
                  (fun x ->
                    Inl (comp (comp (comp (comp ia
    (showsl_lit ": error when closing critical pairs of "))
                                      (showsl_tp
(showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char) i tp))
                                (showsl_literal "\092"))
                          x))))
    | a, ia, i, j, r, Persistent_Decomposition (siga, ps) ->
        debug ia "Persistent Decomposition"
          (let checks =
             map (map_prod id
                   (fun prf iaa p ->
                     check_cr_proof (_B1, _B2, _B3, _B4) a iaa i j p prf))
               ps
             in
            bindb (catch_error
                    (forallM
                      (fun (n, (prf, f)) ->
                        catch_error (f (add_index ia n) prf)
                          (fun x ->
                            Inl (comp (comp ia
(showsl_lit ": error while checking confluence of subproblems\092"))
                                  x)))
                      (enumerate one_nata checks))
                    (fun x -> Inl (snd x)))
              (fun _ ->
                catch_error
                  (check_persistence_cr
                    ((equal_lab _B3 (equal_list equal_nat)),
                      (showl_lab _B4 (showl_list showl_nat)))
                    ((ccompare_list ccompare_char), (equal_list equal_char),
                      mapping_impl_list, (showl_list showl_char))
                    ((ccompare_list ccompare_char), (equal_list equal_char),
                      mapping_impl_list, (showl_list showl_char))
                    siga r (map fst ps))
                  (fun x ->
                    Inl (comp (comp (comp ia
                                      (showsl_lit
": error in checking persistent decomposition of "))
                                (showsl_trs
                                  (showl_lab _B4 (showl_list showl_nat))
                                  (showl_list showl_char) r))
                          x))));;

let rec edges_to_adj_fun (_A1, _A2)
  e = precompute_fun _A1
        (fun a ->
          remdups _A2
            (maps (fun ea -> (if eq _A2 (fst ea) a then [snd ea] else [])) e))
        (remdups _A2 (map fst e @ map snd e));;

let rec create_graph_impl (_A1, _A2)
  e = Gen_g_impl_ext
        (membera _A2 (map fst e @ map snd e), edges_to_adj_fun (_A1, _A2) e,
          remdups _A2 (map fst e @ map snd e), ());;

let rec as_singleton _B x = (Array [x], onea _B);;

let rec gi_V0 (Gen_g_impl_ext (gi_V, gi_E, gi_V0, more)) = gi_V0;;

let rec as_is_empty s = equal_nata (snd s) zero_nata;;

let rec rev_append x0 ac = match x0, ac with [], ac -> ac
                     | x :: xs, ac -> rev_append xs (x :: ac);;

let rec glist_delete_aux
  eq x xa2 asa = match eq, x, xa2, asa with eq, x, [], asa -> asa
    | eq, x, y :: ys, asa ->
        (if eq x y then rev_append asa ys
          else glist_delete_aux eq x ys (y :: asa));;

let rec glist_delete eq x l = glist_delete_aux eq x l [];;

let rec select_edge_tr (_A1, _A2)
  s = (let (a, (aa, (ab, bb))) = s in
        (if as_is_empty bb then (None, (a, (aa, (ab, bb))))
          else (let (ac, bc) = as_top bb in
                 (if less_eq_nat
                       (as_get aa (minus_nata (as_length aa) one_nata)) ac
                   then (let xa = gen_pick (fun x -> foldli (id x)) bc in
                         let xb = glist_delete (eq _A1) xa bc in
                         let xc =
                           (if is_Nil xb then as_pop bb
                             else as_set bb (minus_nata (as_length bb) one_nata)
                                    (ac, xb))
                           in
                          (Some xa, (a, (aa, (ab, xc)))))
                   else (None, (a, (aa, (ab, bb))))))));;

let rec as_empty _B uu = (Array [], zerob _B);;

let rec stat_start x = (fun _ -> ()) x;;

let rec last_seg_tr _A
  s = (let (a, (aa, (_, _))) = s in
       let (_, bc) =
         whilea
           (fun (xe, _) ->
             less_nat xe
               (if equal_nata
                     (plus_nata (minus_nata (as_length aa) one_nata) one_nata)
                     (as_length aa)
                 then as_length a
                 else as_get aa
                        (plus_nata (minus_nata (as_length aa) one_nata)
                          one_nata)))
           (fun (ac, bc) -> (let xa = as_get a ac in (suc ac, xa :: bc)))
           (as_get aa (minus_nata (as_length aa) one_nata), [])
         in
        bc);;

let rec collapse_tr _A
  v s = (let (a, (aa, (ab, bb))) = s in
         let x = idx_of_tr _A v (a, (aa, (ab, bb))) in
         let xa = as_take (plus_nata x one_nata) aa in
          (a, (xa, (ab, bb))));;

let rec stat_stop x = (fun _ -> ()) x;;

let rec compute_SCC_tr (_A1, _A2)
  g = (let _ = stat_start () in
       let xa = ([], Empty) in
       let a =
         foldli (id (gi_V0 g)) (fun _ -> true)
           (fun xb (a, b) ->
             (if not (match rbt_lookup _A1 b xb with None -> false
                       | Some i ->
                         (if less_eq_int zero_inta i then false else true))
               then (let xc =
                       (a, (as_singleton one_nat xb,
                             (as_singleton one_nat zero_nata,
                               (rbt_insert _A1 xb (int_of_nat zero_nata) b,
                                 (if is_Nil (gi_E g xb)
                                   then as_empty zero_nat ()
                                   else as_singleton one_nat
  (zero_nata, gi_E g xb))))))
                       in
                     let (aa, (_, (_, (ad, _)))) =
                       whilea
                         (fun (_, xf) ->
                           not (as_is_empty (let (xg, (_, (_, _))) = xf in xg)))
                         (fun (aa, ba) ->
                           (match
                             select_edge_tr (_A2, _A1.linorder_compare_order) ba
                             with (None, bb) ->
                               (let xf =
                                  last_seg_tr _A1.linorder_compare_order bb in
                                let xg = pop_tr _A1 bb in
                                let xh = xf :: aa in
                                 (xh, xg))
                             | (Some xf, bb) ->
                               (if (match
                                     rbt_lookup _A1
                                       (let (_, (_, (xl, _))) = bb in xl) xf
                                     with None -> false
                                     | Some i ->
                                       (if less_eq_int zero_inta i then true
 else false))
                                 then (let ab = collapse_tr _A1 xf bb in
(aa, ab))
                                 else (if not
    (match rbt_lookup _A1 (let (_, (_, (xl, _))) = bb in xl) xf
      with None -> false
      | Some i -> (if less_eq_int zero_inta i then false else true))
then (aa, push_code _A1 g xf bb) else (aa, bb)))))
                         xc
                       in
                      (aa, ad))
               else (a, b)))
           xa
         in
       let (aa, _) = a in
       let _ = stat_stop () in
        aa);;

let rec scc_decomp (_A1, _A2, _A3, _A4, _A5)
  e = (let ee =
         set ((ceq_prod _A1 _A1), (ccompare_prod _A2 _A2),
               (set_impl_prod _A5 _A5))
           e
         in
        filtera
          (fun a ->
            (match a with [] -> true
              | [v] ->
                member ((ceq_prod _A1 _A1), (ccompare_prod _A2 _A2)) (v, v) ee
              | _ :: _ :: _ -> true))
          (compute_SCC_tr (_A3, _A4) (create_graph_impl (_A3, _A4) e)));;

let rec check_acyclic (_A1, _A2, _A3, _A4, _A5, _A6)
  r = catch_error
        (catch_error
          (forallM
            (fun scc ->
              Inl (comp (comp (showsl_lit "SCC ") (showsl_list _A6 scc))
                    (showsl_lit " detected ")))
            (scc_decomp (_A1, _A2, _A3, _A4, _A5) r))
          (fun x -> Inl (snd x)))
        (fun x ->
          Inl (comp (comp (comp (showsl_lit "\092graph ")
                            (showsl_list_prod _A6 _A6 r))
                      (showsl_lit " not acyclic\092"))
                x));;

let rec get_conjunctions
  = function Conjunction phis -> maps get_conjunctions phis
    | Atom v -> [Atom v]
    | NegAtom v -> [NegAtom v]
    | Disjunction v -> [Disjunction v];;

let rec get_disjunctions
  = function Disjunction phis -> maps get_disjunctions phis
    | Atom v -> [Atom v]
    | NegAtom v -> [NegAtom v]
    | Conjunction v -> [Conjunction v];;

let rec pos_gctxt (_A1, _A2) _B
  = function
    GCHole ->
      inserta ((ceq_list ceq_nat), (ccompare_list ccompare_nat)) []
        (set_empty ((ceq_list ceq_nat), (ccompare_list ccompare_nat))
          (of_phantom set_impl_lista))
    | GCFun (f, cs) ->
        sup_seta ((ceq_list ceq_nat), (ccompare_list ccompare_nat))
          (inserta ((ceq_list ceq_nat), (ccompare_list ccompare_nat)) []
            (set_empty ((ceq_list ceq_nat), (ccompare_list ccompare_nat))
              (of_phantom set_impl_lista)))
          (sup_setb
            (finite_UNIV_list, cenum_list, (ceq_list ceq_nat),
              (cproper_interval_list ccompare_nat), set_impl_list)
            (image
              ((ceq_prod (ceq_gctxt _A2) ceq_nat),
                (ccompare_prod (ccompare_gctxt _A1 _B) ccompare_nat))
              ((ceq_set
                 (cenum_list, (ceq_list ceq_nat),
                   (cproper_interval_list
                     ccompare_nat).ccompare_cproper_interval)),
                (ccompare_set
                  (finite_UNIV_list, (ceq_list ceq_nat),
                    (cproper_interval_list ccompare_nat), set_impl_list)),
                set_impl_set)
              (fun (c, i) ->
                image ((ceq_list ceq_nat), (ccompare_list ccompare_nat))
                  ((ceq_list ceq_nat), (ccompare_list ccompare_nat),
                    set_impl_list)
                  (fun a -> i :: a) (pos_gctxt (_A1, _A2) _B c))
              (set ((ceq_prod (ceq_gctxt _A2) ceq_nat),
                     (ccompare_prod (ccompare_gctxt _A1 _B) ccompare_nat),
                     (set_impl_prod set_impl_gctxt set_impl_nat))
                (zip cs (upt zero_nata (size_list cs))))));;

let rec inline _B
  i rho =
    (let cs = snd rho in
     let (s, t) = nth cs i in
     let sigma = subst _B (the_Var t) s in
     let csa = take i cs @ drop (suc i) cs in
      ((fst (fst rho), subst_apply_term (snd (fst rho)) sigma),
        map (fun (sa, a) -> (subst_apply_term sa sigma, a)) csa));;

let rec invariant (Art_Node (x1, x2, x3, x4)) = x2;;

let rec location (Art_Node (x1, x2, x3, x4)) = x3;;

let rec initial_nodesa
  (Art_impl_ext (initial_nodes, nodes, more)) = initial_nodes;;

let rec name (Art_Node (x1, x2, x3, x4)) = x1;;

let rec edgea (Art_Node (x1, x2, x3, x4)) = x4;;

let rec nodesb (Art_impl_ext (initial_nodes, nodes, more)) = nodes;;

let rec art_edge_of (_E1, _E2, _E3, _E4)
  pi x1 = match pi, x1 with pi, Cover_Edge (an, uu) -> Cover an
    | pi, Children_Edge ans ->
        Children
          (map (fun (t, a) ->
                 (let (aa, _) = a in
                   (transition_of (_E1, _E2, _E3, _E4) pi t, aa)))
            ans);;

let rec art_nodes ai = map name (nodesb ai);;

let rec art_of (_E1, _E2, _E3, _E4) (_F1, _F2, _F3, _F4)
  pi ai =
    (let ans = nodesb ai in
      Art_ext
        (initial_nodesa ai, art_nodes ai,
          map_of_total (_F1, _F2, _F3)
            (fun a ->
              comp (showsl_lit "error in looking up art edge ") (showsl _F4 a))
            (map (fun a ->
                   (name a, art_edge_of (_E1, _E2, _E3, _E4) pi (edgea a)))
              ans),
          map_of_total (_F1, _F2, _F3)
            (fun a ->
              comp (showsl_lit "error in looking up node location ")
                (showsl _F4 a))
            (map (fun a -> (name a, location a)) ans),
          map_of_total (_F1, _F2, _F3)
            (fun a ->
              comp (showsl_lit "error in looking up node invariant ")
                (showsl _F4 a))
            (map (fun a -> (name a, invariant a)) ans),
          ()));;

let rec check_dp_loop (_B1, _B2, _B3) (_C1, _C2, _C3, _C4, _C5, _C6, _C7)
  i dpp (DP_loop_prf (s, prseq, sigma, c)) =
    (let p = pairsb i dpp in
     let r = rulese i dpp in
     let nfs = nfsc i dpp in
     let q = qc i dpp in
      (if equal_ctxta _B2 _C4 c Hole
        then bindb (check (not (null prseq))
                     (showsl_lit "rewrite sequence must be non-empty"))
               (fun _ ->
                 (if null q
                   then check_qsteps (_B2, _B3) (_C2, _C4, _C5, _C7)
                          (fun _ -> true) nfs p r prseq s
                          (subst_apply_term s
                            (mk_subst _C4 (fun a -> Var a) sigma))
                   else check_qsteps_subst (_B2, _B3) (_C2, _C4, _C5, _C7)
                          (check_NF_iteration (_C1, _C2, _C3, _C4, _C6)
                            (_B1, _B2) sigma q)
                          nfs p r prseq s
                          (subst_apply_term s
                            (mk_subst _C4 (fun a -> Var a) sigma))))
        else check_loop (_B1, _B2, _B3) (_C1, _C2, _C3, _C4, _C5, _C6, _C7) q
               nfs s (map (fun (x, (y, (_, z))) -> (x, (y, z))) prseq) sigma c
               r));;

let rec check_rel_seq (_A1, _A2) (_B1, _B2, _B3, _B4)
  r sa x2 s u b = match r, sa, x2, s, u, b with
    r, sa, [], s, u, true ->
      check (equal_terma _A1 _B2 s u)
        (comp (comp (comp (comp (showsl_lit
                                  "the last term of the rewrite sequence\092")
                            (showsl_terma _A2 _B4 s))
                      (showsl_lit
                        "\092does not correspond to the goal term\092"))
                (showsl_terma _A2 _B4 u))
          (showsl_literal "\092"))
    | r, sa, [], s, u, false ->
        Inl (showsl_lit "did not find strict step in rewrite sequence")
    | ra, sa, (p, (r, (true, t))) :: prts, s, u, b ->
        bindb (check_qrstep (_A1, _A2) (_B1, _B2, _B3, _B4) (fun _ -> true)
                false ra p r s t)
          (fun _ ->
            check_rel_seq (_A1, _A2) (_B1, _B2, _B3, _B4) ra sa prts t u true)
    | ra, sa, (p, (r, (false, t))) :: prts, s, u, b ->
        bindb (check_qrstep (_A1, _A2) (_B1, _B2, _B3, _B4) (fun _ -> true)
                false sa p r s t)
          (fun _ ->
            check_rel_seq (_A1, _A2) (_B1, _B2, _B3, _B4) ra sa prts t u b);;

let rec pat_dv_impl _A _B
  p = (let (_, (sigma, mu)) = p in
        remdups _B
          (map fst (mk_subst_domain _B _A sigma @ mk_subst_domain _B _A mu)));;

let rec set_option (_A1, _A2, _A3)
  = function None -> bot_set (_A1, _A2, _A3)
    | Some x2 -> inserta (_A1, _A2) x2 (bot_set (_A1, _A2, _A3));;

let rec equal_complexity_class (Comp_Poly x) (Comp_Poly ya) = equal_nata x ya;;

let default_nfs_nt_trs : bool = false;;

let rec string_reversal_complete_rel_tt _B _C
  i tp =
    bindb (check (q_emptyb i tp) (showsl_lit "Q is not empty"))
      (fun _ ->
        bindb (check_unary_signature _B _C (rulesd i tp))
          (fun _ ->
            Inr (mkc i default_nfs_nt_trs [] (map rev_rule (rc i tp))
                  (map rev_rule (rwc i tp)))));;

let rec check_to_srs_complete (_A1, _A2) (_B1, _B2)
  v old_new r s =
    (let mu = funas_trs_list r in
     let (d, (da, nu)) = extract_components _B1 mu old_new in
      bindb (check_components (_B1, _B2) mu (d, (da, nu)))
        (fun _ ->
          bindb (check_varcond_subset _B2 (_A1, _A2) s)
            (fun _ ->
              catch_error
                (forallM
                  (fun slr ->
                    (let y = choose_var v (fst slr) in
                     let stra = str d y in
                     let to_slr = (fun (l, ra) -> (stra l, stra ra)) in
                      check (list_ex
                              (fun lr ->
                                equal_proda (equal_term _B1 _A1)
                                  (equal_term _B1 _A1) (to_slr lr) slr)
                              r)
                        (comp (showsl_lit "could not find original rule for ")
                          (showsl_rule _B2 _A2 slr))))
                  s)
                (fun x -> Inl (snd x)))));;

let rec const_to_string_complete_tt (_B1, _B2) (_C1, _C2)
  i tp (Const_string_complete_proof (v, old_new, s)) =
    bindb (check (q_emptyb i tp) (showsl_lit "Q is not empty"))
      (fun _ ->
        bindb (check_to_srs_complete (_C1, _C2) (_B1, _B2) v old_new
                (rulesd i tp) s)
          (fun _ -> Inr (mkc i false [] s [])));;

let rec check_NF_trs_subset (_A1, _A2) (_B1, _B2, _B3)
  r = check_NF_terms_subset (is_NF_trs (_A1, _A2) (_B1, _B2, _B3) r);;

let rec switch_termination_proc (_B1, _B2, _B3)
  i joins_i dpp =
    (let p = pairsb i dpp in
     let r = rulese i dpp in
     let q = qc i dpp in
      (match
        (let cp = critical_pairs_impl _B2 r r in
          bindb (catch_error
                  (forallM
                    (fun (b, _) -> check b (showsl_lit "rules are not overlay"))
                    cp)
                  (fun x -> Inl (snd x)))
            (fun _ ->
              bindb (check_wf_trs _B3
                      ((equal_list equal_char), (showl_list showl_char)) r)
                (fun _ ->
                  bindb (check_critical_pairs (_B2, _B3) r cp joins_i)
                    (fun _ ->
                      bindb (check (null (critical_pairs_impl _B2 p r))
                              (showsl_lit "there are overlaps between P and R"))
                        (fun _ ->
                          catch_error
                            (check_NF_trs_subset (_B1, _B2)
                              ((ccompare_list ccompare_char),
                                (equal_list equal_char), mapping_impl_list)
                              r q)
                            (fun x ->
                              Inl (comp (showsl_terma _B3
  (showl_list showl_char) x)
                                    (showsl_lit
                                      " is not in normal form w.r.t. R"))))))))
        with Inl a -> Inl a
        | Inr _ -> Inr (mkd i (nfsc i dpp) false p [] [] [] r)));;

let rec dp_q_reduction_nonterm (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  (_C1, _C2, _C3, _C4)
  i dpp (DP_q_reduction_nonterm_prf q) =
    (let p = pairsb i dpp in
     let r = rulese i dpp in
     let qa = qc i dpp in
     let nfs = nfsc i dpp in
     let f =
       set ((ceq_prod _B2 ceq_nat), (ccompare_prod _B3 ccompare_nat),
             (set_impl_prod _B6 set_impl_nat))
         (funas_trs_list (p @ r))
       in
     let rQ =
       filtera
         (fun qb ->
           less_eq_set
             ((cenum_prod _B1 cenum_nat), (ceq_prod _B2 ceq_nat),
               (ccompare_prod _B3 ccompare_nat))
             (funas_term (_B2, _B3, _B6) qb) f)
         qa
       in
      bindb (catch_error
              (check_NF_terms_subset (is_NF_terms (_B4, _B5) (_C1, _C2, _C3) q)
                rQ)
              (fun x ->
                Inl (comp (comp (showsl_lit "the term ")
                            (showsl_terma _B7 _C4 x))
                      (showsl_lit " is missing in Q\039 "))))
        (fun _ -> Inr (mkd i nfs false p [] q [] r)));;

let rec check_instance (_A1, _A2) (_B1, _B2, _B3, _B4)
  pa p =
    catch_error
      (forallM
        (fun st ->
          check (list_ex (instance_rule _A1 _B2 (_B1, _B2, _B3) st) pa)
            (comp (showsl_rule _A2 _B4 st)
              (showsl_lit " is not an instance of any original pair")))
        p)
      (fun x -> Inl (snd x));;

let rec instantiation_complete_proc (_B1, _B2) (_C1, _C2, _C3, _C4)
  i dpp (Instantiation_complete_proc_prf p) =
    (let pa = pairsb i dpp in
     let q = qc i dpp in
     let r = rulese i dpp in
     let nfs = nfsc i dpp in
      bindb (check (not nfs || null q)
              (showsl_lit
                "normal form subst. currently not supported for innermost"))
        (fun _ ->
          bindb (check_instance (_B1, _B2) (_C1, _C2, _C3, _C4) pa p)
            (fun _ -> Inr (mkd i nfs false p [] q [] r))));;

let rec rule_removal_nonterm_dp (_B1, _B2) (_C1, _C2)
  i dpp (Rule_removal_nonterm_dp_prf (p, r)) =
    (let prm =
       (if is_none p then []
         else ceta_list_diff
                (compare_order_prod
                  (compare_order_term _B1.compare_compare_order
                    _C1.compare_compare_order)
                  (compare_order_term _B1.compare_compare_order
                    _C1.compare_compare_order))
                (pairsb i dpp) (the p))
       in
     let rrm =
       (if is_none r then []
         else ceta_list_diff
                (compare_order_prod
                  (compare_order_term _B1.compare_compare_order
                    _C1.compare_compare_order)
                  (compare_order_term _B1.compare_compare_order
                    _C1.compare_compare_order))
                (rulese i dpp) (the r))
       in
      Inr (delete_R_Rwc i (delete_P_Pwa i dpp prm prm) rrm rrm));;

let rec mk_rel_tp i (nfs, (q, (r, rw))) = mkc i nfs q r rw;;

let rec mk_dppa
  i (nfs, (m, (p, (pw, (q, (r, rw)))))) = mkd i nfs m p pw q r rw;;

let rec mk_tpa i (nfs, (q, r)) = mkc i nfs q r [];;

let rec check_assmb (_B1, _B2) (_C1, _C2)
  i j tp_check dp_check rtp_check fptp_check unk_check x7 = match
    i, j, tp_check, dp_check, rtp_check, fptp_check, unk_check, x7 with
    i, j, tp_check, dp_check, rtp_check, fptp_check, unk_check,
      Not_SN_assm_proof (t, prf)
      -> tp_check (mk_tpa i t) prf
    | i, j, tp_check, dp_check, rtp_check, fptp_check, unk_check,
        Not_RelSN_assm_proof (t, prf)
        -> rtp_check (mk_rel_tp i t) prf
    | i, j, tp_check, dp_check, rtp_check, fptp_check, unk_check,
        Infinite_assm_proof (t, prf)
        -> dp_check (mk_dppa j t) prf
    | i, j, tp_check, dp_check, rtp_check, fptp_check, unk_check,
        Not_SN_FP_assm_proof (t, prf)
        -> fptp_check t prf
    | i, j, tp_check, dp_check, rtp_check, fptp_check, unk_check,
        Unknown_assm_proof (t, prf)
        -> unk_check t prf
    | i, j, uu, uv, uw, ux, uy, SN_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for termination assumptions in non-termination proof")
    | i, j, uu, uv, uw, ux, uy, Finite_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for termination assumptions in non-termination proof")
    | i, j, uu, uv, uw, ux, uy, SN_FP_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for termination assumptions in non-termination proof")
    | i, j, uu, uv, uw, ux, uy, Complexity_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for termination assumptions in non-termination proof");;

let rec q_increase_nonterm_dp _B _C
  i dpp (Q_increase_nonterm_dp_prf q) =
    (let p = pairsb i dpp in
     let r = rulese i dpp in
     let qa = qc i dpp in
     let nfs = nfsc i dpp in
      Inr (mkd i nfs false p [] (list_union (equal_term _B _C) qa q) [] r));;

let rec check_dpp_subsumesa (_B1, _B2, _B3) (_C1, _C2, _C3)
  j dp dpp =
    (let (nfsa, (_, (p, (pw, (q, (r, rw)))))) = dp in
     let pairsa = p @ pw in
     let rulesa = r @ rw in
     let nfs = nfsc j dpp in
     let pairs = pairsb j dpp in
     let rules = rulese j dpp in
     let qa = qc j dpp in
      catch_error
        (bindb
          (check (if not (null q) then (if nfs then nfsa else true) else true)
            (showsl_lit "incompatible substitutions-in-normal-form flags"))
          (fun _ ->
            bindb (catch_error
                    (check_subseteq
                      (equal_prod
                        (equal_term (equal_lab _B2 _C2) (equal_list equal_char))
                        (equal_term (equal_lab _B2 _C2)
                          (equal_list equal_char)))
                      pairsa pairs)
                    (fun x ->
                      Inl (toomuch "pair"
                            (showsl_rule (showl_lab _B3 _C3)
                              (showl_list showl_char) x))))
              (fun _ ->
                bindb (catch_error
                        (check_subseteq
                          (equal_prod
                            (equal_term (equal_lab _B2 _C2)
                              (equal_list equal_char))
                            (equal_term (equal_lab _B2 _C2)
                              (equal_list equal_char)))
                          rulesa rules)
                        (fun x ->
                          Inl (toomuch "rule"
                                (showsl_rule (showl_lab _B3 _C3)
                                  (showl_list showl_char) x))))
                  (fun _ ->
                    catch_error
                      (check_NF_terms_subset
                        (is_NF_terms
                          ((compare_order_lab _B1.compare_compare_order
                             _C1.compare_compare_order),
                            (equal_lab _B2 _C2))
                          ((ccompare_list ccompare_char),
                            (equal_list equal_char), mapping_impl_list)
                          q)
                        qa)
                      (fun x ->
                        Inl (comp (showsl_lit "NF(Q) differs due to term ")
                              (showsl_terma (showl_lab _B3 _C3)
                                (showl_list showl_char) x)))))))
        (fun x ->
          Inl (comp (showsl_lit
                      "problem is showing subsumption for non-termination\092")
                x)));;

let rec enfc_cand _A
  isQnf r q x3 = match isQnf, r, q, x3 with isQnf, r, q, (uu, Var uv) -> []
    | isQnf, r, q, (s, Fun (f, ts)) ->
        map (fun a -> (s, a)) ts @
          maps (fun (l, ra) ->
                 (if (match
                       mgu_class _A (Fun (f, map (icap_impl _A isQnf r s) ts)) l
                       with None -> false
                       | Some mu ->
                         list_all
                           (fun u ->
                             isQnf (subst_apply_term
                                     (map_term (fun x -> x)
                                       (fun a ->
 Chara (true, false, false, true, true, true, true, false) :: a)
                                       u)
                                     mu))
                           (args l) &&
                           list_all
                             (fun u ->
                               isQnf (subst_apply_term
                                       (map_term (fun x -> x)
 (fun a -> Chara (false, false, false, true, true, true, true, false) :: a) u)
                                       mu))
                             s)
                   then [(args l, ra)] else []))
            r;;

let rec enfc_impl _A
  isQnf isRnf r q s t =
    list_all (fun (a, b) -> enfc_q _A isQnf isRnf r q a b)
      (mk_rtrancl_list
        (equal_proda (equal_list (equal_term _A (equal_list equal_char)))
          (equal_term _A (equal_list equal_char)))
        (enfc_cand _A isQnf r q) [(s, t)]);;

let rec check_nfc (_A1, _A2, _A3)
  inn r q isQnf ss nfs t =
    bindb (check_wf_trs _A3 ((equal_list equal_char), (showl_list showl_char))
            r)
      (fun _ ->
        (if inn then Inr ()
          else catch_error
                 (forallM
                   (fun ta ->
                     check (enfc_impl _A2 isQnf
                             (is_NF_trs (_A1, _A2)
                               ((ccompare_list ccompare_char),
                                 (equal_list equal_char), mapping_impl_list)
                               r)
                             r q ss ta)
                       (comp (showsl_lit " nfc not satisfied for ")
                         (showsl_terma _A3 (showl_list showl_char) ta)))
                   (supteq_list t))
                 (fun x -> Inl (snd x))));;

let rec rewriting_complete_proc (_B1, _B2, _B3)
  i (Rewriting_complete_proc_prf (u_opt, stb, sta, st, lr, p)) dpp =
    (match
      catch_error
        (let s = fst stb in
         let t = snd stb in
         let ta = snd sta in
         let r = rulese i dpp in
          bindb (check_rstep (_B2, _B3)
                  ((ccompare_list ccompare_char), (equal_list equal_char),
                    mapping_impl_list, (showl_list showl_char))
                  r p lr t ta)
            (fun _ ->
              bindb (check (nFQ_subset_NF_rulesc i dpp)
                      (showsl_lit "innermost rewriting required"))
                (fun _ ->
                  bindb (check_rewrite_common_preconditions (_B1, _B2, _B3) i
                          u_opt stb (args s) (args (subt_at t p)) ta lr p false
                          dpp)
                    (fun _ ->
                      bindb (check (not (is_Var s))
                              (showsl_lit "lhs of pair must not be variable"))
                        (fun _ ->
                          bindb (check
                                  (eq_rule_mod_vars _B2
                                    ((ccompare_list ccompare_char),
                                      (equal_list equal_char),
                                      mapping_impl_list)
                                    sta st)
                                  (comp (comp
  (comp (showsl_lit "the rule ") (showsl_rule _B3 (showl_list showl_char) sta))
  (showsl_lit " is not a renamed variant of "))
                                    (showsl_rule _B3 (showl_list showl_char)
                                      st)))
                            (fun _ ->
                              bindb (check
                                      (equal_terma _B2 (equal_list equal_char) s
(fst sta))
                                      (showsl_lit
"left-hand sides of old and new pair differ"))
                                (fun _ ->
                                  (let q = qc i dpp in
                                   let inn =
                                     isOK (check_NF_trs_subset (_B1, _B2)
    ((ccompare_list ccompare_char), (equal_list equal_char), mapping_impl_list)
    r q)
                                     in
                                    bindb (check_nfc (_B1, _B2, _B3) inn r q
    (is_QNFc i dpp) (args s) (nfsc i dpp) (subt_at t p))
                                      (fun _ ->
bindb (catch_error
        (forallM
          (fun (l, _) ->
            check (not (is_Var l)) (showsl_lit "lhss must not be variables"))
          r)
        (fun x -> Inl (snd x)))
  (fun _ ->
    (if not (is_Var t)
      then check (not (not (null (rules_mapc i dpp (the (root t))))))
             (comp (comp (showsl_lit "root of ")
                     (showsl_terma _B3 (showl_list showl_char) t))
               (showsl_lit " must not be defined"))
      else Inr ())))))))))))
        (fun x ->
          Inl (comp (comp (comp (comp (comp
(showsl_lit "error when rewriting the pair\092")
(showsl_rule _B3 (showl_list showl_char) stb))
                                  (showsl_lit "\092 to the pair\092"))
                            (showsl_rule _B3 (showl_list showl_char) st))
                      (showsl_literal "\092"))
                x))
      with Inl a -> Inl a | Inr _ -> Inr (replace_paira i dpp stb [st]));;

let rec rstep_enum_impl _A (_B1, _B2, _B3)
  r ta t p =
    concat
      (map_filter
        (fun x ->
          (if less_eq_list (equal_nat, order_nat) p x
            then Some (let tp = subt_at ta x in
                        (if membera (equal_list equal_nat) (poss_list t) x
                          then (if equal_ctxta _A _B2 (ctxt_of_pos_term x ta)
                                     (ctxt_of_pos_term x t)
                                 then maps
(fun tpa ->
  maps (fun (l, ra) ->
         map (fun mu -> ((l, ra), (mu, x)))
           (option_to_list
             (match_list (_B1, _B2, _B3) _A _B2 (fun a -> Var a)
               [(l, tp); (ra, tpa)])))
    r)
[subt_at t x]
                                 else [])
                          else []))
            else None))
        (poss_list ta));;

let rec narrow_enum_impl _A (_B1, _B2, _B3)
  r sta st p =
    (let (s, t) = sta in
     let (sa, ta) = st in
      maps (fun mu ->
             map (fun (lr, (tau, q)) -> (mu, (lr, (q, tau))))
               (rstep_enum_impl _A (_B1, _B2, _B3) r (subst_apply_term t mu) ta
                 p))
        (option_to_list (matcha _A (_B1, _B2, _B3) sa s)));;

let rec check_narrow (_B1, _B2, _B3)
  inn i dpp r q nfs sta st quad =
    (let ((_, t), ((s, ta), (mu, (lr, (p, _))))) = (sta, (st, quad)) in
      bindb (check_nfc (_B1, _B2, _B3) inn r q (is_QNFc i dpp) (args s) nfs
              (subt_at (subst_apply_term t mu) p))
        (fun _ ->
          check_rewrite_common_preconditions (_B1, _B2, _B3) i None
            (s, subst_apply_term t mu) (args s)
            (args (subt_at (subst_apply_term t mu) p)) ta lr p false dpp));;

let rec narrowing_complete_proc (_B1, _B2, _B3)
  i (Narrowing_complete_proc_prf (st, p, sts)) dpp =
    (match
      catch_error
        (let (s, t) = st in
         let q = qc i dpp in
         let nfs = nfsc i dpp in
         let rules = rulese i dpp in
         let check_ndef =
           check_no_defined_root _B3 (showl_list showl_char)
             (fun fn -> not (null (rules_mapc i dpp fn)))
           in
         let inn =
           isOK (check_NF_trs_subset (_B1, _B2)
                  ((ccompare_list ccompare_char), (equal_list equal_char),
                    mapping_impl_list)
                  rules q)
           in
         let cnarrow = check_narrow (_B1, _B2, _B3) inn i dpp rules q nfs st in
          bindb (check (null q || nFQ_subset_NF_rulesc i dpp)
                  (showsl_lit "full or innermost rewriting required"))
            (fun _ ->
              bindb (if null q then Inr ()
                      else bindb (check_no_var _B3 (showl_list showl_char) s)
                             (fun _ ->
                               bindb (check_no_var _B3 (showl_list showl_char)
                                       t)
                                 (fun _ ->
                                   bindb (check_ndef t)
                                     (fun _ ->
                                       (if nfs
 then bindb (catch_error
              (forallM
                (fun (l, _) -> check_no_var _B3 (showl_list showl_char) l)
                rules)
              (fun x -> Inl (snd x)))
        (fun _ ->
          check (wf_rule (equal_list equal_char) (s, t))
            (comp (showsl_rule _B3 (showl_list showl_char) (s, t))
              (showsl_lit " is not well formed")))
 else Inr ())))))
                (fun _ ->
                  catch_error
                    (forallM
                      (fun sta ->
                        (let quads =
                           narrow_enum_impl _B2
                             ((ccompare_list ccompare_char),
                               (equal_list equal_char), mapping_impl_list)
                             rules st sta p
                           in
                          catch_error
                            (existsM
                              (fun quad ->
                                catch_error
                                  (check (null q) (showsl_lit "q not empty"))
                                  (fun _ -> cnarrow sta quad))
                              quads)
                            (fun x ->
                              Inl (comp (comp
  (showsl_rule _B3 (showl_list showl_char) sta)
  (showsl_lit
    (match quads with [] -> " does not seem to be narrowed pair"
      | _ :: _ -> " violates side conditions for completeness")))
                                    (showsl_list_gen id "" "" "" "" x)))))
                      sts)
                    (fun x -> Inl (snd x)))))
        (fun x ->
          Inl (comp (comp (comp (comp (comp
(showsl_lit "error when narrowing\092")
(showsl_rule _B3 (showl_list showl_char) st))
                                  (showsl_lit "\092 to the pairs\092"))
                            (showsl_trs _B3 (showl_list showl_char) sts))
                      (showsl_literal "\092"))
                x))
      with Inl a -> Inl a | Inr _ -> Inr (replace_paira i dpp st sts));;

let rec is_inverse_renaming_impl _A _B
  sigma =
    (let a = mk_subst_domain _A _B sigma in
      map (fun (x, y) -> (the_Var y, Var x)) a);;

let rec vars_subst_impl _A _B
  sigma =
    (let sigmaa = mk_subst_domain _A _B sigma in
      map fst sigmaa @ maps (comp vars_term_list snd) sigmaa);;

let rec vars_pat_term_impl _A _B
  p = (let (s, (sigma, mu)) = p in
        remdups _B
          (vars_term_list s @
            vars_subst_impl _B _A sigma @ vars_subst_impl _B _A mu));;

let rec is_renaming_impl _A _B
  sigma =
    (let sigmaa = map snd (mk_subst_domain _A _B sigma) in
      list_all is_Var sigmaa && distinct (equal_term _B _A) sigmaa);;

let rec pat_dom_renaming_impl _A _B
  p rho =
    (let rhoa = mk_subst_domain _B _A rho in
     let xs = map (fun a -> Var a) (vars_pat_term_impl _A _B p) in
      is_renaming_impl _B _A rho &&
        (list_all (membera _B (pat_dv_impl _A _B p)) (map fst rhoa) &&
          list_all (fun t -> not (membera (equal_term _A _B) xs t))
            (map snd rhoa)));;

let rec check_pat_eqv_prf (_A1, _A2, _A3) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  x0 x1 = match x0, x1 with
    Pat_Irrelevant (sigmaa, mua), (t, (sigma, mu)) ->
      (let w = w_impl _B5 (mk_subst_domain _B5 _A2 sigma) t in
       let siga = mk_subst _B5 (fun a -> Var a) sigma in
       let sigaa = mk_subst _B5 (fun a -> Var a) sigmaa in
       let mub = mk_subst _B5 (fun a -> Var a) mu in
       let muba = mk_subst _B5 (fun a -> Var a) mua in
        bindb (catch_error
                (catch_error
                  (forallM
                    (fun x ->
                      bindb (check (equal_terma _A2 _B5 (siga x) (sigaa x))
                              (x, (siga x,
                                    (sigaa x,
                                      [Chara
 (false, false, false, false, true, true, true, false);
Chara (true, false, true, false, true, true, true, false);
Chara (true, false, true, true, false, true, true, false);
Chara (false, false, false, false, true, true, true, false);
Chara (true, false, false, true, false, true, true, false);
Chara (false, true, true, true, false, true, true, false);
Chara (true, true, true, false, false, true, true, false)]))))
                        (fun _ ->
                          check (equal_terma _A2 _B5 (mub x) (muba x))
                            (x, (mub x,
                                  (muba x,
                                    [Chara (true, true, false, false, false,
     true, true, false);
                                      Chara
(false, false, true, true, false, true, true, false);
                                      Chara
(true, true, true, true, false, true, true, false);
                                      Chara
(true, true, false, false, true, true, true, false);
                                      Chara
(true, false, false, true, false, true, true, false);
                                      Chara
(false, true, true, true, false, true, true, false);
                                      Chara
(true, true, true, false, false, true, true, false)])))))
                    w)
                  (fun x -> Inl (snd x)))
                (fun x ->
                  Inl (let (xa, (tb, (ta, sub))) = x in
                        comp (comp (comp (comp
   (comp (comp (comp (showsl_lit
                       "error in equivalence (irrelevant): for variable ")
                 (showsl _B7 xa))
           (showsl_lit " obtain different values for "))
     (showsl_lista showl_char sub))
   (showsl_lit " substitution: "))
                                     (showsl_terma _A3 _B7 tb))
                               (showsl_lit " != "))
                          (showsl_terma _A3 _B7 ta))))
          (fun _ -> Inr (t, (sigmaa, mua))))
    | Pat_Simplify (mu_1, mu_2), (t, (sigma, mu)) ->
        bindb (check
                (subst_eq (_B1, _B2, _B3, _B4, _B5, _B6) (_A1, _A2) mu
                  (subst_compose_impl _B5 _A2 mu_1 mu_2))
                (showsl_lit "mu != mu1 mu2"))
          (fun _ ->
            bindb (check
                    (commutes_impl (_B1, _B2, _B3, _B4, _B5, _B6) (_A1, _A2)
                      mu_1 sigma)
                    (showsl_lit "sigma and mu1 do not commute"))
              (fun _ ->
                Inr (subst_apply_term t (mk_subst _B5 (fun a -> Var a) mu_1),
                      (sigma, mu_2))))
    | Pat_Dom_Renaming rho, (t, (sigma, mu)) ->
        bindb (check (pat_dom_renaming_impl _A2 _B5 (t, (sigma, mu)) rho)
                (showsl_lit "rho is not a domain renaming for p"))
          (fun _ ->
            (let i_rho = is_inverse_renaming_impl _B5 _A2 rho in
             let sigmaa =
               mk_subst_case _B5 _A2
                 (map (comp the_Var (mk_subst _B5 (fun a -> Var a) rho))
                   (map fst (mk_subst_domain _B5 _A2 sigma)))
                 (fun x ->
                   subst_apply_term
                     (subst_apply_term
                       (subst_apply_term (Var x)
                         (mk_subst _B5 (fun a -> Var a) i_rho))
                       (mk_subst _B5 (fun a -> Var a) sigma))
                     (mk_subst _B5 (fun a -> Var a) rho))
                 []
               in
             let mua =
               mk_subst_case _B5 _A2
                 (map (comp the_Var (mk_subst _B5 (fun a -> Var a) rho))
                   (map fst (mk_subst_domain _B5 _A2 mu)))
                 (fun x ->
                   subst_apply_term
                     (subst_apply_term (Var x)
                       (mk_subst _B5 (fun a -> Var a) i_rho))
                     (mk_subst _B5 (fun a -> Var a) mu))
                 i_rho
               in
              Inr (subst_apply_term t (mk_subst _B5 (fun a -> Var a) rho),
                    (sigmaa, mua))));;

let rec subst_compose_impla _A _B
  sigma rho =
    map (fun (x, s) -> (x, subst_apply_term s rho))
      (mk_subst_domain _A _B sigma);;

let rec subst_replace_impl _A
  sigma x t = (x, t) :: filtera (fun (y, _) -> not (eq _A y x)) sigma;;

let rec showsl_pat_term (_A1, _A2) (_B1, _B2)
  p = (let (s, (sigma, tau)) = p in
        showsl_prod (showl_term _A2 _B2)
          (showl_prod (showl_list (showl_prod _B2 (showl_term _A2 _B2)))
            (showl_list (showl_prod _B2 (showl_term _A2 _B2))))
          (s, (mk_subst_domain _B1 _A1 sigma, mk_subst_domain _B1 _A1 tau)));;

let rec showsl_pat_rule (_A1, _A2) (_B1, _B2)
  pr = (let (p1, (p2, _)) = pr in
         comp (comp (showsl_pat_term (_A1, _A2) (_B1, _B2) p1)
                (showsl_lit " --> "))
           (showsl_pat_term (_A1, _A2) (_B1, _B2) p2));;

let rec subst_power_impl _A _B
  sigma n =
    (if equal_nata n zero_nata then []
      else subst_compose_impl _A _B sigma
             (subst_power_impl _A _B sigma (minus_nata n one_nata)));;

let rec check_pat_rule_prf (_A1, _A2, _A3)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  ra p x2 = match ra, p, x2 with
    ra, p, Pat_OrigRule ((l, r), isPair) ->
      (if isPair
        then bindb (check
                     (membera
                       (equal_prod (equal_term _A2 _B5) (equal_term _A2 _B5)) p
                       (l, r))
                     (comp (showsl_rule _A3 _B8 (l, r))
                       (showsl_lit " is not a pair")))
               (fun _ -> Inr ((l, ([], [])), ((r, ([], [])), isPair)))
        else bindb (check
                     (membera
                       (equal_prod (equal_term _A2 _B5) (equal_term _A2 _B5)) ra
                       (l, r))
                     (comp (showsl_rule _A3 _B8 (l, r))
                       (showsl_lit " is not a rule")))
               (fun _ -> Inr ((l, ([], [])), ((r, ([], [])), isPair))))
    | r, p, Pat_InitPump (pat, sigma, theta) ->
        bindb (check_pat_rule_prf (_A1, _A2, _A3)
                (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) r p pat)
          (fun (a, b) ->
            (let (s, (siga, tau)) = a in
              (fun (aa, ba) ->
                (let (t, (sigaa, taua)) = aa in
                  (fun bb ->
                    catch_error
                      (bindb
                        (check (null (siga @ tau @ sigaa @ taua))
                          (showsl_lit "substitutions must be empty"))
                        (fun _ ->
                          bindb (check
                                  (equal_terma _A2 _B5
                                    (subst_apply_term s
                                      (mk_subst _B5 (fun ab -> Var ab) theta))
                                    (subst_apply_term t
                                      (mk_subst _B5 (fun ab -> Var ab) sigma)))
                                  (showsl_lit "s theta != t sigma"))
                            (fun _ ->
                              bindb (check
                                      (commutes_impl
(_B1, _B2, _B3, _B4, _B5, _B7) (_A1, _A2) theta sigma)
                                      (showsl_lit
"sigma and theta do not commute"))
                                (fun _ ->
                                  Inr ((s, (sigma, [])),
((t, (theta, [])), bb))))))
                      (fun x ->
                        Inl (comp (comp (showsl_lit
  "problem with initial pumping after deriving correct pattern rule\092")
                                    (showsl_pat_rule (_A2, _A3) (_B5, _B8)
                                      ((s, (siga, tau)),
((t, (sigaa, taua)), bb))))
                              x))))
                  ba))
              b)
    | r, pa, Pat_InitPumpCtxt (pat, sigma, p, z) ->
        bindb (check_pat_rule_prf (_A1, _A2, _A3)
                (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) r pa pat)
          (fun (a, b) ->
            (let (s, (siga, tau)) = a in
              (fun (aa, ba) ->
                (let (t, (sigaa, taua)) = aa in
                  (fun bb ->
                    catch_error
                      (bindb
                        (check (not bb)
                          (showsl_lit "pairs not allowed in init pump ctxt"))
                        (fun _ ->
                          bindb (check (null (siga @ tau @ sigaa @ taua))
                                  (showsl_lit "substitutions must be empty"))
                            (fun _ ->
                              bindb (check (in_poss p t)
                                      (showsl_lit "p is not a valid position"))
                                (fun _ ->
                                  bindb (check
  (equal_terma _A2 _B5 s
    (subst_apply_term (subt_at t p) (mk_subst _B5 (fun ab -> Var ab) sigma)))
  (showsl_lit "s != t |_ p sigma"))
                                    (fun _ ->
                                      bindb
(check
  (not (membera _B5
         (vars_term_list s @ vars_term_list t @ vars_subst_impl _B5 _A2 sigma)
         z))
  (showsl_lit "z is not fresh"))
(fun _ ->
  (let tz = ctxt_apply_term (ctxt_of_pos_term p t) (Var z) in
    Inr ((s, (sigma, [])),
          ((tz, ((z, tz) :: sigma, [(z, subt_at t p)])), bb)))))))))
                      (fun x ->
                        Inl (comp (comp (showsl_lit
  "problem with initial pumping (with ctxt) after deriving correct pattern rule\092")
                                    (showsl_pat_rule (_A2, _A3) (_B5, _B8)
                                      ((s, (siga, tau)),
((t, (sigaa, taua)), bb))))
                              x))))
                  ba))
              b)
    | r, p, Pat_Equiv (pat, left, eqv) ->
        bindb (check_pat_rule_prf (_A1, _A2, _A3)
                (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) r p pat)
          (fun (pleft, (pright, b)) ->
            catch_error
              (bindb
                (check_pat_eqv_prf (_A1, _A2, _A3)
                  (_B1, _B2, _B3, _B4, _B5, _B7, _B8) eqv
                  (if left then pleft else pright))
                (fun pnew ->
                  Inr (if left then (pnew, (pright, b))
                        else (pleft, (pnew, b)))))
              (fun x ->
                Inl (comp (comp (showsl_lit
                                  "problem with pattern equivalence after deriving correct pattern rule\092")
                            (showsl_pat_rule (_A2, _A3) (_B5, _B8)
                              (pleft, (pright, b))))
                      x)))
    | r, pa, Pat_Narrow (pat1, pat2, p) ->
        bindb (check_pat_rule_prf (_A1, _A2, _A3)
                (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) r pa pat1)
          (fun (a, b) ->
            (let (s, (sigma, mu)) = a in
              (fun (aa, ba) ->
                (let (t, (siga, mua)) = aa in
                  (fun b1 ->
                    bindb (check_pat_rule_prf (_A1, _A2, _A3)
                            (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) r pa pat2)
                      (fun (ab, bb) ->
                        (let (u, (sig1, mu1)) = ab in
                          (fun (ac, bc) ->
                            (let (v, (sig2, mu2)) = ac in
                              (fun b2 ->
                                catch_error
                                  (bindb
                                    (check
                                      (subst_eq (_B1, _B2, _B3, _B4, _B5, _B7)
 (_A1, _A2) siga sigma &&
(subst_eq (_B1, _B2, _B3, _B4, _B5, _B7) (_A1, _A2) sig1 sigma &&
  (subst_eq (_B1, _B2, _B3, _B4, _B5, _B7) (_A1, _A2) sig2 sigma &&
    (subst_eq (_B1, _B2, _B3, _B4, _B5, _B7) (_A1, _A2) mua mu &&
      (subst_eq (_B1, _B2, _B3, _B4, _B5, _B7) (_A1, _A2) mu1 mu &&
        subst_eq (_B1, _B2, _B3, _B4, _B5, _B7) (_A1, _A2) mu2 mu)))))
                                      (showsl_lit
"substitutions are not identical"))
                                    (fun _ ->
                                      bindb
(check (in_poss p t) (showsl_lit "p is not a valid position"))
(fun _ ->
  bindb (check (equal_terma _A2 _B5 (subt_at t p) u) (showsl_lit "t |_ p != u"))
    (fun _ ->
      bindb (check (if b2 then null p else true)
              (showsl_lit "there is a P step, so p must be epsilon"))
        (fun _ ->
          Inr ((s, (sigma, mu)),
                ((ctxt_apply_term (ctxt_of_pos_term p t) v, (sigma, mu)),
                  b1 || b2)))))))
                                  (fun x ->
                                    Inl (comp
  (comp (comp (comp (showsl_lit
                      "problem with pattern narrowing after deriving correct pattern rules\092")
                (showsl_pat_rule (_A2, _A3) (_B5, _B8)
                  ((s, (sigma, mu)), ((t, (siga, mua)), b1))))
          (showsl_lit "\092and\092"))
    (showsl_pat_rule (_A2, _A3) (_B5, _B8)
      ((u, (sig1, mu1)), ((v, (sig2, mu2)), b2))))
  x))))
                              bc))
                          bb)))
                  ba))
              b)
    | r, p, Pat_Inst (pat, rho, Pat_Base) ->
        bindb (check_pat_rule_prf (_A1, _A2, _A3)
                (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) r p pat)
          (fun (a, b) ->
            (let (s, (sigma_s, mu_s)) = a in
              (fun (aa, ba) ->
                (let (t, (sigma_t, mu_t)) = aa in
                  (fun bb ->
                    catch_error
                      (let xs =
                         map fst
                           (mk_subst_domain _B5 _A2 sigma_s @
                             mk_subst_domain _B5 _A2 mu_s @
                               mk_subst_domain _B5 _A2 sigma_t @
                                 mk_subst_domain _B5 _A2 mu_t)
                         in
                        bindb (catch_error
                                (forallM
                                  (fun x ->
                                    check (not (membera _B5 xs x))
                                      (showsl_lit "domains not disjoint"))
                                  (vars_subst_impl _B5 _A2 rho))
                                (fun x -> Inl (snd x)))
                          (fun _ ->
                            (let rhoa = mk_subst _B5 (fun ab -> Var ab) rho in
                              Inr ((subst_apply_term s rhoa,
                                     (subst_compose_impla _B5 _A2 sigma_s rhoa,
                                       subst_compose_impla _B5 _A2 mu_s rhoa)),
                                    ((subst_apply_term t rhoa,
                                       (subst_compose_impla _B5 _A2 sigma_t
  rhoa,
 subst_compose_impla _B5 _A2 mu_t rhoa)),
                                      bb)))))
                      (fun x ->
                        Inl (comp (comp (showsl_lit
  "problem with pattern instantiation (base) after deriving correct pattern rule\092")
                                    (showsl_pat_rule (_A2, _A3) (_B5, _B8)
                                      ((s, (sigma_s, mu_s)),
((t, (sigma_t, mu_t)), bb))))
                              x))))
                  ba))
              b)
    | r, p, Pat_Inst (pat, rho, Pat_Pump) ->
        bindb (check_pat_rule_prf (_A1, _A2, _A3)
                (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) r p pat)
          (fun (a, b) ->
            (let (s, (sigma_s, mu_s)) = a in
              (fun (aa, ba) ->
                (let (t, (sigma_t, mu_t)) = aa in
                  (fun bb ->
                    catch_error
                      (bindb
                        (check
                          (commutes_impl (_B1, _B2, _B3, _B4, _B5, _B7)
                            (_A1, _A2) rho sigma_s)
                          (showsl_lit "rho does not commute with sigma_s"))
                        (fun _ ->
                          bindb (check
                                  (commutes_impl (_B1, _B2, _B3, _B4, _B5, _B7)
                                    (_A1, _A2) rho mu_s)
                                  (showsl_lit "rho does not commute with mu_s"))
                            (fun _ ->
                              bindb (check
                                      (commutes_impl
(_B1, _B2, _B3, _B4, _B5, _B7) (_A1, _A2) rho sigma_t)
                                      (showsl_lit
"rho does not commute with sigma_t"))
                                (fun _ ->
                                  bindb (check
  (commutes_impl (_B1, _B2, _B3, _B4, _B5, _B7) (_A1, _A2) rho mu_t)
  (showsl_lit "rho does not commute with mu_t"))
                                    (fun _ ->
                                      Inr
((s, (subst_compose_impl _B5 _A2 sigma_s rho, mu_s)),
  ((t, (subst_compose_impl _B5 _A2 sigma_t rho, mu_t)), bb)))))))
                      (fun x ->
                        Inl (comp (comp (showsl_lit
  "problem with pattern instantiation (pumping) after deriving correct pattern rule\092")
                                    (showsl_pat_rule (_A2, _A3) (_B5, _B8)
                                      ((s, (sigma_s, mu_s)),
((t, (sigma_t, mu_t)), bb))))
                              x))))
                  ba))
              b)
    | r, p, Pat_Inst (pat, rho, Pat_Close) ->
        bindb (check_pat_rule_prf (_A1, _A2, _A3)
                (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) r p pat)
          (fun (a, b) ->
            (let (s, (sigma_s, mu_s)) = a in
              (fun (aa, ba) ->
                (let (t, (sigma_t, mu_t)) = aa in
                  (fun bb ->
                    Inr ((s, (sigma_s, subst_compose_impl _B5 _A2 mu_s rho)),
                          ((t, (sigma_t, subst_compose_impl _B5 _A2 mu_t rho)),
                            bb))))
                  ba))
              b)
    | r, p, Pat_Rewr (pat, rewr, Pat_Base, uu) ->
        bindb (check_pat_rule_prf (_A1, _A2, _A3)
                (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) r p pat)
          (fun (a, b) ->
            (let (s, (sigma_s, mu_s)) = a in
              (fun (aa, ba) ->
                (let (t, (sigma_t, mu_t)) = aa in
                  (fun bb ->
                    catch_error
                      (let (ta, rseq) = rewr in
                       let tb = last (ta :: map (fun (_, (_, tb)) -> tb) rseq)
                         in
                        bindb (check (equal_terma _A2 _B5 t ta)
                                (showsl_lit "terms t do not match"))
                          (fun _ ->
                            bindb (check_rsteps (_A2, _A3) (_B3, _B5, _B6, _B8)
                                    r rseq ta tb)
                              (fun _ ->
                                Inr ((s, (sigma_s, mu_s)),
                                      ((tb, (sigma_t, mu_t)), bb)))))
                      (fun x ->
                        Inl (comp (comp (showsl_lit
  "problem with pattern rewriting (base) after deriving correct pattern rule\092")
                                    (showsl_pat_rule (_A2, _A3) (_B5, _B8)
                                      ((s, (sigma_s, mu_s)),
((t, (sigma_t, mu_t)), bb))))
                              x))))
                  ba))
              b)
    | r, p, Pat_Rewr (pat, rewr, Pat_Pump, x) ->
        bindb (check_pat_rule_prf (_A1, _A2, _A3)
                (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) r p pat)
          (fun (a, b) ->
            (let (s, (sigma_s, mu_s)) = a in
              (fun (aa, ba) ->
                (let (t, (sigma_t, mu_t)) = aa in
                  (fun bb ->
                    catch_error
                      (let (ta, rseq) = rewr in
                       let tb = last (ta :: map (fun (_, (_, tb)) -> tb) rseq)
                         in
                        bindb (check
                                (equal_terma _A2 _B5
                                  (mk_subst _B5 (fun ab -> Var ab) sigma_t x)
                                  ta)
                                (showsl_lit
                                  "sigma_t x does not match starting term"))
                          (fun _ ->
                            bindb (check_rsteps (_A2, _A3) (_B3, _B5, _B6, _B8)
                                    r rseq ta tb)
                              (fun _ ->
                                Inr ((s, (sigma_s, mu_s)),
                                      ((t,
 (subst_replace_impl _B5 sigma_t x tb, mu_t)),
bb)))))
                      (fun xa ->
                        Inl (comp (comp (showsl_lit
  "problem with pattern rewriting (pumping) after deriving correct pattern rule\092")
                                    (showsl_pat_rule (_A2, _A3) (_B5, _B8)
                                      ((s, (sigma_s, mu_s)),
((t, (sigma_t, mu_t)), bb))))
                              xa))))
                  ba))
              b)
    | r, p, Pat_Rewr (pat, rewr, Pat_Close, x) ->
        bindb (check_pat_rule_prf (_A1, _A2, _A3)
                (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) r p pat)
          (fun (a, b) ->
            (let (s, (sigma_s, mu_s)) = a in
              (fun (aa, ba) ->
                (let (t, (sigma_t, mu_t)) = aa in
                  (fun bb ->
                    catch_error
                      (let (ta, rseq) = rewr in
                       let tb = last (ta :: map (fun (_, (_, tb)) -> tb) rseq)
                         in
                        bindb (check
                                (equal_terma _A2 _B5
                                  (mk_subst _B5 (fun ab -> Var ab) mu_t x) ta)
                                (showsl_lit
                                  "sigma_t x does not match starting term"))
                          (fun _ ->
                            bindb (check_rsteps (_A2, _A3) (_B3, _B5, _B6, _B8)
                                    r rseq ta tb)
                              (fun _ ->
                                Inr ((s, (sigma_s, mu_s)),
                                      ((t,
 (sigma_t, subst_replace_impl _B5 mu_t x tb)),
bb)))))
                      (fun xa ->
                        Inl (comp (comp (showsl_lit
  "problem with pattern rewriting (closing) after deriving correct pattern rule\092")
                                    (showsl_pat_rule (_A2, _A3) (_B5, _B8)
                                      ((s, (sigma_s, mu_s)),
((t, (sigma_t, mu_t)), bb))))
                              xa))))
                  ba))
              b)
    | r, p, Pat_Exp_Sigma (pat, k) ->
        bindb (check_pat_rule_prf (_A1, _A2, _A3)
                (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) r p pat)
          (fun (a, b) ->
            (let (s, (sigma_s, mu_s)) = a in
              (fun (aa, ba) ->
                (let (t, (sigma_t, mu_t)) = aa in
                  (fun bb ->
                    Inr ((subst_apply_term s
                            (mk_subst _B5 (fun ab -> Var ab)
                              (subst_power_impl _B5 _A2 sigma_s k)),
                           (sigma_s, mu_s)),
                          ((subst_apply_term t
                              (mk_subst _B5 (fun ab -> Var ab)
                                (subst_power_impl _B5 _A2 sigma_t k)),
                             (sigma_t, mu_t)),
                            bb))))
                  ba))
              b);;

let rec check_non_loop_prf (_A1, _A2, _A3)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  r pa (Non_loop_prf (pat, sigma, mu, m, b, p)) =
    bindb (check_pat_rule_prf (_A1, _A2, _A3)
            (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) r pa pat)
      (fun (a, c) ->
        (let (s, (sigmaa, mua)) = a in
          (fun (aa, ca) ->
            (let (t, (sigma_t, mu_t)) = aa in
              (fun is_pair ->
                catch_error
                  (bindb
                    (check (if is_pair then null p else true)
                      (showsl_lit "p must be empty, since pairs are contained"))
                    (fun _ ->
                      bindb (check
                              (commutes_impl (_B1, _B2, _B3, _B4, _B5, _B7)
                                (_A1, _A2) sigmaa sigma)
                              (showsl_lit "sigma and sigma\039 do not commute"))
                        (fun _ ->
                          bindb (check
                                  (commutes_impl (_B1, _B2, _B3, _B4, _B5, _B7)
                                    (_A1, _A2) mua sigma)
                                  (showsl_lit
                                    "mu and sigma\039 do not commute"))
                            (fun _ ->
                              bindb (check
                                      (subst_eq (_B1, _B2, _B3, _B4, _B5, _B7)
(_A1, _A2) sigma_t
(subst_compose_impl _B5 _A2 (subst_power_impl _B5 _A2 sigmaa m) sigma))
                                      (showsl_lit
"sigma_t != sigma^m sigma\039 "))
                                (fun _ ->
                                  bindb (check
  (subst_eq (_B1, _B2, _B3, _B4, _B5, _B7) (_A1, _A2) mu_t
    (subst_compose_impl _B5 _A2 mua mu))
  (showsl_lit "mu_t != mu mu\039 "))
                                    (fun _ ->
                                      bindb
(check (in_poss p t) (showsl_lit "p is not a position in t"))
(fun _ ->
  check (equal_terma _A2 _B5
          (subst_apply_term s
            (mk_subst _B5 (fun ab -> Var ab)
              (subst_power_impl _B5 _A2 sigmaa b)))
          (subt_at t p))
    (showsl_lit "s sigma^b != t |_ p"))))))))
                  (fun x ->
                    Inl (comp (comp (showsl_lit
                                      "problem with application condition of non-loop theorem after deriving correct pattern rule\092")
                                (showsl_pat_rule (_A2, _A3) (_B5, _B8)
                                  ((s, (sigmaa, mua)),
                                    ((t, (sigma_t, mu_t)), is_pair))))
                          x))))
              ca))
          c);;

let rec check_non_loop_dp_prf (_B1, _B2, _B3)
  (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8)
  i dpp prf =
    (let p = pairsb i dpp in
     let r = rulese i dpp in
      bindb (check (null (qc i dpp))
              (showsl_lit "strategy for non-loops unsupported"))
        (fun _ ->
          check_non_loop_prf (_B1, _B2, _B3)
            (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8) r p prf));;

let rec rule_removal_nonterm_trs (_B1, _B2) (_C1, _C2)
  i tp (Rule_removal_nonterm_trs_prf r) =
    (let ra = rulesd i tp in
     let rrm =
       ceta_list_diff
         (compare_order_prod
           (compare_order_term _B1.compare_compare_order
             _C1.compare_compare_order)
           (compare_order_term _B1.compare_compare_order
             _C1.compare_compare_order))
         ra r
       in
      Inr (delete_R_Rwb i tp rrm rrm));;

let rec showsl_location = function A -> showsl_lit "above"
                          | Ba -> showsl_lit "below"
                          | H -> showsl_lit "here"
                          | Ra -> showsl_lit "right";;

let rec showsl_pattern _A _B
  (c, (s, p)) =
    comp (comp (comp (comp (comp (comp (showsl_lit "(")
                                   (showsl_terma _A _B (ctxt_apply_term c s)))
                             (showsl_lit ", "))
                       (showsl_pos (hole_pos c)))
                 (showsl_lit ", "))
           (showsl_location p))
      (showsl_lit ")");;

let rec fp_R_decide (_A1, _A2) (_B1, _B2, _B3, _B4)
  mu l oo q c t =
    not (bex ((ceq_prod (ceq_term _A2 _B4) (ceq_term _A2 _B4)),
               (ccompare_prod (ccompare_term _A1 _B3) (ccompare_term _A1 _B3)))
          (let h = hole_pos c in
           let q_s = filtera (fun qa -> is_left_of qa q) (poss_list t) in
           let m1 = map (fun qa -> (subt_at t qa, l)) q_s in
           let sterms =
             remdups (equal_term _A2 _B4)
               (map (si_subst mu)
                 (remdups _B4 (maps (fun qa -> si_W mu (subt_at t qa)) q_s)))
             in
           let uterms = maps supteq_list sterms in
           let m2 = map (fun u -> (u, l)) (remdups (equal_term _A2 _B4) uterms)
             in
           let p_s =
             filtera (fun qa -> is_left_of qa h)
               (poss_list (ctxt_apply_term c t))
             in
           let m3 = map (fun p -> (subt_at (ctxt_apply_term c t) p, l)) p_s in
           let stermsa =
             remdups (equal_term _A2 _B4)
               (map (si_subst mu)
                 (remdups _B4
                   (maps (fun p -> si_W mu (subt_at (ctxt_apply_term c t) p))
                     p_s)))
             in
           let utermsa = maps supteq_list stermsa in
           let m4 = map (fun u -> (u, l)) (remdups (equal_term _A2 _B4) utermsa)
             in
            set ((ceq_prod (ceq_term _A2 _B4) (ceq_term _A2 _B4)),
                  (ccompare_prod (ccompare_term _A1 _B3)
                    (ccompare_term _A1 _B3)),
                  (set_impl_prod set_impl_term set_impl_term))
              (m1 @ m2 @ m3 @ m4))
          (match_decision (_A1, _A2) (_B1, _B2, _B3, _B4) mu));;

let rec pos_dec
  p q oo =
    (if null p
      then (match remove_suffix equal_nat oo q with None -> None
             | Some r -> Some (zero_nata, r))
      else (let n0 =
              nat (ceiling floor_ceiling_rat
                    (divide_rata
                      (of_nat semiring_1_rat
                        (minus_nata (size_list oo) (size_list q)))
                      (of_nat semiring_1_rat (size_list p))))
              in
             (match remove_suffix equal_nat oo (power power_list p n0 @ q)
               with None -> None | Some r -> Some (n0, r))));;

let rec ctxt_subst
  uu uv n t =
    (if equal_nata n zero_nata then t
      else ctxt_apply_term uu
             (subst_apply_term (ctxt_subst uu uv (minus_nata n one_nata) t)
               uv));;

let rec h_match_probs (_A1, _A2) (_B1, _B2)
  mu l oo q c t =
    (match pos_dec (hole_pos c) q oo
      with None ->
        set_empty
          ((ceq_prod (ceq_term _B2 _A2) (ceq_term _B2 _A2)),
            (ccompare_prod (ccompare_term _B1 _A1) (ccompare_term _B1 _A1)))
          (of_phantom (set_impl_proda set_impl_term set_impl_term))
      | Some (n, o) ->
        inserta
          ((ceq_prod (ceq_term _B2 _A2) (ceq_term _B2 _A2)),
            (ccompare_prod (ccompare_term _B1 _A1) (ccompare_term _B1 _A1)))
          (subt_at (ctxt_subst c mu n t) o, l)
          (set_empty
            ((ceq_prod (ceq_term _B2 _A2) (ceq_term _B2 _A2)),
              (ccompare_prod (ccompare_term _B1 _A1) (ccompare_term _B1 _A1)))
            (of_phantom (set_impl_proda set_impl_term set_impl_term))));;

let rec fp_H_decide (_A1, _A2) (_B1, _B2, _B3, _B4)
  mu l oo q c t =
    not (bex ((ceq_prod (ceq_term _A2 _B4) (ceq_term _A2 _B4)),
               (ccompare_prod (ccompare_term _A1 _B3) (ccompare_term _A1 _B3)))
          (h_match_probs (_B3, _B4) (_A1, _A2) (si_subst mu) l oo q c t)
          (match_decision (_A1, _A2) (_B1, _B2, _B3, _B4) mu));;

let rec decompositions
  p = map (fun pa -> (pa, the (remove_prefix equal_nat pa p))) (prefix_list p);;

let rec simplify_emp_main (_A1, _A2) (_B1, _B2, _B3, _B4)
  mu_incr x1 solved = match mu_incr, x1, solved with
    mu_incr, (Hole, (Fun (g, ls), (c, (t, [])))), solved ->
      (if gmatch_decision (_A1, _A2) (_B1, _B2, _B3, _B4) mu_incr
            ((t, Fun (g, ls)) :: solved)
        then Some None
        else (if equal_ctxta _A2 _B4 c Hole then None
               else simplify_emp_main (_A1, _A2) (_B1, _B2, _B3, _B4) mu_incr
                      (c, (Fun (g, ls),
                            (subst_apply_ctxt c (si_subst mu_incr),
                              (subst_apply_term t (si_subst mu_incr), []))))
                      solved))
    | mu_incr, (More (f, bef, d, aft), (Fun (g, ls), (c, (t, [])))), solved ->
        bind (guard (eq _A2 f g))
          (fun _ ->
            bind (guard
                   (equal_nata (size_list ls)
                     (suc (plus_nata (size_list bef) (size_list aft)))))
              (fun _ ->
                (let pairs_bef = zip bef (take (size_list bef) ls) in
                 let pairs_aft = zip aft (drop (suc (size_list bef)) ls) in
                  simplify_emp_main (_A1, _A2) (_B1, _B2, _B3, _B4) mu_incr
                    (d, (nth ls (size_list bef),
                          (c, (t, pairs_bef @ pairs_aft))))
                    solved)))
    | mu_incr, (d, (Var x, (c, (t, [])))), solved ->
        Some (Some (d, (Var x, (c, (t, solved)))))
    | mu_incr, (d, (l, (c, (t, (Var x, Fun (f, ls)) :: mp)))), solved ->
        bind (guard (member (_B1, _B2) x (si_v_incr mu_incr)))
          (fun _ ->
            (let m =
               map (fun (s, a) -> (subst_apply_term s (si_subst mu_incr), a)) in
              simplify_emp_main (_A1, _A2) (_B1, _B2, _B3, _B4) mu_incr
                (subst_apply_ctxt d (si_subst mu_incr),
                  (l, (subst_apply_ctxt c (si_subst mu_incr),
                        (subst_apply_term t (si_subst mu_incr),
                          m ((Var x, Fun (f, ls)) :: mp)))))
                (m solved)))
    | mu_incr, (d, (l, (c, (t, (Fun (g, ts), Fun (f, ls)) :: mp)))), solved ->
        bind (guard (eq _A2 f g))
          (fun _ ->
            bind (zip_option ts ls)
              (fun pairs ->
                simplify_emp_main (_A1, _A2) (_B1, _B2, _B3, _B4) mu_incr
                  (d, (l, (c, (t, pairs @ mp)))) solved))
    | mu_incr, (d, (l, (c, (t, (s, Var x) :: mp)))), solved ->
        simplify_emp_main (_A1, _A2) (_B1, _B2, _B3, _B4) mu_incr
          (d, (l, (c, (t, mp)))) ((s, Var x) :: solved);;

let rec simplify_emp (_A1, _A2) (_B1, _B2, _B3, _B4)
  mu_incr emp =
    (let (d, (l, (c, t))) = emp in
      (match
        simplify_emp_main (_A1, _A2) (_B1, _B2, _B3, _B4) mu_incr
          (d, (l, (c, (t, [])))) []
        with None -> Inr false
        | Some a -> (match a with None -> Inr true | Some aa -> Inl aa)));;

let rec eident_prob_to_ident_prob
  (d, (si, (c, t))) = (ctxt_apply_term d t, si);;

let rec eident_prob_of_semp _A _B
  (d, (l, (c, (t, mp)))) =
    bind (map_of (equal_term _A _B) (reverse_rules mp) l)
      (fun si -> Some (d, (si, (c, t))));;

let rec ident_prob_of_semp _A _B
  (d, (l, (c, (t, mp)))) = ident_prob_of_smp _A _B mp;;

let rec ident_prob_of_emp (_A1, _A2) (_B1, _B2, _B3, _B4)
  mu_incr emp =
    (match simplify_emp (_A1, _A2) (_B1, _B2, _B3, _B4) mu_incr emp
      with Inl semp ->
        Some (map eident_prob_to_ident_prob
                (option_to_list (eident_prob_of_semp _A2 _B4 semp)) @
               ident_prob_of_semp _A2 _B4 semp)
      | Inr true -> Some [] | Inr false -> None);;

let rec n0b
  p q oo =
    nat (ceiling floor_ceiling_rat
          (divide_rata
            (of_nat semiring_1_rat
              (minus_nata (suc (size_list oo)) (size_list q)))
            (of_nat semiring_1_rat (size_list p))));;

let rec fp_B_decide (_A1, _A2) (_B1, _B2, _B3, _B4)
  mu l oo q c t =
    not (bex ((ceq_prod (ceq_term _A2 _B4) (ceq_term _A2 _B4)),
               (ccompare_prod (ccompare_term _A1 _B3) (ccompare_term _A1 _B3)))
          (foldr
            (comp (sup_seta
                    ((ceq_prod (ceq_term _A2 _B4) (ceq_term _A2 _B4)),
                      (ccompare_prod (ccompare_term _A1 _B3)
                        (ccompare_term _A1 _B3))))
              (fun qa ->
                h_match_probs (_B3, _B4) (_A1, _A2) (si_subst mu) l oo qa c t))
            (proper_prefix_list q)
            (set_empty
              ((ceq_prod (ceq_term _A2 _B4) (ceq_term _A2 _B4)),
                (ccompare_prod (ccompare_term _A1 _B3) (ccompare_term _A1 _B3)))
              (of_phantom (set_impl_proda set_impl_term set_impl_term))))
          (match_decision (_A1, _A2) (_B1, _B2, _B3, _B4) mu)) &&
      not (bex ((ceq_prod (ceq_ctxt _A2 _B4)
                  (ceq_prod (ceq_term _A2 _B4)
                    (ceq_prod (ceq_ctxt _A2 _B4) (ceq_term _A2 _B4)))),
                 (ccompare_prod (ccompare_ctxt _A1 _B3)
                   (ccompare_prod (ccompare_term _A1 _B3)
                     (ccompare_prod (ccompare_ctxt _A1 _B3)
                       (ccompare_term _A1 _B3)))))
            (let p = hole_pos c in
             let n = (fun pa -> n0b p pa oo) in
             let ps =
               filtera
                 (fun (pa, pb) ->
                   less_pos oo (pb @ power power_list p (n pb)) &&
                     less_pos pa p)
                 (remdups
                   (equal_prod (equal_list equal_nat) (equal_list equal_nat))
                   (decompositions p))
               in
              set ((ceq_prod (ceq_ctxt _A2 _B4)
                     (ceq_prod (ceq_term _A2 _B4)
                       (ceq_prod (ceq_ctxt _A2 _B4) (ceq_term _A2 _B4)))),
                    (ccompare_prod (ccompare_ctxt _A1 _B3)
                      (ccompare_prod (ccompare_term _A1 _B3)
                        (ccompare_prod (ccompare_ctxt _A1 _B3)
                          (ccompare_term _A1 _B3)))),
                    (set_impl_prod set_impl_ctxt
                      (set_impl_prod set_impl_term
                        (set_impl_prod set_impl_ctxt set_impl_term))))
                (map (fun (pb, pa) ->
                       (subt_at_ctxt c pb,
                         (l, (subst_apply_ctxt c (si_subst mu),
                               subst_apply_term
                                 (ctxt_subst c (si_subst mu) (n pa) t)
                                 (si_subst mu)))))
                  (remdups
                    (equal_prod (equal_list equal_nat) (equal_list equal_nat))
                    ps)))
            (fun ep ->
              bex ((ceq_list (ceq_prod (ceq_term _A2 _B4) (ceq_term _A2 _B4))),
                    (ccompare_list
                      (ccompare_prod (ccompare_term _A1 _B3)
                        (ccompare_term _A1 _B3))))
                (set_option
                  ((ceq_list (ceq_prod (ceq_term _A2 _B4) (ceq_term _A2 _B4))),
                    (ccompare_list
                      (ccompare_prod (ccompare_term _A1 _B3)
                        (ccompare_term _A1 _B3))),
                    set_impl_list)
                  (ident_prob_of_emp (_A1, _A2) (_B1, _B2, _B3, _B4) mu ep))
                (list_all (ident_decision (_A1, _A2) (_B3, _B4) mu))));;

let rec n0
  p q oo =
    nat (ceiling floor_ceiling_rat
          (divide_rata
            (of_nat semiring_1_rat (minus_nata (size_list oo) (size_list q)))
            (of_nat semiring_1_rat (size_list p))));;

let rec bounded_postfixes
  p ps =
    map_filter (fun x -> (if not (is_none x) then Some (the x) else None))
      (map (remove_prefix equal_nat p) ps);;

let rec fp_A_decide (_A1, _A2) (_B1, _B2, _B3, _B4)
  mu l oo q c t =
    not (bex ((ceq_prod (ceq_term _A2 _B4) (ceq_term _A2 _B4)),
               (ccompare_prod (ccompare_term _A1 _B3) (ccompare_term _A1 _B3)))
          (match subt_at t q
            with Var _ ->
              set_empty
                ((ceq_prod (ceq_term _A2 _B4) (ceq_term _A2 _B4)),
                  (ccompare_prod (ccompare_term _A1 _B3)
                    (ccompare_term _A1 _B3)))
                (of_phantom (set_impl_proda set_impl_term set_impl_term))
            | Fun (_, _) ->
              (let h = hole_pos c in
               let n = n0 h q oo in
               let hn = power power_list h n in
               let cs = ctxt_subst c (si_subst mu) n t in
               let q_s = bounded_postfixes q (poss_list t) in
               let qoo_s =
                 maps (fun qa ->
                        map (fun a -> (qa, a)) (prefix_list (hn @ q @ qa)))
                   q_s
                 in
               let qoo_sf =
                 filtera (fun qoo -> less_pos (hn @ q) (snd qoo @ oo)) qoo_s in
               let m1 = map (fun qoo -> (subt_at cs (snd qoo), l)) qoo_sf in
               let sterms =
                 remdups (equal_term _A2 _B4)
                   (map (si_subst mu) (si_W mu (subt_at t q)))
                 in
               let uterms =
                 maps (comp (filtera (fun ta -> not (is_Var ta))) supteq_list)
                   sterms
                 in
               let m2 =
                 map (fun u -> (u, l)) (remdups (equal_term _A2 _B4) uterms) in
                set ((ceq_prod (ceq_term _A2 _B4) (ceq_term _A2 _B4)),
                      (ccompare_prod (ccompare_term _A1 _B3)
                        (ccompare_term _A1 _B3)),
                      (set_impl_prod set_impl_term set_impl_term))
                  (m1 @ m2)))
          (match_decision (_A1, _A2) (_B1, _B2, _B3, _B4) mu));;

let rec fp_decide (_A1, _A2) (_B1, _B2, _B3, _B4)
  mu = (fun (q, (c, t)) (l, (la, loc)) ->
         (if equal_locationa loc H
           then fp_H_decide (_A1, _A2) (_B1, _B2, _B3, _B4) mu
                  (ctxt_apply_term l la) (hole_pos l) q c t
           else true) &&
           ((if equal_locationa loc A
              then fp_A_decide (_A1, _A2) (_B1, _B2, _B3, _B4) mu
                     (ctxt_apply_term l la) (hole_pos l) q c t
              else true) &&
             ((if equal_locationa loc Ba
                then fp_B_decide (_A1, _A2) (_B1, _B2, _B3, _B4) mu
                       (ctxt_apply_term l la) (hole_pos l) q c t
                else true) &&
               (if equal_locationa loc Ra
                 then fp_R_decide (_A1, _A2) (_B1, _B2, _B3, _B4) mu
                        (ctxt_apply_term l la) (hole_pos l) q c t
                 else true))));;

let rec fp_valid (_A1, _A2) (_B1, _B2) (_C1, _C2)
  p = ball ((ceq_prod (ceq_ctxt _A2 _B2) (ceq_prod (ceq_term _A2 _B2) _C1)),
             (ccompare_prod (ccompare_ctxt _A1 _B1)
               (ccompare_prod (ccompare_term _A1 _B1) _C2)))
        p (fun (l, (la, _)) -> not (is_Var (ctxt_apply_term l la)));;

let rec check_fploop (_A1, _A2, _A3) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  r p (FP_loop_prf (c, sigma, t, seq)) =
    (let mu = subst_incr (_B1, _B2, _B4, _B6) _A2 sigma in
     let mua = si_subst mu in
      bindb (check (not (null seq))
              (showsl_lit "looping reduction must not be empty"))
        (fun _ ->
          bindb (check
                  (equal_terma _A2 _B4 (let (_, (_, ta)) = last seq in ta)
                    (ctxt_apply_term c (subst_apply_term t mua)))
                  (showsl_lit "last term in sequence is not C[t sigma]"))
            (fun _ ->
              bindb (check
                      (fp_valid (_A1, _A2) (_B3, _B4)
                        (ceq_location, ccompare_location)
                        (set ((ceq_prod (ceq_ctxt _A2 _B4)
                                (ceq_prod (ceq_term _A2 _B4) ceq_location)),
                               (ccompare_prod (ccompare_ctxt _A1 _B3)
                                 (ccompare_prod (ccompare_term _A1 _B3)
                                   ccompare_location)),
                               (set_impl_prod set_impl_ctxt
                                 (set_impl_prod set_impl_term
                                   set_impl_location)))
                          p))
                      (showsl_lit
                        "lhss in forbidden patterns must not be variables"))
                (fun _ ->
                  bindb (catch_error
                          (forallM
                            (fun (l, _) ->
                              check (not (is_Var l))
                                (showsl_lit
                                  "lhss of TRS must not be variables"))
                            r)
                          (fun x -> Inl (snd x)))
                    (fun _ ->
                      (let seqa =
                         zip (t :: map (fun (_, (_, ta)) -> ta) seq) seq in
                        catch_error
                          (forallM
                            (fun (ta, (q, (ra, s))) ->
                              bindb (check_rstep (_A2, _A3) (_B2, _B4, _B5, _B7)
                                      r q ra ta s)
                                (fun _ ->
                                  (let check_fpstep =
                                     fp_decide (_A1, _A2) (_B1, _B2, _B3, _B4)
                                       mu (q, (c, ta))
                                     in
                                    catch_error
                                      (forallM
(fun pt ->
  check (check_fpstep pt)
    (comp (comp (comp (comp (comp (comp (comp
  (showsl_lit "iterating reduction ") (showsl_terma _A3 _B7 ta))
                                    (showsl_lit " -->"))
                              (showsl_pos q))
                        (showsl_lit " "))
                  (showsl_terma _A3 _B7 s))
            (showsl_lit " does not respect forbidden pattern "))
      (showsl_pattern _A3 _B7 pt)))
p)
                                      (fun x -> Inl (snd x)))))
                            seqa)
                          (fun x -> Inl (snd x))))))));;

let rec check_fp_tp_subsumes (_A1, _A2, _A3) (_B1, _B2, _B3)
  ta t =
    (let ((p, r), (pa, ra)) = (ta, t) in
      catch_error
        (bindb
          (check
            (equal_lista
              (equal_prod
                (equal_ctxt (equal_lab _A2 _B2) (equal_list equal_char))
                (equal_prod
                  (equal_term (equal_lab _A2 _B2) (equal_list equal_char))
                  equal_location))
              p pa)
            (showsl_lit "difference in forbidden patterns strategy"))
          (fun _ ->
            catch_error
              (check_subseteq
                (equal_prod
                  (equal_term (equal_lab _A2 _B2) (equal_list equal_char))
                  (equal_term (equal_lab _A2 _B2) (equal_list equal_char)))
                r ra)
              (fun x ->
                Inl (toomuch "rule"
                      (showsl_rule (showl_lab _A3 _B3) (showl_list showl_char)
                        x)))))
        (fun x ->
          Inl (comp (showsl_lit
                      "problem in showing subsumption for non-termination\092")
                x)));;

let rec string_reversal_complete_tt _B _C
  i tp =
    (let r = rulesd i tp in
      bindb (check (q_emptyb i tp) (showsl_lit "Q is not empty"))
        (fun _ ->
          bindb (check_unary_signature _B _C r)
            (fun _ -> Inr (mkc i default_nfs_nt_trs [] (map rev_rule r) []))));;

let rec showsl_lines _A
  desc_empty = showsl_list_gen (showsl _A) desc_empty "" "\092" "";;

let rec showsl_tree_automaton _A _B
  (Tree_Automaton (fin, rules, eps)) =
    comp (comp (comp (comp (comp (comp (showsl_lit "final: ")
                                   (showsl_list _A fin))
                             (showsl_lit "\092rules: "))
                       (showsl_lines (showl_ta_rule _A _B) "empty" rules))
                 (showsl_lit "\092epsilon: "))
           (showsl_list_prod _A _A eps))
      (showsl_literal "\092");;

let rec ta_idx_impl (f, (r, (e, (rhs, (idx, (det, (efcl, eicl))))))) = idx;;

let rec comp_res_of_order = function Eq -> EQUAL
                            | Lt -> LESS
                            | Gt -> GREATER;;

let rec compare_res _A x y = comp_res_of_order (compare _A x y);;

let rec set_iterator_image g it = (fun c f -> it c (fun x -> f (g x)));;

let rec map_iterator_dom it = set_iterator_image fst it;;

let rec list_eq
  eq x1 x2 = match eq, x1, x2 with eq, [], [] -> true
    | eq, aa :: la, a :: l -> (if eq aa a then list_eq eq la l else false)
    | uu, v :: va, [] -> false
    | uu, [], v :: va -> false;;

let rec rule_filter_opt_code _A _B
  a b c =
    (match
      rbt_comp_lookup (compare_proda _B.compare_compare_order compare_nat)
        (ta_idx_impl a) (b, size_list c)
      with None -> Empty
      | Some xc ->
        gen_image (comp map_iterator_dom rm_iterateoi) Empty
          (fun k -> rbt_comp_insert (compare _A.compare_compare_order) k ()) snd
          (gen_filter (comp map_iterator_dom rm_iterateoi) Empty
            (fun k ->
              rbt_comp_insert
                (compare_proda (compare_list _A.compare_compare_order)
                  _A.compare_compare_order)
                k ())
            (fun (xd, _) ->
              list_eq (comp2eq (compare_res _A.compare_compare_order)) xd c)
            xc));;

let rec ta_rhs_states_impl
  (f, (r, (e, (rhs, (idx, (det, (efcl, eicl))))))) = rhs;;

let rec update_all_code _A
  m s v =
    comp map_iterator_dom rm_iterateoi s (fun _ -> true)
      (fun x -> rbt_comp_insert (compare _A) x v) m;;

let rec update_all2_code _A _B
  m s1 s2 v =
    comp map_iterator_dom rm_iterateoi s1 (fun _ -> true)
      (fun x sigma ->
        (let xa =
           update_all_code _B
             (match rbt_comp_lookup (compare _A) sigma x with None -> Empty
               | Some a -> id a)
             s2 v
           in
          rbt_comp_insert (compare _A) x xa sigma))
      m;;

let rec union_image_rs_code _A _B
  s f = comp map_iterator_dom rm_iterateoi s (fun _ -> true)
          (fun x ->
            rbt_comp_union_with_key (compare _B) (fun _ _ rv -> rv) (f x))
          Empty;;

let rec ta_eps_cl_impl
  (f, (r, (e, (rhs, (idx, (det, (efcl, eicl))))))) q = efcl q;;

let rec ta_match_var_ref_code _A _B _C
  ta qsig x q =
    comp map_iterator_dom rm_iterateoi qsig (fun _ -> true)
      (fun xa sigma ->
        (if not (gen_disjoint (comp map_iterator_dom rm_iterateoi)
                  (fun k s ->
                    (match
                      rbt_comp_lookup (compare _A.compare_compare_order) s k
                      with None -> false | Some _ -> true))
                  (ta_eps_cl_impl ta xa) q)
          then rbt_comp_insert
                 (compare_lista (compare_prod _C _A.compare_compare_order))
                 [(x, xa)] () sigma
          else sigma))
      Empty;;

let rec set_App_code _A
  x xs =
    comp map_iterator_dom rm_iterateoi x (fun _ -> true)
      (fun xa ->
        comp map_iterator_dom rm_iterateoi xs (fun _ -> true)
          (fun xaa -> rbt_comp_insert (compare_lista _A) (xa @ xaa) ()))
      Empty;;

let rec concat_listset_code _A
  = rec_list (rbt_comp_insert (compare_lista _A) [] () Empty)
      (fun x _ -> set_App_code _A x);;

let rec map2 f x1 ys = match f, x1, ys with f, [], ys -> []
               | f, v :: va, [] -> []
               | f, x :: xs, y :: ys -> f x y :: map2 f xs ys;;

let rec ta_match_code _A _B _C
  ta qsig x2 q = match ta, qsig, x2, q with
    ta, qsig, Var x, q -> ta_match_var_ref_code _A _B _C ta qsig x q
    | a, b, Fun (c, d), e ->
        (match
          rbt_comp_lookup (compare_proda _B.compare_compare_order compare_nat)
            (ta_idx_impl a) (c, size_list d)
          with None -> Empty
          | Some xe ->
            union_image_rs_code
              (compare_prod (compare_list _A.compare_compare_order)
                _A.compare_compare_order)
              (compare_list (compare_prod _C _A.compare_compare_order)) xe
              (fun (xf, xg) ->
                (if not (gen_disjoint (comp map_iterator_dom rm_iterateoi)
                          (fun k s ->
                            (match
                              rbt_comp_lookup (compare _A.compare_compare_order)
                                s k
                              with None -> false | Some _ -> true))
                          (ta_eps_cl_impl a xg) e)
                  then concat_listset_code
                         (compare_prod _C _A.compare_compare_order)
                         (map2 (fun t q ->
                                 ta_match_code _A _B _C a b t
                                   (rbt_comp_insert
                                     (compare _A.compare_compare_order) q ()
                                     Empty))
                           d xf)
                  else Empty)));;

let rec ta_res_args_aux_code _A _B
  = (fun x xa xb ->
      (match
        rbt_comp_lookup (compare_proda _B.compare_compare_order compare_nat)
          (ta_idx_impl x) (xa, size_list xb)
        with None -> Empty
        | Some xc ->
          union_image_rs_code
            (compare_prod (compare_list _A.compare_compare_order)
              _A.compare_compare_order)
            _A.compare_compare_order xc
            (fun (xd, xe) ->
              (if list_all2
                    (fun xf s ->
                      (match
                        rbt_comp_lookup (compare _A.compare_compare_order) s xf
                        with None -> false | Some _ -> true))
                    xd xb
                then ta_eps_cl_impl x xe else Empty))));;

let rec ta_res_code _A _B
  = (fun x xa ->
      rec_term (fun xb xc -> ta_eps_cl_impl xc xb)
        (fun xb xc xd ->
          (let a = map (fun xe -> snd xe xd) xc in
            ta_res_args_aux_code _A _B xd xb a))
        xa x);;

let rec is_compatible_code _A _B (_C1, _C2)
  ta r =
    comp map_iterator_dom rm_iterateoi r is_Inr
      (fun x sigma ->
        (let (a, b) = x in
         let xa =
           ta_match_code _A _B _C1 ta (ta_rhs_states_impl ta) a
             (ta_rhs_states_impl ta)
           in
          comp map_iterator_dom rm_iterateoi xa is_Inr
            (fun xb sigmaa ->
              (let xc = map_term (fun xg -> xg) (fun_of _C2 xb) a in
               let xd = ta_res_code _A _B ta xc in
                (if gen_isEmptya
                      (gen_balla (comp map_iterator_dom rm_iterateoi)) xd
                  then sigmaa
                  else (let xe = map_term (fun xg -> xg) (fun_of _C2 xb) b in
                        let xf = ta_res_code _A _B ta xe in
                         (if gen_isEmptya
                               (gen_balla (comp map_iterator_dom rm_iterateoi))
                               xf
                           then (let xg =
                                   gen_pick (comp map_iterator_dom rm_iterateoi)
                                     xd
                                   in
                                  Inl (xg, (xc, xe)))
                           else (let aa =
                                   update_all2_code _A.compare_compare_order
                                     _A.compare_compare_order (projr sigmaa) xd
                                     xf (xc, xe)
                                   in
                                  Inr aa))))))
            sigma))
      (Inr Empty);;

let rec ta_rules_impla (f, (r, (e, (rhs, (idx, (det, (efcl, eicl))))))) = r;;

let rec ta_final_impla (f, (r, (e, (rhs, (idx, (det, (efcl, eicl))))))) = f;;

let rec ta_idx_rhs_init_code _A _B _C
  det rs efcl =
    comp map_iterator_dom rm_iterateoi rs (fun _ -> true)
      (fun x (a, (aa, ba)) ->
        (let TA_rule (xg, xh, xi) = x in
         let y = size_list xh in
          (match rbt_comp_lookup (compare_proda _B compare_nat) a (xg, y)
            with None ->
              (rbt_comp_insert (compare_proda _B compare_nat) (xg, y)
                 (rbt_comp_insert (compare_proda (compare_list _A) _A) (xh, xi)
                   () Empty)
                 a,
                (aa, rbt_comp_union_with_key (compare _C) (fun _ _ rv -> rv)
                       (efcl xi) ba))
            | Some xj ->
              (rbt_comp_insert (compare_proda _B compare_nat) (xg, y)
                 (rbt_comp_insert (compare_proda (compare_list _A) _A) (xh, xi)
                   () xj)
                 a,
                ((if aa
                   then gen_balla (comp map_iterator_dom rm_iterateoi) xj
                          (fun (xk, _) ->
                            not (list_eq (comp2eq (compare_res _A)) xh xk))
                   else false),
                  rbt_comp_union_with_key (compare _C) (fun _ _ rv -> rv)
                    (efcl xi) ba)))))
      (Empty, (det, Empty));;

let rec prod_eq eqa eqb x1 x2 = (let (a1, b1) = x1 in
                                 let (a2, b2) = x2 in
                                  eqa a1 a2 && eqb b1 b2);;

let rec glist_member eq x xa2 = match eq, x, xa2 with eq, x, [] -> false
                       | eq, x, y :: ys -> eq x y || glist_member eq x ys;;

let rec glist_insert eq x xs = (if glist_member eq x xs then xs else x :: xs);;

let rec ta_make_code _A _B _C
  f r e =
    (let a =
       ta_idx_rhs_init_code _B.compare_compare_order _C _B.compare_compare_order
         (is_Nil e) r
         (fun xi ->
           impl_of _B.linorder_compare_order (memo_rbt_rtrancl _B e xi))
       in
     let (ab, (aa, ba)) = a in
      (f, (r, (e, (ba, (ab, (aa, ((fun s ->
                                    impl_of _B.linorder_compare_order
                                      (memo_rbt_rtrancl _B e s)),
                                   (fun s ->
                                     impl_of _B.linorder_compare_order
                                       (memo_rbt_rtrancl _B
 (gen_image foldli []
   (glist_insert
     (prod_eq (comp2eq (compare_res _B.compare_compare_order))
       (comp2eq (compare_res _B.compare_compare_order))))
   (fun xc -> (snd xc, fst xc)) e)
 s))))))))));;

let rec ta_eps_impl (f, (r, (e, (rhs, (idx, (det, (efcl, eicl))))))) = e;;

let rec is_None a = (match a with None -> true | Some _ -> false);;

let rec ta_only_res_wits_code _A _B
  a b = (let y =
           gen_filter (comp map_iterator_dom rm_iterateoi) Empty
             (fun k -> rbt_comp_insert (compare _A.compare_compare_order) k ())
             (fun xb ->
               not (is_None
                     (rbt_comp_lookup (compare _A.compare_compare_order) b xb)))
             (ta_final_impla a)
           in
         let y_a =
           gen_filter (comp map_iterator_dom rm_iterateoi) Empty
             (fun k ->
               rbt_comp_insert
                 (compare_ta_rule _A.compare_compare_order
                   _B.compare_compare_order)
                 k ())
             (fun xb ->
               list_all_rec
                 (fun xc ->
                   not (is_None
                         (rbt_comp_lookup (compare _A.compare_compare_order) b
                           xc)))
                 (r_lhs_states xb))
             (ta_rules_impla a)
           in
         let aa =
           filtera
             (fun xb ->
               not (is_None
                     (rbt_comp_lookup (compare _A.compare_compare_order) b
                       (fst xb))))
             (ta_eps_impl a)
           in
          ta_make_code _A.compare_compare_order _A _B.compare_compare_order y
            y_a aa);;

let rec ta_only_prs_wits_code _A _B
  a b = (let y =
           gen_filter (comp map_iterator_dom rm_iterateoi) Empty
             (fun k -> rbt_comp_insert (compare _A.compare_compare_order) k ())
             (fun xb ->
               not (is_None
                     (rbt_comp_lookup (compare _A.compare_compare_order) b xb)))
             (ta_final_impla a)
           in
         let y_a =
           gen_filter (comp map_iterator_dom rm_iterateoi) Empty
             (fun k ->
               rbt_comp_insert
                 (compare_ta_rule _A.compare_compare_order
                   _B.compare_compare_order)
                 k ())
             (fun xb ->
               not (is_None
                     (rbt_comp_lookup (compare _A.compare_compare_order) b
                       (r_rhs xb))))
             (ta_rules_impla a)
           in
         let aa =
           filtera
             (fun xb ->
               not (is_None
                     (rbt_comp_lookup (compare _A.compare_compare_order) b
                       (snd xb))))
             (ta_eps_impl a)
           in
          ta_make_code _A.compare_compare_order _A _B.compare_compare_order y
            y_a aa);;

let rec next_res_wit_code _A _B
  r m = comp map_iterator_dom rm_iterateoi r is_None
          (fun x _ ->
            (let TA_rule (xb, xc, xd) = x in
              bind (mapMa (rbt_comp_lookup (compare _A) m) xc)
                (fun xe -> Some (xd, Fun (xb, xe)))))
          None;;

let rec res_wits_code _A _B
  tAi = (let x = (ta_rules_impla tAi, Empty) in
         let (_, b) =
           whilea
             (fun (xb, _) ->
               not (gen_isEmptya
                     (gen_balla (comp map_iterator_dom rm_iterateoi)) xb))
             (fun (a, b) ->
               (match
                 next_res_wit_code _A.compare_compare_order
                   _B.compare_compare_order a b
                 with None -> (Empty, b)
                 | Some (aa, ba) ->
                   (let xd = ta_eps_cl_impl tAi aa in
                    let xe = update_all_code _A.compare_compare_order b xd ba in
                    let xf =
                      gen_filter (comp map_iterator_dom rm_iterateoi) Empty
                        (fun k ->
                          rbt_comp_insert
                            (compare_ta_rule _A.compare_compare_order
                              _B.compare_compare_order)
                            k ())
                        (fun xf ->
                          not (match
                                rbt_comp_lookup
                                  (compare _A.compare_compare_order) xd
                                  (r_rhs xf)
                                with None -> false | Some _ -> true))
                        a
                      in
                     (xf, xe))))
             x
           in
          b);;

let rec ta_eps_icl_impl
  (f, (r, (e, (rhs, (idx, (det, (efcl, eicl))))))) q = eicl q;;

let rec ctxt_compose
  x0 d = match x0, d with Hole, d -> d
    | More (f, ss1, c, ss2), d -> More (f, ss1, ctxt_compose c d, ss2);;

let rec prs_wits_code _A _B
  tAi = (let x =
           update_all_code _A.compare_compare_order Empty
             (union_image_rs_code _A.compare_compare_order
               _A.compare_compare_order (ta_final_impla tAi)
               (ta_eps_icl_impl tAi))
             Hole
           in
         let xa = (false, (ta_rules_impla tAi, x)) in
         let (_, (_, ba)) =
           whilea (fun (xb, (_, _)) -> not xb)
             (fun (_, (aa, ba)) ->
               comp map_iterator_dom rm_iterateoi aa (fun _ -> true)
                 (fun xc (ab, (ac, bc)) ->
                   (match
                     rbt_comp_lookup (compare _A.compare_compare_order) bc
                       (r_rhs xc)
                     with None ->
                       (ab, (rbt_comp_insert
                               (compare_ta_rule _A.compare_compare_order
                                 _B.compare_compare_order)
                               xc () ac,
                              bc))
                     | Some xd ->
                       (false,
                         (ac, (let TA_rule (xj, xk, _) = xc in
                                rec_list (fun _ xn _ _ _ -> xn)
                                  (fun xm xn xo xp xq xr xs xt ->
                                    (let y_b =
                                       update_all_code _A.compare_compare_order
 xq (ta_eps_icl_impl xp (the_Var xm))
 (ctxt_compose xr (More (xs, xt, Hole, xn)))
                                       in
                                      xo xp y_b xr xs (xt @ [xm])))
                                  (map (fun a -> Var a) xk) tAi bc xd xj [])))))
                 (true, (Empty, ba)))
             xa
           in
          ba);;

let rec trim_ta_wits_code _A _B
  tAi = (let x = res_wits_code _A _B tAi in
         let xa = ta_only_res_wits_code _A _B tAi x in
         let xb = prs_wits_code _A _B xa in
         let xc = ta_only_prs_wits_code _A _B xa xb in
          (xc, (x, xb)));;

let rec is_coh_final_code _A
  fin rel =
    rm_iterateoi rel is_None
      (fun x _ ->
        (let (a, b) = x in
          (if (match rbt_comp_lookup (compare _A) fin a with None -> false
                | Some _ -> true)
            then rm_iterateoi b is_None
                   (fun xa _ ->
                     (let (aa, (ab, bb)) = xa in
                       (if (match rbt_comp_lookup (compare _A) fin aa
                             with None -> false | Some _ -> true)
                         then None else Some (ab, bb))))
                   None
            else None)))
      None;;

let rec map_add2_code _A _B
  m1 m2 =
    rm_iterateoi m2 (fun _ -> true)
      (fun x sigma ->
        (let (a, b) = x in
          (match rbt_comp_lookup (compare _A) sigma a
            with None -> rbt_comp_insert (compare _A) a b sigma
            | Some xc ->
              rbt_comp_insert (compare _A) a
                (rbt_comp_union_with_key (compare _B) (fun _ _ rv -> rv) xc b)
                sigma)))
      m1;;

let rec ta_check_comcoh_code _A _B (_C1, _C2)
  ta r =
    (let a = trim_ta_wits_code _A _B ta in
     let (ab, (aa, ba)) = a in
     let xa = is_compatible_code _A _B (_C1, _C2) ab r in
     let xb =
       (fun xf xg ->
         subst_apply_term
           (ctxt_apply_term
             (the (rbt_comp_lookup (compare _A.compare_compare_order) ba xf))
             xg)
           (fun xh ->
             the (rbt_comp_lookup (compare _A.compare_compare_order) aa xh)))
       in
      (match xa with Inl (aba, (ac, bc)) -> Some (xb aba ac, xb aba bc)
        | Inr ra ->
          (match
            (let (aba, bb) =
               whilea
                 (fun ac ->
                   (match ac with (Inl _, _) -> false
                     | (Inr xj, _) ->
                       not (gen_isEmpty (gen_ball rm_iterateoi) xj)))
                 (fun (aba, bb) ->
                   (let xd =
                      comp map_iterator_dom rm_iterateoi (ta_rules_impla ab)
                        is_Inr
                        (fun xd sigma ->
                          (let TA_rule (xe, y, z) = xd in
                            foldli (upt zero_nata (size_list y)) is_Inr
                              (fun xf s ->
                                (let xg = nth y xf in
                                  (match
                                    rbt_comp_lookup
                                      (compare _A.compare_compare_order)
                                      (projr aba) xg
                                    with None -> s
                                    | Some xh ->
                                      rm_iterateoi xh is_Inr
(fun xi sigmaa ->
  (let (ac, (ad, bd)) = xi in
   let xj = map (fun ae -> Var ae) (take xf y) in
   let xk = map (fun ae -> Var ae) (drop (suc xf) y) in
   let xl = Fun (xe, xj @ ad :: xk) in
   let xm = Fun (xe, xj @ bd :: xk) in
   let xn = list_update y xf ac in
   let xo = rule_filter_opt_code _A _B ab xe xn in
    (if gen_isEmptya (gen_balla (comp map_iterator_dom rm_iterateoi)) xo
      then Inl (z, (xl, xm))
      else (let xp =
              gen_filter (comp map_iterator_dom rm_iterateoi) Empty
                (fun k ->
                  rbt_comp_insert (compare _A.compare_compare_order) k ())
                (fun xv ->
                  is_None
                    (bind (rbt_comp_lookup (compare _A.compare_compare_order) bb
                            z)
                      (fun xw ->
                        rbt_comp_lookup (compare _A.compare_compare_order) xw
                          xv)))
                xo
              in
             (if gen_isEmptya (gen_balla (comp map_iterator_dom rm_iterateoi))
                   xp
               then sigmaa
               else Inr (update_all2_code _A.compare_compare_order
                          _A.compare_compare_order (projr sigmaa)
                          (rbt_comp_insert (compare _A.compare_compare_order) z
                            () Empty)
                          xp (xl, xm)))))))
s)))
                              sigma))
                        (Inr Empty)
                      in
                     (if is_Inr xd
                       then (xd, map_add2_code _A.compare_compare_order
                                   _A.compare_compare_order bb (projr xd))
                       else (xd, bb))))
                 (Inr ra, ra)
               in
              (if is_Inr aba then Inr bb else aba))
            with Inl (aba, (ac, bc)) -> Some (xb aba ac, xb aba bc)
            | Inr raa ->
              (match
                is_coh_final_code _A.compare_compare_order (ta_final_impla ab)
                  raa
                with None -> None
                | Some (aba, bb) ->
                  Some (subst_apply_term aba
                          (fun xj ->
                            the (rbt_comp_lookup
                                  (compare _A.compare_compare_order) aa xj)),
                         subst_apply_term bb
                           (fun xj ->
                             the (rbt_comp_lookup
                                   (compare _A.compare_compare_order) aa
                                   xj)))))));;

let rec rep_ta_code _B _A (Abs_ta_code x) = x;;

let rec check_comcoh_wit _A _B (_C1, _C2)
  x xa =
    ta_check_comcoh_code _A _B (_C1.compare_compare_order, _C2)
      (rep_ta_code _A _B x)
      (impl_of
        (linorder_prod
          (linorder_term _B.compare_compare_order _C1.compare_compare_order)
          (linorder_term _B.compare_compare_order _C1.compare_compare_order))
        xa);;

let rec check_comcoh_wit_ls _A _B (_C1, _C2)
  ta r =
    check_comcoh_wit _A _B (_C1, _C2) ta
      (g_from_list_dflt_basic_oops_rm_basic_ops
        (compare_order_prod
          (compare_order_term _B.compare_compare_order
            _C1.compare_compare_order)
          (compare_order_term _B.compare_compare_order
            _C1.compare_compare_order))
        r);;

let rec ta_make_ls_code _A _B
  = (fun x xa xb ->
      ta_make_code _A.compare_compare_order _A _B.compare_compare_order
        (gen_set Empty
          (fun k -> rbt_comp_insert (compare _A.compare_compare_order) k ()) x)
        (gen_set Empty
          (fun k ->
            rbt_comp_insert
              (compare_ta_rule _A.compare_compare_order
                _B.compare_compare_order)
              k ())
          xa)
        (gen_set []
          (glist_insert
            (prod_eq (comp2eq (compare_res _A.compare_compare_order))
              (comp2eq (compare_res _A.compare_compare_order))))
          xb));;

let rec make_ls _A _B x xa xb = Abs_ta_code (ta_make_ls_code _A _B x xa xb);;

let rec ta_code_make_impl _A _B
  (Tree_Automaton (fin, rs, eps)) = make_ls _A _B fin rs eps;;

let rec ta_rhs_states (_A1, _A2, _A3, _A4, _A5) (_B1, _B2)
  ta = sup_seta (_A2, _A3)
         (image ((ceq_ta_rule _A4 _B2), (ccompare_ta_rule _A3 _B1))
           (_A2, _A3, _A5) r_rhs (ta_rules ta))
         (compute_trancl (_A1, _A2, _A3, _A5)
           (image ((ceq_ta_rule _A4 _B2), (ccompare_ta_rule _A3 _B1))
             (_A2, _A3, _A5) r_rhs (ta_rules ta))
           (ta_eps ta));;

let rec initial_rel (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2, _B3)
  (_C1, _C2, _C3, _C4, _C5, _C6, _C7)
  ta r =
    (let rhs =
       ta_rhs_states (_A2, _A3, _A4.ccompare_cproper_interval, _A5, _A6)
         (_B1, _B3) ta
       in
     let matcha =
       ta_matcha (_A2, _A3, _A4.ccompare_cproper_interval, _A5, _A6) (_B1, _B3)
         (_C3, _C4.ccompare_cproper_interval) ta rhs
       in
     let analyze_rule =
       (fun (l, ra) ->
         (let _ = vars_term (_B2, _B3) (_C1, _C2, _C3, _C4, _C5, _C6, _C7) l in
           sup_setb
             ((finite_UNIV_prod _A1.finite_UNIV_card_UNIV
                (finite_UNIV_set _A1.finite_UNIV_card_UNIV)),
               (cenum_prod _A2
                 (cenum_set (_A2, _A3, _A4.ccompare_cproper_interval, _A6))),
               (ceq_prod _A3
                 (ceq_set (_A2, _A3, _A4.ccompare_cproper_interval))),
               (cproper_interval_prod _A4
                 (cproper_interval_set (_A1, _A3, _A4, _A6))),
               (set_impl_prod _A6 set_impl_set))
             (image
               ((ceq_list (ceq_prod _C3 _A3)),
                 (ccompare_list
                   (ccompare_prod _C4.ccompare_cproper_interval
                     _A4.ccompare_cproper_interval)))
               ((ceq_set
                  ((cenum_prod _A2
                     (cenum_set
                       (_A2, _A3, _A4.ccompare_cproper_interval, _A6))),
                    (ceq_prod _A3
                      (ceq_set (_A2, _A3, _A4.ccompare_cproper_interval))),
                    (cproper_interval_prod _A4
                      (cproper_interval_set
                        (_A1, _A3, _A4, _A6))).ccompare_cproper_interval)),
                 (ccompare_set
                   ((finite_UNIV_prod _A1.finite_UNIV_card_UNIV
                      (finite_UNIV_set _A1.finite_UNIV_card_UNIV)),
                     (ceq_prod _A3
                       (ceq_set (_A2, _A3, _A4.ccompare_cproper_interval))),
                     (cproper_interval_prod _A4
                       (cproper_interval_set (_A1, _A3, _A4, _A6))),
                     (set_impl_prod _A6 set_impl_set))),
                 set_impl_set)
               (fun sigma ->
                 (let qr =
                    ta_res (_A2, _A3, _A4.ccompare_cproper_interval, _A5, _A6)
                      (_B1, _B3) ta
                      (map_term (fun x -> x) (fun_of _C6 sigma) ra)
                    in
                   image (_A3, _A4.ccompare_cproper_interval)
                     ((ceq_prod _A3
                        (ceq_set (_A2, _A3, _A4.ccompare_cproper_interval))),
                       (ccompare_prod _A4.ccompare_cproper_interval
                         (ccompare_set
                           (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6))),
                       (set_impl_prod _A6 set_impl_set))
                     (fun q -> (q, qr))
                     (ta_res (_A2, _A3, _A4.ccompare_cproper_interval, _A5, _A6)
                       (_B1, _B3) ta
                       (map_term (fun x -> x) (fun_of _C6 sigma) l))))
               (matcha l))))
       in
      sup_setb
        ((finite_UNIV_prod _A1.finite_UNIV_card_UNIV
           (finite_UNIV_set _A1.finite_UNIV_card_UNIV)),
          (cenum_prod _A2
            (cenum_set (_A2, _A3, _A4.ccompare_cproper_interval, _A6))),
          (ceq_prod _A3 (ceq_set (_A2, _A3, _A4.ccompare_cproper_interval))),
          (cproper_interval_prod _A4
            (cproper_interval_set (_A1, _A3, _A4, _A6))),
          (set_impl_prod _A6 set_impl_set))
        (image
          ((ceq_prod (ceq_term _B3 _C6) (ceq_term _B3 _C6)),
            (ccompare_prod (ccompare_term _B2 _C5) (ccompare_term _B2 _C5)))
          ((ceq_set
             ((cenum_prod _A2
                (cenum_set (_A2, _A3, _A4.ccompare_cproper_interval, _A6))),
               (ceq_prod _A3
                 (ceq_set (_A2, _A3, _A4.ccompare_cproper_interval))),
               (cproper_interval_prod _A4
                 (cproper_interval_set
                   (_A1, _A3, _A4, _A6))).ccompare_cproper_interval)),
            (ccompare_set
              ((finite_UNIV_prod _A1.finite_UNIV_card_UNIV
                 (finite_UNIV_set _A1.finite_UNIV_card_UNIV)),
                (ceq_prod _A3
                  (ceq_set (_A2, _A3, _A4.ccompare_cproper_interval))),
                (cproper_interval_prod _A4
                  (cproper_interval_set (_A1, _A3, _A4, _A6))),
                (set_impl_prod _A6 set_impl_set))),
            set_impl_set)
          analyze_rule r));;

let rec initial_relation (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2, _B3)
  (_C1, _C2, _C3, _C4, _C5, _C6, _C7)
  ta r =
    (let q_qs =
       initial_rel (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2, _B3)
         (_C1, _C2, _C3, _C4, _C5, _C6, _C7) ta r
       in
      (if member
            ((ceq_set (_A2, _A3, _A4.ccompare_cproper_interval)),
              (ccompare_set (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6)))
            (bot_set (_A3, _A4.ccompare_cproper_interval, _A6))
            (image
              ((ceq_prod _A3
                 (ceq_set (_A2, _A3, _A4.ccompare_cproper_interval))),
                (ccompare_prod _A4.ccompare_cproper_interval
                  (ccompare_set (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6))))
              ((ceq_set (_A2, _A3, _A4.ccompare_cproper_interval)),
                (ccompare_set (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6)),
                set_impl_set)
              snd q_qs)
        then None
        else Some (sup_setb
                    ((finite_UNIV_prod _A1.finite_UNIV_card_UNIV
                       _A1.finite_UNIV_card_UNIV),
                      (cenum_prod _A2 _A2), (ceq_prod _A3 _A3),
                      (cproper_interval_prod _A4 _A4), (set_impl_prod _A6 _A6))
                    (image
                      ((ceq_prod _A3
                         (ceq_set (_A2, _A3, _A4.ccompare_cproper_interval))),
                        (ccompare_prod _A4.ccompare_cproper_interval
                          (ccompare_set
                            (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A6))))
                      ((ceq_set
                         ((cenum_prod _A2 _A2), (ceq_prod _A3 _A3),
                           (cproper_interval_prod _A4
                             _A4).ccompare_cproper_interval)),
                        (ccompare_set
                          ((finite_UNIV_prod _A1.finite_UNIV_card_UNIV
                             _A1.finite_UNIV_card_UNIV),
                            (ceq_prod _A3 _A3), (cproper_interval_prod _A4 _A4),
                            (set_impl_prod _A6 _A6))),
                        set_impl_set)
                      (fun (q, a) ->
                        image (_A3, _A4.ccompare_cproper_interval)
                          ((ceq_prod _A3 _A3),
                            (ccompare_prod _A4.ccompare_cproper_interval
                              _A4.ccompare_cproper_interval),
                            (set_impl_prod _A6 _A6))
                          (fun aa -> (q, aa)) a)
                      q_qs))));;

let rec decide_coherent_compatible_main (_A1, _A2, _A3, _A4, _A5, _A6)
  (_B1, _B2, _B3) (_C1, _C2, _C3, _C4, _C5, _C6, _C7)
  ta r normalizer =
    (match
      normalizer
        (initial_relation (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2, _B3)
          (_C1, _C2, _C3, _C4, _C5, _C6, _C7) ta r)
      with None -> false
      | Some rel ->
        less_eq_set (_A2, _A3, _A4.ccompare_cproper_interval)
          (imagea (_A3, _A4.ccompare_cproper_interval)
            (_A3, _A4.ccompare_cproper_interval, _A6) rel (ta_final ta))
          (ta_final ta));;

let rec coherent_rule (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2)
  rel rules (TA_rule (f, qs, q)) =
    foldr (comp (sup_seta
                  ((ceq_prod _A3 (ceq_option _A3)),
                    (ccompare_prod _A4.ccompare_cproper_interval
                      (ccompare_option _A4.ccompare_cproper_interval))))
            (fun i ->
              (let qi = nth qs i in
               let qi_s =
                 image ((ceq_prod _A3 _A3),
                         (ccompare_prod _A4.ccompare_cproper_interval
                           _A4.ccompare_cproper_interval))
                   (_A3, _A4.ccompare_cproper_interval, _A6) snd
                   (filter
                     ((ceq_prod _A3 _A3),
                       (ccompare_prod _A4.ccompare_cproper_interval
                         _A4.ccompare_cproper_interval))
                     (fun qq -> eq _A5 (fst qq) qi) rel)
                 in
               let a =
                 sup_setb
                   ((finite_UNIV_option _A1), (cenum_option _A2),
                     (ceq_option _A3), (cproper_interval_option _A4),
                     (set_impl_option _A6))
                   (image (_A3, _A4.ccompare_cproper_interval)
                     ((ceq_set
                        ((cenum_option _A2), (ceq_option _A3),
                          (cproper_interval_option
                            _A4).ccompare_cproper_interval)),
                       (ccompare_set
                         ((finite_UNIV_option _A1), (ceq_option _A3),
                           (cproper_interval_option _A4),
                           (set_impl_option _A6))),
                       set_impl_set)
                     (fun qia ->
                       (let qsa = list_update qs i qia in
                        let rls =
                          filter
                            ((ceq_ta_rule _A5 _B2),
                              (ccompare_ta_rule _A4.ccompare_cproper_interval
                                _B1))
                            (fun (TA_rule (g, qsaa, _)) ->
                              eq _B2 g f && equal_lista _A5 qsaa qsa)
                            rules
                          in
                         (if less_eq_set
                               (cenum_ta_rule, (ceq_ta_rule _A5 _B2),
                                 (ccompare_ta_rule _A4.ccompare_cproper_interval
                                   _B1))
                               rls (set_empty
                                     ((ceq_ta_rule _A5 _B2),
                                       (ccompare_ta_rule
 _A4.ccompare_cproper_interval _B1))
                                     (of_phantom set_impl_ta_rulea))
                           then inserta
                                  ((ceq_option _A3),
                                    (ccompare_option
                                      _A4.ccompare_cproper_interval))
                                  None
                                  (bot_set
                                    ((ceq_option _A3),
                                      (ccompare_option
_A4.ccompare_cproper_interval),
                                      (set_impl_option _A6)))
                           else image ((ceq_ta_rule _A5 _B2),
(ccompare_ta_rule _A4.ccompare_cproper_interval _B1))
                                  ((ceq_option _A3),
                                    (ccompare_option
                                      _A4.ccompare_cproper_interval),
                                    (set_impl_option _A6))
                                  (comp (fun a -> Some a) r_rhs) rls)))
                     qi_s)
                 in
                image ((ceq_option _A3),
                        (ccompare_option _A4.ccompare_cproper_interval))
                  ((ceq_prod _A3 (ceq_option _A3)),
                    (ccompare_prod _A4.ccompare_cproper_interval
                      (ccompare_option _A4.ccompare_cproper_interval)),
                    (set_impl_prod _A6 (set_impl_option _A6)))
                  (fun aa -> (q, aa)) a)))
      (upt zero_nata (size_list qs))
      (bot_set
        ((ceq_prod _A3 (ceq_option _A3)),
          (ccompare_prod _A4.ccompare_cproper_interval
            (ccompare_option _A4.ccompare_cproper_interval)),
          (set_impl_prod _A6 (set_impl_option _A6))));;

let rec new_states (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2)
  ta rel =
    (let rules = ta_rules ta in
      sup_setb
        ((finite_UNIV_prod _A1 (finite_UNIV_option _A1)),
          (cenum_prod _A2 (cenum_option _A2)), (ceq_prod _A3 (ceq_option _A3)),
          (cproper_interval_prod _A4 (cproper_interval_option _A4)),
          (set_impl_prod _A6 (set_impl_option _A6)))
        (image
          ((ceq_ta_rule _A5 _B2),
            (ccompare_ta_rule _A4.ccompare_cproper_interval _B1))
          ((ceq_set
             ((cenum_prod _A2 (cenum_option _A2)),
               (ceq_prod _A3 (ceq_option _A3)),
               (cproper_interval_prod _A4
                 (cproper_interval_option _A4)).ccompare_cproper_interval)),
            (ccompare_set
              ((finite_UNIV_prod _A1 (finite_UNIV_option _A1)),
                (ceq_prod _A3 (ceq_option _A3)),
                (cproper_interval_prod _A4 (cproper_interval_option _A4)),
                (set_impl_prod _A6 (set_impl_option _A6)))),
            set_impl_set)
          (coherent_rule (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2) rel rules)
          rules));;

let rec normalize_main (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2)
  ta rel accu =
    (let newa = new_states (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2) ta rel in
      (if member
            ((ceq_option _A3), (ccompare_option _A4.ccompare_cproper_interval))
            None
            (image
              ((ceq_prod _A3 (ceq_option _A3)),
                (ccompare_prod _A4.ccompare_cproper_interval
                  (ccompare_option _A4.ccompare_cproper_interval)))
              ((ceq_option _A3),
                (ccompare_option _A4.ccompare_cproper_interval),
                (set_impl_option _A6))
              snd newa)
        then None
        else (let new_rel =
                image ((ceq_prod _A3 (ceq_option _A3)),
                        (ccompare_prod _A4.ccompare_cproper_interval
                          (ccompare_option _A4.ccompare_cproper_interval)))
                  ((ceq_prod _A3 _A3),
                    (ccompare_prod _A4.ccompare_cproper_interval
                      _A4.ccompare_cproper_interval),
                    (set_impl_prod _A6 _A6))
                  (fun (x, y) -> (x, the y)) newa
                in
              let new_accu =
                sup_seta
                  ((ceq_prod _A3 _A3),
                    (ccompare_prod _A4.ccompare_cproper_interval
                      _A4.ccompare_cproper_interval))
                  accu rel
                in
              let todo =
                minus_set
                  ((ceq_prod _A3 _A3),
                    (ccompare_prod _A4.ccompare_cproper_interval
                      _A4.ccompare_cproper_interval))
                  new_rel new_accu
                in
               (if less_eq_set
                     ((cenum_prod _A2 _A2), (ceq_prod _A3 _A3),
                       (ccompare_prod _A4.ccompare_cproper_interval
                         _A4.ccompare_cproper_interval))
                     todo
                     (bot_set
                       ((ceq_prod _A3 _A3),
                         (ccompare_prod _A4.ccompare_cproper_interval
                           _A4.ccompare_cproper_interval),
                         (set_impl_prod _A6 _A6)))
                 then Some new_accu
                 else normalize_main (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2)
                        ta todo new_accu))));;

let rec normalizea (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2)
  ta x1 = match ta, x1 with
    ta, Some rel ->
      normalize_main (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2) ta rel
        (bot_set
          ((ceq_prod _A3 _A3),
            (ccompare_prod _A4.ccompare_cproper_interval
              _A4.ccompare_cproper_interval),
            (set_impl_prod _A6 _A6)))
    | ta, None -> None;;

let rec decide_coherent_compatible (_A1, _A2, _A3, _A4, _A5, _A6)
  (_B1, _B2, _B3) (_C1, _C2, _C3, _C4, _C5, _C6, _C7)
  ta r =
    decide_coherent_compatible_main (_A1, _A2, _A3, _A4, _A5, _A6)
      (_B1, _B2, _B3) (_C1, _C2, _C3, _C4, _C5, _C6, _C7) ta r
      (normalizea (_A1.finite_UNIV_card_UNIV, _A2, _A3, _A4, _A5, _A6)
        (_B1, _B3) ta);;

let rec closed_under_rewriting (_A1, _A2, _A3, _A4, _A5, _A6)
  (_B1, _B2, _B3, _B4) (_C1, _C2, _C3, _C4, _C5, _C6, _C7)
  ta r =
    decide_coherent_compatible (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2, _B4)
      (_C1, _C2, _C3, _C4, _C5, _C6, _C7)
      (trim_ta (_A1.finite_UNIV_card_UNIV, _A2, _A3, _A4, _A5, _A6)
        (_B1, _B3, _B4) ta)
      r;;

let rec ta_det_impl (f, (r, (e, (rhs, (idx, (det, x)))))) = det;;

let rec deta _A _B x = ta_det_impl (rep_ta_code _A _B x);;

let rec sorted_ps_ta (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8)
  (_B1, _B2, _B3, _B4, _B5)
  ta = Tree_Automaton
         (sorted_list_of_set
            ((ceq_list _A3), (ccompare_list _A4.ccompare_cproper_interval),
              (equal_list _A6), (linorder_list (_A6, _A7)))
            (image
              ((ceq_set (_A2, _A3, _A4.ccompare_cproper_interval)),
                (ccompare_set (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A8)))
              ((ceq_list _A3), (ccompare_list _A4.ccompare_cproper_interval),
                set_impl_list)
              (sorted_list_of_set
                (_A3, _A4.ccompare_cproper_interval, _A6, _A7))
              (ta_final
                (ps_ta (_A1, _A2, _A3, _A4, _A6, _A8) (_B1, _B2, _B4, _B5)
                  ta))),
           sorted_list_of_set
             ((ceq_ta_rule (equal_list _A6) _B4),
               (ccompare_ta_rule (ccompare_list _A4.ccompare_cproper_interval)
                 _B2),
               (equal_ta_rule (equal_list _A6) _B4),
               (linorder_ta_rule (compare_list _A5) _B3))
             (image
               ((ceq_ta_rule
                  (equal_set (_A2, _A3, _A4.ccompare_cproper_interval, _A6))
                  _B4),
                 (ccompare_ta_rule
                   (ccompare_set (_A1.finite_UNIV_card_UNIV, _A3, _A4, _A8))
                   _B2))
               ((ceq_ta_rule (equal_list _A6) _B4),
                 (ccompare_ta_rule (ccompare_list _A4.ccompare_cproper_interval)
                   _B2),
                 set_impl_ta_rule)
               (fun (TA_rule (g, qs, q)) ->
                 TA_rule
                   (g, map (sorted_list_of_set
                             (_A3, _A4.ccompare_cproper_interval, _A6, _A7))
                         qs,
                     sorted_list_of_set
                       (_A3, _A4.ccompare_cproper_interval, _A6, _A7) q))
               (ta_rules
                 (ps_ta (_A1, _A2, _A3, _A4, _A6, _A8) (_B1, _B2, _B4, _B5)
                   ta))),
           []);;

let rec tree_aut_trs_closed (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7) (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8)
  ta rel r =
    bindb (check_varcond_subset _B7 (_C6, _C8) r)
      (fun _ ->
        catch_error
          (match rel
            with Decision_Proc_Old ->
              bindb (catch_error (check_det _A6 _B5 ta)
                      (fun x ->
                        Inl (comp (showsl_lit
                                    "decision procedure requires det. TA as input\092")
                              x)))
                (fun _ ->
                  check (closed_under_rewriting (_A1, _A2, _A3, _A4, _A6, _A7)
                          (_B2, _B3.compare_compare_order, _B4, _B5)
                          (_C1, _C2, _C3, _C4, _C5.compare_compare_order, _C6,
                            _C7)
                          (ta_of_ta
                            (_A3, _A4.ccompare_cproper_interval, _A6, _A7)
                            (_B2, _B5) ta)
                          (set ((ceq_prod (ceq_term _B5 _C6)
                                  (ceq_term _B5 _C6)),
                                 (ccompare_prod
                                   (ccompare_term _B3.compare_compare_order
                                     _C5.compare_compare_order)
                                   (ccompare_term _B3.compare_compare_order
                                     _C5.compare_compare_order)),
                                 (set_impl_prod set_impl_term set_impl_term))
                            r))
                    (showsl_lit "TA is not closed under rewriting"))
            | Decision_Proc ->
              (let tc = ta_code_make_impl _A5 _B3 ta in
                (if deta _A5 _B3 tc
                  then (match check_comcoh_wit_ls _A5 _B3 (_C5, _C6) tc r
                         with None -> Inr ()
                         | Some (wl, wr) ->
                           Inl (comp (comp (comp
     (comp (showsl_lit "TA is not closed under rewriting\092")
       (showsl_terma _B7 _C8 wl))
     (showsl_lit " is accepted by TA and rewrites to\092"))
                                       (showsl_terma _B7 _C8 wr))
                                 (showsl_lit " which is not accepted by TA")))
                  else (let tca =
                          ta_code_make_impl (compare_order_list (_A5, _A6)) _B3
                            (sorted_ps_ta
                              (_A1, _A2, _A3, _A4, _A5.compare_compare_order,
                                _A6, _A5.linorder_compare_order, _A7)
                              (_B1, _B2, _B3.compare_compare_order, _B5, _B6)
                              (trim_ta
                                (_A1.finite_UNIV_card_UNIV, _A2, _A3, _A4, _A6,
                                  _A7)
                                (_B2, _B4, _B5)
                                (ta_of_ta
                                  (_A3, _A4.ccompare_cproper_interval, _A6, _A7)
                                  (_B2, _B5) ta)))
                          in
                         (match
                           check_comcoh_wit_ls (compare_order_list (_A5, _A6))
                             _B3 (_C5, _C6) tca r
                           with None -> Inr ()
                           | Some (wl, wr) ->
                             Inl (comp (comp
 (comp (comp (showsl_lit "TA is not closed under rewriting\092")
         (showsl_terma _B7 _C8 wl))
   (showsl_lit " is accepted by TA and rewrites to\092"))
 (showsl_terma _B7 _C8 wr))
                                   (showsl_lit
                                     " which is not accepted by TA"))))))
            | Id_Relation ->
              bindb (generate_ta_cond
                      (_A1, _A2, _A3, _A4.ccompare_cproper_interval, _A5, _A6,
                        _A7, _A8)
                      (_B3, _B5, _B7) ta rel)
                (fun taa ->
                  bindb (catch_error
                          (if isOK (check_left_linear_trs _B7
                                     (_C3, _C4.ccompare_cproper_interval, _C7,
                                       _C8)
                                     r)
                            then Inr () else check_det _A6 _B5 ta)
                          (fun x ->
                            Inl (comp (showsl_lit
"could not ensure left-linearity or determinism\092")
                                  x)))
                    (fun _ ->
                      catch_error
                        (state_compatible_eff_list (_A5, _A6) _B3 (_C5, _C6) taa
                          (rel_checker (_A5, _A6) rel) r)
                        (fun x ->
                          Inl (let (lr, (lrq, q)) = x in
                                comp (comp (comp
     (comp (comp (comp (comp (showsl_lit "TA is not compatible with R\092")
                         (showsl_lit "for rule "))
                   (showsl_rule _B7 _C8 lr))
             (showsl_lit "\092which is instantiated by states to "))
       (showsl_rule _B7 _A8 lrq))
     (showsl_lit "\092the state "))
                                       (showsl _A8 q))
                                  (showsl_lit
                                    " is only reachable from the lhs\092")))))
            | Some_Relation _ ->
              bindb (generate_ta_cond
                      (_A1, _A2, _A3, _A4.ccompare_cproper_interval, _A5, _A6,
                        _A7, _A8)
                      (_B3, _B5, _B7) ta rel)
                (fun taa ->
                  bindb (catch_error
                          (if isOK (check_left_linear_trs _B7
                                     (_C3, _C4.ccompare_cproper_interval, _C7,
                                       _C8)
                                     r)
                            then Inr () else check_det _A6 _B5 ta)
                          (fun x ->
                            Inl (comp (showsl_lit
"could not ensure left-linearity or determinism\092")
                                  x)))
                    (fun _ ->
                      catch_error
                        (state_compatible_eff_list (_A5, _A6) _B3 (_C5, _C6) taa
                          (rel_checker (_A5, _A6) rel) r)
                        (fun x ->
                          Inl (let (lr, (lrq, q)) = x in
                                comp (comp (comp
     (comp (comp (comp (comp (showsl_lit "TA is not compatible with R\092")
                         (showsl_lit "for rule "))
                   (showsl_rule _B7 _C8 lr))
             (showsl_lit "\092which is instantiated by states to "))
       (showsl_rule _B7 _A8 lrq))
     (showsl_lit "\092the state "))
                                       (showsl _A8 q))
                                  (showsl_lit
                                    " is only reachable from the lhs\092"))))))
          (fun x ->
            Inl (comp (comp (comp (showsl_lit
                                    "problem when ensuring (state-)compatibility of TRS with TA\092")
                              (showsl_tree_automaton _A8 _B7 ta))
                        (showsl_literal "\092"))
                  x)));;

let rec intersect_ta (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2, _B3, _B4)
  (_C1, _C2, _C3, _C4, _C5, _C6)
  tA1 tA2 =
    prod_ta (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2, _B3, _B4)
      (_C1, _C2, _C3, _C4, _C5, _C6) tA1 tA2
      (productc (_A3, _A4.ccompare_cproper_interval, _A6)
        (_C3, _C4.ccompare_cproper_interval, _C6) (ta_final tA1)
        (ta_final tA2));;

let rec ta_contains_nf (_A1, _A2, _A3, _A4, _A5, _A6)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) (_C1, _C2)
  ta r =
    not (ta_empty
          ((cenum_prod cenum_term _A2),
            (ceq_prod (ceq_term _B7 equal_unit) _A3),
            (ccompare_prod (ccompare_term _B5 compare_unit)
              _A4.ccompare_cproper_interval),
            (equal_prod (equal_term _B7 equal_unit) _A5),
            (set_impl_prod set_impl_term _A6))
          (_B4.ccompare_cproper_interval, _B6, _B7)
          (intersect_ta
            (finite_UNIV_term, cenum_term, (ceq_term _B7 equal_unit),
              (cproper_interval_term _B5 compare_unit),
              (equal_term _B7 equal_unit), set_impl_term)
            (_B3, _B4.ccompare_cproper_interval, _B7, _B8)
            (_A1, _A2, _A3, _A4, _A5, _A6)
            (ta_nf (_B1, _B2, _B3, _B4, _B5, _B7, _B8) (_C1, _C2)
              (image
                ((ceq_prod (ceq_term _B7 _C2) (ceq_term _B7 _C2)),
                  (ccompare_prod (ccompare_term _B5 _C1)
                    (ccompare_term _B5 _C1)))
                ((ceq_term _B7 _C2), (ccompare_term _B5 _C1), set_impl_term) fst
                r)
              (ta_syms (_A4.ccompare_cproper_interval, _A5)
                (_B3, _B4.ccompare_cproper_interval, _B7, _B8) ta))
            ta));;

let rec check_trs_not_wn (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8)
  r (Not_wn_ta_prf (ta, rel)) =
    bindb (check_varcond_subset _A9 (_B6, _B8) r)
      (fun _ ->
        bindb (check_left_linear_trs _A9
                (_B3, _B4.ccompare_cproper_interval, _B7, _B8) r)
          (fun _ ->
            (let tA_trim =
               trim_ta (_C1.finite_UNIV_card_UNIV, _C2, _C3, _C4, _C6, _C7)
                 (_A4.ccompare_cproper_interval, _A6, _A7)
                 (ta_of_ta (_C3, _C4.ccompare_cproper_interval, _C6, _C7)
                   (_A4.ccompare_cproper_interval, _A7) ta)
               in
              bindb (check
                      (not (ta_empty
                             (_C2, _C3, _C4.ccompare_cproper_interval, _C6, _C7)
                             (_A4.ccompare_cproper_interval, _A6, _A7) tA_trim))
                      (showsl_lit "TA is empty"))
                (fun _ ->
                  bindb (tree_aut_trs_closed
                          (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8)
                          (_A3, _A4.ccompare_cproper_interval, _A5, _A6, _A7,
                            _A8, _A9)
                          (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) ta rel r)
                    (fun _ ->
                      check (not (ta_contains_nf
                                   (_C1.finite_UNIV_card_UNIV, _C2, _C3, _C4,
                                     _C6, _C7)
                                   (_A1, _A2, _A3, _A4,
                                     _A5.compare_compare_order, _A6, _A7, _A8)
                                   (_B5.compare_compare_order, _B6) tA_trim
                                   (set ((ceq_prod (ceq_term _A7 _B6)
   (ceq_term _A7 _B6)),
  (ccompare_prod
    (ccompare_term _A5.compare_compare_order _B5.compare_compare_order)
    (ccompare_term _A5.compare_compare_order _B5.compare_compare_order)),
  (set_impl_prod set_impl_term set_impl_term))
                                     r)))
                        (showsl_lit "TA accepts some normal form"))))));;

let rec check_not_wn_ta_prf (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9)
  (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8)
  (_E1, _E2, _E3, _E4, _E5, _E6, _E7, _E8)
  i tp prf =
    (let r = rulesd i tp in
      bindb (check (null (qb i tp))
              (showsl_lit
                "strategy is unsupported for tree automata based nontermination"))
        (fun _ ->
          check_trs_not_wn (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9)
            (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8)
            (_E1, _E2, _E3, _E4, _E5, _E6, _E7, _E8) r prf));;

let rec check_dps (_A1, _A2) (_B1, _B2)
  unshp r p =
    (let d = defined_list r in
      catch_error
        (forallM
          (fun x ->
            (if (let (ll, rr) = x in
                  not (is_Var rr) &&
                    (not (membera (equal_prod _A1 equal_nat) d
                           (the (root rr))) &&
                      list_ex
                        (fun (l, ra) ->
                          equal_terma _A1 _B1 l (sharp_term unshp ll) &&
                            (equal_terma _A1 _B1 ra (sharp_term unshp rr) ||
                              supt_impl _A1 _B1 ra (sharp_term unshp rr)))
                        r))
              then Inr () else Inl x))
          p)
        (fun x -> Inl (snd x)));;

let rec unsharp = function Sharp f -> f
                  | Lab (v, va) -> Lab (v, va)
                  | FunLab (v, va) -> FunLab (v, va)
                  | UnLab v -> UnLab v;;

let rec dp_trans_nontermination_tt (_B1, _B2) (_C1, _C2) (_D1, _D2)
  i j tp (DP_trans_nontermination_tt_prf p) =
    (let r = rulesd i tp in
     let q = qb i tp in
      bindb (check (null q || not (nfsb i tp))
              (showsl_lit "strategies and normal form substitutions problem"))
        (fun _ ->
          bindb (catch_error
                  (check_dps ((equal_lab _B1 _C1), (showl_lab _B2 _C2))
                    (_D1, _D2) unsharp r p)
                  (fun x ->
                    Inl (comp (showsl_lit "problematic rule: ")
                          (showsl_rule (showl_lab _B2 _C2) _D2 x))))
            (fun _ -> Inr (mkd j false false p [] q [] r))));;

let rec switch_termination_tt (_B1, _B2, _B3)
  i joins_i tp =
    (let r = rulesd i tp in
     let q = qb i tp in
      (match
        (let cp = critical_pairs_impl _B2 r r in
          bindb (catch_error
                  (forallM
                    (fun (b, _) -> check b (showsl_lit "rules are not overlay"))
                    cp)
                  (fun x -> Inl (snd x)))
            (fun _ ->
              bindb (check_wf_trs _B3
                      ((equal_list equal_char), (showl_list showl_char)) r)
                (fun _ ->
                  bindb (check_critical_pairs (_B2, _B3) r cp joins_i)
                    (fun _ ->
                      catch_error
                        (check_NF_trs_subset (_B1, _B2)
                          ((ccompare_list ccompare_char),
                            (equal_list equal_char), mapping_impl_list)
                          r q)
                        (fun x ->
                          Inl (comp (showsl_terma _B3 (showl_list showl_char) x)
                                (showsl_lit
                                  " is not in normal form w.r.t. R")))))))
        with Inl a -> Inl a | Inr _ -> Inr (mkc i (nfsb i tp) [] r [])));;

let rec q_increase_nonterm_trs (_B1, _B2, _B3) (_C1, _C2, _C3)
  i dpp (Q_increase_nonterm_trs_prf q) =
    (let r = rulesd i dpp in
     let qa = qb i dpp in
     let nfs = nfsb i dpp in
      Inr (mkc i nfs (list_union (equal_term _B2 _C2) qa q) r []));;

let rec showsl_srs_rule _A _B
  (l, r) = comp (comp (showsl _A l) (showsl_lit " -> ")) (showsl _B r);;

let rec showsl_exp _A _B
  (f, c) = comp (comp (showsl _A f) (showsl_lit "n+ ")) (showsl _B c);;

let rec showsl_p _A _B _C _D _E
  (b, ((f, (c, m)), a)) =
    comp (comp (comp (comp (comp (showsl _A b) (showsl _D m))
                       (showsl_lit " ^ ("))
                 (showsl_exp _B _C (f, c)))
           (showsl_lit ")"))
      (showsl _E a);;

let rec showsl_pat _A _B _C _D _E _F _G _H _I _J
  (p1, p2) =
    comp (comp (showsl_p _A _B _C _D _E p1) (showsl_lit " ->+ "))
      (showsl_p _F _G _H _I _J p2);;

let rec showsl_oc _A _B
  (l, r) = comp (comp (showsl _A l) (showsl_lit " ->+ ")) (showsl _B r);;

let rec normalize_wp _A
  x = (let (l, (aa, r)) = x in
       let (f, aaa) = aa in
        (if equal_nata f zero_nata
          then (let (aab, m) = aaa in
                 (if equal_nata aab zero_nata
                   then ([], ((zero_nata, (zero_nata, [])), l @ r))
                   else normalize_wp _A
                          (l @ concat
                                 (replicate (suc (minus_nata aab one_nata)) m),
                            ((zero_nata, (zero_nata, m)), r))))
          else (if equal_nata (minus_nata f one_nata) zero_nata
                 then (let (aab, m) = aaa in
                        (if equal_nata aab zero_nata
                          then (match m
                                 with [] ->
                                   ([], ((zero_nata, (zero_nata, [])), l @ r))
                                 | a :: ma ->
                                   (match r
                                     with [] ->
                                       (l,
 ((suc zero_nata, (zero_nata, a :: ma)), []))
                                     | b :: ra ->
                                       (if eq _A a b
 then normalize_wp _A (l @ [b], ((suc zero_nata, (zero_nata, ma @ [b])), ra))
 else (l, ((suc zero_nata, (zero_nata, a :: ma)), b :: ra)))))
                          else normalize_wp _A
                                 (l @ concat
(replicate (suc (minus_nata aab one_nata)) m),
                                   ((suc zero_nata, (zero_nata, m)), r))))
                 else (let (aab, m) = aaa in
                        (if equal_nata aab zero_nata
                          then (match m
                                 with [] ->
                                   ([], ((zero_nata, (zero_nata, [])), l @ r))
                                 | vb :: vba ->
                                   normalize_wp _A
                                     (l, ((suc zero_nata,
    (zero_nata,
      concat
        (replicate (suc (suc (minus_nata (minus_nata f one_nata) one_nata)))
          (vb :: vba)))),
   r)))
                          else normalize_wp _A
                                 (l @ concat
(replicate (suc (minus_nata aab one_nata)) m),
                                   ((suc (suc
   (minus_nata (minus_nata f one_nata) one_nata)),
                                      (zero_nata, m)),
                                     r)))))));;

let rec word_pat_equiv _A
  wp1 wp2 =
    equal_proda (equal_list _A)
      (equal_prod (equal_prod equal_nat (equal_prod equal_nat (equal_list _A)))
        (equal_list _A))
      wp1 wp2 ||
      equal_proda (equal_list _A)
        (equal_prod
          (equal_prod equal_nat (equal_prod equal_nat (equal_list _A)))
          (equal_list _A))
        (normalize_wp _A wp1) (normalize_wp _A wp2);;

let rec check_stepb (_A1, _A2, _A3, _A4)
  r step =
    (match step
      with OC1 (uv, _) ->
        check (member
                ((ceq_prod (ceq_list _A1) (ceq_list _A1)),
                  (ccompare_prod (ccompare_list _A2) (ccompare_list _A2)))
                uv r)
          (comp (showsl_srs_rule (showl_list _A4) (showl_list _A4) uv)
            (showsl_lit " is not an original rule"))
      | OC2 (oc_new, oc1, oc2, t, x, l) ->
        (let (wl, tr) = oc_new in
         let (w, tx) = oc1 in
         let (xl, ra) = oc2 in
          catch_error
            (bindb
              (check (equal_lista _A3 xl (x @ l))
                (showsl_lit "problem: xl != x l"))
              (fun _ ->
                bindb (check (equal_lista _A3 tx (t @ x))
                        (showsl_lit "problem: tx != t x"))
                  (fun _ ->
                    bindb (check (equal_lista _A3 wl (w @ l))
                            (showsl_lit "problem: wl != w l"))
                      (fun _ ->
                        check (equal_lista _A3 tr (t @ ra))
                          (showsl_lit "problem: tr != t r")))))
            (fun xa ->
              Inl (comp (comp (comp (comp (comp
    (comp (comp (showsl_lit "problem in checking overlap OC2 of\092")
            (showsl_oc (showl_list _A4) (showl_list _A4) oc1))
      (showsl_literal "\092"))
    (showsl_oc (showl_list _A4) (showl_list _A4) oc2))
                                      (showsl_lit "\092to yield\092"))
                                (showsl_oc (showl_list _A4) (showl_list _A4)
                                  oc_new))
                          (showsl_literal "\092"))
                    xa)))
      | OC2p (oc_new, oc1, oc2, x, t, l) ->
        (let (lw, rt) = oc_new in
         let (w, xt) = oc1 in
         let (lx, ra) = oc2 in
          catch_error
            (bindb
              (check (equal_lista _A3 lx (l @ x))
                (showsl_lit "problem: lx != l x"))
              (fun _ ->
                bindb (check (equal_lista _A3 lw (l @ w))
                        (showsl_lit "problem: lw != l w"))
                  (fun _ ->
                    bindb (check (equal_lista _A3 rt (ra @ t))
                            (showsl_lit "problem: rt != r t"))
                      (fun _ ->
                        check (equal_lista _A3 xt (x @ t))
                          (showsl_lit "problem: xt != x t")))))
            (fun xa ->
              Inl (comp (comp (comp (comp (comp
    (comp (comp (showsl_lit "problem in checking overlap OC2\039 of\092")
            (showsl_oc (showl_list _A4) (showl_list _A4) oc1))
      (showsl_literal "\092"))
    (showsl_oc (showl_list _A4) (showl_list _A4) oc2))
                                      (showsl_lit "\092to yield\092"))
                                (showsl_oc (showl_list _A4) (showl_list _A4)
                                  oc_new))
                          (showsl_literal "\092"))
                    xa)))
      | OC3 (oc_new, oc1, oc2, t1, t2) ->
        (let (w, t1rt2) = oc_new in
         let (wa, t1xt2) = oc1 in
         let (x, ra) = oc2 in
          catch_error
            (bindb
              (check (equal_lista _A3 t1rt2 (t1 @ ra @ t2))
                (showsl_lit "problem: t1_r_t2 != t1 r t2"))
              (fun _ ->
                bindb (check (equal_lista _A3 t1xt2 (t1 @ x @ t2))
                        (showsl_lit "problem: t1_x_t2 != t1 x t2"))
                  (fun _ ->
                    check (equal_lista _A3 w wa)
                      (showsl_lit "problem: w differs"))))
            (fun xa ->
              Inl (comp (comp (comp (comp (comp
    (comp (comp (showsl_lit "problem in checking overlap OC3 of\092")
            (showsl_oc (showl_list _A4) (showl_list _A4) oc1))
      (showsl_literal "\092"))
    (showsl_oc (showl_list _A4) (showl_list _A4) oc2))
                                      (showsl_lit "\092to yield\092"))
                                (showsl_oc (showl_list _A4) (showl_list _A4)
                                  oc_new))
                          (showsl_literal "\092"))
                    xa)))
      | OC3p (oc_new, oc1, oc2, t1, t2) ->
        (let (t1wt2, ra) = oc_new in
         let (t1xt2, raa) = oc1 in
         let (w, x) = oc2 in
          catch_error
            (bindb
              (check (equal_lista _A3 t1wt2 (t1 @ w @ t2))
                (showsl_lit "problem: t1_w_t2 != t1 w t2"))
              (fun _ ->
                bindb (check (equal_lista _A3 t1xt2 (t1 @ x @ t2))
                        (showsl_lit "problem: t1_x_t2 != t1 x t2"))
                  (fun _ ->
                    check (equal_lista _A3 ra raa)
                      (showsl_lit "problem: r differs"))))
            (fun xa ->
              Inl (comp (comp (comp (comp (comp
    (comp (comp (showsl_lit "problem in checking overlap OC3\039 of\092")
            (showsl_oc (showl_list _A4) (showl_list _A4) oc1))
      (showsl_literal "\092"))
    (showsl_oc (showl_list _A4) (showl_list _A4) oc2))
                                      (showsl_lit "\092to yield\092"))
                                (showsl_oc (showl_list _A4) (showl_list _A4)
                                  oc_new))
                          (showsl_literal "\092"))
                    xa)))
      | OCDP1 (p, oc1) ->
        (let (lc, cr) = oc1 in
         let a = p in
         let (aa, b) = a in
          (let (e1, ab) = aa in
           let (ac, ba) = ab in
            (let (f, (d, l)) = ac in
              (fun c1 (c2, ad) ->
                (let (ae, bb) = ad in
                  (let (fa, (da, ra)) = ae in
                    (fun e2 ->
                      catch_error
                        (bindb
                          (check (null e1) (showsl_lit "problem: e1 not empty"))
                          (fun _ ->
                            bindb (check (null e2)
                                    (showsl_lit "problem: e2 not empty"))
                              (fun _ ->
                                bindb (check (equal_lista _A3 c1 c2)
(showsl_lit "problem: c not equal"))
                                  (fun _ ->
                                    bindb (check
    (equal_nata f one_nata &&
      (equal_nata fa one_nata &&
        (equal_nata d zero_nata && equal_nata da zero_nata)))
    (showsl_lit "problem: 0 and 1 conditions not met"))
                                      (fun _ ->
bindb (check (equal_lista _A3 lc (l @ c1)) (showsl_lit "problem: lc != l c"))
  (fun _ ->
    check (equal_lista _A3 cr (c1 @ ra)) (showsl_lit "problem: cr != c r")))))))
                        (fun x ->
                          Inl (comp (comp (comp
    (comp (comp (showsl_lit "problem in checking overlap OCDP1 of\092")
            (showsl_oc (showl_list _A4) (showl_list _A4) oc1))
      (showsl_lit "\092to yield\092"))
    (showsl_pat (showl_list _A4) showl_nat showl_nat (showl_list _A4)
      (showl_list _A4) (showl_list _A4) showl_nat showl_nat (showl_list _A4)
      (showl_list _A4) p))
                                      (showsl_literal "\092"))
                                x))))
                    bb)))
              ba)
            b)
      | OCDP2 (p, oc1) ->
        (let (cl, rc) = oc1 in
         let a = p in
         let (aa, b) = a in
          (let (c1, ab) = aa in
           let (ac, ba) = ab in
            (let (f, (d, l)) = ac in
              (fun e1 (e2, ad) ->
                (let (ae, bb) = ad in
                  (let (fa, (da, ra)) = ae in
                    (fun c2 ->
                      catch_error
                        (bindb
                          (check (null e1 && null e2)
                            (showsl_lit "problem: e1 or e2 not empty"))
                          (fun _ ->
                            bindb (check (equal_lista _A3 c1 c2)
                                    (showsl_lit "problem: c not equal"))
                              (fun _ ->
                                bindb (check
(equal_nata f one_nata &&
  (equal_nata fa one_nata &&
    (equal_nata d zero_nata && equal_nata da zero_nata)))
(showsl_lit "problem: 0 and 1 conditions not met"))
                                  (fun _ ->
                                    bindb (check (equal_lista _A3 cl (c1 @ l))
    (showsl_lit "problem: lc != l c"))
                                      (fun _ ->
check (equal_lista _A3 rc (ra @ c1)) (showsl_lit "problem: cr != c r"))))))
                        (fun x ->
                          Inl (comp (comp (comp
    (comp (comp (showsl_lit "problem in checking overlap OCDP1 of\092")
            (showsl_oc (showl_list _A4) (showl_list _A4) oc1))
      (showsl_lit "\092to yield\092"))
    (showsl_pat (showl_list _A4) showl_nat showl_nat (showl_list _A4)
      (showl_list _A4) (showl_list _A4) showl_nat showl_nat (showl_list _A4)
      (showl_list _A4) p))
                                      (showsl_literal "\092"))
                                x))))
                    bb)))
              ba)
            b)
      | WPEQ (p_new, p) ->
        (let (left, right) = p in
         let (lefta, righta) = p_new in
          catch_error
            (bindb
              (check (word_pat_equiv _A3 left lefta)
                (showsl_lit "problem: lhss are not equivalent"))
              (fun _ ->
                check (word_pat_equiv _A3 right righta)
                  (showsl_lit "problem: rhss are not equivalent")))
            (fun x ->
              Inl (comp (comp (comp (comp (comp
    (showsl_lit "problem in checking equivalence of\092")
    (showsl_pat (showl_list _A4) showl_nat showl_nat (showl_list _A4)
      (showl_list _A4) (showl_list _A4) showl_nat showl_nat (showl_list _A4)
      (showl_list _A4) p))
                                      (showsl_lit "\092and\092"))
                                (showsl_pat (showl_list _A4) showl_nat showl_nat
                                  (showl_list _A4) (showl_list _A4)
                                  (showl_list _A4) showl_nat showl_nat
                                  (showl_list _A4) (showl_list _A4) p_new))
                          (showsl_literal "\092"))
                    x)))
      | Lift (p_new, p) ->
        (let a = p in
         let (aa, b) = a in
          (let (l1, ab) = aa in
           let (ac, ba) = ab in
            (let (f1, (c1, m1)) = ac in
              (fun r1 (l2, ad) ->
                (let (ae, bb) = ad in
                  (let (f2, (c2, m2)) = ae in
                    (fun r2 ->
                      (let af = p_new in
                       let (ag, bc) = af in
                        (let (l1a, ah) = ag in
                         let (ai, bd) = ah in
                          (let (f1a, (c1a, m1a)) = ai in
                            (fun r1a (l2a, aj) ->
                              (let (ak, be) = aj in
                                (let (f2a, (c2a, m2a)) = ak in
                                  (fun r2a ->
                                    catch_error
                                      (bindb
(check (equal_lista _A3 l1 l1a && equal_lista _A3 l2 l2a)
  (showsl_lit "problem: l and l\039 do not match"))
(fun _ ->
  bindb (check (equal_lista _A3 r1 r1a && equal_lista _A3 r2 r2a)
          (showsl_lit "problem: r and r\039 do not match"))
    (fun _ ->
      bindb (check (equal_nata f1 f1a && equal_nata f2 f2a)
              (showsl_lit "problem: f and f\039 do not match"))
        (fun _ ->
          bindb (check (equal_lista _A3 m1 m1a && equal_lista _A3 m2 m2a)
                  (showsl_lit "problem: m and m\039 do not match"))
            (fun _ ->
              bindb (check (equal_nata c1a (plus_nata c1 f1))
                      (showsl_lit
                        "problem: constant factor on the left not properly increased"))
                (fun _ ->
                  check (equal_nata c2a (plus_nata c2 f2))
                    (showsl_lit
                      "problem: constant factor on the right not properly increased")))))))
                                      (fun x ->
Inl (comp (comp (comp (comp (comp (showsl_lit
                                    "problem in checking lifting of\092")
                              (showsl_pat (showl_list _A4) showl_nat showl_nat
                                (showl_list _A4) (showl_list _A4)
                                (showl_list _A4) showl_nat showl_nat
                                (showl_list _A4) (showl_list _A4) p))
                        (showsl_lit "\092to yield\092"))
                  (showsl_pat (showl_list _A4) showl_nat showl_nat
                    (showl_list _A4) (showl_list _A4) (showl_list _A4) showl_nat
                    showl_nat (showl_list _A4) (showl_list _A4) p_new))
            (showsl_literal "\092"))
      x))))
                                  be)))
                            bd)
                          bc)))
                    bb)))
              ba)
            b)
      | DPOC1_1 (p_new, p1, oc1, l, ra) ->
        (let (left, (lxr, (m2, r2))) = p1 in
         let (x, v) = oc1 in
         let (lefta, (lvr, (m2a, r2a))) = p_new in
          catch_error
            (bindb
              (check
                (equal_proda (equal_list _A3)
                  (equal_prod
                    (equal_prod equal_nat
                      (equal_prod equal_nat (equal_list _A3)))
                    (equal_list _A3))
                  left lefta)
                (showsl_lit "problem: lhss are not equal"))
              (fun _ ->
                bindb (check
                        (equal_proda equal_nat
                          (equal_prod equal_nat (equal_list _A3)) m2 m2a)
                        (showsl_lit "problem: m2 and m2\039 do not match"))
                  (fun _ ->
                    bindb (check (equal_lista _A3 r2 r2a)
                            (showsl_lit "problem: r2 and r2\039 do not match"))
                      (fun _ ->
                        bindb (check (equal_lista _A3 lxr (l @ x @ ra))
                                (showsl_lit "problem: l_x_r != l @ x @ r"))
                          (fun _ ->
                            check (equal_lista _A3 lvr (l @ v @ ra))
                              (showsl_lit "problem: l_v_r != l @ v @ r"))))))
            (fun xa ->
              Inl (comp (comp (comp (comp (comp
    (comp (comp (showsl_lit "problem in checking overlap DPOC1_1 of\092")
            (showsl_pat (showl_list _A4) showl_nat showl_nat (showl_list _A4)
              (showl_list _A4) (showl_list _A4) showl_nat showl_nat
              (showl_list _A4) (showl_list _A4) p1))
      (showsl_literal "\092"))
    (showsl_oc (showl_list _A4) (showl_list _A4) oc1))
                                      (showsl_lit "\092to yield\092"))
                                (showsl_pat (showl_list _A4) showl_nat showl_nat
                                  (showl_list _A4) (showl_list _A4)
                                  (showl_list _A4) showl_nat showl_nat
                                  (showl_list _A4) (showl_list _A4) p_new))
                          (showsl_literal "\092"))
                    xa)))
      | DPOC1_2 (p_new, p1, oc1, l, ra, x) ->
        (let a = p1 in
         let (aa, b) = a in
          (let (l1, (m1, r1)) = aa in
            (fun (xr, (m2, r2)) ->
              (let (lx, v) = oc1 in
               let ab = p_new in
               let (ac, ba) = ab in
                (let (ll1, (m1a, r1a)) = ac in
                  (fun (vr, (m2a, r2a)) ->
                    catch_error
                      (bindb
                        (check
                          (equal_proda equal_nat
                             (equal_prod equal_nat (equal_list _A3)) m1 m1a &&
                            equal_proda equal_nat
                              (equal_prod equal_nat (equal_list _A3)) m2 m2a)
                          (showsl_lit "problem: m components modified"))
                        (fun _ ->
                          bindb (check
                                  (equal_lista _A3 r1 r1a &&
                                    equal_lista _A3 r2 r2a)
                                  (showsl_lit "problem: r components modified"))
                            (fun _ ->
                              bindb (check (equal_lista _A3 ll1 (l @ l1))
                                      (showsl_lit "problem: l_l1 != l @ ll1"))
                                (fun _ ->
                                  bindb (check (equal_lista _A3 xr (x @ ra))
  (showsl_lit "problem: x_r != x @ r"))
                                    (fun _ ->
                                      bindb
(check (equal_lista _A3 lx (l @ x)) (showsl_lit "problem: l_x != l @ x"))
(fun _ ->
  check (equal_lista _A3 vr (v @ ra)) (showsl_lit "problem: v_r != v @ r")))))))
                      (fun xa ->
                        Inl (comp (comp (comp
  (comp (comp (comp (comp (showsl_lit
                            "problem in checking overlap DPOC1_2 of\092")
                      (showsl_pat (showl_list _A4) showl_nat showl_nat
                        (showl_list _A4) (showl_list _A4) (showl_list _A4)
                        showl_nat showl_nat (showl_list _A4) (showl_list _A4)
                        p1))
                (showsl_literal "\092"))
          (showsl_oc (showl_list _A4) (showl_list _A4) oc1))
    (showsl_lit "\092to yield\092"))
  (showsl_pat (showl_list _A4) showl_nat showl_nat (showl_list _A4)
    (showl_list _A4) (showl_list _A4) showl_nat showl_nat (showl_list _A4)
    (showl_list _A4) p_new))
                                    (showsl_literal "\092"))
                              xa))))
                  ba)))
            b)
      | DPOC2 (p_new, p1, oc1, l, ra) ->
        (let a = p1 in
         let (left, aa) = a in
         let (l2, ab) = aa in
         let (ac, b) = ab in
          (let (f2, (c2, lxr)) = ac in
            (fun r2 ->
              (let (x, v) = oc1 in
               let ad = p_new in
               let (lefta, ae) = ad in
               let (l2a, af) = ae in
               let (ag, ba) = af in
                (let (f2a, (c2a, lvr)) = ag in
                  (fun r2a ->
                    catch_error
                      (bindb
                        (check
                          (equal_proda (equal_list _A3)
                            (equal_prod
                              (equal_prod equal_nat
                                (equal_prod equal_nat (equal_list _A3)))
                              (equal_list _A3))
                            left lefta)
                          (showsl_lit "problem: left components modified"))
                        (fun _ ->
                          bindb (check (equal_nata f2 f2a)
                                  (showsl_lit "problem: f components modified"))
                            (fun _ ->
                              bindb (check (equal_nata c2 c2a)
                                      (showsl_lit
"problem: c components modified"))
                                (fun _ ->
                                  bindb (check (equal_lista _A3 l2 l2a)
  (showsl_lit "problem: l components modified"))
                                    (fun _ ->
                                      bindb
(check (equal_lista _A3 r2 r2a) (showsl_lit "problem: r components modified"))
(fun _ ->
  bindb (check (equal_lista _A3 lxr (l @ x @ ra))
          (showsl_lit "problem: l_x_r != l @ x @ r"))
    (fun _ ->
      check (equal_lista _A3 lvr (l @ v @ ra))
        (showsl_lit "problem: l_v_r != l @ v @ r"))))))))
                      (fun xa ->
                        Inl (comp (comp (comp
  (comp (comp (comp (comp (showsl_lit
                            "problem in checking overlap DPOC2 of\092")
                      (showsl_pat (showl_list _A4) showl_nat showl_nat
                        (showl_list _A4) (showl_list _A4) (showl_list _A4)
                        showl_nat showl_nat (showl_list _A4) (showl_list _A4)
                        p1))
                (showsl_literal "\092"))
          (showsl_oc (showl_list _A4) (showl_list _A4) oc1))
    (showsl_lit "\092to yield\092"))
  (showsl_pat (showl_list _A4) showl_nat showl_nat (showl_list _A4)
    (showl_list _A4) (showl_list _A4) showl_nat showl_nat (showl_list _A4)
    (showl_list _A4) p_new))
                                    (showsl_literal "\092"))
                              xa))))
                  ba)))
            b)
      | DPOC3_1 (p_new, p1, oc1, l, ra) ->
        (let (left, (l2, (m2, lxr))) = p1 in
         let (x, v) = oc1 in
         let (lefta, (l2a, (m2a, lvr))) = p_new in
          catch_error
            (bindb
              (check
                (equal_proda (equal_list _A3)
                  (equal_prod
                    (equal_prod equal_nat
                      (equal_prod equal_nat (equal_list _A3)))
                    (equal_list _A3))
                  left lefta)
                (showsl_lit "problem: left components modified"))
              (fun _ ->
                bindb (check
                        (equal_proda equal_nat
                          (equal_prod equal_nat (equal_list _A3)) m2 m2a)
                        (showsl_lit "problem: m components modified"))
                  (fun _ ->
                    bindb (check (equal_lista _A3 l2 l2a)
                            (showsl_lit "problem: l components modified"))
                      (fun _ ->
                        bindb (check (equal_lista _A3 lxr (l @ x @ ra))
                                (showsl_lit "problem: l_x_r != l @ x @ r"))
                          (fun _ ->
                            check (equal_lista _A3 lvr (l @ v @ ra))
                              (showsl_lit "problem: l_v_r != l @ v @ r"))))))
            (fun xa ->
              Inl (comp (comp (comp (comp (comp
    (comp (comp (showsl_lit "problem in checking overlap DPOC3_1 of\092")
            (showsl_pat (showl_list _A4) showl_nat showl_nat (showl_list _A4)
              (showl_list _A4) (showl_list _A4) showl_nat showl_nat
              (showl_list _A4) (showl_list _A4) p1))
      (showsl_literal "\092"))
    (showsl_oc (showl_list _A4) (showl_list _A4) oc1))
                                      (showsl_lit "\092to yield\092"))
                                (showsl_pat (showl_list _A4) showl_nat showl_nat
                                  (showl_list _A4) (showl_list _A4)
                                  (showl_list _A4) showl_nat showl_nat
                                  (showl_list _A4) (showl_list _A4) p_new))
                          (showsl_literal "\092"))
                    xa)))
      | DPOC3_2 (p_new, p1, oc1, l, ra, x) ->
        (let a = p1 in
         let (aa, b) = a in
          (let (l1, (m1, r1)) = aa in
            (fun (l2, (m2, lx)) ->
              (let (xr, v) = oc1 in
               let ab = p_new in
               let (ac, ba) = ab in
                (let (l1a, (m1a, r1r)) = ac in
                  (fun (l2a, (m2a, lv)) ->
                    catch_error
                      (bindb
                        (check
                          (equal_proda equal_nat
                             (equal_prod equal_nat (equal_list _A3)) m2 m2a &&
                            equal_proda equal_nat
                              (equal_prod equal_nat (equal_list _A3)) m1 m1a)
                          (showsl_lit "problem: m components modified"))
                        (fun _ ->
                          bindb (check (equal_lista _A3 l1 l1a)
                                  (showsl_lit "problem: l components modified"))
                            (fun _ ->
                              bindb (check (equal_lista _A3 l2 l2a)
                                      (showsl_lit
"problem: l2 components modified"))
                                (fun _ ->
                                  bindb (check (equal_lista _A3 lx (l @ x))
  (showsl_lit "problem: l_x != l @ x"))
                                    (fun _ ->
                                      bindb
(check (equal_lista _A3 xr (x @ ra)) (showsl_lit "problem: x_r != x @ r"))
(fun _ ->
  bindb (check (equal_lista _A3 r1r (r1 @ ra))
          (showsl_lit "problem: r1_r != r1 @ r"))
    (fun _ ->
      check (equal_lista _A3 lv (l @ v))
        (showsl_lit "problem: l_v != l @ v"))))))))
                      (fun xa ->
                        Inl (comp (comp (comp
  (comp (comp (comp (comp (showsl_lit
                            "problem in checking overlap DPOC3_2 of\092")
                      (showsl_pat (showl_list _A4) showl_nat showl_nat
                        (showl_list _A4) (showl_list _A4) (showl_list _A4)
                        showl_nat showl_nat (showl_list _A4) (showl_list _A4)
                        p1))
                (showsl_literal "\092"))
          (showsl_oc (showl_list _A4) (showl_list _A4) oc1))
    (showsl_lit "\092to yield\092"))
  (showsl_pat (showl_list _A4) showl_nat showl_nat (showl_list _A4)
    (showl_list _A4) (showl_list _A4) showl_nat showl_nat (showl_list _A4)
    (showl_list _A4) p_new))
                                    (showsl_literal "\092"))
                              xa))))
                  ba)))
            b)
      | DPDP1_1 (p_new, p1, p2, l, ra) ->
        (let (left, (ll2, (mm, r2r))) = p1 in
         let a = p2 in
         let (aa, b) = a in
          (let (l2, (mma, r2)) = aa in
            (fun (l2a, (mm2, r2a)) ->
              (let (lefta, (ll2a, (mm2a, r2pr))) = p_new in
                catch_error
                  (bindb
                    (check
                      (equal_proda (equal_list _A3)
                        (equal_prod
                          (equal_prod equal_nat
                            (equal_prod equal_nat (equal_list _A3)))
                          (equal_list _A3))
                        left lefta)
                      (showsl_lit "problem: left components modified"))
                    (fun _ ->
                      bindb (check
                              (equal_proda equal_nat
                                (equal_prod equal_nat (equal_list _A3)) mm mma)
                              (showsl_lit "problem: mm components modified"))
                        (fun _ ->
                          bindb (check
                                  (equal_proda equal_nat
                                    (equal_prod equal_nat (equal_list _A3)) mm2
                                    mm2a)
                                  (showsl_lit
                                    "problem: mm2 components modified"))
                            (fun _ ->
                              bindb (check (equal_lista _A3 ll2 (l @ l2))
                                      (showsl_lit "problem: l_l2 != l @ l2"))
                                (fun _ ->
                                  bindb (check (equal_lista _A3 r2r (r2 @ ra))
  (showsl_lit "problem: r2_r != r2 @ r"))
                                    (fun _ ->
                                      bindb
(check (equal_lista _A3 r2pr (r2a @ ra))
  (showsl_lit "problem: r2pr != r2\039 @ r"))
(fun _ ->
  check (equal_lista _A3 ll2a (l @ l2a))
    (showsl_lit "problem: l_l2 != l @ l2 "))))))))
                  (fun x ->
                    Inl (comp (comp (comp (comp
    (comp (comp (comp (showsl_lit "problem in checking overlap DPDP1_1 of\092")
                  (showsl_pat (showl_list _A4) showl_nat showl_nat
                    (showl_list _A4) (showl_list _A4) (showl_list _A4) showl_nat
                    showl_nat (showl_list _A4) (showl_list _A4) p1))
            (showsl_literal "\092"))
      (showsl_pat (showl_list _A4) showl_nat showl_nat (showl_list _A4)
        (showl_list _A4) (showl_list _A4) showl_nat showl_nat (showl_list _A4)
        (showl_list _A4) p2))
    (showsl_lit "\092to yield\092"))
                                      (showsl_pat (showl_list _A4) showl_nat
showl_nat (showl_list _A4) (showl_list _A4) (showl_list _A4) showl_nat showl_nat
(showl_list _A4) (showl_list _A4) p_new))
                                (showsl_literal "\092"))
                          x)))))
            b)
      | DPDP1_2 (p_new, p1, p2, l, ra) ->
        (let a = p1 in
         let (aa, b) = a in
          (let (l1, (mm1, r1)) = aa in
            (fun (ll2, (mm, r1a)) ->
              (let ab = p2 in
               let (ac, ba) = ab in
                (let (l2, (mm2, r1r)) = ac in
                  (fun (l2a, (mm2a, r2)) ->
                    (let ad = p_new in
                     let (ae, bb) = ad in
                      (let (l3, (mm3, r1pr)) = ae in
                        (fun (ll2a, (mm3a, r3)) ->
                          catch_error
                            (bindb
                              (check (equal_lista _A3 l1 l3)
                                (showsl_lit
                                  "problem: l1\039 components modified"))
                              (fun _ ->
                                bindb (check
(equal_proda equal_nat (equal_prod equal_nat (equal_list _A3)) mm1 mm3)
(showsl_lit "problem: mm1\039 components modified"))
                                  (fun _ ->
                                    bindb (check
    (equal_proda equal_nat (equal_prod equal_nat (equal_list _A3)) mm mm2)
    (showsl_lit "problem: mm components modified"))
                                      (fun _ ->
bindb (check
        (equal_proda equal_nat (equal_prod equal_nat (equal_list _A3)) mm2a
          mm3a)
        (showsl_lit "problem: mm2\039 components modified"))
  (fun _ ->
    bindb (check (equal_lista _A3 r2 r3)
            (showsl_lit "problem: r2\039 components modified"))
      (fun _ ->
        bindb (check (equal_lista _A3 ll2 (l @ l2))
                (showsl_lit "problem: l_l2 != l @ l2"))
          (fun _ ->
            bindb (check (equal_lista _A3 r1pr (r1 @ ra))
                    (showsl_lit "problem: r1\039r != r1\039 @ r"))
              (fun _ ->
                bindb (check (equal_lista _A3 r1r (r1a @ ra))
                        (showsl_lit "problem: r1r != r1 @ r"))
                  (fun _ ->
                    check (equal_lista _A3 ll2a (l @ l2a))
                      (showsl_lit "problem: l_l2 != l @ l2 "))))))))))
                            (fun x ->
                              Inl (comp (comp
  (comp (comp (comp (comp (comp (showsl_lit
                                  "problem in checking overlap DPDP1_2 of\092")
                            (showsl_pat (showl_list _A4) showl_nat showl_nat
                              (showl_list _A4) (showl_list _A4) (showl_list _A4)
                              showl_nat showl_nat (showl_list _A4)
                              (showl_list _A4) p1))
                      (showsl_literal "\092"))
                (showsl_pat (showl_list _A4) showl_nat showl_nat
                  (showl_list _A4) (showl_list _A4) (showl_list _A4) showl_nat
                  showl_nat (showl_list _A4) (showl_list _A4) p2))
          (showsl_lit "\092to yield\092"))
    (showsl_pat (showl_list _A4) showl_nat showl_nat (showl_list _A4)
      (showl_list _A4) (showl_list _A4) showl_nat showl_nat (showl_list _A4)
      (showl_list _A4) p_new))
  (showsl_literal "\092"))
                                    x))))
                        bb)))
                  ba)))
            b)
      | DPDP2_1 (p_new, p1, p2, l, ra) ->
        (let a = p1 in
         let (aa, b) = a in
          (let (l1, (mm1, r1)) = aa in
            (fun (l1a, (mm, r2r)) ->
              (let ab = p2 in
               let (ac, ba) = ab in
                (let (ll1, (mm2, r2)) = ac in
                  (fun (l2, (mm2a, r2a)) ->
                    (let ad = p_new in
                     let (ae, bb) = ad in
                      (let (ll1a, (mm3, r3)) = ae in
                        (fun (l3, (mm3a, r2pr)) ->
                          catch_error
                            (bindb
                              (check (equal_lista _A3 r2r (r2 @ ra))
                                (showsl_lit "problem: r2r != r2 @ r"))
                              (fun _ ->
                                bindb (check (equal_lista _A3 ll1 (l @ l1a))
(showsl_lit "problem: l_l1 != l @ l1"))
                                  (fun _ ->
                                    bindb (check (equal_lista _A3 ll1a (l @ l1))
    (showsl_lit "problem: l_l1\039 != l @ l1\039 "))
                                      (fun _ ->
bindb (check
        (equal_proda equal_nat (equal_prod equal_nat (equal_list _A3)) mm1 mm3)
        (showsl_lit "problem: mm1\039 component modified"))
  (fun _ ->
    bindb (check (equal_lista _A3 r1 r3)
            (showsl_lit "problem: r1\039 component modified"))
      (fun _ ->
        bindb (check (equal_lista _A3 l2 l3)
                (showsl_lit "problem: l2\039 component modified"))
          (fun _ ->
            bindb (check
                    (equal_proda equal_nat
                      (equal_prod equal_nat (equal_list _A3)) mm2a mm3a)
                    (showsl_lit "problem: mm2\039 components modified"))
              (fun _ ->
                bindb (check
                        (equal_proda equal_nat
                          (equal_prod equal_nat (equal_list _A3)) mm mm2)
                        (showsl_lit "problem: mm components modified"))
                  (fun _ ->
                    check (equal_lista _A3 r2pr (r2a @ ra))
                      (showsl_lit "problem: r2\039r != r2\039 @ r"))))))))))
                            (fun x ->
                              Inl (comp (comp
  (comp (comp (comp (comp (comp (showsl_lit
                                  "problem in checking overlap DPDP2_1 of\092")
                            (showsl_pat (showl_list _A4) showl_nat showl_nat
                              (showl_list _A4) (showl_list _A4) (showl_list _A4)
                              showl_nat showl_nat (showl_list _A4)
                              (showl_list _A4) p1))
                      (showsl_literal "\092"))
                (showsl_pat (showl_list _A4) showl_nat showl_nat
                  (showl_list _A4) (showl_list _A4) (showl_list _A4) showl_nat
                  showl_nat (showl_list _A4) (showl_list _A4) p2))
          (showsl_lit "\092to yield\092"))
    (showsl_pat (showl_list _A4) showl_nat showl_nat (showl_list _A4)
      (showl_list _A4) (showl_list _A4) showl_nat showl_nat (showl_list _A4)
      (showl_list _A4) p_new))
  (showsl_literal "\092"))
                                    x))))
                        bb)))
                  ba)))
            b)
      | DPDP2_2 (p_new, p1, p2, l, ra) ->
        (let a = p1 in
         let (aa, b) = a in
          (let (l1, (mm1, r1)) = aa in
            (fun (l1a, (mm1a, r1a)) ->
              (let ab = p2 in
               let (ac, ba) = ab in
                (let (ll1, (mm2, r1r)) = ac in
                  (fun right ->
                    (let ad = p_new in
                     let (ae, bb) = ad in
                      (let (ll1a, (mm3, r1pr)) = ae in
                        (fun righta ->
                          catch_error
                            (bindb
                              (check (equal_lista _A3 r1r (r1a @ ra))
                                (showsl_lit "problem: r1r != r1 @ r"))
                              (fun _ ->
                                bindb (check (equal_lista _A3 ll1 (l @ l1a))
(showsl_lit "problem: l_l1 != l @ l1"))
                                  (fun _ ->
                                    bindb (check (equal_lista _A3 ll1a (l @ l1))
    (showsl_lit "problem: l_l1\039 != l @ l1\039 "))
                                      (fun _ ->
bindb (check (equal_lista _A3 r1pr (r1 @ ra))
        (showsl_lit "problem: r1\039_r != r1\039 @ r "))
  (fun _ ->
    bindb (check
            (equal_proda equal_nat (equal_prod equal_nat (equal_list _A3)) mm1
              mm3)
            (showsl_lit "problem: mm1\039 component modified"))
      (fun _ ->
        bindb (check
                (equal_proda equal_nat (equal_prod equal_nat (equal_list _A3))
                  mm1a mm2)
                (showsl_lit "problem: mm components modified"))
          (fun _ ->
            check (equal_proda (equal_list _A3)
                    (equal_prod
                      (equal_prod equal_nat
                        (equal_prod equal_nat (equal_list _A3)))
                      (equal_list _A3))
                    right righta)
              (showsl_lit "problem: right components modified"))))))))
                            (fun x ->
                              Inl (comp (comp
  (comp (comp (comp (comp (comp (showsl_lit
                                  "problem in checking overlap DPDP2_ of\092")
                            (showsl_pat (showl_list _A4) showl_nat showl_nat
                              (showl_list _A4) (showl_list _A4) (showl_list _A4)
                              showl_nat showl_nat (showl_list _A4)
                              (showl_list _A4) p1))
                      (showsl_literal "\092"))
                (showsl_pat (showl_list _A4) showl_nat showl_nat
                  (showl_list _A4) (showl_list _A4) (showl_list _A4) showl_nat
                  showl_nat (showl_list _A4) (showl_list _A4) p2))
          (showsl_lit "\092to yield\092"))
    (showsl_pat (showl_list _A4) showl_nat showl_nat (showl_list _A4)
      (showl_list _A4) (showl_list _A4) showl_nat showl_nat (showl_list _A4)
      (showl_list _A4) p_new))
  (showsl_literal "\092"))
                                    x))))
                        bb)))
                  ba)))
            b));;

let rec prems_ofa
  step =
    (match step with OC1 (_, _) -> []
      | OC2 (_, p, pa, _, _, _) ->
        [(([], ((zero_nata, (zero_nata, [])), fst p)),
           ([], ((zero_nata, (zero_nata, [])), snd p)));
          (([], ((zero_nata, (zero_nata, [])), fst pa)),
            ([], ((zero_nata, (zero_nata, [])), snd pa)))]
      | OC2p (_, p, pa, _, _, _) ->
        [(([], ((zero_nata, (zero_nata, [])), fst p)),
           ([], ((zero_nata, (zero_nata, [])), snd p)));
          (([], ((zero_nata, (zero_nata, [])), fst pa)),
            ([], ((zero_nata, (zero_nata, [])), snd pa)))]
      | OC3 (_, p, pa, _, _) ->
        [(([], ((zero_nata, (zero_nata, [])), fst p)),
           ([], ((zero_nata, (zero_nata, [])), snd p)));
          (([], ((zero_nata, (zero_nata, [])), fst pa)),
            ([], ((zero_nata, (zero_nata, [])), snd pa)))]
      | OC3p (_, p, pa, _, _) ->
        [(([], ((zero_nata, (zero_nata, [])), fst p)),
           ([], ((zero_nata, (zero_nata, [])), snd p)));
          (([], ((zero_nata, (zero_nata, [])), fst pa)),
            ([], ((zero_nata, (zero_nata, [])), snd pa)))]
      | OCDP1 (_, p) ->
        [(([], ((zero_nata, (zero_nata, [])), fst p)),
           ([], ((zero_nata, (zero_nata, [])), snd p)))]
      | OCDP2 (_, p) ->
        [(([], ((zero_nata, (zero_nata, [])), fst p)),
           ([], ((zero_nata, (zero_nata, [])), snd p)))]
      | WPEQ (_, p) -> [p] | Lift (_, p) -> [p]
      | DPOC1_1 (_, p, rl, _, _) ->
        [p; (([], ((zero_nata, (zero_nata, [])), fst rl)),
              ([], ((zero_nata, (zero_nata, [])), snd rl)))]
      | DPOC1_2 (_, p, rl, _, _, _) ->
        [p; (([], ((zero_nata, (zero_nata, [])), fst rl)),
              ([], ((zero_nata, (zero_nata, [])), snd rl)))]
      | DPOC2 (_, p, rl, _, _) ->
        [p; (([], ((zero_nata, (zero_nata, [])), fst rl)),
              ([], ((zero_nata, (zero_nata, [])), snd rl)))]
      | DPOC3_1 (_, p, rl, _, _) ->
        [p; (([], ((zero_nata, (zero_nata, [])), fst rl)),
              ([], ((zero_nata, (zero_nata, [])), snd rl)))]
      | DPOC3_2 (_, p, rl, _, _, _) ->
        [p; (([], ((zero_nata, (zero_nata, [])), fst rl)),
              ([], ((zero_nata, (zero_nata, [])), snd rl)))]
      | DPDP1_1 (_, p1, p2, _, _) -> [p1; p2]
      | DPDP1_2 (_, p1, p2, _, _) -> [p1; p2]
      | DPDP2_1 (_, p1, p2, _, _) -> [p1; p2]
      | DPDP2_2 (_, p1, p2, _, _) -> [p1; p2]);;

let rec check_proof (_A1, _A2, _A3, _A4)
  r delta x2 = match r, delta, x2 with
    r, delta, p :: ps ->
      bindb (check_stepb (_A1, _A2, _A3, _A4) r p)
        (fun _ ->
          bindb (catch_error
                  (forallM
                    (fun pa ->
                      check (member
                              ((ceq_prod
                                 (ceq_prod (ceq_list _A1)
                                   (ceq_prod
                                     (ceq_prod ceq_nat
                                       (ceq_prod ceq_nat (ceq_list _A1)))
                                     (ceq_list _A1)))
                                 (ceq_prod (ceq_list _A1)
                                   (ceq_prod
                                     (ceq_prod ceq_nat
                                       (ceq_prod ceq_nat (ceq_list _A1)))
                                     (ceq_list _A1)))),
                                (ccompare_prod
                                  (ccompare_prod (ccompare_list _A2)
                                    (ccompare_prod
                                      (ccompare_prod ccompare_nat
(ccompare_prod ccompare_nat (ccompare_list _A2)))
                                      (ccompare_list _A2)))
                                  (ccompare_prod (ccompare_list _A2)
                                    (ccompare_prod
                                      (ccompare_prod ccompare_nat
(ccompare_prod ccompare_nat (ccompare_list _A2)))
                                      (ccompare_list _A2)))))
                              pa delta)
                        (comp (showsl_lit
                                "problem: nothing known about premise ")
                          (showsl_pat (showl_list _A4) showl_nat showl_nat
                            (showl_list _A4) (showl_list _A4) (showl_list _A4)
                            showl_nat showl_nat (showl_list _A4)
                            (showl_list _A4) pa)))
                    (prems_ofa p))
                  (fun x -> Inl (snd x)))
            (fun _ ->
              check_proof (_A1, _A2, _A3, _A4) r
                (sup_seta
                  ((ceq_prod
                     (ceq_prod (ceq_list _A1)
                       (ceq_prod
                         (ceq_prod ceq_nat (ceq_prod ceq_nat (ceq_list _A1)))
                         (ceq_list _A1)))
                     (ceq_prod (ceq_list _A1)
                       (ceq_prod
                         (ceq_prod ceq_nat (ceq_prod ceq_nat (ceq_list _A1)))
                         (ceq_list _A1)))),
                    (ccompare_prod
                      (ccompare_prod (ccompare_list _A2)
                        (ccompare_prod
                          (ccompare_prod ccompare_nat
                            (ccompare_prod ccompare_nat (ccompare_list _A2)))
                          (ccompare_list _A2)))
                      (ccompare_prod (ccompare_list _A2)
                        (ccompare_prod
                          (ccompare_prod ccompare_nat
                            (ccompare_prod ccompare_nat (ccompare_list _A2)))
                          (ccompare_list _A2)))))
                  delta
                  (inserta
                    ((ceq_prod
                       (ceq_prod (ceq_list _A1)
                         (ceq_prod
                           (ceq_prod ceq_nat (ceq_prod ceq_nat (ceq_list _A1)))
                           (ceq_list _A1)))
                       (ceq_prod (ceq_list _A1)
                         (ceq_prod
                           (ceq_prod ceq_nat (ceq_prod ceq_nat (ceq_list _A1)))
                           (ceq_list _A1)))),
                      (ccompare_prod
                        (ccompare_prod (ccompare_list _A2)
                          (ccompare_prod
                            (ccompare_prod ccompare_nat
                              (ccompare_prod ccompare_nat (ccompare_list _A2)))
                            (ccompare_list _A2)))
                        (ccompare_prod (ccompare_list _A2)
                          (ccompare_prod
                            (ccompare_prod ccompare_nat
                              (ccompare_prod ccompare_nat (ccompare_list _A2)))
                            (ccompare_list _A2)))))
                    (pat_of p)
                    (set_empty
                      ((ceq_prod
                         (ceq_prod (ceq_list _A1)
                           (ceq_prod
                             (ceq_prod ceq_nat
                               (ceq_prod ceq_nat (ceq_list _A1)))
                             (ceq_list _A1)))
                         (ceq_prod (ceq_list _A1)
                           (ceq_prod
                             (ceq_prod ceq_nat
                               (ceq_prod ceq_nat (ceq_list _A1)))
                             (ceq_list _A1)))),
                        (ccompare_prod
                          (ccompare_prod (ccompare_list _A2)
                            (ccompare_prod
                              (ccompare_prod ccompare_nat
                                (ccompare_prod ccompare_nat
                                  (ccompare_list _A2)))
                              (ccompare_list _A2)))
                          (ccompare_prod (ccompare_list _A2)
                            (ccompare_prod
                              (ccompare_prod ccompare_nat
                                (ccompare_prod ccompare_nat
                                  (ccompare_list _A2)))
                              (ccompare_list _A2)))))
                      (of_phantom
                        (set_impl_proda
                          (set_impl_prod set_impl_list
                            (set_impl_prod
                              (set_impl_prod set_impl_nat
                                (set_impl_prod set_impl_nat set_impl_list))
                              set_impl_list))
                          (set_impl_prod set_impl_list
                            (set_impl_prod
                              (set_impl_prod set_impl_nat
                                (set_impl_prod set_impl_nat set_impl_list))
                              set_impl_list)))))))
                ps))
    | r, delta, [] -> Inr ();;

let rec check_non_loop_srs_proof (_A1, _A2, _A3, _A4)
  ra x1 = match ra, x1 with
    ra, SE_OC ((m, lmr), l, r, steps) ->
      bindb (check
              (membera
                (equal_prod
                  (equal_prod (equal_list _A3)
                    (equal_prod
                      (equal_prod equal_nat
                        (equal_prod equal_nat (equal_list _A3)))
                      (equal_list _A3)))
                  (equal_prod (equal_list _A3)
                    (equal_prod
                      (equal_prod equal_nat
                        (equal_prod equal_nat (equal_list _A3)))
                      (equal_list _A3))))
                (map pat_of steps)
                (([], ((zero_nata, (zero_nata, [])), m)),
                  ([], ((zero_nata, (zero_nata, [])), lmr))))
              (showsl_lit "overlap closure not derived within proof"))
        (fun _ ->
          bindb (check (equal_lista _A3 lmr (l @ m @ r))
                  (showsl_lit "no selfoverlap"))
            (fun _ ->
              check_proof (_A1, _A2, _A3, _A4) ra
                (set_empty
                  ((ceq_prod
                     (ceq_prod (ceq_list _A1)
                       (ceq_prod
                         (ceq_prod ceq_nat (ceq_prod ceq_nat (ceq_list _A1)))
                         (ceq_list _A1)))
                     (ceq_prod (ceq_list _A1)
                       (ceq_prod
                         (ceq_prod ceq_nat (ceq_prod ceq_nat (ceq_list _A1)))
                         (ceq_list _A1)))),
                    (ccompare_prod
                      (ccompare_prod (ccompare_list _A2)
                        (ccompare_prod
                          (ccompare_prod ccompare_nat
                            (ccompare_prod ccompare_nat (ccompare_list _A2)))
                          (ccompare_list _A2)))
                      (ccompare_prod (ccompare_list _A2)
                        (ccompare_prod
                          (ccompare_prod ccompare_nat
                            (ccompare_prod ccompare_nat (ccompare_list _A2)))
                          (ccompare_list _A2)))))
                  (of_phantom
                    (set_impl_proda
                      (set_impl_prod set_impl_list
                        (set_impl_prod
                          (set_impl_prod set_impl_nat
                            (set_impl_prod set_impl_nat set_impl_list))
                          set_impl_list))
                      (set_impl_prod set_impl_list
                        (set_impl_prod
                          (set_impl_prod set_impl_nat
                            (set_impl_prod set_impl_nat set_impl_list))
                          set_impl_list)))))
                steps))
    | ra, SE_DP ((left, right), l, r, steps) ->
        bindb (check
                (membera
                  (equal_prod
                    (equal_prod (equal_list _A3)
                      (equal_prod
                        (equal_prod equal_nat
                          (equal_prod equal_nat (equal_list _A3)))
                        (equal_list _A3)))
                    (equal_prod (equal_list _A3)
                      (equal_prod
                        (equal_prod equal_nat
                          (equal_prod equal_nat (equal_list _A3)))
                        (equal_list _A3))))
                  (map pat_of steps) (left, right))
                (showsl_lit "overlap closure not derived within proof"))
          (fun _ ->
            (let a = left in
             let (l1, aa) = a in
             let (ab, b) = aa in
              (let (f1, (c1, m1)) = ab in
                (fun r1 ->
                  (let ac = right in
                   let (l2, ad) = ac in
                   let (ae, ba) = ad in
                    (let (f2, (c2, m2)) = ae in
                      (fun r2 ->
                        bindb (check
                                (equal_lista _A3 m1 m2 &&
                                  (equal_lista _A3 l2 (l @ l1) &&
                                    equal_lista _A3 r2 (r1 @ r)))
                                (showsl_lit "problem with selfoverlap"))
                          (fun _ ->
                            bindb (check
                                    (less_eq_nat f1 f2 &&
                                      (equal_nata
 (modulo_nata (max ord_nat (minus_nata c2 c1) (minus_nata c1 c2)) f1)
 zero_nata &&
((if less_nat f1 f2 then equal_nata (modulo_nata f2 f1) zero_nata else true) &&
  (if equal_nata f1 f2 then less_eq_nat c1 c2 else true))))
                                    (showsl_lit
                                      "could not ensure fitting condition for selfoverlap"))
                              (fun _ ->
                                check_proof (_A1, _A2, _A3, _A4) ra
                                  (set_empty
                                    ((ceq_prod
                                       (ceq_prod (ceq_list _A1)
 (ceq_prod (ceq_prod ceq_nat (ceq_prod ceq_nat (ceq_list _A1))) (ceq_list _A1)))
                                       (ceq_prod (ceq_list _A1)
 (ceq_prod (ceq_prod ceq_nat (ceq_prod ceq_nat (ceq_list _A1)))
   (ceq_list _A1)))),
                                      (ccompare_prod
(ccompare_prod (ccompare_list _A2)
  (ccompare_prod
    (ccompare_prod ccompare_nat
      (ccompare_prod ccompare_nat (ccompare_list _A2)))
    (ccompare_list _A2)))
(ccompare_prod (ccompare_list _A2)
  (ccompare_prod
    (ccompare_prod ccompare_nat
      (ccompare_prod ccompare_nat (ccompare_list _A2)))
    (ccompare_list _A2)))))
                                    (of_phantom
                                      (set_impl_proda
(set_impl_prod set_impl_list
  (set_impl_prod
    (set_impl_prod set_impl_nat (set_impl_prod set_impl_nat set_impl_list))
    set_impl_list))
(set_impl_prod set_impl_list
  (set_impl_prod
    (set_impl_prod set_impl_nat (set_impl_prod set_impl_nat set_impl_list))
    set_impl_list)))))
                                  steps))))
                      ba)))
                b));;

let rec term_to_string = function Fun (f, [t]) -> f :: term_to_string t
                         | Var v -> []
                         | Fun (v, []) -> []
                         | Fun (v, vb :: vd :: ve) -> [];;

let rec srs_of_trs_impl
  r = maps (fun (l, ra) ->
             (if unary_term l
               then (if unary_term ra
                      then [(term_to_string l, term_to_string ra)] else [])
               else []))
        r;;

let rec check_non_loop_srs_prf (_B1, _B2, _B3, _B4)
  i tp prf =
    (let r = rulesd i tp in
     let s =
       set ((ceq_prod (ceq_list _B1) (ceq_list _B1)),
             (ccompare_prod (ccompare_list _B2) (ccompare_list _B2)),
             (set_impl_prod set_impl_list set_impl_list))
         (srs_of_trs_impl r)
       in
      bindb (check (null (qb i tp))
              (showsl_lit "strategy for non-loops unsupported"))
        (fun _ -> check_non_loop_srs_proof (_B1, _B2, _B3, _B4) s prf));;

let rec check_not_wwf_qtrs (_B1, _B2) (_C1, _C2, _C3)
  i tp =
    bindb (check (null (qb i tp) || not (nfsb i tp))
            (showsl_lit "strategies and normal form substitutions problem"))
      (fun _ ->
        check (not (isOK (check_wwf_qtrs _B2 (_C2, _C3) (is_QNFb i tp)
                           (rulesd i tp))))
          (comp (showsl_lit "The Q-TRS is well formed")
            (showsl_literal "\092")));;

let rec check_tp_subsumesa (_B1, _B2, _B3) (_C1, _C2, _C3)
  i t tp =
    (let (nfsa, (q, rs)) = t in
     let nfs = nfsb i tp in
     let rsa = rulesd i tp in
     let qa = qb i tp in
      catch_error
        (bindb
          (check (if not (null q) then (if nfs then nfsa else true) else true)
            (showsl_lit "incompatible substitutions-in-normal-form flags"))
          (fun _ ->
            bindb (catch_error
                    (check_subseteq
                      (equal_prod
                        (equal_term (equal_lab _B2 _C2) (equal_list equal_char))
                        (equal_term (equal_lab _B2 _C2)
                          (equal_list equal_char)))
                      rs rsa)
                    (fun x ->
                      Inl (toomuch "rule"
                            (showsl_rule (showl_lab _B3 _C3)
                              (showl_list showl_char) x))))
              (fun _ ->
                catch_error
                  (check_NF_terms_subset
                    (is_NF_terms
                      ((compare_order_lab _B1.compare_compare_order
                         _C1.compare_compare_order),
                        (equal_lab _B2 _C2))
                      ((ccompare_list ccompare_char), (equal_list equal_char),
                        mapping_impl_list)
                      q)
                    qa)
                  (fun x ->
                    Inl (comp (showsl_lit "NF(Q) differs due to term ")
                          (showsl_terma (showl_lab _B3 _C3)
                            (showl_list showl_char) x))))))
        (fun x ->
          Inl (comp (showsl_lit
                      "problem in showing subsumption for non-termination\092")
                x)));;

let rec check_non_loop_trs_prf (_B1, _B2, _B3)
  (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8)
  i tp prf =
    (let r = rulesd i tp in
      bindb (check (null (qb i tp))
              (showsl_lit "strategy for non-loops unsupported"))
        (fun _ ->
          check_non_loop_prf (_B1, _B2, _B3)
            (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8) r [] prf));;

let rec uncurry_nonterm_tt_check (_B1, _B2)
  i info fmap check_inj r dpp =
    (let (a, (sml, (u, e))) = info in
     let ra = rulesd i dpp in
     let nfs = nfsb i dpp in
     let sm = sig_list_to_sig_map _B1 a sml fmap in
     let r_eta = e @ ra in
     let uR = uncurry_rules _B1 a sm r_eta in
      (match
        bindb (check (null (qb i dpp))
                (showsl_lit "strategy not supported for uncurrying"))
          (fun _ ->
            (let s = uncurry_of_sig_list a sml sm in
              bindb (only_eta_rules (_B1, _B2)
                      ((ccompare_list ccompare_char), (equal_list equal_char),
                        mapping_impl_list, (showl_list showl_char))
                      e r_eta)
                (fun _ ->
                  bindb (check_inj a (nat_of_integer (Big_int.big_int_of_int 2))
                          sml)
                    (fun _ ->
                      bindb (catch_error
                              (check_CS_subseteq _B1
                                ((ccompare_list ccompare_char),
                                  (equal_list equal_char), mapping_impl_list)
                                u s)
                              (fun x ->
                                Inl (comp (comp (showsl_lit "rule ")
    (showsl_rule _B2 (showl_list showl_char) x))
                                      (showsl_lit " is not an uncurry rule"))))
                        (fun _ ->
                          catch_error
                            (check_subseteq
                              (equal_prod
                                (equal_term _B1 (equal_list equal_char))
                                (equal_term _B1 (equal_list equal_char)))
                              r (u @ uR))
                            (fun x ->
                              Inl (comp (comp (showsl_lit "rule ")
  (showsl_rule _B2 (showl_list showl_char) x))
                                    (showsl_lit
                                      " is neither uncurried rules nor uncurry rule"))))))))
        with Inl aa -> Inl aa | Inr _ -> Inr (mkc i nfs [] r [])));;

let rec uncurry_nonterm_tt (_B1, _B2) (_C1, _C2)
  i (Uncurry_nt_proof ((a, (sml, (u, e))), r)) tp =
    uncurry_nonterm_tt_check ((equal_lab _B1 _C1), (showl_lab _B2 _C2)) i
      (a, (sml, (u, e)))
      (fmap _B1 _C1 a (nat_of_integer (Big_int.big_int_of_int 2)))
      (check_inj (_B1, _B2) (_C1, _C2)) r tp;;

let rec check_trs_loop (_B1, _B2, _B3) (_C1, _C2, _C3, _C4, _C5, _C6, _C7)
  i tp (TRS_loop_prf (s, rseq, sigma, c)) =
    check_loop (_B1, _B2, _B3) (_C1, _C2, _C3, _C4, _C5, _C6, _C7) (qb i tp)
      (nfsb i tp) s rseq sigma c (rulesd i tp);;

let rec rule_removal_nonterm_reltrs (_B1, _B2) (_C1, _C2)
  i tp (Rule_removal_nonterm_reltrs_prf (r, s)) =
    (let rrm =
       (if is_none r then []
         else ceta_list_diff
                (compare_order_prod
                  (compare_order_term _B1.compare_compare_order
                    _C1.compare_compare_order)
                  (compare_order_term _B1.compare_compare_order
                    _C1.compare_compare_order))
                (rc i tp) (the r))
       in
     let srm =
       (if is_none s then []
         else ceta_list_diff
                (compare_order_prod
                  (compare_order_term _B1.compare_compare_order
                    _C1.compare_compare_order)
                  (compare_order_term _B1.compare_compare_order
                    _C1.compare_compare_order))
                (rwc i tp) (the s))
       in
      Inr (delete_R_Rwb i tp rrm srm));;

let rec check_rel_tp_subsumes (_B1, _B2, _B3) (_C1, _C2, _C3)
  i t tp =
    (let (nfsa, (q, (r, rw))) = t in
     let nfs = nfsb i tp in
     let rules = rulesd i tp in
     let ra = rc i tp in
     let qa = qb i tp in
      catch_error
        (bindb
          (check (if not (null q) then (if nfs then nfsa else true) else true)
            (showsl_lit "incompatible substitutions-in-normal-form flags"))
          (fun _ ->
            bindb (catch_error
                    (check_subseteq
                      (equal_prod
                        (equal_term (equal_lab _B2 _C2) (equal_list equal_char))
                        (equal_term (equal_lab _B2 _C2)
                          (equal_list equal_char)))
                      rw rules)
                    (fun x ->
                      Inl (toomuch "rule"
                            (showsl_rule (showl_lab _B3 _C3)
                              (showl_list showl_char) x))))
              (fun _ ->
                bindb (catch_error
                        (check_subseteq
                          (equal_prod
                            (equal_term (equal_lab _B2 _C2)
                              (equal_list equal_char))
                            (equal_term (equal_lab _B2 _C2)
                              (equal_list equal_char)))
                          r ra)
                        (fun x ->
                          Inl (toomuch "rule"
                                (showsl_rule (showl_lab _B3 _C3)
                                  (showl_list showl_char) x))))
                  (fun _ ->
                    catch_error
                      (check_NF_terms_subset
                        (is_NF_terms
                          ((compare_order_lab _B1.compare_compare_order
                             _C1.compare_compare_order),
                            (equal_lab _B2 _C2))
                          ((ccompare_list ccompare_char),
                            (equal_list equal_char), mapping_impl_list)
                          q)
                        qa)
                      (fun x ->
                        Inl (comp (showsl_lit "NF(Q) differs due to term ")
                              (showsl_terma (showl_lab _B3 _C3)
                                (showl_list showl_char) x)))))))
        (fun x ->
          Inl (comp (showsl_lit
                      "problem in showing subsumption for non-termination\092")
                x)));;

let rec check_wf_reltrs _A (_B1, _B2) _C (_D1, _D2)
  (r, s) =
    bindb (check_wf_trs _A (_B1, _B2) r)
      (fun _ ->
        (if null r then Inr () else check_varcond_subset _C (_D1, _D2) s));;

let rec check_not_wf_reltrs (_B1, _B2) (_C1, _C2, _C3)
  i tp =
    bindb (check (q_emptyb i tp)
            (showsl_lit "currently only empty Q is supported"))
      (fun _ ->
        check (not (isOK (check_wf_reltrs _B2 (_C2, _C3) _B2 (_C2, _C3)
                           (rc i tp, rwc i tp))))
          (comp (showsl_lit "The TRSs R and S are well formed")
            (showsl_literal "\092")));;

let rec reltrs_as_trs i tp = (let q = qb i tp in
                              let r = rc i tp in
                              let nfs = nfsb i tp in
                              let a = mkc i nfs q r [] in
                               Inr a);;

let rec check_rel_loop (_A1, _A2) (_B1, _B2, _B3, _B4)
  sa rseq sigma c r s =
    check_rel_seq (_A1, _A2) (_B1, _B2, _B3, _B4) r s rseq sa
      (ctxt_apply_term c
        (subst_apply_term sa (mk_subst _B2 (fun a -> Var a) sigma)))
      false;;

let rec check_rel_trs_loop (_B1, _B2) (_C1, _C2, _C3, _C4)
  i tp (Rel_trs_loop_prf (s, rseq, sigma, c)) =
    bindb (check (q_emptyb i tp) (showsl_lit "Q is not empty"))
      (fun _ ->
        check_rel_loop (_B1, _B2) (_C1, _C2, _C3, _C4) s rseq sigma c (rc i tp)
          (rwc i tp));;

let rec check_reltrs_nontermination_proof (_B1, _B2, _B3, _B4) (_C1, _C2, _C3)
  ia j assms i tp x5 = match ia, j, assms, i, tp, x5 with
    ia, j, assms, i, tp, Rel_Loop p ->
      debug i "Loop"
        (catch_error
          (check_rel_trs_loop ((equal_lab _B3 _C2), (showl_lab _B4 _C3))
            ((ccompare_list ccompare_char), (equal_list equal_char),
              mapping_impl_list, (showl_list showl_char))
            ia tp p)
          (fun x ->
            Inl (comp (comp i (showsl_lit ": error in loop proof\092")) x)))
    | ia, j, assms, i, tp, Rel_R_Not_SN prf ->
        bindb (reltrs_as_trs ia tp)
          (fun tpa ->
            catch_error
              (check_trs_nontermination_proof (_B1, _B2, _B3, _B4)
                (_C1, _C2, _C3) ia j assms (add_index i one_nata) tpa prf)
              (fun x ->
                Inl (comp (comp i
                            (showsl_lit
                              ": error below the nontermination proof for R\092"))
                      x)))
    | ia, j, assms, i, tp, Rel_Not_Well_Formed ->
        debug i "Not Well-Formed"
          (catch_error
            (check_not_wf_reltrs
              ((compare_order_lab _B1.compare_compare_order
                 _C1.compare_compare_order),
                (showl_lab _B4 _C3))
              ((compare_order_list (compare_order_char, equal_char)),
                (equal_list equal_char), (showl_list showl_char))
              ia tp)
            (fun x ->
              Inl (comp (comp i
                          (showsl_lit ": error in not-well-formed proof\092"))
                    x)))
    | ia, j, assms, i, tp, Rel_TRS_String_Reversal prf ->
        debug i "String Reversal"
          (bindb
            (catch_error
              (string_reversal_complete_rel_tt (showl_lab _B4 _C3)
                (showl_list showl_char) ia tp)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit ": error when applying string reversal on\092"))
                                  (showsl_tp (showl_lab _B4 _C3)
                                    (showl_list showl_char) ia tp))
                            (showsl_literal "\092"))
                      x)))
            (fun tpa ->
              catch_error
                (check_reltrs_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata) tpa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the string reversal technique\092"))
                        x))))
    | ia, j, assms, i, tp, Rel_Rule_Removal (p, prf) ->
        debug i "Rule Removal"
          (bindb
            (rule_removal_nonterm_reltrs
              ((compare_order_lab _B1.compare_compare_order
                 _C1.compare_compare_order),
                (showl_lab _B4 _C3))
              ((compare_order_list (compare_order_char, equal_char)),
                (showl_list showl_char))
              ia tp p)
            (fun tpa ->
              catch_error
                (check_reltrs_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata) tpa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit ": error below the rule removal\092"))
                        x))))
    | ia, j, assms, i, tp, Rel_TRS_Assume_Not_SN (t, ass) ->
        debug i "Finiteness Assumption or Unknown Proof"
          (if assms
            then bindb (catch_error
                         (check_rel_tp_subsumes (_B1, _B3, _B4) (_C1, _C2, _C3)
                           ia t tp)
                         (fun x ->
                           Inl (comp (comp (comp i
     (showsl_lit ": error in non-termination assumption or unknown proof\092"))
                                       x)
                                 (showsl_literal "\092"))))
                   (fun _ ->
                     catch_error
                       (catch_error
                         (forallM_index
                           (fun asa ja ->
                             check_assmb (_B1, _B4) (_C1, _C3) ia j
                               (check_trs_nontermination_proof
                                 (_B1, _B2, _B3, _B4) (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               (check_dp_nontermination_proof
                                 (_B1, _B2, _B3, _B4) (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               (check_reltrs_nontermination_proof
                                 (_B1, _B2, _B3, _B4) (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               (check_fp_nontermination_proof
                                 (_B1, _B2, _B3, _B4) (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               (check_unknown_disproof (_B1, _B2, _B3, _B4)
                                 (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               asa)
                           ass)
                         (fun x -> Inl (snd x)))
                       (fun x ->
                         Inl (comp (comp i
                                     (showsl_lit
                                       ": error below unknown proof\092"))
                               x)))
            else Inl (comp i
                       (showsl_lit
                         ": the proof contains an assumption or unknown proof\092")))
and check_trs_nontermination_proof (_B1, _B2, _B3, _B4) (_C1, _C2, _C3)
  ia j assms i tp x5 = match ia, j, assms, i, tp, x5 with
    ia, j, assms, i, tp, TRS_Loop p ->
      debug i "Loop"
        (catch_error
          (check_trs_loop
            ((compare_lab _B1.compare_compare_order _C1.compare_compare_order),
              (equal_lab _B3 _C2), (showl_lab _B4 _C3))
            ((ceq_list ceq_char), (ccompare_list ccompare_char),
              (compare_list compare_char), (equal_list equal_char),
              mapping_impl_list, set_impl_list, (showl_list showl_char))
            ia tp p)
          (fun x ->
            Inl (comp (comp (comp (comp i
                                    (showsl_lit
                                      ": error when checking loop of\092"))
                              (showsl_tp (showl_lab _B4 _C3)
                                (showl_list showl_char) ia tp))
                        (showsl_literal "\092"))
                  x)))
    | ia, j, assms, i, tp, TRS_Nonloop p ->
        debug i "Nonloop"
          (catch_error
            (check_non_loop_trs_prf
              ((compare_lab _B1.compare_compare_order
                 _C1.compare_compare_order),
                (equal_lab _B3 _C2), (showl_lab _B4 _C3))
              (cenum_list, (ceq_list ceq_char), (ccompare_list ccompare_char),
                (compare_list compare_char), (equal_list equal_char),
                mapping_impl_list, set_impl_list, (showl_list showl_char))
              ia tp p)
            (fun x ->
              Inl (comp (comp (showsl_lit
                                "problem in checking possibly non-looping infinite reduction for\092")
                          (showsl_tp (showl_lab _B4 _C3) (showl_list showl_char)
                            ia tp))
                    x)))
    | ia, j, assms, i, tp, TRS_Nonloop_SRS p ->
        debug i "Nonloop SRS"
          (catch_error
            (check_non_loop_srs_prf
              ((ceq_lab _B3 _C2),
                (ccompare_lab _B1.compare_compare_order
                  _C1.compare_compare_order),
                (equal_lab _B3 _C2), (showl_lab _B4 _C3))
              ia tp p)
            (fun x ->
              Inl (comp (comp (showsl_lit
                                "problem in checking possibly non-looping infinite reduction for\092")
                          (showsl_tp (showl_lab _B4 _C3) (showl_list showl_char)
                            ia tp))
                    x)))
    | ia, j, assms, i, tp, TRS_Not_Well_Formed ->
        debug i "Not Well-Formed"
          (catch_error
            (check_not_wwf_qtrs
              ((compare_order_lab _B1.compare_compare_order
                 _C1.compare_compare_order),
                (showl_lab _B4 _C3))
              ((compare_order_list (compare_order_char, equal_char)),
                (equal_list equal_char), (showl_list showl_char))
              ia tp)
            (fun x ->
              Inl (comp (comp i
                          (showsl_lit ": error in not well-formed proof\092"))
                    x)))
    | ia, j, assms, i, tp, TRS_Termination_Switch (p, prf) ->
        debug i "Switch to Termination"
          (bindb
            (catch_error
              (switch_termination_tt
                ((compare_order_lab _B1.compare_compare_order
                   _C1.compare_compare_order),
                  (equal_lab _B3 _C2), (showl_lab _B4 _C3))
                ia p tp)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit ": error in switching to full strategy for the DP-problem\092"))
                                  (showsl_tp (showl_lab _B4 _C3)
                                    (showl_list showl_char) ia tp))
                            (showsl_literal "\092"))
                      x)))
            (fun tpa ->
              catch_error
                (check_trs_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata) tpa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the termination switch technique\092"))
                        x))))
    | ia, j, assms, i, tp, TRS_Rule_Removal (p, prf) ->
        debug i "Rule Removal"
          (bindb
            (rule_removal_nonterm_trs
              ((compare_order_lab _B1.compare_compare_order
                 _C1.compare_compare_order),
                (showl_lab _B4 _C3))
              ((compare_order_list (compare_order_char, equal_char)),
                (showl_list showl_char))
              ia tp p)
            (fun tpa ->
              catch_error
                (check_trs_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata) tpa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit ": error below the rule removal\092"))
                        x))))
    | ia, j, assms, i, tp, TRS_String_Reversal prf ->
        debug i "String Reversal"
          (bindb
            (catch_error
              (string_reversal_complete_tt (showl_lab _B4 _C3)
                (showl_list showl_char) ia tp)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit ": error when applying string reversal on\092"))
                                  (showsl_tp (showl_lab _B4 _C3)
                                    (showl_list showl_char) ia tp))
                            (showsl_literal "\092"))
                      x)))
            (fun tpa ->
              catch_error
                (check_trs_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata) tpa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the string reversal technique\092"))
                        x))))
    | ia, j, assms, i, tp, TRS_Constant_String (p, prf) ->
        debug i "Constants into Unary"
          (bindb
            (catch_error
              (const_to_string_complete_tt
                ((equal_lab _B3 _C2), (showl_lab _B4 _C3))
                ((equal_list equal_char), (showl_list showl_char)) ia tp p)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit ": error when turning constants into strings on\092"))
                                  (showsl_tp (showl_lab _B4 _C3)
                                    (showl_list showl_char) ia tp))
                            (showsl_literal "\092"))
                      x)))
            (fun tpa ->
              catch_error
                (check_trs_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata) tpa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the constants into string technique\092"))
                        x))))
    | ia, j, assms, i, tp, TRS_DP_Trans (p, prf) ->
        debug i "DP Transformation"
          (bindb
            (catch_error
              (dp_trans_nontermination_tt (_B3, _B4) (_C2, _C3)
                ((equal_list equal_char), (showl_list showl_char)) ia j tp p)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit ": error in DP transformation on\092"))
                                  (showsl_tp (showl_lab _B4 _C3)
                                    (showl_list showl_char) ia tp))
                            (showsl_literal "\092"))
                      x)))
            (fun dpp ->
              catch_error
                (check_dp_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata) dpp prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the DP transformation\092"))
                        x))))
    | ia, j, assms, i, tp, TRS_Q_Increase (p, prf) ->
        debug i "Q increase"
          (bindb
            (q_increase_nonterm_trs
              ((compare_order_lab _B1.compare_compare_order
                 _C1.compare_compare_order),
                (equal_lab _B3 _C2), (showl_lab _B4 _C3))
              ((compare_order_list (compare_order_char, equal_char)),
                (equal_list equal_char), (showl_list showl_char))
              ia tp p)
            (fun tpa ->
              catch_error
                (check_trs_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata) tpa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the Q increase technique\092"))
                        x))))
    | ia, j, assms, i, tp, TRS_Uncurry (p, prf) ->
        debug i "Uncurrying"
          (bindb (uncurry_nonterm_tt (_B3, _B4) (_C2, _C3) ia p tp)
            (fun tpa ->
              catch_error
                (check_trs_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata) tpa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the uncurrying technique\092"))
                        x))))
    | ia, j, assms, i, tp, TRS_Not_WN_Tree_Automaton prf ->
        catch_error
          (check_not_wn_ta_prf
            (finite_UNIV_lab, cenum_lab, (ceq_lab _B3 _C2),
              (cproper_interval_lab _B1.compare_compare_order
                _C1.compare_compare_order),
              (compare_order_lab _B1.compare_compare_order
                _C1.compare_compare_order),
              (default_lab _B2), (equal_lab _B3 _C2), set_impl_lab,
              (showl_lab _B4 _C3))
            (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
              (cproper_interval_list ccompare_char),
              (compare_order_list (compare_order_char, equal_char)),
              (equal_list equal_char), set_impl_list, (showl_list showl_char))
            (card_UNIV_list, cenum_list, (ceq_list ceq_char),
              (cproper_interval_list ccompare_char),
              (compare_order_list (compare_order_char, equal_char)),
              (equal_list equal_char), set_impl_list, (showl_list showl_char))
            ia tp prf)
          (fun x ->
            Inl (comp (comp (showsl_lit
                              "error in tree automaton based non-termination proof for\092")
                        (showsl_tp (showl_lab _B4 _C3) (showl_list showl_char)
                          ia tp))
                  x))
    | ia, j, assms, i, tp, TRS_Assume_Not_SN (t, ass) ->
        debug i "Finiteness Assumption or Unknown Proof"
          (if assms
            then bindb (catch_error
                         (check_tp_subsumesa (_B1, _B3, _B4) (_C1, _C2, _C3) ia
                           t tp)
                         (fun x ->
                           Inl (comp (comp (comp i
     (showsl_lit ": error in non-termination assumption or unknown proof\092"))
                                       x)
                                 (showsl_literal "\092"))))
                   (fun _ ->
                     catch_error
                       (catch_error
                         (forallM_index
                           (fun asa ja ->
                             check_assmb (_B1, _B4) (_C1, _C3) ia j
                               (check_trs_nontermination_proof
                                 (_B1, _B2, _B3, _B4) (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               (check_dp_nontermination_proof
                                 (_B1, _B2, _B3, _B4) (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               (check_reltrs_nontermination_proof
                                 (_B1, _B2, _B3, _B4) (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               (check_fp_nontermination_proof
                                 (_B1, _B2, _B3, _B4) (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               (check_unknown_disproof (_B1, _B2, _B3, _B4)
                                 (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               asa)
                           ass)
                         (fun x -> Inl (snd x)))
                       (fun x ->
                         Inl (comp (comp i
                                     (showsl_lit
                                       ": error below unknown proof\092"))
                               x)))
            else Inl (comp i
                       (showsl_lit
                         ": the proof contains an assumption or unknown proof\092")))
    | ia, j, assms, i, tp, FPTRS_Assume_Not_SN (t, ass) ->
        Inl (comp i
              (showsl_lit ": encountered unexpected FPTRS_Assume_Not_SN\092"))
and check_fp_nontermination_proof (_B1, _B2, _B3, _B4) (_C1, _C2, _C3)
  ia j assms i tp x5 = match ia, j, assms, i, tp, x5 with
    ia, j, assms, i, (pa, r), TRS_Loop p ->
      (let TRS_loop_prf (a, b, c, d) = p in
        debug i "Loop"
          (catch_error
            (check_fploop
              ((compare_lab _B1.compare_compare_order
                 _C1.compare_compare_order),
                (equal_lab _B3 _C2), (showl_lab _B4 _C3))
              ((ceq_list ceq_char), (ccompare_list ccompare_char),
                (compare_list compare_char), (equal_list equal_char),
                mapping_impl_list, set_impl_list, (showl_list showl_char))
              r pa (FP_loop_prf (d, c, a, b)))
            (fun x ->
              Inl (comp (comp i
                          (showsl_lit
                            ": error when checking forbidden pattern loop\092"))
                    x))))
    | ia, j, assms, i, (pa, r), TRS_Rule_Removal (p, prf) ->
        debug i "Rule Removal"
          (bindb
            (rule_removal_nonterm_trs
              ((compare_order_lab _B1.compare_compare_order
                 _C1.compare_compare_order),
                (showl_lab _B4 _C3))
              ((compare_order_list (compare_order_char, equal_char)),
                (showl_list showl_char))
              ia (mkc ia false [] r []) p)
            (fun tp ->
              catch_error
                (check_fp_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata)
                  (pa, rulesd ia tp) prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit ": error below the rule removal\092"))
                        x))))
    | ia, j, assms, i, tp, FPTRS_Assume_Not_SN (t, ass) ->
        debug i "Finiteness Assumption or Unknown Proof"
          (if assms
            then bindb (catch_error
                         (check_fp_tp_subsumes (_B1, _B3, _B4) (_C1, _C2, _C3) t
                           tp)
                         (fun x ->
                           Inl (comp (comp (comp i
     (showsl_lit ": error in non-termination assumption or unknown proof\092"))
                                       x)
                                 (showsl_literal "\092"))))
                   (fun _ ->
                     catch_error
                       (catch_error
                         (forallM_index
                           (fun asa ja ->
                             check_assmb (_B1, _B4) (_C1, _C3) ia j
                               (check_trs_nontermination_proof
                                 (_B1, _B2, _B3, _B4) (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               (check_dp_nontermination_proof
                                 (_B1, _B2, _B3, _B4) (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               (check_reltrs_nontermination_proof
                                 (_B1, _B2, _B3, _B4) (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               (check_fp_nontermination_proof
                                 (_B1, _B2, _B3, _B4) (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               (check_unknown_disproof (_B1, _B2, _B3, _B4)
                                 (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               asa)
                           ass)
                         (fun x -> Inl (snd x)))
                       (fun x ->
                         Inl (comp (comp i
                                     (showsl_lit
                                       ": error below unknown proof\092"))
                               x)))
            else Inl (comp i
                       (showsl_lit
                         ": the proof contains an assumption or unknown proof\092")))
    | ia, j, assms, i, tp, TRS_Not_Well_Formed ->
        Inl (comp i (showsl_lit ": encountered unexpected proof\092"))
    | ia, j, assms, i, tp, TRS_String_Reversal v ->
        Inl (comp i (showsl_lit ": encountered unexpected proof\092"))
    | ia, j, assms, i, tp, TRS_Constant_String (v, va) ->
        Inl (comp i (showsl_lit ": encountered unexpected proof\092"))
    | ia, j, assms, i, tp, TRS_DP_Trans (v, va) ->
        Inl (comp i (showsl_lit ": encountered unexpected proof\092"))
    | ia, j, assms, i, tp, TRS_Termination_Switch (v, va) ->
        Inl (comp i (showsl_lit ": encountered unexpected proof\092"))
    | ia, j, assms, i, tp, TRS_Nonloop v ->
        Inl (comp i (showsl_lit ": encountered unexpected proof\092"))
    | ia, j, assms, i, tp, TRS_Nonloop_SRS v ->
        Inl (comp i (showsl_lit ": encountered unexpected proof\092"))
    | ia, j, assms, i, tp, TRS_Q_Increase (v, va) ->
        Inl (comp i (showsl_lit ": encountered unexpected proof\092"))
    | ia, j, assms, i, tp, TRS_Uncurry (v, va) ->
        Inl (comp i (showsl_lit ": encountered unexpected proof\092"))
    | ia, j, assms, i, tp, TRS_Not_WN_Tree_Automaton v ->
        Inl (comp i (showsl_lit ": encountered unexpected proof\092"))
    | ia, j, assms, i, tp, TRS_Assume_Not_SN (v, va) ->
        Inl (comp i (showsl_lit ": encountered unexpected proof\092"))
and check_dp_nontermination_proof (_B1, _B2, _B3, _B4) (_C1, _C2, _C3)
  ia j assms i dpp x5 = match ia, j, assms, i, dpp, x5 with
    ia, j, assms, i, dpp, DP_Loop p ->
      debug i "Loop"
        (catch_error
          (check_dp_loop
            ((compare_lab _B1.compare_compare_order _C1.compare_compare_order),
              (equal_lab _B3 _C2), (showl_lab _B4 _C3))
            ((ceq_list ceq_char), (ccompare_list ccompare_char),
              (compare_list compare_char), (equal_list equal_char),
              mapping_impl_list, set_impl_list, (showl_list showl_char))
            j dpp p)
          (fun x ->
            Inl (comp (comp (comp (comp i
                                    (showsl_lit
                                      ": error in checking loop for the following DP-problem\092"))
                              (showsl_dpp (showl_lab _B4 _C3)
                                (showl_list showl_char) j dpp))
                        (showsl_literal "\092"))
                  x)))
    | ia, j, assms, i, dpp, DP_Nonloop p ->
        debug i "Nonloop"
          (catch_error
            (check_non_loop_dp_prf
              ((compare_lab _B1.compare_compare_order
                 _C1.compare_compare_order),
                (equal_lab _B3 _C2), (showl_lab _B4 _C3))
              (cenum_list, (ceq_list ceq_char), (ccompare_list ccompare_char),
                (compare_list compare_char), (equal_list equal_char),
                mapping_impl_list, set_impl_list, (showl_list showl_char))
              j dpp p)
            (fun x ->
              Inl (comp (comp (comp (comp i
                                      (showsl_lit
": error in checking nonloop for the following DP-problem\092"))
                                (showsl_dpp (showl_lab _B4 _C3)
                                  (showl_list showl_char) j dpp))
                          (showsl_literal "\092"))
                    x)))
    | ia, j, assms, i, dpp, DP_Rule_Removal (p, prf) ->
        debug i "Rule Removal"
          (bindb
            (rule_removal_nonterm_dp
              ((compare_order_lab _B1.compare_compare_order
                 _C1.compare_compare_order),
                (showl_lab _B4 _C3))
              ((compare_order_list (compare_order_char, equal_char)),
                (showl_list showl_char))
              j dpp p)
            (fun dppa ->
              catch_error
                (check_dp_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the pair and rule removal\092"))
                        x))))
    | ia, j, assms, i, dpp, DP_Q_Reduction (p, prf) ->
        debug i "Q reduction"
          (bindb
            (catch_error
              (dp_q_reduction_nonterm
                (cenum_lab, (ceq_lab _B3 _C2),
                  (ccompare_lab _B1.compare_compare_order
                    _C1.compare_compare_order),
                  (compare_order_lab _B1.compare_compare_order
                    _C1.compare_compare_order),
                  (equal_lab _B3 _C2), set_impl_lab, (showl_lab _B4 _C3))
                ((ccompare_list ccompare_char), (equal_list equal_char),
                  mapping_impl_list, (showl_list showl_char))
                j dpp p)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit
  ": error in reducing the innermost lhss in the following DP-problem\092"))
                                  (showsl_dpp (showl_lab _B4 _C3)
                                    (showl_list showl_char) j dpp))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit ": error below the Q reduction\092"))
                        x))))
    | ia, j, assms, i, dpp, DP_Q_Increase (p, prf) ->
        debug i "Q increase"
          (bindb
            (q_increase_nonterm_dp (equal_lab _B3 _C2) (equal_list equal_char) j
              dpp p)
            (fun dppa ->
              catch_error
                (check_dp_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit ": error below the Q increase\092"))
                        x))))
    | ia, j, assms, i, dpp, DP_Termination_Switch (p, prf) ->
        debug i "Switch to Termination"
          (bindb
            (catch_error
              (switch_termination_proc
                ((compare_order_lab _B1.compare_compare_order
                   _C1.compare_compare_order),
                  (equal_lab _B3 _C2), (showl_lab _B4 _C3))
                j p dpp)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit ": error in switching to full strategy for the DP-problem\092"))
                                  (showsl_dpp (showl_lab _B4 _C3)
                                    (showl_list showl_char) j dpp))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the termination switch processor\092"))
                        x))))
    | ia, j, assms, i, dpp, DP_Instantiation (p, prf) ->
        debug i "Instantiation"
          (bindb
            (catch_error
              (instantiation_complete_proc
                ((equal_lab _B3 _C2), (showl_lab _B4 _C3))
                ((ccompare_list ccompare_char), (equal_list equal_char),
                  mapping_impl_list, (showl_list showl_char))
                j dpp p)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit ": error when applying the instantiation processor on\092"))
                                  (showsl_dpp (showl_lab _B4 _C3)
                                    (showl_list showl_char) j dpp))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the instantiation processor\092"))
                        x))))
    | ia, j, assms, i, dpp, DP_Rewriting (p, prf) ->
        debug i "Rewriting"
          (bindb
            (catch_error
              (rewriting_complete_proc
                ((compare_order_lab _B1.compare_compare_order
                   _C1.compare_compare_order),
                  (equal_lab _B3 _C2), (showl_lab _B4 _C3))
                j p dpp)
              (fun x ->
                Inl (comp (comp i
                            (showsl_lit
                              ": error when applying the rewriting processor\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the rewriting processor\092"))
                        x))))
    | ia, j, assms, i, dpp, DP_Narrowing (p, prf) ->
        debug i "Narrowing"
          (bindb
            (catch_error
              (narrowing_complete_proc
                ((compare_order_lab _B1.compare_compare_order
                   _C1.compare_compare_order),
                  (equal_lab _B3 _C2), (showl_lab _B4 _C3))
                j p dpp)
              (fun x ->
                Inl (comp (comp i
                            (showsl_lit
                              ": error when applying the narrowing processor\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_dp_nontermination_proof (_B1, _B2, _B3, _B4)
                  (_C1, _C2, _C3) ia j assms (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the rewriting processor"))
                        x))))
    | ia, j, assms, i, dpp, DP_Assume_Infinite (dp, ass) ->
        debug i "Finiteness Assumption or Unknown Proof"
          (if assms
            then bindb (catch_error
                         (check_dpp_subsumesa (_B1, _B3, _B4) (_C1, _C2, _C3) j
                           dp dpp)
                         (fun x ->
                           Inl (comp (comp (comp i
     (showsl_lit ": error in finiteness assumption or unknown proof\092"))
                                       x)
                                 (showsl_literal "\092"))))
                   (fun _ ->
                     catch_error
                       (catch_error
                         (forallM_index
                           (fun asa ja ->
                             check_assmb (_B1, _B4) (_C1, _C3) ia j
                               (check_trs_nontermination_proof
                                 (_B1, _B2, _B3, _B4) (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               (check_dp_nontermination_proof
                                 (_B1, _B2, _B3, _B4) (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               (check_reltrs_nontermination_proof
                                 (_B1, _B2, _B3, _B4) (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               (check_fp_nontermination_proof
                                 (_B1, _B2, _B3, _B4) (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               (check_unknown_disproof (_B1, _B2, _B3, _B4)
                                 (_C1, _C2, _C3) ia j assms
                                 (add_index i (suc ja)))
                               asa)
                           ass)
                         (fun x -> Inl (snd x)))
                       (fun x ->
                         Inl (comp (comp i
                                     (showsl_lit
                                       ": error below unknown proof\092"))
                               x)))
            else Inl (comp i
                       (showsl_lit
                         ": the proof contains an assumption or unknown proof\092")))
and check_unknown_disproof (_B1, _B2, _B3, _B4) (_C1, _C2, _C3)
  ia j assms i tpa (Assume_NT_Unknown (tp, ass)) =
    debug i "Unknown Proof"
      (if assms
        then bindb (catch_error
                     (check (equal_lista equal_char tpa tp)
                       (comp (comp (comp (showsl_lit
   "unknown problems are not identical:\092")
                                     (showsl_lista showl_char tpa))
                               (showsl_lit "\092 vs\092"))
                         (showsl_lista showl_char tp)))
                     (fun x ->
                       Inl (comp (comp (comp i
 (showsl_lit ": error in nontermination assumption or unknown proof\092"))
                                   x)
                             (showsl_literal "\092"))))
               (fun _ ->
                 catch_error
                   (forallM_index
                     (fun asa ja ->
                       check_assmb (_B1, _B4) (_C1, _C3) ia j
                         (check_trs_nontermination_proof (_B1, _B2, _B3, _B4)
                           (_C1, _C2, _C3) ia j assms (add_index i (suc ja)))
                         (check_dp_nontermination_proof (_B1, _B2, _B3, _B4)
                           (_C1, _C2, _C3) ia j assms (add_index i (suc ja)))
                         (check_reltrs_nontermination_proof (_B1, _B2, _B3, _B4)
                           (_C1, _C2, _C3) ia j assms (add_index i (suc ja)))
                         (check_fp_nontermination_proof (_B1, _B2, _B3, _B4)
                           (_C1, _C2, _C3) ia j assms (add_index i (suc ja)))
                         (check_unknown_disproof (_B1, _B2, _B3, _B4)
                           (_C1, _C2, _C3) ia j assms (add_index i (suc ja)))
                         asa)
                     ass)
                   (fun x -> Inl (snd x)))
        else Inl (comp i
                   (showsl_lit ": the proof contains an unknown proof\092")));;

let rec find_variant_in_trs (_A1, _A2) (_B1, _B2, _B3, _B4)
  ra r =
    catch_error
      (firstM (check_variants_rule (_A1, _A2) (_B1, _B2, _B3, _B4) ra) r)
      (fun x -> Inl (showsl_sep id (showsl_literal "\092") x));;

let rec check_rstep_p _A (_B1, _B2, _B3)
  c rho p s t =
    (if not (membera (equal_list equal_nat) (poss_list t) p)
      then Inl (showsl_literal "no step possible at this position")
      else (let (l, r) = rho in
             (match
               match_list (_B1, _B2, _B3) _A _B2 (fun a -> Var a)
                 [(l, subt_at s p); (r, subt_at t p)]
               with None -> Inl (showsl_literal "rule does not match")
               | Some sigma ->
                 (if equal_terma _A _B2 t
                       (ctxt_apply_term (ctxt_of_pos_term p s)
                         (subst_apply_term r sigma))
                   then c (subst_apply_term l sigma) (subst_apply_term r sigma)
                   else Inl (showsl_literal "result does not match")))));;

let rec check_step_rule _A (_B1, _B2, _B3)
  c rho s t =
    catch_error
      (existsM (fun p -> check_rstep_p _A (_B1, _B2, _B3) c rho p s t)
        (poss_list s))
      (fun _ -> Inl (showsl_literal " is not a reduct"));;

let rec check_step (_A1, _A2) (_B1, _B2, _B3, _B4)
  c x1 s t = match c, x1, s, t with
    c, ea :: e, s, t ->
      catch_error
        (choice
          [check_step_rule _A1 (_B1, _B2, _B3) c ea s t;
            check_step (_A1, _A2) (_B1, _B2, _B3, _B4) c e s t])
        (fun x -> Inl (showsl_sep id (showsl_literal "\092") x))
    | c, [], s, t ->
        Inl (comp (comp (comp (comp (showsl_lit "no step from")
                                (showsl_terma _A2 _B4 s))
                          (showsl_lit " and "))
                    (showsl_terma _A2 _B4 t))
              (showsl_lit " found "));;

let rec sym_list _A
  xs = union (equal_prod _A _A) xs (map (fun (x, y) -> (y, x)) xs);;

let rec check_stepa (_A1, _A2)
  check_ord x1 x2 = match check_ord, x1, x2 with
    check_ord, (e, r), OC_Collapse (s, t, u) ->
      catch_error
        (bindb
          (find_variant_in_trs (_A1, _A2)
            ((ccompare_list ccompare_char), (equal_list equal_char),
              mapping_impl_list, (showl_list showl_char))
            (t, s) r)
          (fun ts ->
            (let ra =
               removeAll
                 (equal_prod (equal_term _A1 (equal_list equal_char))
                   (equal_term _A1 (equal_list equal_char)))
                 ts r
               in
              bindb (catch_error
                      (check_step (_A1, _A2)
                        ((ccompare_list ccompare_char), (equal_list equal_char),
                          mapping_impl_list, (showl_list showl_char))
                        check_ord
                        (ra @ sym_list (equal_term _A1 (equal_list equal_char))
                                e)
                        t u)
                      (fun x ->
                        Inl (comp (comp (comp
  (comp (showsl_lit " no ordstep from ")
    (showsl_terma _A2 (showl_list showl_char) t))
  (showsl_lit " to "))
                                    (showsl_terma _A2 (showl_list showl_char)
                                      u))
                              x)))
                (fun _ -> Inr ((u, s) :: e, ra)))))
        (fun x ->
          Inl (comp (comp (comp (comp (comp
(comp (comp (comp (showsl_lit "error in collapse step from ")
              (showsl_terma _A2 (showl_list showl_char) s))
        (showsl_lit " -> "))
  (showsl_terma _A2 (showl_list showl_char) t))
(showsl_lit " to "))
                                  (showsl_terma _A2 (showl_list showl_char) u))
                            (showsl_lit " = "))
                      (showsl_terma _A2 (showl_list showl_char) t))
                x))
    | check_ord, (e, r), OC_Simplifyr (s, t, u) ->
        catch_error
          (bindb
            (find_variant_in_trs (_A1, _A2)
              ((ccompare_list ccompare_char), (equal_list equal_char),
                mapping_impl_list, (showl_list showl_char))
              (s, t) e)
            (fun st ->
              (let ea =
                 removeAll
                   (equal_prod (equal_term _A1 (equal_list equal_char))
                     (equal_term _A1 (equal_list equal_char)))
                   st e
                 in
                bindb (catch_error
                        (check_step (_A1, _A2)
                          ((ccompare_list ccompare_char),
                            (equal_list equal_char), mapping_impl_list,
                            (showl_list showl_char))
                          check_ord
                          (r @ sym_list (equal_term _A1 (equal_list equal_char))
                                 ea)
                          t u)
                        (fun x ->
                          Inl (comp (comp (comp
    (comp (showsl_lit " no ordstep from ")
      (showsl_terma _A2 (showl_list showl_char) t))
    (showsl_lit " to "))
                                      (showsl_terma _A2 (showl_list showl_char)
u))
                                x)))
                  (fun _ -> Inr ((s, u) :: ea, r)))))
          (fun x ->
            Inl (comp (comp (comp (comp (comp
  (comp (comp (comp (showsl_lit "error in simplifyr step from ")
                (showsl_terma _A2 (showl_list showl_char) s))
          (showsl_lit " = "))
    (showsl_terma _A2 (showl_list showl_char) t))
  (showsl_lit " to "))
                                    (showsl_terma _A2 (showl_list showl_char)
                                      s))
                              (showsl_lit " = "))
                        (showsl_terma _A2 (showl_list showl_char) u))
                  x))
    | check_ord, (e, r), OC_Simplifyl (s, t, u) ->
        catch_error
          (bindb
            (find_variant_in_trs (_A1, _A2)
              ((ccompare_list ccompare_char), (equal_list equal_char),
                mapping_impl_list, (showl_list showl_char))
              (s, t) e)
            (fun st ->
              (let ea =
                 removeAll
                   (equal_prod (equal_term _A1 (equal_list equal_char))
                     (equal_term _A1 (equal_list equal_char)))
                   st e
                 in
                bindb (catch_error
                        (check_step (_A1, _A2)
                          ((ccompare_list ccompare_char),
                            (equal_list equal_char), mapping_impl_list,
                            (showl_list showl_char))
                          check_ord
                          (r @ sym_list (equal_term _A1 (equal_list equal_char))
                                 ea)
                          s u)
                        (fun x ->
                          Inl (comp (comp (comp
    (comp (showsl_lit " no ordstep from ")
      (showsl_terma _A2 (showl_list showl_char) s))
    (showsl_lit " to "))
                                      (showsl_terma _A2 (showl_list showl_char)
u))
                                x)))
                  (fun _ -> Inr ((u, t) :: ea, r)))))
          (fun x ->
            Inl (comp (comp (comp (comp (comp
  (comp (comp (comp (showsl_lit "error in simplifyl step from ")
                (showsl_terma _A2 (showl_list showl_char) s))
          (showsl_lit " = "))
    (showsl_terma _A2 (showl_list showl_char) t))
  (showsl_lit " to "))
                                    (showsl_terma _A2 (showl_list showl_char)
                                      u))
                              (showsl_lit " = "))
                        (showsl_terma _A2 (showl_list showl_char) t))
                  x))
    | check_ord, (e, r), OC_Compose (s, t, u) ->
        catch_error
          (bindb
            (find_variant_in_trs (_A1, _A2)
              ((ccompare_list ccompare_char), (equal_list equal_char),
                mapping_impl_list, (showl_list showl_char))
              (s, t) r)
            (fun st ->
              (let ra =
                 removeAll
                   (equal_prod (equal_term _A1 (equal_list equal_char))
                     (equal_term _A1 (equal_list equal_char)))
                   st r
                 in
                bindb (catch_error
                        (check_step (_A1, _A2)
                          ((ccompare_list ccompare_char),
                            (equal_list equal_char), mapping_impl_list,
                            (showl_list showl_char))
                          check_ord
                          (ra @ sym_list
                                  (equal_term _A1 (equal_list equal_char)) e)
                          t u)
                        (fun x ->
                          Inl (comp (comp (comp
    (comp (showsl_lit " no ordstep from ")
      (showsl_terma _A2 (showl_list showl_char) t))
    (showsl_lit " to "))
                                      (showsl_terma _A2 (showl_list showl_char)
u))
                                x)))
                  (fun _ -> Inr (e, (s, u) :: ra)))))
          (fun x ->
            Inl (comp (comp (comp (comp (comp
  (comp (comp (comp (showsl_lit "error in compose step from ")
                (showsl_terma _A2 (showl_list showl_char) s))
          (showsl_lit " -> "))
    (showsl_terma _A2 (showl_list showl_char) t))
  (showsl_lit " to "))
                                    (showsl_terma _A2 (showl_list showl_char)
                                      s))
                              (showsl_lit " -> "))
                        (showsl_terma _A2 (showl_list showl_char) u))
                  x))
    | check_ord, (e, r), OC_Delete s ->
        catch_error
          (bindb
            (find_variant_in_trs (_A1, _A2)
              ((ccompare_list ccompare_char), (equal_list equal_char),
                mapping_impl_list, (showl_list showl_char))
              (s, s) e)
            (fun ss ->
              Inr (removeAll
                     (equal_prod (equal_term _A1 (equal_list equal_char))
                       (equal_term _A1 (equal_list equal_char)))
                     ss e,
                    r)))
          (fun x ->
            Inl (comp (comp (comp (comp (showsl_lit "error in delete step for ")
                                    (showsl_terma _A2 (showl_list showl_char)
                                      s))
                              (showsl_lit " = "))
                        (showsl_terma _A2 (showl_list showl_char) s))
                  x))
    | check_ord, (e, r), OC_Orientr (s, t) ->
        catch_error
          (bindb (check_ord t s)
            (fun _ ->
              bindb (find_variant_in_trs (_A1, _A2)
                      ((ccompare_list ccompare_char), (equal_list equal_char),
                        mapping_impl_list, (showl_list showl_char))
                      (s, t) e)
                (fun st ->
                  Inr (removeAll
                         (equal_prod (equal_term _A1 (equal_list equal_char))
                           (equal_term _A1 (equal_list equal_char)))
                         st e,
                        (t, s) :: r))))
          (fun x ->
            Inl (comp (comp (comp (comp (showsl_lit
  "error in orientr step for ")
                                    (showsl_terma _A2 (showl_list showl_char)
                                      s))
                              (showsl_lit " -> "))
                        (showsl_terma _A2 (showl_list showl_char) t))
                  x))
    | check_ord, (e, r), OC_Orientl (s, t) ->
        catch_error
          (bindb (check_ord s t)
            (fun _ ->
              bindb (find_variant_in_trs (_A1, _A2)
                      ((ccompare_list ccompare_char), (equal_list equal_char),
                        mapping_impl_list, (showl_list showl_char))
                      (s, t) e)
                (fun st ->
                  Inr (removeAll
                         (equal_prod (equal_term _A1 (equal_list equal_char))
                           (equal_term _A1 (equal_list equal_char)))
                         st e,
                        (s, t) :: r))))
          (fun x ->
            Inl (comp (comp (comp (comp (showsl_lit
  "error in orientl step for ")
                                    (showsl_terma _A2 (showl_list showl_char)
                                      s))
                              (showsl_lit " -> "))
                        (showsl_terma _A2 (showl_list showl_char) t))
                  x))
    | check_ord, (e, r), OC_Deduce (s, t, u) ->
        catch_error
          (let sa = r @ sym_list (equal_term _A1 (equal_list equal_char)) e in
            bindb (catch_error
                    (check_step (_A1, _A2)
                      ((ccompare_list ccompare_char), (equal_list equal_char),
                        mapping_impl_list, (showl_list showl_char))
                      (fun _ _ -> Inr ()) sa s t)
                    (fun x ->
                      Inl (comp (comp (comp
(comp (showsl_lit " no step from ")
  (showsl_terma _A2 (showl_list showl_char) s))
(showsl_lit " to "))
                                  (showsl_terma _A2 (showl_list showl_char) t))
                            x)))
              (fun _ ->
                bindb (catch_error
                        (check_step (_A1, _A2)
                          ((ccompare_list ccompare_char),
                            (equal_list equal_char), mapping_impl_list,
                            (showl_list showl_char))
                          (fun _ _ -> Inr ()) sa s u)
                        (fun x ->
                          Inl (comp (comp (comp
    (comp (showsl_lit " no step from ")
      (showsl_terma _A2 (showl_list showl_char) s))
    (showsl_lit " to "))
                                      (showsl_terma _A2 (showl_list showl_char)
u))
                                x)))
                  (fun _ -> Inr ((t, u) :: e, r))))
          (fun x ->
            Inl (comp (comp (comp (comp (comp
  (comp (showsl_lit "error in deduce step ")
    (showsl_terma _A2 (showl_list showl_char) t))
  (showsl_lit " <- "))
                                    (showsl_terma _A2 (showl_list showl_char)
                                      s))
                              (showsl_lit " -> "))
                        (showsl_terma _A2 (showl_list showl_char) u))
                  x));;

let rec check_same_set _A
  xs ys = bindb (check_subseteq _A xs ys) (fun _ -> check_subseteq _A ys xs);;

let rec check_oc (_A1, _A2)
  check_ord x1 x2 x3 = match check_ord, x1, x2, x3 with
    check_ord, (ea, ra), (e, r), x :: xs ->
      bindb (check_stepa (_A1, _A2) check_ord (ea, ra) x)
        (fun (eaa, raa) -> check_oc (_A1, _A2) check_ord (eaa, raa) (e, r) xs)
    | check_ord, (ea, ra), (e, r), [] ->
        bindb (catch_error
                (check_same_set
                  (equal_prod (equal_term _A1 (equal_list equal_char))
                    (equal_term _A1 (equal_list equal_char)))
                  ea e)
                (fun _ -> Inl (showsl_lit " equation sets do not match")))
          (fun _ ->
            catch_error
              (check_same_set
                (equal_prod (equal_term _A1 (equal_list equal_char))
                  (equal_term _A1 (equal_list equal_char)))
                ra r)
              (fun _ -> Inl (showsl_lit " rule sets do not match")));;

let rec check_ooverlap (_A1, _A2, _A3) (_B1, _B2)
  xvar yvar check_joinable check_instance e r rho_1 rho_2 p =
    (match
      mgu_var_disjoint_generic _A1 _B1 xvar yvar (fst rho_1)
        (subt_at (fst rho_2) p)
      with None -> Inr ()
      | Some (sigma_1, sigma_2) ->
        (let s =
           ctxt_apply_term
             (ctxt_of_pos_term p (subst_apply_term (fst rho_2) sigma_2))
             (subst_apply_term (snd rho_1) sigma_1)
           in
         let t = subst_apply_term (snd rho_2) sigma_2 in
          catch_error
            (choice
              [catch_error (check_instance e r s t)
                 (fun x ->
                   Inl (comp (comp (comp (comp
   (comp (showsl_lit "CP ") (showsl_terma _B2 _A3 s)) (showsl_lit " = "))
                                     (showsl_terma _B2 _A3 t))
                               (showsl_lit " not an instance of an equation"))
                         x));
                catch_error (check_joinable e r s t)
                  (fun x ->
                    Inl (comp (comp (comp (comp
    (comp (showsl_lit "CP ") (showsl_terma _B2 _A3 s)) (showsl_lit " = "))
                                      (showsl_terma _B2 _A3 t))
                                (showsl_lit " is not joinable"))
                          x))])
            (fun x -> Inl (showsl_sep id (showsl_literal "\092") x))));;

let rec check_ECPsa (_A1, _A2, _A3) (_B1, _B2)
  xvar yvar check_joinable check_instance e r =
    (let ea = sym_list (equal_term _B1 _A1) e in
     let s = union (equal_prod (equal_term _B1 _A1) (equal_term _B1 _A1)) r ea
       in
      catch_error
        (forallM
          (fun rho_2 ->
            (let l_2 = fst rho_2 in
              catch_error
                (forallM
                  (fun rho_1 ->
                    catch_error
                      (forallM
                        (check_ooverlap (_A1, _A2, _A3) (_B1, _B2) xvar yvar
                          check_joinable check_instance ea r rho_1 rho_2)
                        (funposs_list l_2))
                      (fun x -> Inl (snd x)))
                  s)
                (fun x -> Inl (snd x))))
          s)
        (fun x -> Inl (snd x)));;

let rec check_joinable _A (_B1, _B2, _B3)
  r s t =
    (match
      (compute_rstep_NF _A (_B1, _B2, _B3) r s,
        compute_rstep_NF _A (_B1, _B2, _B3) r t)
      with (None, _) -> Inl (showsl_lit "error: check_joinable")
      | (Some _, None) -> Inl (showsl_lit "error: check_joinable")
      | (Some u, Some v) ->
        check (equal_terma _A _B2 u v) (showsl_lit "normal forms differ"));;

let rec check_rule_instance _A (_B1, _B2, _B3)
  ra r =
    (match
      match_list (_B1, _B2, _B3) _A _B2 (fun a -> Var a)
        [(fst r, fst ra); (snd r, snd ra)]
      with None -> Inl (showsl_lit "rules do not match") | Some _ -> Inr ());;

let rec check_instancea _A (_B1, _B2, _B3)
  e s t =
    catch_error (existsM (check_rule_instance _A (_B1, _B2, _B3) (s, t)) e)
      (fun x -> Inl (showsl_sep id (showsl_literal "\092") x));;

let rec check_ECPs (_A1, _A2)
  = check_ECPsa
      ((equal_list equal_char), infinite_list, (showl_list showl_char))
      (_A1, _A2)
      (fun a -> Chara (false, false, false, true, true, true, true, false) :: a)
      (fun a -> Chara (true, false, false, true, true, true, true, false) :: a)
      (fun _ ->
        check_joinable _A1
          ((ccompare_list ccompare_char), (equal_list equal_char),
            mapping_impl_list))
      (fun e _ ->
        check_instancea _A1
          ((ccompare_list ccompare_char), (equal_list equal_char),
            mapping_impl_list)
          e);;

let rec valida (Redord_ext (valid, less, min_const, more)) = valid;;

let rec check_FGCR (_D1, _D2)
  ro f e r =
    bindb (valida ro)
      (fun _ ->
        bindb (catch_error
                (check_subseteq (equal_prod _D1 equal_nat)
                  (funas_trs_list
                    (union
                      (equal_prod (equal_term _D1 (equal_list equal_char))
                        (equal_term _D1 (equal_list equal_char)))
                      e r))
                  f)
                (fun x ->
                  Inl (comp (comp (showsl_lit "the function symbol ")
                              (showsl_prod _D2 showl_nat x))
                        (showsl_lit " does not occur in the TRS\092"))))
          (fun _ -> check_ECPs (_D1, _D2) e r));;

let rec lessa (Redord_ext (valid, less, min_const, more)) = less;;

let rec check_FGCR_run (_A1, _A2)
  ro f e_0 r_0 e r steps =
    (let check_ord =
       (fun s t ->
         check (lessa ro s t) (showsl_lit "term pair cannot be oriented"))
       in
      bindb (catch_error (forallM (fun (a, b) -> check_ord a b) r_0)
              (fun x -> Inl (snd x)))
        (fun _ ->
          bindb (catch_error
                  (check_oc (_A1, _A2) check_ord (e_0, r_0) (e, r) steps)
                  (fun x ->
                    Inl (comp (showsl_lit
                                "the oKB run could not be reconstructed\092\092")
                          x)))
            (fun _ ->
              catch_error (check_FGCR (_A1, _A2) ro f e r)
                (fun x ->
                  Inl (comp (showsl_lit
                              "ground confluence could not be verified\092\092")
                        x)))));;

let rec create_KBO_redord (_A1, _A2, _A3)
  pr fs =
    (let (ch, (p, (w, (w0, (lcs, scf))))) =
       prec_weight_repr_to_prec_weight (_A1, _A2, _A3) pr in
     let valid =
       bindb ch
         (fun _ ->
           bindb (catch_error
                   (check_same_set (equal_prod _A2 equal_nat) fs
                     (map fst (fst pr)))
                   (fun _ -> Inl (showsl_literal " signature does not match ")))
             (fun _ ->
               bindb (check (less_nat zero_nata (size_list lcs))
                       (showsl_literal
                         "there must be a minimal constant with weight w0"))
                 (fun _ ->
                   check (distinct equal_nat (map (comp fst snd) (fst pr)))
                     (showsl_literal "the given precedence is not injective"))))
       in
      Redord_ext
        (valid,
          (fun s t ->
            fst (kbo_impl (equal_list equal_char) w w0 p (membera _A2 lcs) scf s
                  t)),
          nth lcs zero_nata, ()));;

let rec check_ordered_completion_proof (_B1, _B2, _B3)
  a ia i j e_0 e r ro (OKB steps) =
    debug ia "OKB"
      (match ro
        with RPO_Input _ -> Inl (showsl_lit "unsupported reduction order ")
        | KBO_Input precw ->
          (let f = funas_trs_list e_0 in
            catch_error
              (check_FGCR_run
                ((equal_lab _B2 (equal_list equal_nat)),
                  (showl_lab _B3 (showl_list showl_nat)))
                (create_KBO_redord
                  ((compare_order_lab _B1.compare_compare_order
                     (compare_list compare_nat)),
                    (equal_lab _B2 (equal_list equal_nat)),
                    (showl_lab _B3 (showl_list showl_nat)))
                  precw f)
                f e_0 [] e r steps)
              (fun x ->
                Inl (comp (comp ia
                            (showsl_lit
                              ": error in ground completeness proof \092"))
                      x))));;

let rec rhs_n
  u (lr, cs) n =
    (if less_nat n (size_list cs)
      then ctxt_apply_term (u (lr, cs) n) (fst (nth cs n)) else snd lr);;

let rec lhs_n
  u (lr, cs) n =
    (if equal_nata n zero_nata then fst lr
      else ctxt_apply_term (u (lr, cs) (minus_nata n one_nata))
             (snd (nth cs (minus_nata n one_nata))));;

let rec rules_impl
  u cr =
    map (fun i -> (lhs_n u cr i, rhs_n u cr i))
      (upt zero_nata (suc (size_list (snd cr))));;

let rec create_ctxts
  r = (match r with [] -> None
        | _ :: rr ->
          bind (mapMa
                 (fun a ->
                   (match a with (Var _, _) -> None | (Fun (_, []), _) -> None
                     | (Fun (u, _ :: ts), _) -> Some (More (u, [], Hole, ts))))
                 rr)
            (fun cs ->
              (let _ = size_list cs in
                Some (fun i ->
                       (if less_nat i (size_list cs) then nth cs i
                         else Hole)))));;

let rec create_U _A _B
  c_rs =
    bind (mapMa
           (fun (cr, rs) ->
             bind (guard (equal_nata (size_list rs) (suc (size_list (snd cr)))))
               (fun _ -> bind (create_ctxts rs) (fun ctxt -> Some (cr, ctxt))))
           c_rs)
      (fun cr_ctxts ->
        (let m =
           map_of
             (equal_prod (equal_prod (equal_term _A _B) (equal_term _A _B))
               (equal_list (equal_prod (equal_term _A _B) (equal_term _A _B))))
             cr_ctxts
           in
          Some (fun cr ->
                 (match m cr with None -> (fun _ -> Hole)
                   | Some ctxt -> ctxt))));;

let rec check_unraveling (_A1, _A2) (_B1, _B2)
  c_rs ctrs =
    catch_error
      (bindb
        (catch_error
          (check_subseteq
            (equal_prod (equal_prod (equal_term _A1 _B1) (equal_term _A1 _B1))
              (equal_list
                (equal_prod (equal_term _A1 _B1) (equal_term _A1 _B1))))
            ctrs (map fst c_rs))
          (fun x ->
            Inl (comp (comp (showsl_lit "did not find rule ")
                        (showsl_crule _A2 _B2 _A2 _B2 x))
                  (showsl_literal "\092"))))
        (fun _ ->
          bindb (match create_U _A1 _B1 c_rs
                  with None ->
                    Inl (showsl_lit "unable to extract unraveling contexts")
                  | Some a -> Inr a)
            (fun u ->
              bindb (catch_error
                      (forallM
                        (fun (c, rs) ->
                          check (equal_lista
                                  (equal_prod (equal_term _A1 _B1)
                                    (equal_term _A1 _B1))
                                  (rules_impl u c) rs)
                            (comp (comp (showsl_lit "problem with rules of ")
                                    (showsl_crule _A2 _B2 _A2 _B2 c))
                              (showsl_literal "\092")))
                        c_rs)
                      (fun x -> Inl (snd x)))
                (fun _ -> Inr (maps snd c_rs)))))
      (fun x -> Inl (comp (showsl_lit "problem in unraveling\092") x));;

let rec check_quasi_reductive_proof (_B1, _B2, _B3, _B4)
  a ia i j ctrs (Unravel (u_info, prf)) =
    debug ia "Unravel"
      (bindb
        (catch_error
          (check_unraveling
            ((equal_lab _B3 (equal_list equal_nat)),
              (showl_lab _B4 (showl_list showl_nat)))
            ((equal_list equal_char), (showl_list showl_char)) u_info ctrs)
          (fun x ->
            Inl (comp (comp ia (showsl_lit ": error in unraveling\092")) x)))
        (fun r ->
          (let tp = mkc i false [] r [] in
            catch_error
              (check_trs_termination_proof (_B1, _B2, _B3, _B4) i j a
                (comp ia (showsl_lit ".1")) tp prf)
              (fun x ->
                Inl (comp (comp ia
                            (showsl_lit
                              ": error below strong normalization + wcr\092"))
                      x)))));;

let rec showsl_gctxt _A
  = function GCHole -> showsl_literal "_"
    | GCFun (f, ts) ->
        comp (showsl _A f)
          (showsl_list_gen id "" "(" ", " ")" (map (showsl_gctxt _A) ts));;

let rec gctxts_to_terms_intern
  iv i x2 = match iv, i, x2 with
    iv, i, GCFun (f, ts) :: cs ->
      (let (i1, res1) = gctxts_to_terms_intern iv i ts in
       let (i2, res2) = gctxts_to_terms_intern iv i1 cs in
        (i2, Fun (f, res1) :: res2))
    | iv, i, GCHole :: cs ->
        (let (ia, res) = gctxts_to_terms_intern iv (plus_nata i one_nata) cs in
          (ia, Var (iv i) :: res))
    | iv, i, [] -> (i, []);;

let rec gc_matcher _A
  c l = map_option fst
          (mgu_var_disjoint_generic (equal_list equal_char) _A
            (fun a ->
              Chara (false, false, false, true, true, true, true, false) :: a)
            (fun a ->
              Chara (true, false, false, true, true, true, true, false) :: a)
            l (hda (snd (gctxts_to_terms_intern
                          (fun i ->
                            Chara (false, false, false, true, true, true, true,
                                    false) ::
                              shows_prec_nat zero_nata i [])
                          zero_nata [c]))));;

let rec check_contains_U0 (_A1, _A2, _A3)
  r u s =
    catch_error
      (forallM
        (fun fts ->
          (match fts with Var _ -> Inr ()
            | Fun (f, ts) ->
              (let tcapb =
                 GCFun (f, map (tcap (_A1, _A2)
                                 ((compare_list compare_char),
                                   (equal_list equal_char))
                                 u)
                             ts)
                 in
                catch_error
                  (forallM
                    (fun lr ->
                      (match gc_matcher _A2 tcapb (fst lr) with None -> Inr ()
                        | Some sigma ->
                          (let irule =
                             (subst_apply_term (fst lr) sigma,
                               subst_apply_term (snd lr) sigma)
                             in
                            check (in_rstep_impl (_A1, _A2)
                                    ((ccompare_list ccompare_char),
                                      (compare_list compare_char),
                                      (equal_list equal_char),
                                      mapping_impl_list)
                                    (fst irule) (snd irule) u)
                              (comp (comp (comp
    (comp (comp (comp (comp (comp (comp (comp
  (showsl_literal "When considering the subterm ")
  (showsl_terma _A3 (showl_list showl_char) fts))
                                    (showsl_literal "\092and the rule "))
                              (showsl_rule _A3 (showl_list showl_char) lr))
                        (showsl_literal "\092the capped subterm is "))
                  (showsl_gctxt _A3 tcapb))
            (showsl_literal "\092leading to the mgu with the lhs: "))
      (showsl_list_prod (showl_list showl_char)
        (showl_term _A3 (showl_list showl_char))
        (map (fun x -> (x, sigma x)) (vars_term_list (fst lr)))))
    (showsl_literal "\092The instantiated rule "))
                                      (showsl_rule _A3 (showl_list showl_char)
irule))
                                (showsl_literal
                                  "\092cannot be simulated by the given set of usable rules")))))
                    r)
                  (fun x -> Inl (snd x)))))
        (supteq_list s))
      (fun x -> Inl (snd x));;

let rec check_usable_instantiation (_A1, _A2, _A3)
  r u s =
    (let uu =
       set ((ceq_prod (ceq_term _A2 (equal_list equal_char))
              (ceq_term _A2 (equal_list equal_char))),
             (ccompare_prod (ccompare_term _A1 (compare_list compare_char))
               (ccompare_term _A1 (compare_list compare_char))),
             (set_impl_prod set_impl_term set_impl_term))
         u
       in
      bindb (catch_error (check_contains_U0 (_A1, _A2, _A3) r uu s)
              (fun x ->
                Inl (comp (showsl_literal "U <= U0(R,s) required\092") x)))
        (fun _ ->
          catch_error
            (forallM
              (fun ra ->
                catch_error (check_contains_U0 (_A1, _A2, _A3) r uu ra)
                  (fun x ->
                    Inl (comp (comp (comp (showsl_literal
    "U <= U0(R,r) for rhs r = ")
                                      (showsl_terma _A3 (showl_list showl_char)
ra))
                                (showsl_literal " required\092"))
                          x)))
              (map snd u))
            (fun x -> Inl (snd x))));;

let rec check_usable_rules_unif (_A1, _A2, _A3)
  r u s =
    catch_error
      (bindb
        (check (ground s || list_all (fun l -> not (is_Var l)) (map fst r))
          (comp (comp (showsl_literal "since ")
                  (showsl_terma _A3 (showl_list showl_char) s))
            (showsl_literal
              " is not ground, left-hand sides of R must not be variables")))
        (fun _ ->
          bindb (check_varcond_subset _A3
                  ((equal_list equal_char), (showl_list showl_char)) u)
            (fun _ ->
              catch_error (check_usable_instantiation (_A1, _A2, _A3) r u s)
                (fun x ->
                  Inl (comp (showsl_literal
                              "closure properties of usable rules not satisfied\092")
                        x)))))
      (fun x ->
        Inl (comp (comp (comp (comp (comp (comp
    (comp (showsl_literal
            "problem in checking validity of usable rules U =\092")
      (showsl_trs _A3 (showl_list showl_char) u))
    (showsl_literal "\092for term "))
                                      (showsl_terma _A3 (showl_list showl_char)
s))
                                (showsl_literal "\092wrt TRS R =\092"))
                          (showsl_trs _A3 (showl_list showl_char) r))
                    (showsl_literal "\092"))
              x));;

let rec sl_c
  (Sl_ops_ext
    (sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS, sl_lgen,
      sl_LS_gen, more))
    = sl_c;;

let rec check_qmodel_rule_ass _A _B _C
  i cge alpha (l, r) =
    (let cl = eval i alpha l in
     let cr = eval i alpha r in
      check (cge cl cr)
        (comp (comp (comp (comp (comp (showsl_lit "rule ")
                                  (showsl_rule _A _C (l, r)))
                            (showsl_lit
                              " violates the model condition, [lhs] = "))
                      (showsl _B cl))
                (showsl_lit ", [rhs] = "))
          (showsl _B cr)));;

let rec check_qmodel_rule _A _B (_C1, _C2)
  i c cge lr =
    catch_error
      (forallM (fun alpha -> check_qmodel_rule_ass _A _B _C2 i cge alpha lr)
        (map (fun_of _C1) (enum_vectors c (insert_vars_rule _C1 lr []))))
      (fun x -> Inl (snd x));;

let rec check_qmodel _A _B (_C1, _C2)
  i c cge r =
    catch_error (forallM (check_qmodel_rule _A _B (_C1, _C2) i c cge) r)
      (fun x -> Inl (snd x));;

let rec check_non_join_model _A _B (_D1, _D2)
  cge gen rs rt s t =
    catch_error
      (bindb (gen (funas_trs_list (rs @ rt)) [])
        (fun ops ->
          (let i = sl_I ops in
           let e = eval i (fun _ -> sl_c ops) in
           let es = e s in
           let et = e t in
            bindb (check (not (cge et es))
                    (comp (comp (comp (comp
(comp (comp (comp (comp (showsl_lit "the inequality must not hold: [")
                    (showsl_terma _B _D2 t))
              (showsl_lit "] = "))
        (showsl _A et))
  (showsl_lit " >= "))
(showsl _A es))
                                  (showsl_lit " = ["))
                            (showsl_terma _B _D2 s))
                      (showsl_lit "]")))
              (fun _ ->
                check_qmodel _B _A (_D1, _D2) i (sl_C ops) cge
                  (reverse_rules rs @ rt)))))
      (fun x ->
        Inl (comp (comp (showsl_lit
                          "problem in disproving non-joinability via interpretations")
                    (showsl_literal "\092"))
              x));;

let rec check_non_join_finite_model (_A1, _A2) (_B1, _B2, _B3, _B4)
  x0 rs rt s t = match x0, rs, rt, s, t with
    Rootlab x, rs, rt, s, t ->
      check_non_join_model (showl_lab _A2 (showl_list showl_nat))
        (showl_lab _A2 (showl_list showl_nat)) (_B2, _B4)
        (equal_laba _A1 (equal_list equal_nat))
        (slm_gen_to_sl_gen (rl_slm _A1 (equal_list equal_nat) x)) rs rt s t
    | Finitelab sli, rs, rt, s, t ->
        check_non_join_model showl_nat (showl_lab _A2 (showl_list showl_nat))
          (_B2, _B4) equal_nata
          (slm_gen_to_sl_gen (fun _ _ -> Inr (sli_to_slm _A1 sli))) rs rt s t
    | QuasiFinitelab (sli, v), rs, rt, s, t ->
        check_non_join_model showl_nat (showl_lab _A2 (showl_list showl_nat))
          (_B2, _B4) qmodel_cge
          (fun f g -> qsli_to_sl (_B1, _B2, _B3) (_A1, _A2) v f g sli) rs rt s
          t;;

let rec match_tcap_below_impl _A _B
  l r x2 = match l, r, x2 with
    l, r, Fun (f, ts) -> matchc _A _B (GCFun (f, map (tcapI _A _B r) ts)) l
    | l, r, Var x -> false;;

let rec usable_rules_reach_U0_impl _A _B
  r t = inductive_set_impl (equal_prod (equal_term _A _B) (equal_term _A _B))
          (equal_term _A _B) r
          (fun ta (l, _) ->
            is_Var l ||
              list_ex
                (fun u -> not (is_Var u) && match_tcap_below_impl _A _B l r u)
                (supteq_list ta))
          (fun lr -> [snd lr]) [t];;

let rec usable_rules_reach_impl (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  r t = (let u0t = usable_rules_reach_U0_impl _A2 _B6 r t in
          (if list_all
                (fun (l, ra) ->
                  less_eq_set (_B2, _B3, _B4.ccompare_cproper_interval)
                    (vars_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
                      ra)
                    (vars_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
                      l))
                u0t
            then u0t else r));;

let rec check_non_join_redpair (_A1, _A2) _B
  rp rs rt sa t =
    catch_error
      (bindb (valid rp)
        (fun _ ->
          bindb (catch_error (forallM (ns rp) (reverse_rules rs @ rt))
                  (fun x -> Inl (snd x)))
            (fun _ -> s rp (sa, t))))
      (fun x ->
        Inl (comp (comp (showsl_lit
                          "problem in disproving non-joinability via discrimination pairs")
                    (showsl_literal "\092"))
              x));;

let rec non_join_with_ta (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7) (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8)
  _D (_E1, _E2, _E3, _E4, _E5, _E6, _E7, _E8)
  (_F1, _F2, _F3, _F4, _F5, _F6, _F7, _F8) _G
  ta1 rel1 r1 t1 ta2 rel2 r2 t2 =
    (let tA1 =
       ta_of_ta (_A3, _A4.ccompare_cproper_interval, _A6, _A7) (_B2, _B5) ta1 in
     let tA2 =
       ta_of_ta (_E3, _E4.ccompare_cproper_interval, _E6, _E7) (_B2, _B5) ta2 in
      bindb (check
              (ta_member (_B2, _B5)
                (_A2, _A3, _A4.ccompare_cproper_interval, _A6, _A7) t1 tA1)
              (comp (showsl_terma _B7 _D t1)
                (showsl_lit " is not accepted by first automaton")))
        (fun _ ->
          bindb (check
                  (ta_member (_B2, _B5)
                    (_E2, _E3, _E4.ccompare_cproper_interval, _E6, _E7) t2 tA2)
                  (comp (showsl_terma _B7 _G t2)
                    (showsl_lit " is not accepted by first automaton")))
            (fun _ ->
              bindb (check
                      (ta_empty
                        ((cenum_prod _A2 _E2), (ceq_prod _A3 _E3),
                          (ccompare_prod _A4.ccompare_cproper_interval
                            _E4.ccompare_cproper_interval),
                          (equal_prod _A6 _E6), (set_impl_prod _A7 _E7))
                        (_B2, _B4, _B5)
                        (intersect_ta
                          (_A1.finite_UNIV_card_UNIV, _A2, _A3, _A4, _A6, _A7)
                          (_B1, _B2, _B5, _B6)
                          (_E1.finite_UNIV_card_UNIV, _E2, _E3, _E4, _E6, _E7)
                          tA1 tA2))
                      (showsl_lit "intersection of automata is non-empty"))
                (fun _ ->
                  bindb (catch_error
                          (tree_aut_trs_closed
                            (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8)
                            (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
                            (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8) ta1 rel1
                            r1)
                          (fun x ->
                            Inl (comp (showsl_lit
"could not ensure closure under rewriting for first automaton\092")
                                  x)))
                    (fun _ ->
                      catch_error
                        (tree_aut_trs_closed
                          (_E1, _E2, _E3, _E4, _E5, _E6, _E7, _E8)
                          (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
                          (_F1, _F2, _F3, _F4, _F5, _F6, _F7, _F8) ta2 rel2 r2)
                        (fun x ->
                          Inl (comp (showsl_lit
                                      "could not ensure closure under rewriting for second automaton\092")
                                x)))))));;

let rec check_non_join (_A1, _A2, _A3, _A4)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  rs rt s t x4 = match rs, rt, s, t, x4 with
    rs, rt, s, t, Diff_NFs ->
      bindb (check
              (not (equal_terma (equal_lab _A3 (equal_list equal_nat))
                     (equal_list equal_char) s t))
              (comp (comp (comp (comp (showsl_lit "the terms ")
                                  (showsl_terma
                                    (showl_lab _A4 (showl_list showl_nat))
                                    (showl_list showl_char) s))
                            (showsl_lit " and "))
                      (showsl_terma (showl_lab _A4 (showl_list showl_nat))
                        (showl_list showl_char) t))
                (showsl_lit " are identical")))
        (fun _ ->
          (let chknf =
             (fun sa r ->
               check (is_NF_trs
                       ((compare_order_lab _A1.compare_compare_order
                          (compare_list compare_nat)),
                         (equal_lab _A3 (equal_list equal_nat)))
                       ((ccompare_list ccompare_char), (equal_list equal_char),
                         mapping_impl_list)
                       r sa)
                 (comp (comp (showsl_lit "the term ")
                         (showsl_terma (showl_lab _A4 (showl_list showl_nat))
                           (showl_list showl_char) sa))
                   (showsl_lit " is not in normal form")))
             in
            bindb (chknf s rs) (fun _ -> chknf t rt)))
    | rs, rt, s, t, Grounding (sigma, prf) ->
        (let sigmaa = mk_subst (equal_list equal_char) (fun a -> Var a) sigma in
          check_non_join (_A1, _A2, _A3, _A4)
            (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) rs rt
            (subst_apply_term s sigmaa) (subst_apply_term t sigmaa) prf)
    | rs, rt, s, t, Subterm_NJ (p, prf) ->
        bindb (check
                (member ((ceq_list ceq_nat), (ccompare_list ccompare_nat)) p
                  (pos_gctxt
                    ((compare_lab _A1.compare_compare_order
                       (compare_list compare_nat)),
                      (equal_lab _A3 (equal_list equal_nat)))
                    (compare_list compare_char)
                    (tcapI (equal_lab _A3 (equal_list equal_nat))
                      (equal_list equal_char) rs s)))
                (comp (comp (comp (showsl_lit "position ") (showsl_pos p))
                        (showsl_lit " not in capped term  of "))
                  (showsl_terma (showl_lab _A4 (showl_list showl_nat))
                    (showl_list showl_char) s)))
          (fun _ ->
            bindb (check
                    (member ((ceq_list ceq_nat), (ccompare_list ccompare_nat)) p
                      (pos_gctxt
                        ((compare_lab _A1.compare_compare_order
                           (compare_list compare_nat)),
                          (equal_lab _A3 (equal_list equal_nat)))
                        (compare_list compare_char)
                        (tcapI (equal_lab _A3 (equal_list equal_nat))
                          (equal_list equal_char) rt t)))
                    (comp (comp (comp (showsl_lit "position ") (showsl_pos p))
                            (showsl_lit " not in capped term  of "))
                      (showsl_terma (showl_lab _A4 (showl_list showl_nat))
                        (showl_list showl_char) t)))
              (fun _ ->
                check_non_join (_A1, _A2, _A3, _A4)
                  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) rs rt (subt_at s p)
                  (subt_at t p) prf))
    | rs, rt, s, t, Tcap_Non_Unif grd_subst ->
        (let sigma = grd_subst s t in
         let cs =
           tcapI (equal_lab _A3 (equal_list equal_nat)) (equal_list equal_char)
             rs (subst_apply_term s sigma)
           in
         let ct =
           tcapI (equal_lab _A3 (equal_list equal_nat)) (equal_list equal_char)
             rt (subst_apply_term t sigma)
           in
          check (is_none (mergea (equal_lab _A3 (equal_list equal_nat)) cs ct))
            (comp (comp (comp (comp (showsl_lit "could not infer that ")
                                (showsl_terma
                                  (showl_lab _A4 (showl_list showl_nat))
                                  (showl_list showl_char) s))
                          (showsl_lit " and "))
                    (showsl_terma (showl_lab _A4 (showl_list showl_nat))
                      (showl_list showl_char) t))
              (showsl_lit " are not joinable")))
    | rs, rt, s, t, Tree_Aut_Intersect_Empty (ta1, rel1, ta2, rel2) ->
        catch_error
          (non_join_with_ta (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
            ((ceq_lab _A3 (equal_list equal_nat)),
              (ccompare_lab _A1.compare_compare_order
                (compare_list compare_nat)),
              (compare_order_lab _A1.compare_compare_order
                (compare_list compare_nat)),
              (default_lab _A2), (equal_lab _A3 (equal_list equal_nat)),
              set_impl_lab, (showl_lab _A4 (showl_list showl_nat)))
            (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
              (cproper_interval_list ccompare_char),
              (compare_order_list (compare_order_char, equal_char)),
              (equal_list equal_char), set_impl_list, (showl_list showl_char))
            (showl_list showl_char) (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
            (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
              (cproper_interval_list ccompare_char),
              (compare_order_list (compare_order_char, equal_char)),
              (equal_list equal_char), set_impl_list, (showl_list showl_char))
            (showl_list showl_char) ta1 rel1 rs s ta2 rel2 rt t)
          (fun x ->
            Inl (comp (comp (comp (comp (comp
  (comp (showsl_lit "could not infer that ")
    (showsl_terma (showl_lab _A4 (showl_list showl_nat)) (showl_list showl_char)
      s))
  (showsl_lit " and "))
                                    (showsl_terma
                                      (showl_lab _A4 (showl_list showl_nat))
                                      (showl_list showl_char) t))
                              (showsl_lit " are not joinable"))
                        (showsl_literal "\092"))
                  x))
    | rs, rt, s, t, Finite_Model_Gt i ->
        catch_error
          (check_non_join_finite_model (_A3, _A4)
            ((ccompare_list ccompare_char), (equal_list equal_char),
              mapping_impl_list, (showl_list showl_char))
            i rs rt s t)
          (fun x ->
            Inl (comp (comp (comp (comp (comp
  (comp (showsl_lit "could not infer that ")
    (showsl_terma (showl_lab _A4 (showl_list showl_nat)) (showl_list showl_char)
      s))
  (showsl_lit " and "))
                                    (showsl_terma
                                      (showl_lab _A4 (showl_list showl_nat))
                                      (showl_list showl_char) t))
                              (showsl_lit " are not joinable"))
                        (showsl_literal "\092"))
                  x))
    | rs, rt, s, t, Reduction_Pair_Gt rp ->
        catch_error
          (check_non_join_redpair
            ((compare_order_lab _A1.compare_compare_order
               (compare_list compare_nat)),
              (showl_lab _A4 (showl_list showl_nat)))
            (showl_list showl_char)
            (get_redtriple
              ((ceq_lab _A3 (equal_list equal_nat)),
                (ccompare_lab _A1.compare_compare_order
                  (compare_list compare_nat)),
                (compare_order_lab _A1.compare_compare_order
                  (compare_list compare_nat)),
                (equal_lab _A3 (equal_list equal_nat)), set_impl_lab,
                (showl_lab _A4 (showl_list showl_nat)))
              rp)
            rs rt s t)
          (fun x ->
            Inl (comp (comp (comp (comp (comp
  (comp (showsl_lit "could not infer that ")
    (showsl_terma (showl_lab _A4 (showl_list showl_nat)) (showl_list showl_char)
      s))
  (showsl_lit " and "))
                                    (showsl_terma
                                      (showl_lab _A4 (showl_list showl_nat))
                                      (showl_list showl_char) t))
                              (showsl_lit " are not joinable"))
                        (showsl_literal "\092"))
                  x))
    | rs, rt, s, t, Usable_Rules_Reach_NJ prf ->
        check_non_join (_A1, _A2, _A3, _A4)
          (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
          (usable_rules_reach_impl
            ((compare_lab _A1.compare_compare_order (compare_list compare_nat)),
              (equal_lab _A3 (equal_list equal_nat)))
            (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
              (cproper_interval_list ccompare_char),
              (compare_list compare_char), (equal_list equal_char),
              set_impl_list)
            rs s)
          (usable_rules_reach_impl
            ((compare_lab _A1.compare_compare_order (compare_list compare_nat)),
              (equal_lab _A3 (equal_list equal_nat)))
            (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
              (cproper_interval_list ccompare_char),
              (compare_list compare_char), (equal_list equal_char),
              set_impl_list)
            rt t)
          s t prf
    | rs, rt, s, t, Usable_Rules_Reach_Unif_NJ (u_sum, prf) ->
        (match u_sum
          with Inl u ->
            bindb (check_usable_rules_unif
                    ((compare_lab _A1.compare_compare_order
                       (compare_list compare_nat)),
                      (equal_lab _A3 (equal_list equal_nat)),
                      (showl_lab _A4 (showl_list showl_nat)))
                    rs u s)
              (fun _ ->
                check_non_join (_A1, _A2, _A3, _A4)
                  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) u rt s t prf)
          | Inr u ->
            bindb (check_usable_rules_unif
                    ((compare_lab _A1.compare_compare_order
                       (compare_list compare_nat)),
                      (equal_lab _A3 (equal_list equal_nat)),
                      (showl_lab _A4 (showl_list showl_nat)))
                    rt u t)
              (fun _ ->
                check_non_join (_A1, _A2, _A3, _A4)
                  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) rs u s t prf))
    | rs, rt, s, t, Argument_Filter_NJ (pi, prf) ->
        (match
          afs_of
            ((ceq_lab _A3 (equal_list equal_nat)),
              (ccompare_lab _A1.compare_compare_order
                (compare_list compare_nat)),
              (compare_order_lab _A1.compare_compare_order
                (compare_list compare_nat)),
              set_impl_lab)
            pi
          with None -> Inl (showsl_lit "invalid argument filter")
          | Some pia ->
            (let af = af_term pia in
             let afs = af_rules pia in
              check_non_join (_A1, _A2, _A3, _A4)
                (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) (afs rs) (afs rt)
                (af s) (af t) prf));;

let rec ru_impl r = map fst r;;

let rec check_conditional_non_cr (_A1, _A2, _A3, _A4)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  r s t u ps qs reason =
    bindb (check_csteps
            ((equal_lab _A3 (equal_list equal_nat)),
              (showl_lab _A4 (showl_list showl_nat)))
            ((ccompare_list ccompare_char), (equal_list equal_char),
              mapping_impl_list, (showl_list showl_char))
            r s t ps)
      (fun _ ->
        bindb (check_csteps
                ((equal_lab _A3 (equal_list equal_nat)),
                  (showl_lab _A4 (showl_list showl_nat)))
                ((ccompare_list ccompare_char), (equal_list equal_char),
                  mapping_impl_list, (showl_list showl_char))
                r s u qs)
          (fun _ ->
            check_non_join (_A1, _A2, _A3, _A4)
              (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) (ru_impl r) (ru_impl r) t
              u reason));;

let rec find_index _A
  i x xa2 = match i, x, xa2 with i, x, [] -> Inl ()
    | i, x, y :: ys ->
        (if eq _A x y then Inr i else find_index _A (suc i) x ys);;

let rec check_inline_conds_rule (_A1, _A2, _A3)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  r rho x2 = match r, rho, x2 with r, rho, [] -> Inr (r, rho)
    | r, rho, c :: cs ->
        catch_error
          (let a = rho in
           let (aa, b) = a in
            (let (l, _) = aa in
              (fun csa ->
                (let (s, t) = c in
                  bindb (check (is_Var t)
                          (showsl_lit "condition with non-variable rhs"))
                    (fun _ ->
                      (let x = the_Var t in
                        bindb (catch_error
                                (find_index
                                  (equal_prod (equal_term _A2 _B6)
                                    (equal_term _A2 _B6))
                                  zero_nata c csa)
                                (fun xa ->
                                  Inl (let () = xa in
showsl_lit "condition does not occur in rule")))
                          (fun i ->
                            bindb (check (not (contains_var_term _B6 x s))
                                    (showsl_lit "occurs check failed"))
                              (fun _ ->
                                bindb (check (not (contains_var_term _B6 x l))
(showsl_lit "inlining not allowed in lhs of rule"))
                                  (fun _ ->
                                    bindb (check
    (not (member (_B3, _B4.ccompare_cproper_interval) x
           (sup_seta (_B3, _B4.ccompare_cproper_interval)
             (sup_setb (_B1, _B2, _B3, _B4, _B7)
               (image ((ceq_term _A2 _B6), (ccompare_term _A1 _B5))
                 ((ceq_set (_B2, _B3, _B4.ccompare_cproper_interval)),
                   (ccompare_set (_B1, _B3, _B4, _B7)), set_impl_set)
                 (vars_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7))
                 (image
                   ((ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6)),
                     (ccompare_prod (ccompare_term _A1 _B5)
                       (ccompare_term _A1 _B5)))
                   ((ceq_term _A2 _B6), (ccompare_term _A1 _B5), set_impl_term)
                   snd (set ((ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6)),
                              (ccompare_prod (ccompare_term _A1 _B5)
                                (ccompare_term _A1 _B5)),
                              (set_impl_prod set_impl_term set_impl_term))
                         (take i csa)))))
             (sup_setb (_B1, _B2, _B3, _B4, _B7)
               (image ((ceq_term _A2 _B6), (ccompare_term _A1 _B5))
                 ((ceq_set (_B2, _B3, _B4.ccompare_cproper_interval)),
                   (ccompare_set (_B1, _B3, _B4, _B7)), set_impl_set)
                 (vars_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7))
                 (image
                   ((ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6)),
                     (ccompare_prod (ccompare_term _A1 _B5)
                       (ccompare_term _A1 _B5)))
                   ((ceq_term _A2 _B6), (ccompare_term _A1 _B5), set_impl_term)
                   snd (set ((ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6)),
                              (ccompare_prod (ccompare_term _A1 _B5)
                                (ccompare_term _A1 _B5)),
                              (set_impl_prod set_impl_term set_impl_term))
                         (drop (suc i) csa))))))))
    (showsl_lit "inlining not allowed in rhss of conditions"))
                                      (fun _ ->
check_inline_conds_rule (_A1, _A2, _A3) (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  (inline _B6 i rho ::
    removeAll
      (equal_prod (equal_prod (equal_term _A2 _B6) (equal_term _A2 _B6))
        (equal_list (equal_prod (equal_term _A2 _B6) (equal_term _A2 _B6))))
      rho r)
  (inline _B6 i rho)
  (map (fun (u, ab) -> (subst_apply_term u (subst _B6 x s), ab)) cs))))))))))
              b)
          (fun x ->
            Inl (comp (comp (comp (comp (comp
  (showsl_lit "error while inlining condition ") (showsl_eqa _A3 _B8 c))
                                    (showsl_lit " of rule "))
                              (showsl_crule _A3 _B8 _A3 _B8 rho))
                        (showsl_literal "\092"))
                  x));;

let rec check_inline_conds_ctrs (_A1, _A2, _A3)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  r x1 = match r, x1 with r, [] -> Inr r
    | ra, (r, cs) :: rcs ->
        catch_error
          (bindb
            (check
              (membera
                (equal_prod
                  (equal_prod (equal_term _A2 _B6) (equal_term _A2 _B6))
                  (equal_list
                    (equal_prod (equal_term _A2 _B6) (equal_term _A2 _B6))))
                ra r)
              (comp (showsl_crule _A3 _B8 _A3 _B8 r)
                (showsl_lit " does not occur in the input CTRS")))
            (fun _ ->
              bindb (check_inline_conds_rule (_A1, _A2, _A3)
                      (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) ra r cs)
                (fun (rb, raa) ->
                  check_inline_conds_ctrs (_A1, _A2, _A3)
                    (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
                    (raa ::
                      removeAll
                        (equal_prod
                          (equal_prod (equal_term _A2 _B6) (equal_term _A2 _B6))
                          (equal_list
                            (equal_prod (equal_term _A2 _B6)
                              (equal_term _A2 _B6))))
                        raa rb)
                    rcs)))
          (fun x ->
            Inl (comp (showsl_lit "error while inlining conditions\092") x));;

let rec check_inline_conds (_A1, _A2, _A3)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  ra r rcs =
    bindb (check_inline_conds_ctrs (_A1, _A2, _A3)
            (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) ra rcs)
      (fun raa ->
        catch_error
          (check_same_set
            (equal_prod (equal_prod (equal_term _A2 _B6) (equal_term _A2 _B6))
              (equal_list
                (equal_prod (equal_term _A2 _B6) (equal_term _A2 _B6))))
            r raa)
          (fun _ ->
            Inl (comp (comp (comp (comp (showsl_lit "error while inlining:")
                                    (showsl_lit
                                      "\092internally computed CTRS\092"))
                              (showsl_ctrs _A3 _B8 _A3 _B8 raa))
                        (showsl_lit "\092but certificate contains CTRS\092"))
                  (showsl_ctrs _A3 _B8 _A3 _B8 r))));;

let rec check_infeasible_rules (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9)
  r = (fun xs ->
        catch_error
          (forallM
            (fun (ra, ps) ->
              catch_error
                (check_infeasibleb (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
                  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9) r (snd ra) ps)
                (fun x ->
                  Inl (comp (comp (comp (showsl_lit "rule ")
                                    (showsl_crule _A9 _B9 _A9 _B9 ra))
                              (showsl_lit " is not infeasible\092"))
                        x)))
            xs)
          (fun x -> Inl (snd x)));;

let rec check_ccr_trans (_A1, _A2, _A3, _A4) (_B1, _B2, _B3)
  (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8, _C9)
  i ra x2 = match i, ra, x2 with
    i, ra, Inline_Conditions_CCRT (r, rcs) ->
      debug i "Inline Conditions"
        (bindb
          (check_inline_conds
            ((compare_lab _A1 _B1), (equal_lab _A3 _B2), (showl_lab _A4 _B3))
            (_C1, _C2, _C3, _C4, _C5.compare_compare_order, _C6, _C8, _C9) ra r
            rcs)
          (fun _ -> Inr r))
    | i, r, Infeasible_Rule_Removal_CCRT rps ->
        debug i "Infeasible Rule Removal"
          (bindb
            (check_infeasible_rules
              (finite_UNIV_lab, cenum_lab, (ceq_lab _A3 _B2),
                (cproper_interval_lab _A1 _B1), (compare_order_lab _A1 _B1),
                (default_lab _A2), (equal_lab _A3 _B2), set_impl_lab,
                (showl_lab _A4 _B3))
              (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8, _C9) r rps)
            (fun _ ->
              Inr (list_diff
                    (equal_prod
                      (equal_prod (equal_term (equal_lab _A3 _B2) _C6)
                        (equal_term (equal_lab _A3 _B2) _C6))
                      (equal_list
                        (equal_prod (equal_term (equal_lab _A3 _B2) _C6)
                          (equal_term (equal_lab _A3 _B2) _C6))))
                    r (map fst rps))));;

let rec check_redundant_rules_ncr (_A1, _A2) (_B1, _B2, _B3, _B4)
  ra r n =
    bindb (catch_error
            (check_subseteq
              (equal_prod (equal_term _A1 _B2) (equal_term _A1 _B2)) ra r)
            (fun _ ->
              Inl (showsl_lit "old TRS is not a subsystem of given TRS")))
      (fun _ ->
        (let s =
           list_diff (equal_prod (equal_term _A1 _B2) (equal_term _A1 _B2)) r ra
           in
         let t =
           list_diff (equal_prod (equal_term _A1 _B2) (equal_term _A1 _B2)) ra r
           in
          bindb (catch_error
                  (forallM
                    (fun (l, rb) ->
                      check (membera (equal_term _A1 _B2)
                              (reachable_terms _A1 (_B1, _B2, _B3) ra l n) rb)
                        (comp (showsl_lit "could not simulate rule ")
                          (showsl_rule _A2 _B4 (l, rb))))
                    s)
                  (fun x -> Inl (snd x)))
            (fun _ ->
              catch_error
                (forallM
                  (fun (l, raa) ->
                    check (membera (equal_term _A1 _B2)
                            (reachable_terms _A1 (_B1, _B2, _B3) r l n) raa)
                      (comp (showsl_lit "could not simulate rule ")
                        (showsl_rule _A2 _B4 (l, raa))))
                  t)
                (fun x -> Inl (snd x)))));;

let rec check_persistence_not_cr (_A1, _A2) (_B1, _B2, _B3, _B4)
  (_C1, _C2, _C3, _C4)
  siga r s =
    bindb (check_persistence1 (_A1, _A2) (_B1, _B2, _B3, _B4)
            (_C1, _C2, _C3, _C4) (_C1, _C2, _C3, _C4) siga r [s])
      (fun (ra, ss) ->
        (let sigF = mk_sigF _A1 siga in
         let types = map (comp snd snd) siga in
         let needed_types = needed_types_code _A2 (_B2, _B4) siga in
          catch_error
            (existsM
              (fun sa ->
                check_litsim_trs (_A1, _A2)
                  ((ccompare_prod _C1 _B1), (equal_prod _C2 _B2),
                    (mapping_impl_prod _C3 _B3), (showl_prod _C4 _B4))
                  sa (hda ss))
              (map (fun ty ->
                     (let tys = needed_types ty in
                       filtera
                         (fun rb -> membera _B2 tys (type_of_rule sigF rb)) ra))
                types))
            (fun _ ->
              Inl (comp (comp (showsl_lit
                                "persistent decomposition: new system is not induced by any type:")
                          (showsl_literal "\092"))
                    (showsl_trs _A2 _C4 s)))));;

let rec check_modularity_ncr (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2)
  ra r =
    bindb (catch_error
            (check_subseteq
              (equal_prod (equal_term _A4 _B1) (equal_term _A4 _B1)) r ra)
            (fun _ ->
              Inl (showsl_lit "new TRS is not a subsystem of given TRS")))
      (fun _ ->
        (let s =
           list_diff (equal_prod (equal_term _A4 _B1) (equal_term _A4 _B1)) ra r
           in
         let f = funas_trs_list r in
         let g = funas_trs_list s in
          bindb (check
                  (less_eq_set
                    ((cenum_prod _A1 cenum_nat), (ceq_prod _A2 ceq_nat),
                      (ccompare_prod _A3 ccompare_nat))
                    (inf_seta
                      ((ceq_prod _A2 ceq_nat), (ccompare_prod _A3 ccompare_nat))
                      (set ((ceq_prod _A2 ceq_nat),
                             (ccompare_prod _A3 ccompare_nat),
                             (set_impl_prod _A5 set_impl_nat))
                        f)
                      (set ((ceq_prod _A2 ceq_nat),
                             (ccompare_prod _A3 ccompare_nat),
                             (set_impl_prod _A5 set_impl_nat))
                        g))
                    (bot_set
                      ((ceq_prod _A2 ceq_nat), (ccompare_prod _A3 ccompare_nat),
                        (set_impl_prod _A5 set_impl_nat))))
                  (showsl_lit "signatures are not disjoint"))
            (fun _ ->
              bindb (check_varcond_subset _A6 (_B1, _B2) r)
                (fun _ ->
                  catch_error
                    (catch_error
                      (forallM
                        (fun x ->
                          (if (let (l, _) = x in not (is_Var l)) then Inr ()
                            else Inl x))
                        s)
                      (fun x -> Inl (snd x)))
                    (fun _ ->
                      Inl (showsl_lit "lhss must not be variables"))))));;

let rec check_non_cr (_A1, _A2, _A3, _A4)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  r s seq1 seq2 reason =
    (let chk =
       check_rsteps_last
         ((equal_lab _A3 (equal_list equal_nat)),
           (showl_lab _A4 (showl_list showl_nat)))
         ((ccompare_list ccompare_char), (equal_list equal_char),
           mapping_impl_list, (showl_list showl_char))
         r s
       in
      bindb (chk seq1)
        (fun _ ->
          bindb (chk seq2)
            (fun _ ->
              check_non_join (_A1, _A2, _A3, _A4)
                (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) r r (rseq_last s seq1)
                (rseq_last s seq2) reason)));;

let rec check_ncr_proof (_B1, _B2, _B3, _B4, _B5)
  a ia i j r x5 = match a, ia, i, j, r, x5 with
    a, ia, i, j, r, SN_NWCR prf ->
      debug ia "SN_NWCR"
        (let tp = mkc i false [] r [] in
          bindb (catch_error
                  (check_trs_termination_proof (_B1, _B2, _B4, _B5) i j a
                    (add_index ia one_nata) tp prf)
                  (fun x ->
                    Inl (comp (comp ia
                                (showsl_lit
                                  ": error below strong normalization + wcr\092"))
                          x)))
            (fun _ ->
              catch_error
                (check
                  (not (isOK (check_critical_pairs_NF
                               ((equal_lab _B4 (equal_list equal_nat)),
                                 (showl_lab _B5 (showl_list showl_nat)))
                               ((ccompare_list ccompare_char),
                                 (equal_list equal_char), mapping_impl_list,
                                 (showl_list showl_char))
                               r (critical_pairs_impl
                                   (equal_lab _B4 (equal_list equal_nat)) r
                                   r))))
                  (showsl_lit "all critical pairs are joinable"))
                (fun x ->
                  Inl (comp (comp (comp (comp ia
  (showsl_lit ": error when disproving local confluence of "))
                                    (showsl_tp
                                      (showl_lab _B5 (showl_list showl_nat))
                                      (showl_list showl_char) i tp))
                              (showsl_literal "\092"))
                        x))))
    | a, ia, i, j, r, Non_Join (s, seq1, seq2, prf) ->
        debug ia "Non_Join"
          (catch_error
            (check_non_cr (_B1, _B3, _B4, _B5)
              (card_UNIV_list, cenum_list, (ceq_list ceq_char),
                (cproper_interval_list ccompare_char),
                (compare_order_list (compare_order_char, equal_char)),
                (equal_list equal_char), set_impl_list, (showl_list showl_char))
              r s seq1 seq2 prf)
            (fun x ->
              Inl (comp (comp (comp (comp ia
                                      (showsl_lit
": error when disproving CR of "))
                                (showsl_trs
                                  (showl_lab _B5 (showl_list showl_nat))
                                  (showl_list showl_char) r))
                          (showsl_literal "\092"))
                    x)))
    | a, ia, i, j, ra, NCR_Disj_Subtrs (r, prf) ->
        debug ia "Modularity"
          (bindb
            (catch_error
              (check_modularity_ncr
                (cenum_lab, (ceq_lab _B4 (equal_list equal_nat)),
                  (ccompare_lab _B1.compare_compare_order
                    (compare_list compare_nat)),
                  (equal_lab _B4 (equal_list equal_nat)), set_impl_lab,
                  (showl_lab _B5 (showl_list showl_nat)))
                ((equal_list equal_char), (showl_list showl_char)) ra r)
              (fun x ->
                Inl (comp (comp (comp (comp ia
(showsl_lit ": error when applying modularity to switch to "))
                                  (showsl_trs
                                    (showl_lab _B5 (showl_list showl_nat))
                                    (showl_list showl_char) r))
                            (showsl_literal "\092"))
                      x)))
            (fun _ ->
              catch_error
                (check_ncr_proof (_B1, _B2, _B3, _B4, _B5) a
                  (add_index ia one_nata) i j r prf)
                (fun x ->
                  Inl (comp (comp ia
                              (showsl_lit
                                ": error below the modular decomposition\092"))
                        x))))
    | a, ia, i, j, r, NCR_Redundant_Rules (rs, n, prf) ->
        debug ia "Redundant Rules"
          (bindb
            (catch_error
              (check_ncr_proof (_B1, _B2, _B3, _B4, _B5) a
                (add_index ia one_nata) i j rs prf)
              (fun x ->
                Inl (comp (comp (comp ia
                                  (showsl_lit
                                    ": error when proving nonconfluence of modified TRS\092"))
                            (showsl_trs (showl_lab _B5 (showl_list showl_nat))
                              (showl_list showl_char) rs))
                      x)))
            (fun _ ->
              catch_error
                (check_redundant_rules_ncr
                  ((equal_lab _B4 (equal_list equal_nat)),
                    (showl_lab _B5 (showl_list showl_nat)))
                  ((ccompare_list ccompare_char), (equal_list equal_char),
                    mapping_impl_list, (showl_list showl_char))
                  r rs n)
                (fun x ->
                  Inl (comp (comp (comp ia
                                    (showsl_lit
                                      ": error in checking redundant rules transformation of the TRS\092"))
                              (showsl_trs (showl_lab _B5 (showl_list showl_nat))
                                (showl_list showl_char) r))
                        x))))
    | a, ia, i, j, r, NCR_Persistent_Decomposition (siga, s, prf) ->
        debug ia "Persistent Decomposition"
          (bindb
            (catch_error
              (check_ncr_proof (_B1, _B2, _B3, _B4, _B5) a
                (add_index ia one_nata) i j s prf)
              (fun x ->
                Inl (comp (comp (comp ia
                                  (showsl_lit
                                    ": error while proving nonconfluence of resulting TRS\092"))
                            (showsl_trs (showl_lab _B5 (showl_list showl_nat))
                              (showl_list showl_char) s))
                      x)))
            (fun _ ->
              catch_error
                (check_persistence_not_cr
                  ((equal_lab _B4 (equal_list equal_nat)),
                    (showl_lab _B5 (showl_list showl_nat)))
                  ((ccompare_list ccompare_char), (equal_list equal_char),
                    mapping_impl_list, (showl_list showl_char))
                  ((ccompare_list ccompare_char), (equal_list equal_char),
                    mapping_impl_list, (showl_list showl_char))
                  siga r s)
                (fun x ->
                  Inl (comp (comp (comp ia
                                    (showsl_lit
                                      ": error in checking persistent decomposition of "))
                              (showsl_trs (showl_lab _B5 (showl_list showl_nat))
                                (showl_list showl_char) r))
                        x))));;

let rec check_conditional_ncr_proof (_B1, _B2, _B3, _B4, _B5)
  a ia i j r x5 = match a, ia, i, j, r, x5 with
    a, ia, i, j, r, Unconditional_CNCR prf ->
      debug ia "Unconditional"
        (bindb
          (catch_error
            (catch_error
              (forallM (fun x -> (if null (snd x) then Inr () else Inl x)) r)
              (fun x -> Inl (snd x)))
            (fun x ->
              Inl (comp (comp (showsl_lit "rule with non-empty conditions")
                          (showsl_literal "\092"))
                    (showsl_crule (showl_lab _B5 (showl_list showl_nat))
                      (showl_list showl_char)
                      (showl_lab _B5 (showl_list showl_nat))
                      (showl_list showl_char) x))))
          (fun _ ->
            check_ncr_proof (_B1, _B2, _B3, _B4, _B5) a ia i j (map fst r) prf))
    | a, ia, i, j, r, Transformation_CNCR (t, prf) ->
        debug ia "CR Preserving Transformation"
          (bindb
            (check_ccr_trans (_B1.compare_compare_order, _B3, _B4, _B5)
              ((compare_list compare_nat), (equal_list equal_nat),
                (showl_list showl_nat))
              (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
                (cproper_interval_list ccompare_char),
                (compare_order_list (compare_order_char, equal_char)),
                (equal_list equal_char), mapping_impl_list, set_impl_list,
                (showl_list showl_char))
              ia r t)
            (fun ra ->
              check_conditional_ncr_proof (_B1, _B2, _B3, _B4, _B5) a ia i j ra
                prf))
    | a, ia, i, j, r, Non_Join_CNCR (s, t, u, ps, qs, prf) ->
        debug ia "Conditional Non-Joinability"
          (catch_error
            (check_conditional_non_cr (_B1, _B3, _B4, _B5)
              (card_UNIV_list, cenum_list, (ceq_list ceq_char),
                (cproper_interval_list ccompare_char),
                (compare_order_list (compare_order_char, equal_char)),
                (equal_list equal_char), set_impl_list, (showl_list showl_char))
              r s t u ps qs prf)
            (fun x ->
              Inl (comp (comp (comp (comp ia
                                      (showsl_lit
"error when disproving CR of "))
                                (showsl_ctrs
                                  (showl_lab _B5 (showl_list showl_nat))
                                  (showl_list showl_char)
                                  (showl_lab _B5 (showl_list showl_nat))
                                  (showl_list showl_char) r))
                          (showsl_literal "\092"))
                    x)));;

let rec check_subsumptions_guided (_A1, _A2) (_B1, _B2, _B3, _B4)
  e x1 = match e, x1 with e, [] -> Inr ()
    | ea, (e, seq) :: convs ->
        bindb (catch_error
                (check_conversiona (_A1, _A2) (_B1, _B2, _B3, _B4) ea seq
                  (fst e) (snd e))
                (fun x ->
                  Inl (comp (comp (comp (showsl_lit
  "problem in conversion for equation ")
                                    (showsl_eq _A2 _B4 e))
                              (showsl_literal "\092"))
                        x)))
          (fun _ ->
            check_subsumptions_guided (_A1, _A2) (_B1, _B2, _B3, _B4) (e :: ea)
              convs);;

let rec check_convertible_instance (_A1, _A2) (_B1, _B2, _B3, _B4)
  eq e convs =
    bindb (catch_error
            (existsM
              (fun c ->
                check (instance_rule _A1 _B2 (_B1, _B2, _B3) (fst c) eq)
                  (comp (showsl_lit "not an instance of ")
                    (showsl_eq _A2 _B4 eq)))
              convs)
            (fun _ -> Inl (showsl_lit " no instance found")))
      (fun _ ->
        check_subsumptions_guided (_A1, _A2) (_B1, _B2, _B3, _B4) e convs);;

let rec check_subsumption_guided (_A1, _A2) (_B1, _B2, _B3, _B4)
  ea e convs =
    bindb (catch_error
            (check_subseteq
              (equal_prod (equal_term _A1 _B2) (equal_term _A1 _B2)) ea
              (map fst convs))
            (fun x ->
              Inl (comp (showsl_lit "could not find conversion for equation ")
                    (showsl_eq _A2 _B4 x))))
      (fun _ ->
        check_subsumptions_guided (_A1, _A2) (_B1, _B2, _B3, _B4) e convs);;

let rec check_subsumption_NF (_A1, _A2) (_B1, _B2, _B3, _B4)
  e r = catch_error
          (forallM
            (fun ea ->
              catch_error
                (check_join_NF (_A1, _A2) (_B1, _B2, _B3, _B4) r (fst ea)
                  (snd ea))
                (fun x ->
                  Inl (comp (comp (comp (showsl_lit "could not join equation ")
                                    (showsl_eq _A2 _B4 ea))
                              (showsl_literal "\092"))
                        x)))
            e)
          (fun x -> Inl (snd x));;

let rec check_subsumption (_A1, _A2) (_B1, _B2, _B3, _B4)
  e r convs_o =
    (match convs_o
      with None -> check_subsumption_NF (_A1, _A2) (_B1, _B2, _B3, _B4) e r
      | Some a ->
        check_subsumption_guided (_A1, _A2) (_B1, _B2, _B3, _B4) e r a);;

let rec check_completion_proof (_B1, _B2, _B3, _B4)
  a ia i j e r (SN_WCR_Eq (joins_i, prf, conv1, conv2)) =
    debug ia "SN_WCR_Eq"
      (let tp = mkc i false [] r [] in
        bindb (catch_error
                (check_trs_termination_proof (_B1, _B2, _B3, _B4) i j a
                  (add_index ia one_nata) tp prf)
                (fun x ->
                  Inl (comp (comp ia
                              (showsl_lit
                                ": error below strong normalization + wcr\092"))
                        x)))
          (fun _ ->
            bindb (catch_error
                    (check_subsumption_guided
                      ((equal_lab _B3 (equal_list equal_nat)),
                        (showl_lab _B4 (showl_list showl_nat)))
                      ((ccompare_list ccompare_char), (equal_list equal_char),
                        mapping_impl_list, (showl_list showl_char))
                      r e conv1)
                    (fun x ->
                      Inl (comp (comp ia
                                  (showsl_lit
                                    ": error when showing that rewrite relation can be simulated by equations\092"))
                            x)))
              (fun _ ->
                bindb (catch_error
                        (check_subsumption
                          ((equal_lab _B3 (equal_list equal_nat)),
                            (showl_lab _B4 (showl_list showl_nat)))
                          ((ccompare_list ccompare_char),
                            (equal_list equal_char), mapping_impl_list,
                            (showl_list showl_char))
                          e r conv2)
                        (fun x ->
                          Inl (comp (comp ia
                                      (showsl_lit
": error when showing that equations can be simulated by rewrite system\092"))
                                x)))
                  (fun _ ->
                    catch_error
                      (check_critical_pairs
                        ((equal_lab _B3 (equal_list equal_nat)),
                          (showl_lab _B4 (showl_list showl_nat)))
                        r (critical_pairs_impl
                            (equal_lab _B3 (equal_list equal_nat)) r r)
                        joins_i)
                      (fun x ->
                        Inl (comp (comp (comp
  (comp ia (showsl_lit ": error when proving local confluence of \092"))
  (showsl_tp (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char) i
    tp))
                                    (showsl_literal "\092"))
                              x))))));;

let rec check_equational_disproof (_B1, _B2, _B3, _B4)
  a ia i j e x5 x6 = match a, ia, i, j, e, x5, x6 with
    a, ia, i, j, e, Equation eq, Completion_and_Normalization_Different (r, p)
      -> debug ia "Completion_and_Normalization"
           (bindb (check_completion_proof (_B1, _B2, _B3, _B4) a ia i j e r p)
             (fun _ ->
               (let s = fst eq in
                let t = snd eq in
                 (match
                   (compute_rstep_NF (equal_lab _B3 (equal_list equal_nat))
                      ((ccompare_list ccompare_char), (equal_list equal_char),
                        mapping_impl_list)
                      r s,
                     compute_rstep_NF (equal_lab _B3 (equal_list equal_nat))
                       ((ccompare_list ccompare_char), (equal_list equal_char),
                         mapping_impl_list)
                       r t)
                   with (None, _) ->
                     Inl (comp (comp (comp (showsl_lit
     "error when computing normal forms of ")
                                       (showsl_terma
 (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char) s))
                                 (showsl_lit " and "))
                           (showsl_terma (showl_lab _B4 (showl_list showl_nat))
                             (showl_list showl_char) t))
                   | (Some _, None) ->
                     Inl (comp (comp (comp (showsl_lit
     "error when computing normal forms of ")
                                       (showsl_terma
 (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char) s))
                                 (showsl_lit " and "))
                           (showsl_terma (showl_lab _B4 (showl_list showl_nat))
                             (showl_list showl_char) t))
                   | (Some sa, Some ta) ->
                     (if not (equal_terma (equal_lab _B3 (equal_list equal_nat))
                               (equal_list equal_char) sa ta)
                       then Inr ()
                       else Inl (comp (comp
(comp (comp (showsl_terma (showl_lab _B4 (showl_list showl_nat))
              (showl_list showl_char) s)
        (showsl_lit " and "))
  (showsl_terma (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char)
    t))
(showsl_lit " have same normal form "))
                                  (showsl_terma
                                    (showl_lab _B4 (showl_list showl_nat))
                                    (showl_list showl_char) sa)))))))
    | a, ia, i, j, e, Inequality eq, Convertible_Instance convs ->
        debug ia "Convertible_Instance"
          (check_convertible_instance
            ((equal_lab _B3 (equal_list equal_nat)),
              (showl_lab _B4 (showl_list showl_nat)))
            ((ccompare_list ccompare_char), (equal_list equal_char),
              mapping_impl_list, (showl_list showl_char))
            eq e convs)
    | a, ia, i, j, e, Inequality v,
        Completion_and_Normalization_Different (va, vb)
        -> Inl (showsl_lit "unsupported disproof")
    | a, ia, i, j, e, Equation va, Convertible_Instance v ->
        Inl (showsl_lit "unsupported disproof");;

let rec match_crule _A (_B1, _B2, _B3)
  rho_1 rho_2 =
    bind (zip_option
           (fst (fst rho_2) ::
             snd (fst rho_2) :: map fst (snd rho_2) @ map snd (snd rho_2))
           (fst (fst rho_1) ::
             snd (fst rho_1) :: map fst (snd rho_1) @ map snd (snd rho_1)))
      (match_list (_B1, _B2, _B3) _A _B2 (fun a -> Var a));;

let rec check_overlapa (_A1, _A2, _A3, _A4, _A5) (_B1, _B2)
  xvar yvar check_infeasible r rho_1 rho_2 p =
    catch_error
      (match
        mgu_var_disjoint_generic _A2 _B1 xvar yvar (subt_at (fst (fst rho_1)) p)
          (fst (fst rho_2))
        with None -> Inr ()
        | Some (sigma_1, sigma_2) ->
          catch_error
            (choice
              [check (null p &&
                       equal_terma _B1 _A2
                         (subst_apply_term (snd (fst rho_1)) sigma_1)
                         (subst_apply_term (snd (fst rho_2)) sigma_2))
                 (showsl_lit "is not a trivial root-overlap");
                check (null p &&
                        (not (is_none
                               (match_crule _B1 (_A1, _A2, _A3) rho_1 rho_2)) &&
                          not (is_none
                                (match_crule _B1 (_A1, _A2, _A3) rho_2 rho_1))))
                  (showsl_lit
                    "is not a root-overlap of variants of the same rule");
                catch_error
                  (check_infeasible r (subst_list sigma_1 (snd rho_1))
                    (subst_list sigma_2 (snd rho_2)))
                  (fun x ->
                    Inl (comp (showsl_lit
                                "could not be shown to be infeasible\092")
                          x))])
            (fun x -> Inl (showsl_sep id (showsl_literal "\092") x)))
      (fun x ->
        Inl (comp (comp (comp (showsl_lit "the ")
                          (showsl_coverlap _B2 _A5 _B2 _A5 _B2 _A5 _B2 _A5 rho_1
                            rho_2 p))
                    (showsl_literal "\092"))
              x));;

let rec check_ao_infeasiblea (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9)
  r cs_1 cs_2 x3 = match r, cs_1, cs_2, x3 with
    r, cs_1, cs_2, AO_Infeasibility_Proof p ->
      check_infeasibleb (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
        (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9) r (cs_1 @ cs_2) p
    | r, cs_1, cs_2, AO_Lhss_Equal (s, t, u, p) ->
        bindb (check
                (membera (equal_prod (equal_term _A7 _B6) (equal_term _A7 _B6))
                  cs_1 (s, t))
                (comp (comp (showsl_eqa _A9 _B9 (s, t))
                        (showsl_lit " is not an equation in "))
                  (showsl_conditions _A9 _B9 cs_1)))
          (fun _ ->
            bindb (check
                    (membera
                      (equal_prod (equal_term _A7 _B6) (equal_term _A7 _B6))
                      cs_2 (s, u))
                    (comp (comp (showsl_eqa _A9 _B9 (s, u))
                            (showsl_lit " is not an equation in "))
                      (showsl_conditions _A9 _B9 cs_2)))
              (fun _ ->
                check_nonjoinable (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
                  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9) (map fst r) t u
                  p));;

let rec check_ao_infeasible (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9)
  css r cs_1 cs_2 =
    catch_error
      (existsM
        (fun (cs_1a, (cs_2a, p)) ->
          (let cs = cs_1 @ cs_2 in
           let csa = cs_1a @ cs_2a in
            bindb (check
                    (equal_nata (size_list cs_1a) (size_list cs_1) &&
                      equal_nata (size_list cs_2a) (size_list cs_2))
                    (showsl_lit "lengths differ"))
              (fun _ ->
                bindb (check
                        (not (is_none
                               (match_rules _A7
                                 (_B4.ccompare_cproper_interval, _B6, _B7) csa
                                 cs)) &&
                          not (is_none
                                (match_rules _A7
                                  (_B4.ccompare_cproper_interval, _B6, _B7) cs
                                  csa)))
                        id)
                  (fun _ ->
                    check_ao_infeasiblea
                      (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
                      (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9) r cs_1a
                      cs_2a p))))
        css)
      (fun x -> Inl (showsl_sep id (showsl_literal "\092") x));;

let rec check_almost_orthogonal_modulo_infeasibilitya
  (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
  css r =
    bindb (check_left_linear_trs _A9
            ((ceq_list ceq_char), (ccompare_list ccompare_char), set_impl_list,
              (showl_list showl_char))
            (map fst r))
      (fun _ ->
        catch_error
          (forallM
            (fun rho_1 ->
              (let l_1 = fst (fst rho_1) in
                catch_error
                  (forallM
                    (fun rho_2 ->
                      catch_error
                        (forallM
                          (check_overlapa
                            ((ccompare_list ccompare_char),
                              (equal_list equal_char), mapping_impl_list,
                              infinite_list, (showl_list showl_char))
                            (_A7, _A9)
                            (fun a ->
                              Chara (false, false, false, true, true, true,
                                      true, false) ::
                                a)
                            (fun a ->
                              Chara (true, false, false, true, true, true, true,
                                      false) ::
                                a)
                            (check_ao_infeasible
                              (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
                              (finite_UNIV_list, cenum_list,
                                (ceq_list ceq_char),
                                (cproper_interval_list ccompare_char),
                                (compare_order_list
                                  (compare_order_char, equal_char)),
                                (equal_list equal_char), mapping_impl_list,
                                set_impl_list, (showl_list showl_char))
                              css)
                            r rho_1 rho_2)
                          (funposs_list l_1))
                        (fun x -> Inl (snd x)))
                    r)
                  (fun x -> Inl (snd x))))
            r)
          (fun x -> Inl (snd x)));;

let rec x_vars (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  rho i =
    sup_seta (_B3, _B4.ccompare_cproper_interval)
      (vars_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7) (fst (fst rho)))
      (sup_setb (_B1, _B2, _B3, _B4, _B7)
        (image ((ceq_term _A2 _B6), (ccompare_term _A1 _B5))
          ((ceq_set (_B2, _B3, _B4.ccompare_cproper_interval)),
            (ccompare_set (_B1, _B3, _B4, _B7)), set_impl_set)
          (vars_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7))
          (image
            ((ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6)),
              (ccompare_prod (ccompare_term _A1 _B5) (ccompare_term _A1 _B5)))
            ((ceq_term _A2 _B6), (ccompare_term _A1 _B5), set_impl_term) snd
            (set ((ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6)),
                   (ccompare_prod (ccompare_term _A1 _B5)
                     (ccompare_term _A1 _B5)),
                   (set_impl_prod set_impl_term set_impl_term))
              (take i (snd rho))))));;

let rec extended_properly_oriented (_A1, _A2)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  r = ball ((ceq_prod (ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6))
              (ceq_list (ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6)))),
             (ccompare_prod
               (ccompare_prod (ccompare_term _A1 _B5) (ccompare_term _A1 _B5))
               (ccompare_list
                 (ccompare_prod (ccompare_term _A1 _B5)
                   (ccompare_term _A1 _B5)))))
        r (fun rho ->
            less_eq_set (_B2, _B3, _B4.ccompare_cproper_interval)
              (vars_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
                (snd (fst rho)))
              (vars_term (_A1, _A2) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
                (fst (fst rho))) ||
              not (all_interval_nat
                    (comp not
                      (fun m ->
                        all_interval_nat
                          (fun i ->
                            less_eq_set
                              (_B2, _B3, _B4.ccompare_cproper_interval)
                              (vars_term (_A1, _A2)
                                (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
                                (fst (nth (snd rho) i)))
                              (x_vars (_A1, _A2)
                                (_B1, _B2, _B3, _B4, _B5, _B6, _B7) rho i))
                          zero_nata m &&
                          all_interval_nat
                            (fun i ->
                              less_eq_set
                                (_B2, _B3, _B4.ccompare_cproper_interval)
                                (inf_seta (_B3, _B4.ccompare_cproper_interval)
                                  (vars_term (_A1, _A2)
                                    (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
                                    (snd (fst rho)))
                                  (vars_rule (_A1, _A2)
                                    (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
                                    (nth (snd rho) i)))
                                (x_vars (_A1, _A2)
                                  (_B1, _B2, _B3, _B4, _B5, _B6, _B7) rho m))
                            m (size_list (snd rho))))
                    zero_nata (suc (size_list (snd rho)))));;

let rec check_extended_properly_oriented (_A1, _A2, _A3)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  r = check (extended_properly_oriented (_A1, _A2)
              (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
              (set ((ceq_prod (ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6))
                      (ceq_list
                        (ceq_prod (ceq_term _A2 _B6) (ceq_term _A2 _B6)))),
                     (ccompare_prod
                       (ccompare_prod (ccompare_term _A1 _B5)
                         (ccompare_term _A1 _B5))
                       (ccompare_list
                         (ccompare_prod (ccompare_term _A1 _B5)
                           (ccompare_term _A1 _B5)))),
                     (set_impl_prod (set_impl_prod set_impl_term set_impl_term)
                       set_impl_list))
                r))
        (showsl_lit "the given CTRS is not extended properly oriented\092");;

let rec funas_crule (_A1, _A2, _A3, _A4, _A5, _A6, _A7) (_B1, _B2)
  rho = sup_seta
          ((ceq_prod _A3 ceq_nat),
            (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat))
          (funas_rule (_A3, _A4.ccompare_cproper_interval, _A7) (fst rho))
          (funas_trs (_A1, _A2, _A3, _A4, _A5, _A6, _A7) (_B1, _B2)
            (set ((ceq_prod (ceq_term _A6 _B2) (ceq_term _A6 _B2)),
                   (ccompare_prod (ccompare_term _A5 _B1)
                     (ccompare_term _A5 _B1)),
                   (set_impl_prod set_impl_term set_impl_term))
              (snd rho)));;

let rec funas_ctrs (_A1, _A2, _A3, _A4, _A5, _A6, _A7) (_B1, _B2)
  r = sup_setb
        ((finite_UNIV_prod _A1 finite_UNIV_nat), (cenum_prod _A2 cenum_nat),
          (ceq_prod _A3 ceq_nat),
          (cproper_interval_prod _A4 cproper_interval_nat),
          (set_impl_prod _A7 set_impl_nat))
        (image
          ((ceq_prod (ceq_prod (ceq_term _A6 _B2) (ceq_term _A6 _B2))
             (ceq_list (ceq_prod (ceq_term _A6 _B2) (ceq_term _A6 _B2)))),
            (ccompare_prod
              (ccompare_prod (ccompare_term _A5 _B1) (ccompare_term _A5 _B1))
              (ccompare_list
                (ccompare_prod (ccompare_term _A5 _B1)
                  (ccompare_term _A5 _B1)))))
          ((ceq_set
             ((cenum_prod _A2 cenum_nat), (ceq_prod _A3 ceq_nat),
               (cproper_interval_prod _A4
                 cproper_interval_nat).ccompare_cproper_interval)),
            (ccompare_set
              ((finite_UNIV_prod _A1 finite_UNIV_nat), (ceq_prod _A3 ceq_nat),
                (cproper_interval_prod _A4 cproper_interval_nat),
                (set_impl_prod _A7 set_impl_nat))),
            set_impl_set)
          (funas_crule (_A1, _A2, _A3, _A4, _A5, _A6, _A7) (_B1, _B2)) r);;

let rec check_constructor_term (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8)
  (_B1, _B2, _B3)
  s r = check (less_eq_set
                ((cenum_prod _A2 cenum_nat), (ceq_prod _A3 ceq_nat),
                  (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat))
                (funas_term (_A3, _A4.ccompare_cproper_interval, _A7) s)
                (minus_set
                  ((ceq_prod _A3 ceq_nat),
                    (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat))
                  (funas_ctrs (_A1, _A2, _A3, _A4, _A5, _A6, _A7) (_B1, _B2)
                    (set ((ceq_prod
                            (ceq_prod (ceq_term _A6 _B2) (ceq_term _A6 _B2))
                            (ceq_list
                              (ceq_prod (ceq_term _A6 _B2)
                                (ceq_term _A6 _B2)))),
                           (ccompare_prod
                             (ccompare_prod (ccompare_term _A5 _B1)
                               (ccompare_term _A5 _B1))
                             (ccompare_list
                               (ccompare_prod (ccompare_term _A5 _B1)
                                 (ccompare_term _A5 _B1)))),
                           (set_impl_prod
                             (set_impl_prod set_impl_term set_impl_term)
                             set_impl_list))
                      r))
                  (set ((ceq_prod _A3 ceq_nat),
                         (ccompare_prod _A4.ccompare_cproper_interval
                           ccompare_nat),
                         (set_impl_prod _A7 set_impl_nat))
                    (defined_list (map fst r)))))
          (comp (comp (showsl_lit "the term ") (showsl_terma _A8 _B3 s))
            (showsl_lit " is not a constructor term\092"));;

let rec check_Ru_NF (_A1, _A2, _A3) (_B1, _B2, _B3, _B4)
  s r = check (is_NF_trs (_A1, _A2) (_B1, _B2, _B3) (map fst r) s)
          (comp (comp (showsl_lit "the term ") (showsl_terma _A3 _B4 s))
            (showsl_lit " is not an Ru normal form\092"));;

let rec check_linear_term _A (_B1, _B2, _B3, _B4)
  s = check (linear_term (_B1, _B2, _B3) s)
        (comp (comp (showsl_literal "the term ") (showsl_terma _A _B4 s))
          (showsl_literal " is not linear\092"));;

let rec check_ground_term _A _B
  s = check (ground s)
        (comp (comp (showsl_literal "the term ") (showsl_terma _A _B s))
          (showsl_literal " is not a ground term\092"));;

let rec check_right_stable (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  r = catch_error
        (catch_error
          (forallM
            (fun ra ->
              catch_error
                (forallM
                  (fun i ->
                    (let t_i = snd (nth (snd ra) i) in
                      bindb (catch_error
                              (check_disjoint _B4
                                (vars_term_list (fst (fst ra)) @
                                  maps (fun (s, _) -> vars_term_list s)
                                    (take (suc i) (snd ra)) @
                                    maps (fun (_, a) -> vars_term_list a)
                                      (take i (snd ra)))
                                (vars_term_list t_i))
                              (fun x ->
                                Inl (comp (comp
    (comp (comp (showsl_lit "variable ") (showsl _B7 x))
      (showsl_lit " in rhs of condition "))
    (showsl_nat i))
                                      (showsl_lit " is not fresh\092"))))
                        (fun _ ->
                          catch_error
                            (choice
                              [bindb (check_linear_term _A8 (_B1, _B2, _B6, _B7)
                                       t_i)
                                 (fun _ ->
                                   check_constructor_term
                                     (_A1, _A2, _A3, _A4,
                                       _A5.compare_compare_order, _A6, _A7, _A8)
                                     (_B3, _B4, _B7) t_i r);
                                bindb (check_ground_term _A8 _B7 t_i)
                                  (fun _ ->
                                    check_Ru_NF (_A5, _A6, _A8)
                                      (_B2, _B4, _B5, _B7) t_i r)])
                            (fun x ->
                              Inl (showsl_sep id (showsl_literal "\092") x)))))
                  (upt zero_nata (size_list (snd ra))))
                (fun x -> Inl (snd x)))
            r)
          (fun x -> Inl (snd x)))
        (fun x ->
          Inl (comp (showsl_lit "the CTRS is not right stable\092") x));;

let rec check_level_confluence_modulo_infeasibilitya
  (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
  css r =
    bindb (check_varcond_no_Var_lhs _A9 (showl_list showl_char) (map fst r))
      (fun _ ->
        bindb (check_type3 _A9
                ((equal_list equal_char), (showl_list showl_char)) r)
          (fun _ ->
            bindb (check_extended_properly_oriented
                    (_A5.compare_compare_order, _A7, _A9)
                    (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
                      (cproper_interval_list ccompare_char),
                      (compare_list compare_char), (equal_list equal_char),
                      set_impl_list, (showl_list showl_char))
                    r)
              (fun _ ->
                bindb (check_right_stable
                        (_A1, _A2, _A3, _A4, _A5, _A7, _A8, _A9)
                        ((ceq_list ceq_char), (ccompare_list ccompare_char),
                          (compare_list compare_char), (equal_list equal_char),
                          mapping_impl_list, set_impl_list,
                          (showl_list showl_char))
                        r)
                  (fun _ ->
                    check_almost_orthogonal_modulo_infeasibilitya
                      (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9) css r))));;

let rec trancl_of_list _A
  xs = maps (fun x -> map (fun a -> (x, a)) (trancl_list_impl _A xs [x]))
         (map fst xs);;

let rec check_infeasiblea (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9)
  css r cs_1 cs_2 =
    catch_error
      (existsM
        (fun (cs, p) ->
          (let csa = cs_1 @ cs_2 in
            bindb (check
                    (not (is_none
                           (match_rules _A7
                             (_B4.ccompare_cproper_interval, _B6, _B7) cs
                             csa)) &&
                      not (is_none
                            (match_rules _A7
                              (_B4.ccompare_cproper_interval, _B6, _B7) csa
                              cs)))
                    id)
              (fun _ ->
                check_infeasibleb (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
                  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8, _B9) r
                  (trancl_of_list (equal_term _A7 _B6) cs) p)))
        css)
      (fun x -> Inl (showsl_sep id (showsl_literal "\092") x));;

let rec check_almost_orthogonal_modulo_infeasibility
  (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
  css r =
    bindb (check_left_linear_trs _A9
            ((ceq_list ceq_char), (ccompare_list ccompare_char), set_impl_list,
              (showl_list showl_char))
            (map fst r))
      (fun _ ->
        catch_error
          (forallM
            (fun rho_1 ->
              (let l_1 = fst (fst rho_1) in
                catch_error
                  (forallM
                    (fun rho_2 ->
                      catch_error
                        (forallM
                          (check_overlapa
                            ((ccompare_list ccompare_char),
                              (equal_list equal_char), mapping_impl_list,
                              infinite_list, (showl_list showl_char))
                            (_A7, _A9)
                            (fun a ->
                              Chara (false, false, false, true, true, true,
                                      true, false) ::
                                a)
                            (fun a ->
                              Chara (true, false, false, true, true, true, true,
                                      false) ::
                                a)
                            (check_infeasiblea
                              (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
                              (finite_UNIV_list, cenum_list,
                                (ceq_list ceq_char),
                                (cproper_interval_list ccompare_char),
                                (compare_order_list
                                  (compare_order_char, equal_char)),
                                (equal_list equal_char), mapping_impl_list,
                                set_impl_list, (showl_list showl_char))
                              css)
                            r rho_1 rho_2)
                          (funposs_list l_1))
                        (fun x -> Inl (snd x)))
                    r)
                  (fun x -> Inl (snd x))))
            r)
          (fun x -> Inl (snd x)));;

let rec check_level_confluence_modulo_infeasibility
  (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9)
  css r =
    bindb (check_varcond_no_Var_lhs _A9 (showl_list showl_char) (map fst r))
      (fun _ ->
        bindb (check_type3 _A9
                ((equal_list equal_char), (showl_list showl_char)) r)
          (fun _ ->
            bindb (check_extended_properly_oriented
                    (_A5.compare_compare_order, _A7, _A9)
                    (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
                      (cproper_interval_list ccompare_char),
                      (compare_list compare_char), (equal_list equal_char),
                      set_impl_list, (showl_list showl_char))
                    r)
              (fun _ ->
                bindb (check_right_stable
                        (_A1, _A2, _A3, _A4, _A5, _A7, _A8, _A9)
                        ((ceq_list ceq_char), (ccompare_list ccompare_char),
                          (compare_list compare_char), (equal_list equal_char),
                          mapping_impl_list, set_impl_list,
                          (showl_list showl_char))
                        r)
                  (fun _ ->
                    check_almost_orthogonal_modulo_infeasibility
                      (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8, _A9) css r))));;

let rec check_almost_orthogonal (_A1, _A2)
  r = bindb (check_left_linear_trs _A2
              ((ceq_list ceq_char), (ccompare_list ccompare_char),
                set_impl_list, (showl_list showl_char))
              (map fst r))
        (fun _ ->
          catch_error
            (forallM
              (fun rho_1 ->
                (let l_1 = fst (fst rho_1) in
                  catch_error
                    (forallM
                      (fun rho_2 ->
                        catch_error
                          (forallM
                            (check_overlapa
                              ((ccompare_list ccompare_char),
                                (equal_list equal_char), mapping_impl_list,
                                infinite_list, (showl_list showl_char))
                              (_A1, _A2)
                              (fun a ->
                                Chara (false, false, false, true, true, true,
true, false) ::
                                  a)
                              (fun a ->
                                Chara (true, false, false, true, true, true,
true, false) ::
                                  a)
                              (fun _ _ _ ->
                                Inl (showsl_lit
                                      "infeasibility check not supported"))
                              r rho_1 rho_2)
                            (funposs_list l_1))
                          (fun x -> Inl (snd x)))
                      r)
                    (fun x -> Inl (snd x))))
              r)
            (fun x -> Inl (snd x)));;

let rec check_level_confluence (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8)
  r = bindb (check_varcond_no_Var_lhs _A8 (showl_list showl_char) (map fst r))
        (fun _ ->
          bindb (check_type3 _A8
                  ((equal_list equal_char), (showl_list showl_char)) r)
            (fun _ ->
              bindb (check_extended_properly_oriented
                      (_A5.compare_compare_order, _A6, _A8)
                      (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
                        (cproper_interval_list ccompare_char),
                        (compare_list compare_char), (equal_list equal_char),
                        set_impl_list, (showl_list showl_char))
                      r)
                (fun _ ->
                  bindb (check_right_stable
                          (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8)
                          ((ceq_list ceq_char), (ccompare_list ccompare_char),
                            (compare_list compare_char),
                            (equal_list equal_char), mapping_impl_list,
                            set_impl_list, (showl_list showl_char))
                          r)
                    (fun _ -> check_almost_orthogonal (_A6, _A8) r))));;

let rec funs_crule_list r = add_funs_rule (fst r) (funs_trs_list (snd r));;

let rec funs_ctrs_list trs = maps funs_crule_list trs;;

let rec check_source_preserving _A (_B1, _B2)
  crs zv =
    catch_error
      (forallM
        (fun cr ->
          catch_error
            (forallM
              (fun i ->
                catch_error
                  (catch_error
                    (check_subseteq _B1 (vars_term_list (fst (fst cr)))
                      (zv cr i))
                    (fun _ ->
                      Inl (comp (comp (showsl_lit
"Some variable in lhs does not occur in Z_")
                                  (showsl_nat i))
                            (showsl_lit ". \092"))))
                  (fun x ->
                    Inl (comp (comp (comp (showsl_lit
    "The unraveling is not source preserving for rule ")
                                      (showsl_crule _A _B2 _A _B2 cr))
                                (showsl_literal "\092"))
                          x)))
              (upt zero_nata (size_list (snd cr))))
            (fun x -> Inl (snd x)))
        crs)
      (fun x -> Inl (snd x));;

let rec y_impl
  cr i =
    vars_term_list (snd (fst cr)) @
      vars_term_list (snd (nth (snd cr) i)) @
        vars_trs_list (drop (suc i) (snd cr));;

let rec check_Z_vars _A (_B1, _B2)
  crs z =
    catch_error
      (catch_error
        (forallM
          (fun cr ->
            catch_error
              (forallM
                (fun i ->
                  catch_error
                    (bindb
                      (catch_error
                        (check_subseteq _B1
                          (list_inter _B1 (x_impl cr i) (y_impl cr i)) (z cr i))
                        (fun x ->
                          Inl (comp (comp (comp
    (comp (showsl_lit "Variable ") (showsl _B2 x))
    (showsl_lit " does not occur in variable list of "))
                                      (showsl_nat (suc i)))
                                (showsl_lit ". U-symbol\092"))))
                      (fun _ ->
                        check (distinct _B1 (z cr i))
                          (showsl_lit
                            " variables in additional arguments of U-symbols are not distinct.\092")))
                    (fun x ->
                      Inl (comp (comp (comp
(showsl_lit "conditions for variable-lists in U-symbols for ")
(showsl_crule _A _B2 _A _B2 cr))
                                  (showsl_lit " are violated.\092"))
                            x)))
                (upt zero_nata (size_list (snd cr))))
              (fun x -> Inl (snd x)))
          crs)
        (fun x -> Inl (snd x)))
      (fun x ->
        Inl (comp (showsl_lit
                    "The CTRS does not fulfill the condition on Z variables.\092")
              x));;

let rec check_U_cond (_A1, _A2) (_B1, _B2)
  u crs f z =
    catch_error
      (catch_error
        (forallM
          (fun cr ->
            catch_error
              (forallM
                (fun i ->
                  (match u cr i
                    with Hole -> Inl (showsl_lit " Unexpected empty context.")
                    | More (fa, [], Hole, aft) ->
                      catch_error
                        (bindb
                          (catch_error (check_disjoint _A1 [fa] f)
                            (fun _ ->
                              Inl (comp (comp
  (showsl_lit "The function symbol ") (showsl _A2 fa))
                                    (showsl_lit " is not fresh.\092"))))
                          (fun _ ->
                            bindb (check
                                    (equal_lista (equal_term _A1 _B1) aft
                                      (map (fun a -> Var a) (z cr i)))
                                    (showsl_lit " U does not map to Z vars "))
                              (fun _ ->
                                check_f (_A1, _A2) (_B1, _B2) cr i fa crs u)))
                        (fun x ->
                          Inl (comp (comp (comp
    (comp (comp (showsl_lit "Conditions for ")
            (showsl_crule _A2 _B2 _A2 _B2 cr))
      (showsl_lit " at  "))
    (showsl_nat i))
                                      (showsl_lit " are violated.\092"))
                                x))
                    | More (_, [], More (_, _, _, _), _) ->
                      Inl (showsl_lit " Unexpected empty context.")
                    | More (_, _ :: _, _, _) ->
                      Inl (showsl_lit " Unexpected empty context.")))
                (upt zero_nata (size_list (snd cr))))
              (fun x -> Inl (snd x)))
          crs)
        (fun x -> Inl (snd x)))
      (fun x ->
        Inl (comp (showsl_lit
                    "The CTRS does not fulfill the condition on the U symbols.\092")
              x));;

let rec create_Umap_cr
  cr r =
    (match r with [] -> None
      | _ :: rr ->
        mapMa (fun (a, b) ->
                (let (l, _) = a in
                  (fun i ->
                    (match l with Var _ -> None | Fun (_, []) -> None
                      | Fun (u, _ :: _) -> Some (u, (cr, i)))))
                  b)
          (zip rr (upt zero_nata (size_list rr))));;

let rec create_Umap _A
  c_rs =
    (match mapMa (fun (a, b) -> create_Umap_cr a b) c_rs
      with None -> (fun _ -> None) | Some u -> map_of _A (concat u));;

let rec create_zs
  r = (match r with [] -> None
        | _ :: rr ->
          bind (mapMa
                 (fun a ->
                   (match a with (Var _, _) -> None | (Fun (_, []), _) -> None
                     | (Fun (_, _ :: ts), _) -> Some (map the_Var ts)))
                 rr)
            (fun cs ->
              Some (fun i ->
                     (if less_nat i (size_list cs) then nth cs i else []))));;

let rec create_Z _A _B
  c_rs =
    bind (mapMa (fun (cr, rs) -> bind (create_zs rs) (fun zs -> Some (cr, zs)))
           c_rs)
      (fun cr_zs ->
        (let mc =
           map_of
             (equal_prod (equal_prod (equal_term _A _B) (equal_term _A _B))
               (equal_list (equal_prod (equal_term _A _B) (equal_term _A _B))))
             cr_zs
           in
          Some (fun cr ->
                 (match mc cr with None -> (fun _ -> []) | Some zs -> zs))));;

let rec check_sp_unraveling (_A1, _A2) (_B1, _B2, _B3, _B4, _B5)
  c_rs ctrs =
    catch_error
      (bindb
        (catch_error
          (check_same_set
            (equal_prod (equal_prod (equal_term _A1 _B3) (equal_term _A1 _B3))
              (equal_list
                (equal_prod (equal_term _A1 _B3) (equal_term _A1 _B3))))
            ctrs (map fst c_rs))
          (fun x ->
            Inl (comp (comp (showsl_lit "did not find rule ")
                        (showsl_crule _A2 _B5 _A2 _B5 x))
                  (showsl_literal "\092"))))
        (fun _ ->
          bindb (match create_U _A1 _B3 c_rs
                  with None ->
                    Inl (showsl_lit "unable to extract unraveling contexts")
                  | Some a -> Inr a)
            (fun u ->
              bindb (match create_Z _A1 _B3 c_rs
                      with None ->
                        Inl (showsl_lit "unable to extract Z variables")
                      | Some a -> Inr a)
                (fun z ->
                  bindb (Inr (create_Umap _A1 c_rs))
                    (fun _ ->
                      bindb (check_U_cond (_A1, _A2) (_B3, _B5) u ctrs
                              (funs_ctrs_list ctrs) z)
                        (fun _ ->
                          bindb (check_Z_vars _A2 (_B3, _B5) ctrs z)
                            (fun _ ->
                              bindb (check_dctrs _A2 (_B3, _B5) ctrs)
                                (fun _ ->
                                  bindb (check_type3 _A2 (_B3, _B5) ctrs)
                                    (fun _ ->
                                      bindb
(catch_error
  (forallM
    (fun (c, rs) ->
      check (equal_lista (equal_prod (equal_term _A1 _B3) (equal_term _A1 _B3))
              (rules_impl u c) rs)
        (comp (comp (showsl_lit "problem with rules of ")
                (showsl_crule _A2 _B5 _A2 _B5 c))
          (showsl_literal "\092")))
    c_rs)
  (fun x -> Inl (snd x)))
(fun _ ->
  bindb (catch_error
          (check_left_linear_trs _A2 (_B1, _B2, _B4, _B5) (maps snd c_rs))
          (fun x ->
            Inl (comp (showsl_lit "the unraveled TRS is not left-linear\092")
                  x)))
    (fun _ ->
      bindb (catch_error (check_wf_ctrs _A2 (_B3, _B5) ctrs)
              (fun x ->
                Inl (comp (showsl_lit "the CTRS is not well-formed\092") x)))
        (fun _ ->
          bindb (catch_error (check_source_preserving _A2 (_B3, _B5) ctrs z)
                  (fun x ->
                    Inl (comp (showsl_lit
                                "unraveling is not source preserving\092")
                          x)))
            (fun _ -> Inr (maps snd c_rs))))))))))))))
      (fun x ->
        Inl (comp (showsl_lit
                    "preconditions on the unraveling are not satisfied\092")
              x));;

let rec check_conditional_cr_proof (_B1, _B2, _B3, _B4, _B5)
  a ia i j ctrs x5 = match a, ia, i, j, ctrs, x5 with
    a, ia, i, j, ctrs, Unconditional_CR prf ->
      debug ia "Unconditional"
        (bindb
          (catch_error
            (catch_error
              (forallM (fun x -> (if null (snd x) then Inr () else Inl x)) ctrs)
              (fun x -> Inl (snd x)))
            (fun x ->
              Inl (comp (comp (showsl_lit "rule with non-empty conditions")
                          (showsl_literal "\092"))
                    (showsl_crule (showl_lab _B5 (showl_list showl_nat))
                      (showl_list showl_char)
                      (showl_lab _B5 (showl_list showl_nat))
                      (showl_list showl_char) x))))
          (fun _ ->
            check_cr_proof (_B1, _B2, _B4, _B5) a ia i j (map fst ctrs) prf))
    | a, ia, i, j, ctrs, Unravel_CR (u_info, prf) ->
        debug ia "Unravel"
          (bindb
            (catch_error
              (check_sp_unraveling
                ((equal_lab _B4 (equal_list equal_nat)),
                  (showl_lab _B5 (showl_list showl_nat)))
                ((ceq_list ceq_char), (ccompare_list ccompare_char),
                  (equal_list equal_char), set_impl_list,
                  (showl_list showl_char))
                u_info ctrs)
              (fun x ->
                Inl (comp (comp (comp ia (showsl_lit ": error in unraveling"))
                            (showsl_literal "\092"))
                      x)))
            (fun r ->
              catch_error
                (check_cr_proof (_B1, _B2, _B4, _B5) a (add_index ia one_nata) i
                  j r prf)
                (fun x ->
                  Inl (comp (comp (comp ia
                                    (showsl_lit
                                      ": error below confluence proof"))
                              (showsl_literal "\092"))
                        x))))
    | a, ia, i, j, ctrs, Transformation_CR (t, prf) ->
        debug ia "CR Reflecting Transformation"
          (bindb
            (check_ccr_trans (_B1.compare_compare_order, _B3, _B4, _B5)
              ((compare_list compare_nat), (equal_list equal_nat),
                (showl_list showl_nat))
              (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
                (cproper_interval_list ccompare_char),
                (compare_order_list (compare_order_char, equal_char)),
                (equal_list equal_char), mapping_impl_list, set_impl_list,
                (showl_list showl_char))
              ia ctrs t)
            (fun ctrsa ->
              check_conditional_cr_proof (_B1, _B2, _B3, _B4, _B5) a ia i j
                ctrsa prf))
    | a, ia, i, j, ctrs, Almost_Orthogonal_CR ->
        debug ia "Almost-Orthogonal"
          (check_level_confluence
            (finite_UNIV_lab, cenum_lab, (ceq_lab _B4 (equal_list equal_nat)),
              (cproper_interval_lab _B1.compare_compare_order
                (compare_list compare_nat)),
              (compare_order_lab _B1.compare_compare_order
                (compare_list compare_nat)),
              (equal_lab _B4 (equal_list equal_nat)), set_impl_lab,
              (showl_lab _B5 (showl_list showl_nat)))
            ctrs)
    | a, ia, i, j, ctrs, Almost_Orthogonal_Modulo_Infeasibility_CR cps ->
        debug ia "Almost-Orthogonal modulo Infeasibility"
          (check_level_confluence_modulo_infeasibility
            (finite_UNIV_lab, cenum_lab, (ceq_lab _B4 (equal_list equal_nat)),
              (cproper_interval_lab _B1.compare_compare_order
                (compare_list compare_nat)),
              (compare_order_lab _B1.compare_compare_order
                (compare_list compare_nat)),
              (default_lab _B3), (equal_lab _B4 (equal_list equal_nat)),
              set_impl_lab, (showl_lab _B5 (showl_list showl_nat)))
            cps ctrs)
    | a, ia, i, j, ctrs, Almost_Orthogonal_Modulo_Infeasibility_CRa cps ->
        debug ia "Almost-Orthogonal modulo Infeasibility + meet-to-join"
          (check_level_confluence_modulo_infeasibilitya
            (finite_UNIV_lab, cenum_lab, (ceq_lab _B4 (equal_list equal_nat)),
              (cproper_interval_lab _B1.compare_compare_order
                (compare_list compare_nat)),
              (compare_order_lab _B1.compare_compare_order
                (compare_list compare_nat)),
              (default_lab _B3), (equal_lab _B4 (equal_list equal_nat)),
              set_impl_lab, (showl_lab _B5 (showl_list showl_nat)))
            cps ctrs)
    | a, ia, i, j, ctrs, AL94_CR (qrp, cj, icp, ucp) ->
        debug ia "AL94"
          (bindb
            (check_quasi_reductive_proof (_B1, _B2, _B4, _B5) a ia i j ctrs qrp)
            (fun _ ->
              check_al94
                (finite_UNIV_lab, cenum_lab,
                  (ceq_lab _B4 (equal_list equal_nat)),
                  (cproper_interval_lab _B1.compare_compare_order
                    (compare_list compare_nat)),
                  (compare_order_lab _B1.compare_compare_order
                    (compare_list compare_nat)),
                  (default_lab _B3), (equal_lab _B4 (equal_list equal_nat)),
                  set_impl_lab, (showl_lab _B5 (showl_list showl_nat)))
                cj icp ucp ctrs));;

let rec check_fptrs_termination_proof (_B1, _B2, _B3, _B4)
  ia j a i tp prf =
    check_fptrs_termination_proof_main (_B1, _B2, _B3, _B4) ia j a i tp
      (unlab_to_split_otrs (_B1, _B3)
        ((compare_order_list (compare_order_nat, equal_nat)),
          (equal_list equal_nat))
        ((compare_order_list (compare_order_char, equal_char)),
          (equal_list equal_char))
        prf);;

let rec check_E_reachable _A (_B1, _B2, _B3)
  e s t =
    membera (equal_term _A _B2)
      (mk_rtrancl_list (equal_terma _A _B2) (rewrite _A (_B1, _B2, _B3) e) [s])
      t;;

let rec check_AC_rule (_A1, _A2, _A3, _A4, _A5) _B
  lr = (let (l, r) = lr in
         (if is_Var l then false
           else (let f = fst (the (root l)) in
                  set_eq
                    ((cenum_prod _A1 cenum_nat), (ceq_prod _A2 ceq_nat),
                      (ccompare_prod _A3 ccompare_nat))
                    (funas_term (_A2, _A3, _A5) l)
                    (inserta
                      ((ceq_prod _A2 ceq_nat), (ccompare_prod _A3 ccompare_nat))
                      (f, nat_of_integer (Big_int.big_int_of_int 2))
                      (bot_set
                        ((ceq_prod _A2 ceq_nat),
                          (ccompare_prod _A3 ccompare_nat),
                          (set_impl_prod _A5 set_impl_nat)))) &&
                    (set_eq
                       ((cenum_prod _A1 cenum_nat), (ceq_prod _A2 ceq_nat),
                         (ccompare_prod _A3 ccompare_nat))
                       (funas_term (_A2, _A3, _A5) r)
                       (inserta
                         ((ceq_prod _A2 ceq_nat),
                           (ccompare_prod _A3 ccompare_nat))
                         (f, nat_of_integer (Big_int.big_int_of_int 2))
                         (bot_set
                           ((ceq_prod _A2 ceq_nat),
                             (ccompare_prod _A3 ccompare_nat),
                             (set_impl_prod _A5 set_impl_nat)))) &&
                      (eq (equal_multiset _B) (vars_term_ms _B l)
                         (vars_term_ms _B r) &&
                        eq (equal_multiset _A4) (funs_term_ms _A4 l)
                          (funs_term_ms _A4 r))))));;

let rec check_AC_theory (_A1, _A2, _A3, _A4, _A5, _A6) (_B1, _B2)
  e = catch_error
        (forallM
          (fun lr ->
            check (check_AC_rule (_A1, _A2, _A3, _A4, _A5) _B1 lr)
              (comp (comp (showsl_lit "rule ") (showsl_rule _A6 _B2 lr))
                (showsl_lit " violates AC-property")))
          e)
        (fun x -> Inl (snd x));;

let rec check_symmetric_AC_theory (_A1, _A2, _A3, _A4, _A5, _A6)
  (_B1, _B2, _B3, _B4)
  e = bindb (check_AC_theory (_A1, _A2, _A3, _A4, _A5, _A6) (_B2, _B4) e)
        (fun _ ->
          catch_error
            (catch_error
              (forallM
                (fun (l, r) ->
                  check (check_E_reachable _A4 (_B1, _B2, _B3) e r l)
                    (comp (comp (comp (showsl_lit "rhs ")
                                  (showsl_terma _A6 _B4 r))
                            (showsl_lit " does not rewrite to lhs "))
                      (showsl_terma _A6 _B4 l)))
                e)
              (fun x -> Inl (snd x)))
            (fun x ->
              Inl (comp (showsl_lit "theory is not symmetric\092") x)));;

let rec re (Ac_tp_ops_ext (ac_tp, r, a, c, mk, delete_rules, e, more)) = r;;

let rec c (Ac_tp_ops_ext (ac_tp, r, a, c, mk, delete_rules, e, more)) = c;;

let rec a (Ac_tp_ops_ext (ac_tp, r, a, c, mk, delete_rules, e, more)) = a;;

let rec check_only_C_rule (_A1, _A2, _A3) _B
  oc lr =
    (match lr with (Var _, _) -> true | (Fun (_, []), _) -> true
      | (Fun (_, [_]), _) -> true
      | (Fun (f, [s; t]), r) ->
        (if member (_A1, _A2) f oc then equal_terma _A3 _B r (Fun (f, [t; s]))
          else true)
      | (Fun (_, _ :: _ :: _ :: _), _) -> true);;

let rec check_only_C_theory (_A1, _A2, _A3, _A4) (_B1, _B2)
  oc e =
    catch_error
      (forallM
        (fun lr ->
          check (check_only_C_rule (_A1, _A2, _A3) _B1 oc lr)
            (comp (comp (showsl_lit "rule ") (showsl_rule _A4 _B2 lr))
              (showsl_lit " violates only-C-property")))
        e)
      (fun x -> Inl (snd x));;

let rec check_AC_same_as_E (_A1, _A2, _A3, _A4) (_B1, _B2, _B3, _B4, _B5)
  x y z a c e =
    catch_error
      (let xa = Var x in
       let ya = Var y in
       let za = Var z in
        bindb (catch_error
                (catch_error
                  (forallM
                    (fun f ->
                      check (check_E_reachable _B3 (_A1, _A2, _A3) e
                              (Fun (f, [xa; ya])) (Fun (f, [ya; xa])))
                        f)
                    c)
                  (fun xb -> Inl (snd xb)))
                (fun xb ->
                  Inl (comp (comp (showsl_lit "could not simulate C-rules for ")
                              (showsl _B5 xb))
                        (showsl_lit " by E"))))
          (fun _ ->
            bindb (catch_error
                    (catch_error
                      (forallM
                        (fun f ->
                          check (check_E_reachable _B3 (_A1, _A2, _A3) e
                                  (Fun (f, [xa; Fun (f, [ya; za])]))
                                  (Fun (f, [Fun (f, [xa; ya]); za])))
                            f)
                        a)
                      (fun xb -> Inl (snd xb)))
                    (fun xb ->
                      Inl (comp (comp (showsl_lit
"could not simulate A-rules for ")
                                  (showsl _B5 xb))
                            (showsl_lit " by E"))))
              (fun _ ->
                bindb (catch_error
                        (catch_error
                          (forallM
                            (fun f ->
                              check (check_E_reachable _B3 (_A1, _A2, _A3) e
                                      (Fun (f, [Fun (f, [xa; ya]); za]))
                                      (Fun (f, [xa; Fun (f, [ya; za])])))
                                f)
                            a)
                          (fun xb -> Inl (snd xb)))
                        (fun xb ->
                          Inl (comp (comp (showsl_lit
    "could not simulate A-rules for ")
                                      (showsl _B5 xb))
                                (showsl_lit " by E"))))
                  (fun _ ->
                    catch_error
                      (catch_error
                        (forallM
                          (fun (l, r) ->
                            check (equal_acterma _B3 _A2
                                    (aocnf (_B1, _B2, _B3) _A2
                                      (set (_B1, _B2, _B4) a)
                                      (set (_B1, _B2, _B4) c) l)
                                    (aocnf (_B1, _B2, _B3) _A2
                                      (set (_B1, _B2, _B4) a)
                                      (set (_B1, _B2, _B4) c) r))
                              (l, r))
                          e)
                        (fun xb -> Inl (snd xb)))
                      (fun xb ->
                        Inl (comp (comp (showsl_lit "equation ")
                                    (showsl_rule _B5 _A4 xb))
                              (showsl_lit " is not AC-equivalent")))))))
      (fun xa ->
        Inl (comp (showsl_lit
                    "could not ensure that equations simulate AC-equivalence\092")
              xa));;

let rec check_ext_rule3 _A (_B1, _B2, _B3)
  uu f xb x3 = match uu, f, xb, x3 with
    (la, ra), f, xb,
      (Fun (ga, [Fun (ha, [Var xa; l]); Var ya]),
        Fun (g, [Fun (h, [Var x; r]); Var y]))
      -> equal_proda (equal_term _A _B3)
           (equal_prod (equal_term _A _B3)
             (equal_prod _A
               (equal_prod _A
                 (equal_prod _A (equal_prod _A (equal_prod _B3 _B3))))))
           (la, (ra, (f, (f, (f, (f, (xa, ya)))))))
           (l, (r, (ga, (g, (ha, (h, (x, y))))))) &&
           (not (eq _B3 x y) &&
             (not (member (_B1, _B2) x xb) && not (member (_B1, _B2) y xb)))
    | uu, uv, uw, (Var vb, va) -> false
    | uu, uv, uw, (Fun (vb, []), va) -> false
    | uu, uv, uw, (Fun (vb, Var vf :: ve), va) -> false
    | uu, uv, uw, (Fun (vb, Fun (vf, []) :: ve), va) -> false
    | uu, uv, uw, (Fun (vb, Fun (vf, Fun (vj, vk) :: vi) :: ve), va) -> false
    | uu, uv, uw, (Fun (vb, Fun (vf, [vh]) :: ve), va) -> false
    | uu, uv, uw, (Fun (vb, Fun (vf, vh :: vj :: vl :: vm) :: ve), va) -> false
    | uu, uv, uw, (Fun (vb, [vd]), va) -> false
    | uu, uv, uw, (Fun (vb, vd :: Fun (vh, vi) :: vg), va) -> false
    | uu, uv, uw, (Fun (vb, vd :: vf :: vh :: vi), va) -> false
    | uu, uv, uw, (v, Var vb) -> false
    | uu, uv, uw, (v, Fun (vb, [])) -> false
    | uu, uv, uw, (v, Fun (vb, Var vf :: ve)) -> false
    | uu, uv, uw, (v, Fun (vb, Fun (vf, []) :: ve)) -> false
    | uu, uv, uw, (v, Fun (vb, Fun (vf, Fun (vj, vk) :: vi) :: ve)) -> false
    | uu, uv, uw, (v, Fun (vb, Fun (vf, [vh]) :: ve)) -> false
    | uu, uv, uw, (v, Fun (vb, Fun (vf, vh :: vj :: vl :: vm) :: ve)) -> false
    | uu, uv, uw, (v, Fun (vb, [vd])) -> false
    | uu, uv, uw, (v, Fun (vb, vd :: Fun (vh, vi) :: vg)) -> false
    | uu, uv, uw, (v, Fun (vb, vd :: vf :: vh :: vi)) -> false;;

let rec check_ext_rule2 _A (_B1, _B2, _B3)
  uu f xb x3 = match uu, f, xb, x3 with
    (la, ra), f, xb, (Fun (ga, [Var xa; l]), Fun (g, [Var x; r])) ->
      equal_proda (equal_term _A _B3)
        (equal_prod (equal_term _A _B3) (equal_prod _A (equal_prod _A _B3)))
        (la, (ra, (f, (f, xa)))) (l, (r, (ga, (g, x)))) &&
        not (member (_B1, _B2) xa xb)
    | uu, uv, uw, (Var vb, va) -> false
    | uu, uv, uw, (Fun (vb, []), va) -> false
    | uu, uv, uw, (Fun (vb, Fun (vf, vg) :: ve), va) -> false
    | uu, uv, uw, (Fun (vb, [vd]), va) -> false
    | uu, uv, uw, (Fun (vb, vd :: vf :: vh :: vi), va) -> false
    | uu, uv, uw, (v, Var vb) -> false
    | uu, uv, uw, (v, Fun (vb, [])) -> false
    | uu, uv, uw, (v, Fun (vb, Fun (vf, vg) :: ve)) -> false
    | uu, uv, uw, (v, Fun (vb, [vd])) -> false
    | uu, uv, uw, (v, Fun (vb, vd :: vf :: vh :: vi)) -> false;;

let rec check_ext_rule1 _A (_B1, _B2, _B3)
  uu f xb x3 = match uu, f, xb, x3 with
    (la, ra), f, xb, (Fun (ga, [l; Var xa]), Fun (g, [r; Var x])) ->
      equal_proda (equal_term _A _B3)
        (equal_prod (equal_term _A _B3) (equal_prod _A (equal_prod _A _B3)))
        (la, (ra, (f, (f, xa)))) (l, (r, (ga, (g, x)))) &&
        not (member (_B1, _B2) xa xb)
    | uu, uv, uw, (Var vb, va) -> false
    | uu, uv, uw, (Fun (vb, []), va) -> false
    | uu, uv, uw, (Fun (vb, [vd]), va) -> false
    | uu, uv, uw, (Fun (vb, vd :: Fun (vh, vi) :: vg), va) -> false
    | uu, uv, uw, (Fun (vb, vd :: vf :: vh :: vi), va) -> false
    | uu, uv, uw, (v, Var vb) -> false
    | uu, uv, uw, (v, Fun (vb, [])) -> false
    | uu, uv, uw, (v, Fun (vb, [vd])) -> false
    | uu, uv, uw, (v, Fun (vb, vd :: Fun (vh, vi) :: vg)) -> false
    | uu, uv, uw, (v, Fun (vb, vd :: vf :: vh :: vi)) -> false;;

let rec check_ext_rule (_A1, _A2, _A3, _A4) (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  rext a c lr =
    (let (l, _) = lr in
      (if is_Var l ||
            (not (equal_nata (size_list (args l))
                   (nat_of_integer (Big_int.big_int_of_int 2))) ||
              not (member (_A1, _A2) (fst (the (root l))) a))
        then true
        else (let f = fst (the (root l)) in
              let x =
                vars_rule (_A3, _A4) (_B1, _B2, _B3, _B4, _B5, _B6, _B7) lr in
               list_ex
                 (check_ext_rule1 _A4 (_B3, _B4.ccompare_cproper_interval, _B6)
                   lr f x)
                 rext &&
                 (if not (member (_A1, _A2) f c)
                   then list_ex
                          (check_ext_rule2 _A4
                            (_B3, _B4.ccompare_cproper_interval, _B6) lr f x)
                          rext &&
                          list_ex
                            (check_ext_rule3 _A4
                              (_B3, _B4.ccompare_cproper_interval, _B6) lr f x)
                            rext
                   else true))));;

let rec check_ext_trs (_A1, _A2, _A3, _A4, _A5, _A6)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  r a c rext =
    (let aa = set (_A1, _A2, _A5) a in
     let ca = set (_A1, _A2, _A5) c in
      catch_error
        (catch_error
          (forallM
            (fun lr ->
              check (check_ext_rule (_A1, _A2, _A3, _A4)
                      (_B1, _B2, _B3, _B4, _B5, _B6, _B7) rext aa ca lr)
                (comp (comp (comp (showsl_lit
                                    "could not find extended rules for rule l -> r:\092  ")
                              (showsl_rule _A6 _B8 lr))
                        (showsl_lit
                          "\092  expecting rule f(l,x) -> f(r,x) for all A and AC symbols,\092"))
                  (showsl_lit
                    "and rules f(x,l) -> f(x,r) and f(f(x,l),y) -> f(f(x,r),y) for all A symbols")))
            r)
          (fun x -> Inl (snd x)))
        (fun x ->
          Inl (comp (showsl_lit
                      "could not ensure validity of AC-extended system\092")
                x)));;

let rec ac_dependency_pairs_checks (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8, _C9)
  i shp x y z (AC_dependency_pairs_proof (e, dpr, dpe, rext)) tp =
    catch_error
      (let aa = a i tp in
       let ca = c i tp in
       let r = re i tp in
       let oc = list_diff _B5 ca aa in
       let d = defined_list r in
       let _ =
         set ((ceq_prod _B2 ceq_nat), (ccompare_prod _B3 ccompare_nat),
               (set_impl_prod _B6 set_impl_nat))
           d
         in
       let da = defined_list (r @ e) in
        bindb (catch_error
                (forallM
                  (fun (f, n) ->
                    check (not (membera (equal_prod _B5 equal_nat) da
                                 (shp f, n)))
                      (comp (comp (comp (showsl_lit "sharping ") (showsl _B7 f))
                              (showsl_lit " yields the defined symbol "))
                        (showsl _B7 (shp f))))
                  da)
                (fun xa -> Inl (snd xa)))
          (fun _ ->
            bindb (catch_error (check_wf_trs _B7 (_C6, _C9) (r @ e))
                    (fun xa ->
                      Inl (comp (showsl_lit
                                  "TRS or equations are not well-formed")
                            xa)))
              (fun _ ->
                bindb (catch_error
                        (check_symmetric_AC_theory
                          (_B1, _B2, _B3, _B5, _B6, _B7)
                          (_C4.ccompare_cproper_interval, _C6, _C7, _C9) e)
                        (fun xa ->
                          Inl (comp (showsl_lit
                                      "equations do not form a symmetric AC-theory\092")
                                xa)))
                  (fun _ ->
                    bindb (catch_error
                            (check_only_C_theory (_B2, _B3, _B5, _B7) (_C6, _C9)
                              (set (_B2, _B3, _B6) oc) e)
                            (fun xa ->
                              Inl (comp (showsl_lit
  "equations do not form AC_C-theory\092")
                                    xa)))
                      (fun _ ->
                        bindb (catch_error
                                (check_subseteq _B5 (funs_trs_list e) (aa @ ca))
                                (fun xa ->
                                  Inl (comp
(comp (showsl_lit "equations contain symbol ") (showsl _B7 xa))
(showsl_lit " which is not AC-symbol"))))
                          (fun _ ->
                            bindb (catch_error
                                    (check_AC_same_as_E
                                      (_C4.ccompare_cproper_interval, _C6, _C7,
_C9)
                                      (_B2, _B3, _B5, _B6, _B7) x y z aa ca e)
                                    (fun xa ->
                                      Inl
(comp (showsl_lit
        "could not ensure that equations correspond to AC equivalence\092")
  xa)))
                              (fun _ ->
                                bindb (catch_error
(check_subseteq (equal_prod (equal_term _B5 _C6) (equal_term _B5 _C6))
  (dP_list _B5 _C6 shp r d) dpr)
(fun xa ->
  Inl (comp (showsl_lit "could not find DP for R: ") (showsl_rule _B7 _C9 xa))))
                                  (fun _ ->
                                    bindb (catch_error
    (check_subseteq (equal_prod (equal_term _B5 _C6) (equal_term _B5 _C6))
      (dP_list _B5 _C6 shp e d) dpe)
    (fun xa ->
      Inl (comp (showsl_lit "could not find DP for E: ")
            (showsl_rule _B7 _C9 xa))))
                                      (fun _ ->
catch_error
  (check_ext_trs (_B2, _B3, _B4, _B5, _B6, _B7)
    (_C1, _C2, _C3, _C4, _C5, _C6, _C8, _C9) r aa ca rext)
  (fun xa ->
    Inl (comp (showsl_lit "could not ensure validity of extended TRS R_ext\092")
          xa)))))))))))
      (fun xa ->
        Inl (comp (showsl_lit "problem in applying AC-dependency pairs\092")
              xa));;

let rec mkf
  (Ac_dpp_ops_ext
    (ac_dpp, p, pw, pairs, r, rw, rules, e, mk, eq_rules_map,
      reverse_eq_rules_map, delete_pairs_rules, eq_rules_no_left_var,
      eq_rules_non_collapsing, intersect_pairs, more))
    = mk;;

let rec ac_dependency_pairs_proc_simple (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8, _C9)
  i j shp x y z (AC_dependency_pairs_proof (e, dpr, dpe, rext)) tp =
    bindb (ac_dependency_pairs_checks (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
            (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8, _C9) i shp x y z
            (AC_dependency_pairs_proof (e, dpr, dpe, rext)) tp)
      (fun _ ->
        (let r = re i tp in
          Inr (mkf j
                (dpr @
                  map (fun (l, ra) -> (sharp_term shp l, sharp_term shp ra))
                    rext)
                dpe [] r e)));;

let rec delete_pairs_rulesa
  (Ac_dpp_ops_ext
    (ac_dpp, p, pw, pairs, r, rw, rules, e, mk, eq_rules_map,
      reverse_eq_rules_map, delete_pairs_rules, eq_rules_no_left_var,
      eq_rules_non_collapsing, intersect_pairs, more))
    = delete_pairs_rules;;

let rec eq_rules_map
  (Ac_dpp_ops_ext
    (ac_dpp, p, pw, pairs, r, rw, rules, e, mk, eq_rules_map,
      reverse_eq_rules_map, delete_pairs_rules, eq_rules_no_left_var,
      eq_rules_non_collapsing, intersect_pairs, more))
    = eq_rules_map;;

let rec rulesf
  (Ac_dpp_ops_ext
    (ac_dpp, p, pw, pairs, r, rw, rules, e, mk, eq_rules_map,
      reverse_eq_rules_map, delete_pairs_rules, eq_rules_no_left_var,
      eq_rules_non_collapsing, intersect_pairs, more))
    = rules;;

let rec pairsc
  (Ac_dpp_ops_ext
    (ac_dpp, p, pw, pairs, r, rw, rules, e, mk, eq_rules_map,
      reverse_eq_rules_map, delete_pairs_rules, eq_rules_no_left_var,
      eq_rules_non_collapsing, intersect_pairs, more))
    = pairs;;

let rec rwe
  (Ac_dpp_ops_ext
    (ac_dpp, p, pw, pairs, r, rw, rules, e, mk, eq_rules_map,
      reverse_eq_rules_map, delete_pairs_rules, eq_rules_no_left_var,
      eq_rules_non_collapsing, intersect_pairs, more))
    = rw;;

let rec rf
  (Ac_dpp_ops_ext
    (ac_dpp, p, pw, pairs, r, rw, rules, e, mk, eq_rules_map,
      reverse_eq_rules_map, delete_pairs_rules, eq_rules_no_left_var,
      eq_rules_non_collapsing, intersect_pairs, more))
    = r;;

let rec ec
  (Ac_dpp_ops_ext
    (ac_dpp, p, pw, pairs, r, rw, rules, e, mk, eq_rules_map,
      reverse_eq_rules_map, delete_pairs_rules, eq_rules_no_left_var,
      eq_rules_non_collapsing, intersect_pairs, more))
    = e;;

let rec ac_mono_ur_redpair_proc (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  (_C1, _C2, _C3, _C4, _C5)
  i rp p_remove r_remove ur dpp =
    (match
      catch_error
        (let p = pairsc i dpp in
         let _ = rulesf i dpp in
         let e = ec i dpp in
         let premove =
           set ((ceq_prod (ceq_term _B5 _C3) (ceq_term _B5 _C3)),
                 (ccompare_prod (ccompare_term _B4 _C2)
                   (ccompare_term _B4 _C2)),
                 (set_impl_prod set_impl_term set_impl_term))
             p_remove
           in
         let rremove =
           set ((ceq_prod (ceq_term _B5 _C3) (ceq_term _B5 _C3)),
                 (ccompare_prod (ccompare_term _B4 _C2)
                   (ccompare_term _B4 _C2)),
                 (set_impl_prod set_impl_term set_impl_term))
             r_remove
           in
         let us =
           foldr (comp (sup_seta
                         ((ceq_prod _B2 ceq_nat),
                           (ccompare_prod _B3 ccompare_nat)))
                   (comp (funas_term (_B2, _B3, _B6)) snd))
             (p @ ur)
             (bot_set
               ((ceq_prod _B2 ceq_nat), (ccompare_prod _B3 ccompare_nat),
                 (set_impl_prod _B6 set_impl_nat)))
           in
         let filt =
           (fun lr ->
             ball ((ceq_prod _B2 ceq_nat), (ccompare_prod _B3 ccompare_nat))
               (funas_term (_B2, _B3, _B6) (fst lr))
               (fun f ->
                 member
                   ((ceq_prod _B2 ceq_nat), (ccompare_prod _B3 ccompare_nat)) f
                   us))
           in
         let (ps, pns) =
           partition
             (fun lr ->
               member
                 ((ceq_prod (ceq_term _B5 _C3) (ceq_term _B5 _C3)),
                   (ccompare_prod (ccompare_term _B4 _C2)
                     (ccompare_term _B4 _C2)))
                 lr premove &&
                 filt lr)
             p
           in
         let (urs, urns) =
           partition
             (fun lr ->
               member
                 ((ceq_prod (ceq_term _B5 _C3) (ceq_term _B5 _C3)),
                   (ccompare_prod (ccompare_term _B4 _C2)
                     (ccompare_term _B4 _C2)))
                 lr rremove &&
                 filt lr)
             ur
           in
         let rm = eq_rules_map i dpp in
          bindb (valid rp)
            (fun _ ->
              bindb (mono rp (ps @ urs @ urns @ pns))
                (fun _ ->
                  bindb (catch_error
                          (check_symmetric_AC_theory
                            (_B1, _B2, _B3, _B5, _B6, _B7) (_C1, _C3, _C4, _C5)
                            e)
                          (fun x ->
                            Inl (comp (showsl_lit
"usable rules demand symmetric AC theory\092")
                                  x)))
                    (fun _ ->
                      bindb (catch_error
                              (forallM
                                (fun (l, _) ->
                                  check (not (is_Var l))
                                    (showsl_lit
                                      "variables as lhss not allowed"))
                                (rf i dpp @ rwe i dpp))
                              (fun x -> Inl (snd x)))
                        (fun _ ->
                          bindb (check_ur_P_closed_rm_af (_B5, _B7) (_C3, _C5)
                                  rm ur full_af p)
                            (fun _ ->
                              bindb (catch_error
                                      (catch_error (forallM (ns rp) urns)
(fun x -> Inl (snd x)))
                                      (fun x ->
Inl (comp (showsl_lit "problem when orienting usable rules\092") x)))
                                (fun _ ->
                                  bindb (catch_error
  (catch_error (forallM (s rp) urs) (fun x -> Inl (snd x)))
  (fun x ->
    Inl (comp (showsl_lit "problem when orienting usable rules\092") x)))
                                    (fun _ ->
                                      bindb
(catch_error (catch_error (forallM (ns rp) pns) (fun x -> Inl (snd x)))
  (fun x -> Inl (comp (showsl_lit "problem when orienting DPs\092") x)))
(fun _ ->
  catch_error (catch_error (forallM (s rp) ps) (fun x -> Inl (snd x)))
    (fun x ->
      Inl (comp (showsl_lit "problem when orienting DPs\092") x)))))))))))
        (fun x ->
          Inl (comp (comp (comp (showsl_lit
                                  "could not apply the monotonic AC reduction pair processor with usable rules and the following\092")
                            (desc rp))
                      (showsl_literal "\092"))
                x))
      with Inl a -> Inl a
      | Inr _ -> Inr (delete_pairs_rulesa i dpp p_remove r_remove));;

let rec pwc
  (Ac_dpp_ops_ext
    (ac_dpp, p, pw, pairs, r, rw, rules, e, mk, eq_rules_map,
      reverse_eq_rules_map, delete_pairs_rules, eq_rules_no_left_var,
      eq_rules_non_collapsing, intersect_pairs, more))
    = pw;;

let rec pc
  (Ac_dpp_ops_ext
    (ac_dpp, p, pw, pairs, r, rw, rules, e, mk, eq_rules_map,
      reverse_eq_rules_map, delete_pairs_rules, eq_rules_no_left_var,
      eq_rules_non_collapsing, intersect_pairs, more))
    = p;;

let rec ac_dpp_trivial_check
  i dpp =
    bindb (check (null (pc i dpp)) (showsl_lit "there are strict pairs"))
      (fun _ ->
        check (null (pwc i dpp) || null (rf i dpp))
          (showsl_lit "there are weak pairs and strict rules"));;

let rec ac_ur_redpair_proc (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  (_C1, _C2, _C3, _C4, _C5)
  i rp p_remove ur dpp =
    (match
      catch_error
        (let p = pairsc i dpp in
         let _ = rulesf i dpp in
         let e = ec i dpp in
         let premove =
           set ((ceq_prod (ceq_term _B5 _C3) (ceq_term _B5 _C3)),
                 (ccompare_prod (ccompare_term _B4 _C2)
                   (ccompare_term _B4 _C2)),
                 (set_impl_prod set_impl_term set_impl_term))
             p_remove
           in
         let (ps, pns) =
           partition
             (fun lr ->
               member
                 ((ceq_prod (ceq_term _B5 _C3) (ceq_term _B5 _C3)),
                   (ccompare_prod (ccompare_term _B4 _C2)
                     (ccompare_term _B4 _C2)))
                 lr premove)
             p
           in
         let rm = eq_rules_map i dpp in
         let pi = af rp in
          bindb (valid rp)
            (fun _ ->
              bindb (catch_error
                      (check_symmetric_AC_theory (_B1, _B2, _B3, _B5, _B6, _B7)
                        (_C1, _C3, _C4, _C5) e)
                      (fun x ->
                        Inl (comp (showsl_lit
                                    "usable rules demand symmetric AC theory\092")
                              x)))
                (fun _ ->
                  bindb (catch_error
                          (forallM
                            (fun (l, _) ->
                              check (not (is_Var l))
                                (showsl_lit "variables as lhss not allowed"))
                            (rf i dpp @ rwe i dpp))
                          (fun x -> Inl (snd x)))
                    (fun _ ->
                      bindb (check_ur_P_closed_rm_af (_B5, _B7) (_C3, _C5) rm ur
                              pi p)
                        (fun _ ->
                          bindb (catch_error
                                  (catch_error (forallM (ns rp) ur)
                                    (fun x -> Inl (snd x)))
                                  (fun x ->
                                    Inl (comp
  (showsl_lit "problem when orienting usable rules\092") x)))
                            (fun _ ->
                              bindb (catch_error
                                      (catch_error (forallM (nst rp) pns)
(fun x -> Inl (snd x)))
                                      (fun x ->
Inl (comp (showsl_lit "problem when orienting DPs\092") x)))
                                (fun _ ->
                                  catch_error
                                    (catch_error (forallM (s rp) ps)
                                      (fun x -> Inl (snd x)))
                                    (fun x ->
                                      Inl
(comp (showsl_lit "problem when orienting DPs\092") x)))))))))
        (fun x ->
          Inl (comp (comp (comp (showsl_lit
                                  "could not apply the AC reduction pair processor with usable rules and the following\092")
                            (desc rp))
                      (showsl_literal "\092"))
                x))
      with Inl a -> Inl a
      | Inr _ -> Inr (delete_pairs_rulesa i dpp p_remove []));;

let rec check_size_preserving_trs _A (_B1, _B2)
  e = catch_error
        (catch_error
          (forallM
            (fun (l, r) ->
              check (equal_nata (num_symbs l) (num_symbs r) &&
                      eq (equal_multiset _B1) (vars_term_ms _B1 l)
                        (vars_term_ms _B1 r))
                (comp (comp (showsl_lit "rule ") (showsl_rule _A _B2 (l, r)))
                  (showsl_lit " is not size preserving")))
            e)
          (fun x -> Inl (snd x)))
        (fun x -> Inl (comp (showsl_lit "TRS is not size-preserving\092") x));;

let rec ac_subterm_proc (_B1, _B2, _B3, _B4, _B5, _B6) (_C1, _C2, _C3)
  i pi p_remove dpp =
    (match
      catch_error
        (let p = pairsc i dpp in
         let r = rulesf i dpp in
         let e = ec i dpp in
         let re = r @ e in
         let f = map fst pi in
         let ff =
           set ((ceq_prod _B1 ceq_nat), (ccompare_prod _B2 ccompare_nat),
                 (set_impl_prod _B5 set_impl_nat))
             f
           in
         let pi_opt = status_of _B4 pi in
          bindb (check (not (is_none pi_opt))
                  (showsl_lit "argument filter lists invalid positions"))
            (fun _ ->
              (let pia = the pi_opt in
               let premove =
                 set ((ceq_prod (ceq_term _B4 _C2) (ceq_term _B4 _C2)),
                       (ccompare_prod (ccompare_term _B3 _C1)
                         (ccompare_term _B3 _C1)),
                       (set_impl_prod set_impl_term set_impl_term))
                   p_remove
                 in
               let (ps, pns) =
                 partition
                   (fun lr ->
                     member
                       ((ceq_prod (ceq_term _B4 _C2) (ceq_term _B4 _C2)),
                         (ccompare_prod (ccompare_term _B3 _C1)
                           (ccompare_term _B3 _C1)))
                       lr premove)
                   p
                 in
                bindb (catch_error
                        (forallM
                          (fun fa ->
                            check (not (null (status pia fa)))
                              (comp (comp (showsl_lit "status of symbol ")
                                      (showsl_prod _B6 showl_nat fa))
                                (showsl_lit " in F must be non-empty")))
                          f)
                        (fun x -> Inl (snd x)))
                  (fun _ ->
                    bindb (catch_error
                            (check_size_preserving_trs _B6 (_C2, _C3) e)
                            (fun x ->
                              Inl (comp (showsl_lit
  "E is not size preserving\092")
                                    x)))
                      (fun _ ->
                        bindb (catch_error
                                (forallM
                                  (fun (l, _) ->
                                    check (not (is_Var l))
                                      (showsl_lit
"variables as lhss not allowed"))
                                  re)
                                (fun x -> Inl (snd x)))
                          (fun _ ->
                            bindb (catch_error
                                    (catch_error
                                      (forallM
(check_supteqproj_pred (_B1, _B2, _B3, _B4, _B6) (_C1, _C2, _C3) pia ff)
(filtera
  (fun lr ->
    member ((ceq_prod _B1 ceq_nat), (ccompare_prod _B2 ccompare_nat))
      (the (root (fst lr))) ff)
  re))
                                      (fun x -> Inl (snd x)))
                                    (fun x ->
                                      Inl
(comp (showsl_lit "problem when orienting rules with root in F\092") x)))
                              (fun _ ->
                                bindb (catch_error
(catch_error
  (forallM
    (check_supteqproj_pred (_B1, _B2, _B3, _B4, _B6) (_C1, _C2, _C3) pia ff)
    pns)
  (fun x -> Inl (snd x)))
(fun x -> Inl (comp (showsl_lit "problem when orienting DPs\092") x)))
                                  (fun _ ->
                                    catch_error
                                      (catch_error
(forallM (check_suptproj_pred (_B1, _B2, _B3, _B4, _B6) (_C1, _C2, _C3) pia ff)
  ps)
(fun x -> Inl (snd x)))
                                      (fun x ->
Inl (comp (showsl_lit "problem when orienting DPs\092") x))))))))))
        (fun x ->
          Inl (comp (showsl_lit "could not apply the AC subterm processor\092")
                x))
      with Inl a -> Inl a
      | Inr _ -> Inr (delete_pairs_rulesa i dpp p_remove []));;

let rec intersect_pairsc
  (Ac_dpp_ops_ext
    (ac_dpp, p, pw, pairs, r, rw, rules, e, mk, eq_rules_map,
      reverse_eq_rules_map, delete_pairs_rules, eq_rules_no_left_var,
      eq_rules_non_collapsing, intersect_pairs, more))
    = intersect_pairs;;

let rec graph_approx_edg_rt_sym (_A1, _A2)
  m (uu, (uv, ct)) = graph_approx_rt_sym_main (_A1, _A2) m ct;;

let rec eq_rules_non_collapsinga
  (Ac_dpp_ops_ext
    (ac_dpp, p, pw, pairs, r, rw, rules, e, mk, eq_rules_map,
      reverse_eq_rules_map, delete_pairs_rules, eq_rules_no_left_var,
      eq_rules_non_collapsing, intersect_pairs, more))
    = eq_rules_non_collapsing;;

let rec reverse_eq_rules_map
  (Ac_dpp_ops_ext
    (ac_dpp, p, pw, pairs, r, rw, rules, e, mk, eq_rules_map,
      reverse_eq_rules_map, delete_pairs_rules, eq_rules_no_left_var,
      eq_rules_non_collapsing, intersect_pairs, more))
    = reverse_eq_rules_map;;

let rec reverse_tcapRM_ac_dpp _B _C
  i dpp =
    tcapRM _B _C (eq_rules_non_collapsinga i dpp) (reverse_eq_rules_map i dpp);;

let rec eq_rules_no_left_vara
  (Ac_dpp_ops_ext
    (ac_dpp, p, pw, pairs, r, rw, rules, e, mk, eq_rules_map,
      reverse_eq_rules_map, delete_pairs_rules, eq_rules_no_left_var,
      eq_rules_non_collapsing, intersect_pairs, more))
    = eq_rules_no_left_var;;

let rec tcapRM_ac_dpp _B _C
  i dpp = tcapRM _B _C (eq_rules_no_left_vara i dpp) (eq_rules_map i dpp);;

let rec check_ac_dep_graph_proc (_B1, _B2, _B3, _B4, _B5)
  i dpp dps =
    (let c = tcapRM_ac_dpp _B3 (equal_list equal_char) i dpp in
     let rc = reverse_tcapRM_ac_dpp _B3 (equal_list equal_char) i dpp in
     let p = pairsc i dpp in
      bindb (catch_error
              (check_subseteq
                (equal_prod (equal_term _B3 (equal_list equal_char))
                  (equal_term _B3 (equal_list equal_char)))
                p (maps snd dps))
              (fun x ->
                Inl (comp (comp (comp (showsl_lit "Dependency Pair ")
                                  (showsl_rule _B5 (showl_list showl_char) x))
                            (showsl_lit " is missing in decomposition"))
                      (showsl_literal "\092"))))
        (fun _ ->
          catch_error
            (check_graph_decomp
              (comp (showsl_prod (showl_term _B5 (showl_list showl_char))
                      (showl_term _B5 (showl_list showl_char)))
                fst)
              (emptyd
                ((ccompare_option (ccompare_prod _B1 ccompare_nat)),
                  (mapping_impl_option
                    (mapping_impl_prod _B4 mapping_impl_nat)))
                (comp (comp root fst) fst))
              (graph_approx_edg_rt_sym (_B1, _B3))
              (insertd
                ((ccompare_option (ccompare_prod _B1 ccompare_nat)),
                  (equal_option (equal_prod _B3 equal_nat))))
              (fun (a, b) ->
                (let (aa, ba) = a in
                  (let (_, t) = aa in
                    (fun (_, ct) (ab, bb) ->
                      (let (u, _) = ab in
                        (fun (cu, _) ->
                          matchc _B3 (equal_list equal_char) ct u &&
                            matchc _B3 (equal_list equal_char) cu t))
                        bb))
                    ba)
                  b)
              (map (fun (real, cs) ->
                     (not (is_none real),
                       map (fun (s, t) -> ((s, t), (rc s, c t))) cs))
                dps))
            (fun x ->
              Inl (comp (comp (comp (showsl_lit
                                      "our estimation (EDG***) could not show that you have a valid decomposition ")
                                (showsl_lit "due to the following reason"))
                          (showsl_literal "\092"))
                    x))));;

let rec ac_dep_graph_proc (_B1, _B2, _B3, _B4, _B5)
  i d dps =
    (match check_ac_dep_graph_proc (_B1, _B2, _B3, _B4, _B5) i d dps
      with Inl a -> Inl a
      | Inr _ ->
        Inr (map_filter
              (fun x ->
                (if not (is_none (fst x))
                  then Some (the (fst x), intersect_pairsc i d (snd x))
                  else None))
              dps));;

let rec showsl_ac_dpp _B _C
  i d = (let p = pc i d in
         let pw = pwc i d in
         let r = rf i d in
         let rw = rwe i d in
         let e = ec i d in
          comp (comp (comp (comp (if null p then id
                                   else showsl_trsa (showsl _B) (showsl _C)
  "pairs:" " -> " p)
                             (if null pw then id
                               else showsl_trsa (showsl _B) (showsl _C)
                                      "weak pairs:" " ->= " pw))
                       (if null r then id
                         else showsl_trsa (showsl _B) (showsl _C)
                                "strict rules:" " ->! " r))
                 (if null rw then id
                   else showsl_trsa (showsl _B) (showsl _C) "rules:" " -> " rw))
            (if null e then id
              else showsl_trsa (showsl _B) (showsl _C) "equations:" " -> " e));;

let rec check_ac_dp_termination_proof (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  ia i dpp x3 = match ia, i, dpp, x3 with
    ia, i, dpp, AC_P_is_Empty ->
      debug i "P is empty"
        (catch_error (ac_dpp_trivial_check ia dpp)
          (fun x ->
            Inl (comp (comp (comp i
                              (showsl_lit
                                "problem in applying trivial check on\092"))
                        (showsl_ac_dpp _B8 (showl_list showl_char) ia dpp))
                  x)))
    | ia, i, dpp, AC_Redpair_UR_Proc (redp, del_p, ur, prf) ->
        debug i "AC_Redpair_UR_Proc"
          (bindb
            (catch_error
              (ac_ur_redpair_proc
                (_B1, _B2, _B3, _B4.compare_compare_order, _B5, _B7, _B8)
                ((ccompare_list ccompare_char), (compare_list compare_char),
                  (equal_list equal_char), mapping_impl_list,
                  (showl_list showl_char))
                ia (get_redtriple (_B2, _B3, _B4, _B5, _B7, _B8) redp) del_p ur
                dpp)
              (fun x ->
                Inl (comp (comp (comp (comp
(comp (comp i
        (showsl_lit
          ": error when applying the ac-reduction pair processor to remove from the DP problem\092"))
  (showsl_ac_dpp _B8 (showl_list showl_char) ia dpp))
(showsl_lit "\092 the pairs\092"))
                                  (showsl_rules _B8 (showl_list showl_char)
                                    del_p))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_ac_dp_termination_proof
                  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) ia
                  (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the ac-reduction pair processor\092"))
                        x))))
    | ia, i, dpp, AC_Subterm_Proc (pi, del_p, prf) ->
        debug i "AC_Subterm_Proc"
          (bindb
            (catch_error
              (ac_subterm_proc
                (_B2, _B3, _B4.compare_compare_order, _B5, _B7, _B8)
                ((compare_list compare_char), (equal_list equal_char),
                  (showl_list showl_char))
                ia pi del_p dpp)
              (fun x ->
                Inl (comp (comp (comp (comp
(comp (comp i
        (showsl_lit
          ": error when applying the ac-subterm processor to remove from the DP problem\092"))
  (showsl_ac_dpp _B8 (showl_list showl_char) ia dpp))
(showsl_lit "\092 the pairs\092"))
                                  (showsl_rules _B8 (showl_list showl_char)
                                    del_p))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_ac_dp_termination_proof
                  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) ia
                  (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the ac-reduction pair processor\092"))
                        x))))
    | ia, i, dpp, AC_Mono_Redpair_UR_Proc (redp, del_p, del_r, ur, prf) ->
        debug i "AC_Mono_Redpair_UR_Proc"
          (bindb
            (catch_error
              (ac_mono_ur_redpair_proc
                (_B1, _B2, _B3, _B4.compare_compare_order, _B5, _B7, _B8)
                ((ccompare_list ccompare_char), (compare_list compare_char),
                  (equal_list equal_char), mapping_impl_list,
                  (showl_list showl_char))
                ia (get_redtriple (_B2, _B3, _B4, _B5, _B7, _B8) redp) del_p
                del_r ur dpp)
              (fun x ->
                Inl (comp (comp (comp (comp
(comp (comp (comp (comp i
                    (showsl_lit
                      ": error when applying the monotone ac-reduction pair processor to remove from the DP problem\092"))
              (showsl_ac_dpp _B8 (showl_list showl_char) ia dpp))
        (showsl_lit "\092 the pairs\092"))
  (showsl_rules _B8 (showl_list showl_char) del_p))
(showsl_lit "\092 and the rules\092"))
                                  (showsl_rules _B8 (showl_list showl_char)
                                    del_r))
                            (showsl_literal "\092"))
                      x)))
            (fun dppa ->
              catch_error
                (check_ac_dp_termination_proof
                  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) ia
                  (add_index i one_nata) dppa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the monotone ac-reduction pair processor\092"))
                        x))))
    | ia, i, dpp, AC_Dep_Graph_Proc edpts ->
        debug i "Dep_Graph_Proc"
          (bindb
            (catch_error
              (ac_dep_graph_proc (_B3, _B4, _B5, _B6, _B8) ia dpp edpts)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit
  ": error while trying to perform Sctxt_closure-decomposition  on\092"))
                                  (showsl_ac_dpp _B8 (showl_list showl_char) ia
                                    dpp))
                            (showsl_literal "\092"))
                      x)))
            (fun pdpps ->
              catch_error
                (catch_error
                  (forallM_index
                    (fun (prof, dppa) j ->
                      check_ac_dp_termination_proof
                        (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) ia
                        (add_index i (suc j)) dppa prof)
                    pdpps)
                  (fun x -> Inl (snd x)))
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the dependency graph processor\092"))
                        x))));;

let rec ac_dependency_pairs_proc (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8, _C9)
  i j shp x y z (AC_dependency_pairs_proof (e, dpr, dpe, rext)) tp =
    bindb (ac_dependency_pairs_checks (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
            (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8, _C9) i shp x y z
            (AC_dependency_pairs_proof (e, dpr, dpe, rext)) tp)
      (fun _ ->
        (let r = re i tp in
          Inr (mkf j dpr dpe [] r e,
                mkf j (map (fun (l, ra) ->
                             (sharp_term shp l, sharp_term shp ra))
                        rext)
                  dpe [] r e)));;

let rec delete_rulesa
  (Ac_tp_ops_ext (ac_tp, r, a, c, mk, delete_rules, e, more)) = delete_rules;;

let rec eb (Ac_tp_ops_ext (ac_tp, r, a, c, mk, delete_rules, e, more)) = e;;

let rec ac_rule_removal (_B1, _B2, _B3) (_C1, _C2, _C3)
  i rp r_remove tp =
    (match
      catch_error
        (let r = re i tp in
         let e = eb i tp in
         let rremove =
           set ((ceq_prod (ceq_term _B2 _C2) (ceq_term _B2 _C2)),
                 (ccompare_prod (ccompare_term _B1 _C1)
                   (ccompare_term _B1 _C1)),
                 (set_impl_prod set_impl_term set_impl_term))
             r_remove
           in
         let rns =
           filtera
             (fun lr ->
               not (member
                     ((ceq_prod (ceq_term _B2 _C2) (ceq_term _B2 _C2)),
                       (ccompare_prod (ccompare_term _B1 _C1)
                         (ccompare_term _B1 _C1)))
                     lr rremove))
             r
           in
          bindb (valid rp)
            (fun _ ->
              bindb (mono rp (r_remove @ rns @ e))
                (fun _ ->
                  bindb (catch_error
                          (catch_error (forallM (ns rp) rns)
                            (fun x -> Inl (snd x)))
                          (fun x ->
                            Inl (comp (showsl_lit
"problem when orienting weak rules\092")
                                  x)))
                    (fun _ ->
                      bindb (catch_error
                              (catch_error (forallM (ns rp) e)
                                (fun x -> Inl (snd x)))
                              (fun x ->
                                Inl (comp (showsl_lit
    "problem when orienting equations\092")
                                      x)))
                        (fun _ ->
                          catch_error
                            (catch_error (forallM (s rp) r_remove)
                              (fun x -> Inl (snd x)))
                            (fun x ->
                              Inl (comp (showsl_lit
  "problem when orienting strict rules\092")
                                    x)))))))
        (fun x ->
          Inl (comp (comp (comp (showsl_lit
                                  "could not apply AC rule removal with the following\092")
                            (desc rp))
                      (showsl_literal "\092"))
                x))
      with Inl a -> Inl a | Inr _ -> Inr (delete_rulesa i tp r_remove));;

let rec showsl_ac_tp _B _C
  i t = (let r = re i t in
         let aa = a i t in
         let ca = c i t in
          comp (comp (showsl_trsa (showsl _B) (showsl _C) "rules:" " -> " r)
                 (if null aa then id
                   else comp (comp (showsl_lit "A-symbols: ")
                               (showsl_lista _B aa))
                          (showsl_literal "\092")))
            (if null ca then id
              else comp (comp (showsl_lit "C-symbols: ") (showsl_lista _B ca))
                     (showsl_literal "\092")));;

let rec check_ac_termination_proof (_B1, _B2, _B3) (_C1, _C2, _C3)
  ia j i tp x4 = match ia, j, i, tp, x4 with
    ia, j, i, tp, AC_DP_Trans (info, prf1, prf2) ->
      debug i "AC Dependency Pairs"
        (bindb
          (catch_error
            (ac_dependency_pairs_proc
              (cenum_lab, (ceq_lab _B2 _C2),
                (ccompare_lab _B1.compare_compare_order
                  _C1.compare_compare_order),
                (compare_lab _B1.compare_compare_order
                  _C1.compare_compare_order),
                (equal_lab _B2 _C2), set_impl_lab, (showl_lab _B3 _C3))
              (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
                (cproper_interval_list ccompare_char),
                (compare_list compare_char), (equal_list equal_char),
                mapping_impl_list, set_impl_list, (showl_list showl_char))
              j ia (fun a -> Sharp a)
              [Chara (false, false, false, true, true, true, true, false)]
              [Chara (true, false, false, true, true, true, true, false)]
              [Chara (false, true, false, true, true, true, true, false)] info
              tp)
            (fun x ->
              Inl (comp (comp (comp i
                                (showsl_lit
                                  ": error when applying the ac-dependency pair processor on\092"))
                          (showsl_ac_tp (showl_lab _B3 _C3)
                            (showl_list showl_char) j tp))
                    x)))
          (fun (dp1, dp2) ->
            bindb (catch_error
                    (check_ac_dp_termination_proof
                      (cenum_lab, (ceq_lab _B2 _C2),
                        (ccompare_lab _B1.compare_compare_order
                          _C1.compare_compare_order),
                        (compare_order_lab _B1.compare_compare_order
                          _C1.compare_compare_order),
                        (equal_lab _B2 _C2), mapping_impl_lab, set_impl_lab,
                        (showl_lab _B3 _C3))
                      ia (add_index i one_nata) dp1 prf1)
                    (fun x ->
                      Inl (comp (comp i
                                  (showsl_lit
                                    ": error below the ac-dependency pair processor\092"))
                            x)))
              (fun _ ->
                catch_error
                  (check_ac_dp_termination_proof
                    (cenum_lab, (ceq_lab _B2 _C2),
                      (ccompare_lab _B1.compare_compare_order
                        _C1.compare_compare_order),
                      (compare_order_lab _B1.compare_compare_order
                        _C1.compare_compare_order),
                      (equal_lab _B2 _C2), mapping_impl_lab, set_impl_lab,
                      (showl_lab _B3 _C3))
                    ia (add_index i (nat_of_integer (Big_int.big_int_of_int 2)))
                    dp2 prf2)
                  (fun x ->
                    Inl (comp (comp i
                                (showsl_lit
                                  ": error below the ac-dependency pair processor\092"))
                          x)))))
    | ia, j, i, tp, AC_DP_Trans_Single (info, prf1) ->
        debug i "AC Dependency Pairs"
          (bindb
            (catch_error
              (ac_dependency_pairs_proc_simple
                (cenum_lab, (ceq_lab _B2 _C2),
                  (ccompare_lab _B1.compare_compare_order
                    _C1.compare_compare_order),
                  (compare_lab _B1.compare_compare_order
                    _C1.compare_compare_order),
                  (equal_lab _B2 _C2), set_impl_lab, (showl_lab _B3 _C3))
                (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
                  (cproper_interval_list ccompare_char),
                  (compare_list compare_char), (equal_list equal_char),
                  mapping_impl_list, set_impl_list, (showl_list showl_char))
                j ia (fun a -> Sharp a)
                [Chara (false, false, false, true, true, true, true, false)]
                [Chara (true, false, false, true, true, true, true, false)]
                [Chara (false, true, false, true, true, true, true, false)] info
                tp)
              (fun x ->
                Inl (comp (comp (comp i
                                  (showsl_lit
                                    ": error when applying the ac-dependency pair processor on\092"))
                            (showsl_ac_tp (showl_lab _B3 _C3)
                              (showl_list showl_char) j tp))
                      x)))
            (fun dp1 ->
              catch_error
                (check_ac_dp_termination_proof
                  (cenum_lab, (ceq_lab _B2 _C2),
                    (ccompare_lab _B1.compare_compare_order
                      _C1.compare_compare_order),
                    (compare_order_lab _B1.compare_compare_order
                      _C1.compare_compare_order),
                    (equal_lab _B2 _C2), mapping_impl_lab, set_impl_lab,
                    (showl_lab _B3 _C3))
                  ia (add_index i one_nata) dp1 prf1)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the ac-dependency pair processor\092"))
                        x))))
    | ia, j, i, tp, AC_Rule_Removal (redp, del_r, prf) ->
        debug i "AC_Mono_Redpair_UR_Proc"
          (bindb
            (catch_error
              (ac_rule_removal
                ((compare_lab _B1.compare_compare_order
                   _C1.compare_compare_order),
                  (equal_lab _B2 _C2), (showl_lab _B3 _C3))
                ((compare_list compare_char), (equal_list equal_char),
                  (showl_list showl_char))
                j (get_redtriple
                    ((ceq_lab _B2 _C2),
                      (ccompare_lab _B1.compare_compare_order
                        _C1.compare_compare_order),
                      (compare_order_lab _B1.compare_compare_order
                        _C1.compare_compare_order),
                      (equal_lab _B2 _C2), set_impl_lab, (showl_lab _B3 _C3))
                    redp)
                del_r tp)
              (fun x ->
                Inl (comp (comp (comp (comp
(comp (comp i
        (showsl_lit
          ": error when applying ac-rule removal to remove from the AC termination problem\092"))
  (showsl_ac_tp (showl_lab _B3 _C3) (showl_list showl_char) j tp))
(showsl_lit "\092 the rules\092"))
                                  (showsl_rules (showl_lab _B3 _C3)
                                    (showl_list showl_char) del_r))
                            (showsl_literal "\092"))
                      x)))
            (fun tpa ->
              catch_error
                (check_ac_termination_proof (_B1, _B2, _B3) (_C1, _C2, _C3) ia j
                  (add_index i one_nata) tpa prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit ": error below ac-rule removal\092"))
                        x))))
    | ia, j, i, tp, AC_R_is_Empty ->
        debug i "AC_R_is_Empty"
          (catch_error
            (check (null (re j tp)) (showsl_lit "The TRS is not empty"))
            (fun x ->
              Inl (comp (comp (comp (comp i
                                      (showsl_lit
": error when applying the R-is-Empty check on the AC termination problem\092"))
                                (showsl_ac_tp (showl_lab _B3 _C3)
                                  (showl_list showl_char) j tp))
                          (showsl_literal "\092"))
                    x)));;

let rec equal_strategy _A _B
  x0 x1 = match x0, x1 with Innermost, Innermost_Q x3 -> false
    | Innermost_Q x3, Innermost -> false
    | No_Strategy, Innermost_Q x3 -> false
    | Innermost_Q x3, No_Strategy -> false
    | No_Strategy, Innermost -> false
    | Innermost, No_Strategy -> false
    | Innermost_Q x3, Innermost_Q y3 -> equal_lista (equal_term _A _B) x3 y3
    | Innermost, Innermost -> true
    | No_Strategy, No_Strategy -> true;;

let rec showsl_transition _D
  showsl_tatom (Transition (s, t, phi)) =
    comp (comp (comp (comp (showsl _D s) (showsl_lit " ---> ")) (showsl _D t))
           (showsl_lit ": "))
      (showsl_formula showsl_tatom phi);;

let rec transition_hintsa
  (Hinter_ext (nodes, succ_trans_list, cover_hints, transition_hints, more)) =
    transition_hints;;

let rec node_invariant
  (Art_ext (initial_nodes, nodes, edge, node_location, node_invariant, more)) =
    node_invariant;;

let rec rename_vars_exp r = map_term id (map_prod r id);;

let rec nodesa
  (Hinter_ext (nodes, succ_trans_list, cover_hints, transition_hints, more)) =
    nodes;;

let rec map_formula
  f x1 = match f, x1 with f, Atom x1 -> Atom (f x1)
    | f, NegAtom x2 -> NegAtom (f x2)
    | f, Conjunction x3 -> Conjunction (map (map_formula f) x3)
    | f, Disjunction x4 -> Disjunction (map (map_formula f) x4);;

let rec edge
  (Art_ext (initial_nodes, nodes, edge, node_location, node_invariant, more)) =
    edge;;

let rec check_children_edges_cond (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2) _E
  _F tc2 sa2 ne2 a p h =
       (let lc = assertion p in
         catch_error
           (catch_error
             (forallM
               (fun aa ->
                 (match edge a aa with Cover _ -> Inr ()
                   | Children children ->
                     (match transition_hintsa h aa
                       with None -> Inl (showsl_lit "not yet")
                       | Some hints ->
                         catch_error
                           (bindb
                             (check
                               (equal_nata (size_list hints)
                                 (size_list children))
                               (showsl_lit
                                 "the number of hints differs from the number of children"))
                             (fun _ ->
                               catch_error
                                 (forallM
                                   (fun (b, c) ->
                                     (let (tau, ba) = b in
                                       (fun ha ->
 (let Transition (l, _, phi) = tau in
   catch_error
     (check_formula _B1 ((equal_trans_var _C1), (showl_trans_var _C2)) _D1
       (_A1, _A2) sa2 tc2 ne2 ha
       (Disjunction
         [map_formula (rename_vars_exp (fun ab -> Post ab))
            (node_invariant a ba);
           form_not
             (map_formula (rename_vars_exp (fun ab -> Pre ab))
               (node_invariant a aa));
           form_not (map_formula (rename_vars_exp (fun ab -> Pre ab)) (lc l));
           form_not phi]))
     (fun x ->
       Inl (comp (comp (comp (comp (comp (comp
   (comp (showsl_lit "problem in checking edge ") (showsl _F aa))
   (showsl_lit " --> "))
                                     (showsl _F ba))
                               (showsl_lit " for transition\092"))
                         (showsl_transition _E sa2 tau))
                   (showsl_literal "\092"))
             x)))))
                                       c)
                                   (zip children hints))
                                 (fun x -> Inl (snd x))))
                           (fun x ->
                             Inl (comp (comp
 (comp (showsl_lit "problem in checking transitions of node ") (showsl _F aa))
 (showsl_literal "\092"))
                                   x)))))
               (nodesa h))
             (fun x -> Inl (snd x)))
           (fun x ->
             Inl (comp (showsl_lit
                         "could not ensure transition edge condition of ART\092")
                   x)));;

let rec is_cover_nodea = function Cover x1 -> true
                         | Children x2 -> false;;

let rec cover_hintsa
  (Hinter_ext (nodes, succ_trans_list, cover_hints, transition_hints, more)) =
    cover_hints;;

let rec node_location
  (Art_ext (initial_nodes, nodes, edge, node_location, node_invariant, more)) =
    node_location;;

let rec nodes
  (Art_ext (initial_nodes, nodes, edge, node_location, node_invariant, more)) =
    nodes;;

let rec check_cover_edges_cond (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2) _E
  (_F1, _F2)
  tc sa ne a p h =
    (let _ = assertion p in
      catch_error
        (catch_error
          (forallM
            (fun aa ->
              (match edge a aa
                with Cover b ->
                  catch_error
                    (bindb
                      (check (membera _F1 (nodes a) b)
                        (showsl_lit "target node is not listed as art-node"))
                      (fun _ ->
                        bindb (check
                                (eq _E (node_location a aa) (node_location a b))
                                (showsl_lit "node-locations differ"))
                          (fun _ ->
                            bindb (check (not (is_cover_nodea (edge a b)))
                                    (showsl_lit
                                      "target node must not have cover edge"))
                              (fun _ ->
                                check_formula _B1 (_C1, _C2) _D1 (_A1, _A2) sa
                                  tc ne (cover_hintsa h aa)
                                  (Disjunction
                                    [node_invariant a b;
                                      form_not (node_invariant a aa)])))))
                    (fun x ->
                      Inl (comp (comp (comp
(showsl_lit "problem in checking cover edge of node ") (showsl _F2 aa))
                                  (showsl_literal "\092"))
                            x))
                | Children _ -> Inr ()))
            (nodesa h))
          (fun x -> Inl (snd x)))
        (fun x ->
          Inl (comp (showsl_lit
                      "could not ensure cover edge condition of ART\092")
                x)));;

let rec succ_trans_list
  (Hinter_ext (nodes, succ_trans_list, cover_hints, transition_hints, more)) =
    succ_trans_list;;

let rec initial_nodes
  (Art_ext (initial_nodes, nodes, edge, node_location, node_invariant, more)) =
    initial_nodes;;

let rec check_simulation_cond (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2)
  (_E1, _E2, _E3, _E4, _E5, _E6) (_F1, _F2)
  tc sa ne a p h =
    catch_error
      (bindb
        (check
          (less_eq_set (_E1, _E2, _E3) (initial p)
            (set (_E2, _E3, _E5) (map (node_location a) (initial_nodes a))))
          (showsl_lit
            "not all initial nodes of LTS are represented by initial nodes in ART"))
        (fun _ ->
          catch_error
            (forallM
              (fun aa ->
                (match edge a aa with Cover _ -> Inr ()
                  | Children children ->
                    (let l = node_location a aa in
                      catch_error
                        (catch_error
                          (forallM
                            (fun tau ->
                              check (list_ex
                                      (fun (taua, b) ->
equal_transition_rule _B1 _C1 _D1 _E4 taua tau &&
  (eq _E4 (node_location a b) (target tau) && membera _F1 (nodes a) b))
                                      children)
                                (comp (comp
(comp (comp (comp (showsl_lit
                    "could not find matching transition in ART for node ")
              (showsl _F2 aa))
        (showsl_lit " and transition "))
  (showsl _E6 (source tau)))
(showsl_lit " --> "))
                                  (showsl _E6 (target tau))))
                            (succ_trans_list h l))
                          (fun x -> Inl (snd x)))
                        (fun e1 ->
                          catch_error
                            (check_valid_formula _B1 (_C1, _C2) _D1 (_A1, _A2)
                              sa tc ne (form_not (node_invariant a aa)))
                            (fun _ ->
                              Inl (comp e1
                                    (showsl_lit
                                      "\092and could not prove unsatisfiability of the node invariant")))))))
              (nodesa h))
            (fun x -> Inl (snd x))))
      (fun x ->
        Inl (comp (showsl_lit
                    "could not ensure simulation condition of ART\092")
              x));;

let rec check_initial_cond (_A1, _A2) (_B1, _B2) (_C1, _C2) (_E1, _E2)
  sa a =
    catch_error
      (forallM
        (fun init ->
          bindb (check (membera _E1 (nodes a) init)
                  (comp (comp (showsl_lit "initial node of A (")
                          (showsl _E2 init))
                    (showsl_lit ") is not mentioned as node of ART")))
            (fun _ ->
              check (equal_formulaa (equal_term _A1 (equal_prod _B1 _C1))
                      (node_invariant a init) (Conjunction []))
                (comp (comp (comp (showsl_lit
                                    "the node invariant for the initial ART node (")
                              (showsl _E2 init))
                        (showsl_lit ") must be TRUE,  but it is "))
                  (showsl_formula sa (node_invariant a init)))))
        (initial_nodes a))
      (fun x -> Inl (snd x));;

let rec art _A (_B1, _B2, _B3, _B4)
  type_of_fun bool_types a =
    list_all
      (fun aa ->
        formula (is_bool (_B1, _B2, _B3) type_of_fun bool_types)
          (node_invariant a aa))
      (nodes a);;

let rec check_art _A (_B1, _B2, _B3, _B4)
  type_of_fun bool_types a =
    check (art _A (_B1, _B2, _B3, _B4) type_of_fun bool_types a)
      (showsl_lit "ill-formed invariant");;

let rec check_art_invariants (_A1, _A2) (_B1, _B2, _B3, _B4) (_C1, _C2)
  (_D1, _D2) (_E1, _E2, _E3, _E4, _E5, _E6) (_F1, _F2)
  type_of_fun bool_types tc tc2 sa sa2 ne ne2 a p h =
    catch_error
      (bindb (check_art _A2 (_B1, _B2, _B3, _B4) type_of_fun bool_types a)
        (fun _ ->
          bindb (check_initial_cond (_A1, _A2) (_D1, _D2) (_B3, _B4) (_F1, _F2)
                  sa a)
            (fun _ ->
              bindb (check_simulation_cond (_C1, _C2) (_A1, _A2) (_D1, _D2)
                      (_B3, _B4) (_E1, _E2, _E3, _E4, _E5, _E6) (_F1, _F2) tc sa
                      ne a p h)
                (fun _ ->
                  bindb (check_cover_edges_cond (_C1, _C2) (_A1, _A2) (_D1, _D2)
                          (_B3, _B4) _E4 (_F1, _F2) tc sa ne a p h)
                    (fun _ ->
                      check_children_edges_cond (_C1, _C2) (_A1, _A2) (_D1, _D2)
                        (_B3, _B4) _E6 _F2 tc2 sa2 ne2 a p h)))))
      (fun x ->
        Inl (comp (showsl_lit "could not ensure validity of art-graph:\092")
              x));;

let rec is_cover_node
  = function Art_Node (uu, uv, uw, Cover_Edge (ux, uy)) -> true
    | Art_Node (uz, va, vb, Children_Edge vc) -> false;;

let rec transition_hints _E
  ai = map_of _E
         (map_filter
           (fun x ->
             (if not (is_cover_node x)
               then Some (name x,
                           (let Children_Edge a = edgea x in
                             map (fun (_, (_, h)) -> h) a))
               else None))
           (nodesb ai));;

let rec cover_hints (_E1, _E2, _E3) _G
  ai = map_of_default (_E1, _E2, _E3) (default _G)
         (map_filter
           (fun x ->
             (if is_cover_node x
               then Some (name x, (let Cover_Edge (_, h) = edgea x in h))
               else None))
           (nodesb ai));;

let rec make_hinter _D (_J1, _J2, _J3) _L
  pi ai =
    Hinter_ext
      (art_nodes ai, succ_transitions _D pi, cover_hints (_J1, _J2, _J3) _L ai,
        transition_hints _J2 ai, ());;

let rec check_art_invariants_impl (_A1, _A2, _A3) (_B1, _B2, _B3, _B4)
  (_C1, _C2) (_D1, _D2, _D3) (_E1, _E2, _E3, _E4, _E5, _E6, _E7)
  (_F1, _F2, _F3, _F4) (_G1, _G2, _G3, _G4, _G5, _G6)
  type_of_fun bool_types tc tc2 sa sa2 ne ne2 ai pi =
    check_art_invariants (_A2, _A3) (_B1, _B2, _B3, _B4) (_C1, _C2) (_D2, _D3)
      (_E1, _E2, _E3, _E4, _E6, _E7) (_F2, _F4) type_of_fun bool_types tc tc2 sa
      sa2 ne ne2 (art_of (_G2, _G3, _G4, _G6) (_F1, _F2, _F3, _F4) pi ai)
      (lts_of (_A1, _A2) (_D1, _D2) (_B2, _B3) (_E2, _E3, _E4, _E5, _E6)
        (_G1, _G2, _G5) pi)
      (make_hinter _E4 (_F1, _F2, _F3) (default_hint _C1) pi ai);;

let rec check_unique_names _E
  ai = check (distinct _E (map name (nodesb ai)))
         (showsl_lit "Nodes in art graph must have unique names");;

let rec restrict_invariants (_A1, _A2)
  i la l = (if member (_A1, _A2) l la then i l else Conjunction []);;

let rec showsl_labeled_transition _D _E
  showsl_tatom (lab, tran) =
    comp (comp (showsl _D lab) (showsl_lit ": "))
      (showsl_transition _E showsl_tatom tran);;

let rec showsl_lts _D _E
  showsl_atom showsl_tatom (Lts_Impl (i, tran, lc)) =
    comp (comp (comp (comp (comp (comp (showsl_lit
 "LTS:\092Initial locations: ")
                                   (showsl_list _D i))
                             (showsl_lit "\092Transitions\092"))
                       (showsl_sep
                         (showsl_labeled_transition _E _D showsl_tatom)
                         (showsl_literal "\092") tran))
                 (showsl_lit "\092Location conditions"))
           (showsl_sep
             (fun (l, f) ->
               comp (comp (showsl _D l) (showsl_lit ": "))
                 (showsl_formula showsl_atom f))
             (showsl_literal "\092") lc))
      (showsl_literal "\092");;

let rec get_disj_invariant _D
  a l = Disjunction
          (map_filter
            (fun x ->
              (if eq _D (node_location a x) l && not (is_cover_nodea (edge a x))
                then Some (node_invariant a x) else None))
            (nodes a));;

let rec check_trivial_implication _A
  phi psi =
    list_all
      (fun phia ->
        (let c_phis = get_conjunctions phia in
          list_ex
            (fun psia ->
              list_all (membera (equal_formula _A) c_phis)
                (get_conjunctions psia))
            (get_disjunctions psi)))
      (get_disjunctions phi);;

let rec showsl_art_node _D _E _F
  = function
    Art_Node (n, uu, l, Cover_Edge (m, uv)) ->
      comp (comp (comp (comp (showsl _E n) (showsl_lit "(@ ")) (showsl _D l))
             (showsl_lit "): covered by "))
        (showsl _E m)
    | Art_Node (n, uw, l, Children_Edge ls) ->
        comp (comp (comp (comp (showsl _E n) (showsl_lit "(@ ")) (showsl _D l))
               (showsl_lit "): goes to "))
          (default_showsl_list
            (fun (tr, (na, _)) ->
              comp (comp (comp (showsl_lit "-") (showsl _F tr))
                     (showsl_lit "->"))
                (showsl _E na))
            ls);;

let rec showsl_art _D _E _F
  a = comp (comp (comp (comp (showsl_lit "ART:\092Initial node: ")
                         (showsl_list _E (initial_nodesa a)))
                   (showsl_lit "\092Arcs\092"))
             (showsl_sep (showsl_art_node _D _E _F) (showsl_literal "\092")
               (nodesb a)))
        (showsl_literal "\092");;

let rec invariant_proof_checker (_A1, _A2, _A3) (_B1, _B2, _B3, _B4) (_C1, _C2)
  (_D1, _D2, _D3) (_E1, _E2, _E3, _E4, _E5, _E6, _E7)
  (_F1, _F2, _F3, _F4, _F5, _F6) (_G1, _G2, _G3, _G4)
  type_of_fun bool_types tc tc2 sa sa2 ne ne2 pi (Impact (ii, ai)) =
    debug id "invariant checking: Impact"
      (let i = map_of_default (_E3, _E4, _E5) (Conjunction []) ii in
        (match
          catch_error
            (bindb (check_unique_names _G2 ai)
              (fun _ ->
                bindb (debug id
                        "provided invariants against deduced invariants from ART graph"
                        (catch_error
                          (forallM
                            (fun l ->
                              (let il = i l in
                                bindb (check
(formula (is_bool (_B1, _B2, _B3) type_of_fun bool_types) il)
(comp (showsl_lit "ill-formed formula for location ") (showsl _E7 l)))
                                  (fun _ ->
                                    check (check_trivial_implication
    (equal_term _A2 (equal_prod _D2 _B3))
    (get_disj_invariant _E4
      (art_of (_F2, _F3, _F4, _F6) (_G1, _G2, _G3, _G4) pi ai) l)
    il)
                                      (comp
(showsl_lit
  "could not match provided invariant with invariant extracted from art-graph at location ")
(showsl _E7 l)))))
                            (nodes_lts_impl _E4 pi))
                          (fun x -> Inl (snd x))))
                  (fun _ ->
                    debug id "checking ART graph"
                      (check_art_invariants_impl (_A1, _A2, _A3)
                        (_B1, _B2, _B3, _B4) (_C1, _C2) (_D1, _D2, _D3)
                        (_E1, _E2, _E3, _E4, _E5, _E6, _E7) (_G1, _G2, _G3, _G4)
                        (_F1, _F2, _F3, _F4, _F5, _F6) type_of_fun bool_types tc
                        tc2 sa sa2 ne ne2 ai pi))))
            (fun x ->
              Inl (comp (comp (comp (comp (showsl_lit
    "problem in ensuring invariants for ")
                                      (showsl_lts _E7 _F6 sa sa2 pi))
                                (showsl_lit "\092via "))
                          (showsl_art _E7 _G4 _F6 ai))
                    x))
          with Inl a -> Inl a
          | Inr _ ->
            Inr (restrict_invariants (_E2, _E3) i
                  (set (_E2, _E3, _E6) (nodes_lts_impl _E4 pi)))));;

let rec safe_by_assertion_checker _A (_B1, _B2) _C (_D1, _D2)
  (_E1, _E2, _E3, _E4)
  showsl_atom logic_checker negate_atom pi err =
    catch_error
      (forallM
        (fun l ->
          catch_error
            (check_valid_formula _A (_B1, _B2) _C (_D1, _D2) showsl_atom
              logic_checker negate_atom
              (form_not (assertion_of (_E1, _E2, _E3) pi l)))
            (fun _ ->
              Inl (comp (comp (showsl_lit
                                "could not deduce from assertion that ")
                          (showsl _E4 l))
                    (showsl_lit " is unreachable"))))
        err)
      (fun x -> Inl (snd x));;

let rec fix_invariants (_D1, _D2, _D3, _D4) _E
  p phi =
    catch_error
      (let ls = nodes_lts_impl _D2 p in
       let _ = assertion_of (_D1, _D2, _D3) p in
        Inr (Lts_Impl
              (initiala p, transitions_impl p, map (fun l -> (l, phi l)) ls)))
      (fun x ->
        Inl (comp (showsl_lit
                    "problem when fixing invariants as assertions\092")
              x));;

let rec check_safety_proof (_A1, _A2, _A3) (_B1, _B2, _B3, _B4) (_C1, _C2)
  (_D1, _D2, _D3) (_E1, _E2, _E3, _E4, _E5, _E6, _E7)
  (_F1, _F2, _F3, _F4, _F5, _F6) (_G1, _G2, _G3, _G4)
  type_of_fun bool_types tc tc2 sa sa2 ne ne2 pi err x10 = match
    type_of_fun, bool_types, tc, tc2, sa, sa2, ne, ne2, pi, err, x10 with
    type_of_fun, bool_types, tc, tc2, sa, sa2, ne, ne2, pi, err,
      Invariant_Assertion (inv_prf, inner)
      -> debug id "Add Invariants"
           (bindb
             (invariant_proof_checker (_A1, _A2, _A3) (_B1, _B2, _B3, _B4)
               (_C1, _C2) (_D1, _D2, _D3) (_E1, _E2, _E3, _E4, _E5, _E6, _E7)
               (_F1, _F2, _F3, _F4, _F5, _F6) (_G1, _G2, _G3, _G4) type_of_fun
               bool_types tc tc2 sa sa2 ne ne2 pi inv_prf)
             (fun i ->
               bindb (fix_invariants (_E3, _E4, _E5, _E7) _F6 pi i)
                 (fun qi ->
                   check_safety_proof (_A1, _A2, _A3) (_B1, _B2, _B3, _B4)
                     (_C1, _C2) (_D1, _D2, _D3)
                     (_E1, _E2, _E3, _E4, _E5, _E6, _E7)
                     (_F1, _F2, _F3, _F4, _F5, _F6) (_G1, _G2, _G3, _G4)
                     type_of_fun bool_types tc tc2 sa sa2 ne ne2 qi err inner)))
    | type_of_fun, bool_types, tc, tc2, sa, sa2, ne, ne2, pi, err, Trivial ->
        debug id "Unsatisfiable Error states"
          (safe_by_assertion_checker _A2 (_D2, _D3) _B3 (_C1, _C2)
            (_E3, _E4, _E5, _E7) sa tc ne pi err);;

let rec check_safety (_A1, _A2, _A3) (_B1, _B2, _B3, _B4) (_C1, _C2)
  (_D1, _D2, _D3) (_E1, _E2, _E3, _E4, _E5, _E6, _E7)
  (_F1, _F2, _F3, _F4, _F5, _F6) (_G1, _G2, _G3, _G4)
  type_of_fun bool_types tc tc2 sa sa2 ne ne2 pi err prf =
    bindb (debug id "init - Check well-formedness"
            (check_lts_impl (_B1, _B2, _B3) type_of_fun bool_types pi))
      (fun _ ->
        check_safety_proof (_A1, _A2, _A3) (_B1, _B2, _B3, _B4) (_C1, _C2)
          (_D1, _D2, _D3) (_E1, _E2, _E3, _E4, _E5, _E6, _E7)
          (_F1, _F2, _F3, _F4, _F5, _F6) (_G1, _G2, _G3, _G4) type_of_fun
          bool_types tc tc2 sa sa2 ne ne2 pi err prf);;

let rec check_single_subsumption (_A1, _A2) (_B1, _B2, _B3, _B4)
  eq e convs =
    bindb (check
            (membera (equal_prod (equal_term _A1 _B2) (equal_term _A1 _B2))
              (map fst convs) eq)
            (comp (showsl_lit "could not find conversion for equation ")
              (showsl_eq _A2 _B4 eq)))
      (fun _ ->
        check_subsumptions_guided (_A1, _A2) (_B1, _B2, _B3, _B4) e convs);;

let rec showsl_ln
  i = comp (comp (showsl_literal "\092") (showsl_nat i)) (showsl_lit ": ");;

let rec eq_proof_lines _A _B
  x0 i = match x0, i with
    Refl s, i ->
      (comp (comp (showsl_ln (suc i)) (showsl_eq _A _B (s, s)))
         (showsl_lit " [refl]"),
        (suc i, (s, s)))
    | Sym p, i ->
        (let (s, (ia, (l, r))) = eq_proof_lines _A _B p i in
          (comp (comp (comp (comp (comp s (showsl_ln (suc ia)))
                              (showsl_eq _A _B (r, l)))
                        (showsl_lit " [sym "))
                  (showsl_nat ia))
             (showsl_lit "]"),
            (suc ia, (r, l))))
    | Trans (p1, p2), i ->
        (let (s1, (i1, (s, _))) = eq_proof_lines _A _B p1 i in
         let (s2, (i2, (_, v))) = eq_proof_lines _A _B p2 i1 in
          (comp (comp (comp (comp (comp (comp
  (comp (comp s1 s2) (showsl_ln (suc i2))) (showsl_eq _A _B (s, v)))
                                    (showsl_lit " [trans "))
                              (showsl_nat i1))
                        (showsl_lit ", "))
                  (showsl_nat i2))
             (showsl_lit "]"),
            (suc i2, (s, v))))
    | Assm ((l, r), sigma), i ->
        (let eq = (subst_apply_term l sigma, subst_apply_term r sigma) in
          (comp (comp (comp (comp (showsl_ln (suc i)) (showsl_eq _A _B eq))
                        (showsl_lit " [assm "))
                  (showsl_eq _A _B (l, r)))
             (showsl_lit "]"),
            (suc i, eq)))
    | Cong (f, ps), i ->
        (let (s, (is, (ls, rs))) = eq_proofs_lines _A _B ps i in
         let eq = (Fun (f, ls), Fun (f, rs)) in
         let ia = last is in
         let isa = butlast is in
          (comp (comp (comp s (showsl_ln (suc ia))) (showsl_eq _A _B eq))
             (showsl_list_gen showsl_nat " [cong]" " [cong " ", " "]" isa),
            (suc ia, eq)))
and eq_proofs_lines _A _B
  x0 i = match x0, i with [], i -> (id, ([i], ([], [])))
    | p :: ps, i ->
        (let (s1, (ia, (l, r))) = eq_proof_lines _A _B p i in
         let (s2, (is, (ls, rs))) = eq_proofs_lines _A _B ps ia in
          (comp s1 s2, (ia :: is, (l :: ls, r :: rs))));;

let rec check_proves (_A1, _A2) (_B1, _B2)
  e x1 = match e, x1 with e, Refl s -> Inr (s, s)
    | e, Sym p ->
        bindb (check_proves (_A1, _A2) (_B1, _B2) e p)
          (fun (s, t) -> Inr (t, s))
    | e, Trans (p1, p2) ->
        bindb (check_proves (_A1, _A2) (_B1, _B2) e p1)
          (fun (s, t) ->
            bindb (check_proves (_A1, _A2) (_B1, _B2) e p2)
              (fun (ta, u) ->
                (if equal_terma _A1 _B1 t ta then Inr (s, u)
                  else Inl (comp (comp (comp
 (comp (comp (showsl_lit "the error occurs in the following part \092")
         (fst (eq_proof_lines _A2 _B2 (Trans (p1, p2)) zero_nata)))
   (showsl_lit "\092\092"))
 (showsl_terma _A2 _B2 t))
                                   (showsl_lit " is not equal to "))
                             (showsl_terma _A2 _B2 ta)))))
    | e, Assm ((l, r), sigma) ->
        (if membera (equal_prod (equal_term _A1 _B1) (equal_term _A1 _B1)) e
              (l, r)
          then Inr (subst_apply_term l sigma, subst_apply_term r sigma)
          else Inl (comp (comp (comp (comp (comp
     (showsl_lit "the error occurs in the following part \092")
     (fst (eq_proof_lines _A2 _B2 (Assm ((l, r), sigma)) zero_nata)))
                                       (showsl_lit "\092\092"))
                                 (showsl_eqa _A2 _B2 (l, r)))
                           (showsl_lit " is not in the "))
                     (showsl_eqs _A2 _B2 e)))
    | e, Cong (f, ps) ->
        bindb (mapM (check_proves (_A1, _A2) (_B1, _B2) e) ps)
          (fun sts -> Inr (Fun (f, map fst sts), Fun (f, map snd sts)));;

let rec check_eq_proof (_A1, _A2) (_B1, _B2)
  e p eq =
    catch_error
      (bindb (check_proves (_A1, _A2) (_B1, _B2) e p)
        (fun eqa ->
          (if equal_proda (equal_term _A1 _B1) (equal_term _A1 _B1) eq eqa
            then Inr ()
            else Inl (showsl_lit "the proof does not fit the goal"))))
      (fun x ->
        Inl (comp (comp (comp (comp (comp (comp
    (comp (showsl_lit "there is an error in the equational logic proof\092")
      (fst (eq_proof_lines _A2 _B2 p zero_nata)))
    (showsl_lit "\092\092for proving the equation\092\092"))
                                      (showsl_eqa _A2 _B2 eq))
                                (showsl_lit "\092\092using the "))
                          (showsl_eqs _A2 _B2 e))
                    (showsl_literal "\092"))
              x));;

let rec check_equational_proof (_B1, _B2, _B3, _B4)
  a ia i j e x5 va = match a, ia, i, j, e, x5, va with
    a, ia, i, j, e, Equation eq, Equational_Proof_Tree p ->
      debug ia "Equational_Proof_Tree"
        (check_eq_proof
          ((equal_lab _B3 (equal_list equal_nat)),
            (showl_lab _B4 (showl_list showl_nat)))
          ((equal_list equal_char), (showl_list showl_char)) e p eq)
    | a, ia, i, j, e, Equation eq, Conversion eseq ->
        debug ia "Conversion"
          (check_conversiona
            ((equal_lab _B3 (equal_list equal_nat)),
              (showl_lab _B4 (showl_list showl_nat)))
            ((ccompare_list ccompare_char), (equal_list equal_char),
              mapping_impl_list, (showl_list showl_char))
            e eseq (fst eq) (snd eq))
    | a, ia, i, j, e, Equation eq, Conversion_With_History convs ->
        debug ia "Conversion with History"
          (check_single_subsumption
            ((equal_lab _B3 (equal_list equal_nat)),
              (showl_lab _B4 (showl_list showl_nat)))
            ((ccompare_list ccompare_char), (equal_list equal_char),
              mapping_impl_list, (showl_list showl_char))
            eq e convs)
    | a, ia, i, j, e, Equation eq, Completion_and_Normalization (r, p) ->
        debug ia "Completion_and_Normalization"
          (bindb (check_completion_proof (_B1, _B2, _B3, _B4) a ia i j e r p)
            (fun _ ->
              (let s = fst eq in
               let t = snd eq in
                (match
                  (compute_rstep_NF (equal_lab _B3 (equal_list equal_nat))
                     ((ccompare_list ccompare_char), (equal_list equal_char),
                       mapping_impl_list)
                     r s,
                    compute_rstep_NF (equal_lab _B3 (equal_list equal_nat))
                      ((ccompare_list ccompare_char), (equal_list equal_char),
                        mapping_impl_list)
                      r t)
                  with (None, _) ->
                    Inl (comp (comp (comp (showsl_lit
    "error when computing normal forms of ")
                                      (showsl_terma
(showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char) s))
                                (showsl_lit " and "))
                          (showsl_terma (showl_lab _B4 (showl_list showl_nat))
                            (showl_list showl_char) t))
                  | (Some _, None) ->
                    Inl (comp (comp (comp (showsl_lit
    "error when computing normal forms of ")
                                      (showsl_terma
(showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char) s))
                                (showsl_lit " and "))
                          (showsl_terma (showl_lab _B4 (showl_list showl_nat))
                            (showl_list showl_char) t))
                  | (Some sa, Some ta) ->
                    (if equal_terma (equal_lab _B3 (equal_list equal_nat))
                          (equal_list equal_char) sa ta
                      then Inr ()
                      else Inl (comp (comp (comp
     (showsl_terma (showl_lab _B4 (showl_list showl_nat))
       (showl_list showl_char) s)
     (showsl_lit " and "))
                                       (showsl_terma
 (showl_lab _B4 (showl_list showl_nat)) (showl_list showl_char) t))
                                 (showsl_lit
                                   " have different normal forms")))))))
    | uu, uv, uw, ux, uy, Inequality v, va ->
        Inl (showsl_lit "unsupported equational proof ");;

let set_impl_ty : (ty, set_impla) phantom = Phantom Set_RBT;;

let rec check_dp_termination_proof (_B1, _B2, _B3, _B4)
  ia j a i dpp prf =
    check_dp_termination_proof_main (_B1, _B2, _B3, _B4) ia j a i dpp
      (fst (unlab_to_split_dp (_B1, _B3)
             ((compare_order_list (compare_order_nat, equal_nat)),
               (equal_list equal_nat))
             ((compare_order_list (compare_order_char, equal_char)),
               (equal_list equal_char))
             prf));;

let rec equal_claim
  x0 x1 = match x0, x1 with Ordered_Completed, Anything -> false
    | Anything, Ordered_Completed -> false
    | Completed, Anything -> false
    | Anything, Completed -> false
    | Completed, Ordered_Completed -> false
    | Ordered_Completed, Completed -> false
    | Nonconfluent, Anything -> false
    | Anything, Nonconfluent -> false
    | Nonconfluent, Ordered_Completed -> false
    | Ordered_Completed, Nonconfluent -> false
    | Nonconfluent, Completed -> false
    | Completed, Nonconfluent -> false
    | Confluent, Anything -> false
    | Anything, Confluent -> false
    | Confluent, Ordered_Completed -> false
    | Ordered_Completed, Confluent -> false
    | Confluent, Completed -> false
    | Completed, Confluent -> false
    | Confluent, Nonconfluent -> false
    | Nonconfluent, Confluent -> false
    | Nonterminating, Anything -> false
    | Anything, Nonterminating -> false
    | Nonterminating, Ordered_Completed -> false
    | Ordered_Completed, Nonterminating -> false
    | Nonterminating, Completed -> false
    | Completed, Nonterminating -> false
    | Nonterminating, Nonconfluent -> false
    | Nonconfluent, Nonterminating -> false
    | Nonterminating, Confluent -> false
    | Confluent, Nonterminating -> false
    | Upperbound x4, Anything -> false
    | Anything, Upperbound x4 -> false
    | Upperbound x4, Ordered_Completed -> false
    | Ordered_Completed, Upperbound x4 -> false
    | Upperbound x4, Completed -> false
    | Completed, Upperbound x4 -> false
    | Upperbound x4, Nonconfluent -> false
    | Nonconfluent, Upperbound x4 -> false
    | Upperbound x4, Confluent -> false
    | Confluent, Upperbound x4 -> false
    | Upperbound x4, Nonterminating -> false
    | Nonterminating, Upperbound x4 -> false
    | Terminating, Anything -> false
    | Anything, Terminating -> false
    | Terminating, Ordered_Completed -> false
    | Ordered_Completed, Terminating -> false
    | Terminating, Completed -> false
    | Completed, Terminating -> false
    | Terminating, Nonconfluent -> false
    | Nonconfluent, Terminating -> false
    | Terminating, Confluent -> false
    | Confluent, Terminating -> false
    | Terminating, Nonterminating -> false
    | Nonterminating, Terminating -> false
    | Terminating, Upperbound x4 -> false
    | Upperbound x4, Terminating -> false
    | No, Anything -> false
    | Anything, No -> false
    | No, Ordered_Completed -> false
    | Ordered_Completed, No -> false
    | No, Completed -> false
    | Completed, No -> false
    | No, Nonconfluent -> false
    | Nonconfluent, No -> false
    | No, Confluent -> false
    | Confluent, No -> false
    | No, Nonterminating -> false
    | Nonterminating, No -> false
    | No, Upperbound x4 -> false
    | Upperbound x4, No -> false
    | No, Terminating -> false
    | Terminating, No -> false
    | Yes, Anything -> false
    | Anything, Yes -> false
    | Yes, Ordered_Completed -> false
    | Ordered_Completed, Yes -> false
    | Yes, Completed -> false
    | Completed, Yes -> false
    | Yes, Nonconfluent -> false
    | Nonconfluent, Yes -> false
    | Yes, Confluent -> false
    | Confluent, Yes -> false
    | Yes, Nonterminating -> false
    | Nonterminating, Yes -> false
    | Yes, Upperbound x4 -> false
    | Upperbound x4, Yes -> false
    | Yes, Terminating -> false
    | Terminating, Yes -> false
    | Yes, No -> false
    | No, Yes -> false
    | Upperbound x4, Upperbound y4 -> equal_nata x4 y4
    | Anything, Anything -> true
    | Ordered_Completed, Ordered_Completed -> true
    | Completed, Completed -> true
    | Nonconfluent, Nonconfluent -> true
    | Confluent, Confluent -> true
    | Nonterminating, Nonterminating -> true
    | Terminating, Terminating -> true
    | No, No -> true
    | Yes, Yes -> true;;

let rec type_of_fun = function LessF -> ([IntT; IntT], BoolT)
                      | LeF -> ([IntT; IntT], BoolT)
                      | EqF -> ([IntT; IntT], BoolT)
                      | SumF n -> (replicate n IntT, IntT)
                      | ConstF uu -> ([], IntT)
                      | ProdF n -> (replicate n IntT, IntT);;

let rec mke (Ac_tp_ops_ext (ac_tp, r, a, c, mk, delete_rules, e, more)) = mk;;

let rec fresh_variable_checker _A (_B1, _B2, _B3, _B4) (_C1, _C2) _D _E
  type_of_fun bool_types definability_checker sa2 p f x ty =
    catch_error
      (bindb
        (catch_error
          (forallM
            (fun (l, fa) ->
              check (not (in_vars_formula (equal_prod _C1 _B3) (x, ty) fa))
                (comp (comp (showsl _C2 x)
                        (showsl_lit
                          " is not fresh, it occurs in location condition of "))
                  (showsl _D l)))
            (assertion_impl p))
          (fun xa -> Inl (snd xa)))
        (fun _ ->
          catch_error
            (forallM
              (fun (tr, tau) ->
                (let psi = f tr in
                  bindb (check
                          (formula
                            (is_bool (_B1, _B2, _B3) type_of_fun bool_types)
                            psi)
                          (showsl_lit
                            "new transition formula seems to be not well-formed"))
                    (fun _ ->
                      catch_error
                        (let Transition (_, _, phi) = tau in
                          bindb (definability_checker (Post x) ty psi)
                            (fun _ ->
                              bindb (check
                                      (not
(in_vars_formula (equal_prod (equal_trans_var _C1) _B3) (Post x, ty) phi))
                                      (comp
(showsl_lit "Post x in transition formula") (showsl_formula sa2 phi)))
                                (fun _ ->
                                  check (not
  (in_vars_formula (equal_prod (equal_trans_var _C1) _B3) (Pre x, ty) phi))
                                    (comp (showsl_lit
    "Pre x in transition formula")
                                      (showsl_formula sa2 phi)))))
                        (fun xa ->
                          Inl (comp (comp (comp
    (showsl_lit "problem in transition formula of transition ") (showsl _E tr))
                                      (showsl_literal "\092"))
                                xa)))))
              (transitions_impl p))
            (fun xa -> Inl (snd xa))))
      (fun xa ->
        Inl (comp (showsl_lit "fresh_variable_checker failed\092") xa));;

let rec showsl_formulaa _A
  = function Atom a -> showsl _A a
    | NegAtom a -> comp (comp (showsl_lit "! (") (showsl _A a)) (showsl_lit ")")
    | Conjunction a ->
        showsl_list_gen id "TRUE" "(" " & " ")" (map (showsl_formulaa _A) a)
    | Disjunction a ->
        showsl_list_gen id "FALSE" "(" " | " ")" (map (showsl_formulaa _A) a);;

let rec refine_transition_formula
  (Transition (l, r, phi)) psi = Transition (l, r, form_and phi psi);;

let rec refine_transition_formulas
  p f = Lts_Impl
          (initiala p,
            map (fun (tr, tau) -> (tr, refine_transition_formula tau (f tr)))
              (transitions_impl p),
            assertion_impl p);;

let rec fresh_variable_addition _A (_B1, _B2, _B3, _B4) (_C1, _C2) _D
  (_E1, _E2, _E3, _E4)
  type_of_fun bool_types definability_checker sa2 p
    (Fresh_Variable_Addition_Info (x, ty, forms)) =
    (let m = of_alist (_E1, _E2, _E3) forms in
     let f = lookup_default (_E1, _E2) (Conjunction []) m in
      (match
        catch_error
          (bindb
            (catch_error
              (forallM
                (fun fa ->
                  check (formula
                          (is_bool (_B1, _B2, _B3) type_of_fun bool_types) fa)
                    (comp (showsl_formulaa
                            (showl_term _A
                              (showl_prod (showl_trans_var _C2) _B4))
                            fa)
                      (showsl_lit " is not a valid formula")))
                (map snd forms))
              (fun xa -> Inl (snd xa)))
            (fun _ ->
              fresh_variable_checker _A (_B1, _B2, _B3, _B4) (_C1, _C2) _D _E4
                type_of_fun bool_types definability_checker sa2 p f x ty))
          (fun xa ->
            Inl (comp (comp (comp (showsl_lit
                                    "problem in adding fresh variable ")
                              (showsl _C2 x))
                        (showsl_literal "\092"))
                  xa))
        with Inl a -> Inl a | Inr _ -> Inr (refine_transition_formulas p f)));;

let rec is_sharp = function Flat uu -> false
                   | Sharpa uv -> true;;

let rec trivial_termination_checker _D _E
  showsl_atom showsl_tatom p =
    check (null (filtera (fun tau -> is_sharp (source tau))
                  (map snd (transitions_impl p))))
      (comp (showsl_lit "there are remaining sharp transitions in ")
        (showsl_lts (showl_sharp _D) _E showsl_atom showsl_tatom p));;

let rec lex_less_formula _A _B _C
  less_eq_formula less_formula uu uv uw = match
    less_eq_formula, less_formula, uu, uv, uw with
    less_eq_formula, less_formula, uu, [va], v :: vb :: vc -> Disjunction []
    | less_eq_formula, less_formula, uu, [], v :: vb :: vc -> Disjunction []
    | less_eq_formula, less_formula, [va], uv, v :: vb :: vc -> Disjunction []
    | less_eq_formula, less_formula, [], uv, v :: vb :: vc -> Disjunction []
    | less_eq_formula, less_formula, uu, uv, [] -> Disjunction []
    | less_eq_formula, less_formula, uu, v :: vb :: vc, [va] -> Disjunction []
    | less_eq_formula, less_formula, [va], v :: vb :: vc, uw -> Disjunction []
    | less_eq_formula, less_formula, [], v :: vb :: vc, uw -> Disjunction []
    | less_eq_formula, less_formula, uu, [], uw -> Disjunction []
    | less_eq_formula, less_formula, v :: vb :: vc, uv, [va] -> Disjunction []
    | less_eq_formula, less_formula, v :: vb :: vc, [va], uw -> Disjunction []
    | less_eq_formula, less_formula, [], uv, uw -> Disjunction []
    | less_eq_formula, less_formula, ba :: b :: bs, xa :: x :: xs, ya :: y :: ys
        -> form_or (form_and (less_formula xa ya) (less_eq_formula ba ya))
             (Conjunction
               [less_eq_formula xa ya;
                 lex_less_formula _A _B _C less_eq_formula less_formula
                   (b :: bs) (x :: xs) (y :: ys)])
    | less_eq_formula, less_formula, [b], [x], [y] ->
        form_and (less_formula x y) (less_eq_formula b y);;

let rec bound_exp (Transition_removal_info (x1, x2, x3, x4, x5)) = x4;;

let rec hinter (Transition_removal_info (x1, x2, x3, x4, x5)) = x5;;

let rec rank (Transition_removal_info (x1, x2, x3, x4, x5)) = x1;;

let rec check_lex_strict (_A1, _A2) (_B1, _B2, _B3, _B4) (_C1, _C2) (_D1, _D2)
  (_E1, _E2, _E3) _F
  type_of_fun bool_types showsl_atom logic_checker negate_atom less_eq_formula
    less_formula info pi (tr, Transition (l, r, phi)) =
    (let psi =
       lex_less_formula _A2 _C2 _B4 less_eq_formula less_formula
         (map (rename_vars_exp (fun a -> Pre a)) (bound_exp info))
         (map (rename_vars_exp (fun a -> Post a)) (rank info r))
         (map (rename_vars_exp (fun a -> Pre a)) (rank info l))
       in
      bindb (check
              (formula (is_bool (_B1, _B2, _B3) type_of_fun bool_types) psi)
              (comp (comp (showsl_lit "lex-less does not encode valid formula")
                      (showsl_literal "\092"))
                (showsl_formulaa
                  (showl_term _A2 (showl_prod (showl_trans_var _C2) _B4)) psi)))
        (fun _ ->
          check_formula _A1 ((equal_trans_var _C1), (showl_trans_var _C2)) _B3
            (_D1, _D2) showsl_atom logic_checker negate_atom (hinter info tr)
            (Disjunction
              [psi; Disjunction [];
                form_not
                  (map_formula (rename_vars_exp (fun a -> Pre a))
                    (assertion_of
                      ((ccompare_sharp _E1), (equal_sharp _E2),
                        mapping_impl_sharp)
                      pi l));
                form_not phi])));;

let rec lex_leq_formula _A _B _C
  less_eq_formula less_formula uu uv uw = match
    less_eq_formula, less_formula, uu, uv, uw with
    less_eq_formula, less_formula, uu, [va], v :: vb :: vc -> Disjunction []
    | less_eq_formula, less_formula, uu, [], v :: vb :: vc -> Disjunction []
    | less_eq_formula, less_formula, [va], uv, v :: vb :: vc -> Disjunction []
    | less_eq_formula, less_formula, [], uv, v :: vb :: vc -> Disjunction []
    | less_eq_formula, less_formula, uu, v :: va, [] -> Disjunction []
    | less_eq_formula, less_formula, v :: va, uv, [] -> Disjunction []
    | less_eq_formula, less_formula, uu, v :: vb :: vc, [va] -> Disjunction []
    | less_eq_formula, less_formula, [va], v :: vb :: vc, uw -> Disjunction []
    | less_eq_formula, less_formula, [], v :: vb :: vc, uw -> Disjunction []
    | less_eq_formula, less_formula, uu, [], v :: va -> Disjunction []
    | less_eq_formula, less_formula, v :: va, [], uw -> Disjunction []
    | less_eq_formula, less_formula, v :: vb :: vc, uv, [va] -> Disjunction []
    | less_eq_formula, less_formula, v :: vb :: vc, [va], uw -> Disjunction []
    | less_eq_formula, less_formula, [], uv, v :: va -> Disjunction []
    | less_eq_formula, less_formula, [], v :: va, uw -> Disjunction []
    | less_eq_formula, less_formula, [], [], [] -> Conjunction []
    | less_eq_formula, less_formula, ba :: b :: bs, xa :: x :: xs, ya :: y :: ys
        -> form_or (form_and (less_formula xa ya) (less_eq_formula ba ya))
             (Conjunction
               [less_eq_formula xa ya;
                 lex_leq_formula _A _B _C less_eq_formula less_formula (b :: bs)
                   (x :: xs) (y :: ys)])
    | less_eq_formula, less_formula, [b], [x], [y] -> less_eq_formula x y;;

let rec check_lex_weak (_A1, _A2) (_B1, _B2, _B3, _B4) (_C1, _C2) (_D1, _D2)
  (_E1, _E2, _E3) _F
  type_of_fun bool_types showsl_atom logic_checker negate_atom less_eq_formula
    less_formula info pi (tr, Transition (l, r, phi)) =
    (let psi =
       lex_leq_formula _A2 _C2 _B4 less_eq_formula less_formula
         (map (rename_vars_exp (fun a -> Pre a)) (bound_exp info))
         (map (rename_vars_exp (fun a -> Post a)) (rank info r))
         (map (rename_vars_exp (fun a -> Pre a)) (rank info l))
       in
      bindb (check
              (formula (is_bool (_B1, _B2, _B3) type_of_fun bool_types) psi)
              (comp (comp (showsl_lit "lex-leq does not encode valid formula")
                      (showsl_literal "\092"))
                (showsl_formulaa
                  (showl_term _A2 (showl_prod (showl_trans_var _C2) _B4)) psi)))
        (fun _ ->
          check_formula _A1 ((equal_trans_var _C1), (showl_trans_var _C2)) _B3
            (_D1, _D2) showsl_atom logic_checker negate_atom (hinter info tr)
            (Disjunction
              [psi; Disjunction [];
                form_not
                  (map_formula (rename_vars_exp (fun a -> Pre a))
                    (assertion_of
                      ((ccompare_sharp _E1), (equal_sharp _E2),
                        mapping_impl_sharp)
                      pi l));
                form_not phi])));;

let rec update_transitions_impl
  (Lts_Impl (i, tsa, lc)) ts = Lts_Impl (i, ts, lc);;

let rec del_transitions_impl (_E1, _E2, _E3)
  p td =
    update_transitions_impl p
      (diff_by_label (_E1, _E2) (transitions_impl p) (set (_E1, _E2, _E3) td));;

let rec processor (_A1, _A2, _A3, _A4, _A5)
  check_weak check_strict td pi =
    catch_error
      (bindb
        (catch_error
          (forallM
            (fun (tr, tau) ->
              (if membera _A3 td tr
                then bindb (check (is_sharp (source tau))
                             (comp (comp (showsl_lit "non-sharp transition ")
                                     (showsl _A5 tr))
                               (showsl_lit " cannot be removed")))
                       (fun _ ->
                         catch_error (check_strict (tr, tau))
                           (fun x ->
                             Inl (comp (comp
 (comp (showsl_lit "Failed to strictly orient transition ") (showsl _A5 tr))
 (showsl_literal "\092"))
                                   x)))
                else (if is_sharp (source tau)
                       then catch_error (check_weak (tr, tau))
                              (fun x ->
                                Inl (comp (comp
    (comp (showsl_lit "Failed to weakly orient transition ") (showsl _A5 tr))
    (showsl_literal "\092"))
                                      x))
                       else Inr ())))
            (transitions_impl pi))
          (fun x -> Inl (snd x)))
        (fun _ -> Inr (del_transitions_impl (_A1, _A2, _A4) pi td)))
      (fun x ->
        Inl (comp (comp (comp (showsl_lit "Failed to eliminate transitions ")
                          (showsl_list _A5 td))
                    (showsl_lit ":\092"))
              x));;

let rec removed (Transition_removal_info (x1, x2, x3, x4, x5)) = x2;;

let rec lex_processor (_A1, _A2) (_B1, _B2, _B3, _B4) (_C1, _C2) (_D1, _D2)
  (_E1, _E2, _E3) (_F1, _F2, _F3, _F4, _F5)
  type_of_fun bool_types showsl_atom logic_checker negate_atom dom_type
    less_eq_formula less_formula is_constant info pi =
    bindb (catch_error
            (forallM
              (fun l ->
                catch_error
                  (forallM
                    (fun e ->
                      check (has_type _B3 type_of_fun e dom_type)
                        (comp (showsl_lit "Unexpected type of expression:\092")
                          (showsl_terma _A2 (showl_prod _C2 _B4) e)))
                    (rank info l))
                  (fun x -> Inl (snd x)))
              (nodes_lts_impl (equal_sharp _E2) pi))
            (fun x -> Inl (snd x)))
      (fun _ ->
        bindb (catch_error
                (forallM
                  (fun e ->
                    check (has_type _B3 type_of_fun e dom_type)
                      (comp (showsl_lit "Unexpected type of bound: ")
                        (showsl_terma _A2 (showl_prod _C2 _B4) e)))
                  (bound_exp info))
                (fun x -> Inl (snd x)))
          (fun _ ->
            bindb (catch_error
                    (forallM
                      (fun e ->
                        check (is_constant e)
                          (comp (showsl_lit "Non-constant bound: ")
                            (showsl_terma _A2
                              (showl_prod (showl_trans_var _C2) _B4) e)))
                      (map (rename_vars_exp (fun a -> Pre a)) (bound_exp info)))
                    (fun x -> Inl (snd x)))
              (fun _ ->
                processor (_F1, _F2, _F3, _F4, _F5)
                  (check_lex_weak (_A1, _A2) (_B1, _B2, _B3, _B4) (_C1, _C2)
                    (_D1, _D2) (_E1, _E2, _E3) _F5 type_of_fun bool_types
                    showsl_atom logic_checker negate_atom less_eq_formula
                    less_formula info pi)
                  (check_lex_strict (_A1, _A2) (_B1, _B2, _B3, _B4) (_C1, _C2)
                    (_D1, _D2) (_E1, _E2, _E3) _F5 type_of_fun bool_types
                    showsl_atom logic_checker negate_atom less_eq_formula
                    less_formula info pi)
                  (removed info) pi)));;

let rec check_skip_transition _A (_B1, _B2) _C (_D1, _D2)
  showsl_atom logic_checker negate_atom tau =
    (let Transition (_, _, phi) = tau in
      check_valid_formula _A (_B1, _B2) _C (_D1, _D2) showsl_atom logic_checker
        negate_atom (map_formula (rename_vars_exp untrans_var) phi));;

let rec change_source
  l (Transition (src, tgt, phi)) = Transition (l, tgt, phi);;

let rec location_addition_outgoing _A (_B1, _B2, _B3) (_C1, _C2) (_D1, _D2)
  (_E1, _E2, _E3) _F
  type_of_fun bool_types showsl_atom logic_checker negate_atom p
    (Location_Addition_Info (old, newa, skip_ID, skip)) =
    bindb (check
            (not (membera (equal_sharp _E2) (nodes_lts_impl (equal_sharp _E2) p)
                   newa))
            (comp (comp (showsl_lit "location-id ") (showsl_sharp _E3 newa))
              (showsl_lit " is not fresh")))
      (fun _ ->
        bindb (check (is_sharp newa)
                (comp (comp (showsl_lit "new location ")
                        (showsl_sharp _E3 newa))
                  (showsl_lit " must be sharp location")))
          (fun _ ->
            bindb (check (is_sharp old)
                    (comp (comp (showsl_lit "copied location ")
                            (showsl_sharp _E3 old))
                      (showsl_lit " must be sharp location")))
              (fun _ ->
                bindb (check
                        (transition_rule (_B1, _B2, _B3) type_of_fun bool_types
                          skip)
                        (comp (comp (showsl_lit "new transition ")
                                (showsl _F skip_ID))
                          (showsl_lit " seems to be ill-formed")))
                  (fun _ ->
                    bindb (catch_error
                            (check_skip_transition _A (_C1, _C2) _B3 (_D1, _D2)
                              showsl_atom logic_checker negate_atom skip)
                            (fun x ->
                              Inl (comp (comp
  (comp (showsl_lit "new transition ") (showsl _F skip_ID))
  (showsl_lit " must be skip transition\092"))
                                    x)))
                      (fun _ ->
                        bindb (check
                                (equal_sharpa _E2 (source skip) old &&
                                  equal_sharpa _E2 (target skip) newa)
                                (comp (comp
(comp (comp (comp (showsl_lit "new skip transition ") (showsl _F skip_ID))
        (showsl_lit " must be from "))
  (showsl_sharp _E3 old))
(showsl_lit " to "))
                                  (showsl_sharp _E3 newa)))
                          (fun _ ->
                            (let trans = transitions_impl p in
                             let (sharp, flat) =
                               partition
                                 (fun tau -> is_sharp (source (snd tau))) trans
                               in
                             let (sharp_modify, sharp_keep) =
                               partition
                                 (fun tau ->
                                   equal_sharpa _E2 (source (snd tau)) old)
                                 sharp
                               in
                             let q =
                               Lts_Impl
                                 (initiala p,
                                   (skip_ID, skip) ::
                                     flat @
                                       sharp_keep @
 map (fun tau -> (fst tau, change_source newa (snd tau))) sharp_modify,
                                   (newa,
                                     assertion_of
                                       ((ccompare_sharp _E1), (equal_sharp _E2),
 mapping_impl_sharp)
                                       p old) ::
                                     assertion_impl p)
                               in
                              bindb (catch_error
                                      (forallM
(fun l ->
  check (equal_formulaa (equal_term _A (equal_prod _C1 _B3))
          (assertion_of
            ((ccompare_sharp _E1), (equal_sharp _E2), mapping_impl_sharp) p l)
          (assertion_of
            ((ccompare_sharp _E1), (equal_sharp _E2), mapping_impl_sharp) q l))
    (comp (comp (showsl_lit "location condition of initial state ")
            (showsl_sharp _E3 l))
      (showsl_lit " has been changed")))
(initiala p))
                                      (fun x -> Inl (snd x)))
                                (fun _ -> Inr q))))))));;

let rec change_target
  l (Transition (src, tgt, phi)) = Transition (src, l, phi);;

let rec location_addition_incoming _A (_B1, _B2, _B3) (_C1, _C2) (_D1, _D2)
  (_E1, _E2, _E3) _F
  type_of_fun bool_types showsl_atom logic_checker negate_atom p
    (Location_Addition_Info (newa, old, skip_ID, skip)) =
    bindb (check
            (not (membera (equal_sharp _E2) (nodes_lts_impl (equal_sharp _E2) p)
                   newa))
            (comp (comp (showsl_lit "location-id ") (showsl_sharp _E3 newa))
              (showsl_lit " is not fresh")))
      (fun _ ->
        bindb (check (is_sharp newa)
                (comp (comp (showsl_lit "new location ")
                        (showsl_sharp _E3 newa))
                  (showsl_lit " must be sharp location")))
          (fun _ ->
            bindb (check (is_sharp old)
                    (comp (comp (showsl_lit "copied location ")
                            (showsl_sharp _E3 old))
                      (showsl_lit " must be sharp location")))
              (fun _ ->
                bindb (check
                        (transition_rule (_B1, _B2, _B3) type_of_fun bool_types
                          skip)
                        (comp (comp (showsl_lit "new transition ")
                                (showsl _F skip_ID))
                          (showsl_lit " seems to be ill-formed")))
                  (fun _ ->
                    bindb (catch_error
                            (check_skip_transition _A (_C1, _C2) _B3 (_D1, _D2)
                              showsl_atom logic_checker negate_atom skip)
                            (fun x ->
                              Inl (comp (comp
  (comp (showsl_lit "new transition ") (showsl _F skip_ID))
  (showsl_lit " must be skip transition\092"))
                                    x)))
                      (fun _ ->
                        bindb (check
                                (equal_sharpa _E2 (source skip) newa &&
                                  equal_sharpa _E2 (target skip) old)
                                (comp (comp
(comp (comp (comp (showsl_lit "new skip transition ") (showsl _F skip_ID))
        (showsl_lit " must be from "))
  (showsl_sharp _E3 newa))
(showsl_lit " to "))
                                  (showsl_sharp _E3 old)))
                          (fun _ ->
                            (let trans = transitions_impl p in
                             let (sharp, flat) =
                               partition
                                 (fun tau -> is_sharp (source (snd tau))) trans
                               in
                             let (sharp_modify, sharp_keep) =
                               partition
                                 (fun tau ->
                                   equal_sharpa _E2 (target (snd tau)) old)
                                 sharp
                               in
                             let q =
                               Lts_Impl
                                 (initiala p,
                                   (skip_ID, skip) ::
                                     flat @
                                       sharp_keep @
 map (fun tau -> (fst tau, change_target newa (snd tau))) sharp_modify,
                                   (newa,
                                     assertion_of
                                       ((ccompare_sharp _E1), (equal_sharp _E2),
 mapping_impl_sharp)
                                       p old) ::
                                     assertion_impl p)
                               in
                              bindb (catch_error
                                      (forallM
(fun l ->
  check (equal_formulaa (equal_term _A (equal_prod _C1 _B3))
          (assertion_of
            ((ccompare_sharp _E1), (equal_sharp _E2), mapping_impl_sharp) p l)
          (assertion_of
            ((ccompare_sharp _E1), (equal_sharp _E2), mapping_impl_sharp) q l))
    (comp (comp (showsl_lit "location condition of initial state ")
            (showsl_sharp _E3 l))
      (showsl_lit " has been changed")))
(initiala p))
                                      (fun x -> Inl (snd x)))
                                (fun _ -> Inr q))))))));;

let rec location_addition _A (_B1, _B2, _B3) (_C1, _C2) (_D1, _D2)
  (_E1, _E2, _E3) _F
  type_of_fun bool_types showsl_atom logic_checker negate_atom p info =
    (let Location_Addition_Info (src, _, _, _) = info in
      (if not (membera (equal_sharp _E2) (nodes_lts_impl (equal_sharp _E2) p)
                src)
        then location_addition_incoming _A (_B1, _B2, _B3) (_C1, _C2) (_D1, _D2)
               (_E1, _E2, _E3) _F type_of_fun bool_types showsl_atom
               logic_checker negate_atom p info
        else location_addition_outgoing _A (_B1, _B2, _B3) (_C1, _C2) (_D1, _D2)
               (_E1, _E2, _E3) _F type_of_fun bool_types showsl_atom
               logic_checker negate_atom p info));;

let rec call_graph_sharp_impl _D
  r = remdups (equal_prod (equal_sharp _D) (equal_sharp _D))
        (maps (fun tau ->
                (if is_sharp (source (snd tau))
                  then [(source (snd tau), target (snd tau))] else []))
          (transitions_impl r));;

let rec scc_decomposition (_D1, _D2, _D3, _D4)
  cp sccs_info =
    (let cg = call_graph_sharp_impl _D3 cp in
     let sccs =
       scc_decomp
         ((ceq_sharp _D3), (ccompare_sharp _D1),
           (compare_order_sharp _D2.compare_compare_order), (equal_sharp _D3),
           set_impl_sharp)
         cg
       in
      bindb (catch_error
              (forallM
                (fun c ->
                  check (list_ex
                          (fun d ->
                            set_eq
                              (cenum_sharp, (ceq_sharp _D3),
                                (ccompare_sharp _D1))
                              (set ((ceq_sharp _D3), (ccompare_sharp _D1),
                                     set_impl_sharp)
                                c)
                              (set ((ceq_sharp _D3), (ccompare_sharp _D1),
                                     set_impl_sharp)
                                d))
                          sccs_info)
                    (comp (comp (showsl_lit "could not find SCC ")
                            (showsl_lista (showl_sharp _D4) c))
                      (showsl_lit " in provided decomposition")))
                sccs)
              (fun x -> Inl (snd x)))
        (fun _ ->
          (let trans = transitions_impl cp in
           let (sharp, flat) =
             partition (fun tau -> is_sharp (source (snd tau))) trans in
           let cPs =
             map (fun c ->
                   (let l =
                      set ((ceq_sharp _D3), (ccompare_sharp _D1),
                            set_impl_sharp)
                        c
                      in
                     Lts_Impl
                       (initiala cp,
                         flat @
                           filtera
                             (fun tau ->
                               member ((ceq_sharp _D3), (ccompare_sharp _D1))
                                 (source (snd tau)) l &&
                                 member ((ceq_sharp _D3), (ccompare_sharp _D1))
                                   (target (snd tau)) l)
                             sharp,
                         assertion_impl cp)))
               sccs_info
             in
            catch_error (Inr cPs)
              (fun x ->
                Inl (comp (comp (comp (showsl_lit
"error is SCC-decomposition w.r.t. sccs_info\092")
                                  (showsl_lista (showl_list (showl_sharp _D4))
                                    sccs_info))
                            (showsl_literal "\092"))
                      x)))));;

let rec cut_transition_split (_A1, _A2, _A3, _A4, _A5) (_E1, _E2, _E3)
  (Cut_Transition_Split_Info ct_ids) cp =
    catch_error
      (let (p, r) =
         partition (comp (fun tau -> is_sharp (source tau)) snd)
           (transitions_impl cp)
         in
       let (ct, rr) = partition (comp (comp is_sharp target) snd) r in
       let slist = map (comp source snd) p in
       let s = set ((ceq_sharp _E2), (ccompare_sharp _E1), set_impl_sharp) slist
         in
       let cT_ids = set (_A1, _A2, _A4) (concat ct_ids) in
        bindb (catch_error
                (forallM
                  (fun l ->
                    check (not (is_sharp l))
                      (comp (comp (showsl_lit "initial state ")
                              (showsl_sharp _E3 l))
                        (showsl_lit " must not be sharped")))
                  (initiala cp))
                (fun x -> Inl (snd x)))
          (fun _ ->
            bindb (catch_error
                    (forallM
                      (fun (t_id, cta) ->
                        check (if member ((ceq_sharp _E2), (ccompare_sharp _E1))
                                    (target cta) s
                                then member (_A1, _A2) t_id cT_ids else true)
                          (comp (comp (comp
(showsl_lit "did not find cut-transition ") (showsl _A5 t_id))
                                  (showsl_lit
                                    " in partition\092relevant cut-points are: "))
                            (showsl_list_sharp _E3 slist)))
                      ct)
                    (fun x -> Inl (snd x)))
              (fun _ ->
                (let rrp = rr @ p in
                  Inr (map (fun ct_idsa ->
                             Lts_Impl
                               (initiala cp,
                                 filtera
                                   (fun cta -> membera _A3 ct_idsa (fst cta))
                                   ct @
                                   rrp,
                                 assertion_impl cp))
                        ct_ids)))))
      (fun x ->
        Inl (comp (showsl_lit "error in splitting cut transitions on LTS\092")
              x));;

let rec showsl_cooperation_program _D _E
  (Lts_Impl (i, tran, lc)) =
    comp (let trana = filtera (fun (_, tt) -> is_sharp (target tt)) tran in
          let tranb =
            (let (ss, fs) =
               partition (fun (_, tt) -> is_sharp (source tt)) trana in
              fs @ ss)
            in
           comp (showsl_lit "Cooperation program (only sharp-part)\092")
             (showsl_sep
               (fun (t, tt) ->
                 comp (comp (comp (comp (showsl _E t) (showsl_lit ": "))
                              (showsl_sharp _D (source tt)))
                        (showsl_lit " --> "))
                   (showsl_sharp _D (target tt)))
               (showsl_literal "\092") tranb))
      (showsl_literal "\092");;

let rec check_cooperation_proof (_A1, _A2, _A3) (_B1, _B2, _B3, _B4) (_C1, _C2)
  (_D1, _D2, _D3) (_E1, _E2, _E3, _E4) (_F1, _F2, _F3, _F4, _F5, _F6)
  type_of_fun bool_types tc tc2 sa sa2 ne ne2 dom_type less_eq_formula
    less_formula is_constant definability_checker i cPi x15 = match
    type_of_fun, bool_types, tc, tc2, sa, sa2, ne, ne2, dom_type,
      less_eq_formula, less_formula, is_constant, definability_checker, i, cPi,
      x15
    with
    type_of_fun, bool_types, tc, tc2, sa, sa2, ne, ne2, dom_type,
      less_eq_formula, less_formula, is_constant, definability_checker, i, cPi,
      Cut_Transition_Split scc_proofs
      -> debug i "Cut Transition Split"
           (bindb
             (catch_error
               (cut_transition_split (_F1, _F2, _F3, _F5, _F6) (_E1, _E3, _E4)
                 (Cut_Transition_Split_Info (map fst scc_proofs)) cPi)
               (fun x ->
                 Inl (comp (comp (comp i
                                   (showsl_lit
                                     ": error in cut-transition split\092"))
                             (showsl_cooperation_program _E4 _F6 cPi))
                       x)))
             (fun sccs ->
               catch_error
                 (forallM_index
                   (fun (scc, prof) j ->
                     catch_error
                       (check_cooperation_proof (_A1, _A2, _A3)
                         (_B1, _B2, _B3, _B4) (_C1, _C2) (_D1, _D2, _D3)
                         (_E1, _E2, _E3, _E4) (_F1, _F2, _F3, _F4, _F5, _F6)
                         type_of_fun bool_types tc tc2 sa sa2 ne ne2 dom_type
                         less_eq_formula less_formula is_constant
                         definability_checker (add_index i (suc j)) scc prof)
                       (fun x ->
                         Inl (comp (comp i
                                     (showsl_lit
                                       ": error below cut-transition split\092"))
                               x)))
                   (zip sccs (map snd scc_proofs)))
                 (fun x -> Inl (snd x))))
    | type_of_fun, bool_types, tc, tc2, sa, sa2, ne, ne2, dom_type,
        less_eq_formula, less_formula, is_constant, definability_checker, i,
        cPi, Scc_Decomp scc_proofs
        -> debug i "SCC Decomp"
             (bindb
               (catch_error
                 (scc_decomposition (_E1, _E2, _E3, _E4) cPi
                   (map fst scc_proofs))
                 (fun x ->
                   Inl (comp (comp (comp i
                                     (showsl_lit
                                       ": error in Scc decomposition\092"))
                               (showsl_cooperation_program _E4 _F6 cPi))
                         x)))
               (fun sccs ->
                 catch_error
                   (forallM_index
                     (fun (scc, prof) j ->
                       catch_error
                         (check_cooperation_proof (_A1, _A2, _A3)
                           (_B1, _B2, _B3, _B4) (_C1, _C2) (_D1, _D2, _D3)
                           (_E1, _E2, _E3, _E4) (_F1, _F2, _F3, _F4, _F5, _F6)
                           type_of_fun bool_types tc tc2 sa sa2 ne ne2 dom_type
                           less_eq_formula less_formula is_constant
                           definability_checker (add_index i (suc j)) scc prof)
                         (fun x ->
                           Inl (comp (comp i
                                       (showsl_lit
 ": error below Scc decomposition\092"))
                                 x)))
                     (zip sccs (map snd scc_proofs)))
                   (fun x -> Inl (snd x))))
    | type_of_fun, bool_types, tc, tc2, sa, sa2, ne, ne2, dom_type,
        less_eq_formula, less_formula, is_constant, definability_checker, i,
        cPi, Fresh_Variable_Addition (info, iproof)
        -> debug i "Fresh Variable Addition"
             (bindb
               (catch_error
                 (fresh_variable_addition _A3 (_B1, _B2, _B3, _B4) (_D2, _D3)
                   (showl_sharp _E4) (_F2, _F3, _F4, _F6) type_of_fun bool_types
                   definability_checker sa2 cPi info)
                 (fun x ->
                   Inl (comp (comp (comp i
                                     (showsl_lit
                                       ": error in fresh variable addition\092"))
                               (showsl_cooperation_program _E4 _F6 cPi))
                         x)))
               (fun q ->
                 catch_error
                   (check_cooperation_proof (_A1, _A2, _A3) (_B1, _B2, _B3, _B4)
                     (_C1, _C2) (_D1, _D2, _D3) (_E1, _E2, _E3, _E4)
                     (_F1, _F2, _F3, _F4, _F5, _F6) type_of_fun bool_types tc
                     tc2 sa sa2 ne ne2 dom_type less_eq_formula less_formula
                     is_constant definability_checker (add_index i one_nata) q
                     iproof)
                   (fun x ->
                     Inl (comp (comp i
                                 (showsl_lit
                                   ": error below fresh variable addition\092"))
                           x))))
    | type_of_fun, bool_types, tc, tc2, sa, sa2, ne, ne2, dom_type,
        less_eq_formula, less_formula, is_constant, definability_checker, i,
        cPi, Location_Addition (info, iproof)
        -> debug i "Location Addition"
             (bindb
               (catch_error
                 (location_addition _A2 (_B1, _B2, _B3) (_D2, _D3) (_C1, _C2)
                   (_E1, _E3, _E4) _F6 type_of_fun bool_types sa tc ne cPi info)
                 (fun x ->
                   Inl (comp (comp (comp i
                                     (showsl_literal
                                       ": error in location addition\092"))
                               (showsl_cooperation_program _E4 _F6 cPi))
                         x)))
               (fun q ->
                 catch_error
                   (check_cooperation_proof (_A1, _A2, _A3) (_B1, _B2, _B3, _B4)
                     (_C1, _C2) (_D1, _D2, _D3) (_E1, _E2, _E3, _E4)
                     (_F1, _F2, _F3, _F4, _F5, _F6) type_of_fun bool_types tc
                     tc2 sa sa2 ne ne2 dom_type less_eq_formula less_formula
                     is_constant definability_checker (add_index i one_nata) q
                     iproof)
                   (fun x ->
                     Inl (comp (comp i
                                 (showsl_literal
                                   ": error below location addition\092"))
                           x))))
    | type_of_fun, bool_types, tc, tc2, sa, sa2, ne, ne2, dom_type,
        less_eq_formula, less_formula, is_constant, definability_checker, i,
        cPi, Transition_Removal (info, iproof)
        -> debug i "Transition Removal"
             (bindb
               (catch_error
                 (lex_processor (_A2, _A3) (_B1, _B2, _B3, _B4) (_D2, _D3)
                   (_C1, _C2) (_E1, _E3, _E4) (_F1, _F2, _F3, _F5, _F6)
                   type_of_fun bool_types sa2 tc2 ne2 dom_type less_eq_formula
                   less_formula is_constant info cPi)
                 (fun x ->
                   Inl (comp (comp (comp i
                                     (showsl_lit
                                       ": error in transition removal\092"))
                               (showsl_cooperation_program _E4 _F6 cPi))
                         x)))
               (fun cPia ->
                 catch_error
                   (check_cooperation_proof (_A1, _A2, _A3) (_B1, _B2, _B3, _B4)
                     (_C1, _C2) (_D1, _D2, _D3) (_E1, _E2, _E3, _E4)
                     (_F1, _F2, _F3, _F4, _F5, _F6) type_of_fun bool_types tc
                     tc2 sa sa2 ne ne2 dom_type less_eq_formula less_formula
                     is_constant definability_checker (add_index i one_nata)
                     cPia iproof)
                   (fun x ->
                     Inl (comp (comp i
                                 (showsl_lit
                                   ": error below transition removal\092"))
                           x))))
    | type_of_fun, bool_types, tc, tc2, sa, sa2, ne, ne2, dom_type,
        less_eq_formula, less_formula, is_constant, definability_checker, i,
        cPi, Invariants_Update (iproof, cproof)
        -> debug i "Invariants_Update"
             (bindb
               (catch_error
                 (invariant_proof_checker (_A1, _A2, _A3) (_B1, _B2, _B3, _B4)
                   (_C1, _C2) (_D1, _D2, _D3)
                   (cenum_sharp, (ceq_sharp _E3), (ccompare_sharp _E1),
                     (equal_sharp _E3), mapping_impl_sharp, set_impl_sharp,
                     (showl_sharp _E4))
                   (_F1, _F2, _F3, _F4, _F5, _F6)
                   ((ccompare_list ccompare_char), (equal_list equal_char),
                     mapping_impl_list, (showl_list showl_char))
                   type_of_fun bool_types tc tc2 sa sa2 ne ne2 cPi iproof)
                 (fun x ->
                   Inl (comp (comp (comp i
                                     (showsl_lit
                                       ": error in invariant update\092"))
                               (showsl_cooperation_program _E4 _F6 cPi))
                         x)))
               (fun ia ->
                 bindb (fix_invariants
                         ((ccompare_sharp _E1), (equal_sharp _E3),
                           mapping_impl_sharp, (showl_sharp _E4))
                         _F6 cPi ia)
                   (fun q ->
                     catch_error
                       (check_cooperation_proof (_A1, _A2, _A3)
                         (_B1, _B2, _B3, _B4) (_C1, _C2) (_D1, _D2, _D3)
                         (_E1, _E2, _E3, _E4) (_F1, _F2, _F3, _F4, _F5, _F6)
                         type_of_fun bool_types tc tc2 sa sa2 ne ne2 dom_type
                         less_eq_formula less_formula is_constant
                         definability_checker (add_index i one_nata) q cproof)
                       (fun x ->
                         Inl (comp (comp i
                                     (showsl_lit
                                       ": error below invariant update\092"))
                               x)))))
    | type_of_fun, bool_types, tc, tc2, sa, sa2, ne, ne2, dom_type,
        less_eq_formula, less_formula, is_constant, definability_checker, i,
        cPi, Triviala
        -> debug i "Trivial CP"
             (catch_error (trivial_termination_checker _E4 _F6 sa sa2 cPi)
               (fun x ->
                 Inl (comp (comp i
                             (showsl_lit
                               ": error in trivial cooperation termination checker\092"))
                       x)));;

let rec check_exists_cut _A (_B1, _B2) _C (_D1, _D2) (_E1, _E2)
  showsl_atom logic_checker negate_atom taus n =
    check (list_ex
            (fun tau ->
              equal_sharpa _E1 (source tau) (Flat n) &&
                (equal_sharpa _E1 (target tau) (Sharpa n) &&
                  isOK (check_skip_transition _A (_B1, _B2) _C (_D1, _D2)
                         showsl_atom logic_checker negate_atom tau)))
            taus)
      (comp (showsl_lit "missing skip transition for ") (showsl _E2 n));;

let rec sharp_transition
  (Transition (l, r, phi)) = Transition (Sharpa l, Sharpa r, phi);;

let rec flat_transition
  (Transition (l, r, phi)) = Transition (Flat l, Flat r, phi);;

let rec make_copy_prog
  (Lts_Impl (init, tau_s, lc)) cutpoints =
    Lts_Impl
      (map (fun a -> Flat a) init,
        map (fun (tr, tau) -> (Flat tr, flat_transition tau)) tau_s @
          cutpoints @
            map (fun (tr, tau) -> (Sharpa tr, sharp_transition tau)) tau_s,
        map (fun (tr, a) -> (Flat tr, a)) lc @
          map (fun (tr, a) -> (Sharpa tr, a)) lc);;

let rec call_graph_impl _D
  r = remdups (equal_prod _D _D)
        (map (fun tau -> (source (snd tau), target (snd tau)))
          (transitions_impl r));;

let rec check_cut_points (_A1, _A2, _A3, _A4, _A5, _A6)
  r x = check_acyclic (_A1, _A2, _A3, _A4, _A5, _A6)
          (filtera
            (fun ab ->
              not (member (_A1, _A2) (fst ab) x) &&
                not (member (_A1, _A2) (snd ab) x))
            r);;

let rec natural = function Sharpa l -> l
                  | Flat l -> l;;

let rec create_initial_cp_prog _A (_B1, _B2, _B3) (_C1, _C2) (_D1, _D2)
  (_E1, _E2, _E3, _E4, _E5, _E6) _F
  type_of_fun bool_types showsl_atom logic_checker negate_atom p cp_trans_list =
    (match
      (let cp_trans = concat cp_trans_list in
       let cut_points =
         remdups _E4 (map (fun (_, tau) -> natural (source tau)) cp_trans) in
        bindb (catch_error
                (check_cut_points (_E1, _E2, _E3, _E4, _E5, _E6)
                  (call_graph_impl _E4 p) (set (_E1, _E2, _E5) cut_points))
                (fun x ->
                  Inl (comp (showsl_lit
                              "problem in ensuring validity of cutpoints\092")
                        x)))
          (fun _ ->
            bindb (catch_error
                    (forallM
                      (fun (n, cp) ->
                        check (transition_rule (_B1, _B2, _B3) type_of_fun
                                bool_types cp)
                          (comp (showsl_sharp _F n)
                            (showsl_lit " is non valid transition rule")))
                      cp_trans)
                    (fun x -> Inl (snd x)))
              (fun _ ->
                catch_error
                  (forallM
                    (check_exists_cut _A (_C1, _C2) _B3 (_D1, _D2) (_E4, _E6)
                      showsl_atom logic_checker negate_atom (map snd cp_trans))
                    cut_points)
                  (fun x -> Inl (snd x)))))
      with Inl a -> Inl a
      | Inr _ -> Inr (map (make_copy_prog p) cp_trans_list));;

let rec check_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3, _B4) (_C1, _C2)
  (_D1, _D2, _D3) (_E1, _E2, _E3, _E4, _E5, _E6, _E7, _E8)
  (_F1, _F2, _F3, _F4, _F5, _F6)
  type_of_fun bool_types tc tc2 sa sa2 ne ne2 dom_type less_eq_formula
    less_formula is_constant definability_checker i pi x15 = match
    type_of_fun, bool_types, tc, tc2, sa, sa2, ne, ne2, dom_type,
      less_eq_formula, less_formula, is_constant, definability_checker, i, pi,
      x15
    with
    type_of_fun, bool_types, tc, tc2, sa, sa2, ne, ne2, dom_type,
      less_eq_formula, less_formula, is_constant, definability_checker, i, pi,
      Via_Cooperation cp_proofs
      -> debug i "Switch to Cooperation Program"
           (bindb
             (catch_error
               (create_initial_cp_prog _A2 (_B1, _B2, _B3) (_D2, _D3) (_C1, _C2)
                 (_E2, _E3, _E4, _E5, _E7, _E8) _F6 type_of_fun bool_types sa tc
                 ne pi (map fst cp_proofs))
               (fun x ->
                 Inl (comp (comp i
                             (showsl_lit
                               ": error in creating initial cooperation program\092"))
                       x)))
             (fun cPi ->
               catch_error
                 (forallM_index
                   (fun (r, prf) n ->
                     catch_error
                       (check_cooperation_proof (_A1, _A2, _A3)
                         (_B1, _B2, _B3, _B4) (_C1, _C2) (_D1, _D2, _D3)
                         (_E3, _E4, _E5, _E8)
                         ((ceq_sharp _F3), (ccompare_sharp _F2),
                           (equal_sharp _F3), mapping_impl_sharp,
                           set_impl_sharp, (showl_sharp _F6))
                         type_of_fun bool_types tc tc2 sa sa2 ne ne2 dom_type
                         less_eq_formula less_formula is_constant
                         definability_checker (add_index i n) r prf)
                       (fun x ->
                         Inl (comp (comp i
                                     (showsl_lit
                                       ": error below switching to initial cooperation program\092"))
                               x)))
                   (zip cPi (map snd cp_proofs)))
                 (fun x -> Inl (snd x))))
    | type_of_fun, bool_types, tc, tc2, sa, sa2, ne, ne2, dom_type,
        less_eq_formula, less_formula, is_constant, definability_checker, i, pi,
        Invariants_Update_LTS (iproof, cproof)
        -> debug i "Invariant Update"
             (bindb
               (catch_error
                 (invariant_proof_checker (_A1, _A2, _A3) (_B1, _B2, _B3, _B4)
                   (_C1, _C2) (_D1, _D2, _D3)
                   (_E1, _E2, _E3, _E5, _E6, _E7, _E8)
                   (_F1, _F2, _F3, _F4, _F5, _F6)
                   ((ccompare_list ccompare_char), (equal_list equal_char),
                     mapping_impl_list, (showl_list showl_char))
                   type_of_fun bool_types tc tc2 sa sa2 ne ne2 pi iproof)
                 (fun x ->
                   Inl (comp (comp i
                               (showsl_lit ": error in invariant update\092"))
                         x)))
               (fun ia ->
                 bindb (fix_invariants (_E3, _E5, _E6, _E8) _F6 pi ia)
                   (fun q ->
                     catch_error
                       (check_termination_proof (_A1, _A2, _A3)
                         (_B1, _B2, _B3, _B4) (_C1, _C2) (_D1, _D2, _D3)
                         (_E1, _E2, _E3, _E4, _E5, _E6, _E7, _E8)
                         (_F1, _F2, _F3, _F4, _F5, _F6) type_of_fun bool_types
                         tc tc2 sa sa2 ne ne2 dom_type less_eq_formula
                         less_formula is_constant definability_checker
                         (add_index i one_nata) q cproof)
                       (fun x ->
                         Inl (comp (comp i
                                     (showsl_lit
                                       ": error below invariant update\092"))
                               x)))))
    | type_of_fun, bool_types, tc, tc2, sa, sa2, ne, ne2, dom_type,
        less_eq_formula, less_formula, is_constant, definability_checker, i, pi,
        Trivialb
        -> debug i "Trivial Termination"
             (catch_error
               (check (null (transitions_impl pi))
                 (showsl_lit
                   "transition rules remains at trivial termination proof"))
               (fun x ->
                 Inl (comp (comp i
                             (showsl_lit
                               ": error in trivial termination checker\092"))
                       x)));;

let rec checka (_A1, _A2, _A3) (_B1, _B2, _B3, _B4) (_C1, _C2) (_D1, _D2, _D3)
  (_E1, _E2, _E3, _E4, _E5, _E6, _E7, _E8) (_F1, _F2, _F3, _F4, _F5, _F6)
  type_of_fun bool_types tc tc2 sa sa2 ne ne2 dom_type less_eq_formula
    less_formula is_constant definability_checker pi prf =
    bindb (debug id "init - Check well-formedness"
            (catch_error
              (check_lts_impl (_B1, _B2, _B3) type_of_fun bool_types pi)
              (fun x ->
                Inl (comp (showsl_lit "input LTS is not well-formed") x))))
      (fun _ ->
        check_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3, _B4) (_C1, _C2)
          (_D1, _D2, _D3) (_E1, _E2, _E3, _E4, _E5, _E6, _E7, _E8)
          (_F1, _F2, _F3, _F4, _F5, _F6) type_of_fun bool_types tc tc2 sa sa2 ne
          ne2 dom_type less_eq_formula less_formula is_constant
          definability_checker id pi prf);;

let rec def_checker_main _A
  x ty l1 r1 l2 r2 =
    is_bool (ceq_ty, ccompare_ty, equal_ty) type_of_fun
      (inserta (ceq_ty, ccompare_ty) BoolT
        (set_empty (ceq_ty, ccompare_ty) (of_phantom set_impl_ty)))
      (Fun (LeF, [l1; r1])) &&
      (has_type equal_ty type_of_fun r1 ty &&
        (is_bool (ceq_ty, ccompare_ty, equal_ty) type_of_fun
           (inserta (ceq_ty, ccompare_ty) BoolT
             (set_empty (ceq_ty, ccompare_ty) (of_phantom set_impl_ty)))
           (Fun (LeF, [l2; r2])) &&
          (equal_terma equal_sig (equal_prod _A equal_ty) (Var (x, ty)) l1 &&
            (equal_terma equal_sig (equal_prod _A equal_ty) l1 r2 &&
              (equal_terma equal_sig (equal_prod _A equal_ty) l2 r1 &&
                not (contains_var_term (equal_prod _A equal_ty) (x, ty)
                      r1))))));;

let rec def_checker_maina _A
  x ty l1 r1 =
    is_bool (ceq_ty, ccompare_ty, equal_ty) type_of_fun
      (inserta (ceq_ty, ccompare_ty) BoolT
        (set_empty (ceq_ty, ccompare_ty) (of_phantom set_impl_ty)))
      (Fun (EqF, [l1; r1])) &&
      (has_type equal_ty type_of_fun r1 ty &&
        (equal_terma equal_sig (equal_prod _A equal_ty) (Var (x, ty)) l1 &&
          not (contains_var_term (equal_prod _A equal_ty) (x, ty) r1)));;

let rec def_checker_eq _A
  x ty eq =
    (match eq with Atom (Var _) -> Inl (showsl_lit "Invalid literal")
      | Atom (Fun (LessF, _)) -> Inl (showsl_lit "Invalid literal")
      | Atom (Fun (LeF, _)) -> Inl (showsl_lit "Invalid literal")
      | Atom (Fun (SumF _, _)) -> Inl (showsl_lit "Invalid literal")
      | Atom (Fun (ConstF _, _)) -> Inl (showsl_lit "Invalid literal")
      | Atom (Fun (ProdF _, _)) -> Inl (showsl_lit "Invalid literal")
      | Atom (Fun (EqF, [])) -> Inl (showsl_lit "Invalid literal")
      | Atom (Fun (EqF, [_])) -> Inl (showsl_lit "Invalid literal")
      | Atom (Fun (EqF, [l1; r1])) ->
        check (def_checker_maina _A x ty l1 r1 ||
                def_checker_maina _A x ty r1 l1)
          (showsl_lit "Non-equation not supported")
      | Atom (Fun (EqF, _ :: _ :: _ :: _)) -> Inl (showsl_lit "Invalid literal")
      | NegAtom _ -> Inl (showsl_lit "Invalid clause")
      | Conjunction _ -> Inl (showsl_lit "Invalid clause")
      | Disjunction _ -> Inl (showsl_lit "Invalid clause"));;

let rec def_checker (_A1, _A2, _A3)
  x ty phi =
    catch_error
      (match phi with Atom _ -> def_checker_eq _A1 x ty phi
        | NegAtom _ -> def_checker_eq _A1 x ty phi | Conjunction [] -> Inr ()
        | Conjunction [eq] -> def_checker_eq _A1 x ty eq
        | Conjunction [Atom (Var _); _] ->
          Inl (showsl_lit "Invalid 1st literal")
        | Conjunction [Atom (Fun (LessF, _)); _] ->
          Inl (showsl_lit "Invalid 1st literal")
        | Conjunction [Atom (Fun (LeF, [])); _] ->
          Inl (showsl_lit "Invalid 1st literal")
        | Conjunction [Atom (Fun (LeF, [_])); _] ->
          Inl (showsl_lit "Invalid 1st literal")
        | Conjunction [Atom (Fun (LeF, [l1; r1])); lt2] ->
          (match lt2 with Atom (Var _) -> Inl (showsl_lit "Invalid 2nd literal")
            | Atom (Fun (LessF, _)) -> Inl (showsl_lit "Invalid 2nd literal")
            | Atom (Fun (LeF, [])) -> Inl (showsl_lit "Invalid 2nd literal")
            | Atom (Fun (LeF, [_])) -> Inl (showsl_lit "Invalid 2nd literal")
            | Atom (Fun (LeF, [l2; r2])) ->
              check (def_checker_main _A1 x ty l1 r1 l2 r2 ||
                      def_checker_main _A1 x ty l2 r2 l1 r1)
                (showsl_lit "Non-equation not supported")
            | Atom (Fun (LeF, _ :: _ :: _ :: _)) ->
              Inl (showsl_lit "Invalid 2nd literal")
            | Atom (Fun (SumF _, _)) -> Inl (showsl_lit "Invalid 2nd literal")
            | Atom (Fun (ConstF _, _)) -> Inl (showsl_lit "Invalid 2nd literal")
            | Atom (Fun (ProdF _, _)) -> Inl (showsl_lit "Invalid 2nd literal")
            | Atom (Fun (EqF, _)) -> Inl (showsl_lit "Invalid 2nd literal")
            | NegAtom _ -> Inl (showsl_lit "Invalid 2nd clause")
            | Conjunction _ -> Inl (showsl_lit "Invalid 2nd clause")
            | Disjunction _ -> Inl (showsl_lit "Invalid 2nd clause"))
        | Conjunction [Atom (Fun (LeF, _ :: _ :: _ :: _)); _] ->
          Inl (showsl_lit "Invalid 1st literal")
        | Conjunction [Atom (Fun (SumF _, _)); _] ->
          Inl (showsl_lit "Invalid 1st literal")
        | Conjunction [Atom (Fun (ConstF _, _)); _] ->
          Inl (showsl_lit "Invalid 1st literal")
        | Conjunction [Atom (Fun (ProdF _, _)); _] ->
          Inl (showsl_lit "Invalid 1st literal")
        | Conjunction [Atom (Fun (EqF, _)); _] ->
          Inl (showsl_lit "Invalid 1st literal")
        | Conjunction [NegAtom _; _] -> Inl (showsl_lit "Invalid 1st clause")
        | Conjunction [Conjunction _; _] ->
          Inl (showsl_lit "Invalid 1st clause")
        | Conjunction [Disjunction _; _] ->
          Inl (showsl_lit "Invalid 1st clause")
        | Conjunction (_ :: _ :: _ :: _) ->
          Inl (showsl_lit "Conjunction of more than two")
        | Disjunction _ -> def_checker_eq _A1 x ty phi)
      (fun xa ->
        Inl (comp (comp (comp (comp (comp (showsl_lit
    "error in definibility checker for ")
                                      (showsl _A3 x))
                                (showsl_lit " on formula\092  "))
                          (showsl_formula (showsl_IA_exp (_A1, _A2, _A3)) phi))
                    (showsl_literal "\092"))
              xa));;

let rec is_constant = function Fun (uu, []) -> true
                      | Var v -> false
                      | Fun (v, vb :: vc) -> false;;

let rec check_termination (_A1, _A2, _A3, _A4, _A5, _A6)
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) (_C1, _C2, _C3, _C4, _C5, _C6)
  = checka (ccompare_sig, equal_sig, showl_sig)
      (ceq_ty, ccompare_ty, equal_ty, showl_ty) (default_hints, showl_hints)
      (_A1, _A3, _A6) (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
      (_C1, _C2, _C3, _C4, _C5, _C6) type_of_fun
      (inserta (ceq_ty, ccompare_ty) BoolT
        (set_empty (ceq_ty, ccompare_ty) (of_phantom set_impl_ty)))
      (check_clause (_A1, _A3, _A4, _A5, _A6))
      (check_clause
        ((ccompare_trans_var _A1), (equal_trans_var _A3),
          mapping_impl_trans_var, (linorder_trans_var _A2),
          (showl_trans_var _A6)))
      (showsl_IA_exp (_A3, _A5, _A6))
      (showsl_IA_exp
        ((equal_trans_var _A3), (linorder_trans_var _A2),
          (showl_trans_var _A6)))
      negate negate IntT (fun s t -> Atom (Fun (LeF, [s; t])))
      (fun s t -> Atom (Fun (LessF, [s; t]))) is_constant
      (def_checker
        ((equal_trans_var _A3), (linorder_trans_var _A2),
          (showl_trans_var _A6)));;

let rec get_args_impl x0 t = match x0, t with true, t -> args t
                        | false, t -> [t];;

let rec innermost_repl_map_impl (_A1, _A2, _A3)
  r ecap p =
    remdups (equal_prod (equal_prod _A2 equal_nat) equal_nat)
      (maps (fun (a, b) ->
              (let (l, ra) = a in
                (fun ba ->
                  maps (fun u ->
                         (if not (is_Var u)
                           then maps (fun rs ->
                                       maps
 (fun f ->
   maps (fun n ->
          maps (fun i ->
                 (if contains_var_term
                       (equal_sum equal_unit (equal_list equal_char)) (Inl ())
                       (ecap (get_args_impl ba l) (nth rs i))
                   then [(f, i)] else []))
            (upt zero_nata n))
     [snd f])
 [the (root u)])
                                  [args u]
                           else []))
                    (supteq_list ra)))
                b)
        (map (fun lr -> (lr, true)) r @ map (fun st -> (st, false)) p));;

let rec mu_i_P_impl (_A1, _A2, _A3)
  r ecap p =
    (let fis = innermost_repl_map_impl (_A1, _A2, _A3) r ecap p in
     let fs = remdups (equal_prod _A2 equal_nat) (map fst fis) in
     let mu =
       (fun f ->
         set (ceq_nat, ccompare_nat, set_impl_nat)
           (map_filter
             (fun x ->
               (if (let (g, _) = x in equal_proda _A2 equal_nat g f)
                 then Some (snd x) else None))
             fis))
       in
      (fs, (precompute_fun (compare_order_prod _A1 compare_order_nat) mu fs,
             "innermost URM")));;

let rec mu_i_impl (_A1, _A2, _A3)
  r ecap = mu_i_P_impl (_A1, _A2, _A3) r ecap [];;

let rec default_fs (_A1, _A2) r = funas_trs_list r;;

let rec full_empty _A
  fs = (let fsa = filtera (fun (_, n) -> not (equal_nata n zero_nata)) fs in
         (fsa, ((fun f ->
                  (if membera (equal_prod _A equal_nat) fsa f then full_af f
                    else set_empty (ceq_nat, ccompare_nat)
                           (of_phantom set_impl_nata))),
                 "full AF")));;

let rec get_fs_mu (_A1, _A2, _A3, _A4, _A5, _A6, _A7)
  r ecap inn x3 = match r, ecap, inn, x3 with
    r, ecap, inn, Runtime_Complexity (c, d) ->
      (if inn &&
            less_eq_set
              ((cenum_prod _A1 cenum_nat), (ceq_prod _A2 ceq_nat),
                (ccompare_prod _A3 ccompare_nat))
              (inf_seta
                ((ceq_prod _A2 ceq_nat), (ccompare_prod _A3 ccompare_nat))
                (set ((ceq_prod _A2 ceq_nat), (ccompare_prod _A3 ccompare_nat),
                       (set_impl_prod _A6 set_impl_nat))
                  c)
                (set ((ceq_prod _A2 ceq_nat), (ccompare_prod _A3 ccompare_nat),
                       (set_impl_prod _A6 set_impl_nat))
                  (defined_list r)))
              (bot_set
                ((ceq_prod _A2 ceq_nat), (ccompare_prod _A3 ccompare_nat),
                  (set_impl_prod _A6 set_impl_nat)))
        then mu_i_impl (_A4, _A5, _A7) r ecap
        else full_empty _A5
               (remdups (equal_prod _A5 equal_nat)
                 (c @ d @ default_fs (_A4, _A7) r)))
    | r, ecap, inn, Derivational_Complexity f ->
        full_empty _A5
          (remdups (equal_prod _A5 equal_nat) (f @ default_fs (_A4, _A7) r));;

let rec is_Fsharp_term (_A1, _A2, _A3, _A4)
  uu uv x2 = match uu, uv, x2 with uu, uv, Var uw -> false
    | fs, fa, Fun (f, ts) ->
        member ((ceq_prod _A2 ceq_nat), (ccompare_prod _A3 ccompare_nat))
          (f, size_list ts) fs &&
          less_eq_set
            ((cenum_prod _A1 cenum_nat), (ceq_prod _A2 ceq_nat),
              (ccompare_prod _A3 ccompare_nat))
            (set ((ceq_prod _A2 ceq_nat), (ccompare_prod _A3 ccompare_nat),
                   (set_impl_prod _A4 set_impl_nat))
              (maps funas_term_list ts))
            fa;;

let rec add_funas_mctxt
  x0 fs = match x0, fs with
    MFun (f, cs), fs -> (f, size_list cs) :: foldr add_funas_mctxt cs fs
    | MVar v, fs -> fs
    | MHole, fs -> fs;;

let rec funas_mctxt_list c = add_funas_mctxt c [];;

let rec split_term
  p x1 = match p, x1 with
    p, Var x -> (if p (Var x) then (MHole, [Var x]) else (MVar x, []))
    | p, Fun (f, ts) ->
        (if p (Fun (f, ts)) then (MHole, [Fun (f, ts)])
          else (let us = map (split_term p) ts in
                 (MFun (f, map fst us), maps snd us)));;

let rec split_DP (_A1, _A2)
  fs = (fun r ->
         split_term
           (fun t ->
             not (is_Var t) &&
               member ((ceq_prod _A1 ceq_nat), (ccompare_prod _A2 ccompare_nat))
                 (the (root t)) fs)
           (snd r));;

let rec check_DP_complexity (_A1, _A2, _A3, _A4, _A5, _A6, _A7) _B
  p x1 = match p, x1 with
    p, Derivational_Complexity uu ->
      Inl (showsl_lit "require runtime complexity")
    | p, Runtime_Complexity (c, fs) ->
        (let fsa =
           set ((ceq_prod _A3 ceq_nat),
                 (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat),
                 (set_impl_prod _A6 set_impl_nat))
             fs
           in
         let (rs, r) =
           partition
             (fun lr ->
               member
                 ((ceq_prod _A3 ceq_nat),
                   (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat))
                 (the (root (fst lr))) fsa)
             p
           in
         let cs_ts = map (split_DP (_A3, _A4.ccompare_cproper_interval) fsa) rs
           in
         let cp =
           remdups (equal_prod _A5 equal_nat)
             (maps (comp funas_mctxt_list fst) cs_ts)
           in
         let cpa =
           set ((ceq_prod _A3 ceq_nat),
                 (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat),
                 (set_impl_prod _A6 set_impl_nat))
             cp
           in
         let f =
           remdups (equal_prod _A5 equal_nat)
             (c @ funas_trs_list r @
                    concat
                      (maps (fun (fsb, _) -> map funas_term_list (args fsb))
                        rs) @
                      concat
                        (maps (fun (_, a) ->
                                maps (fun t -> map funas_term_list (args t)) a)
                          cs_ts))
           in
         let fa =
           set ((ceq_prod _A3 ceq_nat),
                 (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat),
                 (set_impl_prod _A6 set_impl_nat))
             f
           in
          bindb (check
                  (is_empty
                     ((card_UNIV_prod _A1 card_UNIV_nat),
                       (ceq_prod _A3 ceq_nat),
                       (cproper_interval_prod _A4 cproper_interval_nat))
                     (inf_seta
                       ((ceq_prod _A3 ceq_nat),
                         (ccompare_prod _A4.ccompare_cproper_interval
                           ccompare_nat))
                       fa fsa) &&
                    (is_empty
                       ((card_UNIV_prod _A1 card_UNIV_nat),
                         (ceq_prod _A3 ceq_nat),
                         (cproper_interval_prod _A4 cproper_interval_nat))
                       (inf_seta
                         ((ceq_prod _A3 ceq_nat),
                           (ccompare_prod _A4.ccompare_cproper_interval
                             ccompare_nat))
                         fa cpa) &&
                      is_empty
                        ((card_UNIV_prod _A1 card_UNIV_nat),
                          (ceq_prod _A3 ceq_nat),
                          (cproper_interval_prod _A4 cproper_interval_nat))
                        (inf_seta
                          ((ceq_prod _A3 ceq_nat),
                            (ccompare_prod _A4.ccompare_cproper_interval
                              ccompare_nat))
                          fsa cpa)))
                  (showsl_lit "symbols are not disjoint"))
            (fun _ ->
              bindb (check
                      (list_all
                        (fun lr ->
                          is_Fsharp_term
                            (_A2, _A3, _A4.ccompare_cproper_interval, _A6) fsa
                            fa (fst lr))
                        rs)
                      (showsl_lit "lhss of RS are not sharp terms"))
                (fun _ -> Inr (rs, (r, (cp, (fs, f)))))));;

let rec get_fs_mu_DP (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8)
  r ecap inn s cm =
    (let (fs, (mu, info)) =
       get_fs_mu (_A2, _A3, _A4.ccompare_cproper_interval, _A5, _A6, _A7, _A8) r
         ecap inn cm
       in
      (match
        check_DP_complexity (_A1, _A2, _A3, _A4, _A6, _A7, _A8)
          (showl_list showl_char) r cm
        with Inl _ -> (fs, (mu, info))
        | Inr (rs, (_, (cp, (_, _)))) ->
          (if list_all
                (membera
                  (equal_prod (equal_term _A6 (equal_list equal_char))
                    (equal_term _A6 (equal_list equal_char)))
                  rs)
                s
            then (list_inter (equal_prod _A6 equal_nat) fs cp,
                   ((fun f ->
                      (if membera (equal_prod _A6 equal_nat) cp f then mu f
                        else set_empty (ceq_nat, ccompare_nat)
                               (of_phantom set_impl_nata))),
                     info ^ " with DPs"))
            else (fs, (mu, info)))));;

let rec showsl_complexity_class
  c = (if equal_nata (degree c) zero_nata then showsl_lit "O(1)"
        else (if equal_nata (degree c) one_nata then showsl_lit "O(n)"
               else comp (comp (showsl_lit "O(n^") (showsl_nat (degree c)))
                      (showsl_lit ")")));;

let rec is_ur_closed_af_impl_tp_mv_impl (_A1, _A2) (_B1, _B2)
  ic qnf r pi =
    (let urc =
       (fun s ->
         is_ur_closed_term_af_impl (_A1, _A2) (_B1, _B2) qnf (ic s) pi r)
       in
      (fun u s ->
        (let sa =
           map (map_term (fun x -> x)
                 (fun a ->
                   Chara (false, false, false, true, true, true, true, false) ::
                     a))
             s
           in
          (fun t ->
            urc sa u sa
              (map_term (fun x -> x)
                (fun a ->
                  Chara (false, false, false, true, true, true, true, false) ::
                    a)
                t)))));;

let rec extract_rt_C_D
  = function Runtime_Complexity (c, d) -> Inr (c, d)
    | Derivational_Complexity v ->
        Inl (showsl_lit "runtime complexity required");;

let rec cpx
  (Redtriple_ext
    (valid, s, ns, nst, af, mono_af, mono, desc, not_ws_ns, cpx, more))
    = cpx;;

let rec rule_shift_complexity_urm_ur_tt (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  i rp rdelete ur cm cc tp =
    (let rb = rulesd i tp in
     let r = rc i tp in
     let rw = rwc i tp in
     let r2 =
       ceta_list_diff
         (compare_order_prod
           (compare_order_term _B5.compare_compare_order
             (compare_list compare_char))
           (compare_order_term _B5.compare_compare_order
             (compare_list compare_char)))
         r rdelete
       in
     let rremain = rw @ r2 in
     let isnf = is_QNFb i tp in
     let inn = nFQ_subset_NF_rulesb i tp in
      (match
        catch_error
          (bindb
            (catch_error
              (check_subseteq
                (equal_prod (equal_term _B6 (equal_list equal_char))
                  (equal_term _B6 (equal_list equal_char)))
                rdelete rb)
              (fun x ->
                Inl (comp (comp (showsl_lit "rule ")
                            (showsl_rule _B8 (showl_list showl_char) x))
                      (showsl_lit
                        " should be deleted, but does not occur in problem"))))
            (fun _ ->
              bindb (check_wf_trs _B8
                      ((equal_list equal_char), (showl_list showl_char)) rb)
                (fun _ ->
                  bindb (check (nFQ_subset_NF_rulesb i tp)
                          (showsl_lit "innermost required"))
                    (fun _ ->
                      (let (fs, (mu, info)) =
                         get_fs_mu_DP (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
                           rb (icap_impl _B6 isnf rb) inn rdelete cm
                         in
                       let (_, (pi, _)) =
                         get_fs_mu_DP (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
                           rb (icap_impl _B6 isnf rb) inn rremain cm
                         in
                        bindb (extract_rt_C_D cm)
                          (fun (cl, dl) ->
                            (let c =
                               set ((ceq_prod _B3 ceq_nat),
                                     (ccompare_prod
                                       _B4.ccompare_cproper_interval
                                       ccompare_nat),
                                     (set_impl_prod _B7 set_impl_nat))
                                 cl
                               in
                             let d =
                               set ((ceq_prod _B3 ceq_nat),
                                     (ccompare_prod
                                       _B4.ccompare_cproper_interval
                                       ccompare_nat),
                                     (set_impl_prod _B7 set_impl_nat))
                                 dl
                               in
                              bindb (check
                                      (null
(list_inter (equal_prod _B6 equal_nat) cl (defined_list rb)))
                                      (comp
(comp (showsl_lit "constructors ") (showsl_lista (showl_prod _B8 showl_nat) cl))
(showsl_lit " must not be defined")))
                                (fun _ ->
                                  bindb (valid rp)
                                    (fun _ ->
                                      bindb
(catch_error
  (forallM
    (fun (l, ra) ->
      check (not (member
                    ((ceq_prod _B3 ceq_nat),
                      (ccompare_prod _B4.ccompare_cproper_interval
                        ccompare_nat))
                    (the (root l)) d &&
                   less_eq_set
                     ((cenum_prod _B2 cenum_nat), (ceq_prod _B3 ceq_nat),
                       (ccompare_prod _B4.ccompare_cproper_interval
                         ccompare_nat))
                     (foldr
                       (comp (sup_seta
                               ((ceq_prod _B3 ceq_nat),
                                 (ccompare_prod _B4.ccompare_cproper_interval
                                   ccompare_nat)))
                         (funas_term (_B3, _B4.ccompare_cproper_interval, _B7)))
                       (args l)
                       (bot_set
                         ((ceq_prod _B3 ceq_nat),
                           (ccompare_prod _B4.ccompare_cproper_interval
                             ccompare_nat),
                           (set_impl_prod _B7 set_impl_nat))))
                     c))
        (comp (showsl_rule _B8 (showl_list showl_char) (l, ra))
          (showsl_lit " should be usable")))
    (list_diff
      (equal_prod (equal_term _B6 (equal_list equal_char))
        (equal_term _B6 (equal_list equal_char)))
      rb ur))
  (fun x -> Inl (snd x)))
(fun _ ->
  (let ic = icap_impl_tp _B6 i tp in
   let qnf = is_QNFb i tp in
   let ra = rulesd i tp in
   let uu =
     set ((ceq_prod (ceq_term _B6 (equal_list equal_char))
            (ceq_term _B6 (equal_list equal_char))),
           (ccompare_prod
             (ccompare_term _B5.compare_compare_order
               (compare_list compare_char))
             (ccompare_term _B5.compare_compare_order
               (compare_list compare_char))),
           (set_impl_prod set_impl_term set_impl_term))
       ur
     in
   let is_urc =
     is_ur_closed_af_impl_tp_mv_impl (_B5.compare_compare_order, _B6)
       ((ceq_term _B6 (equal_list equal_char)),
         (ccompare_term _B5.compare_compare_order (compare_list compare_char)))
       ic qnf ra mu uu
     in
   let pia = af_inter (af rp) pi in
   let is_urc_pi =
     is_ur_closed_af_impl_tp_mv_impl (_B5.compare_compare_order, _B6)
       ((ceq_term _B6 (equal_list equal_char)),
         (ccompare_term _B5.compare_compare_order (compare_list compare_char)))
       ic qnf ra pia uu
     in
    bindb (catch_error
            (forallM
              (fun (l, rd) ->
                check (is_urc (args l) rd && is_urc_pi (args l) rd)
                  (comp (comp (showsl_lit
                                "problem with closure properties of usable rule ")
                          (showsl_rule _B8 (showl_list showl_char) (l, rd)))
                    (showsl_lit ": rhs is not closed under usable rules")))
              ur)
            (fun x -> Inl (snd x)))
      (fun _ ->
        bindb (catch_error
                (catch_error
                  (forallM
                    (fun f ->
                      check (subset
                              (card_UNIV_nat, cenum_nat, ceq_nat, ccompare_nat)
                              (mu f) (mono_af rp f))
                        (comp (comp (comp (comp
    (comp (showsl_lit "error in monotonicity: strict order for ")
      (showsl_prod _B8 showl_nat f))
    (showsl_lit " ensures monotonicity in positions "))
                                      (showsl_position_set f (mono_af rp f)))
                                (showsl_lit
                                  "\092but usable replacement map is "))
                          (showsl_position_set f (mu f))))
                    fs)
                  (fun x -> Inl (snd x)))
                (fun x ->
                  Inl (comp (comp (comp (comp
  (comp x (showsl_lit "\092the computed usable replacement map ("))
  (showsl_literal info))
                                    (showsl_lit ") is\092"))
                              (showsl_sep
                                (fun f ->
                                  comp (comp
 (comp (showsl_lit "mu(") (showsl_prod _B8 showl_nat f)) (showsl_lit ") = "))
                                    (showsl_position_set f (mu f)))
                                (showsl_literal "\092") fs))
                        (showsl_lit
                          "\092and mu(f) = {} for all other symbols f"))))
          (fun _ ->
            bindb (catch_error
                    (catch_error
                      (forallM (s rp)
                        (list_inter
                          (equal_prod (equal_term _B6 (equal_list equal_char))
                            (equal_term _B6 (equal_list equal_char)))
                          rdelete ur))
                      (fun x -> Inl (snd x)))
                    (fun x ->
                      Inl (comp (showsl_lit
                                  "problem when orienting strict TRS\092")
                            x)))
              (fun _ ->
                bindb (catch_error
                        (catch_error
                          (forallM (ns rp)
                            (list_inter
                              (equal_prod
                                (equal_term _B6 (equal_list equal_char))
                                (equal_term _B6 (equal_list equal_char)))
                              rremain ur))
                          (fun x -> Inl (snd x)))
                        (fun x ->
                          Inl (comp (showsl_lit
                                      "problem when orienting non-strict TRS\092")
                                x)))
                  (fun _ ->
                    catch_error (cpx rp cm cc)
                      (fun x ->
                        Inl (comp (showsl_lit
                                    "problem when ensuring complexity of order\092")
                              x)))))))))))))))))
          (fun x ->
            Inl (comp (comp (comp (comp (comp
  (comp (comp (showsl_lit "could not derive the intended complexity ")
          (showsl_complexity_class cc))
    (showsl_lit " from the following\092"))
  (desc rp))
                                    (showsl_literal "\092"))
                              x)
                        (showsl_lit "\092with usable rules\092"))
                  (showsl_trs _B8 (showl_list showl_char) ur)))
        with Inl a -> Inl a
        | Inr _ ->
          Inr (mkc i (nfsb i tp) (qb i tp) r2
                (list_union
                  (equal_prod (equal_term _B6 (equal_list equal_char))
                    (equal_term _B6 (equal_list equal_char)))
                  rw rdelete))));;

let rec inn_usable_rules_wf_tp (_B1, _B2)
  i d nfs =
    inn_usable_rules_wf (_B1, _B2) (is_QNFb i d) (icap_impl_tp _B2 i d)
      (rulesd i d) nfs;;

let rec x_1_to_x_n
  x = fresh_strings_list
        [Chara (false, false, false, true, true, true, true, false)] one_nata []
        x;;

let rec get_innermost_strict_repl_map_rc (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  i d s x3 = match i, d, s, x3 with
    i, d, s, Derivational_Complexity f ->
      full_empty _B5
        (remdups (equal_prod _B5 equal_nat)
          (f @ default_fs (_B4, _B7) (rulesd i d)))
    | i, da, s, Runtime_Complexity (c, d) ->
        (let r = rulesd i da in
          (if nFQ_subset_NF_rulesb i da &&
                less_eq_set
                  ((cenum_prod _B1 cenum_nat), (ceq_prod _B2 ceq_nat),
                    (ccompare_prod _B3 ccompare_nat))
                  (inf_seta
                    ((ceq_prod _B2 ceq_nat), (ccompare_prod _B3 ccompare_nat))
                    (set ((ceq_prod _B2 ceq_nat),
                           (ccompare_prod _B3 ccompare_nat),
                           (set_impl_prod _B6 set_impl_nat))
                      c)
                    (set ((ceq_prod _B2 ceq_nat),
                           (ccompare_prod _B3 ccompare_nat),
                           (set_impl_prod _B6 set_impl_nat))
                      (defined_list r)))
                  (bot_set
                    ((ceq_prod _B2 ceq_nat), (ccompare_prod _B3 ccompare_nat),
                      (set_impl_prod _B6 set_impl_nat)))
            then (let isNF = is_QNFb i da in
                  let u = inn_usable_rules_wf_tp (_B4, _B5) i da true in
                  let a = icap_impl_tp _B5 i da in
                   mu_approx_impl (_B4, _B5) r
                     (maps (fun (f, n) ->
                             maps (fun xs ->
                                    map (fun lr -> (xs, (Fun (f, xs), lr))) s)
                               [map (fun aa -> Var aa) (x_1_to_x_n n)])
                       d)
                     (fun ss t -> u (ss, t)) isNF a)
            else full_empty _B5
                   (remdups (equal_prod _B5 equal_nat)
                     (c @ d @ default_fs (_B4, _B7) r))));;

let rec get_innermost_strict_repl_map_rc_DP
  (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  i d s t =
    (let (fs, (mu, info)) =
       get_innermost_strict_repl_map_rc
         (_B2, _B3, _B4.ccompare_cproper_interval, _B5, _B6, _B7, _B8) i d s t
       in
      (match
        check_DP_complexity (_B1, _B2, _B3, _B4, _B6, _B7, _B8)
          (showl_list showl_char) (rulesd i d) t
        with Inl _ -> (fs, (mu, info))
        | Inr (rs, (_, (cp, (_, _)))) ->
          (if list_all
                (membera
                  (equal_prod (equal_term _B6 (equal_list equal_char))
                    (equal_term _B6 (equal_list equal_char)))
                  rs)
                s
            then (list_inter (equal_prod _B6 equal_nat) fs cp,
                   ((fun f ->
                      (if membera (equal_prod _B6 equal_nat) cp f then mu f
                        else set_empty (ceq_nat, ccompare_nat)
                               (of_phantom set_impl_nata))),
                     info ^ " with DPs"))
            else (fs, (mu, info)))));;

let rec rule_shift_complexity_urm_tt (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  i rp rdelete cm cc tp =
    (let rb = rulesd i tp in
     let r = rc i tp in
     let rw = rwc i tp in
     let r2 =
       ceta_list_diff
         (compare_order_prod
           (compare_order_term _B5.compare_compare_order
             (compare_list compare_char))
           (compare_order_term _B5.compare_compare_order
             (compare_list compare_char)))
         r rdelete
       in
     let _ = qb i tp in
      (match
        catch_error
          (bindb
            (catch_error
              (check_subseteq
                (equal_prod (equal_term _B6 (equal_list equal_char))
                  (equal_term _B6 (equal_list equal_char)))
                rdelete rb)
              (fun x ->
                Inl (comp (comp (showsl_lit "rule ")
                            (showsl_rule _B8 (showl_list showl_char) x))
                      (showsl_lit
                        " should be deleted, but does not occur in problem"))))
            (fun _ ->
              bindb (check_wf_trs _B8
                      ((equal_list equal_char), (showl_list showl_char)) rb)
                (fun _ ->
                  (let (fs, (mu, info)) =
                     get_innermost_strict_repl_map_rc_DP
                       (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) i tp rdelete cm
                     in
                    bindb (valid rp)
                      (fun _ ->
                        bindb (catch_error
                                (catch_error
                                  (forallM
                                    (fun f ->
                                      check
(subset (card_UNIV_nat, cenum_nat, ceq_nat, ccompare_nat) (mu f) (mono_af rp f))
(comp (comp (comp (comp (comp (showsl_lit
                                "error in monotonicity: strict order for ")
                          (showsl_prod _B8 showl_nat f))
                    (showsl_lit " ensures monotonicity in positions "))
              (showsl_position_set f (mono_af rp f)))
        (showsl_lit "\092but usable replacement map is "))
  (showsl_position_set f (mu f))))
                                    fs)
                                  (fun x -> Inl (snd x)))
                                (fun x ->
                                  Inl (comp
(comp (comp (comp (comp x
                    (showsl_lit "\092the computed usable replacement map ("))
              (showsl_literal info))
        (showsl_lit ") is\092"))
  (showsl_sep
    (fun f ->
      comp (comp (comp (showsl_lit "mu(") (showsl_prod _B8 showl_nat f))
             (showsl_lit ") = "))
        (showsl_position_set f (mu f)))
    (showsl_literal "\092") fs))
(showsl_lit "\092and mu(f) = {} for all other symbols f"))))
                          (fun _ ->
                            bindb (catch_error
                                    (catch_error (forallM (s rp) rdelete)
                                      (fun x -> Inl (snd x)))
                                    (fun x ->
                                      Inl
(comp (showsl_lit "problem when orienting strict TRS\092") x)))
                              (fun _ ->
                                bindb (catch_error
(catch_error (forallM (ns rp) (rw @ r2)) (fun x -> Inl (snd x)))
(fun x ->
  Inl (comp (showsl_lit "problem when orienting non-strict TRS\092") x)))
                                  (fun _ ->
                                    catch_error (cpx rp cm cc)
                                      (fun x ->
Inl (comp (showsl_lit "problem when ensuring complexity of order\092")
      x))))))))))
          (fun x ->
            Inl (comp (comp (comp (comp (comp
  (showsl_lit "could not derive the intended complexity ")
  (showsl_complexity_class cc))
                                    (showsl_lit " from the following\092"))
                              (desc rp))
                        (showsl_literal "\092"))
                  x))
        with Inl a -> Inl a
        | Inr _ ->
          Inr (mkc i (nfsb i tp) (qb i tp) r2
                (list_union
                  (equal_prod (equal_term _B6 (equal_list equal_char))
                    (equal_term _B6 (equal_list equal_char)))
                  rw rdelete))));;

let rec smart_rule_shift_complexity (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8)
  i rp rdelete x3 cm cc tp = match i, rp, rdelete, x3, cm, cc, tp with
    i, rp, rdelete, Some ur, cm, cc, tp ->
      rule_shift_complexity_urm_ur_tt (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) i
        rp rdelete ur cm cc tp
    | i, rp, rdelete, None, cm, cc, tp ->
        rule_shift_complexity_urm_tt (_B1, _B2, _B3, _B4, _B5, _B6, _B7, _B8) i
          rp rdelete cm cc tp;;

let rec is_ur_closed_af_impl_tp_mv (_B1, _B2, _B3)
  i d pi u =
    (let ic = icap_impl_tp _B2 i d in
     let qnf = is_QNFb i d in
     let r = rulesd i d in
     let urc =
       (fun s ->
         is_ur_closed_term_af_impl (_B1.compare_compare_order, _B2)
           ((ceq_term _B2 (equal_list equal_char)),
             (ccompare_term _B1.compare_compare_order
               (compare_list compare_char)))
           qnf (ic s) pi r
           (set ((ceq_prod (ceq_term _B2 (equal_list equal_char))
                   (ceq_term _B2 (equal_list equal_char))),
                  (ccompare_prod
                    (ccompare_term _B1.compare_compare_order
                      (compare_list compare_char))
                    (ccompare_term _B1.compare_compare_order
                      (compare_list compare_char))),
                  (set_impl_prod set_impl_term set_impl_term))
             u))
       in
      (fun s ->
        (let sa =
           map (map_term (fun x -> x)
                 (fun a ->
                   Chara (false, false, false, true, true, true, true, false) ::
                     a))
             s
           in
          (fun t ->
            urc sa sa
              (map_term (fun x -> x)
                (fun a ->
                  Chara (false, false, false, true, true, true, true, false) ::
                    a)
                t)))));;

let rec usable_rules_complexity_innermost (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  i nUr cm cc cp =
    catch_error
      (let s = rc i cp in
       let w = rwc i cp in
       let r = s @ w in
        bindb (check (nFQ_subset_NF_rulesb i cp)
                (showsl_lit "innermost required"))
          (fun _ ->
            bindb (check_wf_trs _B7
                    ((equal_list equal_char), (showl_list showl_char)) r)
              (fun _ ->
                bindb (extract_rt_C_D cm)
                  (fun (cl, dl) ->
                    (let c =
                       set ((ceq_prod _B2 ceq_nat),
                             (ccompare_prod _B3 ccompare_nat),
                             (set_impl_prod _B6 set_impl_nat))
                         cl
                       in
                     let d =
                       set ((ceq_prod _B2 ceq_nat),
                             (ccompare_prod _B3 ccompare_nat),
                             (set_impl_prod _B6 set_impl_nat))
                         dl
                       in
                     let isnf = is_QNFb i cp in
                      bindb (check
                              (null (list_inter (equal_prod _B5 equal_nat) cl
                                      (defined_list r)))
                              (comp (comp (showsl_lit "constructors ")
                                      (showsl_lista (showl_prod _B7 showl_nat)
cl))
                                (showsl_lit " must not be defined")))
                        (fun _ ->
                          bindb (catch_error
                                  (check_subseteq
                                    (equal_prod
                                      (equal_term _B5 (equal_list equal_char))
                                      (equal_term _B5 (equal_list equal_char)))
                                    nUr r)
                                  (fun _ ->
                                    Inl (comp
  (comp (showsl_lit "rule ") (showsl_rules _B7 (showl_list showl_char) nUr))
  (showsl_lit " does not occur in problem"))))
                            (fun _ ->
                              (let ur =
                                 list_diff
                                   (equal_prod
                                     (equal_term _B5 (equal_list equal_char))
                                     (equal_term _B5 (equal_list equal_char)))
                                   r nUr
                                 in
                                bindb (catch_error
(forallM
  (fun (l, ra) ->
    check (not (member
                  ((ceq_prod _B2 ceq_nat), (ccompare_prod _B3 ccompare_nat))
                  (the (root l)) d &&
                 less_eq_set
                   ((cenum_prod _B1 cenum_nat), (ceq_prod _B2 ceq_nat),
                     (ccompare_prod _B3 ccompare_nat))
                   (foldr
                     (comp (sup_seta
                             ((ceq_prod _B2 ceq_nat),
                               (ccompare_prod _B3 ccompare_nat)))
                       (funas_term (_B2, _B3, _B6)))
                     (args l)
                     (bot_set
                       ((ceq_prod _B2 ceq_nat),
                         (ccompare_prod _B3 ccompare_nat),
                         (set_impl_prod _B6 set_impl_nat))))
                   c))
      (comp (showsl_rule _B7 (showl_list showl_char) (l, ra))
        (showsl_lit " should be usable")))
  nUr)
(fun x -> Inl (snd x)))
                                  (fun _ ->
                                    (let (_, (mu, _)) =
                                       get_fs_mu
 (_B1, _B2, _B3, _B4, _B5, _B6, _B7) r (icap_impl _B5 isnf r) true cm
                                       in
                                     let is_urc =
                                       is_ur_closed_af_impl_tp_mv
 (_B4, _B5, _B7) i cp mu ur
                                       in
                                      bindb
(catch_error
  (forallM
    (fun (l, ra) ->
      check (is_urc (args l) ra)
        (comp (comp (showsl_lit
                      "problem with closure properties of usable rule ")
                (showsl_rule _B7 (showl_list showl_char) (l, ra)))
          (showsl_lit ": rhs is not closed under usable rules")))
    ur)
  (fun x -> Inl (snd x)))
(fun _ ->
  Inr (mkc i (nfsb i cp) (qb i cp)
        (list_diff
          (equal_prod (equal_term _B5 (equal_list equal_char))
            (equal_term _B5 (equal_list equal_char)))
          s nUr)
        (list_diff
          (equal_prod (equal_term _B5 (equal_list equal_char))
            (equal_term _B5 (equal_list equal_char)))
          w nUr)))))))))))))
      (fun x ->
        Inl (comp (showsl_lit
                    "error when restricting to innermost usable rules\092")
              x));;

let rec get_signature_of_cm = function Derivational_Complexity f -> f
                              | Runtime_Complexity (c, d) -> c @ d;;

let rec usable_rules_complexity_usymbols (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  (_C1, _C2, _C3)
  i nUr cm cc cp =
    catch_error
      (let s = rc i cp in
       let w = rwc i cp in
       let r = s @ w in
        bindb (catch_error
                (check_subseteq
                  (equal_prod (equal_term _B5 _C2) (equal_term _B5 _C2)) nUr r)
                (fun _ ->
                  Inl (comp (comp (showsl_lit "rule ")
                              (showsl_rules _B7 _C3 nUr))
                        (showsl_lit " does not occur in problem"))))
          (fun _ ->
            (let ur =
               list_diff (equal_prod (equal_term _B5 _C2) (equal_term _B5 _C2))
                 r nUr
               in
             let us =
               set ((ceq_prod _B2 ceq_nat), (ccompare_prod _B3 ccompare_nat),
                     (set_impl_prod _B6 set_impl_nat))
                 (maps (comp funas_term_list snd) ur @ get_signature_of_cm cm)
               in
             let urs =
               set ((ceq_prod (ceq_term _B5 _C2) (ceq_term _B5 _C2)),
                     (ccompare_prod (ccompare_term _B4 _C1)
                       (ccompare_term _B4 _C1)),
                     (set_impl_prod set_impl_term set_impl_term))
                 ur
               in
              bindb (check_varcond_subset _B7 (_C2, _C3) ur)
                (fun _ ->
                  bindb (catch_error
                          (forallM
                            (fun lr ->
                              check (if less_eq_set
  ((cenum_prod _B1 cenum_nat), (ceq_prod _B2 ceq_nat),
    (ccompare_prod _B3 ccompare_nat))
  (funas_term (_B2, _B3, _B6) (fst lr)) us
                                      then member
     ((ceq_prod (ceq_term _B5 _C2) (ceq_term _B5 _C2)),
       (ccompare_prod (ccompare_term _B4 _C1) (ccompare_term _B4 _C1)))
     lr urs
                                      else true)
                                (comp (comp (showsl_lit "rule ")
(showsl_rule _B7 _C3 lr))
                                  (showsl_lit " should be usable")))
                            r)
                          (fun x -> Inl (snd x)))
                    (fun _ ->
                      Inr (mkc i (nfsb i cp) (qb i cp)
                            (list_diff
                              (equal_prod (equal_term _B5 _C2)
                                (equal_term _B5 _C2))
                              s nUr)
                            (list_diff
                              (equal_prod (equal_term _B5 _C2)
                                (equal_term _B5 _C2))
                              w nUr)))))))
      (fun x ->
        Inl (comp (showsl_lit
                    "error when restricting to usable rules w.r.t. usable symbols\092")
              x));;

let rec usable_rules_complexity (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
  i nUr cm cc cp =
    (match
      usable_rules_complexity_usymbols
        (_B1, _B2, _B3, _B4.compare_compare_order, _B5, _B6, _B7)
        ((compare_list compare_char), (equal_list equal_char),
          (showl_list showl_char))
        i nUr cm cc cp
      with Inl e ->
        (match
          usable_rules_complexity_innermost (_B1, _B2, _B3, _B4, _B5, _B6, _B7)
            i nUr cm cc cp
          with Inl ea ->
            Inl (comp (comp (comp (comp (showsl_lit
  "neither of the usable rules processors is applicable:\092")
                                    (showsl_lit
                                      "the one via usable symbols complains as follows\092"))
                              e)
                        (showsl_lit
                          "\092\092and the one via icap and innermost says\092"))
                  ea)
          | Inr a -> Inr a)
      | Inr a -> Inr a);;

let rec split_proc_complexity (_B1, _B2) (_C1, _C2)
  i s1 cp =
    catch_error
      (let s = rc i cp in
       let w = rwc i cp in
       let nfs = nfsb i cp in
       let q = qb i cp in
        bindb (catch_error
                (check_subseteq
                  (equal_prod (equal_term _B1 _C1) (equal_term _B1 _C1)) s1 s)
                (fun x ->
                  Inl (comp (comp (showsl_lit "rule ") (showsl_rule _B2 _C2 x))
                        (showsl_lit " is not a strict rule"))))
          (fun _ ->
            (let s2 =
               list_diff (equal_prod (equal_term _B1 _C1) (equal_term _B1 _C1))
                 s s1
               in
              Inr (mkc i nfs q s1 (s2 @ w), mkc i nfs q s2 (s1 @ w)))))
      (fun x ->
        Inl (comp (showsl_lit "error when splitting complexity problem\092")
              x));;

let rec showsl_complexity_measure (_A1, _A2)
  = function
    Runtime_Complexity (c, d) ->
      comp (comp (comp (comp (showsl_lit
                               "basic terms f(c1,..,cn) where f in\092")
                         (showsl_lista (showl_prod _A2 showl_nat)
                           (sort_key (linorder_prod _A1 linorder_nat)
                             (fun x -> x) d)))
                   (showsl_lit "\092 and ci is term over signature\092 "))
             (showsl_lista (showl_prod _A2 showl_nat)
               (sort_key (linorder_prod _A1 linorder_nat) (fun x -> x) c)))
        (showsl_literal "\092")
    | Derivational_Complexity f ->
        comp (comp (showsl_lit "all terms over signature\092")
               (showsl_lista (showl_prod _A2 showl_nat)
                 (sort_key (linorder_prod _A1 linorder_nat) (fun x -> x) f)))
          (showsl_literal "\092");;

let rec showsl_complexityLL (_A1, _A2) (_B1, _B2) (_C1, _C2)
  (q, (s, (w, (cm, cc)))) =
    comp (comp (comp (comp (comp (comp (comp
 (comp (comp (comp (showsl_lit "strict rules\092")
               (showsl_rules (showl_lab _A2 _B2) _C2
                 (sort_key
                   (linorder_prod
                     (linorder_term
                       (compare_lab _A1.compare_compare_order
                         _B1.compare_compare_order)
                       _C1.compare_compare_order)
                     (linorder_term
                       (compare_lab _A1.compare_compare_order
                         _B1.compare_compare_order)
                       _C1.compare_compare_order))
                   (fun x -> x) s)))
         (showsl_lit "\092\092weak rules\092"))
   (showsl_rules (showl_lab _A2 _B2) _C2
     (sort_key
       (linorder_prod
         (linorder_term
           (compare_lab _A1.compare_compare_order _B1.compare_compare_order)
           _C1.compare_compare_order)
         (linorder_term
           (compare_lab _A1.compare_compare_order _B1.compare_compare_order)
           _C1.compare_compare_order))
       (fun x -> x) w)))
 (showsl_lit "\092\092innermost lhss (Q)\092"))
                                   (showsl_lines
                                     (showl_term (showl_lab _A2 _B2) _C2)
                                     "empty"
                                     (sort_key
                                       (linorder_term
 (compare_lab _A1.compare_compare_order _B1.compare_compare_order)
 _C1.compare_compare_order)
                                       (fun x -> x) q)))
                             (showsl_lit "\092\092starting terms: "))
                       (showsl_complexity_measure
                         ((linorder_lab _A1.compare_compare_order
                            _B1.compare_compare_order),
                           (showl_lab _A2 _B2))
                         cm))
                 (showsl_lit "\092intended complexity: "))
           (showsl_complexity_class cc))
      (showsl_literal "\092");;

let rec check_terms_of_main (_A1, _A2)
  x0 x1 = match x0, x1 with
    Derivational_Complexity f, Derivational_Complexity g ->
      catch_error (check_subseteq (equal_prod _A1 equal_nat) f g)
        (fun x ->
          Inl (comp (showsl_lit "consider of symbol ")
                (showsl_prod _A2 showl_nat x)))
    | Runtime_Complexity (f, h), Derivational_Complexity g ->
        catch_error (check_subseteq (equal_prod _A1 equal_nat) (f @ h) g)
          (fun x ->
            Inl (comp (showsl_lit "consider of symbol ")
                  (showsl_prod _A2 showl_nat x)))
    | Runtime_Complexity (f, h), Runtime_Complexity (f1, h1) ->
        bindb (catch_error (check_subseteq (equal_prod _A1 equal_nat) f f1)
                (fun x ->
                  Inl (comp (showsl_lit "consider symbol ")
                        (showsl_prod _A2 showl_nat x))))
          (fun _ ->
            catch_error (check_subseteq (equal_prod _A1 equal_nat) h h1)
              (fun x ->
                Inl (comp (showsl_lit "consider symbol ")
                      (showsl_prod _A2 showl_nat x))))
    | Derivational_Complexity g, Runtime_Complexity (f1, h1) ->
        Inl (showsl_lit
              "mixing runtime complexity and derivational complexity");;

let rec check_terms_of_nat (_A1, _A2)
  cm1 cm2 =
    catch_error (check_terms_of_main (_A1, _A2) cm1 cm2)
      (fun x -> Inl (comp (showsl_lit "error comparing start terms\092") x));;

let rec check_complexity_subsumes (_A1, _A2, _A3) (_B1, _B2, _B3)
  (_C1, _C2, _C3, _C4, _C5)
  (q1, (s1, (w1, (cm1, cc1)))) (q2, (s2, (w2, (cm2, cc2)))) =
    catch_error
      (bindb
        (catch_error
          (check_subseteq
            (equal_prod (equal_term (equal_lab _A2 _B2) _C3)
              (equal_term (equal_lab _A2 _B2) _C3))
            s2 s1)
          (fun x ->
            Inl (comp (comp (showsl_lit "strict rule ")
                        (showsl_rule (showl_lab _A3 _B3) _C5 x))
                  (showsl_lit " is missing"))))
        (fun _ ->
          bindb (catch_error
                  (check_subseteq
                    (equal_prod (equal_term (equal_lab _A2 _B2) _C3)
                      (equal_term (equal_lab _A2 _B2) _C3))
                    w2 w1)
                  (fun x ->
                    Inl (comp (comp (showsl_lit "weak rule ")
                                (showsl_rule (showl_lab _A3 _B3) _C5 x))
                          (showsl_lit " is missing"))))
            (fun _ ->
              bindb (catch_error
                      (check_NF_terms_subset
                        (is_NF_terms
                          ((compare_order_lab _A1.compare_compare_order
                             _B1.compare_compare_order),
                            (equal_lab _A2 _B2))
                          (_C1, _C3, _C4) q2)
                        q1)
                      (fun x ->
                        Inl (comp (showsl_lit "NF(Q) differs due to term ")
                              (showsl_terma (showl_lab _A3 _B3) _C5 x))))
                (fun _ ->
                  bindb (check (less_eq_complexity_class cc1 cc2)
                          (showsl_lit "complexity classes do not match"))
                    (fun _ ->
                      check_terms_of_nat
                        ((equal_lab _A2 _B2), (showl_lab _A3 _B3)) cm2 cm1)))))
      (fun x ->
        Inl (comp (comp (comp (comp (comp (showsl_lit
    "could not ensure that assumption matches current complexity problem\092")
                                      x)
                                (showsl_lit "\092\092assumption is\092"))
                          (showsl_complexityLL (_A1, _A3) (_B1, _B3) (_C2, _C5)
                            (q1, (s1, (w1, (cm1, cc1))))))
                    (showsl_lit "\092\092current problem is\092"))
              (showsl_complexityLL (_A1, _A3) (_B1, _B3) (_C2, _C5)
                (q2, (s2, (w2, (cm2, cc2)))))));;

let rec check_tup (_A1, _A2) t x1 = match t, x1 with t, Var x -> false
                               | t, Fun (f, ts) -> member (_A1, _A2) f t;;

let rec dPos_impl (_A1, _A2)
  shp d x2 = match shp, d, x2 with shp, d, Var x -> []
    | shp, d, Fun (f, ts) ->
        (let n = size_list ts in
          (if member ((ceq_prod _A1 ceq_nat), (ccompare_prod _A2 ccompare_nat))
                (f, n) d
            then (fun a -> ([], Fun (shp f, ts)) :: a) else id)
            (maps (fun (i, ti) ->
                    map (fun (p, a) -> (i :: p, a))
                      (dPos_impl (_A1, _A2) shp d ti))
              (zip (upt zero_nata n) ts)));;

let rec check_rule_dt (_A1, _A2, _A3, _A4) (_B1, _B2)
  shp d ds =
    (fun (a, b) ->
      (let (l, r) = a in
        (fun (dl, dr) ->
          catch_error
            (let sl = sharp_term shp l in
              bindb (check (equal_terma _A3 _B1 sl dl)
                      (comp (comp (comp (showsl_lit "wrong lhs, expected ")
                                    (showsl_terma _A4 _B2 sl))
                              (showsl_lit " but got "))
                        (showsl_terma _A4 _B2 dl)))
                (fun _ ->
                  (let pts = dPos_impl (_A1, _A2) shp d r in
                   let spts = map snd pts in
                   let (_, dts) = split_term (check_tup (_A1, _A2) ds) dr in
                    check (equal_multiseta (equal_term _A3 _B1)
                            (mset (equal_term _A3 _B1) dts)
                            (mset (equal_term _A3 _B1) spts))
                      (showsl_lit
                        "multiset of subterms with defined roots differs"))))
            (fun x ->
              Inl (comp (comp (comp (comp (comp
    (showsl_lit "could not ensure that ") (showsl_rule _A4 _B2 (dl, dr)))
                                      (showsl_lit
" is a valid dependency tuple for "))
                                (showsl_rule _A4 _B2 (l, r)))
                          (showsl_literal "\092"))
                    x))))
        b);;

let rec dt_transformation (_A1, _A2, _A3, _A4, _A5, _A6) (_C1, _C2, _C3, _C4)
  shp i info cm cc cp =
    catch_error
      (let DT_Transformation_Info (s_DT_s, w_DT_w, q) = info in
        (match cm
          with Derivational_Complexity _ ->
            Inl (showsl_lit
                  "only runtime complexity supported for dependency tuples")
          | Runtime_Complexity (c, d) ->
            (let s = rc i cp in
             let w = rwc i cp in
             let sa = map fst s_DT_s in
             let wa = map fst w_DT_w in
             let r = sa @ wa in
             let dd = defined_list r in
             let dda =
               set ((ceq_prod _A2 ceq_nat),
                     (ccompare_prod _A3.ccompare_cproper_interval ccompare_nat),
                     (set_impl_prod _A5 set_impl_nat))
                 d
               in
              bindb (catch_error
                      (forallM
                        (fun lr ->
                          check (list_ex
                                  (eq_rule_mod_vars _A4 (_C1, _C2, _C3) lr) sa)
                            (comp (showsl_lit
                                    "could not find DT for strict rule ")
                              (showsl_prod (showl_term _A6 _C4)
                                (showl_term _A6 _C4) lr)))
                        s)
                      (fun x -> Inl (snd x)))
                (fun _ ->
                  bindb (catch_error
                          (forallM
                            (fun lr ->
                              check (list_ex
                                      (eq_rule_mod_vars _A4 (_C1, _C2, _C3) lr)
                                      wa)
                                (comp (showsl_lit
"could not find DT for weak rule ")
                                  (showsl_prod (showl_term _A6 _C4)
                                    (showl_term _A6 _C4) lr)))
                            w)
                          (fun x -> Inl (snd x)))
                    (fun _ ->
                      bindb (catch_error
                              (forallM
                                (fun f ->
                                  check (member
  ((ceq_prod _A2 ceq_nat),
    (ccompare_prod _A3.ccompare_cproper_interval ccompare_nat))
  f dda)
                                    (comp (comp (showsl_lit "defined symbol ")
    (showsl_prod _A6 showl_nat f))
                                      (showsl_lit
" does not occur in defined symbols from RC")))
                                dd)
                              (fun x -> Inl (snd x)))
                        (fun _ ->
                          (let dTs = map snd s_DT_s in
                           let dTw = map snd w_DT_w in
                           let da =
                             set ((ceq_prod _A2 ceq_nat),
                                   (ccompare_prod _A3.ccompare_cproper_interval
                                     ccompare_nat),
                                   (set_impl_prod _A5 set_impl_nat))
                               dd
                             in
                           let shpf = (fun (f, a) -> (shp f, a)) in
                           let ds =
                             image ((ceq_prod _A2 ceq_nat),
                                     (ccompare_prod
                                       _A3.ccompare_cproper_interval
                                       ccompare_nat))
                               ((ceq_prod _A2 ceq_nat),
                                 (ccompare_prod _A3.ccompare_cproper_interval
                                   ccompare_nat),
                                 (set_impl_prod _A5 set_impl_nat))
                               shpf da
                             in
                           let ddd =
                             image ((ceq_prod _A2 ceq_nat),
                                     (ccompare_prod
                                       _A3.ccompare_cproper_interval
                                       ccompare_nat))
                               (_A2, _A3.ccompare_cproper_interval, _A5)
                               (comp shp fst) da
                             in
                           let f = funas_trs_list r @ c @ d in
                           let fs =
                             set ((ceq_prod _A2 ceq_nat),
                                   (ccompare_prod _A3.ccompare_cproper_interval
                                     ccompare_nat),
                                   (set_impl_prod _A5 set_impl_nat))
                               f
                             in
                            bindb (catch_error
                                    (forallM
                                      (fun qa ->
check (not (is_Var qa) &&
        not (member
              ((ceq_prod _A2 ceq_nat),
                (ccompare_prod _A3.ccompare_cproper_interval ccompare_nat))
              (the (root qa)) fs))
  (comp (comp (showsl_lit "new Q-term ") (showsl_terma _A6 _C4 qa))
    (showsl_lit " not allowed")))
                                      q)
                                    (fun x -> Inl (snd x)))
                              (fun _ ->
                                bindb (check_wf_trs _A6 (_C2, _C4) r)
                                  (fun _ ->
                                    bindb (catch_error
    (check_NF_terms_subset (is_QNFb i cp) (map fst r))
    (fun _ -> Inl (showsl_lit "innermost required")))
                                      (fun _ ->
bindb (catch_error
        (forallM
          (fun fa ->
            check (not (member
                         ((ceq_prod _A2 ceq_nat),
                           (ccompare_prod _A3.ccompare_cproper_interval
                             ccompare_nat))
                         fa ds))
              (comp (showsl_prod _A6 showl_nat fa)
                (showsl_lit " as sharped symbol is not fresh")))
          f)
        (fun x -> Inl (snd x)))
  (fun _ ->
    bindb (check
            (is_empty
              ((card_UNIV_prod _A1 card_UNIV_nat), (ceq_prod _A2 ceq_nat),
                (cproper_interval_prod _A3 cproper_interval_nat))
              (inf_seta
                ((ceq_prod _A2 ceq_nat),
                  (ccompare_prod _A3.ccompare_cproper_interval ccompare_nat))
                (set ((ceq_prod _A2 ceq_nat),
                       (ccompare_prod _A3.ccompare_cproper_interval
                         ccompare_nat),
                       (set_impl_prod _A5 set_impl_nat))
                  c)
                da))
            (showsl_lit
              "constructors of RC and defined symbols of TRSs are not disjoint"))
      (fun _ ->
        bindb (catch_error
                (forallM
                  (check_rule_dt (_A2, _A3.ccompare_cproper_interval, _A4, _A6)
                    (_C2, _C4) shp da ddd)
                  s_DT_s)
                (fun x -> Inl (snd x)))
          (fun _ ->
            bindb (catch_error
                    (forallM
                      (check_rule_dt
                        (_A2, _A3.ccompare_cproper_interval, _A4, _A6)
                        (_C2, _C4) shp da ddd)
                      w_DT_w)
                    (fun x -> Inl (snd x)))
              (fun _ ->
                Inr (Runtime_Complexity (c, map shpf d),
                      mkc i false (qb i cp @ q) dTs (r @ dTw))))))))))))))))
      (fun x ->
        Inl (comp (showsl_lit "error when switching to dependency tuples\092")
              x));;

let rec showsl_mctxt _A _B
  = function MHole -> showsl_lit "[]"
    | MVar x -> showsl _B x
    | MFun (f, cs) ->
        comp (showsl _A f)
          (showsl_list_gen id "" "(" ", " ")" (map (showsl_mctxt _A _B) cs));;

let rec ground_mctxt = function MVar uu -> false
                       | MHole -> true
                       | MFun (f, cs) -> list_all ground_mctxt cs;;

let rec funas_mctxt (_A1, _A2, _A3, _A4, _A5, _A6, _A7) (_B1, _B2)
  = function
    MFun (f, cs) ->
      sup_seta
        ((ceq_prod _A3 ceq_nat),
          (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat))
        (inserta
          ((ceq_prod _A3 ceq_nat),
            (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat))
          (f, size_list cs)
          (bot_set
            ((ceq_prod _A3 ceq_nat),
              (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat),
              (set_impl_prod _A7 set_impl_nat))))
        (sup_setb
          ((finite_UNIV_prod _A1 finite_UNIV_nat), (cenum_prod _A2 cenum_nat),
            (ceq_prod _A3 ceq_nat),
            (cproper_interval_prod _A4 cproper_interval_nat),
            (set_impl_prod _A7 set_impl_nat))
          (image ((ceq_mctxt _A6 _B2), (ccompare_mctxt _A5 _B1))
            ((ceq_set
               ((cenum_prod _A2 cenum_nat), (ceq_prod _A3 ceq_nat),
                 (cproper_interval_prod _A4
                   cproper_interval_nat).ccompare_cproper_interval)),
              (ccompare_set
                ((finite_UNIV_prod _A1 finite_UNIV_nat), (ceq_prod _A3 ceq_nat),
                  (cproper_interval_prod _A4 cproper_interval_nat),
                  (set_impl_prod _A7 set_impl_nat))),
              set_impl_set)
            (funas_mctxt (_A1, _A2, _A3, _A4, _A5, _A6, _A7) (_B1, _B2))
            (set ((ceq_mctxt _A6 _B2), (ccompare_mctxt _A5 _B1), set_impl_mctxt)
              cs)))
    | MVar v ->
        bot_set
          ((ceq_prod _A3 ceq_nat),
            (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat),
            (set_impl_prod _A7 set_impl_nat))
    | MHole ->
        bot_set
          ((ceq_prod _A3 ceq_nat),
            (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat),
            (set_impl_prod _A7 set_impl_nat));;

let rec is_compound_context (_A1, _A2, _A3, _A4, _A5, _A6, _A7) (_B1, _B2)
  cComp c =
    ground_mctxt c &&
      less_eq_set
        ((cenum_prod _A2 cenum_nat), (ceq_prod _A3 ceq_nat),
          (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat))
        (funas_mctxt (_A1, _A2, _A3, _A4, _A5, _A6, _A7) (_B1, _B2) c) cComp;;

let rec if_Fun_in_set (_A1, _A2)
  f = (fun t ->
        is_Var t ||
          member ((ceq_prod _A1 ceq_nat), (ccompare_prod _A2 ccompare_nat))
            (the (root t)) f);;

let rec uncap_till
  p x1 = match p, x1 with p, Var x -> (if p (Var x) then [Var x] else [])
    | p, Fun (f, ts) ->
        (if p (Fun (f, ts)) then [Fun (f, ts)] else maps (uncap_till p) ts);;

let rec check_rule_wdp (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8) (_B1, _B2, _B3)
  shp cComp =
    (fun (a, b) ->
      (let (l, r) = a in
        (fun (p, q) ->
          catch_error
            (let la = sharp_term shp l in
              bindb (check (equal_terma _A6 _B2 la p)
                      (comp (comp (comp (showsl_lit "wrong lhs, expected ")
                                    (showsl_terma _A8 _B3 la))
                              (showsl_lit " but got "))
                        (showsl_terma _A8 _B3 p)))
                (fun _ ->
                  (let us =
                     uncap_till
                       (if_Fun_in_set (_A3, _A4.ccompare_cproper_interval)
                         (uminus_set cComp))
                       r
                     in
                   let (c, usa) =
                     split_term
                       (if_Fun_in_set (_A3, _A4.ccompare_cproper_interval)
                         (uminus_set cComp))
                       q
                     in
                    bindb (check
                            (equal_lista (equal_term _A6 _B2)
                              (map (sharp_term shp) us) usa)
                            (showsl_lit
                              "lists of maximal subterms with defined root differ"))
                      (fun _ ->
                        check (is_compound_context
                                (_A1, _A2, _A3, _A4, _A5, _A6, _A7) (_B1, _B2)
                                cComp c)
                          (comp (comp (showsl_mctxt _A8 _B3 c)
                                  (showsl_lit
                                    " is not a proper compound context of "))
                            (showsl_terma _A8 _B3 q))))))
            (fun x ->
              Inl (comp (comp (comp (comp (comp
    (showsl_lit "could not ensure that ") (showsl_rule _A8 _B3 (p, q)))
                                      (showsl_lit
" is a valid weak dependency pair for "))
                                (showsl_rule _A8 _B3 (l, r)))
                          (showsl_literal "\092"))
                    x))))
        b);;

let rec check_wdp_trans (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8)
  (_C1, _C2, _C3, _C4, _C5)
  shp i info cm cc cp =
    catch_error
      (let WDP_Trans_Info (compa, s_wdps, w_wdps, q) = info in
        (match cm
          with Derivational_Complexity _ ->
            Inl (showsl_lit
                  "only runtime complexity supported for weak dependency pairs")
          | Runtime_Complexity (c, d) ->
            (let s = rc i cp in
             let w = rwc i cp in
             let sa = map fst s_wdps in
             let wa = map fst w_wdps in
             let r = sa @ wa in
             let fs = funas_trs_list r in
             let ds = defined_list r in
              bindb (catch_error
                      (forallM
                        (fun ra ->
                          check (list_ex
                                  (eq_rule_mod_vars _A6 (_C1, _C3, _C4) ra) sa)
                            (comp (showsl_lit
                                    "could not find weak dependency pair for strict rule ")
                              (showsl_prod (showl_term _A8 _C5)
                                (showl_term _A8 _C5) ra)))
                        s)
                      (fun x -> Inl (snd x)))
                (fun _ ->
                  bindb (catch_error
                          (forallM
                            (fun ra ->
                              check (list_ex
                                      (eq_rule_mod_vars _A6 (_C1, _C3, _C4) ra)
                                      wa)
                                (comp (showsl_lit
"could not find weak dependency pair for weak rule ")
                                  (showsl_prod (showl_term _A8 _C5)
                                    (showl_term _A8 _C5) ra)))
                            w)
                          (fun x -> Inl (snd x)))
                    (fun _ ->
                      (let wDP_S = map snd s_wdps in
                       let wDP_W = map snd w_wdps in
                       let shpf = (fun (f, a) -> (shp f, a)) in
                       let f = fs @ c @ d in
                       let f_sharps = map shpf f in
                       let _ =
                         image ((ceq_prod _A3 ceq_nat),
                                 (ccompare_prod _A4.ccompare_cproper_interval
                                   ccompare_nat))
                           ((ceq_prod _A3 ceq_nat),
                             (ccompare_prod _A4.ccompare_cproper_interval
                               ccompare_nat),
                             (set_impl_prod _A7 set_impl_nat))
                           (fun (fa, a) -> (shp fa, a))
                           (set ((ceq_prod _A3 ceq_nat),
                                  (ccompare_prod _A4.ccompare_cproper_interval
                                    ccompare_nat),
                                  (set_impl_prod _A7 set_impl_nat))
                             f)
                         in
                       let cComp =
                         sup_seta
                           ((ceq_prod _A3 ceq_nat),
                             (ccompare_prod _A4.ccompare_cproper_interval
                               ccompare_nat))
                           (minus_set
                             ((ceq_prod _A3 ceq_nat),
                               (ccompare_prod _A4.ccompare_cproper_interval
                                 ccompare_nat))
                             (minus_set
                               ((ceq_prod _A3 ceq_nat),
                                 (ccompare_prod _A4.ccompare_cproper_interval
                                   ccompare_nat))
                               (set ((ceq_prod _A3 ceq_nat),
                                      (ccompare_prod
_A4.ccompare_cproper_interval ccompare_nat),
                                      (set_impl_prod _A7 set_impl_nat))
                                 fs)
                               (set ((ceq_prod _A3 ceq_nat),
                                      (ccompare_prod
_A4.ccompare_cproper_interval ccompare_nat),
                                      (set_impl_prod _A7 set_impl_nat))
                                 (defined_list r)))
                             (set ((ceq_prod _A3 ceq_nat),
                                    (ccompare_prod _A4.ccompare_cproper_interval
                                      ccompare_nat),
                                    (set_impl_prod _A7 set_impl_nat))
                               d))
                           compa
                         in
                        bindb (catch_error
                                (forallM
                                  (fun qa ->
                                    check (not (is_Var qa) &&
    not (membera (equal_prod _A6 equal_nat) f (the (root qa))))
                                      (comp
(comp (showsl_lit "new Q-term ") (showsl_terma _A8 _C5 qa))
(showsl_lit " not allowed")))
                                  q)
                                (fun x -> Inl (snd x)))
                          (fun _ ->
                            bindb (check_wf_trs _A8 (_C3, _C5) r)
                              (fun _ ->
                                bindb (catch_error
(forallM
  (fun fa ->
    check (not (member
                 ((ceq_prod _A3 ceq_nat),
                   (ccompare_prod _A4.ccompare_cproper_interval ccompare_nat))
                 fa cComp))
      (comp (showsl_prod _A8 showl_nat fa)
        (showsl_lit " clashes with sharp symbols")))
  f_sharps)
(fun x -> Inl (snd x)))
                                  (fun _ ->
                                    bindb (catch_error
    (forallM
      (fun fa ->
        check (not (member
                     ((ceq_prod _A3 ceq_nat),
                       (ccompare_prod _A4.ccompare_cproper_interval
                         ccompare_nat))
                     fa cComp))
          (comp (showsl_prod _A8 showl_nat fa)
            (showsl_lit " clashes with defined symbols of RC")))
      d)
    (fun x -> Inl (snd x)))
                                      (fun _ ->
bindb (catch_error
        (forallM
          (fun fa ->
            check (not (member
                         ((ceq_prod _A3 ceq_nat),
                           (ccompare_prod _A4.ccompare_cproper_interval
                             ccompare_nat))
                         fa cComp))
              (comp (showsl_prod _A8 showl_nat fa)
                (showsl_lit " clashes with defined symbols")))
          ds)
        (fun x -> Inl (snd x)))
  (fun _ ->
    bindb (catch_error
            (forallM
              (check_rule_wdp (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8)
                (_C2, _C3, _C5) shp cComp)
              s_wdps)
            (fun x -> Inl (snd x)))
      (fun _ ->
        bindb (catch_error
                (forallM
                  (check_rule_wdp (_A1, _A2, _A3, _A4, _A5, _A6, _A7, _A8)
                    (_C2, _C3, _C5) shp cComp)
                  w_wdps)
                (fun x -> Inl (snd x)))
          (fun _ ->
            Inr (Runtime_Complexity (c, map shpf d),
                  mkc i (nfsb i cp) (qb i cp @ q) (wDP_S @ sa)
                    (wDP_W @ wa)))))))))))))))
      (fun x ->
        Inl (comp (comp (showsl_lit
                          "error when switching to weak dependency pairs")
                    (showsl_literal "\092"))
              x));;

let rec equal_boundstype x0 x1 = match x0, x1 with Roof, Match -> false
                           | Match, Roof -> false
                           | Match, Match -> true
                           | Roof, Roof -> true;;

let rec boundstype (Bounds_Info (x1, x2, x3, x4, x5)) = x1;;

let rec stackable_of_cm = function Derivational_Complexity f -> f
                          | Runtime_Complexity (c, d) -> c;;

let rec bounds_complexity_rel (_B1, _B2, _B3)
  (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8)
  (_D1, _D2, _D3, _D4, _D5, _D6, _D7, _D8)
  i info rdelete cm cc tp =
    catch_error
      (let r = rc i tp in
       let rw = rwc i tp in
       let r2 =
         ceta_list_diff
           (compare_order_prod
             (compare_order_term _B1.compare_compare_order
               _C5.compare_compare_order)
             (compare_order_term _B1.compare_compare_order
               _C5.compare_compare_order))
           r rdelete
         in
        bindb (catch_error
                (check_subseteq
                  (equal_prod (equal_term _B2 _C6) (equal_term _B2 _C6)) rdelete
                  (rulesd i tp))
                (fun x ->
                  Inl (comp (comp (showsl_lit "could not find rule ")
                              (showsl_rule _B3 _C8 x))
                        (showsl_lit " in current complexity problem"))))
          (fun _ ->
            bindb (check (less_eq_complexity_class (Comp_Poly one_nata) cc)
                    (showsl_lit "can only ensure linear complexity"))
              (fun _ ->
                bindb (check (equal_boundstype (boundstype info) Match)
                        (showsl_lit
                          "complexity analysis requires boundstype match"))
                  (fun _ ->
                    (let _ = rulesd i tp in
                      bindb (check_bounds_generic (_B1, _B2, _B3)
                              (_D1, _D2, _D3, _D4, _D5, _D6, _D7, _D8)
                              (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8) info
                              rdelete (rw @ r2) (stackable_of_cm cm)
                              (roots_of_cm cm))
                        (fun _ ->
                          Inr (mkc i (nfsb i tp) (qb i tp) r2
                                (list_union
                                  (equal_prod (equal_term _B2 _C6)
                                    (equal_term _B2 _C6))
                                  rw rdelete))))))))
      (fun x ->
        Inl (comp (comp (comp (comp (showsl_lit
                                      "problem in ensuring match-RT boundedness of\092")
                                (showsl_tp _B3 _C8 i tp))
                          (showsl_lit "\092with deletion of rules\092"))
                    (showsl_trs _B3 _C8 rdelete))
              x));;

let rec bounds_complexity (_B1, _B2, _B3)
  (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8)
  (_D1, _D2, _D3, _D4, _D5, _D6, _D7, _D8)
  i info cm cc tp =
    catch_error
      (bindb
        (check (less_eq_complexity_class (Comp_Poly one_nata) cc)
          (showsl_lit "can only ensure linear complexity"))
        (fun _ ->
          bindb (check (equal_boundstype (boundstype info) Match)
                  (showsl_lit "complexity analysis requires boundstype match"))
            (fun _ ->
              check_bounds_generic (_B1, _B2, _B3)
                (_D1, _D2, _D3, _D4, _D5, _D6, _D7, _D8)
                (_C1, _C2, _C3, _C4, _C5, _C6, _C7, _C8) info (rulesd i tp) []
                (stackable_of_cm cm) (roots_of_cm cm))))
      (fun x ->
        Inl (comp (comp (comp (showsl_lit
                                "problem in ensuring match boundedness of\092")
                          (showsl_tp _B3 _C8 i tp))
                    (showsl_literal "\092"))
              x));;

let rec check_assm (_B1, _B2) (_C1, _C2)
  i cpx_check x2 = match i, cpx_check, x2 with
    i, cpx_check, Complexity_assm_proof (cp, prf) ->
      cpx_check (mk_cpx (_B1, _B2) (_C1, _C2) i cp) prf
    | i, uu, SN_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for termination or non-termination assumptions in complexity proof")
    | i, uu, Finite_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for termination or non-termination assumptions in complexity proof")
    | i, uu, SN_FP_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for termination or non-termination assumptions in complexity proof")
    | i, uu, Not_SN_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for termination or non-termination assumptions in complexity proof")
    | i, uu, Infinite_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for termination or non-termination assumptions in complexity proof")
    | i, uu, Not_RelSN_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for termination or non-termination assumptions in complexity proof")
    | i, uu, Not_SN_FP_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for termination or non-termination assumptions in complexity proof")
    | i, uu, Unknown_assm_proof (v, va) ->
        Inl (showsl_lit
              "no support for termination or non-termination assumptions in complexity proof");;

let rec check_complexity_proof (_B1, _B2, _B3) (_C1, _C2, _C3)
  ia assms i x3 x4 = match ia, assms, i, x3, x4 with
    ia, assms, i, (tp, (cm, cc)),
      Rule_Shift_Complexity (redp, rdelete, ur_opt, prf)
      -> debug i "Rule Removal Complexity"
           (bindb
             (catch_error
               (smart_rule_shift_complexity
                 (card_UNIV_lab, cenum_lab, (ceq_lab _B2 _C2),
                   (cproper_interval_lab _B1.compare_compare_order
                     _C1.compare_compare_order),
                   (compare_order_lab _B1.compare_compare_order
                     _C1.compare_compare_order),
                   (equal_lab _B2 _C2), set_impl_lab, (showl_lab _B3 _C3))
                 ia (get_redtriple
                      ((ceq_lab _B2 _C2),
                        (ccompare_lab _B1.compare_compare_order
                          _C1.compare_compare_order),
                        (compare_order_lab _B1.compare_compare_order
                          _C1.compare_compare_order),
                        (equal_lab _B2 _C2), set_impl_lab, (showl_lab _B3 _C3))
                      redp)
                 rdelete ur_opt cm cc tp)
               (fun x ->
                 Inl (comp (comp (comp (comp i
 (showsl_lit ": error when applying the rule shifting technique on\092"))
                                   (showsl_tp (showl_lab _B3 _C3)
                                     (showl_list showl_char) ia tp))
                             (showsl_literal "\092"))
                       x)))
             (fun tpa ->
               catch_error
                 (check_complexity_proof (_B1, _B2, _B3) (_C1, _C2, _C3) ia
                   assms (add_index i one_nata) (tpa, (cm, cc)) prf)
                 (fun x ->
                   Inl (comp (comp i
                               (showsl_lit
                                 ": error below the rule shifting processor\092"))
                         x))))
    | ia, assms, i, (tp, (cm, cc)), RisEmpty_Complexity ->
        debug i "R is empty for complexity"
          (check (null (rc ia tp))
            (comp (comp i (showsl_lit ": R is not empty in\092"))
              (showsl_tp (showl_lab _B3 _C3) (showl_list showl_char) ia tp)))
    | ia, assms, i, (tp, (cm, cc)),
        Remove_Nonapplicable_Rules_Complexity (r, prf)
        -> debug i "Removing non-applicable rules"
             (let _ = rc ia tp in
               bindb (catch_error (check_non_applicable_rules (is_QNFb ia tp) r)
                       (fun x ->
                         Inl (comp (comp i
                                     (showsl_lit
                                       ": error when removing non-applicable rules\092"))
                               (comp (showsl_rule (showl_lab _B3 _C3)
                                       (showl_list showl_char) x)
                                 (showsl_lit " is applicable")))))
                 (fun _ ->
                   (let tpa = delete_R_Rwb ia tp r r in
                     catch_error
                       (check_complexity_proof (_B1, _B2, _B3) (_C1, _C2, _C3)
                         ia assms (add_index i one_nata) (tpa, (cm, cc)) prf)
                       (fun x ->
                         Inl (comp (comp i
                                     (showsl_lit
                                       ": error below the non-applicable rules removal\092"))
                               x)))))
    | ia, assms, i, (tp, (cm, cc)), Matchbounds_Complexity info ->
        debug i "Matchbounds"
          (catch_error
            (bounds_complexity
              ((compare_order_lab _B1.compare_compare_order
                 _C1.compare_compare_order),
                (equal_lab _B2 _C2), (showl_lab _B3 _C3))
              (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
                (cproper_interval_list ccompare_char),
                (compare_order_list (compare_order_char, equal_char)),
                (equal_list equal_char), set_impl_list, (showl_list showl_char))
              (card_UNIV_list, cenum_list, (ceq_list ceq_char),
                (ccompare_list ccompare_char),
                (compare_order_list (compare_order_char, equal_char)),
                (equal_list equal_char), set_impl_list, (showl_list showl_char))
              ia info cm cc tp)
            (fun x ->
              Inl (comp (comp i
                          (showsl_lit ": error when applying matchbounds\092"))
                    x)))
    | ia, assms, i, (tp, (cm, cc)), Matchbounds_Rel_Complexity (info, rdel, prf)
        -> debug i "Matchbounds-Rel"
             (bindb
               (catch_error
                 (bounds_complexity_rel
                   ((compare_order_lab _B1.compare_compare_order
                      _C1.compare_compare_order),
                     (equal_lab _B2 _C2), (showl_lab _B3 _C3))
                   (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
                     (cproper_interval_list ccompare_char),
                     (compare_order_list (compare_order_char, equal_char)),
                     (equal_list equal_char), set_impl_list,
                     (showl_list showl_char))
                   (card_UNIV_list, cenum_list, (ceq_list ceq_char),
                     (ccompare_list ccompare_char),
                     (compare_order_list (compare_order_char, equal_char)),
                     (equal_list equal_char), set_impl_list,
                     (showl_list showl_char))
                   ia info rdel cm cc tp)
                 (fun x ->
                   Inl (comp (comp i
                               (showsl_lit
                                 ": error when applying relative matchbounds\092"))
                         x)))
               (fun tpa ->
                 catch_error
                   (check_complexity_proof (_B1, _B2, _B3) (_C1, _C2, _C3) ia
                     assms (add_index i one_nata) (tpa, (cm, cc)) prf)
                   (fun x ->
                     Inl (comp (comp i
                                 (showsl_lit
                                   ": error below applying relative matchbounds\092"))
                           x))))
    | ia, assms, i, (tp, (cm, cc)), DT_Transformation (info, prf) ->
        debug i "DT Transformation"
          (bindb
            (catch_error
              (dt_transformation
                (card_UNIV_lab, (ceq_lab _B2 _C2),
                  (cproper_interval_lab _B1.compare_compare_order
                    _C1.compare_compare_order),
                  (equal_lab _B2 _C2), set_impl_lab, (showl_lab _B3 _C3))
                ((ccompare_list ccompare_char), (equal_list equal_char),
                  mapping_impl_list, (showl_list showl_char))
                (fun a -> Sharp a) ia info cm cc tp)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit ": error when applying the DT transformation on\092"))
                                  (showsl_tp (showl_lab _B3 _C3)
                                    (showl_list showl_char) ia tp))
                            (showsl_literal "\092"))
                      x)))
            (fun (cma, tpa) ->
              catch_error
                (check_complexity_proof (_B1, _B2, _B3) (_C1, _C2, _C3) ia assms
                  (add_index i one_nata) (tpa, (cma, cc)) prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the DT transformation\092"))
                        x))))
    | ia, assms, i, (tp, (cm, cc)), WDP_Transformation (info, prf) ->
        debug i "WDP Transformation"
          (bindb
            (catch_error
              (check_wdp_trans
                (finite_UNIV_lab, cenum_lab, (ceq_lab _B2 _C2),
                  (cproper_interval_lab _B1.compare_compare_order
                    _C1.compare_compare_order),
                  (compare_lab _B1.compare_compare_order
                    _C1.compare_compare_order),
                  (equal_lab _B2 _C2), set_impl_lab, (showl_lab _B3 _C3))
                ((ccompare_list ccompare_char), (compare_list compare_char),
                  (equal_list equal_char), mapping_impl_list,
                  (showl_list showl_char))
                (fun a -> Sharp a) ia info cm cc tp)
              (fun x ->
                Inl (comp (comp (comp (comp i
(showsl_lit ": error when applying the WDP transformation on\092"))
                                  (showsl_tp (showl_lab _B3 _C3)
                                    (showl_list showl_char) ia tp))
                            (showsl_literal "\092"))
                      x)))
            (fun (cma, tpa) ->
              catch_error
                (check_complexity_proof (_B1, _B2, _B3) (_C1, _C2, _C3) ia assms
                  (add_index i one_nata) (tpa, (cma, cc)) prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below the WDP transformation\092"))
                        x))))
    | ia, assms, i, (tp, (cm, cc)), Usable_Rules_Complexity (ur, prf) ->
        debug i "Usable Rules"
          (bindb
            (catch_error
              (usable_rules_complexity
                (cenum_lab, (ceq_lab _B2 _C2),
                  (ccompare_lab _B1.compare_compare_order
                    _C1.compare_compare_order),
                  (compare_order_lab _B1.compare_compare_order
                    _C1.compare_compare_order),
                  (equal_lab _B2 _C2), set_impl_lab, (showl_lab _B3 _C3))
                ia ur cm cc tp)
              (fun x ->
                Inl (comp (comp i
                            (showsl_lit
                              ": error when applying usable rules\092"))
                      x)))
            (fun tpa ->
              catch_error
                (check_complexity_proof (_B1, _B2, _B3) (_C1, _C2, _C3) ia assms
                  (add_index i one_nata) (tpa, (cm, cc)) prf)
                (fun x ->
                  Inl (comp (comp i
                              (showsl_lit
                                ": error below applying usable rules\092"))
                        x))))
    | ia, assms, i, (tp, (cm, cc)), Complexity_Assumption (a, ass) ->
        debug i "Complexity Assumption"
          (if assms
            then bindb (catch_error
                         (check_complexity_subsumes (_B1, _B2, _B3)
                           (_C1, _C2, _C3)
                           ((ccompare_list ccompare_char),
                             (compare_order_list
                               (compare_order_char, equal_char)),
                             (equal_list equal_char), mapping_impl_list,
                             (showl_list showl_char))
                           a (qb ia tp, (rc ia tp, (rwc ia tp, (cm, cc)))))
                         (fun x ->
                           Inl (comp (comp i
                                       (showsl_lit
 ": error in complexity assumption or unknown proof\092"))
                                 x)))
                   (fun _ ->
                     catch_error
                       (catch_error
                         (forallM_index
                           (fun asa j ->
                             check_assm (_B1, _B3) (_C1, _C3) ia
                               (check_complexity_proof (_B1, _B2, _B3)
                                 (_C1, _C2, _C3) ia assms (add_index i (suc j)))
                               asa)
                           ass)
                         (fun x -> Inl (snd x)))
                       (fun x ->
                         Inl (comp (comp i
                                     (showsl_lit
                                       ": error below unknown proof\092"))
                               x)))
            else Inl (showsl_lit
                       "the proof contains an assumption or unknown proof which have to be manually allowed"))
    | ia, assms, i, (tp, (cm, cc)), Split_Complexity (info, prf1, prf2) ->
        debug i "Split"
          (bindb
            (catch_error
              (split_proc_complexity ((equal_lab _B2 _C2), (showl_lab _B3 _C3))
                ((equal_list equal_char), (showl_list showl_char)) ia info tp)
              (fun x ->
                Inl (comp (comp i
                            (showsl_lit
                              ": error when applying split processor\092"))
                      x)))
            (fun (tp1, tp2) ->
              bindb (catch_error
                      (check_complexity_proof (_B1, _B2, _B3) (_C1, _C2, _C3) ia
                        assms (add_index i one_nata) (tp1, (cm, cc)) prf1)
                      (fun x ->
                        Inl (comp (comp i
                                    (showsl_lit
                                      ": error below applying split processor\092"))
                              x)))
                (fun _ ->
                  catch_error
                    (check_complexity_proof (_B1, _B2, _B3) (_C1, _C2, _C3) ia
                      assms
                      (add_index i (nat_of_integer (Big_int.big_int_of_int 2)))
                      (tp2, (cm, cc)) prf2)
                    (fun x ->
                      Inl (comp (comp i
                                  (showsl_lit
                                    ": error below applying split processor\092"))
                            x)))));;

let rec translate_complexity_claim _A
  r s start =
    (let f = funas_trs_list (r @ s) in
     let d = defined_list (r @ s) in
     let c = filtera (fun fa -> not (membera (equal_prod _A equal_nat) d fa)) f
       in
      (match start with Full -> Derivational_Complexity f
        | Constructor_Based -> Runtime_Complexity (c, d)));;

let rec check_unknown_proof (_B1, _B2, _B3, _B4)
  ia j a i u prf =
    check_unknown_proof_main (_B1, _B2, _B3, _B4) ia j a i u
      (unlab_to_split_unknown (_B1, _B3)
        ((compare_order_list (compare_order_nat, equal_nat)),
          (equal_list equal_nat))
        ((compare_order_list (compare_order_char, equal_char)),
          (equal_list equal_char))
        prf);;

let rec fp_strategy_to_fp_impl _A
  strat r =
    (match strat with Outermost -> o_to_fp_impl (map fst r)
      | Context_Sensitive a -> mu_to_fp_impl _A a
      | Forbidden_Patterns fp -> fp);;

let rec proof_to_string
  proof =
    (match proof with TRS_Termination_Proof _ -> "TRS termination proof"
      | Complexity_Proof _ -> "complexity proof"
      | DP_Termination_Proof _ -> "DP termination proof"
      | DP_Nontermination_Proof _ -> "DP nontermination proof"
      | TRS_Nontermination_Proof _ -> "TRS nontermination proof"
      | FP_Termination_Proof _ -> "FP TRS termination proof"
      | Relative_TRS_Nontermination_Proof _ ->
        "relative TRS nontermination proof"
      | TRS_Confluence_Proof _ -> "TRS confluence proof"
      | TRS_Non_Confluence_Proof _ -> "TRS nonconfluence proof"
      | Completion_Proof _ -> "completion proof"
      | Ordered_Completion_Proof _ -> "ordered completion proof"
      | Equational_Proof _ -> "equational proof"
      | Equational_Disproof _ -> "equational disproof"
      | Quasi_Reductive_Proof _ -> "quasi-reductive proof"
      | Conditional_CR_Proof _ -> "conditional confluence proof"
      | Conditional_Non_CR_Proof _ -> "conditional nonconfluence proof"
      | Tree_Automata_Closed_Proof _ -> "tree automata closed proof"
      | AC_Termination_Proof _ -> "AC termination proof"
      | LTS_Termination_Proof _ -> "LTS termination proof"
      | LTS_Safety_Proof _ -> "LTS safety proof"
      | Unknown_Proof _ -> "unknown proof"
      | Unknown_Disproof _ -> "unknown disproof");;

let rec input_to_string
  input =
    (match input with DP_input (_, _, _, _) -> "DP input"
      | Inn_TRS_input (_, _, _, _) -> "TRS input"
      | CPX_input (_, _, _, _, _) -> "Complexity input"
      | COMP_input (_, _) -> "Completion input"
      | OCOMP_input (_, _, _, _) -> "Ordered Completion input"
      | EQ_input (_, _) -> "Equational input"
      | FP_TRS_input (_, _) -> "FP-TRS input" | CTRS_input _ -> "CTRS input"
      | TA_input (_, _) -> "TA input" | AC_input (_, _, _) -> "AC-TRS input"
      | LTS_input _ -> "LTS input"
      | LTS_safety_input (_, _) -> "LTS safety input"
      | Unknown_input _ -> "Unknown input");;

let rec claim_to_string
  claim =
    (match claim with Yes -> "yes" | No -> "no" | Terminating -> "termination"
      | Upperbound _ -> "upperbound complexity"
      | Nonterminating -> "nontermination" | Confluent -> "confluence"
      | Nonconfluent -> "nonconfluence" | Completed -> "completion"
      | Ordered_Completed -> "ordered completion" | Anything -> "anything");;

let rec check_cert (_B1, _B2, _B3, _B4, _B5)
  i j k l a input claim proof =
    (let mismatch =
       showsl_lit
         ((((("Claiming " ^ claim_to_string claim) ^ " of ") ^
             input_to_string input) ^
            " by ") ^
           proof_to_string proof)
       in
      (match input
        with DP_input (m, p, q, r) ->
          (match proof with TRS_Termination_Proof _ -> Inl mismatch
            | Complexity_Proof _ -> Inl mismatch
            | DP_Termination_Proof prf ->
              (if equal_claim claim Terminating || equal_claim claim Anything
                then check_dp_termination_proof (_B1, _B2, _B4, _B5) i j a
                       (showsl_lit "1")
                       (mkd j default_nfs_dp m p []
                         (strategy_to_Q (equal_lab _B4 (equal_list equal_nat))
                           (equal_list equal_char) q r)
                         [] r)
                       prf
                else Inl mismatch)
            | DP_Nontermination_Proof prf ->
              (if equal_claim claim Nonterminating || equal_claim claim Anything
                then check_dp_nontermination_proof (_B1, _B3, _B4, _B5)
                       ((compare_order_list (compare_order_nat, equal_nat)),
                         (equal_list equal_nat), (showl_list showl_nat))
                       i j a (showsl_lit "1")
                       (mkd j default_nfs_nt_dp m p []
                         (strategy_to_Q (equal_lab _B4 (equal_list equal_nat))
                           (equal_list equal_char) q r)
                         [] r)
                       prf
                else Inl mismatch)
            | TRS_Nontermination_Proof _ -> Inl mismatch
            | FP_Termination_Proof _ -> Inl mismatch
            | Relative_TRS_Nontermination_Proof _ -> Inl mismatch
            | TRS_Confluence_Proof _ -> Inl mismatch
            | TRS_Non_Confluence_Proof _ -> Inl mismatch
            | Completion_Proof _ -> Inl mismatch
            | Ordered_Completion_Proof _ -> Inl mismatch
            | Equational_Proof _ -> Inl mismatch
            | Equational_Disproof _ -> Inl mismatch
            | Quasi_Reductive_Proof _ -> Inl mismatch
            | Conditional_CR_Proof _ -> Inl mismatch
            | Conditional_Non_CR_Proof _ -> Inl mismatch
            | Tree_Automata_Closed_Proof _ -> Inl mismatch
            | AC_Termination_Proof _ -> Inl mismatch
            | LTS_Termination_Proof _ -> Inl mismatch
            | LTS_Safety_Proof _ -> Inl mismatch
            | Unknown_Proof _ -> Inl mismatch
            | Unknown_Disproof _ -> Inl mismatch)
        | Inn_TRS_input (q, r, s, start) ->
          (match proof
            with TRS_Termination_Proof prf ->
              (if equal_claim claim Terminating || equal_claim claim Anything
                then check_trs_termination_proof (_B1, _B2, _B4, _B5) i j a
                       (showsl_lit "1")
                       (mkc i default_nfs_trs
                         (strategy_to_Q (equal_lab _B4 (equal_list equal_nat))
                           (equal_list equal_char) q r)
                         r s)
                       prf
                else Inl mismatch)
            | Complexity_Proof prf ->
              (match claim with Yes -> Inl mismatch | No -> Inl mismatch
                | Terminating -> Inl mismatch
                | Upperbound ub ->
                  (let cm =
                     translate_complexity_claim
                       (equal_lab _B4 (equal_list equal_nat)) r s start
                     in
                   let cc = Comp_Poly ub in
                    check_complexity_proof (_B1, _B4, _B5)
                      ((compare_order_list (compare_order_nat, equal_nat)),
                        (equal_list equal_nat), (showl_list showl_nat))
                      i a (showsl_lit "1")
                      (mkc i true
                         (strategy_to_Q (equal_lab _B4 (equal_list equal_nat))
                           (equal_list equal_char) q (r @ s))
                         r s,
                        (cm, cc))
                      prf)
                | Nonterminating -> Inl mismatch | Confluent -> Inl mismatch
                | Nonconfluent -> Inl mismatch | Completed -> Inl mismatch
                | Ordered_Completed -> Inl mismatch | Anything -> Inl mismatch)
            | DP_Termination_Proof _ -> Inl mismatch
            | DP_Nontermination_Proof _ -> Inl mismatch
            | TRS_Nontermination_Proof prf ->
              (if equal_claim claim Nonterminating || equal_claim claim Anything
                then bindb (check (equal_start_term start Full) mismatch)
                       (fun _ ->
                         bindb (check (null s) mismatch)
                           (fun _ ->
                             check_trs_nontermination_proof (_B1, _B3, _B4, _B5)
                               ((compare_order_list
                                  (compare_order_nat, equal_nat)),
                                 (equal_list equal_nat), (showl_list showl_nat))
                               i j a (showsl_lit "1")
                               (mkc i default_nfs_nt_trs
                                 (strategy_to_Q
                                   (equal_lab _B4 (equal_list equal_nat))
                                   (equal_list equal_char) q r)
                                 r [])
                               prf))
                else Inl mismatch)
            | FP_Termination_Proof _ -> Inl mismatch
            | Relative_TRS_Nontermination_Proof prf ->
              (if equal_claim claim Nonterminating || equal_claim claim Anything
                then bindb (check (equal_start_term start Full) mismatch)
                       (fun _ ->
                         check_reltrs_nontermination_proof (_B1, _B3, _B4, _B5)
                           ((compare_order_list (compare_order_nat, equal_nat)),
                             (equal_list equal_nat), (showl_list showl_nat))
                           i j a (showsl_lit "1")
                           (mkc i default_nfs_nt_trs
                             (strategy_to_Q
                               (equal_lab _B4 (equal_list equal_nat))
                               (equal_list equal_char) q r)
                             r s)
                           prf)
                else Inl mismatch)
            | TRS_Confluence_Proof prf ->
              (if equal_claim claim Confluent || equal_claim claim Anything
                then bindb (check (equal_start_term start Full)
                             (showsl_lit
                               "Confluence with start term not supported"))
                       (fun _ ->
                         bindb (check
                                 (equal_strategy
                                   (equal_lab _B4 (equal_list equal_nat))
                                   (equal_list equal_char) q No_Strategy)
                                 (showsl_lit
                                   "Confluence under strategy not supported"))
                           (fun _ ->
                             bindb (check (null s)
                                     (showsl_lit
                                       "Relative confluence not supported"))
                               (fun _ ->
                                 check_cr_proof (_B1, _B2, _B4, _B5) a
                                   (showsl_lit "1") i j r prf)))
                else Inl mismatch)
            | TRS_Non_Confluence_Proof prf ->
              (if equal_claim claim Nonconfluent || equal_claim claim Anything
                then bindb (check (equal_start_term start Full)
                             (showsl_lit
                               "Nonconfluence with start term not supported"))
                       (fun _ ->
                         bindb (check
                                 (equal_strategy
                                   (equal_lab _B4 (equal_list equal_nat))
                                   (equal_list equal_char) q No_Strategy)
                                 (showsl_lit
                                   "Confluence under strategy not supported"))
                           (fun _ ->
                             bindb (check (null s)
                                     (showsl_lit
                                       "Relative nonconfluence not supported"))
                               (fun _ ->
                                 check_ncr_proof (_B1, _B2, _B3, _B4, _B5) a
                                   (showsl_lit "1") i j r prf)))
                else Inl mismatch)
            | Completion_Proof _ -> Inl mismatch
            | Ordered_Completion_Proof _ -> Inl mismatch
            | Equational_Proof _ -> Inl mismatch
            | Equational_Disproof _ -> Inl mismatch
            | Quasi_Reductive_Proof _ -> Inl mismatch
            | Conditional_CR_Proof _ -> Inl mismatch
            | Conditional_Non_CR_Proof _ -> Inl mismatch
            | Tree_Automata_Closed_Proof _ -> Inl mismatch
            | AC_Termination_Proof _ -> Inl mismatch
            | LTS_Termination_Proof _ -> Inl mismatch
            | LTS_Safety_Proof _ -> Inl mismatch
            | Unknown_Proof _ -> Inl mismatch
            | Unknown_Disproof _ -> Inl mismatch)
        | CPX_input (q, r, s, cm, cc) ->
          (match proof with TRS_Termination_Proof _ -> Inl mismatch
            | Complexity_Proof prf ->
              bindb (match claim with Yes -> Inl mismatch | No -> Inl mismatch
                      | Terminating -> Inl mismatch
                      | Upperbound ub ->
                        check (equal_complexity_class cc (Comp_Poly ub))
                          (showsl_lit "complexity class mismatch")
                      | Nonterminating -> Inl mismatch
                      | Confluent -> Inl mismatch | Nonconfluent -> Inl mismatch
                      | Completed -> Inl mismatch
                      | Ordered_Completed -> Inl mismatch | Anything -> Inr ())
                (fun _ ->
                  check_complexity_proof (_B1, _B4, _B5)
                    ((compare_order_list (compare_order_nat, equal_nat)),
                      (equal_list equal_nat), (showl_list showl_nat))
                    i a (showsl_lit "1")
                    (mkc i true
                       (strategy_to_Q (equal_lab _B4 (equal_list equal_nat))
                         (equal_list equal_char) q (r @ s))
                       r s,
                      (cm, cc))
                    prf)
            | DP_Termination_Proof _ -> Inl mismatch
            | DP_Nontermination_Proof _ -> Inl mismatch
            | TRS_Nontermination_Proof _ -> Inl mismatch
            | FP_Termination_Proof _ -> Inl mismatch
            | Relative_TRS_Nontermination_Proof _ -> Inl mismatch
            | TRS_Confluence_Proof _ -> Inl mismatch
            | TRS_Non_Confluence_Proof _ -> Inl mismatch
            | Completion_Proof _ -> Inl mismatch
            | Ordered_Completion_Proof _ -> Inl mismatch
            | Equational_Proof _ -> Inl mismatch
            | Equational_Disproof _ -> Inl mismatch
            | Quasi_Reductive_Proof _ -> Inl mismatch
            | Conditional_CR_Proof _ -> Inl mismatch
            | Conditional_Non_CR_Proof _ -> Inl mismatch
            | Tree_Automata_Closed_Proof _ -> Inl mismatch
            | AC_Termination_Proof _ -> Inl mismatch
            | LTS_Termination_Proof _ -> Inl mismatch
            | LTS_Safety_Proof _ -> Inl mismatch
            | Unknown_Proof _ -> Inl mismatch
            | Unknown_Disproof _ -> Inl mismatch)
        | COMP_input (e, r) ->
          (match proof with TRS_Termination_Proof _ -> Inl mismatch
            | Complexity_Proof _ -> Inl mismatch
            | DP_Termination_Proof _ -> Inl mismatch
            | DP_Nontermination_Proof _ -> Inl mismatch
            | TRS_Nontermination_Proof _ -> Inl mismatch
            | FP_Termination_Proof _ -> Inl mismatch
            | Relative_TRS_Nontermination_Proof _ -> Inl mismatch
            | TRS_Confluence_Proof _ -> Inl mismatch
            | TRS_Non_Confluence_Proof _ -> Inl mismatch
            | Completion_Proof prf ->
              (if equal_claim claim Completed || equal_claim claim Anything
                then check_completion_proof (_B1, _B2, _B4, _B5) a
                       (showsl_lit "1") i j e r prf
                else Inl mismatch)
            | Ordered_Completion_Proof _ -> Inl mismatch
            | Equational_Proof _ -> Inl mismatch
            | Equational_Disproof _ -> Inl mismatch
            | Quasi_Reductive_Proof _ -> Inl mismatch
            | Conditional_CR_Proof _ -> Inl mismatch
            | Conditional_Non_CR_Proof _ -> Inl mismatch
            | Tree_Automata_Closed_Proof _ -> Inl mismatch
            | AC_Termination_Proof _ -> Inl mismatch
            | LTS_Termination_Proof _ -> Inl mismatch
            | LTS_Safety_Proof _ -> Inl mismatch
            | Unknown_Proof _ -> Inl mismatch
            | Unknown_Disproof _ -> Inl mismatch)
        | OCOMP_input (e_0, e, r, ord) ->
          (match proof with TRS_Termination_Proof _ -> Inl mismatch
            | Complexity_Proof _ -> Inl mismatch
            | DP_Termination_Proof _ -> Inl mismatch
            | DP_Nontermination_Proof _ -> Inl mismatch
            | TRS_Nontermination_Proof _ -> Inl mismatch
            | FP_Termination_Proof _ -> Inl mismatch
            | Relative_TRS_Nontermination_Proof _ -> Inl mismatch
            | TRS_Confluence_Proof _ -> Inl mismatch
            | TRS_Non_Confluence_Proof _ -> Inl mismatch
            | Completion_Proof _ -> Inl mismatch
            | Ordered_Completion_Proof prf ->
              (if equal_claim claim Ordered_Completed ||
                    equal_claim claim Anything
                then check_ordered_completion_proof (_B1, _B4, _B5) a
                       (showsl_lit "1") i j e_0 e r ord prf
                else Inl mismatch)
            | Equational_Proof _ -> Inl mismatch
            | Equational_Disproof _ -> Inl mismatch
            | Quasi_Reductive_Proof _ -> Inl mismatch
            | Conditional_CR_Proof _ -> Inl mismatch
            | Conditional_Non_CR_Proof _ -> Inl mismatch
            | Tree_Automata_Closed_Proof _ -> Inl mismatch
            | AC_Termination_Proof _ -> Inl mismatch
            | LTS_Termination_Proof _ -> Inl mismatch
            | LTS_Safety_Proof _ -> Inl mismatch
            | Unknown_Proof _ -> Inl mismatch
            | Unknown_Disproof _ -> Inl mismatch)
        | EQ_input (e, eq) ->
          (match proof with TRS_Termination_Proof _ -> Inl mismatch
            | Complexity_Proof _ -> Inl mismatch
            | DP_Termination_Proof _ -> Inl mismatch
            | DP_Nontermination_Proof _ -> Inl mismatch
            | TRS_Nontermination_Proof _ -> Inl mismatch
            | FP_Termination_Proof _ -> Inl mismatch
            | Relative_TRS_Nontermination_Proof _ -> Inl mismatch
            | TRS_Confluence_Proof _ -> Inl mismatch
            | TRS_Non_Confluence_Proof _ -> Inl mismatch
            | Completion_Proof _ -> Inl mismatch
            | Ordered_Completion_Proof _ -> Inl mismatch
            | Equational_Proof prf ->
              (if equal_claim claim Yes || equal_claim claim Anything
                then check_equational_proof (_B1, _B2, _B4, _B5) a
                       (showsl_lit "1") i j e eq prf
                else Inl mismatch)
            | Equational_Disproof prf ->
              (if equal_claim claim No || equal_claim claim Anything
                then check_equational_disproof (_B1, _B2, _B4, _B5) a
                       (showsl_lit "1") i j e eq prf
                else Inl mismatch)
            | Quasi_Reductive_Proof _ -> Inl mismatch
            | Conditional_CR_Proof _ -> Inl mismatch
            | Conditional_Non_CR_Proof _ -> Inl mismatch
            | Tree_Automata_Closed_Proof _ -> Inl mismatch
            | AC_Termination_Proof _ -> Inl mismatch
            | LTS_Termination_Proof _ -> Inl mismatch
            | LTS_Safety_Proof _ -> Inl mismatch
            | Unknown_Proof _ -> Inl mismatch
            | Unknown_Disproof _ -> Inl mismatch)
        | FP_TRS_input (strat, r) ->
          (match proof with TRS_Termination_Proof _ -> Inl mismatch
            | Complexity_Proof _ -> Inl mismatch
            | DP_Termination_Proof _ -> Inl mismatch
            | DP_Nontermination_Proof _ -> Inl mismatch
            | TRS_Nontermination_Proof prf ->
              (if equal_claim claim Nonterminating || equal_claim claim Anything
                then check_fp_nontermination_proof (_B1, _B3, _B4, _B5)
                       ((compare_order_list (compare_order_nat, equal_nat)),
                         (equal_list equal_nat), (showl_list showl_nat))
                       i j a (showsl_lit "1")
                       (fp_strategy_to_fp_impl
                          (equal_lab _B4 (equal_list equal_nat)) strat r,
                         r)
                       prf
                else Inl mismatch)
            | FP_Termination_Proof prf ->
              (if equal_claim claim Terminating || equal_claim claim Anything
                then check_fptrs_termination_proof (_B1, _B2, _B4, _B5) i j a
                       (showsl_lit "1")
                       (fp_strategy_to_fp_impl
                          (equal_lab _B4 (equal_list equal_nat)) strat r,
                         r)
                       prf
                else Inl mismatch)
            | Relative_TRS_Nontermination_Proof _ -> Inl mismatch
            | TRS_Confluence_Proof _ -> Inl mismatch
            | TRS_Non_Confluence_Proof _ -> Inl mismatch
            | Completion_Proof _ -> Inl mismatch
            | Ordered_Completion_Proof _ -> Inl mismatch
            | Equational_Proof _ -> Inl mismatch
            | Equational_Disproof _ -> Inl mismatch
            | Quasi_Reductive_Proof _ -> Inl mismatch
            | Conditional_CR_Proof _ -> Inl mismatch
            | Conditional_Non_CR_Proof _ -> Inl mismatch
            | Tree_Automata_Closed_Proof _ -> Inl mismatch
            | AC_Termination_Proof _ -> Inl mismatch
            | LTS_Termination_Proof _ -> Inl mismatch
            | LTS_Safety_Proof _ -> Inl mismatch
            | Unknown_Proof _ -> Inl mismatch
            | Unknown_Disproof _ -> Inl mismatch)
        | CTRS_input ctrs ->
          (match proof with TRS_Termination_Proof _ -> Inl mismatch
            | Complexity_Proof _ -> Inl mismatch
            | DP_Termination_Proof _ -> Inl mismatch
            | DP_Nontermination_Proof _ -> Inl mismatch
            | TRS_Nontermination_Proof _ -> Inl mismatch
            | FP_Termination_Proof _ -> Inl mismatch
            | Relative_TRS_Nontermination_Proof _ -> Inl mismatch
            | TRS_Confluence_Proof _ -> Inl mismatch
            | TRS_Non_Confluence_Proof _ -> Inl mismatch
            | Completion_Proof _ -> Inl mismatch
            | Ordered_Completion_Proof _ -> Inl mismatch
            | Equational_Proof _ -> Inl mismatch
            | Equational_Disproof _ -> Inl mismatch
            | Quasi_Reductive_Proof prf ->
              (if equal_claim claim Terminating || equal_claim claim Anything
                then check_quasi_reductive_proof (_B1, _B2, _B4, _B5) a
                       (showsl_lit "1") i j ctrs prf
                else Inl mismatch)
            | Conditional_CR_Proof prf ->
              (if equal_claim claim Confluent || equal_claim claim Anything
                then check_conditional_cr_proof (_B1, _B2, _B3, _B4, _B5) a
                       (showsl_lit "1") i j ctrs prf
                else Inl mismatch)
            | Conditional_Non_CR_Proof prf ->
              (if equal_claim claim Nonconfluent || equal_claim claim Anything
                then check_conditional_ncr_proof (_B1, _B2, _B3, _B4, _B5) a
                       (showsl_lit "1") i j ctrs prf
                else Inl mismatch)
            | Tree_Automata_Closed_Proof _ -> Inl mismatch
            | AC_Termination_Proof _ -> Inl mismatch
            | LTS_Termination_Proof _ -> Inl mismatch
            | LTS_Safety_Proof _ -> Inl mismatch
            | Unknown_Proof _ -> Inl mismatch
            | Unknown_Disproof _ -> Inl mismatch)
        | TA_input (ta, r) ->
          (match proof with TRS_Termination_Proof _ -> Inl mismatch
            | Complexity_Proof _ -> Inl mismatch
            | DP_Termination_Proof _ -> Inl mismatch
            | DP_Nontermination_Proof _ -> Inl mismatch
            | TRS_Nontermination_Proof _ -> Inl mismatch
            | FP_Termination_Proof _ -> Inl mismatch
            | Relative_TRS_Nontermination_Proof _ -> Inl mismatch
            | TRS_Confluence_Proof _ -> Inl mismatch
            | TRS_Non_Confluence_Proof _ -> Inl mismatch
            | Completion_Proof _ -> Inl mismatch
            | Ordered_Completion_Proof _ -> Inl mismatch
            | Equational_Proof _ -> Inl mismatch
            | Equational_Disproof _ -> Inl mismatch
            | Quasi_Reductive_Proof _ -> Inl mismatch
            | Conditional_CR_Proof _ -> Inl mismatch
            | Conditional_Non_CR_Proof _ -> Inl mismatch
            | Tree_Automata_Closed_Proof prf ->
              (if equal_claim claim Yes || equal_claim claim Anything
                then tree_aut_trs_closed
                       (card_UNIV_list, cenum_list, (ceq_list ceq_char),
                         (cproper_interval_list ccompare_char),
                         (compare_order_list (compare_order_char, equal_char)),
                         (equal_list equal_char), set_impl_list,
                         (showl_list showl_char))
                       ((ceq_lab _B4 (equal_list equal_nat)),
                         (ccompare_lab _B1.compare_compare_order
                           (compare_list compare_nat)),
                         (compare_order_lab _B1.compare_compare_order
                           (compare_list compare_nat)),
                         (default_lab _B3),
                         (equal_lab _B4 (equal_list equal_nat)), set_impl_lab,
                         (showl_lab _B5 (showl_list showl_nat)))
                       (finite_UNIV_list, cenum_list, (ceq_list ceq_char),
                         (cproper_interval_list ccompare_char),
                         (compare_order_list (compare_order_char, equal_char)),
                         (equal_list equal_char), set_impl_list,
                         (showl_list showl_char))
                       ta prf r
                else Inl mismatch)
            | AC_Termination_Proof _ -> Inl mismatch
            | LTS_Termination_Proof _ -> Inl mismatch
            | LTS_Safety_Proof _ -> Inl mismatch
            | Unknown_Proof _ -> Inl mismatch
            | Unknown_Disproof _ -> Inl mismatch)
        | AC_input (r, aa, c) ->
          (match proof with TRS_Termination_Proof _ -> Inl mismatch
            | Complexity_Proof _ -> Inl mismatch
            | DP_Termination_Proof _ -> Inl mismatch
            | DP_Nontermination_Proof _ -> Inl mismatch
            | TRS_Nontermination_Proof _ -> Inl mismatch
            | FP_Termination_Proof _ -> Inl mismatch
            | Relative_TRS_Nontermination_Proof _ -> Inl mismatch
            | TRS_Confluence_Proof _ -> Inl mismatch
            | TRS_Non_Confluence_Proof _ -> Inl mismatch
            | Completion_Proof _ -> Inl mismatch
            | Ordered_Completion_Proof _ -> Inl mismatch
            | Equational_Proof _ -> Inl mismatch
            | Equational_Disproof _ -> Inl mismatch
            | Quasi_Reductive_Proof _ -> Inl mismatch
            | Conditional_CR_Proof _ -> Inl mismatch
            | Conditional_Non_CR_Proof _ -> Inl mismatch
            | Tree_Automata_Closed_Proof _ -> Inl mismatch
            | AC_Termination_Proof prf ->
              (if equal_claim claim Terminating || equal_claim claim Anything
                then check_ac_termination_proof (_B1, _B4, _B5)
                       ((compare_order_list (compare_order_nat, equal_nat)),
                         (equal_list equal_nat), (showl_list showl_nat))
                       l k (showsl_lit "1") (mke k r aa c) prf
                else Inl mismatch)
            | LTS_Termination_Proof _ -> Inl mismatch
            | LTS_Safety_Proof _ -> Inl mismatch
            | Unknown_Proof _ -> Inl mismatch
            | Unknown_Disproof _ -> Inl mismatch)
        | LTS_input r ->
          (match proof with TRS_Termination_Proof _ -> Inl mismatch
            | Complexity_Proof _ -> Inl mismatch
            | DP_Termination_Proof _ -> Inl mismatch
            | DP_Nontermination_Proof _ -> Inl mismatch
            | TRS_Nontermination_Proof _ -> Inl mismatch
            | FP_Termination_Proof _ -> Inl mismatch
            | Relative_TRS_Nontermination_Proof _ -> Inl mismatch
            | TRS_Confluence_Proof _ -> Inl mismatch
            | TRS_Non_Confluence_Proof _ -> Inl mismatch
            | Completion_Proof _ -> Inl mismatch
            | Ordered_Completion_Proof _ -> Inl mismatch
            | Equational_Proof _ -> Inl mismatch
            | Equational_Disproof _ -> Inl mismatch
            | Quasi_Reductive_Proof _ -> Inl mismatch
            | Conditional_CR_Proof _ -> Inl mismatch
            | Conditional_Non_CR_Proof _ -> Inl mismatch
            | Tree_Automata_Closed_Proof _ -> Inl mismatch
            | AC_Termination_Proof _ -> Inl mismatch
            | LTS_Termination_Proof prf ->
              (if equal_claim claim Terminating || equal_claim claim Anything
                then check_termination
                       ((ccompare_list ccompare_char),
                         (compare_list compare_char), (equal_list equal_char),
                         mapping_impl_list,
                         (linorder_list (equal_char, linorder_char)),
                         (showl_list showl_char))
                       (cenum_list, (ceq_list ceq_char),
                         (ccompare_list ccompare_char),
                         (compare_order_list (compare_order_char, equal_char)),
                         (equal_list equal_char), mapping_impl_list,
                         set_impl_list, (showl_list showl_char))
                       ((ceq_list ceq_char), (ccompare_list ccompare_char),
                         (equal_list equal_char), mapping_impl_list,
                         set_impl_list, (showl_list showl_char))
                       r prf
                else Inl mismatch)
            | LTS_Safety_Proof _ -> Inl mismatch
            | Unknown_Proof _ -> Inl mismatch
            | Unknown_Disproof _ -> Inl mismatch)
        | LTS_safety_input (r, e) ->
          (match proof with TRS_Termination_Proof _ -> Inl mismatch
            | Complexity_Proof _ -> Inl mismatch
            | DP_Termination_Proof _ -> Inl mismatch
            | DP_Nontermination_Proof _ -> Inl mismatch
            | TRS_Nontermination_Proof _ -> Inl mismatch
            | FP_Termination_Proof _ -> Inl mismatch
            | Relative_TRS_Nontermination_Proof _ -> Inl mismatch
            | TRS_Confluence_Proof _ -> Inl mismatch
            | TRS_Non_Confluence_Proof _ -> Inl mismatch
            | Completion_Proof _ -> Inl mismatch
            | Ordered_Completion_Proof _ -> Inl mismatch
            | Equational_Proof _ -> Inl mismatch
            | Equational_Disproof _ -> Inl mismatch
            | Quasi_Reductive_Proof _ -> Inl mismatch
            | Conditional_CR_Proof _ -> Inl mismatch
            | Conditional_Non_CR_Proof _ -> Inl mismatch
            | Tree_Automata_Closed_Proof _ -> Inl mismatch
            | AC_Termination_Proof _ -> Inl mismatch
            | LTS_Termination_Proof _ -> Inl mismatch
            | LTS_Safety_Proof prf ->
              (if equal_claim claim Yes || equal_claim claim Anything
                then check_safety (ccompare_sig, equal_sig, showl_sig)
                       (ceq_ty, ccompare_ty, equal_ty, showl_ty)
                       (default_hints, showl_hints)
                       ((ccompare_list ccompare_char), (equal_list equal_char),
                         (showl_list showl_char))
                       (cenum_list, (ceq_list ceq_char),
                         (ccompare_list ccompare_char), (equal_list equal_char),
                         mapping_impl_list, set_impl_list,
                         (showl_list showl_char))
                       ((ceq_list ceq_char), (ccompare_list ccompare_char),
                         (equal_list equal_char), mapping_impl_list,
                         set_impl_list, (showl_list showl_char))
                       ((ccompare_list ccompare_char), (equal_list equal_char),
                         mapping_impl_list, (showl_list showl_char))
                       type_of_fun
                       (inserta (ceq_ty, ccompare_ty) BoolT
                         (set_empty (ceq_ty, ccompare_ty)
                           (of_phantom set_impl_ty)))
                       (check_clause
                         ((ccompare_list ccompare_char),
                           (equal_list equal_char), mapping_impl_list,
                           (linorder_list (equal_char, linorder_char)),
                           (showl_list showl_char)))
                       (check_clause
                         ((ccompare_trans_var (ccompare_list ccompare_char)),
                           (equal_trans_var (equal_list equal_char)),
                           mapping_impl_trans_var,
                           (linorder_trans_var (compare_list compare_char)),
                           (showl_trans_var (showl_list showl_char))))
                       (showsl_IA_exp
                         ((equal_list equal_char),
                           (linorder_list (equal_char, linorder_char)),
                           (showl_list showl_char)))
                       (showsl_IA_exp
                         ((equal_trans_var (equal_list equal_char)),
                           (linorder_trans_var (compare_list compare_char)),
                           (showl_trans_var (showl_list showl_char))))
                       negate negate r e prf
                else Inl mismatch)
            | Unknown_Proof _ -> Inl mismatch
            | Unknown_Disproof _ -> Inl mismatch)
        | Unknown_input ur ->
          (match proof with TRS_Termination_Proof _ -> Inl mismatch
            | Complexity_Proof _ -> Inl mismatch
            | DP_Termination_Proof _ -> Inl mismatch
            | DP_Nontermination_Proof _ -> Inl mismatch
            | TRS_Nontermination_Proof _ -> Inl mismatch
            | FP_Termination_Proof _ -> Inl mismatch
            | Relative_TRS_Nontermination_Proof _ -> Inl mismatch
            | TRS_Confluence_Proof _ -> Inl mismatch
            | TRS_Non_Confluence_Proof _ -> Inl mismatch
            | Completion_Proof _ -> Inl mismatch
            | Ordered_Completion_Proof _ -> Inl mismatch
            | Equational_Proof _ -> Inl mismatch
            | Equational_Disproof _ -> Inl mismatch
            | Quasi_Reductive_Proof _ -> Inl mismatch
            | Conditional_CR_Proof _ -> Inl mismatch
            | Conditional_Non_CR_Proof _ -> Inl mismatch
            | Tree_Automata_Closed_Proof _ -> Inl mismatch
            | AC_Termination_Proof _ -> Inl mismatch
            | LTS_Termination_Proof _ -> Inl mismatch
            | LTS_Safety_Proof _ -> Inl mismatch
            | Unknown_Proof b ->
              check_unknown_proof (_B1, _B2, _B4, _B5) i j a (showsl_lit "1") ur
                b
            | Unknown_Disproof b ->
              check_unknown_disproof (_B1, _B3, _B4, _B5)
                ((compare_order_list (compare_order_nat, equal_nat)),
                  (equal_list equal_nat), (showl_list showl_nat))
                i j a (showsl_lit "1") ur b)));;

let rec intersect_values _A _B
  key vs m = foldr (aux _A _B key m) vs (empty _B.linorder_compare_order);;

let rec intersect_rules (_A1, _A2) _B
  rs = intersect_values
         (equal_prod equal_bool
           (equal_prod (equal_term _A2 _B) (equal_term _A2 _B)))
         (compare_order_prod _A1 compare_order_nat) key
         (map (fun a -> (true, a)) rs @ map (fun a -> (false, a)) rs);;

let rec impl_ofe _B (AC_DPP x) = x;;

let rec rw_implb (p, (pw, (r, (rw, (e, uu))))) = rw;;

let rec rwb _A xa = rw_implb (impl_ofe _A xa);;

let rec r_implb (p, (pw, (r, (rw, (e, uu))))) = r;;

let rec rb _A xa = r_implb (impl_ofe _A xa);;

let rec e_impl (p, (pw, (r, (rw, (e, uu))))) = e;;

let rec e _A xa = e_impl (impl_ofe _A xa);;

let rec eq_rules_non_collapsing _A
  d = (let t = list_all (fun (_, r) -> not (is_Var r)) in
        t (rb _A d) && (t (rwb _A d) && t (e _A d)));;

let rec eq_rules_no_left_var _A
  d = (let t = list_all (fun (l, _) -> not (is_Var l)) in
        t (rb _A d) && (t (rwb _A d) && t (e _A d)));;

let rec mk_implb (_A1, _A2) _B
  p pw r rw e =
    (p, (pw, (r, (rw, (e, insert_rules equal_unit (_A1, _A2) _B () (r @ rw @ e)
                            (empty
                              (linorder_prod _A1.linorder_compare_order
                                linorder_nat)))))));;

let rec delete_pairs_rules_impl (_A1, _A2) _B
  (p, (pw, (r, (rw, (e, m))))) pd rd =
    (if null rd
      then (list_diff (equal_prod (equal_term _A2 _B) (equal_term _A2 _B)) p pd,
             (list_diff (equal_prod (equal_term _A2 _B) (equal_term _A2 _B)) pw
                pd,
               (r, (rw, (e, m)))))
      else mk_implb (_A1, _A2) _B
             (list_diff (equal_prod (equal_term _A2 _B) (equal_term _A2 _B)) p
               pd)
             (list_diff (equal_prod (equal_term _A2 _B) (equal_term _A2 _B)) pw
               pd)
             (list_diff (equal_prod (equal_term _A2 _B) (equal_term _A2 _B)) r
               rd)
             (list_diff (equal_prod (equal_term _A2 _B) (equal_term _A2 _B)) rw
               rd)
             e);;

let rec delete_pairs_rules (_A1, _A2) _B
  xc xd xe =
    AC_DPP (delete_pairs_rules_impl (_A1, _A2) _B (impl_ofe _A1 xc) xd xe);;

let rec reverse_rules_mapa (_A1, _A2)
  d fn =
    maps (fun (l, r) ->
           (if equal_optiona (equal_prod _A2 equal_nat) (root r) (Some fn)
             then [(r, l)] else []))
      (rb _A1 d @ rwb _A1 d @ e _A1 d);;

let rec intersect_pairs_impla (_A1, _A2) _B
  (p, (pw, (r, (rw, (e, m))))) pd =
    (list_inter (equal_prod (equal_term _A2 _B) (equal_term _A2 _B)) p pd,
      (list_inter (equal_prod (equal_term _A2 _B) (equal_term _A2 _B)) pw pd,
        (r, (rw, (e, m)))));;

let rec intersect_pairsa (_A1, _A2) _B
  xb xc = AC_DPP (intersect_pairs_impla (_A1, _A2) _B (impl_ofe _A1 xb) xc);;

let rec rules_map_implb _A
  (p, (pw, (r, (rw, (e, m))))) fn =
    (match lookup (compare_order_prod _A compare_order_nat) m fn with None -> []
      | Some a -> map snd a);;

let rec rules_mapb _A xa = rules_map_implb _A (impl_ofe _A xa);;

let rec ac_dpp_impl (_A1, _A2) (_B1, _B2)
  (p, (pw, (r, (rw, (e, uu))))) =
    (set ((ceq_prod (ceq_term _A2 _B2) (ceq_term _A2 _B2)),
           (ccompare_prod (ccompare_term _A1.compare_compare_order _B1)
             (ccompare_term _A1.compare_compare_order _B1)),
           (set_impl_prod set_impl_term set_impl_term))
       p,
      (set ((ceq_prod (ceq_term _A2 _B2) (ceq_term _A2 _B2)),
             (ccompare_prod (ccompare_term _A1.compare_compare_order _B1)
               (ccompare_term _A1.compare_compare_order _B1)),
             (set_impl_prod set_impl_term set_impl_term))
         pw,
        (set ((ceq_prod (ceq_term _A2 _B2) (ceq_term _A2 _B2)),
               (ccompare_prod (ccompare_term _A1.compare_compare_order _B1)
                 (ccompare_term _A1.compare_compare_order _B1)),
               (set_impl_prod set_impl_term set_impl_term))
           r,
          (set ((ceq_prod (ceq_term _A2 _B2) (ceq_term _A2 _B2)),
                 (ccompare_prod (ccompare_term _A1.compare_compare_order _B1)
                   (ccompare_term _A1.compare_compare_order _B1)),
                 (set_impl_prod set_impl_term set_impl_term))
             rw,
            set ((ceq_prod (ceq_term _A2 _B2) (ceq_term _A2 _B2)),
                  (ccompare_prod (ccompare_term _A1.compare_compare_order _B1)
                    (ccompare_term _A1.compare_compare_order _B1)),
                  (set_impl_prod set_impl_term set_impl_term))
              e))));;

let rec ac_dpp (_A1, _A2) (_B1, _B2)
  xa = ac_dpp_impl (_A1, _A2) (_B1, _B2) (impl_ofe _A1 xa);;

let rec rules_implc (p, (pw, (r, (rw, (e, uu))))) = r @ rw;;

let rec rulesc _A xa = rules_implc (impl_ofe _A xa);;

let rec pairs_impla (p, (pw, (r, (rw, (e, uu))))) = p @ pw;;

let rec pairsa _A xa = pairs_impla (impl_ofe _A xa);;

let rec mkb (_A1, _A2) _B
  xc xd xe xf xg = AC_DPP (mk_implb (_A1, _A2) _B xc xd xe xf xg);;

let rec pw_impla (p, (pw, (r, (rw, (e, uu))))) = pw;;

let rec pwa _A xa = pw_impla (impl_ofe _A xa);;

let rec p_impla (p, (pw, (r, (rw, (e, uu))))) = p;;

let rec pa _A xa = p_impla (impl_ofe _A xa);;

let rec ac_dpp_rbt_impl (_A1, _A2) (_B1, _B2)
  = Ac_dpp_ops_ext
      (ac_dpp (_A1, _A2) (_B1, _B2), pa _A1, pwa _A1, pairsa _A1, rb _A1,
        rwb _A1, rulesc _A1, e _A1, mkb (_A1, _A2) _B2, rules_mapb _A1,
        reverse_rules_mapa (_A1, _A2), delete_pairs_rules (_A1, _A2) _B2,
        eq_rules_no_left_var _A1, eq_rules_non_collapsing _A1,
        intersect_pairsa (_A1, _A2) _B2, ());;

let rec c_rules_impl
  c = (let x = Var [Chara (false, false, false, true, true, true, true, false)]
         in
       let y = Var [Chara (true, false, false, true, true, true, true, false)]
         in
        map (fun f -> (Fun (f, [x; y]), Fun (f, [y; x]))) c);;

let rec a_trs_impl
  a = (let x = Var [Chara (false, false, false, true, true, true, true, false)]
         in
       let y = Var [Chara (true, false, false, true, true, true, true, false)]
         in
       let z = Var [Chara (false, true, false, true, true, true, true, false)]
         in
        map (fun f ->
              (Fun (f, [Fun (f, [x; y]); z]), Fun (f, [x; Fun (f, [y; z])])))
          a @
          map (fun f ->
                (Fun (f, [x; Fun (f, [y; z])]), Fun (f, [Fun (f, [x; y]); z])))
            a);;

let rec aC_trs_impl a c = a_trs_impl a @ c_rules_impl c;;

let rec ac_tp_list_impl (_A1, _A2, _A3, _A4, _A5)
  = Ac_tp_ops_ext
      ((fun (r, a) ->
         (let (aa, c) = a in
           (set ((ceq_prod (ceq_term _A4 (equal_list equal_char))
                   (ceq_term _A4 (equal_list equal_char))),
                  (ccompare_prod (ccompare_term _A3 (compare_list compare_char))
                    (ccompare_term _A3 (compare_list compare_char))),
                  (set_impl_prod set_impl_term set_impl_term))
              r,
             (set (_A1, _A2, _A5) aa, set (_A1, _A2, _A5) c)))),
        (fun (r, (_, _)) -> r), (fun (_, a) -> (let (aa, _) = a in aa)),
        (fun (_, (_, c)) -> c), (fun r a c -> (r, (a, c))),
        (fun (r, a) ->
          (let (aa, c) = a in
            (fun dr ->
              (list_diff
                 (equal_prod (equal_term _A4 (equal_list equal_char))
                   (equal_term _A4 (equal_list equal_char)))
                 r dr,
                (aa, c))))),
        (fun (_, a) -> (let (aa, b) = a in aC_trs_impl aa b)), ());;

let rec rules_non_collapsing_impl
  (uu, (uv, (uw, (ux, (uy, (uz, (nc, va))))))) = nc;;

let rec impl_ofd _B (DPP x) = x;;

let rec rules_non_collapsing _A d = rules_non_collapsing_impl (impl_ofd _A d);;

let rec nFQ_subset_NF_rules_impla (uu, (uv, (uw, (ux, (uy, (b, uz)))))) = b;;

let rec nFQ_subset_NF_rulesa _A d = nFQ_subset_NF_rules_impla (impl_ofd _A d);;

let rec rules_no_left_var_impl
  = function (uu, (uv, (uw, (ux, (uy, (uz, (va, ([], ([], vb))))))))) -> true
    | (v, (vb, (ve, (vg, (vi, (vk, (vm, (vq :: vr, vp)))))))) -> false
    | (v, (vb, (ve, (vg, (vi, (vk, (vm, (vo, (vs :: vt, vr))))))))) -> false;;

let rec rules_no_left_var _A d = rules_no_left_var_impl (impl_ofd _A d);;

let rec reverse_rules_map_impl _A
  (uu, (uv, (uw, (ux, (uy, (uz, (va, (vb, (vc, (vd, (m, ve))))))))))) fn =
    (match lookup (compare_order_prod _A compare_order_nat) m fn with None -> []
      | Some a -> map snd a);;

let rec reverse_rules_map _A d = reverse_rules_map_impl _A (impl_ofd _A d);;

let rec is_NF_trs_subset is_Q_nf r = is_NF_subset is_Q_nf (map fst r);;

let rec wwf_qtrs_impl _B
  nf r =
    list_all (fun ra -> wf_rule _B ra || not (applicable_rule_impl nf ra)) r;;

let rec intersect_rules_impl (_A1, _A2) _B
  d ri =
    (let (nfs, (mi, (p, (pw, (q, (nfq, (nc,
 (vR, (vRw, (m, (rm, (wwf, isnf))))))))))))
       = d in
     let (vri, ria) = partition (comp is_Var fst) ri in
     let vr =
       list_inter (equal_prod (equal_term _A2 _B) (equal_term _A2 _B)) vR vri in
     let vrw =
       list_inter (equal_prod (equal_term _A2 _B) (equal_term _A2 _B)) vRw vri
       in
     let ma = intersect_rules (_A1, _A2) _B ria m in
     let rma = intersect_rules (_A1, _A2) _B (reverse_rules ri) rm in
     let rs =
       vr @ vrw @ map snd
                    (values
                      (linorder_prod _A1.linorder_compare_order linorder_nat)
                      ma)
       in
      (nfs, (mi, (p, (pw, (q, (nfq || is_NF_trs_subset isnf rs,
                                (nc || list_all (fun r -> not (is_Var (snd r)))
 rs,
                                  (vr, (vrw,
 (ma, (rma, (wwf || wwf_qtrs_impl _B isnf rs, isnf)))))))))))));;

let rec intersect_rulesa (_A1, _A2) _B
  d rs = DPP (intersect_rules_impl (_A1, _A2) _B (impl_ofd _A1 d) rs);;

let rec intersect_pairs_impl _A _B
  (nfs, (mi, (p, (pw, rest)))) ps =
    (nfs, (mi, (list_inter (equal_prod (equal_term _A _B) (equal_term _A _B)) p
                  ps,
                 (list_inter (equal_prod (equal_term _A _B) (equal_term _A _B))
                    pw ps,
                   rest))));;

let rec intersect_pairs (_A1, _A2) _B
  d ps = DPP (intersect_pairs_impl _A2 _B (impl_ofd _A1 d) ps);;

let rec replace_pair_impl _A _B
  (nfs, (mi, (p, (pw, rest)))) pair ps =
    (nfs, (mi, (replace_impl (equal_prod (equal_term _A _B) (equal_term _A _B))
                  pair ps p,
                 (replace_impl
                    (equal_prod (equal_term _A _B) (equal_term _A _B)) pair ps
                    pw,
                   rest))));;

let rec replace_pair (_A1, _A2) _B
  d pair ps = DPP (replace_pair_impl _A2 _B (impl_ofd _A1 d) pair ps);;

let rec rules_implb _A
  (uu, (uv, (uw, (ux, (uy, (uz, (va, (vr, (vrw, (m, vb)))))))))) =
    vr @ vrw @ map snd
                 (values (linorder_prod _A.linorder_compare_order linorder_nat)
                   m);;

let rec split_rules_impla (_A1, _A2) _B
  d rs =
    partition (membera (equal_prod (equal_term _A2 _B) (equal_term _A2 _B)) rs)
      (rules_implb _A1 d);;

let rec split_rulesa (_A1, _A2) _B
  d = split_rules_impla (_A1, _A2) _B (impl_ofd _A1 d);;

let rec pairs_impl (uu, (uv, (p, (pw, uw)))) = p @ pw;;

let rec split_pairs_impl _A _B
  d ps =
    partition (membera (equal_prod (equal_term _A _B) (equal_term _A _B)) ps)
      (pairs_impl d);;

let rec split_pairs (_A1, _A2) _B d = split_pairs_impl _A2 _B (impl_ofd _A1 d);;

let rec delete_R_Rw_impla (_A1, _A2) _B
  d r rw =
    (let (nfs, (mi, (p, (pw, (q, (nfq, (nc,
 (vR, (vRw, (m, (rm, (wwf, isnf))))))))))))
       = d in
     let (vr, ra) = partition (comp is_Var fst) r in
     let (vrw, rwa) = partition (comp is_Var fst) rw in
     let vra =
       list_diff (equal_prod (equal_term _A2 _B) (equal_term _A2 _B)) vR vr in
     let vrwa =
       list_diff (equal_prod (equal_term _A2 _B) (equal_term _A2 _B)) vRw vrw in
     let ma =
       delete_rules equal_bool (_A1, _A2) _B true ra
         (delete_rules equal_bool (_A1, _A2) _B false rwa m)
       in
     let rma =
       delete_rules equal_bool (_A1, _A2) _B true (reverse_rules r)
         (delete_rules equal_bool (_A1, _A2) _B false (reverse_rules rw) rm)
       in
     let rs =
       vra @ vrwa @
               map snd
                 (values (linorder_prod _A1.linorder_compare_order linorder_nat)
                   ma)
       in
      (nfs, (mi, (p, (pw, (q, (nfq || is_NF_trs_subset isnf rs,
                                (nc || list_all
 (fun rb -> not (is_Var (snd rb))) rs,
                                  (vra, (vrwa,
  (ma, (rma, (wwf || wwf_qtrs_impl _B isnf rs, isnf)))))))))))));;

let rec delete_R_Rwa (_A1, _A2) _B
  d r rw = DPP (delete_R_Rw_impla (_A1, _A2) _B (impl_ofd _A1 d) r rw);;

let rec delete_P_Pw_impl _A _B
  (nfs, (mi, (p, (pw, rest)))) pd pwd =
    (nfs, (mi, (list_diff (equal_prod (equal_term _A _B) (equal_term _A _B)) p
                  pd,
                 (list_diff (equal_prod (equal_term _A _B) (equal_term _A _B))
                    pw pwd,
                   rest))));;

let rec delete_P_Pw (_A1, _A2) _B
  d p pw = DPP (delete_P_Pw_impl _A2 _B (impl_ofd _A1 d) p pw);;

let rec rules_map_impla _A
  (uu, (uv, (uw, (ux, (uy, (uz, (va, (vb, (vc, (m, vd)))))))))) fn =
    (match lookup (compare_order_prod _A compare_order_nat) m fn with None -> []
      | Some a -> map snd a);;

let rec rules_mapa _A d = rules_map_impla _A (impl_ofd _A d);;

let rec wwf_rules_impl
  (uu, (uv, (uw, (ux, (uy, (uz, (va, (vb, (vc, (vd, (ve, (wwf, vf)))))))))))) =
    wwf;;

let rec wwf_rules _A d = wwf_rules_impl (impl_ofd _A d);;

let rec q_empty_impla (uu, (uv, (uw, (ux, (q, uy))))) = null q;;

let rec q_emptya _A d = q_empty_impla (impl_ofd _A d);;

let rec is_QNF_impla
  (uu, (uv, (uw, (ux, (uy, (uz, (va, (vb, (vc, (vd, (ve, (vf, isnf)))))))))))) =
    isnf;;

let rec is_QNFa _A d = is_QNF_impla (impl_ofd _A d);;

let rec rulesb _A d = rules_implb _A (impl_ofd _A d);;

let rec pairs _A d = pairs_impl (impl_ofd _A d);;

let rec dpp_impl (_A1, _A2) (_B1, _B2)
  (nfs, (mi, (p, (pw, (q, (uu, (uv, (vr, (vrw, (m, uw)))))))))) =
    (nfs, (mi, (set ((ceq_prod (ceq_term _A2 _B2) (ceq_term _A2 _B2)),
                      (ccompare_prod
                        (ccompare_term _A1.compare_compare_order _B1)
                        (ccompare_term _A1.compare_compare_order _B1)),
                      (set_impl_prod set_impl_term set_impl_term))
                  p,
                 (set ((ceq_prod (ceq_term _A2 _B2) (ceq_term _A2 _B2)),
                        (ccompare_prod
                          (ccompare_term _A1.compare_compare_order _B1)
                          (ccompare_term _A1.compare_compare_order _B1)),
                        (set_impl_prod set_impl_term set_impl_term))
                    pw,
                   (set ((ceq_term _A2 _B2),
                          (ccompare_term _A1.compare_compare_order _B1),
                          set_impl_term)
                      q,
                     (set ((ceq_prod (ceq_term _A2 _B2) (ceq_term _A2 _B2)),
                            (ccompare_prod
                              (ccompare_term _A1.compare_compare_order _B1)
                              (ccompare_term _A1.compare_compare_order _B1)),
                            (set_impl_prod set_impl_term set_impl_term))
                        (vr @ rules_with _A1.linorder_compare_order id m),
                       set ((ceq_prod (ceq_term _A2 _B2) (ceq_term _A2 _B2)),
                             (ccompare_prod
                               (ccompare_term _A1.compare_compare_order _B1)
                               (ccompare_term _A1.compare_compare_order _B1)),
                             (set_impl_prod set_impl_term set_impl_term))
                         (vrw @
                           rules_with _A1.linorder_compare_order not m)))))));;

let rec dpp (_A1, _A2) (_B1, _B2)
  d = dpp_impl (_A1, _A2) (_B1, _B2) (impl_ofd _A1 d);;

let rec nfs_impla (nfs, uu) = nfs;;

let rec nfsa _A d = nfs_impla (impl_ofd _A d);;

let rec mk_impla (_A1, _A2) (_B1, _B2, _B3)
  nfs mi p pw q r rw =
    (let (vr, ra) = partition (comp is_Var fst) r in
     let (vrw, rwa) = partition (comp is_Var fst) rw in
     let rs = r @ rw in
     let isnf = is_NF_terms (_A1, _A2) (_B1, _B2, _B3) q in
      (nfs, (mi, (p, (pw, (q, (is_NF_trs_subset isnf rs,
                                (list_all (fun rb -> not (is_Var (snd rb))) rs,
                                  (vr, (vrw,
 (insert_rules equal_bool (_A1, _A2) _B2 true ra
    (insert_rules equal_bool (_A1, _A2) _B2 false rwa
      (empty (linorder_prod _A1.linorder_compare_order linorder_nat))),
   (insert_rules equal_bool (_A1, _A2) _B2 true (reverse_rules r)
      (insert_rules equal_bool (_A1, _A2) _B2 false (reverse_rules rw)
        (empty (linorder_prod _A1.linorder_compare_order linorder_nat))),
     (wwf_qtrs_impl _B2 isnf rs, isnf)))))))))))));;

let rec mka (_A1, _A2) (_B1, _B2, _B3)
  nfs mi p pw q r rw =
    DPP (mk_impla (_A1, _A2) (_B1, _B2, _B3) nfs mi p pw q r rw);;

let rec rw_impla _A
  (uu, (uv, (uw, (ux, (uy, (uz, (va, (vb, (vrw, (m, vc)))))))))) =
    vrw @ rules_with _A.linorder_compare_order not m;;

let rec rwa _A d = rw_impla _A (impl_ofd _A d);;

let rec pw_impl (uu, (uv, (uw, (pw, ux)))) = pw;;

let rec pw _A d = pw_impl (impl_ofd _A d);;

let rec r_impla _A
  (uu, (uv, (uw, (ux, (uy, (uz, (va, (vr, (vb, (m, vc)))))))))) =
    vr @ rules_with _A.linorder_compare_order id m;;

let rec ra _A d = r_impla _A (impl_ofd _A d);;

let rec q_impla (uu, (uv, (uw, (ux, (q, uy))))) = q;;

let rec qa _A d = q_impla (impl_ofd _A d);;

let rec p_impl (uu, (uv, (p, uw))) = p;;

let rec p _A d = p_impl (impl_ofd _A d);;

let rec m_impl (uu, (mi, uv)) = mi;;

let rec m _A d = m_impl (impl_ofd _A d);;

let rec dpp_rbt_impl (_A1, _A2) (_B1, _B2, _B3, _B4)
  = Dpp_ops_ext
      (dpp (_A1, _A2) (_B2, _B3), p _A1, pw _A1, pairs _A1, qa _A1, ra _A1,
        rwa _A1, rulesb _A1, q_emptya _A1, rules_no_left_var _A1,
        rules_non_collapsing _A1, is_QNFa _A1, nFQ_subset_NF_rulesa _A1,
        rules_mapa _A1, reverse_rules_map _A1, intersect_pairs (_A1, _A2) _B3,
        replace_pair (_A1, _A2) _B3, intersect_rulesa (_A1, _A2) _B3,
        delete_P_Pw (_A1, _A2) _B3, delete_R_Rwa (_A1, _A2) _B3,
        split_pairs (_A1, _A2) _B3, split_rulesa (_A1, _A2) _B3,
        mka (_A1, _A2) (_B1, _B3, _B4), m _A1, nfsa _A1, wwf_rules _A1, ());;

let rec nFQ_subset_NF_rules_impl (uu, (uv, (b, uw))) = b;;

let rec impl_ofc _B _A (TP x) = x;;

let rec nFQ_subset_NF_rules _A _B
  tp = nFQ_subset_NF_rules_impl (impl_ofc _A _B tp);;

let rec rules_impla _A
  (uu, (uv, (uw, (vR, (vRw, (m, ux)))))) =
    vR @ vRw @ map snd
                 (values (linorder_prod _A.linorder_compare_order linorder_nat)
                   m);;

let rec split_rules_impl _A _B
  tp rs =
    (let m =
       ceta_set_of
         (compare_order_prod
           (compare_order_term _A.compare_compare_order
             _B.compare_compare_order)
           (compare_order_term _A.compare_compare_order
             _B.compare_compare_order))
         rs
       in
      partition m (rules_impla _A tp));;

let rec split_rules _A _B tp = split_rules_impl _A _B (impl_ofc _A _B tp);;

let rec delete_R_Rw_impl (_A1, _A2) _B
  (nfs, (q, (uu, (vR, (vRw, (m, isnf)))))) r rw =
    (let (vr, ra) = partition (comp is_Var fst) r in
     let (vrw, rwa) = partition (comp is_Var fst) rw in
     let vra =
       list_diff (equal_prod (equal_term _A2 _B) (equal_term _A2 _B)) vR vr in
     let vrwa =
       list_diff (equal_prod (equal_term _A2 _B) (equal_term _A2 _B)) vRw vrw in
     let ma =
       delete_rules equal_bool (_A1, _A2) _B true ra
         (delete_rules equal_bool (_A1, _A2) _B false rwa m)
       in
      (nfs, (q, (is_NF_trs_subset isnf
                   (vra @
                     vrwa @
                       map snd
                         (values
                           (linorder_prod _A1.linorder_compare_order
                             linorder_nat)
                           ma)),
                  (vra, (vrwa, (ma, isnf)))))));;

let rec delete_R_Rw (_A1, _A2) (_B1, _B2)
  tp r rw = TP (delete_R_Rw_impl (_A1, _A2) _B2 (impl_ofc _A1 _B1 tp) r rw);;

let rec rules_map_impl _A
  (uu, (uv, (uw, (ux, (uy, (m, uz)))))) fn =
    (match lookup (compare_order_prod _A compare_order_nat) m fn with None -> []
      | Some a -> map snd a);;

let rec rules_map _A _B tp = rules_map_impl _A (impl_ofc _A _B tp);;

let rec qreltrs_impl (_A1, _A2) (_B1, _B2)
  (nfs, (q, (uu, (vR, (vRw, (m, isnf)))))) =
    (nfs, (set ((ceq_term _A2 _B2),
                 (ccompare_term _A1.compare_compare_order _B1), set_impl_term)
             q,
            (set ((ceq_prod (ceq_term _A2 _B2) (ceq_term _A2 _B2)),
                   (ccompare_prod (ccompare_term _A1.compare_compare_order _B1)
                     (ccompare_term _A1.compare_compare_order _B1)),
                   (set_impl_prod set_impl_term set_impl_term))
               (vR @ rules_with _A1.linorder_compare_order id m),
              set ((ceq_prod (ceq_term _A2 _B2) (ceq_term _A2 _B2)),
                    (ccompare_prod (ccompare_term _A1.compare_compare_order _B1)
                      (ccompare_term _A1.compare_compare_order _B1)),
                    (set_impl_prod set_impl_term set_impl_term))
                (vRw @ rules_with _A1.linorder_compare_order not m))));;

let rec qreltrs (_A1, _A2) (_B1, _B2)
  tp = qreltrs_impl (_A1, _A2) (_B1.compare_compare_order, _B2)
         (impl_ofc _A1 _B1 tp);;

let rec q_empty_impl (uu, (q, uv)) = null q;;

let rec q_empty _A _B tp = q_empty_impl (impl_ofc _A _B tp);;

let rec is_QNF_impl (uu, (uv, (uw, (ux, (uy, (uz, isnf)))))) = isnf;;

let rec is_QNF _A _B tp = is_QNF_impl (impl_ofc _A _B tp);;

let rec rulesa _A _B tp = rules_impla _A (impl_ofc _A _B tp);;

let rec nfs_impl _A (nfs, uu) = nfs;;

let rec nfs _A _B tp = nfs_impl _A (impl_ofc _A _B tp);;

let rec mk_impl (_A1, _A2) (_B1, _B2, _B3)
  nfs q r rw =
    (let (vr, ra) = partition (comp is_Var fst) r in
     let (vrw, rwa) = partition (comp is_Var fst) rw in
     let isnf = is_NF_terms (_A1, _A2) (_B1, _B2, _B3) q in
      (nfs, (q, (is_NF_trs_subset isnf (r @ rw),
                  (vr, (vrw, (insert_rules equal_bool (_A1, _A2) _B2 true ra
                                (insert_rules equal_bool (_A1, _A2) _B2 false
                                  rwa (empty
(linorder_prod _A1.linorder_compare_order linorder_nat))),
                               isnf)))))));;

let rec mk (_A1, _A2) (_B1, _B2, _B3, _B4)
  nfs q r rw = TP (mk_impl (_A1, _A2) (_B1, _B3, _B4) nfs q r rw);;

let rec rw_impl _A
  (uu, (uv, (uw, (ux, (vRw, (m, uy)))))) =
    vRw @ rules_with _A.linorder_compare_order not m;;

let rec rw _A _B tp = rw_impl _A (impl_ofc _A _B tp);;

let rec r_impl _A
  (uu, (uv, (uw, (vR, (ux, (m, uy)))))) =
    vR @ rules_with _A.linorder_compare_order id m;;

let rec r _A _B tp = r_impl _A (impl_ofc _A _B tp);;

let rec q_impl (uu, (q, uv)) = q;;

let rec q _A _B tp = q_impl (impl_ofc _A _B tp);;

let rec tp_rbt_impl (_A1, _A2) (_B1, _B2, _B3, _B4)
  = Tp_ops_ext
      (qreltrs (_A1, _A2) (_B2, _B3), q _A1 _B2, r _A1 _B2, rw _A1 _B2,
        rulesa _A1 _B2, q_empty _A1 _B2, is_QNF _A1 _B2,
        nFQ_subset_NF_rules _A1 _B2, rules_map _A1 _B2,
        delete_R_Rw (_A1, _A2) (_B2, _B3), split_rules _A1 _B2,
        mk (_A1, _A2) (_B1, _B2, _B3, _B4), nfs _A1 _B2, ());;

let rec certify_cert_problem (_B1, _B2, _B3, _B4, _B5)
  i j k l a input claim proof =
    (match check_cert (_B1, _B2, _B3, _B4, _B5) i j k l a input claim proof
      with Inl err -> Error (err "") | Inr _ -> Certified);;

let rec invariant_proof_parser _F
  location_parser trans_parser atom_parser hint_parser i =
    xml_change
      (art_parser _F xml_text location_parser trans_parser atom_parser
        hint_parser)
      (fun prf -> xml_return (Impact (i, prf)));;

let rec sharp_location_parser
  location_id_parser =
    xml_or
      (xml_change (location_id_parser "locationDuplicate")
        (comp xml_return (fun a -> Sharpa a)))
      (xml_change (location_id_parser "locationId")
        (comp xml_return (fun a -> Flat a)));;

let rec invariant_parser
  location_parser atom_parser =
    xml_do "invariant"
      (xml_take (xml_do "location" (xml_take location_parser xml_return))
        (fun l ->
          xml_take
            (xml_do "formula"
              (xml_take (formula_parser atom_parser) xml_return))
            (fun phi -> xml_return (l, phi))));;

let rec invariants_parser
  location_parser atom_parser =
    xml_do "invariants"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (invariant_parser location_parser atom_parser) xml_return);;

let rec cooperation_proof_parser (_A1, _A2) _E (_F1, _F2, _F3)
  location_id_parser exp_parser atom_parser tatom_parser variable_parser
    hint_parser type_parser dom_type trans_parser xs =
    xml_or (xml_do "trivial" (xml_return Triviala))
      (xml_or
        (xml_do "newInvariants"
          (xml_take
            (invariants_parser (sharp_location_parser location_id_parser)
              atom_parser)
            (fun i ->
              xml_take
                (invariant_proof_parser _E
                  (sharp_location_parser location_id_parser) trans_parser
                  atom_parser hint_parser i)
                (fun p ->
                  xml_take
                    (cooperation_proof_parser (_A1, _A2) _E (_F1, _F2, _F3)
                      location_id_parser exp_parser atom_parser tatom_parser
                      variable_parser hint_parser type_parser dom_type
                      trans_parser)
                    (fun cp -> xml_return (Invariants_Update (p, cp)))))))
        (xml_or
          (xml_do "transitionRemoval"
            (xml_take
              (xml_do "rankingFunctions"
                (xml_take_many_sub [] zero_nata Infinity_enat
                  (xml_do "rankingFunction"
                    (xml_take
                      (xml_do "location"
                        (xml_take (sharp_location_parser location_id_parser)
                          xml_return))
                      (fun l ->
                        xml_take
                          (xml_do "expression"
                            (xml_take_many_sub [] one_nata Infinity_enat
                              exp_parser xml_return))
                          (fun es -> xml_return (l, es)))))
                  xml_return))
              (fun rs ->
                xml_take
                  (xml_do "bound"
                    (xml_take_many_sub [] zero_nata Infinity_enat exp_parser
                      xml_return))
                  (fun bounds ->
                    xml_take
                      (xml_do "remove"
                        (xml_take_many_sub [] zero_nata Infinity_enat
                          trans_parser xml_return))
                      (fun removed ->
                        xml_take_default (fun _ -> default_hinta _E)
                          (xml_do "hints"
                            (xml_take_many_sub [] zero_nata Infinity_enat
                              (xml_do "hint"
                                (xml_take trans_parser
                                  (fun tr ->
                                    xml_take_default (default_hinta _E)
                                      (hints_parser _E hint_parser)
                                      (fun hint -> xml_return (tr, hint)))))
                              (fun pairs ->
                                xml_return
                                  (map_of_default (_F1, _F2, _F3)
                                    (default_hinta _E) pairs))))
                          (fun hinter ->
                            xml_take
                              (cooperation_proof_parser (_A1, _A2) _E
                                (_F1, _F2, _F3) location_id_parser exp_parser
                                atom_parser tatom_parser variable_parser
                                hint_parser type_parser dom_type trans_parser)
                              (fun inner ->
                                (let rf =
                                   map_of_default
                                     ((ccompare_sharp _A1), (equal_sharp _A2),
                                       mapping_impl_sharp)
                                     [] rs
                                   in
                                  xml_return
                                    (Transition_Removal
                                      (Transition_removal_info
 (rf, removed, dom_type, bounds, hinter),
inner))))))))))
          (xml_or
            (xml_do "locationAddition"
              (xml_take
                (transition_parser (sharp_location_parser location_id_parser)
                  trans_parser tatom_parser)
                (fun tr ->
                  xml_take
                    (cooperation_proof_parser (_A1, _A2) _E (_F1, _F2, _F3)
                      location_id_parser exp_parser atom_parser tatom_parser
                      variable_parser hint_parser type_parser dom_type
                      trans_parser)
                    (fun prof ->
                      (let (tr_id, tau) = tr in
                        xml_return
                          (Location_Addition
                            (Location_Addition_Info
                               (source tau, target tau, tr_id, tau),
                              prof)))))))
            (xml_or
              (xml_do "freshVariableAddition"
                (xml_take variable_parser
                  (fun x1 ->
                    xml_take type_parser
                      (fun x2 ->
                        xml_take
                          (xml_do "additionalFormulas"
                            (xml_take_many_sub [] zero_nata Infinity_enat
                              (xml_do "additionalFormula"
                                (xml_take trans_parser
                                  (fun tr ->
                                    xml_take (formula_parser tatom_parser)
                                      (fun phi -> xml_return (tr, phi)))))
                              xml_return))
                          (fun x3 ->
                            xml_take
                              (cooperation_proof_parser (_A1, _A2) _E
                                (_F1, _F2, _F3) location_id_parser exp_parser
                                atom_parser tatom_parser variable_parser
                                hint_parser type_parser dom_type trans_parser)
                              (fun prof ->
                                xml_return
                                  (Fresh_Variable_Addition
                                    (Fresh_Variable_Addition_Info (x1, x2, x3),
                                      prof))))))))
              (xml_or
                (xml_do "cutTransitionSplit"
                  (xml_take_many_sub [] zero_nata Infinity_enat
                    (xml_do "cutTransitionsWithProof"
                      (xml_take
                        (xml_do "cutTransitions"
                          (xml_take_many_sub [] zero_nata Infinity_enat
                            trans_parser xml_return))
                        (fun cuts ->
                          xml_take
                            (cooperation_proof_parser (_A1, _A2) _E
                              (_F1, _F2, _F3) location_id_parser exp_parser
                              atom_parser tatom_parser variable_parser
                              hint_parser type_parser dom_type trans_parser)
                            (fun prof -> xml_return (cuts, prof)))))
                    (fun ps -> xml_return (Cut_Transition_Split ps))))
                (xml_do "sccDecomposition"
                  (xml_take_many_sub [] zero_nata Infinity_enat
                    (xml_do "sccWithProof"
                      (xml_take
                        (xml_do "scc"
                          (xml_take_many_sub [] one_nata Infinity_enat
                            (sharp_location_parser location_id_parser)
                            xml_return))
                        (fun scc ->
                          xml_take
                            (cooperation_proof_parser (_A1, _A2) _E
                              (_F1, _F2, _F3) location_id_parser exp_parser
                              atom_parser tatom_parser variable_parser
                              hint_parser type_parser dom_type trans_parser)
                            (fun prof -> xml_return (scc, prof)))))
                    (fun sccs -> xml_return (Scc_Decomp sccs)))))))))
      xs;;

let rec cut_points_to_transitions
  ts x1 = match ts, x1 with ts, [] -> ts
    | ts, (l, (tr, phi)) :: cps ->
        cut_points_to_transitions
          ((Flat tr, Transition (Flat l, Sharpa l, phi)) :: ts) cps;;

let rec cutPoints_parser
  location_id_parser trans_id_parser tatom_parser =
    xml_do "cutPoints"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "cutPoint"
          (xml_take (location_id_parser "locationId")
            (fun l ->
              xml_take (trans_id_parser "skipId")
                (fun tr ->
                  xml_take
                    (xml_do "skipFormula"
                      (xml_take (formula_parser tatom_parser) xml_return))
                    (fun phi -> xml_return (l, (tr, phi)))))))
        (fun tuples -> xml_return (cut_points_to_transitions [] tuples)));;

let rec sharp_trans_id
  trans_id_parser =
    xml_or
      (xml_change (trans_id_parser "transitionDuplicate")
        (comp xml_return (fun a -> Sharpa a)))
      (xml_change (trans_id_parser "transitionId")
        (comp xml_return (fun a -> Flat a)));;

let rec termination_proof_parser (_A1, _A2) (_B1, _B2) _F
  location_id_parser trans_id_parser exp_parser atom_parser tatom_parser
    variable_parser hint_parser type_parser dom_type xml =
    xml_or (xml_do "trivial" (xml_return Trivialb))
      (xml_or
        (xml_do "newInvariants"
          (xml_take
            (invariants_parser (location_parser location_id_parser) atom_parser)
            (fun i ->
              xml_take
                (invariant_proof_parser _F (location_parser location_id_parser)
                  (trans_id trans_id_parser) atom_parser hint_parser i)
                (fun p ->
                  xml_take
                    (termination_proof_parser (_A1, _A2) (_B1, _B2) _F
                      location_id_parser trans_id_parser exp_parser atom_parser
                      tatom_parser variable_parser hint_parser type_parser
                      dom_type)
                    (fun cp -> xml_return (Invariants_Update_LTS (p, cp)))))))
        (xml_do "switchToCooperationTermination"
          (xml_take
            (cutPoints_parser location_id_parser trans_id_parser tatom_parser)
            (fun cp ->
              xml_take
                (cooperation_proof_parser (_A1, _A2) _F
                  ((ccompare_sharp _B1), (equal_sharp _B2), mapping_impl_sharp)
                  location_id_parser exp_parser atom_parser tatom_parser
                  variable_parser hint_parser type_parser dom_type
                  (sharp_trans_id trans_id_parser))
                (fun p -> xml_return (Via_Cooperation [(cp, p)]))))))
      xml;;

let rec tatom_parser
  x = xml_change (bexp_parser trans_var_parser)
        (comp xml_return (fun a -> Atom a)) x;;

let rec type_parser
  x = xml_or (xml_do "int" (xml_return IntT)) (xml_do "bool" (xml_return BoolT))
        x;;

let rec hint_parser
  x = xml_do "linearCombination"
        (xml_take_many_sub [] zero_nata Infinity_enat (xml_int "constant")
          (fun xs -> xml_return (Hints xs)))
        x;;

let rec atom_parser
  x = xml_change (bexp_parser variable_parser)
        (comp xml_return (fun a -> Atom a)) x;;

let rec exp_parser x = exp_parsera variable_parser x;;

let rec lts_termination_proof_parser
  x = termination_proof_parser
        ((ccompare_list ccompare_char), (equal_list equal_char))
        ((ccompare_list ccompare_char), (equal_list equal_char)) default_hints
        xml_text xml_text exp_parser atom_parser tatom_parser variable_parser
        hint_parser type_parser IntT x;;

let rec safety_proof_parser _F
  location_parser trans_parser atom_parser hint_parser =
    xml_do "safetyViaInvariants"
      (xml_take (invariants_parser location_parser atom_parser)
        (fun i ->
          xml_take
            (invariant_proof_parser _F location_parser trans_parser atom_parser
              hint_parser i)
            (fun p -> xml_return (Invariant_Assertion (p, Trivial)))));;

let rec lts_safety_proof_parser
  x = safety_proof_parser default_hints (location_parser xml_text)
        (trans_id xml_text) atom_parser hint_parser x;;

let rec xml2dp_nontermination_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name x =
    xml_do "dpNonterminationProof"
      (xml_take
        (xml_or
          (xml_do "dpRuleRemoval"
            (xml_take_optional
              (xml_do "dps"
                (xml_take (rules (showl_lab _A3 _B3) xml2name)
                  (fun xa -> xml_return (id xa))))
              (fun p ->
                xml_take_optional
                  (xml_do "trs"
                    (xml_take (rules (showl_lab _A3 _B3) xml2name)
                      (fun xa -> xml_return (id xa))))
                  (fun r ->
                    xml_take
                      (xml2dp_nontermination_proof (_A1, _A2, _A3)
                        (_B1, _B2, _B3) xml2name)
                      (fun prf ->
                        xml_return
                          (DP_Rule_Removal
                            (Rule_removal_nonterm_dp_prf (p, r), prf)))))))
          (xml_or
            (xml_change (loop (showl_lab _A3 _B3) xml2name)
              (fun (s, (rseq, (sigma, c))) ->
                xml_return (DP_Loop (DP_loop_prf (s, rseq, sigma, c)))))
            (xml_or
              (xml_change (nonloop (_A1, _A3) (_B1, _B3) xml2name)
                (comp xml_return (fun a -> DP_Nonloop a)))
              (xml_or
                (xml_do "innermostLhssRemovalProc"
                  (xml_take (innermostLhss (showl_lab _A3 _B3) xml2name)
                    (fun a ->
                      xml_take
                        (xml2dp_nontermination_proof (_A1, _A2, _A3)
                          (_B1, _B2, _B3) xml2name)
                        (fun b ->
                          xml_return
                            (DP_Q_Reduction
                              (DP_q_reduction_nonterm_prf a, b))))))
                (xml_or
                  (xml_do "innermostLhssIncreaseProc"
                    (xml_take (innermostLhss (showl_lab _A3 _B3) xml2name)
                      (fun a ->
                        xml_take
                          (xml2dp_nontermination_proof (_A1, _A2, _A3)
                            (_B1, _B2, _B3) xml2name)
                          (fun b ->
                            xml_return
                              (DP_Q_Increase
                                (Q_increase_nonterm_dp_prf a, b))))))
                  (xml_or
                    (xml_do "instantiationProc"
                      (xml_take
                        (xml_do "dps"
                          (xml_take (rules (showl_lab _A3 _B3) xml2name)
                            (fun xa -> xml_return (id xa))))
                        (fun a ->
                          xml_take
                            (xml2dp_nontermination_proof (_A1, _A2, _A3)
                              (_B1, _B2, _B3) xml2name)
                            (fun b ->
                              xml_return
                                (DP_Instantiation
                                  (Instantiation_complete_proc_prf a, b))))))
                    (xml_or
                      (xml_do "narrowingProc"
                        (xml_take (rule (showl_lab _A3 _B3) xml2name)
                          (fun a ->
                            xml_take pos
                              (fun b ->
                                xml_take
                                  (xml_do "narrowings"
                                    (xml_take
                                      (rules (showl_lab _A3 _B3) xml2name)
                                      (fun xa -> xml_return (id xa))))
                                  (fun c ->
                                    xml_take
                                      (xml2dp_nontermination_proof
(_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
                                      (fun d ->
xml_return (DP_Narrowing (Narrowing_complete_proc_prf (a, b, c), d))))))))
                      (xml_or
                        (xml_do "rewritingProc"
                          (xml_take (rule (showl_lab _A3 _B3) xml2name)
                            (fun (s, t) ->
                              xml_take (rstep (showl_lab _A3 _B3) xml2name)
                                (fun (p, (lr, ta)) ->
                                  xml_take_optional
                                    (rule (showl_lab _A3 _B3) xml2name)
                                    (fun c ->
                                      xml_take_optional
(xml_do "usableRules"
  (xml_take (rules (showl_lab _A3 _B3) xml2name)
    (fun xa -> xml_return (id xa))))
(fun uopt ->
  xml_take
    (xml2dp_nontermination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
    (fun prf ->
      (let st = (match c with None -> (s, ta) | Some v -> v) in
        xml_return
          (DP_Rewriting
            (Rewriting_complete_proc_prf (uopt, (s, t), (s, ta), st, lr, p),
              prf))))))))))
                        (xml_or
                          (xml_do "switchFullStrategyProc"
                            (xml_take (wcr_proof (showl_lab _A3 _B3) xml2name)
                              (fun a ->
                                xml_take
                                  (xml2dp_nontermination_proof (_A1, _A2, _A3)
                                    (_B1, _B2, _B3) xml2name)
                                  (fun b ->
                                    xml_return
                                      (DP_Termination_Switch (a, b))))))
                          (xml_or
                            (xml_do "infinitenessAssumption"
                              (xml_take
                                (xml2dp_inputa
                                  ((equal_lab _A2 _B2), (showl_lab _A3 _B3))
                                  xml2name false)
                                (fun xa ->
                                  xml_return (DP_Assume_Infinite (xa, [])))))
                            (xml_do "unknownProof"
                              (xml_take (xml_text "description")
                                (fun _ ->
                                  xml_take
                                    (xml2dp_inputa
                                      ((equal_lab _A2 _B2), (showl_lab _A3 _B3))
                                      xml2name false)
                                    (fun b ->
                                      xml_take_many_sub [] zero_nata
Infinity_enat (fun xa -> right (fst xa))
(fun _ -> xml_return (DP_Assume_Infinite (b, [])))))))))))))))))
        (fun xa -> xml_return (id xa)))
      x;;

let rec xml2inn_nt_trs_assm (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name =
    xml_change (xml2_trs_input (showl_lab _A3 _B3) xml2name)
      (fun (Inn_TRS_input (inn, r, s, start)) ->
        (if not (null s)
          then xml_error "(innermost) TRS without relative rules expected"
          else (if not (equal_start_term start Full)
                 then xml_error "start term is not allowed here"
                 else xml_return
                        (default_nfs_nt_trs,
                          (strategy_to_Q (equal_lab _A2 _B2)
                             (equal_list equal_char) inn r,
                            r)))));;

let rec xml2trs_nontermination_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name x =
    xml_do "trsNonterminationProof"
      (xml_take
        (xml_or (xml_leaf "variableConditionViolated" TRS_Not_Well_Formed)
          (xml_or
            (xml_change (loop (showl_lab _A3 _B3) xml2name)
              (fun (s, (rseq, (sigma, c))) ->
                xml_return
                  (TRS_Loop
                    (TRS_loop_prf
                      (s, map (fun (xa, (y, (_, z))) -> (xa, (y, z))) rseq,
                        sigma, c)))))
            (xml_or
              (xml_change (nonloop (_A1, _A3) (_B1, _B3) xml2name)
                (comp xml_return (fun a -> TRS_Nonloop a)))
              (xml_or
                (xml_change (nonloop_srs (_A1, _A3) (_B1, _B3) xml2name)
                  (comp xml_return (fun a -> TRS_Nonloop_SRS a)))
                (xml_or
                  (xml_do "ruleRemoval"
                    (xml_take
                      (xml_do "trs"
                        (xml_take (rules (showl_lab _A3 _B3) xml2name)
                          (fun xa -> xml_return (id xa))))
                      (fun a ->
                        xml_take
                          (xml2trs_nontermination_proof (_A1, _A2, _A3)
                            (_B1, _B2, _B3) xml2name)
                          (fun b ->
                            xml_return
                              (TRS_Rule_Removal
                                (Rule_removal_nonterm_trs_prf a, b))))))
                  (xml_or
                    (xml_do "dpTrans"
                      (xml_take
                        (xml_do "dps"
                          (xml_take (rules (showl_lab _A3 _B3) xml2name)
                            (fun xa -> xml_return (id xa))))
                        (fun a ->
                          xml_take (xml_bool "markedSymbols")
                            (fun _ ->
                              xml_take
                                (xml2dp_nontermination_proof (_A1, _A2, _A3)
                                  (_B1, _B2, _B3) xml2name)
                                (fun c ->
                                  xml_return
                                    (TRS_DP_Trans
                                      (DP_trans_nontermination_tt_prf a,
c)))))))
                    (xml_or
                      (xml_do "stringReversal"
                        (xml_take
                          (xml_do "trs"
                            (xml_take (rules (showl_lab _A3 _B3) xml2name)
                              (fun xa -> xml_return (id xa))))
                          (fun _ ->
                            xml_take
                              (xml2trs_nontermination_proof (_A1, _A2, _A3)
                                (_B1, _B2, _B3) xml2name)
                              (fun b -> xml_return (TRS_String_Reversal b)))))
                      (xml_or
                        (xml_do "constantToUnary"
                          (xml_take plain_var
                            (fun a ->
                              xml_take (renaming (showl_lab _A3 _B3) xml2name)
                                (fun b ->
                                  xml_take
                                    (xml_do "trs"
                                      (xml_take
(rules (showl_lab _A3 _B3) xml2name) (fun xa -> xml_return (id xa))))
                                    (fun c ->
                                      xml_take
(xml2trs_nontermination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
(fun d ->
  xml_return
    (TRS_Constant_String (Const_string_complete_proof (a, b, c), d))))))))
                        (xml_or
                          (xml_do "innermostLhssIncrease"
                            (xml_take
                              (innermostLhss (showl_lab _A3 _B3) xml2name)
                              (fun a ->
                                xml_take
                                  (xml2trs_nontermination_proof (_A1, _A2, _A3)
                                    (_B1, _B2, _B3) xml2name)
                                  (fun b ->
                                    xml_return
                                      (TRS_Q_Increase
(Q_increase_nonterm_trs_prf a, b))))))
                          (xml_or
                            (xml_do "switchFullStrategy"
                              (xml_take (wcr_proof (showl_lab _A3 _B3) xml2name)
                                (fun a ->
                                  xml_take
                                    (xml2trs_nontermination_proof
                                      (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
                                    (fun b ->
                                      xml_return
(TRS_Termination_Switch (a, b))))))
                            (xml_or
                              (xml_do "uncurry"
                                (xml_take
                                  (uncurry_info (showl_lab _A3 _B3) xml2name)
                                  (fun a ->
                                    xml_take
                                      (xml_do "trs"
(xml_take (rules (showl_lab _A3 _B3) xml2name) (fun xa -> xml_return (id xa))))
                                      (fun b ->
xml_take (xml2trs_nontermination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
  (fun c -> xml_return (TRS_Uncurry (Uncurry_nt_proof (a, b), c)))))))
                              (xml_or
                                (xml_change
                                  (not_wn_ta (_A1, _A3) (_B1, _B3) xml2name)
                                  (comp xml_return
                                    (fun a -> TRS_Not_WN_Tree_Automaton a)))
                                (xml_or
                                  (xml_do "nonterminationAssumption"
                                    (xml_take
                                      (xml2inn_nt_trs_assm (_A1, _A2, _A3)
(_B1, _B2, _B3) xml2name)
                                      (fun xa ->
xml_return (TRS_Assume_Not_SN (xa, [])))))
                                  (xml_do "unknownProof"
                                    (xml_take (xml_text "description")
                                      (fun _ ->
xml_take (xml2inn_nt_trs_assm (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
  (fun b ->
    xml_take_many_sub [] zero_nata Infinity_enat (fun xa -> right (fst xa))
      (fun _ -> xml_return (TRS_Assume_Not_SN (b, []))))))))))))))))))))
        (fun xa -> xml_return (id xa)))
      x;;

let rec xml2inn_rel_nt_trs_assm (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name =
    xml_change (xml2_trs_input (showl_lab _A3 _B3) xml2name)
      (fun (Inn_TRS_input (inn, r, s, start)) ->
        (if not (equal_start_term start Full)
          then xml_error "start term is not allowed here"
          else xml_return
                 (default_nfs_nt_trs,
                   (strategy_to_Q (equal_lab _A2 _B2) (equal_list equal_char)
                      inn r,
                     (r, s)))));;

let rec xml2reltrs_nontermination_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name x =
    xml_do "relativeNonterminationProof"
      (xml_take
        (xml_or (xml_leaf "variableConditionViolated" Rel_Not_Well_Formed)
          (xml_or
            (xml_change (loop (showl_lab _A3 _B3) xml2name)
              (fun (s, (rseq, (sigma, c))) ->
                xml_return (Rel_Loop (Rel_trs_loop_prf (s, rseq, sigma, c)))))
            (xml_or
              (xml_change
                (xml2trs_nontermination_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
                  xml2name)
                (comp xml_return (fun a -> Rel_R_Not_SN a)))
              (xml_or
                (xml_do "ruleRemoval"
                  (xml_take
                    (xml_do "trs"
                      (xml_take (rules (showl_lab _A3 _B3) xml2name)
                        (fun xa -> xml_return (Some xa))))
                    (fun a ->
                      xml_take
                        (xml_do "trs"
                          (xml_take (rules (showl_lab _A3 _B3) xml2name)
                            (fun xa -> xml_return (Some xa))))
                        (fun b ->
                          xml_take
                            (xml2reltrs_nontermination_proof (_A1, _A2, _A3)
                              (_B1, _B2, _B3) xml2name)
                            (fun c ->
                              xml_return
                                (Rel_Rule_Removal
                                  (Rule_removal_nonterm_reltrs_prf (a, b),
                                    c)))))))
                (xml_or
                  (xml_do "stringReversal"
                    (xml_take
                      (xml_do "trs"
                        (xml_take (rules (showl_lab _A3 _B3) xml2name)
                          (fun xa -> xml_return (Some xa))))
                      (fun _ ->
                        xml_take
                          (xml_do "trs"
                            (xml_take (rules (showl_lab _A3 _B3) xml2name)
                              (fun xa -> xml_return (Some xa))))
                          (fun _ ->
                            xml_take
                              (xml2reltrs_nontermination_proof (_A1, _A2, _A3)
                                (_B1, _B2, _B3) xml2name)
                              (fun c ->
                                xml_return (Rel_TRS_String_Reversal c))))))
                  (xml_or
                    (xml_do "nonterminationAssumption"
                      (xml_take
                        (xml2inn_rel_nt_trs_assm (_A1, _A2, _A3) (_B1, _B2, _B3)
                          xml2name)
                        (fun xa ->
                          xml_return (Rel_TRS_Assume_Not_SN (xa, [])))))
                    (xml_do "unknownProof"
                      (xml_take (xml_text "description")
                        (fun _ ->
                          xml_take
                            (xml2inn_rel_nt_trs_assm (_A1, _A2, _A3)
                              (_B1, _B2, _B3) xml2name)
                            (fun b ->
                              xml_take_many_sub [] zero_nata Infinity_enat
                                (fun xa -> right (fst xa))
                                (fun _ ->
                                  xml_return
                                    (Rel_TRS_Assume_Not_SN (b, [])))))))))))))
        (fun xa -> xml_return (id xa)))
      x;;

let rec xml2ordered_completion_step (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_or
      (xml_do "deduce"
        (xml_take (term (showl_lab _A2 _B2) xml2name)
          (fun a ->
            xml_take (term (showl_lab _A2 _B2) xml2name)
              (fun b ->
                xml_take (term (showl_lab _A2 _B2) xml2name)
                  (fun c -> xml_return (OC_Deduce (a, b, c)))))))
      (xml_or
        (xml_do "orientl"
          (xml_take (term (showl_lab _A2 _B2) xml2name)
            (fun a ->
              xml_take (term (showl_lab _A2 _B2) xml2name)
                (fun b -> xml_return (OC_Orientl (a, b))))))
        (xml_or
          (xml_do "orientr"
            (xml_take (term (showl_lab _A2 _B2) xml2name)
              (fun a ->
                xml_take (term (showl_lab _A2 _B2) xml2name)
                  (fun b -> xml_return (OC_Orientr (a, b))))))
          (xml_or
            (xml_do "delete"
              (xml_take (term (showl_lab _A2 _B2) xml2name)
                (fun x -> xml_return (OC_Delete x))))
            (xml_or
              (xml_do "compose"
                (xml_take (term (showl_lab _A2 _B2) xml2name)
                  (fun a ->
                    xml_take (term (showl_lab _A2 _B2) xml2name)
                      (fun b ->
                        xml_take (term (showl_lab _A2 _B2) xml2name)
                          (fun c -> xml_return (OC_Compose (a, b, c)))))))
              (xml_or
                (xml_do "simplifyl"
                  (xml_take (term (showl_lab _A2 _B2) xml2name)
                    (fun a ->
                      xml_take (term (showl_lab _A2 _B2) xml2name)
                        (fun b ->
                          xml_take (term (showl_lab _A2 _B2) xml2name)
                            (fun c -> xml_return (OC_Simplifyl (a, b, c)))))))
                (xml_or
                  (xml_do "simplifyr"
                    (xml_take (term (showl_lab _A2 _B2) xml2name)
                      (fun a ->
                        xml_take (term (showl_lab _A2 _B2) xml2name)
                          (fun b ->
                            xml_take (term (showl_lab _A2 _B2) xml2name)
                              (fun c -> xml_return (OC_Simplifyr (a, b, c)))))))
                  (xml_do "collapse"
                    (xml_take (term (showl_lab _A2 _B2) xml2name)
                      (fun a ->
                        xml_take (term (showl_lab _A2 _B2) xml2name)
                          (fun b ->
                            xml_take (term (showl_lab _A2 _B2) xml2name)
                              (fun c ->
                                xml_return (OC_Collapse (a, b, c)))))))))))));;

let rec xml2ordered_completion_proof (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "orderedCompletionProof"
      (xml_take
        (xml_do "run"
          (xml_take_many_sub [] zero_nata Infinity_enat
            (xml_do "orderedCompletionStep"
              (xml_take
                (xml2ordered_completion_step (_A1, _A2) (_B1, _B2) xml2name)
                (fun x -> xml_return (id x))))
            (fun a -> xml_return (id a))))
        (fun x -> xml_return (OKB x)));;

let rec unraveling_info (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "unravelingInformation"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "unravelingEntry"
          (xml_take (crule (showl_lab _A2 _B2) xml2name "conditionalRule")
            (fun a ->
              xml_take_many_sub [] zero_nata Infinity_enat
                (rule (showl_lab _A2 _B2) xml2name)
                (fun b -> xml_return (a, b)))))
        (fun a -> xml_return (id a)));;

let rec xml2quasi_reductive_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name =
    xml_do "quasiReductiveProof"
      (xml_take
        (xml_do "unraveling"
          (xml_take (unraveling_info (_A1, _A3) (_B1, _B3) xml2name)
            (fun a ->
              xml_take
                (xml2trs_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
                  xml2name)
                (fun b -> xml_return (Unravel (a, b))))))
        (fun x -> xml_return (id x)));;

let rec map_r_states
  f r = TA_rule (r_root r, map f (r_lhs_states r), f (r_rhs r));;

let rec map_states_impl
  f (Tree_Automaton (qs, ts, eps)) =
    Tree_Automaton
      (map f qs, map (map_r_states f) ts, map (fun (p, q) -> (f p, f q)) eps);;

let rec xml2nonreachable_etac_info (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "nonreachableEtac"
      (xml_take (cPFsignature (showl_lab _A2 _B2) xml2name)
        (fun a ->
          xml_take xml2name
            (fun b ->
              xml_take xml2name
                (fun c ->
                  xml_take (tree_automaton (ta_normal_lhs xml2name))
                    (fun d ->
                      xml_take (xml2state_map (_A1, _A2) (_B1, _B2) xml2name)
                        (fun e ->
                          xml_return
                            (Nonreachable_ETAC
                              (a, b, c, map_states_impl e d))))))));;

let rec xml2nonreachability_proof (_A1, _A2) (_B1, _B2)
  xml2name x =
    xml_do "nonreachabilityProof"
      (xml_take
        (xml_or (xml_leaf "nonreachableTcap" Nonreachable_Gtcap)
          (xml_or (xml2nonreachable_etac_info (_A1, _A2) (_B1, _B2) xml2name)
            (xml_or
              (xml_do "nonreachableSubstApprox"
                (xml_take (rules (showl_lab _A2 _B2) xml2name)
                  (fun a ->
                    xml_take
                      (xml2nonreachability_proof (_A1, _A2) (_B1, _B2) xml2name)
                      (fun b ->
                        xml_return (Nonreachable_Subst_Approx (a, b))))))
              (xml_do "nonreachableReverse"
                (xml_take
                  (xml2nonreachability_proof (_A1, _A2) (_B1, _B2) xml2name)
                  (fun xa -> xml_return (Nonreachable_Reverse xa)))))))
        (fun xa -> xml_return (id xa)))
      x;;

let rec xml2nonjoinability_proof (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "nonjoinabilityProof"
      (xml_take
        (xml_or (xml_leaf "nonjoinableTcap" Nonjoinable_Tcap)
          (xml_change (xml2nonreachability_proof (_A1, _A2) (_B1, _B2) xml2name)
            (comp xml_return (fun a -> Nonjoinable_Ground_NF a))))
        (fun x -> xml_return (id x)));;

let rec xml2infeasibility_proof (_A1, _A2) (_B1, _B2)
  xml2name x =
    xml_do "infeasibilityProof"
      (xml_take
        (xml_or
          (xml_do "infeasibleCompoundConditions"
            (xml_take xml2name
              (fun a ->
                xml_take
                  (xml2nonreachability_proof (_A1, _A2) (_B1, _B2) xml2name)
                  (fun b ->
                    xml_return (Infeasible_Compound_Conditions (a, b))))))
          (xml_or
            (xml_do "infeasibleEquation"
              (xml_take (rule (showl_lab _A2 _B2) xml2name)
                (fun a ->
                  xml_take
                    (xml2nonreachability_proof (_A1, _A2) (_B1, _B2) xml2name)
                    (fun b ->
                      xml_return
                        ((let (aa, ba) = a in
                           (fun c -> Infeasible_Equation (aa, ba, c)))
                          b)))))
            (xml_or
              (xml_do "infeasibleSubset"
                (xml_take (rules (showl_lab _A2 _B2) xml2name)
                  (fun a ->
                    xml_take
                      (xml2infeasibility_proof (_A1, _A2) (_B1, _B2) xml2name)
                      (fun b -> xml_return (Infeasible_Subset (a, b))))))
              (xml_or
                (xml_do "infeasibleRhssEqual"
                  (xml_take (term (showl_lab _A2 _B2) xml2name)
                    (fun a ->
                      xml_take (term (showl_lab _A2 _B2) xml2name)
                        (fun b ->
                          xml_take (term (showl_lab _A2 _B2) xml2name)
                            (fun c ->
                              xml_take
                                (xml2nonjoinability_proof (_A1, _A2) (_B1, _B2)
                                  xml2name)
                                (fun d ->
                                  xml_return
                                    (Infeasible_Rhss_Equal (a, b, c, d))))))))
                (xml_do "infeasibleTrans"
                  (xml_take (term (showl_lab _A2 _B2) xml2name)
                    (fun a ->
                      xml_take (term (showl_lab _A2 _B2) xml2name)
                        (fun b ->
                          xml_take (term (showl_lab _A2 _B2) xml2name)
                            (fun c ->
                              xml_take
                                (xml2nonreachability_proof (_A1, _A2) (_B1, _B2)
                                  xml2name)
                                (fun d ->
                                  xml_return
                                    (Infeasible_Trans (a, b, c, d))))))))))))
        (fun xa -> xml_return (id xa)))
      x;;

let rec xml2infeasible_rules_info (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "infeasibleRules"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "infeasibleRule"
          (xml_take (crule (showl_lab _A2 _B2) xml2name "rule")
            (fun a ->
              xml_take (xml2infeasibility_proof (_A1, _A2) (_B1, _B2) xml2name)
                (fun b -> xml_return (a, b)))))
        (fun a -> xml_return (id a)));;

let rec xml2inline_cond_info (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "inlinedRules"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "inlinedRule"
          (xml_take (crule (showl_lab _A2 _B2) xml2name "rule")
            (fun rule ->
              xml_take
                (conditions (showl_lab _A2 _B2) xml2name "inlinedConditions")
                (fun conds -> xml_return (rule, conds)))))
        xml_return);;

let rec xml2conditional_ncr_proof
  xml2name x =
    xml_do "conditionalCrDisproof"
      (xml_take
        (xml_or
          (xml_do "unconditional"
            (xml_take (xml2ncr_proof xml2name)
              (fun xa -> xml_return (Unconditional_CNCR xa))))
          (xml_or
            (xml_do "inlineConditions"
              (xml_take
                (xml_do "rules"
                  (xml_take_many_sub [] zero_nata Infinity_enat
                    (crule
                      (showl_lab (showl_list showl_char) (showl_list showl_nat))
                      xml2name "rule")
                    xml_return))
                (fun a ->
                  xml_take
                    (xml2inline_cond_info
                      ((compare_order_list (compare_order_char, equal_char)),
                        (showl_list showl_char))
                      ((compare_order_list (compare_order_nat, equal_nat)),
                        (showl_list showl_nat))
                      xml2name)
                    (fun b ->
                      xml_take (xml2conditional_ncr_proof xml2name)
                        (fun c ->
                          xml_return
                            (comp (fun aa ba -> Transformation_CNCR (aa, ba))
                              (fun ba -> Inline_Conditions_CCRT (a, ba)) b
                              c))))))
            (xml_or
              (xml_do "infeasibleRuleRemoval"
                (xml_take
                  (xml2infeasible_rules_info
                    ((compare_order_list (compare_order_char, equal_char)),
                      (showl_list showl_char))
                    ((compare_order_list (compare_order_nat, equal_nat)),
                      (showl_list showl_nat))
                    xml2name)
                  (fun a ->
                    xml_take (xml2conditional_ncr_proof xml2name)
                      (fun b ->
                        xml_return
                          (comp (fun aa ba -> Transformation_CNCR (aa, ba))
                            (fun aa -> Infeasible_Rule_Removal_CCRT aa) a b)))))
              (xml_do "nonJoinableFork"
                (xml_take
                  (xml_do "terms"
                    (xml_take
                      (term (showl_lab (showl_list showl_char)
                              (showl_list showl_nat))
                        xml2name)
                      (fun a ->
                        xml_take
                          (term (showl_lab (showl_list showl_char)
                                  (showl_list showl_nat))
                            xml2name)
                          (fun b ->
                            xml_take
                              (term (showl_lab (showl_list showl_char)
                                      (showl_list showl_nat))
                                xml2name)
                              (fun c -> xml_return (a, (b, c)))))))
                  (fun a ->
                    xml_take
                      (csteps
                        ((compare_order_list (compare_order_char, equal_char)),
                          (showl_list showl_char))
                        ((compare_order_list (compare_order_nat, equal_nat)),
                          (showl_list showl_nat))
                        xml2name)
                      (fun b ->
                        xml_take
                          (csteps
                            ((compare_order_list
                               (compare_order_char, equal_char)),
                              (showl_list showl_char))
                            ((compare_order_list
                               (compare_order_nat, equal_nat)),
                              (showl_list showl_nat))
                            xml2name)
                          (fun c ->
                            xml_take (xml2non_join_info xml2name)
                              (fun d ->
                                xml_return
                                  ((let (s, aa) = a in
                                    let (ab, ba) = aa in
                                     (fun ca da e ->
                                       Non_Join_CNCR (s, ab, ba, ca, da, e)))
                                     b
                                     c
                                    d))))))))))
        (fun xa -> xml_return (id xa)))
      x;;

let rec xml2context_joinable_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name =
    xml_do "contextJoinabilityProof"
      (xml_take (term (showl_lab _A3 _B3) xml2name)
        (fun a ->
          xml_take (csteps (_A1, _A3) (_B1, _B3) xml2name)
            (fun b ->
              xml_take (csteps (_A1, _A3) (_B1, _B3) xml2name)
                (fun c ->
                  xml_take
                    (xml2const_map (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
                    (fun d ->
                      xml_return
                        (Contextual_Join
                          (orig_term d a, map (orig_cstep d) b,
                            map (orig_cstep d) c)))))));;

let rec xml2context_joinable_ccps (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name =
    xml_do "contextJoinableCCPs"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "contextJoinableCCP"
          (xml_take (term (showl_lab _A3 _B3) xml2name)
            (fun s ->
              xml_take (term (showl_lab _A3 _B3) xml2name)
                (fun t ->
                  xml_take
                    (conditions (showl_lab _A3 _B3) xml2name "conditions")
                    (fun cs ->
                      xml_take
                        (xml2context_joinable_proof (_A1, _A2, _A3)
                          (_B1, _B2, _B3) xml2name)
                        (fun p -> xml_return (s, (t, (cs, p)))))))))
        xml_return);;

let rec xml2ao_infeasibility_proof (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "aoInfeasibilityProof"
      (xml_take
        (xml_or
          (xml_change (xml2infeasibility_proof (_A1, _A2) (_B1, _B2) xml2name)
            (comp xml_return (fun a -> AO_Infeasibility_Proof a)))
          (xml_do "aoLhssEqual"
            (xml_take (term (showl_lab _A2 _B2) xml2name)
              (fun a ->
                xml_take (term (showl_lab _A2 _B2) xml2name)
                  (fun b ->
                    xml_take (term (showl_lab _A2 _B2) xml2name)
                      (fun c ->
                        xml_take
                          (xml2nonjoinability_proof (_A1, _A2) (_B1, _B2)
                            xml2name)
                          (fun d ->
                            xml_return (AO_Lhss_Equal (a, b, c, d)))))))))
        (fun x -> xml_return (id x)));;

let rec xml2ao_infeasible_conds (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "aoInfeasibleConditions"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "aoInfeasibleCondition"
          (xml_take (conditions (showl_lab _A2 _B2) xml2name "conditions")
            (fun a ->
              xml_take (conditions (showl_lab _A2 _B2) xml2name "conditions")
                (fun b ->
                  xml_take
                    (xml2ao_infeasibility_proof (_A1, _A2) (_B1, _B2) xml2name)
                    (fun c -> xml_return (a, (b, c)))))))
        (fun a -> xml_return (id a)));;

let rec xml2infeasible_conds (_A1, _A2) (_B1, _B2)
  xml2name =
    xml_do "infeasibleConditions"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "infeasibleCondition"
          (xml_take (conditions (showl_lab _A2 _B2) xml2name "conditions")
            (fun a ->
              xml_take (xml2infeasibility_proof (_A1, _A2) (_B1, _B2) xml2name)
                (fun b -> xml_return (a, b)))))
        (fun a -> xml_return (id a)));;

let rec xml2unfeasible_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name =
    xml_do "unfeasibilityProof"
      (xml_take
        (xml_do "terms"
          (xml_take (term (showl_lab _A3 _B3) xml2name)
            (fun a ->
              xml_take (term (showl_lab _A3 _B3) xml2name)
                (fun b ->
                  xml_take (term (showl_lab _A3 _B3) xml2name)
                    (fun c -> xml_return (a, (b, c)))))))
        (fun a ->
          xml_take (csteps (_A1, _A3) (_B1, _B3) xml2name)
            (fun b ->
              xml_take (csteps (_A1, _A3) (_B1, _B3) xml2name)
                (fun c ->
                  xml_take
                    (xml_do "rules"
                      (xml_take (crule (showl_lab _A3 _B3) xml2name "rule")
                        (fun aa ->
                          xml_take (crule (showl_lab _A3 _B3) xml2name "rule")
                            (fun ba -> xml_return (aa, ba)))))
                    (fun d ->
                      xml_take
                        (xml2const_map (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
                        (fun e ->
                          xml_return
                            ((let (t, (u, v)) = a in
                               (fun ps qs (r, ra) m ->
                                 UnfeasibleOverlap
                                   (orig_term m t, orig_term m u, orig_term m v,
                                     map (orig_cstep m) ps,
                                     map (orig_cstep m) qs, r, ra)))
                               b
                               c
                               d
                              e)))))));;

let rec xml2unfeasible_ccps (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name =
    xml_do "unfeasibleCCPs"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "unfeasibleCCP"
          (xml_take (substa (showl_lab _A3 _B3) xml2name)
            (fun a ->
              xml_take
                (xml2unfeasible_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
                (fun b -> xml_return (subst_of (equal_list equal_char) a, b)))))
        (fun a -> xml_return (id a)));;

let rec xml2conditional_cr_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name x =
    xml_do "conditionalCrProof"
      (xml_take
        (xml_or
          (xml_do "unconditional"
            (xml_take (xml2cr_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
              (fun xa -> xml_return (Unconditional_CR xa))))
          (xml_or
            (xml_do "unraveling"
              (xml_take (unraveling_info (_A1, _A3) (_B1, _B3) xml2name)
                (fun a ->
                  xml_take
                    (xml2cr_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
                    (fun b -> xml_return (Unravel_CR (a, b))))))
            (xml_or
              (xml_do "inlineConditions"
                (xml_take
                  (xml_do "rules"
                    (xml_take_many_sub [] zero_nata Infinity_enat
                      (crule (showl_lab _A3 _B3) xml2name "rule") xml_return))
                  (fun a ->
                    xml_take
                      (xml2inline_cond_info (_A1, _A3) (_B1, _B3) xml2name)
                      (fun b ->
                        xml_take
                          (xml2conditional_cr_proof (_A1, _A2, _A3)
                            (_B1, _B2, _B3) xml2name)
                          (fun c ->
                            xml_return
                              (comp (fun aa ba -> Transformation_CR (aa, ba))
                                (fun ba -> Inline_Conditions_CCRT (a, ba)) b
                                c))))))
              (xml_or
                (xml_do "infeasibleRuleRemoval"
                  (xml_take
                    (xml2infeasible_rules_info (_A1, _A3) (_B1, _B3) xml2name)
                    (fun a ->
                      xml_take
                        (xml2conditional_cr_proof (_A1, _A2, _A3)
                          (_B1, _B2, _B3) xml2name)
                        (fun b ->
                          xml_return
                            (comp (fun aa ba -> Transformation_CR (aa, ba))
                              (fun aa -> Infeasible_Rule_Removal_CCRT aa) a
                              b)))))
                (xml_or (xml_leaf "almostOrthogonal" Almost_Orthogonal_CR)
                  (xml_or
                    (xml_do "almostOrthogonalModuloInfeasibility"
                      (xml_take
                        (xml2ao_infeasible_conds (_A1, _A3) (_B1, _B3) xml2name)
                        (fun xa ->
                          xml_return
                            (Almost_Orthogonal_Modulo_Infeasibility_CRa xa))))
                    (xml_do "al94"
                      (xml_take
                        (xml2quasi_reductive_proof (_A1, _A2, _A3)
                          (_B1, _B2, _B3) xml2name)
                        (fun a ->
                          xml_take
                            (xml2context_joinable_ccps (_A1, _A2, _A3)
                              (_B1, _B2, _B3) xml2name)
                            (fun b ->
                              xml_take
                                (xml2infeasible_conds (_A1, _A3) (_B1, _B3)
                                  xml2name)
                                (fun c ->
                                  xml_take
                                    (xml2unfeasible_ccps (_A1, _A2, _A3)
                                      (_B1, _B2, _B3) xml2name)
                                    (fun d ->
                                      xml_return
(AL94_CR (a, b, c, d))))))))))))))
        (fun xa -> xml_return (id xa)))
      x;;

let rec xml2ac_dp_termination_proof (_A1, _A2) (_B1, _B2)
  xml2name x =
    xml_do "acDPTerminationProof"
      (xml_take
        (xml_or
          (xml_do "acRedPairProc"
            (xml_take
              (ordering_constraint_proof (showl_lab _A2 _B2) xml2name false)
              (fun a ->
                xml_take
                  (xml_do "dps"
                    (xml_take (rules (showl_lab _A2 _B2) xml2name)
                      (fun xa -> xml_return (id xa))))
                  (fun b ->
                    xml_take
                      (xml_do "usableRules"
                        (xml_take (rules (showl_lab _A2 _B2) xml2name)
                          (fun xa -> xml_return (id xa))))
                      (fun c ->
                        xml_take
                          (xml2ac_dp_termination_proof (_A1, _A2) (_B1, _B2)
                            xml2name)
                          (fun d ->
                            xml_return (AC_Redpair_UR_Proc (a, b, c, d))))))))
          (xml_or
            (xml_do "acSubtermProc"
              (xml_take (multiset_af (showl_lab _A2 _B2) xml2name)
                (fun a ->
                  xml_take
                    (xml_do "dps"
                      (xml_take (rules (showl_lab _A2 _B2) xml2name)
                        (fun xa -> xml_return (id xa))))
                    (fun b ->
                      xml_take
                        (xml2ac_dp_termination_proof (_A1, _A2) (_B1, _B2)
                          xml2name)
                        (fun c -> xml_return (AC_Subterm_Proc (a, b, c)))))))
            (xml_or (xml_leaf "acTrivialProc" AC_P_is_Empty)
              (xml_or
                (xml_do "acMonoRedPairProc"
                  (xml_take
                    (ordering_constraint_proof (showl_lab _A2 _B2) xml2name
                      false)
                    (fun a ->
                      xml_take
                        (xml_do "dps"
                          (xml_take (rules (showl_lab _A2 _B2) xml2name)
                            (fun xa -> xml_return (id xa))))
                        (fun b ->
                          xml_take
                            (xml_do "trs"
                              (xml_take (rules (showl_lab _A2 _B2) xml2name)
                                (fun xa -> xml_return (id xa))))
                            (fun c ->
                              xml_take
                                (xml_do "usableRules"
                                  (xml_take (rules (showl_lab _A2 _B2) xml2name)
                                    (fun xa -> xml_return (id xa))))
                                (fun d ->
                                  xml_take
                                    (xml2ac_dp_termination_proof (_A1, _A2)
                                      (_B1, _B2) xml2name)
                                    (fun e ->
                                      xml_return
(AC_Mono_Redpair_UR_Proc (a, b, c, d, e)))))))))
                (xml_do "acDepGraphProc"
                  (xml_take_many_sub [] zero_nata Infinity_enat
                    (xml_do "component"
                      (xml_take
                        (xml_do "dps"
                          (xml_take (rules (showl_lab _A2 _B2) xml2name)
                            (fun xa -> xml_return (id xa))))
                        (fun dps ->
                          xml_take_optional (xml_bool "realScc")
                            (fun _ ->
                              xml_take_optional
                                (xml2ac_dp_termination_proof (_A1, _A2)
                                  (_B1, _B2) xml2name)
                                (fun prfOpt -> xml_return (prfOpt, dps))))))
                    (fun ret -> xml_return (AC_Dep_Graph_Proc ret))))))))
        (fun xa -> xml_return (id xa)))
      x;;

let rec xml2ac_termination_proof (_A1, _A2) (_B1, _B2)
  xml2name x =
    xml_do "acTerminationProof"
      (xml_take
        (xml_or
          (xml_do "acDependencyPairs"
            (xml_take
              (xml_do "equations"
                (xml_take (rules (showl_lab _A2 _B2) xml2name)
                  (fun xa -> xml_return (id xa))))
              (fun e ->
                xml_take
                  (xml_do "dpEquations"
                    (xml_take (rules (showl_lab _A2 _B2) xml2name)
                      (fun xa -> xml_return (id xa))))
                  (fun dpe ->
                    xml_take
                      (xml_do "dps"
                        (xml_take (rules (showl_lab _A2 _B2) xml2name)
                          (fun xa -> xml_return (id xa))))
                      (fun dp ->
                        xml_take
                          (xml_do "extensions"
                            (xml_take (rules (showl_lab _A2 _B2) xml2name)
                              (fun xa -> xml_return (id xa))))
                          (fun rext ->
                            xml_take
                              (xml2ac_dp_termination_proof (_A1, _A2) (_B1, _B2)
                                xml2name)
                              (fun p1 ->
                                xml_take_optional
                                  (xml2ac_dp_termination_proof (_A1, _A2)
                                    (_B1, _B2) xml2name)
                                  (fun a ->
                                    (match a
                                      with None ->
xml_return
  (AC_DP_Trans_Single (AC_dependency_pairs_proof (e, dp, dpe, rext), p1))
                                      | Some p2 ->
xml_return
  (AC_DP_Trans (AC_dependency_pairs_proof (e, dp, dpe, rext), p1, p2)))))))))))
          (xml_or
            (xml_do "acRuleRemoval"
              (xml_take
                (ordering_constraint_proof (showl_lab _A2 _B2) xml2name false)
                (fun a ->
                  xml_take
                    (xml_do "trs"
                      (xml_take (rules (showl_lab _A2 _B2) xml2name)
                        (fun xa -> xml_return (id xa))))
                    (fun b ->
                      xml_take
                        (xml2ac_termination_proof (_A1, _A2) (_B1, _B2)
                          xml2name)
                        (fun c -> xml_return (AC_Rule_Removal (a, b, c)))))))
            (xml_leaf "acRIsEmpty" AC_R_is_Empty)))
        (fun xa -> xml_return (id xa)))
      x;;

let rec subsumption_proof _A
  xml2name =
    xml_do "subsumptionProof"
      (xml_take_many_sub [] zero_nata Infinity_enat
        (xml_do "ruleSubsumptionProof"
          (xml_take (rule _A xml2name)
            (fun r ->
              xml_take (conversion _A xml2name)
                (fun (_, e) -> xml_return (r, e)))))
        xml_return);;

let rec xml2completion_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name =
    xml_do "completionProof"
      (xml_take (wcr_proof (showl_lab _A3 _B3) xml2name)
        (fun w ->
          xml_take
            (xml2trs_termination_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
            (fun t ->
              xml_take
                (xml_do "equivalenceProof"
                  (xml_take (subsumption_proof (showl_lab _A3 _B3) xml2name)
                    (fun s1 ->
                      xml_take_optional
                        (subsumption_proof (showl_lab _A3 _B3) xml2name)
                        (fun s2 -> xml_return (s1, s2)))))
                (fun (s1, s2) -> xml_return (SN_WCR_Eq (w, t, s1, s2))))));;

let rec xml2equational_disproof (_A1, _A2, _A3)
  xml2name =
    xml_do "equationalDisproof"
      (xml_take
        (xml_or
          (xml_do "completionAndNormalization"
            (xml_take
              (xml_do "trs"
                (xml_take
                  (rules (showl_lab _A3 (showl_list showl_nat)) xml2name)
                  (fun x -> xml_return (id x))))
              (fun a ->
                xml_take
                  (xml2completion_proof (_A1, _A2, _A3)
                    ((compare_order_list (compare_order_nat, equal_nat)),
                      (equal_list equal_nat), (showl_list showl_nat))
                    xml2name)
                  (fun b ->
                    xml_return
                      (Completion_and_Normalization_Different (a, b))))))
          (xml_do "convertibleInstance"
            (xml_take
              (subsumption_proof (showl_lab _A3 (showl_list showl_nat))
                xml2name)
              (fun x -> xml_return (Convertible_Instance x)))))
        (fun x -> xml_return (id x)));;

let rec xml2unknown_disproof (_A1, _A2) (_B1, _B2)
  xml2name x =
    xml_do "unknownInputProof"
      (xml_take
        (xml_or
          (xml_do "unknownAssumption"
            (xml_take xml2unknown_input
              (fun xa -> xml_return (Assume_NT_Unknown (xa, [])))))
          (xml_do "unknownProof"
            (xml_take (xml_text "description")
              (fun _ ->
                xml_take xml2unknown_input
                  (fun b ->
                    xml_take_many_sub [] zero_nata Infinity_enat
                      (fun xa -> right (fst xa))
                      (fun _ -> xml_return (Assume_NT_Unknown (b, []))))))))
        (fun xa -> xml_return (id xa)))
      x;;

let rec xml2equational_proof (_A1, _A2, _A3)
  xml2name =
    xml_do "equationalProof"
      (xml_take
        (xml_or
          (xml_do "equationalProofTree"
            (xml_take
              (xml2eq_proof (showl_lab _A3 (showl_list showl_nat)) xml2name)
              (fun x -> xml_return (Equational_Proof_Tree x))))
          (xml_or
            (xml_change
              (conversion (showl_lab _A3 (showl_list showl_nat)) xml2name)
              (fun (_, e) -> xml_return (Conversion e)))
            (xml_or
              (xml_change
                (subsumption_proof (showl_lab _A3 (showl_list showl_nat))
                  xml2name)
                (comp xml_return (fun a -> Conversion_With_History a)))
              (xml_do "completionAndNormalization"
                (xml_take
                  (xml_do "trs"
                    (xml_take
                      (rules (showl_lab _A3 (showl_list showl_nat)) xml2name)
                      (fun x -> xml_return (id x))))
                  (fun a ->
                    xml_take
                      (xml2completion_proof (_A1, _A2, _A3)
                        ((compare_order_list (compare_order_nat, equal_nat)),
                          (equal_list equal_nat), (showl_list showl_nat))
                        xml2name)
                      (fun b ->
                        xml_return (Completion_and_Normalization (a, b)))))))))
        (fun x -> xml_return (id x)));;

let rec xml2complexity_inputa (_A1, _A2)
  xml2name =
    xml_change (xml2complexity_input _A2 xml2name)
      (fun (CPX_input (q, s, w, cm, cc)) ->
        xml_return
          (strategy_to_Q _A1 (equal_list equal_char) q (s @ w),
            (s, (w, (cm, cc)))));;

let rec xml2complexity_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
  xml2name x =
    xml_do "complexityProof"
      (xml_take
        (xml_or
          (xml_do "ruleShifting"
            (xml_take
              (ordering_constraint_proof (showl_lab _A3 _B3) xml2name true)
              (fun rp ->
                xml_take
                  (xml_do "trs"
                    (xml_take (rules (showl_lab _A3 _B3) xml2name)
                      (fun xa -> xml_return (id xa))))
                  (fun del ->
                    xml_take_optional
                      (xml_do "usableRules"
                        (xml_take (rules (showl_lab _A3 _B3) xml2name)
                          (fun xa -> xml_return (id xa))))
                      (fun ur ->
                        xml_take
                          (xml2complexity_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
                            xml2name)
                          (fun prf ->
                            xml_return
                              (Rule_Shift_Complexity (rp, del, ur, prf))))))))
          (xml_or
            (xml_do "usableRules"
              (xml_take
                (xml_do "nonUsableRules"
                  (xml_take (rules (showl_lab _A3 _B3) xml2name)
                    (fun xa -> xml_return (id xa))))
                (fun a ->
                  xml_take
                    (xml2complexity_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
                      xml2name)
                    (fun b -> xml_return (Usable_Rules_Complexity (a, b))))))
            (xml_or
              (xml_do "split"
                (xml_take
                  (xml_do "trs"
                    (xml_take (rules (showl_lab _A3 _B3) xml2name)
                      (fun xa -> xml_return (id xa))))
                  (fun a ->
                    xml_take
                      (xml2complexity_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
                        xml2name)
                      (fun b ->
                        xml_take
                          (xml2complexity_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
                            xml2name)
                          (fun c -> xml_return (Split_Complexity (a, b, c)))))))
              (xml_or
                (xml_do "removeNonApplicableRules"
                  (xml_take
                    (xml_do "trs"
                      (xml_take (rules (showl_lab _A3 _B3) xml2name)
                        (fun xa -> xml_return (id xa))))
                    (fun a ->
                      xml_take
                        (xml2complexity_proof (_A1, _A2, _A3) (_B1, _B2, _B3)
                          xml2name)
                        (fun b ->
                          xml_return
                            (Remove_Nonapplicable_Rules_Complexity (a, b))))))
                (xml_or
                  (xml_change (bounds_info xml2name)
                    (comp xml_return (fun a -> Matchbounds_Complexity a)))
                  (xml_or
                    (xml_do "relativeBounds"
                      (xml_take (bounds_info xml2name)
                        (fun a ->
                          xml_take
                            (xml_do "trs"
                              (xml_take (rules (showl_lab _A3 _B3) xml2name)
                                (fun xa -> xml_return (id xa))))
                            (fun b ->
                              xml_take
                                (xml2complexity_proof (_A1, _A2, _A3)
                                  (_B1, _B2, _B3) xml2name)
                                (fun c ->
                                  xml_return
                                    (Matchbounds_Rel_Complexity (a, b, c)))))))
                    (xml_or (xml_leaf "rIsEmpty" RisEmpty_Complexity)
                      (xml_or
                        (xml_do "dtTransformation"
                          (xml_take
                            (rule_pairs (showl_lab _A3 _B3) xml2name "strictDTs"
                              "ruleWithDT")
                            (fun a ->
                              xml_take
                                (rule_pairs (showl_lab _A3 _B3) xml2name
                                  "weakDTs" "ruleWithDT")
                                (fun b ->
                                  xml_take
                                    (innermostLhss (showl_lab _A3 _B3) xml2name)
                                    (fun c ->
                                      xml_take
(xml2complexity_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
(fun d ->
  xml_return (DT_Transformation (DT_Transformation_Info (a, b, c), d))))))))
                        (xml_or
                          (xml_do "wdpTransformation"
                            (xml_take
                              (symbols (showl_lab _A3 _B3) xml2name
                                "compoundSymbols")
                              (fun a ->
                                xml_take
                                  (rule_pairs (showl_lab _A3 _B3) xml2name
                                    "strictWDPs" "ruleWithWDP")
                                  (fun b ->
                                    xml_take
                                      (rule_pairs (showl_lab _A3 _B3) xml2name
"weakWDPs" "ruleWithWDP")
                                      (fun c ->
xml_take (innermostLhss (showl_lab _A3 _B3) xml2name)
  (fun d ->
    xml_take (xml2complexity_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
      (fun e ->
        xml_return
          (WDP_Transformation
            (WDP_Trans_Info
               (set ((ceq_prod (ceq_lab _A2 _B2) ceq_nat),
                      (ccompare_prod
                        (ccompare_lab _A1.compare_compare_order
                          _B1.compare_compare_order)
                        ccompare_nat),
                      (set_impl_prod set_impl_lab set_impl_nat))
                  a,
                 b, c, d),
              e)))))))))
                          (xml_or
                            (xml_do "unknownProof"
                              (xml_take (xml_text "description")
                                (fun _ ->
                                  xml_take
                                    (xml2complexity_inputa
                                      ((equal_lab _A2 _B2), (showl_lab _A3 _B3))
                                      xml2name)
                                    (fun b ->
                                      xml_take_many_sub [] zero_nata
Infinity_enat
(xml_do "subProof"
  (xml_take
    (xml2complexity_inputa ((equal_lab _A2 _B2), (showl_lab _A3 _B3)) xml2name)
    (fun a ->
      xml_take (xml2complexity_proof (_A1, _A2, _A3) (_B1, _B2, _B3) xml2name)
        (fun ba -> xml_return (Complexity_assm_proof (a, ba))))))
(fun c -> xml_return (Complexity_Assumption (b, c)))))))
                            (xml_do "complexityAssumption"
                              (xml_take
                                (xml2complexity_inputa
                                  ((equal_lab _A2 _B2), (showl_lab _A3 _B3))
                                  xml2name)
                                (fun xa ->
                                  xml_return
                                    (Complexity_Assumption
                                      (xa, [])))))))))))))))
        (fun xa -> xml_return (id xa)))
      x;;

let rec xml2unknown_proof (_A1, _A2) (_B1, _B2)
  xml2name x =
    xml_do "unknownInputProof"
      (xml_take
        (xml_or
          (xml_do "unknownAssumption"
            (xml_take xml2unknown_input
              (fun xa -> xml_return (Assume_Unknown (xa, [])))))
          (xml_do "unknownProof"
            (xml_take (xml_text "description")
              (fun _ ->
                xml_take xml2unknown_input
                  (fun b ->
                    xml_take_many_sub [] zero_nata Infinity_enat
                      (fun xa -> right (fst xa))
                      (fun _ -> xml_return (Assume_Unknown (b, []))))))))
        xml_return)
      x;;

let rec xml2cert_problem
  xml2uname xml2name =
    xml_do "certificationProblem"
      (xml_take_optional
        (xml2input (showl_lab (showl_list showl_char) (showl_list showl_nat))
          xml2name)
        (fun inp ->
          xml_take_optional (xml_text "cpfVersion")
            (fun _ ->
              xml_take
                (xml_do "proof"
                  (xml_take
                    (xml_or
                      (xml_change
                        (xml2trs_termination_proof
                          ((compare_order_list
                             (compare_order_char, equal_char)),
                            (equal_list equal_char), (showl_list showl_char))
                          ((compare_order_list (compare_order_nat, equal_nat)),
                            (equal_list equal_nat), (showl_list showl_nat))
                          xml2name)
                        (comp xml_return (fun a -> TRS_Termination_Proof a)))
                      (xml_or
                        (xml_change
                          (xml2trs_nontermination_proof
                            ((compare_order_list
                               (compare_order_char, equal_char)),
                              (equal_list equal_char), (showl_list showl_char))
                            ((compare_order_list
                               (compare_order_nat, equal_nat)),
                              (equal_list equal_nat), (showl_list showl_nat))
                            xml2name)
                          (comp xml_return
                            (fun a -> TRS_Nontermination_Proof a)))
                        (xml_or
                          (xml_change
                            (xml2reltrs_nontermination_proof
                              ((compare_order_list
                                 (compare_order_char, equal_char)),
                                (equal_list equal_char),
                                (showl_list showl_char))
                              ((compare_order_list
                                 (compare_order_nat, equal_nat)),
                                (equal_list equal_nat), (showl_list showl_nat))
                              xml2name)
                            (comp xml_return
                              (fun a -> Relative_TRS_Nontermination_Proof a)))
                          (xml_or
                            (xml_change
                              (xml2cr_proof
                                ((compare_order_list
                                   (compare_order_char, equal_char)),
                                  (equal_list equal_char),
                                  (showl_list showl_char))
                                ((compare_order_list
                                   (compare_order_nat, equal_nat)),
                                  (equal_list equal_nat),
                                  (showl_list showl_nat))
                                xml2name)
                              (comp xml_return
                                (fun a -> TRS_Confluence_Proof a)))
                            (xml_or
                              (xml_change (xml2ncr_proof xml2name)
                                (comp xml_return
                                  (fun a -> TRS_Non_Confluence_Proof a)))
                              (xml_or
                                (xml_change
                                  (xml2dp_termination_proof
                                    ((compare_order_list
                                       (compare_order_char, equal_char)),
                                      (equal_list equal_char),
                                      (showl_list showl_char))
                                    ((compare_order_list
                                       (compare_order_nat, equal_nat)),
                                      (equal_list equal_nat),
                                      (showl_list showl_nat))
                                    xml2name)
                                  (comp xml_return
                                    (fun a -> DP_Termination_Proof a)))
                                (xml_or
                                  (xml_change
                                    (xml2dp_nontermination_proof
                                      ((compare_order_list
 (compare_order_char, equal_char)),
(equal_list equal_char), (showl_list showl_char))
                                      ((compare_order_list
 (compare_order_nat, equal_nat)),
(equal_list equal_nat), (showl_list showl_nat))
                                      xml2name)
                                    (comp xml_return
                                      (fun a -> DP_Nontermination_Proof a)))
                                  (xml_or
                                    (xml_change
                                      (xml2completion_proof
((compare_order_list (compare_order_char, equal_char)), (equal_list equal_char),
  (showl_list showl_char))
((compare_order_list (compare_order_nat, equal_nat)), (equal_list equal_nat),
  (showl_list showl_nat))
xml2name)
                                      (comp xml_return
(fun a -> Completion_Proof a)))
                                    (xml_or
                                      (xml_change
(xml2ordered_completion_proof
  ((compare_order_list (compare_order_char, equal_char)),
    (showl_list showl_char))
  ((compare_order_list (compare_order_nat, equal_nat)), (showl_list showl_nat))
  xml2name)
(comp xml_return (fun a -> Ordered_Completion_Proof a)))
                                      (xml_or
(xml_change
  (xml2equational_proof
    ((compare_order_list (compare_order_char, equal_char)),
      (equal_list equal_char), (showl_list showl_char))
    xml2name)
  (comp xml_return (fun a -> Equational_Proof a)))
(xml_or
  (xml_change
    (xml2equational_disproof
      ((compare_order_list (compare_order_char, equal_char)),
        (equal_list equal_char), (showl_list showl_char))
      xml2name)
    (comp xml_return (fun a -> Equational_Disproof a)))
  (xml_or
    (xml_change
      (xml2complexity_proof
        ((compare_order_list (compare_order_char, equal_char)),
          (equal_list equal_char), (showl_list showl_char))
        ((compare_order_list (compare_order_nat, equal_nat)),
          (equal_list equal_nat), (showl_list showl_nat))
        xml2name)
      (comp xml_return (fun a -> Complexity_Proof a)))
    (xml_or
      (xml_change
        (xml2quasi_reductive_proof
          ((compare_order_list (compare_order_char, equal_char)),
            (equal_list equal_char), (showl_list showl_char))
          ((compare_order_list (compare_order_nat, equal_nat)),
            (equal_list equal_nat), (showl_list showl_nat))
          xml2name)
        (comp xml_return (fun a -> Quasi_Reductive_Proof a)))
      (xml_or
        (xml_change
          (xml2conditional_cr_proof
            ((compare_order_list (compare_order_char, equal_char)),
              (equal_list equal_char), (showl_list showl_char))
            ((compare_order_list (compare_order_nat, equal_nat)),
              (equal_list equal_nat), (showl_list showl_nat))
            xml2name)
          (comp xml_return (fun a -> Conditional_CR_Proof a)))
        (xml_or
          (xml_change (xml2conditional_ncr_proof xml2name)
            (comp xml_return (fun a -> Conditional_Non_CR_Proof a)))
          (xml_or
            (xml_do "treeAutomatonClosedProof"
              (xml_take closed_criterion
                (fun x -> xml_return (Tree_Automata_Closed_Proof x))))
            (xml_or
              (xml_change
                (xml2ac_termination_proof
                  ((compare_order_list (compare_order_char, equal_char)),
                    (showl_list showl_char))
                  ((compare_order_list (compare_order_nat, equal_nat)),
                    (showl_list showl_nat))
                  xml2name)
                (comp xml_return (fun a -> AC_Termination_Proof a)))
              (xml_or
                (xml_do "ltsTerminationProof"
                  (xml_take lts_termination_proof_parser
                    (fun x -> xml_return (LTS_Termination_Proof x))))
                (xml_or
                  (xml_do "ltsSafetyProof"
                    (xml_take lts_safety_proof_parser
                      (fun x -> xml_return (LTS_Safety_Proof x))))
                  (xml_or
                    (xml_change
                      (xml2unknown_proof
                        ((compare_order_list (compare_order_char, equal_char)),
                          (showl_list showl_char))
                        ((compare_order_list (compare_order_nat, equal_nat)),
                          (showl_list showl_nat))
                        xml2name)
                      (comp xml_return (fun a -> Unknown_Proof a)))
                    (xml_change
                      (xml2unknown_disproof
                        ((compare_order_list (compare_order_char, equal_char)),
                          (showl_list showl_char))
                        ((compare_order_list (compare_order_nat, equal_nat)),
                          (showl_list showl_nat))
                        xml2name)
                      (comp xml_return
                        (fun a -> Unknown_Disproof a)))))))))))))))))))))))
                    xml_return))
                (fun p ->
                  xml_take_optional (fun x -> right (fst x))
                    (fun _ -> xml_return (inp, p))))));;

let rec parse_cert_problem
  x = parse_xmlfile
        (debug (showsl_literal "0") "parsing xml to certification problem"
          (xml2cert_problem (xml_text "name") plain_name))
        x;;

let rec certify_proof
  a xtc_o claim_o cpf =
    (let input_ret1 =
       (match xtc_o with None -> right None
         | Some xtc ->
           bind2 (parse_xmlfile
                   (problem
                     (showl_lab (showl_list showl_char) (showl_list showl_nat))
                     plain_name)
                   xtc)
             (comp left implode) (comp right (fun aa -> Some aa)))
       in
      bind2a input_ret1 (fun aa -> Unsupported aa)
        (fun input_o1 ->
          (let claim_ret =
             (match claim_o with Inl aa -> right aa
               | Inr aa -> parse_claim plain_name aa)
             in
            bind2a claim_ret (comp (fun aa -> Unsupported aa) implode)
              (fun claim ->
                bind2a (parse_cert_problem cpf)
                  (comp (fun aa -> Unsupported aa) implode)
                  (fun (input_o2, proof) ->
                    (match (input_o1, input_o2)
                      with (None, None) -> Unsupported "missing input problem"
                      | (None, Some input) ->
                        certify_cert_problem
                          ((compare_order_list
                             (compare_order_char, equal_char)),
                            (countable_list countable_char), default_list,
                            (equal_list equal_char), (showl_list showl_char))
                          (tp_rbt_impl
                            ((compare_order_lab (compare_list compare_char)
                               (compare_list compare_nat)),
                              (equal_lab (equal_list equal_char)
                                (equal_list equal_nat)))
                            ((ccompare_list ccompare_char),
                              (compare_order_list
                                (compare_order_char, equal_char)),
                              (equal_list equal_char), mapping_impl_list))
                          (dpp_rbt_impl
                            ((compare_order_lab (compare_list compare_char)
                               (compare_list compare_nat)),
                              (equal_lab (equal_list equal_char)
                                (equal_list equal_nat)))
                            ((ccompare_list ccompare_char),
                              (compare_list compare_char),
                              (equal_list equal_char), mapping_impl_list))
                          (ac_tp_list_impl
                            ((ceq_lab (equal_list equal_char)
                               (equal_list equal_nat)),
                              (ccompare_lab (compare_list compare_char)
                                (compare_list compare_nat)),
                              (compare_lab (compare_list compare_char)
                                (compare_list compare_nat)),
                              (equal_lab (equal_list equal_char)
                                (equal_list equal_nat)),
                              set_impl_lab))
                          (ac_dpp_rbt_impl
                            ((compare_order_lab (compare_list compare_char)
                               (compare_list compare_nat)),
                              (equal_lab (equal_list equal_char)
                                (equal_list equal_nat)))
                            ((compare_list compare_char),
                              (equal_list equal_char)))
                          a input claim proof
                      | (Some input, _) ->
                        certify_cert_problem
                          ((compare_order_list
                             (compare_order_char, equal_char)),
                            (countable_list countable_char), default_list,
                            (equal_list equal_char), (showl_list showl_char))
                          (tp_rbt_impl
                            ((compare_order_lab (compare_list compare_char)
                               (compare_list compare_nat)),
                              (equal_lab (equal_list equal_char)
                                (equal_list equal_nat)))
                            ((ccompare_list ccompare_char),
                              (compare_order_list
                                (compare_order_char, equal_char)),
                              (equal_list equal_char), mapping_impl_list))
                          (dpp_rbt_impl
                            ((compare_order_lab (compare_list compare_char)
                               (compare_list compare_nat)),
                              (equal_lab (equal_list equal_char)
                                (equal_list equal_nat)))
                            ((ccompare_list ccompare_char),
                              (compare_list compare_char),
                              (equal_list equal_char), mapping_impl_list))
                          (ac_tp_list_impl
                            ((ceq_lab (equal_list equal_char)
                               (equal_list equal_nat)),
                              (ccompare_lab (compare_list compare_char)
                                (compare_list compare_nat)),
                              (compare_lab (compare_list compare_char)
                                (compare_list compare_nat)),
                              (equal_lab (equal_list equal_char)
                                (equal_list equal_nat)),
                              set_impl_lab))
                          (ac_dpp_rbt_impl
                            ((compare_order_lab (compare_list compare_char)
                               (compare_list compare_nat)),
                              (equal_lab (equal_list equal_char)
                                (equal_list equal_nat)))
                            ((compare_list compare_char),
                              (equal_list equal_char)))
                          a input claim proof))))));;

let rec certifyproof x = certify_proof false None (Inl Anything) (explode x);;

end;; (*struct Ceta*)
